;;; -*- Mode:Common-Lisp; Package:BFIG; Base:10 -*-

;;; Copyright (c) 1999 - 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :BFIG)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Activation
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; As assignments are made current or disallowed,
;;; this code incrementally maintains,  whether 
;;; each variable and conflict is active or inactive.

;;; For conditional CSPs:
;;;  A conflict is resolved if supported OR if CONFLICT IS INACTIVE.

;;;  A conflict is active IFF all variables of its inconsistent 
;;;  assignments are active.

;;;   - Maintain a flag for a conflict being active.  

;;;   - Maintain for each conflict a count of inactive variables
;;;     (inactive-variable-count).

;;; To perform incremental update of variable/conflict activity,
;;; conflict support and resolution:

;;;   - Incremental Conflict Activity (in activation.lisp - done):
;;;     When a variable becomes (in)active, check if any conflict mentioning
;;;     variable becomes (in)active.

;;;     - Update conflict's active? flag when 
;;;        - all variables are active OR
;;;      . - if one variable becomes inactive.

;;;     - To determine conflict activity.
;;;       - (inc/)decrement inactive-variable-count.
;;;       - when count becomes (non-)zero conflict becomes
;;;         (in)active.

;;;     - alternatively, do the equivalent of watched literals
;;;       (i.e., put a watch on the first inactive variable of conflict).

;;;   - Incremental Conflict Resolution 2 (in activation.lisp):

;;;     When a conflict becomes (in)active, 
;;;     - when conflict is unsupported, 
;;;       conflict becomes (resolved)unresolved.

;;;   - Incremental Conflict Support 2 (in activation.lisp):

;;;     When an assignment becomes (in)active, 
;;;     Update the support of any conflict for which assignment 
;;;     is a kernel.
;;;
;;;     - Conflict becomes supported by an activated assignment
;;;       - if assignment is a kernel of conflict, and 
;;;       - if conflict is unsupported.

;;;     - Conflict becomes unsupported by a deactivated assignment
;;;       - if conflict is supported by assignment, 
;;;       - and conflict can not be resupported (see below).

;;;  .. and repeated from conflicts.lisp:

;;;   - Incremental Conflict Support 1 (in conflicts.lisp):
;;;     When an assignment becomes (dis)allowed,
;;;     Update the support of any conflict for which assignment is a
;;;     member of conflict's inconsistent assignments.

;;;     - Conflict becomes supported by a disallowed assignment
;;;       - if assignment is a member of the inconsistent assignments 
;;;         of conflict, AND 
;;;       - if conflict is unsupported.

;;;     - Conflict becomes unsupported by an assignment that 
;;;       has gone from disallowed to allowed, 
;;;       - if conflict is supported by assignment being disallowed, and
;;;          conflict cannot be resupported (see below).

;;;   - Incremental Conflict Resupport (in conflicts.lisp):
;;;     When conflict becomes unsupported by assignment,

;;;     - try to resupport conflict by another
;;;       - current assignment OR
;;;       - DISALLOWED assignment.

;;;     - To avoid oscillation between support and unsupport,
;;;       only try resupporting after 

;;;       1) removing any assignments that are no longer 
;;;          current or disallowed.

;;;       2) adding any new assignments that are current
;;;          or disallowed.

;;;   - Incremental Conflict Resolution 1 (in conflicts.lisp):

;;;     When a conflict becomes (un)supported
;;;     (and cannot be resupported).

;;;     - When conflict is active,
;;;     - Update conflict as (un)resolved.

;;;   - Incremental Resolved Conflicts Count (in conflicts.lisp
;;;;                                          and activation.lisp).

;;;     - When a conflict becomes(un)resolved, 
;;;     - then (inc/dec)rement the unresolved conflict count.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do:
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; o Implement Incremental Conflict Resolution 2.

;;; o Implement Incremental Conflict Support 2.

;;; o Write function to determine conflict-active?

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Optimization:
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; o Incrementally determine that conflict is active
;;;   using watched literals.

;;; o Implement Incremental Conflict Resupport.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Querying Conflict Activity
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Added.
;;; o Currently defined in bfig-defs.lisp
;;;(defmethod conflict-active?
;;;	
;;;	((conflict conditional-conflict))
;;;
;;;  "Returns true iff CONFLICT is active.
;;;   A conflict is active iff all of its variables are active.
;;;   Unconditional variables are always active."
;;;  
;;;  (with-slots (inactive-variable-count) conflict
;;;	(eql inactive-variable-count 0)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Selection of Unassigned Variables During Search
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; SETUP-VARIABLE-SEARCH-ORDER! pseudo code:

;;; - Create a priority queue, sorted by increasing
;;;   size of variable domain.
;;; - Mark all variables as unqueued.
;;; - Initialize queue with all variables with no guard.
;;;    - mark a variable as queued, as it is enqueued.
;;; - Repeat until queue is empty.
;;;   - Pop smallest element e from queue.
;;;   - collect e on the list of variables to assign in order.
;;;   - Let v be the variable in the guard of e.
;;;   - If variable is not queued, then queue it.
;;; - Record collected variables as variables in order.  

(defmethod setup-variable-search-order!

	((search-problem conditional-search-problem))
  
  "Reorders variables of SEARCH-PROBLEM in the order in which
   the variables are to be split during search.

   Variables are ordered so that each conditional variable
   appears AFTER the variables that activate it.
   All else being equal, variables are ordered by increasing
   domain size, based on the most constrained variable heuristic."
  
  ;; Order variables by increasing domain size, with the additional
  ;; constraint that each conditional variable appears after the
  ;; variable that activates it.

  ;; Queue contains all active variables yet to be placed in order.
  
  (let ((Q (make-ordered-list-queue :key #'state-variable-domain-size)))
	
	(with-slots (variables) search-problem

	  ;; Start with all unconditional (unguarded) variables queued.
	  ;; Unmark the remaining variables for good measure.
	  (loop for v in variables
		  do (cond ((state-variable-unguarded? v)
					(enqueue! Q v)
					(mark-state-variable! v))
				   (:otherwise
					(unmark-state-variable! v))))

	  ;;; Create a list, ORDERED-VARIABLES, of variables,
	  ;;; ordered as described above.

	  ;;; fix this
	  ;;; wasn't working w/o explicit call
	  ;;; -- gustavop
	  (loop until (queue-empty? Q)
		  for v = (dequeue! q)
		  collecting v into ordered-variables
		  do (loop for a in (state-variable-domain v)
				 do (loop for cv in (state-assignment-variables-activated a)
						unless (state-variable-marked? cv)
						do (enqueue! Q cv)
						   (mark-state-variable! cv)))
		  finally (setq variables ordered-variables))
	  
	  (debug-print-for :BFIG t
					   "  Variables to be split in the order ~a." variables)

	  ;; Leave variables unmarked.
	  (clear-state-variable-marks! search-problem)))

  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;s;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Internalize Conditional Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Added method.
(defmethod make-implicit-assignments-explicit

    ((state-space conditional-state-space) assignments)
  
  "Given a list of ASSIGNMENT objects,
   Returns this list, augmented by any assignments implied.

   For a conditional STATE-SPACE the implicit assignments,
   are all assignments that appear in guard conditions that
   activate ASSIGNMENTS."
  
  ;; Q processes all implicit assignments.
  (loop with q = assignments
	  with augmented? = nil
	  until (null q)
			;; Stop when all implicit assignments processed.
	  for a = (pop q)
	  for v = (state-assignment-variable a)

	  unless (state-variable-marked? v)
			 ;; Accumulate all explicit assignments
	  collect a into explicit-assignments
					 ;; Use mark to ensure there are no duplicates.
	  and do (mark-state-variable! v)
		  ;; When assignment is to a conditional variable, 
		  ;; add the guard condition to the explicit assignments.
	  and unless (state-variable-unguarded? v)
	  do (setq augmented? T)
		 (push (state-variable-guard v) Q)
	  finally 
		;; Cleanup marks before 
		;; returning all explicit assignments.
		(clear-state-variable-marks! state-space)
		(when augmented?
		  (debug-print-for :bfig t "Augmenting conditional conflict ~~~a with guards, ~%        producing ~a."
						   assignments explicit-assignments))
		(return explicit-assignments)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;s;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Initialization of Current Assignments
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Generalize clear-current-assignments!
;;;   so that they apply to conditional-state-space w/o conflicts.
;;;   For example, write a maintain-current-assignment mixin.

;;; o Do we need to clear the effects of propagating any other
;;;   activations, other than conflicts?

;;; Changed:
;;; o New method.
(defmethod clear-effects-of-current-assignments!
	
	((state-space conditional-state-space-w-conflicts))
  
  "Clear the effects propagated for any current 
   assignments for STATE-SPACE."
  
  ;; Clear active conflicts before resolved conflicts,
  ;; since conflict activity determines the number
  ;; of unresolved conflicts.
  (clear-activated-conflicts! state-space)
  (clear-resolved-conflicts! state-space))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; (Un)Record Assignment as Current
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod propagate-current-assignment!

	((state-space conditional-state-space)
	 (new-assignment conditional-state-assignment))
  

  "Identifies all effects of recording NEW-ASSIGNMENT 
   as current in STATE-SPACE.

   Assumptions:

   - NEW-ASSIGNMENT is recorded as a current assignment,
     BUT does not support any consequences.
  
   Side Effects:  

   -  Records all consequences of NEW-ASSIGNMENT."

  ;; Determines which variables and conflicts become
  ;; active.
  (propagate-activation! state-space new-assignment)
  T)

(defmethod undo-consequences-of-current-assignment!

	((state-space conditional-state-space)
	 (old-assignment conditional-state-assignment))

  "Identifies all effects of removing OLD-ASSIGNMENT 
   as current in STATE-SPACE.

   Assumptions:

   - OLD-ASSIGNMENT is still recorded as a current assignment,
     AND consequences of assignment are recorded.
  
   Side Effects:  

   -  Removes all consequences of OLD-ASSIGNMENT.
   -  Leaves OLD-ASSIGNMENT recorded as current."

  (propagate-activation! state-space old-assignment))

(defmethod propagate-current-assignment!

	((state-space conditional-state-space-w-conflicts)
	 (new-assignment conditional-state-assignment-w-conflicts))
  

  "Identifies all effects of recording NEW-ASSIGNMENT 
   as current in STATE-SPACE.

   Assumptions:

   - NEW-ASSIGNMENT is recorded as a current assignment,
     BUT does not support any consequences.
  
   Side Effects:  

   -  Records all consequences of NEW-ASSIGNMENT."

  ;; Determines which variables and conflicts become
  ;; active.
  (propagate-activation! state-space new-assignment)
  
  ;; Updates conflict resolution as the result of new-assignment
  ;; and consequently conflicts becoming active.
  (support-conflicts-by-assignment! state-space new-assignment)
  T)

(defmethod undo-consequences-of-current-assignment!

	((state-space conditional-state-space-w-conflicts)
	 (old-assignment conditional-state-assignment-w-conflicts))

  "Identifies all effects of removing OLD-ASSIGNMENT 
   as current in STATE-SPACE.

   Assumptions:

   - OLD-ASSIGNMENT is still recorded as a current assignment,
     AND consequences of assignment are recorded.
  
   Side Effects:  

   -  Removes all consequences of OLD-ASSIGNMENT.
   -  Leaves OLD-ASSIGNMENT recorded as current."
  
  ;; Remove activation of variables and conflicts that were
  ;; activated by OLD-ASSIGNMENT.
  (propagate-activation! state-space old-assignment)
  
  ;; Update record of supported and resolved conflicts, as a result
  ;; of removing OLD-ASSIGNMENT, and consequently, variables and
  ;; conflicts becoming deactivated.
  (unsupport-conflicts-by-assignment! state-space old-assignment))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Propagating Activation for Assignments
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod propagate-activation!
	((state-space conditional-state-space)
	 (assignment conditional-state-assignment))

  "Propagate the effect of a change in ASSIGNMENT's 
   activation (current/not current)
   to variables that assigment ACTIVATES.

   Note: This function is called whenever an assignment 
         is added or removed from current."

  (with-slots (variables-activated) assignment
	(loop for v in variables-activated
		do (update-activation! state-space v))
	T))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Propagating Activation for Variables
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod update-activation! 

	((state-space conditional-state-space)
	 (variable conditional-state-variable))

  "Update VARIABLE's active? flag and propagate changes to consequences.

   Assumes that CURRENT-CDB-ASSIGNMENT has been updated to reflect
   the current (changed) state.

   Assumes this is ONLY called on variables with guards."
  (with-slots (active? guard) variable
	(let* ((gvariable (state-assignment-variable guard))
		   (current (state-variable-current-cdb-assignment gvariable))
		   (new-active? (eql guard current)))
	  (debug-print-for :bfig t "    ~a: guard = ~a, current assignment = ~a."
					   gvariable guard current)
	  (cond ((eql active? new-active?)
			 (if active?
				 (debug-print-for :bfig t "    ~a remains active."
								  variable)
			   (debug-print-for :bfig t "    ~a remains inactive."
								variable))
			 )
			(:else
			 (debug-print-progn-for 
			  :bfig
			  (if new-active?
				  (debug-print-for :bfig t "    ~a becomes active."
								   variable)
				(debug-print-for :bfig t "    ~a becomes inactive."
								 variable)))
			 (setq active? new-active?)
			 (propagate-activation! state-space variable))))))

;;; To Do:
;;; o Inform the search algorithm that variable has become
;;;   active.
(defmethod propagate-activation! 
	((state-space conditional-state-space)
	 (variable conditional-state-variable))

  "Propagate the effect of a change in VARIABLE's activation."

;;;  (debug-print-for :bfig t "propagating ~a does nothing yet."
;;;						   variable)

  T)

;;; Optimization:
;;; o When incrementally determining conflict-active?,
;;;   use a watched literal approach.
(defmethod propagate-activation! :after
		   ((state-space conditional-state-space-w-conflicts)
			(variable conditional-state-variable-w-conflicts))

  "Propagate the effect of a change in VARIABLE's activation
   to conflicts whose inconsistent assignments include variable."

  (with-slots (active?) variable

	(let ((activated-conflicts
		   (state-variable-conflicts-activated variable)))
	
;;;	  (debug-print-for :bfig t "propagating ~a to conflicts ~a."
;;;					   variable activated-conflicts)
	
	  (loop for c in activated-conflicts
		  do (update-activation-from-change! state-space c active?))
	
	  T)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Setting Up Conditional Conflicts for Activation
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod setup-conflict-activation!
	((state-space conditional-state-space-w-conflicts)
	 (conflict conditional-conflict))
  
  "Setup the link between CONFLICT and the variables that activate it."

  (with-slots (inconsistent-assignments)
	  conflict
	(loop
		for a in inconsistent-assignments
		for v = (state-assignment-variable a)
		unless (state-variable-active? v)
		do (push conflict 
				 (state-variable-conflicts-activated v))))
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Removing Conditional Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod remove-all-conflicts! :after

		   ((state-space conditional-state-space-w-conflicts))
  
  "Removes the link between all conflicts in STATE-SPACE
   and the variables that activate them.

   Note: The function does not delete the conflict objects."

  ;; Remove pointers from each variable to all conflicts
  ;; activated by variable.
  (with-slots (variables) state-space
	(loop for variable in variables
		do (setf (state-variable-conflicts-activated variable)
			 '())))
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Initializing Activation of Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod clear-activated-conflicts! 

	((state-space conditional-state-space-w-conflicts))
  
  "Clear the record of all conflicts that are activated by the
   current assignment of STATE-SPACE."
  
  (loop for conflict in (state-space-conflicts state-space)
      do (initialize-conflict-activation! state-space conflict))
  T)

;;; Optimization:
;;; o Incrementally determine conflict-active?
;;;   using a watched literal approach.
(defmethod initialize-conflict-activation!
	((state-space conditional-state-space-w-conflicts)
	 (conflict conditional-conflict))
  
  "Determine if CONFLICT is initially active, based on the 
   activity of the variables of CONFLICT's inconsistent assignments.
   
   A conflict is active iff all variables mentioned in its 
   inconsistent assignments is active."

  (with-slots (inconsistent-assignments inactive-variable-count)
	  conflict
	(loop 
		with count = 0
		for a in inconsistent-assignments
		for v = (state-assignment-variable a)
		unless (state-variable-active? v)
		do (incf count)
		finally (setq inactive-variable-count count))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Propagating Activation for Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod update-activation-from-change!
	((state-space conditional-state-space-w-conflicts) 
	 (conflict conditional-conflict)
	 active?)

  "Update the activation of CONFLICT and propagate consequences, 
   given that one of the variables listed in conflict's
   inconsistent assignments has changed its activation to ACTIVE?.

   A conflict is active iff all variables mentioned in its 
   inconsistent assignments is active."
  
  (with-slots (inactive-variable-count) conflict
	(cond (active?
		   (decf inactive-variable-count)

		   (when (conflict-active? conflict)
			 
			 ;; Conflict just became active.
			 (debug-print-for :bfig t "    ~a becomes active."
							  conflict)
			 
			 (unless (conflict-supported? state-space conflict)
			   			   
			   ;; Conflict went from resolved to unresolved
			   (debug-print-for :bfig t "    ~a becomes unresolved."
								conflict)
			   (incf (state-space-unresolved-conflict-count
					  state-space)))))

		  (:otherwise
		   (let ((was-active? (conflict-active? conflict)))
			 (incf inactive-variable-count)

			 (when was-active?

			   ;; Conflict just became inactive.
			   (debug-print-for :bfig t "    ~a becomes inactive."
								conflict)
			   (unless (conflict-supported? state-space conflict)
			   
				 ;; Conflict went from unresolved to resolved
				 (debug-print-for :bfig t "    ~a -> resolved."
								  conflict)
				 (decf (state-space-unresolved-conflict-count
						state-space))))
			 )))))
