;;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :BFIG)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			equal-kernels
;;;			set-contains-kernel
;;;			equal-kernel-sets
;;;			equal-kernel-sequences
;;;			equal-state-assignments)
;;;		  :BFIG))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                               ;;;
;;;  Interface Functions:    	                                  ;;;
;;;     Comparing Kernel and State Variable Assignments           ;;;
;;;                                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (EQUAL-KERNELS KERNEL1 KERNEL2)
;;;
;;;   Returns true if two KERNELS are equivalent.
;;;   Each kernel is a set of assignments to 
;;;   state variables that resolve all conflicts.

;;; (SET-CONTAINS-KERNEL KERNEL-SET KERNEL)
;;;  
;;;   Returns true if KERNEL is contained in the
;;;   set of kernels KERNEL-SET.  Each kernel
;;;   is a set of assignments to state variables
;;;   that resolve all conflicts.

;;; (EQUAL-KERNEL-SETS KERNEL-SET1 KERNEL-SET2)
;;;
;;;   Returns true iff KERNEL-SET1 and KERNEL-SET2
;;;   denote the same sets of kernels.  Each kernel
;;;   is a set of assignments to state variables
;;;   that resolve all conflicts.

;;; (EQUAL-KERNEL-SEQUENCES KERNEL-SEQ1 KERNEL-SEQ2)
;;;
;;;   Returns true iff KERNEL-SEQ1 and KERNEL-SEQ2
;;;   denote the same sequences of kernels.  Each kernel
;;;   is a set of assignments to state variables that 
;;;   resolve all conflicts.

;;; (EQUAL-ASSIGNMENTS A1 A2)
;;;   Returns true iff and only if A1 and A2 denote the same
;;;   truth assignment. a1 and a2 are of the form (p truth),
;;;   where p is a proposition described by a proposition structure
;;;   or a symbol (the proposition description) and truth is
;;;   a truth value.

;;; (EQUAL-STATE-ASSIGNMENTS A1 A2)
;;;
;;;   Returns true iff and only if A1 and A2 denote the same
;;;   assignments to state variables.  A1 and A2 are lists of
;;;   assignments to individual state variables.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                               ;;;
;;;  To Do:                                                       ;;;
;;;                                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Optimization:
;;; o None of these comparison functions are designed
;;;   for efficiency.

;;; To Do:
;;; o This file is essentially a copy of compare.lisp for ISAT.
;;;   It would be more elegant to collapse the two sets of functions
;;;   into one set.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                               ;;;
;;;         Comparing Kernel and Variable Assignments             ;;;
;;;                                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun equal-kernels 
	(kernel1 kernel2)

  "Returns true if two KERNELS are equivalent.
   Each kernel is a set of assignments to 
   state-variables that resolve all conflicts."
  
  (equal-state-assignments kernel1 kernel2))

(defun set-contains-kernel 
	(kernel-set kernel)
  
  "Returns true if KERNEL is contained in the
   set of kernels KERNEL-SET.  Each kernel
   is a set of assignments to state-variables
   that resolve all conflicts."

  (member kernel kernel-set :test #'equal-kernels))

(defun equal-kernel-sets 
	(kernel-set1 kernel-set2)

  "Returns true iff KERNEL-SET1 and KERNEL-SET2
   denote the same sets of kernels.  Each kernel
   is a set of assignments to state-variables
   that resolve all conflicts."

  (and (subsetp kernel-set1 kernel-set2 :test #'equal-kernels)
       (subsetp kernel-set2 kernel-set1 :test #'equal-kernels)))

(defun equal-kernel-sequences 
	(kernel-seq1 kernel-seq2)

  "Returns true iff KERNEL-SEQ1 and KERNEL-SEQ2
   denote the same sequences of kernels.  Each kernel
   is a set of assignments to state-variables that 
   resolve all conflicts."
  
  (every #'equal-kernels kernel-seq1 kernel-seq2))

(defun equal-state-assignments 
	(a1 a2)

  "Returns true iff and only if A1 and A2 denote the same
   assignments to state-variables.  A1 and A2 are lists of
   assignments to individual state-variables."

  (and (subsetp a1 a2 :test #'equal-state-assignment)
       (subsetp a2 a1 :test #'equal-state-assignment)))

(defun equal-state-assignment 
	(l1 l2)

  "Returns true iff and only if A1 and A2 denote the same
   assignment to some decision variable."

  (equal-candgen-sexpression l1 l2))

(defun equal-candgen-sexpression 
	(l1 l2)
  
  "Returns true if and only if L1 and L2 denote the same
   decision variable assignments. L1 and L2 can be arbitrary nested lists.
   decision assignment objects in L1 and L2 are dereferenced to 
   their assignment descriptions during comparison."
  
  (if (null l1)
	  
      (null l2)
	
    (let ((l1-pd (extract-state-assignment-description l1))
		  (l2-pd (extract-state-assignment-description l2)))

      (cond ((listp l1-pd)
			 
			 (if (listp l2-pd)
				 
				 (and (equal-candgen-sexpression (first l1-pd)(first l2-pd))
					  (equal-candgen-sexpression (rest l1-pd) (rest l2-pd)))
			   
			   nil))
			
			(:otherwise
			 
			 (equal l1-pd l2-pd))))))

(defun extract-state-assignment-description 
	(a)

  "A is a decision assignment object or a decision assignment description.
   Returns an assignment description for A."
  
  (if (typep a 'state-assignment)

      (state-assignment-description a)

    a))
