;;; -*- Mode:Common-Lisp; Package:BFIG; Base:10 -*-

;;; Copyright (c) 1999 - 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :BFIG)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;;;;			  conflict
;;;;;;			  new-conflict
;;;			   insert-conflict!
;;;            insert-new-conflict		
;;;			   remove-all-conflicts!)
;;;		     :BFIG)
;;;  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Maintaining Conflicts and Constituent Kernels
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This file contains routines for maintaining the conflicts of a search
;;; problem, and their corresponding constituent kernels.
;;; Incrementally maintains a record of the current search state and the
;;; corresponding conflicts that are resolved by this state.

;;; To perform incremental update of conflict support and resolution:

;;;   - Incremental Conflict Support 1:

;;;     When an assignment becomes (dis)allowed,
;;;     Update the support of any conflict for which assignment is a
;;;     member of conflict's inconsistent assignments.

;;;     - Conflict becomes supported by a disallowed assignment
;;;       - if assignment is a member of the inconsistent assignments 
;;;         of conflict, AND 
;;;       - if conflict is unsupported.

;;;     - Conflict becomes unsupported by an assignment that 
;;;       has gone from disallowed to allowed, 
;;;       - if conflict is supported by assignment being disallowed, and
;;;          conflict cannot be resupported (see below).

;;;   - Incremental Conflict Resupport:
;;;     When conflict becomes unsupported by assignment,

;;;     - try to resupport conflict by another
;;;       - current assignment OR
;;;       - DISALLOWED assignment.

;;;     - To avoid oscillation between support and unsupport,
;;;       only try resupporting after 

;;;       1) removing any assignments that are no longer 
;;;          current or disallowed.

;;;       2) adding any new assignments that are current
;;;          or disallowed.

;;;   - Incremental Conflict Resolution 1:

;;;     When a conflict becomes (un)supported
;;;     (and cannot be resupported).

;;;     - When conflict is active,
;;;       update conflict as (un)resolved.

;;;     - Note: If variables are unconditional,
;;;             Then conflict is always active,
;;;             Else conflict is active if all
;;;               variables in its inconsistent 
;;;               assignments are active.

;;;   - Incremental Resolved Conflicts Count:

;;;     - When a conflict becomes(un)resolved, 
;;;     - then (inc/dec)rement the unresolved conflict count.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; BFIG Interface 
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; *** Adding and Removing Conflicts:

;;; (MAKE-INSTANCE 'CONFLICT :DECISION-SPACE DECISION-SPACE
;;;      :INCONSISTENT-ASSIGNMENTS INCONSISTENT-ASSIGNMENTS)
;;;  
;;;   Asserts a new conflict for search-problem.  A conflict is a partial
;;;   set, INCONSISTENT-ASSIGNMENTS, of variable assignments that are 
;;;   inconsistent.  A conflict contains at most one assignment 
;;;   for each variable in the search problem.  INCONSISTENT-ASSIGNMENTS 
;;;   can either be a list of state-assignment objects, or an
;;;   assignment description of the form (= variable-name value).

;;; (INSERT-CONFLICT! (CONFLICT CONFLICT) 
;;;     (STATE-SPACE STATE-SPACE-W-CONFLICTS)))
;;;
;;;   Inserts CONFLICT object into STATE-SPACE, where
;;;   conflicts are maintained in increasing order of
;;;   cardinality. Updates the record of whether or not 
;;;   conflict is resolved,and returns conflict.
;;;   Note: does not try to detect and remove redundant 
;;;   conflicts.

;;; (INSERT-NEW-CONFLICT (SPACE STATE-SPACE-W-CONFLICTS)
;;;    :INCONSISTENT-ASSIGNMENTS INCONSISTENT-ASSIGNMENTS)
;;;  
;;;   Asserts a new conflict for search-problem.  A conflict is a set, 
;;;   INCONSISTENT-ASSIGNMENTS, of variable assignments that are mutually 
;;;   inconsistent.  A conflict contains at most one assignment 
;;;   for each variable in the search problem.  INCONSISTENT-ASSIGNMENTS 
;;;   can either be a list of decision-assignment objects or 
;;;   an assignment description of the form (= variable-name value).
  
;;; (REMOVE-ALL-CONFLICTS!
;;;    (STATE-SPACE STATE-SPACE-W-CONFLICTS))
;;;  
;;;   Removes the set of all conflicts currently associated with
;;;   STATE-SPACE and any information recorded about the 
;;;   status of these conflicts within STATE-SPACE.
;;;   Note: The function does not delete the conflict objects.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Additional Conflicts Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; *** Updating the Current State:

;;; (CLEAR-CURRENT-ASSIGNMENTS! (STATE-SPACE STATE-SPACE-W-CONFLICTS))
;;;   Clear the record of any current assignments for STATE-SPACE and 
;;;   the record of any conflicts it resolves.

;;; (RECORD-ASSIGNMENT-AS-CURRENT! (STATE-SPACE STATE-SPACE-W-CONFLICTS)
;;;	    (NEW-ASSIGNMENT STATE-ASSIGNMENT-W-CONFLICTS))
;;;   Record NEW-ASSIGNMENT as current in STATE-SPACE, and 
;;;   identify all conflicts resolved by the current assignments.
;;;
;;;   Assumptions:
;;;   - NEW-ASSIGNMENT is not one of the current assignments,
;;;     and does not support any conflict.
;;;  
;;;   Side Effects:  
;;;   -  Records NEW-ASSIGNMENT on its respective 
;;;      STATE-VARIABLE-CURRENT-CDB-ASSIGNMENTs
;;;      of STATE-SPACE.
;;;
;;;   -  Each conflict resolved by NEW-ASSIGNMENT, is
;;;      supported by new-assignment or some other current assignment."

;;;  (UNRECORD-ASSIGNMENT-AS-CURRENT! (state-space state-space-w-conflicts)
;;;	    (OLD-ASSIGNMENT STATE-ASSIGNMENT-W-CONFLICTS))
;;;   Removes the record of OLD-ASSIGNMENT as current in STATE-SPACE,
;;;   and the record of conflicts resolved by this assignment.
;;;
;;;  Assumptions:
;;;   - OLD-ASSIGNMENT is included in the current assignments of STATE-SPACE.
;;;   - Those conflicts of STATE-SPACE that are resolved by current 
;;;     allowed/disallowed assignments are supported by one of the 
;;;     assignments in the extension.
;;;  
;;;  Side Effects:  
;;;   - Clears OLD-ASSIGNMENT from current assignments.
;;;   - Clears the conflicts supported by OLD-ASSIGNMENT."

;;;  (RECORD-DISALLOWED-ASSIGNMENT-AS-CURRENT! 
;;;	    (STATE-SPACE STATE-SPACE-W-CONFLICTS)
;;;	    (DISALLOWED-ASSIGNMENT STATE-ASSIGNMENT-W-CONFLICTS))
;;;   Record DISALLOWED-ASSIGNMENT as current in STATE-SPACE, and 
;;;   identify all conflicts resolved by the disallowed assignment.
;;;
;;;   Assumptions:
;;;   - DISALLOWED-ASSIGNMENT is not one of the current diasallowed
;;;     or allowed assignments, and does not support any conflict.
;;;  
;;;   Side Effects:  
;;;   -  Records DISALLOWED-ASSIGNMENT on its respective 
;;;      STATE-VARIABLE-CURRENT-CDB-DISALLOWED-ASSIGNMENTs
;;;      of STATE-SPACE.
;;;
;;;   -  Each conflict resolved by DISALLOWED-ASSIGNMENT, is
;;;      supported by DISALLOWED-ASSIGNMENT or some other 
;;;      current assignment.

;;;  (UNRECORD-DISALLOWED-ASSIGNMENT-AS-CURRENT! 
;;;	   (STATE-SPACE STATE-SPACE-W-CONFLICTS)
;;;	   (DISALLOWED-ASSIGNMENT STATE-ASSIGNMENT-W-CONFLICTS))
;;;
;;;   Removes the record of DISALLOWED-ASSIGNMENT as current in STATE-SPACE,
;;;   and the record of conflicts resolved by this assignment.
;;;
;;;  Assumptions:
;;;
;;;   - DISALLOWED-ASSIGNMENT is included in the current disallowed
;;;     assignments of STATE-SPACE.
;;;
;;;   - Those conflicts of STATE-SPACE that are resolved by current 
;;;     allowed/disallowed assignments are supported by one of the 
;;;     assignments in the extension.
;;;  
;;;  Side Effects:  
;;;
;;;   - Clears DISALLOWED-ASSIGNMENT from current assignments.
;;;   - Clears conflicts supported by DISALLOWED-ASSIGNMENT."


;;; *** [FYI: Search State Update Functions in the Candidate Generator:

;;;    The following is the equivalent interface on the candidate
;;;    generator side.

;;; (CLEAR-CURRENT-SEARCH-NODE! (SEARCH-PROBLEM SEARCH-PROBLEM))
;;;   Clear the assignments recorded as the current search state
;;;   for SEARCH-PROBLEM.  Updates the record of conflicts resolved.

;;; (RECORD-SEARCH-NODE-AS-CURRENT! 
;;;    (SEARCH-PROBLEM SEARCH-PROBLEM) 
;;;    (NEW-SEARCH-NODE SEARCH-NODE)
;;;    DISJOINT?)
;;;
;;;   Records NEW-SEARCH-NODE as the current search node.
;;;   Updates the current (partial) search state based on NEW-SEARCH-NODE
;;;   and updates the record of conflicts resolved.
;;;
;;;   Assumptions:
;;;
;;;   - SEARCH-PROBLEM-CURRENT-SEARCH-NODE, 
;;;     STATE-VARIABLE-CURRENT-CDB-ASSIGNMENT
;;;     and the support of the conflicts of search-problem are
;;;     set to nil.
;;;  
;;;   Side Effects:  
;;;
;;;   -  Records new-search node as SEARCH-PROBLEM-CURRENT-SEARCH-NODE
;;;      for SEARCH-PROBLEM.
;;;   -  Records each assignment of the new kernel on its respective
;;;      STATE-VARIABLE-CURRENT-CDB-ASSIGNMENT.
;;;   -  For each conflict resolved by the new kernel, SUPPORTS the
;;;      conflict by the assignment that resolves it."

;;; (UNRECORD-CURRENT-SEARCH-NODE! (SEARCH-PROBLEM SEARCH-PROBLEM) DISJOINT?)
;;;
;;;   Removes the record of the current-search-node and 
;;;   updates the set of resolved conflicts accordingly.
;;;
;;;  Assumptions:
;;;
;;;   - SEARCH-PROBLEM-CURRENT-SEARCH-NODE is set to an search node.
;;;   - The current kernel is the best cost extension of the current 
;;;     search-node.
;;;   - The current variable assignments, 
;;;     STATE-VARIABLE-CURRENT-CDB-ASSIGNMENT,
;;;     are set to the assignments of the current search node.
;;;   - Those conflicts of SEARCH-PROBLEM that are resolved by the best cost
;;;     extension are supported by one of the assignments in the extension.
;;;  
;;;  Side Effects:  
;;;
;;;   - Clears SEARCH-PROBLEM-CURRENT-SEARCH-NODE of SEARCH-PROBLEM.
;;;   - Clears the current variable assignments.
;;;   - Clears the conflict support.

;;; END FYI]

;;; *** Querying the Status of Conflicts:

;;; (ALL-CONFLICTS-RESOLVED? 
;;;    (DECISION-SPACE DECISION-SPACE-W-CONFLICTS))
;;;  
;;;  Returns True IFF all inserted conflicts are resolved by the current 
;;;  search state.

;;; (SMALLEST-UNRESOLVED-CONFLICT-OF-CURRENT-STATE
;;;    (DECISION-SPACE DECISION-SPACE-W-CONFLICTS))
;;;  
;;;   Returns the conflict of SEARCH-PROBLEM that:
;;;     1) is not resolved by the current search state, and
;;;     2) has the smallest number of constituent kernel assignments
;;;        (i.e., prime implicants).  
;;;   Returns nil if all conflicts are resolved.


;;; *** Extracting (Constitutent) Kernels of an Unresolved Conflict

;;; (KERNELS-OF-THE-SMALLEST-UNRESOLVED-CONFLICT
;;;    (decision-space decision-space-w-conflicts) CURRENT-STATE-ASSIGNMENTS
;;;    &OPTIONAL DISALLOWED-ASSIGNMENTS)
;;;  
;;;   Finds the constituent kernels of a conflict of SEARCH PROBLEM 
;;;   such that: 

;;;     1) conflict is not resolved by CURRENT-STATE-ASSIGNMENTS.
;;;     2) conflict contains the smallest number of assignments.  

;;;   Returns the kernels sorted in decreasing order of cost.  
;;;   Trims each leading kernel that is inconsistent with 
;;;   CURRENT-STATE-ASSIGNMENTS.  
;;;   Note: some inconsistent constituent kernels may still reside 
;;;   in the list.
;;;   
;;;   If all conflicts are resolved then the function returns nil.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do:
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Fix:
;;; o In new-conflict, the constituent kernels of the conflicts 
;;;    are properly sorted; however, the cost function used is not correct.
;;;    See the the JDAM paper, and implement properly.

;;; To Do:
;;; o Implement resupport, as described in the initial notes,
;;;   so that conflict management is incremental.

;;; o Factor out conflict management, as a separate object 
;;;   and set of methods.

;;;   Include several implementations:
;;;   - current conflict management.
;;;   - conflict management with domain info recorded.
;;;   - conflicts replaced with full propositional clauses.

;;; o Reason about variable domains within the conflict data base:
;;;   - Add in conflicts representing:
;;;     a) collective exhaustion and 
;;;     b) mutual exclusion of variable domains (needed?).
;;;   - Detecting resolved conflicts will require incremental unit 
;;;     propagation.
;;;  
;;; o If possible, separate this module from search-problem as a distinct
;;;   class, or a subclass of search-problem.

;;; o For conditional CSPs:
;;;;  - Modify implementation of suboptimality conflicts.
;;;   - Test.

;;;   - Change the definition of resolved so that a conflict is resolved 
;;;     if supported OR if CONFLICT IS INACTIVE.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; BFIG Interface Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Adding and Removing Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Optimization:  
;;; o This function allows redundant conflicts because the cost 
;;;   of looking for redundant conflicts is assumed to be greater 
;;;   than the benefit of removing them.  This requires confirmation.
;;;   Redundant conflicts can be removed by implementing a TRIE
;;;   structure, as in the IPIA algorithm of de Kleer.

;;; Note - Correction made from Livingstone:
;;; o Within Livingstone, the inconsistent-assignments of a conflict
;;;   were sorted in best first order, and assigned during search in 
;;;   this order.  To search in best first order, instead the constituent 
;;;   kernels should be sorted in best first order, since this is the set 
;;;   that is being selected from during search.  
;;;   This bug is corrected below.

(defmethod insert-conflict! 

	((conflict conflict) 
	 (state-space state-space-w-conflicts)
	 &optional (disjoint? t))
  (declare (ignore disjoint?)) ;;TODO remove

  "Inserts CONFLICT object into STATE-SPACE where
   conflicts of STATE-SPACE are maintained in increasing order of
   cardinality.  Updates the record of whether or not 
   conflict is resolved.
   RETURNS conflict.

   Note: does not try to detect and remove redundant conflicts.
   DISJOINT? is True iff search is generating search nodes denoting
   disjoint sets of candidates (e.g., as opposed to overlapping kernels).

   Side Effect: Maintains the list of conflicts for
   STATE-SPACE ordered from smallest to largest set of
   constituent kernels.  This is used by search to implement
   a most constrained variable heuristic, when splitting to
   resolve a conflict."
  
  (debug-print-for :bfig t "Inserting conflict ~a." 
				   conflict)

  ;; Maintain the list of conflicts for search problem
  ;; in increasing order of cardinality.  
  ;;   (setf (state-space-conflicts state-space)
  ;;     (insert-in-order
  ;;      conflict
  ;;      (state-space-conflicts state-space)
  ;;      :order #'<=
  ;;      :key #'(lambda (conf)
  ;; 			  (length (conflict-constituent-kernels conf)))))
  
  (with-slots (conflicts) state-space

	;;Unless there is no previous conflict, the new one is one higher
	(let ((last-conflict (first conflicts)))
	  (unless (null last-conflict)
		(setf (conflict-number conflict) (+ 1 (conflict-number last-conflict)))))

	(incf (state-space-conflict-num state-space))
	;;Keep the conflicts in order.
	(push conflict conflicts))
    
  ;; Record conflict on each of its constituent kernels (implicants).
  ;; Used to detect resolved conflicts.
  ;;   (loop for assignment in (conflict-constituent-kernels conflict)
  ;;       do (push conflict (state-assignment-resolved-conflicts assignment)))
  
  ;; Record conflict on each of its assignments.
  ;; Used to detect resolved conflicts.
  ;;   (loop for assignment in (conflict-inconsistent-assignments conflict)
  ;; 	 do (push conflict (state-assignment-conflicts assignment)))

;;;Removed because the conflict list is not maintained in order correctly anymore
  ;; Record if conflict is resolved by the current state.
  ;;  (try-supporting-conflict! state-space conflict disjoint?)
  
  ;; and update count of conflicts remaining to be resolved.
  ;;  (unless (conflict-supported? state-space conflict)
  ;;    (incf (state-space-unresolved-conflict-count state-space)))
  t)

(defmethod insert-new-conflict 

    ((state-space state-space-w-conflicts)
	 &key inconsistent-assignments)
  
  "Inserts a new conflict into STATE-SPACE.  A conflict is a set
   of variable assignments, INCONSISTENT-ASSIGNMENTS, that are mutually 
   inconsistent.  A conflict contains at most one assignment 
   for each variable in STATE-SPACE.  INCONSISTENT-ASSIGNMENTS 
   can be a list of either state-assignment objects or 
   assignment descriptions, of the form (= variable-name value)."

  (with-slots (conflict-type) state-space
	(let ((conflict (make-instance conflict-type
					  :state-space state-space
					  :inconsistent-assignments inconsistent-assignments
					  ;;Set the initial activity
					  :activity (state-space-conflict-bump-inc state-space))))
	
	  (insert-conflict! conflict state-space)

	  ;;When a new conflict is learned, decay the activities
	  (decay-conflict-activity! state-space)
	  ;;When growing the conflict list, check if it's time to 
	  ;;remove conflicts.
	  (manage-conflict-list! state-space )

	  conflict)))

(defmethod remove-all-conflicts!

    ((state-space state-space-w-conflicts))
  
  "Removes the set of all conflicts that are currently 
   associated with STATE-SPACE.  Updates any information 
   recorded about the status of these conflicts within 
   STATE-SPACE.

   Note: The function does not delete the conflict objects."

  ;;; Clear the record of current search node assignments 
  ;;; supporting resolved conflicts.
  (clear-resolved-conflicts! state-space)
  
  ;;; Clear the set of conflicts.
  (setf (state-space-conflicts state-space) nil)

  ;; Remove pointers from each assignment to each conflict resolved
  ;; by assignment or containing assignment.
  (with-slots (variables) state-space
	(loop for variable in variables
		do (loop for assignment in (state-variable-domain variable)
			   do (setf (state-assignment-resolved-conflicts assignment) nil)
				  (setf (state-assignment-conflicts assignment) nil))))

  ;; Record that all conflicts are resolved (since there are no conflicts).
  (setf (state-space-unresolved-conflict-count state-space) 
    0)
  
  state-space)

(defmethod remove-conflict! 
	((conflict conflict) (state-space state-space-w-conflicts))
  
  "Remove a conflict from the problem. Remove it from the central repository
   in the state-space and also from the assignments."

  ;;Remove it from the list of current conflicts.
  ;; (with-slots (conflicts) state-space
  ;; 	(setf conflicts (delete conflict conflicts)))

  ;;Mark that it has decayed.
  (setf (conflict-decayed? conflict) t)
  (decf (state-space-conflict-num state-space))

  ;; Unrecord conflict on each of its constituent kernels (implicants).
  ;; Used to detect resolved conflicts.
  ;;  (loop for assignment in (conflict-constituent-kernels conflict)
  ;; 	 do (setf (state-assignment-resolved-conflicts assignment)
  ;; 			  (remove conflict
  ;;                 (state-assignment-resolved-conflicts assignment))))
  
  ;; Unrecord conflict on each of its assignments.
  ;; Used to detect resolved conflicts.
  ;;   (loop for assignment in (conflict-inconsistent-assignments conflict)
  ;; 	 do (setf (state-assignment-conflicts assignment)
  ;; 			  (remove conflict (state-assignment-conflicts assignment))))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Additional Conflict Module Interface functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Updating The Current Search State
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Updated comments.
;;; o Factored out clearing of effects.

;;; Changed: 
;;; o Updated To Do.

;;; To Do:
;;; o Does variable activation rely on the binding of
;;;   state-variable-current-cdb-assignment?
;;;   If so, then
;;;   - generalize the following from state-space-w-conflicts 
;;;     to state-space and move to a different file.
;;;   - Move slot current-cdb-assignment from
;;;     state-variable-w-conflicts, to state-variable
;;;     and rename -cdb- to something like -candgen-.
(defmethod clear-current-assignments!

	((state-space state-space))
  
  "Clear the record of any current assignments
   for STATE-SPACE and the record of any conflicts 
   they support."
  
  ;; ...the record of its assignments and
  (loop for variable in (state-space-variables state-space)
      do (setf (state-variable-current-cdb-assignment 
				variable) 
		   nil))

  ;; ... all effects propagated by those assignments.
  (clear-effects-of-current-assignments! state-space)
  T)

;;; Changed:
;;; o New method.
(defmethod clear-effects-of-current-assignments!
	
	((state-space state-space))
  
  "Clear the effects propagated for any current 
   assignments for STATE-SPACE."
  
  T)

;;; Changed:
;;; o New method.
(defmethod clear-effects-of-current-assignments!
	
	((state-space state-space-w-conflicts))
  
  "Clear the effects propagated for any current 
   assignments for STATE-SPACE."
  
  (clear-resolved-conflicts! state-space))

;;; Changed:
;;; o Calls propagate-current-assignment! as an
;;;   intermediate abstraction from conflicts.
;;; o Generalized to state-space.
;;; o Updated To Do.

(defmethod record-assignment-as-current! 

	((state-space state-space)
	 (new-assignment state-assignment))


  "Record NEW-ASSIGNMENT as current in STATE-SPACE, and 
   identifies all conflicts supported by the current assignments.

   Assumptions:

   - NEW-ASSIGNMENT is not one of the current assignments,
     and does not support any consequences.
  
   Side Effects:  

   -  Records NEW-ASSIGNMENT on its respective 
      STATE-VARIABLE-CURRENT-CDB-ASSIGNMENTs
      of STATE-SPACE.

   -  All consequences of NEW-ASSIGNMENT, are recorded."

  (with-slots (variable) new-assignment

	(setf (state-variable-current-cdb-assignment variable) 
	  new-assignment)
		 
	(propagate-current-assignment! state-space new-assignment)))

;;; Changed:
;;; o New method.
(defmethod propagate-current-assignment! 

	((state-space state-space)
	 (new-assignment state-assignment))

  "Identifies all effects of recording NEW-ASSIGNMENT 
   as current in STATE-SPACE.

   Assumptions:

   - NEW-ASSIGNMENT is recorded as a current assignment,
     BUT does not support any consequences.
  
   Side Effects:  

   -  Records all consequences of NEW-ASSIGNMENT."

  T)

;;; Changed:
;;; o New method.
(defmethod propagate-current-assignment! 

	((state-space state-space-w-conflicts)
	 (new-assignment state-assignment-w-conflicts))

  "Identifies all effects of recording NEW-ASSIGNMENT 
   as current in STATE-SPACE.

   Assumptions:

   - NEW-ASSIGNMENT is recorded as a current assignment,
     BUT does not support any consequences.
  
   Side Effects:  

   -  Records all consequences of NEW-ASSIGNMENT."

  (support-conflicts-by-assignment! state-space new-assignment))

;;; Changed:
;;; o Calls undo-consequences-of-current-assignment! as an
;;;   intermediate abstraction from conflicts.
;;; o Generalized to state-space.
;;; o Updated comments.
;;; o Updated To Do.

;;; Note: UNDO-CONSEQUENCES-OF-CURRENT-ASSIGNMENT!
;;;       must be called AFTER clearing the current-cdb-assignment
;;;       slot.
(defmethod unrecord-assignment-as-current! 

	((state-space state-space)
	 (old-assignment state-assignment))

  "Removes the record of OLD-ASSIGNMENT as current in STATE-SPACE,
   and the record of conflicts supported by this assignment.

  Assumptions:

   - OLD-ASSIGNMENT is included in the current assignments of STATE-SPACE.

   - Those conflicts of STATE-SPACE whose inconsistent assignments
     are ruled out by current allowed/disallowed assignments are 
     supported by one of the assignments in the extension.
  
  Side Effects:  

   - Clears OLD-ASSIGNMENT from current assignments.
   - Clears the conflicts supported by OLD-ASSIGNMENT."

  (with-slots (variable) old-assignment

	;; Update the current assignment to reflect the removal of OLD-ASSIGNMENT.
	(setf (state-variable-current-cdb-assignment variable) 
	  nil)

	;; Note: undo-consequences-of-current-assignment!
	;;       requires that current-cdb-assignment be up to date when called.
	(undo-consequences-of-current-assignment! state-space old-assignment)
	
	T))

;;; Changed:
;;; o New method.
(defmethod undo-consequences-of-current-assignment! 

	((state-space state-space)
	 (old-assignment state-assignment))

  "Identifies all effects of removing OLD-ASSIGNMENT 
   as current in STATE-SPACE.

   Assumptions:

   - OLD-ASSIGNMENT is still recorded as a current assignment,
     AND consequences of assignment are recorded.
  
   Side Effects:  

   -  Removes all consequences of OLD-ASSIGNMENT.
   -  Leaves OLD-ASSIGNMENT recorded as current."

  T)

;;; Changed:
;;; o New method.
(defmethod undo-consequences-of-current-assignment! 

	((state-space state-space-w-conflicts)
	 (old-assignment state-assignment-w-conflicts))

  "Identifies all effects of removing OLD-ASSIGNMENT 
   as current in STATE-SPACE.

   Assumptions:

   - OLD-ASSIGNMENT is still recorded as a current assignment,
     AND consequences of assignment are recorded.
  
   Side Effects:  

   -  Removes all consequences of OLD-ASSIGNMENT.
   -  Leaves OLD-ASSIGNMENT recorded as current."

  (unsupport-conflicts-by-assignment! state-space old-assignment))

;;; Changed:
;;; o Updated comments.
(defmethod record-disallowed-assignment-as-current! 

	((state-space state-space-w-conflicts)
	 (disallowed-assignment state-assignment-w-conflicts))

  "Record DISALLOWED-ASSIGNMENT as current in STATE-SPACE, 
   and support all unsupported conflicts in which one of 
   its inconsistentassignments is rules out by DISALLOWED-ASSIGNMENT.

   Assumptions:

   - DISALLOWED-ASSIGNMENT is not one of the current diasallowed
     or allowed assignments, and does not support any conflict.
  
   Side Effects:  

   -  Records DISALLOWED-ASSIGNMENT on its respective 
      STATE-VARIABLE-CURRENT-CDB-DISALLOWED-ASSIGNMENTs
      of STATE-SPACE.

   -  Each conflict resolved by DISALLOWED-ASSIGNMENT, is
      supported by DISALLOWED-ASSIGNMENT or some other 
      current assignment."

  (with-slots (variable) disallowed-assignment

	(setf (state-variable-current-cdb-disallowed-assignment variable) 
	  disallowed-assignment)
		 
	(support-conflicts-by-disallowed-assignment! 
	 state-space disallowed-assignment)))

;;; Changed:
;;; o Updated comments.
(defmethod unrecord-disallowed-assignment-as-current! 

	((state-space state-space-w-conflicts)
	 (disallowed-assignment state-assignment-w-conflicts))

  "Remove the record of DISALLOWED-ASSIGNMENT as current in 
   STATE-SPACE,and the record of conflicts supported by this 
   assignment.

  Assumptions:

   - DISALLOWED-ASSIGNMENT is included in the current disallowed
     assignments of STATE-SPACE.

   - Those conflicts of STATE-SPACE whose inconsistent assignments
     are ruled out by current allowed/disallowed assignments are 
     supported by one of the assignments in the extension.
  
  Side Effects:  

   - Clears DISALLOWED-ASSIGNMENT from current assignments.
   - Clears conflicts supported by DISALLOWED-ASSIGNMENT."

  (with-slots (variable) disallowed-assignment
	
	(unsupport-conflicts-by-disallowed-assignment! 
	 state-space disallowed-assignment)

	(setf (state-variable-current-cdb-disallowed-assignment variable) 
	  nil)
	T))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Query the Status of Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



;;; Changed:
;;; o Updated comments.
(defmethod unresolvable-conflict-in-ancestors?

    ((search-problem search-problem-w-conflicts))
  
  "Returns True IFF any known conflicts cannot be resolved by any child
   of the current search node. This is true if all of the inconsistent assignments
   in the conflict are actually present in the current assignment of the node."

  (with-slots (current-search-node conflicts) 
	  search-problem
	
	(with-slots (ancestor-assignments ancestor-assignments-bit) 
		current-search-node
	  
	  ;;There is an unresolvable conflict with the current node if the ancestor
	  ;;assignments exactly match a conflict
	  (loop for assignment in ancestor-assignments
		  thereis (loop for conflict in (state-assignment-conflicts assignment)
					  thereis (and (conflict-active? conflict)
								   (bit-rep-unresolvable-conflict? 
									(conflict-inconsistent-assignments-bit
									 conflict)
									ancestor-assignments-bit search-problem)))))
	
	;; (let ((current-assignments 
	;; 	   (remove nil
	;; 			   (append (search-node-ancestor-assignments current-search-node) 
	;; 					   (list (search-node-preferred-child-assignment current-search-node))))))
	;;   ;;  (format t "new way: ~a~%" current-assignments)
  
	;;   ;;Try all the assignments
	;;   (not (loop for assignment in current-assignments
	;; 		  ;;Every assignment
	;; 		  always (loop for conflict in (state-assignment-conflicts assignment)
	;; 					;; 		  do (format t "new way conflict: ~a~%" conflict)
	;; 					;;does not meet any conflicts
	;; 					never (and (conflict-active? conflict)
	;; 							   (loop for inconsistent-assignment in (conflict-inconsistent-assignments conflict)
	;; 								  ;;meaning that all the inconsistent assigns are in the current
	;; 								  always (member inconsistent-assignment current-assignments))))))
  

	)
  ;;   ;;Check all the conflicts explicitly
  ;;   ;(null (smallest-unresolved-conflict-of-current-state search-problem))
  ;;   ;; 	(format t "~a ~a ~%" new-way old-way)

  ;; )
  ;;Old method
  ;;  (zerop (state-space-unresolved-conflict-count state-space))
  )

(defmethod unresolvable-conflict-in-full-assignment?

    ((search-problem search-problem-w-conflicts))
  
  "Returns True IFF any known conflicts cannot be resolved by any child
of the current search node. This is true if all of the inconsistent assignments
in the conflict are actually present in the current assignment of the node."

  (with-slots (current-search-node conflicts assignment-bit-ht) search-problem
	(with-slots (ancestor-assignments-bit) current-search-node

	  (let ((current-assignments 
			 (remove nil
					 (append (search-node-ancestor-assignments current-search-node)
							 (list (search-node-preferred-child-assignment
									current-search-node)))))
			(current-assignment-bit		;get the bit vector version of the 
			 (append-assignment-to-bit-clause 
			  (search-node-ancestor-assignments-bit current-search-node)
			  (gethash (search-node-preferred-child-assignment
						current-search-node) assignment-bit-ht))))
		
		;;There is an unresolvable conflict with the current node if the ancestor
		;;assignments exactly match a conflict
		(loop for assignment in current-assignments
			thereis (loop for conflict in (state-assignment-conflicts assignment)
										  ;; (loop for conflict in conflicts
						thereis (and (conflict-active? conflict)
									 (bit-rep-unresolvable-conflict? 
									  (conflict-inconsistent-assignments-bit
									   conflict)
									  current-assignment-bit search-problem) 
									 search-problem)))
										;)
		))
	
	;; 	(let ((current-assignments 
	;; 	   (remove nil
	;; 			   (append (search-node-ancestor-assignments current-search-node) 
	;; 					   (list (search-node-preferred-child-assignment current-search-node))))))
	;;  (format t "new way: ~a~%" current-assignments)
  
	;;   ;;Try all the assignments
	;;   (not (loop for assignment in current-assignments
	;; 		  ;;Every assignment
	;; 		  always (loop for conflict in (state-assignment-conflicts assignment)
	;; 					;; 		  do (format t "new way conflict: ~a~%" conflict)
	;; 					;;does not meet any conflicts
	;; 					never (and (conflict-active? conflict)
	;; 							   (loop for inconsistent-assignment in (conflict-inconsistent-assignments conflict)
	;; 								  ;;meaning that all the inconsistent assigns are in the current
	;; 								  always (member inconsistent-assignment current-assignments))))))
  

	
	;;   ;;Check all the conflicts explicitly
	;;   ;(null (smallest-unresolved-conflict-of-current-state search-problem))
	;;   ;; 	(format t "~a ~a ~%" new-way old-way)

	;; )
	;;Old method
	;;  (zerop (state-space-unresolved-conflict-count state-space))
	))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Extracting (Constitutent) Kernels of an Unresolved Conflict
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Notes: 

;;; o When expanding a node, in order to resolve a conflict, 
;;;   it is guaranteed that none of the (constituent) kernels
;;;   of the conflict are mentioned by node's ancestor assignments. 
;;;   This is because we check that conflict is unresolved 
;;;   by the current state.

;;; o Bug fixed from Livingstone implementation:
;;;   - In Livingstone kernels-of-the-smallest-unresolved-conflict
;;;   determined the conflicts resolved by the current-search-state, 
;;;   not the current partial-kernel.  In Livingstone the 
;;;   current-search-state differs from the current partial kernel,
;;;   in that any variable unassigned in the partial kernel is given 
;;;   its best cost assignment.  These additional assignments may resolve 
;;;   conflicts that are not resolved by the partial kernel.
;;;   Hence the routine may return an assignment as a kernel, even
;;;   though some conflicts are unresolved.  In this case the returned
;;;   assignment can be extended to infeasible states.
;;;   
;;;   - Also note that because Livingstone use the current search state
;;;     as a full assignment, during trimming it eliminated the best cost 
;;;     assignment from the constituent kernels.  As a result, Livingstone
;;;     would not generate some valid kernels, hence it could eliminate
;;;     feasible states.

;;; Check:
;;; o Do we need to pass in current-assignments, or can we pick them up
;;;   from state-space?
;;; o Is this returning the smallest unresolved conflict, or the
;;;   unresolved conflict with the smallest list of kernels?
;;;   We want the latter.

;;; FIX:
;;; o Compute the cost of kernels correctly according to the JDAM paper.
(defmethod kernels-of-the-smallest-unresolved-conflict

    ((state-space state-space-w-conflicts) 
	 current-assignments 
	 smallest-conflict
     &optional (disallowed-assignments '()))
  
  "Returns the constituent kernels of a conflict of STATE-SPACE
   such that: 

     1) conflict is not resolved by CURRENT-ASSIGNMENTS.
     2) conflict contains the smallest number of assignments.  

   Returns the constituent kernels, while trimming leading kernels 
   that are inconsistent with CURRENT-ASSIGNMENTS or 
   DISALLOWED-ASSIGNMENTS.  

   If STATE-SPACE is a DECISION-SPACE, then the constitutent kernels
   are sorted in decreasing order of best cost.

   Note: Some inconsistent constituent kernels may reside in the list.

   If all conflicts are resolved then nil is returned.

   If DISJOINT? is true, then splitting during search ensures that 
   sibling search nodes denote disjoint sets of candidate assignments.

   ASSUMES, upon entry, that CURRENT-ASSIGNMENT has already been 
   installed as the current search state."

  ;;Take the smallest conflict as an input to avoid rechecking
  (let ((conflict smallest-conflict
;;;               (smallest-unresolved-conflict-of-current-state state-space)
				  ))
	
	;; Get the smallest unresolved conflict.
    (cond ((not (null conflict))
		   ;; Get conflict's kernel assignments, 
		   ;; listed in decreasing best cost, and with
		   ;; leading inconsistent kernels trimmed.
		   (let ((kernels
				  (trim-leading-inconsistent-assignments
				   (conflict-constituent-kernels conflict)
				   current-assignments
				   disallowed-assignments)))
			 (debug-print-for :bfig t "    Smallest unresolved conflict = ~a." 
							  conflict)
			 (debug-print-for :bfig t "    Kernels of unresolved = ~a."
							  kernels)
			 kernels))
      
		  ;; Otherwise, all conflicts are resolved, return nil.
		  (:otherwise
		   (debug-print-for :bfig t "    No unresolved conflicts." 
							conflict)
		   nil))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The (Constituent) Kernels of a Conflict
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod kernels-of-conflict 
	
	((conflict conflict))

  "Takes as input a CONFLICT object.
   RETURNS a new list of assignments, called constitutent kernels,
   each of which resolves CONFLICT.  A constituent kernel resolves a 
   conflict if, given kernel, at least one of the assignments in the 
   conflict can not hold.

   IMPORTANT:  The constitutent kernels returned by this function
   are NOT SORTED in order of cost.

   NOTE: Since the list is new, it may be destructively modified;
   for example, by sort."
  
  (loop for assignment in (conflict-inconsistent-assignments conflict)
      for alternate-assignments = 
		(state-assignment-alternate-assignments assignment)
		;; Since ASSIGNMENT is in CONFLICT, and ALTERNATE-ASSIGNMENTS is 
		;; inconsistent with ASSIGNMENT, it follows that, if 
		;; one of ALTERNATE-ASSIGNMENTS holds, then at least one of the
		;; assignments in CONFLICT cannot occur.  Hence each member of
		;; ALTERNATE-ASSIGNMENTS resolves CONFLICT -- its a kernel.

      nconcing (copy-list alternate-assignments) 
      into conflict-kernels
		   
      finally 
		(return (delete-duplicates conflict-kernels :test #'eql))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Identifying Unresolved Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Updated comments.
;;; o Modified unresolved count to active conflicts.
(defmethod clear-resolved-conflicts! 

	((state-space state-space-w-conflicts))
  
  "Clear the record of all conflicts that are resolved by the current
   search node of STATE-SPACE."
  
  (loop with conflict-count = 0
      for conflict in (state-space-conflicts state-space)
					  
      do (setf (conflict-support conflict) nil)
		 (when (conflict-active? conflict)
		   (incf conflict-count))
		 
	  finally
		;; All active conflicts are unresolved.
		(setf (state-space-unresolved-conflict-count state-space)
		  conflict-count))
  
  t)

;;; The following functions maintain information related to the conflicts of 
;;; a search problem.  For each conflict RECORD-CONSTITUENT-KERNELS-OF-CONFLICT
;;; computes a set of assignments, called constitutent kernels, each of which 
;;; resolves the conflict (i.e., are prime implicants).  These constituent 
;;; kernels are returned sorted in order of decreasing best cost.  They are
;;; also stored with conflict, and may be retrieved in sorted order using 
;;; CONFLICT-CONSTITUENT-KERNELS.  This ordering is exploited by the best 
;;; first candidate generator.

;;; Given a CURRENT SEARCH NODE and corresponding PARTIAL STATE 
;;; (a set of assignments), a conflict 
;;; (v1 = m1, v2 = m2, ...., vn = mn) is resolved precisely when at least 
;;; one of the state's variables vj has a value in the state different from mj.
;;; A node's PARTIAL STATE consists of its ancestor assignments plus its 
;;; best child assignment.
;;; Whether or not a conflict is resolved is determined by calling 
;;; CONFLICT-RESOLVED?.  ALL-CONFLICTS-RESOLVED? determines whether or not 
;;; all conflicts associated with SEARCH PROBLEM are resolved by CURRENT 
;;; PARTIAL STATE.  If one or more conflicts are not resolved, then 
;;; the lowest cardinality unresolved conflict may be retrieved using
;;; SMALLEST-UNRESOLVED-CONFLICT.  

;;; The resolution of a conflict is computed by maintaining an assignment in 
;;; the current partial state that resolves the conflict, called the 
;;; CONFLICT SUPPORT.  Support of all conflicts is determined by incrementally
;;; maintaining a count of all unsupported conflicts.

;;; Note:
;;;  The slot STATE-VARIABLE-CURRENT-CDB-ASSIGNMENT is
;;;  updated whenever changing between current search nodes, 
;;;  so that current-search-node-best-extension-has-assignment?
;;;  can be tested in constant time, rather than time that is linear 
;;;  in the the number of assignments made by the current search node.
;;;  The gain for the little added complexity may be small.

;;; Optimization:
;;; o The test to determine if any of the disallowed assignments D 
;;    conflict with assignments C is O(|C||D|).  This could
;;;   be reduced to  O(|C|+ |D|) by recording each current 
;;;   disallowed assignment with its corresponding variable.  
;;;   The gain for this optimization may be negligible.

;;; Changed:
;;; o Updated comments.
(defmethod try-supporting-conflict! 

	((state-space state-space-w-conflicts)
;;;	 (search-problem informed-search-problem-w-conflicts)
	 (conflict conflict)
	 disjoint?)
  
  "If the current search state contains an assignment 
   that supports CONFLICT, then record this assignment as
   CONFLICT's current support; otherwise, record NIL as
   CONFLICT's support.

   An assignment supports a CONFLICT if it ensures that at 
   least one of CONFLICT's inconsistent assignments can not 
   arise.

   DISJOINT? is TRUE iff search is generating search-nodes 
   that denote disjoint sets of candidates."

  (unless (or (support-with-current-assignments conflict)

			  (and disjoint?
				   (support-with-current-disallowed-assignments
					;; Fix delete comments when works
;;;					search-problem
					conflict)))
	
	;; No support found, record that there is none.
    (setf (conflict-support conflict) nil)))

;;; Changed:
;;; o Updated comments.
(defmethod support-with-current-assignments
	
	((conflict conflict))

  "Tries to support CONFLICT using one of the current assignments
   inserted into the conflict database.  
   A conflict is SUPPORTED by an assignment if assignment ensures
   that at least one of CONFLICT's inconsistent assignments cannot hold.
   Returns True iff conflict is supported, otherwise NIL."
  
  (loop for assignment in (conflict-constituent-kernels conflict)

      when (cdb-current-assignments-contain-assignment? assignment)
		   ;; One of the constituent kernels of conflict is in the 
		   ;; assignments selected by the current search node, hence 
		   ;; the search node resolves the conflict.
		   
      do (setf (conflict-support conflict) assignment)
		 (debug-print-for :bfig t "    ~a resolves ~a." 
						  assignment conflict)
		 (return T)
		 
      finally (return nil)))

;;; Changed:
;;; o Updated comments.
(defmethod support-with-current-disallowed-assignments
	
	((conflict conflict))

  "Tries to support CONFLICT using one of the current disallowed
   assignments inserted into the conflict database.
   A conflict is SUPPORTED by an assignment if assignment ensures
   that at least one of CONFLICT's inconsistent assignments cannot hold.
   Returns True iff conflict is resolved, otherwise NIL."

  (loop for disallowed in (conflict-inconsistent-assignments conflict)
	  when (cdb-current-disallowed-assignments-contain-assignment? 
			disallowed)
		   
		   ;; One of conflict's assignments is disallowed 
		   ;; by the current search node, hence 
		   ;; the search node resolves the conflict.
	  do (setf (conflict-support conflict) disallowed)
		 (debug-print-for :bfig t "    disallowed ~a resolves ~a." 
						  disallowed conflict)
		 (return T)
	  finally (return nil)))

;;; Fix delete comments when works

;;; To Do:
;;; o Generalize method from search-problem to state-space, by
;;;   - associating disallowed-assignments for the current state
;;;     with the conflict database.
;;;   - Then mimic support-with-current-assignments 
;;;(defmethod support-with-current-disallowed-assignments 
;;;	
;;;	((search-problem informed-search-problem) 
;;;	 (conflict conflict))
;;;  
;;;  "Tries to resolve CONFLICT with one of the assignments
;;;   recorded as disallowed by the current search node of SEARCH-PROBLEM  
;;;   Returns True if conflict is resolved, otherwise NIL."
;;;
;;;  (let ((current (search-problem-current-search-node search-problem)))
;;;
;;;    (unless (null current)
;;;
;;;      (let ((disallowed-assignments (search-node-disallowed-assignments
;;;									 current)))
;;;
;;;		(support-conflict-with-disallowed-assignments
;;;		 conflict disallowed-assignments)))))

;;;(defmethod support-conflict-with-disallowed-assignments
;;;
;;;	((conflict conflict) disallowed-assignments)
;;;  
;;;  "Tries to resolve CONFLICT based on DISALLOWED-ASSIGNMENTS.
;;;   Returns True if conflict is resolved, otherwise NIL."
;;;		  
;;;  (unless (null disallowed-assignments)
;;;
;;;	(loop for conflict-assign in (conflict-inconsistent-assignments
;;;								  conflict)
;;;
;;;		when (member conflict-assign disallowed-assignments :test #'eql)
;;;			 ;; One of the disallowed assignments of 
;;;			 ;; current-search-node is in conflict, and hence 
;;;			 ;; resolves conflict.
;;;
;;;		do 
;;;		  (setf (conflict-support conflict) conflict-assign)
;;;		  ;;(debug-print-for :bfig t "    disallowed ~a resolves ~a." 
;;;		  ;;		      assignment conflict)
;;;		  (return T)
;;;				 
;;;		finally (return nil))))

(defmethod cdb-current-assignments-contain-assignment? 
	((assignment state-assignment))
  
  "Returns true iff ASSIGNMENT is asserted as a current
   assignment within the confict database of state space."
  
  (eql assignment (state-variable-current-cdb-assignment 
                   (state-assignment-variable assignment))))


;;; Changed:
;;; o Updated comments.
(defmethod cdb-current-disallowed-assignments-contain-assignment? 
	((assignment state-assignment-w-conflicts))
  
  "Returns true iff ASSIGNMENT is asserted as a current
   disallowed assignment within the confict database of state space."
  
  (eql assignment (state-variable-current-cdb-disallowed-assignment 
                   (state-assignment-variable assignment))))

;;; Changed:
;;; o Updated comments.
;;; o Modified decf.
(defmethod support-conflicts-by-assignment! 

	((state-space state-space-w-conflicts) 
	 (assignment state-assignment-w-conflicts))

  "Tries supporting each of the unsupported conflicts of STATE-SPACE
   with ASSIGNMENT. A conflict is supported by ASSIGNMENT
   if ASSIGNMENT ensures that the conflict's set of inconsistent 
   assignments can't arise."

  ;; Note: In general conflict-support could be either a current
  ;;       assignment or disallowed assignment, however an assignment
  ;;       can't be both made and disallowed by the current node.
  ;;       Hence if assignment is current, and assignment supports
  ;;       conflict, it must be a support as a resulting of being
  ;;       a current assignment
  
  ;; Iterate over each conflict that can be resolved by assignment.
  (loop for conflict in (state-assignment-resolved-conflicts assignment)
	  unless (conflict-supported? state-space conflict)

	  do (setf (conflict-support conflict) assignment)
		 
		 ;; If conflict is active (all variables active), 
		 ;; then conflict just went from unresolved to resolved.
		 ;; Note: Unconditional variables are always active.
		 (when (conflict-active? conflict)
		   (decf (state-space-unresolved-conflict-count state-space)))))

;;; To Do:
;;; o Make this incremental.  When support is lost, try to resupport.
;;;   Only resupport AFTER all current and disallowed assignments have
;;;   been removed.

;;; Optimization:
;;; o Rather than maintaining support at all times, we could
;;;   only maintain support while conflict is active.
;;;   - Every time conflict goes from inactive to active,
;;;     recompute its support from scratch.

;;; Changed:
;;; o Updated comments.
;;; o Modified test on incf.

;;; Note:
;;; o When a conflict becomes unsupported, we do not try to find
;;;   an alternative source of support.  This is because the conflict
;;;   support algorithm is not designed to be incremental.
(defmethod unsupport-conflicts-by-assignment! 
	((state-space state-space-w-conflicts) 
	 (assignment state-assignment-w-conflicts))

  "For each resolved conflict of STATE-SPACE, if conflict is 
   supported by ASSIGNMENT, then removes that support.

   The function DOES NOT try to find an alternative source 
   of support for a conflict that becomes unsupported."
 
  ;; Note: In general conflict-support could be either a 
  ;;       current assignment or disallowed assignment, however 
  ;;       an assignment can't be both made and disallowed by 
  ;;       the current node.  Hence if assignment is current, and 
  ;;       assignment supports conflict, it must be a support as a 
  ;;       result of being a current assignment.
  (loop for conflict in 
		(state-assignment-resolved-conflicts assignment)

	  when (eql assignment (conflict-support conflict))

	  do (setf (conflict-support conflict) nil)
		 
		 ;; If conflict is active (all variables active),
		 ;; then conflict went from resolved to unresolved.
		 ;; Note: Unconditional variables are always active.
		 (when (conflict-active? conflict)
		   (incf (state-space-unresolved-conflict-count state-space)))))

;;; Changed:
;;; o Updated comments.
;;; o Modified test on decf.
(defmethod support-conflicts-by-disallowed-assignment! 
    ((state-space state-space-w-conflicts) 
	 (disallowed-assignment state-assignment-w-conflicts))

  "Tries supporting each of the conflicts of STATE-SPACE
   using ASSIGNMENT.  A conflict is supported by ASSIGNMENT
   if ASSIGNMENT  ensures that the conflict's inconsistent 
   assignments can't arise."

  ;; Note: In general conflict-support could be either a current
  ;;       assignment or disallowed assignment, however an assignment
  ;;       can't be both made and disallowed by the current node.
  ;;       Hence if assignment is disallowed, and assignment supports
  ;;       conflict, it must be a support as a resulting of being
  ;;       disallowed.
  (loop for conflict in (state-assignment-conflicts disallowed-assignment)

	  unless (conflict-supported? state-space conflict)

	  do (setf (conflict-support conflict) disallowed-assignment)

		 ;; If conflict is active (all variables active),
		 ;; then conflict went from unresolved to resolved.
		 ;; Note: Unconditional variables are always active.
		 (when (conflict-active? conflict)
		   (decf (state-space-unresolved-conflict-count state-space)))))

;;; To Do:
;;; o Make this incremental.  When support is lost, try to resupport.
;;;   Only resupport AFTER all current and disallowed assignments have
;;;   been removed.

;;; Changed:
;;; o Updated comments.
;;; o Modified test on incf.
(defmethod unsupport-conflicts-by-disallowed-assignment!
    ((state-space state-space-w-conflicts) 
	 (disallowed-assignment state-assignment-w-conflicts))

  "For each conflict of STATE-SPACE, if conflict is 
   supported by ASSIGNMENT, then remove that support.

   The function DOES NOT try to find an alternative source 
   of support for a conflict that becomes unsupported,
   since the implementation is not incremental."
 
  ;; Note: In general conflict-support could be either a current
  ;;       assignment or disallowed assignment, however an assignment
  ;;       can't be both made and disallowed by the current node.
  ;;       Hence if assignment is disallowed, and assignment supports
  ;;       conflict, it must be a support as a resulting of being
  ;;       disallowed.
  (loop for conflict in (state-assignment-conflicts disallowed-assignment)
						
	  when (eql disallowed-assignment (conflict-support conflict))
		   
	  do (setf (conflict-support conflict) nil)
		 ;; If conflict is active (all variables active),
		 ;; then conflict went from resolved to unresolved.
		 ;; Note: Unconditional variables are always active.
		 (when (conflict-active? conflict)
		   (incf (state-space-unresolved-conflict-count state-space)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Returning Unresolved Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; An unresolved conflict is found by recording
;;; the current node's best assignment as the current 
;;; state and checking which conflicts are supported.

;;; Optimization:
;;; o Might be able to detect the first unresolved conflict
;;;   faster by an incremental algorithm based on switching
;;;   current nodes.
;;; o Should be able to identifing the first unresolved conflict
;;;   w/o having to start at the beginning of the list of all conflicts.

;;; Changed:
;;; o modified comments.
;;; o modified defn of conflict-resolved?
(defmethod smallest-unresolved-conflict-of-current-state
    ((state-space state-space-w-conflicts))
  
  "Returns the conflict of STATE-SPACE that:

     1) is not resolved by the current search state, and

     2) has the smallest number of constituent kernel assignments
        (i.e., prime implicants).  

  Returns nil if all conflicts are resolved.

  Assumptions: Conflicts of STATE-SPACE are sorted from
               shortest to longest."

  (with-slots (current-search-node assignment-bit-ht ) state-space
  
	(let ((shortest-conflict nil)
		  (shortest-conflict-length 0)
		  (candidate-found? nil)
		  (current-conflict-length nil)
		  (current-assignments 
		   (remove nil
				   (append (search-node-ancestor-assignments current-search-node) 
						   (list (search-node-preferred-child-assignment
								  current-search-node)))))
		  (current-assignment-bit		;get the bit vector version.
		   (append-assignment-to-bit-clause 
			(search-node-ancestor-assignments-bit current-search-node)
			(gethash (search-node-preferred-child-assignment current-search-node)
					 assignment-bit-ht)))

		  ;; (current-assignments 
		  ;; 		   (remove nil
		  ;; 				   (append (search-node-ancestor-assignments 
		  ;;                      current-search-node) 
		  ;; 					   (list (search-node-preferred-child-assignment 
		  ;;                          current-search-node)))))
		  )
	  
	  (block find-shortest
		;;		(loop for assignment in current-assignments
		;; Try every assignment
		;;		   do 
		(loop for assignment in current-assignments
			do (loop for conflict in (state-assignment-conflicts assignment)
									 ;; (loop for conflict 
									 ;; in (state-space-conflicts state-space) 
									 ;; (state-assignment-conflicts assignment)
									 ;; if any conflicts are unresolved
				   when (and (conflict-active? conflict)
							 (bit-rep-resolvable-conflict?
							  (conflict-inconsistent-assignments-bit conflict)
							  current-assignment-bit
							  state-space)
;;; 	   (loop for kernel-assignment in (conflict-constituent-kernels conflict)
;;; 		  meaning that none of the kernels are in the assignment
;;; 		  never (member kernel-assignment current-assignments))
							 )
						;;See if you should keep it
				   do (setf current-conflict-length
						(length (conflict-inconsistent-assignments conflict)))
					  (when (or (not candidate-found?) ; Keep candidate if this 
										; is the first.
								(< current-conflict-length 
								   shortest-conflict-length)) ; or it's shorter.
						(setf candidate-found? t)
						(setf shortest-conflict-length current-conflict-length)
						(setf shortest-conflict conflict)
										; Use a non-local return to jump back if 
										; you find a unit conflict
										; since you can't do better.
						(when (>= 2 current-conflict-length) 
						  (return-from find-shortest)))
										;)
					  )))
										;)

	
	
	  ;; (loop for conflict in (state-space-conflicts state-space)
			  
	  ;; 	   unless (conflict-resolved? state-space conflict)
	  ;; 	   ;; Conflict is resolved if supported or inactive.
			  
	  ;; 	   ;; This is the first resolved conflict, and since the list of 
	  ;; 	   ;; conflicts is sorted by increasing cardinality of constituent
	  ;; 	   ;; kernels, it is also the smallest unresolved conflict.
			  
	  ;; 	   return conflict
	  ;;     finally (unless (all-conflicts-resolved? state-space)
	  ;;				(error "System error: all conflicts resolved, but ~
	  ;;                       counter is non-zero")
	  ;;)
	  ;; 	   )
	  (unless (null shortest-conflict)
		(bump-conflict! shortest-conflict state-space))
	  shortest-conflict)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Suboptimality Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; A suboptimality conflict is a partial assignment, all of whose
;;; extensions are guaranteed to be worse than the incumbent.

;;; Fix:
;;; o Cost is not correct, per JDAM paper.

;;; o Generalize to conditional CSPs:
;;;   - Take into consideration variable activation.
;;;     The assignments appearing in the suboptimality conflict must all be 
;;;     active.
;;;     - Sort variables in topological order of variable activation
;;;       using a depth-first numbering.
;;;     - To bias towards plcing nodes in decreasing order of cost,
;;;       select among the children the child with the greatest cost.
;;;     - Reverse the order of the list of assignments, and then generate
;;;       the conflicts as described below.

;;; To Do:

;;; o Does this function belong here, or in a separate decision-space file?
;;; o Whenever the incumbent-cost is updated, the suboptimality conflicts might
;;;   be improved by deleting assignments from best to worst value.
;;;   - Consider updating the conflict when 
;;;     determining whether a suboptimality conflict can be resolved.
;;;   - To update a suboptimality conflict efficiently, 
;;;     - maintain with the conflict:
;;;        a) its cost and 
;;;        b) its assignments, in order of cost from best to worst
;;;           (already done?).
;;;     - when updating, 
;;;        - greedily delete best assignments in the conflict and
;;;        - and update cost.
;;;        - stop just before cost becomes better than the incumbent.
;;;        - you'll need to rebuild the constituent kernels at this point.
(defmethod suboptimality-conflict-from-assignments
	((decision-space decision-space-w-conflicts) assignments incumbent-cost)
  
  "Returns a sub-optimality conflict for ASSIGNMENTS of DECISION-SPACE
   with respect to INCUMBENT-COST.  This is a minimal subset of ASSIGNMENTS, 
   whose cost is no better than the incumbent."

  ;; Generate and return a suboptimality conflict.
  (with-slots ((order-fn cost-order-function)
			   binary-objective-function
			   total-cost
			   cost-offset)
	  decision-space
			  
	(let ((ordered-assignments
		   ;; Sort node's assignments from worst to best.
		   (nreverse (sort assignments order-fn
						   :key #'decision-assignment-cost))))

	  ;; Make conflict by greedily accumulating the worst assignments,
	  ;; until their accumulated cost is no better than the incumbent.

	  ;; The accumulated cost is the combination of the cost of the 
	  ;; assigments, the cost-offsets, AND the best cost of all 
	  ;; unassigned variables.
			  
	  (loop 
		  with selected-assignment-cost = cost-offset
		  with conflict-cost
					   
		  for assigns on ordered-assignments
		  for assignment = (first assigns)
		  for assignment-cost = (decision-assignment-cost assignment)

		  for remaining = (rest assigns)
						  ;; compute the best cost of all variables left out
						  ;; thus far.

		  for remaining-cost = (assignments-best-cost remaining 
													  binary-objective-function)

		  do
			;; Accumulate asssigment costs from worst to best.
			;; = cost-offset 
			;;   + objective-function(assignment-cost1,...assignmentn-cost)
			(setq selected-assignment-cost 
			  (funcall binary-objective-function
					   assignment-cost selected-assignment-cost))
					
			;; Add in the best cost of the unassigned variables.
			(setq conflict-cost (+ selected-assignment-cost
								   remaining-cost))
				
			;; Simultaneously accumulate assignments ..
		  collecting assignment
						 
					 ;; Stop as soon as conflict cost is no better
					 ;; than the incumbent and return the assignments
					 ;; accumulated to date.
		  until (funcall order-fn incumbent-cost conflict-cost)))))

;;; To Do:
;;; o Put all code supporting the cost of assignments in one file.

;;; o For conditional CSPs,
;;;   the best cost depends on which variables are active.
;;;   Need to determine how to compute this.

(defun assignments-best-cost (assignments binary-objective-function)

  "Returns the cost of the best cost ASSIGNMENTS to
   the variables of ASSIGNMENTS.  Assignment costs are
   combined pairwise using BINARY-OBJECTIVE-FUNCTION."
  
  (loop with accumulated-cost = 0
	  for assignment in assignments
	  for variable = (state-assignment-variable assignment)
	  for best-assignment = (decision-variable-best-assignment variable)
	  for best-cost = (decision-assignment-cost best-assignment)
	  do (setq accumulated-cost
		   (funcall binary-objective-function 
					accumulated-cost
					best-cost))
	  finally (return accumulated-cost)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conflict decay
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod bump-conflict! 
	((conflict conflict) (state-space state-space-w-conflicts))
  
  "Bump the activity of the conflict because it was used. Rescale as necessary."

  (with-slots (activity) conflict
	(setf activity (+ activity (state-space-conflict-bump-inc state-space)))

	(when (<= 1d27 activity)
	  (rescale-conflict-activities! state-space))))


(defmethod rescale-conflict-activities! ((state-space state-space-w-conflicts))

  "Rescale all the conflict activities."

  (loop for one-conflict in (state-space-conflicts state-space)
	  do (setf (conflict-activity one-conflict)
		   (/ (conflict-activity one-conflict) 1d27)))
  
  (setf (state-space-conflict-bump-inc state-space) 
	(/ (state-space-conflict-bump-inc state-space) 1d27)))

(defmethod decay-conflict-activity! ((state-space state-space-w-conflicts))

  "Decay the activity of all conflicts by scaling the bump increase."

  (setf (state-space-conflict-bump-inc state-space)
	(* (state-space-conflict-decay state-space)
	   (state-space-conflict-bump-inc state-space))))

(defmethod manage-conflict-list! ((state-space state-space-w-conflicts))

  (with-slots (conflicts conflict-max-number conflict-learned-growth
			   conflict-bump-inc) state-space
	
	(let ((num-conflicts (state-space-conflict-num state-space)))
	  
	  ;; When there are too many conflicts,
	  (when (>= num-conflicts
				conflict-max-number)

		;; remove conflicts you don't need.
		(let ((activity-lim (/ conflict-bump-inc num-conflicts))
			  (activity-sorted-conflicts 
			   (delete nil (sort (copy-list conflicts)
								 #'< :key #'conflict-activity)
					   :test #'(lambda (x y) (declare (ignore x))
									   (conflict-decayed? y)))))
		  
		  ;;	 	  (break)
		  
		  ;; Remove at least half the conflicts.
		  (loop for i from 1 to (floor (/ num-conflicts 2))

								;; Remove this conflict from the sorted list and 
								;; the problem by popping and removing.
			  do (remove-conflict! (pop activity-sorted-conflicts) state-space))

		  ;; Also remove any that don't meet the threshold.
		  ;; 		  (break)
		  (loop 
			  until (null activity-sorted-conflicts) 
					;; Can't happen, but just to be safe.
					
			  until (> (conflict-activity (first activity-sorted-conflicts))
					   activity-lim)
					
					;; Remove this conflict from the sorted list and the problem 
					;; by popping and removing.
;;;			  do (format t "popping ~a ~a ~%"
;;;						 (conflict-activity (first activity-sorted-conflicts))
;;;						 activity-lim)

			  do (remove-conflict! (pop activity-sorted-conflicts) state-space))

		  ;;and finally, increase the number of conflicts to keep in the future.
		  (setf conflict-max-number
			(* conflict-learned-growth conflict-max-number))
;;; 		  (format t "Now using conflicts: ~a ~a ~a ~a ~a~%"
;;;				  (state-space-conflict-num state-space)
;;;				  (length (state-space-conflicts state-space))
;;;  				  conflict-max-number activity-lim conflict-bump-inc)
		  )
		))))

(defmethod record-split-on-conflict ((conflict conflict) search-node search-problem)
  (declare (ignore search-node search-problem))
  T)

(defmethod drop-node-due-to-conflicts! (current-node search-problem)
  (declare (ignore search-problem))
  (debug-print-for :bfig t "    Discarding entire node ~a" current-node))

(defmethod record-node-inconsistent-with-conflict! (search-node 
												   (new-conflict conflict) 
												   search-problem)
  "Default definition of method to record node inconsistencies with
   conflicts. Does nothing."
  (declare (ignore search-problem search-node))
  T)
