;;; -*- Mode:Common-Lisp; Package:BFIG; Base:10 -*-

;;; Copyright (c) 1999 - 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :BFIG)

;;; FIX:
;;; o mtu::declare-debug-print-system!
;;;   has been exported from MTU in its package definition,
;;;   and included in BFIG.  Why isn't the function external (: vs ::)
;;;   and why isn't it part of BFIG?

;;; To Do:
;;; o Test Branch-and-Bound
;;; o Test dfs and bfs.

(mtu::declare-debug-print-system! :BFIG)

;;;(eval-when (compile load eval)

;;;  (export '(2
;;;            *default-search-problem*
;;;			;;            *bfs-debug*
;;;			state-space-source
;;;			state-space-variables

;;;			decision-space-binary-objective-function

;;;			search-problem
;;;			make-search-problem
;;;			search-problem-solutions
;;;			search-problem-best-solution-cost
;;;			search-problem-solution-count
;;;			search-problem-candidate-count
;;;
;;;			first-solution-found?
;;;
;;;			search-node-best-cost
;;;			search-node-assignments
;;;			search-node-ancestor-assignments
;;;			search-node-preferred-child-assignment
;;;			search-node-preferred-expansion
;;;	    
;;;			state-variable-domain
;;;			state-variable-previous-assignment
;;;			state-variable-current-assignment
;;;			state-variable-best-assignment
;;;	    
;;;			state-assignment-variable
;;;			state-assignment-value
;;;			state-assignment-to-constraint-assignment
;;;
;;;			get-assignment-from-hash
;;;			)
;;;          :BFIG)
;;;
;;;  )

;;; This file defines objects used to generate the kernels (implicants)
;;; of a set of conflict.  The objects defined in this file correspond
;;; to a:

;;;    Search Problem,               (value variables plus conflicts)
;;;    Decision Variable,            (finite domain variable w utility)
;;;    Variable Assignment,          (valid assignment to a decision variable)
;;;    Conflict and                  (inconsistent partial assignment to vars)
;;;    Search Node                   (node in a search tree, used to
;;;                                    generate a kernel (prime implicant
;;;                                    of the conflicts).

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Fix:
;;; o We are not computing the cost of the constituent
;;;   kernels correctly in best-expansion.  See the JDAM paper.

;;; To Do:
;;; o Look for opportunities for sharing between this system
;;;   and the constraint systems, for variable, 
;;;   assignment, search node and other structures.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Space
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Tighten types of methods that operate on
;;;   decision-space.
(defclass state-space ()
  
  ((name
	:initarg :name
	:initform (gentemp "STATE-SPACE-")
	:accessor state-space-name		
	:documentation "The name of a state space.")
   
   (source
	:initarg :source
	:initform nil
	:accessor state-space-source
	:documentation "Object that created state space.")

   (variable-type
	:allocation :class
	:reader state-space-variable-type
	:initform 'state-variable
	:documentation "Class type of variables of state space.")

   (variables
	:initform nil
	:accessor state-space-variables
	:documentation "The variables of the state space.")
										
   (assignment-ht
	:initform (make-hash-table :test #'equal)
	:accessor state-space-assignment-ht
	:documentation "A hash table mapping assignment ~
                    expressions to assignment objects.")
   
   (assignment-bit-ht
	:initform (make-hash-table :test #'equal)
	:accessor state-space-assignment-bit-ht
	:documentation "Holds the lookup from each assignment to the bit vector
                    representation.")

   (zero-bit-vector
	:accessor state-space-zero-bit-vector
	:documentation "Holds a 0 bit vector of the right length for equal checking.")
   )
  
  (:documentation   "A state space, defined as a set of variables,
                     their corresponding domains."))

(defmethod print-object 

	((state-space state-space) stream)

  (with-slots (name) state-space
	(format stream "<SS: ~a>" name)))

(defmacro get-assignment-from-hash 

	(assignment assignment-ht)

  "Returns the bsf-assignment object of ASSIGNMENT-HT 
   corresponding to ASSIGNMENT expression.  
   ASSIGNMENT-HT is an assignment hash table.
   ASSIGNMENT is an expression (= variable value)."
  
  `(gethash ,assignment ,assignment-ht))

(defmacro get-state-assignment 

	(state-assignment state-space)

  "Returns the bsf-assignment object of STATE-SPACE
   corresponding to ASSIGNMENT.  
   ASSIGNMENT is an expression (= variable-name value)."
  
  `(get-assignment-from-hash
    ,state-assignment (state-space-assignment-ht
					   ,state-space)))

;;; Changed
;;; - Added macro to compare constituent kernels from different variables.
;;;   -- gustavop

(defmacro better-kernel? 
    (ordering-function binary-function)

  "True if assignment-1 is a better kernel than assignment-2."
    
  `(lambda (assign-1 assign-2) 
	 (let ((cost1 (decision-assignment-cost assign-1))
		   (cost2 (decision-assignment-cost assign-2)))
	   
	   (if (equal (state-assignment-variable assign-1)
				  (state-assignment-variable assign-2))
		   
		   (funcall ,ordering-function cost1 cost2)
		 
		 (funcall ,ordering-function
				  
				  (funcall ,binary-function cost1
						   (decision-assignment-cost
							(decision-variable-best-assignment 
							 (state-assignment-variable assign-2))))
				  
				  (funcall ,binary-function
						   (decision-assignment-cost
							(decision-variable-best-assignment 
							 (state-assignment-variable assign-1)))
						   cost2)))))
  )


;;; Changed:
;;; - Added method.
(defmethod initialize-instance :after 

		   ((state-space state-space) &key)

  "Create a new state-space named NAME, created by SOURCE.
   SOURCE is the object that created search problem.
   NAME is a string."

  (initialize-state-space! state-space)
  )

;;; Changed:
;;; - Added method.
(defmethod initialize-state-space! ((state-space state-space))
  
  "Initializes STATE-SPACE."

  (clear-state-variable-marks! state-space)
  )
  
;;; Changed:
;;; - Added method.
(defmethod clear-state-variable-marks! ((state-space state-space))

  "Unmarks all variables of STATE-SPACE."
  
  (with-slots (variables) state-space
	(loop for v in variables
		do (unmark-state-variable! v))
	T))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Space with Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass state-space-w-conflicts (state-space)
  
  ((variable-type
	:initform 'state-variable-w-conflicts)
   
   (conflict-type
	:initform 'conflict)
   
   (conflicts 
	:initform nil
	:accessor state-space-conflicts
	:documentation "The current set of conflicts,  maintained in ~
                    increasing order of number of assignments per conflict.")

   (unresolved-conflict-count
	:initform 0
	:accessor state-space-unresolved-conflict-count
	:documentation "The number of conflicts not resolved by the current ~
                    search state (partial kernel).")

   (conflict-decay
	:initform (/ 1.0 0.98)
	:accessor state-space-conflict-decay
	:documentation "How fast to decay the activity of a conflict.")
  
   (conflict-bump-inc
	:initform 1
	:accessor state-space-conflict-bump-inc
	:documentation "The current amount to increase a conflict's activity by.")
   
   (conflict-max-number
	:initform 100
	:accessor state-space-conflict-max-number
	:documentation "The current maximum number of conflicts to maintain."
	)

   (conflict-learned-growth
	:initform 1.5
	:accessor state-space-conflict-learned-growth
	:documentation "How fast to grow the number of conflicts.")

   (conflict-num
	:initform 0
	:accessor state-space-conflict-num
	:documentation "The number of current conflicts.")
   )
  
  (:documentation   "A Decision Space whose feasibility is constrained 
                     by conflicts."))

;;;(defmethod print-object 
;;;	((sswc state-space-w-conflicts) stream)
;;;  (with-slots (name) sswc
;;;	(format stream "<SSwC: ~a>" name)))

(defmethod order-constituent-kernels
	
	((space state-space-w-conflicts)
	 constituent-kernels)

  "Establishes an order for CONSTITUENT-KERNELS of a conflict of STATE-SPACE.
   This is the order in which kernels are split on during search.
   Returns ordered CONSTITUENT-KERNELS.
   Note: Currently the order of constitutent-kernels isn't modified."

  constituent-kernels)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Decision Space
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Consider folding these three slots instead into an 
;;;   objective that can be called to provide the different
;;;   basic operations for evaluating cost.

;;; Fixed.
;;; o  missing :initargs were causing troubles when setting up
;;;    minimization problem.
;;; -- gustavop
(defclass decision-space (state-space)
  
  ((name
	:initform (gentemp "decision-space-")
	:accessor decision-space-name
	:documentation "Name of a Decision Space.")
   
   (variable-type
	:initform 'decision-variable)
   
   (cost-offset 
	:initform 0
	:initarg :cost-offset
	:accessor decision-space-cost-offset
	:documentation "A constant added to all assignment costs.")

   (binary-objective-function 
	:initform #'+
	:initarg :binary-objective-function
	:accessor decision-space-binary-objective-function
	:documentation "Function for computing global cost, ~
                    composing attribute costs.")

   (cost-order-function 
	:initform #'>
	:initarg :cost-order-function
	:accessor decision-space-cost-order-function
	:documentation "Predicate for determining the order of attribute ~
                    (assignment) costs , >= for decreasing and <= ~
                    for increasing.")
   )
  
  (:documentation   "A valued state space, defined as a set of variables,
                     their corresponding domains and costs for each
                     domain element."))

(defmethod print-object 

	((decision-space decision-space) stream)

  (with-slots (name) decision-space
	(format stream "<DS: ~a>" name)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Decision Space with Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; To Do:
;;; o Should conflicts be provided as a mix in?
(defclass decision-space-w-conflicts 
	
	(decision-space state-space-w-conflicts)
  
  ((variable-type
	:initform 'decision-variable-w-conflicts))

  (:documentation   "A Decision Space whose feasibility is constrained 
                     by conflicts."))

;;;(defmethod print-object 
;;;	
;;;	((dswc decision-space-w-conflicts) stream)
;;;  
;;;  (with-slots (name) dswc
;;;	(format stream "<DSwC: ~a>" name)))

;;; Fixed.
;;; o order-constituent-kernels now correctly compares
;;;   kernels from different variables.
;;; -- gustavop
(defmethod order-constituent-kernels 
	
	((space decision-space-w-conflicts)
	 constituent-kernels)

  "Orders CONSTITUENT-KERNELS of a conflict of DECISION-SPACE from best
   to worst cost.  This is the order in which kernels are split on, and
   is essential so that the generator searches the kernels in order of 
   best cost.
   Returns ordered CONSTITUENT-KERNELS."

  ;; Note: Destructively modifies the list of constitutent-kernels.
  ;;    This is okay because kernels-of-conflict made sure that
  ;;    constituent kernels is a new list.
  (sort constituent-kernels
		(better-kernel? (decision-space-cost-order-function space)
						(decision-space-binary-objective-function space))))
		;(decision-space-cost-order-function space)
		;:key #'decision-assignment-cost))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Variable Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:

;;; o The conflict database and Opsat each have a current-assignment.
;;;   Can they be shared, or do they need to be distinct?

;;; o Current-assignment and previous-assignment are only
;;;   used by opsat, not bfig.  
;;;    - the name should be qualified to indicate that
;;;      the current-assignment is wrt opsat 
;;;      (e.g., opsat-current-assignment), since
;;;      there is the notion of more than one current assignment.
;;;    - They should be moved to a separate class or mixin.
;;;    - This class  should be moved to the opsat package, 
;;;      until more generally employed.
(defclass state-variable ()
  
  ((name 
	:initarg :name
	:accessor state-variable-name
	:documentation "Name of the decision variable.")

   (domain
	:initform nil
	:accessor state-variable-domain
	:documentation "Possible assignments to variable.")
   
   (assignment-type
	:allocation :class
	:reader state-variable-assignment-type
	:initform 'state-assignment
	:documentation "Class type of assignments to variable.")
   
   ;;; Slots related to candidate generation.
   
   ;; For identification of active variables and conflicts
   ;; and supported conflicts.
   (current-cdb-assignment
	;; Records the partial assignments specified by
	;; the current search node.  This is used by the 
	;; candidate generator to identify unresolved conflicts.
	:initform nil
	:accessor state-variable-current-cdb-assignment
	:documentation "Assignment to variable that is currently ~
                    inserted within the conflict data base.")
      
   ;; For OPSAT testing of candidate consistency, 
   ;; maintains the current and previous candidate state.
   (current-assignment
	:initform nil
	:accessor state-variable-current-assignment
	:documentation "Assignment to variable in the current search state.")

   (previous-assignment
	:initform nil
	:accessor state-variable-previous-assignment
	:documentation "Assignment to variable in the ~
                    previous search state.")
   
   ;; General Maintenance Slots
   (mark
	:initform nil
	:reader state-variable-marked?
	:documentation "Maintenance flag, can be True or nil.")
   )

  (:documentation "A variable of a state space with ~
                   associated domain."))

(defmethod print-object 

	((variable state-variable) stream)
  
  "Prints short descriptor for a variable."
  
  (with-slots (name) variable
	(format stream "~a" name)))

;;; Changed:
;;; o value-descriptions -> domain
(defmethod initialize-instance :after 
		   ((variable state-variable) 
			&key state-space domain)
  
  "Creates a new variable for STATE-SPACE with the 
   specified name.  DOMAIN is a list of domain 
   element descriptions.
   Each DOMAIN element description is either an atom, 
   denoting an element, or list of the form:

   (ELEMENT . <keyword-value-pair>*)
  
   where ELEMENT is the value of an assignment, and the 
   keyword-value-pairs correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of VARIABLE."

  (let ((intern-domain (make-variable-domain
						state-space variable
						domain)))
    
	;; Setup the assignments in variable's domain.
    (state-variable-install-domain! variable 
									intern-domain)

	;; Add VARIABLE to the variables of the search problem
    (push variable
		  (state-space-variables state-space))

	T))

;;; Changed:
;;; o Added method.
(defmethod mark-state-variable! ((v state-variable))

  "Marks state variable V."
  
  (with-slots (mark) v
	(setq mark T)))

;;; Changed:
;;; o Added method.
(defmethod unmark-state-variable! ((v state-variable))

  "Unmarks state variable V."

  (with-slots (mark) v
	(setq mark nil)))

;;; Changed:
;;; o Added method.
(defmethod state-variable-active? ((state-variable state-variable))

  "Returns true iff STATE-VARIABLE is active.
   A variable is active iff its guard condition is
   satisfied by the current assignment.

   Unconditional state variables, that is, those without guard
   conditions are always active."

  T)

(defmethod state-variable-domain-size ((v state-variable))

  "Returns the number of elements in state variable V's domain."

  (with-slots (domain) v
	(length domain)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Variable with Conflicts Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Some of the methods that are currently defined over
;;;   decision-variable-w-conflicts need to be split, 
;;;   or tailored to work on state-variable-w-conflicts.
(defclass state-variable-w-conflicts (state-variable)
  
  ((assignment-type
	:initform 'state-assignment-w-conflicts)

   (current-cdb-disallowed-assignment
	;; Records the disallowed assignments specified by
	;; the current search node.  This is used by the 
	;; candidate generator to identify unresolved conflicts.
	:initform nil
	:accessor state-variable-current-cdb-disallowed-assignment
	:documentation "Assignment to variable that is currently
                    disallowed within the conflict data base."))

  (:documentation "A state variable whose assignment is used to ~
                   determine (un)resolved conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Decision Variable
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Check:  
;;; o Is there any presumption that the domain is in order 
;;;   of best to worst cost?

;;; To Do:
;;; o Separate slots are maintained for candidate-generator's
;;;   best-extension and conflict database's current assignment.
;;;   Can these be collapsed, or do they need to be maintained
;;;   separately?
(defclass decision-variable (state-variable)
  
  ((assignment-type
	:initform 'decision-assignment)
   
   (best-assignment
	:initform nil
	:accessor decision-variable-best-assignment
	:documentation "Domain assignment of variable with the best cost.")
  
   (cg-best-extension
	:initform nil
	:accessor decision-variable-cg-best-extension
	:documentation "Assignment to variable that is part of the best ~
                    extension to candidate generator's current ~
                    partial assignment."))

  (:documentation "A decision variable with domain and associated ~
                   cost for each domain element."))

;;; Changed:
;;; o Updated documentation regarding domain.
(defmethod initialize-instance :after 

		   ((variable decision-variable) 
			&key state-space)
  
  "Creates a new variable for STATE-SPACE with the 
   specified name.  DOMAIN is a list of domain 
   element descriptions.
   Each DOMAIN element description is either an atom, 
   denoting an element, or list of the form:

   (ELEMENT . <keyword-value-pair>*)
  
   where ELEMENT is the value of an assignment, and the 
   keyword-value-pairs correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of VARIABLE.
  
   See method MAKE-VARIABLE-ASSIGNMENT for further documentation. 

   Ensures that variable's domain is placed in best first order."

  ;; Place the domain in best first order.  Also set the best and
  ;; current assignments to the best element of this domain.
  (variable-order-domain! state-space 
						  variable))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Decision Variable with Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass decision-variable-w-conflicts 
	(decision-variable state-variable-w-conflicts)
  
  ((assignment-type
	:initform 'decision-assignment-w-conflicts))

  (:documentation "A decision variable whose assignment is used to ~
                   determine resolved conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Assignment Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o to-constraint-assignment should not be part of basic bfig, and
;;;   instead should be a class extension to assignments in OPSAT.
(defclass state-assignment ()

  ((variable
	:initarg :variable
	:accessor state-assignment-variable
	:documentation "The variable of this assignment.")

   (value
	:initarg :value
	:accessor state-assignment-value
	:documentation "The value of this assignment.")

   (alternate-assignments
	:accessor state-assignment-alternate-assignments
	:initform '()
	:documentation "All other assignments in variable's domain, ~
                    besides this assignment.")

   ;; Link to a constraint system, for generate and test.

   (to-constraint-assignment
	:accessor state-assignment-to-constraint-assignment
	:initform nil
	:documentation "The object in constraint-system that denotes assignment.  
                    This is cached to speed up insertion/removal."))
  
  (:documentation "A variable/value assignment, with associated conflicts."))

(defmethod print-object 

	((state-assignment state-assignment) stream)
  
  "Prints short descriptor for a state variable assignment."
  
  (with-slots (variable value) state-assignment
	(format stream "~a=~a" variable value)))

(defmacro state-assignment-description
	
	(state-assignment)
  
  "Given a STATE-ASSIGNMENT object,
   Returns an external description of assignment, in the form:
     (= variable-name value)."
  
  `(make-assignment
	:variable-name (state-variable-name
					(state-assignment-variable ,state-assignment))
	:value (state-assignment-value ,state-assignment)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Assignment with Conflict Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Check:
;;; - Can this and similar functions be made a mixin of assignment rather 
;;;   than inherit from state-assignment?
(defclass state-assignment-w-conflicts (state-assignment)

  ((resolved-conflicts
	:initform '()
	:accessor state-assignment-resolved-conflicts
	:documentation "All conflicts that are resolved by this assignment.
                    A resolved conflict contains an assignment that is
                    inconsistent with this assignment).  This assignment
                    is a constitutent kernel of each resolved-conflict.")

   (conflicts
	:initform '()
	:accessor state-assignment-conflicts
	:documentation "Conflicts that contain assignment."))
  
  (:documentation "A variable/value assignment, with associated conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Decision Assignment Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass decision-assignment (state-assignment)

  ((cost 
	:initarg :cost
	:initform 0.0
	:accessor decision-assignment-cost
	:documentation "Attribute cost of assigning value to variable."))
  
  (:documentation "A decision variable assignment that can be assigned
                   during candidate generation."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Decision Assignment with Conflicts Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass decision-assignment-w-conflicts 

	(decision-assignment 
	 state-assignment-w-conflicts)

  ()
  
  (:documentation "A decision variable assignment that can be assigned
                   during conflict-directed candidate generation."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conflict Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conflict ()

  ((inconsistent-assignments
	:initarg :inconsistent-assignments
	:accessor conflict-inconsistent-assignments
	:documentation "A list of assignments that are collectively inconsistent.
                    Represents an infeasible subset of the state space,
                    that is, any candidate containing all these assignments 
                    is infeasible.")

   (constituent-kernels
	:initform nil
	:accessor conflict-constituent-kernels
	:documentation "A list of all assignments that resolve this conflict; that 
                    is, any candidate containing one of these assignments,
                    is guaranteed to lie outside the infeasible subspace 
                    represented by the conflict.  Assignments are sorted
                    by cost, from best to worst.")

   (support 
	:initform nil
	:accessor conflict-support 
	:documentation "The assignment in the current search state that resolves 
                    the conflict. When non-NIL, points to the best cost 
                    assignment that resolves conflict")

   (activity
	:initform 0d0
	:initarg :activity
	:accessor conflict-activity
	:documentation "Stores the activity of this conflict - specified as 
                    a double.")

   (inconsistent-assignments-bit
	:accessor conflict-inconsistent-assignments-bit
	:documentation
	"The bit vector representation of the inconsistent assignments 
     in the conflict.")

   (decayed?
	:initform nil
	:accessor conflict-decayed?
	:documentation "Whether the conflict has been removed by decaying. Used because
                    conflicts are tracked in the search nodes. Transitioned from 
                    nil -> t exactly once.")

   (number
	:initform 0
	:accessor conflict-number
	:documentation "The number of the conflict. Used to avoid reprocessing.")
   
   )

  (:documentation 
   "A partial assignment that is inconsistent with problem's constraints."))

(defmethod print-object 
	((conflict conflict) stream)
  
  "Prints short descriptor for a conflict of a candidate generation problem."
  
  (with-slots (inconsistent-assignments) conflict
	(format stream 
;;;			"Cflct:~a>" 
			" ~~~a" 
			inconsistent-assignments)))

;;; FIX 
;;; o The cost used for a constitutent kernel in the call to sort below
;;;   is not correct.  Redefine according to the JDAM paper.

;;; Changed:
;;; o Modified comments.


(defmethod initialize-instance :after 
		   ((conflict conflict)
			&key
			state-space
			inconsistent-assignments
			(inconsistent-constraints nil)
			)
  
  "Asserts a new conflict for STATE-SPACE.  A conflict is a partial
   set of variable assignments, INCONSISTENT-ASSIGNMENTS, that are 
   inconsistent.  A conflict contains at most one assignment 
   for each variable in STATE-SPACE.  INCONSISTENT-ASSIGNMENTS 
   can be a list of either assignment objects or 
   assignment descriptions, of the form (= variable-name value).

   Side Effect: The list of constituent kernels (prime implicants) of
   the conflict is destructively placed in the order in which children
   are visited when splitting on a resolved conflict.  

   For a decision-space kernels are placed in best-first order based 
   on cost.  This order is essential to the search algorithm.  It 
   ensures that search tree nodes are expanded in best first order."

  ;; Creates a list of assignment objects, that make explicit all
  ;; inconsistent assignments described in INCONSISTENT-ASSIGNMENTS.
  (let ((intern-inconsistent-assignments 
		 (internalize-state-assignments state-space
										inconsistent-assignments)))

	;; Record the assignments of conflict.
	(setf (conflict-inconsistent-assignments conflict)
	  intern-inconsistent-assignments)

	;;Record the bit vector representation of this conflict. 
	(setf (conflict-inconsistent-assignments-bit conflict)
	  (clause-to-bit-clause (conflict-inconsistent-assignments conflict)
							state-space))

	(let ((constituent-kernels			; Constituent-Kernels are unordered 
		   (kernels-of-conflict conflict))) ; at this point.

	  ;; Place constituent-kernels in the order that children are visited
	  ;; when splitting on a conflict.
	  (setq constituent-kernels
		(order-constituent-kernels state-space constituent-kernels)

;;;		;; Note: Destructively modifies the list of constitutent-kernels,
;;;		;;    This is okay because kernels-of-conflict makes sure that
;;;		;;    constituent kernels is a new list.
;;;		(sort constituent-kernels 
;;;			  (state-space-cost-order-function decision-space)
;;;			  :key #'decision-assignment-cost)
		)

	  (setf (conflict-constituent-kernels conflict) constituent-kernels)
	  T)))

(defmethod conflict-supported? 

	((state-space state-space-w-conflicts) (conflict conflict))

  "Returns true if an assignment in the current search 
   state has been recorded that resolves CONFLICT."

  (with-slots (current-search-node) state-space
	(if (null current-search-node)		

		;; If there is no current search node, none are supported.
		nil
	  
	  ;;Try the current assignments.
	  (let ((current-assignments 
			 (remove nil
					 (append (search-node-ancestor-assignments current-search-node) 
							 (list (search-node-preferred-child-assignment
									current-search-node))))))
		
		;; 	(format t "~%current node: ~a ~a~%" current-search-node 
		;;     current-assignments)
		
		;; If any of the inconsistent assignments have other assignments to the 
		;; same var in the current assignments, true.
		(not (loop for inconsistent-assignment 
				 in (remove nil (conflict-inconsistent-assignments conflict))
					
				 always (member inconsistent-assignment 
								current-assignments
;;;								:test #'(lambda (x y) 
;;;										  (and (not (eql x y))
;;;								 			   (eql (state-assignment-variable x)
;;;													(state-assignment-variable y)))
;;;										  )
								))))) ;;WORKING
	)
  
  ;; `(not (null (conflict-support ,conflict)))
  )

;;; To Do:
;;; o Delete when the code works.
;;;(defmacro conflict-resolved? 
;;;
;;;	(conflict)
;;;  
;;;  "Returns true if conflict is resolved by the current search state.  
;;;   A conflict is resolved if the assignments extending conflict and
;;;   current search state DO NOT intersect, or equivalently,
;;;   if search state entails the negation of conflict."
;;;  
;;;  `(not (null (conflict-support ,conflict))))

;;; Changed:
;;; o Added.
(defmethod conflict-active?
	
	((conflict conflict))

  "Returns true iff CONFLICT is active.
   A conflict is active iff all of its variables are active.
   Note: unconditional variables and conflicts on unconditional variables
         are ALWAYS active."
  
  t)

;;; Changed:
;;; o Added.
(defmethod conflict-resolved? 

	((state-space state-space-w-conflicts) (conflict conflict))
  
  "Returns true if conflict is resolved by the current search state.
   Given a current state, a conflict is resolved if conflict is
   INACTIVE in the current state or if one of its inconsistent 
   assignments CAN NOT hold."

  (or (not (conflict-active? conflict))
	  (conflict-supported? state-space conflict)
	  ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Search Order
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Delete comments when all search orders are demonstrated as working.
;;;(defvar *allowed-search-order-types*
;;;	'(
;;;	  :depth-first :breadth-first		; Uninformed Search
;;;										; Informed Search
;;;	  :best-first						;   A-star or heuristic.
;;;;;;	  :depth-first-BB :breadth-first-BB ;   Branch and Bound
;;;;;;	  :best-first-BB
;;;	  )
;;;
;;;  "Allowed search orderings for search-tree expansion")

;;;(defmacro allowed-search-order? 
;;;	
;;;	(search-order-type)
;;;  
;;;  "Returns true iff SEARCH-ORDER-TYPE is an
;;;   order of tree expansion that the candidate generator
;;;   supports."
;;;  
;;;  `(member ,search-order-type
;;;		   *allowed-search-order-types*))

;;; To Do:
;;; o Delete when code works.
;;; (defvar *branch-and-bound-search-orders*
;;;	'(:depth-first-BB :breadth-first-BB :best-first-BB)
;;;  "Search orders that involve branch and bound search.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Search Problem Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *default-search-problem* nil
  "The current search problem")

;;;(eval-when (:compile-toplevel :load-toplevel :execute)

(defclass search-problem 
	(state-space)
  
  ((name
	:initform (gentemp "search-problem-")
	:accessor search-problem-name
	:documentation "Name of Search Problem.")
   
   ;; Search Parameters
   
   (allowed-search-orders
	:initform '(:depth-first :breadth-first)
	:allocation :class
	:documentation  "Allowed search orderings for search-tree expansion")
   
   (search-order
	:initarg :search-order
	:initform :DEFAULT
	:accessor search-problem-search-order
	:documentation "A keyword, indicating the order in which search ~
                    nodes are expanded; one of :depth-first, 
                    :breadth-first or :best-first.")
   
   ;;; Search State
   
   (search-queue
	:initform :NONE
	:accessor search-problem-search-queue
	:documentation "Queue of unexpanded search tree nodes. ~
                    For best first search this is sorted in ~
                    best 1st order. The queue order is based on the ~
                    value of the binary-objective-function.")
										  
   (node-type
	:allocation :class
	:reader search-problem-node-type
	:initform 'search-node
	:documentation "Class type of search-tree-node of search-problem.")
   
   (current-search-node 
	:initform nil
	:accessor search-problem-current-search-node
	:documentation "The most recently expanded search-node.")

   (current-search-node-feasible? 
	:initform nil
	:accessor search-problem-current-search-node-feasible?
	:documentation "The most recently expanded search-node is feasible or not.")

   (feasible-nodes
	:initform nil
	:accessor search-problem-feasible-nodes
	:documentation "Feasible nodes recorded for the problem.")

   (feasible-minimal-relaxations
	:initform nil
	:accessor search-problem-feasible-minimal-relaxations
	:documentation "Feasible minimal relaxations recorded for the problem.")

   (old-search-node
	:initform nil
	:accessor search-problem-old-search-node
	:documentation "The last search node.")

   (old-assignments
	:initform nil
	:accessor search-problem-old-assignments
	:documentation "The last set of assignments.")

   (current-minimal-assignments-against-conflicts
	:initform nil
	:accessor search-problem-current-minimal-candidates
	:documentation "The minimal candidates found that resolve all known conflicts")

   ;;; Solutions
   
   (solutions 
	:initform nil 
	:accessor search-problem-solutions
	:documentation "A list of consistent states found thus far.")
   
   ;; Statistics maintained.

   (count-of-search-nodes-created
	:initform 0
	:accessor search-problem-count-of-search-nodes-created
	:documentation "The number of search nodes expanded thus ~
                    far in the search.")
	  
   (count-of-search-nodes-considered 
	:initform 0
	:accessor search-problem-count-of-search-nodes-considered
	:documentation "The number of search nodes expanded thus ~
                    far in the search.")    

   (solution-count 
	:initform 0
	:accessor search-problem-solution-count
	:documentation "Count of solutions found.")
										
   (candidate-count 
	:initform 0 
	:accessor search-problem-candidate-count
	:documentation "Count of candidates tested for consistency.")
   )
  
  (:documentation   "State space search problem and systematic search method."))

(defmethod initialize-instance :after 

		   ((search-problem search-problem) &key)

  "Create a new search-problem named NAME, created by SOURCE.
   SOURCE is the object that created search problem.
   NAME is a string.
   Key SEARCH-ORDER is a keyword, specifying the order in 
     which the candidate is explored.  SEARCH-ORDER is one of 
     *ALLOWED-SEARCH-ORDER-TYPES*, or :DEFAULT.  :DEFAULT is
     equivalent to :depth-first if uninformed search or branch and bound 
     are specified, and :best-first, otherwise.  
     SEARCH-ORDER defaults to :DEFAULT."

  (setq *default-search-problem* search-problem)

  (initialize-search-order! search-problem)

;;; Note: The candidate generator needs to be initialized
;;;       AFTER the variables are recorded,  in order to determine
;;;       the order in which variables are split.  Pull initialize
;;;       into this function only when add-variables is also
;;;       incorporated.
;;;  (initialize-candidate-generator! search-problem 
;;;								   :search-order :DEFAULT)
  )

(defmethod print-object 

	((search-problem search-problem) stream)

  (with-slots (name) search-problem
	(format stream "<SP: ~a>" name)))

(defmethod first-solution-found? 

	((search-problem search-problem))

  "Returns True if at least one solution to SEARCH-PROBLEM
   has been found."

  (with-slots (best-solution-cost) search-problem
	(not (null best-solution-cost))))

(defmethod allowed-search-order? 
	
	(search-order (search-problem search-problem))
  
  "Returns true iff SEARCH-ORDER is an order of tree expansion 
   that the candidate generator supports."
  
  (with-slots (allowed-search-orders) search-problem
	(member search-order allowed-search-orders)))

(defmethod initialize-search-order! 

	((search-problem search-problem))

  "Determine and sets the search queue SEARCH-ORDER slot of SEARCH-PROBLEM.
   SEARCH-ORDER specifies the order in searcn-nodes are expanded.
   SEARCH-ORDER is one of *ALLOWED-SEARCH-ORDER-TYPES*, or :DEFAULT.  
   :DEFAULT is equivalent to :depth-first.  SEARCH-ORDER defaults to :DEFAULT."

  (with-slots (search-order) 
	  search-problem
	(when (eql search-order :DEFAULT)
	  (setq search-order :DEPTH-FIRST)))
  T)

(defmethod performing-best-first-search? 
	
	((search-problem search-problem))

  "Returns null unless SEARCH-PROBLEM is
   exploring the queue in best-first order.

   Note: SEARCH-PROBLEM can not perform BFS 
   unless it is an informed search=problem."

  (with-slots (search-order) search-problem
	(eql search-order :best-first)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Search Problem w Conflicts Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Where possible, generalize methods on search-problem-w-conflicts 
;;;   to search-problem.

(defclass search-problem-w-conflicts 
  	(search-problem state-space-w-conflicts)
  
  ()
  
  (:documentation   "State space search problem and 
                     conflict-directed systematic search method."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Informed Search Problem Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Split into w and wo-conflicts.
;;; o This naming is too complicated.
;;;   informed-search-problem -> decision-problem or optimization-problem

(defclass informed-search-problem (decision-space search-problem)
  
  ((name
	:initform (gentemp "informed-search-problem-")
	:documentation "Name of Informed Search Problem.")
   
   ;; Search Parameters

   (allowed-search-orders
	:initform '(:depth-first :breadth-first :best-first))
   
   (branch-and-bound?
	:initarg :branch-and-bound?
	:initform nil
;;;	:accessor search-problem-branch-and-bound?
	:documentation "A boolean flag indicating whether or not
                    Branch and Bound search is being performed.")
   
   ;;; Search State
   
   (node-type
	:initform 'informed-search-node
	:documentation "Class type of search-tree-node of search-problem.")

   (incumbent-node 
	:initform :NONE
	:documentation "The best cost solution found thus far. ~
                    Incumbent is always :NONE, when NOT ~
                    performing branch and bound search.")
   
   ;;; Solutions
   
   (best-solution-cost 
	:initform nil 
	:accessor search-problem-best-solution-cost
	:documentation "The cost of the best solution."))
  
  (:documentation "An informed search problem."))

(defmethod print-object 

	((search-problem informed-search-problem) stream)

  (with-slots (name) search-problem
	(format stream "<ISP: ~a>" name)))

(defmethod initialize-search-order! 

	((search-problem informed-search-problem))

  "Determine and sets the search queue SEARCH-ORDER slot of SEARCH-PROBLEM.
   Operates on the values of the following two slots:
   BRANCH-AND-BOUND? is a flag, which is non-nil iff
     Branch and bound search is to be performed.  Defaults to nil.
   SEARCH-ORDER is a keyword, specifying the order in 
     which the candidate is explored.  SEARCH-ORDER is one of 
     *ALLOWED-SEARCH-ORDER-TYPES*, or :DEFAULT.  :DEFAULT is
     equivalent to :depth-first if branch and bound is specified,
     and :best-first, otherwise.  SEARCH-ORDER defaults to :DEFAULT."

  (with-slots (search-order branch-and-bound?) 
	  search-problem

	(when (eql search-order :default)
	  
	  (if branch-and-bound?
		  
		  (setq search-order :depth-first)
		
		(setq search-order :best-first))))
  T)

(defmethod performing-branch-and-bound? 

	((search-problem informed-search-problem)) 
  
  "Returns True IFF performing branch and bound search
   for SEARCH-PROBLEM."
  
  (with-slots (branch-and-bound?) search-problem
	(not (null branch-and-bound?))))

(defmethod no-incumbent? 
	
	((search-problem informed-search-problem))
  
  "Returns true if no incumbent has been identified for
   SEARCH-PROBLEM."
  
  (with-slots (incumbent-node) search-problem
	(eql incumbent-node :none)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Informed Search Problem w Conflicts Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o generalize methods on informed-search-problem-w-conflicts
;;;   to informed-search-problem.
(defclass informed-search-problem-w-conflicts 
	(decision-space-w-conflicts informed-search-problem search-problem-w-conflicts)
  
  ((node-type
	:initform 'informed-search-node-w-conflicts
	:documentation "Class type of search-tree-node of search-problem."))
  
  (:documentation "An informed search problem 
                   and conflict-directed search method."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Search Tree Node Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; delete references to this variable when code works.
;;;(defvar *search-node-counter* 0)

(defclass search-node ()

  ((search-problem
	:initarg :search-problem
	:initform nil
	:accessor search-node-search-problem
	:documentation "Search-problem being explored.")
   
   (index
;;;	:initform (incf *search-node-counter*)
	:accessor search-node-index
	:documentation "Unique index for search node, a non-negative integer.")
   
   (ancestor-assignments
	:initarg :ancestor-assignments
	:accessor search-node-ancestor-assignments
	:initform nil
    :documentation "Assignments already made by the candidate generator.
                    (e.g., constituent kernels of a subset of the conflicts).

                    Note: This list builds upon its ancestor nodes, and hence
                    must be copied before modifying.")

   (source-minimal-relaxations
	:initarg :source-minimal-relaxations
	:accessor search-node-source-minimal-relaxations
	:initform nil
    :documentation "The minimal relaxation that the node is made from.")

   (ancestor-assignments-bit
	:accessor search-node-ancestor-assignments-bit
	:documentation "The bit vector representation of the clause given by the 
                    ancestor assignments. Does not include the children because 
                    those change a lot and are easily added.")
   
   (child-assignments
	:initarg :child-assignments
	:accessor search-node-child-assignments
	:initform '()
	:documentation "A set of alternative assignments to be used to extend ~
                    the ancestor assignments of node.  Child-assignments are ~
                    listed in best first order.  When search node splits on ~
                    an unresolved conflict, the child assignments ~
                    are the constituent kernels of that conflict.")
   
   (unassigned-ancestor-variables
	:initarg :unassigned-ancestor-variables
	:accessor search-node-unassigned-ancestor-variables
    :documentation "All variables not listed in an ANCESTOR-ASSIGNMENT.")
	
   (unassigned-variables
	:initarg :unassigned-variables
	:accessor search-node-unassigned-variables
    :documentation "All variables not listed in either an ANCESTOR-ASSIGNMENT 
                    OR the BEST CHILD ASSIGNMENT.")


   ;; Slots related to conflicts.
   
   (disallowed-assignments
	:initarg :disallowed-assignments
	:accessor search-node-disallowed-assignments
	:initform '()
    :documentation "Assignments that cannot appear in any descendant. These ~
                    are used to ensure that the search tree is systematic.")
    
   (split-on-conflict?
	:initarg :split-on-conflict?
	:accessor search-node-split-on-conflict?
	:initform nil ;; Defaults to not split on conflict.
	:documentation "True if children are constituent kernels, which are ~
                    selected resolve a conflict.")
   )

  (:documentation "An uninformed search tree node, denoting a snapshot of one
                   of the candidate generator's search states.  For 
                   conflict-directed search, if a node splits on a conflict,
                   then the children are the constituent kernels of an
                   unresolved conflict."))

(defmethod print-object 

	((search-node search-node) stream)
  
  "Prints short descriptor for a candidate generator search node."

  (with-slots (index) search-node
	(format stream "N~a" index)))

(defmethod initialize-instance :after 
		   
		   ((search-node search-node) 
			&key)

  "Creates and returns a new SEARCH-NODE for SEARCH-PROBLEM.
   SEARCH-NODE denotes an expanded search node whose
   state is ANCESTOR-ASSIGNMENTS, and whose unexpanded children 
   are CHILD-ASSIGNMENTS, where ANCESTOR and CHILD.
   ASSIGNMENTS are lists of state variable assignment objects.
   SEARCH-NODE is given a unique ID.

   UNASSIGNED-ANCESTOR-VARIABLES is a list of state variable objects
     containing all variables not assigned in ancestor-assignment,
     it may include assigned variables, but the first must not be 
     assigned.

   UNASSIGNED-VARIABLES is a list of state variable objects
     containing all variables not assigned by the current 
     assignment = ancestors + best child, it may include assigned 
     variables, but the first must not be assigned.

   DISALLOWED-ASSIGNMENTS is a list of assignments that cannot
     appear in a candidate of search-node.

   SPLIT-ON-CONFLICT? is true if the children of search-node
     is a set of constituent kernels, used to resolve a conflict."

  (with-slots (split-on-conflict? index search-problem ancestor-assignments ancestor-assignments-bit) 
	  search-node
	
	;;Record the bit vector version of the ancestor assignments
	(setf ancestor-assignments-bit 
	  (if (null ancestor-assignments)	;if there aren't any
		  ;;create the default - no assignments
		  (list (copy-seq (state-space-zero-bit-vector search-problem)) 
				(copy-seq (state-space-zero-bit-vector search-problem)))
		;;Otherwise actually calculate it
		(clause-to-bit-clause ancestor-assignments search-problem)))

	(setq 
		;; Records flag on whether node's children resolve
		;; a conflict.
		split-on-conflict? (not (null split-on-conflict?)))
	
	(setq index
	  (incf (search-problem-count-of-search-nodes-created
			 search-problem)))
	T))

(defmacro search-node-split-on-conflict! 
	(search-node)

  "Declares that the children of SEARCH-NODE are selected
   to resolve an unresolved conflict."
  
  `(setf (search-node-split-on-conflict? ,search-node) t))

;;; Optimization:
;;; o May want to throw away disallowed assignments that 
;;;   are ruled out by ancestor-assignments.
(defmacro search-node-add-disallowed-assignment! 

	(assignment search-node)
  
  "Add ASSIGNMENT to the list of assignments that cannot appear in
   a candidate of SEARCH-NODE."
  
  `(push ,assignment
		 (search-node-disallowed-assignments ,search-node)))

(defmethod search-node-current-assignments ((search-node search-node))

  "Return a list with the full current assignment of a node. This is the 
   ancestors of and the current preferred child."
  
  (with-slots (ancestor-assignments-bit) search-node
	(let ((current-assignments
		   (append (search-node-ancestor-assignments search-node) 
				   (list (search-node-preferred-child-assignment search-node)))))
	  (when (member nil current-assignments)
		
		;;TODO Remove this if it doesn't come up.
		(error current-assignments "There shouldn't be a nil.") 
		)
	  current-assignments)))

(defmethod search-node-current-assignments-bit 
	
	((search-node search-node) (search-problem search-problem-w-conflicts))
  
  "Return the bit vector representation of the current assignment. Calculate this
   by appending the current child to the bit vector version of the ancestors."
  
  (with-slots (assignment-bit-ht) search-problem
	(with-slots (ancestor-assignments-bit) search-node
	  (append-assignment-to-bit-clause 
	   ancestor-assignments-bit
	   (gethash (search-node-preferred-child-assignment search-node)
				assignment-bit-ht)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Informed Search Tree Node Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass informed-search-node (search-node)

  ((best-cost 
	:initform nil
	:accessor search-node-best-cost
	:documentation "The best cost of all states denoted by ~
                    search-node (i.e., leaf states rooted at this node). ~
                    This cost is computed based on the assignments: ~
                    BEST-CHILD-ASSIGNMENT and ANCESTOR ASSIGNMENTS."))

  (:documentation "An informed search tree node, denoting a snapshot of one
                   of the candidate generator's search states.  For 
                   conflict-directed search, if a node splits on a conflict,
                   then the children are the constituent kernels of an
                   unresolved conflict."))

;;; Changed:
;;; o Modified print out.

(defmethod print-object 

	((search-node informed-search-node) stream)
  
  "Prints short descriptor for a candidate generator informed search node."

;;;  (with-slots (index best-cost) search-node
;;;	(format stream "N~a[c=~a]" index best-cost))
  (with-slots (index) search-node
	(format stream "N~a" index))
  )

(defmethod initialize-instance :after

		   ((search-node informed-search-node) 
			&key search-problem)

  "Creates and returns a new SEARCH-NODE for SEARCH-PROBLEM.
   SEARCH-NODE denotes an expanded search node whose
   state is ANCESTOR-ASSIGNMENTS, and whose unexpanded children 
   are CHILD-ASSIGNMENTS, where ANCESTOR and CHILD.
   ASSIGNMENTS are lists of decision variable assignment objects.
   SEARCH-NODE is given a unique ID and an optimistic bound
   on its cost.

   UNASSIGNED-ANCESTOR-VARIABLES is a list of decision variable objects
     containing all variables not assigned in ancestor-assignment,
     it may include assigned variables, but the first must not be 
     assigned.

   UNASSIGNED-VARIABLES is a list of decision variable objects
     containing all variables not assigned by the current 
     assignment = ancestors + best child, it may include assigned 
     variables, but the first must not be assigned.

   DISALLOWED-ASSIGNMENTS is a list of assignments that cannot
     appear in a candidate of search-node.

   SPLIT-ON-CONFLICT? is true if the children of search-node
     is a set of constituent kernels, used to resolve a conflict.

   Computes BEST-COST for each node, the least upper bound on
   the cost of the states under node."

  (with-slots (best-cost) search-node
	
	;; Records least bound on cost.
	(setq best-cost (compute-search-node-best-cost
					 search-problem search-node))
	T))

;;;(defmacro search-node-best-child-assignment 
;;;
;;;	(search-node)
;;;
;;;  "Returns the best cost assignment that resolves
;;;   the working conflict of SEARCH-NODE, given that 
;;;   at least one child of SEARCH-NODE exist."
;;;
;;;  `(children-best-assignment 
;;;	(search-node-child-assignments ,search-node)))

;;;(defmacro search-node-sibling-assignments 
;;;
;;;	(search-node)
;;;
;;;  "Sibling assignments to NODE's best child assignment."
;;;  
;;;  `(lesser-sibling-assignments
;;;    (search-node-child-assignments ,search-node)))

(defmacro uninformed-preferred-assignment 

	(child-assignments)

  "Returns the best cost assignment of a list of
   CHILD-ASSIGNMENTS."

  `(first ,child-assignments))

(defmacro uninformed-lesser-sibling-assignments 

	(child-assignments)

  "Returns all but the best cost assignment of a list of
   CHILD-ASSIGNMENTS."

  `(rest ,child-assignments))

(defmethod search-node-preferred-child-assignment 

	((search-node search-node))

  "Returns the most promising child of SEARCH-NODE
   to be expanded during uninformed search."

  (uninformed-preferred-assignment
   (search-node-child-assignments search-node)))

(defmethod search-node-preferred-child-among

	((search-node search-node) child-assignments)

  "Returns the most promising child of SEARCH-NODE
   to be expanded during uninformed search."

  (uninformed-preferred-assignment
   child-assignments))

(defmethod search-node-lesser-sibling-assignments 

	((search-node search-node))

  "Sibling assignments to the first child 
   assignment of SEARCN-NODE that uninformed
   search expands."
  
  (uninformed-lesser-sibling-assignments
   (search-node-child-assignments search-node)))

(defmacro children-best-assignment 

	(child-assignments)

  "Returns the best cost assignment of a list of
   CHILD-ASSIGNMENTS."

  `(first ,child-assignments))

(defmacro lesser-sibling-assignments 

	(child-assignments)

  "Returns all but the best cost assignment of a list of
   CHILD-ASSIGNMENTS."

  `(rest ,child-assignments))

(defmethod search-node-preferred-child-assignment 

	((search-node informed-search-node))

  "Returns the best cost assignment that resolves
   the working conflict of SEARCH-NODE, given that 
   at least one child of SEARCH-NODE exist."

  (children-best-assignment 
   (search-node-child-assignments search-node)))

(defmethod search-node-preferred-child-among

	((search-node informed-search-node) child-assignments)

  "Returns the best cost assignment that resolves
   the working conflict of SEARCH-NODE, given that 
   at least one child of SEARCH-NODE exist."

  (children-best-assignment 
   child-assignments))

(defmethod search-node-lesser-sibling-assignments 

	((search-node informed-search-node))

  "Sibling assignments to NODE's best child assignment."
  
  (lesser-sibling-assignments
   (search-node-child-assignments search-node)))

;;; Changed:
;;; o Modified print out.

;;; To Do:
;;; o Generalize this to either informed or uninformed search-node.
;;; o Standardize on the name of a print function for presenting
;;;   the details of an object.
(defun print-search-node-details 

    (search-node &optional stream depth)
  
  "Prints a detailed description of a candidate generator search node."

  (declare (ignore depth))

  ;;; problems when printing this out
  ;;; Error: formatter found unsyntactic format string:
  ;;; Unknown format directive
  ;;; -- gustavop
  (format stream 
;;;		  "~a:~a s:~a d:~a u:~a ua:~a"
		  "            Node ~a: ~%              best cost            = ~a ~%              preferred expansion  = ~a  ~%              siblings of ~a        = ~a  ~%              disallowed           = ~a  ~%              unassigned           = ~a ~%              unassigned ancestors = ~a"
										;  ""
		  search-node
		  (ignore-errors (search-node-best-cost
						  search-node))
		  (ignore-errors (search-node-preferred-expansion 
						  search-node))
		  (ignore-errors (search-node-preferred-child-assignment
						  search-node))
		  (ignore-errors (search-node-lesser-sibling-assignments
						  search-node))
		  (ignore-errors (search-node-disallowed-assignments
						  search-node))
		  (ignore-errors (search-node-unassigned-variables 
						  search-node))
		  (ignore-errors (search-node-unassigned-ancestor-variables 
						  search-node))
		  ))

;;;(defun print-search-node-details 
;;;
;;;    (search-node &optional stream depth)
;;;  
;;;  "Prints a detailed description of a candidate generator search node."
;;;
;;;  (declare (ignore depth))
;;;  (format stream 
;;;;;		  "~a:~a s:~a d:~a u:~a ua:~a"
;;;		  "~a = ~a  (sibs of ~a = ~a) (dis=~a unasgn=~a un-anc=~a)"
;;;		  search-node
;;;		  (ignore-errors (search-node-preferred-expansion 
;;;						  search-node))
;;;		  (ignore-errors (search-node-preferred-child-assignment
;;;						  search-node))
;;;		  (ignore-errors (search-node-lesser-sibling-assignments
;;;						  search-node))
;;;		  (ignore-errors (search-node-disallowed-assignments
;;;						  search-node))
;;;		  (ignore-errors (search-node-unassigned-variables 
;;;						  search-node))
;;;		  (ignore-errors (search-node-unassigned-ancestor-variables 
;;;						  search-node))
;;;		  ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Maintaining a Search-Nodes' List of Unassigned Variables
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Moved to bfig-generator.lisp
;;;   - Delete this and the above banner.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Comparing Search Nodes based on Cost
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod better-search-node? 
	
	((node1 informed-search-node)
	 (node2 informed-search-node)
	 (search-problem informed-search-problem-w-conflicts))
  
  "Returns true IFF search NODE1 of SEARCH-PROBLEM
   has a better cost than NODE2 of SEARCH-PROBLEM."
  
  (with-slots (cost-order-function) search-problem
	(funcall cost-order-function
			 (search-node-best-cost node1)
			 (search-node-best-cost node2))))

(defmethod better-than-incumbent? 
	
	((node informed-search-node)
	 (search-problem informed-search-problem-w-conflicts))
  
  "Returns true IFF search NODE of SEARCH-PROBLEM
   has a worse cost than the INCUMBENT of SEARCH-PROBLEM."
  
  (with-slots (incumbent-node) search-problem
	(better-search-node? node incumbent-node search-problem)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Informed Search Tree Node Object w conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass informed-search-node-w-conflicts (informed-search-node)

  ((resolvable-conflicts
	:initarg :resolvable-conflicts
	:accessor search-node-resolvable-conflicts
	:initform nil
	:documentation "A list of all conflicts not resolved by the ancestors
                    which can be resolved by some further assignment. Sorted by 
                    increasing length of the conflict clause.")
   
   (newest-processed-conflict
	:initarg :newest-processed-conflict
	:initform -1						;always less than the default 0
	:accessor search-node-newest-processed-conflict
	:documentation "The number of the conflict processed most recently - used to 
                    avoid reprocessing conflicts that have been avoided."))

  (:documentation "An informed search tree node, adding slots for managing
                   conflicts to the informed search node class."))

(defmethod process-new-conflicts ((search-node informed-search-node) 
								  (search-problem search-problem-w-conflicts))

  "Consider any new conflicts. Return whether the node still appears to be valid. 
   If false,the node is totally inconsistent and should be discarded.

   One of three things can happen:
   1. Resolved by the ancestors. Ignore.
   2. Unresolvable with the ancestors - this node and all children are 
      inconsistent. Stop and return false.
  3. Not resolved yet, but not all the variables of the conflict are unassigned so
     it may be resolved in the future. Insert into the resolvable-conflicts list, 
     sorted by the length of the conflict. "
  
  (with-slots (newest-processed-conflict resolvable-conflicts) search-node

	;;(format t "processing ~%")
  
	(loop for new-conflict in (state-space-conflicts search-problem)
;;;	  do (format t "considering conflict: ~a~%" new-conflict)

							  ;;Loop over all conflicts newer than the ones 
							  ;; you've seen.
							
		while (and (not (null new-conflict))
				   (> (conflict-number new-conflict) newest-processed-conflict))
			
			  ;;Test if it was avoided already.
		do (if (ancestors-avoid-conflict? search-node new-conflict search-problem)
			 
			   (progn						
				 ;; (format t "avoided already ~%")
				 nil)					;conflict resolved, do nothing.
		   
			 ;;Check if they were all assigned to see if there's a problem.
			 (if (ancestors-assign-all-vars? search-node new-conflict)
			   
				 (progn					; All assigned, so definitely a conflict.

				   ;; Check if the conflict can be resolved continuously

				   ;; Yes -- Resolve continuously


				   ;; No -- Unresolvable.
				   (bump-conflict! new-conflict search-problem) ; Bump the conflict.
				   (record-node-inconsistent-with-conflict! search-node new-conflict search-problem)
				   (return-from process-new-conflicts nil)) ;Unresolvable conflict, 
										; node is inconsistent.
			   ;;Not all assigned, so put into unresolved list.
			   (progn
				 ;; (format t "inserted~%")
				 (setf resolvable-conflicts
				   (insert-in-order
					new-conflict
					resolvable-conflicts
					:order #'<=
					:key #'(lambda (conf)
							 (length (conflict-constituent-kernels conf)))))))))

	;; Update the most recently processed one.
	(unless (null (state-space-conflicts search-problem))
	
	  (setf newest-processed-conflict
		(conflict-number (first (state-space-conflicts search-problem))))))

  ;; If it got this far, there were no inconsistent assignments so return true.
  t)

(defmethod test-ancestors-against-resolvable

	((search-node informed-search-node-w-conflicts)
	 (search-problem search-problem-w-conflicts))

  "Check the ancestors against the resolvable conflicts. Used when a node is 
   created by splitting, so make sure that the new assignment, which resolved at 
   least that one conflict, doesn't violate any others."

  (with-slots (resolvable-conflicts) search-node
	
	(loop for conflict in resolvable-conflicts
		with active?
		do (setf active? (null (conflict-decayed? conflict)))
		   
		if active? 
		   
		do (if (ancestors-avoid-conflict? search-node conflict search-problem)
			   
			   ;; Conflict resolved, may discard now.
			   (progn					
				 ;; (format t "toss~%") 
				 (setf resolvable-conflicts
				   (delete conflict resolvable-conflicts)))
			 
			 ;;If it wasn't avoided, see if all were assigned.
			 (if  (ancestors-assign-all-vars? search-node conflict)
				 
				 ;; Unresolvable conflict, node is inconsistent.
				 (return-from test-ancestors-against-resolvable nil) 
			   
			   ;; Not all are assigned, stays on the unresolved list.
			   nil))
		   
		   ;; Get rid of the decayed conflict so stop checking.
		else	 
		do
		;; (format t "not active")
		(setf resolvable-conflicts (delete conflict resolvable-conflicts))
		(bump-conflict! conflict search-problem)))
  
  ;;Made it this far, no problem, return true.
  
  t)

;;; Fixed!
;;; We should return only **active** conflicts. Else we may break the compatible
;;; search order for conditional search computed in setup-variable-search-order!
(defmethod shortest-resolvable-conflict
	((search-node informed-search-node-w-conflicts)
	 (search-problem search-problem-w-conflicts))
  
  "The shortest conflict is the first one in the list that the current assignment 
   doesn't resolve.  Remember that the current assignment becomes the ancestor 
   and you shouldn't split on anything resolved by that ancestor. The 
   resolvable-conflicts list includes things that might be resolved by the 
   current child."
  
  ;;   (format t "shortest: ~a ~% current~a~%" 
  ;;          (search-node-resolvable-conflicts search-node)
  ;; 		  (search-node-ancestor-assignments search-node))
  
  (let ((first-conflict 
		 (loop for conflict in (search-node-resolvable-conflicts search-node)
			 thereis (when (and (not (current-avoid-conflict?
									  search-node conflict search-problem))
								(conflict-active? conflict))
					   
					   conflict))))
	
	(unless (null first-conflict)
	  
	  (bump-conflict! first-conflict search-problem))
	
	first-conflict))

(defmethod test-current-against-resolvable
	((search-node informed-search-node-w-conflicts)
	 (search-problem search-problem-w-conflicts)) 

  "Check if the current assignment creates any unresolvable conflicts. 
   The current assignments are temporary, so don't record anything."
  
  (with-slots (resolvable-conflicts) search-node

	;;(format t "conflicts: ~a~%" resolvable-conflicts)

	(loop for conflict in resolvable-conflicts
		with active?
		do (setf active? (null (conflict-decayed? conflict)))
		if active? 
		do (when (and (current-assign-all-vars?
					   search-node conflict search-problem)
					  ;;If ancestors are assigned, check for inconsistency
					  (not (current-avoid-conflict?
							search-node conflict search-problem)))
			 
			 ;; Unresolvable conflict, node is inconsistent.
			 (return-from test-current-against-resolvable nil)
			 )
		else							

		  ;; Get rid of the decayed conflict so you stop checking.
		do (setf resolvable-conflicts (delete conflict resolvable-conflicts))
		(bump-conflict! conflict search-problem)))
  
  ;;Otherwise no problem
  t)





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Probems with Conditional Variables
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Space with Conditional Variables
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-state-space (state-space)
  
  ((variable-type
	:initform 'conditional-state-variable))
  
  (:documentation   "A state space, defined by a set of conditional
                     variables, and their corresponding domains."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional State Space with Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-state-space-w-conflicts
	(conditional-state-space 
	 state-space-w-conflicts)
  
  ((variable-type
	:initform 'conditional-state-variable-w-conflicts)

   (conflict-type
	:initform 'conditional-conflict))
  
  (:documentation "A state space, defined by a set of conditional
                   variables, with associated conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Decision Space with Conditional Variables
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-decision-space 
	(conditional-state-space decision-space)
  
  ((variable-type
	:initform 'conditional-decision-variable))
  
  (:documentation   "A valued state space, defined by a set of conditional
                     variables, their corresponding domains and costs for 
                     each domain element."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Decision Space with Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-decision-space-w-conflicts
	(conditional-decision-space
	 conditional-state-space-w-conflicts
	 decision-space-w-conflicts 
	 )
  
  ((variable-type
	:initform 'conditional-decision-variable-w-conflicts))
  
  (:documentation   "A valued state space, defined by a set of conditional
                     variables, with associated conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional State Variable
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Removed reader state-variable-active?
;;; o Removed guard as an init arg, and added it as
;;;   a make-instance keyword.
(defclass conditional-state-variable (state-variable)
  
  ((assignment-type
	:initform 'conditional-state-assignment)
	  
   (guard
;;;	:initarg :guard
	:accessor state-variable-guard
	:documentation "A state assignment used to activate state-variable.")
  
   (active?
	:initform nil
;;;	:reader state-variable-active?
	:documentation "True iff if guard is satisfied.  A variable must be
                   assigned a domain element iff its guard is satisfied."))

  (:documentation "A conditional variable of a state space, activated
                   by an assignment to another variable."))

;;; Changed:
;;;o Added method.

(defmethod state-variable-unguarded? 
	
	((state-variable state-variable))

  "Returns true iff STATE-VARIABLE is always active."

  T)

;;;(defconstant *always-active* *true*
;;;  "The guard value corresponding to variable being active.")

;;; Changed:
;;;o Added method.
(defmethod state-variable-unguarded?
	
	((state-variable conditional-state-variable))

  "Returns true iff STATE-VARIABLE is always active."

  (with-slots (guard) state-variable
	(eql guard *always-active*)))

;;; Changed:
;;;o Added method.
(defmethod state-variable-active? 
	
	((state-variable conditional-state-variable))

  "Returns true iff STATE-VARIABLE is active.
   A variable is active iff its guard condition is
   satisfied by the current assignment."

  (with-slots (active?) state-variable
	active?))

(defmethod print-object 

	((variable conditional-state-variable) stream)
  
  "Print short descriptor for a variable."
  
  (if (state-variable-unguarded? variable)
	  (with-slots (name) variable
		(format stream "~a" name))
	(with-slots (name guard) variable
	  (format stream "~a(~a)" name guard))))

;;; Changed:
;;; o Updated documentation regarding DOMAIN.
;;; o Guard now processed as an initialize-instance keyword.

;;; To Do:
;;; o Error check that guard is an assignment to
;;;   a defined variable.

;;; o This requires any variable mentioned in the guard
;;;   to already be created.  Consider relaxing this
;;;   requirement, for example, by internalizing
;;;   guards on demand, or when the problem is
;;;   initialized.

;;; Note:
;;; o This assumes that all variables are created
;;;   before any assignments are made.
(defmethod initialize-instance :after 
		   ((variable conditional-state-variable) 
			&key (guard *always-active*) state-space)
  
  "Create a conditional variable for STATE-SPACE with the specified name.  
   Guard is an assignment expression:

            (= variable-name value),

   an assignment object, or *always-active* (= *true*).
   If *true*, then variable is always active; otherwise, variable is
   active iff assignment is current.

   Note, the variable mentioned in a guard must already
   be created by the time it is mentioned in a guard.

   Each DOMAIN element description is either an atom, 
   denoting an element, or list of the form:

   (ELEMENT . <keyword-value-pair>*)
  
   where ELEMENT is the value of an assignment, and the 
   keyword-value-pairs correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of VARIABLE.

   Assumes that, when a variable is created, no current assignments
   exist."

  (with-slots ((intern-guard guard) active?) variable
	(cond ((eql guard *always-active*)
		   (setq intern-guard *always-active*
				 ;; Variable must be active if it has no
				 ;; guard condition.
				 active? t))

		  (:otherwise

		   ;; Internalize the guard expression.
		   (setq intern-guard
			 (internalize-state-assignment state-space guard)
			 ;; Initially there are no current assignments, and
			 ;; hence a guarded variable is inactive.
			 active? nil)
		   
		   ;; Link variable to the assignment that activates it.
		   (push variable
				 (state-assignment-variables-activated intern-guard))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional State Variable with Conflicts Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-state-variable-w-conflicts 
	(conditional-state-variable state-variable-w-conflicts)
  
  ((assignment-type
	:initform 'conditional-state-assignment-w-conflicts)
   
   (conflicts-activated
	:initform '()
	:accessor state-variable-conflicts-activated
	:documentation "conflicts whose inconsistent assignments
                    include variable"))

  (:documentation "A conditional state variable whose assignments
                   can participate in conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Decision Variable
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-decision-variable 
	(decision-variable conditional-state-variable)
  
  ((assignment-type
	:initform 'conditional-decision-assignment))

  (:documentation "A conditional variable of a decision space, 
                   activated by an assignment to another variable."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Decision Variable with Conflicts Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-decision-variable-w-conflicts 
	(conditional-decision-variable
	 decision-variable-w-conflicts
	 conditional-state-variable-w-conflicts)
  
  ((assignment-type
	:initform 'conditional-decision-assignment-w-conflicts))

  (:documentation "A conditional decision variable whose assignments
                   can participate in conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional State Assignment Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-state-assignment (state-assignment)

  ((variables-activated
	:initform '()
	:accessor state-assignment-variables-activated
	:documentation "Conflicts that contain assignment."))
  
  (:documentation "A variable/value assignment to a conditional variable."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Decision Assignment Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-decision-assignment 
	(decision-assignment
	 conditional-state-assignment)
  ()
  (:documentation "A valued variable/value assignment to a 
                   conditional decision variable."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional State Assignment w Conflicts Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-state-assignment-w-conflicts
	(conditional-state-assignment
	 state-assignment-w-conflicts)
  ()
  (:documentation "A variable/value assignment to a conditional variable,
                   with associated conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Decision Assignment w Conflicts Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-decision-assignment-w-conflicts
	(conditional-decision-assignment
	 decision-assignment-w-conflicts
	 conditional-state-assignment-w-conflicts)

  ()
  
  (:documentation "A valued variable/value assignment to a 
                   conditional decision variable,
                   with associated conflicts."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Conflict Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Added.
(defclass conditional-conflict (conflict)

  ((inactive-variable-count
	:documentation "Number of variables in conflict's inconsistent
                    assignments that are not active.")
   
;;;   (active?
;;;	:initform nil
;;;	:reader conflict-active?
;;;	:documentation "True if and only if all variables mentioned in conflict
;;;                    are active.")
   )
  
  (:documentation 
   "A partial assignment to conditional variables 
    that is inconsistent with problem's constraints."))

;;; Changed:
;;; o Added.

;;; FIX: 
;;; o If variable is active, then do we need to update the
;;;   unresolved conflict count?

(defmethod initialize-instance :after 
		   ((conflict conditional-conflict)
			&key
			state-space)
  
  "Setup to track the activation of CONFLICT.  CONFLICT is ACTIVE
   iff the variables of its inconsistent assignments are active.

   Assumptions:  INCONSISTENT-ASSIGNMENTS have been setup when called."

  (setup-conflict-activation! state-space conflict)
  (initialize-conflict-activation! state-space conflict))

;;; Changed:
;;; o Added.
(defmethod conflict-active? 
	((conflict conditional-conflict))
  
  "Returns true iff CONFLICT is active.  CONFLICT is active IFF
   all of the variables in its inconsistent assignments are active."

  (with-slots (inactive-variable-count)
	  conflict
	(eql inactive-variable-count 0)))

;;; Changed:
;;; o Added.
;; (defmethod conflict-resolved? 

;; 	((conflict conditional-conflict))
  
;;   "Returns true if conflict is resolved by the current search state.
;;    Given a curent state, a conflict is resolved if conflict is INACTIVE
;;    or if one of its inconsistent assignments CAN NOT hold."
  
;;   (or (not (conflict-active? conflict))
;; 	  (conflict-supported? conflict)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Search Problem Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conditional-search-problem
	(conditional-state-space search-problem)
  
  ((name
	:initform (gentemp "conditional-search-problem-")))
  
  (:documentation "A search problem with conditional variables."))

(defmethod print-object 

	((search-problem conditional-search-problem) stream)

  (with-slots (name) search-problem
	(format stream "<CSP: ~a>" name)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Search Problem w Conflicts Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Generalize methods on conditional-search-problem-w-conflicts
;;;   to conditional-search-problem.
(defclass conditional-search-problem-w-conflicts
	(conditional-state-space-w-conflicts
	 conditional-search-problem
	 search-problem-w-conflicts)
  
  ()
  
  (:documentation "A search problem with conditional variables."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Decision Problem Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass informed-conditional-search-problem
	(conditional-decision-space 
	 conditional-search-problem
	 informed-search-problem)
  
  ((name
	:initform (gentemp "conditional-search-problem-")))
  
  (:documentation "A search problem with conditional variables."))

(defmethod print-object 

	((search-problem informed-conditional-search-problem) stream)

  (with-slots (name) search-problem
	(format stream "<ICSP: ~a>" name)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Conditional Decision Problem with Conflicts Object
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass informed-conditional-search-problem-w-conflicts
	(conditional-decision-space-w-conflicts
	 informed-conditional-search-problem
	 informed-search-problem-w-conflicts)
  
  ((name
	:initform (gentemp "conditional-search-problem-")))
  
  (:documentation "A search problem with conditional variables."))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Definitions for Soft Constraints Handling
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; -- gustavop
;;;
;;;  ** Class definitions to represent a soft-constraint. **
;;; 
;;; A soft-constraint is a cost function that maps assignments to the variables in its
;;; scope to values. The scope is simply the domain of the soft-constraint, i.e., the 
;;; set of variables mentioned in an assignment. Every assignment of the soft-constraint
;;; is also called a tuple, and it containts an assignment to each variable in the scope.
;;;	As opposed to a hard-constraint, a tuple of a soft-constraint may take values other
;;; than 0 (satisfaction) and infinity (unacceptable violation).
;;;
;;; An objective function is composed of one or more soft-constraints, i.e., one or more
;;; smaller cost functions. Given a full-assignment to the variables in a problem, we may
;;; evaluate the objective function by "composing" the valuations from each soft-constraint.
;;; (where "composing" means applying an operator which is associative, commutative an 
;;; monotonic. E.g.: +, *)
;;;
;;; e.g.: Variables = A, B, C. A soft constraint f1 with scope (A,B,C) may be defined by:
;;; f1(A,B,C) = '((((A = a-val-1) (B = b-val-2) (C = c-val-5)) reward-1)
;;;		  (((A = a-val-4) (B = b-val-3) (C = c-val-6)) reward-2) ....(other tuples)...)
;;; That is f1 is a cost function which associates the value reward-1 with the set of assignments
;;; ((A = a-val-1) (B = b-val-2) (C = c-val-5)) and reward-2 with ((A = a-val-4) (B = b-val-3) (C = c-val-6))
;;;  
;;;
;;; Note: OpSat standard problem is an OCSP. To make the extension easier, the new 
;;; definitions extend the ones related to a problem in which MIP does apply.
;;; Nonetheless, conceptually speaking, shouldn't we decribe an OCSP (w/ MIP) as an
;;; extension of a Valued-CSP where a soft-constraint must have just one variable within
;;; its scope?

(defclass decision-variable-w-conflicts-vcsp 
    (decision-variable-w-conflicts)
  
  ((assignment-type 
	:initform 'decision-assignment-w-conflicts))
  
  (:documentation "A decision-variable-w-conflicts whose assignments, 
		   standing by themselves, have no determined metric. 
		   When composing a tuple assignment, then there may be 
		   preference among its possible assignments."))


(defclass decision-space-w-conflicts-vcsp
    (decision-space-w-conflicts)
  
  ((variable-type
    :initform 'decision-variable-w-conflicts-vcsp)
   
   (variables-ht
	:initform (make-hash-table :test #'equal)
	:accessor state-space-variables-ht
	:documentation "A hash table mapping assignment
                    expressions to assignment objects. 
				    This hashtable is used when estimating the cost
				    of a node. Before bounding the cost of the node, its
					assignments will have been recored in the hash-table."))
  
  (:documentation "A Decision Space With Conflicts where MPI does not hold."))


(defclass informed-search-problem-w-conflicts-vcsp
    (informed-search-problem-w-conflicts decision-space-w-conflicts-vcsp)
  
  ((soft-constraint-system
	:initarg :soft-constraint-system
	:initform nil
	:accessor opsat-problem-soft-constraint-system
	:documentation "A soft-constraint system containing the soft-constraints of the
                    Optimal CSP."))
  
  (:documentation "An informed-seach-problem-w-conflicts with non-separable 
  		           functions i.e. soft-costraints."))


(defclass soft-tuple-assignment ()
  
  ((scope
	:initarg :scope
	:accessor scope-of-soft-tuple
	:documentation "Variables related to this tuple assignment, in same
			order.")

   (variables-assignments
	:initarg :variables-assignments
	:accessor assignments-of-soft-tuple
	:documentation
	"Assignments composing an element of a relation (soft-constraint).")
   
   (metric-value
	:initarg :metric-value
	:accessor metric-value
	:documentation "Metric value associated with this assignment of variables."))
  
  (:documentation "A set of assignments within an specified scope with a constant 
		   metric-value"))


(defclass soft-constraint ()
  
  ((name
	:initarg :name
	:initform (gentemp "SOFT-CONSTRAINT-")
	:accessor soft-constraint-name		
	:documentation "The name of a soft-constraint.")
   
   (ordered-tuples-of-assignments
	:initform nil
	:accessor tuples-of-soft-constraint 
	:documentation
	"Sequence of of tuples of assignments ordered by associated costs")
   
   (scope
	:initarg :scope
	:initform nil
	:accessor soft-constraint-scope
	:documentation
	"The scope of a soft constraint, i.e., the set of related variables.")))

(defmethod print-object 

	((soft-constraint soft-constraint) stream)

  (with-slots (name ordered-tuples-of-assignments scope) soft-constraint
	(format stream "<SC: ~a ~a ~a>" name scope ordered-tuples-of-assignments)))


(defmethod print-object

	((soft-tuple-assignment soft-tuple-assignment) stream)

  (with-slots (variables-assignments metric-value) soft-tuple-assignment
	(format stream "<ST: (~a ~a)>" variables-assignments metric-value))) 

(defmethod new-soft-constraint-list
    (scope tuples-with-costs metric-ordering-fn)
  
  "Creates and returns a valued-constraint, composed of a set of assignments
   with specific cost associated with each assignment tuple."
  
  (let ((soft-cons (make-instance 'soft-constraint :scope scope)))
    
    (loop for tuple-assignment in tuples-with-costs
		do (add-soft-tuple! soft-cons tuple-assignment :scope scope))
    (setf (tuples-of-soft-constraint soft-cons) 
	  (sort (tuples-of-soft-constraint soft-cons) 
			metric-ordering-fn :key  #'metric-value))
			      
    soft-cons))


(defmethod add-soft-tuple! 
    ((soft-constraint soft-constraint)
     tuple-assign
     &optional &key (scope nil))
  
  "Adds the tuple to the provided soft-constraint."
  
  (make-instance 'soft-tuple-assignment 
    :variables-assignments (car tuple-assign)
    :metric-value (cadr tuple-assign)
    :scope scope
    :related-soft-constraint  soft-constraint))


(defmethod initialize-instance :after
		   ((tuple-assignment soft-tuple-assignment)
			&key related-soft-constraint)
  
  "Inserts the tuple assignment in the soft-constraint.
   Side effect: tuple-assignment added to elements of related-soft-constraint."
  
  (push tuple-assignment  (tuples-of-soft-constraint related-soft-constraint)))



;; **********************************************************************
;; 
;;      Bit vector representation
;;
;; **********************************************************************

(defmethod initialize-bit-ht! ((state-space state-space))

  "Prepare the ht that provides the lookup between an assignment and it's bit
   vector representation. The representation is a list of two bit vectors. 
   The second is 0s except for the nth bit representing any active assignments. 
   The first one is 0s with 1s on the locations of all assignments made to any 
   assigned variable. The pair allows us to distinguish true/false assignments 
   from not being assigned.

   For example. Assume you had x={1,2} and y={1,2,3}, then the bit positions of 
   each bit vector are {x=1, x=2, y=1, y=2, y=3}. The construction is based on 
   the order of the vars and domains when this function is called, but thereafter,
   the bit-assignment-ht holds stable versions, so those lists may be
   reordered later. It is tricky, and unsupported currently, to change the number 
   of vars or domain elements because the representation requires fixed positions 
   of the data. 

   Note that if we try to represent x=1 with a single vector, say 10000, it is 
   impossible to distinguish whether y is unassigned, or whether y does not equal 
   2 but does equal 1, while only looking at the 4th element. Hence,we create 2 
   vectors, the first where the y=2 slot tells us if y is assigned at all, 
   and the second vector, whose y=2 slot tells us if the value, if assigned, is 
   actually 2.

   x=1 is represented by (#*11000 #*10000) because the first var, with 2 elements 
   is assigned, to its first value.  Similarly, y=2 is represented by 
   (#*00111 #*00010).

   To create clauses, simply OR component clauses two together. 
   x=2, y=3 is (#*11111 #*01001) 

   A value like (#*11110 #*01001) is illegal because in the first bit vector, y 
   is assigned for the first two domain elements and not assigned for the last 
   one. The value (#*11111 #11000) is not allowed because x is assigned the value 
   1 and 2. Also, it claims that y is given a value, but no value is selected in 
   the second vector.

  NOTE: This code assumes that domain elements of the variables are unique, i.e., 
  that the length of the assignment-ht is the same as the sum of the length of 
  all the domains. A vector out of bounds error can be caused in this function 
  by violating that assumption, say,by duplicating (= x 1) (= x 1) in the domain 
  of x. "

  (with-slots (assignment-bit-ht assignment-ht variables zero-bit-vector)
	  state-space
	
	(let ((num-assignments (hash-table-count assignment-ht))
										; The length of the bit vectors.
		  (curr-bit-index 0)			; The bit of the next assignment.
		  (curr-var-assigned-bit nil)	; The bit vector for the variable 
										; (first one).
		  (curr-assignment-bit nil))	; Bit vector of the assignment(second one).

	  (debug-print-progn-for 
	   :bfig (format t "Init bit vectors, total bits: ~a~%" num-assignments))

	  ;;Set aside a 0 vector.
	  (setf zero-bit-vector (make-array num-assignments 
										:element-type 'bit
										:initial-element 0))

	  ;;Do this for each variable.
;;	  (format t "~s~%" (length variables))
	  (loop for variable in variables
							;;Make a clean vector for the var.
		  do (setf curr-var-assigned-bit (make-array num-assignments 
													 :element-type 'bit
													 :initial-element 0))

			 ;;Mark off all the assignments for this var.
			 (loop for i from 0 to (+ -1 (length (state-variable-domain variable)))					
				 do (setf (bit curr-var-assigned-bit (+ i curr-bit-index))
					  1)
					)

			 ;;For each assignment,
			 (loop for assignment in (state-variable-domain variable)
									 ;; make the vector for this assignment 
									 ;; and set the bit.
				 do (setf curr-assignment-bit (make-array num-assignments 
														  :element-type 'bit
														  :initial-element 0))

					(setf (bit curr-assignment-bit  curr-bit-index)
					  1)
					(incf curr-bit-index) ; Go to the next bit,
				  
					;;and save the pair in the ht.
					(setf (gethash assignment assignment-bit-ht) 
					  (list curr-var-assigned-bit curr-assignment-bit))))
	  )))

(defun clause-to-bit-clause (clause state-space)

  "Take a clause, a list of assignments, and return the equivalent bit 
   representation. ASSUMES that the clause is valid!"
  
  (with-slots (assignment-bit-ht) state-space
	(loop for assignment in (rest clause)
							;;Start with the first assignment
		with building-bit-clause = (gethash (first clause) assignment-bit-ht)
								   ;;for all other elements, append the 
								   ;; assignment.
								   
		do (setf building-bit-clause 
			 (append-assignment-to-bit-clause
			  building-bit-clause (gethash assignment assignment-bit-ht)))
		   
		   ;;return the finished bit-clause.
		finally (return building-bit-clause))))

(defun append-assignment-to-bit-clause (bit-clause bit-assignment)

  "Add a single assignment to the clause if both are in bit form. OR the two
   parts of the bit representation."

  (if (null bit-assignment)				;if there is no assignment

	  bit-clause						;return the clause

	;;else calculate the real result
	(list (bit-ior (first bit-clause) (first bit-assignment))
		  (bit-ior (second bit-clause) (second bit-assignment)))))

(defun bit-rep-unresolvable-conflict? (conflict clause state-space)

  "There is an unresolvable conflict if all the variables are assigned and the 
   conflict is not avoided."

  (and (bit-rep-all-vars-assigned? conflict clause)
	   (not (bit-rep-conflict-avoided? conflict clause state-space))))

(defun bit-rep-resolvable-conflict? (conflict clause state-space)
  
  "There is a resolvable conflict the conflict is not explicitly avoided
   but there is some var of the variable that is not assigned yet."
  
  (and (not (bit-rep-all-vars-assigned? conflict clause))
	   (not (bit-rep-conflict-avoided? conflict clause state-space))))

(defmethod ancestors-assign-all-vars? 
	((search-node informed-search-node-w-conflicts)
	 (conflict conflict))

  "Method wrapper for the bit rep function."

  (bit-rep-all-vars-assigned? (conflict-inconsistent-assignments-bit conflict)
							  (search-node-ancestor-assignments-bit search-node)))

(defmethod current-assign-all-vars? ((search-node informed-search-node-w-conflicts)
									 (conflict conflict)
									 (search-problem search-problem-w-conflicts))

  "Method wrapper for the bit rep function."

  (bit-rep-all-vars-assigned? (conflict-inconsistent-assignments-bit conflict)
							  (search-node-current-assignments-bit 
							   search-node 
							   search-problem)))

(defun bit-rep-all-vars-assigned? (conflict clause)

  "Return true if all the variables assigned in the conflict are also assigned
   in the clause."

  (equal (first conflict)
		 (bit-and (first conflict) (first clause))))

(defmethod ancestors-avoid-conflict? 
	((search-node informed-search-node-w-conflicts)
	 (conflict conflict)
	 (state-space state-space))

  "Method wrapper for the bit rep function."
  
  (bit-rep-conflict-avoided? (conflict-inconsistent-assignments-bit conflict)
							 (search-node-ancestor-assignments-bit search-node)
							 state-space))

(defmethod current-avoid-conflict? ((search-node informed-search-node-w-conflicts)
									(conflict conflict)
									(state-space state-space))

  "Method wrapper for the bit rep function."

  (bit-rep-conflict-avoided? (conflict-inconsistent-assignments-bit conflict)
							 (search-node-current-assignments-bit
							  search-node state-space)
							 state-space))

(defun bit-rep-conflict-avoided? (conflict clause state-space)

  "The conflict is avoided if the clause explicitly makes a different
   assignment to some var in the conflict."

  (with-slots (zero-bit-vector) state-space
	
	(not								
										; Invert the following.
	 (equal zero-bit-vector				; Is equal to 0s if nothing breaks the 
										; conflict.
			(bit-and (first conflict)	;The conflict breaks if they share a var 
										; assigned,
					 (bit-and (first clause)
							  (bit-xor (second conflict) ;and the assignment is 
										; different (xor).
									   (second clause))))))))

(defmethod is-ordered? (cost1 cost2 (space decision-space))
  "Returns COST1 COST-ORDER-FN COST2"
  (funcall (decision-space-cost-order-function space) cost1 cost2))