;;; -*- Mode:Common-Lisp; Package:BFIG; Base:10 -*-

;;; Copyright (c) 1999 - 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :BFIG)

;;;(eval-when (compile load eval)
;;;  (export '(initialize-candidate-generator!
;;; 			next-candidate
;;;             candidate-is-solution!
;;;	    		)
;;;		  :BFIG)
;;;  )

;;; To Do:
;;; o Split into an informed and uninformed candidate
;;;   generator (also a kernel generator).
;;;   - modifications noted at top of functions to be modified.

;;; o Extend to conditional CSP.
;;;   - modifications noted at top of functions to be modified.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Best First Kernel (Implicant) Generator (BFIG)
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;  BFIG takes a search-problem and generates a set of kernels
;;;  (strictly speaking, "small" implicants) for the conflicts of 
;;;  search-problem.  It is guaranteed that the set of kernels 
;;;  returned imply all complete search states that are consistent 
;;;  with the conflicts of search-problem.

;;;  The cost of a kernel is the best cost of the search states that
;;;  are extensions of the kernel.  Kernels are generated best first. 
 
;;;  Kernels are generated on demand.  The generator tries to 
;;;  minimize the amount of work expended towards generating the next 
;;;  kernel.

;;;  The addition of conflicts to search problem may be interleaved 
;;;  with requests for kernels.  Each kernel is guaranteed to 
;;;  cover the set of conflicts that have been added to search-problem 
;;;  at the time of the request.  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; *** Best First Kernel (Prime Implicants) Generation:

;;;  (ALL-CANDIDATES
;;;	     (SEARCH-PROBLEM SEARCH-PROBLEM)
;;;	     &OPTIONAL &KEY
;;;			   (CUTOFF-COST NIL)
;;;			   (MAXIMUM-VISITED-CANDIDATE-COUNT NIL)
;;;			   (FULL-ASSIGNMENT? T)
;;;			   (DISJOINT? T)
;;;			   (SEARCH-ORDER :DEFAULT))
;;;  
;;;   Returns a list of all candidate solutions to SEARCH-PROBLEM, 
;;;   listed in SEARCH-ORDER.  Each solution is a list of 
;;;   externalized assignments.
;;;
;;;   Search terminates when it reaches a cutoff or if the 
;;;   candidates are exhausted.
;;;
;;;   If search is performed in best-first order, 
;;;   then CUTOFF-COST is the worst cost considered.  
;;;
;;;   MAXIMUM-VISITED-CANDIDATE-COUNT should be a positive
;;;   integer or nil.  If not null, then search will generate
;;;   a maximum of MAXIMUM-VISITED-CANDIDATE-COUNT candidates.
;;;
;;;   If FULL-ASSIGNMENT? is true, then search-node
;;;   specifies assignments to all decision-variables; otherwise, 
;;;   it returns a KERNEL that is (nearly) minimal.
;;;
;;;   If DISJOINT? is true, then the kernels returned denote disjoint
;;;   sets of assignments; otherwise, the kernels can overlap.
;;;
;;;   SEARCH-ORDER is is one of the keywords in 
;;;   *ALLOWED-SEARCH-ORDER-TYPES* (for example 
;;;   :BEST-FIRST), or :DEFAULT.  If :DEFAULT, then
;;;   search-order is the last one specified.

;;; (INITIALIZE-CANDIDATE-GENERATOR 
;;;    (SEARCH-PROBLEM INFORMED-SEARCH-PROBLEM)
;;;    &key (SEARCH-ORDER :BEST-FIRST))

;;;    Initializes the queue of search-nodes for 
;;;    SEARCH-PROBLEM with an search-node denoting all 
;;;    search-problem states.  The best cost state denoted by this 
;;;    set is the one that selects the best cost assignment for 
;;;    every variable.  
;;;    SEARCH-ORDER is is one of the keywords in 
;;;    *ALLOWED-SEARCH-ORDER-TYPES* (for example 
;;;    :BEST-FIRST), or :DEFAULT.  If :DEFAULT, THEN
;;;    SEARCH-ORDER IS THE LAST ONE SPECIFIED.
;;;    Called for effect, not value.

;;; (NEXT-CANDIDATE SEARCH-PROBLEM CUTOFF-COST 
;;;    MAXIMUM-VISITED-STATE-COUNT ...)
;;;  
;;;   Returns an SEARCH-NODE containing the next best cost 
;;;   kernel of the conflicts of SEARCH-PROBLEM
;;;   (the best cost assignment is one of its children).
;;;   If search is performed in decreasing (increasing) order, 
;;;   then the CUTOFF-COST is the least (greatest) cost considered.  
;;;   If search is cutoff or no kernel is found, the 
;;;   function returns NIL.  Otherwise, an search-node object 
;;;   is returned.


;;;  (SEARCH-NODE-PREFERRED-EXPANSION (SEARCH-NODE SEARCH-NODE))
;;;
;;;   Returns the preferred state of SEARCH-NODE, which consists
;;;   of its ancestor-assignments, augmented by the
;;;   child-assignment that is preferred to be expanded first.
;;;
;;;   Note: children are currently listed, left to right,
;;;   from most preferred to least preferred.
;;;
;;;   IMPORTANT: The preferred expansion is not a new list.
;;;   Hence before destructively modifying preferred-expansion, 
;;;   for example using sort, make sure to COPY the list first.
;;;
;;;   Note: If search-node splits on an unresolved conflict, then 
;;;   the children supplied must be the constituent-kernels of that 
;;;   conflict.

;;;  (CANDIDATE-IS-SOLUTION?
;;;     (NEW-SOLUTION SEARCH-NODE)
;;;	    (SEARCH-PROBLEM INFORMED-SEARCH-PROBLEM))
  
;;;   Informs the candidate generator that search-node NEW-SOLUTION
;;;   is a solution to SEARCH-PROBLEM.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do:
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Fix:

;;; o When all conflicts are removed, shouldn't the 
;;;   search state also be cleared?  Check the callers.  
;;;   - I'm concerned that the external interface allows 
;;;     the system to be placed into an inconsistent state.

;;; To Do:
;;; o Identify conflicts that are resolved because
;;;   of disallowed assignments.
;;; o Extend to perform weighted model counting.
;;; o Extend to maintain the search tree, either for 
;;;   debugging or for compilation.
;;; o Currently this generates kernels that are not minimal.
;;;   - Fixing this requires implementing a fast subset test.
;;;     This can be done, for example using the TRIE data
;;;     structure in deKleer's IPIA algorithm.  See de Kleer's
;;;     building problem solver text.

;;; o Update the export list.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Data structures for best first kernel (prime implicant) generation
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; BFIG uses a set of SEARCH-NODES to compactly describe all unexplored 
;;; SEARCH STATES that are not ruled out by a set of CONFLICTS.  The set of 
;;; SEARCH-NODES and corresponding costs are stored in 
;;; SEARCH-PROBLEM-SEARCH-QUEUE.

;;; A search problem includes a set of VARIABLES.  Each variable has
;;; a DOMAIN, consisting of an exclusive set of possible ASSIGNMENTS.  
;;; Each assignment has an associated COST.  All variables and their domain 
;;; are assumed to be specified before specifying any CONFLICT of a search 
;;; problem.

;;; Each state has an associated COST.  States are placed in best first order 
;;; based on ORDERING-FUNCTION.  Each state is a set of ASSIGNMENTS.  
;;; Each assignment has a COST associated with it.  
;;; The cost of a state is calculated from the cost of its assignments using 
;;; a BINARY-OBJECTIVE-FUNCTION.  

;;; We assume the objective satisfies the properties below.  In the following 
;;; we assume that the ordering-function is >=, so that best first order
;;; is in decreasing order; the changes for increasing order are noted 
;;; parenthetically:

;;;   - let L be a set of assignments, and let I be an assignment.
;;;   - the cost of (I . L) is calculated by applying a binary function to 
;;;     the cost of I and the cost of L (e.g., as the sum, product, etc.).
;;;   - the cost of L is greater (less) than or equal to the cost of (I . L).
;;;   - if I1 and I2 are assignments and the cost of I1 is less than the cost
;;;     of I2, then the cost of (I1 . L) is less than the cost of 
;;;     (I2 . L).

;;; Note: The above assumptions capture objective functions that calculate the
;;;       cost of L as the sum or product of the cost of its assignments.

;;; The binary-objective-function calculates the cost of a state recursively by 
;;; computing the cost of (I . L) from the cost of I and the computed cost 
;;; of L.

;;; A CONFLICT is a list representing a set of inconsistent assignments 
;;;   (v1 = m1, v2 = m2, ...., vn = mn), where the vi's are variables and mi's 
;;; are values for these variables.  The conflict contains at most one 
;;; assignment per variable.  

;;; The states not ruled out by the set of conflicts is described by a set 
;;; of kernels ("small" implicants) of the conflicts.  Any state that is an
;;; extension of a kernel is  guaranteed to be consistent with the set of 
;;; conflicts, and is said to be contained within the kernel.  The best 
;;; cost state within an kernel is the state that assigns the best 
;;; assignment to any variable unspecified in the kernel.  The cost of 
;;; a kernel is the cost of its best state.  Hence the best cost state 
;;; can be identified by generating the kernels in order of best to worst cost.

;;; Kernels are generated by a set covering algorithm on the set of 
;;; constituent kernels of the conflicts.  This algorithm generates the 
;;; kernels in best first order.  The 
;;; algorithm generates a kernel by repeatedly extending a set 
;;; of partial kernels, extending a partial kernel only if it is 
;;; possible for its extended value to be greater than all other partial 
;;; kernels.  The current partial kernels of a covering are 
;;; stored in SEARCH-QUEUE of SEARCH-PROBLEM and are encoded into 
;;; data structures called SEARCH-NODES as follows:

;;; Consider a list of partial kernels (I1 . L), (I2 . L), ..., (In . L).
;;; Assume that the cost of Ii is greater (less) than or equal to the cost 
;;; of Ij whenever i < j.  This means that (I1 . L) would always be 
;;; considered before the other partial kernels in that list.  Hence, 
;;; instead of generating all the above kernels explicitly, it suffices 
;;; to generate the common assignments L and a list of the alternate 
;;; assignments (l1 I2 I3 ... In), sorted in best-first order.  Once partial
;;; kernel (I1 . L) is processed, the set of alternate assignments is reduced 
;;; to (I2 I3 ... In).  The advantage of this approach is that it minimizes 
;;; the partial kernels explicitly maintained.  Each search-node 
;;; encodes such a list.  In particular, the list L is stored in the
;;; ASSIGNMENTS field of an search-node, the alternate assignments 
;;; (l1 I2 I3 ... In) are stored in the CHILD-ASSIGNMENTS field.  The next
;;; assignment with the best cost (l1) is accessed by BEST-NEXT-ASSIGNMENT.

;;; Also note that, during expansion:

;;;   o Kernel generation uses the classic heuristic from CSP of selecting 
;;;     a next variable to search over that has the smallest domain.  In this
;;;     case the domain is the constituent kernels of each conflict.  To support 
;;;     this conflicts are sorted by increasing cardinality of the number
;;;     of constituent kernels.

;;;   o The constituent kernels are sorted by "best" cost.  This is used by 
;;;     the best first search to make sure that the assignments added to 
;;;     states are the next ones in the best first order.  

;;; See the JDAM paper by Williams and Ragno for further details.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Generating Candidate Assignments
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Optimization:
;;; o Search-node should keep around the cost of ancestor 
;;;   assignments.  Then it should only have to compute the 
;;;   incremental cost of adding a new assignment.

;;; o Modularize.
;;;   - Put the routines related to current state that are used
;;;     in an initialization routine, in the conflicts package.
;;;     
;;;   - Put the routines for creating a root node in a separate fn.

;;; To Do:
;;; o FIX Delete this comment when works.
;;;   Split function into the part w-conflicts and w-o-conflicts.
;;;   For example, setup-variable-search-order! is for w-o-conflicts.
;;;    - Its not clear that any of this fn is special to w-conflicts,
;;;      if not, drop -w-conflicts from type.

;;; Changed:
;;; o Generalized to search-problem.
;;; o Added call (setup-variable-search-order! search-problem)
;;; o Updated to do.
(defmethod initialize-candidate-generator!

	((search-problem search-problem
;;; Fix delete when works					 
;;;					 search-problem-w-conflicts
					 )
	 &key (search-order :DEFAULT) 
	 &allow-other-keys)

  "Initializes the search and search statistics of SEARCH-PROBLEM.
   This function is called for effect, not for value.
   SEARCH-ORDER is :DEFAULT, one of the keywords in 
   *ALLOWED-SEARCH-ORDER-TYPES* (for example 
   :BEST-FIRST, :DEPTH-FIRST or :BREADTH-FIRST).  If :DEFAULT, then
   search-order is the last one specified.

   Note:  Make sure to call at least once AFTER all variables
          have been recorded.  This is required in order to set
          the order in which variables are split."

  (debug-print-for :bfig t "Initialize ~a." 
				   search-problem)

  ;; Clear-out old.



  ;; Clear record of the current search node.
  (clear-current-search-node! search-problem)

  ;; Clear variables
  (clear-state-variable-marks! search-problem)

  ;; Clear statistics of the search algorithm
  ;; performance.
  (clear-search-statistics! search-problem)

  ;; Sets up the order in which variables are split.
  (setup-variable-search-order! search-problem)
  
  ;; Setup queue for new search.
  (initialize-problem-search-queue!
   search-problem search-order)

  T)

;;; Changed:
;;; o Updated comments.

;;; Check:
;;; o During tests, why is this called three times at the start?
;;;   Could make this a subroutine all, rather than an after method.
;;; o Why is this restricted to -w-conflicts?  Shouldn't
;;;   it apply to informed-search-problem?

;;; Changed.
;;; o I also assumed that this method should apply to
;;;   informed-search-problem. Need to confirm.
;;; -- gustavop
(defmethod initialize-candidate-generator! :after

		   ((search-problem informed-search-problem)
			;; informed-search-problem-w-conflicts)
			&key &allow-other-keys)
    
  "Clear incumbent of SEARCH-PROBLEM."
  
  (clear-problem-incumbent! search-problem)
  T)

;;; Changed.
;;; o Same as in initialize-candidate-generator!
;;; -- gustavop

(defmethod all-candidates
	
	((search-problem search-problem)
	 ;; search-problem-w-conflicts)
	 &optional &key
			   (cutoff-cost nil)
			   (maximum-visited-candidate-count nil)
			   (full-assignment? T)
			   (disjoint? T)
			   (search-order :DEFAULT))
  
  
  "Returns a list of all candidate solutions to SEARCH-PROBLEM, 
   listed in SEARCH-ORDER.  Each solution is a list of 
   externalized assignments.

   Search terminates when it reaches a cutoff or if the 
   candidates are exhausted.

   If search is performed in best-first order, 
   then CUTOFF-COST is the worst cost considered.  

   MAXIMUM-VISITED-CANDIDATE-COUNT should be a positive
   integer or nil.  If not null, then search will generate
   a maximum of MAXIMUM-VISITED-CANDIDATE-COUNT candidates.

   If FULL-ASSIGNMENT? is true, then search-node
   specifies assignments to all decision-variables; otherwise, 
   it returns a KERNEL that is (nearly) minimal.

   If DISJOINT? is true, then the kernels returned denote disjoint
   sets of assignments; otherwise, the kernels can overlap.

   SEARCH-ORDER is is one of the keywords in 
   *ALLOWED-SEARCH-ORDER-TYPES* (for example 
   :BEST-FIRST), or :DEFAULT.  If :DEFAULT, then
   search-order is the last one specified."
  
  (initialize-candidate-generator!
   search-problem
   :search-order search-order)
  
  (loop for assignment = (next-assignment 
						  search-problem
						  :cutoff-cost cutoff-cost
						  :full-assignment? full-assignment?
						  :disjoint? disjoint?)
	  for count from 1 by 1

	  until (or (null assignment)
				(and maximum-visited-candidate-count
					 (>= count maximum-visited-candidate-count)))

	  collecting assignment))

;;; Changed.
;;; o Same as in initialize-candidate-generator!
;;; -- gustavop

(defmethod next-assignment
	
	((search-problem search-problem)
	 ;; search-problem-w-conflicts)
	 &optional &key
			   (cutoff-cost nil)
			   (maximum-visited-state-count nil)
			   (full-assignment? T)
			   (disjoint? T))

  "Returns a list of externalized assignments denoting the next 
   best candidate of SEARCH-PROBLEM.  The candidate consists of the
   node's  ancestor-assignments together with its best child 
   assignment.

   If search is cutoff or no candidate is found, then the 
   function returns NIL.  Otherwise, a list of assignment
   exprfessions is returned.  

   If search is performed in best-first order, 
   then CUTOFF-COST is the worst cost considered.  

   If FULL-ASSIGNMENT? is true, then search-node
   specifies assignments to all decision-variables; otherwise, 
   it returns a KERNEL that is (nearly) minimal.

   If DISJOINT? is true, then the kernels returned denote disjoint
   sets of assignments; otherwise, the kernels can overlap."

  (let ((candidate (next-candidate search-problem 
								   cutoff-cost 
								   maximum-visited-state-count
								   full-assignment?
								   disjoint?)))

	(unless (null candidate)
	  (externalize-state-assignments 
	   (search-node-preferred-expansion candidate)))
	))

;;; Optimization:

;;;  o expand-search-node-and-enqueue! non-deterministically 
;;;    selects one of the constituent kernels resolving each conflict.
;;;    These coverings may have redundant assignments and may not 
;;;    be minimal.  

;;;    o Is it worth trying to throw out redundant assignments 
;;;      and non-minimal coverings (probably not)?

;;;    o Implement according to the JDAM paper, to ensure
;;;      systematicity.

;;; To Do:
;;; o Separate into w vs w/o conflicts.
;;;   - update (typep ...) test.

;;; o In the future we will probably want to generalize opsat
;;;   to handle a range of objective functions.  To facilitate this,
;;;   - make the objective function a class, 
;;;   - collect together all methods, fns and macros operating on this 
;;;     objective into a separate file.
;;;   - convert appropriate fns and macros to methods.

;;; o Implement the generation of only kernels that are minimal.

;;; o This tracks the number of kernels generated. This info
;;;   should be displayed by the test system.

;;; Changed:
;;; o Added call to update unassigned variables of next-node.
;;; o Debug-print statements.

;;; Changed.
;;; o Same as in initialize-candidate-generator!
;;;   (applicable for search-problem)
;;; -- gustavop


(defmethod next-candidate

    ((search-problem search-problem)
	 ;;search-problem-w-conflicts)
	 cutoff-cost 
     maximum-visited-state-count
     full-assignment?
     disjoint?)
  
  "Returns a search-node denoting the next best candidate of
   SEARCH-PROBLEM.  The candidate consists of the node's 
   ancestor-assignments together with its best child assignment.

   If search is cutoff or no candidate is found, then the 
   function returns NIL.  Otherwise, a search-node object 
   is returned.  

   If search is performed in best-first order, 
   then CUTOFF-COST is the worst cost considered.  

   If FULL-ASSIGNMENT? is true, then search-node
   specifies assignments to all decision-variables; otherwise, 
   it returns a KERNEL that is (nearly) minimal.

   If DISJOINT? is true, then the kernels returned denote disjoint
   sets of assignments; otherwise, the kernels can overlap."

  ;;; Important Note:
  ;;; SEARCH-PROBLEM-CURRENT-SEARCH-NODE POINTS holds the 
  ;;; most recently considered node (except during the first round).
  ;;; This current node IS NOT EXPANDED, until the next round.

  (debug-print-for :bfig t "Find next candidate.")

  (loop 
	  with next-node
		   
		   ;; Finish up with last round:
		   ;; Expand the previous search node if one exists, 
		   ;; placing expansions on queue.
	  for previous-node = (search-problem-current-search-node search-problem)

						  ;; If there is a previous node that hasn't been 
						  ;; expanded yet:
	  do (unless  (null previous-node) 


		 
		   ;; Expand the previously visited search-node P and 
		   ;; place it on search queue.  The reason we 
		   ;; wait until the next iteration to expand P is 
		   ;; that testing the consistency of P's state 
		   ;; assigments might generate a new conflict.  When 
		   ;; this occurs we want to expand P after the conflict 
		   ;; has been recorded, so that it might be used for 
		   ;; expansion.
		   (expand-and-queue-current-search-node! search-problem
												  full-assignment?
												  disjoint?))
	  

		 ;; Start the next round, by dequeuing the next best node.
	  do (setq next-node (dequeue-search-node search-problem))
	   
		 ;; Terminate WITHOUT returning a candidate
		 ;; when no more search nodes are queued, 
		 ;; or one of the termination conditions is met.
		 (when (end-search? search-problem 
							cutoff-cost 
							maximum-visited-state-count
							next-node
							(search-problem-count-of-search-nodes-considered
							 search-problem))
	 
		   (return nil))
		 
		 ;; Evaluate the node just dequeued, called NEXT-NODE:

	   (multiple-value-bind (candidate-found? return-value) 
		   (evaluate-next-node next-node 
							   previous-node 
							   search-problem
							   cutoff-cost
							   maximum-visited-state-count
							   full-assignment?
							   disjoint?)
		 
		 (when candidate-found? (return return-value)))))


(defmethod evaluate-next-node :around (next-node
									   previous-node
									   (search-problem search-problem)
									   cutoff-cost
									   maximum-visited-state-count
									   full-assignment?
									   disjoint?)
  "This specialization updates the count of search nodes considered
   statistic. Then, passes control on to next method and uses its
   return value as the return value for this method."
  
  (declare (ignore next-node previous-node cutoff-cost maximum-visited-state-count full-assignment? disjoint?))
  
  ;; Keep statistics on the number of search nodes considered.
  (incf (search-problem-count-of-search-nodes-considered 
		 search-problem))
  
  ;; Now pass control on to the next method.
  (call-next-method))



(defmethod evaluate-next-node (next-node
							   previous-node
							   (search-problem informed-search-problem-w-conflicts)
							   cutoff-cost
							   maximum-visited-state-count
							   full-assignment?
							   disjoint?)
  "Specializes on INFORMED-SEARCH-PROBLEM-W-CONFLICTS. If we're
   performing a branch and bound search, this specialization records
   suboptimality conflicts. If a suboptimality conflict needs to be
   recorded, NEXT-NODE is not a candidate and the search should
   continue. Otherwise, calls the next specialization."
  
  (declare (ignore cutoff-cost maximum-visited-state-count full-assignment?))
  
  ;; Evaluate the node just dequeued, called NEXT-NODE:
  
  (if (and (performing-branch-and-bound? search-problem)
		   (not (no-incumbent? search-problem))
		   (not (better-than-incumbent? next-node search-problem)))
	  (progn
		;; Extract and record a suboptimality conflict.
		(let ((suboptimality-conflict-assignments 
			   
			   (search-problem-node-suboptimality-conflict 
				search-problem next-node)))
		  
		  (insert-new-conflict search-problem
							   :inconsistent-assignment 
							   suboptimality-conflict-assignments))
		
		;; Remove node as current, 
		;; so that it isn't expanded on the next round.
		
		(debug-print-for :bfig t "  Shift off Node ~a."
						 previous-node)
		
		(unrecord-current-search-node! search-problem
									   :disjoint? disjoint?)
		(values nil nil))
	  ;; Else
	  (call-next-method)))

(defmethod evaluate-next-node (next-node
							   previous-node
							   (search-problem search-problem)
							   cutoff-cost
							   maximum-visited-state-count
							   full-assignment?
							   disjoint?)
  "Specializes on general SEARCH-PROBLEM. Checks if NEXT-NODE is a
   candidate."

  (declare (ignore cutoff-cost maximum-visited-state-count previous-node))
  ;; Evaluate the node just dequeued, called NEXT-NODE:
  
  ;; Make NEXT-NODE the current node.  
  ;; This updates the current state to the preferred state that
  ;; extends this node, identifies active variables and conflicts,
  ;; and supports all conflicts resolved by this node.
  
  (debug-print-for :bfig t "  Shift to Node ~a.~2%~a~%"
				   next-node
				   (print-search-node-details next-node nil))
  
  
  (record-search-node-as-current! search-problem next-node
								  :disjoint? disjoint?)
  
  ;; Sets the variables that are unassigned by NEXT-NODE.
  ;; Note: Waited until node was recorded as current, 
  ;; in order to identify all active variables.
  (set-unassigned-variables-of-current-search-node! search-problem)
  
  (cond ((candidate-found? search-problem next-node full-assignment?)
		 
		 ;; Return if the best extension of NEXT-NODE is a candidate,
		 ;; that is, if 1) it resolves all conflicts or 
		 ;; 2) is a full assignment when FULL-ASSIGNMENT? is True, or
		 ;; 3) there is no NEXT-NODE.
		 (let ((count
				(incf (search-problem-candidate-count 
					   search-problem))))
		   
		   ;; All conflicts resolved, return as a candidate.
		   (debug-print-for :bfig t 
							"  Candidate ~a = Node ~a."
;;;									 "  Candidate ~a = Node ~a.~2%~a~%"
							count next-node
;;;									 (print-search-node-details
;;;									  next-node nil)
							)
		   (values T next-node)
										;(return next-node)
		   ))
		
		(:else
		 
		 ;; The termination condition is not met.
		 ;; keep expanding.
		 (values nil nil)
										;T
		 ))  
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Minimal Candidate Generation function
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; The previous minimal candidate generator in BFIG is probably a fake/faulty one.
;; It generates (nearly) minimal candidates.
;; The following functions implements Peng's Master thesis work on minimal relaxation enumeration.
;; Although it is called relaxation candidate, it can be applied to all OPSAT problems 
;; and return minimal candidates and conflicts (the real ones!).

(defmethod next-minimal-relaxation-candidate

    ((search-problem search-problem)
	 ;;search-problem-w-conflicts)
	 cutoff-cost 
     maximum-visited-state-count
     full-assignment?
     disjoint?)
  
  (declare (ignore cutoff-cost))
  (declare (ignore maximum-visited-state-count))
  (declare (ignore full-assignment?))
  (debug-print-for :bfig t "Find next candidate.")



  (if (= (search-problem-candidate-count search-problem) 0)
	  ;; The first candidate: Choose the best one.
	  (let ((variables-to-be-assigned nil)
			(current-minimal-candidate (dequeue! (search-problem-search-queue search-problem)))
			(activated-variables nil))

		(format t "~%Generating Candidate #~A  ~%" (search-problem-candidate-count 
													search-problem))
		(format t "Ancestor Assignment ~A~%" (search-node-ancestor-assignments current-minimal-candidate))
		(format t "Unassigned variables: ~A~2%" (search-node-unassigned-variables current-minimal-candidate))


		(loop for variable in (search-node-unassigned-variables current-minimal-candidate)
		   do (when (state-variable-active? variable)
				(push variable variables-to-be-assigned)
				(push variable activated-variables)))


		(loop 
		   with variable-tba 
		   with best-assignment
			 
		   do (setf variable-tba (pop variables-to-be-assigned))
			 (setf best-assignment (decision-variable-best-assignment variable-tba))

			 (format t "Assigning ~A   ~A~%" variable-tba best-assignment)
			 (push best-assignment (search-node-ancestor-assignments current-minimal-candidate))
			 (setf (state-variable-current-assignment variable-tba) best-assignment)
			 (setf (state-variable-current-cdb-assignment variable-tba) best-assignment)
			 (propagate-current-assignment! search-problem best-assignment)

			 (loop for variable in (search-node-unassigned-variables current-minimal-candidate)
				;;	do  (format t "checking ~A ~A~%" variable (state-variable-active? variable))
				do (when (and (state-variable-active? variable) (not (member variable activated-variables)))
					 (push variable variables-to-be-assigned)
					 (push variable activated-variables)
					 ;;	 (format t "Activating ~a~%" variable)
					 ))

		   until (null variables-to-be-assigned))

		(record-search-node-as-current! search-problem current-minimal-candidate
										:disjoint? disjoint?)
		(set-unassigned-variables-of-current-search-node! search-problem)
		(incf (search-problem-candidate-count search-problem))
		(values current-minimal-candidate))
















	  ;; Following candidate: generated from conflicts

	  (let ((current-conflicts (state-space-conflicts search-problem))
			)

		(format t "~3%Generating Candidate #~A ~%" (search-problem-candidate-count 
												   search-problem))
		;; Prepare for this round
		;; Move current variable assignments to previous
		;; clear the candidate queue
		(loop for variable in (state-space-variables search-problem)
		   do (setf (state-variable-previous-assignment variable) (state-variable-current-assignment variable)))


		(when (not (search-problem-current-search-node-feasible? search-problem))


		  (queue-clear! (search-problem-search-queue search-problem))
		  
		  (format t "Current conflicts ~A~%" current-conflicts)

		  (let ((minimal-relaxations nil)
				(conflict (first current-conflicts)))

			;;	(format t "New conflict: ~A~%" conflict)
			;;	(format t "Conflicting Assignment ~A~%" (conflict-inconsistent-assignments conflict))

			;; Generate constituent relaxations to the newly discovered conflict
			(let ((constituent-relaxations-to-current-conflict (list '())))

			  (loop for assignment in (conflict-inconsistent-assignments conflict)


				 ;;  do (format t "Alternate assignments of ~A are:    ~A~%" assignment (state-assignment-alternate-assignments assignment))
				 ;; Use alternate assignments to the decision variable as constituent kernels
				 do (setf constituent-relaxations-to-current-conflict (append constituent-relaxations-to-current-conflict (state-assignment-alternate-assignments assignment))))

			  ;; Remove the initial () in the list
			  (pop constituent-relaxations-to-current-conflict)

			  (format t "New constituent relaxations  ~A ~%" constituent-relaxations-to-current-conflict)
			  (format t "Previous minimal relaxations  ~A~%" (search-problem-current-minimal-candidates search-problem))
			  ;; Update the candidates we have for the problem using newly generated constituent kernels.
			  (setf minimal-relaxations (incremental-min-set-covering constituent-relaxations-to-current-conflict (search-problem-current-minimal-candidates search-problem) search-problem)))

			(format t "Newly generated minimal relaxations: ~A~%" minimal-relaxations)

			(setf (search-problem-current-minimal-candidates search-problem) minimal-relaxations)

			(loop for minimal-relaxation in minimal-relaxations
			   ;; Enqueue updated minimal candidates, making sure that they have been expanded to their best covered full candidates.
				 

			   do (let ((minimal-candidate (expand-to-candidate minimal-relaxation search-problem))) 
					
					;;	 (format t "Checking if ~A has conflicting assignments~%" (search-node-ancestor-assignments minimal-candidate))

					(when (candidate-not-in-conflict? minimal-candidate)

					  ;;	 (format t " NO!!! ~%")

					  (enqueue! (search-problem-search-queue search-problem) minimal-candidate))))))
		
		;;	(format t "Dequeue the best candidate ~%")

		;; Terminate if there is nothing in the queue
		(when (queue-empty? (search-problem-search-queue search-problem))

		  (return-from next-minimal-relaxation-candidate NIL))

		

		;; Dequeue the best candidate for testing
		(let ((current-minimal-candidate (dequeue! (search-problem-search-queue search-problem))))

		  ;; Dequeue the best minimal candidate for the problem, instantiate it and prepare the problem for consistency check.

		  (record-search-node-as-current! search-problem current-minimal-candidate
										  :disjoint? disjoint?)
		  (set-unassigned-variables-of-current-search-node! search-problem)
		  (incf (Search-problem-candidate-count search-problem))


		  (loop for variable in (state-space-variables search-problem)
			 do (setf (state-variable-current-assignment variable) nil))

		  (loop for assignment in (search-node-ancestor-assignments current-minimal-candidate)

			 do (setf (state-variable-current-assignment (state-assignment-variable assignment)) assignment)
			 ;;	 do (setf (state-variable-previous-assignment (state-assignment-variable assignment)) assignment)
			   )

		  (format t "Returning ~A~%" (search-node-ancestor-assignments current-minimal-candidate))


		  (values current-minimal-candidate))))
  )


(defmethod expand-to-candidate 
	(minimal-relaxation search-problem)

  "Create a search node based on current minimal relaxations."


  ;; Initialize the candidate with some basic parameters: node type, search-problem, unassigned variables, 
  (let ((current-minimal-candidate (make-instance (search-problem-node-type search-problem)
												  :search-problem search-problem
												  :unassigned-variables (state-space-variables search-problem)
												  :unassigned-ancestor-variables (state-space-variables search-problem)
												  :source-minimal-relaxations (copy-list minimal-relaxation)
											;;	  :ancestor-assignments minimal-relaxation
												  )))

	;; Set the current assignments of the node as the minimal relaxation candidate.

	;;(record-search-node-as-current! search-problem current-minimal-candidate)
	;;(set-unassigned-variables-of-current-search-node! search-problem)


	;; Going through all unassigned variables, picking out those are activated but are unassigned.

;;	(format t "~%Expanding ~A~%" minimal-relaxation)

	;; make all implicit assignments explicit

	;; Initialize the activation of variables
	(loop for variable in (state-space-variables search-problem)

	   do (with-slots (active?) variable (setf active? NIL))
		 (when (state-variable-unguarded? variable)
		   (with-slots (active?) variable (setf active? T)))
		 
	 ;;  do (format t " ~A is activated?  ~A~%" variable (state-variable-active? variable))
		 )

	(setf minimal-relaxation (make-implicit-assignments-explicit search-problem minimal-relaxation))
;;	(format t "Initial Assignments:  ~A~%" minimal-relaxation)
	(setf (search-node-ancestor-assignments current-minimal-candidate) minimal-relaxation)





	;; Propagate the activation of minimal candidates
	(loop for assignment in minimal-relaxation
		 do (setf (state-variable-current-assignment (state-assignment-variable assignment)) assignment)
		 (setf (state-variable-current-cdb-assignment (state-assignment-variable assignment)) assignment)
		 (propagate-current-assignment! search-problem assignment))


	(let ((variables-to-be-assigned nil)
		  (activated-variables nil)
		  (unassigned-variables nil))

	  ;; Locate unassigned variables
	  (loop for variable in (state-space-variables search-problem)

		 do (let ((assigned? nil))

			  (loop for assignment in minimal-relaxation
				 do (when (equal (state-assignment-variable assignment) variable)

					  (setf assigned? T)
					  (push variable activated-variables)))

			  (when (equal assigned? nil)
				(push variable unassigned-variables))))


	  (setf (search-node-unassigned-variables current-minimal-candidate) unassigned-variables)
	  (setf (search-node-unassigned-ancestor-variables current-minimal-candidate) unassigned-variables)

	;;  (format t "Unassigned variables ~A~%" (search-node-unassigned-variables current-minimal-candidate))



	  ; Initialize variables to be assigned

	  (loop for variable in (search-node-unassigned-variables current-minimal-candidate)
		 do (when (state-variable-active? variable)
			  ;; Keeps two lists here are necessary. 
			  ;; One for variables that needs to be assigned, which will finally got cleared.
			  ;; One for variables that are activated, which keeps increasing since newly made assignments may activate new variables.
			  ;;	(format t "Activating ~A~%" variable)
			  (push variable variables-to-be-assigned)
			  (push variable activated-variables)))

	 ;; (format t "Starting activated variables: ~A~%" activated-variables)

	  ;; Loop with the list of to be assigned variables. Stop until it becomes empty.
	  (loop 
		 with variable-tba 
		 with best-assignment
		   
		 ;; Stop when nothing more needs to be assigned.
		 until (null variables-to-be-assigned)

		 ;; Pop one variable that needs assignment from the list.
		 do (setf variable-tba (pop variables-to-be-assigned))

		 ;; Make its best possible assignment.
		   (setf best-assignment (decision-variable-best-assignment variable-tba))
	;;	   (format t "Assigning ~A~%" best-assignment)
		 ;;(format t "Assigning ~A   ~A~%" variable-tba best-assignment)

		 ;; Push it to assigned variables list.
		   (push best-assignment (search-node-ancestor-assignments current-minimal-candidate))
		   
		 ;; Record the current assignment for the variable
		   (setf (state-variable-current-assignment variable-tba) best-assignment)
		   (setf (state-variable-current-cdb-assignment variable-tba) best-assignment)

		 ;; Propagate the activations caused by this assignment.
		   (propagate-current-assignment! search-problem best-assignment)


		 ;; Add any newly activated variables to the variables-to-be-assigned list. 
		   (loop for variable in (search-node-unassigned-variables current-minimal-candidate)
			  ;;	do  (format t "checking ~A ~A~%" variable (state-variable-active? variable))

			  ;; Make sure that the variable is NEWLY activated.
			  do (when (and (state-variable-active? variable) (not (member variable activated-variables)))
				   (push variable variables-to-be-assigned)
				   (push variable activated-variables)
				   ;;	 (format t "Activating ~a~%" variable)
				   ))


		 ))

;;	(format t "Result ~A~%" (search-node-ancestor-assignments current-minimal-candidate))
;;	(setf (search-node-current-assignments current-minimal-candidate) (search-node-ancestor-assignments current-minimal-candidate))

;;	(format t "Expand  ~A  to ~A~%" minimal-relaxation (search-node-ancestor-assignments current-minimal-candidate))

	(setf (search-node-best-cost current-minimal-candidate)
		  (compute-search-node-best-cost search-problem current-minimal-candidate))

	(values current-minimal-candidate)

))


(defmethod incremental-min-set-covering (new-constituent-relaxations previous-minimal-relaxations search-problem)

  "Update minimal relaxations based on the constituent relaxations to known conflicts"

  

  (let ((minimal-relaxations (list '()))
		(not-superset nil))

	(pop minimal-relaxations)

	(loop for constituent-relaxation in new-constituent-relaxations
		 
		 ;; If there is no previous relaxations, then all current constituent relaxations become the minimal candidate.
	   do (if (null previous-minimal-relaxations)

			  (push (list constituent-relaxation) minimal-relaxations)
			  
			  ;; Else, make a cross product between existing candidates and new constituent relaxations. 
			  
			  (progn

				(loop for previous-minimal-relaxation in previous-minimal-relaxations

				   do (let ((new-minimal-relaxation (add-relaxation previous-minimal-relaxation constituent-relaxation)))

						;; Keep if the candidate is not the superset of any other candidates, and does not have internal conflicts between its conflicts.
						
						(setf new-minimal-relaxation (remove-duplicates new-minimal-relaxation :test #'equal))

						(setf not-superset
							(not-superset? new-minimal-relaxation minimal-relaxations))


						(when (and (first not-superset) (not (null new-minimal-relaxation)))
						 ;; (format t "Updating ~A~%" (second not-superset))
						  (setf minimal-relaxations (second not-superset))
						;;  (format t "Add ~A to ~A   to    " new-minimal-relaxation minimal-relaxations)
						  
						  (when (first (not-superset? new-minimal-relaxation
											   (search-problem-feasible-minimal-relaxations 
												search-problem)))

							(push new-minimal-relaxation minimal-relaxations)
							   

							   )

						;;  (format t "~A~%" minimal-relaxations)
						  ))))
			  ))
	
	



	(values minimal-relaxations)))



(defun not-superset? (new-relaxation minimal-relaxations)

  "Check if new-relaxation is the superset of any relaxations in 
the relaxation-set"

  (when (not (null minimal-relaxations))

	(let ((relaxation-set-copy (copy-list minimal-relaxations))
		  (relaxation nil))



	  (loop while (not (null relaxation-set-copy))

		 do (setf relaxation (pop relaxation-set-copy))

	;;	 do	(format t "is  ~A  the superset of ~A   " new-relaxation relaxation)

		 do (if (is-superset? new-relaxation relaxation)
				;; If new-relaxation is the superset of any relaxation, then not-superset? should return nil.
				(progn
			;;	  (format t "Yes  ~%")
				  (return-from not-superset? (list NIL minimal-relaxations)))

				;; If new-relaxation is the subset of any relaxation, then remove the superset relaxation from relaxation-ser/
				(progn

			;;	  (format t "No  ~%")
				  (when (is-superset? relaxation new-relaxation)
					
			;;		(format t "deleting ~A from ~A   to   " relaxation minimal-relaxations)
					(setf minimal-relaxations (remove relaxation minimal-relaxations :test #'equal))
			;;		(format t "~A~%" minimal-relaxations)
					;;(delete relaxation relaxation-set)
					;;	  (push relaxation relaxations-to-be-removed)
					))))))


 ;; (format t "DC  ~A~%" minimal-relaxations)


  (values (list T minimal-relaxations)))


(defun is-superset? (relaxation1 relaxation2)


  "Check if relaxation1 is the superset of relaxation2"

  ;; Going through all the assignments, making sure that all assignments in relaxation2 appear in relaxation1.


  (loop for assignment in relaxation2

	   

	 do (when (not (member assignment relaxation1 :test #'equal))
			;; If relaxation1 does not have assignment, then it won't be the superset of relaxation2.

		 ;; (format t "~A  is not a member of ~A~%" assignment relaxation1)

		  (return-from is-superset? NIL)))

  (values T))




(defmethod add-relaxation (previous-relaxations new-relaxation)


  "Add new constituent relaxation to existing relaxations"

  ;; If the new relaxation is contained in previous-relaxations, then return the previous one.
  (when (member new-relaxation previous-relaxations)
	(return-from add-relaxation previous-relaxations))

  (loop for relaxation in previous-relaxations
	 ;; if the new relaxation is in conflict with previous relaxations, then do not add.
	 do (when (in-conflict? relaxation new-relaxation)
		  (return-from add-relaxation nil)))


  ;; Attach new-relaxation to previous-relaxations
  (values (append previous-relaxations (list new-relaxation))))


(defun candidate-not-in-conflict? (minimal-candidate)


  "Checks if there are conflicting assignment in the minimal-candidate"



  (loop for assignment1 in (search-node-ancestor-assignments minimal-candidate)

	 do (loop for assignment2 in (search-node-ancestor-assignments minimal-candidate)

		   do (when (not (equal assignment1 assignment2))

			;;	(format t " ~A and ~A are in conflict ~% " assignment1 assignment2)

				;; If assignment1 and assignment2 are in conflict
				(when (in-conflict? assignment1 assignment2)
				  
				  (return-from candidate-not-in-conflict? nil))


				;; If assignment1 has guard, check if its guard is in conflict with assignment2
				(when (null (state-variable-unguarded? (state-assignment-variable assignment1)))
				  (when (in-conflict? (state-variable-guard 
									   (state-assignment-variable assignment1))
									  assignment2)

					(return-from candidate-not-in-conflict? nil)))

				;; If assignment2 has guard, check if its guard is in conflict with assignment1
				(when (null (state-variable-unguarded? (state-assignment-variable assignment2)))
				  (when (in-conflict? (state-variable-guard 
									   (state-assignment-variable assignment2))
									  assignment1)

					(return-from candidate-not-in-conflict? nil)))

				;; If both assignments have guard, check if their guards are consistent
				(when (and (null (state-variable-unguarded? (state-assignment-variable assignment1)))
						   (null (state-variable-unguarded? (state-assignment-variable assignment2)))) 


				  (when (in-conflict? (state-variable-guard 
									   (state-assignment-variable assignment1))
									  (state-variable-guard 
									   (state-assignment-variable assignment2)))

					(return-from candidate-not-in-conflict? nil)))
			;;	(format t " NO!!! ~%")

				)

			 

			 ))


  (values T))



(defun in-conflict? (relaxation1 relaxation2)


  "Check if two relaxations are in conflict with each other."

  (when (or  (not (equal (state-assignment-variable relaxation1) 
						 (state-assignment-variable relaxation2)))
			 (equal relaxation1 relaxation2))

	(return-from in-conflict? NIL))

  (values T))


(defmethod search-node-preferred-expansion 

	((search-node search-node))

  "Returns the preferred state of SEARCH-NODE, which consists
   of its ancestor-assignments, augmented by the
   child-assignment that is preferred to be expanded first.

   Note: children are currently listed, left to right,
   from most preferred to least preferred.

   IMPORTANT: The preferred expansion is not a new list.
   Hence before destructively modifying preferred-expansion, 
   for example using sort, make sure to COPY the list first.

   Note: If search-node splits on an unresolved conflict, then 
   the children supplied must be the constituent-kernels of that 
   conflict."

  ;; Copy so that a destructive change to the expansion won't
  ;; disturb the search tree.
  (copy-list (search-node-preferred-expansion-internal search-node)))


;;; Changed.
;;; o Same as in initialize-candidate-generator!
;;;   (applicable for search-problem)
;;; -- gustavop

(defmethod candidate-is-solution!

    ((new-solution search-node)
	 (search-problem search-problem))
  ;; search-problem-w-conflicts))
  
  "Informs the candidate generator that search-node NEW-SOLUTION
   is a solution to SEARCH-PROBLEM."

  ;; By default, new-solution is ignored.
  T)

(defmethod candidate-is-solution!

    ((new-solution informed-search-node)
	 (search-problem informed-search-problem-w-conflicts))
  
  "Informs the candidate generator that search-node NEW-SOLUTION
   is a solution to SEARCH-PROBLEM."

  ;; Update incumbent when performing Branch and Bound search.
  (when (performing-branch-and-bound? search-problem)
	
	(with-slots (incumbent-node) search-problem
	  
	  (when (or (eql incumbent-node :none)
				(better-than-incumbent? new-solution search-problem))
		
		(setq incumbent-node new-solution)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Supporting functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Initializing Search
;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed: To Do:
;;; o Generalize from search-problem-w-conflicts to
;;;   search-problem.

;;; Optimization:
;;; o Unassign only the assignments of search node.

;;; Check whether clear-current-assignments is applicable to
;;; state-space without conflicts. This was not the case in
;;; my version. 10/2/08
;;; -- gustavop
(defmethod clear-current-search-node! 

	((search-problem search-problem))
  
  "Clear the record of any search node as current 
   for SEARCH-PROBLEM and the record of any conflicts 
   it resolves."
  
  ;; Clear the node under current consideration, including...
  (setf (search-problem-current-search-node search-problem) nil)
  
  (clear-current-assignments! search-problem))

;;; Changed:
;;; o New method.
(defmethod setup-variable-search-order!

	((search-problem search-problem))
  
  "Sets up selection of which variable to split on."

  T)

;;; Changed.
;;; o Same as in initialize-candidate-generator!
;;;   (applicable to informed-search-problem)
;;; -- gustavop

(defmethod clear-problem-incumbent! 
	
	((search-problem informed-search-problem))
  ;; informed-search-problem-w-conflicts))

  "Sets up the incumbent search node of SEARCH-PROBLEM.
   The incumbent-node is the best node found to date that
   denotes a solution.  Incumbent begins as :none."
  
  (with-slots (incumbent-node) search-problem
	(setq incumbent-node :none)))

;;; To Do:
;;; o Make *search-node-counter* an attribute of search-problem.
(defmethod clear-search-statistics! 
	
	((search-problem search-problem))
  
  "Clears the search statistics maintained for SEARCH-PROBLEM."

  ;; Reset the count of search tree nodes created...
  ;;  (setq *search-node-counter* 0)
  (setf (search-problem-count-of-search-nodes-created
		 search-problem)
    0)
    
  ;; expanded, ...
  (setf (search-problem-count-of-search-nodes-considered 
		 search-problem) 
    0)
  
  ;; and candidates found.
  (setf (search-problem-candidate-count search-problem) 0)
  T)

;;; Changed:
;;; o Debug-print statements.
;;; o Generalized to search-problem.
(defmethod initialize-problem-search-queue! 
	
	((search-problem search-problem
;;; Fix delete when works					 
;;;					 search-problem-w-conflicts
					 )
	 search-order)

  "Sets up a search-queue for SEARCH-PROBLEM that 
   enumerates solutions based on SEARCH-ORDER.
   Queue starts with a (root) search-node, 
   denoting all SEARCH-PROBLEM states.
 
   SEARCH-ORDER is is one of the keywords in 
   *ALLOWED-SEARCH-ORDER-TYPES* (for example 
   :BEST-FIRST), or :DEFAULT.  If :DEFAULT, then
   search-order is the last one specified."

  ;; Determine and record the current search order
  ;; and put in place the corresponding search-queue.
  (multiple-value-bind (new-search-order changed?)
	  (record-new-search-order
	   search-problem search-order)
	
	(when (or changed?
			  (no-search-queue? search-problem))
	  
	  (add-search-queue! search-problem new-search-order)))

  ;; and its corresponding search queue.
  ;; Clear...
  (clear-search-queue! search-problem)

  ;; and queue a new search tree root.
  (let ((root-node
		 (new-root-search-node search-problem)))

	(debug-print-for :bfig t 
					 "  Queue root ~a." 
;;;					 "  Queue root ~a.~2%~a~%" 
					 root-node
;;;					 (print-search-node-details root-node nil)
					 )

	(enqueue-search-node! search-problem 
						  root-node))

  T)

;;; Changed:
;;; o Generalized to search-problem.
(defmethod new-root-search-node

	((search-problem search-problem
;;; Fix delete when works					 
;;;					 search-problem-w-conflicts
					 ))

  "Creates and returns a search-node that denotes the 
     root of a search tree for SEARCH-PROBLEM."
  
  ;; Note: For search-nodes with slot :split-on-conflict?
  ;;       This exploits the default init of nil.
	
  (make-instance (search-problem-node-type search-problem)
	:search-problem search-problem
	:unassigned-ancestor-variables 
	(state-space-variables search-problem)
	:unassigned-variables 
	(state-space-variables search-problem)
;;;	:split-on-conflict? nil
	:search-problem search-problem))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Maintaining the Search Order
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod record-new-search-order

	((search-problem search-problem) 
	 search-order)

  "Update the SEARCH-ORDER of SEARCH-PROBLEM. 
   Returns (VALUES NEW-SEARCH-ORDER CHANGED?).

   SEARCH-ORDER is one of the keywords in 
   ALLOWED-SEARCH-ORDER of SEARCH-PROBLEM 
   (for example :DEPTH-FIRST), or :DEFAULT.  
   If :DEFAULT, then NEW-SEARCH-ORDER is the same as
   previously specified for SEARCH-PROBLEM, otherwise
   its SEARCH-ORDER.

   CHANGED? is nil IFF NEW-SEARCH-ORDER is the
   same as the previously recorded search-order for 
   SEARCH-PROBLEM."

  (cond ((eql search-order :DEFAULT)

		 ;; :DEFAULT search-order to the previous
		 ;; search-order specified for SEARCH-PROBLEM.
		 (values (search-problem-search-order search-problem)
				 nil))
		  
		((allowed-search-order? search-order search-problem)

		 ;; Specified a valid search order, record it.
		 (let ((old-order 
				(search-problem-search-order search-problem)))
		   
		   (setf (search-problem-search-order search-problem) 
			 search-order)

		   (values search-order
				   (not (eql search-order old-order)))))
		  
		(:otherwise
		 
		 (with-slots (allowed-search-orders) search-problem
		   (error "Search order ~a not allowed for ~a.  Must be one of ~a"
				  search-order search-problem allowed-search-orders)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Maintaining the Search Queue
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Generalized to search-problem.
;;; done. --gustavop
(defmethod no-search-queue? 
	
	((search-problem 
	  search-problem
	  ;; search-problem-w-conflicts
	  ))

  "Returns true IFF SEARCH-PROBLEM does not have a 
   search-queue in place."

  (eql (search-problem-search-queue search-problem) :none))

;;; Check:
;;; o Test this function for
;;;   search-order = :depth-first and :breadth-first.

;;; To Do:
;;; o Confirm that search-order has already been checked.

;;; Changed:
;;; o Generalized to search-problem.
(defmethod add-search-queue! 
	
	((search-problem search-problem
;;; Fix delete when works					 
;;;					 search-problem-w-conflicts
					 )
	 search-order)
  
  "Associates with SEARCH-PROBLEM a search-queue that corresponds
   to SEARCH-ORDER.
   This functions presumes that SEARCH-ORDER has already been
   checked and determined suitable to the type of SEARCH-PROBLEM."
  
  (setf (search-problem-search-queue search-problem)

	(case search-order

	  (:depth-first
	   (make-smart-list :fifo-queue? nil))
	  
	  (:breadth-first
	   (make-smart-list :fifo-queue? t))
	  
	  (:best-first
	   ;; Keyword only supplied for informed search problem.
	   (make-ordered-list-queue :order #'(lambda (x y) (best-first-queue-tiebreaking 
														x y #'search-node-best-cost 
														(decision-space-cost-order-function
														 search-problem)))))
	  
	  (otherwise
	   (error "Undefined search-order ~a." search-order)))))

(defmethod best-first-queue-tiebreaking ((node1 search-node) (node2 search-node) 
										 primary-key primary-ordering-fn)

  (let ((val1 (apply primary-key (list node1)))
		(val2 (apply primary-key (list node2))))
	(cond ((apply primary-ordering-fn (list val1 val2))
		   
		   t)
		  
		  ((eql val1 val2) 
		   
		   (>= (length (search-node-ancestor-assignments node1)) 
			   (length (search-node-ancestor-assignments node2))))
		  
		  (t 
		   
		   nil)))
  )

;;; Changed:
;;; o Generalized to search-problem.
(defmethod clear-search-queue! 
	
	((search-problem search-problem
;;; Fix delete when works					 
;;;					 search-problem-w-conflicts
					 
					 ))

  "Side Effect:
   Initializes the search-queue of SEARCH-PROBLEM to be empty."

  (with-slots (search-queue) search-problem
	(queue-clear! search-queue)))

;;; Changed:
;;; o Generalized to search-problem.
(defmethod enqueue-search-node! 
	
    ((search-problem search-problem
;;; Fix delete when works					 
;;;					 search-problem-w-conflicts
					 ) 
	 (search-node search-node))
  
  "Side Effect:
   Adds SEARCH-NODE to the search-queue of SEARCH-PROBLEM.
   Assumes that SEARCH-NODE specifies its cost."
  
  (with-slots (search-queue) search-problem
	(enqueue! search-queue search-node)))

;;; Changed:
;;; o Generalized to search-problem.
(defmethod dequeue-search-node 

    ((search-problem search-problem
;;; Fix delete when works					 
;;;					 search-problem-w-conflicts
					 ))
  
  "Returns the next search-node on search-queue of 
   SEARCH-PROBLEM to be expanded.

  Side Effect:  Removes the returned search-node from 
  search-queue of SEARCH-PROBLEM."

  (with-slots (search-queue) search-problem
	(dequeue! search-queue)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Maintaining Unassigned Variables of a Search Node
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Changed:
;;; o Moved here from bfig-defs.lisp
;;; o Renamed from:
;;;      Maintaining a Search-Nodes' List of Unassigned Variables
;;; o Reordered the following set of functions.

;;; Changed:
;;; - Made a method.
;;; - updated comment string.
;;; - Made a method, added state- to name.
(defmethod all-state-variables-assigned? 

	((search-node search-node))

  "Returns True iff SEARCH-NODE contains no unassigned 
   variables.

   Assumes that for the state of SEARCH-NODE, all variables 
   that are assigned or inactive have been removed from the 
   prefix of UNASSIGNED-VARIABLES.  This can be performed by 
   recording SEARCH-NODE as current and calling
   SET-UNASSIGNED-VARIABLES-OF-CURRENT-SEARCH-NODE!."
  
  (null (search-node-unassigned-variables search-node)))

;;; Changed:
;;; - Made a method, added state- to name.
(defmethod state-variable-unassigned? 

    ((variable state-variable) assignments)
  
  "Returns true if VARIABLE is not assigned by 
   ASSIGNMENTS."
  
  (loop for a in assignments
      for v = (state-assignment-variable a)
      always (not (eql v variable))))

;;; Changed:
;;; - New function, derived from update-unassigned-variables.

(defmethod set-unassigned-variables-of-current-search-node!

	((search-problem search-problem))
  
  "Sets the unassigned variables of the current search node
   of SEARCH-PROBLEM.

   Assumes that the current-search-node is set
   for SEARCH-PROBLEM, using RECORD-SEARCH-NODE-AS-CURRENT!,
   This is required in order to identify which variables
   active, based on current node's assignment.
  
   The UNASSIGNED-VARIABLES of current search-node is a list
   reduced from the UNASSIGNED-ANCESTOR-VARIABLES of current 
   search node, that is, current node's parent's unassigned-variables, 
   with  any leading variables trimmed that are assigned or 
   inactive in the PREFERRED-ASSIGNMENTS of the current search 
   node.

   Guarantees that the first variable on the list of 
   UNASSIGNED-VARIABLES is unassigned and active.

   Tries to order variables from most to least preferred.

   Note that this function will trim leading unassigned variables 
   that can never become active.  If these variables were inactive
   at the time this function was called, then they can never
   become active, since the inactive variable's guard variable 
   appeared earlier in the list, and hence was either assigned a 
   value that doesn't satisfy the guard of the inactive variable, or 
   that guard variable itself can never become active."

  ;; Note: New-assigment is null when expanding the root node,
  ;;       since it initially has no children.
  
  (with-slots (current-search-node) search-problem
	(when (null current-search-node)
	  (error "No current-search-node for ~A, shouldn't happen." 
			 search-problem))
	
	(with-slots (unassigned-ancestor-variables
				 unassigned-variables)
		current-search-node

	  (let ((preferred-child-assignment
			 (search-node-preferred-child-assignment 
			  current-search-node)
			 ))

		(cond ((and (not (null preferred-child-assignment))
					(eql (state-assignment-variable preferred-child-assignment)
						 (next-unassigned-variable search-problem
												   unassigned-ancestor-variables))
					)
			   
			   ;; If the current search node has a best child assignment
			   ;; and it assigns the next unassigned variable of its parent,
		 
			   ;; Then create current search node's unassigned variables by 
			   ;; trimming forward to the next active unassigned variable,
			   (let ((current-assignments
					  (search-node-preferred-expansion-internal
					   current-search-node)))
				 
				 (setq unassigned-variables
				   (trim-leading-assigned-variables 
					search-problem
					(remaining-unassigned-variables search-problem
													unassigned-ancestor-variables)
					current-assignments))

				;; (format t "      Unassigned Ancestor Variables: ~A~%" unassigned-ancestor-variables)
				;; (format t "      Remaining Unassigned Variables: ~A~%" (remaining-unassigned-variables search-problem unassigned-ancestor-variables))
				;; (format t "      Unassigned Variables (EXCL Child Assmt): ~A~%" unassigned-variables)
				 ))
		
			  (:else
			   
			   ;; Otherwise, current search node's unassigned variables are
			   ;; inherited as the unassigned variables of its parent node.
			   (setq unassigned-variables
				 unassigned-ancestor-variables)))))))

;;; Changed:
;;; o Added search-problem to arguments of
;;;   call to next-unassigned-variable.
;;; o Added search-problem to arguments of
;;;   call to trim-leading-assigned-variables.
;;; o Commented out update-unassigned-variables,
;;;   replacing it with SET-UNASSIGNED-VARIABLES-OF-CURRENT-SEARCH-NODE!.

;;; To Do:
;;; o Delete when works.
;;;(defmethod update-unassigned-variables 
;;;
;;;    ((search-problem search-problem)
;;;	 unassigned-variables new-assignment augmented-assignments)
;;;  
;;;  "Update the list of UNASSIGNED-VARIABLES of an assignment, 
;;;   given that NEW-ASSIGNMENT has been added to produce
;;;   AUGMENTED-ASSIGNMENTS.
;;;
;;;   Returns UNASSIGNED-VARIABLES with any leading variables
;;;   trimmed that are assigned in AUGMENTED-ASSIGNMENTS.
;;;
;;;   NEW-ASSIGNMENT is a state-assignment object, or nil
;;;     if no assignment has been made.
;;;
;;;   AUGMENTED-ASSIGNMENTS is a list of state-assignment objects
;;;   and must include NEW-ASSIGNMENT.
;;;
;;;   UNASSIGNED-VARIABLES is a list of decision-variable objects.
;;;     The first variable of UNASSIGNED-VARIABLES can 
;;;     appear in NEW-ASSIGNMENT, but not in any other
;;;     assignment of AUGMENTED-ASSIGNMENTS."
;;;
;;;  ;; Note: New-assigment is null when expanding the root node,
;;;  ;;       since it initially has no children.
;;;  (cond ((and (not (null new-assignment))
;;;			  (eql (state-assignment-variable new-assignment)
;;;				   (next-unassigned-variable search-problem
;;;											 unassigned-variables)))
;;;		 ;; When new assignment eliminates the 
;;;		 ;; next unassigned variable, 
;;;		 ;; trim forward to the next unassigned variable.
;;;		 (trim-leading-assigned-variables 
;;;		  search-problem
;;;		  (remaining-unassigned-variables search-problem
;;;										  unassigned-variables)
;;;		  augmented-assignments))
;;;		(:else
;;;		 unassigned-variables)))

;;; Changed:
;;; o Added search-problem to arguments of
;;;   next-unassigned-variable and made a method.
;;; o Updated documentation.
(defmethod next-unassigned-variable 

	((search-problem search-problem) unassigned-variables)
  
  
  "Given UNASSIGNED-VARIABLES, containing all variables of the current 
   search node of SEARCH-PROBLEM that have not been assigned,

   Returns the next active variable in UNASSIGNED-VARIABLES to be assigned.

   Assumes that SEARCH-PROBLEM is focussed on a current search-node
   and that all variables assigned or inactive in this current search-node
   have been removed from the prefix of UNASSIGNED-VARIABLES.
   This can be achieved by recording the search node as current and 
   by calling SET-UNASSIGNED-VARIABLES-OF-CURRENT-SEARCH-NODE!."

  ;;   Note that UNASSIGNED-VARIABLES may include one or more variables
  ;;   that is assigned or inactive in the current node; 
  ;;   however, its first variable is guaranteed to be active an unassigned.

  (first unassigned-variables))

;;; Changed:
;;; o Added search-problem to arguments of
;;;   remaining-unassigned-variables and made a method.
;;; o Updated documentation string.
(defmethod remaining-unassigned-variables 

	((search-problem search-problem) unassigned-variables)
  
  
  "Given the UNASSIGNED-VARIABLES of the current search node of
   SEARCH-PROBLEM, and given that we have split on the next
   unassigned variable in UNASSIGNED-VARIABLES,

   Returns UNASSIGNED-VARIABLES with this next variable removed.

   Assumes that all variables assigned or inactive in the state of
   the current search node of SEARCH-PROBLEM have been removed 
   from the prefix of UNASSIGNED-VARIABLES.  This can be achieved
   by recording this search node as current and by calling 
   SET-UNASSIGNED-VARIABLES-OF-CURRENT-SEARCH-NODE!. "

  unassigned-variables)

;;; Changed:
;;; o Added search-problem to arguments of
;;;   trim-leading-assigned-variables, and made a method.
;;; o Added the condition that state-variable must
;;;   be active.
;;; o Updated documentation.
(defmethod trim-leading-assigned-variables 

    ((search-problem search-problem) 
	 unassigned-variables assignments &key (uvs (list '())))

  "Given UNASSIGNED-VARIABLES and ASSIGNMENTS of the current
   search node of SEARCH-PROBLEM,

   Returns UNASSIGNED-VARIABLES with all leading variables 
   removed that are inactive or assigned for ASSIGNMENTS.

   Assumes that search-node is recorded as current when called.
   "

;;;  (loop for uvs on unassigned-variables
;;;      for variable = (first uvs)
;;;					 
;;;      until (or (null uvs)
;;;				;; Stops when variables exhausted,
;;;				
;;;				;; Or when an active variable is
;;;				;; found that isn't assigned.
;;;				(and (state-variable-active? variable)
;;;					 (state-variable-unassigned? 
;;;					  variable
;;;					  assignments)))
;;;			
;;;   finally (return uvs))

;;;  (setf uvs (list '()))

  (declare (ignore unassigned-variables))

  ;;Modified by Peng. Should use all variables instead of unassigned variables here.
  (loop for variable in (state-space-variables search-problem) ;;(butlast unassigned-variables)
					 
       do (when (and (state-variable-active? variable)
                   (state-variable-unassigned? 
                    variable
                    assignments))
          (push variable uvs)
		  ;;(format t "      Adding: ~A~%" variable)
		  )
			
      finally (return (reverse (butlast uvs))))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Compute Bound on Search Tree Node Cost
;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Determine search node best cost for a conditional CSP.
;;;   The cost of inactive variables is zero.  Can we prove that
;;;   some variables must be active under all assignments?

;;; o Provide a tighter bound, for example, using mini-buckets.
;;; Note: mini-buckets should be applicable just when conflicts are not
;;; used during search (because a static variable ordering is guaranteed).
;;; --gustavop

;;; Changed
;;;   o Generalized to informed-search-problem.
;;; --gustavop
(defmethod compute-search-node-best-cost
	
    ((search-problem informed-search-problem)
	 ;; informed-search-problem-w-conflicts) 
	 (search-node informed-search-node))
  
  " Returns an admissible bound on the best cost of all states 
    represented by search-node: a least upper bound, if 
    maximizing, and a greatest lower bound if minimizing.  

    To compute total cost, the attribute cost of each assignment
    is computed using DECISION-ASSIGNMENT-COST.  
    Attributes are combined using SEARCH-PROBLEM-BINARY-OBJECTIVE-FUNCTION,
    and cost is offset using DECISION-SPACE-COST-OFFSET."

  ;; Determine the best cost state that extends the partial assignment
  ;; of search-node, and record it on the cg-best-extension of each variable.
  (record-search-node-best-extension! search-problem search-node)


  (let ((assignments (search-node-ancestor-assignments search-node)))


	(loop 
	   with objective-function = (decision-space-binary-objective-function
								  search-problem)
	   with total-cost = (decision-space-cost-offset search-problem)
	   for variable in (state-space-variables search-problem)
	   for variable-cost = (decision-assignment-cost
							(decision-variable-cg-best-extension variable))
		 
		do (when (not (state-variable-deactivated-by-assignments? variable assignments))


		   do (setq total-cost
					(funcall objective-function variable-cost total-cost)))
		 
	   finally (return total-cost))


	))

(defmethod state-variable-deactivated-by-assignments?

	((variable state-variable) assignments)

  "Determine if a variable is deactivated by a list of assignments"


nil)

(defmethod state-variable-deactivated-by-assignments?

	((variable conditional-state-variable) assignments)

  "Determine if a variable is deactivated by a list of assignments"


  (let ((gassignment (state-variable-guard variable)))	

	(when (equal gassignment *always-active*)
	  (return-from state-variable-deactivated-by-assignments? nil)
	  )

	(let ((gvariable (state-assignment-variable gassignment)))

	(loop for assignment in assignments
		 
	   do (when (equal gvariable (state-assignment-variable assignment))

			(when (not (equal (state-assignment-value gassignment)
							  (state-assignment-value assignment)))

			(return-from state-variable-deactivated-by-assignments? t)

			)

		 ))))


nil)


;;; To Do:
;;; o Compute the best cost extension for a conditional CSP.
;;;   Its not clear if this can be done without search.

;;; Changed
;;; o Generalized to informed-search-problem.
;;; -- gustavop
(defmethod record-search-node-best-extension!
	
    ((search-problem informed-search-problem)
	 ;; informed-search-problem-w-conflicts) 
	 (search-node informed-search-node))

  "Sets the cg-best-extension of the decision variables of SEARCH-PROBLEM 
   to the best cost extension of the partial assignment of SEARCH-NODE."
  
  ;; This function generates the best cost extension of the
  ;; partial assignment specified by SEARCH-NODE, by assigning
  ;; every unassigned variable its best assignment.
  
  ;; We setup this best state on the test assigments of problem's
  ;; variables.
  
  ;; First we set the cg-best-extension of every variable to its 
  ;; best assignment.

  (loop for variable in (state-space-variables search-problem)

      do (setf (decision-variable-cg-best-extension variable)
		   (decision-variable-best-assignment variable)))
  
  ;; We then overwrite the test assignment of every assigned variables,
  ;; as specified in each ancestor-assignment plus ...
  (loop for new-assignment in
		(search-node-ancestor-assignments search-node)
		
      for variable = (state-assignment-variable new-assignment)
					 
      do (setf (decision-variable-cg-best-extension variable) 
		   new-assignment))

  ;; ... the best child assignment.
  (let ((new-assignment
		 (search-node-preferred-child-assignment search-node)))
	
    (unless (null new-assignment)
	  
      (setf (decision-variable-cg-best-extension
			 (state-assignment-variable new-assignment))
		new-assignment)))
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Computing the Best Partial Assignment of a Search Tree Node
;;;
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod search-node-preferred-expansion-internal
	
	((search-node search-node))

  "Returns the preferred state of SEARCH-NODE, which consists
   of its ancestor-assignments, augmented by the
   child-assignment that is preferred to be expanded first.

   Note: children are currently listed, left to right,
   from most preferred to least preferred.

   IMPORTANT: The preferred expansion is not a new list.
   Hence before destructively modifying preferred-expansion, 
   for example using sort, make sure to COPY the list first.

   Note: If search-node splits on an unresolved conflict, then 
   the children supplied must be the constituent-kernels of that 
   conflict."

  (let ((preferred-child-assignment
		 (search-node-preferred-child-assignment search-node))

        (ancestor-assignments
		 (search-node-ancestor-assignments search-node)))
	
    (preferred-expansion ancestor-assignments 
						 preferred-child-assignment)))

;;; Fix:
;;; o We are not computing the cost of the constituent
;;;   kernels correctly in children-best-assignment.  
;;;   See the JDAM paper.

(defun preferred-expansion 

	(ancestor-assignments preferred-child-assignment)

  "Returns ancestor-assignments augmented by the
   child-assignment that is preferred to be expanded first.

   Note: children are currently listed, left to right,
   from most preferred to least preferred.

   IMPORTANT: The preferred expansion is not a new list.
   Hence before destructively modifying preferred-expansion, 
   for example using sort, make sure to COPY the list first.

   Note: If search-node splits on an unresolved conflict, then 
   the children supplied must be the constituent-kernels of that 
   conflict."

  (cond ((not (null preferred-child-assignment))

		 ;; Then add this assignment to the common assignments.
		 (extend-assignments preferred-child-assignment
							 ancestor-assignments))

		((null ancestor-assignments)

		 ;; If no best assignment, then
		 ;; the ancestor assignments better be NIL.

		 '())

		(:otherwise
		 ;; Else its an error.

		 (error "No child-assignments provided." ))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Current Search Node: 
;;;      Determining Active Variables and Active and Unresolved Conflicts 
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Record Current Search Node
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Optimization:
;;; o This does not appear to be incremental.
;;;   Incrementally update from previous node to next node.

(defmethod record-search-node-as-current!

    ((search-problem search-problem) 
	 (new-search-node search-node)
	 &key &allow-other-keys)

  "Record NEW-SEARCH-NODE as current.

   Assumptions:

   - SEARCH-PROBLEM-CURRENT-SEARCH-NODE, 
     was already set to nil.
  
   Side Effects:  

   -  Records NEW-SEARCH-NODE as SEARCH-PROBLEM-CURRENT-SEARCH-NODE
      of SEARCH-PROBLEM."

  ;;  (debug-print-for :bfig t "  Partial kernel ~a (~a)"
  ;;	       (search-node-preferred-expansion-internal new-search-node) 
  ;;	       new-search-node)

  (with-slots (current-search-node) search-problem
	(setq current-search-node new-search-node))

  T)

;;; Note:  The following three are identical methods.
;;;        The three are provided because the first two don't
;;;        have a common ancestor for which this method applies.
(defmethod record-search-node-as-current!
	
    ((search-problem search-problem-w-conflicts) 
	 (new-search-node search-node)
	 &key disjoint?)
  
  "Record NEW-SEARCH-NODE as current, and 
   identifies all conflicts resolved by node's search state. "
  (record-search-node-as-current-internal! search-problem 
										   new-search-node disjoint?))

(defmethod record-search-node-as-current!
    ((search-problem conditional-search-problem) 
	 (new-search-node search-node)
	 &key disjoint?)

  "Record NEW-SEARCH-NODE as current, and identify all
   variables and conflicts activated by node's search state."

  (record-search-node-as-current-internal! search-problem 
										   new-search-node disjoint?))

(defmethod record-search-node-as-current!
    ((search-problem conditional-search-problem-w-conflicts) 
	 (new-search-node search-node)
	 &key disjoint?)
  
  "Record NEW-SEARCH-NODE as current, and identify all active
   variables and conflicts and all conflicts resolved by node's 
   search state. "
  
  (record-search-node-as-current-internal! search-problem 
										   new-search-node disjoint?))

;;; CHECK whether this fix is valid
;;; o Fixed:
;;; - Old assignemnts *must* be unrecorded so that variables that were active
;;; (because of previous assignments) may now become inactive
;;; - New-assigments *must* be recorded so that variables that were inactive
;;; may become active because of new assignments
;;; This was part of the reason why test12 wais failing.
;;; gustavop
(defmethod record-search-node-as-current-internal!

    (search-problem new-search-node disjoint?)

  "Record NEW-SEARCH-NODE as current and remove the old search node. 
   Update the set of supported conflicts and disallowed assignments
   based on the difference between the old and new search nodes.

   Assumptions:

   - SEARCH-PROBLEM-CURRENT-SEARCH-NODE, 
     STATE-VARIABLE-CURRENT-CDB-ASSIGNMENT
     and conflict support were already set to nil.
  
   Side Effects:  

   -  Records NEW-SEARCH-NODE as SEARCH-PROBLEM-CURRENT-SEARCH-NODE
      of SEARCH-PROBLEM.

   -  Records each assignment of NEW-SEARCH-NODE on their
      respective STATE-VARIABLE-CURRENT-CDB-ASSIGNMENTs
      of SEARCH-PROBLEM.

   -  For each conflict resolved by the candidate, the conflict is
      supported with the assignment that resolves it."

  ;;  (debug-print-for :bfig t "  Partial kernel ~a (~a)"
  ;;	       (search-node-preferred-expansion-internal new-search-node) 
  ;;	       new-search-node)

  (with-slots (old-search-node current-search-node) search-problem

	;;1. Find the differences in the assignments between the current and new node
	;;   and update the support.

	;;   Find the full sets of assignments - use the if to avoid an error when 
	;;   there isn't a node.
	(let ((old-assignments (search-problem-old-assignments search-problem)
			;; (if (null old-search-node)
							   
			;; 			   	   nil
							 
			;; 			   	 (append (search-node-ancestor-assignments
			;; 			   			  old-search-node)
									 
			;; 			   			 (list (search-node-preferred-child-assignment
			;; 			   					old-search-node))))
			)
		  
		  (new-assignments (append (search-node-ancestor-assignments
									new-search-node) 
								   (list (search-node-preferred-child-assignment
										  new-search-node)))))

	  ;; 1.a. Find assignments of the old node not in the new node and unrecord 
	  ;;      them.
	  (loop for old-assignment in (remove nil old-assignments)
		  unless (member old-assignment new-assignments)
		  do (unrecord-assignment-as-current! search-problem old-assignment)) 
	  ;; We must unrecord to deactivate what was activated.

	  ;; 1.b. Find assignments of the new node not in the old node and record them
	  ;; (loop for new-assignment in (remove nil new-assignments)
	  ;; 		 do (unless (and (member new-assignment old-assignments)
	  ;; 						 (not (eql new-search-node old-search-node)))

	  ;; 			  (record-assignment-as-current! search-problem
	  ;;                 new-assignment)))
	  )


	(when disjoint?
	  ;; 2. Find the differences in the disallowed assignments between the old 
	  ;;    and new node, and update the support.

	  ;; Find the full sets of assignments - use the if to avoid an error 
	  ;; when there isn't a node.
	  (let ((old-disallowed-assignments (if (null old-search-node)
											nil
										  (search-node-disallowed-assignments
										   old-search-node)))
			(new-disallowed-assignments (search-node-disallowed-assignments
										 new-search-node)))

		;; 2.a. Find disallowed assignments of the old node not in the new node 
		;;      and unrecord them.
		(loop for old-disallowed-assignment in old-disallowed-assignments
											   
			unless (member old-disallowed-assignment new-disallowed-assignments)
				   
			do (unrecord-disallowed-assignment-as-current!
				search-problem old-disallowed-assignment))

		;; 2.b. Find disallowed assignments of the new node not in the old node 
		;;      and record them.
		(loop for new-disallowed-assignment in new-disallowed-assignments
			unless (member new-disallowed-assignment old-disallowed-assignments)
				   
			do (record-disallowed-assignment-as-current!
				search-problem new-disallowed-assignment))))    
	
	;; Record the partial state of node's best child as current.
	;; This is used to identify active variables and conflicts and to
	;; resolve conflicts.
	(record-search-node-assignments-as-current!
	 search-problem new-search-node)

	;; Record the current disallowed assignments of new-search-node.
	;; These are used to identify resolved conflicts.
	
	;;	(record-search-node-disallowed-assignments-as-current!
	;;	 search-problem new-search-node :disjoint? disjoint?)

	;;And finally, update the current search node to be the new search node.
	(setq current-search-node new-search-node)
	)
  
  T)

(defmethod record-search-node-assignments-as-current!

	((search-problem search-problem) 
	 (new-search-node search-node))

  "Record as current and propagate the consequences of the 
   partial state of SEARCH-NODE's best child.
   Partial state is comprised of the ancestor assignments +
   the preferred child's assignment."
  
  ;; Record of the current assignment is used to 
  ;; identify active variables and conflicts,
  ;; and to identify resolved conflicts.
  
  (loop for new-assignment in 
		(search-node-ancestor-assignments new-search-node)
	  do (record-assignment-as-current! search-problem new-assignment))
  
  (let ((best-new (search-node-preferred-child-assignment new-search-node)))
    (unless (null best-new)
	  (record-assignment-as-current! search-problem best-new)))
  T)

(defmethod record-search-node-disallowed-assignments-as-current!

	((search-problem search-problem) 
	 (new-search-node search-node)
	 &key disjoint?)

  "Record as current and propagate the consequences of the 
   disallowed assignments of SEARCH-NODE's best child."
  
  ;; Record the current disallowed assignments of new-search-node.
  ;; These are used to identify resolved conflicts.
  (when disjoint?
    (loop for disallowed in 
		  (search-node-disallowed-assignments new-search-node)
		do (record-disallowed-assignment-as-current! 
			search-problem disallowed)))
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Unrecord Current Search Node
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Check:
;;; o How is this different from clear-current-search-node! ?
;;;   Document difference clearly, and/or remove one if redundant.

(defmethod unrecord-current-search-node!

	((search-problem search-problem)
	 &key &allow-other-keys)

  "Removes the record of any search-node as current.

  Assumptions:
   - SEARCH-PROBLEM-CURRENT-SEARCH-NODE is set to a search node.

  Side Effects:  
   - Clears SEARCH-PROBLEM-CURRENT-SEARCH-NODE of SEARCH-PROBLEM."

	;; Clear the node under current consideration.
  (setf (search-problem-current-search-node search-problem) 
	nil)
  T)

;;; Note:  The following three are identical methods.
;;;        The three are provided because the first two don't
;;;        have a common ancestor for which this method applies.
(defmethod unrecord-current-search-node!
	((search-problem search-problem-w-conflicts)
	 &key disjoint?)
  "Removes the record of any search-node as current,
   and any resolved conflicts."
  (unrecord-current-search-node-internal! search-problem disjoint?))

(defmethod unrecord-current-search-node!
	((search-problem conditional-search-problem)
	 &key disjoint?)
  "Removes the record of any search-node as current,
   and and active variables and conflicts."
  (unrecord-current-search-node-internal! search-problem disjoint?))

(defmethod unrecord-current-search-node!
	((search-problem conditional-search-problem-w-conflicts)
	 &key disjoint?)
  "Removes the record of any search-node as current,
   and any active variables and conflicts, and resolved conflicts."
  (unrecord-current-search-node-internal! search-problem disjoint?))

(defun unrecord-current-search-node-internal!

	(search-problem disjoint?)
(declare (ignore disjoint?)) ;;TODO: remove
  "Removes the record of any search-node as current,

  Assumptions:

   - SEARCH-PROBLEM-CURRENT-SEARCH-NODE is set to a search node.

   - The current kernel is the best cost extension of the current 
     search node.

   - The current variable assignments, 
     STATE-VARIABLE-CURRENT-CDB-ASSIGNMENT,
     are set to the assignments of the current search node.

   - Those conflicts of SEARCH-PROBLEM that are resolved by the best cost
     extension are supported by one of the assignments in the extension.
  
  Side Effects:  

   - Clears SEARCH-PROBLEM-CURRENT-SEARCH-NODE of SEARCH-PROBLEM.
   - Clears the current variable assignments.
   - Clears the conflict support."

  (with-slots (old-search-node current-search-node) search-problem
	
	;;Keep the old search node for fast update later when you pick a new one
	(setf old-search-node current-search-node)
	
	;; Clear the node under current consideration.
	(setq current-search-node nil)
	T))

;; (defmethod unrecord-current-search-node-assignments!

;; 	((search-problem search-problem)
;; 	 (current-node search-node))
  
;;   "Remove as current the partial state of the best child of CURRENT-NODE.
;;    The partial state of the best child consists of node's ancestor 
;;    assignments plus its preferred child  assignment."

;;   (loop for current-assignment in 
;; 		(search-node-ancestor-assignments current-node)

;; 	  do (unrecord-assignment-as-current! search-problem 
;; 										  current-assignment))
    
;;   (let ((best-child (search-node-preferred-child-assignment 
;; 					 current-node)))
;; 	(unless (null best-child)
;; 	  (unrecord-assignment-as-current! search-problem best-child)))
;;   T)

(defmethod unrecord-current-search-node-disallowed-assignments!

	((search-problem search-problem-w-conflicts)
	 (current-node search-node)
	 &key disjoint?)
  
  "Remove disallowed assignments of current-search-node of 
   SEARCH-PROBLEM."

  (when disjoint?
	(loop for disallowed in 
		  (search-node-disallowed-assignments current-node)
		do (unrecord-disallowed-assignment-as-current! 
			search-problem disallowed)))
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Search Tree Expansion
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:

;;; o Switching between previous and current search-nodes does
;;;   not appear to be incremental.  This may not matter much for
;;;   best-first search, but will matter for depth-first search.
;;;   Make this incremental.

;;; o Factor out the constraint database separate from the candidate
;;;   generator of bfig.  Give it an interface similar to the other
;;;   incremental constraint databases.

;;; o Add functionality for wiring up a search tree.
;;;   When we want to keep a trace of the search tree,
;;;   we can do this by augmenting the current tree structure.

;;;   - The current structure does a good job of keeping track
;;;     of the remaining assignments to be explored.  In particular,
;;;     the children-assignments of a node denote the unexpanded
;;;     children.

;;;   - To maintain a trace, each node should include a list of
;;;     expanded children (in BFS order) as well as unexpanded
;;;     assignments.  A node should point to its parent node as well.

;;;   - Given a record of this tree structure, we no longer need to
;;;     maintain a debug trace.  We can include the status info of
;;;     leaf nodes on the search-node objects.

;;; To Do:
;;; o Generalize to uninformed-search-problem.

;;; o Divide into w vs w/o conflicts.  Doesn't need to record
;;;   current state if no conflicts.

;;; Check:
;;; o Make sure that, when the current-node is entered into the
;;;   conflict database, it only records the partial assignment
;;;   corresponding to the best child.

;;; Changed:
;;; o Debug-print statements.
(defmethod expand-and-queue-current-search-node!

    ((search-problem 
	  search-problem-w-conflicts
;;;	  informed-search-problem-w-conflicts
	  )
	 &optional full-assignment? disjoint?)

  "Expands the preferred child of the current search-node N of 
   SEARCH-PROBLEM.  
   This function generates a binary expansion, comprised of N's 
   best child and a node denoting the remaining siblings of N 
   yet to be explored (the next best sibling node).

   When an unresolved conflict exists, the children of the expanded
   assignment select add each assignment that resolves some conflict
   (that is, the conflicts constitutent kernels).

   If FULL-ASSIGNMENT? is true and all conflicts are resolved, then
   splits on any unassigned variable.

   If DISJOINT? is true, then splitting ensures that sibling search
   nodes denote disjoint sets of assignments (that is, no leaf node
   of two siblings denote the same variable assignment.

   Assumptions:
   o A search node has been recorded as current,
   o The node's best child has been recorded as the current search state
     and the unresolved conflicts for this assignment have been identified.
   o The current search node is no longer on the SEARCH-QUEUE,
     and can be destructively modified.
  
   Side Effects:
   o Adds to search queue two nodes:
     - a new child node, corresponding to current node's best child, 
       whose child-assignments are the constituent kernels of 
       an unresolved conflict.  The conflict selected is the one with
       the smallest number of constituent kernels.
       If no unresolved conflict exists and full-assignment is true,
       then the set of children is the domain of an unassigned variable,
       where variable is the one with the smallest domain.

     - a new sibling node, which destructively modifies the current 
       search-node by removing its best child and leading inconsistent 
       children, and updating cost.
       If disjoint? is true, then sibling includes the restriction that
       its leaf nodes cannot include the assignments of its preceding 
       siblings."

  (let ((current-node
		 (search-problem-current-search-node 
		  search-problem)))

    (debug-print-for :bfig t 
					 "  Expand Node ~a." 
;;;					 "  Expand Node ~a.~2%~a~%" 
					 current-node
;;;					 (print-search-node-details current-node nil)
					 )

	;; Expand the best cost child of current search-node 
	;; either because unresolved conflicts or unassigned variables remain,
	;; or because current search-node was already examined as a candidate.

	;; This assumes that current-node has been recorded in the conflict
	;; data base and corresponding unresolved conflicts have been identified.

	(if (not (process-new-conflicts current-node search-problem))
		(drop-node-due-to-conflicts! current-node search-problem)
		(block expanding-children
		  ;; 1) Create the preferred child node unless children cannot resolve a
		  ;; currently known conflict. But, still allow the sibling to happen (next
		  ;; to try to find a way out.

		  (let ((child-node 
				 (preferred-child-node search-problem current-node 
									   full-assignment? disjoint?)))

			;; and queue it.
			(unless (or (null child-node)
						;;unless the child has an unforseen problem
						(not (test-ancestors-against-resolvable child-node search-problem)))
			  (enqueue-search-node! search-problem child-node)
			  (debug-print-for :bfig t 
							   "    Best child = ~a." 
;;;						 "    Best child = ~a.~2%~a~%" 
							   child-node
;;;						 (print-search-node-details child-node nil)
							   )))


		  ;; 2) Before this node is destructively modified, we NEED to
		  ;; record the current best
		  ;; assignment. RECORD-SEARCH-NODE-AS-CURRENT-INTERNAL! needs
		  ;; to know the current assignment so that it can compute the
		  ;; correct deltas when changing nodes. It used to do that by
		  ;; looking at the previous node (THIS node), but that
		  ;; doesn't make sense because of the destructive
		  ;; modification in step 3.

		  (let ((current-assignments (append (search-node-ancestor-assignments current-node)
											 (list (search-node-preferred-child-assignment
													current-node)))))
			(setf (search-problem-old-assignments search-problem) current-assignments))

		  ;; 3) Create the next most preferred sibling node.

	  
		  (let ((sibling-node 
				 (next-most-preferred-sibling-node
				  search-problem current-node
				  :disjoint? disjoint?)))
	  
			;; and queue it.
			(cond ((null sibling-node)
				   (debug-print-for :bfig t "    Siblings exhausted."))
				  (:otherwise
				   (enqueue-search-node! search-problem sibling-node)
				   (debug-print-for
					:bfig t 
					"    Best sibling = Node ~a." 
;;;		 "    Best sibling = Node ~a.~2%~a~%" 
					sibling-node
;;;		 (print-search-node-details sibling-node nil)
					))))))

	;; Now that the child has been expanded, we no longer need
	;; to know the unresolved conflicts corresponding to the 
	;; current node, remove it from the conflict data base.
	
    (unrecord-current-search-node! search-problem :disjoint? disjoint?)
	(debug-print-for :bfig t 
					 "    Removed the old one." )

	T))


(defmethod expand-and-queue-current-search-node!

    ((search-problem 
	  search-problem
;;;	  informed-search-problem-w-conflicts
	  )
	 &optional full-assignment? disjoint?)

  "Expands the preferred child of the current search-node N of 
   SEARCH-PROBLEM.  
   This function generates a binary expansion, comprised of N's 
   best child and a node denoting the remaining siblings of N 
   yet to be explored (the next best sibling node).

   When an unresolved conflict exists, the children of the expanded
   assignment select add each assignment that resolves some conflict
   (that is, the conflicts constitutent kernels).

   If FULL-ASSIGNMENT? is true and all conflicts are resolved, then
   splits on any unassigned variable.

   If DISJOINT? is true, then splitting ensures that sibling search
   nodes denote disjoint sets of assignments (that is, no leaf node
   of two siblings denote the same variable assignment.

   Assumptions:
   o A search node has been recorded as current,
   o The node's best child has been recorded as the current search state
     and the unresolved conflicts for this assignment have been identified.
   o The current search node is no longer on the SEARCH-QUEUE,
     and can be destructively modified.
  
   Side Effects:
   o Adds to search queue two nodes:
     - a new child node, corresponding to current node's best child, 
       whose child-assignments are the constituent kernels of 
       an unresolved conflict.  The conflict selected is the one with
       the smallest number of constituent kernels.
       If no unresolved conflict exists and full-assignment is true,
       then the set of children is the domain of an unassigned variable,
       where variable is the one with the smallest domain.

     - a new sibling node, which destructively modifies the current 
       search-node by removing its best child and leading inconsistent 
       children, and updating cost.
       If disjoint? is true, then sibling includes the restriction that
       its leaf nodes cannot include the assignments of its preceding 
       siblings."

  (let ((current-node
		 (search-problem-current-search-node 
		  search-problem)))

    (debug-print-for :bfig t 
					 "  Expand Node ~a." 
;;;					 "  Expand Node ~a.~2%~a~%" 
					 current-node
;;;					 (print-search-node-details current-node nil)
					 )

	;; Expand the best cost child of current search-node 
	;; either because unresolved conflicts or unassigned variables remain,
	;; or because current search-node was already examined as a candidate.

	;; This assumes that current-node has been recorded in the conflict
	;; data base and corresponding unresolved conflicts have been identified.


	;; 1) Create the preferred child node unless children cannot resolve a
	;; currently known conflict. But, still allow the sibling to happen (next
	;; to try to find a way out.
	
	(let ((child-node 
		   (preferred-child-node search-problem current-node 
								 full-assignment? disjoint?)))
	  ;; and queue it.
	  (unless (null child-node)
		(enqueue-search-node! search-problem child-node)
		(debug-print-for :bfig t 
						 "    Best child = ~a." 
;;;						 "    Best child = ~a.~2%~a~%" 
						 child-node
;;;						 (print-search-node-details child-node nil)
						 )))


	;; 2) Create the next most preferred sibling node.

	  
	(let ((sibling-node 
		   (next-most-preferred-sibling-node
			search-problem current-node
			:disjoint? disjoint?)))
	  
	  ;; and queue it.
	  (cond ((null sibling-node)
			 (debug-print-for :bfig t "    Siblings exhausted."))
			(:otherwise
			 (enqueue-search-node! search-problem sibling-node)
			 (debug-print-for
			  :bfig t 
			  "    Best sibling = Node ~a." 
;;;		 "    Best sibling = Node ~a.~2%~a~%" 
			  sibling-node
;;;		 (print-search-node-details sibling-node nil)
			  ))))

	;; Now that the child has been expanded, we no longer need
	;; to know the unresolved conflicts corresponding to the 
	;; current node, remove it from the conflict data base.
	
    (unrecord-current-search-node! search-problem :disjoint? disjoint?)
	(debug-print-for :bfig t 
					 "    Removed the old one." )
	T))

;;; Fixed!
;;; (candidate-found? had been previously changed and test8 worked) 
;;; o Child node may be a full or partial assignment that cannot
;;;   resolve a conflict.  If this is the case, then null
;;;   needs to be returned.  This bug is causing test8 to fail.
;;;   (has this bug been addressed?).
;;; -- gustavop

;;; Changed:
;;; o Added search-problem to arguments of update-unassigned-variables.
;;; o removed call to update-unassigned-variables.
(defmethod preferred-child-node

	((search-problem 
	  search-problem
;;;	  search-problem-w-conflicts
	  ) 
	 (search-node search-node)
	 full-assignment? disjoint?)

  "Expand SEARCH-NODE of SEARCH-PROBLEM,
   Returning a node for the preferred child of SEARCH-NODE.
   If FULL-ASSIGNMENT? is true, then continue splitting until
   all variables are assigned, otherwise, stop as soon as there
   are no unresolved conflicts.
   If DISJOINT? is true, then splitting ensures that sibling search
   nodes denote disjoint sets of assignments.

   ASSUMES, upon entry, that the partial assignment corresponding
   to search-node's preferred child has already been record as the 
   current search state, if needed for search-problem."
  
  ;; The new best child-node's ancestor assignments are comprised of
  ;; SEARCH-NODE's ancestor assignments plus its best child assignment.

  ;; The new best child-node's child assignments are comprised of either
  ;; 1) the constituent kernels of the (smallest) conflict that is not 
  ;;    resolved by SEARCH-NODE, or
  ;; 2) the assignments of an unassigned variable with the smallest
  ;;    domain.
  
  (let* ((
		  ;; Note: the list of child-ancestors builds upon its
		  ;; parents ancestors, hence this list needs to be copied
		  ;; before modifying it.
		  child-ancestor-assignments (search-node-preferred-expansion-internal
									  search-node))

		 ;; Variables not mentioned in an ancestor assignment of child.
		 (child-unassigned-ancestor-vars (search-node-unassigned-variables
										  search-node))

		 ;; Assignments that cannot appear in any leaf of child.
		 ;; Used to ensure that the leaves of sibling nodes are disjoint.
		 ;; Note: Unless splitting on conflicts, disjointness is guaranteed w/o
		 ;; disallowing assignments.
		 (child-disallowed-assignments 
		  (when (and disjoint? 
					 (typep search-problem 'search-problem-w-conflicts))
			(search-node-disallowed-assignments search-node))))

    (multiple-value-bind (child-child-assignments split-on-conflict? conflict-split-on)
		;; Returns the children of the newly created child assignment -
		;; either 1) the constituent kernels of an unresolved conflict,
		;; when working with conflicts, or 2) the domain of an unassigned 
		;; variable.
		;; Returns true for split-on-conflict? if children result from 
		;; resolving a conflict.
		(get-new-child-assignments search-problem
								   child-unassigned-ancestor-vars
								   full-assignment?
								   :current-assignments child-ancestor-assignments
								   :disallowed-assignments child-disallowed-assignments
								   )

	  (if split-on-conflict?
		  (record-split-on-conflict conflict-split-on search-node search-problem))
;(format t "~a~%" child-child-assignments)
;	  (setf child-child-assignments (reverse child-child-assignments))
;(format t "~a~%" child-child-assignments)
	  
      (cond ((null child-child-assignments)
			 ;; When splitting failed to produce children,
			 ;; do NOT create a child node.
			 nil)

;; 			Return a child search-node with the
;; 			above child assignments.
			(:otherwise

;;; Fix Delete the following comments when works.
;;;			 (let* (
;;;					;; extract the preferred grand child of the new grand children.
;;;					(child-preferred-child-assignment 
;;;					 (search-node-preferred-child-among
;;;					  search-node
;;;					  child-child-assignments))

					;; Inherit unassigned variables from new node's parent,
					;; while removing the variable of the preferred child.
;;;					(child-unassigned-variables (update-unassigned-variables
;;;												 search-problem
;;;												 child-unassigned-ancestor-vars
;;;												 child-preferred-child-assignment
;;;												 child-ancestor-assignments))
;;;					)
			   
			   ;; Create the new child node.
			   ;; Wait to determine unassigned variables until search-node
			   ;; is expanded off the queue, when active variables are
			   ;; identified.
			 (if (typep search-problem 'search-problem-w-conflicts)
				 (make-instance (search-problem-node-type search-problem)
								:search-problem search-problem
								:ancestor-assignments child-ancestor-assignments
								:child-assignments child-child-assignments
								:unassigned-ancestor-variables child-unassigned-ancestor-vars
								:unassigned-variables :not-specified
;;;				 child-unassigned-variables
								:disallowed-assignments child-disallowed-assignments
								:split-on-conflict? split-on-conflict?
								:search-problem search-problem
								:newest-processed-conflict (search-node-newest-processed-conflict search-node)
								:resolvable-conflicts (copy-list (search-node-resolvable-conflicts search-node)))
				 (make-instance (search-problem-node-type search-problem)
								:search-problem search-problem
								:ancestor-assignments child-ancestor-assignments
								:child-assignments child-child-assignments
								:unassigned-ancestor-variables child-unassigned-ancestor-vars
								:unassigned-variables :not-specified
;;;				 child-unassigned-variables
								:disallowed-assignments child-disallowed-assignments
								:split-on-conflict? split-on-conflict?
								:search-problem search-problem))
;;;			   )
			 )))))

;;; Changed:
;;; o Added search-problem to arguments of
;;;   next-unassigned-variable.
;;; o Updated comments.
;;; o Deleted to do.

;;; Fixed
;;; If we don't get any conflict-kernels but we do have unresolved
;;; conflicts we should not keep expanding this node. Instead,
;;; the subtree rooted at this node should be pruned.
;;; --gustavop
(defmethod get-new-child-assignments

    ((search-problem search-problem-w-conflicts) 
	 unassigned-variables
     full-assignment?
	 &key
	 current-assignments
	 disallowed-assignments)

  (declare (ignore full-assignment?)) ;Doesn't actually change the action
  "Returns (VALUES CHILD-ASSIGNMENTS SPLIT-ON-CONFLICT? CONFLICT-SPLIT-ON).

   CHILD-ASSIGNMENTS is a list of child assignments that 
   extend CURRENT-ASSIGNMENTS for the current search node
   to a candidate solution to SEARCH-PROBLEM.

   SPLIT-ON-CONFLICT? is true iff CHILD-ASSIGNMENTS are
   kernels that resolve an unresolved conflict.

   CURRENT-ASSIGNMENTS is a partial assignment to the variables of 
     SEARCH-PROBLEM, denoting the state of a current search node.

   UNASSIGNED-VARIABLES contains the variables of SEARCH-PROBLEM that
     are not assigned by CURRENT-ASSIGNMENTS.  

     Note that this list may include assigned variables, but the first 
     variable will not to be assigned.  In addition it may have had
     removed unassigned variables that are guaranteed to be inactive.

   A candidate solution is an assignment to SEARCH-PROBLEM's (active) 
   variables that resolves all conflicts.   

   If FULL-ASSIGNMENT? is true,then the candidate solution must be a 
   full assignment; otherwise, it may be partial.

   If DISALLOWED-ASSIGNMENTS is non-NIL, then ensure disjointness of
   solutions, by ensuring that the extensions to current-assignments
   avoids these assignments.

   ASSUMES, upon entry, that the assignments of a current search node 
   have already been record as the current search state, and that 
   UNASSIGNED-VARIABLES has been updated, using
   SET-UNASSIGNED-VARIABLES-OF-CURRENT-SEARCH-NODE! to eliminate any 
   leading variables that are assigned or inactive."
  
  (let* ((smallest-conflict (shortest-resolvable-conflict (search-problem-current-search-node
														   search-problem)
														  search-problem)
		   ;;(smallest-unresolved-conflict-of-current-state search-problem) old way
		   )
		 (conflict-kernels (kernels-of-the-smallest-unresolved-conflict
						   search-problem 
						   current-assignments 
						   smallest-conflict
						   disallowed-assignments)))
	
    (cond ((not (null conflict-kernels))
		   ;; When an unresolved conflict exists, 
  
		   ;; THEN select the unresolved conflict with the smallest 
		   ;; set of constitutent kernels, and return the constituent 
		   ;; kernels, listed from most to least preferred.
		   
		   (debug-print-for :bfig t "    Split on its kernels = ~a." 
							conflict-kernels)
		   (values conflict-kernels T smallest-conflict))

	  ;;; Fixed
	  ;;; If we don't get any conflict-kernels but we do have unresolved
	  ;;; conflicts we should not keep expanding this node. Instead,
	  ;;; the subtree rooted at this node should be pruned.
	  ;;; --gustavop

	  ((not (null smallest-conflict
;;calculated above, avoid recalculating in "kernels-of-the..." above				 
;(smallest-unresolved-conflict-of-current-state search-problem)
))

	   (values nil nil))
	  

		  (:else ;;Even if we're not looking for a full assignment, you need to extend
		   ;;it so that we don't stop

		   (let ((unassigned-variable
				  (next-unassigned-variable search-problem 
											unassigned-variables)))
			 
			 (cond ((not (null unassigned-variable))
					;; IF an unassigned variable exists, 

					;; THEN select the unassigned variable with the 
					;; smallest domain, and RETURN its domain assignments, 
					;; listed from most to least preferred.

					;; Prune away any of domain assignments that are ruled 
					;; out by the DISALLOWED ASSIGNMENTS of the current
					;; search node.
					(let ((trimmed-assignments

						   (trim-leading-inconsistent-assignments
							(variable-ordered-domain unassigned-variable)
							'()			;When splitting on a variable, we
										;are guaranteed that current-assignment
										;hasn't made an assignment to 
										; UNASSIGNED-VARIABLE.
										;hence we don't check consistency
										;with the ancestor assignment.
							
							disallowed-assignments)))
					  
					  (debug-print-for :bfig t "    Select unassigned variable ~a." 
									   unassigned-variable)
					  
					  (debug-print-for :bfig t "    Split on trimmed assignments ~a." 
									   trimmed-assignments)

					  (values trimmed-assignments nil)))

				   (:otherwise
					
					;; IF no (active) unassigned variable exists,

					;; Then node is terminal, return no children.
					(values nil nil)))))
		  
		  )))



;;; Changed:
;;; o Updated comments.

;;; To Do:
;;; o Can we eliminate further redundancy with the two
;;;   methods named get-new-child-assignments?

(defmethod get-new-child-assignments

    ((search-problem search-problem) 
	 unassigned-variables
     full-assignment?
	 &key &allow-other-keys)
  (declare (ignore full-assignment?)) ;Doesn't change anything.

  "Returns (VALUES CHILD-ASSIGNMENTS SPLIT-ON-CONFLICT?).

   CHILD-ASSIGNMENTS is a list of child assignments that 
     extend the recorded assignments for the current
     search node of SEARCH-PROBLEM.

   SPLIT-ON-CONFLICT? is true iff CHILD-ASSIGNMENTS are
     kernels that resolve an unresolved conflict.

     By default a search problem has no conflicts, 
     in which case NIL is always returned.

   UNASSIGNED-VARIABLES contains the variables of SEARCH-PROBLEM that
     are not assigned by the recorded current-assignments.  

     Note that this list may include assigned variables, but the first 
     variable will not to be assigned.  In addition it may have had
     removed unassigned variables that are guaranteed to be inactive.

   A candidate solution is an assignment to SEARCH-PROBLEM's (active) 
   variables.   

   If FULL-ASSIGNMENT? is true, then the candidate solution must be a 
   full assignment; otherewise, it may be partial.

   If DISALLOWED-ASSIGNMENTS is non-NIL, then this function will 
   ensure disjointness of candidates, by ensuring that the 
   extensions to CURRENT-ASSIGNMENTS avoids these assignments.

   ASSUMES, upon entry, that the assignments of a current search node 
   have already been record as the current search state, and that 
   UNASSIGNED-VARIABLES has been updated, using
   SET-UNASSIGNED-VARIABLES-OF-CURRENT-SEARCH-NODE! to eliminate 
   any leading variables that are assigned or inactive."
  
  (cond
   (:else ;;Always extend the assignment

	(let ((unassigned-variable
		   (next-unassigned-variable search-problem
									 unassigned-variables)))
			 
	  (cond ((not (null unassigned-variable))
			 ;; When an (active) unassigned variable exists,

			 ;; THEN select the unassigned variable with the smallest 
			 ;; domain, and return the domain assignments, listed 
			 ;; from most to least preferred.
			 (let ((assignments 
					(variable-ordered-domain unassigned-variable))

				   ;; Note, if there are no conflicts, then there are 
				   ;; no disallowed assignments and hence no domain 
				   ;; elements to trim.

;;;				   (trimmed-assignments
;;;					(trim-leading-inconsistent-assignments
;;;					 (variable-ordered-domain unassigned-variable)
;;;					 '()				;When splitting on a variable, we
;;;										;are guaranteed that current-assignment
;;;										;hasn't assigned unassigned-variable.
;;;										;hence we don't check consistency
;;;										;with the ancestor assignment.
;;;							
;;;										;disallowed-assignments
;;;					 nil)
;;;					)
				   )
					  
			   (debug-print-for :bfig t "    Unassigned variable ~a." 
								unassigned-variable)

			   (debug-print-for :bfig t "    Split on assignments ~a." 
;;;								trimmed-assignments
								assignments
								)

;;;			   (values trimmed-assignments nil)
			   (values assignments nil)
			   ))

			(:otherwise
			 ;; When there is no (active) unassigned variable,

			 ;; return without extending search node.
			 (values nil nil)
			 ))))
   
   ;;Was used to not extend assignment
;;    (:else
;; 	(values nil nil)
;; 	)
))





;;; To Do:
;;; o Divide into w vs w/o conflicts.  Doesn't need to record
;;;   current state if no conflicts.

;;; o Provide a flag that decides whether or not a node should
;;;   point to its child search nodes as they are generated.

;;; Optimization:
;;; o When a node has no remaining siblings, should it be
;;;   explicitly reclaimed, or should we just let the garbage 
;;;   collector work?

;;; Changed:
;;; o Added search-problem to arguments of update-unassigned-variables.
;;; o Removed call to update-unassigned-variables.
(defmethod next-most-preferred-sibling-node

	((search-problem search-problem-w-conflicts) 
	 (search-node informed-search-node)
	 &key disjoint?)

  "Returns the next best sibling node of SEARCH-NODE, or
   none if no siblings remaining.

   If DISJOINT? is true, then splitting ensures that sibling search
   nodes denote disjoint sets of assignments.

   Side-effect:  Sibling node is generated by DESTRUCTIVELY
                 modifying SEARCH-NODE."

  (let ((preferred-child (search-node-preferred-child-assignment 
						  search-node)))

	;; Update disallowed assignments of SEARCH-NODE.
	
	;; When the set of leaves of sibling nodes should be disjoint, impose
	;; the constraint that the assignment of the first sibling should
	;; not appear in the leaves of the remaining siblings.
    (when (and disjoint?
			   (search-node-split-on-conflict? search-node))
	  
	  ;; Augment the disallowed assignments with the child assignment
	  ;; of the best sibling.
      (search-node-add-disallowed-assignment! preferred-child search-node))
	
    (let* ((ancestor-assignments (search-node-ancestor-assignments
								  search-node))
		   
		   (disallowed-assignments (search-node-disallowed-assignments
									search-node))
		   
		   (other-siblings (search-node-lesser-sibling-assignments
							search-node))
		   
		   ;; Prune the leading assignments of the remaining siblings 
		   ;; that are ruled out by disallowed assignments.
		   (other-consistent-siblings (trim-leading-inconsistent-assignments
									   other-siblings
									   ancestor-assignments
									   disallowed-assignments)))

      (cond ((null other-consistent-siblings)
			 
			 ;; No node is returned if no siblings remain to expand.
			 nil)
			
			(:otherwise
			 
			 ;; Child assignments of SEARCH-NODE are updated by removing 
			 ;; the preferred child assignment, along with the leading 
			 ;; next most preferred siblings that are rules out by
			 ;; the ancestor assignments or disallowed assignments.
    
			 (setf (search-node-child-assignments search-node)
			   other-consistent-siblings)

			 ;; Indicate that unassigned variables are yet to be determined.

			 (setf (search-node-unassigned-variables search-node) 
			   :not-specified)

			 (finish-sibling-node-update! search-problem
										  search-node
;;;              Fix Delete these comments when works.
;;;										  ancestor-assignments
;;;										  other-consistent-siblings
										  )

			 ;;; Fix Delete these comments when works.
;;;			 ;; Update SEARCH-NODE's unassigned variables to expand.
;;;			 (setf (search-node-unassigned-variables search-node)
;;;			   (update-unassigned-variables search-problem
;;;				(search-node-unassigned-ancestor-variables search-node)
;;;				(children-best-assignment other-consistent-siblings)
;;;				(preferred-expansion ancestor-assignments 
;;;									 other-consistent-siblings)))
;;;			 
;;;			 ;; Update SEARCH-NODE's cost.
;;;			 (setf (search-node-best-cost search-node)
;;;			   (compute-search-node-best-cost search-problem search-node))
			 
			 ;; When finished, changed child-assignments,
			 ;; disallowed-assignments, unassigned-variables and best cost.
			 
			 ;; Didn't change index, ancestors and
			 ;; unassigned-ancestor-variables. 
			 search-node)))))



;;; Changed:
;;; o Added search-problem to arguments of
;;;   update-unassigned-variables.
;;; Check:
;;; o Had removed creation of other-consistent-siblings,
;;;   but appeared to be needed, and hence put back.
;;;   Is this correct?  Likewise, needed ancestor-assignments
;;;   and put back in.
(defmethod next-most-preferred-sibling-node

	((search-problem search-problem) 
	 (search-node informed-search-node)
	 &key &allow-other-keys)

  "Returns the next best sibling node of SEARCH-NODE, or
   none if no siblings remaining.

   If DISJOINT? is true, then splitting ensures that sibling search
   nodes denote disjoint sets of assignments.

   Side-effect:  Sibling node is generated by DESTRUCTIVELY
                 modifying SEARCH-NODE."

;;;  (let ((preferred-child (search-node-preferred-child-assignment 
;;;						  search-node)))

	;; Update disallowed assignments of SEARCH-NODE.
	
	;; When the set of leaves of sibling nodes should be disjoint, impose
	;; the constraint that the assignment of the first sibling should
	;; not appear in the leaves of the remaining siblings.
    ; (when (and disjoint?
; 			   (search-node-split-on-conflict? search-node))
	  
; 	  ;; Augment the disallowed assignments with the child assignment
; 	  ;; of the best sibling.
;       (search-node-add-disallowed-assignment! preferred-child search-node))
	
  (let* (
;;; Fix Delete these comments when works.
		 (ancestor-assignments (search-node-ancestor-assignments
								search-node))
		   
		   ; (disallowed-assignments (search-node-disallowed-assignments
; 									search-node))
		   
		   (other-siblings (search-node-lesser-sibling-assignments
							search-node))
		   
		   ;; Prune the leading assignments of the remaining siblings 
		   ;; that are ruled out by disallowed assignments.
		   (other-consistent-siblings (trim-leading-inconsistent-assignments
									   other-siblings
									   ancestor-assignments
										;disallowed-assignments)))
									   nil))
		   )

      (cond ((null other-siblings)
			 
			 ;; No node is returned if no siblings remain to expand.
			 nil)
			
			(:otherwise
			 
			 ;; Child assignments of SEARCH-NODE are updated by removing 
			 ;; the preferred child assignment, along with the leading 
			 ;; next most preferred siblings that are rules out by
			 ;; the ancestor assignments or disallowed assignments.
    
			 (setf (search-node-child-assignments search-node)
			   other-consistent-siblings)

			 ;; Indicate that unassigned variables are yet to be determined.
			 (setf (search-node-unassigned-variables search-node) 
			   :not-specified)

			 (finish-sibling-node-update! search-problem
										  search-node
;;; Fix Delete these comments when works.
;;;										  ancestor-assignments
;;;										  other-siblings
										  )

			 ;;; Fix Delete these comments when works.
;;;			 ;; Update SEARCH-NODE's unassigned variables to expand.
;;;			 (setf (search-node-unassigned-variables search-node)
;;;			   (update-unassigned-variables search-problem
;;;				(search-node-unassigned-ancestor-variables search-node)
;;;				(children-best-assignment other-consistent-siblings)
;;;				(preferred-expansion ancestor-assignments 
;;;									 other-consistent-siblings)))
;;;			 
;;;			 ;; Update SEARCH-NODE's cost.
;;;			 (setf (search-node-best-cost search-node)
;;;			   (compute-search-node-best-cost search-problem search-node))
			 
			 ;; When finished, changed child-assignments and unassigned-variables.
			 
			 ;; Didn't change index, ancestors and
			 ;; unassigned-ancestor-variables. 
			 search-node)))
;;;	)
  )




;;; To Do:
;;; o For conditional CSPs,
;;;   - update the list of active variables.

;;; Changed:
;;; o Added search-problem to arguments of update-unassigned-variables.
;;; o informed-search-problem-w-conflicts seemed
;;;   incorrect, removed informed-.
;;; o This function shouldn't be dependent on conflicts.
;;;   replaced search-problem-w-conflicts with
;;;   search-problem.
;;; o Commented out update-unassigned-variables.
(defmethod finish-sibling-node-update!

	((search-problem search-problem)
	 (search-node search-node)
;;;	 ancestor-assignments
;;;	 other-consistent-siblings
	 )
  
  "Complete the process of updating SEARCH-NODE to a sibling-node
   by making those changes that are specific to the type 
   of SEARCH-NODE."

;;;  (let ((next-preferred-child (uninformed-preferred-assignment
;;;							   other-consistent-siblings)))
;;;
;;;	;; update unassigned variables.
;;;	(setf (search-node-unassigned-variables search-node)
;;;	  
;;;	  (update-unassigned-variables 
;;;	   search-problem
;;;	   (search-node-unassigned-ancestor-variables search-node)
;;;	   next-preferred-child
;;;	   (preferred-expansion ancestor-assignments 
;;;							next-preferred-child))))

  T)

;;; Changed:
;;; o Added search-problem to arguments of update-unassigned-variables.
;;; o This function shouldn't be dependent on conflicts.
;;;   replaced informed-search-problem-w-conflicts with
;;;   informed-search-problem.
;;; o Commented out update-unassigned-variables.
(defmethod finish-sibling-node-update!

	((search-problem informed-search-problem)
	 (search-node informed-search-node)
;;;	 ancestor-assignments
;;;	 other-consistent-siblings
	 )
  
  "Complete the process of updating SEARCH-NODE to a sibling-node
   by making those changes that are specific to the type 
   of SEARCH-NODE."

;;; Delete once works.
;;;  (let ((next-best-child (children-best-assignment 
;;;						  other-consistent-siblings)))
;;;
;;;	;; update unassigned variables.
;;;	(setf (search-node-unassigned-variables search-node)
;;;
;;;	  (update-unassigned-variables 
;;;	   search-problem
;;;	   (search-node-unassigned-ancestor-variables search-node)
;;;	   next-best-child
;;;	   (preferred-expansion ancestor-assignments 
;;;							next-best-child))))
  
  ;; Update cost.
  (setf (search-node-best-cost search-node)
	(compute-search-node-best-cost search-problem search-node))

  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Search Termination - With Candidate
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Fix:
;;; o For Conditional CSP
;;;   - Terminate if all active variables are assigned.
(defmethod candidate-found? 

	((search-problem search-problem-w-conflicts)
	 (search-node search-node)
	 full-assignment?)

  "Returns true iff the current SEARCH-NODE of SEARCH-PROBLEM 
   is a complete candidate.  A complete candidate must resolve
   all conflicts.  In addition, if FULL-ASSIGNMENT? is true
   then the candidate must be a full assignment.

   ASSUMES that when called, SEARCH-NODE has been recorded
   as current, and its unassigned-variables has been updated
   using SET-UNASSIGNED-VARIABLES-OF-CURRENT-SEARCH-NODE!."
  
  ;;If there are no unresolvable-conflicts, test if it should be a full-
  ;;assignment

  (if (test-current-against-resolvable search-node search-problem) 
      (if full-assignment?
		(all-state-variables-assigned? search-node)
		t)
    nil))
  
;  (if full-assignment?
;       (all-state-variables-assigned? search-node)
;     (all-conflicts-resolved? search-problem)))


(defmethod candidate-found?

	((search-problem search-problem)
	 (search-node search-node)
	 full-assignment?)

  "Returns true iff the current SEARCH-NODE of SEARCH-PROBLEM 
   is a complete candidate.  A complete candidate must resolve
   all conflicts.  In addition, if FULL-ASSIGNMENT? is true
   then the candidate must be a full assignment.

   ASSUMES that when called, SEARCH-NODE has been recorded
   as current, and its unassigned-variables has been updated
   using SET-UNASSIGNED-VARIABLES-OF-CURRENT-SEARCH-NODE!. "
  
  (if full-assignment?
      (all-state-variables-assigned? search-node)
	T))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Search Termination - Without Candidate
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Note:
;;; o Search is terminated based on the number of search nodes expanded,
;;;   rather than full kernels generated.  This is because the
;;;   time required to generate each kernel is unpredicatable, but
;;;   node expansion takes constant time, hence the time to create n
;;;   nodes is predictable.

;;; Changed:
;;; o Debug-print statements.

;;; Changed:
;;; o Generalized to search-problem
;;; -- gustavop

(defmethod end-search?

	((search-problem search-problem)
										;search-problem-w-conflicts)
	 cutoff-cost 
	 maximum-visited-state-count 
	 next-search-node
	 count-of-states-considered)
  
  "Returns non-nill iff search for the next candidate should be terminated
   without results.  The returned value is a keyword, indicating the reason
   for terminating search. Returns
   :NO-MORE-CANDIDATES if the search-queue is empty, 
   :VISITED-MAX-STATES if we have examined as many search-nodes 
   as allowed, and
  :CROSSED-CUTOFF if performing best-first search, and the cost of the 
   current search-node, and hence the cost of any candidate to be 
   generated hence forward, is below the cutoff cost.

   Note:  If cutoff-cost or maximum-visited-state-count is null, then
          their respective threshold is ignored.  Likewise, cutoff is
          ignored unless performing best-first search."

  (cond
   ;; no more search-nodes
   ((null next-search-node)
	
    (debug-print-for :bfig t "Cutoff; No more candidates.~%")
    :no-more-candidates)
   
   ((and
	 ;; visited the maximum number of states.
     (not (null maximum-visited-state-count))
     (> count-of-states-considered maximum-visited-state-count))
	
    (debug-print-for :bfig t "  Cutoff: Visited max number states ~a."
					 maximum-visited-state-count)

    :visited-max-states)
   
   ((and (performing-best-first-search? search-problem)
		 
		 ;; Note: will only be called if search-problem is
		 ;; an informed search.
		 (search-node-worse-than-cutoff? search-problem 
										 next-search-node
										 cutoff-cost))

    (debug-print-for :bfig t "  Cutoff: Crossed cutoff cost ~a."
					 cutoff-cost)
    :crossed-cutoff)
   
   (:else
	
	nil)))


;;; Changed.
;;; o Generalized to informed-search-problem.
;;; -- gustavop.
(defmethod search-node-worse-than-cutoff? 

    ((search-problem informed-search-problem)
      ;informed-search-problem-w-conflicts)
	 (search-node informed-search-node)
	 cutoff-cost)
  
  "Returns True if the cost of SEARCH-NODE is guaranteed to be
   worse than the CUTOFF-COST of SEARCH-PROBLEM."
  
  (and
   ;; CUTOFF-COST is specified and
   (not (null cutoff-cost))
   (let ((best-cost (search-node-best-cost search-node)))
     (or 

	  ;; we're searching in decreasing order, 
	  ;; and cost is less than cutoff cost
      (and (eql (decision-space-cost-order-function search-problem) 
				#'>=)
		   
		   (< best-cost cutoff-cost))

	  ;; we're searching in increasing order, 
	  ;; and cost is greater than cutoff cost
      (and (eql (decision-space-cost-order-function search-problem) 
				#'<=)
		   (> best-cost cutoff-cost))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Extracting Suboptimality Conflicts for Fathomed Search Node
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Fix:
;;; o Cost is not correct, per JDAM paper.
(defmethod search-problem-node-suboptimality-conflict

	((search-problem informed-search-problem-w-conflicts)
	 (search-node informed-search-node))
  
  "Returns a sub-optimality conflict for SEARCH-NODE of SEARCH-PROBLEM.

   If the cost of SEARCH-NODE is not better than the incumbent of
   SEARCH-PROBLEM, then returns a minimal subset of the assignments
   of SEARCH-NODE, whose cost is worse than the incumbent.

   If SEARCH-PROBLEM has no incumbent or if search-node has a better
   cost than incumbent, then returns :NONE."

  (if (no-incumbent? search-problem)
	  ;; If no incumbent, THEN no suboptimality conflict.
	  (values :none)

	;; Incumbent, compare to search-node.
	(with-slots ((order-fn cost-order-function) incumbent-node) 
		search-problem
	  (with-slots ((incumbent-cost best-cost)) incumbent-node

		(if (with-slots ((node-cost best-cost)) search-node
			  (funcall order-fn incumbent-cost node-cost))

			;; If SEARCH-NODE is better than or equal to the incumbent, 
			;; Then no suboptimality conflict.

			(values :none)

		  ;; Generate and return a suboptimality conflict.
		  
		  (let ((node-assignments 
				 (search-node-preferred-expansion search-node)))
			
			(suboptimality-conflict-from-assignments
			 search-problem
			 node-assignments
			 incumbent-cost)
			
			))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    Constraint-based A* - conflicts are not considered here
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



;; o Methods to be split:

;; - next-candidate

;; - record-search-node-as-current
;;     o Split for w vs w/o conflicts.  If no conflicts then no
;;     disallowed assignments.

;; - unrecord-search-node-as-current
;; o If no conflicts then no disallowed assignments 

;; - expand-and-queue-current-search-node!
;; o Doesn't need to record current state if no conflicts

;; - preferred-child-node
;; o Doesn't need to record current state if no conflicts

;; - get-new-child-assignments
;; o Doesn't need to record current state if no conflicts

;; - next-most-prerred-sibling-node
;; o Doesn't need to record current state if no conflicts

;; - candidate-found?
;; o Doesn't need to record current state if no conflicts
 


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Methods related to expansion and cost computation in VCSP
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; In order to perform CDA* on Valued CSPs two methods are provided:
;;; (they interface with the rest of the code):
;;;    o expand-and-queue-current-search-node!
;;;      This method creates and enqueues ALL children of the current node.
;;;      It first generates a child (not necessarily the best) and then repeatedly
;;;      asks for the next sibling, so as to exhaustively enumerate all children.
;;;      The following are help methods for expand-and-queue-current-search-node!,
;;;      they exhaustively enumerate the remaining children once the first node is generated.
;;;        o  create-and-enqueue-nodes-for-remaining-children!
;;;        o  next-child-node-to-be-enqueued
;;;
;;;    o compute-search-node-best-cost
;;;      At the moment of creation, this function is called so that the node can be evaluated before
;;;      being enqueued. It estimates the cost of the node using a heuristic described further below.
;;;      The following are help methods for compute-search-node-best-cost:
;;;        o record-search-node-best-extension!
;;;        o soft-cons-admissible-bound
;;;        o tuple-assignment-consistent-with-current-node?
;;;
;;;    
;;;      

;;; Fix:
;;; o Problem in test33 might be related to new update-unassigned-variables
;;; scheme. The :not-specified is misleading the evaluation of a give node.
;;; -- gustavop
(defmethod expand-and-queue-current-search-node!

    ((search-problem informed-search-problem-w-conflicts-vcsp)
	 &optional full-assignment? disjoint?)

  "Expands ALL children of current search-node N of SEARCH-PROBLEM.

   When solving Valued CSPs (i.e. MPI does not hold),
   we can't delay expansion of children (i.e. perform binary expansion,
   as described in JDAM paper). Hence, each time a node is expanded, 
   all children must be generated and evaluated through a heuristic
   function.

   When an unresolved conflict exists, the children of the expanded
   assignment select add each assignment that resolves some conflict
   (that is, the conflicts constitutent kernels).

   If FULL-ASSIGNMENT? is true and all conflicts are resolved, then
   splits on any unassigned variable.

   If DISJOINT? is true, then splitting ensures that sibling search
   nodes denote disjoint sets of assignments (that is, no leaf node
   of two siblings denote the same variable assignment.

   Assumptions:
   o A search node has been recorded as current,
   o The node's child has been recorded as the current search state
     and the unresolved conflicts for this assignment have been identified.
   o The current search node is no longer on the SEARCH-QUEUE,
     and can be destructively modified.
  
   Side Effects:
   o Adds to search queue all children of current search node"
  
  ;; Unlike a Standard OCSP, in which it is enough to expand one child 
  ;; (the "preferred one"),
  ;; here we need to create a new node for each possible child (and then evaluate 
  ;; each node through heuristics.
  
  ;; Also notice that, since we don't delay node expansion here, there is no creation
  ;; or update of sibling node as in the OCSP case.
  
  (let* ((current-node
		 (search-problem-current-search-node 
		  search-problem)))

    (debug-print-for :bfig t "  Expand ~a." 
					 (print-search-node-details current-node nil))
	
	;; For a VCSP, we expand ALL children, since we don't know
	;; beforehand which child will have the best cost estimate
	;;
	;; First we create a sample child, called first-child
	;; Next we exhaustively generate all children, by asking
	;; for the next sibling.
	
	(if (not (process-new-conflicts current-node search-problem))
		(debug-print-for :bfig t "    Discarding entire node ~a" current-node)
	  (block expanding-children
		
		;; To generate the first child, we use the *preferred-child-node* method from the OCSP code.
		;; Even though the function name says "preferred-", here, in the context of VCSPs, the node is not
		;; necessarily the best, it is just the child that happen to be created first.
		(let ((first-child
			   (preferred-child-node search-problem current-node
									 full-assignment? disjoint?)))
		  ;; and queue it.
		  (unless (or (null first-child)
					  (not (test-ancestors-against-resolvable first-child search-problem)))
			(enqueue-search-node! search-problem first-child)
			(debug-print-for :bfig t "    First child = ~a." 
							 (print-search-node-details first-child nil))
			
			;; We need to create and enqueue the remaining children, since the *first-child* node is 
			;; not necessarily the best
			(create-and-enqueue-nodes-for-remaining-children! search-problem 
															  first-child 
															  :disjoint? disjoint?)))))

	;; Now that the child has been expanded, we no longer need
	;; to know the unresolved conflicts corresponding to the 
	;; current node, remove it from the conflict data base.
	
    (unrecord-current-search-node! search-problem :disjoint? disjoint?)))


(defmethod create-and-enqueue-nodes-for-remaining-children! 
	((search-problem informed-search-problem-w-conflicts-vcsp)
	 (first-child-node informed-search-node)
	 &key disjoint?)
  
  "Using previously created first-child node, creates and enqueues a new node for
   each remaining possible assignment in first-node sibling.
	
   Side-effect: Creates and Enqueues one node for each possible assignment in 
   first-node's sibling."
  
  (let ((first-in-the-sibling (next-child-node-to-be-enqueued search-problem
															  first-child-node
															  :disjoint? disjoint?)))
	
	;; Create all the children by asking for the "next-child-node" until there is no more
	(do ((following-child-node first-in-the-sibling (next-child-node-to-be-enqueued 
													 search-problem
													 following-child-node
													 :disjoint? disjoint?)))
		((null following-child-node) 'done)
	  ;; and enqueue each created child
	  (enqueue-search-node! search-problem following-child-node)
	  (debug-print-for
	   :bfig t "    following child in the sibling = ~a." 
	   (print-search-node-details following-child-node nil)))))


;;; The following method is fairly similar to *next-most-preferred-sibling-node*, only
;;  that we are CREATING a new node from the siblings, rather than updating a node to 
;;; get the next best sibling

(defmethod next-child-node-to-be-enqueued
    ((search-problem informed-search-problem-w-conflicts-vcsp)
     (last-enqueued-child-node informed-search-node)
     &key disjoint?)
  
  "Create next node from the remaining possible assignments 
   in the sibling
   
   If DISJOINT? is true, then splitting ensures that sibling search
   nodes denote disjoint sets of assignments.

   Side-effect: CREATES a new node based on the remaining possible
             sibling assignments in LAST-ENQUEUED-CHILD-NODE."
  
  (let* ((ancestor-assignments (search-node-ancestor-assignments
								last-enqueued-child-node))
		 
		 (disallowed-assignments (search-node-disallowed-assignments
								  last-enqueued-child-node))
		 
		 (other-siblings (search-node-lesser-sibling-assignments
						  last-enqueued-child-node))
		 
		 ;; Prune the leading assignments of the remaining siblings 
		 ;; that are ruled out by disallowed assignments.
		 (other-consistent-siblings (trim-leading-inconsistent-assignments
									 other-siblings
									 ancestor-assignments
									 disallowed-assignments)))
	

	(cond ((null other-consistent-siblings)
		   
		   ;; No node is returned if no siblings remain to expand.
		   nil)
		  
		  (:otherwise
		   (let* ((last-child-assignment 
				   (search-node-preferred-child-assignment 
					last-enqueued-child-node))
				  (new-node (make-instance (search-problem-node-type 
											search-problem)
							  :search-problem 
							  (search-node-search-problem last-enqueued-child-node)
							  :ancestor-assignments 
							  (search-node-ancestor-assignments last-enqueued-child-node)
							  :child-assignments 
							  other-consistent-siblings
							  :unassigned-ancestor-variables 
							  (search-node-unassigned-ancestor-variables
							   last-enqueued-child-node)
							  :unassigned-variables
							  :not-specified
							  :disallowed-assignments 
							  (search-node-disallowed-assignments 
							   last-enqueued-child-node)
							  :split-on-conflict? (search-node-split-on-conflict? 
												   last-enqueued-child-node))))
			 
			 ;; Update disallowed assignments of SEARCH-NODE.
			 ;; When the set of leaves of sibling nodes should be disjoint, impose
			 ;; the constraint that the assignment of the last sibling should
			 ;; not appear in the leaves of the remaining siblings.
			 
			 (when (and disjoint?
						(search-node-split-on-conflict? last-enqueued-child-node))
			   
			   ;; Augment the disallowed assignments with the child assignment
			   ;; of the last sibling.
			   (search-node-add-disallowed-assignment! last-child-assignment new-node))
			 new-node)))))




;;
;; - The folling methods are used to compute an admissible bound on the cost
;;   of a node in the search tree of a VCSP.
;;
;; - We compute a very simple bound on the cost of a node:
;;   Let F be the objective function for a VCSP, composed of smaller cost functions
;;   f1, f2, ... fk. In the context of VCSPs, we call these "soft-constraints".
;;   o We bound the overall objective F by bounding each fk *individually*
;;     and then composing these values. Clearly, this is an optimistic estimation, 
;;     hence an admissible bound.
;;   o Given a node, a bound on a soft-constraint (or cost function) fk is simply the 
;;     value of the best tuple that is consistent with the assignments made in the node.
;;
;; - We may view this bound computation as simplistic version of the mini-bucket
;;   heuristics, where we only have a single function inside each bucket. If we want
;;   a more accurate bound, we may place more soft-constraints inside
;;   a bucket. That is, inside a bucket, we would take into account the shared variables between 
;;   the soft-constraints, i.e., we don't maximize or minimize each soft-cons individually.
;;



(defmethod compute-search-node-best-cost
	
    ((search-problem informed-search-problem-w-conflicts-vcsp)
	 (search-node informed-search-node))
  
  " Returns an admissible bound on the best cost of all states 
    represented by search-node: a least upper bound, if 
    maximizing, and a greatest lower bound if minimizing.  

    To compute total cost, we compute an admissible bound on each-soft-constraint.
    This can be viewed as an instance of mini-bucket heuristics where we place a single
    soft-constraint inside each bucket."

  ;; Record each assignment made in search-node in the state-space-variables-ht
  ;; This info will be used to bound a soft-constraint. This is the VCSP counterpart of
  ;; *record-search-node-best-extension!*
  (record-search-node-assignments! search-problem search-node)

  ;; Accumulate a bound on the cost of each variable.
  ;;; = cost-offset + objective-function(assignment-cost1,...assignment-cost)
  
  (let ((cost-offset (decision-space-cost-offset search-problem))
		(identity-element 0))
	(if (equal (decision-space-binary-objective-function search-problem) #'*)
		(setf identity-element 1))		; for #'* the identity element is 1
	
	;; compute the bound for each soft-constraint and compose the bounds
	;; with the objective-functions-binary-operator
	;; finally, add the cost-offset
	(loop 
		with total-soft-cost =  identity-element
		with objective-function = (decision-space-binary-objective-function
								   search-problem)
								  
		for soft-constraint in (opsat-problem-soft-constraint-system search-problem)
		for soft-cost = (soft-cons-admissible-bound soft-constraint search-problem)
						
						
		do (setq total-soft-cost 
			 (funcall objective-function total-soft-cost soft-cost))
		finally (return (+ total-soft-cost cost-offset)))))


(defmethod soft-cons-admissible-bound 
    ((soft-constraint soft-constraint)
     (search-problem informed-search-problem-w-conflicts-vcsp))
  
  "Returns an admissible bound on the cost of all leaves under current search-node. 
   Basically, it returns the best-rewarding tuple-assignment that is consistent with
   the assignments in the current search node."
  
    (loop for tuple-assignment in (tuples-of-soft-constraint soft-constraint)
								  
		when  (tuple-assignment-consistent-with-node? search-problem 
															  tuple-assignment 
															  (soft-constraint-scope soft-constraint))
		return (metric-value tuple-assignment)
			   
			   ;; If no tuple-assignment within the soft-constraint is consistent with the node assignments,
			   ;; there must be an error with the provided the soft-constraint. For example,
			   ;; some tuples were not listed.
		finally (error "Problem with soft-constraints tuples:
                        No tuple in soft-constraint with scope ~a is consistent with assignments in node" 
					   (soft-constraint-scope soft-constraint))))


(defmethod tuple-assignment-consistent-with-node?  
    ((search-problem informed-search-problem-w-conflicts-vcsp)
     (tuple-assignment soft-tuple-assignment)
     scope)
  
  "Tests consistency of assignments in search-node under consideration against
   tuple-assignment."
  
 
  
  ;; for each variable in the scope of the related soft-constraint,
  ;;   check if the assignment in the tuple is consistent with the assignment in the node
  (loop for variable in scope
      for index from 0
					 
					 ;; an inconsistency happens if:
					 ;; - the variable has been assigned in the node under consideration 
					 ;;   (i.e. assignment != nil)
					 ;; AND
					 ;; - the assignment in the tuple is not NIL 
					 ;;   (remember NIL means don't care)
					 ;; AND
					 ;; - the assignment in the variable is different from the 
					 ;; assignment in the tuple
      when (and (not (null
					  (gethash variable 
							   (state-space-variables-ht search-problem))))
				(not (null (nth index (assignments-of-soft-tuple tuple-assignment))))
				(not (equal 
					  (gethash variable 
							   (state-space-variables-ht search-problem))
					  
					  (nth index (assignments-of-soft-tuple tuple-assignment)))))
	  return nil
      finally (return T)))


(defmethod record-search-node-assignments!
	
    ((search-problem  informed-search-problem-w-conflicts-vcsp)
	 (search-node informed-search-node))

  "Sets the decision variables of SEARCH-PROBLEM to the assignments 
   made in SEARCH-NODE."
  
  ;; The assignments made in SEARCH-NODE are set in the hash-table
  ;; state-space-variables-ht. This info will be used when
  ;; estimating the cost of a node.
  ;; This method is the VCSP counterpart of *record-search-node-best-extension!*
  
  (loop for variable in (state-space-variables search-problem)
	
						;; First, clear any previous assignment in the hashtable.				
      do (setf (gethash (state-variable-name variable) (state-space-variables-ht search-problem)) 
		   nil))
  
  ;; We then overwrite the test assignment of every assigned variables,
  ;; as specified in each ancestor-assignment plus ...
  (loop for new-assignment in
		(search-node-ancestor-assignments search-node)
      for variable = (state-assignment-variable new-assignment)
      do (setf (gethash (state-variable-name variable) (state-space-variables-ht search-problem)) 
		   (state-assignment-value new-assignment)))
  
  ;; ... the best child assignment.
  (let ((new-assignment
		 (search-node-preferred-child-assignment search-node)))
    (unless (null new-assignment)

	  (setf (gethash (state-variable-name (state-assignment-variable new-assignment)) 
					 (state-space-variables-ht search-problem)) 
		(state-assignment-value new-assignment))))
  T)






