;;; -*- Mode:Common-Lisp; Package:BFIG; Base:10 -*-

;;; Copyright (c) 1999 - 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :BFIG)

;;;(eval-when (compile load eval)
;;;  (export '(
;;; 		   add-variable
;;;	 		   add-objective-function!
;;;
;;;			   externalize-state-assignments
;;;            )
;;;		  :BFIG)
;;;  )

;;; Fix:
;;; o Having NEW-INFORMED-SEARCH-PROBLEM maximize cost by default is not
;;;   what is expected, and causes bugs.  Either change to minimize, and
;;;   adjust all callers, OR change the name "cost" to "utility."

;;; To Do:
;;; o Do we want to split into problems with and without conflicts?
;;; o change the externalize functions to a generic external-description
;;;   function for all systems (look at other files in ITC, ISAT ... for
;;;   the best name).

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Framing a Best-first Kernel (Implicant) Generation Problem
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This file contains functions for building an informed search problem, 
;;; whose solution employs forward, conflict-directed search.  
;;; This forward search process includes best first kernel generation.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; *** Setting Up A Search Problem:

;;; (NEW-INFORMED-SEARCH-PROBLEM
;;;     DECISION-VARIABLE-DESCRIPTIONS CONFLICT-DESCRIPTIONS
;;;     &OPTIONAL &KEY
;;;			   (NAME (GENTEMP "INFORMED-SEARCH-PROBLEM-"))
;;;            (use-conflicts? T)
;;;			   (SOURCE NIL)
;;;			   (SEARCH-ORDER :BEST-FIRST)
;;;			   (BINARY-OBJECTIVE-FUNCTION #'+)  
;;;			   (COST-OFFSET 0)
;;;			   (COST-ORDER-FUNCTION #'>=))
;;;  
;;;   Creates and returns a combinatorial optimization problem with 
;;;   conflicts as constraints.  Defaults to a problem that 
;;;   maximizes (minimizes?) the sum of the decision variable assignment costs.  
;;;   The objective function can be modified by calling add-objective-function!.
;;;
;;;   SEARCH-ORDER is a keyword, specifying the order in which the
;;;   candidate is explored.  SEARCH-ORDER is one of 
;;;   *ALLOWED-SEARCH-ORDER-TYPES*.
;;;
;;;   USE-CONFLICTS?  A boolean.  If true, then perform conflict-directed 
;;;   search.

;;;   DECISION-VARIABLE-DESCRIPTIONS is a list:
;;;     (VARIABLE-NAME (VALUE :COST COST)*)
;;;    where there is a value/cost description for each value in the domain of 
;;;    variable-name.
;;;
;;;   The multi-attribute objective function of the optimization problem 
;;;   is defined over the decision variables.  The attribute cost of
;;;   each decision variable value is specified in DECISION-VARIABLE-DESCRIPTIONS.
;;;
;;;   The composition of attribute costs specified jointly by 
;;;   BINARY-OBJECTIVE-FUNCTION, COST-OFFSET and COST-ORDER-FUNCTION. 
;;;   BINARY-OBJECTIVE-FUNCTION specifies how attribute costs are pair-wise
;;;   combined, and defaults to +.  This function must satisfy preferential
;;;    independence.
;;;  COST-OFFSET specifies a constant value to be added to all costs.
;;;  COST-ORDER-FUNCTION places a total order on values, and is used to
;;;    define the best value.  The default, >= , corresponds to a solution
;;;    that maximizes cost (utility).
;;;
;;;   CONFLICT-DESCRIPTIONS is a list of conflicts, where each
;;;    conflict is a partial assignment to decision variables that
;;;    is inconsistent.
;;;
;;;   SOURCE is an object that invoked the creation of the problem.
;;;  
;;;   NAME is an sexpression naming the problem.

;;;   (MAKE-INSTANCE 'SEARCH-PROBLEM-W-CONFLICTS
;;;       &KEY SOURCE (NAME (GENTEMP "SEARCH-PROBLEM-"))
;;;      (SEARCH-ORDER :BEST-FIRST))
;;;   Creates a new search-problem named NAME, created by SOURCE.
;;;   SOURCE is the object that created search problem.
;;;   NAME is a string.
;;;   SEARCH-ORDER is a keyword, specifying the order in which the
;;;   candidate is explored.  SEARCH-ORDER is one of 
;;;   *ALLOWED-SEARCH-ORDER-TYPES*.
  
;;; (ADD-OBJECTIVE-FUNCTION! (DECISION-SPACE DECISION-SPACE)
;;;    &OPTIONAL
;;;    (BINARY-OBJECTIVE-FUNCTION #'+)
;;;    (COST-OFFSET 0)
;;;    (COST-ORDER-FUNCTION #'>=))
;;;
;;;  Adds cost function to a search problem.
;;;  SEARCH-PROBLEM is a search problem object.
;;;  BINARY-OBJECTIVE-FUNCTION maps from pairs of numbers to numbers,
;;;    and is used to combine attribute cost of the individual
;;;    variable assignments.
;;;  COST-OFFSET is a constantnumber, and is added to all costs.
;;;  COST-ORDER-FUNCTION is used to sort the cost of a conflict from best 
;;;  to worst.  Use #'<= for increasing order and #'>= for decreasing order.

;;; (ADD-VARIABLE (STATE-SPACE STATE-SPACE) 
;;;    VARIABLE-NAME DOMAIN-DESCRIPTIONS)
;;;
;;;   Creates a new variable for STATE-SPACE with the 
;;;   specified name.  Variable will be of the type specified
;;;   in the VARIABLE-TYPE slot of STATE-SPACE.
;;;
;;;   VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
;;;    supplied to make-instance of variable.  It currently consists of
;;;    one required keyword :DOMAIN and one optional keyword :GUARD:
;;;
;;;      (:DOMAIN (DOMAIN-ELEMENT-DESCRIPTION+) 
;;;       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *always-active*])
;;;
;;;    where each DOMAIN-ELEMENT-DESCRIPTION is either an atom, denoting 
;;;    a domain element, or list of the form:
;;;
;;;      (DOMAIN-ELEMENT <KEYWORD-VALUE-PAIR>*)
;;;  
;;;   where DOMAIN-ELEMENT is the value of the assignment, and  
;;;   KEYWORD-VALUE-PAIRS correspond to the remaining keyword 
;;;   arguments to make-instance for ASSIGNMENT-TYPE of the VARIABLE class.
;;;   This currently consists of :COST.
;;;
;;;   :GUARD is optional, and denotes an assignment that
;;;   activates variable.
;;;
;;;   Note: If variable is of type decision-variable, 
;;;   then this function ensures that variable's domain is placed in best 
;;;   first order."

;;; (EXTERNALIZE-STATE-ASSIGNMENTS STATE-ASSIGNMENTS)
;;;  
;;;   Given a list of STATE-ASSIGNMENT objects,
;;;   Returns the external assignment description 
;;;   for each object, in the form:
;;;     (= variable-name value).

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Additional Decision-Space Module Interface Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (INTERNALIZE-STATE-ASSIGNMENTS
;;;     (SEARCH-PROBLEM  SEARCH-PROBLEM) STATE-ASSIGNMENT-DESCRIPTIONS)
;;;  
;;;   Maps a list of STATE-ASSIGNMENT-DESCRIPTIONS to
;;;   their corresponding state-assignment objects. An assignment
;;;   description is an sexpression of the form:
;;;      (= variable-name value) 
;;;   or a state-assignment object.  
;;;   When a state-assignment object is provided, it is simply passed through.

;;; (TRIM-LEADING-INCONSISTENT-ASSIGNMENTS
;;;    ASSIGNMENTS STATE-ASSIGNMENTS &OPTIONAL DISALLOWED-ASSIGNMENTS)
;;;
;;;   Takes as input ASSIGNMENTS,STATE-ASSIGNMENTS and DISALLOWED-ASSIGNMENTS, 
;;;   each of which is a list of state-assignment objects.
;;;   Returns ASSIGNMENTS with all LEADING assignments
;;;   removed that are inconsistent with STATE- or DISALLOWED-ASSIGNMENTS.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:

;;; o Extend to conditional CSP.
;;;   - See comments on functions to be modified below.

;;; o Some calls to constraint system using search-problem-*, 
;;;   in functions like add-variable, should be migrated to OPSAT.

;;; o New-search-problem was broken into a set of add functions. 
;;;   Document these as interface functions.  What function creates 
;;;   a search-problem object?  What function inserts a conflict?

;;; Check:

;;; o Should CONSISTENT-DECISION-ASSIGNMENTS? be part of the interface?

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface Function Definitions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Formulating an Informed Search Problem
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Consider adding this functionality to make-instance for
;;;   informed-search-problem.  Then the make-instance for
;;;   opsat problem etc must be changed.

;;; o Check decision-variable-descriptions to make sure that
;;;   keywords are all legal.

;;; o Below initialize is called twice, once in the following function,
;;;   and once by make-instance.  However, deleting initialize in the
;;;   following function breaks the function.  Collapse to one call
;;;   to initialize.

;;; Changed:
;;; o value-cost-alist -> variable-keyword-list
;;; o added guard.

;;; Fix:
;;; o Calls initialize-candidate-generator!
;;;   The fix in the comments didn't work.  Try again.

;;; o Document and the allowed values of search-order.
;;;   Have code check that search-order is legal.

(defun new-informed-search-problem

    (decision-variable-descriptions
     conflict-descriptions
     &optional &key
			   ;; Defaults to testing consistency on
			   ;; the base constraints alone.
			   
			   (name (gentemp "informed-search-problem-"))
			   (use-conflicts? t)
			   (conditional? nil)
			   (source nil)
			   (search-order :best-first)
			   (binary-objective-function #'+)  
			   (cost-offset 0)
			   (cost-order-function #'>=))
  
  "Creates and returns a combinatorial optimization problem with 
   conflicts as constraints.  Defaults to a problem that 
   maximizes the sum of the decision variable assignment costs.  

   The objective function can be modified by calling ADD-OBJECTIVE-FUNCTION!.

   USE-CONFLICTS?  
     A boolean.  If true, then performs forward, conflict-directed search.

   CONDITIONAL?  
     A BOOLEAN.  If true, then variables are guarded by assignments,.
     A guard is specified within a variable description by 
       :GUARD ' (= <VAR> <VAL>).  
     Use :guard *always-active* (= *true*) to specify a variable 
     that is permanently active.

   SEARCH-ORDER specifies the order in which search tree nodes 
     (partial assignments) are visited.  Defaults to :BEST-FIRST search.
     can be one of ???.

   DECISION-VARIABLE-DESCRIPTIONS is a list:
     (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)
 
    where variable-keyword-list is:
      (:DOMAIN ((VALUE :COST COST)*) 
       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *always-active*])
    where there is a value cost description for each value in the domain of 
    variable-name.  :GUARD is optional, and denotes an assignment that
    activates variable.

   The multi-attribute objective function of the optimization problem 
   is defined over the decision variables.  The attribute cost of
   each decision variable value is specified in DECISION-VARIABLE-DESCRIPTIONS.

   The composition of attribute costs is specified jointly by 
   BINARY-OBJECTIVE-FUNCTION, COST-OFFSET and COST-ORDER-FUNCTION. 

   BINARY-OBJECTIVE-FUNCTION specifies how attribute costs are pair-wise
   combined, and defaults to +.  This function must satisfy preferential
    independence.

  COST-OFFSET specifies a constant value to be added to all costs.

  COST-ORDER-FUNCTION places a total order on values, and is used to
    define the best value.  The default, >=, corresponds to a solution
    that maximizes cost (utility).

   CONFLICT-DESCRIPTIONS is a list of conflicts, where each
    conflict is a partial assignment to decision variables that
    is inconsistent.

   SOURCE is an object that invoked the creation of the problem.
  
   NAME is an sexpression naming the problem."
  
  (let* ((problem-type 
		  (if use-conflicts?
			  
			  (if conditional?
				  'informed-conditional-search-problem-w-conflicts
				'informed-search-problem-w-conflicts)
			
			(if conditional?
				'informed-conditional-search-problem
			  'informed-search-problem)))
		 
		 (problem (make-instance problem-type
					:name name
					:source source
					:search-order search-order
					)))
  
	(add-objective-function! problem 
							 binary-objective-function
							 cost-offset
							 cost-order-function)

	;; Create and add the decision variables to the problem,
	;; along with associated domains and valuations (utilities).
	(loop for (variable-name . variable-keyword-list) 
		in decision-variable-descriptions
		do (add-variable 
			problem variable-name variable-keyword-list))

	;;Set up the bit representation for conflicts.
	(initialize-bit-ht! problem)
	
	(loop for conflict in conflict-descriptions
		do (insert-new-conflict problem
								:inconsistent-assignments
								conflict))

	;; To Do: Initialize is called twice, once below and once
	;;        by make-instance.  However, deleting the following
	;;        line breaks the function.  Collapse to one call
	;;        to initialize.
	(initialize-candidate-generator! problem
									 :search-order 
									 :DEFAULT)
	problem))

;;; To Do:
;;; o ??? Rename to add-decision-variable, then combine with 
;;;   add-decision-variable in opsat using method combination????

(defmethod add-variable 

	((state-space state-space)
	 variable-name variable-keyword-plist)

  "Creates a new variable for STATE-SPACE with the 
   specified name.  Variable will be of the type specified
   in the VARIABLE-TYPE slot of STATE-SPACE.

   VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of variable.  It currently consists of
    one required keyword :DOMAIN and one optional keyword :GUARD:

      (:DOMAIN (DOMAIN-ELEMENT-DESCRIPTION+) 
       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *always-active*])

    where each DOMAIN-ELEMENT-DESCRIPTION is either an atom, denoting 
    a domain element, or list of the form:

      (DOMAIN-ELEMENT <KEYWORD-VALUE-PAIR>*)
  
   where DOMAIN-ELEMENT is the value of the assignment, and  
   KEYWORD-VALUE-PAIRS correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of the VARIABLE class.
   This currently consists of :COST.

   :GUARD is optional, and denotes an assignment that
   activates variable.

   Note: If variable is of type decision-variable, 
   then this function ensures that variable's domain is placed in best 
   first order."
  
  (apply 'make-instance (state-space-variable-type state-space)
		 :name variable-name
		 :state-space state-space
		 variable-keyword-plist)
  )

;;; To Do:
;;; o Define an objective-function class with an appropriate set of
;;;   interface functions.  Then perform the following addition
;;;   via make-instance.
(defmethod add-objective-function!

	((decision-space decision-space)
	 &optional
	 (binary-objective-function #'+)
	 (cost-offset 0)
	 (cost-order-function #'>=))

  "Adds cost function to SEARCH PROBLEM.
   DECISION-SPACE is a search problem object.
   BINARY-OBJECTIVE-FUNCTION maps pairs of numbers to numbers,
     and is used to combine attribute costs of the individual
     variable assignments.
   COST-OFFSET is a constant number, and is added to all costs.
   COST-ORDER-FUNCTION is used to sort a set of assignments from
   best cost to worst cost.  
   Use #'<= for increasing order and #'>= for decreasing order."
  
  (setf (decision-space-binary-objective-function decision-space)
    binary-objective-function)

  (setf (decision-space-cost-offset decision-space)
    cost-offset)

  (setf (decision-space-cost-order-function decision-space)
    cost-order-function)

  ;; Place the domain of each variable in best first order, based
  ;; on the newly specified cost function.
  ;; Also update the best and current assignment of each variable to be 
  ;; the current best element of their respective domain.
  (loop for variable in (state-space-variables decision-space)
      do (variable-order-domain! decision-space variable))
  
  decision-space)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  Exporting Decision Assignments
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Make this a method on state-space.
(defun externalize-state-assignments 
	(state-assignments)
  
  "Given a list of STATE-ASSIGNMENT objects,
   returns a list of corresponding external assignment 
   descriptions, one for each object.  This is of the form:
     (= variable-name value)." 
  
  (loop for state-assignment in state-assignments
      collect (state-assignment-description state-assignment)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Problem Module Interface Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Manipulating Decision Variable Assignments
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Can state assignments and variables be merged with 
;;;   CSP variables and assignments?

;;; Note:
;;; o This is split into three functions in order to support
;;;   conditional state spaces.
(defmethod internalize-state-assignments

    ((state-space state-space) assignment-descriptions)
  
  "Returns a list of assignment objects in state-space, 
   corresponding to ASSIGNMENT-DESCRIPTIONS.  This includes 
   assignments explicitly listed and assignments implied, 
   for example, through implicit variables.

   An assignment description is an sexpression of the form:

      (= variable-name value) 

   or an assignment object.  When an ssignment object is provided, 
   it is simply passed through."

  (make-implicit-assignments-explicit
   state-space
   (internalize-state-assignments-explicitly-listed 
	state-space assignment-descriptions)))

;;; Changed:
;;; o Added
(defmethod internalize-state-assignments-explicitly-listed

    ((state-space state-space) assignment-descriptions)
  
  "Converts a list of ASSIGNMENT-DESCRIPTIONS to
   their corresponding assignment objects.

   RETURNS a list of assignment objects.

   An assignment description is an sexpression of the form:

      (= variable-name value) 

   or an assignment object.  
   When an assignment object is provided, it is simply passed through."

  (with-slots (assignment-ht) state-space
	
	(loop
		for assignment in assignment-descriptions
						  ;; Internalize assignments, if necessary.
		for assignment-object = (if (typep assignment 'state-assignment)
									assignment
								  (get-assignment-from-hash 
								   assignment 
								   assignment-ht))

		when (null assignment-object)

		do (error "Assignment ~a  doesn't specify a value in variable's domain"
				  assignment)

		collecting assignment-object)))

(defmethod internalize-state-assignment

    ((state-space state-space) assignment-description)
  
  "Converts an ASSIGNMENT-DESCRIPTION to
   its corresponding assignment object. An assignment
   description is an sexpression of the form:

      (= variable-name value) 

   or an assignment object.  
   When an ssignment object is provided, it is simply passed through."

  (with-slots (assignment-ht) state-space

	;; Internalize assignment if necessary.
	(let ((assignment-object 
		   
		   (if (typep assignment-description 'state-assignment)
			   
			   assignment-description
			 
			 (get-assignment-from-hash 
			  assignment-description
			  assignment-ht))))

	  (when (null assignment-object)
		(error "Assignment ~a is not (yet) defined."
			   assignment-description))

	  assignment-object)))

;;; Changed:
;;; o Added
(defmethod make-implicit-assignments-explicit

    ((state-space state-space) assignments)
  
  "Given a list of ASSIGNMENT objects,
   Returns this list, augmented by any
   assignments implied.

   An unconditional STATE-SPACE has no implicit
   assignments."
  
  assignments)

;;; To Do:
;;; o Make this a method on state-space.
(defun trim-leading-inconsistent-assignments
    (assignments state-assignments &optional (disallowed-assignments '()))

  "Takes as input ASSIGNMENTS, STATE-ASSIGNMENTS and DISALLOWED-ASSIGNMENTS, 
   each of which is a list of state-assignment objects.
   Returns a suffix of ASSIGNMENTS with all LEADING assignments
   removed that are inconsistent with either STATE-ASSIGNMENTS
   or DISALLOWED-ASSIGNMENTS."
  
  (loop 
      for remaining-assignments on assignments
      for first-assignment = (first remaining-assignments)
							 ;; Stop as soon as we find an
							 ;; assignment that is consistent 
							 ;; with STATE-ASSIGNMENTS.
							 
      until (and (or (null state-assignments)
					 (consistent-and-distinct-with-assignments?
					  first-assignment state-assignments))
				 (not (member first-assignment disallowed-assignments)))

      finally (return remaining-assignments)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Installing Decision Variable Domains
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod make-variable-domain
    ((state-space state-space) 
	 (variable state-variable)
	 value-descriptions)

  "Takes as input a STATE VARIABLE of STATE-SPACE, whose
   (valued) domain is described by VALUE-DESCRIPTIONS.
   VALUE-DESCRIPTIONS is a list of elements, each of
   which describes one value.  

   Each VALUE-DESCRIPTION is either
   an atom, denoting a value, or list of the form:
   (VALUE . <keyword-value-pair>*)  where VALUE is the
   value of the assignment, and the keyword-value-pairs
   correspond to the remaining keyword arguments to 
   make-instance for ASSIGNMENT-TYPE of VARIABLE.

   RETURNS a new list of assignment objects for VARIABLE, 
   that corresponds to VALUE-DESCRIPTIONS.

   Note: This list may be destructively modified; 
   for example, by sort."
      
  ;; create the assignments and add them to the assignment 
  ;; lookup hashtable.
  (loop 
      with ht = (state-space-assignment-ht state-space)
      with variable-name = (state-variable-name variable)

      for value-description in value-descriptions
      for assignment = (make-variable-assignment
						variable value-description)
	  for value = (state-assignment-value assignment)
				  
      do (setf (get-assignment-from-hash
				`(= ,variable-name ,value) ht)
		   assignment)
		 
      collecting assignment))

(defmethod make-variable-assignment 
	((variable state-variable)
	 value-description)
  
  "Given VARIABLE and VALUE-DESCRIPTION, returns
   an object denoting an assignment of variable to value.

   VALUE-DESCRIPTION is either an atom, denoting a value, 
   or list of the form:

   (VALUE . <keyword-value-pair>*)
  
   where VALUE is the value of the assignment, and the 
   keyword-value-pairs correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of VARIABLE."

  (let ((assignment-type
		 (state-variable-assignment-type variable)))
	
	(if (atom value-description)

		(make-instance assignment-type
		  :variable variable
		  :value value-description)
	  
	  (destructuring-bind (value . keyword-value-pairs)
		  value-description
		
;;;		(format t "~% value = ~a, keyword pairs = ~a"
;;;				value keyword-value-pairs)
		
		(apply 'make-instance
			   `(,assignment-type
				 :variable ,variable
				 :value ,value
				 . ,keyword-value-pairs)))
	  )))

;;; Check:
;;; o Before throwing away the assignments of a variable's 
;;;   existing domain, do we need to clear any conflicts they 
;;;   resolve, and any conflicts supported by them?

;;; o When a new domain is installed, can the assignments
;;;   already be mentioned by conflicts?  If so, when do we
;;;   link the conflicts to their corresponding assignment?

(defmethod state-variable-install-domain! 
	((variable state-variable) domain)
  
  " Connects a previously defined VARIABLE object to DOMAIN,
    a previously defined list of assignments object."

  ;; Point each assignment to its variable, and
  ;; clear any previous alternate assignments.
  (loop for assignment in domain

      do (setf (state-assignment-variable assignment) 
		   variable)
		 (setf (state-assignment-alternate-assignments assignment)
		   nil))

  ;; Point variable to its new domain.
  (setf (state-variable-domain variable) domain)
  T)

(defmethod state-variable-install-domain! :after
		   ((variable state-variable-w-conflicts) domain)
  
  " Connects a previously defined VARIABLE object to DOMAIN,
    which is a previously defined list of assignments object.

    Sets up the consistutent kernels and conflicts involving
    assignments in domain."

  ;; Put in pointers from each assignment to other members of its 
  ;; new domain.  This is used to generate constitutent kernels.
  (loop for val1 in domain
					
      do (loop for val2 in domain
						   
			 unless (eql val1 val2)
					
			 do (push val2 (state-assignment-alternate-assignments val1)))

		 ;; Clear the conflicts resolved by each assignment,
		 ;; or containing each assignment.
		 ;; These are used to identify unresolved conflicts.
		 (setf (state-assignment-resolved-conflicts val1) nil)
		 (setf (state-assignment-conflicts val1) nil))
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Comparing Variable Assignment Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o For conditional CSPS
;;;   - check to see if this changes.
(defmethod consistent-and-distinct-with-assignments? 
	((assignment state-assignment) assignments)
  
  "Let ASSIGNMENT be a decision assignment object and ASSIGNMENTS
   be a list of decision assignment objects.

   Returns True iff ASSIGNMENT is consistent with ASSIGNMENTS,
   and is not contained in ASSIGNMENTS."
  
  (loop for assignment2 in assignments
	  always (consistent-distinct-assignments?
			  assignment
			  assignment2)))

(defmethod consistent-distinct-assignments? 

	((assignment1 state-assignment) 
	 (assignment2 state-assignment))
  
  "Let ASSIGNMENT1 and ASSIGNMENT2 be state-assignment objects.

   Returns TRUE iff ASSIGNMENT1 and 2 are consistent and distinct."
  
  ;; The two assignments must be to different variables -
  ;; Two assignments are indistinct if they assign the same variable
  ;;   the same value.  
  ;; Two assignments are mutually INconsistent if they assign different 
  ;;   to the same variable.
  
  (not (eql (state-assignment-variable assignment1)
			(state-assignment-variable assignment2))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Ordering Decision Variable Domains
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod variable-order-domain!
	((decision-space decision-space)
	 (variable decision-variable))

  "Destructively places VARIABLE's list of domain assignments 
   in best first order, according to the current cost-order 
   function of DECISION-SPACE.  Updates VARIABLE's best assignment 
   based on this ordering."
   
  (let ((order-fn (decision-space-cost-order-function decision-space)))

	(unless (null order-fn)
	  
	  ;; Leave unordered if no order function.
	  (let* ((domain (state-variable-domain variable))
			 (sorted-domain 
			  (sort domain order-fn :key #'decision-assignment-cost))
			 (best-assignment (first sorted-domain)))

		;; Destructively modifies the list of assignments denoting
		;; variable-domain.  This is okay, since make-decision-variable-
		;; domain creates a new list of assignments.
		(setf (state-variable-domain variable) sorted-domain)
		(setf (decision-variable-best-assignment variable) best-assignment)
		
		T))))

(defmethod variable-ordered-domain 
	((variable state-variable))

  "Returns' VARIABLE's domain as a list of assignments
   that is sorted from most to least preferred assignment
   to split on.

   Presumes that the domain was created
   using variable-order-domain!"
  
  ;; Variable's domain is placed in the preferred order of
  ;; splitting when it is created; hence
  ;; we simply return this sorted domain.
  (with-slots (domain) variable
	domain))

(defmethod variable-ordered-domain 
	((variable decision-variable))

  "Returns' VARIABLE's domain as a list of assignments
   that is sorted from best to worse cost.

   Presumes that the domain was created
   using variable-order-domain!"
  
  ;; Variable's domain is sorted in order, either when it is created
  ;; or whenever a cost function is specified or modified.  Hence
  ;; we simply return this sorted domain.
  (with-slots (domain) variable
	domain))
