;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package :mtk-csp-tcsp)

;;; For references relating to this code see:
;;;   [1] R. Dechter. Constraint Processing. Morgan Kaufmann, San Francisco,
;;;   California, 2003.

(defclass stn-with-distance-graph (stn)
  ((distance-graph :type weighted-directed-graph
		   :accessor get-distance-graph
		   :initform (make-instance 'weighted-directed-graph
			       :default-weight (make-inf-upper-bound))
		   :documentation "")
   (distance-graph-edges-ht-edge :type hash-table
				 :accessor get-distance-graph-edges-ht-edge
				 :initform (make-hash-table)
				 :documentation "")))

(defmethod get-lower-and-upper-bound-edges ((e_i directed-edge)
					    (stn stn-with-distance-graph))
  (gethash e_i (get-distance-graph-edges-ht-edge stn)))

(defmethod mtk-graph-algorithms:floyd-warshall ((stn stn-with-distance-graph)
						&key
						(add-function #'bound+)
						(min-function #'bound-min)
						(inf (make-inf-upper-bound)))
  " FLOYD-WARSHALL(W) 
    1 n <- rows[W] 
    2 D^{(0)} <- W 
    3 for k <- 1 to n 
    4   do for i <- 1 to n 
    5     do for j <- 1 to n 
    6       do d^{(k)}_{ij} <- min(d^{(k-1)}_{ij},d^{(k-1)}_{ik}
                                                  + d^{(k-1)}_{kj})
    7 return D^{(n)"
  (let ((g (mtk-csp-tcsp:get-distance-graph stn)))
    (call-next-method g
		      :weight-function #'(lambda (e_i) (get-weight e_i g))
		      :add-function add-function
		      :min-function min-function
		      :inf inf)))

;;-------------------------------------------------------------------------------
;; STN to distance graph
;;-------------------------------------------------------------------------------

(defmethod stn-to-distance-graph ((stn stn-with-distance-graph))
  (get-distance-graph stn))

;;-------------------------------------------------------------------------------
;; Delete vertex/vertices from a stn
;;-------------------------------------------------------------------------------

(defmethod delete-variable (variable (n stn-with-distance-graph))
  (delete-vertex variable (get-distance-graph n))
  (call-next-method variable n))

(defmethod delete-vertex (vertex (n stn-with-distance-graph))
  (delete-variable vertex n))

;;-------------------------------------------------------------------------------
;; Add constraint to a STN
;;-------------------------------------------------------------------------------

(defmethod add-constraint ((c_i constraint) (stn stn-with-distance-graph))
  ;; Only binary temporal constraints are supported.
  (assert (eq (length (get-scope c_i)) 2))
  ;; Disjunctive intervals are not allowed in a STN.
  (assert (eq (length (get-relation c_i)) 1))
  (multiple-value-bind (out-c_i e_i) (call-next-method c_i stn)
    (let* ((g (get-distance-graph stn))
	   (interval (first (get-relation c_i)))
	   (e_ub (add-edge (get-scope c_i) g))
	   (ub (get-upper-bound interval))
	   (e_lb (add-edge (reverse (get-scope c_i)) g))
	   (lb (make-bound (* -1 (get-value (get-lower-bound interval)))
			   :inequality
			   (if (eql (get-inequality (get-lower-bound interval))
				    #'>=)
			       #'<= #'<))))
      (set-weight e_ub ub g)
      (set-weight e_lb lb g)
      (setf (gethash e_i (get-distance-graph-edges-ht-edge stn))
			(list e_lb e_ub))
      (values out-c_i e_i (list e_lb e_ub)))))

;;-------------------------------------------------------------------------------
;; Delete temporal constraint(s) from a STN
;;-------------------------------------------------------------------------------

(defmethod delete-constraint ((c_i constraint) (stn stn-with-distance-graph))
  (multiple-value-bind (c_i e_i) (call-next-method c_i stn)
    (remhash e_i (get-temporal-constraint-ht-edge stn))
    (let* ((g (get-distance-graph STn))
	   (lb-ub (get-lower-and-upper-bound-edges e_i stn))
	   (e_lb (first lb-ub))
	   (e_ub (second lb-ub)))
      (delete-weighted-edge e_lb g)
      (delete-weighted-edge e_ub g)
      (remhash e_i (get-distance-graph-edges-ht-edge stn))
      (values c_i e_i lb-ub))))