;;;; Copyright (c) 2012 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the oftware.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Andrew Wang

(in-package :mtk-csp-tcsp)


;;;-------------------------------------------------------------------------------
;;; Duration CLASS (and subclasses)
;;;   * duration
;;;     * uncontrollable-duration
;;;       * probabilistic-duration
;;;     * controllable-duration <--> temporal-constraint
;;;-------------------------------------------------------------------------------

(defclass duration ()
  ((s :type list
      :accessor get-scope
      :initarg :scope
      :initform nil
      :documentation "The scope, a list of start and end time points.")
   (r :accessor get-relation
      :initarg :relation
      :documentation "The relation, charaterizing the feasible and/or allowable difference
                      between the start and end time points.")))

(defmethod print-object ((d_i duration) stream)
  (with-slots ((s s)
               (r r)) d_i
    (format stream "(:scope ~A :relation ~A)" s r)))

;;; There are varying degrees of controllability.  Not all are implemented here.
;;; See Vidal's paper titled 'Controllability characterization and checking in
;;; Contingent Temporal Constraint Networks.'"

;; (defclass controllable-duration (duration)
;;   ((r :type interval
;;       :documentation "A duration whose length you can control to some extent.")))

;; (defclass uncontrollable-duration (duration)
;;   ((r :type interval
;;       :documentation "A duration whose length is selected by Nature upon activation,
;;                       but is nominally guaranteed to be within an interval.")))

;; (defclass probabilistic-duration (uncontrollable-duration)
;;   ((r :type random-variable
;;       :documentation "An uncontrollable duration whose length is determined by a
;;                       probabilistic model.")))

(defun make-uncontrollable-duration (&key ((:scope scope)) ((:relation intervals)))
  ;; Must provide the scope and the intervals
  (assert (and scope intervals))
  ;; Only binary durations are supported.
  (assert (eq (length scope) 2))
  ;; Only simple intervals (non-disjunctive) are supported.
  (assert (eq (length intervals) 1))
  (let ((relation nil))
    (dolist (i_i intervals)
      (push (apply #'make-interval i_i) relation))
    (make-instance 'duration :scope scope :relation relation)))


;;;-------------------------------------------------------------------------------
;;; Durations CLASS
;;;-------------------------------------------------------------------------------

(defclass durations ()
  ((d-list :type list
           :accessor get-duration-list
           :initform nil
           :documentation "The list of durations.")
   (d-ht-time-point :type hash-table
                  :accessor get-durations-ht-time-point
                  :initform (make-hash-table)
                  :documentation "Look up which durations a time-point is involved in.")))

(defmethod print-object ((d durations) stream)
  (format stream "~A" (get-duration-list d)))


;;;-------------------------------------------------------------------------------
;;; STN with Uncertainty (STNU) CLASS
;;;-------------------------------------------------------------------------------

(defclass stnu (stn)
  ((u :type list
      :accessor get-uncontrollable-variables
      :initarg :u
      :documentation "The list of uncontrollable time points.")
   (d :type durations
      :accessor get-uncontrollable-durations
      :initarg :d
      :documentation "The list of uncontrollable durations.")))

;; (defmethod print-object ((n_u stnu) stream)
;;   (format stream "~&(:controllable-variables ~A~% 
;;                      :uncontrollable-variables ~A~% 
;;                      :constraints ~A~% 
;;                      :durations ~A)"
;;           (mapcar #'(lambda (v_i)
;;                       (format nil "(~A :domain ~A)" v_i (get-domain v_i n_u)))
;;                   (get-controllable-variables n_u))
;;           (mapcar #'(lambda (v_i)
;;                       (format nil "(~A :domain ~A)" v_i (get-domain v_i n_u)))
;;                   (get-uncontrollable-variables n_u))
;;           (get-constraints n_u)
;;           (get-uncontrollable-durations n_u)))


;;;-------------------------------------------------------------------------------
;;; Probabilistic STN (pSTN) CLASS
;;;-------------------------------------------------------------------------------

;; (defclass pstn (stnu)
;;   ())


;;-------------------------------------------------------------------------------
;; Add (uncontrollable) durations(s) to an STNU.
;;-------------------------------------------------------------------------------

;; (defmethod add-duration ((d_i duration) (n_u stnu))
;;   ;; A duration is a binary temporal constraint.
;;   (assert (eq (length (get-scope d_i)) 2))
;;   ;; An STNU accepts only uncontrollable durations with a single interval domain.
;;   (if (not (typep n_u pstn)) ; don't recheck for the pSTN case
;;     (progn
;;       (assert (typep d_i uncontrollable-duration))
;;       (let ((rel (get-relation d_i))
;;             (mod (model-of rel))
;;             (dom (domain-of mod) )))))

;;   ;; STN does not allow disjunctive intervals.
;;   (assert (eq (length (get-relation c_i)) 1))
;;   (call-next-method c_i n))

;; (defmethod add-duration ((d_i duration) (n_p pstn))
;;   ;; A pSTN accepts only probabilistic durations.
;;   (assert (typep d_i probabilistic-duration))
;;   (call-next-method d_i n_p))
