;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

;;; For references relating to this code see:
;;;   [1] R. Dechter. Constraint Processing. Morgan Kaufmann, San Francisco,
;;;   California, 2003.

(in-package :mtk-csp-tcsp)


;;;--------------------------------------------------------------------------------
;;; Temporal Constraint Satisfaction Problem (TCSP) CLASS
;;;--------------------------------------------------------------------------------

;; Note that the symbol used for the variables of a CSP must be the same symbol
;; used for the vertices of a GRAPH. This allows a tcsp to be both a temporal
;; network as well as a csp.
(defclass tcsp (csp mtk-graph:directed-graph)
  ((temporal-constraint-ht-edge :type hash-table
                                :accessor get-temporal-constraint-ht-edge
                                :initform (make-hash-table)
                                :documentation "")))

(defmethod get-time-points ((tcsp tcsp))
  (mtk-graph:get-vertices tcsp))

(defmethod get-temporal-constraint ((e_i mtk-graph:directed-edge) (tcsp tcsp))
  (gethash e_i (get-temporal-constraint-ht-edge tcsp)))

;;;-------------------------------------------------------------------------------
;;; Make a temporal constraint from a constraint description list.
;;;-------------------------------------------------------------------------------

(defmethod make-temporal-constraint (&key ((:scope scope)) ((:relation intervals)))
  "Valid constraint description form:
      <scope>         := (<variable> <variable>)
      <relation>      := (<interval>+)
      <interval>      := (:lower-bound <lower-bound> :upper-bound <upper-bound>)
      <lower-bound>   := (<number> :inequality <lb-inequality>) | (<number>)
      <upper-bound>   := (<number> :inequality <ub-inequality>) | (<number>)
      <lb-inequality> := #'> | #'>=
      <ub-inequality> := #'< | #'<=

      Assumptions on relations:
        * Intervals:
            The list of intervals is a non-intersecting disjunction of intervals.
        * Time-Bounds:
            If no :inequality function is specified for the <lower-bound> or 
            <upper-bound>, then the non-strict versions #'>= and #'<= are used
            by default."

  ;; Must provide the scope and the intervals
  (assert (and scope intervals))
  ;; Only binary temporal constraints are supported.
  ;; (But this check is redundant and not general to non-binary temporal constraints.
  ;;  Think about whether this is needed and/or desirable.)
  (assert (eq (length scope) 2))
  (let ((relation nil))
    (dolist (i_i intervals)
      (push (apply #'make-interval i_i) relation))
    (make-instance 'constraint :scope scope :relation relation)))

;;;-------------------------------------------------------------------------------
;;; TCSP to distance graph
;;;-------------------------------------------------------------------------------

(defmethod tcsp-to-distance-graph ((tcsp tcsp))
  (let ((g (make-instance 'weighted-directed-graph
	     :default-weight (make-inf-upper-bound))))
    (add-vertices (get-variables tcsp) g)
    (dolist (c_i (get-constraint-list tcsp))
      (let ((interval (first (get-relation c_i)))
	    (s_i (get-scope tcsp)))
	(add-weighted-edge (list s_i (get-upper-bound interval)) g)
	(add-weighted-edge
	 (list (reverse s_i)
	       (make-bound 
		(* -1 (get-value (get-lower-bound interval)))
		:inequality
		(if (eql (get-inequality (get-lower-bound interval)) #'>=)
		    #'<= #'<)))
	 g)))))

;;;-------------------------------------------------------------------------------
;;; Delete variable(s) from a tcsp
;;;-------------------------------------------------------------------------------

(defmethod delete-variable (variable (tcsp tcsp))
  (delete-edges (find-edges-with-vertex variable tcsp) tcsp)
  (remhash variable (get-temporal-constraint-ht-edge tcsp))
  (call-next-method variable tcsp))

(defmethod delete-variables ((variables list) (tcsp tcsp))
  (dolist (v_i variables) (delete-variable v_i tcsp)))

(defmethod delete-vertex (vertex (tcsp tcsp))
  (delete-variable vertex tcsp))

(defmethod delete-vertices ((vertices list) (tcsp tcsp))
  (delete-variables vertices tcsp))

;;;-------------------------------------------------------------------------------
;;; Add constraint(s) to tcsp
;;;-------------------------------------------------------------------------------

(defmethod add-constraint ((c_i constraint) (tcsp tcsp))
  ;; Only binary temporal constraints are supported.
  (assert (eq (length (get-scope c_i)) 2))
  (call-next-method c_i tcsp)
  (let* ((start (first (get-scope c_i)))
         (end (second (get-scope c_i)))
         (e_i (add-edge `(,start ,end) tcsp)))
    (setf (gethash e_i (get-temporal-constraint-ht-edge tcsp)) c_i)
    e_i))

(defmethod add-constraint ((constraint-description list) (tcsp tcsp))
  "A constraint description takes the following form.
      <constraint> := (:scope <scope> :relation <relation>)"
  (add-constraint (apply #'make-temporal-constraint constraint-description)
                  tcsp))

;; (defmethod add-constraints ((constraints list) (tcsp tcsp))
;;   (dolist (c_i constraints) (add-constraint c_i tcsp)))

;;;-------------------------------------------------------------------------------
;;; Delete temporal constraint(s) from tcsp
;;;-------------------------------------------------------------------------------

(defmethod delete-constraint ((c_i constraint) (tcsp tcsp))
  (call-next-method c_i tcsp)
  (let ((e (find-edges-with-vertex-pair (get-scope c_i) tcsp)))
    (loop
       for e_i in e
       when (eq (gethash e_i (get-temporal-constraint-ht-edge tcsp)) c_i) 
       do
         (remhash e_i (get-temporal-constraint-ht-edge tcsp))
         (delete-edge e_i tcsp)
         (return (values c_i e_i)))))

;; (defmethod delete-constraints ((constraints list) (tcsp tcsp))
;;   (dolist (c_i constraints) (delete-constraints c_i tcsp)))

;; (defmethod delete-constraints ((c constraints) (tcsp tcsp))
;;   (delete-constraints (get-constraint-list c) tcsp))

;;;-------------------------------------------------------------------------------
;;; Find constraints in tcsp
;;;-------------------------------------------------------------------------------

;; (defmethod find-constraints-with-time-point (tp_i (tcsp tcsp))
;;   (find-constraints-with-variable tp_i tcsp))

;; (defmethod find-constraints-with-time-points ((tp list) (tcsp tcsp))
;;   (find-constraints-with-variables tp tcsp))

(defmethod find-constraints-with-time-point (tp_i (tcsp tcsp))
   (mapcar #'(lambda (e_i) (get-temporal-constraint e_i tcsp))
           (find-edges-with-vertex tp_i tcsp)))

(defmethod find-constraints-with-time-points ((tp list) (tcsp tcsp))
   (mapcar #'(lambda (e_i) (get-temporal-constraint e_i tcsp))
           (find-edges-with-vertices tp tcsp)))

(defmethod find-constraints-with-start-time-point (tp_i (tcsp tcsp))
   (mapcar #'(lambda (e_i) (get-temporal-constraint e_i tcsp))
           (find-edges-with-source-vertex tp_i tcsp)))

(defmethod find-constraints-with-end-time-point (tp_i (tcsp tcsp))
   (mapcar #'(lambda (e_i) (get-temporal-constraint e_i tcsp))
           (find-edges-with-target-vertex tp_i tcsp)))

(defmethod find-constraints-with-start-and-end-time-points (tp_i tp_j (tcsp tcsp))
   (mapcar #'(lambda (e_i) (get-temporal-constraint e_i tcsp))
           (find-edges-with-source-and-target-vertices tp_i tp_j tcsp)))
