;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

;;; For references relating to this code see:
;;;   [1] R. Dechter. Constraint Processing. Morgan Kaufmann, San Francisco,
;;;   California, 2003.

(in-package :mtk-csp)


;;;--------------------------------------------------------------------------------
;;; Constraint CLASS
;;;--------------------------------------------------------------------------------

(defclass constraint ()
  ((s :type list
      :accessor get-scope
      :initarg :scope
      :initform nil
      :documentation "")
   (r :accessor get-relation
      :initarg :relation
      :documentation "")))

(defmethod print-object ((c_i constraint) stream)
  (with-slots ((s s) 
               (r r)) c_i
      (format stream "(:scope ~A :relation ~A)" s r)))


;;;--------------------------------------------------------------------------------
;;; Constraints CLASS
;;;--------------------------------------------------------------------------------

(defclass constraints ()
  ((c-list :type list
           :accessor get-constraint-list
           :initform nil
           :documentation "")
   (c-ht-variable :type hash-table
                  :accessor get-constraints-ht-variable
                  :initform (make-hash-table)
                  :documentation "")))

(defmethod print-object ((c constraints) stream)
  (format stream "~A" (get-constraint-list c)))

;;;--------------------------------------------------------------------------------
;;; Constraint Satisfaction Problem (CSP) CLASS
;;;--------------------------------------------------------------------------------

;;; From [1] pg. 25
;;;   Modification:
;;;     * Note that the symbol "v" is used instead of "X" to represent a set of
;;;       variables so that the slot v is the same as the slot for vertices in
;;;       graphs. This allows multiple inheritance of CSP and Graph, in which
;;;       an object may be both a CSP and a type of constraint graph at the
;;;       same time.
(defclass csp ()
  ((v :type list
      :accessor get-variables
      :initform nil
      :documentation "")
   (d :type hash-table
      :accessor get-domain-ht-variable
      :initform (make-hash-table)
      :documentation "")
   (c :type constraints
      :accessor get-constraints
      :initform (make-instance 'constraints)
      :documentation "")))

(defmethod print-object ((p csp) stream)
  (format stream "~&(:variables ~A~% :constraints ~A)"
          (mapcar #'(lambda (v_i)
                      (format nil "(~A :domain ~A)" v_i (get-domain v_i p)))
                  (get-variables p))
          (get-constraints p)))

(defmethod get-domain (variable (p csp))
  (gethash variable (get-domain-ht-variable p)))

(defmethod get-constraint-list ((p csp))
  (get-constraint-list (get-constraints p)))

;;;-------------------------------------------------------------------------------
;;; Generate constraint graph
;;;-------------------------------------------------------------------------------

(defmethod csp-to-primal-constraint-graph ((p csp))
  (let ((v (get-variables p))
	(s (mapcar #'(lambda (c_i) (get-scope c_i))
		   (get-constraint-list p))))
    (generate-primal-constraint-graph v s)))

(defmethod generate-primal-constraint-graph ((variables list) (scopes list))
  (let ((g (make-instance 'mtk-graph:graph)))
    (mtk-graph:add-vertices variables g)
    (dolist (s_i scopes)
      (loop
	 for v on s_i 
	 do (let ((v_j (first v))
		  (v-remaining (rest v)))
	      (dolist (v_k v-remaining)
		(let ((vertices (list v_j v_k)))
		  (assert (listp vertices))
		  (if (and (not (eq v_j v_k))
			   (null (mtk-graph:find-edges-with-vertex-pair
				  vertices
				  g)))
		      (mtk-graph:add-edge vertices g)))))))
    g))

(defmethod csp-to-dual-constraint-graph ((p csp))
  (let ((v (get-variables p))
	(s (mapcar #'(lambda (c_i) (get-scope c_i))
		   (get-constraint-list p))))
    (generate-dual-constraint-graph v s)))

(defmethod generate-dual-constraint-graph ((variables list) (scopes list))
  (let ((g (make-instance 'mtk-graph:hyper-graph)))
    (mtk-graph:add-vertices variables g)
    (dolist (s_i scopes)
      (if (and (> (length s_i) 1)
	       (null (mtk-graph:find-edges-with-vertices variables g)))
	  (mtk-graph:add-edges variables g)))
    g))

;;;-------------------------------------------------------------------------------
;;; Add variable(s) to a csp
;;;-------------------------------------------------------------------------------

(defmethod add-variable (variable (p csp))
  (pushnew variable (get-variables p)))

(defmethod add-variables ((variables list) (p csp))
  (loop for v_i in variables
     collecting (add-variable v_i p)))

;;;-------------------------------------------------------------------------------
;;; Delete variable(s) from a csp
;;;-------------------------------------------------------------------------------

(defmethod delete-variable (variable (p csp))
  (delete-constraints (find-constraints-with-variable variable p) p)
  (remhash variable (get-domain-ht-variable p))
  (setf (get-variables p) (remove variable (get-variables p))))

(defmethod delete-variables ((variables list) (p csp))
  (dolist (v_i variables) (delete-variable v_i p)))

;;;-------------------------------------------------------------------------------
;;; Add constraint(s) to a csp
;;;-------------------------------------------------------------------------------

(defmethod add-constraint ((c_i constraint) (c constraints))
  (push c_i (get-constraint-list c))
  (dolist (v_j (get-scope c_i))
    (push c_i (gethash v_j (get-constraints-ht-variable c))))
  c_i)

(defmethod add-constraint ((c_i constraint) (p csp))
  (add-variables (get-scope c_i) p)
  (add-constraint c_i (get-constraints p)))

(defmethod add-constraints ((constraints list) (p csp))
  (dolist (c_i constraints) (add-constraint c_i p)))

(defmethod add-constraints ((c constraints) (p csp))
  (add-constraints (get-constraint-list c) p))

;;;-------------------------------------------------------------------------------
;;; Delete constraint(s) from a csp
;;;-------------------------------------------------------------------------------

(defmethod delete-constraint ((c_i constraint) (c constraints))
  (setf (get-constraint-list c) (remove c_i (get-constraint-list c)))
  (dolist (v_j (get-scope c_i))
    (setf (gethash v_j (get-constraints-ht-variable c)) 
          (remove c_i (gethash v_j (get-constraints-ht-variable c))))))

(defmethod delete-constraint ((c_i constraint) (p csp))
  (delete-constraint c_i (get-constraints p)))

(defmethod delete-constraints ((constraints list) (p csp))
  (dolist (c_i constraints) (delete-constraints c_i p)))

(defmethod delete-constraints ((c constraints) (p csp))
  (delete-constraints (get-constraint-list c) p))

;;;-------------------------------------------------------------------------------
;;; Find constraints in a csp
;;;-------------------------------------------------------------------------------

(defmethod find-constraints-with-variable (variable (c constraints))
  (gethash variable (get-constraints-ht-variable c)))

(defmethod find-constraints-with-variable (variable (p csp))
  (find-constraints-with-variable variable (get-constraints p)))

(defmethod find-constraints-with-variables ((variables list) (c constraints))
  (reduce #'intersection
          (mapcar #'(lambda (v_i) (find-constraints-with-variable v_i c))
                  variables)))

(defmethod find-constraints-with-variables ((variables list) (p csp))
  (find-constraints-with-variables variables (get-constraints p)))
