(in-package :drake2)

(defmethod construct-dispatchable-graph ((g mtk-graph:labeled-weighted-directed-graph))
  ;; Make a new graph and a hash table to keep the rigid components
  (let* ((new-graph (make-instance 'mtk-graph:labeled-weighted-directed-graph 
								   :state-space (get-state-space g))))

	(multiple-value-bind (rigid-component-ht)
		(johnsons-preprocess-and-contract g new-graph)

	  (add-vertices (get-vertices g) new-graph)

	  ;;now traverse and output the edges
	  (dolist (source-vertex (sort (copy-list (get-vertices g)) #'string<))
		
		(let* ((unweighted-d-pi-ht 
				(bellman-ford g source-vertex 
							  :predecessor-graph? t
							  :test-consistency? nil)))

		  ;;traverse the graph and put all the necessary edges into the new graph
		  (insert-non-dominated-edges! new-graph source-vertex 
									   unweighted-d-pi-ht 
									   rigid-component-ht)))

	  (debug-print-for :drake t "Done collecting edges.")	

	  (debug-print-for :mtk-graph-algorithms t "Final rigid components ~a"
					   (loop for env being the hash-key of rigid-component-ht
						  collecting (list env (gethash env rigid-component-ht)))))	

	;;and return the result
	new-graph))


(defmethod johnsons-preprocess-and-contract ((ldg labeled-weighted-directed-graph)
											 (output-graph labeled-weighted-directed-graph))

  (let ((global-root-node (gensym)))
	;;Add a universal zero weight edge from the root node to every node
	(dolist (vertex (get-vertices ldg))
	  (update-labeled-weighted-edge (list (list global-root-node vertex) 0 
										  (get-universal-clause (get-state-space ldg)))
									ldg))


	(let* ((d-pi-ht (bellman-ford ldg global-root-node
								  :predecessor-graph? t)))
	  ;;check that the problem is feasible.
	  (when (null d-pi-ht)
		(error 'infeasible-problem))

	  (debug-print-for :drake t "SSSP: ~a"
					   (loop for env being the hash-key of d-pi-ht
						  collecting (list env (gethash env d-pi-ht))))

	  ;; TODO this allows Drake to be faster! But it doesn't dispatch correctly if it's "hacked" this way!
	  ;; A good test case is (test-pike-1).

	  (let ( ;(rigid-component-ht (make-hash-table))
			(rigid-component-ht (find-strongly-connected-components ldg d-pi-ht)) ;; TODO This is what is super slow!!!
			)

		(debug-print-for :drake t "Alternate rigid components ~a"
						 (loop for env being the hash-key of rigid-component-ht
							collecting (list env (gethash env rigid-component-ht))))

		(delete-vertex global-root-node ldg)

		(contract-for-rigid-components! ldg rigid-component-ht d-pi-ht output-graph)

		(debug-print-for :drake t "Contracted graph ~a"
						 ldg)

		(debug-print-for :drake t "Prepared output ~a"
						 output-graph)

		(values rigid-component-ht)))))



(defmethod contract-for-rigid-components! ((ldg labeled-weighted-directed-graph)
										   rigid-component-ht
										   d-pi-ht
										   (output-graph labeled-weighted-directed-graph))
  
  ;;moving an edge may put it on a new rigid component, so double loop over the components
  ;;to completely move everything
  
  (loop for env in (sort (loop for all-env being the hash-key of rigid-component-ht
							unless (null (gethash all-env rigid-component-ht))
							collect all-env into useful-envs
							finally (return useful-envs))
						 #'<
						 :key #'state-clause-length)
	 do 
	 (loop for component in (gethash env rigid-component-ht)
		with leader
		do (setf leader (find-rigid-component-leaders 
						 output-graph
						 component
						 env d-pi-ht ))
			
		do (debug-print-for :drake t "All leaders for ~a ~a ~a" component env leader)
		do 
			
		(progn
		  (debug-print-for :drake t
						   "Moving edges for component: ~a leader: ~a env: ~a" component leader env)
			  
		  (loop for non-leader in (remove leader component)
			 with leader-advantage
			 do (setf leader-advantage 
						  
					  ;;leader not the natural leader
					  (- (first (first (query-lvs (gethash non-leader d-pi-ht) 
												  env)))
						 (first (first (query-lvs (gethash leader d-pi-ht) 
												  env)))))
				   
			 (debug-print-for :drake t
							  "Adjusting non-leader: ~a ~a " non-leader leader-advantage)
			 ;;if they're zero-zero, mark in the output graph
										;when (eql leader-advantage 0)
			 ;; (add-rigid-pair! output-graph 
			 ;; 				 leader non-leader 
			 ;; 				 env
			 ;; 				 leader-advantage)
			 ;;put these directly into the output graph
			 do (update-labeled-weighted-edge
				 (list (list leader non-leader)
					   leader-advantage 
					   env)
				 output-graph)
			 (update-labeled-weighted-edge
			  (list (list non-leader leader)
					(- leader-advantage) 
					env)
			  output-graph)
				   
			 ;;now adjust the existing graph
			 do (dolist (inbound-edge (find-edges-with-target-vertex 
									   non-leader (get-edges ldg)))
					  
				  (unless (member (get-source-vertex inbound-edge)
								  component)
					(debug-print-for :drake t
									 "Moving in edge: ~a" inbound-edge)
					(dolist (inbound-pair 
							  (lvs-pairs
							   (get-weight inbound-edge ldg)))
					  (debug-print-for :drake t
									   "Moving in pair: ~a" inbound-pair)
					
					  ;;when this value isn't incompatible with the component
					  (when (clause-valid? (clause-intersection 
											env
											(pair-environment inbound-pair)))
						(update-labeled-weighted-edge 
						 (list (list (get-source-vertex inbound-edge)
									 leader)
							   (- (pair-value inbound-pair) 
								  leader-advantage)
							   (clause-intersection
								(pair-environment inbound-pair)
								env))
						 ldg))))
				  (debug-print-for :drake t
								   "Cleaning in ~a ~a" inbound-edge env)
							 
				  (clear-dependence-on-env! (get-weight inbound-edge 
														ldg) env)

				  )
				   
				   
			 (dolist (outbound-edge (find-edges-with-source-vertex 
									 non-leader (get-edges ldg)))
					 
			   (unless (member (get-target-vertex outbound-edge)
							   component)
				 (debug-print-for :drake t
								  "Moving out edge: ~a" outbound-edge)
				 (dolist (outbound-pair 
						   (lvs-pairs 
							(get-weight outbound-edge ldg)))
				   (debug-print-for :drake t
									"Moving out pair: ~a" outbound-pair)
						 
				   (when (clause-valid? (clause-intersection 
										 env
										 (pair-environment outbound-pair)))	 
					 (update-labeled-weighted-edge 
					  (list (list leader
								  (get-target-vertex outbound-edge))
							(+ (pair-value outbound-pair) 
							   leader-advantage)
							(clause-intersection
							 env
							 (pair-environment outbound-pair)))
					  ldg))))
			   (debug-print-for :drake t
								"Cleaning out ~a ~a" outbound-edge 
								env) 
			   (clear-dependence-on-env! (get-weight outbound-edge ldg) 
										 env)))))))


(defmethod add-zero-group! ((output-graph labeled-weighted-directed-graph) 
							zero-group
							env)
  (unless (eql 1 (length zero-group))
	(debug-print-for :drake t "Documenting zero group ~a" (list zero-group env))
	(pushnew (list zero-group env) (get-zero-components output-graph)
			 :test #'(lambda (x y) (and (eql (second x) (second y))
										(subsetp (first x) (first y))
										(subsetp (first y) (first x)))))))


(defmethod find-rigid-component-leaders ((output-graph labeled-weighted-directed-graph)
										 component
										 (component-env state-clause)
										 d-pi-ht)
  (debug-print-for :drake t "Finding rigid component leader for ~a ~a." 
				   component component-env)
  (loop for vertex in component
	 with leader = nil
	 with leader-val = *inf*
	 with current-val
	 do  (setf current-val (first (first (query-lvs (gethash vertex d-pi-ht) component-env))))
	 ;;swap the leader if it's lower, or equal in value
	 ;;and has other leaders
	 when (< current-val leader-val)
	 
	 do (setf leader vertex)
	 and do (setf leader-val current-val)
	   
	 do (debug-print-for :drake t "Found initial leader: ~a " leader)
	 finally 
	 ;;Try using every vertex as the root of a zero-group. See if any
	 ;;of the other vertices are exactly the same distance as the zero-leader,
	 ;;in which case they're part of the zero-group.
	 (loop for zero-leader in component 
		do 
		(debug-print-for :drake t "Zero leader ~a" zero-leader)
		(loop for zero-member in (remove zero-leader component)
		   do (debug-print-for :drake t "Member ~a distances: ~a ~a ~a" 
							   zero-member
								   
							   (first (first (query-lvs (gethash zero-leader d-pi-ht) 
														component-env)))
							   (first (first (query-lvs (gethash zero-member d-pi-ht) 
														component-env)))
							   (-  
								(first (first (query-lvs (gethash zero-leader d-pi-ht) 
														 component-env)))
								(first (first (query-lvs (gethash zero-member d-pi-ht) 
														 component-env)))))
		   ;;collect the zero related part of the group
		   when (equalp 0 (-  
						   (first (first (query-lvs (gethash zero-leader d-pi-ht) 
													component-env)))
						   (first (first (query-lvs (gethash zero-member d-pi-ht) 
													component-env)))))
		   collect zero-member into zero-component
		   ;;unless we didn't find anything, add the leader to the group
		   ;;and make it a zero group
		   finally (unless (null zero-component)
					 (add-zero-group! output-graph (append (list zero-leader)
														   zero-component)
									  component-env))))
	 (return leader)))


(defmethod insert-non-dominated-edges! ((ldg labeled-weighted-directed-graph)
										start-vertex
										d-pi-ht
										rigid-component-ht
										&key (actually-prune? t))

  (debug-print-for :drake t "~%~%Finding edges for: ~a" start-vertex)
  (debug-print-for :drake t "SSSP: ~a"
				   (loop for env being the hash-key of d-pi-ht
					  collecting (list env (gethash env d-pi-ht))))

  (let ((pruned-edges-ht (make-hash-table :test #'equal)))
	(traverse-predecessor-graph-for-pruned-edges! 
	 ldg
	 d-pi-ht
	 start-vertex
	 start-vertex
	 nil
	 (list (list start-vertex *inf*))
	 (get-universal-clause (get-state-space ldg))
	 *inf*
	 rigid-component-ht
	 pruned-edges-ht)

	;;now collect all the vertices that aren't pruned
	(loop for end-vertex in (sort (copy-list (get-vertices ldg)) #'string<) ;;loop for all the vertices
	   do (debug-print-for :drake t "collecting for end: ~a" end-vertex)
	   (unless (eql start-vertex end-vertex) ;that aren't the start
		 (dolist (path-pair (lvs-pairs 
							 (gethash end-vertex d-pi-ht))) ;try all the pairs
		   (debug-print-for :drake t "candidate: ~a" path-pair)
		   
		   ;;form the description
		   (let ((candidate (list (list start-vertex end-vertex)
								  (first (pair-value path-pair))
								  (pair-environment path-pair))))
			 ;;and unless it's been pruned, or it's *inf*
			 (unless (or (eql *inf* (first (pair-value path-pair)))
						 (and actually-prune?
							  (gethash candidate pruned-edges-ht)))
			   (debug-print-for :drake t "Keep!")
			   ;;update the graph
			   (update-labeled-weighted-edge candidate ldg))))))))


(defun traverse-predecessor-graph-for-pruned-edges!
	(ldg
	 d-pi-ht
	 start-vertex
	 current-vertex
	 most-recent-path
	 current-path
	 current-env
	 predecessor-min-value
	 rigid-component-ht
	 pruned-edges-ht)
  "Do not call with start-vertex=current-vertex."
  (debug-print-for :drake t "current path: ~a, ~a considering ~a ~a" 
				   current-path current-env most-recent-path predecessor-min-value)

  (loop for event-on-path in (mapcar #'first current-path)
	 with event-path-ht = (make-hash-table)
	 do (assert (not (gethash event-on-path event-path-ht)))
	 (setf (gethash event-on-path event-path-ht) t) )
  
  ;;Can only prune on the second step away from the start, not the first
  (unless (or (eql start-vertex current-vertex)
			  (eql 2 (length current-path)))
	;;Split on whether the weight is negative
	(if (<= 0 (first most-recent-path))
	  
		(progn ;;progn for positive edges
		
		  (debug-print-for :drake t "positive case")
		  (when	   ;and predecessors-visited?
			  ;;and the value is either not there or bigger,
			  ;;in which case we don't have evidence to prune
			  ;;recall you can't prune unless this path
			  ;;contains the entirety of the value you're pruning
			  (and (eql (second most-recent-path) current-env)
				   (<= predecessor-min-value (first most-recent-path)))
			
			(debug-print-for :drake t "Prune!")
			
			;;now, save the edge, but not if it's inf
			(unless (eql *inf* (second most-recent-path))
			  (setf (gethash  
										;update-labeled-weighted-edge 
					 (list (list start-vertex current-vertex) 
						   (first most-recent-path)
						   (second most-recent-path))
					 pruned-edges-ht) t))))

		;;depends on whether the destination is a rigid member/leader
		(progn							;handle negative case
			
		  (debug-print-for :drake t "negative case")
		  (when							  
			  ;;and isn't enough evidence to prune, that is,
			  ;;there's no value or it's not negative
			  (and  (eql (second most-recent-path) current-env)
					(> 0 predecessor-min-value))
		   
			(debug-print-for :drake t "Prune!")

			(unless (eql *inf* (second most-recent-path))
			  (setf						;update-labeled-weighted-edge 
			   (gethash (list (list start-vertex current-vertex) 
							  (first most-recent-path)
							  (second most-recent-path))
						pruned-edges-ht)
			   t))))))

  (loop for target-vertex in (sort  (set-difference
									 (get-vertices ldg)
									 (mapcar #'first current-path)) #'string<)
	 do (loop for target-inbound-pair in 
			 (lvs-pairs  (gethash target-vertex d-pi-ht))
			 with new-env
			 with new-path
			 do (setf new-env (clause-intersection 
							   current-env
							   (pair-environment target-inbound-pair)))
			  
			 (setf new-path (append (list (append (list target-vertex)
												  (pair-value target-inbound-pair)))
									current-path))
			 when (and (not (null new-env))
					   (eql current-vertex (second (pair-value target-inbound-pair))))
			  
		   ;;when we're not doubling back for that environment or back onto
		   ;;the current path
			 do 

		   ;;all paths from the first node are fine, otherwise test them
			 (when (and (not (null new-env))
						(clause-subsumed? (pair-environment target-inbound-pair)
										  new-env)
						(or (eql start-vertex current-vertex)
							(not (member target-vertex 
										 (mapcar #'first current-path))))
						(path-valid? new-path new-env d-pi-ht (not (eq new-env current-env ))))
			  
			   (traverse-predecessor-graph-for-pruned-edges!
				ldg
				d-pi-ht
				start-vertex
				target-vertex
				(list (first (pair-value target-inbound-pair))
					  (pair-environment target-inbound-pair))
				new-path
				new-env
				;;update the min value with the path length
				;;to the current vertex
				(min predecessor-min-value (second (first current-path)))
				rigid-component-ht
				pruned-edges-ht)))))


(defmethod find-strongly-connected-components
	((g mtk-graph:labeled-weighted-directed-graph)
	 d-pi-ht)
  "Identify only those SCC that the compiling algorithm cares about!"

  (let ((rigid-component-ht (make-hash-table)))
	(debug-print-for :drake t "SSSP: ~a"
					 (loop for env being the hash-key of d-pi-ht
						collecting (list env (gethash env d-pi-ht))))

	(dolist (start-vertex (get-vertices g))
	  (debug-print-for :drake t "Beginning SCC search ~a" start-vertex)
	  (search-for-start-vertex g 
							   start-vertex 
							   start-vertex
							   (get-universal-clause (get-state-space g))
							   (list (list start-vertex))
							   rigid-component-ht
							   d-pi-ht))
	rigid-component-ht))


(defmethod search-for-start-vertex ((g mtk-graph:labeled-weighted-directed-graph)
									start-vertex
									current-vertex
									current-env
									current-path
									rigid-component-ht
									d-pi-ht)

  (loop for target-vertex in (sort (copy-list (get-vertices g)) #'string<)
	 do (loop for target-inbound-pair in (lvs-pairs (gethash target-vertex d-pi-ht))
			 with new-env
			 with new-path
			 do (setf new-env (clause-intersection current-env
												   (pair-environment target-inbound-pair)))
			 (setf new-path (append (list (append (list target-vertex)
												  (pair-value target-inbound-pair)))
									current-path))
			  
			 when (and (not (null new-env))
					   (eql current-vertex
							(second (pair-value target-inbound-pair))))
			
		   ;;when we're not doubling back for that environment or back onto
		   ;;the current path
			 do 	   
		   ;;all paths from the first node are fine, otherwise test them
		   ;;test that the vertex you're going to agrees under the query
		   ;;with the proposed path

			 (when (and (path-valid? new-path new-env d-pi-ht (not (eq new-env current-env)))
						(or (eql start-vertex current-vertex)
							(eql target-vertex start-vertex)
							(>= 1 (apply #'+ 
										 (mapcar 
										  #'(lambda (x)
											  (if (eql x target-vertex)
												  1
												  0))
										  (mapcar #'first current-path))))))

			   (if (eql start-vertex target-vertex)
				   (progn
					 (debug-print-for :drake t "Loops to start: ~a" target-vertex)
					 (add-rigid-component! rigid-component-ht 
										   (sort (remove-duplicates
												  (append (list current-vertex)
														  (mapcar #'first current-path)))
												 #'string<) 
										   new-env))
				   (search-for-start-vertex
					g
					start-vertex
					target-vertex
					new-env
					new-path
					rigid-component-ht
					d-pi-ht))))))


(defun add-rigid-component! (rigid-component-ht
							 new-component
							 env)
  ;;check that a subsuming, larger component doesn't exist before you try to add
  (loop for component-env being the hash-keys of rigid-component-ht
	 if (and (clause-subsumed? env component-env)
			 (member new-component (gethash component-env rigid-component-ht)
					 :test #'subsetp))
	   do (return-from add-rigid-component!))

  ;;loop again, and look for any that this one removes, that is, which
  ;;the env subsumes and which are smaller or equal to this one
  (loop for component-env being the hash-keys of rigid-component-ht
	 ;;if the input clause subsumes the existing one
	 if (clause-subsumed? component-env env)
	 ;;go through the existing rigid components and remove
	 ;;any that are subsets of the new one
	 do (setf (gethash component-env rigid-component-ht)  
			  (remove-if #'(lambda (x) (subsetp x new-component))
						 (gethash component-env rigid-component-ht))))

  ;;Add the new component iff it is different than the other rigid components known
  (push  new-component (gethash env rigid-component-ht)))


(defun path-valid? (path current-env d-pi-ht test-whole-path?)
  (and (not (null current-env))
	   (clause-valid? current-env)
	   ;;if we don't need to retest the whole path, only try test first one
	   (loop for path-element in (if test-whole-path?  
									 (butlast path) 
									 (list (first path)))
		  always (member (list (second path-element) (third path-element))
						 (query-lvs (gethash (first path-element) d-pi-ht)
									current-env)
						 :test #'equal))))
