(in-package #:drake2)

(defclass online-ldg-dispatch (ldg-dispatch)
  ((mailbox
	:type mtk-utils-threading:mailbox
	:initform (mtk-utils-threading:make-mailbox)
	:initarg :mailbox
	:accessor online-dispatch-mailbox
	:documentation "The IDs of completed activities are sent to this mailbox by an external thread.")
   (lock 
	:initform (bordeaux-threads:make-recursive-lock) 
	:initarg :lock
	:accessor  online-dispatch-lock
	:documentation "Lock used to make mailbox access atomic.")
   (stop-flag
	:type mtk-utils-threading:flag
	:initform (mtk-utils-threading:make-flag :value nil)
	:initarg :stop-flag
	:accessor online-dispatch-stop-flag
	:documentation "An external thread should set this flag to t to terminate the online-execution.")
   (running-flag
	:type mtk-utils-threading:flag
	:initform (mtk-utils-threading:make-flag :value nil)
	:initarg :running-flag
	:accessor online-dispatch-running-flag
	:documentation "Flag used to indicate that the dispatcher is running. t if running. nil if not.")
   (failed-activity-flag
	:type mtk-utils-threading:flag
	:initform (mtk-utils-threading:make-flag :value nil)
	:initarg :failed-activity-flag
	:accessor online-dispatch-failed-activity-flag
	:documentation "If a dispatched activity returns failure, the activity callback should notify the dispatcher by setting this flag to the ID of the failed activity.")
   (activities
	:type list
	:initform '()
	:initarg :activities
	:accessor online-dispatch-activities
	:documentation "A list of activities, where an activity is the list (lb-edge ub-edge activity).
                    When activities are activated (start event executed, they are removed from this
                    list and placed on the list of active-activities.")
   (active-activities
	:type list
	:initform '()
	:initarg :active-activities
	:accessor online-dispatch-active-activities
	:documentation "A list of activities whose start events are being executed. When the end-event
                    of an activity is received, it is removed from this list and discarded.")
   (completed-activities
	:type list
	:initform '()
	:initarg :completed-activities
	:accessor online-dispatch-completed-activities
	:documentation "A list of the activities that have been completed - i.e., their completed ID's
                    were received.")))


;;; ***********************************************************************
;;;
;;;                THREADING
;;;
;;; ***********************************************************************
(defmethod start-dispatcher ((dispatch online-ldg-dispatch) &rest forms)
  "Starts the dispatcher. Takes in the same keyed arguments as the 
   method online-execute!."
  (bordeaux-threads:make-thread 
   (lambda () 
	 (apply #'online-execute! dispatch forms))
   :name (symbol-name (gensym "THREAD-"))))

(defmethod stop-dispatcher ((dispatch online-ldg-dispatch) &key (wait nil))
  "Tells the dispatcher to stop before the next processing loop.
Once called the dispatcher will not run again until clear-stop-dispatcher is called.
If WAIT is nil (default), this method will return immediately.
If WAIT is t, this method will block until the dispatcher stops.
If the dispatcher is not running, the method returns immediately."
  (with-slots (stop-flag running-flag) dispatch
	(mtk-utils-threading:flag! stop-flag t)
	(when wait
	  ;; wait for the dispatcher to finish
	  (mtk-utils-threading:flag? running-flag :wait t :wait-value nil))))

(declaim (inline peek-stop-dispatcher))
(defmethod peek-stop-dispatcher ((dispatch online-ldg-dispatch))
  "Returns whether the dispatcher is supposed to be stopped.
This method is not intended as a public interface."
  (eq (mtk-utils-threading:flag? (online-dispatch-stop-flag dispatch)) t))

(defmethod clear-stop-dispatcher ((dispatch online-ldg-dispatch))
  "Clears any previous calls to stop the dispatcher."
  (mtk-utils-threading:flag! (online-dispatch-stop-flag dispatch) nil))

(defmethod dispatcher-running? ((dispatch online-ldg-dispatch))
  (mtk-utils-threading:flag? (online-dispatch-running-flag dispatch) :wait nil))

(defmacro with-running-flag (running-flag &rest body)
  "Handles the setting of the running flag. And ensures only one instance of body
can be run at any given time.
RUNNING-FLAG must be an instance of mtk-utils-threading:FLAG class."
  (let ((run-flag (gensym "FLAG-"))
		(flag-lock (gensym "FLAG-LOCK-"))
		(run (gensym "RUN-")))
	`(let ((,run-flag ,running-flag)
		   (,flag-lock (mtk-utils-threading:flag-lock ,running-flag))
		   (,run nil))
	   ;; The above two let layers ensure the dispatch argument 
	   ;; is evaluated once, and there is no variable-name clash
	   ;; See "http://www.apl.jhu.edu/~hall/Lisp-Notes/Macros.html"
	   ;; for the challenges of writing macros.
	   (bordeaux-threads:with-recursive-lock-held (,flag-lock)
		 ;; See if something is already running. If not, set running to true.
		 (setf ,run (not (mtk-utils-threading:flag? ,run-flag :wait nil)))
		 (when ,run (mtk-utils-threading:flag! ,run-flag t)))
	   (when ,run
		 ;; Start running the body.
		 ,@body
		 ;; When the body is done, set the run flag to nil.
		 (mtk-utils-threading:flag! ,run-flag nil)))))


;;; ***********************************************************************
;;;
;;;                Online Dispatching
;;;
;;; ***********************************************************************
(defmethod online-execute! ((dispatch online-ldg-dispatch) &key 
							(end-time *inf*)
							(ignore-missed-activities nil)
							(terminate-after-last-event t)
							(event-callback nil)
							(activity-callback #'default-activity-callback)
							(execution-monitor-check-fn #'default-execution-monitor-check-fn)
							(status-callback #'default-status-callback)
							(can-commit-to-events? nil) ; Function pointer that allows something else to restrict event execution (ex., event env's)
							(could-add-conflicts-and-commit-to-events? nil) ; Similar to above function, but checks by adding a conflict first
							(add-conflicts-and-commit-to-events nil)
							(time-fn #'default-time-fn) ; TODO - Change this when we want to use ROS /clock, or other time function
							(dont-reset-activity-ids nil))
  "Dispatches events in a labeled directed graph on-line.

   END-TIME is given in graph time-units. Defaults to infinity.
   IGNORE-MISSED-ACTIVITIES 
     When nil missed activities will cause a planning error.
     When t missed activities will be ignored.
     Default value is nil.
   TERMINATE-AFTER-LAST-EVENT
     When t, the dispatcher will terminate after executing the last event (default).
     When nil, the dispatcher will continue running after the last event is executed.
     This last case is useful if the master-dispatcher instance will be modified on-line.
   EVENT-CALLBACK is the function that notifies events have been executed
    It will be called with two arguments:
     The time the event was executed.
     The list of events executed. May be unspecified / left nil for a default.
   ACTIVITY-CALLBACK is the function used to dispatch activities.
     It will be called with four arguments: 
     id - a numerical id unique to this activity, 
     primitive - a string to be dispatched
     lb - floating point lowerbound time
     ub - floating point upperbound time.
   STATUS-CALLBACK reports status.
    It will be called with two string arguments.
    The first string tells the class of status message.
    The second string gives details on the status
   EXECUTION-MONITOR-CHECK-FN allows execution monitor checking to be be
    be plugged in. It is assumed this method is relatively quick. This
    function will be called with no arguments, and is expected to return
    T if execution is going fine, or nil if a problem has been detected.

   CAN-COMMIT-TO-EVENTS - A function reference that should return T if 
    the given events can be executed, otherwise should return nil.

   TIME-FN - returns the current time (seconds) that the dispatcher will use. By default,
    uses the current system time.

   DONT-RESET-ACTIVITY-IDS - Defaults to nil, so that every time the dispatcher is run
    the automatically-assigned activity ID's are reset. If T, these are NOT reset."

  ;; Make time print outs relative to the start, and (optionally) reset activity ID's!
  (format-time-reset)
  (unless dont-reset-activity-ids
	(setf *unique-activity-id* 0))

  (bordeaux-threads:with-recursive-lock-held ((online-dispatch-lock dispatch))
	;; Make sure only one instance of online-execute! is ever running & set the running flag.
	(with-running-flag (online-dispatch-running-flag dispatch)
	  ;; Start the core dispatcher algorithm.
	  (with-slots (current-time S ldg events-to-execute) dispatch
		(setf current-time 0)
		;; Dispatch the magic "start" event now. (This may enable certain causal links rooted in the initial conditions).
		(default-event-callback current-time (list :|start|) event-callback)
		(loop 
		   with real-start-time = (funcall time-fn)
		   with real-end-time = 0
		   with completed-activity-ids = '()
		   ;; The following until-conditional allows future modification
		   ;; so we can find out why the loop was terminated.
		   until (cond ((null events-to-execute) terminate-after-last-event)
					   ((peek-stop-dispatcher dispatch) t)
					   ((>= current-time end-time) t)
					   (t nil)) do
		   ;; Get a list of all completed events from the mailbox
		   (loop with mail and mail-exists do
				(multiple-value-setq (mail mail-exists) (mtk-utils-threading:mailbox-read (online-dispatch-mailbox dispatch) :wait nil))
				(if mail-exists
					(push mail completed-activity-ids)
					(return)))
		   ;; Check to see if any activities failed
		   (let* ((failed-id (mtk-utils-threading:flag? (online-dispatch-failed-activity-flag dispatch))))
			 (when (not (null failed-id))
			   ;; We've detected that a primitive activity failed!
			   (return-from online-execute! -2)))

		   ;; Remove activities with completed ids.
		   (remove-completed-activities dispatch completed-activity-ids)
		   ;; And removed missed windows
		   (remove-missed-windows! dispatch
								   :add-conflicts-and-commit-to-events add-conflicts-and-commit-to-events)

		   ;; If we are using the execution monitor, check the status of execution and report any problems.
		   (when (not (funcall execution-monitor-check-fn))
			 (return-from online-execute! -1))

		   ;; Get the candidate events to execute. The :event-waiting parameter filters out all events
		   ;; that are waiting on activities!

		   ;; TODO: Probably add a callback below for IR stuff?

		   (let* ((candidate-events (get-events-to-execute dispatch
														   :event-waiting? (lambda (event) 
																			 (event-waiting-on-activity? dispatch event))
														   :can-commit-to-events? can-commit-to-events?
														   :could-add-conflicts-and-commit-to-events? could-add-conflicts-and-commit-to-events?
														   :add-conflicts-and-commit-to-events add-conflicts-and-commit-to-events)))
			 ;; Either dispatch the candidate events or increment time.
			 (if candidate-events
				 (progn		 ; Candidate events
				   ;; Execute the events from the dispatcher
				   (execute-events! dispatch candidate-events)
				   ;; In order to ensure that every causal link is checked at least once by the
				   ;; execution monitoring algorithm (even if it is zero time length - we still 
				   ;; need to check preconditions and effects), partion the now-executing events
				   ;; into ones corresponding to the ends of activities and all other events
				   ;; (corresponding to the beginning of new activities or nothing). Then,
				   ;; update the active causal links with the just-completed-activity events and then
				   ;; update again with all other events, each time checking if the plan is valid
				   ;; each time (If it's not then something changed/some action didn't do what it promised!)
				   (let* ((just-completed-activity-events nil)
						  (all-other-events nil))
					 (loop for event in candidate-events do
						  (if (find-if (lambda (activity) (equalp (activity-end activity) event)) (online-dispatch-completed-activities dispatch)) 
							  (push event just-completed-activity-events)
							  (push event all-other-events)))
					 ;; Update the execution monitor with the events corresponding to just-finished activities (if there are any)
					 (when just-completed-activity-events
					   (default-event-callback current-time just-completed-activity-events event-callback))
					 ;; Check the execution monitor
					 (when (not (funcall execution-monitor-check-fn))
					   (return-from online-execute! -1))
					 ;; Next, run the callback for all other events (if there are any)
					 (when all-other-events
					   (default-event-callback current-time all-other-events event-callback))

					 (when (not (funcall execution-monitor-check-fn))
					   (return-from online-execute! -1)))
				   ;; Dispatch the events that correspond to activities				
				   (dispatch-activities dispatch candidate-events activity-callback))

				 (progn					; No candidate events
				   ;; The time only advances when there are no candidates to execute at the current time.
				   ;; Note: The method get-events-to-execute does NOT necessarily return all of the
				   ;; events that should be executed at this current time.
				   ;; Incrementing time in this loop increases the likelihood that events/activities 
				   ;; that should be executed close to simultaneously, will be.
				   (let ((temp-real-end-time (funcall time-fn))
						 (temp-current-time (- real-end-time real-start-time)))
					 ;; check for critical events that will be missed
					 (multiple-value-bind (missed-events critical) (find-missed-events dispatch :time temp-current-time)
					   ;; loop over the missed events and state which activities were missed
					   (dolist (activity (online-dispatch-active-activities dispatch))
						 (when (and (find (activity-end activity) missed-events) status-callback)
						   (funcall status-callback "drake.missed-activity" (activity-id activity) (activity-primitive activity))))
					   ;; there are critical events and we don't want to ignore missed activities, we need to remove missed activities
					   (if (and critical ignore-missed-activities)	   
						   (remove-missed-activities dispatch missed-events)
						   (progn
							 ;; update the time
							 (setf real-end-time temp-real-end-time)
							 (setf current-time temp-current-time))))))))

		   ;; yield the thread for a few cycles
		   (bordeaux-threads:thread-yield)))
	  ;; Now that execution is theoretically complete, check the execution monitor one last time
	  ;; to see if every constraint is satisifed!
	  (when (not (funcall execution-monitor-check-fn))
		(return-from online-execute! -1)))
	nil))


(defun default-event-callback (current-time executed-events event-callback)
  "Default event callback. Prints out that the events were executed, then calls a user callback."
  (format-time t "Executing event~p ~a at time ~a~%" (length executed-events) (colorize :yellow (format nil "~{~a~^, ~}" executed-events)) (float current-time))
  (when event-callback
	(funcall event-callback current-time executed-events)))

(defun default-activity-callback (id primitive lb ub)
  "Default activity callback when no other is specified - simply prints out the activity."
  (format-time t "Dispatching primitive ~a: [~a, ~a] ~a~%" id lb ub (colorize :purple (format nil "~a" primitive))))

(defun default-status-callback (name &rest data)
  (format-time t "Drake Status Callback: ~s ~s ~%" name data))

(defun default-execution-monitor-check-fn ()
  "Default execution monitor check when no other is specified, that never returns a problem."
  T)

(defun default-time-fn ()
  "Default time function - returns the current system time in seconds."
  (/ (get-internal-real-time) internal-time-units-per-second))

(defmethod notify-complete ((dispatch online-ldg-dispatch) (activity-id number))
  (mtk-utils-threading:mailbox-send (online-dispatch-mailbox dispatch) activity-id))

(defmethod find-missed-events ((ldg-dispatch ldg-dispatch) &key (time nil timep))
  "Returns 2 values: The list of events that were missed, and if events were missed
   the second value will be T if the events resulted in an invalid state space.
   Otherwise returns the empty list and nil.
   Missing an event does not necessarily constitute an invalid state space.

   TIME - if a value for TIME is provided it will be used in place of the
          CURRENT-TIME value in ldg-dispatch.

   Note: This method is almost identical to remove-missed-windows!"
  (with-slots (current-time ldg events-to-execute) ldg-dispatch
	;; decide which time to use
	(when (not (or timep (numberp time)))
		(setf time current-time))
	;; start the search for critical events
	(let ((missed-any? nil)
		  (missed-events nil)
		  (conflicts-needed-to-execute nil))
	  (loop for event in events-to-execute
		   
		 ;;grab the ub
		 do (let ((event-ub (gethash event (dispatch-ub ldg-dispatch))))
			  
			  ;;collect env of violated ubs
			  (loop for ub-pair in (lvs-pairs event-ub)
				 when (< (+ (pair-value ub-pair) (dispatch-time-tol ldg-dispatch)) time)
				 do 
				   (push (pair-environment ub-pair) conflicts-needed-to-execute)
				 ;; save a copy of the offending event
				   (pushnew event missed-events)
				   (setf missed-any? t))))
	  
	  ;;don't test this if nothing changed
	  (if missed-any?
		  (values missed-events (would-conflicts-remove-all-kernels? conflicts-needed-to-execute))
		  (values nil nil)))))


;;; ***********************************************************************
;;;
;;;                ACTIVITIES
;;;
;;; ***********************************************************************

;; A convenience structure for use by insert-new-constraint
(defstruct activities list)

(defclass activity ()
  ((start
	:initform '()
	:initarg :start
	:accessor activity-start
	:documentation "The start event.")
   (end
	:initform '()
	:initarg :end
	:accessor activity-end
	:documentation "The end event.")
   (lowerbound
	:initform '()
	:initarg :lowerbound
	:accessor activity-lowerbound
	:documentation "The lowerbound.")
   (upperbound
	:initform '()
	:initarg :upperbound
	:accessor activity-upperbound
	:documentation "The upperbound.")
   (primitive
	:type string
	:initform nil
	:initarg :primitive
	:accessor activity-primitive
	:documentation "The string primitive to be dispatched.")
   (id
	:type t
	:initform nil
	:initarg :id
 	:accessor activity-id
	:documentation "A unique id for each activity of a given dispatcher. 
A value of nil indicates this activity has not yet been assigned an id. 
nil is NOT a valid id.")))

(defmethod print-object ((activity activity) stream)
  (with-slots (start end lowerbound upperbound primitive id) activity
	  (format stream "<Activity id:~a ~a->~a [~a,~a] ~s>" id start end lowerbound upperbound primitive)))

(defparameter *unique-activity-id* 0
  "The generator used to ensure each activity has a unique id.")

(defun make-activity (start end lb ub primitive &key (id nil))
  "Creates an activity. 
   LB-EDGE and UB-EDGE must be the exact same list instances used 
     to create the labeled-weighted-directed-graph.
   PRIMITIVE will be the form dispatched.
   ID should be a number. If nil, a unique number will be generated
     for this activity."
  (make-instance 'activity :start start :end end
				 :lowerbound lb :upperbound ub
				 :primitive primitive 
				 :id (if id id (incf *unique-activity-id*))))


(defmethod activity-computed-lowerbound ((activity activity) (dispatch online-ldg-dispatch) &key (inf *inf*) (less-function #'<))
  (with-slots (ldg lb ub current-time) dispatch
	(let ((end-lb-lvs (gethash (activity-end activity) lb)) 
		  (uc (get-universal-clause (get-state-space ldg)))
		  (end-lb inf))
	  (unless (lvs-pairs end-lb-lvs)
		(return-from activity-computed-lowerbound (activity-lowerbound activity)))
	  (loop for pair in (lvs-pairs end-lb-lvs)
		 when (and (clause-subsumed? (pair-environment pair) uc)
				   (funcall less-function (pair-value pair) end-lb)) do
		 (setf end-lb (pair-value pair)))
	  (- end-lb current-time))))


(defmethod activity-computed-upperbound ((activity activity) (dispatch online-ldg-dispatch) &key (-inf *-inf*) (greater-function #'>))
  (with-slots (ldg lb ub current-time) dispatch
	(let* ((end-ub-lvs (gethash (activity-end activity) ub))
		   (uc (get-universal-clause (get-state-space ldg)))
		   (end-ub -inf))
	  (unless (lvs-pairs end-ub-lvs)
		(return-from activity-computed-upperbound (activity-upperbound activity)))
	  (loop for pair in (lvs-pairs end-ub-lvs)
		 when (and (clause-subsumed? (pair-environment pair) uc)
				   (funcall greater-function (pair-value pair) end-ub)) do
		 (setf end-ub (pair-value pair)))
	  (- end-ub current-time))))


(defmethod remove-completed-activities ((dispatch online-ldg-dispatch) (completed-activity-ids list))
  "Removes any newly-completed activities from active-activities, and adds them to completed-activities.
   Note that this method knows what activities are newly-completed by using the completed-activity-ids queue which is
   pushed to from the mailbox."
  (with-slots (active-activities completed-activities) dispatch
	(loop for id in completed-activity-ids do
		 (loop for activity in active-activities do
			  ;; If we found the activity corresponding to this id - move it from the active
			  ;; to the completed list
			  (when (equal (activity-id activity) id)
				(setf active-activities (delete activity active-activities))
				(setf completed-activities (push activity completed-activities)))))))


(defmethod remove-missed-activities ((dispatch online-ldg-dispatch) (end-events list))
  "Similar to remove-completed-activities."
  (with-slots (active-activities completed-activities) dispatch
	(loop for event in end-events do
		 (loop for activity in active-activities do
			  ;; If we just found the activity corresponding to this event event - move it
			  ;; to the other list
			  (when (equal (activity-end activity) event)
				(setf active-activities (delete activity active-activities))
				(setf completed-activities (push activity completed-activities)))))))


(defmethod remove-events-waiting-on-activities ((dispatch online-ldg-dispatch) (start-events list))
  "This method takes a list of events about to be executed, and removes from them
   any events waiting on an activity to complete. 

   This method is lazy in that it will remove ALL the events if any single event is waiting
   on an activity. This works because the (get-events-to-execute) method that creates the list
   of events passed to this method, will only return one event at a time or a zero-group."
  (with-slots (active-activities) dispatch	
	(if (and active-activities 
			 (find-if (lambda (activity) (find (activity-end activity) start-events)) active-activities))
		nil
		start-events)))


(defmethod event-waiting-on-activity? ((dispatch online-ldg-dispatch) event)
  "A helper that returns T iff the given event is waiting on activity that is currently
   being executed (and hence the event should not be scheduled), or else nil."
  (with-slots (active-activities activities) dispatch
	(let* ((all-activities (append active-activities activities))
		   (activity-waiting (find-if (lambda (activity) 
										(equalp (activity-end activity) event)) 
									  all-activities)))
	  (if activity-waiting
		  t
		  nil))))
	

(defmethod dispatch-activities ((dispatch online-ldg-dispatch) (start-events list) (callback function))
  "Find all activities with the given start-events, and call the callback to execute those activities.
   Activities are removed from dispatch's activities list and placed on the active-activities list."
  (with-slots (ldg activities active-activities) dispatch
	(loop for activity in activities
	   when (find (activity-start activity) start-events :test #'eql) do
		 ;; Remove the activity from the activities list
		 (setf activities (delete activity activities))
		 ;; Add the activity to the active-activities list
		 (push activity active-activities)
		 ;; Dispatch the activity by calling the callback
		 (when callback 
		   (funcall callback 
					(activity-id activity) 
					(activity-primitive activity) 
					(activity-computed-lowerbound activity dispatch) 
					(activity-computed-upperbound activity dispatch))))))


(defmethod insert-new-constraint ((graph mtk-graph:labeled-weighted-directed-graph) 
								  (activities activities) &key 
								  (start nil) 
								  (end nil) 
								  (lb 0) 
								  (ub 0) 
								  (guard nil) 
								  (primitive nil) &allow-other-keys)
  (with-slots (state-space) graph
	;; The hash-tables used in the graph structures use #'eql as their test. 
	;; This means an instance of a vertex list uniquely identifies an edge.
	;; i.e. Two vertex lists '(a b) and '(a b) are not #'eql.
	(let ((lb-edge (list end start))
		  (ub-edge (list start end)))
	  (mtk-graph:add-weighted-edge 
	   `(,lb-edge 
		 ,(mtk-graph:make-labeled-value-set-from-description state-space 'mtk-graph:strict-labeled-value-set #'< #'eql `((,(- lb) ,guard)))) graph)
	  (mtk-graph:add-weighted-edge 
	   `(,ub-edge 
		 ,(mtk-graph:make-labeled-value-set-from-description state-space 'mtk-graph:strict-labeled-value-set #'< #'eql `((,ub ,guard)))) graph)
	  (when primitive
		(push (make-activity start end lb ub primitive) (activities-list activities))))))


(defmethod computed-lowerbound ((end-event symbol)(lowerbound number)(current-time number)(dispatch online-ldg-dispatch) &key (inf *inf*) (less-function #'<))
  (with-slots (ldg lb ub) dispatch
	(let ((end-lb-lvs (gethash end-event lb)) 
		  (uc (get-universal-clause (get-state-space ldg)))
		  (end-lb inf))
	  (unless (lvs-pairs end-lb-lvs)
		(return-from computed-lowerbound lowerbound))
	  (loop for pair in (lvs-pairs end-lb-lvs)
		 when (and (clause-subsumed? (pair-environment pair) uc)
				   (funcall less-function (pair-value pair) end-lb)) do
		   (setf end-lb (pair-value pair)))
	  (- end-lb current-time))))


(defmethod computed-upperbound ((end-event symbol)(upperbound number)(current-time number)(dispatch online-ldg-dispatch) &key (-inf *-inf*) (greater-function #'>))
  (with-slots (ldg lb ub) dispatch
	(let* ((end-ub-lvs (gethash end-event ub))
		   (uc (get-universal-clause (get-state-space ldg)))
		   (end-ub -inf))
	  (unless (lvs-pairs end-ub-lvs)
		(return-from computed-upperbound upperbound))
	  (loop for pair in (lvs-pairs end-ub-lvs)
		 when (and (clause-subsumed? (pair-environment pair) uc)
				   (funcall greater-function (pair-value pair) end-ub)) do
		   (setf end-ub (pair-value pair)))
	  (- end-ub current-time))))
