(in-package :drake2)

(defclass ldg-dispatch ()

  ((ldg 
	:type mtk-graph:labeled-weighted-directed-graph
	:initarg :ldg
	:accessor dispatch-ldg)
   
   (current-time
	:initform 0
	:accessor dispatch-current-time
	:documentation "The current time of the dispatcher.")

   (time-tol
	:initform 0
	:initarg :time-tol
	:accessor dispatch-time-tol
	:documentation "The allowed tolerance after an event should be dispatched 
when an event can still be executed without being considered missed.")

   (lb
	:initform (make-hash-table)
	:accessor dispatch-lb
	:documentation "A hash table mapping from events to the known
lower bounds.")

   (ub
	:initform (make-hash-table)
	:accessor dispatch-ub
	:documentation "A hash table mapping from events to the known
upper bounds.")

   (start-event
	:initarg :start
	:accessor dispatch-start
	:documentation "The start vertex of the graph.")

   (S
	:initform nil
	:accessor dispatch-s
	:documentation "The set of executed events.")
   
   (events-to-execute
	:initform nil
	:accessor dispatch-events-to-execute
	:documentation "The events left to execute.")

   (trace
	:initform nil
	:accessor dispatch-trace
	:documentation "A list of the times of executions. (event time)")
))

(defmethod initialize-instance :after ((ldg-dispatch ldg-dispatch) &key)

  (with-slots (lb ub ldg events-to-execute) ldg-dispatch

	;;copy the events
	(setf events-to-execute (copy-list (get-vertices ldg)))
	
	;;initialize all the lb/ub to be -/+ inf
	(loop for event in (get-vertices ldg)
	   do (let ((state-space (get-state-space ldg)))
			
			(setf (gethash event lb) (make-instance 'mtk-graph:strict-labeled-value-set 
													:state-space state-space
													:strict-order #'>
													:value-equalp #'eql))
;			(add-pair-to-lvs! (gethash event lb)
;							  (make-instance 'labeled-pair 
;											 :environment 
;											 (get-universal-clause state-space)
;											 :value *-inf*))
			
			(setf (gethash event ub) (make-instance 'mtk-graph:strict-labeled-value-set 
										  :state-space state-space
										  :strict-order #'<
										  :value-equalp #'eql))
;			(add-pair-to-lvs! (gethash event ub)
;							  (make-instance 'labeled-pair 
;											 :environment 
;											 (get-universal-clause state-space)
;											 :value *inf*))
))))



(defmethod offline-execute! ((ldg-dispatch ldg-dispatch) &key (max-iterations *inf*))
  (debug-print-for :drake t "~%~% Beginning dispatch!******************************")
  (with-slots (current-time S ldg events-to-execute) ldg-dispatch
	(loop 
	   for iterations from 1 to max-iterations
	   until (null events-to-execute) 
	   ;;for each decision loop, keep the max time
	   do (measure-one-step-latency   
			(remove-missed-windows! ldg-dispatch) ;;technically not needed yet
			(let ((candidate-events (get-events-to-execute ldg-dispatch)))
			  (if (null candidate-events)
				  (setf current-time (+ 1 current-time))
				  (execute-events! ldg-dispatch candidate-events)))))))


(defmethod execute-events! ((ldg-dispatch ldg-dispatch) events)
  (with-slots (events-to-execute current-time trace S ldg) ldg-dispatch
	(debug-print-for :drake t "Executing ~a at ~a, S=~a remaining: ~a events-to-execute~%" events current-time S events-to-execute)

	;;move from to-executed to already executed
	(loop for event in events
	   do (push event S)
	   (setf events-to-execute (remove event events-to-execute)))
	   
	;;handle each event, includes error checking in case there are edges
	;;between the events executed simultaneously
	(loop for event in  events
	   do 
	   (propagate-bounds! ldg-dispatch event 0)
	   (push (list event (+ 0 current-time))
			 trace))))


(defmethod propagate-bounds! ((ldg-dispatch ldg-dispatch) event delay)
  (with-slots (current-time ldg lb ub S execution-time-ht) ldg-dispatch
	(let* ((state-space (get-state-space ldg))
		   (current-time-lvs (make-instance 'strict-labeled-value-set
											:state-space state-space
											:strict-order #'<
											:value-equalp #'eql)))
	  ;;initialize the current time lvs with a single element, the current time
	  (add-pair-to-lvs! current-time-lvs 
						(make-instance 'labeled-pair
									   :environment (get-universal-clause
													 state-space)
									   :value (+ delay current-time)))
	 
	  ;;propagate to successors
	  (dolist (edge (find-edges-with-source-vertex event ldg))
		
		(let* ((target-event (first (get-target-vertices edge)))
			   (target-ub (gethash target-event ub))
			   (edge-weight (get-weight edge ldg)))
		  (debug-print-for :drake t "Propagate ub edge ~a ~a into ~a~%"
						   edge edge-weight target-ub)
		  (unless (member target-event S)
			(update-lvs-with-binary-operation! target-ub edge-weight 
											   current-time-lvs #'+ )
			(debug-print-for :drake t "giving new ub for ~a: ~a~%" target-event target-ub))))

	  ;;propagate to predecessors
	  (dolist (edge (find-edges-with-target-vertex event ldg))
		
		(let* ((source-event (first (get-source-vertices edge)))
			   (source-lb (gethash source-event lb))
			   (edge-weight (get-weight edge ldg)))
		  (debug-print-for :drake t "Propagate lb edge ~a ~a into ~a~%"
						   edge edge-weight source-lb)
		  (unless (member source-event S)
			(update-lvs-with-binary-operation! source-lb 
											   current-time-lvs edge-weight 
											   #'- )
			(debug-print-for :drake t "giving new lb for ~a: ~a~%" source-event source-lb)))))))


(defmethod remove-missed-windows! ((ldg-dispatch ldg-dispatch) &key
								   (add-conflicts-and-commit-to-events nil))
  (with-slots (current-time ldg events-to-execute) ldg-dispatch

	(let ((missed-any? nil))
	  (loop for event in events-to-execute
		 
		 ;;grab the ub
		 do (let ((event-ub (gethash event (dispatch-ub ldg-dispatch))))
			
			  ;;collect env of violated ubs
			  (loop for ub-pair in (lvs-pairs event-ub)
				 when (< (+ (pair-value ub-pair) (dispatch-time-tol ldg-dispatch)) current-time) do 
				 (add-conflict! (get-state-space ldg) (pair-environment ub-pair))
				 (if add-conflicts-and-commit-to-events
				 	 (funcall add-conflicts-and-commit-to-events (list (pair-environment ub-pair)) nil))
				 (debug-print-for :drake t
								  "Missed event ~a window: ~a pair: ~a~%" 
								  event event-ub ub-pair)
				 (setf missed-any? t))))

	  ;;don't test this if nothing changed
	  (when missed-any?
		(assert (not (state-space-invalid? (get-state-space ldg))))
		;;assuming we haven't failed, prune for the new conflicts
		(prune-dispatch-of-invalid-environments! ldg-dispatch)))))

(defmethod prune-dispatch-of-invalid-environments! ((ldg-dispatch ldg-dispatch))
  "Prune the distance graph and the lb/ubs of any invalid environments."
  (with-slots (ub lb ldg) ldg-dispatch

	;;Strip out zero components invalidated
	(setf (get-zero-components ldg)
		  (remove-if #'(lambda (x) (not (state-clause-valid? (second x))))
					 (get-zero-components ldg)))

	;;Tell every lvs in the lb, ub, and edge weight hash tables to prune.
	(loop for lb-lvs being the hash-value of lb
		 do (prune-lvs-of-invalid-environments! lb-lvs))
	(loop for ub-lvs being the hash-value of ub
		 do (prune-lvs-of-invalid-environments! ub-lvs))
	(loop for edge-lvs being the hash-value of (get-weight-ht-edge ldg)
		 do (prune-lvs-of-invalid-environments! edge-lvs))))


(defmethod get-events-to-execute ((ldg-dispatch ldg-dispatch) &key
								  (event-waiting? nil)
								  (can-commit-to-events? nil) ;; Function pointer that allows something else to restrict event execution (ex., event env's)
								  (could-add-conflicts-and-commit-to-events? nil) ;; Similar to above function, but checks by adding a conflict first
								  (add-conflicts-and-commit-to-events nil)) ;; A callback to actually add conflicts and then commit.)
  "Get the next event to execute, if there is one at the current time. If not,
return nil.

This function has been modified to take in an optional argument, event-waiting?,
which should be a function that will test if each event can actually be executed
now or not (or, perhaps, it is waiting because it's the end event of an activity)"

  (let ((found-events nil))
	(debug-print-for :drake t "Finding something to execute, ~a remaining."
					 (dispatch-events-to-execute ldg-dispatch))

	;;now try zero-groups, ensuring that all the events in the zero-group
	;;are still ones we need to execute
	;;(format t "Dispatch events to execute: ~a~%" (dispatch-events-to-execute ldg-dispatch))

	;;(loop for zg in (get-zero-components (dispatch-ldg ldg-dispatch)) do
	;;	 (format t "   Zero: ~a~%" zg))

	(loop for zero-group in (get-zero-components (dispatch-ldg ldg-dispatch))
	   when (and (subsetp (first zero-group) (dispatch-events-to-execute ldg-dispatch))
				 (if event-waiting? ; Only applies if this function pointer passed in
				 	 (notany event-waiting? (first zero-group)) 
				 	 t)
				 (events-executable? ldg-dispatch (first zero-group)
									 :can-commit-to-events? can-commit-to-events?
									 :could-add-conflicts-and-commit-to-events? could-add-conflicts-and-commit-to-events?
									 :add-conflicts-and-commit-to-events add-conflicts-and-commit-to-events))
	   
	   do (debug-print-for :drake t "Group executable.")
	   (return-from get-events-to-execute (first zero-group)))

	;;try individual events
	(loop for event in (dispatch-events-to-execute ldg-dispatch)
	   when (and (if event-waiting? ; Only applies if this function pointer passed in
					 (not (funcall event-waiting? event))
					 t)
				 (events-executable? ldg-dispatch (list event)
									 :can-commit-to-events? can-commit-to-events?
									 :could-add-conflicts-and-commit-to-events? could-add-conflicts-and-commit-to-events?
									 :add-conflicts-and-commit-to-events add-conflicts-and-commit-to-events))
	   do (debug-print-for :drake t "Individual executable.")
	   (return-from get-events-to-execute  (list event) ))

	found-events))

(defmethod events-executable? ((ldg-dispatch ldg-dispatch) events &key 
							   (can-commit-to-events? nil) ;; Function pointer that allows something else to restrict event execution (ex., event env's)
							   (could-add-conflicts-and-commit-to-events? nil) ;; Similar to above function, but checks by adding a conflict first
							   (add-conflicts-and-commit-to-events nil)) ;; A callback to actually add conflicts and then commit.
  "Determine whether the event is executable at the current time."

  (with-slots (current-time ldg S) ldg-dispatch
	(debug-print-for :drake t "Testing executable? ~a at ~a" events
					 current-time)
	(let ((conflicts-needed-to-execute nil)
		  (result nil))
	  
	  ;; Check to make sure we can commit to all of the events
	  (when can-commit-to-events? ;; Only if this function pointer has been passed in
		(when (not (funcall can-commit-to-events? events))
		  ;;(format t "Can't initially commit to ~a~%" events)
		  (return-from events-executable? nil)))

	  ;;loop over pairs of events in the group, any negative edges between them
	  ;;would need to be discarded
	  (loop for event in events			;loop over pairs of events
		 do (loop for event2 in (remove event events)
			   do (let ((edge (first (find-edges-with-source-and-target-vertices
									  event event2 ldg))))
					(unless (null edge)
					  (loop for edge-pair in ;loop over the edge weights
						   (lvs-pairs (get-weight 
									   edge 
									   ldg))
						 ;;when there's a negative weight, set it as a conflict
						   when (> 0 (pair-value edge-pair))
						   do (push (pair-environment edge-pair) 
									conflicts-needed-to-execute))))))

	  ;;loop over all the events, in case it's a group
	  (loop for event in events
		 with event-lb
		 with event-ub
		 do (setf event-lb (gethash event (dispatch-lb ldg-dispatch)))
		 do (setf event-ub (gethash event (dispatch-ub ldg-dispatch)))

		 ;;collect environments of violated lbs
		 do (loop for lb-pair in (lvs-pairs event-lb)
			   when (> (- (pair-value lb-pair) (dispatch-time-tol ldg-dispatch)) current-time)
			   do (push (pair-environment lb-pair) conflicts-needed-to-execute)
			   (debug-print-for :drake t "Need to discard lb: ~a" lb-pair))

		 ;;collect env of violated ubs
		 do (loop for ub-pair in (lvs-pairs event-ub)
			   when (< (+ (pair-value ub-pair) (dispatch-time-tol ldg-dispatch))  current-time)
			   do (push (pair-environment ub-pair) conflicts-needed-to-execute)
			   (debug-print-for :drake t "Need to discard ub: ~a" ub-pair))

		 ;;check for zero-groups (group env)
		 do (loop for zero-group in (get-zero-components ldg)
			   ;;If the zero-group is not a subset of the events we're considering executing,
			   ;;it contains the event in question, and not all the other ones have been
			   ;;executed already
			   when (and (not (subsetp (first zero-group) events))
						 (member event (first zero-group)))

			   do (push (second zero-group) conflicts-needed-to-execute)
			   and do (debug-print-for :drake t "Need to discard zero: ~a" 
									   zero-group))

		 ;;collect unexecuted predecessors
		 do (loop for edge in (find-edges-with-source-vertex event ldg)
			   ;;unless this predecessor is executed
			   unless (or (member (first (get-target-vertices edge)) S)
						  ;;or it's one of the group you're considering executing at once
						  (member (first (get-target-vertices edge)) events))
			   ;;collect the env of all negative edge weights
			   do (loop for edge-pair in (lvs-pairs (get-weight edge ldg))
					 ;; This comparison must be ">=" and not just ">"
					 ;; to make sure ordering with 0 time-bounds is still included.
					 when (> 0 (pair-value edge-pair)) 
					 do (push (pair-environment edge-pair) conflicts-needed-to-execute)
					 (debug-print-for :drake t 
									  "Need to discard ordering: ~a ~a" edge edge-pair))))

	  ;;now test if it's executable


	  ;; Also: would the conflict remove all ATMS solutions for IR?



	  ;; This section down here needs to be totally changed.


	  (if could-add-conflicts-and-commit-to-events? ;; Function pointer exists?
		  (setf result (not (funcall could-add-conflicts-and-commit-to-events? conflicts-needed-to-execute events)))
		  (setf result (would-conflicts-remove-all-kernels? conflicts-needed-to-execute)))

	  ;;(format t "Conflicts: ~a~%" conflicts-needed-to-execute)
	  ;;(format t "Result: ~a~%" result)

	  (if result
		  nil			 ;makes problem infeasible, return nil
		  (progn		 ;is feasible, make the conflicts and return t
			(dolist (conflict conflicts-needed-to-execute)
			  
			  (add-conflict! (get-state-space ldg) conflict))

			;; Also add it to the ATMS 
			(when add-conflicts-and-commit-to-events
			  (funcall add-conflicts-and-commit-to-events conflicts-needed-to-execute events))
			


			;;if there were new conflicts, prune the graph and lb/ub
			(unless (null conflicts-needed-to-execute)
			  (prune-dispatch-of-invalid-environments! ldg-dispatch))



			;; Also! Filter out the events left to execute now that we've added more constraints. I.e., we 
			;; wont' be executing them all - this ensure that the dispatcher will actually stop when we're done ;-)
			(when can-commit-to-events?
			  (with-slots (events-to-execute S) ldg-dispatch
				(let* ((events-no-longer (loop for event in events-to-execute
											when (not (funcall can-commit-to-events? (list event))) collect event)))  
				  ;;(format t "No longer going to execute events: ~a~%" events-no-longer)
				  (loop for event in events-no-longer do
					   (push event S)
					   (setf events-to-execute (remove event events-to-execute))))))

			t)))))

