(in-package :drake2)

;;; Provides code to create random DTPs and TPNs

;;; Both of these methods try to write out the random problem created to local directories
;;; in the lisp ~ directory.

;;; Crucially, these methods allow you to rerun the last random problem. That way, you can actually
;;; run random instances until you find an issue and then re-run that instance until you fix it.
;;; This implementation cannot save the random seed through lisp crashing, however. Setting
;;; new-problem? to nil recreates the last problem.

;;; These return a list structure that's not really clean, see the code or the unpacking in 
;;; dtp-benchmarking.lisp to see what goes where. Also, you can probably guess what the data
;;; is from the values.

;;; Interface functions
;;; random-structured-dtp: create a dtp with some parameters. The search distances 
;;;     typically should scale with the number of activities - see run-benchmarking
;;;     for a reasonable scaling.
;;; random-tpn: Returns random TPN.

;; Used to store the random seed so that problems can be re-run.
(defvar *last-random-problem-seed*)

(defun random-structured-ldg (num-activities num-disjuncts disjunct-order constraints-per-node 
							  activity-duration cutoff-rad search-rad plan-length plan-width
							  disjunct-search-radius new-problem? controllable? &key (problem-number 0))
	
  ;;if there isn't a state, make one
  (when (null *random-state*)
	(setf *last-random-problem-seed* (make-random-state t)))

  ;;if we're not supposed to use a new one, restore the old one
  (when (not new-problem?)
	(setf *random-state* *last-random-problem-seed*))

  ;;always stash the current one, and make a copy
  (setf *last-random-problem-seed* *random-state*)
  (setf *random-state* (make-random-state *last-random-problem-seed*))

  (let* ((state-space (make-instance 'state-space-w-conflicts))
		 (ldg (make-instance 'labeled-weighted-directed-graph :state-space state-space))
		 (events nil)
		 (constraints nil))
	
	
	;;Figure out the events and their locations + activity based constraints.
	(loop for i from 1 to num-activities
	   do (let* ((this-one-controllable?  (or controllable? ;if uncontrollable, half the activities are uncontrollable
											  (not (> (bounded-random 1 2) 1))))
				 (x (bounded-random 0 plan-length))
				 (y (bounded-random (- 0 plan-width) plan-width))
				 (ub (bounded-random 1 activity-duration))
				 (lb (bounded-random 0 ub))
				 (start-event-symbol (intern (string-upcase (concatenate 'string (write-to-string i) "s"))))
				 (end-event-symbol (intern (string-upcase (concatenate 'string (write-to-string i) "e"))))
				 )

										;(format t "controllable? ~a ~a ~a ~%" (not (<= i (/ num-activities 2))) i (/ num-activities 2))
			
			
			(when (equal 1 i)
			  (setf x 0)
			  (setf y 0))
			(push (list start-event-symbol x y) events)
			(push (list end-event-symbol (+ x ub) y) events)
			(push (list start-event-symbol end-event-symbol lb ub nil this-one-controllable?) constraints)))

										;	(print (reverse events))

	;;Genrate the basic non-activity contraints.
	(loop for event-a in events
	   do (loop for i from 1 to (1- constraints-per-node) ;try to generate the right number
			 do (let ((event-b (find event-a (shuffle-list events) :test #'(lambda (x y) (event-close-and-unconstrained x y constraints search-rad))))
					  (start nil)
					  (end nil))
				  (unless (null event-b)
					(if (< (second event-a) (second event-b))
						(block a-start
						  (setf start event-a)
						  (setf end event-b))
						(block b-start
						  (setf start event-b)
						  (setf end event-a)))
					(let* ((dist (bound-num (norm-distance-events start end) cutoff-rad))
						   (ub (bounded-random (/ dist 2) (* 2 dist)))
						   (lb (bounded-random (- 0 dist) (/ dist 2))))

					  (push (list (first start) (first end) lb ub nil t) constraints))))))


	(let ((current-variable-symbol nil)
		  (disjunct-focus-event nil))
	  (loop for disjunct from 1 to num-disjuncts
		  
		 do (setf current-variable-symbol (intern (string-upcase (concatenate 'string "v" (write-to-string disjunct)))))
		 do (setf disjunct-focus-event (first (shuffle-list events))) ;pick an event to place the disjuncts near
		 do (add-variables-from-descriptions! state-space
											  (list (list current-variable-symbol 
														  ':domain (loop for i from 1 to disjunct-order collecting (list i))))
											  ) ;make the state variable
		 do (loop for assignment from 1 to disjunct-order ;place the choices
			   do (let ((constraint-to-modify  (find disjunct-focus-event (shuffle-list constraints) 
													 :test #'(lambda (x y) (constraint-near-focus x y events disjunct-search-radius)))))
					(when (null constraint-to-modify) ;if it didn't find one close, just use the first
					  (setf constraint-to-modify (first constraints))
					  (debug-print-for :cdtn t "falling back to default constraint"))

					(if (null (fifth constraint-to-modify))
						(setf (fifth constraint-to-modify) (get-clause-from-descriptions state-space (list (list '= current-variable-symbol assignment))))
						(block make-new-constraint
						  (let* ((start (find (first constraint-to-modify) events :test #'(lambda (x y) (eql x (first y)))))
								 (end (find (second constraint-to-modify) events :test #'(lambda (x y) (eql x (first y)))))
								 (dist (bound-num (norm-distance-events start end) cutoff-rad))
								 (ub (bounded-random (/ dist 2) (* 2 dist)))
								 (lb (bounded-random (- 0 dist) (/ dist 2))))
							(push (list (first start) (first end) lb ub 
										(get-clause-from-descriptions state-space (list (list '= current-variable-symbol assignment)))
										t) constraints))))))

		   		  
		 ))

	(loop for constraint in constraints
	   do (debug-print-for :cdtn t "inserting random constraint ~a" constraint)	 
	   do (when (null (fifth constraint))
			(setf (fifth constraint) (get-universal-clause state-space)))
	   do (update-labeled-weighted-edge (list (list (first constraint) 
													(second constraint))
											  (fourth constraint) 
											  (fifth constraint))
										ldg)
	   (update-labeled-weighted-edge (list (list (second constraint) 
												 (first constraint))
										   (- (third constraint)) 
										   (fifth constraint))
									 ldg)
	   )

	;;:start (first constraint)
	;;:end (second constraint)
	;;:lb (third constraint)
	;;:ub (fourth constraint)
	;;:guard (fifth constraint)
	;;:controllable? (sixth constraint)
											  
		 

	(with-open-file (stream (if controllable?  (format nil "~~/lstnBenchmarks/example~a_~a_~a.dtp" disjunct-order
													   num-activities problem-number)
								(format nil "~~/dtpuBenchmarks/example~a_~a_~a.dtp" disjunct-order
										num-activities problem-number)) :direction :output :if-exists :supersede)
	  (format stream "~a ~a ~a ~a ~a ~a ~a ~a ~a ~a ~a ~a ~%" num-activities num-disjuncts disjunct-order constraints-per-node 
			  activity-duration cutoff-rad search-rad plan-length plan-width
			  disjunct-search-radius new-problem? controllable?)
	  (format stream "~a~%" constraints))

										;(print (nreverse constraints))
										;	(dot-output-dgraph dtp)
	(values ldg constraints)
	))

(defun constraint-near-focus (focus-event constraint events disjunct-radius)
  (let ((constraint-start-event (find (first constraint) events :test #'(lambda (x y) (eql x (first y))))))
	(and (< (norm-distance-events focus-event constraint-start-event) disjunct-radius))))


(defun bound-num (num bound)
  (if (> num bound)
	  bound
	  num))

(defun event-close-and-unconstrained (event candidate constraints search-rad)
  (and (not (eql (first event) (first candidate)))
	   (<=  (norm-distance-events event candidate) search-rad)
	   (not (member (list (first event) (first candidate)) constraints :test #'eq-random-constraints))))

(defun eq-random-constraints (c1 c2)
  (or (and (eql (first c1) (first c2))
		   (eql (second c1) (second c2)))
	  (and (eql (first c1) (second c2))
		   (eql (second c1) (first c2)))))

(defun norm-distance-events (e1 e2)
  (norm-distance (second e1) (second e2) 
				 (third e1) (third e2)))

(defun norm-distance (x1 x2 y1 y2)
  (sqrt (+ (expt (- x1 x2) 2)
		   (expt (- y1 y2) 2))))

(defun bounded-random (lb ub)
	"Generates a random bounded integer. Take the floor of the inputs to be sure."
  (let ((lbi (truncate lb))
		(ubi (truncate ub)))
	(+ lbi (random (1+ (- ubi lbi))))))


(defmethod execution-trace-valid? ((ldg labeled-weighted-directed-graph) constraints trace)
  "Test an execution trace against the cdtn it was supposed to come from. Make sure all the constraints
were met. Return t if valid and nil otherwise."

  (let ((exec-time-ht (make-hash-table))
		(full-assignment (get-valid-full-assignment (get-state-space ldg)))
		)

	(debug-print-for :drake t
									"Checked against: ~a ~%" full-assignment)
	(loop for exec in trace
	   do (setf (gethash (first exec) exec-time-ht) (second exec)))

	(loop for constraint in constraints
	   ;;check that e_xy <= y-x
	   do (debug-print-for :drake t "Checking ~a constraint: ~a"
						   constraint
						   (- (gethash (second constraint) exec-time-ht)
							  (gethash (first constraint) exec-time-ht)))
	   always (or (not (clause-subsumed? full-assignment (fifth constraint)))
				
				  (and (>= (fourth constraint)
						   (- (gethash (second constraint) exec-time-ht)
							  (gethash (first constraint) exec-time-ht)))
					   (<= (third constraint)
						   (- (gethash (second constraint) exec-time-ht)
							  (gethash (first constraint) exec-time-ht))))))))


(defun shuffle-list (l)
  (loop for i below (length l) do
	(rotatef
	 (elt l i)
	 (elt l (random (length l)))))
  l)