(in-package :drake2)


(defmethod tpn-to-ldg ((tpn tpn::temporal-plan-network) &key 
					   (zero-offset 0.0001) 
					   (event-to-symbol-function #'default-event-to-symbol-function))
  "Takes a TPN and converts it to an LDG and activity list. 

   It is assumed the TPN has no choices (i.e. is an STN).
   This is a reasonable assumption, because a TPN has no way
   of expression the notion of uncontrollable choice that
   Drake uses.

   ZERO-OFFSET is the value used for the lowerbound when constraints of the form [0,x], for x>0.
   Read the explanation below.

   EVENT-TO-SYMBOL-FUNCTION is a function, given a TPN event object, returns a symbol that will be used 
   in the ldg to uniquely identify that event.

   Returns two values:
   1. A labeled-weighted-directed graph representing the timebounds with their labeled-value-sets.
   2. An object containing a list of activities."

  (let* ( ;; Create the labeled-weighted-directed graph
		 (var-descriptions '())
		 (state-space (make-instance 'state-space-w-conflicts :variable-descriptions var-descriptions))
		 (graph (make-instance 'mtk-graph:labeled-weighted-directed-graph :state-space state-space))
		 ;; Create the data structure to hold onto activities
		 (activities (make-activities)) )

	(loop for episode in (tpn::tpn-episodes tpn) do
		 (let ((lowerbound (to-pinf (tpn::episode-lowerbound episode)))
			   (upperbound (to-pinf (tpn::episode-upperbound episode)))
			   (primitive (tpn::episode-primitive episode)))
		   ;; When a temporal constraint takes the form [0,x], where x is a positive, non-zero value,
		   ;; we need to replace it with the constraint [0+,x], where 0+ represents a slightly positive value.
		   ;;   Reason: Given the events, a->b->c->d, let's say b->c is temporally related by [0,x]. 
		   ;; The bounds a->b and c->d have primitives, but we aren't concerned with their timebounds.
		   ;; Since Drake dispatches using the lowertime bound, it's possible that the function 
		   ;; get-events-to-execute will return event c before it returns b. In which case we would 
		   ;; start executing the c->d primitive, even before the a->b primitive has reported that it completed.
		   (when (and (= lowerbound 0)
					  (> upperbound 0))
			 (setf lowerbound (min zero-offset upperbound)))
		   ;; for each episode, insert a new lb and ub edge into the ldg
		   (insert-new-constraint graph activities
								  :start (funcall event-to-symbol-function (tpn::episode-start episode))
								  :end (funcall event-to-symbol-function (tpn::episode-end episode))
								  :lb lowerbound
								  :ub upperbound
								  :guard nil
								  :primitive primitive ;; should be nil if there is no primitive)
								  )
		   ;; (format t "(insert graph :start ~a :end ~a :lb ~a :ub ~a :guard '())~%" 
		   ;; 		 (funcall event-to-symbol-function (tpn::episode-start episode))
		   ;; 		 (funcall event-to-symbol-function (tpn::episode-end episode))
		   ;; 		 lowerbound
		   ;; 		 upperbound )
		   )
		 )
	(values graph activities)))


;; Added by Steve Levine 5/22/2013
(defmethod tpn-to-ldg-with-choice ((tpn tpn::temporal-plan-network) &key 
								   (zero-offset 0.0001) 
								   (event-to-symbol-function #'default-event-to-symbol-function)
								   (variable-name-fn nil)
								   (value-name-fn nil))
  "Takes a TPN and converts it to an LDG and activity list. 

   Unlike the above function, this is capable of dealing with choices. In other words, it is
   capable of converting a true TPN (not just an STN) to an LDG.

   ZERO-OFFSET is the value used for the lowerbound when constraints of the form [0,x], for x>0.
   Read the explanation below.

   EVENT-TO-SYMBOL-FUNCTION is a function, given a TPN event object, returns a symbol that will be used 
   in the ldg to uniquely identify that event.

   VARIABLE-NAME-FN is a function which, given a decision event (converted before to a symbol),
   returns a string for what the corresponding choice variable should be called.

   VALUE-NAME-FN is a function which, given a decision event e-d (symbol) and a second event e-c (symbol)
   such that there's an edge e-d --> e-c (i.e., e-c is the result of making one of the
   decisions), returns a value name for the choice corresponding to e-d. This can be 
   a number (for things like x_3 = 1, 2, 3, ...), a string, or other.

   Returns two values:
   1. A labeled-weighted-directed graph representing the timebounds with their labeled-value-sets.
   2. An object containing a list of activities."

  (let* ((var-descriptions nil)
		 (state-space nil)
		 (graph nil)
		 ;; Create the data structure to hold onto activities
		 (activities (make-activities))
		 ;; Other useful variables
		 (var-ht-choicenode (make-hash-table :test #'equalp))
		 (variables nil)
		 (domain-ht-var (make-hash-table :test #'equalp))
		 (event-env-ht (make-hash-table :test #'equalp))
		 (episode-env-ht (make-hash-table :test #'equalp)))

	;; Fill in the default variable naming function, it if wasn't specified
	;; (note that this could possibly be a keyword argument, but for consistency
	;; with the following method it isn't.
	(when (null variable-name-fn)
	  (setf variable-name-fn 
			#'(lambda (e)
				(format nil "x_~a" e))))
	;; Similarly, set the default variable's value (i.e, domain values) naming
	;; function. This is harder to make a keyword argument since it needs state -
	;; here acccomplished via a closure. This default value naming function
	;; names values incrementally in the order they were listed (arbitrary).
	(when (null value-name-fn)
	  (let* ((var-counter-ht (make-hash-table :test #'equalp)))
		(setf value-name-fn
			  #'(lambda (e-d e-v)
				  (declare (ignore e-v))
				  (if (gethash e-d var-counter-ht)
					  (setf (gethash e-d var-counter-ht) (incf (gethash e-d var-counter-ht)))
					  (setf (gethash e-d var-counter-ht) 1))))))
	

	;; Choose symbolic variables for each of the TPN decisions (i.e., x=1, y=2, etc.)
	(loop for node in (tpn::tpn-decisions tpn)
	   with variable = nil do
	   ;;(setf variable (intern (format nil "x_~a" (funcall event-to-symbol-function node) :keyword)))
	   (setf variable (intern (funcall variable-name-fn (funcall event-to-symbol-function node)) :keyword))
	   (setf (gethash node var-ht-choicenode) variable)
	   (push variable variables))

	;; Initialize some variables
	(loop for event in (tpn::tp-events tpn) do
		 (setf (gethash event event-env-ht) nil))
	(loop for episode in (tpn::tpn-episodes tpn) do
		 (setf (gethash episode episode-env-ht) nil))

	;; Now, iterate over all decision variables and label all sub-trees starting at those variables
	;; with that assignment. Later on, we'll remove ones with multiple conflicting assignments and also
	;; check for errors.
	(loop for decision-event in (tpn::tpn-decisions tpn)
	   with outgoing-episodes
	   with variable do
	   (setf variable (gethash decision-event var-ht-choicenode))
	   (setf outgoing-episodes (get-tpn-episodes-starting-at-event tpn decision-event))
	   (loop for episode in outgoing-episodes
		  with value do
		  (setf value (funcall value-name-fn 
							   (funcall event-to-symbol-function decision-event)
							   (funcall event-to-symbol-function (tpn::episode-end episode))))

		  ;; Push this value to the variable's domain
		  (setf (gethash variable domain-ht-var)
				(push value (gethash variable domain-ht-var)))
		  ;; Visit all events and episodes in the subtree and add the assignment var=val
		  (visit-subtree-starting-at-episode tpn
											 episode
											 :event-visit-fn (lambda (event)
															   (setf (gethash event event-env-ht) 
																	 (update-assignments (gethash event event-env-ht)
																						 variable
																						 value)))
											 :episode-visit-fn (lambda (episode)
																 (setf (gethash episode episode-env-ht) 
																	   (update-assignments (gethash episode episode-env-ht)
																						   variable
																						   value))))))
	;; Now that we've visited and labeled every subtree of every decision variable, we have enough information to 
	;; determine the labeling of every variable. For each label that has exactly one assignment to a given variable, ex
	;; just x=1, then that's part of the label for that variable. If it has the whole domain, like x=1, x=2, x=3, then 
	;; it means it's after that variable happened, so it's actually none of them and remove them. If it has just some of 
	;; them - ex., x=1 and x=2 but not x=3, then the TPN isn't well-formed (all decision nodes didn't merge back at same
	;; point), and throw an error.
	(loop for episode in (tpn::tpn-episodes tpn)
	   with assignments do
	   (setf assignments (adjust-assignments (gethash episode episode-env-ht) variables domain-ht-var))
	   (if (equalp assignments -1)
		   (error "Error in TPN-->LDG conversion: malformed TPN, some decision threads do not merge back at the same event.")
		   (setf (gethash episode episode-env-ht) assignments)))
	;; Now, do the same thing for events.
	(loop for event in (tpn::tp-events tpn)
	   with assignments do
	   (setf assignments (adjust-assignments (gethash event event-env-ht) variables domain-ht-var))
	   (if (equalp assignments -1)
		   (error "Error in TPN-->LDG conversion: malformed TPN, some decision threads do not merge back at the same event.")
		   (setf (gethash event event-env-ht) assignments)))

	;; At this point, we've finished computing the environment of every episode and event in the TPN (and soon to be LDG).
	;; Put this in a variable description form for Drake.
	(setf var-descriptions (loop for variable in variables 
							  collecting (list variable :domain (loop for val in (gethash variable domain-ht-var) 
																   collecting (list val)))))

	;;(format t "~a~%" var-descriptions)
	(setf state-space (make-instance 'state-space-w-conflicts :variable-descriptions var-descriptions))
	(setf graph (make-instance 'mtk-graph:labeled-weighted-directed-graph :state-space state-space))

	(loop for episode in (tpn::tpn-episodes tpn) do
		 (let ((lowerbound (to-pinf (tpn::episode-lowerbound episode)))
			   (upperbound (to-pinf (tpn::episode-upperbound episode)))
			   (primitive (tpn::episode-primitive episode)))
		   ;; When a temporal constraint takes the form [0,x], where x is a positive, non-zero value,
		   ;; we need to replace it with the constraint [0+,x], where 0+ represents a slightly positive value.
		   ;;   Reason: Given the events, a->b->c->d, let's say b->c is temporally related by [0,x]. 
		   ;; The bounds a->b and c->d have primitives, but we aren't concerned with their timebounds.
		   ;; Since Drake dispatches using the lowertime bound, it's possible that the function 
		   ;; get-events-to-execute will return event c before it returns b. In which case we would 
		   ;; start executing the c->d primitive, even before the a->b primitive has reported that it completed.
		   (when (and (= lowerbound 0)
					  (> upperbound 0))
			 (setf lowerbound (min zero-offset upperbound)))


		   ;;(break)


		   ;; for each episode, insert a new lb and ub edge into the ldg
		   (insert-new-constraint graph activities
								  :start (funcall event-to-symbol-function (tpn::episode-start episode))
								  :end (funcall event-to-symbol-function (tpn::episode-end episode))
								  :lb lowerbound
								  :ub upperbound
								  :guard (gethash episode episode-env-ht)
								  :primitive primitive))) ;; Will be nil if there's no primitive

	;; Now, convert event-env-ht to be in terms of symbols, and not events!
	(let* ((event-env-ht-symbolized (make-hash-table :test #'equalp)))
	  (loop for event being the hash-key in event-env-ht
		 with env do
		 (setf env (gethash event event-env-ht))
		 (setf (gethash (funcall event-to-symbol-function event) event-env-ht-symbolized) (get-clause-from-descriptions state-space env)))
	  (values graph activities event-env-ht-symbolized episode-env-ht domain-ht-var))))

  

;; Some helper functions
(defun get-tpn-episodes-starting-at-event (tpn event)
  (loop for episode in (tpn::tpn-episodes tpn)
	   when (equalp (tpn::episode-start episode) event)
	   collect episode))

;; Recursive function that traverses a graph starting at the given event.
;; Note: assumes the graph is acylic! Otherwise it will run forever.
(defun visit-subtree-starting-at-event (tpn event &key
										(event-visit-fn #'(lambda (event) 
															(declare (ignore event))
															nil))
										(episode-visit-fn #'(lambda (episode) 
															  (declare (ignore episode))
															  nil)))
  (funcall event-visit-fn event)
  (loop for episode in (get-tpn-episodes-starting-at-event tpn event) do
	   (visit-subtree-starting-at-episode tpn 
										  episode
										  :event-visit-fn event-visit-fn
										  :episode-visit-fn episode-visit-fn)))

;; Along with aove, traverses a graph starting at an episode. Graph must be
;; acyclic.
(defun visit-subtree-starting-at-episode (tpn episode &key
										  (event-visit-fn #'(lambda (event) 
															  (declare (ignore event))
															  nil))
										  (episode-visit-fn #'(lambda (episode)
																(declare (ignore episode))
																nil)))
  (funcall episode-visit-fn episode)
  (visit-subtree-starting-at-event tpn 
								   (tpn::episode-end episode)
								   :event-visit-fn event-visit-fn
								   :episode-visit-fn episode-visit-fn))

;; Helper function that, given a set of assignments, also adds the assignment
;; variable=value if it isn't already in that list.
(defun update-assignments (assignments variable value)
  (let* ((assignment `(= ,variable ,value)))
	(unless (find assignment assignments :test #'equalp)
	  (push assignment assignments)))
  assignments)

;; Helper function that removes inconsistent assignments, like x=1 and x=2.
;; It's a bit special in that it deals with some consistencies like you'd expect,
;; but others will throw an error - it's designed to work with the TPN to LDG conversions above.
(defun adjust-assignments (assignments variables domain-ht-var)
  (loop for variable in variables
	 with occurences do
	 (setf occurences (loop for a in assignments when (equalp (second a) variable) collect a))
	 (cond 
	   ;; All the domain values are assigned - this means we're after the variable applies, so remove
	   ;; its assignments.
	   ((= (length occurences) (length (gethash variable domain-ht-var))) 
		(setf assignments (loop for a in assignments unless (equalp (second a) variable) collect a)))
	   ;; There are no occurences of variable - that's fine
	   ((= (length occurences) 0)
		nil)
	   ;; There is exactly one occurence - also fine
	   ((= (length occurences) 1)
		nil)
	   ;; Anything else - there are *some* but not *all* - this is an error!
	   (t
		(return-from adjust-assignments -1))))
  assignments)

(defun default-event-to-symbol-function (event)
  (let* ((event-name (tpn::event-name event))
		 (event-symbol (intern event-name :keyword)))
	event-symbol))


