;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package mtk-graph-algorithms)

;; For references relating to this code see:
;;   [1] T. H. Cormen, C. E. Leiserson, R. L. Rivest, and C. Stein. 
;;       Introduction to Algorithms. The MIT Press,
;;       Cambridge, Massachusetts, 2nd edition, 2001.

;;; From[1] pg. 630
;;;
;;; FLOYD-WARSHALL(W) 
;;;   Modifications:
;;;     * The snapshot of the distance is not kept for every iteration.
(defmethod floyd-warshall ((g mtk-graph:weighted-directed-graph)
						   &key
						   (add-function #'+)
						   (min-function #'min)
						   (zero 0)
						   (inf *inf*))
  " FLOYD-WARSHALL(W) 
    1 n <- rows[W] 
    2 D^{(0)} <- W 
    3 for k <- 1 to n 
    4   do for i <- 1 to n 
    5     do for j <- 1 to n 
    6       do d^{(k)}_{ij} <- min(d^{(k-1)}_{ij},d^{(k-1)}_{ik}
                                                  + d^{(k-1)}_{kj})
    7 return D^{(n)"
  (call-next-method g
					:weight-function #'(lambda (e_i)
										 (mtk-graph:get-weight e_i g))
					:add-function add-function
					:min-function min-function
					:zero zero
					:inf inf))
(defmethod floyd-warshall ((g mtk-graph:directed-graph)
						   &key
						   (weight-function nil)
						   (add-function #'+)
						   (min-function #'min)
						   (zero 0)
						   (inf *inf*))
  " FLOYD-WARSHALL(W) 
    1 n <- rows[W] 
    2 D^{(0)} <- W 
    3 for k <- 1 to n 
    4   do for i <- 1 to n 
    5     do for j <- 1 to n 
    6       do d^{(k)}_{ij} <- min(d^{(k-1)}_{ij},d^{(k-1)}_{ik}
                                                  + d^{(k-1)}_{kj})
    7 return D^{(n)"
  (assert weight-function)
  (assert (functionp weight-function))
  (let ((d (make-hash-table :test #'equal))
		(v (mtk-graph:get-vertices g))
		(e (mtk-graph:get-edge-list (mtk-graph:get-edges g))))
;;; 1 n <- rows[W] 
;;; 2 D^{(0)} <- W 
    (dolist (i v)
      (dolist (j v)
		(if (equal i j)
			(setf (gethash (list i j) d) zero)
			(setf (gethash (list i j) d) inf))))
    (dolist (e_i e)
      (dolist (i (mtk-graph:get-source-vertices e_i))
		(dolist (j (mtk-graph:get-target-vertices e_i))
		  (setf (gethash (list i j) d) 
				(funcall min-function
						 (gethash (list i j) d)
						 (let ((w_i (funcall weight-function e_i)))
						   (if (null w_i)
							   inf
							   w_i)))))))
;;; 3 for k <- 1 to n 
    (dolist (k v)
;;; 4   do for i <- 1 to n 
      (dolist (i v)
;;; 5     do for j <- 1 to n 
		(dolist (j v)
;;; 6       do d^{(k)}_{ij} <- min(d^{(k-1)}_{ij},d^{(k-1)}_{ik} + d^{(k-1)}_{kj})
		  (setf (gethash (list i j) d)
				(funcall min-function (gethash (list i j) d)
						 (funcall add-function
								  (gethash (list i k) d)
								  (gethash (list k j) d)))))))
;;; 7 return D^{(n)}
    d))

;;; From[1] pg. 633
;;;
;;; TRANSITIVE-CLOSURE(G) 
;;; 1  n <- |V[G]| 
;;; 2  for i <- 1 to n 
;;; 3    do for j <- 1 to n 
;;; 4      do if i = j or (i,j) \in E[G] 
;;; 5         then t^{(0)}_{ij} <- 1 
;;; 6         else t^{(0)}_{ij} <- 0 
;;; 7  for k <- 1 to n 
;;; 8    do for i <- 1 to n 
;;; 9      do for j <- 1 to n 
;;; 10       do t^{(k)}_{ij} <- t^{(k-1)}_{ij} \vee 
;;;                             (t^{(k-1)}_{ik} \wedge t^{(k-1)}_{kj}) 
;;; 11 return T (n) 

;;; TRANSITIVE-CLOSURE(G) 
(defmethod transitive-closure ((g mtk-graph:directed-graph))
  (let ((t-ht (make-hash-table :test #'equal))
	(v (mtk-graph:get-vertices g)))
;;; 1  n <- |V[G]| 
;;; 2  for i <- 1 to n 
    (dolist (i v)
;;; 3    do for j <- 1 to n 
      (dolist (j v)
;;; 4      do if i = j or (i,j) \in E[G] 
	(if (or (eq i j)
		(mtk-graph:find-edges-with-source-and-target-vertices i j g))
;;; 5         then t^{(0)}_{ij} <- 1 
	    (setf (gethash (list i j) t-ht) t)
;;; 6         else t^{(0)}_{ij} <- 0 
	    (setf (gethash (list i j) t-ht) nil))))
;;; 7  for k <- 1 to n 
    (dolist (k v)
;;; 8    do for i <- 1 to n 
      (dolist (i v)
;;; 9      do for j <- 1 to n 
	(dolist (j v)
;;; 10       do t^{(k)}_{ij} <- t^{(k-1)}_{ij} \vee 
;;;                             (t^{(k-1)}_{ik} \wedge t^{(k-1)}_{kj}) 
	  (setf (gethash (list i j) t-ht) 
		(or (gethash (list i j) t-ht)
		    (and (gethash (list i k) t-ht)
			 (gethash (list k j) t-ht)))))))
;;; 11 return T (n) 
    (values #'(lambda (i j) (gethash (list i j) t-ht)) t-ht)))
