;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package mtk-graph-algorithms)

;;; For references relating to this code see:
;;;   [1] T. H. Cormen, C. E. Leiserson, R. L. Rivest, and C. Stein. 
;;;       Introduction to Algorithms. The MIT Press,
;;;       Cambridge, Massachusetts, 2nd edition, 2001.

;;; From [1] pg. 541
;;;
;;; DFS(G) 
;;; 1 for each vertex u \in V[G] 
;;; 2   do color[u] <- WHITE 
;;; 3      \pi[u] <- NIL 
;;; 4 time <- 0 
;;; 5 for each vertex u \in V[G] 
;;; 6   do if color[u] = WHITE 
;;; 7       then DFS-VISIT(u) 

;;; DFS(G) 
;;    Modifications:
;;      * "vertices" key was added to enable user defined search ordering at the
;;        top level.
;;      * Added a hash table that keeps track of the children of the search
;;        tree.
(defmethod dfs ((G mtk-graph:graph) &key vertices)
  (if (null vertices) (setf vertices (mtk-graph:get-vertices G)))
  (let (;; Color of the vertex:
	;;   white = not discovered, gray = discovered, and black = finished
	(color (make-hash-table))
	;; Parent vertex
	(p (make-hash-table))
	;; Global time
	(time 0)
	;; Discovery time
	(d (make-hash-table))
	;; Finishing time
	(f (make-hash-table)) 
	;; Keep track of the children during the search
	(c (make-hash-table)))
;;; 1 for each vertex u \in V[G] 
    (dolist (u (mtk-graph:get-vertices G))
;;; 2   do color[u] <- WHITE 
      (setf (gethash u color) 'white)
      ;; Modification: Initialize children
      (setf (gethash u c) nil)
;;; 3      \pi[u] <- NIL 
      (setf (gethash u p) nil))
;;; 4 time <- 0 
    ;; This is done in let (above).
;;; 5 for each vertex u \in V[G] 
    (dolist (u vertices)
;;; 6   do if color[u] = WHITE 
      (if (eq (gethash u color) 'white)
;;; 7       then DFS-VISIT(u) 
	  (setf time (dfs-visit u color p c time d f g))))
  (values p c d f color)))

;;; DFS-VISIT(u) 
;;; 1 color[u] <- GRAY ;;White vertex u has just been discovered. 
;;; 2 time <- time + 1 
;;; 3 d[u] <- time 
;;; 4 for each v \in Adj[u] ;;Explore edge(u,v). 
;;; 5   do if color[v] = WHITE 
;;; 6       then \pi[v] <- u 
;;; 7            DFS-VISIT(v) 
;;; 8 color[u] <- BLACK ;;Blacken u; it is finished. 
;;; 9 f[u] <- time <- time + 1

;;; DFS-VISIT(u)
(defmethod dfs-visit (u color p c time d f (g mtk-graph:graph))
;;; 1 color[u] <- GRAY ;;White vertex u has just been discovered. 
  (setf (gethash u color) 'gray)
;;; 2 time <- time + 1 
  (incf time 1)
;;; 3 d[u] <- time
  (setf (gethash u d) time)
;;; 4 for each v \in Adj[u] ;;Explore edge(u,v). 
  (dolist (v (mtk-graph:find-children-of-vertex u g))
;;; 5   do if color[v] = WHITE 
    (when (eq (gethash v color) 'white)
;;; 6       then \pi[v] <- u 
	  (setf (gethash v p) u)
		;; Modification: Add v as a child of u
	  (push v (gethash u c)) 
;;; 7            DFS-VISIT(v)	   
	  (setf time (dfs-visit v color p c time d f g))))
;;; 8 color[u] <- BLACK ;;Blacken u; it is finished. 
  (setf (gethash u color) 'black)
;;; 9 f[u] <- time <- time + 1
  (setf (gethash u f) (incf time 1))
  time
  )