(in-package :mtk-graph-algorithms)

(defmethod initialize-single-source ((g mtk-graph:labeled-weighted-directed-graph)
									 source-vertex
									 &key
									 (inf *inf*)
									 (predecessor-graph? nil)
									 &allow-other-keys)
"Call next method, provide a wrapper that puts the value in a
labeled pair with a universal clause from the graph's state-space. Can output
a predecessor tree or graph."

  (call-next-method 
   g source-vertex :inf inf
   :value-wrapper #'(lambda (x) 
	   (let* ((lvs (make-instance (if predecessor-graph?
									  'mtk-graph:weak-labeled-value-set
									  'mtk-graph:strict-labeled-value-set)
								  :strict-order #'(lambda (x y) (< (first x) (first y)))
								  :value-equalp (if predecessor-graph?
													#'(lambda (x y) (equalp x y))
													#'(lambda (x y) (eql (first x)
																		 (first y))))
								  :state-space (get-state-space g)))
			  (initial-pair  ;create the initial pair
			   (make-instance 
				'labeled-pair
				;;with a universal clause
				:environment (get-universal-clause (get-state-space g))
				;;and the input value
				:value x)))
		 (add-pair-to-lvs! lvs initial-pair)
		 lvs))))

(defmethod initialize-single-source ((g mtk-graph:directed-graph)
									 source-vertex
									 &key
									 (inf *inf*)
									 (value-wrapper
									  #'(lambda (x) x))
									 &allow-other-keys
)
"Implements Initialize-Single-Source of Cormen, Lieserson, Rivest, and Stein.
Allows the user to specify zeros, infinities,k and predecessor-wrappers to
allow this code to work for labeled distance graphs. The inf should be a function
that returns unique copies. Multi-value return the hash-tables that map from
vertices to initialized distances and predecessor lists."

;;Allocate space, the hash table maps v -> (d[v] pi[v])
(let ((d-pi-ht (make-hash-table)))

  ;;1 for each vertex g \in V[G]
  (dolist (v (mtk-graph:get-vertices g))
	;;2 do d[v] <- \inf
	;;3 pi[v] <- nil
	(setf (gethash v d-pi-ht) (funcall value-wrapper (list inf nil))))
  
  ;;4. d[s] <- 0
  (setf (gethash source-vertex d-pi-ht) (funcall value-wrapper (list 0 nil)))
  
  d-pi-ht))

(defmethod relax! ((g weighted-directed-graph) u v w-uv d-pi-ht)
"Standard graph relaxation from Cormen, Lieserson, Rivest, and Stein. 
Modifies the hash-table passed in. Doesn't actually know how to compute
a full predecessor-graph."


  ;;fetch values and compute d[u]+w(u,v)
  (let* ((d-pi-u (gethash u d-pi-ht)) ;d[u]
		 (d-pi-v (gethash v d-pi-ht)) ;d[v]
		 (candidate-weight (+ (first d-pi-u) w-uv))) ;d[u] + w(u,v)

	;;1. if d[v] > d[u] + w(u,v)
	(when (> (first d-pi-v) candidate-weight)
	  ;;2. then d[v] <- d[u] + w(u,v)
	  ;;3. pi[v] <- u
	  (setf (gethash v d-pi-ht) (list candidate-weight u)))))

(defmethod relax! ((g labeled-weighted-directed-graph) u v w-uv d-pi-ht)
"Relaxation specialized to labeled distance graphs. 
Modifies the hash-table passed in."

 (let* ((d-pi-u (gethash u d-pi-ht))			;d[u]
		(d-pi-v (gethash v d-pi-ht))			;d[v]
		(candidate-weights 			;d[u] + w(u,v)
		 (lvs-binary-function d-pi-u w-uv ;figure out new weights
							  ;;by applying a function that adds the weight
							  ;;and puts u as the new predecessor
							  #'(lambda (x y) 
								  (list (+ (first x) y) u)))))

;   		 (debug-print-for :mtk-graph-algorithms t
;						  "relaxing u: ~a v: ~a 
;lvs u: ~a~% through: ~a~% candidates: ~a~% against: ~a~%"
;						  u v d-pi-u w-uv candidate-weights d-pi-v)

   ;;now update as necessary, updates both d and pi
   (merge-lvs-in-place! d-pi-v candidate-weights)))


(defmethod bellman-ford ((g mtk-graph:weighted-directed-graph)
						 source-vertex
						 &key
						 (inf *inf*)
						 (predecessor-graph? nil))
  "Compute Bellman-Ford on a weighted directed graph."
  ;;Call through to the next method, providing the proper
  ;;weight function.
  (call-next-method g source-vertex
					:weight-function #'(lambda (e_i)
										 (mtk-graph:get-weight e_i g))
					:inf inf
					:predecessor-graph? predecessor-graph?))


(defmethod bellman-ford ((g mtk-graph:directed-graph)
						 source-vertex
						 &key
						 (weight-function nil)
						 (inf *inf*)
						 (predecessor-graph? nil)
						 (test-consistency? t))
  "Compute Bellman-Ford on a weighted directed graph."
  (debug-print-for :mtk-graph-algorithms t "Computing SSSP for ~a" source-vertex)
  
  ;;1. initialize-single-source(G,s)
  (let ((d-pi-ht (initialize-single-source g source-vertex 
								   :inf inf
								   :predecessor-graph? predecessor-graph?)))

	;;2. for i <- 1 to |V(G)|-1
	(loop for i from 1 to (- (length (get-vertices g)) 1)
	   ;;3. do for each edge (u,v) \in E[G]
	   do (dolist (edge-uv (get-edge-list (get-edges g)))
				   ;;pull out u,v
				   
			(let ((u (first (get-vertices edge-uv)))
				  (v (second (get-vertices edge-uv))))
			  ;;4. do Relax(u,v,w)
			  (relax! g u v (funcall weight-function edge-uv) d-pi-ht))))


	   ;;5. for each edge (u,v) \in E[G]

	(let ((consistent? (or (not test-consistency?)
						   (test-bellman-ford-consistency g d-pi-ht weight-function))))
	  (if consistent?
		  d-pi-ht
		  nil))))

(defmethod test-bellman-ford-consistency 
	((g mtk-graph:weighted-directed-graph)
	 d-pi-ht
	 weight-function)

  ;;Loop over the edges and check that the relaxation worked
  (loop for edge-uv in (get-edge-list (get-edges g))
	   never 
	   (let* ((u (first (get-vertices edge-uv)))
			 (v (second (get-vertices edge-uv)))
			  (du (first (gethash u d-pi-ht)))
			  (dv (first (gethash v d-pi-ht)))
			  (w-uv (funcall weight-function edge-uv)))
		 (> dv (+ du w-uv)))))

(defmethod test-bellman-ford-consistency 
	((g mtk-graph:labeled-weighted-directed-graph)
	 d-pi-ht
	 weight-function)
"The labeled case."

  ;;Loop over the edges and check that the relaxation worked
  (loop for edge-uv in (get-edge-list (get-edges g))
	   never 
	   (let* ((u (first (get-vertices edge-uv)))
			  (v (second (get-vertices edge-uv)))
			  (w-uv (funcall weight-function edge-uv))
			  (d-pi-u (gethash u d-pi-ht)) ;d[u]
			  (d-pi-v (gethash v d-pi-ht)) ;d[v]
			  (u-plus-w		   ;d[u] + w(u,v)
			   (lvs-binary-function d-pi-u w-uv ;figure out new weights
									;;by applying a function that adds the weight
									;;and puts u as the new predecessor
									#'(lambda (x y) 
										(list (+ (first x) y) u)))))

;		 (debug-print-for :mtk-graph-algorithms t
;						  "trying to relax ~a->~a ~a ~% through ~a, ~% giving ~a ~% against ~a"
;						  u v d-pi-u w-uv u-plus-w d-pi-v)

		 ;;Create a conflict for the environment of all the values 
		 ;;of d[v] that have a query result in the sum d[u]+w(u,v)
		 ;;where d[v] > d[u] + w(u,v). We don't care about values of d[v] that
		 ;;don't actually have a query result because they're unreachable
		 ;;and just haven't been pruned out. 

		 ;;Recall that we're really looking for complete environments where 
		 ;;d[v] > d[u]+w(u,v), the RHS pairs provide the value the LHS needs
		 ;;to beat for particular environments.
		 (loop for dv-pair in (lvs-pairs d-pi-v)
			  do (loop for uw-pair in (lvs-pairs u-plus-w)
					  do (when (and (> (first (pair-value dv-pair))
									   (first (pair-value uw-pair)))
									(clause-valid? (clause-intersection 
														  (pair-environment dv-pair)
														  (pair-environment uw-pair))))
						   (debug-print-for :mtk-graph-algorithms t 
											"Possible conflict ~a ~a" dv-pair uw-pair)
						   (let ((conflicts (split-clause-around-clauses 
											 (clause-intersection 
											  (pair-environment dv-pair)
											  (pair-environment uw-pair))
											 (loop for dv-pair2 in (lvs-pairs d-pi-v)
												when (<= (first (pair-value dv-pair2))
														 (first (pair-value uw-pair)))
												collect (pair-environment dv-pair2)))))
							 (debug-print-for :mtk-graph-algorithms t "Found conflicts: ~a" conflicts)
							 ;;and make the conflicts conflicts
							 (loop for conflict in conflicts
								do (add-conflict! (get-state-space g) 
												  conflict)))


						   )))))


  (loop for du being the hash-values of d-pi-ht
	   do (remove-invalid-pairs! du))

  ;;Return t if the above have not invalidated the state-space
  (when (state-space-invalid? (get-state-space g))
	(error 'infeasible-problem :text "SSSP removes all kernels."))
  t)

(define-condition infeasible-problem (error)
  ((text :initarg :text :reader :text)))

;; (defmethod labeled-dijkstra ((ldg labeled-weighted-directed-graph)
;; 							 source-vertex)
;;   "This method does not currently work!"
  
;;   (assert nil)

;;     (let ((d-pi-ht (initialize-single-source ldg source-vertex 
;; 								   :inf *inf*
;; 								   :predecessor-graph? t))
;; 		  (distance-heap (make-instance 'cl-heap:fibonacci-heap 
;; 										:key #'first)))

;; 	  ;;The heap contains entries (distance env node) where the distance
;; 	  ;;needs to be propagated from node with env
;; 	  (cl-heap:add-to-heap distance-heap (list  0 
;; 												(get-universal-clause
;; 												 (get-state-space ldg))
;; 												source-vertex))

;; 	  (loop until (cl-heap:is-empty-heap-p distance-heap)
;; 		   with val-to-propagate
;; 		   do (setf val-to-propagate (cl-heap:pop-heap distance-heap))
;; 		   do (loop for edge in (find-edges-with-source-vertex 
;; 								 (third val-to-propagate)
;; 								 ldg)
;; 				 do (debug-print-for :mtk-graph-algorithms t "prop ~a through ~a ~a"
;; 									 val-to-propagate edge (get-weight edge ldg))
;; 				 do (let* ((prop-lvs 
;; 							(make-instance 
;; 							 'mtk-graph:strict-labeled-value-set
;; 							 :strict-order #'(lambda (x y) (< (first x) (first y)))
;; 							 :value-equalp  #'(lambda (x y) (equalp x y))
;; 							 :state-space (get-state-space ldg)))
;; 						   (prop-val-pair (make-instance 
;; 										   'labeled-pair 
;; 										   :environment (second val-to-propagate)
;; 										   :value  (first val-to-propagate))))
;; 					  (add-pair-to-lvs! prop-lvs prop-val-pair)

;; 					  ;;figure out how this one propagates
;; 					  (let ((candidate-pairs (lvs-binary-function-candidates 
;; 											  prop-lvs (get-weight edge ldg)
;; 											  #'(lambda (x y)
;; 												  (list
;; 												   (+  x  y)
;; 												   (third val-to-propagate))))))
;; 						(debug-print-for :mtk-graph-algorithms t 
;; 										 "found candidates: ~a" candidate-pairs)
						
						  
;; 						;;Add the candidates to their target and if
;; 						;;they're kept, they'll need to be propagated
;; 						;;again.
;; 						(loop for candidate-pair in candidate-pairs
;; 						   unless (null (pair-environment candidate-pair))
						   
;; 							 when (add-pair-to-lvs! (gethash (get-target-vertex edge)
;; 														d-pi-ht)
;; 											   candidate-pair)
;; 						   do (cl-heap:add-to-heap 
;; 							   distance-heap 
;; 							   (list (first (pair-value candidate-pair))
;; 									 (pair-environment candidate-pair)
;; 									 (get-target-vertex edge))))))))
;; 	 d-pi-ht ))