;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package mtk-graph-algorithms)

;;; For references relating to this code see:
;;;   [1] T. H. Cormen, C. E. Leiserson, R. L. Rivest, and C. Stein. 
;;;       Introduction to Algorithms. The MIT Press,
;;;       Cambridge, Massachusetts, 2nd edition, 2001.

;;; From [1] pg. 554
;;;
;;; STRONGLY-CONNECTED-COMPONENTS(G) 
;;; 1 call DFS(G) to compute finishing times f[u] for each vertex u 
;;; 2 compute G^T 
;;; 3 call DFS(G^T), but in the main loop of DFS, consider the vertices
;;;   in order of decreasing f[u] (as computed in line 1) 
;;; 4 output the vertices of each tree in the depth-first forest formed in
;;;   line 3 as a separate strongly connected component 

;;; STRONGLY-CONNECTED-COMPONENTS(G) 
(defmethod strongly-connected-components((G mtk-graph:directed-graph))
  (let ((p nil)
	(d nil)
	(f nil)
	(c nil)
	(G^T (mtk-graph:transpose G)))
	;; this allows us to ignore compiler style-warnings, concerning 
	;; the assignment to d, without reading from it.
	;; Note: ignore and ignorable have two different meanings. 'ignore'
	;; means the variable should not be read from or written to.
    (declare (ignorable d))
;;; 1 call DFS(G) to compute finishing times f[u] for each vertex u 
    (multiple-value-setq (p c d f) (dfs G))
;;; 2 compute G^T 
    ;; Done in let.
;;; 3 call DFS(G^T), but in the main loop of DFS, consider the vertices
;;;   in order of decreasing f[u] (as computed in line 1) 
    (multiple-value-setq (p c)
      (dfs G^T
	   :vertices (sort (copy-seq (mtk-graph:get-vertices G^T))
			   #'(lambda (x y) (> (gethash x f) (gethash y f))))))
;;; 4 output the vertices of each tree in the depth-first forest formed in
;;;   line 3 as a separate strongly connected component 
    (loop for v_i in (mtk-graph:get-vertices g^t)
       when (null (gethash v_i p))
       collect (cons v_i (labels ((descendants (u_i) 
				    (let ((children (gethash u_i c)))
				      (append children
					      (apply #'append
					       (mapcar #'(lambda (w_i)
							   (descendants w_i))
						       children))))))
			   (descendants v_i))))))

(defmethod generate-component-graph ((G mtk-graph:directed-graph))
  (let* ((V^SCC (strongly-connected-components G))
	 (G^SCC (make-instance 'mtk-graph:directed-graph))
	 (V^SCC-ht-v (make-hash-table))
	 (e nil))
    (dolist (V^SCC_i V^SCC)
      (dolist (v_i V^SCC_i)
	(setf (gethash v_i V^SCC-ht-v) V^SCC_i)))
    (mtk-graph:add-vertices V^SCC G^SCC)
    (dolist (e_i (mtk-graph:get-edge-list (mtk-graph:get-edges G)))
      (let ((v^scc_i (gethash (mtk-graph:get-source-vertex e_i)
			      V^SCC-ht-v))
	    (v^scc_j (gethash (mtk-graph:get-target-vertex e_i)
			      V^SCC-ht-v)))
	(if (not (eq v^scc_i v^scc_j))
	    (push (list v^scc_i v^scc_j) e))))
    (mtk-graph:add-edges (remove-duplicates e :test #'equal) G^SCC)
    G^SCC))
   