;; This algorithm enumerates all of the possible topological sorts of 
;; a directed acyclic graph. It's based on the paper "Generating Linear
;; Extensions Fast" by Gara Pruesse and Frank Ruskey, 1994. It is asymptotically
;; optimal - namely, each linear extension (topological sorting) is computed
;; in constant amortized time.
;;
;; TOPOLOGICAL SORTS! WOOOOO!!!!! :-D
;;
;; Author: Steve Levine (sjlevine@mit.edu)
;;

(in-package :mtk-graph-algorithms)

;; The topdata class stores useful information and parameters needed
;; for the topological sort enumeration algorithm.
(defclass topdata ()
  ((p :initform nil
	  :initarg :p
	  :accessor p
	  :documentation "Stores the directed acyclic graph for which the topological sorts are based.")
   (le :type  array
	   :initform (make-array 0 :fill-pointer t :adjustable t)
	   :initarg :le
	   :accessor le
	   :documentation "Stores the current linear extension.")
   (li :type hash-table
	   :initform (make-hash-table :test #'equal)
	   :accessor li
	   :documentation "A hash table mapping elements to their indices in le.")
   (a :type array 
	  :initform (make-array 0 :fill-pointer t :adjustable t)
	  :initarg :a
	  :accessor a
	  :documentation "Stores the a array of elements from the algorithm.")
   (b :type array 
	  :initform (make-array 0 :fill-pointer t :adjustable t)
	  :initarg :b
	  :accessor b
	  :documentation "Stores the b array of elements from the algorithm.")
   (max-pair :type num
			 :initform 0
			 :accessor max-pair
			 :documentation "The index of the max pair.")
   (plus
	:initform t
	:initarg :plus
	:accessor plus
	:documentation "Whether this is a +/- node in the graph to be traversed.")
   (N :type num
	  :initform 0
	  :accessor N
	  :documentation "The number of elements in le (constant).")
   (extensions :type list
			   :initform nil
			   :accessor extensions
			   :documentation "A list of computed extensions!")
   (relatable-function
	:initform nil
	:initarg :relatable-function
	:accessor relatable-function
	:documentation "A function of two arguments that returns t if the two vertices are comparable in the poset, and nil otherwise.")))



;; This is the main method to call. Once called with a directed acyclic graph as input, it
;; preprocesses the graph as specified in the paper. It then calls the recurisve gen-le function
;; to generate all of the linear extensions of the graph, and returns them all in a list.
(defmethod topological-sorts-all ((g directed-graph))
  (let* ((topdata (preprocess-graph g))
		 (m (max-pair topdata)))
	(gen-le m topdata)
	(top-sort-switch m topdata)
	(gen-le m topdata)
	(extensions topdata)))


;; A method that ensures that, for each forward directed edge, there is also a backwards
;; directed edge. For each directed edge in the graph, a reversed edge
;; is added if one does not already exist. Note that this makes the graph cyclic!
;; This is useful first step for determining the comparability of elements in the graph, if we 
;; interpret it at as partially ordered set (poset). By adding back edges and then computing a
;; transitive closure, we guarantee that the transitive closure will return T iif the two vertices
;; are comparable, and nil otherwise.
(defmethod bidirectional-graph ((g directed-graph))
  (let ((g^new (make-instance (class-of g)))) 
	(add-vertices (get-vertices g) g^new)
	(dolist (e_i (get-edge-list (get-edges g)))
	  (add-edge e_i g^new)
	  (when (not (find-edges-with-source-and-target-vertices (get-target-vertex e_i) (get-source-vertex e_i) g )) 
		(add-edge (list (get-target-vertex e_i) (get-source-vertex e_i)) g^new)))
	g^new))


;; Return a list of minimal vertices of the directed graph, or namely a list of vertices such that
;; no edge in the graph has that vertex as it's target. Put another way, the in-degree of these minimal
;; vertices are 0.
(defmethod get-minimal-vertices ((g directed-graph))
  (loop for v in (get-vertices g) 
	 if (not (find-edges-with-target-vertex v g)) collect v))


;; This is the preprocessing stage of the algorithm.
(defmethod preprocess-graph ((g directed-graph))
  (let ((topdata (make-instance 'topdata))
		(Q (copy-graph g)))
	(setf (p topdata) g)
	;; Remove pairs of minimal vertices to construct an initial linear extension
	(loop while (get-vertices Q) do
		 (let* ((minimal-vertices (get-minimal-vertices Q)))
		   (if (eql (length minimal-vertices) 1)
			   (progn
				 (let* ((x (first minimal-vertices))) 
				   (vector-push-extend x (le topdata))
				   (delete-vertex x Q)))
			   (progn
				 (let* ((ai (first minimal-vertices))
						(bi (second minimal-vertices)))
				   (vector-push-extend ai (a topdata))
				   (vector-push-extend bi (b topdata))
				   (vector-push-extend ai (le topdata))
				   (vector-push-extend bi (le topdata))
				   (delete-vertex ai Q)
				   (delete-vertex bi Q))))))
	;; Construct li, a hash table that maps objects to indices in le
	(loop for i from 0 to (1- (length (le topdata))) do
		 (setf (gethash (aref (le topdata) i) (li topdata)) i))

	;; Construct the relatable function - which returns t if the two objects in the
	;; directed graph are comparable, and nil otherwise (recall that not all elements
	;; in a partially ordered set are comparable, by definition! This returns what can
	;; be compared).
	(setf (relatable-function topdata) (transitive-closure (bidirectional-graph g)))
	;; Set some other useful fields in topdata
	(setf (N topdata) (length (le topdata)))
	(setf (max-pair topdata) (length (a topdata)))
	(setf (plus topdata) t)
	(top-sort-process topdata)
	topdata))

;; Implementation fo the main GenLE algorithm as defined the paper (see it
;; for details). Note that in the original PASCAL implementation, indices started
;; from 1 - in LISP they start from 0. This is what the get-a and get-b functions do
;; - they esentially abstract away the indices so that the algorithm below can operate
;; as if it had one-based indices.
(defmethod gen-le(i (topdata topdata))
  (when (> i 0)
	(gen-le (1- i) topdata)
	(let* ((mrb 0)
		   (mra 0)
		   (mla 0)
		   (cont t)
		   (typical nil))
	  (loop while (top-sort-right i topdata :ab 'b) do
		   (incf mrb)
		   (top-sort-move (get-b i topdata) topdata :direction 'right)
		   (gen-le (1- i) topdata)
		   (setf mra 0)
		   (when (top-sort-right i topdata :ab 'a)
			 (setf typical t)
			 (setf cont t)
			 (loop while cont do
				  (incf mra)
				  (top-sort-move (get-a i topdata) topdata :direction 'right)
				  (gen-le (1- i) topdata)
				  (setf cont (top-sort-right i topdata :ab 'a))
				  ))
		   (when typical
			 (top-sort-switch (1- i) topdata)
			 (gen-le (1- i) topdata)
			 (if (oddp mrb)
				 (setf mla (1- mra))
				 (setf mla (1+ mra)))
			 (loop for x from 1 to mla do
				  (top-sort-move (get-a i topdata) topdata :direction 'left)
				  (gen-le (1- i) topdata))))

	  (if (and typical (oddp mrb))
		  (top-sort-move (get-a i topdata) topdata :direction 'left)
		  (top-sort-switch (1- i) topdata))

	  (gen-le (1- i) topdata)

	  (loop for x from 1 to mrb do
		   (top-sort-move (get-b i topdata) topdata :direction 'left)
		   (gen-le (1- i) topdata))
	 
	  )))

(defmethod top-sort-switch (i (topdata topdata))
  (if (eql i 0)
	  (setf (plus topdata) (not (plus topdata)))
	  (progn 
		(let* ((index-a (gethash (get-a i topdata) (li topdata)))
			   (index-b (gethash (get-b i topdata) (li topdata))))
		  (transpose-elements index-a index-b topdata)
		  ;; Also switch the a and b
		  (let* ((temp (aref (a topdata) (1- i))))
			(setf (aref (a topdata) (1- i)) (aref (b topdata) (1- i)))
			(setf (aref (b topdata) (1- i)) temp)))))
  (top-sort-process topdata))

(defmethod top-sort-right (i (topdata topdata) &key (ab 'a))
  (if (eql ab 'b)
	  (progn
		(let* ((item (get-b i topdata))
			   (index-le (gethash item (li topdata))))
		  (if (< index-le (1- (N topdata)))
			  (let* ((item-right (aref (le topdata) (1+ index-le))))
				(not (funcall (relatable-function topdata) item item-right)))
			  nil)))
	  
	  (progn ;; ab == 'a
		(let* ((item (get-a i topdata))
			   (index-le (gethash item (li topdata))))
		  (if (< index-le (1- (N topdata)))
			  (let* ((item-right (aref (le topdata) (1+ index-le))))
				(and (not (funcall (relatable-function topdata) item item-right)) (not (equal item-right (get-b i topdata)))))
			  nil)))))

(defmethod get-a (i (topdata topdata))
  (aref (a topdata) (1- i)))

(defmethod get-b (i (topdata topdata))
  (aref (b topdata) (1- i)))

;; A hook method that takes the current lienar extension in le
;; and adds it to the list of all topological sorts if appropriate
(defmethod top-sort-process ((topdata topdata))
  (when (plus topdata)
	(push (copy-seq (le topdata)) (extensions topdata))))

(defmethod top-sort-move (x (topdata topdata) &key (direction 'left))
  (let* ((index-offset (if (eql direction 'left) -1 1))
		 (index-le (gethash x (li topdata)))
		 (index-switch (+ index-le index-offset)))
	(transpose-elements index-le index-switch topdata)
	(top-sort-process topdata)
	))

(defmethod transpose-elements (index-1 index-2 (topdata topdata))
  (let* ((item-1 (aref (le topdata) index-1))
		 (item-2 (aref (le topdata) index-2)))
	;; Swap them
	(setf (aref (le topdata) index-1) item-2)
	(setf (aref (le topdata) index-2) item-1)
	;; Also swap the associated indices
	(setf (gethash item-1 (li topdata)) index-2)
	(setf (gethash item-2 (li topdata)) index-1)))

