;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung
;;;;   David C. Wang

(in-package :mtk-graph)


;;-------------------------------------------------------------------------------
;; Edge CLASS:
;;   * Directed Hyper Edge
;;-------------------------------------------------------------------------------

(defclass directed-hyper-edge (hyper-edge)
  ((source-vertices  :type list
                     :accessor get-source-vertices
                     :initarg :source-vertices
                     :initform (error "Must specify a set of source vertices using :source-vertices key.")
					 :documentation "Add documentation here.")
   (target-vertices :type list
                    :accessor get-target-vertices
                    :initarg :target-vertices
                    :initform (error "Must specify a set of target vertices using :vertices key.")
                    :documentation "Add documentation here.")))

(defmethod print-object ((e directed-hyper-edge) stream)
  (with-slots ((svs source-vertices)
			   (tvs target-vertices)) e
	(format stream "~a->~a" svs tvs)))

;; CONSTRUCTOR
(defmethod make-directed-hyper-edge (&key sources targets)
  (assert sources)
  (assert targets)
  (assert (null (intersection sources targets :test #'eql)))
  (let ((vertices (append sources targets)))
    (make-instance 'directed-hyper-edge
      :vertices vertices
      :source-vertices sources
      :target-vertices targets)))


;;-------------------------------------------------------------------------------
;; Edges CLASS:
;;   * Directed Hyper Edges
;;-------------------------------------------------------------------------------

(defclass directed-hyper-edges (hyper-edges)
  ((edges-ht-source-vertex :type hash-table
                           :accessor get-edges-ht-source-vertex
                           :initform (make-hash-table)
                           :documentation "Add documentation here.")
   (edges-ht-target-vertex :type hash-table
                           :accessor get-edges-ht-target-vertex
                           :initform (make-hash-table)
                           :documentation "Add documentation here.")))

(defmethod print-object ((es directed-hyper-edges) stream)
  (with-slots ((el edge-list) 
			   (sev edges-ht-source-vertex)
			   (tev edges-ht-target-vertex)) es
	(format stream "Edges:~%")
	(dolist (edge el) (format stream "~a~%" edge))
	(format stream "by Sources~%")
	(loop 
	   for item being the hash-key of sev  
	   do (format stream "~a: ~a;~%" item (gethash item sev)))
	(format stream "by Targets~%")
	(loop 
	   for item being the hash-key of tev  
	   do (format stream "~a: ~a;~%" item (gethash item tev)))))

;;----------------------------------------
;; Find Vertex/Vertices

(defmethod find-parents-of-vertex (vertex (e hyper-edges))
  (remove vertex
	  (remove-duplicates
	   (apply #'append (mapcar #'get-source-vertices
			   (find-edges-with-target-vertex vertex e))))))

(defmethod find-children-of-vertex (vertex (e hyper-edge))
  (find-neighbors-of-vertex vertex e))

(defmethod find-children-of-vertex (vertex (e directed-hyper-edges))
  (remove vertex
	  (remove-duplicates
	   (apply #'append (mapcar #'get-target-vertices
			   (find-edges-with-source-vertex vertex e))  ))))

;;----------------------------------------
;; Add Edges

(defmethod add-edge ((e_i directed-hyper-edge) (e directed-hyper-edges))
  (call-next-method e_i e)
  (dolist (v_j (get-source-vertices e_i))
    (push e_i (gethash v_j (get-edges-ht-source-vertex e))))
  (dolist (v_j (get-target-vertices e_i))
    (push e_i (gethash v_j (get-edges-ht-target-vertex e))))
  e_i)

(defmethod add-edge ((edge-description list) (e directed-hyper-edges))
  "Adds an edge to a set of directed hyper edges, where the edge-description
is a sequence of source vertices list and target vertices list the edge
connects."
  (let* ((source-vertices (first edge-description))
	 (target-vertices (second edge-description))
	 (vertices (append source-vertices target-vertices))
	 (e_i (make-instance 'directed-hyper-edge
			     :vertices vertices
			     :source-vertices source-vertices
			     :target-vertices target-vertices)))
    (add-edge e_i e)))

;;----------------------------------------
;; Delete Edges

(defmethod delete-edge ((e_i directed-hyper-edge) (e directed-hyper-edges))
  (call-next-method e_i e)
  (dolist (v_j (get-source-vertices e_i))
    (setf (gethash v_j (get-edges-ht-source-vertex e))
	  (delete e_i (gethash v_j (get-edges-ht-source-vertex e)))))
  (dolist (v_j (get-target-vertices e_i))
    (setf (gethash v_j (get-edges-ht-target-vertex e))
	  (delete e_i (gethash v_j (get-edges-ht-target-vertex e))))))

;;----------------------------------------
;; Find Edges

(defmethod find-edges-with-source-vertex (vertex (e directed-hyper-edges))
  (gethash vertex (get-edges-ht-source-vertex e)))

(defmethod find-edges-with-target-vertex (vertex (e directed-hyper-edges))
  (gethash vertex (get-edges-ht-target-vertex e)))

(defmethod find-edges-with-source-and-target-vertices (source-vertex
						       target-vertex
						       (e directed-hyper-edges))
  (intersection (find-edges-with-source-vertex source-vertex e)
		(find-edges-with-target-vertex target-vertex e)))


;;-------------------------------------------------------------------------------
;; Graph CLASS:
;;   * Directed Hyper Graph
;;-------------------------------------------------------------------------------

(defclass directed-hyper-graph (hyper-graph)
  ((e :type directed-hyper-edges
      :initform (make-instance 'directed-hyper-edges)
      :documentation "Add documentation here.")))

(defmethod transpose ((g directed-hyper-graph))
  (let ((g^t (make-instance (class-of g))))
    (add-vertices (get-vertices g) g^t)
    (dolist (e_i (get-edge-list (get-edges g)))
      (add-edge (list (get-target-vertices e_i) (get-source-vertices e_i)) g^t))
    g^t))

;;----------------------------------------
;; Find Vertex/Vertices

(defmethod find-parents-of-vertex (vertex (g hyper-graph))
  (find-parents-of-vertex vertex (get-edges g)))

(defmethod find-children-of-vertex (vertex (g hyper-graph))
  (find-neighbors-of-vertex vertex g))

(defmethod find-children-of-vertex (vertex (g directed-hyper-graph))
  (find-children-of-vertex vertex (get-edges g)))

;;----------------------------------------
;; Find Edges

(defmethod find-edges-with-source-vertex (vertex (g directed-hyper-graph))
  (find-edges-with-source-vertex vertex (get-edges g)))

(defmethod find-edges-with-target-vertex (vertex (g directed-hyper-graph))
  (find-edges-with-target-vertex vertex (get-edges g)))

(defmethod find-edges-with-source-and-target-vertices (source-vertex
						       target-vertex
						       (g directed-hyper-graph))
  (find-edges-with-source-and-target-vertices source-vertex
					      target-vertex
					      (get-edges g)))