;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung
;;;;   David C. Wang
;;;;   Yi (Andrew) Huang

(in-package :mtk-graph)


;;-------------------------------------------------------------------------------
;; Edge CLASS:
;;   * Hyper Edge (undirected)
;;-------------------------------------------------------------------------------

(defclass hyper-edge ()
  ((vertices :type list
             :accessor get-vertices
             :initarg :vertices 
             :initform (error "Must specify a set of edges using :vertices key.")
             :documentation "Add documentation here.")))

(defmethod print-object ((e hyper-edge) stream)
  (with-slots ((vs vertices)) e
	(format stream "~a" vs)))

;; CONSTRUCTOR
(defmethod make-hyper-edge (&key vertices)
  (assert vertices)
  (make-instance 'hyper-edge :vertices vertices))


;;-------------------------------------------------------------------------------
;; Edges CLASS:
;;   * Hyper Edges (undirected)
;;-------------------------------------------------------------------------------

(defclass hyper-edges ()
  ((edge-list :type list
              :accessor get-edge-list
              :initform nil
              :documentation "")
   (edges-ht-vertex :type hash-table
                    :accessor get-edges-ht-vertex
                    :initform (make-hash-table :test #'equal)
                    :documentation "")))

(defmethod print-object ((es hyper-edges) stream)
  (with-slots ((el edge-list) 
			   (ev edges-ht-vertex)) es
	(format stream "Edges:~%")
	(dolist (edge el) (format stream "~a~%" edge))
	(format stream "by Vertices:~%")
	(loop 
	   for item being the hash-key of ev  
	   do (format stream "~a: ~a;~%" item (gethash item ev)))))

;;----------------------------------------
;; Find Vertex/Vertices

(defmethod find-neighbors-of-vertex (vertex (e hyper-edges))  
  (remove vertex
		  ;; Collect all vertices adjacent to vertex, including itself.
		  (remove-duplicates
		   (apply #'append (mapcar #'get-vertices
								   (find-edges-with-vertex vertex e))))
		  :test #'equal))

;;----------------------------------------
;; Add Edges

(defmethod add-edge ((e_i hyper-edge) (e hyper-edges))
  (push e_i (get-edge-list e))
  (dolist (v_j (get-vertices e_i))
    (push e_i (gethash v_j (get-edges-ht-vertex e))))
  e_i)

(defmethod add-edge ((edge-description list) (e hyper-edges))
  "The edge-description is a list of vertices that the edge connects."
  (let ((e_i (make-instance 'hyper-edge :vertices edge-description)))
    (add-edge e_i e)))

;;----------------------------------------
;; Delete Edges

(defmethod delete-edge ((e_i hyper-edge) (e hyper-edges))
  (setf (get-edge-list e) (delete e_i (get-edge-list e)))
  (dolist (v_j (get-vertices e_i))
    (setf (gethash v_j (get-edges-ht-vertex e))
	  (delete e_i (gethash v_j (get-edges-ht-vertex e))))))

;;----------------------------------------
;; Find Edges

(defmethod find-edges-with-vertex (vertex (e hyper-edges))
  (gethash vertex (get-edges-ht-vertex e)))

(defmethod find-edges-with-vertices ((vertices list) (e hyper-edges))
  (reduce #'intersection
	  (mapcar #'(lambda (v_i) (find-edges-with-vertex v_i e)) vertices)))


;;-------------------------------------------------------------------------------
;; Graph CLASS:
;;   * Hyper Graph (undirected)
;;-------------------------------------------------------------------------------

(defclass hyper-graph ()
  ((v :type list
      :accessor get-vertices
      :initform nil
      :documentation "")
   (e :type hyper-edges
      :accessor get-edges
      :initform (make-instance 'hyper-edges)
      :documentation "")))

(defmethod print-object ((g hyper-graph) stream)
  (with-slots ((v v) 
			   (e e)) g
	(format stream "Vertices: ~a~%" v)
	(format stream "~a~%" e)))

(defmethod get-edge-list ((g hyper-graph))
  (get-edge-list (get-edges g)))

;; COPY CONSTRUCTOR
(defmethod copy-graph ((g hyper-graph))
  (let ((g-copy (make-instance (class-of g))))
    (add-vertices (get-vertices g) g-copy)
    (add-edges (get-edges g) g-copy)
    g-copy))

;;----------------------------------------
;; Add Vertex/Vertices

(defmethod add-vertex (vertex (g hyper-graph))
  (pushnew vertex (get-vertices g) :test #'equal))

(defmethod add-vertices ((vertices list) (g hyper-graph))
  (dolist (v_i vertices)
    (add-vertex v_i g)))

;;----------------------------------------
;; Delete Vertex/Vertices

(defmethod delete-vertex (vertex (g hyper-graph))
  (delete-edges (find-edges-with-vertex vertex g) g)
  (setf (get-vertices g) (delete vertex (get-vertices g)))
  (remhash vertex (get-edges-ht-vertex (get-edges g))))

(defmethod delete-vertices ((vertices list) (g hyper-graph))
  (dolist (v_i vertices)
    (delete-vertex v_i g)))

;;----------------------------------------
;; Find Vertex/Vertices

(defmethod find-neighbors-of-vertex (vertex (g hyper-graph))
  (find-neighbors-of-vertex vertex (get-edges g)))

;;----------------------------------------
;; Add Edges

(defmethod add-edges ((e list) (g hyper-graph))
  (dolist (e_i e)
	(add-edge e_i g)))

(defmethod add-edges ((e hyper-edges) (g hyper-graph))
  (dolist (e_i (get-edge-list e)) (add-edge e_i g)))

(defmethod add-edge (vertices (g hyper-graph))
  (add-vertices vertices g)
  (add-edge vertices (get-edges g)))

(defmethod add-edge ((e_i hyper-edge) (g hyper-graph))
  (add-vertices (get-vertices e_i) g)
  (add-edge e_i (get-edges g)))

;;----------------------------------------
;; Delete Edges

(defmethod delete-edges ((e list) (g hyper-graph))
  (dolist (e_i e) (delete-edge e_i g)))

(defmethod delete-edges ((e hyper-edges) (g hyper-graph))
  (dolist (e_i (get-edge-list e)) (delete-edge e_i g)))

(defmethod delete-edge ((e_i hyper-edge) (g hyper-graph))
  (delete-edge e_i (get-edges g)))

;;----------------------------------------
;; Find Edges

(defmethod find-edges-with-vertex (vertex (g hyper-graph))
  (find-edges-with-vertex vertex (get-edges g)))

(defmethod find-edges-with-vertices ((vertices list) (g hyper-graph))
  (find-edges-with-vertices vertices (get-edges g)))
