(in-package :mtk-graph)

(defclass labeled-pair ()
  ((value
	 :accessor pair-value
	 :initarg :value
	 :documentation "The value of the pair.")

   (environment
	:accessor pair-environment
	:type state-space:state-clause
	:initarg :environment
	:documentation "A state-space clause specifying when the value holds.")))

(defmethod print-object ((pair labeled-pair) stream)
  (format stream "<PAIR: ~a ~a>" (pair-value pair) (pair-environment pair)))

(defclass labeled-value-set ()
  ((pairs
	:accessor lvs-pairs
	:initform nil
	:documentation "A list of the labeled value set pairs.")


   (strict-order
	:accessor lvs-strict-order
	:initarg :strict-order
	:documentation "A function used to determine ordering of the LVS, used
to prune as available. e.g. b< preferentially keeps small values.")

   (value-equalp
	:accessor lvs-value-equalp
	:initarg :value-equalp
	:documentation "A function provides an equality test for values.")

   (state-space
	:type state-space:state-space
	:accessor lvs-state-space
	:initarg :state-space
	:documentation "The state-space defining environments for the LVS.")))

(defclass strict-labeled-value-set (labeled-value-set)
  ())

(defclass weak-labeled-value-set (labeled-value-set)
  ())

(defmethod print-object ((lvs labeled-value-set) stream)
  (format stream "<LVS: ~a>" (lvs-pairs lvs)))

(defmethod copy-lvs ((lvs labeled-value-set))
  "Make a copy of the lvs."
  (let ((new-lvs (make-instance (type-of lvs) 
								:strict-order (lvs-strict-order lvs)
								:value-equalp (lvs-value-equalp lvs)
								:state-space (lvs-state-space lvs))))
	(setf (lvs-pairs new-lvs) (copy-list (lvs-pairs lvs)))
	new-lvs ))

(defmethod make-labeled-value-set-from-description ((state-space state-space) 
													lvs-type
													strict-order
													value-equalp
													description)
  "Construct an entire lvs from description, for example:
 (make-labeled-value-set-from-description #'< state-space '((2 ((= a 1))) (4 ())))"
  ;;make the new lvs
  (let ((lvs (make-instance lvs-type 
							:strict-order strict-order
							:value-equalp value-equalp
							:state-space state-space)))
  
	;;add all the pairs
	(dolist (pair-description description)
	  (add-pair-to-lvs-by-description! lvs pair-description))

	;;return the result
	lvs))


(defmethod lvs-length ((lvs labeled-value-set))
  "Return the number of pairs in the LVS."
  (length (lvs-pairs lvs)))

(defmethod query-lvs-pairs ((lvs labeled-value-set) (query-environment state-clause))
  "A function that returns the value associated with the environment. If there is an ordering,
returns the dominant subsuming value, if not, returns some subsuming value."
  (loop for pair in (lvs-pairs lvs)		;loop over the pairs
	 with dominant-value

	 ;;and return the first applicable one you find
	 when (and (null dominant-value)
			   (clause-subsumed? query-environment (pair-environment pair))) ;when the new one is applicable
	 do (setf dominant-value (pair-value pair))

	 ;;keep going until you find a dominant value, then go past it to higher
	 ;;values
	 until (and (not (null dominant-value))
				(funcall (lvs-strict-order lvs) dominant-value (pair-value pair)))
	 ;;once you've found the dominant value, keep all the variants that
	 ;;also subsume the input env
	 when (and (not (null dominant-value))
			   (clause-subsumed? query-environment (pair-environment pair))) 
	 collect pair into result
	 finally (return-from query-lvs-pairs result))
  )										;return the dominant value

(defmethod query-lvs ((lvs labeled-value-set) (query-environment state-clause))
  (let ((dominant-pairs (query-lvs-pairs lvs query-environment)))
	(unless (null dominant-pairs)
	  (mapcar #'pair-value dominant-pairs))))

(defmethod query-lvs-by-description ((lvs labeled-value-set) environment-description)
  "Call query-lvs, but create the environment from a description first.
A convenience function, mostly for testing purposes."
  (query-lvs lvs (get-clause-from-descriptions (lvs-state-space lvs) 
											   environment-description)))

(defmethod add-pair-to-lvs-by-description! ((lvs labeled-value-set) pair-description)
  "Takes a description of the pair, a list (value environment), where the environment
is a description of a state-clause from the state-space."

  ;;Construct the pair and add it
  (add-pair-to-lvs! lvs 
					(make-instance 'labeled-pair :value (first pair-description) 
								   :environment (get-clause-from-descriptions (lvs-state-space lvs) (second pair-description)))))

(defmethod add-pair-to-lvs! ((lvs labeled-value-set) (new-pair labeled-pair))
  "Add the new-pair to the lvs, modifying it as necessary to maintain minimality.
Return t if the pair is added, nil otherwise."

										;(format t "adding: ~a ~a ~a ~a~%" new-pair (not (null (pair-environment new-pair))) 
										;		(mapcar  #'(lambda (old-pair) (not (is-pair-dominant? lvs old-pair new-pair))) (lvs-pairs lvs))
										;		(apply #'and (mapcar  #'(lambda (old-pair) (not (is-pair-dominant? lvs old-pair new-pair))) (lvs-pairs lvs))))
  ;;Keep the new pair when either there's no order, or none of the old pairs are dominant
  (unless (null (pair-environment new-pair))
	(if (null (lvs-pairs lvs))
		(progn (setf (lvs-pairs lvs) (list new-pair))
			   (return-from add-pair-to-lvs! t))
		(loop for existing-pair in (lvs-pairs lvs)
		   with place-found? = nil

		   ;;when you newly find the place where it goes, keep this one
		   when (and (not place-found?)
					 (is-pair-value-dominant? lvs new-pair existing-pair))
		   collect new-pair into kept-pairs
		   and do (setf place-found? t)

		   ;;until you find the place, nothing can subsume the new one
		   never (and (not place-found?)
					  (does-pair-replace? lvs new-pair existing-pair))
	   
		   ;;until we find the new place, keep everything
		   when (and (not place-found?)
					 (keep-dominant-pair? lvs new-pair existing-pair)) 
		   collect existing-pair into kept-pairs 
		 
		   ;;after the place is found, keep only those that aren't dominated
		   when (and place-found?
					 (keep-non-dominant-pair? lvs new-pair existing-pair))
		   collect existing-pair into kept-pairs

		   ;;finally, if no problems found, save the new version
		   finally 
		   (if place-found?
			   (setf (lvs-pairs lvs) kept-pairs)
			   (setf (lvs-pairs lvs) (append kept-pairs (list new-pair))))
		   
		   (return-from add-pair-to-lvs! t))))
  nil)

(defmethod add-pairs-to-lvs! ((lvs labeled-value-set) pairs)
  (dolist (pair pairs)
	(add-pair-to-lvs! lvs pair)))


(defmethod keep-dominant-pair? ((lvs weak-labeled-value-set)
								(new-pair labeled-pair)
								(old-pair labeled-pair))
  ;;since it's weak, keep the dominant one, unless it's only dominant
  ;;because it's equal in value, even though it's more specific
  (unless (and (funcall (lvs-value-equalp lvs) 
						(pair-value new-pair)
						(pair-value old-pair))
			   (clause-subsumed? (pair-environment old-pair)
								 (pair-environment new-pair)))
	t))


(defmethod keep-dominant-pair? ((lvs strict-labeled-value-set)
									(new-pair labeled-pair)
									(old-pair labeled-pair))
  ;;always keep
  t)

(defmethod keep-non-dominant-pair? ((lvs weak-labeled-value-set)
									(new-pair labeled-pair)
									(old-pair labeled-pair))
  (or (not (clause-subsumed? (pair-environment old-pair) 
							 (pair-environment new-pair)))
	  (funcall (lvs-value-equalp lvs) 
			   (pair-value new-pair) 
			   (pair-value old-pair))))

(defmethod keep-non-dominant-pair? ((lvs strict-labeled-value-set)
									(new-pair labeled-pair)
									(old-pair labeled-pair))
  (not (clause-subsumed? (pair-environment old-pair) 
						 (pair-environment new-pair))))


(defmethod does-pair-replace? ((lvs labeled-value-set) 
							   (replaced-pair labeled-pair) 
							   (replaced-by labeled-pair))
  "The pair-maybe-dominant is dominant, returning true, over pair-other, if the 
environment of other is subsumed and the other value does not pass lvs-order."
										;(format t "dominant test: ~a ~a ~%" (is-pair-applicable? lvs pair-other pair-maybe-dominant)
										;	 (not (is-pair-value-dominant? lvs pair-other pair-maybe-dominant)))

  (and (clause-subsumed? (pair-environment replaced-pair)
						 (pair-environment replaced-by))
	   (or (is-pair-value-dominant? lvs replaced-by replaced-pair)
		   (funcall (lvs-value-equalp lvs)
					(pair-value replaced-pair)
					(pair-value replaced-by)))))



(defmethod is-pair-value-dominant? ((lvs labeled-value-set) (pair-maybe-dominant labeled-pair) (pair-other labeled-pair))
  "A pair is value dominant if the ordering over values indicates it is."  
  (funcall (lvs-strict-order lvs) 
		   (pair-value pair-maybe-dominant) 
		   (pair-value pair-other)))


(defmethod merge-lvs ((x labeled-value-set) (y labeled-value-set))
  "Take the pairs from the two inputs and produce a new lvs."

  (let ((merged-lvs (make-instance (type-of x)
								   :strict-order (lvs-strict-order x) 
								   :value-equalp (lvs-value-equalp x) 
								   :state-space (lvs-state-space x))))
	(dolist (x-pair (lvs-pairs x))
	  (add-pair-to-lvs! merged-lvs x-pair))
	(dolist (y-pair (lvs-pairs y))
	  (add-pair-to-lvs! merged-lvs y-pair))

	merged-lvs))


(defmethod merge-lvs-in-place! ((merge-into labeled-value-set) (values-to-merge labeled-value-set))
  "Take the pairs from the old one and modify the new one"
  (dolist (new-pair (lvs-pairs values-to-merge))
	(add-pair-to-lvs! merge-into new-pair)))


(defmethod lvs-binary-function ((lvs-a labeled-value-set) (lvs-b labeled-value-set) f)
  "Return a new lvs with the same properties as lvs-a, and apply f(lvs-a, lvs-b). f takes
two arguments."
  (assert (functionp f))
  (let ((new-lvs (make-instance (type-of lvs-a)
								:strict-order (lvs-strict-order lvs-a) 
								:value-equalp (lvs-value-equalp lvs-a)
								:state-space (lvs-state-space lvs-a))))

	(loop for candidate-pair in (lvs-binary-function-candidates lvs-a lvs-b f)
	   do (add-pair-to-lvs! new-lvs candidate-pair))
	
	new-lvs))


(defmethod lvs-trinary-function ((lvs-a labeled-value-set) (lvs-b labeled-value-set) (lvs-c labeled-value-set) f )
  "Return a new lvs with the same properties as lvs-a, and apply f(lvs-a, lvs-b). f takes
two arguments."
  (assert (functionp f))
  (let ((new-lvs (make-instance (type-of lvs-a)
								:strict-order (lvs-strict-order lvs-a) 
								:value-equalp (lvs-value-equalp lvs-a)
								:state-space (lvs-state-space lvs-a))))

	;; (loop for pair-a in (lvs-pairs lvs-a)
	;;    do (add-pair-to-lvs! 
	;; 				   new-lvs
	;; 				   (make-instance
	;; 					'labeled-pair 
	;; 					:environment (pair-environment pair-a)
	;; 					:value
	;; 					(funcall f
	;; 							 (pair-value pair-a)
	;; 							 (first (query-lvs lvs-b (pair-environment pair-a)))
	;; 							 (first (query-lvs lvs-c (pair-environment pair-a))))))
	;; 	 )
	(loop for pair-a in (lvs-pairs lvs-a)

	   do (loop for pair-b in (lvs-pairs lvs-b)
			 do (loop for pair-c in (lvs-pairs lvs-b)
				   with new-env
				   do (setf new-env (clause-intersection 
									 (pair-environment pair-c)
									 (clause-intersection 
									  (pair-environment pair-a)
									  (pair-environment pair-b))))
				   when 
				   (not (null new-env))
				   do (add-pair-to-lvs! 
					   new-lvs
					   (make-instance
						'labeled-pair 
						:environment new-env
						:value
						(funcall f
								 (first (query-lvs lvs-a new-env))
								 (first (query-lvs lvs-b new-env))
								 (first (query-lvs lvs-c new-env)))))

				   ))
	   )
	
	new-lvs))


(defmethod update-lvs-with-binary-operation! ((lvs-to-update labeled-value-set)
											  (lvs-a labeled-value-set)
											  (lvs-b labeled-value-set)
											  f)
  "Apply f(a,b)->to-update."

  ;;compute the candidates and add them to the lvs.
  (loop for new-pair in (lvs-binary-function-candidates lvs-a lvs-b f)
	 do (add-pair-to-lvs! lvs-to-update new-pair)))


(defmethod lvs-binary-function-candidates ((lvs-a labeled-value-set) 
									   (lvs-b labeled-value-set)
									   f)
;;Collect the candidates
(loop for pair-a in (lvs-pairs lvs-a)
   appending (loop for pair-b in (lvs-pairs lvs-b)
				collecting 
				  (make-instance 'labeled-pair 
								 :value (funcall f (pair-value pair-a)
												 (pair-value pair-b))
								 :environment (clause-intersection 
											   (pair-environment pair-a)
											   (pair-environment pair-b))))))


(defmethod remove-invalid-pairs! ((lvs labeled-value-set))
  (setf (lvs-pairs lvs)
		(remove-if #'(lambda (pair)
					   (not (state-clause-valid? (pair-environment pair))))
				   (lvs-pairs lvs))))


(defmethod prune-lvs-of-invalid-environments! ((lvs labeled-value-set))
  "Remove all pairs with invalid environments"
  (with-slots (pairs) lvs
	(setf pairs (delete-if-not #'(lambda (pair) 
								   (state-clause-valid? (pair-environment pair)))
							   pairs))))

(defmethod remove-subsumed-pairs! ((lvs labeled-value-set)
								   (environment state-clause))
  (debug-print-for :mtk-graph t "Clean up: ~a of ~a" lvs environment)
  (with-slots (pairs) lvs
	(setf pairs (remove-if #'(lambda (x) 
							   (clause-subsumed? (pair-environment x) environment))
						   pairs)))

  (debug-print-for :mtk-graph t "Cleaned: ~a of ~a" lvs environment)
  lvs)


(defmethod remove-equal-pairs! ((lvs labeled-value-set)
								(environment state-clause))
  (debug-print-for :mtk-graph t "Clean up: ~a of ~a" lvs environment)
  (with-slots (pairs) lvs
	(setf pairs (remove-if #'(lambda (x) 
							   (eql (pair-environment x) environment))
						   pairs)))
  (debug-print-for :mtk-graph t "Cleaned: ~a of ~a" lvs environment)
  lvs)


(defmethod clear-dependence-on-env! ((lvs labeled-value-set) (env state-clause))
  ;;first, remove everything completely subsumed by the env
  (remove-subsumed-pairs! lvs env)

  ;;now take anything that remains and split around the undesirable one
  (loop for pair in (lvs-pairs lvs)
										;if (clause-subsumed? env (pair-environment pair))
	 append  (mapcar #'(lambda (new-env) 
						 (make-instance 
						  'labeled-pair
						  :environment (clause-intersection (pair-environment pair)
															new-env)
						  :value (pair-value pair)))
					 (state-space::kernels-of-conflict env))
	 into new-pairs
										;else append (list pair) into new-pairs
	 finally (setf (lvs-pairs lvs) nil)
	 (add-pairs-to-lvs! lvs new-pairs)
	 )
  
  (debug-print-for :mtk-graph t "Fully cleaned: ~a of ~a ~a" lvs env (state-space::kernels-of-conflict env)))


(defmethod apply-stable-unary-operator! ((lvs labeled-value-set)										 f)
  "Apply a function on each value, replacing the old one."
  (loop for pair in (lvs-pairs lvs)
	 do (setf (pair-value pair) (funcall f (pair-value pair)))))