;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package :mtk-graph)

;;-------------------------------------------------------------------------------
;; Weighted Graph types
;;   * Weighted Hyper Graph (undirected)
;;   * Weighted Directed Hyper Graph
;;   * Weighted Graph (undirected)
;;   * Weighted Directed Graph
;;-------------------------------------------------------------------------------

(defclass edge-weights ()
  ((weight-ht-edge :type hash-table
		   :accessor get-weight-ht-edge
		   :initform (make-hash-table)
		   :documentation "")
   (default-weight :accessor get-default-weight
     :initarg :default-weight
     :initform :default-weight
     :documentation "")))

(defclass weighted-hyper-graph (hyper-graph
                                edge-weights) ())

;;; added by Yi (Andrew) Huang
;;; print the information of a weighted-hyper-graph object
(defmethod print-object ((g weighted-hyper-graph) stream)
  (with-slots ((v v) 
			   (e e)
			   (w weight-ht-edge)) g
	(format stream "Vertices: ~a~%" v)
	(format stream "~a" e)
	(format stream "Weights:~%")
	(loop 
	   for item being the hash-key of w  
	   do (format stream "~a: ~a;~%" item (gethash item w)))
	))


(defclass weighted-directed-hyper-graph (weighted-hyper-graph 
										 directed-hyper-graph) ())
(defclass weighted-graph (weighted-hyper-graph
						  graph) ())
(defclass weighted-directed-graph (weighted-graph
								   directed-graph) ())

(defmethod get-weight ((e_i hyper-edge) (w edge-weights))
  (multiple-value-bind (weight present-p) (gethash e_i (get-weight-ht-edge w))
    (if present-p
	weight
      (get-default-weight w))))
 
(defmethod set-weight ((e_i hyper-edge) weight (w edge-weights))
  (setf (gethash e_i (get-weight-ht-edge w)) weight))

(defmethod remove-weight ((e_i hyper-edge) (w edge-weights))
  (remhash e_i (get-weight-ht-edge w)))

;;-------------------------------------------------------------------------------
;; Delete vertex/vertices in a graph
;;-------------------------------------------------------------------------------

(defmethod delete-vertex (vertex (g weighted-hyper-graph))
  (delete-weighted-edges (find-edges-with-vertex vertex g) g)
  (setf (get-vertices g) (delete vertex (get-vertices g))))

;;-------------------------------------------------------------------------------
;; Add edge(s) to a graph
;;-------------------------------------------------------------------------------

(defmethod add-weighted-edge ((weighted-edge-description list)
			      (g weighted-hyper-graph))
  (assert (= (length weighted-edge-description) 2))
  (let* ((edge-description (first weighted-edge-description))
	 (weight (second weighted-edge-description))
	 (e_i (add-edge edge-description g)))
    (set-weight e_i weight g)
    e_i))

(defmethod add-weighted-edges ((weighted-edge-descriptions list)
			       (g weighted-hyper-graph))
  (dolist (weighted-edge-description weighted-edge-descriptions)
    (add-weighted-edge weighted-edge-description g)))

;;-------------------------------------------------------------------------------
;; Delete edge(s) in a graph
;;-------------------------------------------------------------------------------

(defmethod delete-weighted-edge ((e_i hyper-edge) (g weighted-hyper-graph))
  (remove-weight e_i g)
  (delete-edge e_i (get-edges g)))

(defmethod delete-weighted-edges ((e list) (g weighted-hyper-graph))
  (dolist (e_i e) (delete-weighted-edge e_i g)))
