;;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)

;;; To Do:

;;; o Should remove set-inconsistent-clauses and supporting functions, once
;;;   its callers have been modified.

;;; ***********************************************************************
;;;
;;;                Inconsistency and Extracting Conflicts from Theory
;;;
;;; ***********************************************************************

;;; This module contains functions for examining inconsistent theories, 
;;; inconsistent clauses and inconsistent subsets of a theory. 

;;; We define a conflict as a partial truth assignment that is
;;; inconsistent with the clauses of theory, where each assignment
;;; has been provided by ISAT splitting or an external source
;;; (i.e., not deduced by unit propagation).
  
;;; Note that early implementations returned as a conflict a subset of the 
;;; clauses of theory that are inconsistent.  It then implemented 
;;  assignments as unit clauses.  Extracting the set of 
;;; inconsistent clauses may be harder to derive when the inconsistency 
;;; is found through systematic search, hence we have removed this 
;;; functionality.

;;; External Module interface functions consist of the following:

;;; (CONFLICT-OF-INCONSISTENT-THEORY THEORY 
;;;    &OPTIONAL (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT))

;;;    Given inconsistent THEORY, returns a CONFLICT.  Only call this if
;;;    the theory is known to be inconsistent.  A CONFLICT is 
;;;    generated by applying the function KEY to each clause supporting an 
;;;    inconsistent clause of THEORY, and by collecting all non-nil,
;;;    irredundant values returned by key.

;;; (UNIT-PROPAGATION-IMPLICANT-OF-CLAUSES CLAUSES-TO-ENTAIL 
;;;    &OPTIONAL (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT))
  
;;;    Given CLAUSES-TO-ENTAIL, supported by a theory, 
;;;    returns an IMPLICANT for CLAUSES-TO-ENTAIL 
;;;    based on unit propagation.  

;;;    An implicant is created by applying KEY to each clause supporting
;;;    the entailment of one of the clauses in CLAUSES-TO-ENTAIL, and by
;;;    collecting the non-nil non-duplicate returned values of key.

;;;    CLAUSES-TO-ENTAIL may include clauses that have been removed 
;;;    from theory, in which case this function only tries to 
;;;    entail clauses that are currently inserted into theory.

;;;(eval-when (compile load eval)
;;;  (export '(conflict-of-inconsistent-theory ; New Module Interface
;;;			unit-propagation-implicant-of-clauses ; New Module Interface
;;;
;;;			;; Retire the following
;;;			theory-conflicting-assignments ; Old ISAT interface
;;;			theory-conflict				; Old ISAT interface
;;;            consistent-theory?			; Old ISAT interface
;;;			)
;;;		  :ISAT)
;;;  )

;;; ***********************************************************************
;;;
;;;                  Interface
;;;
;;; ***********************************************************************

(defmethod conflict-of-inconsistent-theory
    ((theory theory) &optional (key #'proposition-to-truth-assignment))
  
  " Given an inconsistent THEORY, returns a CONFLICT.  

    Only call this function if THEORY is known to be inconsistent.  

    A CONFLICT denotes a partial truth assignment that is inconsistent.
    It is generated by applying function KEY to each externally assigned 
    proposition that supports an inconsistent clause of THEORY,
    where an externally assigned proposition is one that is not 
    supported by unit propagation.

    All duplicate and nil values are removed from CONFLICT.

    By default, KEY takes PROPOSITION as input and returns assignment
   (= PROPOSITION TRUTH-VALUE) where TRUTH-VALUE is *true* or *false*."

  (debug-print-for :isat t "Looking for conflict of inconsistent theory ~A." 
				   theory)
  
  (let ((inconsistent-clause
		 (lhs-get-first (theory-inconsistent-clauses theory))))
    
    (debug-progn
     (when (null inconsistent-clause)
	   ;; No inconsistent clause in inconsistent theory;
	   ;; this shouldn't happen.
       (error t "Supplied an allegedly inconsistent theory ~A, ~
                that has no inconsistent clauses." theory)))

    (debug-print-for :itms t "Extracting external assignments ~
                    ~% supporting inconsistent clause ~A." 
					 inconsistent-clause)
	
	;; Trace over the support for INCONSISTENT-CLAUSE,
	;; applying KEY to each externally assigned proposition
	;; and collecting the non-nil, non-redundant values.  
	;; Return the accumulated values.

	;; An externally assigned proposition is one that is not
	;; assigned by unit-propagation.
    (let ((conflict
		   (unit-propagation-implicant-of-clause
			inconsistent-clause nil key '())))
	  
      (debug-print-for :isat t "Conflict is ~A." conflict)
      conflict)))

(defun unit-propagation-implicant-of-clauses
    (clauses-to-entail &optional (key #'proposition-to-truth-assignment))
  
  " Given CLAUSES-TO-ENTAIL, supported by a theory, returns an 
    IMPLICANT for CLAUSES-TO-ENTAIL based on unit propagation.  
  
    An IMPLICANT denotes a partial truth assignment that together 
    with the rest of theory entails CLAUSES-TO-ENTAIL.
    It is generated by applying function KEY to each externally 
    assigned proposition that supports an inconsistent clause of 
    THEORY, where an externally assigned proposition is one that 
    is not supported by unit propagation.

    Duplicate and null values of IMPLICANT are removed.

    By default, KEY takes PROPOSITION as input and returns 
    assignment (= PROPOSITION TRUTH-VALUE) where TRUTH-VALUE 
    is *true* or *false*.

    CLAUSES-TO-ENTAIL may include clauses that have been removed 
    from theory, in which case this function only tries to 
    entail clauses that are currently inserted into theory."

  (let ((implicant '()))
    (loop with proposition
		for clause-to-entail in clauses-to-entail
		unless (constraint-removed? clause-to-entail)
		do (debug-print-for :itms t "Finding implicant for clause ~A." 
							clause-to-entail)
		   (setq proposition
			 (literal-description-proposition
			  (find-proposition-of-true-literal clause-to-entail)))
		   (setq implicant
			 (unit-propagation-implicant-of-proposition
			  proposition key implicant)))
    implicant))

;;; ***********************************************************************
;;;
;;;                  Support Functions for Extracting Conflicts
;;;
;;; ***********************************************************************

;;; To Do:
;;; o Simplify the following two functions using the macros for 
;;;   iterating over clause propositions.

(defmethod find-proposition-of-true-literal ((clause clause))
  
  " Searches the literals in CLAUSE and returns the proposition
    of a literal that is true.   If there is no true literal 
    it returns NIL.  Note: there may be several true literals;
    this only returns the proposition for one."
  
  (or (dolist (proposition (clause-positive-propositions clause))
		(when (eql (proposition-truth-value proposition) *true*)
		  (return proposition)))

      (dolist (proposition (clause-negative-propositions clause))
		(when (eql (proposition-truth-value proposition) *false*)
		  (return proposition)))))

;;; Optimization:

;;; o Unless the set of support is a tree, this
;;;   will visit propositions more than once.
;;;   Use a mark and sweep algorithm to avoid
;;;   duplicate visits.

;;; o This performs recursion.  Replace with a
;;;   stack discipline.

(defmethod unit-propagation-implicant-of-clause 
    ((clause clause) proposition key implicant)
  
  " IMPLICANT is a list of values already extracted.

    Returns IMPLICANT augmented with the assignments
    produced by applying the function KEY to each 
    externally supported proposition that supports
    a literal of CLAUSE, with the exception of the 
    literal containing PROPOSITION.  Removes non-nil 
    and duplicate values; duplicates are removed 
    using EQUAL.

    A proposition is externally supported if its 
    assignment is not supported by unit propagation.

    PROPOSITION is either a proposition instance or nil.

    CLAUSE is either the support for the truth assignment 
    to PROPOSITION, or an inconsistent clause, in which 
    case PROPOSITION must be assigned nil. If CLAUSE is the 
    support for PROPOSITION, then the literal in CLAUSE 
    containing PROPOSITION must be assigned *true*, 
    while its remaining literals are assigned *false*."

  ;; Trace back through the support for the 
  ;; positive literals.
  (debug-print-for :itms t "Tracing support for literals of ~A." 
				   clause)
  (dolist (prop (clause-positive-propositions clause))
	(unless (eql prop proposition)
	  ;; Don't trace over PROPOSITION since it is the 
	  ;; consequent that is being supported.
	  (setq implicant 
		(unit-propagation-implicant-of-proposition 
		 prop key implicant))))

  ;; Next trace back through the support for the 
  ;; negative literals.
  (dolist (prop (clause-negative-propositions clause))
	(unless (eql prop proposition)
	  ;; Don't trace over PROPOSITION since it is the 
	  ;; consequent that is being supported.
	  (setq implicant
		(unit-propagation-implicant-of-proposition 
		 prop key implicant))))

  ;; Finally, return the collected IMPLICANT.
  (values implicant))

;;; Optimization:

;;; o Unless the set of support is a tree, this
;;;   will visit propositions more than once.
;;;   Use a mark and sweep algorithm to avoid
;;;   duplicate visits.

;;; o This performs recursion.  Replace with a
;;;   stack discipline.

;;; o Is there a more efficient means of removing
;;;   redundant assignments?  For example, can this
;;;   fall out of using a mark and sweep?

(defmethod unit-propagation-implicant-of-proposition
    ((proposition proposition) key implicant)
  
  "Returns IMPLICANT augmented with assignments
   produced by applying the function KEY to each 
   proposition that is externally supported
   and supports the truth assignment of PROPOSITION.  
   Duplicate and null values are removed from 
   IMPLICANT.

   The input IMPLICANT is a list of assignments 
   already extracted.    

   A proposition is externally supported if its truth
   assignment is not supported by unit propagation."

  (let ((support-clause (proposition-support proposition)))
    (cond ((unit-propagation-support? support-clause)

		   ;; When PROPOSITION has a support clause, ...
		   (debug-print-for :itms t "Tracing support for proposition ~A." 
							proposition)

		   ;; Extract an implicant for the antededent
		   ;; literals of support-clause added to
		   ;; the IMPLICANT accumulated thus far...
		   (setq implicant
			 (unit-propagation-implicant-of-clause
			  support-clause proposition key implicant)))
	  
		  (:otherwise

		   ;; Proposition is externally supported, 
		   ;; add it to IMPLICANT...
		   (let ((implicant-assignment (funcall key proposition)))
			 (debug-print-for 
			  :itms t "Extracting assignment ~A from ~A using ~A."  
			  implicant-assignment proposition key)
			 (unless (null implicant-assignment)
			   (debug-print-for :itms t "Adding ~A to implicant" 
								implicant-assignment)
			   (pushnew implicant-assignment
						implicant :test #'equal)))))

	;; and return this augmented IMPLICANT.
    (values implicant)))

;;; ***********************************************************************
;;;
;;;                  Generic Functions for Tracing Support
;;;
;;; ***********************************************************************

;;; The following were used for conflict extraction.  They are no
;;; longer used, but are generally useful utility functions.

;;; To Do:
;;; o Update using the macros for iterating over clause propositions.

(defmethod trace-clause-dependencies 
    ((clause clause) (proposition proposition) key collected-values)
  
  " COLLECTED-VALUES is a list of values already extracted.

    Returns COLLECTED-VALUES augmented with the values
    produced by applying the function KEY to each clause
    supporting a literal of CLAUSE, except the one 
    containing PROPOSITION.  Removes non-nil and duplicate
    values.  Duplicates are removed using EQUAL.

    CLAUSE is either the support for the truth assignment 
    to PROPOSITION, or an inconsistent clause, in which 
    case PROPOSITION must be assigned nil. If CLAUSE is the 
    support for PROPOSITION, then the literal in CLAUSE 
    containing PROPOSITION must be assigned *true*, while
    its remaining literals are assigned *false*."

  ;; First see if KEY applied to conflict returns a non-NIL value, 
  ;; and if so add that value to COLLECTED-VALUES

  (let ((collected-value (funcall key clause)))
    (debug-print-for :itms t "Extracted value ~A ~
                    from ~A using ~A."  collected-value clause key)

    (when collected-value
      (debug-print-for :itms t "Adding ~A to collecte-values" collected-value)
      (pushnew collected-value collected-values :test #'equal))

	;; Now trace back through the dependencies for the 
	;; positive literals.
    (debug-print-for :itms t "Tracing support for literals of ~A." 
				 clause)
    (dolist (prop (clause-positive-propositions clause))
      (unless (eql prop proposition)
		;; Don't trace over PROPOSITION since it is the 
		;; consequent that is being supported.
		(setq collected-values 
		  (trace-proposition-dependencies 
		   prop key collected-values))))

	;; Then trace back through the dependencies 
	;; for the negative literals
    (dolist (prop (clause-negative-propositions clause))
      (unless (eql prop proposition)
		;; Don't trace over PROPOSITION since it is the 
		;; consequent that is being supported.
		(setq collected-values
		  (trace-proposition-dependencies
		   prop key collected-values))))

	;; And, finally, return the collected COLLECTED-VALUES.
    (values collected-values)))

(defmethod trace-proposition-dependencies 
    ((proposition proposition) key collected-values)
  
  " COLLECTED-VALUES is a list of values already extracted.

    Returns COLLECTED-VALUES augmented with the non-nil 
    values produced by applying the function KEY to 
    each clause supporting the truth assignment of 
    PROPOSITION."

  (let ((support-clause (proposition-support proposition)))
    (unless (null support-clause)

	  ;; When PROPOSITION has a support clause, ...
      (debug-print-for :itms t "Tracing support for proposition ~A." 
				   proposition)

	  ;; Extract and add a value for that clause
	  ;; to the collected-values accumulated thus far, ..
      (setq collected-values
		(trace-clause-dependencies 
		 support-clause proposition key collected-values))
	  ;; and return this augmented set of values.
      (values collected-values))))
