;;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)

;;; ***************************************************************************
;;;
;;;                          Checking Consistency
;;;
;;; ***************************************************************************

;;; ***************************************************************************
;;;
;;;                           External Interface
;;;
;;; ***************************************************************************


;;;   (CONSISTENT?
;;;    (THEORY THEORY) &OPTIONAL &KEY
;;;  	         (PERFORM-SEARCH? T)
;;;	             (RETURN-CONFLICT? T)
;;;	             (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT))
  
;;;    Checks consistency of THEORY.
;;;    Returns (values CONSISTENT? CONFLICT).

;;;    CONSISTENT? is non-NIL iff an assignment is found, or
;;;    perform-search? is false and unit propagation does not 
;;;    find an inconsistency.

;;;    If theory is found inconsistent based on unit-propagation 
;;;    alone, and RETURN-CONFLICT? is TRUE, then a CONFLICT 
;;;    is returned.

;;;    A CONFLICT denotes a partial truth assignment that is 
;;;    inconsistent.  It is generated by applying function 
;;;    KEY to each externally assigned proposition that 
;;;    supports an inconsistent constraint of THEORY, and by 
;;;    collecting the non-nil, irredundant values returned by KEY.
;;;    An externally assigned proposition is one that is not 
;;;    supported by unit propagation.

;;;    By default, KEY takes PROPOSITION as input and returns 
;;;    assignment (= PROPOSITION TRUTH-VALUE) where 
;;;    TRUTH-VALUE is *true* or *false*.
   
;;;    If unit propagation does not find an inconsistency, 
;;;    and PERFORM-SEARCH? is true, then this function 
;;;    searches for a consistent ASSIGNMENT.

;;;    After an initial call with INITIALIZE? = T,
;;;    then additional assignments may be found by calling 
;;;    this function with INITIALIZE? = nil.  In particular, 
;;;    if INITIALIZE? is T, then searches from beginning; 
;;;    otherwise, searches from the current assignment, 
;;;    which should be the last consistent assignment found.

;;;    Side Effect:  Leaves propositions assigned to the 
;;;    truth values of the current search state.  

;;;    The effects of this search can be cleaned up by 
;;;    calling (clear-search-state! theory).


;;;    (CONSISTENT-WITH?
;;;     (theory theory) &optional &key 
;;; 		      (added-constraints nil)
;;; 		      (added-assignments nil)
;;; 		      (source nil)
;;; 		      (perform-search? t)
;;; 		      (return-conflict? t))
;;; 
;;;    Tests the consistency of the current constraints and 
;;;    assignments in THEORY, conjoined with ADDED-CONSTRAINTS 
;;;    and ADDED-ASSIGNMENTS.
;;; 
;;;    Returns (values CONSISTENT? CONFLICT).  
;;;    CONSISTENT? is true unless theory is proven inconsistent.  
;;;    It first tests consistency by unit propagation.  
;;; 
;;;    Each constraint in ADDED-CONSTRAINTS is a constraint object.
;;; 
;;;    Each assignment in ADDED-ASSIGNMENTS is a list:
;;; 
;;;        (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *true*))
;;; 
;;;    and specifies that proposition (= variable value) has TRUTH-VALUE.
;;;    
;;;    If PERFORM-SEARCH? is TRUE, and unit propagation is consistent, 
;;;    then it continues with a full DPLL-style satisfiability test.  
;;; 
;;;    If RETURN-CONFLICT? is TRUE and an inconsistency is 
;;;    found by unit propagation, then returns a subset of the 
;;;    current inserted assignments that lead to an inconsistency.


;;;   (INCONSISTENT-WITH? (theory theory) &optional &KEY 
;;; 		      (added-constraints nil)
;;; 		      (added-assignments nil)
;;; 		      (source nil)
;;; 		      (perform-search? T)
;;; 		      (return-conflict? T)
;;; 		      (base-theory-consistent-with-all-assignments? T))
;;; 
;;;    Tests the inconsistency of the current constraints and 
;;;    assignments in THEORY,
;;;    conjoined with ADDITIONAL-CONSTRAINTS.
;;; 
;;;    Returns (values INCONSISTENT? CONFLICT).  INCONSISTENT? 
;;;    is false unless theory is proven inconsistent.  
;;;    It first tests consistency by unit propagation.  Then if
;;;    if PERFORM-SEARCH? is TRUE, it continues with a full 
;;;    DPLL-style satisfiability test.  
;;; 
;;;    Each clause in ADDED-CONSTRAINTS is a clause object.
;;; 
;;;    Each assignment in ADDED-ASSIGNMENTS is a list:
;;; 
;;;        (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *true*))
;;; 
;;;    and specifies that proposition (= variable value) has TRUTH-VALUE.
;;; 
;;;    If RETURN-CONFLICT? is TRUE and theory can be proven 
;;;    necessarily true by unit propagation,
;;;    then returns a subset of the current assignments sufficient 
;;;    to guarantee that theory is satisfied.
;;; 
;;;    If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, then 
;;;    assumes that the inserted assignments are consistent with the 
;;;    clauses in theory (minus added-constraints).
;;;    In this case a conflict is a subset of the assignments that, 
;;;    together with THEORY (minus assignments) entail ADDED-CONSTRAINTS.
;;; 
;;;    A conflict is created by collecting a subset of the 
;;;    current clauses that support the truth 
;;;    of all ADDED-CONSTRAINTS, or all non-unit clauses.  
;;;    Next the function KEY is applied to each clause in the set, 
;;;    collecting the non-NIL, non-duplicate returned 
;;;    values as the CONFLICT.

(eval-when (compile load eval)
  (export '(;consistent?
			consistent-with?
			inconsistent-with?)
		  :ISAT));

;;; ***************************************************************************
;;;
;;;                  External Interface: 
;;;
;;; ***************************************************************************

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                  External Interface: Consistent?
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
;;; To check consistency:
;;; 1) Perform unit propagation.

;;; 2) If inconsistent return conflict:
;;;   o search for the assignment clauses supporting a
;;;     proposition and its negation.

;;; 3) Else test DPLL consistency.
;;;   o Doesn't return a conflict.
;;;   o Augmentation: Should be able to return a conflict.

;; To Do:

;;; o Extend ISAT to generate conflicts during chronological search 
;;;   as well as during clause.  These conflicts should be generated
;;;   in the style of CBJ.  Once this extension is made, we no
;;;   longer need to call unit propagation at the top level as a 
;;;   special case.

;;; o When search is called, do we want to restrict the amount of
;;;   search performed, and if so how?

;;; o Do we want a flag for clearing the search state at the end?

;;; o Generalize to arbitrary constraint systems:
;;;   o Split into support for search-based and pure
;;;     propagation-based sat.
;;;   o Generalize call to propagate in test for consistency.
;;;   o Generalize (clear-search-state! theory) for
;;;     search based sat engines, and turn off otherwise.
;;;   o Generalize call to (search-forward theory) for 
;;;     search-based consistency.

;;; Check:
;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't nil denote the
;;;   conflict of empty assignments, in which case everything is
;;;   inconsistent?  Check and correct this.

;;; To Do:
;;; o Document return-new-clauses?.  What does it do?
(defmethod consistent?
    ((theory theory) &optional &key
							   (perform-search? T)
							   (return-conflict? T)
							   (return-new-clauses? nil)
							   (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT))

  "Checks consistency of THEORY.
   Returns (values CONSISTENT? CONFLICT NEW-CLAUSES).

   CONSISTENT? is non-NIL iff an assignment is found, or
   perform-search? is false and unit propagation does not 
   find an inconsistency.

   If theory is found inconsistent based on unit-propagation 
   alone, and RETURN-CONFLICT? is TRUE, then a CONFLICT 
   is returned.

   A CONFLICT denotes a partial truth assignment that is 
   inconsistent.  It is generated by applying function 
   KEY to each externally assigned proposition that 
   supports an inconsistent clause of THEORY, and by 
   collecting the non-nil, irredundant values returned by KEY.
   An externally assigned proposition is one that is not 
   supported by unit propagation.

   If perform-search? is true, NEW-CLAUSES is the set of 
   clauses learned by conflict-directed search. Otherwise,
   it is nil.

   By default, KEY takes PROPOSITION as input and returns 
   assignment (= PROPOSITION TRUTH-VALUE) where 
   TRUTH-VALUE is *true* or *false*.
   
   If unit propagation does not find an inconsistency, 
   and PERFORM-SEARCH? is true, then this function 
   searches for a consistent ASSIGNMENT.

   After an initial call with INITIALIZE? = T,
   then additional assignments may be found by calling 
   this function with INITIALIZE? = nil.  In particular, 
   if INITIALIZE? is T, then searches from beginning; 
   otherwise, searches from the current assignment, 
   which should be the last consistent assignment found.

   Side Effect:  Leaves propositions assigned to the 
   truth values of the current search state.  

   The effects of this search can be cleaned up by 
   calling (clear-search-state! theory)."

  (debug-print-for :isat t "Clearing search state of ~A." theory)
  (clear-search-state! theory)

  (debug-print-for :isat t "Propagating ~A." theory)

  ;; Check consistency based on propagation alone.
  (let ((consistent? (propagate theory)))

    (cond ((not consistent?)
	   
		   ;; Unit-propagation-inconsistent
		   (debug-print-for :isat t
							"~A found inconsistent by unit propagation." 
							theory)

		   (return-inconsistency theory return-conflict? key))

		  ;; Consistent by propagation
		  (:otherwise

		   (debug-print-for :isat t "~A is propagation consistent." theory)

		   (cond (perform-search?

				  ;; Search for a consistent assignment
				  (debug-print-for 
				   :isat 
				   t "Searching for a model (consistent assignment).")
	       
				  ;; :change - if inconsistent, search-forward 
				  ;;           should try to return a conflict.
				  (let ((assignment-found? nil)
						(new-clauses-from-search nil))

					(multiple-value-setq
						(assignment-found? new-clauses-from-search)
					  (search-forward theory :return-new-clauses?
									  return-new-clauses?))

					(debug-print-progn-for
					 :isat
					 (if assignment-found?
						 (debug-print-for :isat t "Found model (~A) for ~A."
										  assignment-found? theory)
		     
					   (debug-print-for :isat t "No model found," ;; theory ~A ~
										;; is inconsistent." theory
										)
					   ))

					(values assignment-found? nil new-clauses-from-search)))

				 ;; Directed not to perform search, 
				 ;; return consistent based on unit propagation alone.
				 (:else

				  (values T nil nil)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                  External Interface:  Consistent-With?
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Fix:
;;;  o Check that callers pass in key where appropriate.

;;;  o Change OpSAT so that it passes in all assignments
;;;    as added-assignments, rather than clauses.
;;;    Make sure that these assignments are passed in
;;;    properly.

;;; Check:
;;;  o Should insert-constraint and remove-constraint take a 
;;;    source as input?

;;;  o Review the following function.

;;; Optimization:
;;;   o Before adding each assignment, check if the 
;;;     proposition has already been assigned the
;;;     opposite truth value.  If so, then trace the 
;;;     support of the proposition to extract a conflict.  
;;;     This will be faster than recording the 
;;;     assignment and updating the clauses to reflect
;;;     the inconsistency.

;;; To Do:
;;;   o Change the format of the returned conflict to be more
;;;     amenable to the needs of the caller, and to be
;;;     consistent with the format of the assignments supplied.

;;;   o Reconsider the format of added-assignments
;;;     to be proposition object/truth-value pairs.

;;; o Generalize to arbitrary constraint systems:
;;;   o Replace key with a constraint-system specific method?
;;;   o Generalize (Un)assert-assignments!.

;;; Check:
;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't nil denote the
;;;   conflict of empty assignments, in which case everything is
;;;   inconsistent?  Check and correct this.
(defmethod consistent-with?
    ((theory theory) &optional &key 
	                           (added-constraints nil)
							   (added-assignments nil)
							   (source nil)
							   (perform-search? t)
							   (return-conflict? t)
	                           (save-learned-conflicts? nil)
							   (key #'proposition-to-truth-assignment)
	 &allow-other-keys)

  "Tests the consistency of the current constraints and assignments in THEORY,
   conjoined with ADDED-CONSTRAINTS and ADDED-ASSIGNMENTS.

   Returns (values CONSISTENT? CONFLICT).  CONSISTENT? is true unless theory
   is proven inconsistent.  It first tests consistency by propagation.  

   Each clause in ADDED-CONSTRAINTS is a clause object.

   Each assignment in ADDED-ASSIGNMENTS is a list:

       (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *true*))

   and specifies that assignment proposition (= variable value) has TRUTH-VALUE.
   
   If PERFORM-SEARCH? is TRUE, and propagation is consistent, then it 
   continues with a full search-based satisfiability test,
   when one exists.

   If RETURN-CONFLICT? is TRUE and an inconsistency is found by
   propagation, then returns a conflict.  

   A CONFLICT denotes a subset of ADDED-ASSIGNMENTS that is inconsistent
   with theory and ADDED-CONSTRAINTS.  Each assignment is represented
   internally as an assignment object.  The conflict is generated by 
   applying function KEY to each such assignment that supports an 
   inconsistency of THEORY, and by collecting the non-nil, 
   irredundant values returned by KEY.

   By default, KEY takes a PROPOSITION object as input and returns 
   assignment (= PROPOSITION TRUTH-VALUE) where TRUTH-VALUE is 
   *true* or *false*."

  (declare (ignore initialize?))

  ;; Insert additional clauses and assignments
  (loop for constraint in added-constraints
      do (insert-constraint theory constraint))

  (assert-assignments! theory source added-assignments)
  
  ;; Test Consistency by propagation.

  ;; Conflict is a list of truth value assignments
  ;; (= PROPOSITION TRUTH-VALUE), where truth-value is
  ;; *true* or *false*, and PROPOSITION is a proposition
  ;; object whose proposition description is an expression
  ;; (= STATE-VARIABLE VALUE).
  
  (multiple-value-bind (consistent? conflict new-clauses-from-search)
      (consistent? theory
				   :perform-search? perform-search?
				   :return-conflict? return-conflict?
				   :return-new-clauses? (not save-learned-conflicts?)
				   :key key)

	;; Remove any clauses added by conflict learning
	(unless (null new-clauses-from-search)

	  (debug-print t "Remove conflict based clauses added to theory.")
	  (loop for clause in new-clauses-from-search
		  do (remove-constraint theory clause)))

	;; Remove the inserted constraints and assignments
    (loop for constraint in added-constraints
		do (remove-constraint theory constraint))

    (unassert-assignments! theory source added-assignments)
    
	;; Return the results

    (values consistent? conflict)))

;;; ***************************************************************************
;;;
;;;                  External Interface: Inconsistent-With?
;;;
;;; ***************************************************************************

;;; To Do: 

;;;   o Change the format of the returned conflict to be more
;;;     amenable to the needs of the caller, and to be
;;;     consistent with the format of the assignments supplied.

;;;   o Also reconsider the format of added-assignments
;;;     to be proposition object/truth-value pairs.

;;;  o When checking the entailment of clauses, the algorithm
;;;    should mark the two types of clauses, and maintain a count 
;;;    of how many of each type exists.  It should incrementally 
;;;    modify this count to reflect how many of them are not yet 
;;;    unit propagation entailed.

;;;  o We need more fine grained control over which clauses must 
;;;    be satisfied.  For example, clauses defining the exclusion 
;;;    and exhaustion of state variable assignments should always 
;;;    be satisfied by the assignments considered by OPSAT.

;;;  o Mark each clause that needs to be entailed,
;;;    or maintain them as a separate list within OPSAT.  
;;;    For this to be an explicit concept utilized within ISAT, 
;;;    we need to introduce into ISAT a concept of design variables
;;;    (analogous to assumptions). 

;;;  o Similar to above, OpSat should only use added-assignments
;;;    to make assignments, not added-constraints.  Update the calls
;;;    from Opsat similar to the comments for consistent-with?

;;;   o Generalize implicant extraction beyond implicants
;;;     generated via unit-propagation.

;;;  o Review the following function. It may be checking 
;;;    some conditions only by unit propagation, when that 
;;;    also need to be checked by search.

;;; Check:
;;;  o Should insert-constraint and remove-constraint take a source 
;;;    as input?  Alternative should the insertion and
;;;    removal be triggered by a boolean variable, or should
;;;    it be implemented as a conditional CSP?

;;;  o When updating the conflict extraction code, make 
;;;    sure that unit-propagation-implicant-of-clauses 
;;;    is correct.

;;; Optimization:
;;;   o Before adding each assignment, check it the proposition
;;;     already has the opposite assignment and trace the support
;;;     of the proposition to extract a conflict.  This should
;;;     be much faster than recording  the assignment and 
;;;     updating the constraints accordingly.  Note: this comments was
;;;     written for consistent-with?  Optimizing for inconsistent-with?
;;;     will be somewhat different.

;;; Fix:
;;; o Generalize to arbitrary constraint systems:
;;;   o Alter key to method?
;;;   o Divide into propagation-based and search based
;;;     (in)consistency tests.
;;;   o Generalize clear-search-state!, propagate,
;;;     theory-clauses, clauses-are-entailed?,
;;;     unit-propagation-implicant-of-clauses?
;;;     and search-forward?
;;;   o What does propagation entailment mean for
;;;     systems like ITC?

;;; Check:
;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't nil denote the
;;;   conflict of empty assignments, in which case everything is
;;;   inconsistent?  Check and correct this.

;;; Fixed:
;;; o After performing propagation to test inconsistency, we should clear
;;; the effects of it before starting to check whether
;;; " theory + added-assignments entail added-constraints ".
;;; This was not being done before, causing ISAT to generates wrong conflicts.
;;; gustavop
(defmethod inconsistent-with? 
    ((theory theory) &optional &KEY 
							   (added-constraints nil)
							   (added-assignments nil)
							   (source nil)
							   (perform-search? T)
							   (return-conflict? T)
	                           (save-learned-conflicts? nil)
							   (key #'proposition-to-truth-assignment)
							   (base-theory-consistent-with-all-assignments? T)
	 &allow-other-keys)

  "Tests the inconsistency of the current constraints and assignments in THEORY,
   conjoined with ADDED-CONSTRAINTS.

   Returns (values INCONSISTENT? CONFLICT).  INCONSISTENT? is false unless 
   theory is proven inconsistent.  It first tests consistency by
   propagation.  Then if PERFORM-SEARCH? is TRUE, it continues with a full 
   search-based satisfiability test, if one exists.  

   Each constraint in ADDED-CONSTRAINTS is a constraint object.

   Each assignment in ADDED-ASSIGNMENTS is a list:

       (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *TRUE*))

   and specifies that assignment proposition (= variable value) has TRUTH-VALUE.

   If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, 
   then assumes that ADDED-ASSIGNMENTS are consistent with the 
   constraints of theory (minus ADDED-CONSTRAINTS).

   If RETURN-CONFLICT? is TRUE, then
   If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is false
   and theory can be proven necessarily true by propagation, 
   then returns a subset of ADDED-ASSIGNMENTS that is
   sufficient to guarantee that theory is satisfied.

   If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, 
   then CONFLICT is a subset of ADDED-ASSIGNMENTS that
   together with THEORY entails ADDED-CONSTRAINTS.

   An assignment is represented internally as a proposition object.  
   The conflict is generated by applying function KEY to each such 
   proposition corresponding to the conflicting assignments,
   and by collecting the non-nil, irredundant values returned by KEY.

   By default, KEY takes PROPOSITION as input and returns assignment
   (= PROPOSITION TRUTH-VALUE) where TRUTH-VALUE is *true* or *false*."

  (debug-print-progn-for :isat
   (debug-print-for :isat t "Testing inconsistency of theory ~A"
				theory)
   (debug-print-for :isat t "    with added assignments ~A"
				added-assignments)
   (debug-print-for :isat t "    and added constraints ~A"
				added-constraints)

   (debug-print-for :isat t "(Re)Initializing the search state of ~A."
				theory))
  
  (clear-search-state! theory)

  ;; Add in the assignments
  (unless (null added-assignments)
	(debug-print-for :isat t "Inserting added assignments ~A."
				 added-assignments)
	(assert-assignments! theory source added-assignments))

  ;; Insert additional constraints
  (unless (null added-constraints)
	(debug-print-for :isat t "Inserting added constraints ~A."
				 added-constraints)
	(loop for constraint in added-constraints
		do (insert-constraint theory constraint)))

  (debug-print-for :isat t
				   "Testing inconsistency using unit propagation, ..." )

  ;; First test inconsistency by unit propagation
  (let ((inconsistent? (not (propagate theory)))
		(conflict nil)
		(new-clauses-from-search nil))

    (if inconsistent?
	    
		(debug-print-for :isat t " ... theory inconsistent by propagation.")

	  ;; Consistent by propagation, 

      (let ((constraints-to-entail
			 (if base-theory-consistent-with-all-assignments?
				 added-constraints
			   (constraint-system-constraints theory))))

		(debug-print-for :isat t " ... theory is consistent by propagation.")
	
		(debug-print-for 
		 :isat t "Check if theory + add-assignments propagation entail ~
                        added-constraints ~A..."
					 constraints-to-entail)

		;; These two loops make sure the effects
		;; of propagation that has just been performed
		;; are cleared.
		;; (This was causing test39 to fail -
		;; ISAT would output a wrong conflict
		;; when manipulating unsat-clauses)
		(loop for constraint in added-constraints
		do (remove-constraint theory constraint))
		(loop for constraint in added-constraints
		do (insert-constraint theory constraint))
		

		(cond ((clauses-are-entailed? constraints-to-entail)
	    
			   ;; If constraints are entailed by propagation, then they 
			   ;; are necessarily satisfied -- UNSAT fails.

			   (debug-print-for
				:isat t " theory + added-assignments entail added-constraints ~a,"
							constraints-to-entail)

			   (debug-print-for 
				:isat t " thus theory + added-assignments + added-constraints ~
                               CAN'T be inconsistent.")
	      
			   (when return-conflict?

				 ;; When looking for a conflict, return a subset of 
				 ;; ADDED-ASSIGNMENTS that entail CONSTRAINTS-TO-ENTAIL.
				 (debug-print-for :isat t "Looking for conflict...")

				 (setq conflict
				   (unit-propagation-implicant-of-clauses
					constraints-to-entail key))
				 
				 (debug-print-for :isat t "Found conflict ~A." conflict)))
			  
			  ;; CASE 2
			  (:otherwise

			   ;; Otherwise, might still be inconsistent, test inconsistency
			   ;; by exhaustive search.

			   (debug-print-for 
				:isat t "Theory is neither propagation inconsistent ~
                               or propagation entailed.")

			   (when perform-search?
				 (debug-print-for 
				  :isat t "Search for a model of theory + added-assigns + added-constraints...")
				 ;; Look for a consistent assignment
				 (multiple-value-bind (success? new-clauses)
					 ;; we need the new clauses so we can undo any clauses
					 ;; introduced by conflict learning
					 (search-forward theory :return-new-clauses? (not save-learned-conflicts?))
				   (setf inconsistent? (not success?))
				   (setf new-clauses-from-search new-clauses))

				 (debug-print-progn-for :isat
				  (if inconsistent?
					  (debug-print-for
					   :isat t " ... found model, consistent.")
					(debug-print-for
					 :isat t " ... no model found, inconsistent."))))))))

    (debug-print t "Restore the state of theory.")

	;; Remove any clauses added by conflict learning
	(unless (null new-clauses-from-search)
	  (debug-print t "Remove conflict based clauses added to theory.")
	  (loop for clause in new-clauses-from-search
		 do (remove-constraint theory clause)) )

	;; Remove the inserted constraints
	(unless (null added-constraints)
	  (debug-print t "Remove constraints added to theory.")
	  (loop for constraint in added-constraints
		  do (remove-constraint theory constraint)))
  
	;; Remove assignments
	(unless (null added-assignments)
	  (debug-print t "4b) Remove assignments added to theory.")
	  (unassert-assignments! theory source added-assignments))

	;; Return the results
    (debug-print t "Return theory + assigns + constraints inconsistent? = ~A"
				 inconsistent?)
	
    (values inconsistent? conflict)))

;;; ***************************************************************************
;;; ***************************************************************************
;;;
;;;                  Supporting Functions
;;;
;;; ***************************************************************************
;;; ***************************************************************************

;;; ***************************************************************************
;;;
;;;                  Support for INCONSISTENT-WITH?
;;;
;;; ***************************************************************************

;;; Optimization:
;;; o We may be able to detect the entailed constraints faster
;;;   by using an idea like watched literals.
(defun clauses-are-entailed?
    (added-clauses)

  "Returns True iff every inserted clause in ADDED-CLAUSES is 
   currently satisfied (removed clauses are ignored)."
  
  ;; All clauses are satisfied.
  (loop for clause in added-clauses
      always (or (constraint-removed? clause)
				 (find-true-literal clause))))

;;; To Do:
;;; o The following routines are not currently used.
;;;   Should they be used somewhere, or thrown away?

;;; Optimization:
;;; o Can this be performed faster using a bit encoding, 
;;;   and by applying bit-vector operations?

;;;(defun assignments-are-entailed?
;;;    (theory added-assignments)
;;;
;;;  "Returns True iff every assignment in ADDED-ASSIGNMENTS 
;;;   has the specified truth value.  
;;;
;;;   Each assignment in ADDED-ASSIGNMENTS is a list:
;;;
;;;       (VARIABLE VALUE &optional (TRUTH-VALUE *TRUE*))
;;;
;;;   and specifies that proposition (= variable value) 
;;;   has TRUTH-VALUE."
;;;  
;;;  ;; All assignments have the specified truth value.
;;;  (loop for (variable-name value . truth-value-list) 
;;;       in added-assignments
;;;      for truth-value = (or (first truth-value-list)
;;;			    *true*)
;;;      for proposition = (get-assignment theory variable-name value)
;;;      always (eql (proposition-truth-value proposition) truth-value)))
;;;

;;; Optimization:
;;; o Can assignments be more efficiently encoded as an array?

;;;(defun assignments-are-inconsistent? (theory added-assignments)
;;;
;;;  "Returns True iff a least on assignment in ADDED-ASSIGNMENTS 
;;;   has the opposite truth value from the specified truth value.  
;;;
;;;   Each assignment in ADDED-ASSIGNMENTS is a list:
;;;
;;;       (VARIABLE VALUE &optional (TRUTH-VALUE *TRUE*))
;;;
;;;   and specifies that proposition (= variable value) has TRUTH-VALUE."
;;;  
;;;  ;; At least one assignment does not have the specified truth value.
;;;  (loop for (variable-name value . truth-value-list) in added-assignments
;;;      for truth-value = (or (first truth-value-list)
;;;			    *true*)
;;;      for proposition = (get-assignment theory variable-name value)
;;;      unless (eql (proposition-truth-value proposition) truth-value)
;;;      return nil
;;;
;;;      finally (return *true*)))
    
;;; Optimization:
;;; o Can find-true-literal be performed more quickly
;;;   by using an idea similar to watched literals.
(defmethod find-true-literal ((clause clause))
  
  " Searches the literals in CLAUSE and returns the proposition
    of a true literal.   If there is no true literal it returns NIL."

  (do-clause-propositions (proposition polarity clause)
  (when (eql (proposition-truth-value proposition) polarity)
    (return-from find-true-literal proposition))))

(defmethod assert-assignments! 
	((theory theory) source assignment-descriptions)
  
  "Asserts every assignment in assignment-descriptions.  Each
   assignment description is a list: 

      (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))

   Denoting assignment (= VARIABLE-NAME VALUE) having TRUTH-VALUE
   in THEORY."

  (loop for (variable-name value . truth-value-list) in assignment-descriptions
      for truth-value = (or (first truth-value-list) *true*)
      do (assign! theory variable-name value truth-value source)))

(defmethod unassert-assignments!
	((theory theory) source assignment-descriptions)

  "Unasserts every assignment in assignment-descriptions.  Each
   assignment description is a list: 

      (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))

   Denoting assignment (= VARIABLE-NAME VALUE) having TRUTH-VALUE
   in THEORY."

  (loop for (variable-name value . truth-value-list) in assignment-descriptions
      for truth-value = (or (first truth-value-list) *true*)
      do (unassign! theory variable-name value truth-value source)))
