;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)

;;; Things to do:

;;; o Review this code and comments.
;;; o Look at interface specification for related, missing functions.

;;; ***********************************************************************
;;;
;;;                ISAT Development Interface
;;;
;;; ***********************************************************************

;;; This Module contains a set of functions to support manipulating a theory,
;;; examining its consequences, and evaluating its performance using a
;;; Lisp interpreter.  This is not considered part of the real-time system.
;;;
;;; This module is comprised of the following external interface functions:
;;;
;;;   THEORY-WITH-CLAUSES (BASE-CLAUSES &OPTIONAL &KEY (NAME 'THEORY))
;;;      Returns a THEORY comprised of BASE-CLAUSES.
;;;      BASE-CLAUSES is a list of clauses, where each clause 
;;;      is a literal, and a literal is of the form P or (not P),
;;;      where P is a symbol denoting a proposition."
;;;
;;;   THEORY-WITH-STATE-CLAUSES (VARIABLE-DOMAINS BASE-CLAUSES
;;;         &optional &key (name 'theory))
;;;      Returns a THEORY comprised of BASE-CLAUSES.
;;;      VARIABLE-DOMAINS is a list of (VARIABLE DOMAIN) pairs,
;;;      where VARIABLE is a symbol and DOMAIN is a list of 
;;;      symbols that variable can take on.
;;;      BASE-CLAUSES is a list of clauses, where each clause 
;;;      is a literal, and a literal is of the form P or (not P),
;;;      where P is a symbol denoting a proposition."
;;;
;;;   FIND-CLAUSE ((THEORY THEORY) SOURCE CLAUSE-DESCRIPTION)
;;;    Takes as input a CLAUSE-DESCRIPTION, and finds the
;;;    corresponding clause object within THEORY.  It makes sure that
;;;    the clause object that is returned has its source set to
;;;    SOURCE.  Correctly treats CLAUSE-DESCRIPTION as a set of literals, and
;;;    hence is able to handle different literal orderings.  If no such
;;;    clause object exists, it returns NIL.  If more than one 
;;;    such clause object exists, it merely returns the first one it finds.  
;;;    It returns either removed or inserted clause objects, without any 
;;;    preference ordering.  The function assumes that clause objects 
;;;    are stored in THEORY without any duplicates, and also assumes that 
;;;    CLAUSE-DESCRIPTION has no duplicate literals.
;;;
;;;   FIND-CLAUSES-FROM-SOURCE ((THEORY THEORY) SOURCE)
;;;     Returns a list of all inserted clauses in THEORY
;;;     whose source is SOURCE.
;;;
;;;   REMOVE-CLAUSES-FROM-SOURCE ((THEORY THEORY) SOURCE)
;;;     Removes all clauses in THEORY whose source is SOURCE.  The
;;;     function returns T.
;;;
;;;   FIND-PROPOSITION-TRUTH-VALUE (NAME &OPTIONAL (THEORY *DEFAULT-THEORY*))
;;;     Returns the truth value of proposition called NAME associated
;;;     with THEORY.  If there is no truth value or if no such proposition
;;;     exists in THEORY (checkable via function above), it returns NIL.
;;;
;;;   THEORY-ASSIGNMENT ((THEORY THEORY))
;;;     Returns the truth assignment for all propositions in theory.
;;;
;;;   CONSISTENT-CLAUSES? (BASE-CLAUSES &OPTIONAL &key
;;;        (NAME 'THEORY) (PERFORM-SEARCH? T) (RETURN-CONFLICT? T)
;;;        (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT))
;;;     Test consistency of BASE-CLAUSES, where BASE-CLAUSES is
;;;     a list of clauses, each clause is a literal, a literal is
;;;     of the form P or (not P), where P is a symbol denoting a
;;;     proposition.  
;;;     Returns (values CONSISTENT? CONFLICT).
;;;     See the function CONSISTENT, in isat.lisp, for a description
;;;     of the remaining arguments and returned values.

;;;   FIND-MODEL-OF-CLAUSES (BASE-CLAUSES &OPTIONAL &KEY (NAME 'THEORY)
;;;      (INITIALIZE? 'T) (RETURN-ASSIGNMENT? 'T) (RETURN-CONFLICT? 'T)
;;;      (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
;;;     Returns a model (i.e., consistent assignment) of BASE-CLAUSES, 
;;;     where BASE-CLAUSES is a list of clauses, each clause 
;;;     is a literal, a literal is of the form P or (not P) where 
;;;     P is a symbol, denoting a proposition.  
;;;     Returns the pair: (FOUND-ASSIGNMENT CONFLICT).
;;;     See the function FIND-CONSISTENT-ASSIGNMENT for a description 
;;;     of the remaining arguments and returned values.
;;;
;;;   CONSISTENT-WITH-CLAUSES? (BASE-CLAUSES &OPTIONAL &KEY 
;;;         (ADDED-CLAUSES NIL) (ADDED-ASSIGNMENTS NIL) 
;;;         (SOURCE NIL) (PERFORM-SEARCH? T) (RETURN-CONFLICT? T)
;;;         (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT))
;;;      Test consistency of BASE-CLAUSES, conjoined with
;;;      ADDED-CLAUSES and ADDED ASSIGNMENTS.
;;;      Returns: (values CONSISTENT? CONFLICT).

;;;      See CONSISTENT-WITH in isat.lisp for a description
;;;      of the remaining arguments and returned values."

;;;   INCONSISTENT-WITH-CLAUSES? (BASE-CLAUSES &OPTIONAL &KEY 
;;;         (ADDED-CLAUSES NIL) (ADDED-ASSIGNMENTS NIL)
;;;         (SOURCE NIL) (PERFORM-SEARCH? T) (RETURN-CONFLICT? T)
;;;         (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
;;;         (BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? T))
;;;      Test the in consistency of BASE-CLAUSES, conjoined with
;;;      ADDED-CLAUSES and ADDED ASSIGNMENTS.
;;;      Returns: (values CONSISTENT? CONFLICT), where
;;;      INCONSISTENT? is false unless theory is proven inconsistent.

;;;      See the function INCONSISTENT-WITH in isat.lisp for a description
;;;      of the remaining arguments and returned values."

;;;   CONSISTENT-STATE-CLAUSES? (VARIABLE-DOMAINS BASE-CLAUSES
;;;         &OPTIONAL &KEY (NAME 'THEORY) (PERFORM-SEARCH? T) 
;;;         (RETURN-CONFLICT? T)
;;;         KEY #' TRUTH-ASSIGNMENT-WHEN-SINGLE-PROPOSITION-CLAUSE))
;;;      Test consistency of BASE-CLAUSES.
;;;      VARIABLE-DOMAINS is a list of (variable domain) pairs,
;;;      where variable is a symbol and domain is a list of 
;;;      symbols that variable can take on.
;;;      BASE-CLAUSES is a list of clauses, where each clause 
;;;      is a literal, a literal is of the form P or (not P),
;;;      and P is a symbol denoting a proposition.  
;;;      Returns: (values CONSISTENT? CONFLICT).
;;;      See the function CONSISTENT in isat.lisp for a description
;;;      of the remaining arguments and returned values.
;;;
;;;   FIND-MODEL-OF-STATE-CLAUSES (VARIABLE-DOMAINS BASE-CLAUSES
;;;         &optional &key (name 'theory) (initialize? 'T)
;;;         (return-conflict? 'T)
;;;         (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT))

;;;      Finds a model (consistent assignment) of BASE-CLAUSES.
;;;      VARIABLE-DOMAINS is a list of (variable domain) pairs,
;;;      where variable is a symbol and domain is a list of 
;;;      symbols that variable can take on.
;;;      BASE-CLAUSES is a list of clauses, where each clause 
;;;      is a literal, and a literal is of the form P or (not P),
;;;      where P is a symbol denoting a proposition.  
;;;      Returns: (Values FOUND-ASSIGNMENT CONFLICT).
;;;      See the function FIND-CONSISTENT-ASSIGNMENT
;;;      for a description of the remaining arguments 
;;;      and returned values.

;;;   CONSISTENT-WITH-STATE-CLAUSES? (VARIABLE-DOMAINS
;;;         BASE-CLAUSES &OPTIONAL &KEY (ADDED-CLAUSES NIL)
;;;         (ADDED-ASSIGNMENTS NIL) (SOURCE NIL)
;;;         (PERFORM-SEARCH? T) (RETURN-CONFLICT? T)
;;;         (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT))
;;;      Test consistency of BASE-CLAUSES, conjoined with
;;;      ADDED-CLAUSES and ADDED ASSIGNMENTS.  All clauses
;;;      are state clauses over domains VARIABLE-DOMAINS.
;;;      Returns: (values CONSISTENT? CONFLICT).

;;;      VARIABLE-DOMAINS is a list of (variable domain) pairs,
;;;      where variable is a symbol and domain is a list of 
;;;      symbols that variable can take on.
;;;      Each clauses is a list of clauses, where each clause 
;;;      is a literal, and a literal is of the form P or (not P),
;;;      where P is a symbol denoting a proposition.  

;;;      See the function CONSISTENT-WITH in isat.lisp for a description
;;;      of the remaining arguments and returned values."

;;;   INCONSISTENT-WITH-STATE-CLAUSES? (VARIABLE-DOMAINS
;;;         BASE-CLAUSES &OPTIONAL &KEY 
;;;         (ADDED-CLAUSES NIL) (ADDED-ASSIGNMENTS NIL)
;;;         (SOURCE NIL) (PERFORM-SEARCH? T) (RETURN-CONFLICT? T)
;;;         (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
;;;         (BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? T))
;;;      Test the in consistency of BASE-CLAUSES, conjoined with
;;;      ADDED-CLAUSES and ADDED ASSIGNMENTS.  All clauses
;;;      are state clauses over domains VARIABLE-DOMAINS.
;;;      Returns: (values CONSISTENT? CONFLICT), where
;;;      INCONSISTENT? is false unless theory is proven inconsistent.

;;;      VARIABLE-DOMAINS is a list of (variable domain) pairs,
;;;      where variable is a symbol and domain is a list of 
;;;      symbols that variable can take on.
;;;      Each clauses is a list of clauses, where each clause 
;;;      is a literal, and a literal is of the form P or (not P),
;;;      where P is a symbol denoting a proposition.  

;;;      See the function INCONSISTENT-WITH in isat.lisp for a description
;;;      of the remaining arguments and returned values."

;;;  LOG-ISAT-STATISTICS ()
;;;     Start logging ISAT statistics (with counters initialized).
;;;
;;;  PRINT-ISAT-STATISTICS ()
;;;     Prints all measures of ISAT performance.
;;;
;;;  ISAT-STATISTICS-OFF ()
;;;     Stop logging statistics for ISAT.
;;;
;;;  ISAT-STATISTICS-ON ()
;;;     Continue logging statistics for ISAT.
;;;
;;;  INITIALIZE-ISAT-STATISTICS ()
;;;     Resets all counters measuring ISAT performance.
;;;
;;;  PRINT-ASSIGNMENT ((THEORY THEORY))
;;;     Prints the truth assignment for all propositions in theory.
;;;
;;;   INITIALIZE-THEORY ((THEORY THEORY))
;;;     Initializes a theory, its propositions, and its
;;;     clauses. This function should rarely be called, but is here
;;;     primarily for debugging purposes.
;;;
;;;   INITIALIZE-INCONSISTENT-CLAUSES
;;;     Initializies the current set of inconsistent clauses of theory.
;;;     Assumes that the SUPPORT status and the
;;;     UNASSIGNED-LITERAL-COUNT of each clause in the THEORY is correctly
;;;     set.  This function is not expected to be used very often.  It does
;;;     not return anything meaningful. 

;;;(eval-when (compile load eval)
;;;  (export '(
;;;
;;;			;; Interface functions
;;;			theory-with-clauses
;;;			theory-with-state-clauses
;;;
;;;			find-proposition
;;;			get-proposition
;;;
;;;			find-clause
;;;			find-clauses-from-source
;;;			remove-clauses-from-source
;;;
;;;			find-proposition-truth-value
;;;			theory-assignment
;;;	    
;;;			consistent-clauses
;;;			find-model-of-clauses
;;;			consistent-with-clauses?
;;;			inconsistent-with-clauses?
;;;	    
;;;			consistent-state-clauses
;;;			find-model-of-state-clauses
;;;			consistent-with-state-clauses?
;;;			inconsistent-with-state-clauses?
;;;	    
;;;			isat-statistics-on
;;;            log-isat-statistics
;;;			initialize-isat-statistics
;;;            print-isat-statistics
;;;			isat-statistics-off
;;;			print-assignment
;;;	    
;;;			initialize-theory
;;;			initialize-inconsistent-clauses ; (should this be external?)
;;;			)
;;;		  :ISAT)
;;;  )

;;; ***********************************************************************
;;;
;;;                  External Interface
;;;
;;; ***********************************************************************

;;; ***********************************************************************
;;;
;;;                  Building Theories
;;;
;;; ***********************************************************************

(defun theory-with-clauses (base-clauses &optional &key (name 'theory))
  "  Returns a THEORY comprised of BASE-CLAUSES.
     BASE-CLAUSES is a list of clauses, where each clause 
     is a literal, and a literal is of the form P or (not P),
     where P is a symbol denoting a proposition."

  (let ((theory (make-instance 'theory :name name)))
	(dolist (description base-clauses)
	  (insert-new-clause theory :source name :clause description))
    theory))

(defun theory-with-state-clauses
    (variable-domains base-clauses
     &optional &key (name 'theory))

  "  Returns a THEORY comprised of BASE-CLAUSES.
     VARIABLE-DOMAINS is a list of (VARIABLE DOMAIN) pairs,
     where VARIABLE is a symbol and DOMAIN is a list of 
     symbols that variable can take on.
     BASE-CLAUSES is a list of clauses, where each clause 
     is a literal, and a literal is of the form P or (not P),
     where P is a symbol denoting a proposition."

  (let ((theory (make-instance 'theory :name name)))

    (loop for (variable domain) in variable-domains
		do (define-state-variable theory name variable domain))

	(dolist (description base-clauses)
	  (insert-new-state-clause theory :source name :clause description))

    theory))

;;; ***********************************************************************
;;;
;;;                  Examining Theories
;;;
;;; ***********************************************************************

(defmethod find-clause
    ((theory theory) source clause-description)
  
  " Takes as input a CLAUSE-DESCRIPTION, and finds the
    corresponding clause object within THEORY.  It makes sure that
    the clause object that is returned has its source set to
    SOURCE.  Correctly treats CLAUSE-DESCRIPTION as a set of literals, and
    hence is able to handle different literal orderings.  If no such
    clause object exists, it returns NIL.  If more than one 
    such clause object exists, it merely returns the first one it finds.  
    It returns either removed or inserted clause objects, without any 
    preference ordering.  The function assumes that clause objects 
    are stored in THEORY without any duplicates, and also assumes that 
    CLAUSE-DESCRIPTION has no duplicate literals."
  
  (assert (no-duplicate-literals clause-description))
  (let ((propn nil))
    (multiple-value-bind 
		(proposition positivep)
		(literal-description-proposition (first clause-description))
      (when proposition
		;; Proceed only when there is a proposition.  If there is no
		;; PROPOSITION here, then it means the clause-description is empty.
		(setq propn (find-proposition theory proposition))
		(when propn
		  ;; Proceed only when there is a proposition.  If there isn't a
		  ;; PROPN here, it means that the clause-description couldn't 
		  ;; possibly be in the THEORY.
		  (if positivep
			  ;; then look only in clause objects in which it occurs positively
			  (find-clause-in-clauses 
			   clause-description
			   (proposition-positive-clauses propn)
			   source
			   theory)
			;; else look only in clause objects in which it occurs negatively
			(find-clause-in-clauses 
			 clause-description
			 (proposition-negative-clauses propn)
			 source
			 theory)))))))

(defmethod find-clauses-from-source ((theory theory) source)
  
  " Returns a list of all inserted clauses in THEORY whose source is SOURCE."
  
  (let ((result nil))
    (do-inserted-clauses (clause (constraint-system-constraints theory))
	                 (when (equal (constraint-source clause) source)
	                   (push clause result)))
    (values result)))

(defmethod remove-clauses-from-source ((theory theory) source)
  
  " Removes all clauses in THEORY whose source is SOURCE.  Returns T."

  (do-inserted-clauses (clause (constraint-system-constraints theory))
	(when (equal (constraint-source clause) source)
	  (remove-constraint theory clause)))
  (values t))

(defmethod find-proposition-truth-value ((theory theory) pname)
  (let ((propn (find-proposition theory pname)))
    (if propn
		(proposition-truth-value propn))))


(defmethod theory-assignment ((theory theory))

  "Returns the truth assignment for all propositions in theory."
  
  (loop for proposition in (theory-propositions theory)
      for truth-value = (proposition-truth-value proposition)
      unless (eql truth-value *none*)
      collecting (list proposition
		       (proposition-truth-value proposition))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                  Testing Theories
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun consistent-clauses?
    (base-clauses
     &optional 
     &key (name 'theory) (perform-search? T) (return-conflict? T)
     (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
	 ;;			    (additional-clauses '())
	 ;;			    (violated-clauses '())
     )

  "Test consistency of BASE-CLAUSES, where BASE-CLAUSES is
     a list of clauses, each clause is a literal, a literal is
     of the form P or (not P) where P is a symbol, denoting a
     proposition.  
     Returns: (values CONSISTENT? CONFLICT).

     See the function CONSISTENT in isat.lisp for a description
     of the remaining arguments and returned values."

  (consistent? (theory-with-clauses base-clauses :name name)
			   :perform-search? perform-search?
			   :return-conflict? return-conflict?
			   :key key))

(defun find-model-of-clauses
    (base-clauses
     &optional 
     &key (name 'theory)
     (initialize? 'T)
     (return-conflict? 'T)
     (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT))

  "Finds a model (consistent assignment) of BASE-CLAUSES,
   where BASE-CLAUSES is a list of clauses, each clause 
   is a literal, a literal is of the form P or (not P),
   where P is a symbol denoting a proposition.  
   Returns: (Values FOUND-ASSIGNMENT CONFLICT).

   See the function FIND-CONSISTENT-ASSIGNMENT
   for a description of the remaining arguments 
   and returned values."

  (find-consistent-assignment 
   (theory-with-clauses base-clauses :name name)
   initialize?
   'return-assignment!
   return-conflict?
   key))

(defun consistent-with-clauses?
    (base-clauses
     &optional 
     &key 
     (added-clauses nil)
     (added-assignments nil)
     (source nil)
     (perform-search? t)
     (return-conflict? t)
     (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT))
  
  "Test consistency of BASE-CLAUSES, conjoined with
     ADDED-CLAUSES and ADDED ASSIGNMENTS.
     Returns: (values CONSISTENT? CONFLICT).

     See the function CONSISTENT-WITH in isat.lisp for a description
     of the remaining arguments and returned values."

  (let* ((theory (theory-with-clauses
		   base-clauses :name source))
	 (intern-added-clauses
	  (loop for c in added-clauses
	      collecting (make-instance 'clause 
					   :constraint-system theory
					   :source source
					   :clause c))))
    (consistent-with? theory
		      :added-constraints intern-added-clauses
		      :added-assignments added-assignments
		      :source source
		      :perform-search? perform-search?
		      :return-conflict? return-conflict?
		      :key key)))

(defun inconsistent-with-clauses?
    (base-clauses
     &optional
     &KEY 
     (added-clauses nil)
     (added-assignments nil)
     (source nil)
     (perform-search? T)
     (return-conflict? T)
     (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
     (base-theory-consistent-with-all-assignments? T))
  
  "Test the in consistency of BASE-CLAUSES, conjoined with
   ADDED-CLAUSES and ADDED ASSIGNMENTS.  All clauses
   are state clauses over domains variable-domains.

   Returns: (values CONSISTENT? CONFLICT), where
   INCONSISTENT? is false unless theory is proven inconsistent.

   VARIABLE-DOMAINS is a list of (variable domain) pairs,
   where variable is a symbol and domain is a list of 
   symbols that variable can take on.
   Each clauses is a list of clauses, where each clause 
   is a literal, and a literal is of the form P or (not P),
   where P is a symbol denoting a proposition.  

   See the function INCONSISTENT-WITH in isat.lisp for a description
   of the remaining arguments and returned values."

  (let* ((theory (theory-with-clauses base-clauses :name source))
	 (intern-added-clauses
	  (loop for c in added-clauses
	      collecting (make-instance 'clause 
					   :constraint-system theory
					   :source source
					   :clause c))))
    (inconsistent-with? theory
			:added-constraints intern-added-clauses
			:added-assignments added-assignments
			:source source
			:perform-search? perform-search?
			:return-conflict? return-conflict?
			:key key
			:base-theory-consistent-with-all-assignments?
			base-theory-consistent-with-all-assignments?
			)))

(defun consistent-state-clauses?
    (variable-domains
     base-clauses
     &optional 
     &key (name 'theory) (perform-search? T) (return-conflict? T)
     (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
	 ;;			    (additional-clauses '())
	 ;;			    (violated-clauses '())
     )
  "  Test consistency of BASE-CLAUSES.
     VARIABLE-DOMAINS is a list of (variable domain) pairs,
     where variable is a symbol and domain is a list of 
     symbols that variable can take on.
     BASE-CLAUSES is a list of clauses, where each clause 
     is a literal, a literal is of the form P or (not P),
     and P is a symbol denoting a proposition.  
     Returns: (values CONSISTENT? CONFLICT).

     See the function CONSISTENT in isat.lisp for a description
     of the remaining arguments and returned values."

  (consistent? 
   (theory-with-state-clauses variable-domains  base-clauses :name name)
   :perform-search? perform-search?
   :return-conflict? return-conflict?
   :key key))

(defun find-model-of-state-clauses
    (variable-domains
     base-clauses
     &optional 
     &key
     (name 'theory)
     (initialize? 'T)
     (return-conflict? 'T)
     (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT))

  "Finds a model (consistent assignment) of BASE-CLAUSES.
   VARIABLE-DOMAINS is a list of (variable domain) pairs,
   where variable is a symbol and domain is a list of 
   symbols that variable can take on.
   BASE-CLAUSES is a list of clauses, where each clause 
   is a literal, and a literal is of the form P or (not P),
   where P is a symbol denoting a proposition.  
   Returns: (Values FOUND-ASSIGNMENT CONFLICT).

   See the function FIND-CONSISTENT-ASSIGNMENT
   for a description of the remaining arguments 
   and returned values."

  (find-consistent-assignment
   (theory-with-state-clauses variable-domains  base-clauses :name name)
   initialize?
   'return-assignment!
   return-conflict?
   key))

(defun consistent-with-state-clauses?
    (variable-domains
     base-clauses
     &optional 
     &key 
     (added-clauses nil)
     (added-assignments nil)
     (source nil)
     (perform-search? t)
     (return-conflict? t)
     (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT))
  
  "Test consistency of BASE-CLAUSES, conjoined with
   ADDED-CLAUSES and ADDED ASSIGNMENTS.  All clauses
   are state clauses over domains variable-domains.
   Returns: (values CONSISTENT? CONFLICT).

   VARIABLE-DOMAINS is a list of (variable domain) pairs,
   where variable is a symbol and domain is a list of 
   symbols that variable can take on.
   Each clauses is a list of clauses, where each clause 
   is a literal, and a literal is of the form P or (not P),
   where P is a symbol denoting a proposition.  

   See the function CONSISTENT-WITH in isat.lisp for a description
   of the remaining arguments and returned values."
  
  (let* ((theory (theory-with-state-clauses
		  variable-domains base-clauses :name source))
	 (intern-added-clauses
	  (loop for c in added-clauses
	      collecting (make-instance 'clause 
					   :constraint-system theory
					   :source source
					   :clause c))))
    (consistent-with? theory
		      :added-constraints intern-added-clauses
		      :added-assignments added-assignments
		      :source source
		      :perform-search? perform-search?
		      :return-conflict? return-conflict?
		      :key key)))

(defun inconsistent-with-state-clauses?
    (variable-domains
     base-clauses
     &optional 
     &key 
     (added-clauses nil)
     (added-assignments nil) 
     (source nil)
     (perform-search? t)
     (return-conflict? t)
     (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
     (base-theory-consistent-with-all-assignments? T))
  
  "Test the in consistency of BASE-CLAUSES, conjoined with
   ADDED-CLAUSES and ADDED ASSIGNMENTS.  All clauses
   are state clauses over domains variable-domains.

   Returns: (values CONSISTENT? CONFLICT), where
   INCONSISTENT? is false unless theory is proven inconsistent.

   VARIABLE-DOMAINS is a list of (variable domain) pairs,
   where variable is a symbol and domain is a list of 
   symbols that variable can take on.
   Each clauses is a list of clauses, where each clause 
   is a literal, and a literal is of the form P or (not P),
   where P is a symbol denoting a proposition.  

   See the function INCONSISTENT-WITH in isat.lisp for a description
   of the remaining arguments and returned values."

  (let* ((theory (theory-with-state-clauses
		  variable-domains base-clauses :name source))
	 (intern-added-clauses
	  (loop for c in added-clauses
	      collecting (make-instance 'clause 
					   :constraint-system theory
					   :source source
					   :clause c))))
    (inconsistent-with? theory
			:added-constraints intern-added-clauses
		        :added-assignments added-assignments
			:source source
			:perform-search? perform-search?
			:return-conflict? return-conflict?
			:key key
			:base-theory-consistent-with-all-assignments?
			base-theory-consistent-with-all-assignments?)))

;;; ***********************************************************************
;;;
;;;                  Initializing Theories for Search and Inference
;;;
;;; ***********************************************************************

(defmethod initialize-theory ((theory theory))
  
  " Sets up a theory, its propositions, and its clauses
    for unit propagation.  PROPAGATE must then be explicitly called to
    perform the unit propagation.

    This function should rarely be called, but is here primarily for
    debugging purposes."
  
  ;; Initialize all the propositions
  (do-ht (name propn (theory-proposition-ht theory))
    (declare (ignore name))
    (setf (proposition-truth-value propn) *none*)
    )
  ;; Initialize all the clauses
  (do-inserted-clauses (cl (constraint-system-constraints theory))
    ;; in an initialized theory, all clause literals are potentially true.
    ;; Note that this assumes that there are no duplicate literals.
    (setf (clause-unassigned-literal-count cl)
          (+ (length (clause-positive-propositions cl))
	     (length (clause-negative-propositions cl))))
    ;; The clause does not have any consequences and does not currently
    ;; have a support
    (let ((p (clause-consequent cl)))
      (unless (null p)
	(clear-support-relation! p cl)))
    ;; Clause no longer has a proposition satisfying it.
    (setf (clause-satisfying-proposition cl) nil)
    ;; The clause is not in a conflict
    (setf (clause-in-conflict? cl) nil)    
    ;; The in-theory-unit-clause? and inconsistent? flags are updated later.
    )
  ;; Initialize the unit-clauses and the set of inconsistent? clauses
  (initialize-unit-clauses theory)
  (initialize-inconsistent-clauses theory)
  )

(defmethod initialize-inconsistent-clauses ((theory theory))
  
  " Initializes the current set of inconsistent clauses of THEORY.
    Assumes that the CONSEQUENT, SATISFYING-PROPOSITION and 
    UNASSIGNED-LITERAL-COUNT of each clause in THEORY is correctly
    set.  

    This function is not expected to be used very often. "
  
  (lhs-clear (theory-inconsistent-clauses theory))
  (do-inserted-clauses (clause (constraint-system-constraints theory))
    (if (inconsistent-clause? clause)
	(progn
	  (lhs-add (theory-inconsistent-clauses theory) clause)
	  (setf (clause-inconsistent? clause) t))
      ;; Else it isn't satisfiable
      (setf (clause-inconsistent? clause) nil))))

(defmethod initialize-unit-clauses ((theory theory))
  
  " Initializes the unit clauses of theory involves going through the theory
    and queuing all unit clauses.  Assumes that the CONSEQUENT
    SATISFYING-PROPOSITION and UNASSIGNED-LITERAL-COUNT of each 
    clause in THEORY are correctly set.  

    This function is not expected to be used very often."
    
  (lhs-clear (theory-unit-clauses theory))
  (do-inserted-clauses (cl (constraint-system-constraints theory))
    (if (unit-clause? cl)
		;; its a unit clause, so add it to the unit clauses of theory
		(progn
		  (lhs-add (theory-unit-clauses theory) cl)
		  (setf (clause-in-theory-unit-clause? cl) t))
	  ;; else it isn't a unit clause, so clear its in-theory-unit-clause? flag
      (setf (clause-in-theory-unit-clause? cl) nil))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                  Maintaining Statistics
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
(defun log-ISAT-statistics ()
  "Start logging ISAT statistics 
  (with counters initialized)"

  (isat-statistics-on)
  (initialize-isat-statistics))

(defun ISAT-statistics-on ()
  "Continue logging statistics for ISAT"
  (setq *isat-statistics-p* 't))

(defun ISAT-statistics-off ()
  "Stop logging statistics for ISAT"
  (setq *isat-statistics-p* nil))

(defun initialize-ISAT-statistics ()
  
  "Resets all counters measuring ISAT performance to 0."
  
  (setq *propagate-counter* 0)
  (setq *update-proposition-clause-counter* 0)
  (setq *remove-clause-counter* 0)
  (setq *update-clause-for-unassigned-proposition-counter* 0)
  (setq *assign-truth-value-counter* 0)
  (setq *unassign-truth-value-counter* 0)
  )

(defun stop-and-print-ISAT-statistics ()
  (ISAT-statistics-off)
  (print-ISAT-statistics))

(defun print-ISAT-statistics ()

  "Prints all tracked ISAT statistics"

  (format t "~%Propagate counter ~a." *propagate-counter*)
  (format t "~%Update-proposition-clause ~a." 
		  *update-proposition-clause-counter*)
  (format t "~%Remove-clause ~a."  *remove-clause-counter*)
  (format t "~%Update-clause-for-unassigned-proposition ~a." 
	      *update-clause-for-unassigned-proposition-counter*)
  (format t "~%Assign-truth-value ~a." *assign-truth-value-counter*)
  (format t "~%Unassign-truth-value ~a." *unassign-truth-value-counter*)
  )

(defmethod print-assignment ((theory theory))

  "Prints the truth assignment for all propositions in theory."
  
  (loop for proposition in (theory-propositions theory)
      for truth-value = (proposition-truth-value proposition)
      unless (eql truth-value *none*)
      do (format t "~% ~A = ~A" 
				 proposition
				 (proposition-truth-value proposition)))
  (print 'Done))

;;; ***********************************************************************
;;;
;;;                  Support for Find-clauses
;;;
;;; ***********************************************************************

(defmethod no-duplicate-literals ((clause clause))
  
  " Returns T if and only if CLAUSE, a list of literals, has
    no duplicate literals."
  
  (do ((remaining-clause clause (cdr remaining-clause)))
      ((null (cdr remaining-clause))
	   ;; then there are no duplicate literals
       (values t))
    (when (member (car remaining-clause)
				  (cdr remaining-clause)
				  :test #'equal)
	  ;; Then there is a duplicate literal
      (return-from no-duplicate-literals 
		(values nil)))))

(defmethod find-clause-in-clauses 
	(clause-description clauses source (theory theory))
  
  " Finds the clause called CLAUSE-DESCRIPTION in CLAUSES.  If the
    clause exists it returns it, else it returns NIL.  
    Assumes that the first literal in CLAUSE-DESCRIPTION already exists in
    CLAUSES."
  
  (dolist (clause clauses)
	;; Check that CLAUSE has the right SOURCE and the right description
    (when (and (equal source (constraint-source clause))
			   (same-clause-description clause-description clause theory))
	  ;; then we're done
      (return-from find-clause-in-clauses
		(values clause))))
  ;; If we come here, then no clause has the right set of literals
  (values nil))

(defmethod same-clause-description 
	(clause-description (clause clause) (theory theory))
  
  " Checks to see if CLAUSE has the same literals as CLAUSE-DESCRIPTION.
	
    First check that the lengths are correct.  If the lengths are
    equal, and there are no duplicates, then it suffices to check that
    each literal in CLAUSE-DESCRIPTION occurs in CLAUSE."

  (when (= (length clause-description)
		   (+ (length (clause-positive-propositions clause))
			  (length (clause-negative-propositions clause))))
	;; If the lengths are not the same, do nothing.
	;; Otherwise, check to see if all literals in CLAUSE-DESCRIPTION are in
	;; CLAUSE.  Note that we could have started with the second literal in
	;; CLAUSE-DESCRIPTION, since the first literal has already known to be in
	;; CLAUSE.  However, we will incur the small overhead of rechecking the
	;; first literal to make this function be self contained.
    (dolist (literal clause-description)
      (multiple-value-bind 
		  (proposition positive?)
		  (literal-description-proposition literal)
		(let ((propn (find-proposition theory proposition)))
		  (unless 
			  (and propn
				   (if positive?
					   (member propn (clause-positive-propositions clause))
					 (member propn (clause-negative-propositions clause))))
			;; the proposition isn't in the clause, so return NIL
			(return-from same-clause-description (values nil))))))
	;; Then all literals are in CLAUSE, so return T
    (values t)))