
(in-package :isat)

(defstruct (linked-hashset-node (:conc-name lhs-node-))  
  (element nil)
  (previous nil)
  (next nil))

(defstruct (linked-hashset (:conc-name lhs-))
  (table nil :type hash-table)
  (first nil :type linked-hashset-node)
  (last nil :type linked-hashset-node))

(defun make-lhashset (&key (test 'eql) (size 67)
				 (rehash-size 1.2) (rehash-threshold 0.6407767)
				 (hash-function nil) (values t)
				 (weak-keys nil))
  (make-linked-hashset
   :table
   (make-hash-table :test test :size size :rehash-size rehash-size
		    :rehash-threshold rehash-threshold
		    :hash-function hash-function :values values
		    :weak-keys weak-keys)
   :first nil
   :last nil))


(defun lhs-add (lhashset key)
  "Returns true if the set did not originally
contain the key and nil if it did. i.e. Returns
true if a change was made to the set"
  (multiple-value-bind (node there?)
	  (gethash key (lhs-table lhashset) nil)
	(cond 
	  ;; if the key already exists, do nothing.
	  (there? nil)
	  ;; if the key does not exist, add it.
	  (:otherwise 
	   (setf node (make-linked-hashset-node
			  :element key
			  :previous (lhs-last lhashset)
			  :next nil))
	   (setf (gethash key (lhs-table lhashset)) node)
	   (if (equal (lhs-size lhashset) 1)
		   (setf (lhs-first lhashset) node)
		   (setf (lhs-last lhashset) node))
	   (if (not (null (lhs-node-previous node)))
		   (setf (lhs-node-next (lhs-node-previous node)) node))
	   (setf (lhs-last lhashset) node)
	   t)
	  )))


(defun lhs-remove (lhashset key)
  "Returns the item if it exists, or nil
if it doesn't."
  (multiple-value-bind (node there?)
	  (gethash key (lhs-table lhashset) nil)
	(cond ((not there?) nil) 
		  (:otherwise
		   (let ((element nil))
			 ;; save off the element
			 (setq element (lhs-node-element node))
			 ;; delete it from the list
			 (remhash key (lhs-table lhashset))
			 ;; update the previous pointer
			 (cond ((null (lhs-node-previous node))
					;; it's the first node
					(setf (lhs-first lhashset) 
						  (lhs-node-next node)))
				   (:otherwise
					;; update previous' next pointers
					(setf (lhs-node-next (lhs-node-previous node)) 
						  (lhs-node-next node))))
			 ;; update the next pointer
			 (cond ((null (lhs-node-next node))
					;; its the last node
					(setf (lhs-last lhashset) 
						  (lhs-node-previous node)))
				   (:otherwise
					(setf (lhs-node-previous (lhs-node-next node)) 
						  (lhs-node-previous node))))
			 element))
		  )))

(defun lhs-remove-first (lhashset)
  (let ((first (lhs-first lhashset)))
	(if (null first)
		nil
		(lhs-remove lhashset (lhs-node-element first)))
))

(defun lhs-remove-last (lhashset)
  (let ((last (lhs-last lhashset)))
	(if (null last)
		nil
		(lhs-remove lhashset (lhs-node-element last)))
))

(defun lhs-get-first (lhashset)
  (let ((first (lhs-first lhashset)))
	(if (null first)
		nil
		(lhs-node-element first))))

(defun lhs-get-last (lhashset)
  (let ((last (lhs-last lhashset)))
	(if (null last)
		nil
		(lhs-node-element last))))

(defun lhs-clear (lhashset) 
  "Clears the hash set of all contents"
  ;; clear the table
  (clrhash (lhs-table lhashset))
  ;; reset the pointers
  (setf (lhs-first lhashset) nil)
  (setf (lhs-last lhashset) nil)
)

(defun lhs-contains (lhashset key)
  "Returns true if the set contains the key
and nil if it does not."
  (multiple-value-bind (orig-value there?)
	  (gethash key (lhs-table lhashset))
	(declare (ignore orig-value))
	there?))

(defun lhs-size (lhashset) 
  "Returns the size of the hash table"
  (hash-table-count (lhs-table lhashset)))

(defun lhs-empty? (lhashset)
  (equal (lhs-size lhashset) 0))

(defmethod print-object ((obj linked-hashset) stream) 
  (let ((node (lhs-first obj))
		(first-loop t))
	(loop 
	   (if (null node) (return))
	   (cond (first-loop
			  (setf first-loop nil)
			  (format stream "~a" (lhs-node-element node)))
			 (:otherwise 
			  (format stream "~%~a" (lhs-node-element node))))
	   (setf node (lhs-node-next node)))
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ITERATOR

(defstruct (linked-hashset-iterator (:conc-name lhs-it-))
  (current nil :type linked-hashset-node)
  (forward-direction t))


(defun make-lhashset-iterator (lhashset forward-direction)
  (make-linked-hashset-iterator
   :current
   (if forward-direction
	   (lhs-first lhashset)
	   (lhs-last lhashset))
   :forward-direction
   forward-direction
))

(defun lhs-it-next (lhashset-iterator)
  (let* ((current (lhs-it-current lhashset-iterator))
		 (element (lhs-node-element current)))
	(setf (lhs-it-current lhashset-iterator) 
		  (if (lhs-it-forward-direction lhashset-iterator)
			  (lhs-node-next current)
			  (lhs-node-previous current)))
	element))

(defun lhs-it-hasnext (lhashset-iterator)
  (not (null (lhs-it-current lhashset-iterator))))