;;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)

;;; ***************************************************************************
;;;
;;;                   Variables, Macros and Procedures to
;;;                        to Support Model Finding
;;;
;;; ***************************************************************************


;;;(eval-when (compile load eval)
;;;  (export '(conflict-learning-on
;;;			conflict-learning-off
;;;			dynamic-backtracking-on
;;;			dynamic-backtracking-off
;;;			random-assignment-on
;;;			random-assignment-off
;;;			)
;;;		  :ISAT))

;;;   (CONFLICT-LEARNING-ON)
;;;   (CONFLICT-LEARNING-OFF)

;;;    Turns ON/OFF conflict-learning within ISAT.
;;;    Conflicts are partial inconsistent assignments,
;;;    and are used to aid forward and backward search.

;;;   (DYNAMIC-BACKTRACKING-ON)
;;;   (DYNAMIC-BACKTRACKING-OFF)

;;;    Turns ON/OFF dynamic-backtracking within ISAT.
;;;    During backtracking, ISAT then avoids undoing 
;;;    assignments that are not mentioned in an active 
;;;    conflict.  

;;;    Dynamic backtracking, requires conflict-learning,
;;;    hence conflict-learning is on as well as long
;;;    as dynamic-backtracking is on.  Turning off
;;;    dynamic backtracking, however, does not turn
;;;    off conflict learning.

;;;   (RANDOM-ASSIGNMENT-ON)
;;;   (RANDOM-ASSIGNMENT-OFF)

;;;    Tells ISAT to TURN ON/OFF randomized assignment.
;;;    That is, for each proposition, ISAT assign its 
;;;    truth values in a random order.

;;; ***************************************************************************
;;;
;;;                           External Interface
;;;
;;; ***************************************************************************

;;; ***************************************************************************
;;;
;;;                  Flags for Controlling Search
;;;
;;; ***************************************************************************

(defvar *conflict-learning-on* t
  "When set to true, ISAT learns conflicts, which are
   are partial inconsistent assignments.  Conflicts are
   used during forward unit propagation and during 
   conflict-directed backtracking.")

(defun conflict-learning-on ()

  "Turns ON conflict-learning within ISAT.
   Conflicts are partial inconsistent assignments,
   and are used to aid forward and backward search."
  
  (setq *conflict-learning-on* t))

(defun conflict-learning-off ()
  
  "Turns OFF conflict-learning within ISAT.
   Conflicts are partial inconsistent assignments,
   and are used to aid forward and backward search."
  
  (setq *conflict-learning-on* nil))

(defvar *dynamic-backtracking-on* t

  "If set to True, then during backtracking,
   ISAT avoids undoing assignments that are not 
   mentioned in an active conflict.  

   Dynamic backtracking, requires conflict-learning,
   hence conflict-learning is on as well as long
   as dynamic-backtracking is on.  Turning off
   dynamic backtracking, however, does not turn
   off conflict learning.")

(defun dynamic-backtracking-on ()

  "Turns ON dynamic-backtracking within ISAT.
   During backtracking, ISAT then avoids undoing 
   assignments that are not mentioned in an active conflict.  
   Dynamic backtracking can only be performed if 

   Dynamic backtracking, requires conflict-learning,
   hence conflict-learning is ON as well as long
   as dynamic-backtracking is ON"
  
  (setq *dynamic-backtracking-on* t)
  (setq *conflict-learning-on* t))

(defun dynamic-backtracking-off ()

  "Turns OFF dynamic-backtracking within ISAT.
   During backtracking, ISAT then avoids undoing 
   assignments that are not mentioned in an active conflict.  

   Dynamic backtracking, requires conflict-learning,
   hence conflict-learning is ON as well as long
   as dynamic-backtracking is ON.  Turning off
   dynamic backtracking, however, does not turn
   off conflict learning."
  
  (setq *dynamic-backtracking-on* nil)
  (setq *conflict-learning-on* nil))

(defmacro perform-chronological-backtracking? ()

  "If TRUE then ISAT backs up through solutions chronologically."
  
  `(not (or *dynamic-backtracking-on*
			*conflict-learning-on*)))

(defmacro perform-conflict-learning? ()

  "If TRUE then ISAT extracts conflicts during search
   and adds them to theory."

  `(or *dynamic-backtracking-on*
       *conflict-learning-on*))

(defmacro perform-conflict-directed-backtracking? ()

  "If TRUE then ISAT backs up to an assignment that 
   contributes to the current inconsistency."

  `(or *conflict-learning-on*
       *dynamic-backtracking-on*))

(defmacro perform-dynamic-backtracking? ()

  "If TRUE then during backup, ISAT only retracts
   assignments that will resolve the current inconsistency."

  `(not (null *dynamic-backtracking-on*)))
  
(defvar *random-assignment-on* nil 

  "If TRUE then for each proposition, ISAT assigns 
   its truth values in a random order.")

(defun random-assignment-on ()

  "Tells ISAT to TURN ON randomized assignment.  That is,
   for each proposition, ISAT assign its truth values 
   in a random order."

  (setq *random-assignment-on* t))

(defun random-assignment-off ()

  "Tells ISAT to TURN OFF randomized assignment.  That is,
   for each proposition, ISAT assign its truth values 
   in a consistent order - True followed by False"

  (setq *random-assignment-on* nil))

(defmacro perform-random-assignment? ()

  "If True is returned, then ISAT Assigns truth values 
   to the current proposition in a random order."

  `(not (null *random-assignment-on*)))

;;; ***************************************************************************
;;;
;;; *** Module Support Functions
;;;
;;; ***************************************************************************

;;; ***************************************************************************
;;;
;;; *** Managing the Queue of Propositions to be Assigned
;;;
;;; ***************************************************************************

;;; Optimization:
;;; o push/pop creates wasted conses and should be replaced.
;;;   
;;;   - If we implement theory-propositions as a growable array,
;;;     then this array can also be used as the queue of assigned 
;;;     and unassigned, by moving index up and down.

;;; o Copylist isn't required here, since push and pop are 
;;;   non-destructive, correct?
(defmacro clear-assignment-queue! (theory)
  
  "Clears the queue of propositions to be assigned
   by ISAT for THEORY."
  `(progn
	 ; clear assigned propositions
     (setf (theory-assigned-propositions ,theory) nil)
	 ; copy unassigned propositions
	 (let ((unassigned (theory-unassigned-propositions , theory)))
	   (heap-clear unassigned)
	   (loop for prop in (theory-propositions ,theory) do
			(if (proposition-searchable prop) 
				(heap-insert unassigned prop)))
	   )
))

(defmacro new-proposition-to-assign (theory)

  "Returns a proposition in THEORY that has no 
   assigned truth value, or NIL if none exists."

  `(heap-peek (theory-unassigned-propositions ,theory)))

(defmacro last-assigned-proposition (theory)

  "Returns the proposition in THEORY that has most 
   recently been assigned a truth value."

  `(first (theory-assigned-propositions ,theory)))

;;; Optimization:
;;; o The following two functions should reuse the 
;;;   conses so that they don't create garbage 
;;;   when moving between lists.  

;;;   - Alternatively this could be implemented as 
;;;     an array or doubly linked list.

;;; o Many of the following functions should be 
;;;   implemented as the equivalent to Zetalisp's defsubst, 
;;;   not defmacros.  How do we get the effect of defsubst
;;;   in commonlisp?  Through an inline declaration?
(defmacro push-new-proposition-to-assigned (theory)

  "Moves the proposition in THEORY at the top of the 
   UNASSIGNED proposition list to the top of the
   list of ASSIGNED propositions."

  `(push (heap-remove (theory-unassigned-propositions ,theory))
         (theory-assigned-propositions ,theory)))

(defmacro pop-last-proposition-from-assigned (theory)

  "Moves the proposition in THEORY at the top of the 
   ASSIGNED proposition list to the top of the
   list of UNASSIGNED propositions."

  `(heap-insert (theory-unassigned-propositions ,theory)
				(pop (theory-assigned-propositions ,theory))))
  
(defmacro all-propositions-assigned? (theory)
  
  "Returns True IFF all propositions of THEORY 
   are assigned truth values."
  
  `(heap-empty-p (theory-unassigned-propositions ,theory)))

(defmacro has-assigned-propositions? (theory)

  "Returns True IFF at least one proposition in THEOORY
   is assigned a truth value."

  `(not (null (theory-assigned-propositions ,theory))))

(defmacro all-propositions-unassigned? (theory)
  
  "Returns True IFF no proposition within THEORY 
   has an assigned truth value."
  
  `(null (theory-assigned-propositions ,theory)))

;;; Optimization:
;;; o Its more efficent to encode theory proposition 
;;;   assignments using an array or bit-vector.
(defmethod current-assignment ((theory theory))
  
  "Returns a description of the current (partial) assignment
   to THEORY.  Description is a list of proposition/truth-value
   pairs, one for each assigned proposition."
  
  (loop for proposition in (theory-assigned-propositions theory)
      for truth-value = (proposition-truth-value proposition)
      collecting `(,proposition ,truth-value)))

;;; Added
;;; o Returns a theory-model object, which separates true-propositions
;;; from false-propositions. (This is convenient if, for example,
;;; ISAT is being used as a CSP solver. In such situations we
;;; should be interested in accessing just the true-propositions
;;; in a  model.

(defmethod current-model ((theory theory))
  "RETURNS a theory-model object which basically separates true-propositions
   from false ones."

  (loop for proposition in (theory-assigned-propositions theory)
			   
      for truth-value = (proposition-truth-value proposition)

      if (true-value? truth-value)
	collecting `(,proposition ,truth-value) into true-propositions
      else
	collecting `(,proposition ,truth-value) into false-propositions
						     
      finally (return (make-instance 'theory-model
			:true-propositions true-propositions
			:false-propositions false-propositions))))
  

;;; ***************************************************************************
;;;
;;; ***   Proposition Truth Assignment and Support
;;;
;;; ***************************************************************************

;;; ***************************************************************************
;;; ***
;;; ***          A Proposition's State of Assignment
;;; ***
;;; ***************************************************************************

;;; During search, ISAT distinguishes between four states of 
;;; a propositional variable:

;;; 1) an unassigned proposition.
;;; 2) a proposition assigned by ISAT splitting.
;;; 3) a proposition assigned via unit propagation.
;;; 4) a proposition assigned
;;;    by some routine outside of ISAT.

;;; The file propositional-theories.lisp provides macros for 
;;; distinguishing the first two.  Below are macros for indicating 
;;; and query support provided by ISAT splitting.

;;; Support external to ISAT can be identified by querying 1-3.  
;;; Macros, commented out, are provided below for testing external 
;;; support as well.

(defconstant *split-by-isat* :isat

  "Support used to indicate that ISAT assigned the truth-value 
   during variable splitting.")

(defmacro isat-support? (support)

  "Returns T if SUPPORT is provided by ISAT variable splitting."

  `(eql ,support *split-by-isat*))

(defmacro split-on-by-isat? (proposition)

  "Returns T if PROPOSITION was assigned by
   ISAT variable splitting.  This holds if
   the current truth assignment of proposition
   is supported by ISAT variable splitting."

  `(isat-support? (proposition-support ,proposition)))

;(defmacro support-external-to-ISAT? (support)
;
;  "Returns T if SUPPORT is provided by an external source
;   other than ISAT."

;  `(not (or (unit-propagation-support? ,support)
;	    (isat-support? ,support))))

;(defmacro supported-external-to-ISAT? (proposition)
;
;  "Returns T if PROPOSITION's truth assignment is
;  supported by an external source other than ISAT."
;
;  `(support-external-to-ISAT? (proposition-support ,proposition)))

;;; ***************************************************************************
;;; ***
;;; ***  Keeping Track of Remaining Values During Proposition Splitting
;;; ***
;;; ***************************************************************************

;;; When ISAT first assigns a proposition a truth value by splitting, 
;;; there will be a remaining assignment that hasn't been tried yet.
;;; Once the first assignment is tried, the remaining value is tried,
;;; and no assignment remains.  This information is maintained by
;;; the following functions.  The remaining value of a proposition
;;; is one of:
;;;   *true*, *false*, *none* 
;;; if proposition is currently being split by ISAT, 
;;; The value is NIL if the proposition is not being split.

(defmacro set-remaining-value! (proposition truth-value)

  "Records that VALUE is the remaining truth value
   to be tried by ISAT to search over PROPOSITION.
   If PROPOSITION is being searched, then value must 
   be one of *true*, *false* or *none*.  If not being
   searched, then value must be NIL."
  
  `(setf (proposition-remaining-value ,proposition)
     ,truth-value))

(defmacro no-remaining-value! (proposition)

  "Records that there are no remaining truth values for
   ISAT to search for PROPOSITION."
  
  `(setf (proposition-remaining-value ,proposition)
     *none*))

(defmacro no-remaining-value? (proposition)
  
  "Returns TRUE if ISAT has no remaining truth values to
   be searched for PROPOSITION."

  `(eql (proposition-remaining-value ,proposition)
		*none*))

(defmacro has-remaining-value? (proposition)
  
  "Returns TRUE if a truth value remains to
   be searched for PROPOSITION by ISAT.
   Only valid to call if (SPLIT-ON-BY-ISAT? PROPOSITION)
   returns TRUE."

  `(not (no-remaining-value? ,proposition)))

;;; ***************************************************************************
;;;
;;;        Routines to Support Conflict Learning
;;;
;;; ***************************************************************************

;;; The following are interface functions betweeen model finding
;;; and with the truth maintenance systems (proposition theory with
;;; incremental unit propagation) that support the extraction, querying
;;  and recording of conflicts:

(defmacro extract-conflicting-assignment (theory)
  
  "Returns a conflict - a list of propositions 
   that support the current inconsistency and 
   are assigned by ISAT or externally 
   (i.e., not by unit propagation).
   Conflict  is guaranteed to not contain duplicate
   propositions."
   
  `(conflict-of-inconsistent-theory ,theory #'identity))

(defun conflict-contains-proposition-split-by-isat? 
    (conflict)
  
  "Returns true if the truth assignment of at least 
   one proposition in CONFLICT results from ISAT
   splitting on that proposition."
   
  (loop for p in conflict
      when (split-on-by-isat? p)
      return T
      finally (return nil)))

;;; Note:
;;; o The following is here for backward compatibility.
;;;   Use make-instance instead.
(defmethod new-clause-from-conflict 
    ((theory theory) source conflict)
  
  "Creates and returns a new clause for THEORY with SOURCE
   that denotes CONFLICT.

   CONFLICT is a list of mutually inconsistent 
   propositions.  Conflict MUST NOT contain duplicate
   propositions

   This function creates a new clause, and associates it
   with THEORY, but does not insert it into THEORY."
  
  ;; Note, as with NEW-CLAUSE, this function does not 
  ;; update any of the status slots of the clause.  
  ;; This is performed when the clause is inserted into 
  ;; the theory. 

  (debug-print-for :isat t "Creating clause for conflict ~A." 
				   conflict)

  (setq conflict (mapcar (lambda (prop) `(not ,(proposition-description prop)))
						 conflict))

  ;;  (setq conflict (mapcar (lambda (prop) (proposition-description prop)) conflict ))
  
  ;;  (make-instance 'clause :source source :clause conflict)
  (make-instance 'clause :constraint-system theory
				 :source source
				 :clause conflict))

(defmethod insert-conflict-as-new-clause! 
	((theory theory) conflict)

  " Create a clause representing CONFLICT and 
   insert it into THEORY.
 
   CONFLICT is a list of propositions, denoting an
   inconsistent assignment.  CONFLICT must be
   guaranteed to have no duplicates.

   Returns the inserted clause"
  
  ;; create the clausev
  (let ((cl (new-clause-from-conflict
			 theory :isat conflict)))
	;; add it to the theory
    (insert-constraint theory cl)
	cl
))


;;; ***************************************************************************
;;;
;;;        Definition -  Model supporting a Theory
;;;
;;; ***************************************************************************

;;; Added
;;; o We may be interested just in extracting the positive literals of a model,
;;; i.e. the 'candgen-assignments' (for OPSat purposes at least).
;;; This definition makes more straight-forward the extraction of such
;;; assignments.

(defclass theory-model ()

  ((name
	:initarg :name
	:initform (gentemp "theory-model-")
	:accessor theory-model-name
	:documentation "The name for a model supporting a theory")

   (true-propositions
	:initarg :true-propositions
	:initform nil
	:accessor theory-model-true-propositions
	:documentation "The propositions set to true in this model")

   (false-propositions
	:initarg :false-propositions
	:initform nil
	:accessor theory-model-false-propositions
	:documentation "The propositions set to false in this model"))

  (:documentation " A model supporting a theory"))

(defmethod print-object

    ((model theory-model) stream)

  (with-slots (name true-propositions false-propositions) model
    (format stream "<Model ~a: ~% ~a ~% ~a>" name true-propositions
	    false-propositions)))
					    
	    