;;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)

;;; ***************************************************************************
;;;
;;;                Model Finding via Systematic Search
;;;
;;; ***************************************************************************

;;; This module implements the core routines for model finding of a
;;; propositional theory.  It supports the consistency query functions
;;; provided in the file ISAT.LISP

;;; Model finding is implemented as an extension of the 
;;; Davis-Putnam (i.e., spliting plus unit propagation)
;;; with several significant extensions:
;;;      a) incremental unit propagation.
;;;      b) conflict learning, 
;;;      c) dynamic backtracking and
;;;      d) randomized assignment.

;;; ***************************************************************************
;;;
;;;                           External Interface
;;;
;;; ***************************************************************************

;;; Queries to theory for consistency and model finding:

;;;   (CLEAR-SEARCH-STATE! THEORY)

;;;    Removes any propositions that are currently
;;;    assigned by the current search, and sets up the queue 
;;;    of propositions to search over for THEORY.

;;;   (FIND-MODEL
;;;    &OPTIONAL (THEORY *DEFAULT-THEORY*)
;;;	         (INITIALIZE? T)
;;;	         (RETURN-MODEL? T)
;;;	         (RETURN-CONFLICT? T)
;;;	         (KEY #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
;;;	         (SEARCH-RESULTS-FORMATTER-FN #'CURRENT-ASSIGNMENT))
  
;;;    Searches for a consistent full truth assignment 
;;;    the propositions of THEORY.  
;;;    Returns the pair: (MODEL CONFLICT).

;;;    MODEL is non-nil iff an assignment is found.

;;;    If RETURN-MODEL? is TRUE then MODEL
;;;    is the consistent assignment found, and consists 
;;;    of a list of proposition/truth-value pairs.

;;;    If theory is found inconsistent based on unit-propagation 
;;;    alone, and RETURN-CONFLICT? is TRUE, then a CONFLICT is 
;;;    returned.
 
;;;    A CONFLICT denotes a partial truth assignment that is 
;;;    inconsistent.  It is generated by applying function 
;;;    KEY to each externally assigned proposition that 
;;;    supports an inconsistent clause of THEORY, and by 
;;;    collecting the non-nil, irredundant values returned by KEY.
;;;    An externally assigned proposition is one that is not 
;;;    supported by unit propagation.

;;;    By default, KEY takes PROPOSITION as input and returns 
;;;    assignment (= PROPOSITION TRUTH-VALUE) where 
;;;    TRUTH-VALUE is *true* or *false*.

;;;    If theory is found consistent by unit propagation, and 
;;;    PERFORM-SEARCH? is true,then this function searches for a 
;;;    MODEL (consistent assignment).

;;;    The first time this function is called, INITIALIZE? 
;;;    must be T. After the initial call additional consistent 
;;;    assignments may be found by calling this function with 
;;;    INITIALIZE? = NIL.  

;;;    In particular, if INITIALIZE? is T, then this function
;;;    searches from the beginning of the search tree; otherwise, 
;;;    it searches starting from the last consistent assignment 
;;;    found (i.e., the current search state).

;;;    The returned MODEL is generated by applying function
;;;    SEARCH-RESULTS-FORMATTER-FN to theory. It defaults to
;;;    the #'CURRENT-ASSIGNMENT of THEORY. 

;;;    Side Effect:  Leaves propositions assigned to truth values 
;;;    of the current search state."

;;;  Search Control:  Flags for controlling the search strategy 
;;;  and routines for modifying the state of these flags are 
;;;  provided in the file  search-defs.lisp.  Documentation for 
;;;  these interface functions is repeated below:

;;;   (CONFLICT-LEARNING-ON)
;;;   (CONFLICT-LEARNING-OFF)

;;;    Turns ON/OFF conflict-learning within ISAT.
;;;    Conflicts are partial inconsistent assignments,
;;;    and are used to aid forward and backward search.

;;;   (DYNAMIC-BACKTRACKING-ON)
;;;   (DYNAMIC-BACKTRACKING-OFF)

;;;    Turns ON/OFF dynamic-backtracking within ISAT.
;;;    During backtracking, ISAT then avoids undoing 
;;;    assignments that are not mentioned in an active 
;;;    conflict.  

;;;    Dynamic backtracking, requires conflict-learning,
;;;    hence conflict-learning is on as well as long
;;;    as dynamic-backtracking is on.  Turning off
;;;    dynamic backtracking, however, does not turn
;;;    off conflict learning.

;;;   (RANDOM-ASSIGNMENT-ON)
;;;   (RANDOM-ASSIGNMENT-OFF)

;;;    Tells ISAT to TURN ON/OFF randomized assignment.
;;;    That is, for each proposition, ISAT assign its 
;;;    truth values in a random order.

;;;(eval-when (compile load eval)
;;;  (export '(clear-search-state!
;;;			find-model
;;;			;; Calls to the following should
;;;			;; be replaced with calls to
;;;			;; find-model
;;;			find-consistent-assignment	
;;;
;;;			)
;;;		  :ISAT))

;;; ***************************************************************************
;;;
;;;                  External Interface: 
;;;
;;; ***************************************************************************

;;; ***************************************************************************
;;;
;;;                  External Interface: Clear-Search-State!
;;;
;;; ***************************************************************************

(defmethod clear-search-state! ((theory theory))

  "Removes any propositions that are assigned by
   isat for the current search, and sets up the queue 
   of propositions to search over for THEORY."

  (debug-print-for :isat t "Initialize ~A." theory)
  (loop for proposition in (theory-propositions theory)
      when (split-on-by-isat? proposition)
      do (set-remaining-value! proposition nil)
		 (unassign-truth-value proposition *split-by-isat* theory))
  (clear-assignment-queue! theory)
  T)

;;; ***************************************************************************
;;;
;;;                  External Interface: Find-Model
;;;
;;; ***************************************************************************

;;; To Do:
;;; o find-consistent-assignment is a legacy name.
;;;   In opsat, replace any calls to this function
;;;   with calls to find-model, while changing the
;;;   key word return-assignment? to return model.
;;;   Once this is completed, delete the following
;;;   function definition.

;;; To Do:
;;; o remove optional and make a method.
(defun find-consistent-assignment
    (&optional (theory *default-theory*)
			   (initialize? T)
			   (return-assignment? T)
			   (return-conflict? T)
			   (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT))
  (find-model theory initialize? return-assignment?
			  return-conflict? key))

;;; To Do:

;;; o Introduce indentations to make it easier to read 
;;;   debugging traces.

;;; o Conflict extraction:
;;;   - Extend conflict extraction to find a conflict 
;;;     resulting from search, not just unit propagation.
;;;   
;;;   - Change the last expression of the search routine
;;;     so that it returns a conflict after search has 
;;;     been exhausted, and initialize? and return-conflict? 
;;;     a are both true.  In this case, when clauses are 
;;;     added, the returned theory should be in an 
;;;     inconsistent state, and a call to conflict-of-inconsistent 
;;;     theory should produce the desired result(???).

;;; o Review callers of find-consistent-assignment. 
;;;     o should probably always return an assignment.
;;;     o should probably not return a conflict by default.
;;;     o should allow the return of multiple assignments 
;;;       or all assignments.

;;; Check:

;;; o When backtrack chronologically returns inconsistent, 
;;;   what does this mean?  
;;;   - Does it always mean that there is a conflict due to 
;;;     unit propagation?  
;;;   - Is it also returned if model exists?
;;;   - Is it returned if a model was previously returned, but 
;;;     the set of solutions has been exhausted?
;;;   - The returned value should distinguish between an 
;;;   - inconsistent theory and an exhausted domain.
;;;   
;;; o Are there other side effects of the search algorithm 
;;;   that should be documented?

;;; Optimizations:

;;; o Conflicts:
;;;   - The current code uses key to extract conflicts.  
;;;     This is probably inefficient, meter this and 
;;;     correct if need be.
;;;     
;;;   - Is there a more efficient representation for the 
;;;     returned conflict?

;;; o Documentation strings:
;;;   - What effect does document strings have on performance. 
;;;     Do they need to be commented out?  Can they be decached
;;;     from the load through a compiler flag?

;;; o Debugging information:
;;;   -  Does it affect performance to compile with or without 
;;;      debugging information? Can debugging statements be 
;;;      eliminated during compilation through a compile time flag?

;;; A note about conflicts:

;;; Find model and the consistency checking functions
;;  define a conflict as a partial truth assignment that is
;;; inconsistent with the clauses of theory, where each assignment
;;; has been provided by ISAT splitting or by an external source
;;; (i.e., not deduced by unit propagation).
  
;;; Note that early implementations returned as a conflict a 
;;; subset of the  clauses of theory that are inconsistent.  
;;  It then implemented assignments as unit clauses.  
;;; Extracting the set of  inconsistent clauses may be harder 
;;; to derive when the inconsistency is found through systematic 
;;; search, hence we have removed this functionality.

;;; Check:
;;; o Changed theory from an optional to a required parameter.
;;;   Check callers.

;;; Changed:
;;; o Added keyword ''search-results-formatter-fn'' which defaults to
;;; #'current-assignment (i.e., by default, does the same as before).
;;; This change allows us to optionally get the found-model as an object
;;; of type theory-model
;;; (When OpSAT asks for the models supporting consistency of a partial
;;; assignment, it is convenient to have  true-propositions separated
;;; from false-propositions). 

(defmethod find-model
    ((theory theory) &optional
					 (initialize? T)
					 (return-model? T)
					 (return-conflict? T)
					 (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
					 (search-results-formatter-fn #'current-assignment))
  
  "Searches for a consistent full truth assignment 
   the propositions of THEORY.  
   Returns the pair: (MODEL CONFLICT).

   MODEL is non-nil iff a model (consistent assignment)
   has been found.

   If RETURN-MODEL? is TRUE then MODEL 
   is the consistent assignment found, and consists 
   of a list of proposition/truth-value pairs.

   If theory is found inconsistent based on unit-propagation 
   alone, and RETURN-CONFLICT? is TRUE, then a CONFLICT is 
   returned.
 
   A CONFLICT denotes a partial truth assignment that is 
   inconsistent.  It is generated by applying function 
   KEY to each externally assigned proposition that 
   supports an inconsistent clause of THEORY, and by 
   collecting the non-nil, irredundant values returned by KEY.
   An externally assigned proposition is one that is not 
   supported by unit propagation.

   By default, KEY takes PROPOSITION as input and returns 
   assignment (= PROPOSITION TRUTH-VALUE) where 
   TRUTH-VALUE is *true* or *false*.

   If theory is found consistent by unit propagation, and 
   PERFORM-SEARCH? is true,then this function searches for a 
  consistent ASSIGNMENT.

   The first time this function is called, INITIALIZE? 
   must be T. After the initial call additional consistent 
   assignments may be found by calling this function with 
   INITIALIZE? = NIL.  

   In particular, if INITIALIZE? is T, then this function
   searches from the beginning of the search tree; otherwise, 
   it searches starting from the last consistent assignment 
   found (i.e., the current search state).

   The returned MODEL is generated by applying function
   SEARCH-RESULTS-FORMATTER-FN to theory. It defaults to
   the #'CURRENT-ASSIGNMENT of THEORY. 

   Side Effect:  Leaves propositions assigned to truth values 
   of the current search state."

  (debug-print-progn-for 
   :isat
   (format t "~%")
   (debug-print-for :isat t "Find model for ~A." 
					theory))

  ;; A) Setup search

  (let ((consistent? (setup-search theory initialize?)))
    (cond ((not consistent?)

		   ;; Inconsistency found at start by unit propagation, 
		   ;; return failure.
		   (debug-print-for :isat t "~A inconsistent by unit propagation." 
							theory)
		   (return-inconsistency theory 
								 (and initialize?
									  return-conflict?)
								 key))

		  (:else

		   ;; B) Invoke Search Loop and return the results.
		   (return-search-results theory 
								  (search-forward theory) ; Search here!
								  return-model?
								  search-results-formatter-fn)))))


;;; Added:
;;; o Method to find all models for a theory.
;;; Fix?
;;; o We may want to change find-model so as initialization
;;; for search it may either propagate, backtrack chronologically
;;; or just do nothing. Right now we just have the first two ones.
;;; (for which reason variable clear?, declared inside the loop,
;;; needs to be T for the 1st iteration and false afterwards.
(defmethod find-all-models
    ((theory theory) &key
					 (return-model? T)
					 (return-conflict? T)
					 (key #'PROPOSITION-TO-TRUTH-ASSIGNMENT)
					 (search-results-formatter-fn #'current-model))
    
    "Exhaustively enumerates all models consistent with theory.
     Returns a pair (LIST-OF-MODELS CONFLICT) where
	i) LIST-OF-MODELS is a list of all MODELs consistent with theory.
	ii) CONFLICT is an inconsistency
	    (for detailed defs refer to documentation of find-model)
    Side Effect:  Leaves propositions assigned to truth values 
    of the current search state. "
    ;(setup-search theory initialize?)
    (loop
		with model
		with conflict
		with clear? = T
		
		do (multiple-value-setq (model conflict)
			 (find-model theory clear?
						 return-model?
						 return-conflict?
						 key 
						 search-results-formatter-fn))
		   (setf clear? nil) ;; clear? needs to be T for 1st iteration
										;do (incf common-lisp-user::*model-counter*)		     	
		when (null model)
		return (values list-of-models conflict)
	       
	
		collecting model into list-of-models))



 (defmethod insert-constraints-and-find-all-models-if-any
     ((theory theory) &optional &key
								(added-constraints nil)
								(added-assignments nil)
								(source nil)
								;(perform-search? t)
								(return-conflict? t)
								(key #'proposition-to-truth-assignment))

   "Same spirit as consistent-with? but now attempts to find all consistent models for
     theory+added-constraints.

   FINDS all consistent models of the current constraints and assignments in THEORY,
   conjoined with ADDED-CONSTRAINTS and ADDED-ASSIGNMENTS.

   RETURNS (values LIST-OF-MODELS CONFLICT).  LIST-OF-MODELS is non-nil unless theory
   is proven inconsistent.  It first tests consistency by propagation.  

   Each clause in ADDED-CONSTRAINTS is a clause object.

   Each assignment in ADDED-ASSIGNMENTS is a list:

       (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *true*))

   and specifies that assignment proposition (= variable value) has TRUTH-VALUE.
   
   If PERFORM-SEARCH? is TRUE, and propagation is consistent, then it 
   continues with a full search-based satisfiability test,
   when one exists.

   If RETURN-CONFLICT? is TRUE and an inconsistency is found by
   propagation, then returns a conflict.  

   A CONFLICT denotes a subset of ADDED-ASSIGNMENTS that is inconsistent
   with theory and ADDED-CONSTRAINTS.  Each assignment is represented
   internally as an assignment object.  The conflict is generated by 
   applying function KEY to each such assignment that supports an 
   inconsistency of THEORY, and by collecting the non-nil, 
   irredundant values returned by KEY.

   By default, KEY takes a PROPOSITION object as input and returns 
   assignment (= PROPOSITION TRUTH-VALUE) where TRUTH-VALUE is 
   *true* or *false*."
   ;; Insert additional clauses and assignments
   (loop for constraint in added-constraints
	   do (insert-constraint theory constraint))
   
   (assert-assignments! theory source added-assignments)
   
   ;; Test Consistency by propagation.
   
   ;; Conflict is a list of truth value assignments
   ;; (= PROPOSITION TRUTH-VALUE), where truth-value is
   ;; *true* or *false*, and PROPOSITION is a proposition
   ;; object whose proposition description is an expression
   ;; (= STATE-VARIABLE VALUE).
   
   (multiple-value-bind (list-of-models conflict) ;consistent? conflict)
	   (find-all-models theory
						:return-conflict? return-conflict?
						:key key)
	 
	 ;; Remove the inserted constraints and assignments
	 (loop for constraint in added-constraints
		 do (remove-constraint theory constraint))
	 
	 (unassert-assignments! theory source added-assignments)
	 
	 ;; Return the results
	 (values list-of-models conflict)))

				     
;;; ***************************************************************************
;;; ***************************************************************************
;;;
;;;                  Support Functions for Model Finding
;;;
;;; ***************************************************************************
;;; ***************************************************************************

(defmethod setup-search ((theory theory) initialize?)

  "Backup search state to a consistent, unit propagated partial assignment.
   If initialize? is non-nil, then use the empty partial assignment.
   Returns TRUE if consistent, and NIL if inconsistent."
  
  ;; Optionally reinitialize the search state and
  ;; Move search to a consistent, unit propagated 
  ;; partial assignment.
  (cond (initialize?
	 (clear-search-state! theory)
	 (propagate theory))

	;; If continuing search, then currently at a full assignment, 
	;;  backtrack in preparation to move to the next full assignment.
		; WHY do we backtrack chonologically - when there are other backtrack methods?
	(:else
	 (debug-print-for :isat t "Search from last assignment found.")
;;;	 (debug-print-for :isat t "Backing up.")
	 (backtrack-chronologically theory))))

;;; ***************************************************************************
;;;
;;;                  Forward Search
;;;
;;; ***************************************************************************

;;; Check:

;;; To Do:
;;; o This function is four and a half pages long;
;;    modularize it by adding appropriate subroutines.

;;; Optimizations:

;;; o ISAT maintains a search queue of assigned and
;;;   unassigned propostions. We may want to encode the mark of 
;;;   assigned vs unassigned simply as an array 
;;;   and a pointer, such that all unassigned propositions 
;;;   are guaranteed to be to the right of pointer.  If 
;;;   this is a bit array, then multiple propositions
;;;   can be tested in one swoop.

;;; o Forward and backtrack search seem more 
;;;   complicated than necessary.  Can this be 
;;;   simplified?

;;;   - Search may be partially simplied by
;;;     folding the initial unit propagation into
;;;     the standard iteration.

;;;   - This search can be partially simplified by 
;;;     folding part of the loop into the function 
;;;     that finds the next and preceding proposition 
;;;     to be assigned/unassigned.

;;;   - Can the search be structured closer to standard 
;;;     code, which places forward and backward search 
;;;     within the same loop.  

;;; o Might want to change new-proposition-to-assign so that it
;;;   skips to the next proposition that needs to be
;;;   assigned.  This might help reduce the number of clauses 
;;;   in the COND within search-forward, and may make it easier 
;;;   to implement faster strategies for jumping forward in a 
;;;   transparent manner.

(defmethod search-forward ((theory theory) &key (return-new-clauses? nil))
  
  "Assigns unassigned propositions until all are assigned 
   or an inconsistency is found.  

   Search assumes that theory starts in a consistent, 
   unit propagated assignment.

   Returns (VALUES SUCCESS? NEW-CLAUSES)
   SUCCESS? is T if a consistent full assignment is found, 
   and NIL if no search states remain. Leaves assignments in place.

   If RETURN-NEW-CLAUSES? is true, any new clauses added 
   by conflict-learning (specifically, dynamic backtracking) 
   will be returned. Otherwise, returns nil."

  ;; Currently there is a partial assignment that is consistent 
  ;; and unit propagated.
  ;; 
  ;; Note: Search tries *TRUE* before *FALSE*, unless randomized
  ;; assignment is enabled.
  
  ;; First find a proposition to which a value can be assigned
  (debug-print-for :isat t "Search forward.")

  ;; Quit immediately if there are no unassigned propositions.
  (if (eql (heap-count (theory-unassigned-propositions theory)) 0)
	  (return-from search-forward (values t nil)) ;; RETURN!
	  )

  (loop 
	 ;; Holds the value to assign to the new
	 ;; unassigned proposition to be assign.
	 with selected-truth-value
	 ;; holds all of the clauses (conflicts)
	 ;; added by this search
	 with new-clauses
		   ;; Optimization:
		   ;; o Might want to change new-proposition-to-assign so that it
		   ;;   skips to the next proposition that needs to be
		   ;;   assigned.  This might help reduce the number of clauses 
		   ;;   in the following COND, and may make it easier to implement
		   ;;   faster strategies for jumping forward in a transparent
		   ;;   manner.
	 for proposition = (new-proposition-to-assign theory)
	 do 
;	   (format t "==========MAIN SEARCH LOOP===========~%")
;	   (format t "num clauses : ~a~%" (length (theory-clauses theory)))
;	   (format t "prop: ~a ~%" proposition)

	   (cond ((all-propositions-assigned? theory)
			  
			  ;; CASE 1: There are no more unassigned propositions, return success.
			  (debug-print-for :isat t "  All propositions assigned ~
                                and consistent.")
			  
			  (return (values t new-clauses)) ;; RETURN!
			  )
			 
			 ;; CASE 2: the currently selected proposition has 
			 ;; already been assigned (by unit prop or external
			 ;; assignment)
			 ((assigned? proposition)
			  ;; update the theory by moving the latest proposition
			  ;; from the unassigned to assigned list.
			  (push-new-proposition-to-assigned theory)

			  ;; Proposition already has a truth value, move forward.
			  ;; If debugging is on, then indicate where we are.
			  (debug-print-progn-for :isat
									 (let ((support (proposition-support proposition)))
									   (cond ((unit-propagation-support? support)
											  (debug-print-for 
											   :isat t 
											   "  Skip ~a, ~a by propagation."
											   Proposition
											   (proposition-truth-value Proposition)))
											 ((isat-support? support)
											  
											  ;; If performing DBT, then proposition 
											  ;; might have already been split on.
											  (if (perform-dynamic-backtracking?)
												  (debug-print-for 
												   :isat t "  Skip ~a, ~a by splitting."
												   proposition (proposition-truth-value Proposition))
												  
												  ;; but this shouldn't occur for chronological BT.
												  (warn "Proposition already split on, this ~
                                   shouldn't have happened, since we ~
                                   performing dynamic backtracking.")))
											 ((no-support? support)
											  (warn "Proposition ~A assigned, but has no ~
                                support.  This shouldn't happen."
													proposition))
											 (:otherwise
											  ;; Support is external to ISAT
											  (debug-print-for :isat t "  Skip ~a, ~a by ~a." 
															   proposition 
															   (proposition-truth-value proposition)
															   support))
											 ))))
			 
			 ;; CASE 3
			 ;; Else proposition is unassigned;
			 ;; select, assign and propagate a truth value 
			 ;; and check if it is consistent.
			 ((progn
				(setq selected-truth-value 
					  (select-truth-value proposition theory))
				;; whether the "assign-truth-value" call is successful
				;; or not, we need to make sure the current proposition
				;; is moved the assigned list - because the backtrack
				;; methods use the assigned list to undo all assignments.
				(push-new-proposition-to-assigned theory)
				(assign-truth-value proposition 
									selected-truth-value
									*split-by-isat*
									theory))

			  ;; Assigning SELECTED-TRUTH-VALUE to proposition is 
			  ;; consistent, move forward.
			  (debug-print-for :isat t "  ~A = ~A consistent, ~
                                          ~a remains."
							   proposition selected-truth-value
							   (opposite-truth selected-truth-value))
			  
			  ;; Record remaining value to assign, ONLY if searching for 
			  ;; multiple models, or if searching chronologically.
			  ;; When searching for one model and using
			  ;; full conflict-learning, if a proposition is assigned
			  ;; through splitting, then the other value always remains
			  ;; as an option to assign.
				(set-remaining-value! 
				 proposition
				 (opposite-truth selected-truth-value))
			  )
			 
			 ;; CASE 4:
			 ;; Assigning the selected value failed.
			 
			 ;; If not performing conflict learning, 
			 ;; then remove selected assignment, and
			 ;; try assigning the opposite truth value 
			 ;; and propagating.
			 ((and
			   (not (perform-conflict-learning?))
			   (progn
				 (unassign-truth-value proposition 
									   *split-by-isat*
									   theory)
				 (assign-truth-value proposition 
									 (opposite-truth selected-truth-value) 
									 *split-by-isat*
									 theory))
			   )
			  
			  ;; Assigning the opposite truth value to proposition 
			  ;; is consistent, record that their are no remaining 
			  ;; values and move forward.
			  (debug-print-progn-for :isat
									 (debug-print-for :isat t "  ~A = ~A failed, but" 
													  proposition selected-truth-value)
									 (debug-print-for :isat t "  ~a = ~a consistent."
													  proposition 
													  (opposite-truth selected-truth-value)))
			  ;; set that there are no remaining values to try.
			  (no-remaining-value! proposition))
			 
			 ;; CASE 5:
			 ;; For chronological search, the second 
			 ;; assignment to proposition failed, so 
			 ;; we backtrack chronologically.
			 
			 ;; If conflict learning is on, then the alternate
			 ;; assignment to PROPOSITION will be assigned
			 ;; automatically by unit propagation, once the
			 ;; conflict is recorded.  We now perform
			 ;; conflict-directed bactracking in order to
			 ;; unassign proposition, record the conflict
			 ;; as a clause and propagate.
			 
			 ;; Backtrack: 
			 ;;  (conflict-directed or chronological).
			 
			 ((if (perform-conflict-learning?)
				  (progn  ;(format t "conflict-directed-backtrack~%")
					;; Save off the learned clauses. 
					;; Return if backtrack was a success
					(multiple-value-bind (success? conflict new-clauses-fromcdb)
						(conflict-directed-backtrack theory :return-new-clauses? return-new-clauses?)
					  (declare (ignore conflict))
					  (if return-new-clauses?
						  (setf new-clauses (append new-clauses new-clauses-fromcdb)))
					  success?))

				  (progn  ;(format t "backtrack-chronologically~%")
					(backtrack-chronologically theory) )
				  )

			  ;; Backtracking succeeded, continue moving forward.
			  (debug-print-for :isat t "  ~a = ~a and ~a fail. ~
                                ~%  Backtracking succeeds, move forward."
							   proposition)
			  )

			 ;; CASE 6: Backtracking failed
			 (:else
			  ;; Backtracking failed, hence no more 
			  ;; consistent assignments remain; 
			  ;; return failure from search-forward.
			  
			  (debug-print-for :isat t "  ~a = ~a and ~a fail. ~
                                          ~%  Backtracking failed."
							   proposition *true* *false*)
			  (return (values nil new-clauses)) ;; RETURN!
			  ))
	   ))

;;; To Do:
;;; o Also implement the standard SAT heuristics for 
;;;   value selection.
(defmethod select-truth-value 
	((proposition proposition) (theory theory))

  "Selects a truth-value from *true* and *false*,
   to be assigned to proposition."

;;;  (declare (ignore proposition theory))
  (if (perform-random-assignment?)
      (random-truth)
    *true*))

;;; ***************************************************************************
;;;
;;;                  Chronological Backtracking
;;;
;;; ***************************************************************************

;;; Check:
;;; o May need to extract the conflict within 
;;;   this function and return it,
;;;   in order to be consistent with the function
;;;   conflict-directed-backtracking.

;;; Fix:
;;; o In the following, when unassign-truth-value
;;;   is called, will it make sure that the unit 
;;;   propagation queue is run?  If not, then make
;;;   an explicit call to propagate.

;;; To Do:
;;; o Implement conflict extraction similar to
;;;   conflict-directed backjumping.

(defmethod backtrack-chronologically ((theory theory))

    
  "Called on a theory that has an inconsistent
   current truth assignment and has been unit
   propagated.

   backs up to a partial assignment that is consistent 
   with theory, and then move forward one consistent 
   assignment.  

   Returns SUCCESS? where
   Success? is true if the function was able to
      backtrack to a consistent next partial state; 
      otherwise, it is nil."

  (debug-print-for :isat t "Backtrack.")
  
  (loop for proposition = (last-assigned-proposition theory)

      when (all-propositions-unassigned? theory)

		   ;; Search has no propositions that are assigned, 
		   ;; hence the inconsistency can't be removed;
		   ;; return failure.
	   
		   ;; Change: Decide whether to return failure 
		   ;;         or to extract the conflict here.
      do (debug-print-for :isat t "  No propositions to back over.")
      and return nil
	       
      do (cond ((not (split-on-by-isat? proposition))

				;; PROPOSITION was not assigned by splitting
				;; hence it can't be unassigned; back over it.
				;; Backup is performed by the expression after
				;; the COND.
				(debug-print-for :isat
								 t
								 "  Back over ~A, ~a by propagation."
								 proposition (proposition-truth-value proposition)))
	       
			   ((has-remaining-value? proposition)

				;; PROPOSITION was assigned by splitting and
				;; has a value remaining, so try it out.
				(let ((success? (try-remaining-value proposition theory)))
				  (when success? (return t))))

			   (:otherwise 

				;; PROPOSITION was split upon, but does not have a 
				;; remaining  value to try to assign.  Remove the 
				;; current assignment to PROPOSITION, and 
				;; backup over PROPOSITION.  Backup is
				;; performed by the expression after the COND.

				;; FIX After calling UNASSIGN we probably need to
				;;     call PROPAGATE to ensure that any unit 
				;;     clauses put on the queue (e.g., for
				;;     resupport) have been propagated, and
				;;     any new inconsistencies have been exposed.

				(debug-print-for :isat t
								 "  Back over ~A, has no remaining assignments."
								 proposition)
				(unassign-truth-value proposition *split-by-isat* theory)))

		 ;; The routine wasn't able to consistently reassign 
		 ;; PROPOSITION for one of the reasons in the COND;
		 ;; back over proposition and loop over its assigned predecessor.
		 (set-remaining-value! proposition nil)
		 (pop-last-proposition-from-assigned theory)))

;;; Fix:
;;; o In the following, when unassign and assign-truth-value
;;;   are called, will they make sure that the unit propagation
;;;   queue is run?  We will want to perform both unassign 
;;;   AND assign before propagating, but then we do need to 
;;;   make sure to propagate.
(defmethod try-remaining-value ((proposition proposition) (theory theory))

  "PROPOSITION was assigned by splitting and
    has a value remaining, so try assigning it.

    RETURNS True iff assignment is consistent by
    unit propagation."

  (debug-print-for :isat t "  Switch ~A to ~A." 
				   proposition
				   (proposition-remaining-value proposition))
		
  ;; First remove the current assignment to PROPOSITION
  ;; and its consequences, ...
  (unassign-truth-value proposition *split-by-isat* theory)
		
  ;; ..then assign the remaining value, propagate
  ;;   and check the results.
		
  ;; FIX After completing both calls to UNASSIGN and ASSIGN,
  ;;     we probably need to call propagate to ensure that 
  ;;     any unit clauses put on the queue during the
  ;;     unassign-truth-value have been propagated.
  ;;  (debug-print-for :isat t "  Assigning ~A = ~A." 
;;				   proposition 
  ;;				   (proposition-remaining-value proposition))
  (cond ((assign-truth-value proposition
							 (proposition-remaining-value
							  proposition)
							 *split-by-isat*
							 theory)

		 ;; Unit propagation on the remaining assignment 
		 ;; is consistent.  Return success, which will 
		 ;; instigate forward search.
		 (debug-print-for :isat t "  ~A = ~A consistent." 
						  proposition 
						  (proposition-remaining-value proposition))

		 ;; Indicate that both assignments have been tried,
		 ;; and that none remain.
		 (no-remaining-value! proposition)
		       
		 ;; Return from backtrack-chronologically
		 ;; with success.
		 'T)
		      
		(:otherwise

		 ;; Assigning the remaining truth value for
		 ;; PROPOSITION led to an inconsistency,
		 ;; remove the assignment and backup.
		 ;; This routine will back over PROPOSITION
		 ;; using the expression after the COND.

		 ;; FIX After calling UNASSIGN we probably need to
		 ;;     call PROPAGATE to ensure that any unit 
		 ;;     clauses put on the queue (e.g., for
		 ;;     resupport) have been propagated, and
		 ;;     any new inconsistencies have been exposed.
		 (debug-print-for :isat t
						  "  ~A = ~A inconsistent, back track."
						  proposition 
						  (proposition-remaining-value proposition))
		 (unassign-truth-value proposition *split-by-isat* 
							   theory)
	  
		 ;; Return failure
		 nil
		 )))

;;; ***************************************************************************
;;;
;;;                  Conflict-directed Backtracking
;;;
;;; ***************************************************************************

;;; ***************************************************************************
;;;
;;;                  Algorithm Sketch
;;;
;;; ***************************************************************************

;;; The following notes provide a sketch of conflict-directed
;;; backtracking  routine that uses clause learning,
;;; incremental unit propagation and dynamic backup.

;;; This sketch flags parts of the implementation that
;;; need to be fixed, checked, optimized or extended.

;;; 1) Start by extracting a conflict.

;;;  o Check: Only extract a conflict when an inconsistency 
;;;    is found, not when we are moving backwards to look for 
;;;    additional models.  Will this function only be
;;;    called in response to an inconsistency?  

;;;  o The conflict contains propositions corresponding to 
;;;    assignments introduced by ISAT splitting AND OPSAT 
;;;    splitting.

;;;    - Check: Have we properly generalized the extraction
;;;             and return of conflicts to include
;;;             OPSAT assignments?

;;;    - Note that if we didn't include propositions 
;;;      for OPSAT splitting, then the learned clauses could 
;;;      not be carried forward from one OPSAT problem call
;;;      to the next.

;;;    - To Do: A conflict depends on inserted clauses, as well
;;;      as external assignments.  If OPSAT is allowed to remove
;;;      clauses, then we need to add a mechanism to remove
;;;      the conflicts depending on the removed clauses.

;;;      - A simple solution is to introduce a proposition P
;;;        that is true iff clause C is inserted, and replace
;;;        C with P -> C.

;;;      - The most clean alternative is to extend the 
;;;        approach to a conditional CSP formulation.

;;;    - Optimization: We might represent the conflict as 
;;;      a bit array, in order to avoid consing.

;;; 3) If the conflict contains no propositions that were split 
;;;    on by ISAT, then backtracking should return failure.  
;;;    In this case the conflict denotes an inconsistent opsat
;;;    assignment.  Return it.

;;;      - Check: that this is implemented properly.

;;; 4) Otherwise,

;;;   o Record the conflict as a new clause, 

;;;   o Select a proposition assigned by ISAT that is in the 
;;;     conflict and backup directly to it.
;;;     - To Do: What is the criteria for selecting 
;;;       this proposition?
;;;       - Most recently assigned?  Random?  One that satisfies 
;;;         a standard value selection heuristic?  Try each.
;;;         
;;;   o Unassign proposition and propagate.
;;;     - This will assign some unassigned propositions
;;;       including the one that has just been unassigned
;;;       (due to the conflict).

;;;     - In addition, it will detect if there are any 
;;;       additional inconsistencies.

;;;  5) If theory is consistent, 

;;;     o  THEN move forward to the next unassigned proposition,

;;;     o  OTHERWISE backup by recursing on conflict-directed 
;;;        backtracking.

;;;      - FIX: This needs to be implemented.
;;;      - Optimizaton: To avoid nested recursion, we would 
;;;        like to return from backtracking after step 4) and 
;;;        then have the calling loop decide to move forward 
;;;        or backup.

;;; ***************************************************************************
;;;
;;;                  Implementation
;;;
;;; ***************************************************************************

;;; FIX:

;;; o Implement the DDBT-wo-DBT option.  This 
;;;   involves unassigning variables along the 
;;;   way, and updating their remaining values 
;;;   similar to chronological backtracking.

;;; o Change this function and its callers so 
;;    that this function can return a conflict.

;;; o Handle the case where BT resolves the current 
;;;   conflict but theory is still inconsistent. 

;;; o A conflict depends on the existence of clauses,
;;;   in a theory.  If a clause is retracted, then
;;;   some of the conflicts may no longer hold.
;;;   The conflict mechanism needs to be modified to
;;;   address this problem.  Several options are 
;;;   possible.  At least the simplest needs to be
;;;   implemented:

;;;   - Remove all conflicts when any clause is removed.

;;;   - Maintain a stack of clauses that are pushed and
;;;     popped.  Corresponding conflicts are popped as well.

;;;   - A TMS is used to decide which conflicts to add 
;;;     and delete.

;;; Check:

;;  o When backtrack is called just after search 
;;;   has been started, make sure to backtrack 
;;    chronologically, and to determine properly 
;;    the next assignment for each proposition 
;;    backtracked on.
	       
;;  o When a global inconsistency has been found,
;;    should BT leave theory in an inconsistent 
;;    state, so that a conflict can be extracted,
;;    or should BT return the conflict?

;;; To Do:

;;;  o Break this code into several cases and meter 
;;;    separately:
;;;     A) conflict-directed backtracking using 
;;;        conflicts, but without using conflicts 
;;;        during forward search.
;;;        
;;;     B) Use conflicts to aid forward search, 
;;;        but not to aid backtracking.

;;;     A&B combined)

;;;     C) Dynamic backup based on conflicts,
;;;        i) without conflicts during forward search.
;;;        ii) with conflicts during forward search.

;;; o The following code currently backs up to the first 
;;;   proposition that it finds in conflict; however, it 
;;;   may be more efficient to select another proposition, 
;;;   either randomly or through a variable selection 
;;;   heuristic. Implement this.

;;; o Decide if a function calling conflict-directed-backtracking
;;;   will ever need to operate over the subset of the 
;;;   clauses that lead to an inconsistency.  If this is 
;;;   needed, then implement the functionality.  However, 
;;;   it is not easy, and that functionality has currently 
;;;   been removed.

;;; FIX After calling unassign, we need to call propagate
;;;     to ensure that all consequences have propagated
;;;     forward, and have exposed any inconsistencies.

;;; FIX Theory may be inconsistent after retracting an assignment
;;;     in the conflict.  How does this backup if that occurs?

(defmethod conflict-directed-backtrack ((theory theory) &key (return-new-clauses? nil))
  
  "Called on a theory that has an inconsistent
   current truth assignment and has been unit
   propagated.

   Returns (VALUES SUCCESS? CONFLICT NEW-CLAUSES), where
   SUCCESS? is true if the function was able to
      backtrack to a consistent next partial state; 
      otherwise, it is nil.

   If SUCCESS? is nil then CONFLICT is an inconsistent
      set of propositions that were assigned external
      to ISAT.  Otherwise, CONFLICT is nil.

   If return-new-clauses? is true NEW-CLAUSES is the list 
      of clauses that were addded to theory as a result of 
      backtracking. These clauses represent the clausal 
      form of the conflicts learned. Otherwise, nil 
      is returned."

  (debug-print-for :isat t "Backtrack with conflicts.")

  ;; Only call on an inconsistent theory.
  (debug-progn
   (when (no-inconsistent-clauses? theory)
     (error "Conflict-directed backtracking should only be called ~
            with an inconsistent theory, but ~A isn't." theory)))

  ;; Extract the conflicting assignment for inconsistent
  ;; theory as a set of propositions.
  (let* ((conflict (extract-conflicting-assignment theory))
		 ;; Even if a conflict is assigned external to ISAT,
		 ;; we should still learn from it by adding it to the
		 ;; theory.
		 (conflict-clause (insert-conflict-as-new-clause! theory conflict))
		 (new-clauses (if return-new-clauses? `(,conflict-clause) nil)))

	;; If conflict contains no proposition split on by ISAT, 
	;; then theory is globally inconsistent, return conflict.
    (If (not (conflict-contains-proposition-split-by-isat? conflict))
		;; if TRUE
		;; The returned conflict is comprised of a set of
		;; propositions assigned external to ISAT.
		(values nil conflict new-clauses) ;; RETURN!
  
		;; if FALSE
		(loop for proposition = (last-assigned-proposition theory)
			do 
			     
			 (cond ((all-propositions-unassigned? theory)

					;; This is possible if the first assignement 
					;; was done externally, and therefore cannot
					;; be undone by backtracking.
					(return (values nil conflict new-clauses)) ;; RETURN!
					)
		  
				   ;; CASE 2
				   ((or (not (split-on-by-isat? proposition))
						(not (member proposition conflict)))			

					;; Proposition not in conflict, back over it.
					(debug-print-for :isat t
									 "  Back over ~A, not in conflict ~A."
									 proposition conflict)
					(pop-last-proposition-from-assigned theory)
					)

				   ;; CASE 3
				   (:otherwise

					;; Proposition was assigned by ISAT and is 
					;; mentioned in the conflict.
		      
					;; For DDB with clause learning, proposition will 
					;; ALWAYS have a remaining assignment to try.
					;; This assignment will be derived automatically 
					;; via unit propagation, once conflict is inserted
					;; into theory as a clause.
					(debug-progn
					 (unless (has-remaining-value? proposition)
					   (warn "Proposition ~A in conflict ~A should have ~
                              a remaining assignment, but doesn't."
							 proposition conflict)))
					(setf (proposition-remaining-value proposition) nil)

					(unassign-truth-value proposition *split-by-isat* theory)

					;; After unassigning the variable, we need to propagate
					;; again. It is possible that the conflict extracted
					;; was not the minimal conflict, and therefore the theory 
					;; will still be inconsistent. If propagation is sucessful, quit.
					(if (propagate theory)
						;; if TRUE
						;; The theory is consistent. Clean up the assigned list
						;; of propositions unassigned during propagation.
						;; Then return the answer
						(progn
						  (pop-unassigned-from-assigned theory)
						  (return (values T nil new-clauses))) ;; RETURN!
						;; if FALSE
						;; The theory is INconsistent. Remove the previous conflict
						;; Extract a new conflict. Add it into the theory.
						(progn
						  (setf conflict (extract-conflicting-assignment theory))
						  (setf conflict-clause (insert-conflict-as-new-clause! theory conflict))
						  (if return-new-clauses? (push conflict-clause new-clauses))
						  )
						)			
					))
			 ))))



(defun pop-unassigned-from-assigned (theory)
  "Moves the propositions that are unassigned from the 
assigned list and places them back on the unassigned list."
  (let ((prev-ass nil)
		(curr-ass (theory-assigned-propositions theory))
		(prop nil))
	(loop
	   (if (null curr-ass) (return))
	   (setq prop (car curr-ass))
	   (cond ((eql *NONE* (proposition-truth-value prop))
			  ;; remove from the assigned list
			  (if (null prev-ass) 
				  ;; removing from the front of the list
				  (progn
					;; move forward
					(setf curr-ass (cdr curr-ass))
					;; remove
					(setf (theory-assigned-propositions theory) curr-ass))
				  ;; removing from the middle
				  (progn
					;;remove
					(setf (cdr prev-ass) (cdr curr-ass))
					;; move forward
					(setf curr-ass (cdr curr-ass))
					)
				  )
			  ;; add to the unassigned list
			  (heap-insert (theory-unassigned-propositions theory) prop))
			 (:otherwise
			  (setf prev-ass curr-ass)
			  (setf curr-ass (cdr curr-ass)))))
	;; add to assigned list
	(setf (theory-assigned-propositions theory) (reverse (theory-assigned-propositions theory)))
	;(setf (theory-assigned-propositions theory) (theory-assigned-propositions theory))
))


;;; ***************************************************************************
;;;
;;;                  Returning the Results of Search
;;;
;;; ***************************************************************************

(defmethod return-inconsistency ((theory theory) return-conflict? key)

  "Theory is inconsistent; flag an inconsistency.

   Returns (values CONSISTENT? CONFLICT)

   CONSISTENT? is nil, indicating that no model was found.
   CONFLICT is returned RETURN-CONFLICT? is TRUE,
   or NIL if none.

   A CONFLICT is a list of propositions that support
   an inconsistent clause and are externally assigned.
   KEY is applied to each proposition in the conflict
   and a list of all non-NIL results is returned."
  
  (cond (return-conflict?

		 ;; conflict desired, return it.
;;;	 (debug-print-for :isat t "Extracting conflict.")
		 (let ((conflict (conflict-of-inconsistent-theory
						  theory key)))

		   (debug-print-for :isat t "Return conflict ~A." conflict)

		   (values nil conflict)))
		(:else

		 ;; Just flag inconsistency.
		 (values nil nil nil))))

;;; To Do:
;;; o Return a conflict when a complete search has been
;;;   performed and all states have been exhausted.

;;; Check:
;;; o Make sure that the function current-assignment 
;;;   works correctly.
(defmethod return-search-results 
    ((theory theory) model-found? return-model? format-results-fn)
  
  "Search of Theory has terminated at its current state,
  Return (values Assignment Conflict) for this state." 

  ;; Change:  Should return a conflict if requested, 
  ;;          when a complete search was performed and 
  ;;          all states have been exhausted.
  (debug-print-progn
   (if model-found?
       (debug-print-for :isat t "Found model ~A."  
		    (current-assignment theory))
     (debug-print-for :isat t "No model found.")))

  (if model-found?
	    
      ;; Assignment found
      (if return-model?
		
	  ;; and want assignment returned
	  (values (funcall format-results-fn theory); (current-assignment theory)
		  nil)
	      
	;; Not Returning Assignment
	(values T nil))

    ;; Completed systematic search
    ;; Calling routine must determine if this means that
    ;; the theory is inconsistent.  In the case that no
    ;; consistent assignment is found, a conflict IS NOT
    ;; returned.
    (values nil nil)))
