;;; -*- Mode:Common-Lisp; Package:COMMON-LISP-USER; Base:10 -*-

;;; Copyright (c) 1999 - 2009 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :COMMON-LISP-USER)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  			Propositional Satisfiability 
;;; 	      with Incremental Unit Propagation 
;;;              and Dynamic Backtracking
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; ISAT manipulates a propositional THEORY.  A theory is a set of
;;; CLAUSES, where each clause is a disjunction of LITERALS.  A literal is
;;; either a proposition or the negation of a proposition.  A proposition
;;; is a primitive token that can be either true or false.  A positive
;;; literal is true if and only if the corresponding proposition is true.
;;; A negative literal is true if and only if the corresponding proposition
;;; is false.  A clause is satisfied if and only if at least one of its
;;; literals is true.  A theory is satisfiable if and only if there is a
;;; truth assignment to all propositions such that every clause is
;;; satisfied.
;;;
;;; The algorithm is an enhancement of the DPLL algorithm that
;;; - performs incremental unit propagation,
;;    using agressive resupport (nayak/williams) and watched literals ().
;;; - learns conflicts as added clauses
;;; - performs a variant of dynamic backtracking (ginsberg)
;;;   that uses dependency management within incremental unit propagation
;;;   to manage eliminating explanations.

;;; Note:
;;;  This implementation DOES NOT include the ITMS algorithm
;;;  by Nayak and Williams, or its extension, by Shen Qu and Williams.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(pushnew :ISAT *features*)

(defpackage :ISAT (:use "CS" "MTK-UTILS" 
						;;;"CLOS" 
						"COMMON-LISP")
			(:export 
			 ;; From propositional-theories.lisp
			 ;; truth
;;;			 *true* true
;;;			 *false* false
;;;			 *none* none

			 ;; Interface variable
			 "*DEFAULT-THEORY*"
			 
			 ;; THEORY Object
			 "THEORY-NAME"
			 "THEORY-CLAUSES"
			 "THEORY-INCONSISTENT-CLAUSES"
			 "THEORY-STATE-VARIABLE-HT"
			 "THEORY-PROPOSITION-HT"
			 "THEORY-UNIT-CLAUSES"

			 ;; PROPOSITION Object
			 "PROPOSITION-DESCRIPTION"
			 "PROPOSITION-TRUTH-VALUE"
			 "PROPOSITION-POSITIVE-CLAUSES"
			 "PROPOSITION-NEGATIVE-CLAUSES"
	    
			 ;; proposition related functions
			 "PROPOSITION-TO-TRUTH-ASSIGNMENT"
			 "PROPOSITION-SUPPORT"

			 ;; CLAUSE Object
			 "CLAUSE-SOURCE"
			 "CLAUSE-POSITIVE-PROPOSITIONS"
			 "CLAUSE-NEGATIVE-PROPOSITIONS"
			 "CLAUSE-CONSEQUENT"
			 "CLAUSE-SATISFYING-PROPOSITION"
			 "CLAUSE?"
			 "CLAUSE-REMOVED?"
			 "CLAUSE-IN-THEORY-UNIT-CLAUSE?"
			 "CLAUSE-INCONSISTENT?"
			 "CLAUSE-SATISFIED-BY"
			 "CLAUSE-IN-CONFLICT?"
;;; 		 print-clause

			 ;; CLAUSE related functions
			 "CLAUSE-DESCRIPTION"
			 "SET-SUPPORT-RELATION!"
			 "CLEAR-SUPPORT-RELATION!"
			 "DO-INSERTED-CLAUSES" ;; Macro
			 "DO-PROPOSITION-INSERTED-CLAUSES" ;; Macro
			 "DO-PROPOSITION-INSERTED-CLAUSES-W-POLARITY" ;; Macro
			 "DO-CLAUSE-PROPOSITIONS" ;; Macro
			 "UNIT-CLAUSE?"
			 "INCONSISTENT-CLAUSE?"
			 "NO-INCONSISTENT-CLAUSES?"
			 "GET-INCONSISTENT-CLAUSE"
			 "ADD-INCONSISTENT-CLAUSE"	;(should this be in the interface?)
			 "REMOVE-INCONSISTENT-CLAUSE" ;(should this be in the interface?)
			
			 ;; From prop-theory-manipulation.lisp
			 "THEORY"
			 "CLAUSE"
			 "PROPOSITION"
;;; 		 new-theory
			
			 "CLAUSE"
			 "NEW-CLAUSE"

			 "INSERT-CLAUSE"
;;;			 insert-constraint
			 "REMOVE-CLAUSE"
			 "INSERT-NEW-CLAUSE"
			 "INSERT-NEW-CLAUSES"

			 "GET-PROPOSITION"
			 "FIND-PROPOSITION"
	    
;;;			 assert!
;;;			 unassert!
			 "CLAUSE-TRUTH-ASSIGNMENT"
	    
			 "TRUTH-VALUE"
			 "SUPPORT"
			 "CONSEQUENCES"

;;;			 define-state-variable
;;;;;;			find-state-variable
;;;;;;	        insert-new-clauses-defining-variable!
	    
			 "STATE-CLAUSE"
			 "NEW-STATE-CLAUSE"
;;;			 new-constraint

;;;			 get-assignment
			 "FIND-ASSIGNMENT"

;;;			 assign!
;;;			 unassign!
			 
			 ;; From model-finding-defs.lisp
			 "CONFLICT-LEARNING-ON"
			 "CONFLICT-LEARNING-OFF"
			 "DYNAMIC-BACKTRACKING-ON"
			 "DYNAMIC-BACKTRACKING-OFF"
			 "RANDOM-ASSIGNMENT-ON"
			 "RANDOM-ASSIGNMENT-OFF"

			 ;; From model-finding.lisp
			 "CLEAR-SEARCH-STATE!"
			 "FIND-MODEL"
			 ;; Calls to the following should
			 ;; be replaced with calls to
			 ;; find-model
			 "FIND-CONSISTENT-ASSIGNMENT"	

			 ;; From consistency.lisp
			 "CONSISTENT?"
;;;			 consistent-with?
;;;			 inconsistent-with?
			 
			 ;; from conflicts.lisp
			 "CONFLICT-OF-INCONSISTENT-THEORY" ; New Module Interface
			 "UNIT-PROPAGATION-IMPLICANT-OF-CLAUSES" ; New Module Interface

			 ;; Retire the following
			 "THEORY-CONFLICTING-ASSIGNMENTS" ; Old ISAT interface
			 "THEORY-CONFLICT"			; Old ISAT interface
			 "CONSISTENT-THEORY?"		; Old ISAT interface
			 
			 ;; From compare.lisp
			 "EQUAL-MODELS"
			 "SET-CONTAINS-MODEL"
			 "EQUAL-MODEL-SETS"
			 "EQUAL-MODEL-SEQUENCES"
			 "EQUAL-CONFLICTS"
			 "EQUAL-CONFLICT"
			 "EQUAL-ASSIGNMENTS"

			 ;; From development.lisp
			 "THEORY-WITH-CLAUSES"
			 "THEORY-WITH-STATE-CLAUSES"

			 "FIND-PROPOSITION"
			 "GET-PROPOSITION"

			 "FIND-CLAUSE"
			 "FIND-CLAUSES-FROM-SOURCE"
			 "REMOVE-CLAUSES-FROM-SOURCE"

			 "FIND-PROPOSITION-TRUTH-VALUE"
			 "THEORY-ASSIGNMENT"
	    
			 "CONSISTENT-CLAUSES"
			 "FIND-MODEL-OF-CLAUSES"
			 "CONSISTENT-WITH-CLAUSES?"
			 "INCONSISTENT-WITH-CLAUSES?"
	    
			 "CONSISTENT-STATE-CLAUSES"
			 "FIND-MODEL-OF-STATE-CLAUSES"
			 "CONSISTENT-WITH-STATE-CLAUSES?"
			 "INCONSISTENT-WITH-STATE-CLAUSES?"
	    
			 "ISAT-STATISTICS-ON"
			 "LOG-ISAT-STATISTICS"
			 "INITIALIZE-ISAT-STATISTICS"
			 "PRINT-ISAT-STATISTICS"
			 "ISAT-STATISTICS-OFF"
			 "PRINT-ASSIGNMENT"
	    
			 "INITIALIZE-THEORY"
			 "INITIALIZE-INCONSISTENT-CLAUSES" ; (should this be external?)
			 ))
