;;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)

;;; ***********************************************************************
;;;
;;;                Incrementally Manipulating a Clausal Theory
;;;
;;; ***********************************************************************

;;; This file contains functions that insert and remove clauses from a
;;; Theory.  It depends on the structures and macros defined in 
;;; propositional-theories.  It also depends on unit-propagation to 
;;; update the logical consequences of insertion and removal.

;;; The external interface functions defined in this file are:
;;;
;;; o Creating New Theories
;;;
;;;   (MAKE-INSTANCE 'THEORY &KEY (NAME (GENTEMP "THEORY-")))

;;;     Creates a new theory named NAME.  The new theory is empty, i.e., 
;;;     it has no clauses and no propositions associated with it.  The
;;;     function has the side effect of setting *DEFAULT-THEORY* to be 
;;;     the newly created theory.  The function returns the newly 
;;;     created theory.

;;; o Associating a New Clauses with a Theory, WITHOUT inserting
;;;
;;;   NEW-CLAUSE (THEORY THEORY) SOURCE CLAUSE
;;;     Creates a new clause in THEORY whose source is SOURCE. 
;;;     CLAUSE is a list of literals, where each literal is either
;;;     <proposition-description> or (NOT <proposition-description>).
;;;     <proposition-description>, the description of a proposition, is an 
;;;     arbitary lisp structure, typically a list.  The function creates a new 
;;;     clause, but DOES NOT insert CLAUSE into THEORY, i.e., the 
;;;     clause is a removed clause.  A clause is inserted by calling 
;;;     INSERT-CLAUSE.  Note that this function does not check for 
;;;     duplicate clauses, but merely proceeds to create additional 
;;;     copies.  Any required propositions are created by the function, 
;;;     and pointers between the clause and its propositions are put 
;;;     in place.  The function returns the newly created clause.

;;; o Inserting and Removing Clauses that are associated with Theory
;;;
;;;   INSERT-CLAUSE ((THEORY THEORY) (CLAUSE CLAUSE))
;;;   INSERT-CONSTRAINT ((THEORY THEORY) (CLAUSE CLAUSE))

;;;     Inserts a currently removed CLAUSE, into THEORY,
;;;     checking the clause's consistency, and updating the unit 
;;;     propagation queue.  The CLAUSE itself must have been created 
;;;     earlier using NEW-CLAUSE.  If the clause is already inserted, 
;;;     it does nothing.  The function does not return anything 
;;;     meaningful. 

;;;   REMOVE-CLAUSE (THEORY THEORY) (CLAUSE CLAUSE)
;;;   REMOVE-CONSTRAINT (THEORY THEORY) (CLAUSE CLAUSE)

;;;     Removes CLAUSE from THEORY, and any 
;;;     consequences of this clause.  Finally, it updates the unit 
;;;     clauses of theory, which when PROPAGATE is called, is used to 
;;;     find alternative support for the removed consequences.  If the
;;;     clause is already removed, it does nothing.  The function does 
;;;     not return anything meaningful.

;;; o Creating AND Inserting Clauses into Theory:
;;;
;;;   INSERT-NEW-CLAUSE (THEORY THEORY) &key SOURCE CLAUSE
;;;        
;;;     Creates a new clause specified by CLAUSE, and INSERTS it 
;;;     into THEORY.  
;;;     CLAUSE must be a list or disjunction of literals, where 
;;;     each literal is <proposition-description> or 
;;;     (NOT <proposition-description>).  
;;;     <proposition-description>  is any Lisp object.  SOURCE can be any Lisp
;;;     object, and is typically used to store the source of the clause.  The 
;;;     function returns the  newly created clause.  Note that it does not 
;;;     check for duplicate  clauses, but merely proceeds to create additional 
;;;     copies.

;;;   INSERT-NEW-CLAUSES (THEORY THEORY) SOURCE CLAUSES
;;;         
;;;     Exactly like INSERT-NEW-CLAUSE, except that it creates and inserts 
;;;     a set of new clauses in THEORY with source being SOURCE.  The function 
;;;     always returns T.

;;; o Creating and Retrieving Propositions

;;;   GET-PROPOSITION THEORY PROPOSITION-DESCRIPTION
;;;     Returns a proposition, described by PROPOSITION-DESCRIPTION and 
;;;     associated with THEORY.  If a proposition with PROPOSITION-DESCRIPTION 
;;;     already exists in THEORY, it merely returns that proposition.  If no 
;;;     such proposition exists, then it creates a new proposition associated 
;;;     with THEORY that has the required description and returns it.

;;;   FIND-PROPOSITION (THEORY THEORY) PROPOSITION-DESCRIPTION
;;;     Returns the proposition with PROPOSITION-DESCRIPTION associated with 
;;;     THEORY.  If no such proposition exists in THEORY, it returns NIL.

;;; o (Un)Assigning the Truth of Propositions in Theory

;;;   ASSERT! (THEORY THEORY)(PROPOSITION PROPOSITION) 
;;;        &OPTIONAL (TRUTH-VALUE *TRUE*) SOURCE 

;;;     Assigns TRUTH-VALUE to PROPOSITION in THEORY, based on SOURCE.  
;;;     TRUTH-VALUE must be one of *true* or *false*.
;;;     Note: Avoids duplicate clauses with the same source.

;;;   UNASSERT! (THEORY THEORY)(PROPOSITION PROPOSITION) 
;;;        &OPTIONAL (TRUTH-VALUE *TRUE*) SOURCE 

;;;     Unassigns TRUTH-VALUE to PROPOSITION in THEORY, based on SOURCE.  
;;;     TRUTH-VALUE must be one of *true* or *false*.
;;;     Note: Avoids duplicate clauses with the same source.

;;;   CLAUSE-TRUTH-ASSIGNMENT CLAUSE
  
;;;     If CLAUSE asserts a truth assignment to a proposition, then
;;;     returns (list PROPOSITION TRUTH-VALUE)  
;;;     otherwise nil.

;;; o Looking up the Status of Propositions

;;;   TRUTH-VALUE (THEORY THEORY) PROPOSITION

;;;     Returns truth value of proposition, resulting from the closure of unit
;;;     clause resolution.

;;;   SUPPORT PROPOSITION &OPTIONAL (THEORY *DEFAULT-THEORY*)

;;;     Returns clause used during unit propagation
;;;     to determine the truth assignment to proposition.

;;;   CONSEQUENCES PROPOSITION &OPTIONAL (THEORY *DEFAULT-THEORY*)

;;;     Returns clauses that contain PROPOSITION and are used to
;;;     support another proposition.

;;; o Support for the State Logic Abstraction

;;;   State logic augments propositional logic with a set of
;;    finite domain variables and domains for each variable.
;;;   Propositions in state logic are assignments of the declared
;;;   variables to elements of their domain.
;;;   Propositional state logic automatically enforces exclusion and
;;;   exhaustion of the variable domains.

;;;   o State Variables

;;;   DEFINE-STATE-VARIABLE (THEORY THEORY) SOURCE STATE-VARIABLE-NAME DOMAIN 

;;;     Defining a new state variable for theory, ranging over domain.
;;;     VARIABLE-NAME is any lisp object, denoting a variable, and DOMAIN 
;;;     is a list of values that the variable may take on.
;;;     SOURCE is the object that created variable, and THEORY is a
;;;     propositional state logic theory that variable is part of.
;;;     Creates and inserts into THEORY a list of new clauses, stating 
;;;     that the values in DOMAIN denote mutually exclusive and 
;;;     collectively exhaustive assignments for VARIABLE-NAME.

;;;   FIND-STATE-VARIABLE THEORY STATE-VARIABLE-NAME 

;;;     Returns object representing STATE-VARIABLE-NAME for THEORY.
;;;     If STATE-VARIABLE-NAME is not defined in THEORY, then returns NIL.

;;;   o State Clauses

;;;   NEW-STATE-CLAUSE (THEORY THEORY) SOURCE &key CLAUSE
;;;
;;;   NEW-CONSTRAINT (THEORY THEORY) &key SOURCE CLAUSE
;;;     Creates and returns a new state clause for THEORY 
;;;     whose source is SOURCE. 

;;;     CLAUSE is a list of literals, each literal is a state assignment or its
;;;     negation, that is:
;;;     (= <variable> <value>) or (NOT (= <variable> <value>)), 
;;;     where <variable> is a state variable of THEORY and <value> is an 
;;;     element of the domain of <variable>.

;;;     This function creates a new clause, 
;;;     but DOES NOT insert CLAUSE into THEORY, i.e., the clause is a removed 
;;;     clause.  A clause is inserted by calling INSERT-CLAUSE.  Note 
;;;     this function that it does not check for duplicate clauses, 
;;;     but merely proceeds to create additional copies.  Any required 
;;;     propositions are created by the function, and pointers between
;;;     the clause and its propositions are put in place.  The function
;;;     returns the newly created clause.

;;;   o State Assignment Propositions

;;;   GET-ASSIGNMENT (THEORY THEORY) VARIABLE-NAME VALUE 

;;;     Returns the proposition object for (= VARIBLE-NAME VALUE) in THEORY.  
;;;     If the proposition already exists, it merely returns it.  Otherwise it
;;;     creates a new proposition and returns it.  This is the preferred way
;;;     of retrieving an assignment since it creates it only if it doesn't
;;;     already exist.

;;;   FIND-ASSIGNMENT THEORY VARIABLE-NAME VALUE

;;;     Returns the proposition object for (= VARIBLE-NAME VALUE) in THEORY.  
;;;     If no such proposition exists in THEORY, it returns NIL.

;;;   o Making and Removing Assignments.

;;;   ASSIGN! (THEORY THEORY) VARIABLE-NAME VALUE
;;;		&OPTIONAL (TRUTH-VALUE *TRUE*) (SOURCE NIL) 
;;;             
    
;;;     Makes assignment (= VARIABLE-NAME VALUE) have TRUTH-VALUE in THEORY, 
;;;     based on SOURCE.  

;;;     VARIABLE-NAME must be a defined variable in THEORY.
;;;     VALUE must be an element of the domain of VARIABLE-NAME.
;;;     TRUTH-VALUE must be one of *TRUE* or *FALSE*."

;;;   UNASSIGN! (THEORY THEORY) VARIABLE-NAME VALUE
;;;		&OPTIONAL (TRUTH-VALUE *TRUE*) (SOURCE NIL) 
;;;             
;;;     Removes truth-value for assignment (= VARIABLE-NAME VALUE) in THEORY, 
;;;     based on SOURCE.  

;;;     VARIABLE-NAME must be a defined variable in THEORY.
;;;     VALUE must be an element of the domain of VARIABLE-NAME.
;;;     TRUTH-VALUE must be one of *TRUE* or *FALSE*.

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			theory
;;;;;;			new-theory
;;;			
;;;			clause
;;;			new-clause
;;;
;;;			insert-clause
;;;			insert-constraint
;;;			remove-clause
;;;            insert-new-clause
;;;			insert-new-clauses
;;;
;;;			get-proposition
;;;			find-proposition
;;;	    
;;;			assert!
;;;			unassert!
;;;			clause-truth-assignment
;;;	    
;;;            truth-value
;;;            support
;;;            consequences
;;;
;;;			define-state-variable
;;;;;;			find-state-variable
;;;;;;	        insert-new-clauses-defining-variable!
;;;	    
;;;			state-clause
;;;			new-state-clause
;;;			new-constraint
;;;
;;;			get-assignment
;;;			find-assignment
;;;
;;;			assign!
;;;			unassign!
;;;			)
;;;		  :ISAT)
;;;  )

;;; To Do:
;;; o Move the code for state logic into a separate file and module.
;;; o The purpose of (CLAUSE-TRUTH-ASSIGNMENT CLAUSE) is unclear in
;;;   the interface documentation.  Clarify.

;;; o To inform unit propagation about assignments,
;;;   ISAT uses the back door of (un)assign-truth-value, 
;;;   while opsat uses (un)assert! together with find-assignment.

;;;   This raises three issues:
;;    o ISAT's back door needs to be a documented interface.
;;;   o The use of two different mechanisms for assigning truth
;;;     to propositions is needlessly complex, and leads to
;;;     confusion.
;;;   o (un)assert! is inefficient.  These fns record an assignment
;;;     by inserting unit clauses.  A clause needlessly consumes
;;;     a significant amount of memory, and the steps taken to go
;;;     from a clause insertion to a call to (un)assign-truth-value
;;;     is quite lengthy.  ISAT's internal approach of calling
;;;     (un)assign-truth-value directly, with a few added bookeeping steps
;;;     is far preferred.  Collapse these two approaches.  In doing so:
;;;       - When assign-truth-value is called, it needs to record the
;;;         source of the truth assignment (e.g., SAT, OPSAT etc.).
;;;       - The conflict extraction routine must be rewritten to extract
;;;         all assignments, based on this information.

;;; o Propositions are assigned by Assert and Unassert by inserting
;;;   unit clauses, this introduces a lot of over head.  ISAT, on the
;;;   other hand, uses assign-truth-value and unassign-truth-value.

;;; Things to do from top of file:

;;; o Add a function:
;;;      (check-assignment variable-name value &optional (theory ..))
;;;   that checks that all propositions specify assignments
;;;   to defined variables and their assigned values are from the
;;;   variable's declare domain.

;;;   Call this function whenever an assignment is being looked up, 
;;;   or when a state clause is being introduced.

;;; ***********************************************************************
;;;
;;;                  External Interface
;;;
;;; ***********************************************************************

(defmethod get-proposition 
    ((theory theory) proposition-description)

  " Returns the proposition object of theory described 
    by PROPOSITION-DESCRIPTION. If the object already exists, 
    it is returned; otherwise, a new proposition object
    is created, associated with theory and returned.  

    Note: This is the preferred way to create a new proposition 
    since it avoids duplication."

  (or (find-proposition theory proposition-description)
	  (make-instance 'proposition
		:theory theory
		:description proposition-description)))

(defmethod new-constraint
    ((theory theory) &key source clause (can-be-relaxed? t))
  
  "Creates and returns a propositional clause that is
   associated with THEORY.
   CLAUSE describe is a description of the clause.
   SOURCE specifies the creator of clause.
   CAN-RELAX? is true iff clause can be relaxed."

  (make-instance 'state-clause 
	:source source
	:constraint-system theory
	:clause clause
	:can-be-relaxed? can-be-relaxed?))

(defmethod insert-constraint ((theory theory)
							  (clause clause))
  
  "This function inserts a removed CLAUSE into THEORY. It checks
   CLAUSE's consistency, and updates the unit propagation queue,
   which is propagated during the next call to PROPAGATE, not
   performed by this function.

   CLAUSE itself must have been created earlier using NEW-CLAUSE.
   If the clause is already inserted, it does nothing.  

   The function is called for effect, it does not return anything 
   meaningful."

  ;; Initialize the status of clause: 
  ;; the CONSEQUENT, SATISFYING-PROPOSITION 
	 ;;; and UNASSIGNED-LITERAL-COUNT.
  (setup-inserted-clause-status clause)

  ;; Finally, update the set of UNIT-CLAUSES and 
  ;; INCONSISTENT-CLAUSES.  

  ;; Clear flags IN-THEORY-UNIT-CLAUSE? and INCONSISTENT?.
  (setf (clause-in-theory-unit-clause? clause) nil)
  (setf (clause-inconsistent? clause) nil)
	 
  ;; Now check and update theory unit and inconsistent clauses
  ;; for clause.
  (if (unit-clause? clause)
	  ;; If it is a unit clause, add it to the unit clauses of theory.
	  (add-to-unit-clauses clause theory)

	;; Clause is not unit, see if it is inconsistent
	(when (inconsistent-clause? clause)

	  ;; It is, add clause to the list of inconsistent clauses.
	  (add-inconsistent-clause clause theory))))

;;; To Do:
;;; o For clause learning, all learned clauses derived from CLAUSE,
;;;   must be removed when CLAUSE is removed.  How expensive
;;;   is it to perform this update?

(defmethod remove-constraint
	((theory theory) (clause clause))
  
  "Removes CLAUSE from THEORY, removes any consequences of
   CLAUSE, and updates the unit clauses of theory,
   which when PROPAGATE is called, is used to find alternative 
   support for the removed consequences.  Propagate IS NOT
   called by this function.

   If CLAUSE is already removed, this function does nothing.

   The function is called for effect, it does not return 
   anything meaningful."

  ;; Nothing needs to be done if the clause is already removed.
  (when *isat-statistics-p*
	(incf *remove-clause-counter*))

  ;; undo any propagations from the clause,
  (undo-clause-propagation clause theory)

  ;; and make sure that clause is removed from the set of unit clauses and
  ;; the set of unsatisfied clauses.
  (remove-from-unit-clauses clause theory)
  (remove-inconsistent-clause clause theory)
    
  ;; Note that when this function returns, the unassigned-
  ;; literal-count of clause is no longer up to date.
  T)

;;;(defmethod insert-clause ((theory theory) 
;;;						  (clause clause))
;;;  (insert-constraint theory clause))

;;;(defmethod remove-clause ((theory theory) (clause clause))
;;;  (remove-constraint theory clause))


;;; Note:
;;; o Use make-instance instead.
;;;(defun new-theory (&key (name (gentemp "THEORY-"))
;;;						(volatile-clause-sources nil))
;;;  
;;;  "Creates a new theory named NAME.  The new theory is 
;;;   empty,that is, it has no clauses or propositions 
;;;   associated with it.
;;;
;;;   The function has the side effect of setting 
;;;   *DEFAULT-THEORY* to be the newly created theory."
;;;
;;;  (declare (ignore volatile-clause-sources))
;;;  (make-instance 'theory
;;;	:name name))		   

;;; Note:
;;; o Use make-instance instead.
;;;(defmethod new-clause
;;;    ((theory theory) source &key clause)
;;;  (new-constraint theory :source source :clause clause))

;;; Note:
;;; o Use make-instance instead.
;;;(defmethod new-state-clause
;;;    ((theory theory) &key source clause)
;;;  (make-instance 'state-clause
;;;	:constraint-system theory
;;;	:source source
;;;	:clause clause))

(defmethod truth-value 
	((theory theory) (proposition proposition))
  
  " Returns truth value of proposition, resulting from unit
    propagation (the closure of unit clause resolution)."

  ;; First make sure the unit propagation is up to date.
  (when (has-unit-clauses? theory)
    (propagate theory))
  (proposition-truth-value proposition))

(defmethod support 
	((proposition proposition) &optional (theory *DEFAULT-THEORY*))
  
  " Returns clause used during unit propagation (the closure 
    of unit clause resolution) to determine the truth 
    assignment to proposition."

  ;; First make sure the unit propagation is up to date.
  (when (has-unit-clauses? theory)
    (propagate theory))
  (proposition-support proposition))

;;; To Do:
;;; o Introduce the following function once proposition-consequences 
;;;   has been written.
;;;(defun consequences (proposition &optional (theory *DEFAULT-THEORY*))
;;;  
;;;  " Returns clauses that are the support for some proposition, 
;;;    resulting from the closure of unit clause resolution, 
;;;    and that contain proposition as an antecedent."
;;;  ;; First make sure the unit propagation is up to date.
;;;  (when (has-unit-clauses? theory)
;;;    (propagate theory))
;;;  (proposition-consequences proposition))

(defmethod insert-new-clause 
    ((theory theory) &key source clause)
  
  "Creates a new clause specified by CLAUSE, and 
   inserts it into THEORY.
   CLAUSE is a list of literals or disjunction (OR . <literals>), 
   where each literal is <proposition-description>
   or (NOT <proposition-description>).  <proposition-description> is 
   any Lisp object.  
   SOURCE can be any Lisp object, and is typically used to store 
   the source of the clause.
   The function returns the newly created clause.  Note that it 
   does not check for duplicate clauses, but merely proceeds to 
   create additional copies."
  
  ;; create the clause
    
  (let ((cl (make-instance 'clause 
			  :source source
			  :constraint-system theory
			  :clause clause)))

	;; add it into the theory
	(insert-constraint theory cl)

	;; and return it
    (values cl)))

;;; Optimization:
;;; o rewrite the following so that it assigns using assign-truth-value,
;;;   rather than by inserting a unit clause.

(defmethod assert!
    ((theory theory)(proposition proposition)
	 &optional (truth-value *true*) (source nil))

  "Assigns TRUTH-VALUE to PROPOSITION in THEORY, based on SOURCE.  
   Note: Avoids duplicate clauses with the same source.

   TRUTH-VALUE must be one of *true* or *false*."

  (debug-print-for :isat t "Asserting ~A as ~A in ~A, by ~A." 
				   proposition truth-value theory source)

  (assign-truth-value proposition truth-value source theory))

;;; Optimization:
;;; o rewrite the following so that it unassigns using unassign-truth-value,
;;;   rather than by removing a unit clause.

;;; To Do:
;;; o unassign-truth-value doesn't require the truth value to be provided.
;;;   Likewise, is there any reason it needs to be provided here?

(defmethod unassert!
    ((theory theory)(proposition proposition)
	 &optional (truth-value *true*) (source nil))

  "Unassigns TRUTH-VALUE to PROPOSITION in THEORY, based on SOURCE.  
   Note: Avoids duplicate clauses with the same source.

   TRUTH-VALUE must be one of *true* or *false*."

  (debug-print-for :isat t "Removing assertion ~A as ~A in ~A, by ~A." 
				   proposition truth-value theory source)

  (unassign-truth-value proposition source theory))

;;;   *****************************************************************
;;;   ***
;;;   *** Manipulating expressions in a state logic
;;;   ***
;;;   *****************************************************************

(defmacro find-state-variable 
    (theory state-variable-name)
  
  "Returns object representing STATE-VARIABLE-NAME for THEORY.
   If STATE-VARIABLE-NAME is not defined in THEORY, then returns
   NIL"
  `(gethash ,state-variable-name
			(theory-state-variable-ht ,theory)))

(defmethod define-state-variable 
    ((theory theory) source variable-name domain)

  "Defining a new state variable for theory, ranging over domain.
   VARIABLE-NAME is any lisp object, denoting a variable, and DOMAIN 
   is a list of values that the variable may take on.
   SOURCE is the object that created variable, and THEORY is a
   propositional state logic theory that variable is part of.
   Creates and inserts into THEORY a list of new clauses, stating 
   that the values in DOMAIN denote mutually exclusive and 
   collectively exhaustive assignments for VARIABLE-NAME."

  ;; Check for Old and Warn.
  (let ((old-state-variable (find-state-variable theory variable-name)))
    (unless (null old-state-variable)
      (warn "State variable ~A already defined for theory ~A, ~
             to have domain ~A and source ~A. ~
             ~% Redefining to have domain ~A and source ~A." 
			variable-name theory 
			(state-variable-domain old-state-variable)
			(state-variable-source old-state-variable)
			domain source)))
  
  ;; Record new.
  (debug-print-for :isat t "Defining state variable ~A with domain ~a for ~A, ~
                  by source ~A." 
				   variable-name domain theory source)
  (let ((state-variable (make-instance 'state-variable
						  :name variable-name
						  :domain domain
						  :source source)))
    (setf (find-state-variable theory variable-name)
      state-variable)
    (insert-new-clauses-defining-variable! 
     theory source variable-name domain)
    state-variable))

;;; Optimization:
;;; o may only want to call check-state-clause-description? in 
;;;   debugging mode.
;;; o If check-state-clause-description? fails, shouldn't this
;;;   generate an error?

(defmethod insert-new-state-clause 
    ((theory theory) &key source clause)
  
  "Creates a new clause specified by CLAUSE and inserts it 
   into THEORY. CLAUSE is a list of state-literals or a 
   disjunction (OR . <state-literals>).  Each state-literal is a state 
   assignment or its negation, that is:
   (= <variable> <value>) or (NOT (= <variable> <value>)), 
   where <variable> is a state variable of THEORY and <value> is an element
   of the domain of <variable>.

   SOURCE can be any Lisp object, and is typically used to store the source 
   of the clause.
   The function returns the newly created clause.  Note that it does 
   not check for duplicate clauses, but merely proceeds to create 
   additional copies."
  
  ;; create the clause
    
  (let ((cl (make-instance 'state-clause 
			  :constraint-system theory
			  :source source
			  :clause clause)))

	;; add it into the theory
    (insert-constraint theory cl)

	;; and return it
    (values cl)))

(defmacro find-assignment
    (theory variable-name value)

  " Returns an existing proposition object for (= VARIBLE-NAME VALUE) 
    in THEORY. If the assignment does not exist, it returns NIL."
  
  `(find-proposition ,theory
					 (make-assignment :variable-name ,variable-name
									  :value ,value)))

(defmethod get-assignment
    ((theory theory) variable-name value)

  " Returns a proposition object for (= VARIABLE-NAME VALUE) in THEORY.  
    If the proposition already exists, it merely returns it.  Otherwise it
    creates a new proposition and returns it.  This is the preferred way
    of retrieving an assignment since it creates one only if it doesn't
    already exist.

   VARIABLE-NAME must be a defined variable in THEORY.
   VALUE must be an element of the domain of VARIABLE-NAME.

   If these conditions are not satisfied, then a warning is issued and nil
   is returned."
  
  (when (check-state-variable-assignment? theory variable-name value)
    (get-proposition theory
					 (make-assignment :variable-name variable-name
									  :value value))))

(defmethod assign! ((theory theory) variable-name value
					&optional (truth-value *true*) (source nil))
    
  "Makes assignment (= VARIABLE-NAME VALUE) have TRUTH-VALUE in THEORY, 
   based on SOURCE.  

   VARIABLE-NAME must be a defined variable in THEORY.
   VALUE must be an element of the domain of VARIABLE-NAME.
   TRUTH-VALUE must be one of *TRUE* or *FALSE*.

  If these conditions are not satisfied, then an error is issued,
  no assertion is made.

  Returns T iff success."

  (let ((assignment (get-assignment theory variable-name value)))
    (unless (null assignment)
      (cond ((not (truth-value? truth-value))
			 (error "~A is not a valid truth value while assigning ~A to ~A ~
                 for theory ~A by ~A."  truth-value value variable-name 
				 theory source)
			 nil)
			(:else (assert! theory assignment truth-value source)
				   T)))))

(defmethod unassign! ((theory theory) variable-name value
					  &optional (truth-value *true*) (source nil))
    
  "Removes TRUTH-VALUE for assignment (= VARIABLE-NAME VALUE) in THEORY, 
   based on SOURCE.  

   VARIABLE-NAME must be a defined variable in THEORY.
   VALUE must be an element of the domain of VARIABLE-NAME.
   TRUTH-VALUE must be one of *TRUE* or *FALSE*.

  If these conditions are not satisfied, then an error is issued,
  no assertion is removed

  Returns T iff success."

  (let ((assignment (get-assignment theory variable-name value)))
    (unless (null assignment)
      (cond ((not (truth-value? truth-value))
			 (error "~A is not a valid truth value while unassigning ~A to ~A ~
                 for theory ~A by ~A."  truth-value value variable-name 
				 theory source)
			 nil)
			(:else (unassert! theory assignment truth-value source)
				   T)))))

;;; ***********************************************************************
;;;
;;;                            Support functions
;;;
;;; ***********************************************************************

;;; Note:
;;; o The following is here for backward compatibility.
;;;   Use make-instance instead.
(defmethod new-proposition ((theory theory) description)
  
  " Creates a new proposition in THEORY.  If it already exists, then it is
    overwritten, and a new proposition is created.  Hence, the caller is
    required to make sure that there is no existing proposition with that
    name.  It returns the new proposition object."
  
  ;; Create the proposition
  (make-instance 'proposition 
	:description description
	:theory theory))

(defmethod connect-propositions-to-clause! ((clause clause))
  
  "Point from the propositions occurring in CLAUSE to CLAUSE."
  
  (dolist (proposition (clause-positive-propositions clause))
	;; PROPOSITION occurs positively in CLAUSE
    (pushnew clause (proposition-positive-clauses proposition) 
             :test #'eql)
	;;(incf (proposition-num-positive-clauses proposition))
	(setf (proposition-num-positive-clauses proposition)
	  (length (proposition-positive-clauses proposition)))
	)
  (dolist (proposition (clause-negative-propositions clause))
	;; PROPOSITION occurs negative in CLAUSE
    (pushnew clause (proposition-negative-clauses proposition) 
             :test #'eql)
	;;(incf (proposition-num-negative-clauses proposition))
	(setf (proposition-num-negative-clauses proposition)
	  (length (proposition-negative-clauses proposition)))
	)
  T)
 
(defun literal-description-proposition (literal-description)

  " A positive LITERAL-DESCRIPTION is <proposition-description>, 
    while a negative LITERAL-DESCRIPTION is (<NOT <proposition-description>).
    Function returns two values: 
      1) The proposition contained in LITERAL, and
      2) the polarity of LITERAL, T if and only if LITERAL is positive."

  (if (negation? literal-description)
      (values (negand literal-description) nil)
    (values literal-description t)))

;;; Note:  The following sets up unit propagation and needs
;;;        to be modified if a different unit propagation
;;;        algorithm is employed, such as watched literals.

;;; To Do:
;;; o put this function or the part related to unit propagation
;;;   setup in unit-propagation.lisp.
(defmethod setup-inserted-clause-status ((clause clause))
  
  "Sets up clause status, under the assumption that clause has not
   been propagated yet.
 
   Clears the consequent of CLAUSE.  
   Sets the unassigned-literal-count and satisfied-by slots 
   of CLAUSE,based on the truth values of the propositions in 
   CLAUSE."
  
  ;; Clear the clause's CONSEQUENT

  (debug-print-for :isat t "Initializing status of clause ~A." clause)

  (let ((p (clause-consequent clause)))
    (unless (null p)
      (debug-print-for :itms t
					   "Removing clause as support for proposition ~A." 
					   p)
      (clear-support-relation! p clause)))

  ;; Counts the number of unassigned literals 
  ;; and finds a proposition satisfying clause.
  (debug-print-for :itms t
				   "Counting unassigned literals and checking satisfaction.")
  (let ((unassigned-literal-count 0)
		(satisfied-by nil)
		(truth-value nil))

    (do-clause-propositions (proposition polarity clause)
      
      (setq truth-value
		(proposition-truth-value proposition))
      
      (cond ((eql truth-value polarity)
			 ;; Literal is true, hence satisfying clause.
			 (when (null satisfied-by)
			   (debug-print-for :itms t "Satisfied by literal ~A."
								proposition)
			   (setq satisfied-by proposition)))
	    
			((eql truth-value (opposite-truth polarity))
			 ;; literal is false, do nothing.
			 (debug-print-for :itms t "Literal ~A is false."
							  proposition))
	    
			(:otherwise
			 ;; Literal is unassigned.
			 (debug-print-for :itms t "Literal ~A is unassigned."
							  proposition)
			 (incf unassigned-literal-count))))

	;; Record unassigned literal count and satisfying proposition.
	;; Note: the satisfying proposition CANNOT be a consequent, since its 
	;; truth value wasn't determined by propagating CLAUSE.
    (debug-print-for :itms t
					 "Clause satisfied-by ~A, unassigned literals = ~A."
					 satisfied-by unassigned-literal-count)
    (setf (clause-unassigned-literal-count clause)
      unassigned-literal-count)
    (setf (clause-satisfying-proposition clause) satisfied-by)))

;;; Optimization:
;;; o Make sure that calls to hash, such as these, are not performed
;;;   by an inner loop function.
;;; o Currently it is called by Opsat.

(defmacro find-truth-assignment-clause
    (theory proposition &optional (truth-value *true*) (source nil))

  "Returns an existing clause in THEORY whose source is SOURCE, and which 
   states that PROPOSITION has TRUTH-VALUE.  If the 
   literal clause does not exist, it returns NIL.
   PROPOSITION is a proposition object.
   truth-value must be one of *TRUE* or *FALSE*."

  `(gethash (list ,proposition ,truth-value ,source) 
			(theory-truth-assignment-clause-ht ,theory)))

(defmethod get-truth-assignment-clause
    ((theory theory)(proposition proposition)
	 &optional (truth-value *true*) (source nil))

  "Returns a clause in THEORY whose source is SOURCE, and which 
   states that PROPOSITION has TRUTH-VALUE.  The function creates a new 
   clause, but DOES NOT insert CLAUSE into THEORY.  Unlike 
   NEW-TRUTH-ASSIGNMENT-CLAUSE, this avoids duplication by 
   returning a canonical clause.
  
   PROPOSITION is a proposition object.
   truth-value must be one of *TRUE* or *FALSE*."

  (or (find-truth-assignment-clause theory proposition truth-value source)
      (setf 
		  (find-truth-assignment-clause theory proposition truth-value source)
		(make-instance 'unit-clause 
		  :constraint-system theory
		  :source source
		  :truth-value truth-value
		  :proposition proposition))))

;;; Note:
;;; o The following is here for backward compatibility.
;;;   Use make-instance instead.
(defmethod new-truth-assignment-clause
    ((theory theory) (proposition proposition)
	 &optional (truth-value *true*) (source nil))
  
  "Returns a new clause in THEORY whose source is SOURCE, and which 
   states that PROPOSITION has TRUTH-VALUE.  The function creates a new 
   clause, but DOES NOT insert CLAUSE into THEORY.

   PROPOSITION is a proposition object.
   TRUTH-VALUE must be one of *TRUE* or *FALSE*."
  
  (make-instance 'unit-clause 
	:constraint-system theory
	:source source
	:truth-value truth-value
	:proposition proposition))

(defmethod clause-truth-assignment ((clause clause))
  
  "If CLAUSE asserts a truth assignment to a proposition, then
   returns (values proposition truth-value), otherwise (values nil nil)."

  (let ((positive-propositions (clause-positive-propositions clause))
		(negative-propositions (clause-negative-propositions clause)))
    
    (if (null positive-propositions)
	
		(if (length-1? negative-propositions)
			(list (first negative-propositions) *false*)
		  nil)
      
      (if (and (length-1? positive-propositions)
			   (null negative-propositions))
	  
		  (list (first positive-propositions) *true*)
	
		nil))))

(defmethod check-state-clause-description? 
	((theory theory) state-clause-description)
  
  "Checks that STATE-CLAUSE-DESCRIPTION describes a valid state clause
   over the state variables of THEORY.  Returns T if valid, otherwise nil.

  A clause is valid if all propositions are state assignments 
  (= <variable> <value>), where <variable> is a defined state variable for
  THEORY, and <value> is an element of the domain of <variable>."

  (loop with state-clause? = T
      with state-variable
      with value
      for literal-description in state-clause-description
      for proposition-description = 
		(literal-description-proposition literal-description)

      do (if (not (assignment-expression? proposition-description))
	     
			 (progn
			   (error "Proposition ~A is not an assignment in clause ~A"
					  proposition-description state-clause-description)
			   (setq state-clause? nil))
	   
		   (progn
			 (setq state-variable
			   (find-state-variable 
				theory
				(assignment-variable proposition-description)))

			 (if (null state-variable)
		 
				 (progn
				   (error "Variable-name ~A, appearing in clause ~A, is not a ~
                          defined state variable for theory ~A" 
						  (assignment-variable proposition-description) 
						  state-clause-description theory)
				   (setq state-clause? nil))
	       
			   (progn
				 (setq value (assignment-value proposition-description))
				 (unless (value-in-domain? state-variable value)
				   (error "Value ~A, appearing in assignment ~A of clause ~
                          ~A, is not a member of the domain of ~
                          ~% state variable ~A for theory ~A." 
						  value proposition-description
						  state-clause-description state-variable theory)
				   (setq state-clause? nil))))))
      finally (return state-clause?)))

(defmethod check-state-variable-assignment? 
	((theory theory) state-variable-name value)
  (let ((state-assignment? T)
		(state-variable (find-state-variable theory state-variable-name)))
    (when (null state-variable)
      (setq state-assignment? nil)
      (error "Variable ~A, being assigned ~A, is not a defined state ~
                variable for theory ~A" state-variable-name value theory))

    (unless (value-in-domain? state-variable value)
      (setq state-assignment? nil)
      (error "Value ~A, appearing in an assignment to state variable ~A, ~
              is not a member of ~A 's domain, for theory ~A." 
			 value state-variable state-variable theory))
	state-assignment?))

(defmethod insert-new-clauses-defining-variable!
    ((theory theory) source variable-name domain)

  " VARIABLE-NAME is any lisp object, denoting a variable, and DOMAIN 
    is a list of values that the variable may take on.
    SOURCE is the object that created variable, and THEORY is a
    propositional state logic theory that variable is part of.
    Creates and inserts into THEORY a list of new clauses, stating 
    that the values in DOMAIN denote mutually exclusive and 
    collectively exhaustive assignments for VARIABLE-NAME."

  (loop for clause-description
      in (clause-descriptions-defining-variable variable-name domain)
      do (insert-constraint
		  theory
		  (make-instance 'state-clause 
			:source source
			:constraint-system theory
			:clause clause-description))))

(defun clause-descriptions-defining-variable (variable-name domain)
  
  " VARIABLE-NAME is any lisp object, denoting a variable, and DOMAIN 
    is a list of values that the variable may take on.
    Returns a list of clause descriptions, stating that the values in 
    DOMAIN denote mutually exclusive and collectively exhaustive
    assignments for VARIABLE-NAME."

  (loop with clause-descriptions = nil
      for value in domain
      for assignment = (make-assignment :variable-name variable-name
										:value value)
      collecting assignment into assignments
      collecting `(NOT ,assignment) into negated-assignments
      finally
		;; States that variable needs to take on one of  the values in domain
		;; (i.e., they are collectively exhaustive)
		(push assignments clause-descriptions)
		;; States that variable cannot take on two or more values in domain.
		;; (i.e., they are mutually exclusive)
		(loop for remaining-neg-assignments on negated-assignments
			for na1 = (first remaining-neg-assignments)
			do (loop for na2 in (rest remaining-neg-assignments)
				   do (push (list na1 na2) clause-descriptions)))
		(return clause-descriptions)))

