;;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)

;;; Things To Do:
;;; o Some of the functions at the end of this file should be taken out
;;;   of the external interface.
;;; o Be clear in the documentation here which functions are part of the module
;;;   interface, and which ones are part of the external interface.
;;; o Replace :not-split-on with a constant variable.
;;; o Look for macros and supporting functions contained in other
;;;   files that belong here.
;;; o Document any undocumented slots, as indicated below.

;;; ***********************************************************************
;;;
;;;                Clausal Theory and related Objects and Accessors
;;;
;;; ***********************************************************************

;;; This file contains objects, macros and functions for constructing 
;;; and accessing parts of a propositional theories.  The file depends on 
;;; the macros defined in bit-manipulation.lisp and utils.lisp.
;;; 
;;; A THEORY is a set of CLAUSES, where each clause is a disjunction of 
;;; LITERALS.  A literal is either a proposition or the negation of a 
;;; proposition. A proposition is a primitive token that can be either 
;;; true or false.  A positive literal is true if and only if the 
;;; corresponding proposition is true.
;;;
;;; The Constants, Variables, Objects and accessors defined in this file are:
;;;
;;; Constants:
;;;   *TRUE*           : a symbol denoting the true truth assignment.
;;;   *FALSE*           : a symbol denoting a false truth assignment.

;;; Variables:
;;;
;;;   *DEFAULT-THEORY* : a variable that holds the most recently created 
;;;                      theory.
;;;
;;; Objects:

;;;   This file defines objects for THEORY, PROPOSITION, and CLAUSE, 
;;;   and associated slot accessors:

;;;   THEORY object slots:
;;;
;;;     NAME                  Lisp object denoting the theory's name.
;;;     PROPOSITION-HT        A hash table that maps proposition names 
;;;                           to proposition data structures.
;;;     CLAUSES               List of all clauses associated with the
;;;                           theory, including clauses that are 
;;;                           currently in the theory, and clauses that 
;;;                           are currently removed.
;;;     UNIT-CLAUSES          List of inserted clauses that have exactly
;;;                           one unassigned literal, with all other
;;;                           literals being false.
;;;     INCONSISTENT-CLAUSES  List of inserted clauses that have all false
;;;                           literals in the current truth assignment
;;;     ASSIGNED-PROPOSITIONS Propositions that have been assigned a truth 
;;;                           value the satisfiability procedure.
;;;     UNASSIGNED-PROPOSITIONS Propositions that have not yet been assigned a
;;;                            truth value by the satisfiability procedure.
;;;
;;;   PROPOSITION object slots:
;;;
;;;     NAME                : Arbitary lisp object denoting PROPOSITION's name.
;;;                           Typically a list.  A theory may have at most one
;;;                           proposition of a given name.
;;;     POSITIVE-CLAUSES    
;;;     NEGATIVE-CLAUSES    : Lists of clauses containing proposition as a
;;;                           positive or negative literal, respectively.
;;;                           The lists include both clauses that are 
;;;                           currently in the theory, and those
;;;                           that have been removed.  
;;;     TRUTH-VALUE         : The current truth assignment to 
;;;                           PROPOSITION.  Its possible values are *TRUE*
;;;                           if its assigned true, *FALSE* if its assigned
;;;                           false, and NIL if it is currently unassigned.
;;;     REMAINING-VALUE       Used by the Satisfiability procedure to keep 
;;;                           track of the truth values yet to be explored.  
;;;                           Value is :not-searched (can take on either value,
;;;                           *TRUE*, *FALSE* or nil (no remaining value).
;;;
;;;   Additional Proposition operations:
;;;
;;;   PROPOSITION-SUPPORT (PROPOSITION)
;;;     This function returns the clause whose CONSEQUENT is PROPOSITION,
;;;     if such a clause exists, and NIL otherwise.
;;;
;;;   CLAUSE object slots:
;;;
;;;     SOURCE               :Any Lisp object, specifies the source of clause,
;;;                           may, for example, be an object that clause is 
;;;;                          part of or the type of the clause.
;;;     POSITIVE-PROPOSITIONS
;;;     NEGATIVE-PROPOSITIONS : Lists of propositions that occur positively
;;;                           or negatively, respectively, in CLAUSE.
;;;     UNASSIGNED-LITERAL-COUNT : The number of literals in CLAUSE
;;;                           that are currently not assigned a truth
;;;                           value.  The value can be any non-negative
;;;                           integer.  Currently, the maximum value of
;;;                           this integer is 256.
;;;     CONSEQUENT            : Proposition whose truth assignment was deduced
;;;                           from clause by unit clause resolution.  It is
;;;                           NIL if clause hasn't been unit resolved.
;;;                           
;;;     SATISFYING-PROPOSITION : Proposition in CLAUSE whose truth value
;;;                           satisfies CLAUSE.  If CONSEQUENT is
;;;                           non-NIL, then the SATISFYING-PROPOSITION is 
;;;                           always the same as CONSEQUENT, and cannot be 
;;;                           changed independent of the CONSEQUENT..
;;;     REMOVED?            : Says whether the clause is currently removed
;;;                           from its theory or not.  Can be T, meaning
;;;                           removed, or NIL, meaning inserted.
;;;     IN-THEORY-UNIT-CLAUSE? : Says whether or not CLAUSE is currently
;;;                           stored in the unit clauses of its theory.
;;;     INCONSISTENT?        : Says whether or not the current assignment to
;;;                           truth values makes CLAUSE inconsistent.
;;;     IN-CONFLICT?         : Says whether or not CLAUSE is currently
;;;                           in a conflict.  A conflict is a set of
;;;                           inconsistent clauses, i.e., a set of clauses
;;;                           that together entail an inconsistency.
;;;
;;; Additional Clause operations:
;;;
;;;   DO-INSERTED-CLAUSES ((CLAUSE CLAUSES) &BODY BODY)
;;;     A macro that executes BODY for each inserted clause in CLAUSES.  
;;;     Behaves exactly like dolist, and permits early return.
;;;
;;;   CLAUSE-DESCRIPTION (CLAUSE)
;;;     Returns the description for CLAUSE.  The CLAUSE's description is
;;;     the list of descriptions of literals in the CLAUSE.  The description is
;;;     consed up from the CLAUSE's literals whenever required, and is not 
;;;     explicitly stored in the CLAUSE.  This function is not expected to be 
;;;     used much, hence we've chosen a version that creates the description 
;;;     on the fly rather than storing it.
;;;
;;;   DO-CLAUSE-PROPOSITIONS ((PROPOSITION CLAUSES) &BODY BODY)
;;;     Executes BODY for each proposition in CLAUSE.
;;;     Behaves exactly like dolist.
;;;
;;; o Managing the inconsistent clauses of theory
;;;
;;;   INCONSISTENT-CLAUSE? (CLAUSE)
;;;     Checks to see if CLAUSE is an inconsistent clause.  A
;;;     clause is inconsistent if all its literals are false.  
;;;     Does not check to see whether or not CLAUSE is currently inserted.
;;;
;;;   GET-INCONSISTENT-CLAUSE (THEORY)
;;;     Returns an inconsistent clause of THEORY.  If there is no
;;;     inconsistent clause, it returns NIL.
;;;
;;;   NO-INCONSISTENT-CLAUSES? (THEORY)
;;;     Returns T if and only if there are no inconsistent
;;;     clauses in theory.
;;;
;;;   ADD-INCONSISTENT-CLAUSE (CLAUSE THEORY)
;;;     Adds CLAUSE to the set of inconsistent clauses of THEORY, if it 
;;;     isn't already in the set.  Does not return anything meaningful.
;;;
;;;   REMOVE-INCONSISTENT-CLAUSE (CLAUSE THEORY)
;;;     Removes CLAUSE from the set of unsatisfied clauses of THEORY, if it
;;;     is currently in the set.  Does not return anything meaningful.
;;;

;;; ***********************************************************************
;;;
;;;                Additional Module Interface Functions
;;;
;;; ***********************************************************************
;;;
;;; o Managing the unit clauses

;;;  HAS-UNIT-CLAUSES? (THEORY)
;;;     Returns T if theory contains unit clauses, and NIL
;;;     otherwise.
;;;
;;;   UNIT-CLAUSE? (CLAUSE)
;;;     Checks to see if CLAUSE is a unit clause.  A clause
;;;     is a unit clause if it has one unassigned literal, and all other
;;;     literals are assigned false.  The function does not check to see 
;;;     whether or not CLAUSE is currently inserted.
;;;
;;;   GET-UNIT-CLAUSE (THEORY)
;;;     Returns a clause in the unit clauses of THEORY.  If there is
;;;     no unit clause in theory, it returns NIL.
;;;
;;;   add-to-unit-clauses (CLAUSE THEORY)
;;;     Adds CLAUSE to the set of unit clauses of THEORY, if it isn't 
;;;     already in the set.  Does not return anything meaningful.
;;;
;;;   remove-from-unit-clauses (CLAUSE THEORY)
;;;     Removes CLAUSE from the unit clauses of THEORY, if it is in
;;;     theory's unit clauses.  Does not return anything meaningful.
;;;

;;; FIX:
;;; o mtu::declare-debug-print-system!
;;;   has been exported from MTU in its package definition,
;;;   and included in bfig.  Why isn't the function external (: vs ::)
;;;   and why isn't it part of :ISAT?

(mtu::declare-debug-print-system! :ITMS)
(mtu::declare-debug-print-system! :ISAT)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			;;; truth
;;;			*true* true
;;;			*false* false
;;;			*none* none
;;;
;;;			;; Interface variable
;;;            *DEFAULT-THEORY*
;;;
;;;			;; THEORY Object
;;;			theory-name
;;;			theory-clauses
;;;			theory-inconsistent-clauses
;;;			theory-state-variable-ht
;;;			theory-proposition-ht
;;;			theory-unit-clauses
;;;
;;;			;; PROPOSITION Object
;;;			proposition-description
;;;			proposition-truth-value
;;;			proposition-positive-clauses
;;;			proposition-negative-clauses
;;;	    
;;;			;; PROPOSITION related functions
;;;			proposition-to-truth-assignment
;;;			proposition-support
;;;
;;;			;; CLAUSE Object
;;;			clause-source
;;;			clause-positive-propositions
;;;			clause-negative-propositions
;;;			clause-consequent
;;;			clause-satisfying-proposition
;;;			clause?
;;;			clause-removed?
;;;			clause-in-theory-unit-clause?
;;;			clause-inconsistent?
;;;			clause-satisfied-by
;;;			clause-in-conflict?
;;;;;;			print-clause
;;;
;;;			;; CLAUSE related functions
;;;			clause-description
;;;			set-support-relation!
;;;			clear-support-relation!
;;;			do-inserted-clauses ;; Macro
;;;			do-proposition-inserted-clauses ;; Macro
;;;			do-proposition-inserted-clauses-w-polarity ;; Macro
;;;			do-clause-propositions ;; Macro
;;;			unit-clause?
;;;			inconsistent-clause?
;;;			no-inconsistent-clauses?
;;;			get-inconsistent-clause
;;;			add-inconsistent-clause		;(should this be in the interface?)
;;;			remove-inconsistent-clause	;(should this be in the interface?)
;;;			)
;;;		  :ISAT)
;;;  )

;;; ***********************************************************************
;;;
;;;                Variables
;;;
;;; ***********************************************************************

(defvar *DEFAULT-THEORY* nil
  "The most recently created theory.")

;;; Variables To Track Performance.

(defvar *isat-statistics-p* nil
  "Controls whether or not isat collects performance
   statistics.  Statistics are only collected if the variable is non-NIL."
  )

(defvar *propagate-counter* 0
  "Number of times PROPAGATE is called."
  )

(defvar *assign-truth-value-counter* 0
  "Number of times ASSIGN-TRUTH-VALUE is called.")

(defvar *unassign-truth-value-counter* 0
  "Number of times UNASSIGN-TRUTH-VALUE is called.")

(defvar *update-proposition-clause-counter* 0
  "Number of times UPDATE-PROPAGATION-CLAUSE is called.")

(defvar *remove-clause-counter* 0
  "Number of times REMOVE-CONSTRAINT is called.")

(defvar *update-clause-for-unassigned-proposition-counter* 0
  "Number of times UPDATE-CLAUSE-FOR-UNASSIGNED-PROPOSITION is called.")


;;; ***********************************************************************
;;;
;;;                Theory Objects
;;;
;;; ***********************************************************************

;;;   Object that holds all information about a theory.  A theory is
;;;   a set of clauses.  The THEORY structure provides access to the
;;;   theory's name, its propositions and clauses, and clauses that are
;;;   currently unit clauses and clauses that are currently unsatisfied.

(defclass theory (essential-constraint-system)

  ((name 
	:initform (gentemp "THEORY-"))

   ;; CLAUSES is the list of all clauses associated with THEORY, including 
   ;; clauses that are currently in the theory, and clauses that are 
   ;; currently removed.

;;;   (clauses :initform '()
;;;			:reader theory-clauses)
   
   ;; STATE-VARIABLE-HT holds the hash table that maps state-variable names to
   ;; state-variable data structures.  This is used when new assignments or
   ;; state-clauses are created, to map quickly from state-variable names
   ;; to their corresponding data structures.
  
   (state-variable-ht :initform (make-hash-table :test #'equal)
					  :accessor theory-state-variable-ht)
  
   ;; PROPOSITION-HT holds the hash table that maps proposition names to
   ;; proposition data structures.  This is useful when new clauses are
   ;; created, to map quickly from a proposition description within a
   ;; clause to the proposition's corresponding data structure.

   (proposition-ht :initform (make-hash-table :test #'equal) 
				   :accessor theory-proposition-ht)

   ;; PROPOSITION list of proposition structures.

   (propositions :initform '()
				 :accessor theory-propositions)
  
   ;; LITERAL-HT holds the hash table that maps a literal description and
   ;; SOURCE for that literal to a clause asserting the truth of literal.
   ;; This is useful when inserting and removing an assignment for a literal
   ;; many times through an external call by SOURCE.

   ;; Note: this is really a unit clause, but the following use of unit-clause
   ;; preempts this.  In addition note that if all clauses are canonicalized
   ;; then this special functionality won't be required.
										;  (literal-clause-ht nil)

   ;; Check:  Was literal-clause-ht renamed to the following?
   (truth-assignment-clause-ht :initform (make-hash-table :test #'equal)
							   :accessor theory-truth-assignment-clause-ht)

   ;; UNIT-CLAUSES is the list of unit clauses that are inserted into theory.  
   ;; A unit clause is any clause that has one unassigned literal, with all 
   ;; other literals assigned false.  This list is updated as truth assignments
   ;; of propositions are changed by unit propagation.

   (unit-clauses :initform (make-lhashset) ;'()
				 :accessor theory-unit-clauses)
  
   ;; INCONSISTENT-CLAUSES is the list of inserted clauses that have all
   ;; false literals.  This list is updated as the truth
   ;; assignments of propositions are changed by unit propagation.

   (inconsistent-clauses  :initform (make-lhashset) ;'()
						  :accessor theory-inconsistent-clauses)
  
   ;; ASSIGNED-PROPOSITIONS Propositions that have been assigned a truth value
   ;; by the satisfiability procedure.

   (assigned-propositions :initform '() 
						  ;;(make-linked-hashset 
						  ;; :test #'equal 
						  ;; :hash-function 
						  ;; (lambda (x) (sxhash (proposition-description x))))
						  :accessor theory-assigned-propositions)
  
   ;; UNASSIGNED-PROPOSITIONS Propositions that have not yet been assigned a
   ;; truth value by the satisfiability procedure.

   (unassigned-propositions :initform (create-heap
									   (lambda (p1 p2) 
										 (< (+ (proposition-num-positive-clauses p1)
											   (proposition-num-negative-clauses p1)) 
											(+ (proposition-num-positive-clauses p2)
											   (proposition-num-negative-clauses p2))
											))) ;'()

							:accessor theory-unassigned-propositions))

  (:documentation "NAME is a lisp object denoting the theory's name."))

(defmethod initialize-instance :after ((theory theory) &rest extra-args)

  "Side effect of setting *DEFAULT-THEORY* to be the newly created theory."

  (declare (ignore extra-args))
  (setq *default-theory* theory))

;;; ***********************************************************************
;;;
;;;                Proposition Objects
;;;
;;; ***********************************************************************

;;;   The proposition object holds information about a proposition,
;;;   including its description, the clauses it occurs in both positively and
;;;   negatively, and its current truth assignment.
(defclass proposition ()
  ;; DESCRIPTION is the proposition's description  A theory may have at most 
  ;; one proposition with a given description  DESCRIPTION can be any Lisp 
  ;; object, typically a list.

  ((description 
	:initarg :description
	:accessor proposition-description)
  
   ;; POSITIVE-CLAUSES and NEGATIVE-CLAUSES is the list of clauses which
   ;; contain proposition as a positive or negative literal, respectively.  
   ;; The lists include both clauses that are currently in the theory, and 
   ;; those that have been removed.

   (positive-clauses 
	:initform '()
	:accessor proposition-positive-clauses)

   (negative-clauses 
	:initform '()
	:accessor proposition-negative-clauses)
  
   ;; NUM-POSITIVE-CLAUSES and NUM-NEGATIVE-CLAUSES should be updated with
   ;; the length of the positive and negative list lengths.

   (num-positive-clauses
	:initform 0
	:accessor proposition-num-positive-clauses)

   (num-negative-clauses
	:initform 0
	:accessor proposition-num-negative-clauses)

   ;; SEARCHABLE is a boolean variable. If true it indicates that this proposition
   ;; should be searched over (assigned a value by) ISAT. If false, ISAT will only
   ;; assign true/false to this proposition via propagation, and not by search.

   (searchable
	:initform t
	:accessor proposition-searchable)

   ;; TRUTH-VALUE stores the current truth assignmen to PROPOSITION.  Its 
   ;; possible values are *TRUE*, *FALSE*, or *NONE*  When it is *NONE*, 
   ;; the truth value is unassigned.
   (truth-value 
	:initform *none*
	:accessor proposition-truth-value)

   ;; Support for proposition's truth assignment.  If unit propagated,
   ;; a clause.  If split on by isat, then *isat* otherwise an
   ;; externally supplied sexp.

   (support 
	:initform nil
	:accessor proposition-support)

   ;; REMAINING-VALUE is used by the Satisfiability procedure to keep track
   ;; of the truth values yet to be explored.  Can take 
   ;; on either value, *TRUE*, *FALSE* or *NONE* (no remaining value).

   (remaining-value 
	:initform nil
	:accessor proposition-remaining-value)
   )
  
  (:documentation "Proposition object holds information about a proposition."))

;;; Note:
;;; o Added to parallel function in Constraint-System. and ITC, and used to
;;;   Externalize an OPSAT solution.  Hopefully this is consistent.
(defmethod constraint-system-assignment-description
	((assignment proposition))
  
  "Returns a description of proposition ASSIGNMENT,
   denoting an assignment:
       (= <variable-name> <value>) "

  (proposition-description assignment))

(defmacro find-proposition
    (theory proposition-description)
  
  " Looks up and returns the proposition object associated 
    with PROPOSITION-DESCRIPTION.  If the proposition object
    does not exist,it returns NIL."
  
  `(gethash ,proposition-description
			(theory-proposition-ht ,theory)))

(defmethod print-object ((obj proposition) stream)
  
  " Print a proposition using its name."

  (with-slots (description) obj 
    (format stream "~a" description)))

;;; Optimization:
;;; o If we can iterate over the hash table efficiently we don't need
;;;   to create a separate proposition list.

(defmethod initialize-instance :after ((proposition proposition) 
									   &key theory &allow-other-keys)
  
  " Creates a new proposition in THEORY.  If it already exists, then it is
    overwritten, and a new proposition is created.  Hence, the caller is
    required to make sure that there is no existing proposition with that
    name."

  (with-slots (description) proposition
	;; insert it into the hash table
    (setf (find-proposition theory description)
      proposition)
    (push proposition (theory-propositions theory))
	T))

;;; Default method defined in mtk-utils/compare.lisp
(defmethod extract-proposition-description ((p proposition))

  "Return the proposition description for proposition p."

  (proposition-description p))


;;;(defstruct (proposition
;;;	    (:print-function print-proposition))
;;;
;;;  "Proposition object holds information about a proposition."
;;;  
;;;  ;; DESCRIPTION is the proposition's description  A theory may have at most 
;;;  ;; one proposition with a given description  DESCRIPTION can be any Lisp 
;;;  ;; object, typically a list.
;;;
;;;  (description nil)
;;;  
;;;  ;; POSITIVE-CLAUSES and NEGATIVE-CLAUSES is the list of clauses which
;;;  ;; contain proposition as a positive or negative literal, respectively.  
;;;  ;; The lists include both clauses that are currently in the theory, and 
;;;  ;; those that have been removed.
;;;
;;;  (positive-clauses nil)
;;;  (negative-clauses nil)
;;;  
;;;  ;; TRUTH-VALUE stores the current truth assignmen to PROPOSITION.  Its 
;;;  ;; possible values are *TRUE*, *FALSE*, or *NONE*  When it is *NONE*, 
;;;  ;; the truth value is unassigned.
;;;  (truth-value *none*)
;;;
;;;  ;; Support for proposition's truth assignment.  If unit propagated,
;;;  ;; a clause.  If split on by isat, then *isat* otherwise an
;;;  ;; externally supplied sexp.
;;;
;;;  (support nil)
;;;
;;;  ;; REMAINING-VALUE is used by the Satisfiability procedure to keep track
;;;  ;; of the truth values yet to be explored.  Can take 
;;;  ;; on either value, *TRUE*, *FALSE* or *NONE* (no remaining value).
;;;
;;;  (remaining-value nil)
;;;  )
;;;
;;;(defun print-proposition (prop stream depth)
;;;  
;;;  " Print a proposition using its name."
;;;  
;;;  (declare (ignore depth))
;;;  (format stream "~a" (ignore-errors (proposition-description prop))))

;;; ***********************************************************************
;;;
;;;                State Variable Objects
;;;
;;; ***********************************************************************

(defclass state-variable ()

  
  ;; NAME is any lisp object or expression that names variable.  
  ;; A theory may have at most one variable with a given NAME.  

  ((name :initarg :name
		 :accessor state-variable-name)

   ;; A list of possible values that may be assigned to STATE-VARIABLE.

   (domain :initarg :domain
		   :accessor state-variable-domain)

   ;; The object that created STATE-VARIABLE.

   (source  :initarg :source
			:accessor state-variable-source))
  
  (:documentation "State variable object holds information 
                   about a state variable.")
  )

(defmethod print-object ((obj state-variable) stream)
  
  " Print a state-variable using its name."
  
  (with-slots (name) obj 
    (format stream "~a" name)))

;;;(defstruct (state-variable
;;;	    (:print-function print-state-variable))
;;;
;;;  "Proposition object holds information about a proposition."
;;;  
;;;  ;; NAME is any lisp object or expression that names variable.  
;;;  ;; A theory may have at most one variable with a given NAME.  
;;;
;;;  (name nil)
;;;
;;;  ;; A list of possible values that may be assigned to STATE-VARIABLE.
;;;
;;;  (Domain nil)
;;;
;;;  ;; The object that created STATE-VARIABLE.
;;;
;;;  (source nil)
;;;  )

;;;(defun print-state-variable (state-variable stream depth)
;;;  
;;;  " Print a state-variable using its name."
;;;  
;;;  (declare (ignore depth))
;;;  (format stream "~a" (ignore-errors (state-variable-name state-variable))))

(defmacro value-in-domain? (state-variable value)

  "Is VALUE an element of the domain of STATE-VARIABLE?"

  `(member ,value
	   (state-variable-domain ,state-variable)))

;;; ***********************************************************************
;;;
;;;                Propositions Denoting Assignments
;;;
;;; ***********************************************************************

(defmacro assignment-proposition? (proposition)
  
  "Returns T iff proposition describes an assignment."
  
  `(assignment-expression?
    (proposition-description ,proposition)))

;;; ***********************************************************************
;;;
;;;                Clause Objects
;;;
;;; ***********************************************************************

;;;   This object holds information about a clause.  A clause is a
;;;   disjunction of literals, where a literal is either a proposition or
;;;   the negation of a proposition.  The clause structure holds
;;;   information about the clause, including a user-specified source, the
;;;   literals, the proposition that currently satisfies it, and various
;;;   bookkeeping flags used during propagation and search.

;;; To Do:
;;; clause-clause is ill named.  it should be clause-description.
(defclass clause (essential-constraint)

  (
   ;; CLAUSE stores a description of clause, denoted as a conjunction
   ;; or list of literal-descriptions.  A literal description is a proposition
   ;; description or its negation, where a proposition description is a
   ;; proposition object or a lisp expression.

   (clause :initarg :clause
		   :accessor clause-clause)

   ;; POSITIVE-PROPOSITIONS and NEGATIVE-PROPOSITIONS stores the propositions
   ;; of the positive and negative literals in CLAUSE, respectively."

   (positive-propositions :initform '()
						  :accessor clause-positive-propositions)
   
   (negative-propositions  :initform '()
						   :accessor clause-negative-propositions)

   ;; SATISFIED-BY stores a proposition that currently satisfies CLAUSE,
   ;; i.e., a positively occurring proposition that is true or a negatively
   ;; occurring proposition that is false.  If a proposition of CLAUSE was
   ;; assigned based on unit clause resolution, then it will always be 
   ;; specified under satisfied-by for that clause.

   (satisfied-by :initform nil
				 :accessor clause-satisfied-by)
  
   ;; UNASSIGNED-LITERAL-COUNT indicates how many unassigned literals occur in
   ;; the clause.

   (unassigned-literal-count :initform 0
							 :accessor clause-unassigned-literal-count)

   ;;  True iff CLAUSE is currently removed from theory. 

   ;; True iff CLAUSE is contained in its theory's list of
   ;; unit clauses.
   (in-theory-unit-clause? :initform nil
						   :accessor clause-in-theory-unit-clause?)

   ;; True iff CLAUSE was used to assign a proposition by 
   ;; unit propagation.

   (propagated? :initform nil
				:accessor clause-propagated?)

   ;; True iff all of CLAUSE's literals are false. 

   (inconsistent? :initform nil
				  :accessor clause-inconsistent?)

   ;; Used during conflict extraction.  Assigned TRUE iff
   ;; CLAUSE supports an inconsistent clause.

   (in-conflict?  :initform nil
				  :accessor clause-in-conflict?))
  
  (:documentation   "This object holds information about a clause."))

(defmethod print-object ((obj clause) stream)
  
  "Prints out clause as:
      (negative propositions) => (positive propositions)
   where the antecedent is a conjunction of propositions 
   and the consequent is a disjunction of propositions.
  
   An empty antecedent list denotes *true*, while
   an empty consequent list denotes *false*."
  
  (with-slots (positive-propositions
			   negative-propositions) obj
	(let ((ref-pos (if (null (rest positive-propositions))
					   (first positive-propositions)
					 positive-propositions))
		  (ref-neg (if (null (rest negative-propositions))
					   (first negative-propositions)
					 negative-propositions)))
	  (cond ((null ref-neg)
			 (if (null ref-pos)
				 (format stream "~a" *true*)
			   (format stream "~a" ref-pos)))
			((null ref-pos)
			 (format stream "not ~a" ref-neg))
			(:else
			 (format stream "~a => ~a" ref-neg ref-pos)
;;;			 (if (clause-removed? clause) "Out" "In")
			 )))))

(defmethod initialize-instance 
	:after ((cl clause) &key constraint-system &allow-other-keys)
  
  "Creates a new clause for CONSTRAINT-SYSTEM, whose source is SOURCE.

   CLAUSE is a list of literals or a disjunction
   (OR . <literals>), where each literal is either
   <proposition-description> or (NOT <proposition-description>).
   <proposition-description>, the description of a proposition, 
   is an arbitary lisp structure, typically a list.  

   This function creates a new clause, but DOES NOT insert the 
   clause into CONSTRAINT-SYSTEM, that is, the clause is said to be removed. 
   Clause can be later inserted by calling INSERT-CLAUSE.  

   Note that this function does not check for duplicate clauses, 
   but merely proceeds to create additional copies.  Any required 
   propositions are created by this function, and pointers relating
   clause to its propositions are put in place."
  
  ;; Note that this function does not update any of the status 
  ;; slots of the clause.  This is performed when the clause 
  ;; is inserted into the constraint-system. 

  ;; First, make sure that the clause isn't too long.
										;  (check-clause-length clause)
  ;; Create the clause object. 
										;  (debug-print t
										;      "Creating ~A for ~A, by ~A." 
										;      clause constraint-system source)

  ;; Strip 'OR.
  (with-slots (clause) cl
	(setq clause (strip-disjunction clause))

	;; Associate clause with CONSTRAINT-SYSTEM (although not inserted).
	
	;; Get, or create when necessary, objects for each of 
	;; CLAUSE's propositions.
	(dolist (literal-description clause)
	  (multiple-value-bind
		  (proposition-description positive?)
		  (literal-description-proposition literal-description)

		;; Get the proposition object corresponding to description.
		(let ((propn (if (typep proposition-description 'proposition)
						 proposition-description
					   (get-proposition constraint-system proposition-description))))
		  ;; In the following we remove duplicate literals 
		  ;; ("P or P" is equivalent to "P").
		  ;; This makes it easier to maintain the unassigned literal count.
		  ;; Note, however, that we do not check for tautologies.
		  (if positive?
			  ;; its a positive literal.
			  (pushnew propn
					   (clause-positive-propositions cl) 
					   :test #'eq)
			;; else its a negative literal
			(pushnew propn 
					 (clause-negative-propositions cl)
					 :test #'eq)))))

	;; Point CLAUSE's propositions to CLAUSE.
	(connect-propositions-to-clause! cl)
	))

(defclass unit-clause (clause)
  ()
  (:documentation   "This object holds information about a 
                     clause comprised of one literal"))

;;; Note:
;;; o The following does additional work and consing that could
;;;   be avoided by directly placing proposition on
;;;    (clause-negative-propositions cl) or 
;;;     (clause-negative-propositions cl) and wiring up.
(defmethod initialize-instance 
	:before ((cl unit-clause) &key truth-value proposition &allow-other-keys)
  
  "Creates a unit clause for THEORY whose source is SOURCE. 
   CLAUSE specifies that PROPOSITION takes on TRUTH-VALUE
   PROPOSITION is a proposition description or object.
   TRUTH-VALUE must be one of *TRUE* or *FALSE*."

;;;  (debug-print t
;;;       "Creating clause asserting ~A is ~A in theory ~A." 
;;;	   proposition truth-value  theory)

  (with-slots (clause) cl
	(let ((clause-description 
		   (list
			(if (eql truth-value *true*)
				proposition
			  (make-negation proposition)))))
	  (setq clause clause-description))))

(defclass conflict-clause (clause)

  ()
  
  (:documentation   "This object holds information about a 
                     clause, comprised of a conjunction of
                     negative literals, denoting an inconsistency"))

;;; Note:
;;; o The following does additional work and consing that could
;;;   be avoided by directly placing conflict on
;;;    (clause-negative-propositions cl) and wiring up.
(defmethod initialize-instance 
	:before ((cl conflict-clause) &key conflict &allow-other-keys)
  
  "Creates a clause for THEORY with SOURCE that denotes CONFLICT.

   CONFLICT is a list of mutually inconsistent 
   propositions.  Conflict MUST NOT contain duplicate
   propositions

   This function creates a new clause, and associates it
   with THEORY, but does not insert it into THEORY."

;;;  (debug-print t
;;;       "Creating conflict ~A for theory ~A." 
;;;	   conflict theory)

  (with-slots (clause) cl
	(let ((clause-description 
		   (loop for l in conflict
			   collecting (make-negation l))))
	  (setq clause clause-description))))

(defclass state-clause (clause)

  ()
  
  (:documentation   "This object holds information about a 
                     state clause, that is a clause whose
                     propositions are variable-value assignments."))

(defmethod initialize-instance 
	:before ((cl state-clause) &key clause constraint-system &allow-other-keys)
  
  "Creates and returns a new state clause for CONSTRAINT-SYSTEM whose source is SOURCE. 
   CLAUSE is a list of state-literals or a disjunction 
   (OR . <state-literals>).  Each state-literal is a state assignment or its
   negation, that is:
   (= <variable> <value>) or (NOT (= <variable> <value>)), 
   where <variable> is a state variable of CONSTRAINT-SYSTEM and <value> is an element
   of the domain of <variable>.

   This function creates a new clause, 
   but DOES NOT insert CLAUSE into CONSTRAINT-SYSTEM, i.e., the clause is a removed 
   clause.  A clause is inserted by calling INSERT-CONSTRAINT.  Note 
   that this function does not check for duplicate clauses, 
   but merely proceeds to create additional copies.  Any required 
   propositions are created by the function, and pointers between
   the clause and its propositions are put in place.  The function
   returns the newly created clause."

;;;  (debug-print t
;;;       "Creating new state clause ~A for theory ~A." 
;;;	   state-clause-description  constraint-system)
  (check-state-clause-description? constraint-system
								   (strip-disjunction clause)))

;;;(defstruct (clause
;;;	    (:print-function print-clause))
;;;
;;;  "This object holds information about a clause."
;;;
;;;  ;; SOURCE stores domain specific information about the clause, and can be 
;;;  ;; any lisp object.  Typically used to specify the source or type of a 
;;;  ;; clause.
;;;
;;;  (source nil)
;;;
;;;  ;; POSITIVE-PROPOSITIONS and NEGATIVE-PROPOSITIONS stores the propositions
;;;  ;; of the positive and negative literals in CLAUSE, respectively."
;;;
;;;  (positive-propositions nil)
;;;  (negative-propositions nil)
;;;
;;;  ;; SATISFIED-BY stores a proposition that currently satisfies CLAUSE,
;;;  ;; i.e., a positively occurring proposition that is true or a negatively
;;;  ;; occurring proposition that is false.  If a proposition of CLAUSE was
;;;  ;; assigned based on unit clause resolution, then it will always be 
;;;  ;; specified under satisfied-by for that clause.
;;;
;;;  (satisfied-by nil)
;;;  
;;;  ;; UNASSIGNED-LITERAL-COUNT indicates how many unassigned literals occur in
;;;  ;; the clause.
;;;
;;;  (unassigned-literal-count 0)
;;;
;;;  ;;  True iff CLAUSE is currently removed from theory. 
;;;
;;;  (removed? nil)
;;;
;;;  ;; True iff CLAUSE is contained in its theory's list of
;;;  ;; unit clauses.
;;;  (in-theory-unit-clause? nil)
;;;
;;;  ;; True iff CLAUSE was used to assign a proposition by 
;;;  ;; unit propagation.
;;;
;;;  (propagated? nil)
;;;
;;;  ;; True iff all of CLAUSE's literals are false. 
;;;
;;;  (inconsistent? nil)
;;;
;;;  ;; Used during conflict extraction.  Assigned TRUE iff
;;;  ;; CLAUSE supports an inconsistent clause.
;;;
;;;  (in-conflict? nil))

;;; Note:
;;; o The following code, which overloads the meaning of
;;;   the satisfied-by slot, was introduced by Pandu Nayak.  
;;;   Its not clear what saving it provides.

;;;  *** Derived fields of CLAUSE ***
;;;
;;; CLAUSE has two additional slots, whose accessor and setf
;;; function are defined below:

;;;   CONSEQUENT : The proposition assigned a truth value as a
;;                 consequence of unit resolution on clause. 
;;                 If the clause does not support proposition by
;;;                unit resolution then the consequent is NIL.

;;;   SATISFYING-PROPOSITION : A proposition whose truth value 
;;;                satisfies the clause.  The satisfying-proposition 
;;;                can only be set if the clause DOES NOT have 
;;;                a consequent.

;;; *** Accessing and Setting the CLAUSE-CONSEQUENT ***

(defmacro clause-consequent (clause)

  " Returns the proposition whose truth value was derived by CLAUSE
    during unit propagation.  If none exists it returns nil."

  ;; This derived CONSEQUENT slot shares the SATISFIED-BY slot, since
  ;; the proposition that satisfies a clause that is used in unit 
  ;; propagation MUST BE the consequent.  We indicate that a 
  ;; satisfied-by proposition is a consequent using the 
  ;; CLAUSE-PROPAGATED? flag.

  ;; It is an (unflagged) error for the CLAUSE-PROPAGATED? flag to 
  ;; be set when unit propagation did not assign the truth value of 
  ;; any proposition.
  `(when (clause-propagated? ,clause)
    ;; When clause was used in a unit propagation,
    ;; the SATISFIED-BY field stores the
    ;; proposition. 
    (clause-satisfied-by ,clause)))

(defmacro setf-clause-consequent (clause proposition)
  
  " Sets the CONSEQUENT field of CLAUSE to PROPOSITION,
    when the truth of PROPOSITION was derived via
    unit propagation using CLAUSE."

  ;; It uses the SATISFIED-BY slot to record PROPOSITION, 
  ;; and indicates that this slot contains a CONSEQUENT 
  ;; by setting PROPAGATED? to T.
  ;; If PROPOSITION is NIL, it ensures that clause's 
  ;; PROPAGATED? flag is NIL.
  ;; Returns PROPOSITION.

  `(progn
     ;; If a clause has a consequent, then the consequent
     ;; also must be clause's satisfying proposition.
    (setf (clause-satisfied-by ,clause) ,proposition)
    (setf (clause-propagated? ,clause) (not (null ,proposition)))))

;;; Allows CLAUSE-CONSEQUENT to be set using setf,
;;; with the setf action performed by function SETF-CLAUSE-CONSEQUENT.
(defsetf clause-consequent setf-clause-consequent)

;;; *** Accessing and Setting the CLAUSE-SATISFYING-PROPOSITION ***

(defmacro clause-satisfying-proposition (clause)
  
  " Returns a PROPOSITION whose truth value satisfies 
    CLAUSE.  If CLAUSE has a consequent, produced by
    unit propagation, then the consequent of clause will 
    always be its satisfying proposition.  Otherwise,
    the satisfying proposition can be any proposition 
    contained in a true literal of CLAUSE."

  `(clause-satisfied-by ,clause))

(defmacro setf-clause-satisfying-proposition (clause proposition)
  
  " Sets CLAUSE's satisfying proposition to PROPOSITION.
    If CLAUSE has a consequent, then consequent is already
    the satisfying proposition, and PROPOSITION is ignored.  
    If CLAUSE has no consequent, then satisfying-proposition 
    is set to PROPOSITION."
  
  `(unless (clause-consequent ,clause)
     ;; Only set if CLAUSE doesn't already have a consequent.
     (setf (clause-satisfied-by ,clause) ,proposition)))

;;; Allows clause-satisfying-proposition to be set using setf,
;;; where the action is performed by function 
;;; SETF-CLAUSE-SATISFYING-PROPOSITION.
(defsetf clause-satisfying-proposition
    setf-clause-satisfying-proposition)

;;; The following two macros allow the implementation of
;;; the support/consequent relation to be modified
;;; in a transparent manner.

(defmacro set-support-relation! (proposition clause)
  "Record that PROPOSITION was assigned based on CLAUSE,
   hence PROPOSITION is the consequent of CLAUSE and
   CLAUSE supports PROPOSITION."
  `(progn
     (setf (proposition-support ,proposition) ,clause)
	 (if (typep ,clause 'clause)
		 (setf (clause-consequent ,clause) ,proposition))))

(defmacro clear-support-relation! (proposition clause)
  "Clear record that CLAUSE is supporting PROPOSITION."

  `(progn
	 (setf (proposition-support ,proposition) nil)
	 ;; Note: Also sets clause-satisfying-proposition
	 ;;       to nil by side-effect.
	 (if (typep ,clause 'clause)
		 (setf (clause-consequent ,clause) nil)))
)

;;;(defmacro clear-support-relation! (proposition clause)
;;;  "Clear record that CLAUSE is supporting
;;;   PROPOSITION."
;;;
;;;  `(progn
;;;     (setf (proposition-support ,proposition) nil)
;;;     ;; Note: Also sets clause-satisfying-proposition
;;;     ;;       to nil by side-effect.
;;;     (setf (clause-consequent ,clause) nil)))

;;;            ***   Printing a clause   ***

;;; To Do:
;;; o Consider a simpler printout for clause.
;;;(defun print-clause (clause stream depth)
;;;  
;;;  "Prints out clause as:
;;;      (negative propositions) => (positive propositions)
;;;   where the antecedent is a conjunction of propositions 
;;;   and the consequent is a disjunction of propositions.
;;;  
;;;   An empty antecedent list denotes *true*, while
;;;   an empty consequent list denotes *false*."
;;;
;;;  (declare (ignore depth))
;;;  (ignore-errors
;;;   (let* ((positive (clause-positive-propositions clause))
;;;		  (negative (clause-negative-propositions clause))
;;;		  (ref-pos (if (null (rest positive))
;;;					   (first positive)
;;;					 positive))
;;;		  (ref-neg (if (null (rest negative))
;;;					   (first negative)
;;;					 negative)))
;;;	 (cond ((null ref-neg)
;;;			(if (null ref-pos)
;;;				(format stream "~a" *true*)
;;;			  (format stream "~a" ref-pos)))
;;;		   ((null ref-pos)
;;;			(format stream "not ~a" ref-neg))
;;;		   (:else
;;;			(format stream "~a => ~a" ref-neg ref-pos)
;;;;;;			 (if (clause-removed? clause) "Out" "In")
;;;			)))))

;;; ***************************************************************************
;;;
;;; *** Proposition Truth Assignment and Support
;;;
;;; ***************************************************************************

(defmacro assigned? (proposition)
  
  "Returns True if PROPOSITION is currently assigned a truth value."
  
  `(not (eql (proposition-truth-value ,proposition)
	     *none*)))

(defun proposition-to-truth-assignment (proposition)

  "Given PROPOSITION, returns expression:
       (= PROPOSITION <truth-value>)
   where truth-value is the current truth assignment
   to PROPOSITION."
  
  `(= ,proposition ,(proposition-truth-value proposition)))

;;; Whenever a proposition is assigned a truth value,
;;; associated with the assignment is a justification for the
;;; assignment, referred to the proposition's support.

;;; In general the support is some sexpression, provided by the
;;; code that derived that truth assignment.
;;;
;;; If the assignment was derived by unit propagation, 
;;; then the support is the clause used during unit clause
;;; resolution to derive proposition's truth value.

;;  If the assignment was provided by another routine or human,
;;; then the support can be anything EXCEPT a clause.

;;; Optimization:
;;; o change this to a macro or inline once debugged.
(defun unit-propagation-support? (support)

  "Returns T if SUPPORT was provided by unit propagation."

  (typep support 'clause))

;;;(defmacro unit-propagation-support? (support)
;;;
;;;  "Returns T if SUPPORT was provided by unit propagation."
;;;
;;;  `(typep ,support 'clause))

(defmacro support-by-unit-propagation? (proposition)

  "Returns T if PROPOSITION's assignment was derived by unit propagation."

  `(unit-propagation-support? (proposition-support ,proposition)))

(defmacro no-support? (support)

  "Returns T if a SUPPORT has not been provided."

  `(null ,support))

(defmacro not-supported? (proposition)

  "Returns T if PROPOSITION has not been provided a support."

  `(no-support? (proposition-support ,proposition)))

;;; ***********************************************************************
;;;
;;;                Clauses Asserting Truth Assignments
;;;
;;; ***********************************************************************

;;; A clause containing a single proposition asserts a
;;; truth assignment to that proposition, based on unit
;;; clause resolution.  The following routines are for
;;; manipulating clauses containing a single proposition and
;;; their corresponding truth assignment.

;;; Note that such a clause is traditionally referred to as a unit
;;; clause.  However, the term unit clause is also used to refer
;;; to a clause that has been reduced to a unit clause through
;;; unit propagation, by assigning false to all literals save one.

;;; In this code we use UNIT-CLAUSE to refer to a clause
;;; that is currently reduced to a unit based on propagation
;;; and splitting.

;;;  We use the term SINGLE-PROPOSITION-CLAUSE to refer to a
;;;  a unit clause that started as a unit clause, before
;;;  splitting and propagation.

;;;  To Do:
;;;  o The above terminology for distinguishing the two types of
;;;    unit clauses is not ideal.  Consider changing to an alternate
;;;    terminology, such as unit-reduced-clause vs x-unit-clause
;;;    (for some x).

;;; Optimization:
;;; o This is highly inefficient, particularly for an inner loop routine.  
;;;   Perform this evaluation when clause
;;;   is created, and record it in a unit-proposition or
;;;   unit-literal slot.
(defun truth-assignment-when-single-proposition-clause (clause)

  "If clause contains a single proposition, returns
   (= PROPOSITION TRUTH-VALUE); otherwise, returns nil.
   PROPOSITION is a proposition structure.
   TRUTH-VALUE is *TRUE* if PROPOSITION appears in a 
   positive literal and *FALSE* if PROPOSITION appears
   in a negative literal."

  (when (typep clause 'clause)
    (let ((pos-propositions (clause-positive-propositions clause))
	  (neg-propositions (clause-negative-propositions clause)))
;;;	(debug-print t "Checking if ~A is a single literal clause."
;;;		clause)
;;;	(debug-print t "Has positive propositions ~A, ~
;;;                     negative propositions ~A."
;;;		pos-propositions neg-propositions)
      (cond ((and (length-1? pos-propositions)
		  (null neg-propositions))
	     (list '= (first pos-propositions) *true*))
	    ((and (length-1? neg-propositions) 
		  (null pos-propositions))
	     (list '= (first neg-propositions) *false*))
	    (:else
	     nil)))))

;;; Optimization:
;;; o This is highly inefficient, particularly for an inner loop routine.  
;;;   Perform this evaluation when clause is created, and record 
;;;   it in a unit-proposition or unit-literal slot.
(defun single-proposition-clause? (clause)

  "If clause contains a single proposition, returns
   proposition; otherwise, returns nil."

  (when (typep clause 'clause)
    (let ((pos-propositions (clause-positive-propositions clause))
	  (neg-propositions (clause-negative-propositions clause)))
      (cond ((and (length-1? pos-propositions)
		  (null neg-propositions))
	     (first pos-propositions))
	    ((and (length-1? neg-propositions)
		  (null pos-propositions))
	     (first neg-propositions))
	    (:else
	     nil)))))

(defun assignment-clause? (clause)
  "Returns T iff CLAUSE positively or negatively asserts
   an assignment."
  
  (let ((proposition (single-proposition-clause? clause)))
    (and proposition
	 (assignment-proposition? proposition))))

(defun pass-state-assignment (clause)

  "Returns CLAUSE IFF it represents a positive or negative state assignment, 
   otherwise returns NIL."
  
  (when (assignment-clause? clause)
    clause))

;;; ***********************************************************************
;;;
;;;                External Interface Functions and Macros
;;;
;;; ***********************************************************************

;;; ***********************************************************************
;;;
;;; *** Operating on Clauses of a Theory
;;;
;;; ***********************************************************************

(defmacro do-inserted-clauses ((clause clauses) &body body)
  
  "Executes BODY for each clause in CLAUSES that is inserted
   in SOME theory.  Behaves exactly like dolist.  Presumes
   that the caller only provides clauses that are relevant
   to the current theory(ies) of interest."
  
  `(dolist (,clause ,clauses)
     (unless (constraint-removed? ,clause)
       ,@body)))

;;; Optimization:
;;; o Confirm that this function is not called by any time
;;;   critical routine.
(defun clause-description (clause)
  
  "Returns the description of the CLAUSE object.  The
   CLAUSE's description is the list of descriptions of literals i
   n the CLAUSE.  

   Efficiency Note: The description is created when required, 
   rather than explicitly stored, under the assumption this 
   isn't invoked often."
  
  (let ((description nil))
    ;; Construct the description from the CLAUSE's positive 
    ;; and negative literals.
    (dolist (proposition (clause-positive-propositions clause))
      (push (proposition-description proposition) description))
    (dolist (proposition (clause-negative-propositions clause))
      (push (list 'not (proposition-description proposition)) description))
    ;; and return DESCRIPTION
    (values description)))

(defmacro do-clause-propositions ((proposition polarity clause) &body body)
  
  "Executes BODY for each proposition in CLAUSE.  Polarity is bound to
   *true* for each proposition appearing in a positive literal and
   *false* for each appearing in a negative literal.
   Behaves exactly like dolist."
  
  `(let ((,polarity *true*))
     (dolist (,proposition (clause-positive-propositions ,clause))
       ,@body)
     (setq ,polarity *false*)
     (dolist (,proposition (clause-negative-propositions ,clause))
       ,@body)))

(defmacro do-proposition-inserted-clauses
    ((clause proposition) &body body)
  
  "Executes BODY for each inserted clause containing PROPOSITION.
   Behaves exactly like dolist."
  
  `(progn
     (do-inserted-clauses (,clause (proposition-positive-clauses ,proposition))
       ,@body)
     (do-inserted-clauses (,clause (proposition-negative-clauses ,proposition))
       ,@body)))

(defmacro do-proposition-inserted-clauses-w-polarity
    ((clause polarity proposition) &body body)
  
  "Executes BODY for each inserted clause containing PROPOSITION  
   Polarity is bound to *true* for each clause containing PROPOSITION
   as a positive literal and *false* for each clause containing
   PROPOSITION as a negative literal.
   Behaves exactly like dolist."
  
  `(let ((,polarity *true*))
     (do-inserted-clauses (,clause (proposition-positive-clauses ,proposition))
       ,@body)
     (setq ,polarity *false*)
     (do-inserted-clauses (,clause (proposition-negative-clauses ,proposition))
       ,@body)))

;;; ***********************************************************************
;;;
;;; *** Managing Unit Reduced and Support Clauses
;;;
;;; ***********************************************************************

;;; For the purpose of ISAT, a UNIT CLAUSE is clause that
;;; CURRENTLY is unsatisfied and has one unsatisfied literal.

;;; The SUPPORT of an assigned PROPOSITION is a clause
;;; that contains 

;;; To Do:
;;; o Does it make sense to leave this in the external interface,
;;;   rather than the module interface?
(defun unit-clause? (clause)
  
  " Checks to see if CLAUSE is a unit clause, that is a clause
    with one unassigned literal and all others assigned false.  
    The result of this function is valid only if clause is inserted."

  ;; A clause is a unit clause if its UNASSIGNED-LITERAL-COUNT slot is 1, and
  ;; it is not currently supported.
  (and (null (clause-satisfying-proposition clause))
       (= (clause-unassigned-literal-count clause) 1)))

;;; Fix:
;;; o The following is replaced by an explicit record of
;;;   support.  Delete the following code as soon as the current
;;;   code passes all tests.

;;; Optimization:
;;; o If this function is called by a time critical routine, 
;;;   then it will be too slow.  Instead proposition should 
;;;   include a slot that points to its support.  
;;;(defun proposition-support (proposition)
  
;;;  "This function returns the clause whose consequent is PROPOSITION.
;;;   If no such clause exists it returns NIL.
;;;
;;;   Efficiency Note: To save space, support of the proposition is not 
;;;   explicitly stored, but is computed by searching through the clauses 
;;;   mentioning PROPOSITION."
;;;
;;;  (do-proposition-inserted-clauses (clause proposition)
;;;    (when (eql (clause-consequent clause) proposition)
;;;      (return-from proposition-support 
;;;	(values clause))))
;;;  
;;;  ;; Else return NIL
;;;  (values nil))

;;; ***********************************************************************
;;;
;;; *** Managing Inconsistent Clauses within a Theory
;;;
;;; ***********************************************************************

;;; To Do:
;;; o Does it make sense to leave this in the external interface,
;;;   rather than the module interface?

;;; Optimization:
;;; o Inline for efficiency?
(defun inconsistent-clause? (clause)
  
  " Checks to see if CLAUSE is an inconsistent clause,
    that is one in which all literals are assigned false.  
    The result of this function is valid only if clause is inserted.

   A clause is inconsistent if its UNASSIGNED-LITERAL-COUNT is 0 and it has
   no satisfying proposition."

  (and (not (clause-satisfying-proposition clause))
       (zerop (clause-unassigned-literal-count clause))))

;;; To Do:
;;; o Does it make sense to leave this in the external interface,
;;;   rather than the module interface?

;;; Optimization:
;;; o Inline for efficiency?
(defun get-inconsistent-clause (theory)
  
  " Returns a clause in THEORY that is inconsistent. If no such 
   clause exists, it returns NIL."
  
  (lhs-get-first (theory-inconsistent-clauses theory)))

;;; To Do:
;;; o Confirm how this is used and whether or not it makes
;;;   sense to be in the external interface.

;;; Optimization:
;;; o Inline for efficiency?
(defun no-inconsistent-clauses? (theory)
  
  " Returns T if there are currently no inconsistent clauses in THEORY,
   and NIL otherwise."
  
  (lhs-empty? (theory-inconsistent-clauses theory)))

;;; To Do:
;;; o Does it make sense to leave this in the external interface,
;;;   rather than the module interface?

;;; Optimization:
;;; o Inline for efficiency?
(defun add-inconsistent-clause (clause theory)
  
  " Adds CLAUSE as a new inconsistent clause to THEORY.  
   Ensures that inconsistent clauses of THEORY contains no duplicates."
  
  (unless (clause-inconsistent? clause)
    ;; If it isn't already in the list, add it in and set the unsatisfied
    ;; flag 
    (lhs-add (theory-inconsistent-clauses theory) clause)
    (setf (clause-inconsistent? clause) t)))
	  
;;; To Do:
;;; o Does it make sense to leave this in the external interface,
;;;   rather than the module interface?
(defun remove-inconsistent-clause (clause theory)
  
  " Removes CLAUSE from the set of inconsistent clauses of THEORY.  No
    error is reported if CLAUSE isn't in the set of inconsistent clauses."
  
  (when (clause-inconsistent? clause)
    ;; proceed only when the clause is in the list of inconsistent clauses
	(lhs-remove (theory-inconsistent-clauses theory) clause)
    (setf (clause-inconsistent? clause) nil)))

;;; ***********************************************************************
;;;
;;;                Internal Module Interface Functions and Macros
;;;
;;; ***********************************************************************

;;; The LTMS-theories module interface is comprised of the following functions,
;;; together with the external interface functions, and the slot accessors 
;;; for the Theory, Clause and Proposition objects.

;;;                 *** Managing the set of unit clauses ***

(defmacro has-unit-clauses? (theory)
  
  " Returns T if theory contains unit clauses, and NIL
    otherwise."
  
  `(not (lhs-empty? (theory-unit-clauses ,theory))))

;;; Optimization:
;;; o inline?
(defun get-unit-clause (theory)
  
  " Returns a clause in the unit clauses of theory, if it exists, and NIL
    otherwise."

  (lhs-get-first (theory-unit-clauses theory)))

(defun add-to-unit-clauses (clause theory)
  
  " Adds CLAUSE to the unit clauses of THEORY.  Makes sure
   that CLAUSE only appears once."

  (unless (clause-in-theory-unit-clause? clause)
    ;; the clause isn't already in the unit clauses, 
    ;; so add it in and set the in-theory-unit-clause? flag.
    (lhs-add (theory-unit-clauses theory) clause)
    (setf (clause-in-theory-unit-clause? clause) t)))

;;; Optimization:
;;; o This could be an expensive operation and is unnecessary.  
;;;   If clause is no longer a unit clause, then it can be 
;;;   detected when clause is removed from the queue.
(defun remove-from-unit-clauses (clause theory)
  
  " Removes CLAUSE from the unit clauses of THEORY.
   If CLAUSE isn't in the set of unit clauses, no error is signaled."
  
  (when (clause-in-theory-unit-clause? clause)
    ;; do anything only if the clause is in the set of unit clauses
	(lhs-remove (theory-unit-clauses theory) clause)
    (setf (clause-in-theory-unit-clause? clause) nil)))
