;;; -*- Mode:Common-Lisp; Package:Isat; Base:10 -*-

;;; Copyright (c) 1989 - 2009 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)
	  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                               ;;;
;;;                  ISAT Tests                                   ;;;
;;;                                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftest-suite :Isat
    (log-ISAT-statistics)
  (stop-and-print-ISAT-statistics)
  )

(deftest (test1 :isat) ()
  (test `(((P ,*false*) (Q ,*false*)) nil)
	 (find-model-of-clauses
	  '((p (not q)) 
	    ((not p)))
	  :name 'test1
	  :initialize? t 
	  :return-conflict? nil
	  )
	:multiple-values t
	:test #'equal-model-conflict-result
	:fail-info "P or not Q, and not P -> Model: Q = false, P = false"
	))

(deftest (test2 :isat) ()
  (test '(nil nil)
		(consistent-clauses?
		 '((p (not q))
		   (q (not r))
		   (r p)
		   ((not p)))
		 :name 'test2
		 :perform-search? t 
		 :return-conflict? nil)
		:multiple-values t
		:test #'equalp					;#'equal-consistent?-conflict-result
		:fail-info "P or not Q, and Q or not R, and R or P, ~
                   and not P -> inconsistent"))

;;; Correct models for test3
;;;   P  Q   R
;;;   --------
;;;   T  T   T
;;;   F  F   F

(deftest (test3 :isat) ()
  (let ((theory (theory-with-clauses
				 '((p (not q))
				   (q (not r))
				   (r (not p)))
				 :name "test3")))

	;;; Assert P
    (let ((c1 (insert-new-clause theory 
								 :source 'test3 
								 :clause '(p)))
		  (c2 (make-instance 'clause 
				:constraint-system theory
				:source 'test3
				:clause '((not p)))))

	  ;; Unit propagation assigns all propositions true
      (propagate theory)
      (test `((P ,*true*)(Q ,*true*)(R ,*true*))
			(theory-assignment theory)
			:test #'equal-models
			:fail-info "P or not Q, Q or not R, R or not P, and P ~
                        -> Unit entailed P = Q = R = true")
      
	  ;; Removing P, all unit propagations should be retract.
      (remove-constraint theory c1)
      (test '()
			(theory-assignment theory)
			:test #'equal-models
			:fail-info "P or not Q, Q or not R, R or not P ~
                        -> No assignments unit entailed")

	  ;; Assert not P
      (insert-constraint theory c2)
	  ;; Unit propagation assigns all propositions false
      (propagate theory)
      (test `((P ,*false*)(Q ,*false*)(R ,*false*))
			(theory-assignment theory)
			:test #'equal-models
			:fail-info "P or not Q, Q or not R, R or not P, and not P ~
                        -> Unit entailed P = Q = R = false")

      (remove-constraint theory c2)
      
	  ;; With P removed, ISAT should find assignments
	  ;; P = Q = R = True (or P = Q = R = False)
      (find-model theory 't)
      (test `(((P ,*true*)(Q ,*true*)(R ,*true*))
			  ((P ,*false*)(Q ,*false*)(R ,*false*)))
			(theory-assignment theory)
			:test #'set-contains-model
			:fail-info "P or not Q, Q or not R, R or not P ~
                        -> 1st model: P = Q = R = true (or false)")
      
	  ;; and second assignment P = Q = R = False (or P = Q = R = True)
      (find-model theory nil)
      (test `(((P ,*false*)(Q ,*false*)(R ,*false*))
			  ((P ,*true*)(Q ,*true*)(R ,*true*)))
			(theory-assignment theory)
			:test #'set-contains-model
			:fail-info "P or not Q, Q or not R, R or not P ~
                        -> 2nd model: P = Q = R = false (or true)"))))

(deftest (test3-1 :isat) ()
  (let ((theory (theory-with-clauses
		 '((p (not q))
		   (q (not r))
		   (r (not p)))
		 :name "test3-1")))
    
    ;;; Assert P
    (let ((c1 (insert-new-clause theory :source 'test3-1 :clause '(p)))
		  (c2 (make-instance 'clause 
				:constraint-system theory
				:source 'test3-1
				:clause '((not p)))))

      ;; Unit propagation assigns all propositions true
      (propagate theory)
      (test `((P ,*true*)(Q ,*true*)(R ,*true*))
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P, and P ~
                        -> Unit entailed P = Q = R = true")
      
      ;; Removing P, all unit propagations should be retract.
      (remove-constraint theory c1)
      (test '()
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P ~
                        -> No assignments unit entailed")

      ;; Assert not P
      (insert-constraint theory c2)
      ;; Unit propagation assigns all propositions false
      (propagate theory)
      (test `((P ,*false*)(Q ,*false*)(R ,*false*))
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P, and not P ~
                        -> Unit entailed P = Q = R = false")
      
      ;; Retract
      (remove-constraint theory c2)
      (test '()
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P ~
                        -> No assignments unit entailed")
      
      ;; Assert P again
      (insert-constraint theory c1)
      ;; Unit propagation assigns all propositions false
      (propagate theory)
      (test `((P ,*true*)(Q ,*true*)(R ,*true*))
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P, and P ~
                        -> Unit entailed P = Q = R = true")
      )))

(deftest (test3-2 :isat) ()
  (let ((theory (theory-with-clauses
		 '((p (not q))
		   (q (not r))
		   (r (not p)))
		 :name "test3-2")))
    
    ;;; Assert not P
    (let ((c1 (make-instance 'clause 
				:constraint-system theory
				:source 'test3-2
				:clause '(p)))
		  (c2 (insert-new-clause theory 
								 :source 'test3-2
								 :clause '((not p)))))

      ;; Unit propagation assigns all propositions true
      (propagate theory)
      (test `((P ,*false*)(Q ,*false*)(R ,*false*))
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P, and P ~
                        -> Unit entailed P = Q = R = false")
      
      ;; Removing not P, all unit propagations should be retract.
      (remove-constraint theory c2)
      (test '()
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P ~
                        -> No assignments unit entailed")

      ;; Assert P
      (insert-constraint theory c1)
      ;; Unit propagation assigns all propositions false
      (propagate theory)
      (test `((P ,*true*)(Q ,*true*)(R ,*true*))
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P, and P ~
                        -> Unit entailed P = Q = R = true")
      
      ;; Retract
      (remove-constraint theory c1)
      (test '()
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P ~
                        -> No assignments unit entailed")
      
      ;; Assert not P again
      (insert-constraint theory c2)
      ;; Unit propagation assigns all propositions false
      (propagate theory)
      (test `((P ,*false*)(Q ,*false*)(R ,*false*))
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not Q, Q or not R, R or not P, and not P ~
                        -> Unit entailed P = Q = R = false")
      )))

(deftest (test3-3 :isat) ()
  (let ((theory (theory-with-clauses
		 '((p (not r))
		   (r (not p)))
		 :name "test3-3")))
    
    ;;; Assert P
    (let ((c1 (insert-new-clause theory :source 'test3-3 :clause '(p)))
		  (c2 (make-instance 'clause 
				:constraint-system theory
				:source 'test3-3
				:clause '((not p)))))

      ;; Unit propagation assigns all propositions true
      (propagate theory)
      (test `((P ,*true*)(R ,*true*))
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not R, R or not P, and P ~
                        -> Unit entailed P = R = true")
      
      ;; Removing P, all unit propagations should be retract.
      (remove-constraint theory c1)
      (test '()
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not R, R or not P ~
                        -> No assignments unit entailed")

      ;; Assert not P
      (insert-constraint theory c2)
      ;; Unit propagation assigns all propositions false
      (propagate theory)
      (test `((P ,*false*)(R ,*false*))
	    (theory-assignment theory)
	    :test #'equal-models
	    :fail-info "P or not R, R or not P, and not P ~
                        -> Unit entailed P = R = false"))))

(deftest (test4 :isat) ()
  (test `((((= P 1) ,*false*)
	   ((= P 2) ,*true*)
	   ((= Q 1) ,*false*)
	   ((= Q 2) ,*true*)) 
	  nil)
	 (find-model-of-state-clauses
	  '((p (1 2))(q (1 2)))
	  '(((= p 1)(not (= q 1)))
	    ((not (= p 1))))
	  :name 'test4
	  :initialize? t 
	  :return-conflict? nil)
	:multiple-values t		
	:test #'equal-model-conflict-result
	:fail-info "P=1 or not Q=1, and not P=1 ~
                    -> Q=1 = false, P=1 = false, Q=2 = true, P=2 = true"
	))

(deftest (test5 :isat) ()

  ;; In the following note that assignments
  ;; ((not (= P 1) (= Q 1)) are inconsistent,
  ;; however, since they are unit clauses, rather
  ;; than external assignments, they are not flagged
  ;; in the conflict.
  (test `(nil nil)
	 (consistent-state-clauses?
	  '((p (1 2))
	    (q (1 2))
	    (r (1 2)))
	  '((or (= p 1)(not (= q 1)))
	    ((= q 1))
	    ((not (= p 1)))
	    (or (= r 1)(not (= q 1))))
	  :name 'test5
	  :perform-search? t 
	  :return-conflict? t
	  :key #'proposition-to-truth-assignment)
	 :multiple-values t
	:test #'equal-consistent?-conflict-result
	:fail-info "P=1 or not Q=1, Q=1, not P=1,R=1 or not Q=1 ~
                    -> Conflict () meaning inconsistent theory.")
  )

(deftest (test6 :isat) ()
  (test `(nil 
	  ((= (= P 1) ,*false*)(= (= Q 1) ,*true*)))
	 (consistent-with-state-clauses?
	  '((p (1 2))
	    (q (1 2))
	    (r (1 2)))
	  '((or (= p 1)(not (= q 1))))
	  :added-clauses  '((or (= r 1)(not (= q 1))))
	  :added-assignments `((q 1)
						   (p 1 ,*false*))
	  :source 'test6
	  :perform-search? t 
	  :return-conflict? t
	  :key #'proposition-to-truth-assignment)
	 :multiple-values t
	:test #'equal-consistent?-conflict-result
	:fail-info "P=1 or not Q=1, R=1 or not Q=1, Q=1, not P=1, ~
                    -> Conflict Q=1 and not P=1."
	))

(deftest (test7 :isat) ()
  (test '(t nil)
	 (inconsistent-with-state-clauses? 
	  '((p (1 2))
	    (q (1 2))
	    (r (1 2)))
	  '((or (= p 1)(not (= q 1))))
	  :added-clauses  '((or (= r 1)(not (= q 1))))
	  :added-assignments `((q 1)
			       (p 1 ,*false*))
	  :return-conflict? T
	  :source 'test7)
	 :multiple-values t
	:test #'equal-consistent?-conflict-result
	:fail-info "P=1 or not Q=1, R=1 or not Q=1, Q=1, not P=1, ~
                    -> Inconsistent"))

(deftest (test8 :isat) ()
  (test
   `(nil 
     ((= (= Q 1) ,*false*)))
    (inconsistent-with-state-clauses? 
     '((p (1 2))
       (q (1 2))
       (r (1 2)))
     '((or (= p 1)(not (= q 1))))
     :added-clauses  '((or (= r 1)(not (= q 1))))
     :added-assignments `((q 1 ,*false*)
			  (p 1 ,*true*))
     :return-conflict? T
     :base-theory-consistent-with-all-assignments? T
     :source 'test8)
    :multiple-values t
   :test #'equal-consistent?-conflict-result
   :fail-info "P=1 or not Q=1, R=1 or not Q=1, not Q=1, P=1, ~
               -> Consistent, Conflict not Q=1"))
