;;; -*- Mode:Common-Lisp; Package:ISAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ISAT)

;;; ***********************************************************************
;;;
;;;                Incremental Unit Propagation
;;;                (aka Truth Maintenance)
;;;
;;; ***********************************************************************

;;; This module implements incremental unit propagation.  
;;; A unit clause is a clause with exactly one unassigned
;;; literal and all other literals are false.  Unit clause resolution 
;;; says that the unassigned literal must be assigned true in order to 
;;; satisfy the clause.  Assigning True to a literal may cause other 
;;; clauses to become unit, enabling further application of unit clause 
;;; resolution.  Unit propagation computes the closure of unit clause 
;;; resolution on Theory through a local graph search.
;;;
;;; This module includes the following interface functions,
;;; used internal to ISAT.

;;; ASSIGN-TRUTH-VALUE (PROPOSITION TRUTH-VALUE SUPPORT THEORY)
;;;   Assigns TRUTH-VALUE to PROPOSITION and then continues
;;;   unit propagation based on this value.  
;;;   SUPPORT is an sexpression indicating the reason for an 
;;;   assignment.  If (and only if) the assignment was derived
;;;   by unit propagation, then support is the clause used in
;;;   the propagation.
;;;   If no inconsistency is found the fn returns T; otherwise, 
;;;   propagation is immediately terminated when an inconsistency 
;;;   is found and the function returns NIL.  
;;;   When the function returns, the set of unit clauses and the 
;;;   set of inconsistent clauses are properly updated for THEORY.


;;; UNASSIGN-TRUTH-VALUE (PROPOSITION SUPPORT THEORY)
;;;   Removes the truth value currently given to PROPOSITION by Support.
;;;   SUPPORT is an sexpression indicating the reason for the
;;;   assignment.  If (and only if) the assignment was derived
;;;   by unit propagation, then support is the clause used in
;;;   the propagation.
;;;   This function recursively unassigns any truth values of 
;;;   propositions that are a consequence of proposition.  
;;;   It also updates the status of all affected clauses.  Returns T.
  
;;; PROPAGATE ((THEORY THEORY))
;;;   Given THEORY, propagates the unit clauses of THEORY until the 
;;;   set of unit clauses become empty, or an inconsistency is detected.  
;;;   If the set of unit clauses becomes empty, then all propositions 
;;;   have truth values that are "unit propagation entailed" by THEORY, 
;;;   and T is returned.  If an inconsistency is detected, the function 
;;;   returns NIL.  
;;;   No truth assignments can be believed if the theory is inconsistent.
;;;   Whenever this function returns, the set of unit clauses are 
;;;   correctly set.

;;; UNDO-CLAUSE-PROPAGATION (CLAUSE THEORY)
  
;;;   Undoes all current truth assignments and support
;;;   that resulted from propagating CLAUSE in THEORY.  It also
;;;   updates any clause status of THEORY that has changed.
;;;   Returns T.

;;; This implementation has the following properties:

;;; o Calls to assign-truth-value, un-assign-truth-value
;;;   and undo-clause-propagation propagate their
;;;   effects immediately.  

;;; o The exception is that, for unassign-truth-value,
;;;   and undo-clause-propagation, resupports are queued 
;;;   as unit clauses, which are propagated when PROPAGATE
;;;   is called.

;;; o Immediate propagation of the effects of assign-truth-value,
;;;   unassign-truth-value and undo-clause propagation is
;;;   performed through recursive function calls.

;;; o Propagation of the effects of assert! and insert-clause is
;;;   delayed until PROPAGATE is called, by using the unit propagation
;;;   queue, 

;;; o The effects of unassert! and remove-constraint are
;;;   propagated immediately, except for resupports.

;;; Removed from external interface.
;;;(eval-when (compile load eval)
;;;  (export '(propagate
;;;			;; PROPAGATE shouldn't be part of the external interface.
;;;			;; Instead, any function that requires that propagation be 
;;;			;; performed should have the appropriate ISAT interface
;;;			;; function call propagate when needed.
;;;
;;;			;; (Un)assign-truth-value are part of the module interface,
;;;			;;  but are not exported outside of ISAT.
;;;            )
;;;		  :ISAT)
;;;  )

;;; To Do:

;;; o There is a set of old functions commented out.  As soon as their
;;;   new versions run through the tests, delete the old versions.

;;; o Propagate shouldn't be exported outside of ITMS.  Find and fix
;;;   any callers.

;;; Optimization:
;;; o This code should not propagate effects through recursive calls,
;;;   it should instead use a set of stacks, built on top of growable
;;;   arrays.
;;;
;;; o This code centers around unit clauses, rather than assignments.
;;;   1) Assignments are made by inserting unit clauses, 2) only unit
;;;   clauses are queued for processing, not assignments, and 
;;;   3) conflicts are taken to be sets of inconsistent clauses.  
;;;   It is probably  more efficient to focus on assignments as first
;;;   class objects in all three cases.

;;; ***********************************************************************
;;;
;;;                Propagate and Its Supporting Functions
;;;
;;; ***********************************************************************

(defmethod propagate ((theory theory))
  
  "Given THEORY, propagates the unit clauses of THEORY until there 
   are no unit clauses remaining, or an inconsistency is detected.  

   If an inconsistency is found, NIL is returned, otherwise T is returned.  

   No truth assignments can be believed if the theory is inconsistent.  
   Whenever this function returns, the set of unit clauses haven been
   correctly set."

  (when *isat-statistics-p*
    (incf *propagate-counter*))
  (debug-print-for :itms t "Propagate.")
  (cond ((no-inconsistent-clauses? theory)
		 ;; Then propagate...
         (loop
		   ;; Loop until there are no more unit clauses.
		   (let ((unit-clause (get-unit-clause theory)))
			 (cond ((null unit-clause)		    

					;; If there are no more unit clauses, 
					;; then return immediately.
					;;   No inconsistency was discovered, 
					;;   hence propagation terminates successfully.
;;;				(debug-print-for :itms t "  Consistent.")
					(return (values t)))

				   (:else
					;; Otherwise, propagate the first unit clause.  
;;;				(debug-print-for :itms t "  Propagate unit clause ~A." 
;;;								 unit-clause)
					(unless (propagate-clause unit-clause theory)
					  ;; Then propagation leads to an inconsistency.
					  (debug-print-for :itms t 
									   "  Inconsistent for unit-clause ~a." unit-clause)
					  ;; and return NIL, indicating an inconsistency is found.
					  (return (values nil))))))))

        (:else
		 ;; Else theory is immediately found to be inconsistent.
		 ;; Return NIL, indicating an inconsistent theory.
		 (debug-print-for :itms t "  Inconsistent immediately.")
		 ;; and return NIL
         (values nil))))

;;; FIX:
;;; Why is values used here?  Only one value is returned.
(defmethod propagate-clause ((clause clause) (theory theory))
  
  " Propagates a unit clause.  If the propagation terminates
    successfully, the function returns T.  If the propagation detects an
    inconsistency, the function returns NIL.  If CLAUSE isn't a unit
    clause it immediately returns T."
  
  (if (unit-clause? clause)

	  ;; When CLAUSE is a unit clause, it can be propagated.
      (multiple-value-bind
		  (propn truth-value)
		  (find-single-unassigned-proposition clause)
		(debug-progn
		 (when (null propn)
		   ;; There should have been exactly one unassigned proposition.
		   ;; This condition can't occur.
		   (error " Unit clause ~a has no unassigned or ~
                             more than one unassigned proposition."
				  clause)))

		;; Assign TRUTH-VALUE to PROPN, and return the result.  
;;;		(debug-print-for :itms t "  ~a = ~a by ~a." propn truth-value clause)

		(values (assign-truth-value propn truth-value clause theory)))

	;; otherwise it isn't a unit clause, so return T
    (values t)))
  
(defmethod find-single-unassigned-proposition ((clause clause))
  
  " Searches the literals in CLAUSE and returns the unique
    unassigned proposition.  As a second value it returns *TRUE* if
    the proposition occurs positively, and *FALSE* if it occurs negatively.
    If there is no unassigned proposition, or more than one unassigned
    proposition, it returns NIL."
  
;;;  (debug-print-for :itms t "  Find unassigned literal in ~A."
;;;				   clause)

  (let ((unassigned-proposition nil)
		(truth-value nil))
    (do-clause-propositions (proposition polarity clause)
      
      (cond ((not (eql (proposition-truth-value proposition)
					   *none*))
;;;			 (debug-print-for :itms t "  Literal ~A with polarity ~A is already assigned."
;;;							  proposition polarity)
			 )
			(:otherwise
			 ;; Then it is unassigned.
;;;			 (debug-print-for :itms t "  Literal ~A with polarity ~A is unassigned."
;;;							  proposition polarity)
			 
			 (cond ((not (null unassigned-proposition))
					;; Then we've discovered a second unassigned proposition,
					;; return failure (NIL).
;;;					(debug-print-for :itms t "  Already found ~A as unassigned, ~
;;;                                    clause has multiple unassigned literals."
;;;									 proposition)

					(return-from find-single-unassigned-proposition 
					  (values nil nil)))
				   (:else
					;; Else indicate the truth value to be given to
					;; UNASSIGNED-proposition
					(setq unassigned-proposition proposition)
					(setq truth-value polarity))))))

	;; and return the values
;;;    (debug-print-progn
;;;     (if (null unassigned-proposition)
;;;		 (debug-print-for :itms t "  No unassigned literals in ~A."
;;;						  clause)
;;;       (debug-print-for :itms t "  Unassigned literal ~A with polarity ~A in ~A."
;;;						unassigned-proposition truth-value clause)))
    
    (values unassigned-proposition truth-value)))

;;; To Do:
;;;  o Update (un)assign/assert calls so that they call
;;;    (un)assign-truth-value rather than inserting unit-clauses.
;;;    At the same time, re-examine the functions for extracting
;;;    conflicts.

;;; Optimization:  
;;;  o It is MUCH more efficient to use an explicit 
;;;    unit-clause queue, maintained by an array or by managing the 
;;;    list of conses, rather than using the program stack.

;;;  o Make this a truth assignment queue, rather than a unit-clause
;;;    queue.  This way we don't have to duplicate effort re-extracting
;;;    the proposition and truth assignment.  If this is the way
;;;    we go, then maintain two queues, one with the proposition and the
;;;    other with its assigned truth value.  This approach won't work
;;;    if we need to re-examine a unit clause to ensure it is unit
;;;    before making the truth assignment.

;;; o The use of clause-consequent to also set clause-satisfied-by,
;;;   appears to make the code more opaque to understand, at the savings
;;;   of one pointer.  Consider removing this optimization.

;;;   o We also may want the ability to delay propagation.

;;; FIX:
;;; o check if proposition is being assigned a truth-value when it
;;;   is already assigned the opposite truth value.  If this happens,
;;;   flag an error.
(defmethod assign-truth-value 
	((proposition proposition) truth-value support (theory theory))

;  (format t "~%~a | ~a | ~a ~%" proposition truth-value support)

  " Assigns TRUTH-VALUE to PROPOSITION in THEORY by reason of SUPPORT,
    and continues unit propagation.

    If no inconsistency is found it returns T; otherwise, 
    propagation is immediately terminated when the inconsistency 
    is found and the function returns NIL.  

    SUPPORT is either a clause, used by unit propagation to derive
    truth-value, or an sexpression, denoting an EXTERNAL reason for support.

    An external support overrides a previous support by unit propagation,
    the purpose of this is to reduce the size of the set of support for
    each proposition.

    It is currently an error to provide multiple external support for
    the same proposition.

    If assignment is inconsistent with a current assignment by unit
    propagation, then theory will be updated to reflect the 
    inconsistency.  However, it is recommended that the caller
    check for an existing inconsistent assignment, and if found,
    not try to make the new assignment.

    When the function returns, the set of unit clauses and set 
    of inconsistent clauses are properly updated for THEORY. "

  (when *isat-statistics-p*
    (incf *assign-truth-value-counter*))

;  (format t "in-assign-truth-value ~a ~a ~a ~%" proposition (proposition-truth-value proposition) (assigned? proposition))

  (cond ((assigned? proposition)

;		 (format t "already assigned~%"); DAVID ADDED

		 ;; Already assigned, update assignment.
		 (debug-print-for :itms t "  ~a = ~a changes support" 
						  proposition truth-value)
		 (debug-print-for :itms t "         to ~a." 
						  support)
		 (values (reassign-truth-value-of-assigned-proposition
				  proposition truth-value support theory)))

		(:otherwise
		 
;		 (format t "not already assigned~%"); DAVID ADDED

		 (debug-print-for :itms t "  ~a = ~a" 
						  proposition truth-value)
		 (debug-print-for :itms t "         by ~a." 
						  support)
		 ;; Was unassigned; assign it.
		 ;; First assign the truth value
		 (setf (proposition-truth-value proposition) truth-value)

		 ;; and setup the support/consequent relationship.
		 (if (unit-propagation-support? support)
			 (set-support-relation! proposition support)
		   (setf (proposition-support proposition) support))

		 (let ((inconsistency-detected? nil))
		   ;; Next, update the status of all clauses containing PROPOSITION.
		   ;; This updates the set of inconsistent clauses and removes any 
		   ;; propagated clause from the set of unit clauses.

		   (if (not (update-proposition-clauses proposition theory))
			   ;; an immediate inconsistency was detected
			   (progn
				 (setq inconsistency-detected? t)
;				 (format t "incon. detected 1 ~%")
				 )
			 
			 ;; Else, the update did not detect an immediate inconsistency,  
			 ;; propagate any new unit clauses.
			 (unless (propagate-proposition-clauses proposition theory)
			   ;; the propagation leads to an inconsistency
			   (setq inconsistency-detected? t)
;			   (format t "incon. detected 2 ~%") 
			   ))

		   ;; Note, that the call to UPDATE-PROPOSITION above DOES NOT add any 
		   ;; new clauses to the set of unit clauses.  Instead, the set of 
		   ;; new unit clauses involving PROPOSITION are immediately 
		   ;; propagated below, by the call to PROPAGATE-PROPOSITION-CLAUSES.  
		   ;; Immediate propagation uses the program stack to keep track 
		   ;; of propagation, rather than an explicit queue.

		   ;; If PROPAGATE-PROPOSITION-CLAUSES returns with consistency, 
		   ;; then the unit-clauses will no longer be unit when
		   ;; propagate-proposition-clauses returns.

		   ;; However, if an inconsistency is detected, then propagation
		   ;; terminates before completion, and we record any clauses
		   ;; containing PROPOSITION that are still unit.
		   (when inconsistency-detected?
			 (add-proposition-clauses-to-unit-clauses proposition theory))

		   ;; Return True unless an inconsistency was detected.
		   (values (not inconsistency-detected?))))))

(defmethod check-new-old-support! 
	((proposition proposition) new-support old-support)

  "It is an error to provide support for an already supported
   proposition unless new support is external and old support is
   via unit propagation."
  
  (if (unit-propagation-support? new-support)
      
	  (error "Proposition ~A already has support ~A, ~
              Shouldn't give it new support ~A by unit propagation."
			 proposition old-support new-support)
	
    (unless (unit-propagation-support? old-support)
      (error "Proposition ~A already has external support ~A, ~
              Shouldn't give it new external support ~A."
			 proposition old-support new-support))))

(defmethod update-proposition-clauses 
	((proposition proposition) (theory theory))
  
 ; (format t "in-update-proposition-clauses before~%")

  " Updates the status of all clauses in which PROPOSITION 
    occurs,including support, unassigned-literal-count, and 
    inconsistent-clauses.  

    Returns T if no inconsistent clauses were 
    discovered,and NIL otherwise.  

    Assumes that PROPOSITION has just been assigned a 
    truth value.

    Requires that the caller take care of propagating 
    proposition's unit clauses, recording them up for
    later propagation when an inconsistency has been found."
  
  (let* ((inconsistent? nil)
		 (truth-value (proposition-truth-value proposition))
		 (positive-literal-true? (eql truth-value *true*))
		 (negative-literal-true? (eql truth-value *false*)))

	;; Update each clause in which the proposition occurs.  
	;; If an inconsistent clause is found, set 
	;; INCONSISTENT? to True.
    (do-proposition-inserted-clauses-w-polarity
		(clause polarity proposition)
      (unless (update-proposition-clause
			   clause
			   proposition
			   (if (eql polarity *true*)
				   positive-literal-true?
				 negative-literal-true?)
			   theory)
		(setq inconsistent? T)))

;	  (format t "in-update-proposition-clauses after~%")

	;; and return T if not inconsistent, else nil.
    (values (not inconsistent?)) 
))

;;; Optimization:
;;; o Call to remove-from-unit-clauses is an expensive operation and 
;;;   is unnecessary.  If clause is no longer unit it can be detected 
;;;   when clause is removed from the queue.
(defmethod update-proposition-clause ((clause clause) 
									  (proposition proposition) 
									  proposition-satisfies-clause?
									  (theory theory))

    
  " Updates the status of CLAUSE,including support, 
    unassigned-literal-count,and inconsistent-clauses, given 
    that its PROPOSITION has just been assigned.

    The exception is that, if clause has just become unit, it DOES NOT
    add clause to the unit clauses of theory, this is performed by 
    ASSIGN-TRUTH-VALUE.  

    The function returns T if an inconsistent clause was NOT discovered, 
    and NIL otherwise.  

    Note that this function assumes that PROPOSITION has just been assigned 
    a truth value."

  (when *isat-statistics-p*
    (incf *update-proposition-clause-counter*))

  ;; If the clause is in the set of unit clauses, the assignment to the final
  ;; unassigned literal definitely removes it from the set of unit clauses, 
  ;; either by satisfying it or by making it inconsistent.
  (remove-from-unit-clauses clause theory)

  ;; Reduce the number of unassigned literals.
  (decf (clause-unassigned-literal-count clause))

  ;; Record satisfier for CLAUSE when it doesn't have one and the 
  ;; newly assigned proposition satisfies it.
  (when (and proposition-satisfies-clause?
			 (null (clause-satisfying-proposition clause)))
    (setf (clause-satisfying-proposition clause) proposition))

  ;; Finally, record clause if inconsistent and return T iff consistent.
  (cond ((inconsistent-clause? clause)
		 ;; add this clause to the list of inconsistent clauses.
		 (add-inconsistent-clause clause theory)
		 ;; Return that an inconsistency has been found.
		 (values nil))
        (:else
         (values t))))

;;; Optimization:
;;; o Propagation should be managed by pushing new unit clauses on
;;;   the unit clause queue and iterating through the queue.  NOT
;;;   through the recursive function call here.  Make sure when
;;;   making this modification that inconsistent clause detection 
;;;   is handled correctly.  Queueing can be performed using 
;;;   ADD-PROPOSITION-CLAUSES-TO-UNIT-CLAUSES.
(defmethod propagate-proposition-clauses 
	((proposition proposition) (theory theory))
  
  " Propagates all unit clauses containing PROPOSITION, which
    is newly assigned PROPOSITION.
    If propagation completes successfully, it returns T.  
    If an inconsistency is detected, it returns NIL."
  
  ;; Try to propagate each clause containing proposition.
  (do-proposition-inserted-clauses (clause proposition)
    (unless (propagate-clause clause theory)
	  ;; an inconsistency has been detected, so return immediately.
      (return-from propagate-proposition-clauses
		(values nil))))
  
  ;; Propagation was successful
  (values t))

;;; Optimization:
;;; o Adding and removing clauses from the list of unit-clauses
;;;   generates garbage.  Can this be done more efficiently?
(defmethod add-proposition-clauses-to-unit-clauses 
	((proposition proposition) (theory theory))
  
  " Adds any unit clause containing PROPOSITION to the set of unit 
    clauses of theory, after proposition has just been assigned a 
    truth value.  T returned."

  ;; If the new assignment to PROPOSITION is *TRUE* then only 
  ;; clauses containing PROPOSITION in a negative literal need 
  ;; to be considered.
  ;; The reason for this is that any clause containing proposition within a
  ;; positive literal will be satisfied by the new assignment, 
  ;; and hence is not unit.

  (let ((clauses-to-check
		 ;; Clauses in which a literal containing proposition went *false*.
		 (if (eql (proposition-truth-value proposition)
				  *true*)
			 (proposition-negative-clauses proposition)
		   (proposition-positive-clauses proposition))))
    (do-inserted-clauses (clause clauses-to-check)
      (when (unit-clause? clause)
		(add-to-unit-clauses clause theory)))
    T))

;;; ***********************************************************************
;;;
;;;         Undo-Clause-Propagation and its Supporting Functions
;;;
;;; ***********************************************************************

;;; Note About Resupport:
;;;   An incremental unit propagation (IUP) algorithm maintains for each
;;;   assigned proposition a support clause used to derive that assignment.
;;;   When a clause can no longer be used to support the truth assignment
;;;   of a proposition, an IUP algorithm tries to resupport the truth 
;;;   assignment using an alternative clause.

;;;   In the following code, resupport is less explicit than in a traditional
;;;   JTMS/LTMS algorithm.  In particular, in this code, when a proposition 
;;;   P loses its truth assignment, incremental unit propagation (IUP) 
;;;   identifies all unit clauses containing P and queues them for unit
;;;   propagation.  These unit clauses are exactly those that can
;;;   be used to resupport P.

;;;   A JTMS/LTMS instead maintains a list of propositions that need to
;;;   be resupported, and it delays checking the clauses of each
;;;   proposition until a time at which that proposition can be safely 
;;;   resupported.  

;;; Optimization:
;;; o The following uses recursion where it should instead use a queue.
;;;   The queue (stack) can be implemented using a growable array.

;;; o The standard TMS approach of identifying relevant clauses 
;;;   just before resupport avoids placing them on a queue, and 
;;;   seems potentially more efficient.

;;; o The assignment functions are implemented by generating unit
;;;   clauses, and is inefficent.  Reimplement these functions on 
;;;   top of assign and unassign-truth-value.
(defmethod undo-clause-propagation 
	((clause clause) (theory theory))
  
  " Undoes all current truth assignments and support
    that resulted from propagating CLAUSE in THEORY.  
    Also updates any clause status of THEORY that 
    has changed.
    
    Called for effect, returns T."
  
  (let ((consequent (clause-consequent clause)))
    (unless (null consequent)
	  ;; There is only something to undo if there is a CONSEQUENT.
	  ;; Reset the support relation between CLAUSE and CONSEQUENT,...
;;;      (debug-print-for :itms t "  ~A support ~A removed." 
;;;					   consequent clause)

;;;	  (debug-print-for :itms t "Remove ~a and consequences." consequent)
	  ;; Unassign CONSEQUENT's truth value and support.
	  ;; This removes all consequences of this assignment as well.
      (unassign-truth-value consequent clause theory)
      T)))

;;; Optimization:
;;;  o If support is a CLAUSE, then it will become
;;;    unit, and will be queued for unit propagation
;;;    and after PROPAGATION is called, support will be
;;;    reinstalled.  In the case where proposition's
;;;    truth value is flipped, queing the clause is
;;;    unnecessary.  Is it worth avoiding this?
(defmethod unassign-truth-value 
	((proposition proposition) support (theory theory))
  
  " Removes the truth value currently given to PROPOSITION, 
    and recursively unassigns any truth values of 
    propositions that are a consequence of proposition.  
    Updates the status of all affected clauses.  

    Called for effect only, returns T."
  
  ;; Check that a correct support is provided.
  (debug-progn
   (unless (eql support
				(proposition-support proposition))
     (warn "~A is not a support for ~A in unassign-truth-value." 
		   support proposition)))
  
  (let ((truth-value (proposition-truth-value proposition)))
    (unless (eql truth-value *none*)
	  ;; Proceed only when there is a truth value to unassign.
      (when *isat-statistics-p*
		(incf *unassign-truth-value-counter*))

	  ;; Remove proposition's truth assignment and support.
      (setf (proposition-truth-value proposition) *none*)
      (debug-print-for :itms t "  ~a = ~a" 
					   proposition truth-value)
      (debug-print-for :itms t "         by ~a removed." 
					   support)
      (if (unit-propagation-support? support)
		  (clear-support-relation! proposition support)
		(setf (proposition-support proposition) nil))

	  ;; Update the status of all inserted clauses that contain proposition.
	  ;; Status involves updating the SATISFYING-PROPOSITION and 
	  ;; UNASSIGNED-LITERAL-COUNT of clause, and the set of unit clauses 
	  ;; and inconsistent clauses of THEORY.

	  ;; Note that if proposition was externally supported, 
	  ;; This tries to RESUPPORT proposition, by
	  ;; propagating clauses that have become unit
	  ;; from this unassignment.
      (do-proposition-inserted-clauses (clause proposition)
		(update-clause-status-given-unassigned-proposition 
		 clause proposition support theory))

	  ;; Finally, undo the consequences of each clause 
	  ;; supported by proposition.
      (if (eql truth-value *true*)
		  ;; Then consequences are within the negative clauses.
		  (do-inserted-clauses (clause (proposition-negative-clauses
										proposition))
			(undo-clause-propagation clause theory))

		;; Else the consequences are within the positive clauses.
		(do-inserted-clauses (clause (proposition-positive-clauses
									  proposition))
		  (undo-clause-propagation clause theory)))
      T)))

;;; Optimization:
;;; o Use literal count to decide when to remove unit and inconsistent
;;;   clauses, as documented below.  Consider whether this removal
;;;   needs to be performed greedily here, or delayed until the clauses 
;;;   are examined for some use.  See comments in the following function.

(defmethod update-clause-status-given-unassigned-proposition
    ((clause clause) unassigned-proposition support (theory theory))
  
  " Updates the status of CLAUSE, given that the truth 
    value of UNASSIGNED-PROPOSITION has just been removed.  

    Called for effect, returning T.

    Updating status includes changing the SATISFYING-PROPOSITION 
    and UNASSIGNED-LITERAL-COUNT of CLAUSE,  and making sure that 
    CLAUSE is correctly added/removed from the set of unit clauses 
    and inconsistent clauses of theory.

    Assumes that UNASSIGNED-PROPOSITION's truth value has 
    already been set to NIL."
     
  (when *isat-statistics-p*
    (incf *update-clause-for-unassigned-proposition-counter*))

  ;; Optimization:
  ;; o Should be able to decide whether or not to perform
  ;;   the following two removals based on literal count.
  
  ;; If CLAUSE was in the set of unit clauses, then removing the assignment of
  ;; UNASSIGNED-PROPOSITION forces CLAUSE out of the set of unit clauses.
  (remove-from-unit-clauses clause theory)

  ;; If CLAUSE was in the set of inconsistent clauses, then removing 
  ;; the assignment of UNASSIGNED-PROPOSITION allows clause to be 
  ;; satisfiable, forcing CLAUSE out of the set of
  ;; inconsistent clauses.
  (remove-inconsistent-clause clause theory)

  ;; The number of unassigned literals increases by one.
  (incf (clause-unassigned-literal-count clause))

  (unless (eql clause support)
	;; If CLAUSE was the SUPPORT for UNASSIGNED-PROPOSITION,
	;; then the support relation and satisfying proposition
	;; for CLAUSE have already been updated (note that if
	;; CLAUSE supports UNASSIGNED-PROPOSITION, then there 
	;; is no alternative proposition that satisfies CLAUSE).
    
	;; Also note that if CLAUSE is the support for PROPOSITION
	;; then the consequences of removing support is handled
	;; by UNASSIGN-TRUTH-VALUE, by the call to 
	;; UNDO-CLAUSE-PROPAGATION on CLAUSE.

	;; Otherwise ...
    
    (when (eql (clause-satisfying-proposition clause) 
			   unassigned-proposition)

	  ;; When UNASSIGNED-PROPOSITION is listed as the satisfying 
	  ;; proposition of CLAUSE, try to find an alternate satisfying 
	  ;; proposition for CLAUSE.
      (setf (clause-satisfying-proposition clause) nil)
    
      (block find-satisfying-proposition
		(do-clause-propositions
			(proposition polarity clause)
		  (when (eql (proposition-truth-value proposition) polarity)
			;; it satisfies the clause.
			(setf (clause-satisfying-proposition clause) proposition)
			(return-from find-satisfying-proposition))))))

  ;; Finally if clause becomes a unit clause as a result of the unassign, 
  ;; then add to theory's set of unit clauses.  This step provides two 
  ;; key functions:
  ;;     o It queues any clause for re-propagation that might be used
  ;;       to resupport a proposition that just lost its support.
  ;;     o It queues a previously inconsistent clause when
  ;;       one of its assignments has been relaxed. 

  ;; Optimization:
  ;; o Should be able to decide whether or not to perform
  ;;   this addition earlier based on literal count.
  (when (unit-clause? clause)
    (add-to-unit-clauses clause theory))
  T)

;;; ***********************************************************************
;;;
;;;         Reassigning an Already Assigned Proposition
;;;
;;; ***********************************************************************

(defmethod reassign-truth-value-of-assigned-proposition 
    ((proposition proposition) truth-value support (theory theory))

  "Assign proposition when proposition has already been assigned."
  
  ;; If already supported, then unless an error,
  ;; old support should from unit propagation, and 
  ;; new support is external.
  (let ((old-truth-value (proposition-truth-value proposition))
		(old-support (proposition-support proposition))
		(consistent? nil))
    (debug-print-for :itms t "  ~a = ~a already,"
					 proposition old-truth-value old-support)
    (debug-print-for :itms t "         by ~a."
					 old-support)

	;; Check that valid support is provided.
	;; Can't double support- caller error.
	;; Can't unit propagation support if already assigned-internal error.
    (debug-progn (check-new-old-support! 
				  proposition support old-support))

    (cond ((eql truth-value old-truth-value)
		   ;; Assignments agree, swap support.
		   (debug-print-for :itms t "    swap support to ~A." support)
		   (swap-proposition-support! proposition support old-support)
		   ;; and return True, since no inconsistency was detected.
		   (setq consistent? T))
	  
		  (:otherwise
		   ;; Assignments disagree. Swap assignments, since an 
		   ;; external assignment over-rides a unit propagation assignment.
	   
		   (debug-print-for :itms t "    disagrees with ~a"
							truth-value)
		   (debug-print-for :itms t "           by ~a."
							support)
		   (debug-print-for :itms t "    Swap, producing inconsistency.")
	   
		   ;; Note:  For the following to work correctly,
		   ;; assign-truth-value MUST be called BEFORE processing
		   ;; any unit clauses that have been pushed on the queue.
		   (unassign-truth-value proposition old-support theory)
		   (setq consistent?
			 (assign-truth-value proposition truth-value support theory))
		   ;; Theory must be inconsistent.
		   (debug-progn
			(when consistent?
			  (error "Swapping truth assignments of ~A should have ~
                      made theory ~a inconsistent but it isn't."
					 proposition theory)))))
    consistent?))

(defmethod swap-proposition-support! 
	((proposition proposition) support old-support)
  
  "Replace proposition's unit propagation support with
   an external support."

  ;; Clause OLD-SUPPORT no longer supports proposition...
  (clear-support-relation! proposition old-support)
   

 ;;  (set-support-relation! proposition support)
  ;; ... but clause OLD-SUPPORT is still satisfied 
  ;; by proposition.
  (if (typep old-support 'clause)
	  (setf (clause-satisfying-proposition old-support)
			proposition))

  ;; Indicate the new external support...
  (setf (proposition-support proposition) support)

  T)

;;; To Do:
;;; o Delete the following as soon as we can swap the assignment of
;;;   a proposition correctly.

;;; To flip the truth assignment of a proposition we would need
;;; to do the following:

;;; Remove the old assignment and its support.
;;; Assign the new truth assignment, and record its support.

;;; Loop for clauses with the opposite polarity from new assignment.
;;;   clause must have been satisfied:
;;;   When the satisfier of clause is proposition,
;;;   Then Remove proposition as satisfier, 
;;;        Look for a new satisfier and if found replace, and
;;;        If unassigned literal count = 0 
;;;        Then move clause to inconsistent.
;;;        Else If unassigned literal count = 1
;;;             Then move to unit clause.
		  
;;; Loop for clauses with the same polarity as new assignment.
;;;   proposition can not have satisfied clause.
;;;   If clause was on inconsistent
;;;   Then remove clause from inconsistent
;;;        record that clause is satisfied by proposition.
;;;   If clause was on unit clauses of theory
;;;   Then remove clause from unit clauses of theory
;;;        record that clause is satisfied by proposition.
;;;   Else If clause does not have a satisfier
;;;        Then record that clause is satisfied by proposition.
