;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2008-2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;; ***********************************************************************
;;;
;;;        To Do:
;;;
;;; ***********************************************************************

;;; o A bug was found in conflict detection by Andrew Yi Huang.
;;;   Code to correct this was added to propagate-ub.lisp.
;;;   This code needs to be 
;;;   - reviewed for correcteness,
;;;   - cleaned up,
;;;   - optimized,
;;;   - renamed,
;;;   - documented, and
;;;   - moved here.

;;; ***********************************************************************
;;;
;;;        Conflict Management for Incremental Temporal Consistency
;;;
;;; ***********************************************************************

;;; Last update: October 11th, 2010.


;;; This module contains functions for managing inconsistencies in a
;;; simple temporal network, including 1) detecting an inconsistency
;;; during propagation, 2) for extracting a set of inconsistent
;;; distance edges or constraints (a conflict), and for 3) rolling back
;;; propagation to remove an inconsistency.  An inconsistency is
;;; detected during upper bound propagation.

;;; A simple temporal network (STN) is inconsistent if and only if there
;;; exists a negative loop in the set of support of the STN's distance graph.
;;; A conflict corresponds to the set of distance edges or corresponding
;;; constraints in the negative loop.  An inconsistency is rolled back by
;;; removing the negative loop and its consequences, and by setting up 
;;; the propagation queue with the events at the fringe of what has been
;;; removed.

;;; Module interface function consists of:

;;;  COMPLETED-UB-CYCLE? (V)
;;;
;;;  Returns True if a negative cycle has been closed at event V
;;;  during upper bound propagation, respectively, 
;;;   else returns nil.  
;;;   V is an event whose upper bound value has just been updated.
;;;   A side effect is that V is marked as ub-visited,
;;;     this is reset by event-ub-initialize!

;;; EXTRACT-UB-CONFLICT (C &OPTIONAL (L '()))
;;;
;;;   Return a conflict, comprised of a set of distance edges
;;;   that form a negative loop.  The negative loop was formed
;;;   during upper bound propagation, respectively.
;;;
;;;   C is an event in a negative loop and
;;;   L is a list of distance edges supporting C within
;;;     that negative loop.

;;; STN-REMOVE-UB-CONFLICT! (CONFLICT STN)
;;;
;;;   Removes a CONFLICT, computed during upper bound propagation,
;;;   and its effects, and sets up upper bound propagation.
;;;   CONFLICT is a list of distance edges that form a negative 
;;;   cycle within STN.

;;; EXTERNALIZE-CONFLICT (CONFLICT KEY)
;;;  
;;;   Maps a conflict from internal to external form.
;;;   Applies key to each element of conflict and returns
;;;   a list of non-null results.

;;; ***********************************************************************
;;;
;;;                Detecting an Inconsistency (Negative Loop)
;;;
;;; ***********************************************************************

;;; The next two routines detect inconsistency and the subsequent
;;; two extract a conflict.

;;;Boolean
;;;CompletedCycle(v)
;;;
;;;  // Returns True if v closes a negative cycle,
;;;  // else return False.
;;;
;;;{21} if v(v) = true;
;;;      // The support of v closes a negative loop
;;;{22}   return true;
;;;
;;;{23} else
;;;      // mark as visited
;;;{24}   v(v)= true;
;;;{25}   return false;

;;; Note:
;; - The following test is incorrect.
;;   If there are two paths to the same event, one providing
;;;  a tighter bound than the second, then completed-ub-cycle?
;;;  will flag a cycle, even though one doesn't exist.
;;;  Andrew Yi Huang found this bug and replacedd it.
;;;  - Brian Williams.
;;;(defmethod completed-ub-cycle? ((x event))
;;;
;;;  "Let X be an event whose upper bound value has just been updated.
;;;
;;;   Returns True if a negative cycle has been closed at event X
;;;   using ub propagation, else returns nil.  
;;;   
;;;   Side Effects:
;;;     V is marked as ub-visited,
;;;     this is reset by event-ub-initialize!"
;;;
;;;  (cond ((event-ub-visited? x)
;;;
;;;		 ;; If event V has already been visited, 
;;;		 ;; then the support of V closes a loop.
;;;		 'T)
;;;
;;;		(:otherwise
;;;
;;;		 ;; This is the first time cycle detection has visited V;
;;;		 ;; mark V as visited.
;;;		 (event-ub-visited! x)
;;;		 
;;;		 ;; No cycle detected yet.
;;;		 nil)))

;;; Note:
;;; o We only detect negative cycles on upper bounds, not lower bounds.

;;; o We haven't derived the correct condition for detecting a lower bound cycle.
;;;   The following is likely in correct.
;;;(defun completed-lb-cycle? (x)
;;;
;;;  "Returns True if a positive cycle has been closed at event X, 
;;;   using lb propagation, else return nil.  
;;;   X is an event whose lower bound value has just been updated.
;;;   A side effect is that X is marked as lb-visited,
;;;     this is reset by event-lb-initialize!"
;;;
;;;  (cond ((event-lb-visited? x)
;;;		 ;; If event X has already been visited, then the support of 
;;;		 ;; V closes a loop.
;;;		 'T)
;;;
;;;		(:otherwise
;;;		 ;; This is the first time cycle detection has visited X;
;;;		 ;; mark X as visited.
;;;		 (event-lb-visited! x)
;;;		 nil)))


;;; ***********************************************************************
;;;
;;;                Extracting a Conflict
;;;
;;; ***********************************************************************

;;;Conflict
;;;ExtractConflict(c,L)
;;;
;;; // Returns a conflict, a set of vertices
;;; // that form a negative loop.
;;; // c is a vertex in a negative loop and
;;; // l is a list of successor vertices of c
;;; //   in that loop.
;;;
;;;{19} if L.contains(c)
;;;       // Collected all vertices in the loop.
;;;{20}   return L;
;;;
;;;{21} else
;;;       // Continue collecting.
;;;{22}   L.add(c);
;;;{23}   ExtractConflict(p(c),L);

(defmethod extract-ub-conflict ((c event) &optional (L '()))

  "Let C be an event in a negative loop and
   L be a list of distance edges supporting C within
     that negative loop.

   Returns a conflict, comprised of a set of distance edges
   that form a negative loop in an upper bound computation."
  
  ;; Collect into L all distance edges in the negative loop,
  ;; working backwards from event C along C's support.
  (let ((e (event-ub-support c)))
	
	(cond ((member e L)
		   
		   ;; Already collected the support for event C,
		   ;; hence the loop has been closed.
		   (debug-print-for :itc t "    Loop closes at ~a" c)
		   
		   ;; Return the collected edges as the conflict 
		   ;; (negative loop).
		   L)

		  (:otherwise
		   
		   ;; Loop hasn't been closed yet.
		   
		   ;; Continue moving back along the support for C,
		   ;; collecting edges.
		   (push e L)
		   (debug-print-for :itc t "  Travere back through ~a." e)
		   
		   (extract-ub-conflict (d-edge-start e) L)))))

;;; Note:
;;; o We don't extract lb-conflicts; we haven't derived the condition
;;;   for detecting lb inconsistency.
;;;(defun extract-lb-conflict (c &optional (L '()))
;;;
;;;  "Return a conflict, comprised of a set of distance edges
;;;   that form a negative loop in a lower bound computation.
;;;
;;;   C is an event in a negative loop and
;;;   L is a list of distance edges supporting C within
;;;     that negative loop."
;;;  
;;;  ;; Collect all distance edges in negative loop L, 
;;;  ;; working backwards from event C along C's support.
;;;  (let ((e (event-lb-support c)))
;;;	(cond ((member e L)
;;;		   ;; C has already been collected, hence the loop is complete,
;;;		   ;; return the collected elements as the loop.
;;;		   L)
;;;
;;;		  (:otherwise
;;;		   ;; Loop hasn't closed yet, continue collecting back along support
;;;		   ;; from C. Support edge e supports its start c.
;;;		   (push e L)
;;;		   (extract-lb-conflict (d-edge-end e) L)))))

;;; ***********************************************************************
;;;
;;;                Rolling Back a Conflict
;;;
;;; ***********************************************************************

;;;The next two routines remove an inconsistency, in preparation for
;;; modifying a constraint, and modifies constraints.

;;;STN-remove-conflict!(set<Node> neg_cycle)
;;;
;;;  // Removes a negative cycle and its effects,
;;;  // and sets up propagation to check consistency.
;;;
;;;{53} set<Node> nodes_reset = {};
;;;{54} for all n in neg_cycle
;;;       // Reset every node in the cycle and their successors.
;;;{55}   d(n) := infinity;
;;;{56}   p(n) := unknown;
;;;{56a}  v(n) := false;
;;;{57}   nodes_reset.insert(n)
;;;{58}   for all m in InvalidateSupports(n)
;;;{58a}    nodes_reset.insert(m);
;;;{58b}    v(m) = false;
;;;
;;;  // Try resupporting all reset nodes.
;;;{59} InsertParents( nodes_reset );

;;; Optimization:
;;; o If edges are added but none are removed, then the 
;;;   conflict still holds, and no propagation is required.
;;;   Exploit this.

(defmethod STN-remove-ub-conflict! (conflict (STN STN))

  "Let CONFLICT be a list of distance edges that form a negative 
   cycle within simple temporal network STN, where
   CONFLICT was computed during upper bound propagation.

   Removes a CONFLICT and its effects, and sets up upper 
   bound propagation."

  ;; List of events that have been reset.
  (let ((events-reset '()))
	
	(debug-print-for :itc t "  Removing conflict ~a." conflict)

	;; Initialize every event supported by an edge in CONFLICT,
	;; along with all successors of those events.
	
	(loop for e in conflict
		for y = (d-edge-end e)
				;; upper bound value on y is now nonsense, initialize it.
				
		do (event-ub-initialize! y)
		   
		   (debug-print-for :itc t "    Initialize event ~a, ~
                                        and reset consequences"
							y)

		   ;; Keep track of reset events.
		   (push y events-reset)
		   
		   ;; The consequences of y are incorrect, reset them.
		   (setq events-reset
			 (reset-all-ub-consequences y STN events-reset)))

	;; Queue predecessors of all reset events to be relaxed.
	;; This is performed in order to look for resupport.
	(debug-print-for :itc t "UB queuing predecessors of ~a:"
					 events-reset)
	
	(queue-predecessors-for-ub-propagation! events-reset STN)
	
	t))

;;; Note:
;;; o We don't detect or remove lb-conflicts; we haven't derived the condition
;;;   for lb inconsistency.
;;;(defun STN-remove-lb-conflict! (conflict STN)
;;;
;;;  "Removes a CONFLICT, computed during lower bound propagation,
;;;   and its effects, and sets up lower bound propagation.
;;;   CONFLICT is a list of distance edges that form a negative 
;;;   cycle within STN."
;;;
;;;  (let ((events-reset '()))
;;;	;; Initialize every event supported by CONFLICT,
;;;	;; along with the successors of those events.
;;;	;; Then setup to relax all initialized events.
;;;	(loop for e in conflict
;;;		for x = (d-edge-start e)
;;;				;; lower bound value on x is now nonsense, initialize it.
;;;		do (event-lb-initialize! x)
;;;		   (push x events-reset)
;;;		   ;; All consequences of n are nonsense, reinitialize them
;;;		   (setq events-reset
;;;			 (reset-all-lb-consequences x STN events-reset)))
;;;
;;;	;; Setup queue to resupport all reset nodes.
;;;	(queue-successors-for-lb-propagation! events-reset STN)
;;;	t))

