;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 1999 - 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;; ***************************************************************************
;;;
;;;                          Checking Temporal Consistency
;;;
;;; ***************************************************************************

;;; ***************************************************************************
;;;
;;;                           External Interface
;;;
;;; ***************************************************************************


;;;   (CONSISTENT?
;;;    (STN STN) &OPTIONAL
;;;  	         (PERFORM-SEARCH? T)
;;;	             (RETURN-CONFLICT? T)
;;;	             (KEY #'STC-GUARD-TO-TRUTH-ASSIGNMENT))
  
;;;    Checks consistency of STN.
;;;    Returns (values CONSISTENT? CONFLICT).

;;;    CONSISTENT? is non-NIL iff there exists an
;;;    assignment to STN events that satisfies all
;;;    active, simple temporal constraints. A constraint
;;;    is active iff the discrete state variable assignments
;;;    satisfy the constraint's guard.

;;;    If stn is found inconsistent and RETURN-CONFLICT? is 
;;;    TRUE, then a CONFLICT is returned.

;;;    A CONFLICT denotes a partial state variable assignment
;;;    that is inconsistent.  It is generated by applying function 
;;;    KEY to each state variable assignment that 
;;;    supports an inconsistent constraint of STN, and by 
;;;    collecting the non-nil, irredundant values returned by KEY.

;;;    By default, KEY takes PROPOSITION as input and returns 
;;;    assignment (= PROPOSITION TRUTH-VALUE) where 
;;;    TRUTH-VALUE is *true* or *false*.
   
;;;    Side Effect:  Compute upper and lower bound
;;;    distance of all events wrt the ITC start.

;;;    (CONSISTENT-WITH?
;;;     (STN STN) &OPTIONAL &KEY 
;;; 		      (ADDED-CONSTRAINTS NIL)
;;; 		      (ADDED-ASSIGNMENTS NIL)
;;; 		      (SOURCE NIL)
;;; 		      (RETURN-CONFLICT? T))
;;; 
;;;    Tests the consistency of the current constraints and 
;;;    assignments in STN, conjoined with ADDED-CONSTRAINTS 
;;;    and ADDED-ASSIGNMENTS.
;;; 
;;;    Returns (values CONSISTENT? CONFLICT).  
;;;    CONSISTENT? is true unless STN is proven inconsistent.  
;;; 
;;;    Each constraint in ADDED-CONSTRAINTS is a constraint object.
;;; 
;;;    Each assignment in ADDED-ASSIGNMENTS is a list:
;;; 
;;;        (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
;;; 
;;;    and specifies that assignment proposition 
;;;    (= VARIABLE VALUE) has TRUTH-VALUE.
;;;    
;;;    If RETURN-CONFLICT? is TRUE and an inconsistency is 
;;;    found, then returns a subset of the current inserted 
;;;    assignments that lead to an inconsistency.

;;;   (INCONSISTENT-WITH? (STN STN) &OPTIONAL &KEY 
;;; 		      (ADDED-CONSTRAINTS NIL)
;;; 		      (ADDED-ASSIGNMENTS NIL)
;;; 		      (SOURCE NIL)
;;; 		      (RETURN-CONFLICT? T)
;;; 		      (BASE-STN-CONSISTENT-WITH-ALL-ASSIGNMENTS? T))
;;; 
;;;    Tests the inconsistency of the current constraints and 
;;;    assignments in STN, conjoined with ADDITIONAL-CONSTRAINTS.
;;; 
;;;    Returns (VALUES INCONSISTENT? CONFLICT).  INCONSISTENT? 
;;;    is false unless STN is proven inconsistent.  
;;; 
;;;    Each clause in ADDED-CONSTRAINTS is a clause object.
;;; 
;;;    Each assignment in ADDED-ASSIGNMENTS is a list:
;;; 
;;;        (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
;;; 
;;;    and specifies that proposition (= VARIABLE VALUE) has TRUTH-VALUE.
;;; 
;;;    If RETURN-CONFLICT? is TRUE and STN can be proven 
;;;    necessarily true, then returns a subset of the current 
;;;    assignments sufficient to guarantee that stn is satisfied.
;;;    Currently this simply returns the complete set of discrete
;;;    state variable assignments as the conflict.
;;; 
;;;    If BASE-STN-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, then 
;;;    assumes that the inserted assignments are consistent with the 
;;;    constraints in STN (minus added-constraints).
;;;    In this case a conflict is a subset of the assignments that, 
;;;    together with STN (minus assignments) entail ADDED-CONSTRAINTS.
;;; 
;;;    A conflict is created by collecting a subset of the 
;;;    current constraints that support the truth 
;;;    of all ADDED-CONSTRAINTS.

;;;    Next the function KEY is applied to each constraint in the set, 
;;;    collecting the non-NIL, non-duplicate returned 
;;;    values as the CONFLICT.

;;;(eval-when (compile load eval)
;;;  (export '(consistent?
;;;			consistent-with?
;;;			inconsistent-with?)
;;;		  :ITC))

;;; ***************************************************************************
;;;
;;;                  External Interface: 
;;;
;;; ***************************************************************************

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                  External Interface: Consistent?
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Check:
;;; o Revised to remove redundant elements in a conflict.  Confirm that
;;;   this is correct.

;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't nil denote the
;;;   conflict of empty assignments, in which case everything is
;;;   inconsistent?  Check and correct this.

;;; Optimization:
;;; o Use return-conflict?, to avoid generating an internal conflict 
;;; if not needed?

;;; To Do:
;;; o The default test should be specialized to an EQUAL-ASSIGNMENT Function.
;;; o Look into replacing KEY default STC-GUARD-TO-TRUTH-ASSIGNMENT
;;;   with STC-GUARD-TO-STATE-ASSIGNMENT-W-TRUTH-DESCRIPTION.
(defmethod consistent?
    ((stn stn) 
	 &key
	 (return-conflict? T)
	 (key #'stc-guard-to-truth-assignment)
	 (test #'equal)
	 (initialize? nil)
	 (return-negative-cycle? nil)
	 &allow-other-keys)

  "Checks consistency of STN.
    Returns (values CONSISTENT? CONFLICT).

    CONSISTENT? is non-NIL iff there exists an
    assignment to STN events that satisfies all
    active, simple temporal constraints. A constraint
    is active iff the discrete state variable assignments
    satisfy the constraint's guard.

    If STN is found inconsistent and RETURN-CONFLICT? is 
    TRUE, then a CONFLICT is returned.

    Internally a CONFLICT is a set of simple-temporal constraints 
    that is inconsistent.  The returned conflict is generated by
    applying KEY to each element of the internal conflict, 
    collecting those result that are not equal to *TRUE*,
    and removing duplicate elements using TEST.

    KEY defaults to STC-GUARD-TO-TRUTH-ASSIGNMENT, 
    Given a constraint of an internal conflict,
    where constraints has guard (= var value), returns 

      (= (= var value) truth-value)
 
    where truth-value is the current truth-value of assignment (= var value), 
    and is one of *true*,*false* or *none*.
   
    TEST takes as input pairs of externalized elements, and returns true
    iff the two elements are equal.  TEST defaults to EQUAL.

    If INITIALIZE? is True, then reset all distances before propagating;
    otherwise, update incrementally from previous values.

    Side Effect:  Computes the upper and lower bound
    distance of all events wrt the ITC start."

;;;  (format t "~2% *** This is my key = ~a" key)
  
  ;; Add by Peng Yu: Record backup of current edges.
;;  (d-edge-backup (STN-events stn))  
;;  (d-edge-print (STN-events stn))  

  ;; Added by Peng Yu: 
  ;;   Relax edges by changing ub & lb to 1,000,000.
  ;;   Remove Removed? mark.
;;;  (d-edge-relaxation (STN-events stn))

  (multiple-value-bind (consistent? internal-conflict)
	  (consistent-stn? STN :initialize? initialize?
					   :return-negative-cycle? return-negative-cycle?)

	(let ((external-conflict
		   (when (and (not consistent?) 
					  return-conflict?) 
			 
			 ;; Provide the conflict in external form.
			 (if (null return-negative-cycle?)
				 (externalize-conflict stn internal-conflict
									   :key key :test test :key-returns-conjunction? t)
				 (externalize-conflict stn (pop internal-conflict)
									   :key key :test test :key-returns-conjunction? t)))))

	  ;; Return both internal and external conflicts if looking for negative cycles

	 (if return-negative-cycle?
	  (values consistent? (list external-conflict (pop internal-conflict)))   
	  (values consistent? external-conflict)))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                  External Interface:  Consistent-With?
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Check:
;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't nil denote the
;;;   conflict of empty assignments, in which case everything is
;;;   inconsistent?  Check and correct this.
(defmethod consistent-with?
    ((stn stn) &optional &key 
	 (added-constraints nil)
	 (added-assignments nil)
	 (source nil)
	 (return-conflict? t)
	 (key #'stc-guard-to-truth-assignment)
	 (initialize? nil)
	 (return-negative-cycle? nil)
	 &allow-other-keys)

  " Tests the consistency of the current constraints and 
    assignments in STN, conjoined with ADDED-CONSTRAINTS 
    and ADDED-ASSIGNMENTS.
 
    Returns (values CONSISTENT? CONFLICT).  
    CONSISTENT? is true unless STN is proven inconsistent.  
 
    Each constraint in ADDED-CONSTRAINTS is a constraint object.
 
    Each assignment in ADDED-ASSIGNMENTS is a list:
 
        (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
 
    and specifies that assignment proposition 
    (= VARIABLE VALUE) has TRUTH-VALUE.
    
    If RETURN-CONFLICT? is TRUE and an inconsistency is 
    found, then returns a subset of the current inserted 
    assignments that lead to an inconsistency.

     If INITIALIZE? is True, then reset all distances before propagating;
     otherwise, update incrementally from previous values.
    "

  (debug-print-progn-for 
   :itc
   (unless (and (null added-assignments)
				(null added-constraints))
	 (debug-print-for :itc t "Testing consistency of STN ~A"
					  stn)
	 (unless (null added-assignments)
	   (debug-print-for :itc t "  with added assignments ~A"
						added-assignments))
	 (unless (null added-constraints)
	   (debug-print-for :itc t "  and added constraints ~A."
						added-constraints))))

  ;; Insert additional constraints and assignments
  (unless (null added-constraints)

	(debug-print-for :itc t "Inserting added constraints ~A."
					 added-constraints)
	(loop for constraint in added-constraints
		do (insert-constraint stn constraint)))

  (unless (null added-assignments)

	(debug-print-for :itc t "Inserting added assignments ~A."
					 added-assignments)
	(assert-assignments! stn source added-assignments))
  
  ;; Test consistency.

  ;; Conflict is a list of truth value assignments
  ;; (= PROPOSITION TRUTH-VALUE), where truth-value is
  ;; *true* or *false*, and PROPOSITION is a proposition
  ;; object whose proposition description is an expression
  ;; (= STATE-VARIABLE VALUE).
  (stn-initialize! stn)
  (multiple-value-bind (consistent? conflict)
      (consistent? stn
				   :return-conflict? return-conflict?
				   :key key
				   :initialize? initialize?
				   :return-negative-cycle? return-negative-cycle?)

	;; Remove the inserted constraints and assignments
	(unless (null added-constraints)

	  (debug-print-for :itc t "Removing added constraints ~A."
					   added-constraints)
	  (loop for constraint in added-constraints
		  do (remove-constraint stn constraint)))

	(unless (null added-assignments)

	  (debug-print-for :itc t "Removing added assignments ~A."
					   added-assignments)
	  (unassert-assignments! stn source added-assignments))
    
	;; Return the results

    (values consistent? conflict)))

;;; ***************************************************************************
;;;
;;;                  External Interface: Inconsistent-With?
;;;
;;; ***************************************************************************

;;; To Do: 

;;;   o Reconsider the format of added-assignments
;;;     to be assignment object/truth-value pairs.

;;;   o OpSat should only use added-assignments
;;;     to make assignments, not added-constraints.

;;;  o Should constraint insertion and removal only be 
;;;    triggered by a variable assignment?

;;; Check:
;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't nil denote the
;;;   conflict of empty assignments, in which case everything is
;;;   inconsistent?  Check and correct this.
(defmethod inconsistent-with? 
    ((stn stn) &optional &key 
						 (added-constraints nil)
						 (added-assignments nil)
						 (source nil)
						 (return-conflict? T)
						 (key #'stc-guard-to-truth-assignment)
						 (initialize? nil)
	 &allow-other-keys)

  "Tests the inconsistency of the current constraints and 
    assignments in STN, conjoined with ADDITIONAL-CONSTRAINTS.
 
    Returns (VALUES INCONSISTENT? CONFLICT).  INCONSISTENT? 
    is false unless STN is proven inconsistent.  

    Each clause in ADDED-CONSTRAINTS is a clause object.
 
    Each assignment in ADDED-ASSIGNMENTS is a list:
 
        (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
 
    and specifies that proposition (= VARIABLE VALUE) has TRUTH-VALUE.
 
    If RETURN-CONFLICT? is TRUE and STN can be proven 
    necessarily true, then returns a subset of the current 
    assignments sufficient to guarantee that stn is satisfied.
    Currently this simply returns the complete set of discrete
    state variable assignments as the conflict.
 
    If BASE-STN-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, then 
    assumes that the inserted assignments are consistent with the 
    constraints in STN (minus added-constraints).
    In this case a conflict is a subset of the assignments that, 
    together with STN (minus assignments) entail ADDED-CONSTRAINTS.
 
    A conflict is created by collecting a subset of the 
    current constraints that support the truth 
    of all ADDED-CONSTRAINTS.

    Next the function KEY is applied to each constraint in the set, 
    collecting the non-NIL, non-duplicate returned 
    values as the CONFLICT.
 
     If INITIALIZE? is True, then reset all distances before propagating;
     otherwise, updates incrementally from previous values. "

  (declare (ignore return-conflict? key))

  (debug-print-progn-for 
   :itc
   (unless (and (null added-assignments)
				(null added-constraints))
	 (debug-print-for :itc t "Testing inconsistency of STN ~A"
					  stn)
	 (unless (null added-assignments)
	   (debug-print-for :itc t "  with added assignments ~A"
						added-assignments))
	 (unless (null added-constraints)
	   (debug-print-for :itc t "  and added constraints ~A."
						added-constraints))))

  ;; Add in the assignments
  (unless (null added-assignments)

	(debug-print-for :itc t "Inserting added assignments ~A."
					 added-assignments)
	(assert-assignments! stn source added-assignments))

  ;; Insert additional constraints
  (cond ((null added-constraints))

		(:otherwise
		 (debug-print-for :itc t "Inserting added constraints ~A."
						  added-constraints)
		 (loop for constraint in added-constraints
			 do (insert-constraint stn constraint))))

  (debug-print-for :itc t "Test inconsistency." )

  ;; Check consistency.
  (multiple-value-bind (consistent? conflict)
	  (consistent-stn? STN :initialize? initialize?)
	(declare (ignore conflict))

	;; Remove the inserted constraints and assignments
	(unless (null added-constraints)

	  (debug-print-for :itc t "Removing added constraints ~A."
					   added-constraints)
	  (loop for constraint in added-constraints
		  do (remove-constraint stn constraint)))

	(unless (null added-assignments)

	  (debug-print-for :itc t "Removing added assignments ~A."
					   added-assignments)
	  (unassert-assignments! stn source added-assignments))

    (if (not consistent?)
		;; Inconsistent
		(debug-print-for :itc t "~A is inconsistent." stn)

	  ;; Consistent
	  (debug-print-for :itc t "~A is consistent." stn))

	;; Conflict of nil means that the current
	;; state is the conflict.
	(values (not consistent?) nil)))
