(in-package :ITC)

;; This file contains functions to dispatch a standard STN without 
;; using incremental techniques.

;;; External Interfaces:
;;; OFFLINE-DISPATCH-STN: Takes an STN object and dispatches it using 
;;; all lower bounds using the dynamic execution algorithm - implemented
;;; for testing purposes.

(defmethod offline-dispatch-stn ((stn stn) &key &allow-other-keys)

  "Dispatch an stn offline by simply taking the lower bound as the next time
   at each time step. This does not actually wait for time to elapse. Returns
   a list of dispach instructions where each element of the returned list is
   (EVENT (LB UB) TIME) where EVENT is the event object, (LB UB) is the range of
   acceptable durations until the next time point must occur, and TIME is the actual
   execution time of EVENT."

  ;; Create a stn-dispatch object and init temp objects
  (let ((dispatch (make-instance 'stn-dispatch :stn stn))
		(curr-event nil) (curr-bounds (list 0 0)) (curr-time 0))

	;;Loop until no more events are scheduled
	(loop 
		do (multiple-value-bind (event-temp bounds-temp time-temp)
			   ;; Call the dispatcher and grab the results.
			   (next-stn-time-point dispatch
									(first curr-bounds))

			 (setf curr-event event-temp) 
			 ;; Store these into vars good outside the value-bind.
			 (setf curr-bounds bounds-temp)
			 (setf curr-time time-temp))
		until (null curr-event)  
			  ;;collect the dispatching instructions.
		collect (list curr-event curr-bounds curr-time)))) 


;;; ***********************************************************************
;;;
;;;                STN-DISPATCH Objects
;;;
;;; ***********************************************************************

;;; Defines an object to hold the book-keeping for dynamic execution of an STN,
;;; which allows the user to ask for the next decision as necessary without 
;;; specifically tracking the STN information. These algorithms leverage the
;;; reasoning algorithms in stn-reasoning.lisp Can also handle episodes, as from a TPN,
;;; in addition to simple constraints. The episode dispatching functions can be found
;;; in tpn/extract-stn.lisp

(defclass stn-dispatch ()
  ((stn 
	:initarg :stn
	:accessor stn
	:initform nil
	:documentation "The STN to dispatch.")

   (current-time
	:initform 0
	:accessor dispatch-time
	:documentation "The current time of the dispatcher.")

   (A
	:initform nil
	:accessor dispatch-a
	:documentation "The set of active stn events.")

   (S
	:initform nil
	:accessor dispatch-s
	:documentation "The set of executed stn events.")

   (incomplete-episodes
	:initform nil
	:initarg :incomplete-episodes
	:accessor dispatch-uncompleted-episodes
	:documentation "A list with the episodes of the STN to be scheduled.")
  
   (waiting-for-episode
	:initform nil
	:accessor dispatch-waiting-for-episode
	:documentation "A list of events that are active except that they are waiting
for an episode to complete."))

  (:documentation "The book-keeping required to do standard dynamic
                   execution of an STN.")
  )

(defmethod initialize-instance :after
		   ((stn-dispatch stn-dispatch) &key &allow-other-keys)

  "Initialize the dispatching object by putting the start of the STN
   in the active event list. Prepares the STN for dispatching by running
   all-pairs shortest path and then filtering the STN of redundant edges."

  (when (string-equal "stn-dispatch" (class-name (class-of stn-dispatch)))
	(with-slots (A stn) stn-dispatch
	  (push (stn-start stn) A)
	  )))

(defmethod next-stn-time-point ((stn-dispatch stn-dispatch) &optional new-time)

  "Implements the core algorithm (without the looping) to determine the next event
   to execute and the time of execution for the stn in the dispatch object, as 
   according to John Stedl's thesis. If the time has changed since the last call, 
   which is zero for the first call, the NEW-TIME should be supplied as the 
   current time."

  (with-slots (stn current-time A S) stn-dispatch

	;;If a new time is provided, update the time 
	(when (not (null new-time))
	  
	  (setf current-time new-time))

	;;Stop when there are no more events to execute.
	(when (eql (length S) (length (stn-events stn)))
	  
	  (return-from  next-stn-time-point (values nil (list 0 0))))

	;; 2. Arbitrarily pick a time point TP in A such that current_time
	;;    belongs to TP's time bound;

	;;    Find the event and hold onto it.
	(let ((exec-event (find-next-exec-event stn-dispatch)))
	  (when (null exec-event)			;if there wasn't an event found, stop.

		(return-from next-stn-time-point
		  (values nil (calc-next-times stn-dispatch) current-time)
		  ))
	  ;; 3a. Set the exec time to current time,
;;;   (format t "exec:~a ~a~%" exec-event current-time)	  
	  (setf (event-lb exec-event) current-time)
	  (setf (event-ub exec-event) current-time)

	  ;; 3b. and add tp to S.
	  (push exec-event S)

	  ;; 4. Propagate time of exec to the immediate neighbors.
	  (propagate-exec-time stn-dispatch exec-event)

	  ;; 5. Put in all time points TPx such that all negative edges starting from
	  ;;   TPx have a destination that is already in S.
	  (loop for i in (stn-events stn)	; For each event,
		  do (when (event-active? stn-dispatch i) ; If it still hasn't happened yet,
			   (push i A)))				; Mark it as active.

	  (setf A (remove-duplicates A))	; Remove duplicates.

	  ;; 6. Determine what the new times may be.
	  ;;    First, make the new event not part of A anymore.

	  (setf A (delete exec-event A))

	  ;;    Calculate the next set of times.
	  (let ((next-times (calc-next-times stn-dispatch)))

		;;  Return the event and the range of possibilities for the next times.
		(values exec-event next-times current-time)))))

(defmethod calc-next-times ((stn-dispatch stn-dispatch))
  (with-slots (A) stn-dispatch
	(let*  ((lb-list (loop for i in A
						 collect (event-lb i)))
			(ub-list (loop for i in A
						 collect (event-ub i))))
	  (list (reduce #'bmin-reduce lb-list) (reduce #'bmin-reduce ub-list)))))

(defun bmin-reduce (&optional (x *infinity*) (y *infinity*))
  (bmin x y))

;;; ***********************************************************************
;;;
;;;                STN Dispatching Utilities
;;;
;;; ***********************************************************************

(defmethod find-next-exec-event ((stn-dispatch stn-dispatch))
  (with-slots (A) stn-dispatch
	(loop for i in A
		thereis (when (current-time-in-event-bounds? stn-dispatch i) i))))

(defmethod event-active? ((stn-dispatch stn-dispatch) (event event))

  "Determines whether the event can be added to the active set, i.e. it
  is not a member of the completed set S and all negative edges beginning 
  from EVENT have destinations in S."

  (with-slots (S) stn-dispatch
	(and (null (member event S)) 
		 ;; It must not already be active
		 ;;and check for prereqs.
		 (not 

		  ;; The next loop tries to find a counter case, so invert the result.
		
		  (loop for i in (event-successors event) 
						 ;; For each outgoing edge,
			  thereis (if (and  (d-edge-lower-bound? i) (b>= 0 (d-edge-ub i))) 
						  ;;if a non-positive weight lower bound,
						  (null (member (d-edge-end i) S)) ;;true if end is not in S
						nil))))) ;;never true if not negative.
  )

(defmethod propagate-exec-time ((stn-dispatch stn-dispatch) (event event))

  "Propagates a time of execution (the current-time of the dispatch) to 
   direct neighbors."

  (with-slots (current-time) stn-dispatch
	
	(with-slots (predecessors successors) event
	  
	  (loop for i in predecessors
		  do (when (b>= 0 (d-edge-ub i))
			   ;; Only propagate neg precceeding edges.
			   ;; Update the lb of the start with this if it is tighter.
			   (setf (event-lb (d-edge-start i))
				 (bmax  (b+ current-time (b- (d-edge-ub i)))
						(event-lb (d-edge-start i))))))
	  
	  (loop for i in successors
		  do (when (b>= (d-edge-ub i) 0) 

			   ;; Only propagate positive succeeding edges.
			   ;; Update the ub of the end with this if it is tigher than 
			   ;; the existing one.
			   (setf (event-ub (d-edge-end i))
				 (bmin (b+ current-time (d-edge-ub i))
					   (event-ub (d-edge-end i)))))) 
	  )))

(defmethod current-time-in-event-bounds? ((stn-dispatch stn-dispatch) (event event))

  "Return whether the input time is within [lb ub] for this event."

  (with-slots (current-time) stn-dispatch
	(and (b>= current-time (event-lb event))
		 (b>= (event-ub event) current-time))))

(defun eq-dispatch? (x y)

  "Checks whether dispatch instructions are the same. The first comes from the 
   dispatch and the second is reference list with symbols instead of real events. 
   Used for testing OFFLINE-DISPATCH-STN."

  (loop for i in x
	  for j in y
	  always (and (equal (event-name (first j)) (first i))
				  (equal (rest i) (rest j)))))
