;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2008-2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;; ***********************************************************************
;;;
;;;        Simple Temporal Network and Related Objects and Accessors
;;;              in Support of Incremental Temporal Inference
;;; 
;;; ***********************************************************************

;;; Last update: September 1st 2010.

;;; This file contains objects, macros and functions for performing
;;; incremental temporal inference on simple temporal networks (STN),
;;; and their corresponding distance graphs (D-Graph or DG).

;;; Objects:

;;;   This file defines specializations of objects for STN, EVENT, 
;;;   and STC, and associated slot accessors.  These specializations
;;;   are preceded by "ITC-."  Specializations include the folloiwng:
;;;
;;;   ITC-STN object slots include:

;;;     STATE-VARIABLE-HT   : Holds the hash table that maps state-variable names to 
;;;                           state-variable objects.  This is used when new assignments 
;;;                           or constraints are created, to map quickly from state-variable 
;;;                           names to their corresponding objects.
;;;  
;;;     STATE-ASSIGNMENT-HT : Holds the hash table that maps state-variable name / value 
;;;                           pairs to state-assignment objects.  This is used when new 
;;;                           assignments or constraints are created, to map quickly from 
;;;                           assignment pairs to their corresponding objects.
;;;
;;;     EVENT-PAIR-QUEUE    : Queue of event pairs, such that at least one edge in each 
;;;                           event pair has changed.
;;;  
;;;     LB-QUEUE            : Queue of events whose lowerbound values are to be 
;;;                           propagated to predecessors.
;;;
;;;     LB-PROPAGATED?      : True if lowerbound consistency has been checked for 
;;;                           STN at least once, since network was last initialized.
;;;
;;;     UB-QUEUE            : Queue of events whose upperbound values are to be 
;;;                           propagated to successors.
;;;
;;;     UB-CONSISTENCY-CHECKED? : True if upperbound consistency has been 
;;;                           checked for STN at least once, since network 
;;;                           was last initialized.
;;;   
;;;     UB-INCONSISTENT?    : True if ub-consistency has been checked, 
;;;                           and proven inconsistent.
;;;
;;;     UB-CONFLICT         : If ub-inconsistent? is True then conflict is a list
;;;                           of d-edges that are inconsistent and nil, otherwise.

;;;   ITC-EVENT object slots include:
;;;     QUEUED?             : True iff one or more d-edges has been inserted/removed
;;;                           and the change needs to be propagated.

;;;   ITC-EVENT object slots include:
;;;
;;;     LB-SUPPORT
;;;     UB-SUPPORT          : Distance edge used to derive the lower/upper 
;;;                           bound value of EVENT.
;;;     LB-CHILDREN
;;;     UB-CHILDREN         : Support for event's lower/upper bound. If bound is inferred,
;;;                           then children are some distance edges.
;;;     LB-VISITED?
;;;     UB-VISITED?         : True if event has been visited by lower-bound propagation.

;;;   ITC SIMPLE TEMPORAL CONSTRAINT (ITC-STC) object slots include:
;;;
;;;     GUARD               : Null or a state-assignment object.  If constraint is inserted, 
;;;                           then to hold, guard must be null or an inserted assignment.  
;;;                           An initarg, guard is provided as an assignment expression or
;;;                           an assignment object.
;;;
;;;     LB-CONSEQUENT
;;;     UB-CONSEQUENT       : An event whose lower/upper bound value was deduced
;;;                           from STC by relaxation.  It is
;;;                           NIL if STC hasn't tightened an event's bound by 
;;;                           relaxation.
;;;
;;;     REMOVED?            : Says whether the constraint is currently removed
;;;                           from STN or not.  Can be T, meaning
;;;                           removed, or NIL, meaning inserted.

;;;   ITC STATE VARIABLE object slots include:

;;;     NAME                : NAME is any lisp object or expression 
;;;                           that names variable.  A STN may have 
;;;                           at most one variable with a given NAME.

;;;     DOMAIN              : A list of possible values that may be 
;;;                           assigned to STATE-VARIABLE.
   
;;;     ASSIGNMENTS         : A list of assignment objects, one for
;;;                           each element of variable's domain.

;;;     SOURCE              : The object that created STATE-VARIABLE.
   
;;;     ORDER               : A number used for sorting the variables in a consistent order.

;;;     OBSERVED?           : True if this variable is observed in a cdtn, false if it 
;;;                           is controlled.

;;;     ASSIGNED-BY-EVENTS  : A list of events for a cdtn when this variable must be assigned.

;;;   ITC STATE ASSIGNMENT object slots include:

;;;     VARIABLE            : VARIABLE is a variable object denoting the 
;;;                           variable of that is being assigned to.

;;;     VALUE               : A lisp object, taken for VARIABLE's domain, 
;;;                           denoting the value that variable is assigned.
   
;;;     CONSTRAINTS         : Constraints that become active when assignment
;;;                           is assigned *true*; that is, assignment is the
;;;                           guard of constraint.
   
;;;     TRUTH-VALUE         : TRUTH-VALUE stores the current truth of STATE-
;;;                           ASSIGNMENT.  Its possible values are *TRUE*, 
;;;                           *FALSE*, or *NONE*  *NONE* corresponds to the 
;;;                           truth value being unassigned.

;;; Additional Constraint Operations:
;;;
;;;   DO-INSERTED-D-EDGES ((EDGE D-EDGES) &BODY BODY)
;;;     A macro that executes BODY for each inserted edge in D-edges.
;;;     Behaves exactly like dolist, and permits early return.
;;;     An edge is inserted if its constraint is inserted and 
;;;     constraint's guard is satisfied.

;;; FIX:
;;; o mtu::declare-debug-print-system!
;;;   has been exported from MTU in its package definition,
;;;   and included in :BFIG.  Why isn't the function external (: vs ::)
;;;   and why isn't it part of :ITC?

(mtu::declare-debug-print-system! :ITC)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			;; ITC EVENT Object
;;;			event-lb-support
;;;			event-ub-support
;;;
;;;			;; ITC Simple Temporal Constraint Object
;;;			stc-lb-consequent
;;;			stc-ub-consequent
;;;			stc-removed?
;;;
;;;			;; STC related functions
;;;			set-lb-support-relation!
;;;			set-ub-support-relation!
;;;			clear-lb-support-relation!
;;;			clear-ub-support-relation!
;;;
;;;			do-inserted-d-edges ;; Macro
;;;			)
;;;		  :ITC)
;;;  )


;;; ***********************************************************************
;;;
;;;                Constants
;;;
;;; ***********************************************************************

;;; FIX:
;;; o Moved here from stn-defs.lisp.  Delete this comment as soon as this
;;;   works.
(eval-when (compile load eval)

;;; Each node now has a status of {unreached, labeled, scanned}.
;;; It is a variation of the label correcting operations.
;;; - Added by Andrew Yi Huang.
  
  (defconstant *unreached* 'unreached
	"The bound of event has not yet been tightened through propagation.")

  (defconstant *labeled* 'labeled
	"A bound for event has been computed through edge relaxation, and the
     edge used has been recorded as support for that event bound.
     It has not yet been checked for a negative loop or propagated.")

  (defconstant *scanned* 'scanned
	"A bound has been computed for event, propagated and checked for
     consistency."))

;;; ***********************************************************************
;;;
;;;                Variables
;;;
;;; ***********************************************************************

;;; Variables To Track Performance.

(defvar *ITC-statistics-p* nil
  "Controls whether or not ITC collects performance
   statistics.  Statistics are only collected if the 
   variable is non-NIL."
  )

(defvar *propagate-counter* 0

  "Number of times PROPAGATE is called.")

(defvar *remove-constraint-counter* 0
  "Number of times REMOVE-CONSTRAINT is called.")

;;; To Do:
;;; o Determine which counters are required for maintaining statistics,
;;;   and writing the statistics functions for the test suite.

;;; (defvar *tighten-bound-counter* 0
;;;  "Number of times ASSIGN-TRUTH-VALUE is called.")

;;; (defvar *unassign-truth-value-counter* 0
;;;  "Number of times UNASSIGN-TRUTH-VALUE is called.")

;;; (defvar *update-proposition-clause-counter* 0
;;;  "Number of times UPDATE-PROPAGATION-CLAUSE is called.")

;;; (defvar *update-clause-for-unassigned-proposition-counter* 0
;;;  "Number of times UPDATE-CLAUSE-FOR-UNASSIGNED-PROPOSITION is called.")



;;; ***********************************************************************
;;;
;;;                FIFO Queue
;;;
;;; ***********************************************************************

;;; Implements a First In First Out Queue object.

(defclass queue ()

  ((first 
	:accessor queue-first
	:initform nil
	:documentation "Points to the list of queue entries, in order to be ~
                    dequeued.")
   (last
	:accessor queue-last
	:initform nil
	:documentation "Points to the last entry on the queue"))
  )

(defmethod clear-queue! ((Q queue))

  "Removes all entries from fifo queue Q."
  
  (setf (queue-first Q) nil)
  (setf (queue-last Q) nil))

(defmethod queue-empty? ((Q queue))
  
  "Returns True iff Q contains no entries."
  
  (null (queue-first Q)))

(defmethod queue-elements ((Q queue))
  
  "Returns a list of the elements on the queue.
   List MUST NOT be destructively modified."
  
  (queue-first Q))

(defmethod queue-insert! (item (Q queue))
  
  "Inserts ITEM onto first-in first-out queue Q."
  
  (let ((item-list (list item)))

	(cond ((null (queue-first Q))
		   
		   (setf (queue-first Q) item-list)
		   (setf (queue-last Q) item-list))
		  
		  (:otherwise
		   
		   (setf (rest (queue-last Q)) item-list)
		   (setf (queue-last Q) item-list)))))

(defmethod queue-pop ((Q queue))
  
  "Removes and returns the next item on first-in first-out queue Q.
   Entries are returned in the order added."
  
  (when (eql (queue-first Q)
			 (queue-last Q))
	
	(setf (queue-last Q) nil))
  
  (pop (queue-first Q)))

;;; check whether a queue contains an item
(defmethod queue-contains (item (Q queue))
  
  "Returns T if the queue contains the item, otherwise NIL."

  (dolist (i (queue-first Q))

	(when (eql item i) 
	  
	  (return T))))

;;; ***********************************************************************
;;;
;;;                ITC Simple Temporal Network (STN) Objects
;;;
;;; ***********************************************************************

;;;   Object holds all information about a simple temporal network STN,
;;;   AND for performing incremental inference on STNS.
;;;   An STN is a set of EVENTS and a set of simple temporal constraints 
;;;   STCs on events.  Each event is a real valued variable.  A simple 
;;;   temporal constraint on events A, B is specified by interval [lb,ub], 
;;;   and is satisfied iff (A - B) in [lb,ub].
;;;   The ITC-STN is an extension of STN to include slots for ITC reasoning.
(defclass ITC-STN (STN)

  ((state-variable-ht :initform (make-hash-table :test #'equal)
					  :accessor STN-state-variable-ht
					  :documentation
					  "STATE-VARIABLE-HT holds the hash table that maps 
                       state-variable names to state-variable objects.  This is
                       used when new assignments or constraints are created, to 
                       map quickly from state-variable names to their 
                       corresponding objects."
					  )
  
   (state-assignment-ht :initform (make-hash-table :test #'equal)
						:accessor STN-state-assignment-ht
						:documentation 
						"STATE-ASSIGNMENT-HT holds the hash table that
                         maps state-variable name/ value pairs to 
                         state-assignment objects.  This is used when 
                         new assignments or constraints are created, to
                         map quickly from assignment pairs to their 
                         corresponding objects."
						)

   (event-pair-queue 
	:accessor STN-event-pair-queue
	:initform (make-instance 'queue)
	:documentation "Queue of event pairs, such that at least one edge in each 
                    event pair has changed.")
  
   (lb-queue 
	:accessor STN-lb-queue
	:initform (make-instance 'queue)
	:documentation
	"Queue of events whose lowerbound values are to be propagated to 
     predecessors.")

   (lb-propagated?
	:accessor STN-lb-propagated?
	:initform nil
	:documentation 
	"True if lowerbound consistency has been checked for STN at least once, 
     since network was last initialized.")

   (ub-queue 
	:accessor STN-ub-queue
	:initform (make-instance 'queue)
	:documentation "Queue of events whose upperbound values are to be 
                    propagated to successors.")

   (ub-consistency-checked? 
	:accessor STN-ub-consistency-checked? 
	:initform nil
	:documentation 
	"True if upperbound consistency has been checked for STN at least once, 
     since network was last initialized.")
  
   (ub-inconsistent? 
	:accessor STN-ub-inconsistent?
	:initform nil
	:documentation
	"True if ub-consistency has been checked, and proven inconsistent.")

   (ub-conflict
	:accessor STN-ub-conflict
	:initform '()
	:documentation "If ub-inconsistent? is True then conflict is a list
                    of d-edges that are inconsistent and nil, otherwise."))
  
  (:documentation "An STN object for which incremental inference can be performed.
                   NAME is a lisp object denoting the STN's name."))

(defmethod initialize-instance :after 
		   ((STN ITC-STN) 
			&key (start-name 'start) &allow-other-keys)

  "Creates a new simple temporal network named NAME.  
   The network is empty, that is, it has no constraints or events
   associated with it, other than the START event.  

   The function has the side effect of setting 
   *DEFAULT-STN* to be the newly created network."
  ;;;Some modufucation has made by Peng to handle different start name in tpn
  ;;;Works in the same principle as modification in stn-defs.lisp. It provides two conditional processes for ITC-STN with and without specified start-event-name slot. Previous tests without start-event-name slot value will use start-name directly.

  (when (not (eql (slot-value STN 'start-event-name) nil));;;Added by Peng;
	(with-slots (name) stn
	  (let ((start
			 (make-instance 'itc-event :name (slot-value STN 'start-event-name)
							:network stn
							:start? T)
			  ))

;;; Associates a start event with STN
		(setf (STN-start STN) start)

;;; Initializes the STN consistency algorithm.
		(STN-wide-initialize! stn)
		
		(queue-insert! start (stn-ub-queue stn))
		(queue-insert! start (stn-lb-queue stn))

		(setq *default-stn* stn))))

  (when (eql (slot-value STN 'start-event-name) nil);;;Added by Peng; End
	(with-slots (name) stn
	  (let ((start 
			 (make-instance 'itc-event :name start-name
							:network stn
							:start? T)
			  ))

	;;; Associates a start event with STN
		(setf (STN-start STN) start)

	  ;; Initializes the STN consistency algorithm.
		(STN-wide-initialize! stn)

		(queue-insert! start (stn-ub-queue stn))
		(queue-insert! start (stn-lb-queue stn))

		(setq *default-stn* stn))))
	'T)

(defmacro find-state-variable 
    (STN state-variable-name)
  
  "Returns object representing STATE-VARIABLE-NAME for STN,
   where state variables are conditional variables, used to
   activate and deactivate simple temporal constraints.
   If STATE-VARIABLE-NAME is not defined in STN, then returns
   NIL"
  `(gethash ,state-variable-name
			(STN-state-variable-ht ,STN)))

(defmacro find-state-assignment
    (STN variable-name value)
  
  "Returns object representing VARIABLE-NAME/VALUE assignment
   in STN.  If assignment is not defined in STN, then returns
   NIL"
  `(gethash (list ,variable-name ,value)
			(STN-state-assignment-ht ,STN)))

;;;   *** STN Record of Previous Propagation ***

(defmacro ITC-STN-lb-propagated! (ITC-STN)

  "Assert that lowerbound propagation has been preformed 
   for simple temporal network ITC-STN at least once since 
   network was last initialized."

  `(setf (STN-lb-propagated? ,ITC-STN) 't))

(defun ITC-STN-lb-not-propagated! (ITC-STN)

  "Asserts that lowerbound consistency has not been checked 
   for simple temporal network ITC-STN since network was last initialized."
  
  (setf (STN-lb-propagated? ITC-STN) nil))

(defmacro STN-ub-consistency-checked! (ITC-STN)

  "Asserts that upperbound consistency has been checked 
   for simple temporal network ITC-STN at least once since 
   network was last initialized."

  `(setf (STN-ub-consistency-checked? ,ITC-STN) 't))

(defmethod ITC-STN-ub-consistency-not-checked! ((ITC-STN ITC-STN))

  "Asserts that upperbound consistency has not been checked 
   for simple temporal network ITC-STN since network was last initialized."
  
  (setf (STN-ub-consistency-checked? ITC-STN) nil)
  (setf (STN-ub-inconsistent? ITC-STN) nil)
  (setf (STN-ub-conflict ITC-STN) nil)
  T)

(defmacro ITC-STN-ub-inconsistent! (ITC-STN)

  "Asserts that upperbound consistency has been checked 
   and STN is inconsistent."

  `(setf (STN-ub-inconsistent? ,ITC-STN) 't))

(defun itc-conflict-constraints (conflict-d-edges)
  
  "Given a list of distance edges that form a conflict,
   that is, the edges form a loop with a negative total edge weight.
   return a list of constraints that the d-edges are part of."
  
  (mapcar #'d-edge-constraint conflict-d-edges))

;;; ***********************************************************************
;;;
;;;                ITC EVENT PAIR Objects
;;;
;;; ***********************************************************************

(defclass itc-event-pair (event-pair)
  
  ;; List of all distance edges from start to end of the event pair.

  ((queued? 
	:initform nil
	:accessor event-pair-queued?
	:documentation "True iff one or more d-edges has been inserted/removed
                    and the change needs to be propagated."))

  (:documentation "Record of all edges from START to END of a network.
                   Used to propagate effects when one or more edges between
                   START and END are inserted and/or removed."))

;;; *** Queue management for changed event pairs.

;;; An event pair (X, Y) is changed if an edge from
;;; X to Y has been inserted or removed.

(defmacro event-pair-queued! (event-pair)

  "Returns true IFF event-pair is queued for
   incremental update.  An event-pair is queued
   iff one or more of its d-edges has been
   inserted and/or removed."
  
  `(setf (event-pair-queued? ,event-pair) T))

(defmacro event-pair-dequeued! (event-pair)

  "Returns true IFF event-pair is NOT queued for
   incremental update."
  
  `(setf (event-pair-queued? ,event-pair) nil))

(defmacro has-event-pairs-queued? (STN)

  "Returns true IFF simple temporal network STN
   has one or more event-pairs queued for
   incremental update."
  
  `(not (queue-empty? 
		 (stn-event-pair-queue ,STN))))

(defmethod has-queued-constraints? ((network ITC-STN))

  "Returns True iff network has one or more constraints 
   queued for propagation."

  (has-event-pairs-queued? network))

(defmethod queue-changed-event-pair! 
	((event-pair event-pair) (STN ITC-STN))

  "Queues EVENT-PAIR if it hasn't already been queued
   for incremental update of simple temporal network STN.   
   An event-pair is queued when one or more of its distance
   edges has been inserted and/or removed."
  
  (unless (event-pair-queued? event-pair)
	
	(event-pair-queued! event-pair)
	(queue-insert! event-pair (stn-event-pair-queue STN))))

(defmethod dequeue-changed-event-pair ((STN ITC-STN))
  
  "Returns an event-pair of simple temporal network STN
   such that one of more of the distance edges Of
   event pair has been inserted and/or removed."
  
  (dequeue-changed-event-pair-from-queue 
   (stn-event-pair-queue STN)))

(defun dequeue-changed-event-pair-from-queue 
	(stn-event-pair-queue)

  "Returns an event-pair of STN-EVENT-PAIR-QUEUE
   such that one of more of the distance edges of
   event pair has been inserted and/or removed."
  
  (let ((event-pair (queue-pop stn-event-pair-queue)))
	
	(cond ((null event-pair) 
		   
		   nil)

		  (:otherwise
		   
		   (event-pair-dequeued! event-pair)
		   event-pair))))

;;; *** Creation and retrieval of event pairs.

(defmethod get-event-pair (x y (STN ITC-STN))y

		   " Returns a unique event pair object for simple temporal network STN 
    for events X to Y.  If the object already exists, this function is 
    returned; otherwise,  a new object is created, associated with STN 
    and returned.  

    Note: This is the preferred way to create a new event-pair object,
    since it avoids duplication."

		   (or (find-event-pair x y STN)
			   
			   (make-instance 'itc-event-pair
				 :start x
				 :end y
				 :network STN)))

;;; ***********************************************************************
;;;
;;;                EVENT Objects
;;;
;;; ***********************************************************************

;;; To Do:
;;; o Document lb/ub-children.  It looks like they wee added by
;;;   Andrew Yi Huang.

;;; Extends events to include slots for itc reasoning.
(defclass itc-event (event)
  
  (
   (lb-support
	:accessor event-lb-support
	:initform *unknown*
	:documentation
	"Support for event's lower bound.  If the bound is infinite, then nil.
     If bound is inferred, then support is a distance edge.")
   
   (ub-support 
	:accessor event-ub-support
	:initform *unknown*
	:documentation
	"Support for event's upper bound.  If the bound is infinite, then nil.
     If bound is inferred, then support is a distance edge.")

   (lb-children
	:accessor event-lb-children
	:initform nil
	:documentation
	"Support for event's lower bound. 
     If bound is inferred, then children are some distance edges.")
   
   (ub-children 
	:accessor event-ub-children
	:initform nil
	:documentation
	"Support for event's upper bound.
     If bound is inferred, then children are some distance edges.")

   (lb-visited? 
	:accessor event-lb-visited?
	:initform nil
	:documentation 
	"True if event has been visited by lower-bound propagation.")
   
   (ub-visited?
	:accessor event-ub-visited?
	:initform nil
	:documentation 
	"True if event has been visited by upper-bound propagation.")
   
   (status
	:accessor event-status
	:initform *unreached*
	:documentation
	"The status of the event can be of 3 types: UNREACHED, LABELED and SCANNED.")
   )
  
  (:documentation
   "Event object holds information about an event variable of an ITC
   simple temporal network."))


;;;(defun print-event-name (event stream depth)
;;;  
;;;  " Print EVENT's name."
;;;  
;;;  (declare (ignore depth))
;;;  (format stream "~a" (ignore-errors (event-name event))))


;;;   *** Marking an Event as Visited During Propagation ***

(defmacro event-lb-visited! (event)
  
  "Records that EVENT has been visited by lower
   bound propagation."
  
  `(setf (event-lb-visited? ,event) 't))

(defmacro event-ub-visited! (event)

  "Records that EVENT has been visited by upper
   bound propagation."
  
  `(setf (event-ub-visited? ,event) 't))

(defmacro event-not-lb-visited! (event)

  "Records that EVENT has NOT been visited by lower
   bound propagation."
  
  `(setf (event-lb-visited? ,event) 'nil))

(defmacro event-not-ub-visited! (event)

  "Records that EVENT has NOT been visited by upper
   bound propagation."
  
  `(setf (event-ub-visited? ,event) 'nil))

;;; FIX:
;;; o Was set to *infinity*, but should have been *-infinity*.
;;;   Once we confirm that the code is correct, delete this comment.
(defmacro event-lb-useless? (s)
  
  "Returns true iff the lower bound on event S
   provides no useful constraint for further propagation."
  
  `(and (eql (event-lb ,s) 
			 ',*-infinity*)
		
		(eql (event-lb-support ,s) 
			 ',*unknown*)))

;;; Check:
;;; Why is it useful if its bound is *infinity* but it is supported.
;;; Can this case occur?  If so, why do we allow a supported bound
;;; to be recorded as infinity?
(defmacro event-ub-useless? (s)
  
  "Returns true iff the upper bound on event S
   provides no useful constraint for further propagation."
  
  `(and (eql (event-ub ,s) 
			 ',*infinity*)
		
		(eql (event-ub-support ,s) 
			 ',*unknown*)))


;;; ***********************************************************************
;;;
;;;                State Variable Objects
;;;
;;; ***********************************************************************

(defclass state-variable ()

  ((name
	:initarg :name
	:accessor state-variable-name

	:documentation "NAME is any lisp object or expression 
                    that names variable.  A STN may have 
                    at most one variable with a given NAME.")

   (domain 
	:initarg :domain
	:accessor state-variable-domain

	:documentation "A list of possible values that may be 
                    assigned to STATE-VARIABLE.")
   
   (assignments
	:initform '()
	:accessor state-variable-assignments
	:documentation "A list of assignment objects, one for
                    each element of variable's domain.")

   (source
	:initarg :source
	:accessor state-variable-source
	:documentation "The object that created STATE-VARIABLE.")
   
   (order
	:accessor state-variable-order
	:documentation
	"A number used for sorting the variables in a consistent order.")

   (observed?
	:initform nil
	:initarg :observed?
	:accessor state-variable-observed?
    :documentation
	"True if this variable is observed in a cdtn, false if it is controlled.")

   (assigned-by-events
	:initarg :assigned-by
	:initform nil
	:accessor state-variable-assigned-by-events
	:documentation
	"A list of events for a cdtn when this variable must be assigned."))
  
  (:documentation "State variable object holds information 
                   about a state variable."))

(defmethod initialize-instance :after 
		   ((state-variable state-variable) 
			&key STN &allow-other-keys)

  "Creates a new state variable object for STN, ranging over 
   DOMAIN.  State variables are conditional variables, used to
   activate and deactivate simple temporal constraints.
   NAME is any lisp object, denoting a variable, 
   and DOMAIN is a list of values that the variable may take on.
   SOURCE is the object that created variable, and STN is a
   simple temporal network that variable is part of.
   Creates a list of assignment objects for STN denoting the
   possible assignments of variable to an element of its domain."

  (with-slots (name domain source order assigned-by-events) state-variable

	;; Check for Old and Warn.
	(let ((old-state-variable 
		   (find-state-variable STN name)))

	  (unless (null old-state-variable)
		(warn "State variable ~A already defined for STN ~A, ~
               with domain ~A and source ~A. ~
             ~% Redefining to have domain ~A and source ~A." 
			  name STN
			  (state-variable-domain old-state-variable)
			  (state-variable-source old-state-variable)
			  domain source)))
  
	(setf order (hash-table-count (STN-state-variable-ht stn)))

	(setf assigned-by-events (mapcar #'(lambda (x) (get-event x stn)) assigned-by-events))

	;; Record new.
	(debug-print-for 
	 :ITC t "New State variable ~A with domain ~a" ;;for ~A, ~
;;;	   by source ~A." 
	 name domain 
;;;	 STN source
	 )

	(setf (find-state-variable STN name)
	  state-variable)

	(setf order (hash-table-count (stn-state-variable-ht stn)))

	(setf (state-variable-assignments state-variable)
	  (loop for value in domain
		  collecting (make-instance 'state-assignment
					   :variable state-variable
					   :value value
					   :STN STN))))
  'T)

(defmethod print-object ((obj state-variable) stream)
  
  " Print a state-variable using its name."
  
  (with-slots (name) obj 
    (format stream "~a" name)))

(defmethod value-in-domain? 
	((state-variable state-variable) value)

  "Is VALUE an element of the domain of STATE-VARIABLE?"

  (member value
		  (state-variable-domain state-variable)))

;;; ***********************************************************************
;;;
;;;                State Variable Assignment Objects
;;;
;;; ***********************************************************************

(defclass state-assignment ()

  ((variable
	:initarg :variable
	:accessor state-assignment-variable
	:documentation "VARIABLE is a variable object denoting the 
                    variable of that is being assigned to.")

   (value
	:initarg :value
	:accessor state-assignment-value
	:documentation "A lisp object, taken for VARIABLE's domain, 
                    denoting the value that variable is assigned.")
   
   (constraints
	:initform '()
	:accessor state-assignment-constraints

	:documentation "Constraints that become active when assignment
                    is assigned *true*; that is, assignment is the
                    guard of constraint.")
   
   (truth-value 
	:initform *none*
	:accessor state-assignment-truth-value
	:documentation "TRUTH-VALUE stores the current truth of STATE-
                   ASSIGNMENT.  Its possible values are *TRUE*, 
                   *FALSE*, or *NONE*  *NONE* corresponds to the 
                   truth value being unassigned."))
     
  (:documentation "SV assignment object holds information 
                   about a state variable assignment."))

(defmethod initialize-instance :after 
		   
		   ((state-assignment state-assignment) 
			&key STN &allow-other-keys)

  "Index state-assignment off of STN."

  (with-slots (variable value) state-assignment
	(let ((name (state-variable-name variable)))
	  (setf (find-state-assignment STN name value)
		state-assignment)))
  'T)

(defmethod print-object ((obj state-assignment) stream)
  
  " Print a state-variable-assignment."
  
  (with-slots (variable value) obj 
    (format stream "~a=~a" variable value)))

;;; Default method defined in mtk-utils/compare.lisp
(defmethod extract-proposition-description
	((A state-assignment))

  "Returns a description of state-assignment object A,
   of the form:
      (= <variable-name> <value>)."
  
  (with-slots (variable value) A
	(make-assignment :variable-name (state-variable-name variable)
					 :value value)))

(defmethod constraint-system-assignment-description
	((A state-assignment))
  
  "Given an assignment object A,
   Returns a description of assignment A of the form:

       (= <variable-name> <value>) "

  (extract-proposition-description A))


;;; To Do:
;;; o Evaluate whether we can replace
;;;     CONSTRAINT-SYSTEM-ASSIGNMENT-DESCRIPTION
;;;   with the following function.  Rethink the best name.

;;; o Move these functions to :CS when state variables and
;;;   state assignments are moved there.
(defmethod constraint-system-assignment-description-w-truth
	((state-assignment state-assignment))

  "Given STATE-ASSIGNMENT object with associated truth value, 
   returns a symbolic expression:

     <assignment-expression>           if truth-value = true

     (not <assignment-expression>)     if truth-value = false

     (unknown <assignment-expression>) if truth-value = none

   where truth-value is the current truth assignment of state-assignment, and 

     <assignment-expression> ::= (= <variable-name> <value>) "
  
  (let ((assignment-expression (constraint-system-assignment-description
								state-assignment))
		(truth-value (state-assignment-truth-value state-assignment)))
	
	(cond ((true-value? truth-value)
		   
		   assignment-expression)
		  
		  ((false-value? truth-value)
		   
		   (make-negation assignment-expression))
		  
		  ((unknown-value? truth-value)
		   
		   (make-unknown assignment-expression))
		  
		  (:otherwise
		   
		   (error "Expected truth value for assignment ~a, but received ~a."
				  state-assignment truth-value)))))

(defmethod assigned? ((state-assignment state-assignment))
  
  "Returns True iff STATE-ASSIGNMENT is currently assigned 
   a truth value."
  
  (not (eql (state-assignment-truth-value state-assignment)
			*none*)))

;;; To Do:
;;; o This seems messy and should be rethought:
;;;   - This only partially translates an assignment to a description,
;;;     by leaving the assignment object in place.
;;;     - Why is this useful?
;;;     - Instead what about replacing the assignment objecti with its description?
;;;   - Naming is inconsistent.  Shouldn't it simply be state-assignment 
;;;     desciption?
;;;   - An expression (= (= x v) true)  seems ad hoc.  Instead represent the
;;;     truth assignment through positive and negative literals:
;;;     (= x v) or (not (= x v)).
(defmethod state-assignment-to-truth-assignment 
	((state-assignment state-assignment))

  "Given STATE-ASSIGNMENT object with associated truth value, 
  returns an truth assignment description:

       (= STATE-ASSIGNMENT <truth-value>)

   where truth-value is the current truth assignment
   STATE-ASSIGNMENT remains a state-assignment object."
  
  `(= ,state-assignment 
	  ,(state-assignment-truth-value state-assignment)))

;;; ***********************************************************************
;;;
;;;                ITC Simple Temporal Constraint Object
;;;
;;; ***********************************************************************

(defclass ITC-STC (STC) 

  (
   ;; Delete once subclass of essential-constraints; change caller.
;;;   (removed?
;;;	:reader stc-removed?
;;;	;; Indicate that constraint starts as removed from STN.
;;;	:initform T
;;;	:documentation
;;;	"True iff CONSTRAINT is currently removed from its simple temporal
;;;     network.")
  
   (guard
	:initarg :guard
	:accessor stc-guard
	:initform *true*
	:documentation 

	"Null or a list object of state-assignments that represents a
     conjunction.  If constraint is inserted, then to hold, guard must
     be null or an inserted assignment.  An initarg, guard is provided
     as an assignment expression or an assignment object."  ))

  (:documentation 
   "An extension of the base STC class to include slots for ITC reasoning
    and guard conditions over state variables, used to activate and deactivate
    constraints.")
  )

(defmethod initialize-instance :after 
		   ((ITC-STC ITC-STC) 
			&rest extra-args)

  "Creates a new simple temporal constraint, used to performa
   incremental temporal consistency.  Constraints include guard
   conditions over state variables, which are used to activate
   and deactivate these constraints."

  (declare (ignore extra-args))
  
  ;; Note that this function does not initialize any of the status 
  ;; slots of the constraint.  This is performed when the constraint
  ;; is inserted into the network. 

  (with-slots (start end ub lb guard constraint-system d-edges) itc-stc

	(check-legal-bounds! start end lb ub)

	(unless (typep start 'event)
	  (setq start (get-event start constraint-system)))
	
	(unless (typep end 'event)
	  (setq end (get-event end constraint-system)))

	(make-instance 'itc-d-edge :start end :end start
				   :ub (b- lb) :constraint itc-stc)

	(make-instance 'itc-d-edge :start start :end end
				   :ub ub :constraint itc-stc)
  	
	;; Record the assignment corresponding to guard.

	(unless (eql guard *true*)
	  
	  ;; For backwards compatibility with previous
	  ;; guard representations not in the form of
	  ;; a list
	   (unless (listp (car guard))
		 ;(format t "Guard is now: ~a" guard)	
		 (setq guard (list guard)))

	  (setf guard (loop with assignment-object
					 for clause in guard
					 for gvariable = (assignment-variable clause)
					 for gvalue = (assignment-value clause)
					 when (typep gvariable 'state-variable)
					 do (setf gvariable (state-variable-name gvariable))
					 end
					 do (setf assignment-object (find-state-assignment constraint-system gvariable gvalue))
					   (push itc-stc (state-assignment-constraints assignment-object))
					 collecting assignment-object)))
	
	;; Initialize STC for propagation.
	(stc-initialize! itc-stc))
  'T)

;;; Default method defined in mtk-utils/compare.lisp
(defmethod extract-proposition-description 
	((stc ITC-STC))

  "Return the proposition description for state-assignment p."
  
  (with-slots (start end lb ub guard) stc

	`(ITC-STC :start ,(event-name start)
			  :end ,(event-name end) 
			  :lb ,lb
			  :ub ,ub 
			  :guard ,(extract-proposition-description guard))))

(defmethod stc-active? ((stc itc-stc))
  
  "Returns true if STC should actively
   hold in its STN."
  
  (and (not (constraint-removed? stc))
	   (stc-guard-satisfied? stc)))

(defmethod stc-guard-satisfied? ((stc itc-stc))

  "Returns non-nil iff there is no guard condition
   for STC, or guard condition evaluates to *true*."
  
  (with-slots (guard) stc
	
	(or (eql guard *true*)
		(every #'(lambda (x) (true-value? (state-assignment-truth-value x))) guard))))

;;; To Do:
;;; o Look to replace all callers of stc-guard-to-truth-assignment with
;;;   stc-guard-to-state-assignment-w-truth-description, below.
(defmethod stc-guard-to-truth-assignment ((stc itc-stc))

  "Given a simple temporal constraint STC, 
   returns an expression denoting the truth assignment 
   to STC's guard.

   If STC's guard always holds (is *true*), 
      then returns *true*.

   If ST'c guard is a state assignment (= variable value) 
      then returns (= (= variable value) truth-assignment)
         where truth-assignment is one of *TRUE*,*FALSE*, or *NONE*
         depending on whether (= variable value) is *true*, *false*
         or unspecified in the constraint system of STC."
  
  (with-slots (guard) stc
	
	(if (eql guard *true*)
		
		guard
		(loop for g in guard
		   collecting
			 (state-assignment-to-truth-assignment g)))))

(defmethod stc-guard-to-state-assignment-w-truth-description
	((stc itc-stc))

  "Give a simple temporal constraint STC, 
   returns an expression denoting the truth assignment 
   to STC's guard.

   If STC's guard always holds (is *true*), 
      then returns *true*.

   If ST'c guard is a state assignment (= variable value) 
      then returns (= (= variable value) truth-assignment)
         where truth-assignment is one of *TRUE*,*FALSE*, or *NONE*
         depending on whether (= variable value) is *true*, *false*
         or unspecified in the constraint system of STC."
  
  (with-slots (guard) stc

	(if (eql guard *true*)
		
		guard
	  
	  (constraint-system-assignment-description-w-truth guard))))


;;;            ***   Printing a constraint   ***

(defmethod print-object ((obj itc-stc) stream)
  (with-slots (lb ub start end guard) obj 
	(call-next-method)
	(unless (eql guard *true*)
	  (format stream "<-~a" guard))))



;;; ***********************************************************************
;;;
;;;               ITC Distance Graph Edge Object
;;;
;;; ***********************************************************************

;;Extends the base d-edge to include consequents for ITC reasoning.

(defclass itc-d-edge (d-edge)


  ((lb-consequent 
	:accessor d-edge-lb-consequent 
	:initform nil
	:documentation
	"LB-CONSEQUENT stores an event whose lb value is determined from 
     distance edge.")
   
   (ub-consequent 
	:accessor d-edge-ub-consequent 
	:initform nil
	:documentation
	"UB-CONSEQUENT stores an event whose ub value is determined from 
     distance edge.")

   (removed? 
	:accessor d-edge-removed?
	;; Indicate that EDGE starts as removed from network.
	:initform T
	:documentation
	"True iff DISTANCE EDGE is currently removed from its simple temporal
	network."))
  
  ;;  True iff DISTANCE EDGE is currently removed from its
  ;;  simple temporal network.


  )


;;;    ***   Support/Consequent Relation between 
;;;          an Event and a Distance Edge   ***

(defmethod queue-changed-d-edge! ((d-edge itc-d-edge))
  
  "Queues up for propagation an insertion or deletion
   of distance edge D-EDGE.  Only queues if performing
   an incremental update (i.e., an initial propagation 
   has been performed)."

  (let ((stn (constraint-constraint-system (d-edge-constraint d-edge))))
	(when (stn-ub-consistency-checked? STN)
	  
	  ;; Queue only if consistency has been checked, 
	  ;; hence were performing incremental updates.
	  (let ((x (d-edge-start d-edge))
			(y (d-edge-end d-edge)))
		
		(let ((event-pair (find-event-pair x y STN)))
		  
		  (when (null event-pair)
			
			(error "An event pair object doesn't exist for d-edge ~a"
				   d-edge))
		  
		  (queue-changed-event-pair! event-pair STN)

		  T)))))

(defmacro event-has-lb-support? (event)

  "True if event has a distance edge supporting
   the derivation of its lower bound value."
  
  `(not (eql (event-lb-support ,event) ,*unknown*)))

(defmacro event-has-ub-support? (event)

  "True if event has a distance edge supporting
   the derivation of its upper bound value."
  
  `(not (eql (event-ub-support ,event) ,*unknown*)))

(defmacro d-edge-has-lb-consequent? (edge)

  "True if distance EDGE is used to support the lower
   bound value of its END."
  
  `(not (eql (d-edge-lb-consequent ,edge) ,*unknown*)))

(defmacro d-edge-has-ub-consequent? (edge)

  "True if distance EDGE is used to support the upper
   bound value of its END."
  
  `(not (eql (d-edge-ub-consequent ,edge) ,*unknown*)))

;;; The following eight macros allow the implementation of
;;; the support/consequent relation to be modified
;;; in a transparent manner.

(defmacro set-lb-support-relation! (event d-edge)

  "Record that EVENT was assigned its lower bound value
   based on D-EDGE, hence EVENT is a lower bound 
   consequence of D-EDGE and D-EDGE lower bound
   supports EVENT."

  `(progn
     (setf (event-lb-support ,event) ,d-edge)
     (setf (d-edge-lb-consequent ,d-edge) ,event)))

(defmacro set-ub-support-relation! (event d-edge)

  "Record that EVENT was assigned its upper bound value
   based on D-EDGE, hence EVENT is an upper bound 
   consequence of D-EDGE and D-EDGE upper bound
   supports EVENT."

  `(progn
     (set-parent ,event ,d-edge)
     (setf (d-edge-ub-consequent ,d-edge) ,event)
	 t))

(defmacro clear-event-lb-support! (event)

  "Clear any edge recorded as supporting the lower
   bound value of EVENT."
  
  `(setf (event-lb-support ,event) ,*unknown*))

;;; FIX:
;;; o Did Andrew Yi Huang change this to remove-parent,
;;;   rather than `(setf (event-ub-support ,event) ,*unknown*)?
;;;   Does this behavior correctly?  This is not good coding style.
(defmacro clear-event-ub-support! (event)

  "Clear any edge recorded as supporting the upper
   bound value of EVENT."

  `(remove-parent ,event))

(defmacro clear-d-edge-lb-consequent! (edge)

  "Clears any event record as the lower bound
   consequence of distance EDGE."

  `(setf (d-edge-lb-consequent ,edge) ,*unknown*))

(defmacro clear-d-edge-ub-consequent! (edge)

  "Clears any event record as the upper bound
   consequence of distance EDGE."

  `(setf (d-edge-ub-consequent ,edge) ,*unknown*))

(defmacro clear-lb-support-relation! (event d-edge)

  "Clear record that D-EDGE supports the lower bound
   value of EVENT, and that EVENT is a lower bound
   consequent of D-EDGE."

  `(progn
	 (clear-event-lb-support! ,event)
	 (clear-d-edge-lb-consequent! ,d-edge)
	 T))

(defmacro clear-ub-support-relation! (event d-edge)

  "Clear record that D-EDGE supports the upper bound
   value of EVENT, and that EVENT is an upper bound
   consequent of D-EDGE."

  `(progn
	 (clear-event-ub-support! ,event)
	 (clear-d-edge-ub-consequent! ,d-edge)
	 T))

;;; ***********************************************************************
;;;
;;;                External Interface Functions and Macros
;;;
;;; ***********************************************************************

;;; ***********************************************************************
;;;
;;; *** Operating on Distance Edges of a Network
;;;
;;; ***********************************************************************

(defmacro do-inserted-d-edges ((d-edge d-edges) &body body)
  
  "Executes BODY with D-EDGE bound to each d-edge in D-EDGES 
   that is inserted in SOME network.  Behaves exactly like dolist.  
   Presumes that the caller only provides distance edges that are 
   relevant to the current network(s) of interest."
  
  `(dolist (,d-edge ,d-edges)
     (unless (d-edge-removed? ,d-edge)
       ,@body)))
