;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2013-2015 James G. Paterson, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; James G. Paterson.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)


(defclass LTNP (ITC-STN)

  ((looping-constraints
	  :initarg :looping-constraints
	  :accessor ltnp-looping-constraints
	  :initform '()
	  :documentation
	  "Holds a list of looping constraints in the LTNP")

   
   (looping-constraints-ht
	  :initarg :looping-constraints
	  :accessor ltnp-looping-constraints-ht
	  :initform (make-hash-table :test #'equal)
	  :documentation
	  "Holds a mapping from an integer key to looping constraints in the LTNP")

   ;; (constraint-to-reward-ht 
   ;; 	:initform (make-hash-table :test #'equal)
   ;; 	:accessor constraint-to-reward-ht
   ;; 	:documentation
   ;; 	"constraint-to-reward-ht holds the hash table that maps 
   ;;  the looping constraint to a reward function")
   )
  (:documentation "Contains a list of all looping temporal contraints in a LTNP"))


(defclass LTCP (ITC-STC)

  ((N-lb 
	:initarg :N-lb
	:accessor N-lb
	:initform nil
	:documentation 
	"Stores the lower bounds of the looping contraint.") 
   (N-ub 
	:initarg :N-ub
	:accessor N-ub
	:initform nil
	:documentation 
	"Stores the lower bounds of the looping contraint.")
   (t-lb 
	:initarg :t-lb
	:accessor t-lb
	:initform nil
	:documentation 
	"Stores the lower bound of the temporal contraint per loop.")
   (t-ub 
	:initarg :t-ub
	:accessor t-ub
	:initform nil
	:documentation 
	"Stores the upper bound of the temporal constraint per loop.")
									
   (reward-function
	:initform (make-instance 'linear-looping-episode-cost)
	:initarg :reward-function
	:accessor LTCP-reward-function
	:documentation
	"Stores the reward function associated with a LTCP."))

  ;; (reward-constant
  ;;  :initform 1
  ;;  :initarg :reward-constant
  ;;  :accessor LTCP-reward-constant
  ;;  :documentation
  ;;  "LTCP-reward-constant is used to calculate reward value.")
  
  (:documentation
   "This object holds information about a simple temporal constraint.
   The simple temporal constraint specifies interval constraint:
    (end - start) in [lb,ub]")) 


(defclass Search-Branch ()

  ((parent   ;; The parent is always consistent!! Been previously checked. 
	:initarg :parent
	:accessor branch-parent
	:initform nil
	:documentation 
	"Stores the path of node numbers to the search node in the search tree.")


   (constraint-split-number  
	:initarg :parent-constraint-split-number 
	:accessor constraint-split-number
	:initform nil
	:documentation 
	"Stores the integer key of the loop range of the node that have have split to get two children.
    Needed so we know which constraint to restore while backtracking.")

   (integer-to-loop-ht 
	:initarg :integer-to-loop-ht 
	:initform (make-hash-table :test #'equal)
	:accessor integer-to-loop-ht
	:documentation
	  "Stores a mapping from looping constraint number to the looping constraint
      range. (see integer-to-constraint-ht under Search-Tree class).")
   
   (reward-ub
	:initarg :reward-ub
	:accessor reward-ub
	:initform 0
	:documentation 
	"Stores the upper bound of the reward range."))

  (:documentation
   "This object holds information about a Search Branch, essentially it's a node in the search space."))

(defclass Search-Tree ()

  (
   (start-branch
		:accessor start-branch
		:initform nil  


	)
   (search-q
	:accessor search-q
	:initform (make-ordered-list-queue :key #'reward-ub :order #'>=)   
	:documentation 
	"Stores search queue, in best-reward-first order."))

  (:documentation
   "This object holds information about search queue, in best-reward-first order."))

(defmethod constraint-reward-range? ((LTCP ltcp) &optional Nl Nu)
  "Returns a 'cons' of the reward range of a looping temporal constraint. If a lower and upper loop 
   bound aren't specified, the loop bounds set in the constraint are used."
  (with-slots (N-lb N-ub)  
	  (let ((range (cons nil nil)))
		(if (and Nl Nu) 		
			(setf range (cons (-(calculate-looping-cost (LTCP-reward-function ltcp) Nl)) (-(calculate-looping-cost (LTCP-reward-function ltcp) Nu))))
			(setf range (cons (-(calculate-looping-cost (LTCP-reward-function ltcp) (N-lb ltcp))) (-(calculate-looping-cost (LTCP-reward-function ltcp) (N-ub ltcp))))))
		(return-from constraint-reward-range? range))))

(defmethod print-object ((obj ltcp) stream)
  (with-slots (lb ub N-lb N-ub t-lb t-ub start end) obj 
	(format stream "~%    (~a,~a)relaxed/temp/loops [~a,~a][~a,~a][~a,~a]" start end lb ub t-lb t-ub N-lb N-ub)))

