;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2008-2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;; ***********************************************************************
;;;
;;;                Manipulation of a Simple Temporal Network
;;;               That Supports Incremental Temporal Inference
;;;
;;; ***********************************************************************

;;; This file contains the major interface functions of a temporal constraint
;;; package that is specialized to incremental temporal consistency checking.  
;;; It includes functions 1) for creating a simple temporal network, 
;;; 2) inserting and removing simple temporal constraints of a network, and 
;;; 3) querying the consistency, event bounds and conflicts of a network.  
;;; This file depends on the objects and macros defined in stn-defs.lisp 
;;; and defs.lisp. It specializes methods in manipulate-stn.lisp for 
;;; modifying STNs. It depends on the algorithms in propagate-ub.lisp and 
;;; propagate-lb.lisp to update the consequences of insertions and removals, 
;;; and files in conflicts-lisp and consistency.lisp to manage inconsistencies 
;;; and conflicts.

;;; The external interface functions defined in this file are:

;;; o Creating Simple Temporal Networks

;;;   (networks, simple temporal constraints and events).
;;;
;;;  MAKE-INSTANCE 'ITC-STN &KEY (NAME (GENTEMP "STN-")) (START-NAME 'START)
;;;  
;;;     Creates a new simple temporal network named NAME.  
;;;     The network is empty, that is, it has no constraints or events
;;;     associated with it, other than the START event.  
;;;     The function returns the newly created network.
;;;     The function has the side effect of setting 
;;;     *DEFAULT-STN* to be the newly created network.

;;; o Retrieving and Creating Events

;;;   GET-EVENT (EVENT-NAME (NETWORK ITC-STN))
;;;     Returns a unique event with EVENT-NAME,associated with NETWORK.  
;;;     Creates the event if it doesn't already exist, and returns it.

;;;   FIND-EVENT EVENT-NAME (NETWORK STN)
;;;     Returns a unique event with EVENT-NAME that is associated with 
;;;     NETWORK  If no such event exists in NETWORK, it returns NIL.

;;; o Creating Discrete State Variables

;;;   o State Variables

;;; To Do: Either generalize and move the next two defns to manipulate-stn.lisp, 
;;;        OR specialize method to ITC-STN.  In the former, we would want to
;;;        add guards to STC constraint objects.

;;;   DEFINE-STATE-VARIABLE (STN STN) SOURCE STATE-VARIABLE-NAME DOMAIN 

;;;     Defining a new state variable for STN, ranging over domain.
;;;     VARIABLE-NAME is any lisp object, denoting a variable, and DOMAIN 
;;;     is a list of values that the variable may take on.
;;;     SOURCE is the object that created variable, and STN is a
;;;     simple temporal nework that variable is part of.

;;;   FIND-STATE-VARIABLE STN STATE-VARIABLE-NAME 

;;;     Returns object representing STATE-VARIABLE-NAME for STN.
;;;     If STATE-VARIABLE-NAME is not defined in STN, then returns NIL.

;;; o Creating and Inserting Simple Temporal Constraints

;;;   INSERT-NEW-CONSTRAINT ITC-STN &KEY START END LB UB &OPTIONAL GUARD SOURCE
;;;     Creates a new simple temporal constraint (STC) from SOURCE, and INSERTS
;;;     it into NETWORK.  START and END are events or names of events, and LB 
;;;     and UB are reals, where LB <= UB.  STC holds whenever GUARD is satisfied,
;;;     where GUARD is currently restricted to a discrete variable assignment.

;;;     The function creates a new STC, but DOES NOT insert it into 
;;;     NETWORK, i.e., the stc is a removed stc.  A STC may be later inserted 
;;;     by calling INSERT-CONSTRAINT.  

;;;     Note that this function does not check for
;;;     duplicate constraints, but merely proceeds to create additional 
;;;     copies.  Any required events are created by the function,
;;;     and pointers between the constraint and its events are put 
;;;     in place.  
;;;     The function returns the newly created constraint.

;;;     Note that this function does not check for
;;;     duplicate constraints, but merely proceeds to create additional
;;;     copies.  Any required events are created by the function, 
;;;     and pointers between the constraint and its events are put 
;;;     in place.  The function returns the newly created constraint.

;;;   MAKE-INSTANCE 'ITC-STC :SOURCE SOURCE :CONSTRAINT-SYSTEM STN :START START
;;;				          :END END :LB LB :UB UB :GUARD GUARD

;;;   NEW-CONSTRAINT (STN ITC-STN) &KEY GUARD START END LB UB SOURCE
;;;
;;;   Creates a new simple temporal constraint (STC) for NETWORK whose source is 
;;;   SOURCE.  START and END are event objects or names of events, and LB and UB 
;;;   are reals, where LB <= UB.  Guard is either *true* or an assignment 
;;;   (= V X) of a state variable V to one of its domain elements X.  V may either 
;;;   by a state variable name or state-variable object of one of STN's state 
;;;   variables.  The default for guard is *true*.
;;;
;;;   If Guard is *true*, then when STC is inserted, it enforces:
;;;
;;;      (START - END) in [l,u]
;;;
;;;   Otherwise, when STC is inserted, it enforces:
;;;
;;;    (= V X) implies (START - END) in [l,u]
;;;
;;;   The function creates a new STC, but DOES NOT insert it into 
;;;   NETWORK, i.e., the STC is a removed STC.  STC may be inserted later
;;;   by calling INSERT-CONSTRAINT.  
;;;
;;;   This function DOES NOT check for duplicate constraints, 
;;;   but merely proceeds to create additional copies.  Events, however,
;;;   are canonicalized.
;;;
;;;   The function returns the newly created constraint.

;;; o Inserting and Removing Simple Temporal Constraints that 
;;;   are already associated with a Network.
;;;
;;;   INSERT-CONSTRAINT ITC-STN ITC-STC
  
;;;     Inserts a removed simple temporal constraint ITC-STC into its 
;;;     network, ITC-STN.  Updates the propagation queue for network, 
;;;     where propagation is used to determine event bounds and 
;;;     consistency when queried.  Propagation is not initiated by this 
;;;     function.

;;;     ITC-STC itself must have been created earlier using NEW-CONSTRAINT.
;;;     If ITC-STC is already inserted, this function does nothing.  

;;;     The function is called for effect, it does not return anything 
;;;     meaningful.

;;;   REMOVE-CONSTRAINT ITC-STN ITC-STC

;;;     Removes constraint ITC-STC from its network, ITC-STN, along with any of its
;;;     logical consequences.  Finally, it updates the queue that PROPAGATE! 
;;;     uses to find alternative support for the removed consequences.  
;;;     If the constraint is already removed, it does nothing.  
;;;     The function does not return anything meaningful.

;;;   o Retrieving State Assignments

;;;   FIND-STATE-ASSIGNMENT STN VARIABLE-NAME VALUE

;;;     Returns the assignment object for (= VARIBLE-NAME VALUE) in STN.  
;;;     If no such assignment exists in STN, it returns NIL.

;;;   o Making and Removing Assignments.

;;;   ASSERT! (STN STN)(ASSIGNMENT STATE-ASSIGNMENT)
;;;	     &OPTIONAL (TRUTH-VALUE *TRUE*) (SOURCE NIL)
;;;
;;;      Makes ASSIGNMENT in STN, based on SOURCE.  
;;;      TRUTH-VALUE must be *true* or *false*
;;;      (*false* currently has no effect).

;;;   UNASSERT! (STN STN)(ASSIGNMENT STATE-ASSIGNMENT)
;;;	     &OPTIONAL (TRUTH-VALUE *TRUE*) (SOURCE NIL)
;;;
;;;      Removes ASSIGNMENT from STN, based on SOURCE.  
;;;      TRUTH-VALUE is one of *true* or *false*.
;;;       (*false* currently has no effect)

;;;   ASSIGN! (STN ITC-STN) VARIABLE-NAME VALUE 
;;;      &OPTIONAL (TRUTH-VALUE *TRUE*) (SOURCE NIL)
;;;    
;;;      Makes assignment (= VARIABLE-NAME VALUE) have TRUTH-VALUE in STC, 
;;;      based on SOURCE.  
;;;
;;;      VARIABLE-NAME must be a defined variable in STC.
;;;      VALUE must be an element of the domain of VARIABLE-NAME.
;;;      TRUTH-VALUE must be one of *TRUE* or *FALSE*.
;;;
;;;      If these conditions are not satisfied, then an error is issued,
;;;      no assertion is made.
;;;
;;;      Returns T iff success.

;;;   UNASSIGN! (STN ITC-STN) VARIABLE-NAME VALUE
;;;		  &OPTIONAL (TRUTH-VALUE *TRUE*) (SOURCE NIL)
;;;    
;;;    Removes TRUTH-VALUE for assignment (= VARIABLE-NAME VALUE) in STC ,
;;;     based on SOURCE.  
;;;
;;;     VARIABLE-NAME must be a defined variable in STC.
;;;     VALUE must be an element of the domain of VARIABLE-NAME.
;;;     TRUTH-VALUE must be one of *TRUE* or *FALSE*.
;;;
;;;     If these conditions are not satisfied, then an error is issued,
;;;     no assertion is removed
;;;
;;;     Returns T iff success.

;;;   ASSERT-ASSIGNMENTS! (STN ITC-STN) SOURCE ASSIGNMENT-DESCRIPTIONS
;;;
;;;     Asserts every assignment in assignment-descriptions.  Each
;;;     assignment description is a list: 

;;;      (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))

;;;     Denoting assignment (= VARIABLE-NAME VALUE) having TRUTH-VALUE
;;;     in STN.

;;;   UNASSERT-ASSIGNMENTS! (STN ITC-STN) SOURCE ASSIGNMENT-DESCRIPTIONS
;;;
;;;     Unasserts every assignment in assignment-descriptions.  Each
;;;     assignment description is a list: 

;;;         (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))

;;;     Denoting assignment (= VARIABLE-NAME VALUE) having TRUTH-VALUE
;;;     in STN.

;;; o Print Functions

;;;   PRINT-STN! ITC-STN

;;;   Enumerates the events, simple temporal constraints,
;;;   and bounds of STN.

;;; o Querying the Bounds on Events based on Current Constraints

;;;   Side Effect:  Leaves the tighest bounds on events computed,
;;;   a set of support for the derivation of each bound.  If network
;;;   is inconsistent, then these bounds are not meaningful, however,
;;;   all inconsistencies are removed from the network."

;;;   EVENT-TIME-BOUNDS (EVENT ITC-EVENT) (NETWORK ITC-STN)

;;;     Returns (values LB UB) where LB and UB are the lower bound and
;;;     upper bound on the time that EVENT occurs, given
;;;     the simple temporal constraints that are currently inserted 
;;;     into NETWORK

;;;   EVENT-PAIR-TIME-BOUND (START ITC-EVENT) (NETWORK ITC-STN)

;;;     Returns (values LB UB) where LB and UB are the lower bound and
;;;     upper bound on the time that EVENT occurs, given
;;;     the simple temporal constraints that are currently inserted 
;;;     into NETWORK.

;;;   EVENT-SUPPORT (EVENT ITC-EVENT) (NETWORK ITC-STN)

;;;     Returns the simple temporal constraint that ITC used to determine 
;;;     the current bounds on EVENT's time, given bounds on neighboring 
;;;     events.

;;; The following is disabled until needed.

;;;   EVENT-CONSEQUENCES (event itc-event) (network ITC-STN)

;;;     Returns all simple temporal constraints that helped compute
;;;     the current bounds of some event, based on the bounds on EVENT.

;;; o Checking Consistency

;;; The following three interface functions, for testing consistency,are defined
;;; in consistency.lisp.

;;;   CONSISTENT? (STN STN) &OPTIONAL (PERFORM-SEARCH? T)
;;;	    (RETURN-CONFLICT? T) (KEY #'STC-GUARD-TO-TRUTH-ASSIGNMENT))

;;;    (CONSISTENT-WITH? (STN STN) &OPTIONAL &KEY (ADDED-CONSTRAINTS NIL)
;;; 	(ADDED-ASSIGNMENTS NIL) (SOURCE NIL) (RETURN-CONFLICT? T))

;;;   (INCONSISTENT-WITH? (STN STN) &OPTIONAL &KEY (ADDED-CONSTRAINTS NIL) 
;;;      (ADDED-ASSIGNMENTS NIL)(SOURCE NIL) (RETURN-CONFLICT? T)
;;;      (BASE-STN-CONSISTENT-WITH-ALL-ASSIGNMENTS? T))

;;; The following is an internal function; call one of the above instead.

;;;   CONSISTENT-STN? (NETWORK ITC-STN)
;;;  
;;;   Checks consistency of simple temporal NETWORK.
;;;   Returns (values CONSISTENT? CONFLICT).
;;;   CONSISTENT? is non-NIL iff a schedule to the events of NETWORK
;;;   exists that is consistent with the network's simple temporal constraints.
;;;   If NETWORK is found inconsistent, then a conflict is returned.
;;;   A CONFLICT is a subset of the constraints of network that are
;;;   inconsistent.

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			new-constraint
;;;			
;;;;;;			find-state-variable
;;;			define-state-variable
;;;			
;;;			get-assignment
;;;			assign!
;;;			unassign!
;;;			
;;;			assert!
;;;			unassert!
;;;
;;;			insert-constraint
;;;			remove-constraint
;;;
;;;			get-event
;;;			find-event
;;;
;;;			print-stn!
;;;	
;;;			consistent-stn?
;;;            event-time
;;;            event-support
;;;;;;            event-consequences
;;;			)
;;;		  :ITC)
;;;  )

;;; To Do:
;;; See to do under Backup and Restoration of the Current State of an STN.


;;;   ***
;;;   *** New Events, Variables, Assignments and Constraints
;;;   ***

;;;   *** Events

(defmethod get-event
    (event-name (network ITC-STN))

  " Returns the event object of NETWORK described 
    by EVENT-NAME.  If the object already exists, 
    it is returned; otherwise, a new event object
    is created, associated with NETWORK and returned.  

    Note: This is the preferred way to create a new 
    event since it avoids duplication."

  (or (find-event event-name network)
	  (make-instance 'itc-event :name event-name
					 :network network
					 :start? nil)))


;;;   *** State Variables

;;; To Do:
;;; o Use a naming convention consistent with other
;;;   creation functions.
(defmethod define-state-variable 
    ((STN STN) source variable-name domain)

  "Creates a new state variable for STN, ranging over DOMAIN.

   NAME is any lisp object, denoting a variable, and DOMAIN 
   is a list of values that the variable may take on.

   SOURCE is the object that created variable, and STN is a
   simple temporal network that variable is part of.

   Creates a list of assignment objects for STN denoting the
   possible assignments of variable to an element of its domain."

  (make-instance 'state-variable
 	:name variable-name
	:domain domain
	:STN STN
	:source source))


;;;   *** State Variable Assignments

;;; Fix:
;;; o If check-state-variable-assignment? returns nil,
;;;   shouldn't we flag an error?  Return nil seems dangerous.

;;; To Do:
;;; o Appears redundant with CHECK-AND-RETURN-VALID-STATE-ASSIGNMENT.
;;;   - Collapse.  Use the name GET-ASSIGNMENT, but the implementation of
;;;     CHECK-AND-RETURN-VALID-STATE-ASSIGNMENT.
(defmethod get-assignment
    ((STN ITC-STN) variable-name value)

  " Returns a state-assignment object for (= VARIABLE-NAME VALUE) 
    in STN.

    VARIABLE-NAME must be a defined variable in STN.
    VALUE must be an element of the domain of VARIABLE-NAME.

    If these conditions are not satisfied, then a warning is
    issued and nil is returned."
  
  ;; Note: All assignments for VARIABLE-NAME have been created when the
  ;;   variable was defined, hence we never need to create the assignment
  ;;   during get.
  
  (when (check-state-variable-assignment? stn variable-name value)
	(find-state-assignment STN variable-name value)))

;;; To Do:
;;; This appears redundant with get-assignment, collapse.
(defmethod check-and-return-valid-state-assignment
	(STN variable-name value &optional (truth-value *true*))
  
  "Returns the assignment object that corresponds to (= VARIABLE-NAME VALUE)
   in STN.  Flags an error if assignment is not defined."

  (let ((variable (find-state-variable STN variable-name)))

	(when (null variable)
	  (error "~A is not a defined variable for STN ~A."  
			 variable-name STN))

	(let ((assignment (find-state-assignment STN variable-name value)))
	  (when (null assignment)
		(error "Value ~A is not a defined domain value for variable ~A in STN ~A."  
			   value variable-name STN))
	  
	  (when (not (truth-value? truth-value))
		(error "~A is not a valid truth value for assignment ~a in STN ~A "  
			   truth-value assignment STN))
	  
	  assignment)))


;;;   *** Constraints

(defmethod new-constraint
	((STN ITC-STN) &key source start end lb ub 
	 (guard *true*) 
	 (lb-relaxable? nil)
	 (lb-cost-ratio nil)
	 (lb-relaxed? nil)
	 (relaxed-lb nil)
	 (ub-relaxable? nil)
	 (ub-cost-ratio nil)
	 (ub-relaxed? nil)
	 (relaxed-ub nil))

  "Creates a new simple temporal constraint (STC) for NETWORK whose source is 
   SOURCE.  START and END are event objects or names of events, and LB and UB 
   are reals, where LB <= UB.  Guard is either *true* or an assignment 
   (= V X) of a state variable V to one of its domain elements X.  V may either 
   by a state variable name or state-variable object of one of STN's state 
   variables.  The default for guard is *true*.

   If Guard is *true*, then when STC is inserted, it enforces:

      (START - END) in [l,u]

   Otherwise, when STC is inserted, it enforces:

    (= V X) implies (START - END) in [l,u]

   The function creates a new STC, but DOES NOT insert it into 
   NETWORK, i.e., the STC is a removed STC.  STC may be inserted later
   by calling INSERT-CONSTRAINT.  

   This function DOES NOT check for duplicate constraints, 
   but merely proceeds to create additional copies.  Events, however,
   are canonicalized.

   The function returns the newly created constraint."
  
 ;; (format t "~A-~A:[~A,~A] ~A:~A~%" start end lb ub lb-relaxable? lb-cost-ratio)

  (make-instance 'itc-stc :source source
				 :constraint-system STN
				 :start start
				 :end end
				 :lb lb
				 :ub ub
				 :guard guard
				 :lb-relaxable? lb-relaxable?
				 :lb-cost-ratio lb-cost-ratio
				 :lb-relaxed? lb-relaxed?
				 :relaxed-lb relaxed-lb
				 :ub-relaxable? ub-relaxable?
				 :ub-cost-ratio ub-cost-ratio
				 :ub-relaxed? ub-relaxed?
				 :relaxed-ub relaxed-ub
				 ))


;;; Fix:
;;; o Make arguments consistent with ISAT.
(defmethod insert-new-constraint
	((constraint-system itc-STN) &key source start end lb ub (guard *true*))

  "Creates a new STC whose source is SOURCE, AND inserts it into 
   CONSTRAINT-SYSTEM.  START and END are names of events, and LB and UB are reals,
   where LB <= UB.

   See documentation for new-constraint.

   Note that this function does not check for duplicate constraints, 
   but merely proceeds to create additional copies.  Any required events 
   are created by the function.  The function returns the newly created 
   constraint."
  ;; create the constraint,
  (let ((stc 
		 (make-instance 'ITC-STC :source source
						:CONSTRAINT-SYSTEM constraint-system
						:start start
						:end end
						:lb lb
						:ub ub
						:guard guard)
		 ))
	;; add it into the constraint-system
    (insert-constraint constraint-system stc)
	;; and return it.
	(values stc)))


;;;   ***
;;;   *** Asserting and Removing Variable Assignments
;;;   ***

;;; Note: The following functions operate directly on assignment objects.
;;; Assign! and unassign! are equivalent user functions that operate on 
;;; variable names and values.

(defmethod assert!
    ((STN STN)(assignment state-assignment)
	 &optional (truth-value *true*) (source nil))

  "Makes ASSIGNMENT in STN, based on SOURCE.  

   TRUTH-VALUE must be *true* or *false*
   (*false* currently has no effect)."

  (debug-print-for :itc t "Asserting ~A in ~A, with source ~A." 
				   assignment STN source)
  
  (when (assigned? assignment)
	(error "Can't assert state-assignment ~a, already ~a."
		   assignment (state-assignment-truth-value assignment)))
  
  (unless (truth-value? truth-value)
	(error "Invalid truth values ~a when asserting ~a."
		   truth-value assignment))

  (setf (state-assignment-truth-value assignment)
	truth-value)

  ;; Propagate the constraints activated by assignment.
  ;; A constraints is active if inserted and its guard,
  ;; ASSIGNMENT, is true.
  (when (true-value? truth-value)
	(loop for stc in (state-assignment-constraints assignment)
		when (stc-active? stc)
		do (debug-print-for :itc t "  ~A activates ~A." 
							assignment stc)
		   (dolist (edge (stc-d-edges stc))
			 (insert-d-edge! edge))))
  'T)

;;; Note:
;;; o Truth value isn't used, but is left in to be consistent
;;;   with unassert! for ISAT.  Could be removed.
(defmethod unassert!
    ((STN STN)(assignment state-assignment)
	 &optional (truth-value *true*) (source nil))

  "Removes ASSIGNMENT from STN, based on SOURCE.  

   TRUTH-VALUE is one of *true* or *false*.
   (*false* currently has no effect)"
  
  (debug-print-for :itc t "Unasserting ~A in ~A, with source ~A." 
				   assignment STN source)
  
  (unless (assigned? assignment)
	(error "Can't unassign ~a, not assigned."
		   assignment))
  
  (unless (truth-value? truth-value)
	(error "Invalid truth values ~a when unassigning ~a."
		   truth-value assignment))
  
  (let ((old-tv (state-assignment-truth-value assignment)))
	(setf (state-assignment-truth-value assignment)
	  *none*)

	(when (true-value? old-tv)
	  ;; Those inserted constraints that are activated by assignment
	  ;; just shifted from active to inactive.
	  (loop for stc in (state-assignment-constraints assignment)
		  unless (constraint-removed? stc)
				 ;; Deactive inserted STC by removing it's d-edges.
		  do (debug-print-for :itc t "  ~A deactivates ~A." 
							  assignment stc)
			 (dolist (edge (stc-d-edges stc))
			   (remove-d-edge! edge))))))



;;;   ***
;;;   *** Inserting and Removing Simple Temporal Constraints
;;;   ***

(defmethod insert-constraint ((constraint-system ITC-STN) (constraint itc-stc))
  
  "Inserts a simple temporal CONSTRAINT into its constraint-system. 
   It queues up incremental propagation, used to determine event bounds 
   and consistency when queried.  Propagation is not initiated by this 
   function.  Note: no queuing is required unless consistency has been
   previously checked.

   CONSTRAINT itself must have been created earlier using NEW-CONSTRAINT.
   If CONSTRAINT is already inserted, this function does nothing.  

   The function is called for effect; it does not return anything 
   meaningful."

  ;; If active, insert its d-edges.
  (when (stc-active? constraint)
	(dolist (edge (stc-d-edges constraint))
	  (insert-d-edge! edge)
	  ;; Restore the initial ub when inserting the constraints
	  (setf (d-edge-ub edge) (d-edge-initial-ub edge))))
  T)

(defmethod remove-constraint ((constraint-system itc-stn)(constraint itc-stc))
  
  "Remove simple temporal CONSTRAINT from its constraint-system. This function 
   queues up incremental propagation, used to update event bounds and 
   consistency, when queried.  Propagation is not initiated by this 
   function.  Note: no queuing is required unless consistency has 
   been previously checked.
  
   If CONSTRAINT is already removed, this function does nothing.

   The function is called for effect, it does not return anything 
   meaningful."

  ;; The remove-constraint :around method from essential-constraint
  ;; ensures that this method is ONLY called if constraint was inserted
  ;; and is being removed (as opposed to being removed when already removed.).

  ;; Nothing needs to be done if the constraint is already removed.

  (when *itc-statistics-p*
	(incf *remove-constraint-counter*))

  (debug-print-for :itc t "Removing constraint ~A." 
				   constraint)

  ;; The distance graph edges of simple temporal constraint C have
  ;; been inserted only if C's guard is satisfied.
  ;; If this is the case, then remove those d-edges.
  (cond ((stc-guard-satisfied? constraint) 
		 (debug-print-for :itc t "  ~A guard satisfied, remove d edges." 
						  constraint)
	
		 (dolist (edge (stc-d-edges constraint))
		   ;;(remove-d-edge! edge) 
		   ;; Instead of removing the edge. Change it to infinity!
		   (setf (d-edge-ub edge) 1000000)
		   (queue-changed-d-edge! edge)		 
		   ))
		(:otherwise
		 (debug-print-for :itc t 
						  "  ~A guard unsatisfied, no d-edges to remove."
						  constraint)))
  
  T)

;;;   ***
;;;   *** Assigning and Unassigning Variables
;;;   ***

;;; The following user functions are equivalent to assign! and unassign!, 
;;; but take a variable name and value, rather than an assignment object.

;;; Note:
;;; o The following function is identical to that for ISAT and could be
;;;   generalized.
(defmethod assign! ((STN ITC-STN) variable-name value
					&optional (truth-value *true*) (source nil))
    
  "Makes assignment (= VARIABLE-NAME VALUE) have TRUTH-VALUE in STC, 
   based on SOURCE.  

   VARIABLE-NAME must be a defined variable in STC.
   VALUE must be an element of the domain of VARIABLE-NAME.
   TRUTH-VALUE must be one of *TRUE* or *FALSE*.

  If these conditions are not satisfied, then an error is issued,
  no assertion is made.

  Returns T iff success."

  (let ((assignment 
		 (check-and-return-valid-state-assignment
		  STN variable-name value truth-value)))
	(assert! stn assignment truth-value source)
	T))

;;; Note:
;;; o The following function is identical to that for ISAT and could be
;;;   generalized.
(defmethod unassign! ((stn itc-stn) variable-name value
					  &optional (truth-value *true*) (source nil))
    
  "Removes TRUTH-VALUE for assignment (= VARIABLE-NAME VALUE) in STC ,
   based on SOURCE.  

   VARIABLE-NAME must be a defined variable in STC.
   VALUE must be an element of the domain of VARIABLE-NAME.
   TRUTH-VALUE must be one of *TRUE* or *FALSE*.

  If these conditions are not satisfied, then an error is issued,
  no assertion is removed

  Returns T iff success."

  (let ((assignment 
		 (check-and-return-valid-state-assignment
		  STN variable-name value truth-value)))
	(unassert! stn assignment truth-value source)
	T))

(defmethod assert-assignments! 
	((stn itc-stn) source assignment-descriptions)
  
  "Asserts every assignment in assignment-descriptions.  Each
   assignment description is a list: 

      (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))

   Denoting assignment (= VARIABLE-NAME VALUE) having TRUTH-VALUE
   in STN."

  (loop for (variable-name value . truth-value-list) 
	  in assignment-descriptions
      for truth-value = (or (first truth-value-list) *true*)
      do (assign! stn variable-name value truth-value source)))

(defmethod unassert-assignments!
	((stn itc-stn) source assignment-descriptions)

  "Unasserts every assignment in assignment-descriptions.  Each
   assignment description is a list: 

      (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))

   Denoting assignment (= VARIABLE-NAME VALUE) having TRUTH-VALUE
   in STN."

  (loop for (variable-name value . truth-value-list) 
	  in assignment-descriptions
      for truth-value = (or (first truth-value-list) *true*)
      do (unassign! stn variable-name value truth-value source)))

;;;   ***
;;;   *** Print Functions
;;;   ***

(defmethod print-stn! ((stn itc-stn) &optional (print-removed? t))

  "For simple temporal network STN, enumerates the events, 
   distance edges, and bounds of STN."
  
  (format t "~2%STN ~a:" stn)
  (format t "~%  Event bounds:~%    ")
  (loop with i = 0
	  for e in (stn-events stn)
	  do (setq i (+ i 1))
		 (format t "~A in [~A,~A], " e (event-lb e)(event-ub e))
	  when (= i 4)
	  do (setq i 0)
		 (format t "~%    ")
		 )
  (format t "~%  Constraints (* denotes active):~%    ")
  (loop with i = 0
	  for c in (constraint-system-constraints stn)
	  do (setq i (+ i 1))
		 (cond ((and (constraint-removed?  c)
					 print-removed?)
				
				(format t "~a [out], " c))
			   
			   ((stc-active? c)
				
				(format t "~a *, " c))
			   
			   (:otherwise
				
				(format t "~a, " c)))
	  when (= i 3)
	  do (setq i 0)
		 (format t "~%    "))
  
  (format t "~%  D-edges:~%    ")
  (loop with i = 0
	  for e in (stn-events stn)
	  do (loop for d in (event-successors e)
			 do (setq i (+ i 1))
				(if (and print-removed?
						 (d-edge-removed? d))
					(format t "~A>~A:~A[out], " 
							(d-edge-start d)(d-edge-end d) (d-edge-ub d))
				  (format t "~A>~A:~A, " 
						  (d-edge-start d)(d-edge-end d) (d-edge-ub d)))
			 when (= i 4)
			 do (setq i 0)
				(format t "~%    ")
				))
  (format t "~%"))

;;;   *** Query Functions

;;; Note: an internal function, called by external functions 
;;; consistent? consistent-with? ...
(defmethod consistent-stn? ((constraint-system itc-stn) &key 
							(initialize? nil)
							(return-negative-cycle? nil))
  
  "Checks consistency of simple temporal CONSTRAINT-SYSTEM.
   Returns (values CONSISTENT? CONFLICT).

   CONSISTENT? is non-NIL iff a schedule to the events of CONSTRAINT-SYSTEM
   exists that is consistent with the constraint-system's simple temporal 
   constraints.

   If CONSTRAINT-SYSTEM is found inconsistent, then a conflict is returned.

   A CONFLICT is a subset of the constraints of constraint-system that are
   inconsistent.

   If INITIALIZE? is True, then reset all distances before propagating;
   otherwise, update incrementally from previous values.

   Side Effect:  Leaves the tighest bounds on events computed,
   a set of support for the derivation of each bound.  If constraint-system
   is inconsistent, then these bounds are not meaningful, however,
   all inconsistencies are removed from the constraint-system."

  (when initialize?
	(stn-initialize! constraint-system))

  (debug-print-for 
   :itc t "Testing consistency of STN ~A." constraint-system)


  ;; (format t "Testing consistency of : ")

  ;; (loop for constraint in (constraint-system-constraints constraint-system)


  ;; 	   do (when (not (constraint- constraint))
  ;; 			(format t " ~A " constraint))
	   
  ;; 	   )

  ;; Check consistency based on propagation.
  ;; First make sure that propagation is up to date.
  (propagate! constraint-system)

  (let ((inconsistent? (stn-ub-inconsistent? constraint-system)))

    (cond (inconsistent?
	   
		   ;; Inconsistent by propagation.
		   (debug-print-for :itc t "  ~A is inconsistent." constraint-system)
		   (let ((conflict (itc-conflict-constraints
							(stn-ub-conflict constraint-system))))
			 
			 (if return-negative-cycle?
				 (values nil (list conflict (stn-ub-conflict constraint-system)))
				 (values nil conflict))
			 ))

		  ;; Consistent by propagation.
		  (:otherwise
		   (debug-print-for :itc t "  ~A is consistent." constraint-system)
		   (values T nil)
		   ))))

(defmethod event-time-bounds
	((event itc-event) (constraint-system itc-STN))
  
  "Returns (values LB UB), where lb and ub are lower and upper bounds, 
   respectively, on the time that EVENT occurs, given the the constraints
   that are currently inserted into CONSTRAINT-SYSTEM."
  
  ;; First make sure that propagation is up to date.
  (propagate! constraint-system)

  ;; Now return the bounds.
  (values (event-lb event) (event-ub event)))

;;; Check:
;;; o Is the following function needed?
(defmethod event-support ((event itc-event) (constraint-system ITC-STN))

  "Returns the two constraints used by ITC to determine the 
   current upper and lower bound on EVENT's time.  EVENT is
   an event of CONSTRAINT-SYSTEM.
   Returns (VALUES LB-CONSTRAINT UB-CONSTRAINT), where
   LB-CONSTRAINT and UB-CONSTRAINT are simple temporal 
   constraints of CONSTRAINT-SYSTEM."
  
  ;; First make sure that propagation is up to date.
  (propagate! constraint-system)

  ;; Now return the constraints used to directly
  ;; determine the bounds on EVENT.
  (values (event-lb-support event)
		  (event-ub-support event)))

;;; Note:
;;;  o This is a loose bound.  A tighter bound can be computed using
;;;    SSSP from start.
(defmethod event-pair-time-bound
	((start itc-event) (end itc-event) (stn itc-stn))
		
  "Given STN with events START and END,
   computes the implied constraint (start,end)[l,b].
   Returns (values l b)."

  ;; Note: The calls to EVENT-TIME-BOUNDS ensures that
  ;;       propagate is called.

  (stn-initialize! STN);;;Added by Peng; Used to initialize STN before propagation. Otherwise infinity occurs in the result.

  (multiple-value-bind (sl su)
	  (event-time-bounds start STN)
	
	(multiple-value-bind (el eu)
		(event-time-bounds end STN)

	  ;; (s,e)[l,u] subset [el-su, eu-sl] = [el, eu] - [sl, su] 

	  (let ((dl (bminus el su))
			(du  (bminus eu sl)))

;;;		(format t "~2%  Event ~a bounds [~a, ~a]" end el eu)
;;;		(format t "~%  Event ~a bounds [~a, ~a]" start sl su)
;;;		(format t "~%  Derived bounds (~a, ~a)[~a, ~a]" 
;;;				start end (bminus el su) (bminus eu sl))

		(values dl du)))))

;;; Check:
;;; o Is the following function needed?
;;;(defmethod event-consequences ((event itc-event) (constraint-system ITC-S))
;;;
;;;  "Returns all simple temporal constraints that helped compute
;;;   the current bounds of some event, based on the bounds on EVENT."
;;;
;;;  ;; First make sure that propagation is up to date.
;;;  (propagate! constraint-system)
;;;
;;;  (event-consequences event))

;;; ***********************************************************************
;;;
;;;                            ITC Support functions
;;;
;;; ***********************************************************************


;;;   *** Initializing a Simple Temporal Network

;;; Reviewed by BCW.
(defmethod STN-wide-initialize! ((stn itc-stn))

  "Let STN be a simple temporal network that supports incremental
   temporal consistency.

   Initialize properties of STN that are constraint-system-wide,
   and support propagation.  Doesn't change or initialize
   the events or constraints contained in STN.

   Note: To check consistency of the STN the first time, 
         it is sufficient to propagate from the start of the STN alone."

  ;; Clears the three queues used for checking STN consistency
  ;; and bounds:  a) the queue of event pairs whose edges have changed,
  ;; b) the queue of upper bounds to be propagated, and c) the queue
  ;; of lower bounds to be propagated.
  (clear-event-pair-queue! stn)
  (clear-queue! (stn-lb-queue stn))
  (clear-queue! (stn-ub-queue stn))

  ;; Record that consistency has not yet been checked.
  ;; Clear related slots (inconsistent? and conflicts).
  (itc-stn-lb-not-propagated! stn)
  (itc-stn-ub-consistency-not-checked! stn)

  T)


;;; Reviewed by BCW.
(defmethod clear-event-pair-queue! ((STN itc-stn))

  "Let STN be a simple temporal network that supports
   incremental temporal consistency checking.

   Removes all event pairs from STN's event-pair queue."
  
  (loop with q = (stn-event-pair-queue STN)

	  while (has-event-pairs-queued? stn)

	  for event-pair = (queue-pop q)
	  do (event-pair-dequeued! event-pair)))


;;;   *** Initializing a Start Event

;;;{05} d(sstart) := 0;
;;;{05a} v(sstart) := false;
;;;{new} s(sstart) := labeled;
;;;NEW: Each node now has a status of {unreached, labeled, scanned}.
;;;     It is a variation of label correcting operation.
;;;     The start event will be given "labeled" status when initialized

;;; Reviewed by BCW.
(defmethod start-initialize! ((event itc-event))
  
  "Let EVENT be the start event of its simple temporal network.

   Initialize testing of consistency and bound computation
   for start EVENT."

  ;; Initialize lower bound propagation.
  (start-lb-initialize! event)

  ;; Initialize upper bound propagation.
  (start-ub-initialize! event)
  
  ;; Initialize event's status 
  ;; (*labeled*, *scanned*, or *unreached*).
  (start-status-initialize! event)
  
  T)

;;;{05} d(sstart) := 0;

(defmethod start-lb-initialize! ((start itc-event))

  "Initialize testing of consistency and lower bound 
   computation for start EVENT."

  (setf (event-lb start) 0)
  (clear-event-lb-support! start)
  (event-lb-visited! start)
  T)

;;;{05a} v(sstart) := false;

;;; Check:
;;;   o Check all callers to see if
;;;;   (start-status-initialize! event) should be called as well.
(defmethod start-ub-initialize! ((start itc-event))

  "Initialize testing of consistency and upper bound 
   computation for start EVENT."

  (setf (event-ub start) 0)
  (clear-event-ub-support! start)
  (event-ub-visited! start)
  T)

;;;{new} s(sstart) := labeled;
;;;NEW: Each node now has a status of {unreached, labeled, scanned}.
;;;     It is a variation of label correcting operation.
;;;     The start event will be given "labeled" status when initialized

;;; Check:
;;; o Do we maintain a status only for UB propagation, or LB as well?
(defmethod start-status-initialize! ((start itc-event))

  "Initialize testing of consistency and upper bound 
   computation for start EVENT."

  (setf (event-status start) *labeled*)
  T)


;;;   *** Initializing an Event ***

;;;{03}   d(s) := infinity;
;;;{04}   p(s) := unknown;
;;;{04a}  v(s) := false;
;;;{new}  s(s) := unreached;
;;;NEW: Each node now has a status of {unreached, labeled, scanned}.
;;;     It is a variation of label correcting operation.
;;;     The other events will be given "unreached" status when initialized
(defmethod event-initialize! ((event itc-event))

  "Initialize testing of consistency and bound computation
   for non-start EVENT."
  
  (event-lb-initialize! event)
  (event-ub-initialize! event)
  (event-status-initialize! event)
  T)

;;;{03'}   d(s) := -infinity;
;;;{04'}   p(s) := unknown;
;;;{04a'}  v(s) := false;
;;;{new}  s(s) := unreached;
(defmethod event-lb-initialize! ((event itc-event))

  "Initialize testing of consistency and lower bound 
   computation for non-start EVENT."

  (setf (event-lb event) 
	*-infinity*)
  (clear-event-lb-support! event)
  (event-not-lb-visited! event)
  T)
	  
;;;{03}   d(s) := infinity;
;;;{04}   p(s) := unknown;
;;;{04a}  v(s) := false;
;;;{new}  s(s) := unreached;
(defmethod event-ub-initialize! ((event itc-event))

  "Initialize testing of consistency and upper bound 
   computation for non-start EVENT."

  (setf (event-ub event) 
	*infinity*)
  (clear-event-ub-support! event)
  (event-not-ub-visited! event)
  T)

;;;{03}   d(s) := infinity;
;;;{04}   p(s) := unknown;
;;;{04a}  v(s) := false;
;;;{new}  s(s) := unreached;
;;;NEW: Each node now has a status of {unreached, labeled, scanned}.
;;;     It is a variation of label correcting operation.
;;;     The other events will be given "unreached" status when initialized
(defmethod event-status-initialize! ((event itc-event))

  "Initialize testing of consistency and upper bound 
   computation for non-start EVENT."

  (setf (event-status event) *unreached*)
  T)


;;;   *** Initializing a Constraint ***

(defmethod STC-initialize! ((stc itc-stc))

  "Initialize properties of simple temporal constraint STC, 
   used for testing consistency and oomputing bounds.

   Leaves inserted/Removed status of constraint unchanged."

  ;; Initialize the two d-edges representing constraint.
  (dolist (edge (stc-d-edges stc))
	(d-edge-initialize! edge))
  T)


;;;   *** Initializing a Distance Edge ***

(defmethod d-edge-initialize! ((edge itc-d-edge))

  "Initialize distance EDGE's properties, used to compute
   consistency and bounds on events.  
   Leaves edge inserted/removed status unchanged."

  ;;; EDGE has no consequences
  (clear-d-edge-lb-consequent! edge)
  (clear-d-edge-ub-consequent! edge)
  
  ;;; and does not participate in an inconsistency.
;;;  (setf (d-edge-in-conflict? edge) nil)

  T)

;;;   ***
;;;   *** Inserting Distance Edges ***
;;;   ***

(defmethod insert-d-edge! ((edge itc-d-edge))
  
  "Inserts removed distance EDGE into its constraint-system.
   This function updates the queues of edge's constraint-system, 
   used to determine event bounds and consistency, when queried.
   Propagation is not initiated by this function.
   Note: edge does not need to be queued unless consistency has 
   been previously checked.

   If EDGE is already inserted, this function does nothing.  

   The function is called for effect, it does not return anything 
   meaningful."

  (debug-print-for :itc t "    Inserting d-edge ~A." edge)
  
  (cond ((not (d-edge-removed? edge))
		 ;; Edge is already in constraint-system, 
		 ;; nothing needs to be done.
		 (debug-print-for :itc t "      ~a already inserted, do nothing."
						  edge)
		 nil)

		(:otherwise

		 ;; Insert d-edge.
		 (setf (d-edge-removed? edge) nil)

		 ;; Queue edge in order to propagate changes, upon query.
		 (queue-changed-d-edge! edge)
		 T)))


;;;   ***
;;;   *** Removing Distance Edges ***
;;;   ***

(defmethod remove-d-edge! ((edge itc-d-edge))
  
  "Removes distance EDGE from its constraint-system.  Sets up the queues
   to find alternative support for removed consequences and for
   propagating changes.
   Propagate! IS NOT called by this function, but on demand.
   Note: edge is not queued unless consistency has been
   previously checked.

   If EDGE is already removed, this function does nothing.

   The function is called for effect, it does not return 
   anything meaningful."

  (debug-print-for :itc t "    Removing d-edge ~A." edge)
  (cond ((d-edge-removed? edge)

		 ;; Edge is already removed from constraint-system, 
		 ;; nothing needs to be done.
		 (debug-print-for :itc t "      ~a already removed, do nothing."
						  edge)
		 nil)

		(:otherwise

		 ;; Set the removed status of EDGE.
		 (setf (d-edge-removed? edge) t)

		 ;; Queue edge to propagate change, upon query.
		 (queue-changed-d-edge! edge)
		 T)))


;;;   ***
;;;   ***   State Variable Assignment
;;;   ***

(defmethod check-state-variable-assignment? 
	((STN ITC-STN) state-variable-name value)
  
  "Returns True iff the assignment of VALUE to variable with name
   STATE-VARIABLE-NAME is a legal assignment for STN.  That is, 
   STATE-VARIABLE-NAME corresponds to a defined state variable of STN and 
   value is an element of its domain."
  
  (let ((state-assignment? T)
		(state-variable (find-state-variable STN state-variable-name)))
	
    (when (null state-variable)
      (setq state-assignment? nil)
      (warn "Variable ~A, being assigned ~A, is not a defined state ~
                variable for STN ~A" state-variable-name value STN))

    (unless (value-in-domain? state-variable value)
	  
      (setq state-assignment? nil)
      (warn "Value ~A, appearing in an assignment to state variable ~A, ~
              is not a member of ~A 's domain, for STN ~A." 
			value state-variable state-variable STN))
	state-assignment?))


;;;   ***
;;;   ***   Event-pair Edges
;;;   ***

(defmethod event-pair-lowest-bound-inserted-d-edge ((event-pair itc-event-pair))

  "Returns the inserted edge of EVENT-PAIR
   that has the least upper bound value.
   Returns nil if no such d-edge is inserted."

  ;; The following exploits the fact that an event-pairs
  ;; distance edges are listed in increasing value of upperbound.
  (loop for d-edge in (event-pair-increasing-d-edges event-pair)
	  unless (d-edge-removed? d-edge)
	  do (return d-edge)
	  finally (return nil)))

(defmethod event-pair-highest-bound-inserted-d-edge 
	((event-pair itc-event-pair))

  "Returns the inserted edge of EVENT-PAIR
   that has the least upper bound value.
   Returns nil if no such d-edge is inserted."

  ;; The following exploits the fact that an event-pairs
  ;; distance edges are listed in decreasing value of upperbound.
  (loop for d-edge in (event-pair-decreasing-d-edges event-pair)
	  unless (d-edge-removed? d-edge)
	  do (return d-edge)
      finally (return nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;           Backup and Restoration of the Current State of an STN
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Author:   
;;;   Peng Yu

;;; Style Revision: 
;;;   Brian Williams

;;; To Do:  Peng, please do the following - BCW.
;;; o Convert to methods and classes.

;;; o Provide complete and consistent documentation strings for all functions, 
;;;   variables and parameters.
;;;    - Document Function performed, Inputs, Outputs, Assumptions, Side Effects.

;;; o Provide consistent naming 
;;;   - Imperative procedures without meaninful results should end with "!"
;;;   - Tests and Flags should end with "?".
;;;   - Underscore _ should be changed to dash -.

;;; o Move to its own suitably named file, and document interface functions
;;;   at the top of the file.

(defvar d-edge_backup nil "Matrix storing backed up input d-edge information.")


(defun d-edge_backup_parameters (start end ub lb removed)
  
  "Add single edge information to a temporary list."
  
  (list :start start :end end :ub ub :lb lb :removed removed))


(defun add-record (d-edge_to_backup)
	
  "Adds the list of single edge information to the matrix d-edge_backup."
	
  (push d-edge_to_backup d-edge_backup))


(defun where (&key start end ub)
  
  "A search function to locate certain edge stored in the matrix."
  
  #'(lambda (d-edge_to_backup)
      (and
       (if start    (equal (getf d-edge_to_backup :start)  start)  t)
       (if end   (equal (getf d-edge_to_backup :end) end) t)
       (if ub   (equal (getf d-edge_to_backup :ub) ub) t)
       )))


(defun select (selector-fn)
  
  "Extracts the edge found by the WHERE function."
  
  (remove-if-not selector-fn d-edge_backup))


(defun delete-rows (selector-fn)
  
  "Remove duplicated edges in the matrix."
  
  (setf d-edge_backup (remove-if selector-fn d-edge_backup)))


;;;  Parameters  used to capture current edge information.

;;; Add documentation strings for all three parameters - BCW
(DEFPARAMETER lowerbound 0)
(DEFPARAMETER upperbound 0)

;;; Removed is a flag and should end with ? - BCW
(DEFPARAMETER removed nil)

(defun return_lb (input_db)
  
  "Return the lower bound of a certain edge.
   Here input_db is a temporary list which contains one edge only"
  
  (dolist (constraint_list input_db)
        
    (setf lowerbound (getf constraint_list :lb))
    )
  
  (return-from return_lb lowerbound)
  )


(defun return_ub (input_db)
 
  "Return the upper bound of a certain edge.
   Here input_db is a temporary list which contains one edge only."

  (dolist (constraint_list input_db)
       
    (setf upperbound (- (getf constraint_list :ub) 1000000))
    )
  
  (return-from return_ub upperbound)
  )

(defun return_removed (input_db)

  "Return the Removed? parameter of a certain edge.
   Here input_db is a temporary list which contains one edge only."

  (dolist (constraint_list input_db)
        
    (setf removed (getf constraint_list :removed))
    )
  
  (return-from return_removed removed)
  )


(defun d-edge-print (events-in-stn)

  "Print d-edges in current STN."
  
  (dolist (event-printing events-in-stn)

;;;	(format t "~%[Backup] event: ~a ub: ~a lb: ~a status: ~a~%"
;;;			event-printing (event-lb event-printing)
;;;			(event-ub event-printing) (event-status event-printing))

	(dolist (d-edge (event-successors event-printing))


	  (when (not (d-edge-removed? d-edge))
	  (format t "Edge: ~a  ub: ~a ~%"
			  d-edge (d-edge-ub d-edge))

		)
	  )
	)
  )

(defun d-edge-backup (events-in-stn)

  "Backup d-edges in current STN."

  (dolist (row-in-backup d-edge_backup)
    (delete-rows (where :start (getf row-in-backup :start)
						:end (getf row-in-backup :end)
						:ub (getf row-in-backup :ub))))
  
  (dolist (event-printing events-in-stn)

;;;	(format t "~%[Backup] event: ~a ub: ~a lb: ~a status: ~a~%"
;;;			event-printing (event-lb event-printing)
;;;			(event-ub event-printing) (event-status event-printing))

	(dolist (d-edge (event-successors event-printing))

	  (format t "[Edge backup] event-successors edges: ~a  ~
                removed?: ~a ub: ~a lb?: ~a  removed?: ~a~%"
			  d-edge (d-edge-removed? d-edge) (d-edge-ub d-edge)
			  (d-edge-lower-bound? d-edge) (d-edge-removed? d-edge))
        
	  (when (and (not (eql (d-edge-ub d-edge) *infinity*))
				 (not (eql (d-edge-ub d-edge) *-infinity*)))
		  
		(add-record (d-edge_backup_parameters (d-edge-start d-edge)
											  (d-edge-end d-edge)
											  (+ 1000000 (d-edge-ub d-edge))
											  (d-edge-lower-bound? d-edge)
											  (d-edge-removed? d-edge)))
		)
	  )
	)
  )



(defun d-edge-recovery (events-in-stn)

  "Restore d-edges of STN to the backed up state."

  (dolist (event-printing events-in-stn)
	
;;;	(format t "~% d-edge_backup is : ~a ~%" d-edge_backup)
;;;	(format t "~% event: ~a ub: ~a lb: ~a status: ~a   ~%"
;;;			event-printing (event-lb event-printing)
;;;			(event-ub event-printing) (event-status event-printing))

    (dolist (d-edge (event-successors event-printing))

      (when (and (not (eql (d-edge-ub d-edge) *infinity*))
				 (not (eql (d-edge-ub d-edge) *-infinity*)))
        
        (when (> (d-edge-ub d-edge) 10000)
		  
		  (setf (d-edge-removed? d-edge)
			(return_removed (select (where :start (d-edge-start d-edge)
										   :end (d-edge-end d-edge)
										   :ub (d-edge-ub d-edge)))))
		  (setf (d-edge-ub d-edge)
			(return_ub (select (where :start (d-edge-start d-edge)
									  :end (d-edge-end d-edge)
									  :ub (d-edge-ub d-edge)))))
          )   
        )
	  
;;;	  (format t "[Edge Recovered] event-successors edges: ~a ~
;;;                 removed?: ~a ub: ~a lb?: ~a  removed?: ~a~%"
;;;			  d-edge (d-edge-removed? d-edge) (d-edge-ub d-edge)
;;;			  (d-edge-lower-bound? d-edge) (d-edge-removed? d-edge))
	  )
	)
  )

;;; FIX:
;;; o I don't understand what this is doing - Brian Williams.
;;;   It needs a clearer explanation and a careful review.
;;;   We shouldn't be approximating infinity with a large number.
;;;   This doesn't seem like a correct solution.

;;; To Do:
;;; o This function doesn't appear to be in the correct file conceptually.
;;;   It it remains, move it.

(defun d-edge-relaxation (events-in-stn)

  "Relax d-edges in current STN.
   Previously, relaxation is performed by switching Removed? mark to T.

   This function catches d-edge with T mark, relax their ub and lb to infinity, 
   and changes Removed? to NIL."
  
  (dolist (event-printing events-in-stn)

;;;	(format t "~%[Relaxation] event: ~a ub: ~a lb: ~a status: ~a~%"
;;;			event-printing (event-lb event-printing)
;;;			(event-ub event-printing) (event-status event-printing))

	(dolist (d-edge (event-successors event-printing))

;;;	  (format t "[edge-pre-processed] event-successors edges: ~a  ~
;;;                 removed?: ~a ub: ~a lb?: ~a  removed?: ~a~%"
;;;			  d-edge (d-edge-removed? d-edge) (d-edge-ub d-edge)
;;;			  (d-edge-lower-bound? d-edge) (d-edge-removed? d-edge))
	  
	  (when (d-edge-removed? d-edge)
          
		(setf (d-edge-removed? d-edge) nil)
		
		(when (and (not (eql (d-edge-ub d-edge) *infinity*))
				   (not (eql (d-edge-ub d-edge) *-infinity*)))
		  
		  (when (d-edge-lower-bound? d-edge)
			
			(setf (d-edge-ub d-edge)
			  (+ 1000000 (d-edge-ub d-edge))))
          
		  (when (not (d-edge-lower-bound? d-edge))
			
			(setf (d-edge-ub d-edge)
			  (+ 1000000 (d-edge-ub d-edge)))
              
			)
		  )
		)
          
;;;	  (format t "[edge-modifed      ] event-successors edges: ~a  ~
;;;                 removed?: ~a ub: ~a lb?: ~a  removed?: ~a~%"
;;;			  d-edge (d-edge-removed? d-edge) (d-edge-ub d-edge) 
;;;			  (d-edge-lower-bound? d-edge) (d-edge-removed? d-edge))
	  )
	)
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;
;; Functions for modifying ITC constraints
;;
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(defun modify-constraint! (itc-constraint l u STN)

  "Change the bounds [l,u] of simple temporal 
    constraint between x and y."

  (setf (stc-lb itc-constraint) l)
  (setf (stc-ub itc-constraint) u)

  (loop for d-edge in (stc-d-edges itc-constraint) do

	   (if (d-edge-lower-bound? d-edge)
		   (STN-modify-arc! d-edge (b- l) STN)
		   (STN-modify-arc! d-edge u STN))
	   ))

(defun STN-modify-arc! (d-edge ub STN)

  " Change to ub the edge weight of arc,
   and queue relevant vertices for update."

  (setf (d-edge-ub d-edge) ub)
  (setf (d-edge-initial-ub d-edge) ub)
 

  (when (STN-ub-consistency-checked? STN)
	;; Only perform incremental update if
	;; consistency already checked at least once.

	(let ((x (d-edge-start d-edge))
		  (y (d-edge-end d-edge)))

	  (let* ((x-ub (event-ub x))
			 (y-ub (event-ub y))
			 (r-y-ub (b+ x-ub ub)))

		;; (format t "x-ub: ~A; y-ub: ~A; r-y-ub: ~A ~%" x-ub y-ub r-y-ub)

		;; (cond ((b> y-ub r-y-ub)
		;; 	   ;; Change improves shortest path
		;; 	   ;; Update and propagate.

		;; 	   (format t "Improving ~A-~A:~A~%" x y ub)

		;; 	   (setf (event-ub y) r-y-ub)
		;; 	   (setf (event-ub-support y) x)
		;; 	   (queue-insert! y (STN-ub-queue STN)))

		;; 	  ((and (b> r-y-u4444b y-ub)
		;; 			(eql (event-ub-support y)
		;; 				 x))
		;; 	   ;; Change worsens shortest path,
		;; 	   ;; Reset value.
		;; 	   (format t "Worsening ~A-~A:~A~%" x y ub)

		;; 	   (event-ub-initialize! y)

		;; 	   ;; Reset consequences.
		;; 	   (let ((nodes-reset '()))
		;; 		 (push y nodes-reset)
		;; 		 (loop for n in (invalidate-support y STN)
		;; 			do (push n nodes-reset))

		;; 		 ;; Setup to resupport.
		;; 		 (insert-parents! nodes-reset STN))
		;; 	   ))


		(let ((event-pair (find-event-pair x y STN)))
		  
		  (when (null event-pair)
			
			(error "An event pair object doesn't exist for d-edge ~a"
				   d-edge))
		  
		  (queue-changed-event-pair! event-pair STN))

))))


(defun invalidate-support (n STN) 

  "Remove n as a support for any s,
   and queue s's predecessors for resupport.
   Returns list of nodes whose d and p values
   have been reset."

  (loop with nodes-reset = '() 
	  for s in (event-successors n)
	  when (eql (event-ub-support s) 
				n)
		   ;; Successor s is supported by n
		   ;; since n reset, reset s as well.

	  do (cond ((eql s (STN-start STN))
				(start-ub-initialize! s)
				(loop for u in (invalidate-support s STN)
					do (push u nodes-reset)))
			   ((not (event-ub-useless? s))
				(event-ub-initialize! s)
				(loop for u in (invalidate-support s STN)
					do (push u nodes-reset)))
			   (:otherwise nil))
	  finally (return nodes-reset)))


(defun insert-parents! (reset-nodes STN) 

  "Rest_nodes lost their support, queue parents
   to resupport."
  (loop for n in reset-nodes
	  do (loop for m in (event-predecessors n)
			 unless (event-ub-useless? m)
			 do (queue-insert! m (STN-ub-queue STN)))))

(defmacro stn-arc-cost (x y STN)
  `(gethash (list ,x ,y)
			(STN-arc-cost-ht ,STN)))
