;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2013-2015 James G. Paterson, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; James G. Paterson.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

(defun run-ltnp-test-1 () 
  "Main function for testing the algorithm and to use as an example of how to use LTNP code."
  (let ((ltnp (make-instance 'LTNP :name  'test17 :start-name 'e_start)))   
	(insert-new-constraint ltnp :source 'test17 :start 'e_start :end 'ea :lb 0 :ub 0)
	(insert-new-constraint ltnp :source 'test17 :start 'ea :end 'eb :lb 0 :ub 0)
	(insert-new-looping-constraint ltnp :source 'test17 :start 'eb :end 'ec :t-lb 30 :t-ub 40 :N-lb 5 :N-ub 10 
								   :reward-function (make-instance 'linear-looping-episode-cost :setup-cost -5 :cost-per-loop -2))

	(insert-new-looping-constraint ltnp :source 'test17 :start 'ed :end 'ee :t-lb 30 :t-ub 40 :N-lb 5 :N-ub 10 
								   :reward-function (make-instance 'linear-looping-episode-cost :setup-cost -5 :cost-per-loop -20))

	(insert-new-constraint ltnp :source 'test17 :start 'ec :end 'ed :lb 0 :ub 0)
	(insert-new-constraint ltnp :source 'test17 :start 'ee :end 'e_end :lb 0 :ub 0)
	(insert-new-constraint ltnp :source 'test17 :start 'e_start :end 'e_end :lb 0 :ub 460)
	
	(solve-ltnp! ltnp)

	(print-stn! ltnp)))


(defun solve-ltnp! (ltnp) 
  "Find best assignment to each looping constraint. First, we prune all loops that can never 
   be consistent. Then we search through the remaining space to find the optimal solution 
   based on the reward functions given."

  ;;(print "Solving Looping Temporal Problem w Preference (LTPP).")

  ;; Prune all loops that can never be consistent   
  (best-integer-assignemnt-to-loops? ltnp)
 ;;

 ;; (when (initial-loop-prune! ltnp)
 ;;  	;; 	;; Then do search for best assignment.
 ;;    (best-integer-assignemnt-to-loops? ltnp)
 ;;  	;; 	;; Run apsp to get updated event bounds and relaxed constraints.
 ;;  	;;(apsp-floyd-warshall! ltnp)
 ;;  	;; 	;;(best-integer-assignemnt-to-loops?-prune-version ltnp)
 ;;   	;;ltnp
 ;; 	)
   ;; 
 )

(defmethod insert-new-looping-constraint
	((constraint-system LTNP) &key source start end t-lb t-ub N-lb N-ub reward-function)

  "Creates a new LTCP whose source is SOURCE, AND inserts it into 
   CONSTRAINT-SYSTEM.  START and END are names of events, and t-lb and t-ub are reals,
   where t-lb<=t-ub and are the temporal bound per loop. N-lb and N-ub are positive integers,
   where N-lb<=N-ub and are the looping bounds.

   The function returns the newly created 
   constraint."

  (let ((ltcp 
		 (make-instance 
		  'LTCP 
		  :source source 
		  :constraint-system constraint-system 
		  :start start 
		  :end end 
		  :lb (* t-lb N-lb) 
		  :ub (* t-ub N-ub)
		  :N-lb N-lb 
		  :N-ub N-ub 
		  :t-lb t-lb 
		  :t-ub t-ub :reward-function reward-function)))
	(insert-constraint constraint-system ltcp)
	(values ltcp)))

(defmethod update-stn-constraints! ((stn stn))
  "Updates the stn constraints (relaxed looping constraints) if the distance edges have changed."  
  (loop for e in (stn-events stn) do 
	   (loop for d in (event-successors e) do
			(when (not (eql nil (d-edge-constraint d)))
			  (if (d-edge-lower-bound? d)
				  (setf (stc-lb (d-edge-constraint d)) (abs(d-edge-ub d)))					
				  (setf (stc-ub (d-edge-constraint d)) (d-edge-ub d))
				  )))))


(defmethod prune-loops-based-on-relaxed-constraints! ((ltnp ltnp))
  "Prunes infeasible loops from a LTNP based on the relaxed temporal constraints.
  Returns true if there may be a feasible solution and nil if there is no feasible solution."  
  (loop for constraint in (ltnp-looping-constraints ltnp) do 
	   (setf (N-ub constraint) (min (N-ub constraint) (floor(/ (stc-ub constraint) (t-lb constraint)))))	   
	   (setf (N-lb constraint) (max (N-lb constraint) (ceiling(/ (stc-lb constraint) (t-ub constraint)))))
	   (when (> (N-lb constraint) (N-ub constraint))
		 (return-from prune-loops-based-on-relaxed-constraints! nil)))
  t)

(defmethod update-relaxed-constraints-and-d-edges! ((ltnp ltnp))
  "Updates the relaxed contraints with product of loop number and temporal bound. Also updates the distance edges."  
  (loop for constraint in (ltnp-looping-constraints ltnp) do 
	   (setf (stc-ub constraint) (* (N-ub constraint) (t-ub constraint)))	   
	   (setf (stc-lb constraint) (* (N-lb constraint) (t-lb constraint)))	   
	   (loop for d-edge in (stc-d-edges constraint) do
			(if (d-edge-lower-bound? d-edge)
				(setf (d-edge-ub d-edge) (- (stc-lb constraint)))
				(setf (d-edge-ub d-edge) (stc-ub constraint) )))))


(defun best-integer-assignemnt-to-loops? (ltnp) 
  "Searches through the ltnp to find the best integer loop assignment to each looping constraint.
  Returns the ltnp with integer assignments to loops."

  "DESTRUCTIVELY modifies ltnp."
  ;; Start search timer.
  ;; Can also use run-time...
  (let ((start-time (get-internal-real-time)))
	(let ((search-tree (initialize-ltnp-search-tree ltnp)))
	  (loop while (not (adt:queue-empty? (search-q search-tree))) do
		    (let ((branch (pop-branch search-tree)))
			 (if (not(consistent-branch? branch ltnp))   ;;(consistent-branch?-and-prune branch ltnp)
				 ;; if not consistent, prune branch (already done by pop) and reset ITC d-edges	 
				 (progn 	 
				   (reset-d-edges! search-tree ltnp))
				 ;; else
				 (progn
				   (when (branch-has-integer-assignment? branch ltnp) ;; We have found the best solution
					 (branch-to-ltnp! branch ltnp)
					 (format t "~%SEARCH TIME: ~a (ms)" (- (get-internal-real-time) start-time))
					 (return-from best-integer-assignemnt-to-loops? ltnp))
				   ;; Two options here: split on reward, or split on interval
				   ;;(split-branch-based-on-reward! branch ltnp)
				   (let ((two-branches (multiple-value-bind (a b) (split-branch-based-on-interval! branch ltnp) (list a b))))
					 (add-branches-to-tree two-branches search-tree)				   ))))
		   ;;(format t "~%QUEUE_SIZE at end of loop: ~a " (coll-size (search-q search-tree)))
		   )))   
  nil)



;;; The function below is a work in progress. It will run apsp to prune loops each time after splitting a loop bound.
(defun best-integer-assignemnt-to-loops?-prune-version (ltnp) 
  "Searches through the ltnp to find the best integer loop assignment to each looping constraint.
  Returns the ltnp with integer assignments to loops."
  ;; Start search timer.
  ;; Can also use run-time...
  (let ((start-time (get-internal-real-time)))
	(let ((search-tree (initialize-ltnp-search-tree ltnp)))
	  (loop while (not (adt:queue-empty? (search-q search-tree))) do
		   (format t "~%NEW LOOP ===>")
		   (let ((branch (pop-branch search-tree)))
			 (if (not(consistent-branch?-and-prune branch ltnp))   ;;(consistent-branch?-and-prune branch ltnp)
				 ;; if not consistent, prune branch (already done by pop) and reset ITC d-edges	 
				 (progn 
				   (print "RESETTING D-EDGES")
				   (print "printing loop ranges")
				   (loop for value being the hash-values of (integer-to-loop-ht branch) do 
						(print value))
				   ;;(reset-d-edges! search-tree ltnp)
				   )
				 ;; else
				 (progn
				   (when (branch-has-integer-assignment? branch ltnp)
					 (print (reward-ub branch))
					 (print (reward-ub (queue-peek(search-q search-tree))))
			 
					 ;;(when reward baranch >= reward of next branch on the queue);; We have found the best solution
					 (when (>= (reward-ub branch) (reward-ub (queue-peek(search-q search-tree))))
					   ;; (print (reward-ub branch))
					   ;; (print (reward-ub (queue-peek(search-q search-tree))))
					   (branch-to-ltnp! branch ltnp)
					   (format t "~%SEARCH TIME: ~a (ms)" (- (get-internal-real-time) start-time))
					   (return-from  best-integer-assignemnt-to-loops?-prune-version (branch-to-ltnp! branch ltnp))))
				   ;; Two options here: split on reward, or split on interval
				   ;;(split-branch-based-on-reward! branch ltnp)
				   (if (branch-has-integer-assignment? branch ltnp)
					   (add-branches-to-tree (list branch) search-tree)
					 (let ((two-branches (multiple-value-bind (a b) (split-branch-based-on-interval! branch ltnp) (list a b))))
					   (add-branches-to-tree two-branches search-tree)				   
					   )))))

		   (format t "~%QUEUE_SIZE at end of loop: ~a " (coll-size (search-q search-tree)))
		   )))   
  nil)


(defun branch-to-ltnp! (branch ltnp)
  "Writes loop-ranges of a search branch to the ltnp and updates the relaxed temporal bounds and the d-edges."
  (loop for key being the hash-keys of (integer-to-loop-ht branch) 
	 using (hash-value loop-range) do 
	   (let ((looping-constraint (gethash key (ltnp-looping-constraints-ht ltnp))))
		 (setf (N-lb looping-constraint) (car loop-range))
		 (setf (N-ub looping-constraint) (cdr loop-range))
		 ))
  (update-relaxed-constraints-and-d-edges! ltnp)
  ltnp)

(defun ltnp-to-branch! (branch ltnp)
  "Writes loop-ranges of a ltnp to the branch and updates the reward."
  (let ( (i 0))
	(loop for constraint in (ltnp-looping-constraints ltnp) do
		 (setf (gethash i (integer-to-loop-ht branch)) (cons (N-lb constraint) (N-ub constraint)))	
		 (setf (gethash i (ltnp-looping-constraints-ht ltnp)) constraint)
		 (incf i))
	(setf (reward-ub branch) (cdr (branch-reward? branch ltnp)))) 
   branch)



(defun initial-loop-prune! (ltnp) 
  "Prune loops that can never be consistent using APSP" 
  (apsp-floyd-warshall! ltnp) ;; Test consistency using APSP and update paths.
  (when (prune-loops-based-on-relaxed-constraints! ltnp) ;; Returns false if not consistent after pruning.
	(update-relaxed-constraints-and-d-edges! ltnp)
	(when(consistent-stn? ltnp :initialize? t)
	  (return-from initial-loop-prune! t)))
  nil)


(defun initialize-ltnp-search-tree (ltnp) 
  "Return a tree object containing a list of Branches (search nodes) that in turn contain
   a subset of the looping ranges of the original LTNP."
  (print "Searching for best assignement to Loops.")
  (let ((start-branch (make-instance 'Search-Branch)) (i 0)(search-tree (make-instance 'Search-Tree)))
	(loop for constraint in (ltnp-looping-constraints ltnp) do
		 (setf (gethash i (integer-to-loop-ht start-branch)) (cons (N-lb constraint) (N-ub constraint)))	
		 (setf (gethash i (ltnp-looping-constraints-ht ltnp)) constraint)
		 (incf i))
	(setf (reward-ub start-branch) (cdr (branch-reward? start-branch ltnp)))
	(setf (branch-parent start-branch) nil)
	(setf (start-branch search-tree) start-branch)
	(enqueue! (search-q search-tree) start-branch)
	search-tree))

(defun pop-branch (tree) 
  "Returns the first (and should be sorted to be the best) branch on the
   search queue."
  (queue-remove! (search-q tree)))

(defun consistent-branch? (branch ltnp) 
  "Check if branch is consistent." 
  (when (eql (branch-parent branch) nil)
	(return-from consistent-branch? (consistent-stn? ltnp :initialize t)))  ;; Already initialized in pre-pruning  
  (let((num (constraint-split-number (branch-parent branch))))	
	(let ((constraint (gethash num (ltnp-looping-constraints-ht ltnp)))
		  (range (gethash num (integer-to-loop-ht branch))))

	  ;; TODO: More efficient would me to just modify the d-edge and add to queue
	  (modify-constraint! constraint (* (car range) (t-lb constraint)) (* (cdr range) (t-ub constraint)) ltnp)
	  (consistent-stn? ltnp :initialize? nil)))
  )

(defun consistent-branch?-and-prune (branch ltnp) 
  "Check if branch is consistent and prune ranges using APSP. WORK IN PROGRESS!"
  (branch-to-ltnp! branch ltnp)
  (when (initial-loop-prune! ltnp)
	(ltnp-to-branch! branch ltnp)	
	(return-from consistent-branch?-and-prune t))
  nil)




(defun branch-has-integer-assignment? (branch ltnp) 
  "Returns true if all active looping constraints have been assigned an integer value."

  (loop for key being the hash-keys of (integer-to-loop-ht branch) 
	 using (hash-value value)
	 do 
	   (when (stc-active? (gethash key (ltnp-looping-constraints-ht ltnp)))
		 (when (not (eql(car value)(cdr value)))
		   (return-from branch-has-integer-assignment? nil))))
  t)

(defun split-branch-based-on-reward! (branch ltnp) 
  "Splits a branch into two by splitting the loop interval of the 
  branch that has the highest reward interval? Or just upper reward bound? Or loop interval?
  Highest upper bound I think, as this has the potential of adding most to the reward..."
  ;; Find number (key) to split on:
  (let ((reward-cons (cons 0 0)))
	(loop for key being the hash-keys of (ltnp-looping-constraints-ht ltnp) 
	   using (hash-value loop-constraint) do 
		 (let ((reward-range (constraint-reward-range? loop-constraint (car(gethash key (integer-to-loop-ht branch))) (cdr(gethash key (integer-to-loop-ht branch)))))) 	   
		   (when (not (eql (car reward-range) (cdr reward-range))) ;; When we don't have an integer assignment
			 (when  (> (cdr reward-range) (cdr reward-cons))
			   (setf reward-cons (cons key (cdr reward-range)))))))
	;; Make two new branches
	(let ((best-range (gethash (car reward-cons) (integer-to-loop-ht branch)))
		  (left-branch (make-instance 'Search-Branch :integer-to-loop-ht (copy-hash-table (integer-to-loop-ht branch))))
		  (right-branch (make-instance 'Search-Branch :integer-to-loop-ht (copy-hash-table (integer-to-loop-ht branch)))))
	  ;; Split the loop range in half
	  (let ((middle (+ (car best-range) (/(- (cdr best-range) (car best-range)) 2))))
		(setf (gethash (car reward-cons) (integer-to-loop-ht left-branch)) (cons (car best-range) (floor middle)))
		(setf (gethash (car reward-cons) (integer-to-loop-ht right-branch)) (cons  (ceiling middle) (cdr best-range)))) 
	  ;; Set the parent
	  (setf (branch-parent left-branch) branch)
	  (setf (branch-parent right-branch) branch)
	  ;; Set the constraint split number (key) of the parent, used for restoring data during backtracking
	  (setf (constraint-split-number branch) (car reward-cons))  	  
	  ;; Set rewards of both branches
	  (setf (reward-ub left-branch) (cdr (branch-reward? left-branch ltnp)))
	  (setf (reward-ub right-branch) (cdr (branch-reward? right-branch ltnp)))
	  ;; Return the two branches in a list
	  (values left-branch right-branch))))

(defun split-branch-based-on-interval! (branch ltnp) 
  "Splits a branch into two by splitting the loop interval of the 
  branch that has the highest loop interval? Or just upper reward bound? Or reward interval?
  Loop interval this time."
  ;; Find number (key) to split on:
  (let ((loop-cons (cons 0 0)))    ;; loop-cons: key and loop interval
	(loop for key being the hash-keys of (ltnp-looping-constraints-ht ltnp) 
	   using (hash-value loop-constraint) do 
		 (when (stc-active? loop-constraint)
		   (let ((loop-interval (gethash key (integer-to-loop-ht branch)))) 
			 (when  (> (- (cdr loop-interval) (car loop-interval)) (cdr loop-cons))
			   (setf loop-cons (cons key (- (cdr loop-interval) (car loop-interval))))))))
	;; Make two new branches
	(let ((best-loop-range (gethash (car loop-cons) (integer-to-loop-ht branch)))
		  (left-branch (make-instance 'Search-Branch :integer-to-loop-ht (copy-hash-table (integer-to-loop-ht branch))))
		  (right-branch (make-instance 'Search-Branch :integer-to-loop-ht (copy-hash-table (integer-to-loop-ht branch)))))
	  ;; Split the loop range in half
	  (let ((middle (+ (car best-loop-range) (/(- (cdr best-loop-range) (car best-loop-range)) 2))))
		(setf (gethash (car loop-cons) (integer-to-loop-ht left-branch)) (cons (car best-loop-range) (floor middle)))
		(setf (gethash (car loop-cons) (integer-to-loop-ht right-branch)) (cons  (ceiling middle) (cdr best-loop-range)))) 
	  ;; Set the parent
	  (setf (branch-parent left-branch) branch)
	  (setf (branch-parent right-branch) branch)
	  ;; Set the constraint split number (key) of the parent, used for restoring data during backtracking
	  (setf (constraint-split-number branch) (car loop-cons))  	  
	  ;; Set rewards of both branches
	  (setf (reward-ub left-branch) (cdr (branch-reward? left-branch ltnp)))
	  (setf (reward-ub right-branch) (cdr (branch-reward? right-branch ltnp)))
	  ;; Return the two branches in a list
	  (values left-branch right-branch))))

(defun branch-reward? (branch ltnp) 
  "Calculates total reward of branch."
  (let ((branch-reward (cons 0 0)))
	(loop for key being the hash-keys of (ltnp-looping-constraints-ht ltnp) 
	   using (hash-value loop-constraint) do 
		 (let ((reward-range (constraint-reward-range? loop-constraint (car(gethash key (integer-to-loop-ht branch))) 
													   (cdr(gethash key (integer-to-loop-ht branch)))))) 	   
		   (setf branch-reward (cons (+ (car branch-reward) (car reward-range)) (+ (cdr branch-reward) (cdr reward-range))))))  
	(return-from branch-reward? branch-reward)))

(defun add-branches-to-tree (branches tree) 
  "Adds branches to tree." 
  (loop for branch in branches do
	   (enqueue! (search-q tree) branch)))

(defun reset-d-edges! (search-tree ltnp)
  ;; Get the parent of the best branch
  (when (not (adt:queue-empty? (search-q search-tree)))
	(let ((parent (branch-parent (queue-peek (search-q search-tree)))))
	  (loop for key being the hash-keys of (integer-to-loop-ht parent) 
		 using (hash-value loop-range) do 
		 ;; Get the looping constraint
		   (let ((looping-constraint (gethash key (ltnp-looping-constraints-ht ltnp))))
			 (let ((l (* (car loop-range) (t-lb looping-constraint)))
				   (u (* (cdr loop-range) (t-ub looping-constraint))))
			   (setf (stc-lb looping-constraint) l) 
			   (setf (stc-ub looping-constraint) u)
			   (loop for d-edge in (stc-d-edges looping-constraint) do
					(if (d-edge-lower-bound? d-edge)
						(progn
						  (setf (d-edge-ub d-edge) (b- l))
						  (setf (d-edge-initial-ub d-edge) (b- l)))
						(progn
						  (setf (d-edge-ub d-edge) u)
						  (setf (d-edge-initial-ub d-edge) u))))))))))


;;; ***********************************************

;;; APSP Methods

;;; **********************************************

(defmethod apsp-floyd-warshall! ((stn itc-stn) &key (stn-mode t))
  "Modifies the input to calculate the all-pairs shortest paths, leaving in
the self-loops for another step to do consistency checking. Uses the standard
Floyd-Warshall algorithm. Uses an arbitrary constraint of the system to label
any edges added by this algorithm.

Stn-mode defaults to true since this is, after all, part of an stn package.
If true, it does not put new edges that go from and to the same event, but uses
them to determine the consistency of the STN. It returns the consistency. If stn-mode
is nil, it returns t and performs the full apsp."
  
  (with-slots (events) stn				;grab the event list
	(loop for k in events				;loop through possible 
	   do (loop for i in events
			 do (loop for j in events
				   ;;it is consistent if it was before and it is not found to have a problem when the edge is updated
					 ;;if it becomes inconsistent (as returned from update), return immediately
					 do (when (not (apsp-update-itc-path! i j k stn stn-mode)) ;update ij with shortcut through k
					 (return-from apsp-floyd-warshall! nil)))))) 
	
  t)

(defmethod apsp-update-itc-path! ((i event) (j event) (k event) (stn stn) stn-mode)
"Consider this triangle of events and update the ij edge if k is a suitable shortcut. Returns
whether an inconsistency was found. If stn-mode is true, it won't put in self-loops, just consider them."
  (let* ((pair-ij (get-event-pair i j stn)) ;;grab pairs and weights
		 (pair-ik (get-event-pair i k stn))
		 (pair-kj (get-event-pair k j stn))
		 (d-ij (get-event-pair-ub pair-ij))
		 (d-ik (get-event-pair-ub pair-ik))
		 (d-kj (get-event-pair-ub pair-kj))
		 (consistent? t))
	(when (and (bound? d-ik) (bound? d-kj)) ;if the alternate path is bound
	  (let ((d-shortcut (b+ d-ik d-kj))) ;compute the shortcut distance
		(when (or (not (bound? d-ij)) (b> d-ij d-shortcut)) ;;if ij is either unbound or longer, 
		  (if (and stn-mode (eql i j)) ;if it's a diagonal update in stn-mode, don't do it
			  (progn (when (b> 0 d-shortcut) ;if the diagonal would become negative, set it inconsistent
					   (setf consistent? nil))) ;and mark the stn as inconsistent
										;else, update the bound

			  (update-itc-event-pair-ub pair-ij d-shortcut nil ;(d-edge-constraint (first (event-pair-increasing-d-edges pair-ik)))
									)))))
	consistent?))


(defmethod update-itc-event-pair-ub ((event-pair event-pair) ub &optional constraint)
  "Updates the d-edge corresponding to an event-pair. It either updates the
edge already present or creates a new one."
  (when (bound? ub)						;error checking
	(let ((edge (car (event-pair-increasing-d-edges event-pair))))
	  (when constraint) ; Does nothing, just remove optional warning

	  (if (null edge)
		  (with-slots (start end) event-pair ;make a new edge
			(make-instance 'itc-d-edge :start start :end end :ub ub :constraint constraint))
		  ;; Else update
		  (progn
			  (setf (d-edge-ub edge) ub)
			 (when (not (eql nil (d-edge-constraint edge)))
			   (if (d-edge-lower-bound? edge)
			 	   (setf (stc-lb (d-edge-constraint edge)) (abs(d-edge-ub edge)))					
			 	   (setf (stc-ub (d-edge-constraint edge)) (d-edge-ub edge))
			 	   )))))))






;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Functions used by kirk/OPSAT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod new-constraint
	((STN LTNP) &key source start end lb ub N-lb N-ub t-lb t-ub
	 (guard *true*) 
	 (lb-relaxable? nil)
	 (lb-cost-ratio nil)
	 (lb-relaxed? nil)
	 (relaxed-lb nil)
	 (ub-relaxable? nil)
	 (ub-cost-ratio nil)
	 (ub-relaxed? nil)
	 (relaxed-ub nil)
	 (reward-function nil))

  "Creates a new simple temporal constraint (STC) for NETWORK whose source is 
   SOURCE.  START and END are event objects or names of events, and LB and UB 
   are reals, where LB <= UB.  Guard is either *true* or an assignment 
   (= V X) of a state variable V to one of its domain elements X.  V may either 
   by a state variable name or state-variable object of one of STN's state 
   variables.  The default for guard is *true*.

   If Guard is *true*, then when STC is inserted, it enforces:

      (START - END) in [l,u]

   Otherwise, when STC is inserted, it enforces:

    (= V X) implies (START - END) in [l,u]

   The function creates a new STC, but DOES NOT insert it into 
   NETWORK, i.e., the STC is a removed STC.  STC may be inserted later
   by calling INSERT-CONSTRAINT.  

   This function DOES NOT check for duplicate constraints, 
   but merely proceeds to create additional copies.  Events, however,
   are canonicalized.

   The function returns the newly created constraint."
  
 ;; (format t "~A-~A:[~A,~A] ~A:~A~%" start end lb ub lb-relaxable? lb-cost-ratio)
  (if (eql N-lb nil)
	  (make-instance 
	   'itc-stc :source source
	   :constraint-system STN
	   :start start
	   :end end
	   :lb lb
	   :ub ub
	   :guard guard
	   :lb-relaxable? lb-relaxable?
	   :lb-cost-ratio lb-cost-ratio
	   :lb-relaxed? lb-relaxed?
	   :relaxed-lb relaxed-lb
	   :ub-relaxable? ub-relaxable?
	   :ub-cost-ratio ub-cost-ratio
	   :ub-relaxed? ub-relaxed?
	   :relaxed-ub relaxed-ub)
	  
	  (make-instance 
	   'LTCP 
	   :source source 
	   :constraint-system STN
	   :start start 
	   :end end 
	   :lb (* t-lb N-lb) 
	   :ub (* t-ub N-ub)
	   :N-lb N-lb 
	   :N-ub N-ub 
	   :t-lb t-lb 
	   :t-ub t-ub 
	   :guard guard
	   :reward-function reward-function)


	  ;; (insert-new-looping-constraint 
	  ;;  STN 
	  ;;  :source source
	  ;;  :start start
	  ;;  :end end
	  ;;  :N-lb N-lb 
	  ;;  :N-ub N-ub 
	  ;;  :t-lb t-lb 
	  ;;  :t-ub t-ub
	  ;;  :reward-function reward-function)
	  ))

(defmethod insert-constraint :after ((constraint-system LTNP) (constraint ltcp))
  
  "Inserts a simple temporal CONSTRAINT into its constraint-system. 
   It queues up incremental propagation, used to determine event bounds 
   and consistency when queried.  Propagation is not initiated by this 
   function.  Note: no queuing is required unless consistency has been
   previously checked.

   CONSTRAINT itself must have been created earlier using NEW-CONSTRAINT.
   If CONSTRAINT is already inserted, this function does nothing.  

   The function is called for effect; it does not return anything 
   meaningful."
  
  (push constraint (ltnp-looping-constraints constraint-system)))

(defmethod consistent?
    ((stn ltnp) 
	 &key	 
	 &allow-other-keys)

  "Checks consistency of LTNP.  Returns (values CONSISTENT?). 

  Currently this is done by solving the LTNP optimally. Instead, 
  it should return t after ANY integer assignment to the LTNP has
  been found, as this is sufficient to check consistency."

  
  (let ((search-tree (initialize-ltnp-search-tree stn)))
	(loop while (not (adt:queue-empty? (search-q search-tree))) do
		 (let ((branch (pop-branch search-tree)))
		   (if (not(consistent-branch? branch stn))   ;;(consistent-branch?-and-prune branch ltnp)
			   ;; if not consistent, prune branch (already done by pop) and reset ITC d-edges	 
			   (progn 	 
				 (reset-d-edges! search-tree stn))
			   ;; else
			   (progn
				 (when (branch-has-integer-assignment? branch stn) ;; We have found the best solution
				   (branch-to-ltnp! (start-branch search-tree) stn)
				   (return-from consistent? t))
				 ;; Two options here: split on reward, or split on interval
				 ;;(split-branch-based-on-reward! branch ltnp)
				 (let ((two-branches (multiple-value-bind (a b) (split-branch-based-on-interval! branch stn) (list a b))))
				   (add-branches-to-tree two-branches search-tree)				   ))))
	   ;;(format t "~%QUEUE_SIZE at end of loop: ~a " (coll-size (search-q search-tree)))
		 ))   
  nil)

 ;;; ***************************************************

;;; Other methods

;;; **************************************************


(defmethod print-stn! ((stn ltnp) &optional (print-removed? t))

  "For simple temporal network STN, enumerates the events, 
   distance edges, and bounds of STN."
  
  (format t "~2%LTNP ~a:" stn)
  (format t "~%  Event bounds:~%    ")
  (loop with i = 0
	  for e in (stn-events stn)
	  do (setq i (+ i 1))
		 (format t "~A in [~A,~A], " e (event-lb e)(event-ub e))
	  when (= i 4)
	  do (setq i 0)
		 (format t "~%   "))
  (format t "~%  Constraints (* denotes active):~%    ")
  (loop with i = 0
	  for c in (constraint-system-constraints stn)
	  do (setq i (+ i 1))
		 (cond ((and (constraint-removed?  c)
					 print-removed?)				
				(format t "~a [out], " c))
			   ((stc-active? c)
				(format t "~a *~% " c))
			   (:otherwise
				(format t "~a ~% " c)))
	  when (= i 3)
	  do (setq i 0)
		 )
  (format t "~%  D-edges:~%    ")
  (loop with i = 0
	 for e in (stn-events stn)
	 do (loop for d in (event-successors e)
		   do (setq i (+ i 1))	
			 (format t "~A>~A:~A, " (d-edge-start d)(d-edge-end d) (d-edge-ub d))
		   when (= i 4)
		   do (setq i 0)
			 (format t "~%    ")
			 ))
  (format t "~%"))


(defun copy-hash-table (table &key key test size
                                   rehash-size rehash-threshold)
  "Returns a copy of hash table TABLE, with the same keys and values
  as the TABLE. The copy has the same properties as the original, unless 
  overridden by the keyword arguments.

  Before each of the original values is set into the new hash-table, KEY
  is invoked on the value. As KEY defaults to CL:IDENTITY, a shallow
  copy is returned by default."
  (setf key (or key 'identity))
  (setf test (or test (hash-table-test table)))
  (setf size (or size (hash-table-size table)))
  (setf rehash-size (or rehash-size (hash-table-rehash-size table)))
  (setf rehash-threshold (or rehash-threshold (hash-table-rehash-threshold table)))
  (let ((copy (make-hash-table :test test :size size
                               :rehash-size rehash-size
                               :rehash-threshold rehash-threshold)))
    (maphash (lambda (k v)
               (setf (gethash k copy) (funcall key v)))
             table)
    copy))