;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2008-10 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;; ***********************************************************************
;;;
;;;                Manipulation of a Generic Simple Temporal Network
;;;
;;; ***********************************************************************

;;; Last update: July 6th, 2008.

;;; This file contains the major interface functions for a generic
;;; simple temporal constraint network.  Please go to manipulate-itc-stn.lisp
;;; for the interface functions for incremental temporal consistency (including
;;; specializations of those provided ehre.

;;; This file includes functions 1) for creating a simple temporal network, and
;;; 2) inserting and removing simple temporal constraints of a network.
;;; This file depends on the objects and macros defined in stn-defs.lisp.

;;; The external interface functions defined in this file are:
;;;
;;;
;;; o Creating Simple Temporal Networks

;;;   (networks, simple temporal constraints and events).
;;;
;;;  MAKE-INSTANCE ('STN &KEY (NAME (GENTEMP "STN-")) (START-NAME 'START))
;;;  
;;;     Creates a new simple temporal network named NAME.  
;;;     The network is empty, that is, it has no constraints or events
;;;     associated with it, other than the START event.  
;;;     The function returns the newly created network.
;;;     The function has the side effect of setting 
;;;     *DEFAULT-STN* to be the newly created network.

;;; o Retrieving and Creating Events

;;;   GET-EVENT (EVENT-NAME (NETWORK STN))
;;;     Returns a unique event with EVENT-NAME,associated with NETWORK.  
;;;     Creates the event if it doesn't already exist, and returns it.

;;;   FIND-EVENT EVENT-NAME (NETWORK STN)
;;;     Returns a unique event with EVENT-NAME that is associated with 
;;;     NETWORK  If no such event exists in NETWORK, it returns NIL.

;;; o Creating and Inserting Simple Temporal Constraints

;;;   INSERT-NEW-CONSTRAINT NETWORK &KEY START END LB UB &OPTIONAL SOURCE
;;;     Creates a new simple temporal constraint (STC) from SOURCE, and INSERTS
;;;     it into NETWORK.  START and END are events or names of events, and LB 
;;;     and UB are reals, where LB <= UB.

;;;     The function creates a new STC, but DOES NOT insert it into 
;;;     NETWORK, i.e., the stc is a removed stc.  A STC may be later inserted 
;;;     by calling INSERT-CONSTRAINT.  

;;;     Note that this function does not check for
;;;     duplicate constraints, but merely proceeds to create additional
;;;     copies.  Any required events are created by the function, 
;;;     and pointers between the constraint and its events are put 
;;;     in place.  The function returns the newly created constraint.

;;;   MAKE-INSTANCE 'STC :SOURCE SOURCE :CONSTRAINT-SYSTEM STN :START START
;;;				          :END END :LB LB :UB UB

;;;   NEW-CONSTRAINT (STN STN) &KEY START END LB UB SOURCE
;;;
;;;   Creates a new simple temporal constraint (STC) for NETWORK whose source is 
;;;   SOURCE.  START and END are event objects or names of events, and LB and UB 
;;;   are reals, where LB <= UB.  
;;;
;;;   When STC is inserted, it enforces:
;;;
;;;      (START - END) in [l,u]
;;;
;;;   The function creates a new STC, but DOES NOT insert it into 
;;;   NETWORK, i.e., the STC is a removed STC.  STC may be inserted later
;;;   by calling INSERT-CONSTRAINT.  
;;;
;;;   This function DOES NOT check for duplicate constraints, 
;;;   but merely proceeds to create additional copies.  Events, however,
;;;   are canonicalized.
;;;
;;;   The function returns the newly created constraint.

;;; o Inserting and Removing Simple Temporal Constraints that 
;;;   are already associated with a Network.
;;;
;;;   INSERT-CONSTRAINT CONSTRAINT
  
;;;     Inserts a removed simple temporal CONSTRAINT into its network.  
;;;     Updates the propagation queue for network, where propagation
;;;     is used to determine event bounds and consistency when queried.
;;;     Propagation is not initiated by this function.

;;;     CONSTRAINT itself must have been created earlier using NEW-STC.
;;;     If CONSTRAINT is already inserted, this function does nothing.  

;;;     The function is called for effect, it does not return anything 
;;;     meaningful.

;;; o Print Functions

;;;   PRINT-STN! STN

;;;   Enumerates the events, simple temporal constraints,
;;;   and bounds of STN.

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			new-constraint
;;;			
;;;			insert-constraint
;;;
;;;			get-event
;;;
;;;			print-stn!
;;;	
;;;;;;            event-consequences
;;;			)
;;;		  :ITC)
;;;  )

;;; Things To Do:

;;; o new-opsat-problem should allow the start name to be passed in as
;;;   an initialization argument.

;;; o The documentation regarding remove constraint may be incorrect.

;;; o Add a reference time.  
;;;   - It is a slot of network as stn-start-event.  
;;;   - Decide how to name it (user supplied, or default name).
;;;   - Create it within new-stn.
;;;   - Bind it to *default-start*.
;;;   - Determine how to make sure that all events are constrained
;;;     to not occur before start.

;;; o Should we allow the ub and lb of events to be externally set?

;;; ***********************************************************************
;;;
;;;                  External Interface
;;;
;;; ***********************************************************************

;;;   ***
;;;   *** Creating Simple Temporal Network Elements
;;;   ***

;;;   *** Simple Temporal Networks

;;;   *** Events

(defmethod get-event
    (event-name (network STN))

  " Returns the event object of NETWORK described 
    by EVENT-NAME.  If the object already exists, 
    it is returned; otherwise, a new event object
    is created, associated with NETWORK and returned.  

    Note: This is the preferred way to create a new 
    event since it avoids duplication."

  (or (find-event event-name network)
	  (progn
		
		(debug-print-for :itc t "New event ~a" ;; for ~A." 
						 event-name
;;;	   			         network
						 )
		  
		(make-instance 'event :name event-name
					   :network network
					   :start? nil))))

;;; Fix:
;;; o Make arguments consistent with ISAT.
(defmethod insert-new-constraint
	((constraint-system STN) &key source start end lb ub)

  "Creates a new STC whose source is SOURCE, AND inserts it into 
   CONSTRAINT-SYSTEM.  START and END are names of events, and LB and UB are reals,
   where LB <= UB.

   See documentation for new-constraint.

   Note that this function does not check for duplicate constraints, 
   but merely proceeds to create additional copies.  Any required events 
   are created by the function.  The function returns the newly created 
   constraint."

  ;; create the constraint,
  (let ((stc 
		 (make-instance 'STC :source source
						:constraint-system constraint-system
						:start start
						:end end
						:lb lb
						:ub ub)
		 ))
	;; add it into the constraint-system
    (insert-constraint constraint-system stc)
	;; and return it.
	(values stc)))


;;;   ***
;;;   *** Inserting and Removing Simple Temporal Constraints
;;;   ***

(defmethod insert-constraint ((constraint-system STN) (constraint stc))
  
  "Inserts a simple temporal CONSTRAINT into its constraint-system. 
   It queues up incremental propagation, used to determine event bounds 
   and consistency when queried.  Propagation is not initiated by this 
   function.  Note: no queuing is required unless consistency has been
   previously checked.

   CONSTRAINT itself must have been created earlier using NEW-CONSTRAINT.
   If CONSTRAINT is already inserted, this function does nothing.  

   The function is called for effect; it does not return anything 
   meaningful."
  (debug-print-for :itc t "Inserting ~A" ;; into ~A." 
				   constraint
;;;				   constraint-system
				   )
		 
  ;; If active, insert its d-edges.
  (dolist (edge (stc-d-edges constraint))
	(insert-d-edge! edge))
  T
  )

;;;   *** Simple Temporal Constraints

(defmethod new-constraint
	((STN STN) &key source start end lb ub (can-be-relaxed? T) &allow-other-keys)

  "Creates a new simple temporal constraint (STC) for CONSTRAINT-SYSTEM whose 
   source is SOURCE.  START and END are event objects or names of events, and 
   LB and UB are reals, where LB <= UB.  CAN-RELAX? is True iff constraint 
   can be relaxed by the problem solver; it defaults to True.

   When STC is inserted, it enforces:

    (= V X) implies (START - END) in [l,u]

   The function creates a new STC, but DOES NOT insert it into 
   CONSTRAINT-SYSTEM, i.e., the STC is a removed STC.  STC may be inserted later
   by calling INSERT-CONSTRAINT.  

   This function DOES NOT check for duplicate constraints, 
   but merely proceeds to create additional copies.  Events, however,
   are canonicalized.

   The function returns the newly created constraint."
  
  (debug-print-for :itc t "New constraint (~a,~a)[~a,~a] " ;; for ~A." 
				   start end lb ub
;;;				   stn
				   )  
  
  (make-instance 'stc :source source
				 :constraint-system STN
				 :start start
				 :end end
				 :lb lb
				 :ub ub
				 :can-be-relaxed? can-be-relaxed?))

;;; Check:
;;; o Does the search algorithm interface require this to
;;;   be a separate function from new-constraint?  If not, collapse 
;;;   this into new-stc, by making optional arguments keywords, 
;;;   and adding a keyword :insert?.


;;;   ***
;;;   *** Print Functions
;;;   ***

(defmethod print-stn! ((stn stn) &optional (print-removed? t))

  "For simple temporal network STN, enumerates the events, 
   distance edges, and bounds of STN."
  ;;make print-remove? error go away
  (when print-removed?)
  (format t "~2%STN ~a:" stn)
  (format t "~%  Event bounds:~%    ")
  (loop with i = 0
	  for e in (stn-events stn)
	  do (setq i (+ i 1))
		 (format t "~A in [~A,~A], " e (event-lb e)(event-ub e))
	  when (= i 4)
	  do (setq i 0)
		 (format t "~%    ")
		 )
  (format t "~%  Constraints:~%    ")
  (loop with i = 0
	  for c in (constraint-system-constraints stn)
	  do (setq i (+ i 1))
		 (format t "~a, " c)
	  when (= i 3)
	  do (setq i 0)
		 (format t "~%    "))
  
  (format t "~%  D-edges:~%    ")
  (loop with i = 0
	  for e in (stn-events stn)
	  do (loop for d in (event-successors e)
			 do (setq i (+ i 1))
				(format t "~A>~A:~A, " 
						(d-edge-start d)(d-edge-end d) (d-edge-ub d))
			 when (= i 4)
			 do (setq i 0)
				(format t "~%    ")
				))
  (format t "~%"))




;;; ***********************************************************************
;;;
;;;                            Support functions
;;;
;;; ***********************************************************************

;;;   ***y
;;;   *** Inserting Distance Edges ***
;;;   ***

(defmethod insert-d-edge! ((edge d-edge))
  
  "Just a place-holder - there's no work to be done for a non-itc d-edge."

  (debug-print-for :itc t "  Inserting d-edge ~A." edge)
  

  
  T)



;;;   ***
;;;   *** Error Checking
;;;   ***

(defun check-legal-bounds! (start end lb ub)
  
  "Returns true if and only if LB and UB are legal event
   upper and lower bounds."
  
  (cond ((not (bound? ub))
		 (error "Illegal constraint between ~a and ~a, ub ~a isn't a real."
				start end ub))
		((not (bound? lb))
		 (error "Illegal constraint between ~a and ~a, lb ~a isn't a real."
				start end lb))
		((b> lb ub)
		 (error "Illegal constraint between ~a and ~a, ub ~a below lb ~a."
				start end ub lb)))
  t)


;;; To Do:
;;; o Put the functions, below, that are related to propagation
;;;   into the propagation.lisp file.

;;;   ***
;;;   *** Initializing an STN for Propagation ***
;;;   ***


;;;   *** Initializing a Start Event

(defmethod start-initialize! ((event event))
  
  "Initialize testing of consistency and bound computation
   for start EVENT."
  T)


;;;   *** Initializing an Event ***

(defmethod event-initialize! ((event event))

  "Initialize testing of consistency and bound computation
   for non-start EVENT."
  T)

;;;   *** Remove an Event ***

(defmethod remove-event! ((event event))

  (debug-print-for :itc t "Removing event ~a" ;; from ~A." 
				   event)
  
  (with-slots (network) event
	(with-slots (events event-ht event-pair-ht) network
	  (loop for second-event in events
		  do (remhash (list event second-event) event-pair-ht)
		  do (remhash (list second-event event) event-pair-ht))
	  (setf events (remove event events)) ;remove it from the event list
	  (remhash (event-name event) event-ht))))

;;;   ***
;;;   *** Maintaining and Manipulating the d-edges of an STN
;;;   ***

(defmethod stn-add-d-edge! ((d-edge d-edge) (STN stn))

  "Adds D-EDGE to lists of all edges of network STN
   that go from event X to Y.
   Maintains one list in increasing order of cost,
   and another in decreasing order."

  (let* ((x (d-edge-start d-edge))
		 (y (d-edge-end d-edge))
		 ;(ub (d-edge-ub d-edge))
		 (event-pair (get-event-pair x y STN)))

	(with-slots (increasing-d-edges decreasing-d-edges) event-pair
	  (setf increasing-d-edges	
			(merge 'list increasing-d-edges (list d-edge) 
				   #'(lambda (i j) (b> (d-edge-ub j) (d-edge-ub i)))))
	  
	  (setf decreasing-d-edges	
			(merge 'list decreasing-d-edges (list d-edge) 
				   #'(lambda (i j) (b> (d-edge-ub i) (d-edge-ub j))))))

	T))


;;;   ***
;;;   *** Removing Distance Edges ***
;;;   ***

(defmethod remove-d-edge! ((edge d-edge))
  
  "Removes distance EDGE from its network. Since there is no mechanism
   to just mark it as inactive, remove it from the events and event-pair."

  (debug-print-for :itc t "  Removing d-edge ~a" ;; from ~A." 
				   edge)

  (with-slots (start end) edge
	(let ((event-pair (get-event-pair start end (event-network start))))

	  (setf (event-pair-increasing-d-edges event-pair) (remove edge (event-pair-increasing-d-edges event-pair)))
	  (setf (event-pair-decreasing-d-edges event-pair) (remove edge (event-pair-decreasing-d-edges event-pair)))
	  (setf (event-successors start) (remove edge (event-successors start)))
	  (setf (event-predecessors end) (remove edge (event-predecessors end))))))


;;;   ***
;;;   ***   Event-pair Edges
;;;   ***
(defmethod find-event-pair-max-value-matching-d-edge ((event-pair event-pair) (d-edge d-edge))
  "For a d-edge, the first one in the event pair is fine."
  (d-edge-ub (first (event-pair-increasing-d-edges event-pair))))


(defmethod event-pair-lowest-bound-inserted-d-edge ((event-pair event-pair))

  "Returns the inserted edge of EVENT-PAIR
   that has the least upper bound value.
   Returns nil if no such d-edge is inserted."

  ;; The following exploits the fact that an event-pairs
  ;; distance edges are listed in increasing value of upperbound.
  (loop for d-edge in (event-pair-increasing-d-edges event-pair)
	  do (return d-edge)
	  finally (return nil)))

(defmethod event-pair-highest-bound-inserted-d-edge 
	((event-pair event-pair))

  "Returns the inserted edge of EVENT-PAIR
   that has the least upper bound value.
   Returns nil if no such d-edge is inserted."

  ;; The following exploits the fact that an event-pairs
  ;; distance edges are listed in decreasing value of upperbound.
  (loop for d-edge in (event-pair-decreasing-d-edges event-pair)
	  do (return d-edge)
	  finally (return nil)))

;;;   ***
;;;   ***   Default STN
;;;   ***

;;; Creates a default STN
(eval-when (load eval)
  (make-instance 'STN :name 'Default-STN
				 :start-name 'START)
  )


