;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2008 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;; ***********************************************************************
;;;
;;;                Incremental Propagation for Temporal Consistency
;;;                               Part II: Lower Bound
;;;
;;; ***********************************************************************

;;; Last update: May 4th, 2008.

;;; This module provides the core algorithms for:
;;; 1) checking temporal consistency of a simple temporal network, and
;;; 2) computing lower bounds on network events, 
;;; based on a generalization of an incremental single-destination, 
;;; shortest path algorithm.

;;; This module is comprised of support functions for lower bound propagation
;;; by PROPAGATE!.  Functions for detecting inconsistency and onflit extraction
;;; are in conflicts.lisp.

;;; This implementation has the following properties, similar to upper
;;; bound propagation:

;;; o Updating the consequences of insertions and deletions is delayed until 
;;;   PROPAGATE is called.

;;; o When switching between edges, we propagate effects after
;;;   ALL insertions and deletions are made, and just before answering a query.
;;;   This allows us to replace a support with a better support, if possible.
;;    Otherwise, we would always have to first remove consequences when the 
;;;   edge is removed, and possibly rederiving these same consequences when
;;;   the insertion is performed.

;;; To Do:

;;; o Update pseudo code for lb propagation rather than upper bound propagation.

;;; Optimization:

;;; o This code should not propagate effects through recursive 
;;;   calls, it should instead use a set of stacks, built on top 
;;;   of growable arrays.

;;; o Some of these functions could be made more compact, by merging them
;;;   with their corresponding upper bound function, and passing in a flag
;;;   for upper bound versus lower bound.

;;;Pseudo Code modified from ITC ICAPS Paper, by Shu, Effinger and Williams:

;;; ***********************************************************************
;;;
;;;     Support for Lower Bound Propagation in Propagate!
;;;
;;; ***********************************************************************

;;; ***********************************************************************
;;;
;;;     Propagation of Event Bounds
;;;
;;; ***********************************************************************

;;; Note: I don't know how to check temporal consistency based on LBs,
;;;       the following, however, is incorrect.

;;;(defun update-lb-temporal-consistency! (network)
;;;
;;;  "Performs lower bound propagation and updates record of
;;;   temporal consistency and any conflicts for network."
;;;
;;;  (let ((lb-conflict (check-lb-temporal-consistency network)))
;;;	(cond ((null lb-conflict)
;;;		   (debug-print-for :itc t "~% Lower bound consistent.")
;;;		   (values T))
;;;		  (:otherwise
;;;		   (debug-print-for :itc "~% Inconsistent: lb conflict = ~a"
;;;							(conflict-constraints lb-conflict))
;;;		   (STN-lb-inconsistent! network)
;;;		   (setf (STN-lb-conflict network) lb-conflict)
;;;
;;;		   ;; (print-stn network)
;;;		   (values nil)))))

;;; The first two routines below compute shortest path values and 
;;; consistency of the initial graph.

;;;Conflict
;;;CheckTemporalConsistency()
;;;
;;;// Determine new shortest paths and check consistency, 
;;;// by propagating changes of all vertices on queue.
;;;// Returns conflict if inconsistent, or {} if consistent.
;;;
;;;{07} while !Q.empty()
;;;{08}   u = Q.pop();
;;;{09}   for v in Succ(u)
;;;{10}     dval = Update(u,v);
;;;
;;;{11}     if(dval) < 0
;;;{12}       if (CompletedCycle(v))
;;;{12b}         return ExtractConf(v, {});
;;;
;;;{13} return {};

(defmethod update-lbs! ((STN STN))

  " Computes updated lower bounds on events, and determines
    consistency."

  (itc-stn-lb-propagated! STN)

  ;; Q contains all events whose lower bounds
  ;; need to be propagated.  This is because either
  ;; 1) the event's bounds have changed, or 2)
  ;; we need to propagate event to compute a new bound
  ;; (resupport) a neighbor.
  ;; Propagate these events
  (debug-print-for :itc t "Processing LB Q.")
  (loop with Q = (stn-lb-queue STN)
	  until (queue-empty? Q)
	  for y = (queue-pop Q)
			  ;; Grab an event
			  
	  do (update-lb-predecessors! y STN)
		 ;; propagate event's bounds to its predecessors.

	  when *itc-statistics-p*
	  do (incf *propagate-counter*)))

(defmethod update-lb-predecessors! ((y event) (STN STN))

  "Propagate event y's bounds to its predecessors within simple temporal
   network STN."

  ;; Relax every inserted EDGE eminating to y.
  (debug-print-for :itc t "Relax edges in to ~a.  Edges = ~a." 
				   y (event-predecessors y))
  (do-inserted-d-edges (edge (event-predecessors y))
	(STN-lb-update! y edge STN))
  T)

;;;value
;;;Update(x,y) 
;;;
;;;  // Relax edge x -> y and propagate
;;;  // successors if d(y) changes.
;;;
;;;{14} if ( d(y) > d(x) + c(x,y))
;;;{15}   d(y) := d(x) + c(x,y);
;;;{16}   p(y) := x;
;;;{17}   Q.insert(y);
;;;{18} return d(y);

(defmethod STN-lb-update! ((y event) (edge d-edge) (STN STN)) 

  "Relax EDGE X -> Y of simple temporal network STN and propagate 
   predecessors if the lower bound of event X changes."

  (let* ((x (d-edge-start edge))
		 (x-lb (event-lb x))
		 (y-lb (event-lb y))
		 (xy-cost (d-edge-ub edge))
		 (new-x-lb (b+ y-lb (b- xy-cost))))
	
	;; Compute a relaxed lower bound value for event X.
	;; x-lb is the old lower bound and new-x-lb is the 
	;; new relaxed value.

	(when (b> new-x-lb x-lb)
	  ;; Relaxation improves the bound,
	  ;; record the new bound and its supporting edge, 
	  (setf (event-lb x) new-x-lb)
	  (set-lb-support-relation! x edge)

	  ;; and queue X in order to propagate its change.
	  (queue-insert! x (stn-lb-queue stn))
	  (debug-print-for :itc t "LB Relax ~a->~a: ~a = ~a -> ~a = ~a - ~a" 
			  x y x x-lb new-x-lb y-lb xy-cost)
	  )
	T))

;;; ***********************************************************************
;;;
;;;     Update after Insertion and Removal of Distance Edges
;;;
;;; ***********************************************************************

;;;  In the following, edge insertions and removals are all accumulated,
;;;  and then the best edge replacement is made when propagate is called.
;;;  This helps limit unnecessary unsupport/resupports.

;;; From the original ITC pseudo code (does not precisely reflect this 
;;; implementation).

;;;void
;;;ModifyConstraint(x,y,l,u)
;;;
;;;   // Change the bounds [l,u] of simple temporal 
;;;   // constraint between x and y.
;;;  
;;;{24} ModifyArc(arc(y,x),-l)
;;;{25} ModifyArc(arc(x,y),u)

;;;(defun modify-constraint! (x y l u STN)
;;;
;;;  "Change the bounds [l,u] of simple temporal 
;;;    constraint between x and y."
;;;
;;;  (STN-modify-arc! y x (b- l) STN)
;;;  (STN-modify-arc! x y u STN))

;;;Void
;;;ModifyArc(arc,c)
;;;
;;;  // Change to c the edge weight of arc,
;;;  // and queue relevant vertices for update.
;;;
;;;{26} setCost(arc,c);
;;;{27} if (d(arc.head) > d(arc.tail) + c) 
;;;       // Change improves shortest path
;;;       // Update and propagate.
;;;
;;;{28}   d(arc.head) := d(arc.tail) + c;
;;;{29}   p(arc.head) := arc.tail;
;;;{30}   Q.Insert(arc.head);
;;;
;;;{31} elseif (d(arc.head) < d(arc.tail) + c
;;;            AND (p(arc.head) == arc.tail)) 
;;;       // Change worsens shortest path,
;;;       // Reset value.
;;;
;;;{32}   d(arc.head) := infinity;
;;;{33}   p(arc.head) := unknown;
;;;
;;;       // Reset consequences.
;;;{34}   set<Node> nodes_reset = {};
;;;{35}   nodes_reset.insert(arc.head);
;;;{36}   for all n in InvalidateSupports(arc.head)
;;;{36}     nodes_reset.insert(n)
;;;
;;;       // Setup to resupport.
;;;{37}   InsertParents(nodes_reset)

(defmethod modify-d-edge-lb! (new-edge (y event) (x event) (STN STN))

  " Incrementally relax NEW-EDGE.  The result may either tighten,
    loosen or have no effect on the lower bound of NEW-EDGE's 
    start event.  When start event changes, queue propagation.

    Note: This function has no effect if shortest paths have
          not previously been computed, since no incremental
          change is required.  Instead the shortest paths are
          computed from scratch upon the first query."
  
  (debug-print-for :itc t "Updating lb ~a -> ~a to ~a."
				   x y new-edge)
  
  (if (null new-edge)  
	  (when (and (event-lb-support x)
				 (eql y
					  (d-edge-end (event-lb-support x))))

		(weaken-d-edge-lb! x stn))

	(when (STN-lb-propagated? STN)
	  ;; Note: This is redundant, its checked by the caller.

	  ;; Only perform incremental update if consistency of the STN
	  ;; has already been checked at least once.

	  (let* ((new-edge-ub (d-edge-ub new-edge))
			 (x-lb (event-lb x))
			 (y-lb (event-lb y))
			 (new-x-lb (b+ y-lb (b- new-edge-ub))))

		;; x-lb is the old lower bound on x, while
		;; new-x-lb is the new, relaxed lower bound on x.
		
		(cond ((b> new-x-lb x-lb)
		  
			   (strengthen-d-edge-lb! new-edge x new-x-lb x-lb stn))

			  ((and (b> x-lb new-x-lb)
					(event-lb-support x)
					(eql y
						 (d-edge-end (event-lb-support x))))
			   
			   (weaken-d-edge-lb! x stn))

			  (:otherwise

			   ;; Change keeps the bound on x the same.
			   
			   (unless (eql new-edge (event-lb-support x))
				 (setf (event-lb-support x) new-edge))))
		T))))

(defmethod weaken-d-edge-lb! ((x event) (stn stn))

  "Change worsens the lower bound of x and support was between the same events 
   as new-edge.  Re-relax x and its consequences, starting from -infinity."

  (debug-print-for :itc t "Weakening ~a."
				   x)
  (let ((events-reset (list x)))
	;; Reset the lower bound value of x to -infinity
	;; and remove its support.
	(event-lb-initialize! x)

	;; Similarly reset all consequences of x.
	(setq events-reset
	  (reset-all-lb-consequences x STN events-reset))

	;; Finally, setup lb propagation queue to compute the
	;; bounds for x AND its consequences.
	(queue-successors-for-lb-propagation! events-reset STN)
	t))

(defmethod strengthen-d-edge-lb! 
	((new-edge d-edge) (x event) new-x-lb x-lb stn)
  
  ;; Change improves the lower bound on x. Update and propagate.

  (debug-print-for :itc t "Edge ~a strengthens LB of ~A to ~A ~
                               from ~A."
				   new-edge x new-x-lb x-lb)

  (setf (event-lb x) new-x-lb)
  (setf (event-lb-support x) new-edge)
  (queue-insert! x (STN-lb-queue STN))
  )

;;;The next two routines implement an arc weight change, by
;;;removing support and setting up propagation, as appropriate.

;;;set<Node>
;;;reset-all-ub-consequences(Node n) 
;;;
;;;  // Remove n as a support for any s,
;;;  // and queue s's predecessors for resupport.
;;;  // Returns list of nodes whose d and p values
;;;  // have been reset.
;;;
;;;{42} set<Node> nodes_reset = {};
;;;{43} for all s in Succ(n)
;;;{44}   if( p(s) == n )
;;;         // Successor s is supported by n
;;;         // since n reset, reset s as well.
;;;
;;;{45}     if( s == sstart )
;;;{46}       d(s) := 0;
;;;{47}       for all t in reset-all-ub-consequences(s)
;;;{47a}        nodes_reset.insert(t);
;;;
;;;{48}     elseif( d(s) != infinity OR p(s) != unknown )
;;;{49}       d(s) := infinity;
;;;{50}       p(s) := unknown;
;;;{51}       for all t in reset-all-ub-consequences(s)
;;;{51a}        nodes_reset.insert(t);
;;;
;;;{52} return nodes_reset;

;;; The above pseudo code does not correctly accumulate the
;;; reset nodes.  It instead needs to insert the s nodes that
;;; have been reset.  This is corrected in the function below.

(defmethod reset-all-lb-consequences ((y event) (STN STN) events-reset)

  "Y is an event of STN whose lower bound value has been reset.
   Reset all event lower bound values and support for events 
   that are consequences of Y.
   EVENTS-RESET is a list of events already reset.
   Returns EVENTS-RESET, augmented with all additional
   events reset within this function."
  
  ;; Remove consequences of X by recursively
  ;; looping through the immediate consequences of X.
  ;; Collect and return all reset events.
  (when (event-predecessors y)
	(debug-print-for :itc t "Resetting LB predecessors of ~a."
					 y)
	(loop
		for edge in (event-predecessors y)
		for x = (d-edge-start edge)
		when (eql (event-lb-support x) 
				  edge)
			 ;; Event x is supported by y through edge;
			 ;; Initialize x and its consequences.

			 ;; Start by removing edge as a support of x.
		do (clear-d-edge-lb-consequent! edge)
		 
		   ;; Reset the lower bound of x and
		   ;; reset all consequences of x,
		   ;; collecting the reset events.
		   (cond ((eql x (STN-start STN))
				  (start-lb-initialize! x)
				  (push x events-reset)
				  (setq events-reset
					(reset-all-lb-consequences x STN events-reset)))
			   
				 ((not (event-lb-useless? x))
				  (event-lb-initialize! x)
				  (push x events-reset)
				  (setq events-reset
					(reset-all-lb-consequences x STN events-reset)))
			   
				 (:otherwise nil))
		 
		finally (return events-reset))))

;;;void
;;;queue-predecessors-for-ub-propagation!(set<Node> reset_nodes) 
;;;
;;;  // rest_nodes lost their support, queue predecessors
;;;  // to resupport.
;;;
;;;
;;;{38} for all n in reset_nodes
;;;{39}   for all m in Pred(n)
;;;{40}     if( d(n) != infinity || p(s) != unknown )
;;;{41}       Q.Insert(m);

;;; Check: 
;;; o L40, changed d(n), p(s) to d(m),p(m).
;;; o Should only need to check if unbounded, 
;;;   not if unsupported, yes?
;;;   Unsupported but bounded should be an error.

(defmethod queue-successors-for-lb-propagation! (reset-events (STN STN)) 

  "RESET-EVENTS are events that have lost their support, 
   prepare to compute a new bound for each reset event by
   queuing the event's successors."

  (unless (null reset-events)
	(loop with lb-q = (STN-lb-queue STN)
		for x in reset-events
				 ;; queue the successor of each reset event X for propagation,
				 ;; unless the successor imposes no constraint.
				 ;; A successor is the end of an edge whose start is X.
		do (debug-print-for :itc t "LB Queueing successors of ~a..."
							x)
		   (loop for edge in (event-successors x)
			   for w = (d-edge-end edge)
			   unless (event-lb-useless? w)
			   do (queue-insert! w lb-q)))))

;;; ***********************************************************************
;;;
;;; Simpler Implementation of Update after Insertion and Removal of D Edges
;;;
;;; ***********************************************************************

;;; The following three functions implement the same functionality
;;; as modify-d-edge!.  They are simpler in that they perform
;;; incremental changes, as separate insertions and removals of d-edges.
;;; They are less efficient in that removed edge loose their support
;;; and consequences, and then new, potentially, stronger inserted
;;; edges are added and propagated.

;;; Note: not used.
;;;(defun relax-inserted-d-edge! (edge)
;;;
;;;  " Incrementally relax inserted EDGE to tighten 
;;;    the bound of edge's end event.  When successful, queue
;;;    successors for propagation.
;;;
;;;    Note: This function has no effect if shortest paths have
;;;          not previously been computed, since no incremental
;;;          change is required.  Instead the shortest paths are
;;;          computed from scratch upon the first query."
;;;
;;;  (let ((STN (stc-network (d-edge-constraint edge))))
;;;	(when (STN-ub-consistency-checked? STN)
;;;
;;;	  ;; Only perform incremental update if consistency of the STN
;;;	  ;; has already been checked at least once.
;;;
;;;	  (let* ((x (d-edge-start edge))
;;;			 (y (d-edge-end edge))
;;;			 (ub (d-edge-ub edge))
;;;			 (x-ub (event-ub x))
;;;			 (y-ub (event-ub y))
;;;			 (r-y-ub (b+ x-ub ub)))
;;;
;;;		;; y-ub is the old upper bound on y, while
;;;		;; r-y-ub is the new, relaxed upper bound on y.
;;;		
;;;		(when (b> y-ub r-y-ub)
;;;		  
;;;		  ;; Change improves shortest path.
;;;		  ;; Update and propagate.
;;;
;;;		  (debug-print-for :itc t "Inserted edge ~a improves ub of event ~A ~
;;;                          from ~A to ~A, propagate."
;;;					   edge y y-ub r-y-ub)
;;;
;;;		  (setf (event-ub y) r-y-ub)
;;;		  (setf (event-ub-support y) edge)
;;;		  (queue-insert! y (STN-ub-queue STN)))))))

;;; Note: not used.
;;;(defun resupport-consequence-of-removed-d-edge! (edge)
;;;
;;;  " If removed EDGE supports its end event, try
;;;    to resupport END, updating and setting up
;;;    the ub-propagation queue as required."
;;;
;;;  (let ((STN (stc-network (d-edge-constraint edge))))
;;;	(when (STN-ub-consistency-checked? STN)
;;;
;;;	  ;; Only perform incremental update if consistency of the STN
;;;	  ;; has been checked at least once.  Otherwise, bounds are
;;;	  ;; computed from scratch upon query.
;;;
;;;	  (let ((y (d-edge-end edge)))
;;;
;;;		(when (eql (event-ub-support y)
;;;				   edge)
;;;		  ;; Edge supports y, then
;;;		  ;; try to find an alternative support that results in
;;;		  ;; the identical upper bound value for y.
;;;
;;;		  (let ((success? (replace-with-equivalent-support y edge)))
;;;			(unless success?
;;;			
;;;			  ;; Couldn't find equivalent alternative support.
;;;
;;;			  ;; Reset the upper bound value of y to infinity.
;;;			  (event-ub-initialize! y)
;;;
;;;			  ;; and reset all consequences of y.
;;;			  (let ((nodes-reset '()))
;;;				(push y nodes-reset)
;;;				(loop for n in (reset-all-ub-consequences y STN)
;;;					do (push n  nodes-reset))
;;;
;;;				;; Finally, setup ub propagation queue to compute the
;;;				;; bounds for y AND its consequences.
;;;				(queue-predecessors-for-ub-propagation! nodes-reset STN))
;;;			  ))))))
;;;  T)

;;; Note: not used.
;;;(defun replace-with-equivalent-support (y edge)
;;;  
;;;  "Let Y be an event that is supported by EDGE.  
;;;   Try resupporting y with an alternative inserted 
;;;   edge whose relaxtion results in the same upper 
;;;   bound value for Y.
;;;   Returns True if and only if success."
;;;  
;;;  (let ((ub (d-edge-ub edge)))
;;;	(loop for new-edge in (event-predecessors y)
;;;		do (when (and (not (d-edge-removed? edge))
;;;					  (not (eql new-edge edge))
;;;					  (eql ub (d-edge-ub new-edge)))
;;;			 (set-ub-support-relation! y new-edge)
;;;			 (return T))
;;;		finally (return nil))))
