;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2008-2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			stn-initialize!
;;;			)
;;;		  :ITC))

;;; ***********************************************************************
;;;
;;;                Incremental Propagation for Temporal Consistency
;;;
;;; ***********************************************************************

;;; Last update: April 27th, 2008.

;;; This module provides the core algorithms for:
;;; 1) checking temporal consistency of a simple temporal network, and
;;; 2) computing time bounds on network events.

;;; This code is based on a generalization of incremental 
;;; single-source/destination, shortest path algorithms.

;;; This module includes two interface functions, STN-INITIALIZE! and
;;; PROPAGATE!.  STN-INITIALIZE! is an external interface function for ITC.
;;; PROPAGATE is NOT part of the external interface.  Instead, 
;;; any interface function that relies on propagation having been performed 
;;; should call PROPAGATE! explicitly:

;;; STN-INITIALIZE! (STN)
;;;   Initialize upper and lower bounds of event times for simple
;;;   temporal network STN, and all propagation information.
;;;   Leaves inserted/removed status of constraints unchanged.

;;; PROPAGATE! (&OPTIONAL (NETWORK *DEFAULT-NETWORK*))
;;;   Given NETWORK, propagates the queued distance edges and events 
;;;   of NETWORK until the propagation-queues become empty, or an 
;;;   inconsistency is detected.  If the propagation queues become 
;;;   empty, then all events have the tightest bounds that logically 
;;;   follow from the inserted simple temporal constraints of NETWORK.

;;;   No bounds can be believed if NETWORK is inconsistent.  Whenever 
;;;   this function returns, the propagation queues have been correctly set.

;;; This implementation has the following properties:

;;; o Updates as a consequence of insertions and deletions is delayed until 
;;;   PROPAGATE is called.

;;; o When modifying edges, we propagate effects after
;;;   ALL insertions and deletions are made, and just before answering a query.
;;;   This allows us to replace a support for an edge with a better support, 
;;;   if possible, before propagation.
;;;   Otherwise, we would always have to first remove consequences when 
;;;   support is removed, and possibly rederive the same consequences when
;;;   new support is inserted.

;;; Pseudo Code listed below is modified from the ITC ICAPS Paper, 
;;; by Shu, Effinger and Williams.

;;; ***********************************************************************
;;;
;;;     To Do:
;;;
;;; ***********************************************************************
;;; To Do:

;;; o Add option to STN to explicitly record that every event follows start.
;;; o See inline comments.

;;; Optimization:

;;; o This code should not propagate effects through recursive 
;;;   calls, it should instead use a set of stacks, built on top 
;;;   of growable arrays.

;;; ***********************************************************************
;;;
;;;     Initialize, Propagate and their Supporting Functions
;;;
;;; ***********************************************************************

;;; ***********************************************************************
;;;
;;;     Initialization for Propagation
;;;
;;; ***********************************************************************

;;;void
;;;Initialize()
;;;
;;; // Set shortest path length for all s but start to infinity.
;;;
;;; // Q - a LIFO Queue
;;; // d(s) - length of shortest path found from sstart to s.
;;; // p(s) - next to last vertex on shortest path from sstart to s.
;;; // v(s) - vertex has been previously visited (relaxed).
;;;
;;;{01} Q := {};
;;;{02} for all s in V(G)
;;;{03}   d(s) := infinity;
;;;{04}   p(s) := unknown;
;;;{04a}  v(s) := false;
;;;{new}  s(s) := unreached;
;;;{05} d(sstart) := 0;
;;;{05a} v(sstart) := false;
;;;{new} s(sstart) := labeled;
;;;{06} Q.insert(sstart);

;;; To Do:
;;; o Move this function to stn-manipulation, so that all
;;;   interface functions are in the same place.  Split the
;;;   function, if needed.


;;; Fix:
;;; This method appears identical to STN-initialize!.
;;; Remove as soon as we confirm that all systems work.
;;;(defmethod initial-STN-initialize! ((STN STN))
;;;  
;;;  "Initialize upper and lower bounds of event times for simple
;;;   temporal network STN, and all propagation information.
;;;
;;;   Leaves inserted/removed status of constraints unchanged."
;;;
;;;  ;; Initializes the STN consistency algorithm.
;;;  (STN-wide-initialize! stn)
;;;
;;;  ;; Initialize the lower and upper bound values and 
;;;  ;; support of all events in STN.
;;;  (dolist (e (STN-events STN))
;;;	(event-initialize! e))
;;;
;;;  ;; Initialize the consequences and in-conflict? flags
;;;  ;; of all constraints.
;;;  (dolist (c (constraint-system-constraints STN))
;;;	(STC-initialize! c))
;;;	
;;;  (let ((ub-queue (STN-ub-queue STN))
;;;		(lb-queue (STN-lb-queue STN))
;;;		(start (STN-start STN)))
;;;	
;;;	;; Initialize the upper and lower bound values and support 
;;;	;; of the START event of STN.
;;;	(start-initialize! start)
;;;	
;;;	;; Re-initialize the upper and lower bound propagation queues.
;;;	;; The first time that upper and lower bounds and consistency 
;;;	;; are computed, propagation emanates from the START event of
;;;	;; STN.
;;;	(queue-insert! start ub-queue)
;;;	(queue-insert! start lb-queue)
;;;    T))

;;; Reviewed by BCW.
(defmethod STN-initialize! ((STN STN))
  "Initialize upper and lower bounds of event times for simple
   temporal network STN, and all propagation information.

   Leaves inserted/removed status of constraints unchanged."

  ;; Initialize STN properties that support consistency checking and propagation,
  ;; except those local to events and constraints.
  (STN-wide-initialize! stn)

  ;; Initialize the lower and upper bound values and 
  ;; support of all events in STN.
  (dolist (e (STN-events STN))
	(event-initialize! e))

  ;; Initialize the consequences and in-conflict? flags
  ;; of all constraints.
  (dolist (c (constraint-system-constraints STN))
	(STC-initialize! c)
	)
	
  (let ((ub-queue (STN-ub-queue STN))
		(lb-queue (STN-lb-queue STN))
		(start (STN-start STN)))

	
	;; Initialize the upper and lower bound values and support 
	;; of the START event of STN.
	(start-initialize! start)
	;; Re-initialize the upper and lower bound propagation queues.
	;; The first time that upper and lower bounds and consistency 
	;; are computed, propagation emanates from the START event of
	;; STN.
	(queue-insert! start ub-queue)
	(queue-insert! start lb-queue)
	
	T))

;;; ***********************************************************************
;;;
;;;     Propagation of Event Bounds
;;;
;;; ***********************************************************************

;;; To Do:
;;; o Add in statistics computation.
;;;   Include a reset of statistics.

;;; Fix:
;;; o What is the purpose of the call to
;;;     (d-edge-relaxation (STN-events network)),
;;;   added by Peng Yu?
;;;   It looks like this fix is addressing a problem that,
;;;   when edges are removed, the graph is no longer connected.

;;;   However to detect inconsistencies, the start event needs to be 
;;;   connected to all events.

;;;   Rather than removing an edge, Peng is relaxing the bound so
;;;   that propagation goes through the edge.
;;;   Peng is relaxing the bound to a large number, rather than infinity.
;;;   - Does this produce the desired effect?
;;;   - Is the desired effect produced if the bound is instead *infinity*?
;;;     Why?
;;;   - Can we allow edges to be added and removed, and instead add a set of
;;;     un-removable arcs from start to each event, with a bound of infinity?

;;; Optimization:
;;; o Only need to perform both propagations if querying both 
;;;   upper and lower bound values.
;;;   For consistency, just propagate the upper bounds.

;;; Reviewed by BCW.
(defmethod propagate! ((network STN))
  
  "Let NETWORK be a simple temporal network.

   Propagate NETWORK'S upper and lower bound queues
   until there are no queued distance-edges remaining, 
   or an inconsistency is detected.

   Returns T.

   If this is the first time propagation is called, 
   then propagation eminates from the network's start event.

   If this is an incremental update, 
   then all changes have been queued for propagation.

   No event bounds can be believed if network is inconsistent.  

   Whenever this function returns, the upper and lower bound 
   propagation queue has been correctly set."
  
  ;;;  (debug-print-for :itc t "Propagating.")
  
  (cond ((or (stn-ub-consistency-checked? network)
			 (stn-lb-propagated? network))
		 
		 ;; STN has already been checked at least once,
		 ;; via upper and / or lower bound propagation.
		 
		 ;; Perform an incremental update.
		 (debug-print-for :itc t "  Propagating incrementally.")

		 (when (has-event-pairs-queued? network)
		   
		   ;; One or more edges are queued for propagation.
		   
		   (when (STN-ub-inconsistent? network)

			 ;; Network was detected as inconsistent 
			 ;; during a previous upper bound propagation.
			 
			 ;; Remove the conflict (negative loop)
			 ;; before propagating.
			 (debug-print-for :itc t 
							  "    Removing conflict ~a before propagating."
							  (STN-ub-conflict network))

			 (STN-remove-ub-conflict! (STN-ub-conflict network)
									  network)

			 ;; Record that consistency has not been checked, 
			 ;; and hence needs to be checked.
			 (itc-stn-ub-consistency-not-checked! network))	   

		   ;; Process the edges that have been changed
		   ;; since the last propagation, and queue for propagation.
		   (process-changed-edges! network)

		   ;; Then propagate consequences ...
		   
		   (unless (STN-ub-inconsistent? network)
			 
			 ;; If network is not already inconsistent, then propagate.
			 ;; If an inconsistency is found then it is recorded,
			 ;; along with a conflict.
			 (update-ub-temporal-consistency! network))
		   
		   ;; If it remains that an inconsistency (negative loop) 
		   ;; has not been found, 
		   ;; then propagate lower bounds as well.
		   (unless (STN-ub-inconsistent? network)
			 
			 (update-lbs! network))))
		
		(:otherwise
		 
		 ;; We haven't propagated yet since the last initialization.
	 

		 ;; Compute bounds and check consistency from scratch.
		 (debug-print-for :itc t "  Propagating for first time.")
		 
		 ;; Propagate out from the START event. 
		 (update-ub-temporal-consistency! network)
		 
		 ;; If no inconsistency has been found, 
		 ;; Then propagate lower bounds as well.
		 (unless (STN-ub-inconsistent? network)

		   (update-lbs! network))))
 
  T)

;;; Reviewed by BCW.
(defmethod update-ub-temporal-consistency! ((network STN))

  "Let NETWORK be a simple temporal network.

   Propagates STN and updates the record of temporal consistency 
   for STN wrt upper bound propagation, as well as any conflicts 
   for STN that were found during UB propagation."

  ;; Record that UB consistency has been checked for network,
  ;; (we are just about to perform this check),
  (stn-ub-consistency-checked! network)

  ;; Propagate to update bounds and to check consistency.
  ;; If inconsistent, a conflict is returned, else nil.
  (let ((ub-conflict (check-ub-temporal-consistency network)))
	
	(cond ((null ub-conflict)
		   
		   ;; No conflict returned, network is consistent.
		   (debug-print-for :itc t "    Network is UB consistent.")

		   ;; Return T, indicating Consistent.
		   (values T))
		  
		  (:otherwise
		   
		   ;; A conflict was returned, network is inconsistent.
		   
		   (debug-print-for :itc t 
							"    Network is UB Inconsistent: conflict = ~a."
							(itc-conflict-constraints ub-conflict))

		   ;; Record that network is inconsistent.
		   (itc-STN-ub-inconsistent! network)
		   
		   ;; Record on NETWORK the conflict that was found.
		   (setf (STN-ub-conflict network) 
			 ub-conflict)

		   ;; (print-stn network)

		   ;; Return nil, indicating that Network is Inconsistent.
		   (values nil)))))

;;; The first two routines below compute shortest path values and 
;;; consistency of the initial graph.

;;;Conflict
;;;CheckTemporalConsistency()
;;;
;;;// Determine new shortest paths and check consistency, 
;;;// by propagating changes of all vertices on queue.
;;;// Returns conflict if inconsistent, or {} if consistent.
;;;
;;;{07} while !Q.empty()
;;;{08}   u = Q.pop();
;;;{09}   for v in Succ(u)
;;;{10}     dval = Update(u,v);
;;;
;;;{New} Removed Steps 11 to 12b and modified Update(u,v), 
;;;      by adding subtree disassembly to check for 
;;;      negative cycles. - Andrew Yi Huang.

;;;{11}     if(dval) < 0
;;;{12}       if (CompletedCycle(v))
;;;{12b}         return ExtractConf(v, {});
;;;
;;;{13} return {};


;;; Reviewed by BCW.
(defmethod check-ub-temporal-consistency ((STN STN))

  "Let STN be a simple temporal network object.

   Compute updated tightest upper bounds on STN events, 
   and determine consistency.

   Returns nil if simple temporal network STN is consistent, 
   and a conflict, otherwise.

   A conflict is a list of distance edges that form a negative cycle. "

  ;; Q contains all events whose upper bounds
  ;; need to be propagated.  This is because either
  ;; 1) the event's bounds have changed, or
  ;; 2) we need to propagate event to compute a new bound
  ;;    (resupport) for a neighbor.

  ;; Propagate events on Q
  (loop with Q = (stn-ub-queue STN)
	  initially (debug-print-for :itc t "  Processing UB Q, contains ~A." 
								 (queue-elements Q))

	  until (queue-empty? Q)

	  for u = (queue-pop Q)
			  ;; Grab the next event off the Q.
			  
	  for conflict = (update-ub-successors u STN)
					 ;; Propagate event's bounds to its successors.

					 ;; Returns a conflict (list of inconsistent d-edges)
					 ;; when propagation generates a negative loop,
					 ;; or NIL, otherwise.

					 ;; Keep track of some statistics for testing.					 
	  when *itc-statistics-p*
	  do (incf *propagate-counter*)

		 ;; As soon as an inconsistency is found, 
		 ;; abort propagation and return a conflict.
	  unless (null conflict)
	  return conflict
						
			 ;; If propagation finishes without an inconsistency, 
			 ;; indicate success by returning NIL.

	  finally (return '())))


;;; Reviewed by BCW.
(defmethod update-ub-successors ((u event) (STN STN))

  "Let U be a temporal event of simple temporal network STN.

   Propagate event U's bound to its neighboring events
   and successors in STN.

   Returns nil if simple temporal network STN is consistent, 
   and a conflict, otherwise.

   A conflict is a list of distance edges that form a negative cycle."

  (let ((conflict nil))

	(when (eql (event-status u) *labeled*)
	  
	  ;; Event U is supported (indicated by *labeled*).
	  ;; It hasn't yet been propagated or checked for consistency.

	  ;; Propagate it.  
	  
	  ;; Relax every EDGE eminating from u
	  ;; that is currently inserted into STN.
	  (do-inserted-d-edges 
		  (edge (event-successors u))
		
		;;{new} Removed Step 11 to 12b and modified Update(u,v) by adding
		;;      "subtree disassembly."  This is performed to check for 
		;;      negative cycles.
		;;      - Andrew Yi Huang.		
		
		;; Relax EDGE and propagate.
		;; If consistent, conflict is nil,
		;; Else conflict is a list of d-edges that form a negative loop.
		(setq conflict
		  (STN-ub-scan u edge STN))

		(when (not (null conflict))

		  ;; Inconsistent, as indicated by the returned conflict.
		  ;; Conflict is a list of D-edges that form a negative loop.
		  
		  ;; Terminate immediately, without processing the remaining edges.
;;;		  (format t "CONFLICT: ~a~%" conflict)
		  
		  (return T)))
	  
	  ;; Indicate that the bound of U has been propagated,
	  ;; and support of event has been "scanned" for a negative cycle.
	  (setf (event-status u) *scanned*))

	;; Return.  If consistent, return nil,
	;;          If inconsistent, return the discovered conflict.
	(values conflict)))

;;;value
;;;Update(u,v) 
;;;
;;;  // Relax edge u -> v and propagate
;;;  // successors if d(v) changes.
;;;
;;;{14} if ( d(v) > d(u) + c(u,v))
;;;{new}  conflict := DisassembleSubtree(u,v)
;;;{new}  if ( conflict == null )
;;;{15}     d(v) := d(u) + c(u,v);
;;;{16}     p(v) := u;
;;;{new}    s(v) := labeled;
;;;{17}     Q.insert(v);
;;;remove step 18
;;;{18} return d(v);
;;;{new}  return conflict

;;; Lines indicated above by {new} were added by Andrew Yi Huang.

;;; To Do:
;;; o "Scan" is ill named.  Replace with a more intuitive name, 
;;;    such as relax-edge-and-queue-successors.
;;;    Was this previously the function named Update?

;;; o "disassemble-subtree" is not an intuitive name.  
;;;   Replace with a name like check-and-extract-negative-loop.

;;; Reviewed by BCW.
(defmethod STN-ub-scan ((u event) (edge d-edge) (STN STN)) 

  " Let STN be a simple temporal network,
         U be an event of STN whose bound has been updated, and
         EDGE: U -> V be an outgoing D-edge of U, yet to be relaxed. 

   Relax EDGE.

   If the upper bound of event V is tightened,
   Then check consistency (absence of a negative cycle),
   and propagate to successors if consistent.

   If propagation detects a negative cycle, 
   then terminate propagation immediately.  

   Returns a conflict if a negative cycle is detected.
   Else it returns nil.
  
   A conflict is a list of D-edges that form  a negative cycle.

   Side Effect:
     If inconsistent, remove the negative cycle, and 
     set up propagation to resupport when called again."

  ;; Compute a relaxed upper bound for successor event V,
  ;; based on U's bound and edge's weight.

  ;;    V-ub is the old upper bound and 
  ;;    new-V-ub is the new relaxed value.

  (let* ((v (d-edge-end edge))
		 (v-ub (event-ub v))
		 (u-ub (event-ub u))
		 (uv-cost (d-edge-ub edge))
		 (new-v-ub (b+ u-ub uv-cost))
		 (conflict nil))
	
	(when (b> v-ub new-v-ub)
	  
	  ;; When relaxation improved V's upper bound.
	  
	  ;; Check if EDGE, the support for the new bound,
	  ;; creates a negative cycle.  

	  ;; If so, then conflict is a list of d-edges that form a negative cycle.
	  ;; Else it is nil.
	  (setq conflict (disassemble-subtree u edge))

	  (when (null conflict)
		
		;; The improved bound is consistent; it doesn't form a negative cycle.
		
		;; Record the new bound for V and support V with EDGE.

		;; Indicate that V has just been supported, through status *labeled*.
		;; V hasn't yet been propagated or checked for consistency.
		(setf (event-status v) *labeled*)
		(setf (event-ub v) new-v-ub)
		(set-parent v edge)
		
		;; Continue propagation by inserting successor V into the queue.
		;; Don't add V if its already in the queue.
		(when (not (queue-contains v (stn-ub-queue stn)))
		  
		  (queue-insert! v (stn-ub-queue stn))))

	  ;; If inconsistent,
	  ;; Then no support is recorded, and
	  ;; propagation is stopped immediately.
	  )

	;; Return: 
	;; If consistent, return nil for success.
	;; If inconsistency detected, return a conflict.
	;; A conflict is a list of d-edges that form a negative cycle.
	(values conflict)))

;;; ***********************************************************************
;;;
;;;     Negative Cycle Detection, Extraction and Removal
;;;
;;; ***********************************************************************

;;; To Do:
;;; o Once cleaned up, move to conflicts.lisp.
;;; o Document and rename.  
;;;   - What does a tree represent? The set of support?  
;;;   - What does it mean to disassemble a subtree?
;;;   - What do u and v denote?
;;;   - Names of functions called for side effect should end with !.
;;;   - What does it mean to set the parent?
;;; o Arguments are incorrect, they are u and edge, not v.

;;;conflict
;;;DisassembleSubtree(u,v) 
;;;
;;;  // Use subtree disassembly to check
;;;  // negative cycles.
;;;
;;;{new} subtreeElements := GetSubtree(v);
;;;{new} if (u is a member of subtreeElements)
;;;{new}   p(v) := u;
;;;{new}   conflict := ExtractConf(v, {});
;;;{new}   p(v) := nil;
;;;{new}   s(v) := unreached;
;;;{new} for e in subtreeElements
;;;{new}   p(e) := nil;
;;;{new}   s(e) := unreached;

;;; o By Andrew Yi Huang.

;;; To Do:
;;; o Rename.  This name is not intuitive.

;;; Reviewed by BCW.
(defmethod disassemble-subtree ((u event) (edge d-edge))

  "Let EDGE be a distance edge, starting at event U,
   which is being relaxed.
   Edge has been determined to tighten the bound of an event V,
   but the update has not yet been made to V and support has not
   yet been recorded.

   Check for a negative cycle involving EDGE.

   Returns a conflict if a cycle exists, else nil.
   A conflict is a list of d edges that form a negative cycle.

   Side Effect:  
     If an inconsistency is found, 
     Removes as support, the edges in the conflict and all its successors.
     Updates the status of unsupported events as *unreached*
     and sets up propagation queue for resupport."

  ;; Edge is from u to v.
  ;; Subtree-elements is a list of all successors of v that appear
  ;;   in V's support tree.
  (let* ((v (d-edge-end edge))
		 (subtree-elements (get-subtree v))
		 (conflict nil))

	(when (member u subtree-elements)
	  
	  ;; A negative loop has been detected.
	  ;;   U, a predecessor of V, is also a successor of V,
	  ;;   hence a loop exists.
	  ;;   In addition the cycle is negative,
	  ;;   since it resulted in a tighter bound for u during relaxation.
	  
	  ;; Record EDGE as the support for V to make explicit the loop, and
	  (set-parent v edge)

	  (debug-print-for :itc t "  Extracting conflict, starting at ~a." 
					   u)

	  ;; construct a conflict, by tracing support from predecessor (parent) U.
	  (setq conflict 
		(extract-ub-conflict u))

	  ;; Clear out the inconsistency from the set of support.
	  
	  ;; Remove support for V and all its successors,
	  ;; and indicate that V and all its successors are unreached by propagation.
	  (remove-parent v)
	  (setf (event-status v) *unreached*))
	
	(dolist (element subtree-elements)
	  
	  (remove-parent element)
	  (setf (event-status element) *unreached*))
	
	;; Return: 
	;; If consistent, return nil for success.
	;; If inconsistency detected, return a conflict.
	;; A conflict is a list of d-edges that form a negative cycle.
	(values conflict)))

;;; To Do:
;;; o Rename to set-of-support-successors.

;;; Optimization:
;;; o Its more efficient to identify duplicates as we go,
;;;   by passing the accumulated list along.
;;; o Its also more efficient to detect duplicate elements and
;;;   subtrees by marking events.
;;; o Do we need to return a list at all, 
;;;   or just check the visited events for some property?

;;; Reviewed by BCW.
(defmethod get-subtree ((v event))

  "Let V be an event.

   Trace forward through the successor support d-edges of V.

   Returns a list of all successor events of V
   (without duplicates)."

  (remove-duplicates
   (flatten
	(event-ub-children v))))

;;; Optimization:
;;; o Flatten unnecessarily conses.
;;; o Iteration over the elements of list is probably faster
;;;   than the tail recursion.

;;; Reviewed by BCW.
(defun flatten (L)
  
  "Let L be a list of events of a simple temporal network.

   Traces forward through the support of the events in L to their successors.

   Returns a list comprised of events and all their successors.

   Note: The returned list may contain duplicate elements."

  (if (null L)
	
	  '()
	
	(let ((1st (first L))
		  (remainder (rest L)))
	  
	  (append 
	   
	   (list 1st)

	   ;; Add all successors of the 1st element,
	   (flatten 
		(event-ub-children 1st))
			  
	   ;; to the successors of the remaining elements of L.
	   (flatten remainder)))))


;;; ***********************************************************************
;;;
;;;     Adding and Removing Support
;;;
;;; ***********************************************************************

;;; The purpose of the next two functions is to add and remove
;;; support.

;;; Let E be a D Edge from U to V,
;;; where E is used to determine the upper bound for V.
;;; E is said to be the support of V,
;;; U is the parent of V, 
;;; and V is one of U's children.

;;; FIX:
;;; o Why does this set the ub-consequent,
;;;   when remove-parent doesn't remove it?
;;;   What is the consequent?   Is it only
;;;   set when edge acts as a support?
;;;   Why is it required?

;;; To Do:
;;; o Give a more suitable name, such as set-support!.

;;; Optimization:
;;; o The call to remove does a lot of consing.
;;;   Replace with the destructive operation delete,
;;;   if no negative side effects.

;;; Reviewed by BCW.
(defmethod set-parent ((v event) (edge d-edge))
  
  "Let EDGE be a distance edge from event U to event V,
   where EDGE was used to derive the current upper bound for V.

   U is considered the parent of V (predecessor) and V a child of U (successor).

   Record Edge as the support for V and add U as a child of V.

   Remove any previous support for V if it exists."

  (let ((u (d-edge-start edge))
		(old-support (event-ub-support v)))
	
	(unless (null old-support)

	  ;; When V is already supported by some edge, remove this support.
	  ;;   Remove edge as a support for V,
	  ;;   remove V as a consequent of edge,
	  ;;   and remove V as a child of the start event of edge.
	  (let ((old-u (d-edge-start old-support)))

		(setf (event-ub-children old-u) 
		  (remove v (event-ub-children old-u)))))
	
	;; Record EDGE as the new support for v,
	(setf (event-ub-support v) edge)

	;; add V as a child (immediate successor) of U, and
	(setf (event-ub-children u) 
	  (add-to-list (event-ub-children u) v))

	;; record v as the consequent of support EDGE.
	(setf (d-edge-ub-consequent edge) v))
  
  T)

;;; Check:
;;; o Why not clear ub-consequent?  Looks like ub-consequent is not
;;;   maintained properly.  Is it still used?
;;; o Why does this set the support to null when its already null?
;;;   It doesn't seem needed.

;;; To Do:
;;; o Give a more suitable name, such as remove-support!.

;;; Reviewed by BCW.
(defmethod remove-parent ((v event))

  "Let V be an event, possibly supported by a D-edge from P to V.

   If D-edge exists,
   Then P is the parent of V and
        V is a child of V.

   Remove V's support, and
   Remove V from the children of P."

  (unless (null (event-ub-support v))
	
	;; Event V is supported,
	
	;; Remove V from the children of its parent, according to V's support.
	(let ((P (d-edge-start (event-ub-support v))))
	
	  (setf (event-ub-children P)
		(remove v (event-ub-children P)))))
  
  ;; Clear Event V's support.
  (setf (event-ub-support v) nil))


;;; The following function is similar to adjoin,
;;; but adds elements to the list's end.

;;; To Do:
;;; - If elements were added to the start of the list, 
;;;   this would be the same as adjoin.  Can adjoin be used?

;;; - The call setf appears to serve no function, 
;;;   and obscures the fact that list is being returned.

;;; Optimization.  
;;; - This copies list L after every addition, and is 
;;;   very space inefficient.

;;; Reviewed by BCW.
(defmethod add-to-list (L value)

  "If VALUE is not already a member of list L, 
   Adds VALUE to the end of L,
   Returns updated list.

   Operation is non-destructive."
  
  (or (member value L)
	  (setf L 
		(append L (list value)))))

;;; ***********************************************************************
;;;
;;;     Update after Insertion and Removal of Distance Edges
;;;
;;; ***********************************************************************

;;;  In the following, edge insertions and removals are all accumulated,
;;;  and then the best edge replacement is made when propagate is called.
;;;  This helps limit unnecessary unsupport/resupports.

;;; Reviewed by BCW.
(defmethod process-changed-edges! ((stn stn))
  
  "Let STN be a simple temporal network.

   Update edge changes that have been made to STN since the last update,
   where each changed edge is update with its tightest current bound.

   Sets up queues to propagate all edge changes."
  
  (cond ((has-event-pairs-queued? stn)

		 ;; Edge changes queued.  Process them.
		 
		 (debug-print-for :itc t "    Processing changed event pairs.")
		 
		 (loop while (has-event-pairs-queued? stn)
					 
					 ;; Process the next pair <x, y>
			 for event-pair = (dequeue-changed-event-pair stn)
			 for x = (event-pair-start event-pair)
			 for y = (event-pair-end event-pair)
					 
					 ;; Get the current tightest upper and lower edge bound weights
					 ;; for pair.
			 for ub-new-edge = (event-pair-lowest-bound-inserted-d-edge 
								event-pair)
							   
			 for lb-new-edge = (event-pair-highest-bound-inserted-d-edge 
								event-pair)
							   
							   ;; Update the upper and lower bound edge weights.
			 do (modify-d-edge-ub! ub-new-edge x y stn)
				(modify-d-edge-lb! lb-new-edge y x stn)))
		
		(:otherwise
		 
		 ;; No edge changes queued.  Do nothing.
		 
		 (debug-print-for :itc t "    No changed event pairs to process.")))
  
  T)

;;; From the original ITC pseudo code (does not precisely reflect this 
;;; implementation).

;;;void
;;;ModifyConstraint(x,y,l,u)
;;;
;;;   // Change the bounds [l,u] of simple temporal 
;;;   // constraint between x and y.
;;;  
;;;{24} ModifyArc(arc(y,x),-l)
;;;{25} ModifyArc(arc(x,y),u)

;;;(defun modify-constraint! (x y l u STN)
;;;
;;;  "Change the bounds [l,u] of simple temporal 
;;;    constraint between x and y."
;;;
;;;  (STN-modify-arc! y x (b- l) STN)
;;;  (STN-modify-arc! x y u STN))

;;;Void => changed to conflict
;;;ModifyArc(arc,c)
;;;
;;;  // Change to c the edge weight of arc,
;;;  // and queue relevant vertices for update.
;;;
;;;{26} setCost(arc,c);
;;;{27} if (d(arc.head) > d(arc.tail) + c) 
;;;       // Change improves shortest path
;;;       // Update and propagate.
;;;{new}  conflict := DisassembleSubtree(u,v)
;;;{new}  if (conflict == null)
;;;{28}     d(arc.head) := d(arc.tail) + c;
;;;{29}     p(arc.head) := arc.tail;
;;;{new}    s(arc.head) := labeled;
;;;{30}     Q.Insert(arc.head);
;;;
;;;{31} elseif (d(arc.head) < d(arc.tail) + c
;;;            AND (p(arc.head) == arc.tail)) 
;;;       // Change worsens shortest path,
;;;       // Reset value.
;;;
;;;{32}   d(arc.head) := infinity;
;;;{33}   p(arc.head) := unknown;
;;;
;;;       // Reset consequences.
;;;{34}   set<Node> nodes_reset = {};
;;;{35}   nodes_reset.insert(arc.head);
;;;{36}   for all n in InvalidateSupports(arc.head)
;;;{36}     nodes_reset.insert(n)
;;;
;;;       // Setup to resupport.
;;;{37}   InsertParents(nodes_reset)
;;;{new}return conflict;

;;; Reviewed by BCW.
(defmethod modify-d-edge-ub! (new-edge (x event) (y event) (STN STN))

  " Let NEW-EDGE be a new edge between events X and Y of
    simple temporal network STN with the tighest upper bound.

    Incrementally relax NEW-EDGE.  The result may either tighten,
    loosen or have no effect on the bound of NEW-EDGE's END event.  
    When END event changes, queue for propagation.

   Returns T.

    Note: This function has no effect if shortest paths have
          not previously been computed, since no incremental
          change is required.  Instead the shortest paths are
          computed from scratch the first time STN is queried."

  (debug-print-for :itc t "      Relax new-edge ~a, starting UB propagation."
				   new-edge)

  (if (null new-edge)  
	  
	  (when (and (event-ub-support y)
				 (eql x
					  (d-edge-start (event-ub-support y))))
		
		(weaken-d-edge-ub! y stn))
	  
	(when (STN-ub-consistency-checked? STN)
	  
	  ;; Only perform incremental update if upper bound consistency of the STN
	  ;; has already been checked at least once.

	  ;; Note: The caller checks that upper or lower bound propagation has
	  ;;   been performed, but not specifically upper bound propagation.
		
	  (let* ((new-edge-ub (d-edge-ub new-edge))

			 ;; x-ub and y-ub are the old upper bounds on x and y,
			 (x-ub (event-ub x))
			 (y-ub (event-ub y))
			 
			 ;; while new-y-ub is the new, relaxed upper bound on y.
			 (new-y-ub (b+ x-ub new-edge-ub))
			 (conflict nil))
		  
		(cond ((b> y-ub new-y-ub)
			   
			   ;; Relaxation with the new edge tightens Y's upper bound.
			   
			   ;; Propagate this tighter bound,
			   ;; which checking for a conflict (negative cycle).
			   (setq conflict
				 (strengthen-d-edge-ub! new-edge y new-y-ub y-ub stn))

			   (unless (null conflict) 
				 
				 ;; A conflict was found,

				 ;; Record that an upper bound inconsistency has been found
				 ;; and record the found conflict.
				 (itc-STN-ub-inconsistent! STN)
				 (setf (STN-ub-conflict STN) conflict)))
				
			  ((and (b> new-y-ub y-ub)
					(and (not (null (event-ub-support y)))
						 (eql x (d-edge-start (event-ub-support y)))))
			   
			   ;; Edge previously supports Y, and 
			   ;; relaxation using the new edge weight 
			   ;; weakens Y's upper bound.

			   ;; Propagate this weaker bound.
			   ;; Note that weakening cannot introduce an inconsistency
			   ;; that didn't previously exist, hence we don't need to
			   ;; consider consistency.
			   (weaken-d-edge-ub! y stn))

			  ;; Otherwise, bound is the same as before,
			  ;; leave support and bound unchanged.
			  
;;;				(:otherwise
;;;				 ;; Change keeps the bound on y the same.
;;;				 (unless (eql new-edge (event-ub-support y))
;;;				   (setf (event-ub-support y) new-edge)))
			  
			  ))))
  
  T)

;;; Reviewed by BCW.
(defmethod weaken-d-edge-ub! ((y event) (stn STN))

  "Let Y be an event of simple temporal network STN
   whose bound was weakened as a result of an edge
   weight change to Y's current support.

   Look to compute the best new bound for Y.
   Relax Y and its consequences to infinity,
   then find a new bound for Y by re-relaxing
   all edges terminating on Y, and propagating forward. "

  (debug-print-for :itc t "      Edge update weakens the UB of ~a; reset." y)

  (let ((events-reset (list y)))

	(debug-print-for
	 :itc t "        Reset to infinity the UB of ~a and its consequences." y)

	;; Reset the upper bound value of y to infinity
	;; and remove its support.
	(event-ub-initialize! y)

;;; FIX: Added status initialization.  Delete this comment when this works.	
	;; Clear status to *unreached*.
	(event-status-initialize! y)

	;; Similarly reset all consequences of y.
	(setq events-reset
	  (reset-all-ub-consequences y STN events-reset))

	;; Finally, setup ub propagation queue to compute new
	;; bounds for y AND its consequences.
	(queue-predecessors-for-ub-propagation! events-reset STN)

	T))


;;; FIX:
;;; o This could be buggy:
;;;   - If an inconsistency is found, then make sure that the STN is updated
;;;     properly, and the queue is set up for further propagation.
;;;   - Make sure that this function mirrors STN-ub-scan.

;;; Reviewed by BCW.
(defmethod strengthen-d-edge-ub! ((new-edge d-edge) 
								  (y event)
								  new-y-ub
								  y-ub
								  (stn STN))
  
  "Let Y be an event of simple temporal network STN
   whose bound was strengthened from Y-UB to NEW-Y-UB 
   as a result of a weight change to NEW-EDGE.

   Check if adding NEW-EDGE as a support introduces
   an inconsistency (negative cycle) and propagate the bounds.

   If consistent, queue Y for propagation.

   Returns a conflict if a negative cycle is detected.
   Else it returns nil.
  
   A conflict is a list of D-edges that form  a negative cycle.

   Side Effect:
     If inconsistent, removes the negative cycle, and 
     sets up propagation to resupport when called again."

  ;; Change improved the bound on y. 
  ;; Update y, check for inconsistency and
  ;; if consistent, queue y to propagate.
  (let* ((conflict nil))
	
	(debug-print-for :itc t "      Edge ~a strengthens UB of ~A from ~A to ~A."
					 new-edge y new-y-ub y-ub)

	;; Determine whether adding new-edge as a support generates a negative cycle.
	(let ((x (d-edge-start new-edge)))
	  (setq conflict (disassemble-subtree x new-edge)))
	
	(when (null conflict)

	  ;; Record Y's new bound and its support.
	  (setf (event-ub y) new-y-ub)
	  (set-parent y new-edge)

	  ;; Indicate that Y has just been supported, through status *labeled*.
	  ;; It hasn't yet been propagated or checked for consistency.
	  (setf (event-status y) *labeled*)

	  ;; Queue to propagate Y's bound to its consequences.
	  (queue-insert! y (STN-ub-queue STN)))

	;; Return conflict if inconsistent, or nil if consistent.
	;; Conflict is a list of d-edges that form a loop.
	(values conflict)))


;;;The next two routines implement an arc weight change, by
;;;removing support and setting up propagation, as appropriate.

;;;set<Node>
;;;reset-all-ub-consequences(Node n) 
;;;
;;;  // Remove n as a support for any s,
;;;  // and queue s's predecessors for resupport.
;;;  // Returns list of nodes whose d and p values
;;;  // have been reset.
;;;
;;;{42} set<Node> nodes_reset = {};
;;;{43} for all s in Succ(n)
;;;{44}   if( p(s) == n )
;;;         // Successor s is supported by n
;;;         // since n reset, reset s as well.
;;;
;;;{45}     if( s == sstart )
;;;{46}       d(s) := 0;
;;;{47}       for all t in reset-all-ub-consequences(s)
;;;{47a}        nodes_reset.insert(t);
;;;
;;;{48}     elseif( d(s) != infinity OR p(s) != unknown )
;;;{49}       d(s) := infinity;
;;;{50}       p(s) := unknown;
;;;{51}       for all t in reset-all-ub-consequences(s)
;;;{51a}        nodes_reset.insert(t);
;;;
;;;{52} return nodes_reset;

;;; The above pseudo code does not correctly accumulate the
;;; reset nodes.  It instead needs to insert the s nodes that
;;; have been reset.  This is corrected in the function below.

;;; Check:
;;; o Consequent is being cleared here, but the functions 
;;;   that record support - stn-ub-scan and strengthen-d-edge-ub! -
;;;   do not seem to be setting up consequences properly.
;;;   Check this through out.  Look into what depends on consequent:
;;;   Does conflict extraction depend on it?
;;; o Is event-ub-visited! mark used?  Is it properly maintained?
;;;   If no longer used, then delete.
;;; o clear-event-ub-support! has been changed so that it calls
;;;   remove-parent.  Is this correct?
;;; o Think through the "useless" condition.  If this case is evaluted
;;;   then Y most be supported, and "useless" will never evaluate to true.
;;;   What was intended?  If event has support but its upper bound is
;;;   *infinity*, should it and its consequences be reset?  Can this
;;;   case occur?

;;; Reviewed by BCW.
(defmethod reset-all-ub-consequences ((x event) (STN STN) events-reset)

  "Let X be an event of simple temporal network STN 
         whose upper bound value has been reset, and
       EVENTS-RESET be a list of events already reset.

   Reset all event upper bound values and support for events 
   that are consequences of X.

   Returns EVENTS-RESET, augmented with all additional
   events that were reset within this function call."
  
  ;; Remove consequences of X by recursively
  ;; looping through the immediate consequences of X.
  ;; Collect and return all reset events.
  (when (event-successors x)
	
	;; One or more successor D-edges exist.
	
	;; Reset those successor edges that are currentlyact as a support.
	(loop
		for edge in (event-successors x)
		for y = (d-edge-end edge)
				
		when (eql (event-ub-support y) 
				  edge)
			 
			 ;; Event y is supported by x through EDGE;
			 
			 ;; Initialize y and its consequences,
			 ;; while collecting the reset events.

		do 
		  ;; First, initialize Y. 

;;; FIX:  Moved to the two cond cases below.  Delete this comment and
;;;       commented out code once this works.
		  ;; Clear Y as the consequent of EDGE.
;;;		  (clear-d-edge-ub-consequent! edge)
		 
		  (cond ((eql y (STN-start STN))

				 ;; Y is the start event.
				 (debug-print-for
				  :itc t "        Reset to 0 the UB of start ~a and to ~
                                   infinity its consequences." y)

				 ;; Reset Y's bound to 0,
				 ;; mark Y as visited.
				 ;; remove EDGE as the support for Y, and
				 (start-ub-initialize! y)
				 
				 ;; Clear Y as the consequent of EDGE.
				 (clear-d-edge-ub-consequent! edge)
				 
;;; FIX: Added status initialization.  Delete this comment when this works.	
				 ;; Clear status to *unreached*.
				 (start-status-initialize! y)

				 (push y events-reset)
				  
				 ;; Reset all consequences of y, collecting the reset events.
				 (setq events-reset
				   (reset-all-ub-consequences y STN events-reset)))
			   
				((not (event-ub-useless? y))
				  
				 ;; Y is a non-start event.
				 (debug-print-for
				  :itc t "        Reset to infinity the UB of ~a and its ~
                                   consequences." y)

				 
				 ;; Reset Y's bound to infinity,
				 ;; mark Y as NOT visited.
				 ;; remove EDGE as the support for Y, and
				 (event-ub-initialize! y)

				 ;; Clear Y as the consequent of EDGE.
				 (clear-d-edge-ub-consequent! edge)
				 
;;; FIX: Added status initialization.  Delete this comment when this works.	
				 ;; Clear status to *unreached*.
				 (event-status-initialize! y)

				 (push y events-reset)
				  
				 ;; Reset all consequences of y, collecting the reset events.
				 (setq events-reset
				   (reset-all-ub-consequences y STN events-reset)))
			   
				(:otherwise
				 
				 ;; Y's upper bound is *infinity*, and it has no support.
				 ;; Don't need to reset.
				 
				 nil))
		 
		finally (return events-reset))))

;;;void
;;;queue-predecessors-for-ub-propagation!(set<Node> reset_nodes) 
;;;
;;;  // rest_nodes lost their support, queue predecessors
;;;  // to resupport.
;;;
;;;
;;;{38} for all n in reset_nodes
;;;{39}   for all m in Pred(n)
;;;{40}     if( d(n) != infinity || p(s) != unknown )
;;;{41}       Q.Insert(m);

;;; Check: 
;;; o L40, changed d(n), p(s) to d(m), p(m).
;;; o Should only need to check if unbounded, 
;;;   not if unsupported, yes?
;;;   Unsupported but bounded should be an error.

;;; Check:
;;; o This will try to propagate an *infinity* bound for Y 
;;;   if Y is supported.  Can this case occur?  How?  
;;;   Should we allow for this case?
;;; o Confirm that an event is checked for consistency when its
;;;   popped off the queue.
;;; o Is it cause a problem that the start event has no support.
;;;   Should it be supported?

;;; FIX:
;;; o Start events have no predecessors.  Should they be queued, in lieu
;;;   of their predecessors?  If start is queued this way, then we shouldn't 
;;;   need Andrew Yi Huang's test of an empty queue.  Implement and confirm.

;;; Reviewed by BCW.
(defmethod queue-predecessors-for-ub-propagation! (reset-events (STN STN)) 

  "Let RESET-EVENTS be events that have lost their support and 
   have had their bound reset to *infinity* (or 0 if the STN start event.

   Prepare to compute a new bound for each reset event by
   queuing the event's predecessors for UB propagation."

  (let ((ub-q (STN-ub-queue STN) ))
	
	(unless (null reset-events)
	
	  ;; Queue for propagation the predecessor of each reset event Y,
	  ;; unless the predecessor imposes no constraint
	  ;; (i.e., is an unsupported event with a bound of *infinity*).
				 
	  ;; A predecessor w is the start of an edge whose end is Y.
				 
	  (loop 
		  for y in reset-events
		  do (debug-print-for 
			  :itc t "      Queue predecessors of ~a for UB propagation..." y)
		 
			 (loop for edge in (event-predecessors y)
				 for w = (d-edge-start edge)

				 unless (event-ub-useless? w)

						;; not a useless event.  Useless events have an UB of
						;; *infinity* and no support.
				 do (debug-print-for :itc t "     ~a" w)

					;; Status *labeled* indicates that W has just been supported.
					(setf (event-status w) *labeled*)

					;; It hasn't yet been propagated or checked for consistency.
					;; This occurs when its taken off the queue.
					(queue-insert! w ub-q))

;;; FIX Delete the following when the code works.
;;;     The following doesn't provide the desired effect.  It checks the queue
;;;     at every iteration, when we only want it to check the queue at the end.			 
			 ;; NEW: 
			 ;; If after resetting events, the queue is empty,
			 ;; then reset the start event and put it on the queue for propagation.

			 ;; The queue will be empty when all events appear in a negative cycle.
			 ;; - Andrew Yi Huang.
;;;		do (when (null (queue-elements ub-q))
;;;
;;;			 (let ((start (stn-start STN)))
;;;
;;;			   (start-ub-initialize! (stn-start STN))
;;;;;; FIX: Replaced with above.  Delete when works.			 
;;;;;;			 (setf (event-ub (stn-start STN)) 0)
;;;
;;;			   ;; Indicate that START is ready for a consistency check
;;;			   ;; and propagation, by recording status *labeled*.
;;;			   (start-status-initialize! start)
;;;;;; FIX:  Replaced with above.  Delete when works.			 
;;;;;;			 (setf (event-status (stn-start STN)) *labeled*)
;;;			 
;;;			   ;; Start hasn't yet been propagated or checked for consistency.
;;;			   (queue-insert! start ub-q)))
			 ))

	;; FIX:
	;; o We should be able to get the same effect as Andrew's patch of
	;;   queueing the start event, by detecting when
	;;   the start event is popped off reset-events, and
	;;   then queue this start event, rather than its predecessors.
	;;   This is conceptually simpler, and a more general solution to
	;;   the problem that Andrew was trying to fix.  Implement.
	
	;; NEW: 
	;; If after resetting events, the queue is empty,
	;; then reset the start event and put it on the queue for propagation.

	;; The queue will be empty when all events appear in a negative cycle.
	;; - Andrew Yi Huang.
	(when (queue-empty? ub-q)

	  (let ((start (stn-start STN)))

		(start-ub-initialize! (stn-start STN))
;;; FIX: Replaced with above.  Delete when works.			 
;;;			 (setf (event-ub (stn-start STN)) 0)

		;; Indicate that START is ready for a consistency check
		;; and propagation, by recording status *labeled*.
		(start-status-initialize! start)
;;; FIX:  Replaced with above.  Delete when works.			 
;;;			 (setf (event-status (stn-start STN)) *labeled*)
			 
		;; Start hasn't yet been propagated or checked for consistency.
		(queue-insert! start ub-q)))
	))

;;; ***********************************************************************
;;;
;;; Simpler Implementation of Update after Insertion and Removal of D Edges
;;;
;;; ***********************************************************************

;;; The following three functions implement the same functionality
;;; as modify-d-edge!.  They are simpler in that they perform
;;; incremental changes, as separate insertions and removals of d-edges.
;;; They are less efficient in that removed edge loose their support
;;; and consequences, and then new, potentially, stronger inserted
;;; edges are added and propagated.

;;; Note: not used.
;;;(defun relax-inserted-d-edge! (edge)
;;;
;;;  " Incrementally relax inserted EDGE to tighten 
;;;    the bound of edge's end event.  When successful, queue
;;;    successors for propagation.
;;;
;;;    Note: This function has no effect if shortest paths have
;;;          not previously been computed, since no incremental
;;;          change is required.  Instead the shortest paths are
;;;          computed from scratch upon the first query."
;;;
;;;  (let ((STN (stc-network (d-edge-constraint edge))))
;;;	(when (STN-ub-consistency-checked? STN)
;;;
;;;	  ;; Only perform incremental update if consistency of the STN
;;;	  ;; has already been checked at least once.
;;;
;;;	  (let* ((x (d-edge-start edge))
;;;			 (y (d-edge-end edge))
;;;			 (ub (d-edge-ub edge))
;;;			 (x-ub (event-ub x))
;;;			 (y-ub (event-ub y))
;;;			 (r-y-ub (b+ x-ub ub)))
;;;
;;;		;; y-ub is the old upper bound on y, while
;;;		;; r-y-ub is the new, relaxed upper bound on y.
;;;		
;;;		(when (b> y-ub r-y-ub)
;;;		  
;;;		  ;; Change improves shortest path.
;;;		  ;; Update and propagate.
;;;
;;;		  (debug-print-for :itc t "Inserted edge ~a improves ub of event ~A ~
;;;                          from ~A to ~A, propagate."
;;;					   edge y y-ub r-y-ub)
;;;
;;;		  (setf (event-ub y) r-y-ub)
;;;		  (setf (event-ub-support y) edge)
;;;		  (queue-insert! y (STN-ub-queue STN)))))))

;;; Note: not used.
;;;(defun resupport-consequence-of-removed-d-edge! (edge)
;;;
;;;  " If removed EDGE supports its end event, try
;;;    to resupport END, updating and setting up
;;;    the ub-propagation queue as required."
;;;
;;;  (let ((STN (stc-network (d-edge-constraint edge))))
;;;	(when (STN-ub-consistency-checked? STN)
;;;
;;;	  ;; Only perform incremental update if consistency of the STN
;;;	  ;; has been checked at least once.  Otherwise, bounds are
;;;	  ;; computed from scratch upon query.
;;;
;;;	  (let ((y (d-edge-end edge)))
;;;
;;;		(when (eql (event-ub-support y)
;;;				   edge)
;;;		  ;; Edge supports y, then
;;;		  ;; try to find an alternative support that results in
;;;		  ;; the identical upper bound value for y.
;;;
;;;		  (let ((success? (replace-with-equivalent-support y edge)))
;;;			(unless success?
;;;			
;;;			  ;; Couldn't find equivalent alternative support.
;;;
;;;			  ;; Reset the upper bound value of y to infinity.
;;;			  (event-ub-initialize! y)
;;;
;;;			  ;; and reset all consequences of y.
;;;			  (let ((nodes-reset '()))
;;;				(push y nodes-reset)
;;;				(loop for n in (reset-all-ub-consequences y STN)
;;;					do (push n  nodes-reset))
;;;
;;;				;; Finally, setup ub propagation queue to compute the
;;;				;; bounds for y AND its consequences.
;;;				(queue-predecessors-for-ub-propagation! nodes-reset STN))
;;;			  ))))))
;;;  T)

;;; Note: not used.
;;;(defun replace-with-equivalent-support (y edge)
;;;  
;;;  "Let Y be an event that is supported by EDGE.  
;;;   Try resupporting y with an alternative inserted 
;;;   edge whose relaxtion results in the same upper 
;;;   bound value for Y.
;;;   Returns True if and only if success."
;;;  
;;;  (let ((ub (d-edge-ub edge)))
;;;	(loop for new-edge in (event-predecessors y)
;;;		do (when (and (not (d-edge-removed? edge))
;;;					  (not (eql new-edge edge))
;;;					  (eql ub (d-edge-ub new-edge)))
;;;			 (set-ub-support-relation! y new-edge)
;;;			 (return T))
;;;		finally (return nil))))
