;;; -*- Mode:Common-Lisp; Package:ITC; Base:10 -*-

;;; Copyright (c) 2008-2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :ITC)

;;; ***********************************************************************
;;;
;;;                Simple Temporal Network and Distance Graph Definitions
;;;
;;; ***********************************************************************

;;; This file contains basic objects for specifying Simple Temporal Networks
;;; (STN) and their corresponding distance graph (DG).

;;; These objects do not contain methods for performing inference on STNs 
;;; and DGs.  These are defined by classes that specialize the classes 
;;; defined here, and are contained in defs.lisp.
;;; 
;;; An STN is comprised of a a set of EVENT variables and 
;;; simple temporal constraints (STCs) between events.
;;; 
;;; An STC is a tuple <START, END, LB, UB>, denoting the interval constraint:
;;;
;;;      (START - END) in [LB,UB]
;;;
;;; where START and END are events, and LB and UB are reals, such that
;;;
;;;   LB <= UB.

;;;  LB may be *-infinity* while UB maybe *infinity*.
;;;
;;; The Constants, Variables, Objects and accessors defined in this file are:
;;;
;;; Constants:
;;;   NONE

;;; Variables:
;;;
;;;   *DEFAULT-STN* : a variable that holds the most recently created 
;;;                   simple temporal network (STN).
;;;
;;; Objects:

;;;   This file defines objects for STN, EVENT, and STC, 
;;;   and associated slot accessors:

;;;   STN object slots include:
;;;
;;;     NAME                  Lisp object denoting the theory's name.
;;;     START                 The start event of STN.  All events must
;;;                           be constrained (implicitly or explicitly)
;;;                           to occur AFTER (i.e., not before) START.
;;;     EVENTS                A list of event variable objects.
;;;     EVENT-HT              A hash table that maps event names 
;;;                           to event objects.
;;;     CONSTRAINTS           List of all constraints associated with the
;;;                           STN, including constraints that are 
;;;                           currently in the STN, and constraints that 
;;;                           are currently removed.
;;;
;;;   EVENT object slots include:
;;;
;;;     NAME                : Arbitary lisp object denoting EVENT's name.
;;;                           Typically a list.  An STN may have at most one
;;;                           event of a given name.
;;;     NETWORK             : The simple temporal network (STN) that EVENT is 
;;;                           part of.
;;;
;;;     LB
;;;     UB                  : Real values, denoting the current upper and lower
;;;                           bound, respectively, on the time of event.

;;;   SIMPLE TEMPORAL CONSTRAINT (STC) object slots include:

;;;     SOURCE               :Any Lisp object, specifies the source of 
;;;                           constraint.  This may, for example, be an object
;;;                           that constraint is part of or the type of the 
;;;                           constraint.
;;;     START
;;;     END                 : start and end events of the STC.
;;;
;;;     LB
;;;     UB                  : Real values, denoting the upper and lower bound
;;;                           of the STC, respectively.

;;; ***********************************************************************
;;;
;;;                To Do:
;;;
;;; ***********************************************************************

;;; ***********************************************************************
;;;
;;;                Constants
;;;
;;; ***********************************************************************

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defconstant *unknown* nil)

;;; FIX:
;;; - Delete this commented out code as soon as tested.
;;;   Moved to itc-stn-defs.lisp.  
;;;     
;;;NEW: Each node now has a status of {unreached, labeled, scanned}.
;;;     It is a variation of the label correcting operations.
;;; - Added by Andrew Yi Huang.
  
;;;  (defconstant *unreached* 'unreached
;;;	"The unreached status of an event.")
;;;
;;;  (defconstant *labeled* 'labeled
;;;	"A bound for event has been computed The labeled status of an event.")
;;;
;;;  (defconstant *scanned* 'scanned
;;;	"The scanned status of an event.")
  
  )


;;; ***********************************************************************
;;;
;;;                Variables
;;;
;;; ***********************************************************************

(defvar *default-STN* nil
  "The most recently created simple temporal network STN.")


;;; ***********************************************************************
;;;
;;;                Simple Temporal Network (STN) Objects
;;;
;;; ***********************************************************************

;;;   Object holds all information about a simple temporal network STN.
;;;   An STN is a set of EVENTS and a set of simple temporal constraints 
;;;   STCs on events.  Each event is a real valued variable.  A simple 
;;;   temporal constraint on events A, B is specified by interval [lb,ub], 
;;;   and is satisfied iff (A - B) in [lb,ub].

(defclass STN (essential-constraint-system-with-explanation)

  ((name
	:initform (gentemp "STN-"))

   ;;;Added by Peng; 
   (start-event-name
	:initarg :start-event-name
	:initform nil)

   (start 
	:accessor STN-start
	:initform nil
	:documentation 
	"START is a distinguised event of STN, denoting the beginning of time.
    Every event must be constrained so that it does not appear before START.")

   (events 
	:accessor STN-events
	:initform '()
	:documentation "EVENTS is the list of all event objects associated with 
                   STN, where an event is a real-value variable denoting the 
                   time when event occurs.")
  
   (event-ht 
	:accessor STN-event-ht
	:initform (make-hash-table :test #'equal)
	:documentation 
	"EVENT-HT holds the hash table that maps EVENT names to event objects.  
     This is used when simple temporal constraints are created, to map 
     quickly from event names to their corresponding object.")

   (event-pair-ht
	:accessor STN-event-pair-ht
	:initform (make-hash-table :test #'equal)
	:documentation 
	"EVENT-PAIR-HT holds a hash table that maps pairs of event names to event 
     pair objects, from the first to the second edge.  Each event pair object 
     includes all edges between those events, sorted by increasing order of 
     upper bound."))
  
  (:documentation "NAME is a lisp object denoting the STN's name."))

;;; To Do:
;;;  o The test (when (string= "STN" (class-name (class-of STN)))
;;;    seems like bad coding style, and will prevent STN from being
;;;    set up properly for other subclasses of STN.  We need to understand
;;;    the rational for this and clean it up.

(defmethod initialize-instance :after 
		   ((STN STN) 
			&key (start-name 'start) &allow-other-keys)

  "Creates a new simple temporal network named NAME.  
   The network is empty, that is, it has no constraints or events
   associated with it, other than the START event.  

   The function has the side effect of setting 
   *DEFAULT-STN* to be the newly created network."
  
  ;; Make sure this one is only called for instances of precisely this class,
  ;; not subclasses of STN.
  ;; Some modification has made by Peng
  ;;Provide two conditional processes for STN with and without specified start-event-name slot. Previous tests without start-event-name slot value will use start-name directly.

  (when (string= "STN" (class-name (class-of stn)))
	(when (eql (slot-value STN 'start-event-name) nil);;;Added by Peng
	(with-slots (name) stn
	  
	  (let ((start 
			 (make-instance 'event :name start-name
							:network stn
							:start? T)
			 ))

;;; Associates a start event with STN
		(setf (STN-start STN) start)
	  
		(setq *default-stn* stn))))
	(when (not (eql (slot-value STN 'start-event-name) nil));;;Added by Peng; 
	  (with-slots (name) stn
		(let ((start
			   (make-instance 'event :name (slot-value STN 'start-event-name)
							  :network stn
							  :start? T)
				))

;;; Associates a start event with STN
		  (setf (STN-start STN) start)
		  (setq *default-stn* stn))));;;Added by Peng;End
	)
  'T)

;;;   *** STN Indices of Events, Variables and Assignments ***

(defmacro find-event
    (event-name network)
  
  " Looks up and returns the event object associated 
    with EVENT-NAME.  If the event object
    does not exist, it returns NIL."
  
  `(gethash ,event-name
			(stn-event-ht ,network)))

;;; ***********************************************************************
;;;
;;;                EVENT PAIR Objects
;;;
;;; ***********************************************************************

(defclass event-pair ()
  
  ;; List of all distance edges from start to end of the event pair.

  ((start 
	:initarg :start
	:accessor event-pair-start
	:initform nil 
	:documentation "START event of all D-EDGES.")
   
   (end 
	:initarg :end
	:accessor event-pair-end
	:initform nil
	:documentation "END event of all D-EDGES.")
   
   (increasing-d-edges
	:initform '()
	:accessor event-pair-increasing-d-edges
	:documentation "List of edges from START to END, sorted in increasing 
                     order of upper bound.")
   
   (decreasing-d-edges
	:initform '()
	:accessor event-pair-decreasing-d-edges
	:documentation "List of edges from START to END, sorted in decreasing 
                    order of upper bound."))

  (:documentation "Record of all edges from START to END of a network.
                   Used to propagate effects when one or more edges between
                   START and END are inserted and/or removed."))

(defmacro find-event-pair (x y STN)

  " Looks up and returns the event-pair object associated 
    with events X and Y for simple temporal network STN.  
    If the objectdoes not exist, this function returns NIL."

  `(gethash (list ,x ,y)
			(STN-event-pair-ht ,STN)))

(defmethod initialize-instance :after 
		   ((event-pair event-pair)
			&key network)

  "Creates and returns a new event pair object for simple temporal
   network STN."

  (with-slots (start end) event-pair
	(setf (find-event-pair start end network) event-pair))
  'T)

(defmethod print-object ((obj event-pair) stream)
  
  " Print an EVENT-PAIR name."
  
  (with-slots (start end) obj 
	(format stream "(~a, ~a)"  start end)))

;;; *** Creation and retrieval of event pairs.

(defmethod get-event-pair (x y (STN STN))

  " Returns a unique event pair object for simple temporal network STN 
    for events X to Y.  If the object already exists, this function is 
    returned; otherwise,  a new object is created, associated with STN 
    and returned.  

    Note: This is the preferred way to create a new event-pair object,
    since it avoids duplication."

  (or (find-event-pair x y STN)
	  (make-instance 'event-pair
		:start x
		:end y
		:network STN)))


;;; ***********************************************************************
;;;
;;;                EVENT Objects
;;;
;;; ***********************************************************************


;;; An event is a real-valued point in time.
(defclass event ()

  ((name 
	:initarg :name
	:accessor event-name
	:initform nil
	:documentation
	"NAME is any lisp object or expression that names event.  A simple temporal
     network may have at most one event with a given NAME.")

   (network 
	:initarg :network
	:accessor event-network
	:initform nil
	:documentation
	"The simple temporal network that EVENT is part of.")
  
   (successors
	:accessor event-successors
	:initform '()
	:documentation 
	"Successor distance edges in the distance graph of the simple temporal 
     network.")
   
   (predecessors
	:accessor event-predecessors
	:initform '()
	:documentation
	"Predecessor distance edges in the distance graph of the simple temporal 
     network.")

   (lb 
	:accessor event-lb
	:initform *-infinity*
	:documentation
	"The tightest lower bound on EVENT, implied by the inserted constraints.")
   
   (ub
	:accessor event-ub
	:initform *infinity*
	:documentation
	"The tightest upper bound on EVENT, implied by the inserted constraints.")

;;; FIX:  Delete when works.  Moved to ITC-EVENT   
;;;   (status
;;;	:accessor event-status
;;;	:initform *unreached*
;;;	:documentation
;;;	"The status of the event can be of 3 types: UNREACHED, LABELED and SCANNED.")
   )

  (:documentation
   "Event object holds information about an event variable of a
   simple temporal network."))


(defmethod initialize-instance :after 
		   ((event event) &key (start? nil) &allow-other-keys)

  "Creates a new temporal event.
   Event is the start event of the network IFF start? is non-nil."
   
  (with-slots (name network) event

	(cond (start?
		   
		   (debug-print-for :itc t "New Start ~a," event)
		   (start-initialize! event))
		  
		  (:otherwise
		   
		   (debug-print-for :itc t "New event ~a," event)
		   
		   ;; Initialize as a non-start event.
		   (event-initialize! event)))

	;; Add to events of network,
	(push event (stn-events network))

	;; insert event into network's hash table,
	(setf (find-event name network)
	  event))

  'T)


(defmethod print-object ((obj event) stream)

  (with-slots (name) obj 
    (format stream "~a" name)))

;;; ***********************************************************************
;;;
;;;                Simple Temporal Constraint Object
;;;
;;; ***********************************************************************

;;;   This object holds information about a simple temporal constraint (STC).  
;;;   An STC is an interval constraint on the temporal distance between two 
;;;   events.  The constraint structure holds information about the constraint,
;;;   including a user-specified source, the start and end events, upper and
;;;   lowerbounds on the duration between events and various
;;;   bookkeeping flags used during propagation and search.

(defclass STC (essential-constraint)

  ((start 
	:initarg :start
	:accessor stc-start
	:initform nil
	:documentation
	"START stores the event that starts the simple temporal constraint.")
   
   (end 
	:initarg :end
	:accessor stc-end
	:initform nil
	:documentation
	"END stores the event that ends the simple temporal constraint.")
  
   (lb 
	:initarg :lb
	:accessor stc-lb
	:initform nil
	:documentation 
	"LB stores the lower bounds of the simple temporal constraint.")

   (lb-relaxable? 
	:initarg :lb-relaxable?
	:accessor stc-lb-relaxable?
	:initform nil
	:documentation 
	"Indicates if the lower bound can be relaxed.")

   (lb-cost-ratio 
	:initarg :lb-cost-ratio
	:accessor stc-lb-cost-ratio
	:initform nil
	:documentation 
	"A linear function gradient: indicates the ratio 
between cost and the relaxation of lb.")

   (lb-relaxed? 
	:initarg :lb-relaxed?
	:accessor stc-lb-relaxed?
	:initform nil
	:documentation 
	"Indicates if the lower bound has been relaxed.")

   (relaxed-lb
	:initarg :relaxed-lb
	:accessor stc-relaxed-lb
	:initform nil
	:documentation 
	"Stores the relaxed value of lb.")
   
   (ub 
	:initarg :ub
	:accessor stc-ub
	:initform nil
	:documentation 
	"UB stores the upper bound of the simple temporal constraint.")

   (ub-relaxable? 
	:initarg :ub-relaxable?
	:accessor stc-ub-relaxable?
	:initform nil
	:documentation 
	"Indicates if the upper bound can be relaxed.")

   (ub-cost-ratio 
	:initarg :ub-cost-ratio
	:accessor stc-ub-cost-ratio
	:initform nil
	:documentation 
	"A linear function gradient: indicates the ratio 
between cost and the relaxation of ub.")

   (ub-relaxed? 
	:initarg :ub-relaxed?
	:accessor stc-ub-relaxed?
	:initform nil
	:documentation 
	"Indicates if the upper bound has been relaxed.")

   (relaxed-ub
	:initarg :relaxed-ub
	:accessor stc-relaxed-ub
	:initform nil
	:documentation 
	"Stores the relaxed value of ub.")


   (d-edges 
	:accessor stc-d-edges
	:initform '()
	:documentation 
	"A list of two distance edges that encode this constraint."))
  
  (:documentation
   "This object holds information about a simple temporal constraint.
   The simple temporal constraint specifies interval constraint:
    (end - start) in [lb,ub]"))

;;; To Do:
;;;  o The test (when (string= "STC" (class-name (class-of stc)))
;;;    seems like bad coding style, and will prevent STC from being
;;;    set up properly for other subclasses of STC.  We need to understand
;;;    the rational for this and clean it up.

(defmethod initialize-instance :after 
		   ((STC STC) 
			&rest extra-args)

  "Creates a new simple temporal constraint."

  (declare (ignore extra-args))
  
  ;; The following initialization is performed only if class is
  ;; an STC, but NOT a subclass of an STC.
  (when (string= "STC" (class-name (class-of stc)))
	
	;; Note that this function does not initialize any of the status 
	;; slots of the constraint.  This is performed when the constraint
	;; is inserted into the constraint-system. 
	(with-slots (start end ub lb  constraint-system) stc

	  (check-legal-bounds! start end lb ub)

	  (unless (typep start 'event)
		(setq start (get-event start constraint-system)))
	
	  (unless (typep end 'event)
		(setq end (get-event end constraint-system)))

;;;	;; Point constraint's events to constraint
;;;    (connect-events-to-constraint! stc)

	  ;; Create new distance graph edges associated with constraint.
	  (make-instance 'd-edge :start end :end start
					 :ub (b- lb) :constraint stc
					 :calculate-arc-type nil
					 :lower-bound? t)

	  (make-instance 'd-edge :start start :end end
					 :ub ub :constraint stc
					 :calculate-arc-type nil
					 :lower-bound? nil)
	  ))
  'T)

;;; Default method defined in mtk-utils/compare.lisp
(defmethod extract-proposition-description 
	((stc STC))

  "Return the proposition description for state-assignment p."
  
  (with-slots (start end lb ub) stc
	`(STC :start ,(event-name start)
		  :end ,(event-name end) 
		  :lb ,lb
		  :ub ,ub)))

;;;            ***   Printing a constraint   ***

(defmethod print-object ((obj stc) stream)
  (with-slots (lb ub start end) obj 
	(format stream "(~a,~a)[~a,~a]" start end lb ub)))



(defclass STC-Relaxation ()

  ((constraint
	:initarg :constraint
	:accessor stc-relaxation-constraint
	:initform nil
	:documentation
	"The stc this relaxation is being applied to.")

   (relaxing-lb?
	:initarg :relaxing-lb?
	:accessor stc-relaxation-lb?
	:initform nil
	:documentation
	"Indicate if the lower bound of the stc is being relaxed.")

   (relaxed-lb
	:initarg :relaxed-lb
	:accessor stc-relaxation-lb
	:initform nil
	:documentation
	"The relaxed lb.")

   (relaxing-ub?
	:initarg :relaxing-ub?
	:accessor stc-relaxation-ub?
	:initform nil
	:documentation
	"Indicate if the upper bound of the stc is being relaxed.")

   (relaxed-ub
	:initarg :relaxed-ub
	:accessor stc-relaxation-ub
	:initform nil
	:documentation
	"The relaxed ub."))
  
  (:documentation
   "This object holds information about the relaxation to a simple temporal constraint."))


(defmethod apply-relaxation (relaxation)

  "Apply the relaxed bounds of this relaxation to its stc"

  (with-slots (constraint relaxing-lb? relaxed-lb relaxing-ub? relaxed-ub) relaxation

	(when relaxing-lb?
	  (setf (stc-lb-relaxed? constraint) t)
	  (setf (stc-relaxed-lb constraint) relaxed-lb))

	(when relaxing-ub?
	  (setf (stc-ub-relaxed? constraint) t)
	  (setf (stc-relaxed-ub constraint) relaxed-ub))

	(loop for d-edge in (stc-d-edges constraint) do

		 (when (and (d-edge-lower-bound? d-edge) relaxing-lb?)
		   ;; need to update the value in the d-edge
		  ;; (format t "updating LB edge ~A to ~A~%" (d-edge-ub d-edge) (b- relaxed-lb))
		   (update-edge-value d-edge (b- relaxed-lb)))

		 (when (and (not (d-edge-lower-bound? d-edge)) relaxing-ub?)
		   ;; need to update the value in the d-edge
		  ;; (format t "updating UB edge ~A to ~A~%" (d-edge-ub d-edge) relaxed-ub)
		   (update-edge-value d-edge relaxed-ub)))))


(defmethod update-edge-value (edge new-value)

  "Update the value and original value of the edge"

  (setf (d-edge-initial-ub edge) new-value)

  (if (b= *infinity* new-value)
	(setf (d-edge-ub edge) 1000000)

	(if (b= *-infinity* new-value)
		(setf (d-edge-ub edge) -1000000)

		(setf (d-edge-ub edge) new-value))))



(defmethod undo-relaxation (relaxation)

  "Undo the relaxations for the stc."

  (with-slots (constraint relaxing-lb? relaxed-lb relaxing-ub? relaxed-ub) relaxation

	(when relaxing-lb?
	  (setf (stc-lb-relaxed? constraint) nil))

	(when relaxing-ub?
	  (setf (stc-ub-relaxed? constraint) t))

	(loop for d-edge in (stc-d-edges constraint) do

		 (if (d-edge-lower-bound? d-edge)
		   ;; need to recover the value of the d-edge
		   (update-edge-value d-edge (b- (stc-lb constraint)))
		   (update-edge-value d-edge (stc-ub constraint))))))



;;; ***********************************************************************
;;;
;;;                Distance Graph Edge Object
;;;
;;; ***********************************************************************

;;;   This object holds information about an edge of the distance graph of STC.
;;;   An edge is a weighted edge between x and y, with weight c, such that
;;;   x-y < c.  The edge structure holds information about the edge,
;;;   including a user-specified source, the start and end events, edge
;;;   weight and various bookkeeping flags used during propagation and search.

(defclass d-edge ()

  ;; START and END are the source and target events of the edge.
  ;; the simple temporal constraint.

  ((start 
	:initarg :start
	:accessor d-edge-start
	:initform nil
	:documentation
	"START stores the events that end the simple temporal constraint.")
   
   (end 
	:initarg :end
	:accessor d-edge-end
	:initform nil
	:documentation
	"END stores the events that end the simple temporal constraint.")
  
   (ub 
	:initarg :ub
	:accessor d-edge-ub
	:initform nil
	:documentation
	"UB stores the upper bound of the distance edge.")	

   (initial-ub 
	:initarg :ub
	:accessor d-edge-initial-ub
	:initform nil
	:documentation
	"initial-UB stores the original upper bound of the distance edge. 
This can be used to recover the initial value of the edge when UB is changed.")	


   (constraint
	:initarg :constraint
	:accessor d-edge-constraint
	:initform nil
	:documentation
	"The CONSTRAINT that distance graph edge is generated from.")
   
   (lower-bound?
	:initform t
	:initarg :lower-bound?
	:accessor d-edge-lower-bound?
	:documentation "Whether this d-edge is a lower or upper bound."))

  (:documentation
   
   "This object holds information about a distance graph edge.
    A distance graph edge specifies that:
       (end - start) < ub."))

(defmethod initialize-instance :after 
		   ((d-edge d-edge) &key (calculate-arc-type t) &allow-other-keys)

  "Creates a new distance-edge of a simple temporal network"

  (with-slots (start end ub initial-ub constraint lower-bound?) d-edge
	
	(let ((STN (event-network start)))

	  (debug-print-for :itc t "New edge ~a," d-edge)
		 
	  ;; Associates edge with its constraint.
	  (stn-add-d-edge! d-edge STN)
	  
	  (unless (null constraint)
		
		(push d-edge (stc-d-edges constraint)))

	  ;; Associates edge with its events.
	  (debug-print-for :itc t 
					   "  made successor of ~a and predecessor of ~a." 
					   start end)
	  
	  (push d-edge (event-successors start))
	  (push d-edge (event-predecessors end))

	  ;; Replace the infinity numbers in the distance graph with a real number
	  ;; to avoid ITC failures.
	  (when (b= *infinity* ub)
	  	  (setf ub 1000000))
	  (when (b= *-infinity* ub)
	  	  (setf ub -1000000))
	  ;; Make a copy of the ub
	  (setf initial-ub ub)
	  
	  ;;Decide whether this is a lower bound.
	  (when calculate-arc-type

		;;grab the opposing edge.
		(let ((opposite-d-edge (find-opposing-d-edge d-edge) )) 
		  
		  (cond 
		   ((b> 0 ub)
			
			(when (and (not (null opposite-d-edge))
					   (not (b> 0 (d-edge-ub opposite-d-edge))))
			  
			  ;; When the opposing edge exists and has a zero weight,
			  ;; then it must be an upper bound because this is definitely a lb.
			  (setf (d-edge-lower-bound? opposite-d-edge) nil)))
		   
		   ((eql 0 ub)
			
			(unless (null opposite-d-edge)

			  ;; If the opposite edge exists
			  ;; this one is whatever that one is not.
			  (setf lower-bound? (not (d-edge-lower-bound? opposite-d-edge)))))
		   
		   ((b> ub 0)
			
			(if (not (null opposite-d-edge))

				;; if this one is positive,
				;; if the opposite exists, 
				;; it must oppose that one because both can be positive
				(setf lower-bound? (not (d-edge-lower-bound? opposite-d-edge))) 

			  ;; otherwise, this is an ub (which is usually true).
			  (setf lower-bound? nil)))) 
		  
;;;		  (format t "setting ub/lb ~a ~a ~a~%" d-edge opposite-d-edge lower-bound?)
		  ))))

  'T)

(defmethod find-opposing-d-edge ((d-edge d-edge))
  
  (with-slots (start end) d-edge
	(first (event-pair-increasing-d-edges
			(get-event-pair end start (event-network (d-edge-start d-edge)))))))

(defmethod d-edge-upper-bound? ((d-edge d-edge))
  
  " A simple accessor function to avoid inversion in other places."
  
  (not (d-edge-lower-bound? d-edge)))

;;;            ***   Printing a distance edge   ***

(defmethod print-object ((obj d-edge) stream)
  
  (with-slots (ub start end lower-bound?) obj 
	(format stream "~a->~a[~a]" start end ub)
	(when lower-bound?
	  (format stream "(lb)"))))


