(in-package :ITC)

;;; This file includes non-incremental reasoning algorithms for STNs.
;;; External interfaces:
;;; consistent-stn? - tests for consistency
;;; apsp-floyd-warshall!
;;; stn-filtering!


(defmethod compile-stn! ((stn stn))
"Compile the stn to minimal dispatchable form by running APSP and filtering the network.
Returns t if the compiled network is dispatchable, nil otherwise. The stn object will 
be transformed into the dispatchable form of the network."

  (when (consistent-stn? stn)
	(stn-filtering! stn)
	t))

(defmethod apsp-floyd-warshall! ((stn stn) &key (stn-mode t))
  "Modifies the input to calculate the all-pairs shortest paths, leaving in
the self-loops for another step to do consistency checking. Uses the standard
Floyd-Warshall algorithm. Uses an arbitrary constraint of the system to label
any edges added by this algorithm.

Stn-mode defaults to true since this is, after all, part of an stn package.
If true, it does not put new edges that go from and to the same event, but uses
them to determine the consistency of the STN. It returns the consistency. If stn-mode
is nil, it returns t and performs the full apsp."
  (print "Doing APSP!!")
  (with-slots (events) stn				;grab the event list
	(loop for k in events				;loop through possible 
	   do (loop for i in events
			 do (loop for j in events
				   ;;it is consistent if it was before and it is not found to have a problem when the edge is updated
					 ;;if it becomes inconsistent (as returned from update), return immediately
					 do (when (not (apsp-update-path! i j k stn stn-mode)) ;update ij with shortcut through k
					 (return-from apsp-floyd-warshall! nil)))))) 
	
  t)				 ;return whether the network is consistent


(defmethod apsp-update-path! ((i event) (j event) (k event) (stn stn) stn-mode)
"Consider this triangle of events and update the ij edge if k is a suitable shortcut. Returns
whether an inconsistency was found. If stn-mode is true, it won't put in self-loops, just consider them."
  (let* ((pair-ij (get-event-pair i j stn)) ;;grab pairs and weights
		 (pair-ik (get-event-pair i k stn))
		 (pair-kj (get-event-pair k j stn))
		 (d-ij (get-event-pair-ub pair-ij))
		 (d-ik (get-event-pair-ub pair-ik))
		 (d-kj (get-event-pair-ub pair-kj))
		 (consistent? t))
	(when (and (bound? d-ik) (bound? d-kj)) ;if the alternate path is bound
	  (let ((d-shortcut (b+ d-ik d-kj))) ;compute the shortcut distance
		(when (or (not (bound? d-ij)) (b> d-ij d-shortcut)) ;;if ij is either unbound or longer, 
		  (if (and stn-mode (eql i j)) ;if it's a diagonal update in stn-mode, don't do it
			  (progn (when (b> 0 d-shortcut) ;if the diagonal would become negative, set it inconsistent
					   (setf consistent? nil))) ;and mark the stn as inconsistent
										;else, update the bound

			  (update-event-pair-ub pair-ij d-shortcut nil ;(d-edge-constraint (first (event-pair-increasing-d-edges pair-ik)))
									)))))
	consistent?))

(defmethod consistent-stn? ((stn stn) &key (initialize? T) &allow-other-keys)
  
  "Determines whether the stn is consistent by calling the apsp algorithm. This
is somewhat inefficient because it re-runs the whole algorithm, but there's no
book-keeping to see if things have changed. As a side effect, the STN is updated 
with the APSP."

  ;; Always reinitializes, whether told to or not.
  (declare (ignore initialize?))
  
  (apsp-floyd-warshall! stn))

(defmethod get-event-pair-ub ((event-pair event-pair))
  "Returns the upper bound of the edge of this pair, else returns nil.
This requires that the pair only has one edge, so cannot be mixed with
ITC routines."

  (let ((edge (car (event-pair-increasing-d-edges event-pair)))) ;;grab the edge itself
	(if (not (null edge)) ;;return the value if it's not nil, otherwise return nila
		(d-edge-ub edge)
		nil)))

(defmethod update-event-pair-ub ((event-pair event-pair) ub &optional constraint)
  "Updates the d-edge corresponding to an event-pair. It either updates the
edge already present or creates a new one."
  (when (bound? ub)						;error checking
	(let ((edge (car (event-pair-increasing-d-edges event-pair))))
	  (when constraint) ; Does nothing, just remove optional warning

	  ;; (unless (null edge) ;unless it doesn't exist, we need to get rid of the old one
	  ;; 	(remove-d-edge! edge)) 
	  ;; (with-slots (start end) event-pair ;always make a new edge
	  ;; 	(make-instance 'd-edge :start start :end end :ub ub :constraint constraint))
	 
	  ;;; Above code altered by James Paterson. It was adding extra edges into the distance graph instead of updating them!

	  (if (null edge)
		  (with-slots (start end) event-pair ;make a new edge
			(make-instance 'd-edge :start start :end end :ub ub :constraint constraint))
		  ;; Else update
		  (setf (d-edge-ub edge) ub))
	  ))) ;with no constraint, since it's just tightening




(defmethod stn-filtering! ((stn stn))
  "Implements the STN filtering algorithm as given in John Stedl's thesis.
Acutally removes any edges not needed to correctly dispatch the STN."
  (let ((to-delete nil))
	(with-slots (events) stn
	  (loop for i in events				;step through every event

		 ;;Use two for loops to step through all pairs of intersecting edges
		 ;;for this event but do not repeat with the reverse order.
		 do (loop for edge-source in (list (event-predecessors i) (event-successors i)
										   ) 
			   do (loop for j in edge-source 
					 for j-count from 1
					 do (loop for k in (split-after j-count edge-source) ;avoid dupes by only doing the later part
						   ;;For this pair of edges, see if either should be deleted
						   do (when (not (eql j k)) 
								(let* ((j-dom-k (and ;;(null (member k to-delete)) 
												 (edge-dominated? j k stn))) ;;check whether either/both edge dominates the other
									   (k-dom-j (and ;;(null (member j to-delete))
												 (edge-dominated? k j stn))))
								  
								  (cond ((and j-dom-k k-dom-j (null (member k to-delete))) 
										 (push j to-delete)) ;;if both are dominated, pick one
										
										((and j-dom-k k-dom-j (null (member j to-delete))) 
										 (push k to-delete))
										(j-dom-k (push j to-delete)) ;;else push the dominated on on - don't worry about dupes
										(k-dom-j (push k to-delete))))))))
		   ))

	(setf to-delete (remove-duplicates to-delete)) ;now go back and remove all the duplicates

	(loop for i in to-delete			;now remove each one
	   do (remove-d-edge! i))))

;Utilities from http://www.ic.unicamp.br/~meidanis/courses/mc336/2006s2/funcional/L-99_Ninety-Nine_Lisp_Problems.html
(defun split (pos org-list &optional (ini 0))
    (if (> pos ini)
        (cons (car org-list) (split pos (cdr org-list) (+ ini 1)))))

(defun split-after (pos org-list &optional (ini 0))
    (if (> pos ini)
        (split-after pos (cdr org-list) (+ ini 1))
        org-list))


(defmethod d-edge-comparable? ((x d-edge) (y d-edge))
  t)

(defmethod edge-dominated? ((x d-edge) (y d-edge) (stn stn))
  "Tests if x is dominated by y as edges of an STN according to 
the algorithm in John Stedl's thesis."

  ;;If the edges aren't comparable, they can't be dominated.
  (if (not (d-edge-comparable? x y))
	  (return-from edge-dominated? nil))

  (let* ((dominated? nil)
		 (dx (d-edge-ub x)) 
		 (dy (d-edge-ub y)))
	;;test for upper domination
	(when (not (eql (d-edge-start x) (d-edge-start y)))
	  (let* ((event-pair  (get-event-pair (d-edge-start x) (d-edge-start y) stn )) ;grab the weight of the third edge of triangle
			 (dxy (find-event-pair-max-value-matching-d-edge event-pair x)))
		(when (and (bound? dxy) (b> dy 0) (b> dx 0) ;(d-edge-upper-bound? x) (d-edge-upper-bound? y) ;non-negative
				   (b= dx (b+ dxy dy)))
		  ;;			(format t "dominated 1: ~a ~a ~a~%" x y (find-event-pair-max-d-edge-matching-d-edge event-pair x))
		  (assert (and (b>= dy 0) (b>= dx 0)) (dx dy) "A negative upper bound?")
		  ;; (format t "upper domination found: ~a by ~a and ~a ~%" x y dxy)
		  (setf dominated? t)
										;	  (format t "up dom:~a ~a ~a~%" x y dxy)
		  )
		(unless dominated?
;;		  (format t "not upper dominated: ~a by ~a ~a~%" x y dxy)
)))

	(when (not dominated?)	 ;only keep going if not already dominated
	  ;;test for lower domination
	  (when (not (eql (d-edge-end y) (d-edge-end x)))
		(let* ((event-pair  (get-event-pair (d-edge-end y) (d-edge-end x) stn )) ;grab the weight of the third edge of triangle
			   (dxy (find-event-pair-max-value-matching-d-edge event-pair x)))
		  (when (and (bound? dxy) ;(d-edge-lower-bound? x) (d-edge-lower-bound? y) 
					 (b> 0 dx) (b> 0 dy) (b= dx (b+ dxy dy))) ;stricly negative
			;(format t "lower domination found: ~a by ~a and ~a ~%" x y dxy)
			(setf dominated? t)
			;;			(format t "dominated 2: ~a ~a ~a~%" x y (find-event-pair-max-d-edge-matching-d-edge event-pair x))
										;			(format t "low dom:~a ~a ~a~%" x y dxy)
			)
		  (unless dominated?
;		  (format t "not lower dominated: ~a by ~a ~a~%" x y dxy)
))))

	(when dominated?
	  (debug-print-for :cdtn t "Dominated: ~a by ~a" x y)
	  )
	
	
	dominated?))
