;;;; Copyright (c) 2009-2010 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:kirk)

;;-------------------------------------------------------------------------------
;; NOTE: The following code does NOT implement a complete reader or writer for
;; the environment file format. Instead, it provides only the functionality
;; needed for the visibility-graph interface.
;;-------------------------------------------------------------------------------

;;-------------------------------------------------------------------------------
;; WRITE Trajectory File
;;-------------------------------------------------------------------------------

;; (defun write-environment-file (env-out-file vehicles features obstacles)
;;   (write-environment-file env-out-file (make-instance 'environment :vehicles vehicles :features features :obstacles obstacles)))

(defun write-environment-file (env-out-file environment)
  (with-open-file (env-out-stream env-out-file 
								  :direction :output
								  :if-exists :supersede
								  :if-does-not-exist :create)
	(write-environment-stream env-out-stream environment)
	  ))

;; (defun write-environment-stream (env-out-stream vehicles features obstacles)
;;   (write-environment-stream env-out-stream (make-instance 'environment :vehicles vehicles :features features :obstacles obstacles)))

(defun write-environment-string (environment)
  (with-output-to-string (env-stream)
	(write-environment-stream env-stream environment)))

(defun write-environment-stream (env-out-stream environment)
  (let ((vehicle-list (get-value-list (env-vehicles environment)))
		(feature-list (get-value-list (env-features environment)))
		(obstacle-list (get-value-list (env-obstacles environment)))
		(env-lxml (make-smart-list)))
	(coll-add! env-lxml :ENVIRONMENT)
	(loop for vehicle in vehicle-list do
		 (coll-add! env-lxml (make-lxml-from-vehicle vehicle)))
	(loop for feature in feature-list do
		 (coll-add! env-lxml (make-lxml-from-feature feature)))
	(loop for obstacle in obstacle-list do
		 (coll-add! env-lxml (make-lxml-from-obstacle obstacle)))
	(print-xml (coll-tolist env-lxml) :stream env-out-stream :pretty t)))

(defun hash-keys (hash-table)
  "Returns the list of keys contained in HASH-TABLE"
  (loop for key being the hash-keys of hash-table
	   collect key))

(defun hash-values (hash-table)
  "Returns the list of values contained in HASH-TABLE"
  (loop for value being the hash-values of hash-table
	   collect value))

(defun get-value-list (object)
  (cond
	((listp object) object)
	((hash-table-p object) (hash-values object))
	(t nil)))

;;-------------------------------------------------------------------------------
;; READ Trajectory File
;;-------------------------------------------------------------------------------

(defun read-environment-file (environment-file)  
  (with-open-file (env-in-stream environment-file
								 :direction :input
								 :if-exists :supersede
								 :if-does-not-exist :create)
	(read-environment-stream env-in-stream)))

(defun read-environment-string (env-string)
  (with-input-from-string (env-in-stream env-string)
	(read-environment-stream env-in-stream)))

(defun read-environment-stream (env-in-stream)
  (let ((environment-xml (parse-xml env-in-stream))
		(obstacle-ht (make-hash-table :test #'equal))
		(feature-ht (make-hash-table :test #'equal))
		(vehicle-ht (make-hash-table :test #'equal)))
	(loop for element in environment-xml do
		 (cond ((and (listp element) (equal :OBSTACLE (first element)))
				(let ((obstacle (make-obstacle-from-lxml element)))
				  (setf (gethash (env-obstacle-name obstacle) obstacle-ht)
						obstacle)))
			   ((and (listp element) (equal :FEATURE (first element)))
				(let ((feature (make-feature-from-lxml element)))
				  (setf (gethash (env-feature-name feature) feature-ht)
						feature)))
			   ((and (listp element) (equal :VEHICLE (first element)))
				(let ((vehicle (make-vehicle-from-lxml element)))
				  (setf (gethash (env-vehicle-name vehicle) vehicle-ht)
						vehicle)) )))
	(values vehicle-ht feature-ht obstacle-ht)
	(make-instance 'environment :vehicles vehicle-ht :features feature-ht :obstacles obstacle-ht)))

;;-------------------------------------------------------------------------------
;; CLASSes
;;-------------------------------------------------------------------------------

(defclass environment ()
  ((vehicles
	:accessor env-vehicles
	:initarg :vehicles
	:initform nil
	:documentation "hashtable of vehicles, parsed from VEHICLE tags")
   (features
	:accessor env-features
	:initarg :features
	:initform nil
	:documentation "hashtable of features, parsed from FEATURE tags")
   (obstacles
	:accessor env-obstacles
	:initarg :obstacles
	:initform nil
	:documentation "hashtable of features, parsed from OBSTACLE tags")))

(defclass env-vehicle ()
  ((name
	:accessor env-vehicle-name
	:initarg :name
	:initform nil
	:documentation "NAME tag")
   (actions
	:accessor env-vehicle-actions
	:initarg :actions
	:initform nil
	:documentation "ACTIONS tag, stored as hash table with the name of the action as the key.")))

(defun make-vehicle-from-lxml (vehicle-lxml)
  (let ((name (get-lxml-value vehicle-lxml :NAME))
		(actions (make-hash-table :test #'equal)))
	(loop for element in vehicle-lxml do
		 (cond ((and (listp element) (equal :ACTION (first element)))
				(let ((action (make-action-from-lxml element)))
				  (setf (gethash (env-vehicle-action-name action) actions)
						action)))))
	(make-instance 'env-vehicle :name name :actions actions)))

(defun make-lxml-from-vehicle (vehicle)
  (let ((vehicle-slist (make-smart-list)))
	(with-slots (name actions) vehicle
	  (append-lxml-element vehicle-slist :NAME name)
	  (loop for action being the hash-values of actions do
		   (coll-add! vehicle-slist (make-lxml-from-action action))) )
	(create-lxml-element :VEHICLE vehicle-slist)))

(defclass env-vehicle-action ()
  ((name
	:accessor env-vehicle-action-name
	:initarg :name
	:initform nil
	:documentation "NAME tag")
   (min-vel
	:accessor env-vehicle-action-minv
	:initarg :min-vel
	:initform nil
	:documentation "MIN_VELOCITY tag")
   (max-vel
	:accessor env-vehicle-action-maxv
	:initarg :max-vel
	:initform nil
	:documentation "MAX_VELOCITY tag")
   (min-acc
	:accessor env-vehicle-action-mina
	:initarg :min-acc
	:initform nil
	:documentation "MIN_ACCELERATION tag")
   (max-acc
	:accessor env-vehicle-action-maxa
	:initarg :max-acc
	:initform nil
	:documentation "MAX_ACCELERATION tag")
   ))

(defun make-action-from-lxml (action-lxml)
  (let ((name (get-lxml-value action-lxml :NAME))
		(minv (get-lxml-number action-lxml :MIN_VELOCITY))
		(maxv (get-lxml-number action-lxml :MAX_VELOCITY))
		(mina (get-lxml-number action-lxml :MIN_ACCELERATION))
		(maxa (get-lxml-number action-lxml :MAX_ACCELERATION)))
	(make-instance 'env-vehicle-action :name name 
				   :min-vel minv :max-vel maxv :min-acc mina :max-acc maxa)))

(defun make-lxml-from-action (action)
  (let ((action-slist (make-smart-list)))
	(with-slots (name min-vel max-vel min-acc max-acc) action
	  (append-lxml-element action-slist :NAME name)
	  (append-lxml-element action-slist :MIN_VELOCITY min-vel)
	  (append-lxml-element action-slist :MAX_VELOCITY max-vel)
	  (append-lxml-element action-slist :MIN_ACCELERATION min-acc)
	  (append-lxml-element action-slist :MAX_ACCELERATION max-acc) )
	(create-lxml-element :ACTION action-slist))) 

(defclass env-obstacle ()
  ((name
	:accessor env-obstacle-name
	:initarg :name
	:initform nil
	:documentation "NAME tag")
   (type
	:accessor env-obstacle-type
	:initarg :type
	:initform nil
	:documentation "TYPE tag")
   (polygon
	:accessor env-obstacle-polygon
	:initarg :polygon
	:initform nil
	:documentation "CORNER tag")))


(defun make-obstacle-from-lxml (obstacle-lxml)
  "Creates a polygon from the lisp representation of the 
   <OBSTACLE></OBSTACLE> XML-segment"
  (let ((name (get-lxml-value obstacle-lxml :NAME))
		(type (get-lxml-value obstacle-lxml :TYPE))
		(poly (mtk-utils-2d-geometry::make-polygon)))
	(loop for element in obstacle-lxml do
		 (when (and (listp element) (equal :CORNER (first element)))
		   (let ((x (get-lxml-number element :X2D))
				 (y (get-lxml-number element :Y2D)))
			 (add-point poly (make-2d-point x y)))))
	(make-instance 'env-obstacle :name name :type type :polygon poly)))

(defun make-lxml-from-obstacle (obstacle)
  (let ((obstacle-slist (make-smart-list)))
	(with-slots (name type polygon) obstacle
	  (append-lxml-element obstacle-slist :NAME name)
	  (append-lxml-element obstacle-slist :TYPE type)
	  (loop for corner in (point-list polygon) do
		   (let ((corner-slist (make-smart-list)))
			 (append-lxml-element corner-slist :X2D (x corner))
			 (append-lxml-element corner-slist :Y2D (y corner))
			 (append-lxml-element obstacle-slist :CORNER corner-slist))) )
	(create-lxml-element :OBSTACLE obstacle-slist)))


(defclass env-feature ()
  ((name
	:accessor env-feature-name
	:initarg :name
	:initform nil
	:documentation "NAME tag")
   (center
	:accessor env-feature-center
	:initarg :center
	:initform nil
	:documentation "CENTER tag")
   (radius
	:accessor env-feature-radius
	:initarg :radius
	:initform 0
	:documentation "RADIUS tag")))

(defun make-feature-from-lxml (feature-xml)
  "Creates a 2d-point from the lisp representation of the 
   <FEATURE></FEATURE> XML-segment"
  (let ((name (get-lxml-value feature-xml :NAME))
		(radius (get-lxml-number feature-xml :RADIUS))
		(center-xml (get-lxml-element feature-xml :CENTER)))
	(let ((x (get-lxml-number center-xml :X2D))
		  (y (get-lxml-number center-xml :Y2D)))
	  (make-instance 'env-feature 
					 :name name 
					 :center (make-2d-point x y) 
					 :radius radius))))

(defun make-lxml-from-feature (feature)
  (let ((feature-slist (make-smart-list)))
	(with-slots (name center radius) feature
	  (append-lxml-element feature-slist :NAME name)
	  (let ((center-slist (make-smart-list)))
		(append-lxml-element center-slist :X2D (x center))
		(append-lxml-element center-slist :Y2D (y center))
		(append-lxml-element feature-slist :CENTER center-slist))
	  (append-lxml-element feature-slist :RADIUS radius) )
	(create-lxml-element :FEATURE feature-slist)))


;;-------------------------------------------------------------------------------
;; Helper Functions
;;-------------------------------------------------------------------------------

(defun get-lxml-element (element-list element-symbol)
  (loop for element-xml in (append element-list '(nil)) 
	 until (and (listp element-xml) (equal element-symbol (first element-xml)))
	 finally (return element-xml)))

(defun get-lxml-value (element-list element-symbol &key (trim t))
  (let ((element (get-lxml-element element-list element-symbol)))
	(if (null element)
		nil
		(if trim
			(string-trim " " (second element))
			(second element)))
	))

(defun get-lxml-number (element-list element-symbol)
  (let ((value (get-lxml-value element-list element-symbol)))
	(if (null value) nil (parse-number value))))

(defmethod append-lxml-element ((element-list smart-list) element-symbol element-value)
  "Appends `(<element-symbol> <element-value>) to element-list. Returns the modified smart list.
   If element-value is null then the unmodified element-list is returned."
  (let ((element (create-lxml-element element-symbol element-value)))
	(if (null element)
		element-list
		(coll-add! element-list element))))

(defun create-lxml-element (element-symbol element-value)
  "Appends `(<element-symbol> <element-value>) to element-list. Returns the modified smart list.
   If element-value is null then the unmodified element-list is returned."
  (cond ((null element-value) 
		 nil)
		((listp element-value)
		 (append `(,element-symbol) element-value) )
		((typep element-value 'ABSTRACT-DATA-TYPES::SMART-LIST)
		 (append `(,element-symbol) (coll-tolist element-value)) )
		((numberp element-value)
		 `(,element-symbol ,(format nil "~s" element-value)))
		(t 
		 `(,element-symbol ,element-value))
		))