;;;; Copyright (c) 2009-2011 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:kirk)

(defmacro error-on-nil (expr format-string &rest args)
  "Macro signals an error when form evaluates to nil."
    (let ((gexpr (gensym "EXPR-")))
	  `(let ((,gexpr ,expr))
		 (if ,gexpr  
			 ,gexpr
			 (error ,format-string ,@args)))
	  ))

#+allegro
(defun make-trajectory-main ()
  (system:with-command-line-arguments
	  ("a:b:c:d:e:f:g:" tpn-in-pathname env-in-pathname tpn-int-pathname tpn-out-pathname env-out-pathname radius explanation-out-pathname)
	(rest)
	;; (continue) will continue executions, but first giving the
	;; person running the app a chance to poke around.
	(declare (ignore rest))
	;;(format t "~s ~s ~s ~s ~s~%" tpn-in-pathname env-in-pathname tpn-out-pathname env-out-pathname radius)
	(kirk-vz-graph-file (pathname tpn-in-pathname) (pathname env-in-pathname) 
						(pathname tpn-out-pathname) (pathname env-out-pathname) 
				   :radius (parse-number radius) :tpn-int-pathname (pathname tpn-int-pathname)
				   :exp-out-pathname (pathname explanation-out-pathname))
	(cl-user::exit 0))
)


(defun kirk-vz-graph-file (tpn-in-pathname env-in-pathname tpn-out-pathname env-out-pathname &key (radius 0) (exp-out-pathname nil) &allow-other-keys)
  "Given an RMPL file and an Environment file, this method will create
  the optimal trajectory file."
  (with-open-file (tpn-in-stream tpn-in-pathname)
	(with-open-file (env-in-stream env-in-pathname)
	  (with-open-file (tpn-out-stream tpn-out-pathname
									  :direction :output
									  :if-exists :supersede
									  :if-does-not-exist :create)
		(with-open-file (env-out-stream env-out-pathname
									  :direction :output
									  :if-exists :supersede
									  :if-does-not-exist :create)
		  (if exp-out-pathname
			  (with-open-file (exp-out-stream exp-out-pathname
											  :direction :output
											  :if-exists :supersede
											  :if-does-not-exist :create)
				(kirk-vz-graph-stream tpn-in-stream env-in-stream tpn-out-stream env-out-stream :radius radius :exp-out-stream exp-out-stream :generate-exp? T))
			  (kirk-vz-graph-stream tpn-in-stream env-in-stream tpn-out-stream env-out-stream :radius radius))
		  ))))
  )

(defun kirk-vz-graph-stream (tpn-in-stream env-in-stream tpn-out-stream env-out-stream &key (radius 0) (tpn-int-stream nil) (exp-out-stream nil) (generate-exp? NIL) &allow-other-keys)
  (let* ((tpns (parse-tpn-stream tpn-in-stream
								 'tpn::get-event 'tpn::make-episode
								 'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns))
		 (environment (read-environment-stream env-in-stream)))
	
	(multiple-value-bind (feasible tpn-out env-out best-choice explanations tpn-int)
		(kirk-vz-graph tpn environment :radius radius :explanations? exp-out-stream :generate-exp? generate-exp?)
	  ;; write the solution tpn
	  (if (null tpn-out)
		 (write-string "No Solution." tpn-out-stream)
		 (write-tpn tpn-out-stream tpn-out :environment :all :force-labeling? t))
	  ;; write out the environment
	  (write-environment-stream env-out-stream env-out)
	  ;; write the intermediate
	  (when tpn-int-stream
		(write-tpn tpn-int-stream tpn-int :force-labeling? t))
	  ;; write the explanations
	  (when (and exp-out-stream explanations)

		(let ((choice-queue (get-decision-queue tpn))) 
		
		(write-explanations exp-out-stream explanations (second choice-queue) (first choice-queue) (third choice-queue)))
	  (values feasible best-choice))
	  )))

(defun kirk-vz-graph (tpn environment &key (radius 0) (generate-exp? nil) &allow-other-keys)
  "Given the TPN and ENVIRONEMNT objects, returns the three values
   the first value is t if the plan is feasible.
   the second value is the new tpn.
   the third value is the new environment.
   the fourth value is the best-choice"
  (let* (;; initialize the hash tables
		 (obstacle-ht (env-obstacles environment))
		 (feature-ht (env-features environment))
		 (vehicle-ht (env-vehicles environment))
		 (vg nil))

	;; (format t "KIRK-INPUT-TPN:~%~a~%" 	(with-output-to-string (tpn-stream)
	;; 									  (tpn::write-tpn tpn-stream tpn :force-labeling? t)))

	;; initialize the visibility graph
	(let ((obstacle-list nil)
		  (feature-list nil))
	  (maphash (lambda (key value) (declare (ignore key)) (push (env-obstacle-polygon value) obstacle-list)) obstacle-ht)
	  (maphash (lambda (key value) (declare (ignore key)) (push (env-feature-center value) feature-list)) feature-ht)
	  ;; create the visibility-graph
	  (setf vg (make-visibility-graph obstacle-list feature-list)))

	(loop for episode in (copy-list (tpn::tpn-episodes tpn)) do
		 (run-vg-on-episode episode tpn environment vg radius))

   		;;(format t "KIRK-INT-TPN: ~s~%" (with-output-to-string (tpn-int-stream2) (write-tpn tpn-int-stream2 tpn :force-labeling? t)))

	;; initialize choice selection problem
	(initialize-tpn-choice-selection! tpn)
	;; find the best choice
	(let* ((best-choice (tpn-best-choice tpn))
		   (feasible (not (null best-choice)))
		   (tpn-out nil)
		   (env-out (make-instance 'environment :vehicles vehicle-ht :features feature-ht :obstacles obstacle-ht))
		   (explanations nil)
		   (tpn-int tpn))

	  ;;(format t "Current TPN: ~S~%" tpn)
	  (format t "Feasible? ~A Best-choice: ~A Generate explanation?: ~a~%" feasible best-choice generate-exp?)
	  


	  ;; write out the tpn
	  (when feasible
		
		(let ((tpn-out-string (make-array '(0) :element-type 'base-char :fill-pointer 0 :adjustable t)))
		  (with-output-to-string (tpn-out-stream tpn-out-string)
			(write-tpn tpn-out-stream
					   tpn
					   ;; Include episodes that are active independent of a choice.
					   :environment (adjoin tpn::*always-active* 
											(tpn::internalize-choice-assignment-descriptions best-choice tpn))
					   :force-labeling? t))
		  (setf tpn-out (first (parse-tpn-string tpn-out-string 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn)))))
	  
	  ;; write out the explanation stream.
	  (when (and (not feasible) generate-exp?)
		(initialize-tpn-choice-selection! tpn)
		(let ((discrete-relaxations (TPN::tpn-explanations tpn :exp-count 3)))

		  (loop for discrete-relaxation in discrete-relaxations


			   do (push (TPN::continuous-constraint-relaxation tpn discrete-relaxation) explanations)))

	  (format t "Explanation: ~A~%" explanations)
	  (format t "TPN-out: ~A~%" tpn-out))


	  (values feasible tpn-out env-out best-choice explanations tpn-int)
	  )
	))

(defun run-vg-on-episode (episode tpn environment vg radius)
  (let ((obstacle-ht (env-obstacles environment))
		(feature-ht (env-features environment))
		(vehicle-ht (env-vehicles environment)))
	(let ((primitive (tpn::episode-primitive episode))) 
	  (when primitive
		(multiple-value-bind (match vehicle-name action-name arguments)
			(match-primitive primitive)
		  (when (and match vehicle-name action-name arguments)
			(let* ((vehicle nil)
				   (action nil)
				   (start-feat-name nil)
				   (end-feat-name nil)
				   (start-feat nil)
				   (end-feat nil))
			  (setf vehicle (gethash vehicle-name vehicle-ht) )
			  (if (null vehicle) 
				  (format t "Warning: Vehicle ~s not found in environment. Not running VZgraph preprocessor on this primitive.~%" vehicle-name) 
				  (progn  
					(setf action (gethash action-name (env-vehicle-actions vehicle)))
					(if (null action)
						(format t  "Warning: Action ~s not found in environment. Not running VZgraph preprocessor on this primitive.~%" action-name)
						(progn 
						  (setf start-feat-name (error-on-nil (string-trim " \"" (first arguments)) 
															  "From location in ~s is nil." primitive))
						  (setf end-feat-name (error-on-nil (string-trim " \"" (second arguments)) 
															"To location in ~s is nil." primitive))
						  (setf start-feat (error-on-nil (env-feature-center (gethash start-feat-name feature-ht)) 
														 "Location ~s not found in environment" start-feat-name))
						  (setf end-feat (error-on-nil (env-feature-center (gethash end-feat-name feature-ht)) 
													   "Location ~s not found in environment" end-feat-name ))
						  
						  (insert-vg-episodes vehicle action start-feat end-feat radius vg episode tpn obstacle-ht feature-ht arguments)
						  ))
					))
			  ))))))
  )

(defun insert-vg-episodes (vehicle action start-feat end-feat radius vg episode tpn obstacle-ht feature-ht orig-arguments)
  (declare (ignore vehicle orig-arguments))
  (let ((min-velocity (env-vehicle-action-minv action))
		(max-velocity (env-vehicle-action-maxv action))
		(path (get-path vg start-feat end-feat))
		(orig-start-event (tpn::episode-start episode))
		(orig-end-event (tpn::episode-end episode)))

	;; There is a bug in Kirk, where the min and max vehicle velocities cannot be the same.
	;; 11/23/2011, David Wang, This bug should be fixed
	;; (if (eql min-velocity max-velocity)
	;; 	(setf max-velocity (* max-velocity 1.00001)))

	;; Delete the primitive from the previous episode if we are not outputting an stn
;;	(delete :PRIMITIVE (rest (tpn::episode-description episode)) :test #'equalp :key (lambda (x) (first x)))

	;; Add the new episodes
	(loop for i from 0 to (- (length path) 2) 
	   with j = 0
	   with (start-event end-event new-node-name new-feat-name distance) do
		 (setf j (+ i 1))
		 (setf distance (get-cost-from-path vg (list (nth i path) (nth j path))))
		 (setf start-event 
			   (if (eql i 0)
				   orig-start-event
				   (tpn::get-event new-node-name tpn)))
		 (if (eql j (- (length path) 1))
			 (setf end-event orig-end-event)
			 (progn (setf new-feat-name (get-feature-name (nth j path) obstacle-ht))
					(setf new-node-name (get-unique-node-name tpn))
					(setf end-event (tpn::get-event new-node-name tpn))
					(setf (gethash new-feat-name feature-ht) 
						  (make-instance 'env-feature :name new-feat-name :center (nth j path) :radius radius)))
			 )

		 ;; modified by Peng. Change the original arc directly instead of adding a new one.
		 (setf (tpn::episode-lowerbound episode) (/ distance max-velocity))
		 (setf (tpn::episode-upperbound episode) (/ distance min-velocity))
		 ;; suppress the unused variable warnings
		 (list start-event end-event)

		 ;; (tpn::add-episode 
		 ;;  start-event end-event
		 ;;  (tpn::make-episode 
		 ;;   start-event end-event
		 ;;   (/ distance max-velocity) ;; lb 
		 ;;   (/ distance min-velocity) ;; ub
		 ;;   0 ;; cost
		 ;;   `(:EPISODE (:PRIMITIVE
		 ;; 			   ,(make-primitive-string (env-vehicle-name vehicle) (env-vehicle-action-name action) (nth i path) (nth j path) feature-ht orig-arguments)))
		 ;;   tpn
		 ;;   :support :CONTROL-PROGRAM))
		 )))


(defun match-primitive (primitive)
  "Matches a primitive string of the form VEHICLE.ACTION(ARG,ARG,...)
   and returns multiple values '(t VEHICLE ACTION (ARG ARG ...)).
   The first argument is t if there is a match, nil if not."
 (let ((match nil)
	   (regs nil))
       (multiple-value-setq (match regs)
         (cl-ppcre:scan-to-strings "\\s*(.*)\\s*\\.\\s*(.*)\\s*\\(\\s*(.*)\\s*\\)\\s*"  primitive))
       ;(format t "~s~%" match)
       ;(format t "~s~%" regs)
	   (if match
		   (let ((vehicle (string-trim '(#\Space #\Tab #\Newline) (aref regs 0)))
				 (action (string-trim '(#\Space #\Tab #\Newline) (aref regs 1)))
				 (arguments (aref regs 2)))
			 (setf vehicle (if (string= vehicle "") nil vehicle))
			 (setf action (if (string= action "") nil action))
			 (setf arguments (cl-ppcre:split "\\s*,\\s*" arguments))
			 (setf arguments (mapcar #'(lambda (x) (string-trim " " x))  arguments))
			 (values t vehicle action arguments))
		   (values nil nil nil nil))
	   ))

(defun get-feature-name (vertex obstacle-ht)
  ;; Creates a unique feature name that does not appear
  ;; in the tpn
  (maphash (lambda (obstacle-name obstacle)
			 ;; Find the feature in the list of obstacles
			 (let ((index (position vertex (point-list (env-obstacle-polygon obstacle)) :test #'equalp-geo)))
			   (when index
				 (return-from get-feature-name (format nil "~a.~a" obstacle-name index)))))
		   obstacle-ht))

(defun get-unique-node-name (tpn)
  (let ((unique-name nil)
		(i 0))
	(loop while t do
		 (setf unique-name (format nil "kirk~s" i))
		 (when (null (gethash unique-name (tpn::tpn-event-ht tpn)))
		   (return unique-name))
		 (incf i)))
  )

(defun make-primitive-string (vehicle-name action-name start-feat end-feat feature-ht orig-arguments)
  (let ((start-feat-name nil)
		(end-feat-name nil))
	(maphash (lambda (key value)
			   (declare (ignore key))
			   (when (equalp-geo start-feat (env-feature-center value)) (setf start-feat-name (env-feature-name value)))) 
			 feature-ht)
	(maphash (lambda (key value) 
			   (declare (ignore key))
			   (when (equalp-geo end-feat (env-feature-center value)) (setf end-feat-name (env-feature-name value)))) 
			 feature-ht)
	;; create the arguments-string
	(let ((arguments-string (concatenate 'string "(\"" start-feat-name "\",\"" end-feat-name "\""))
		  (skip2args 2))
	  (loop for arg-string in orig-arguments do
		   (if (eql skip2args 0)
			   (setf arguments-string (concatenate 'string arguments-string "," arg-string))
			   (decf skip2args)))
	  (setf arguments-string (concatenate 'string arguments-string ")"))	 
	  ;; create the new primitive-string
	  (concatenate 'string 
				   vehicle-name "." action-name arguments-string))))

(defun remove-episode (episode)
  (with-slots (start end tp) episode
	(remove episode (tpn::event-successor-episodes start))
	(remove episode (tpn::event-predecessor-episodes end))
	(remove episode (tpn::tp-episodes tp))
	(with-slots (last-of-episodes) tp
	  (setf last-of-episodes (last (tpn::tp-episodes tp)))))
)
