;;;; Copyright (c) 2009-2010 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:kirk)

(defun kirk-file (tpn-in-pathname tpn-out-pathname &key (exp-out-pathname nil) &allow-other-keys)
  "Given an RMPL file and an Environment file, this method will create
  the optimal trajectory file."
  (with-open-file (tpn-in-stream tpn-in-pathname)
	(with-open-file (tpn-out-stream tpn-out-pathname
									:direction :output
									:if-exists :supersede
									:if-does-not-exist :create)
	  (if exp-out-pathname
		  (with-open-file (exp-out-stream exp-out-pathname
										  :direction :output
										  :if-exists :supersede
										  :if-does-not-exist :create)
			(kirk-stream tpn-in-stream tpn-out-stream :exp-out-stream exp-out-stream))
		  (kirk-stream tpn-in-stream tpn-out-stream))
	  ))
  )

(defun kirk-stream (tpn-in-stream tpn-out-stream &key (exp-out-stream nil) &allow-other-keys)
  "Given TPN-IN-STREAM, an input stream with a TPN in SPEC2 XML representaiton, 
   this method will determine if the trajectory is feasible.
  
   If the plan is feasible, the TPN solution will be written to TPN-OUT-STREAM. Nothing will be written to EXP-OUT-STREAM.
   If the plan is NOT feasible, the explanation will be written to EXP-OUT-STREAM. Nothing will be written to TPN-OUT-STREAM.

   Returns T if temporally feasible, NIl otherwise."
  (let* ((tpns (parse-tpn-stream tpn-in-stream
							   'tpn::get-event 'tpn::make-episode
							   'tpn::add-episode 'tpn::make-tpn))
		 (tpn (first tpns)))
	;; run the kirk algorithm
	(multiple-value-bind (feasible tpn-out explanations) 
		(kirk tpn :explanation? t)
	  ;; write the solution tpn
	  (if (null tpn-out)
		 (write-string "No Solution." tpn-out-stream)
		 (write-tpn tpn-out-stream tpn-out :environment :all :force-labeling? t))
	  ;; write the explanations
	  (when explanations

	   ;; (setf *choice-queue* (get-decision-queue tpn))

		(write-explanations exp-out-stream explanations))
	  
	  feasible)))


(defun kirk (tpn &key (explanation? nil) &allow-other-keys)
  "Given a TPN object, returns three values: (FEASIBLE, TPN-OUT, EXPLANATIONS)

   If the plan is feasible, the first value is T, the second value is the TPN object solution, the third value is nil.
   If the plan is NOT feasible, the first value is NIL, the second value is NIL, the third value is the explanation."

  (unless (not (null tpn))
	(error "Kirk cannot handle a null TPN."))

  ;; initialize choice selection problem (erases all information from previous solutions)
  (initialize-tpn-choice-selection! tpn)
  (print "ddddddddddddddddddddddddddddddddddddd")
  ;; find the best choice
  (let* ((best-choice (tpn-best-choice tpn))
		 (feasible (not (null best-choice))) 
		 (tpn-out nil)
		 (explanations nil))
	;; (format t "Best Choice is ~A~%" best-choice)
	
	  (format t "Feasible? ~A Best-choice: ~A ~%" feasible best-choice)

	;; write out the tpn
	(when feasible
	  (let ((tpn-out-string (make-array '(0) :element-type 'base-char :fill-pointer 0 :adjustable t)))
		(with-output-to-string (tpn-out-stream tpn-out-string)
		  (write-tpn tpn-out-stream
					 tpn
					 ;; Include episodes that are active independent of a choice.
					 :environment (adjoin tpn::*always-active* 
										  (tpn::internalize-choice-assignment-descriptions best-choice tpn))
					 :force-labeling? t))
		(setf tpn-out (first (parse-tpn-string tpn-out-string 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn)))) )

	;; write out the explanation stream.
	(when (and (not feasible) explanation?)
	  (initialize-tpn-choice-selection! tpn)

		(let ((discrete-relaxations (TPN::tpn-explanations tpn :exp-count 5)))

		  (format t "Discrete Explanations: ~A~%" discrete-relaxations)

		  (loop for discrete-relaxation in discrete-relaxations

			   do (push (TPN::continuous-constraint-relaxation tpn discrete-relaxation) explanations)))

	  ;; (setf explanations (list (TPN::continuous-constraint-relaxation tpn (first (TPN::tpn-explanations tpn :exp-count 1)))))
	  )

	(format t "Continuous Explanations: ~A~%" explanations)

	(values feasible tpn-out explanations)
	))