;;;; Copyright (c) 2009-2011 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:kirk)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; KIRKRM-STATE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass kirkrm-state ()
  (
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;; PARAMETERS
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   (time-tolerance 
	:type number
	:initform 0.5 :initarg :time-tolerance
	:accessor kirkrm-time-tolerance
	:documentation "The amount of error tolerated between the current time and an event's 
      exepected execution time, before the event is considered missed.")
   (ignore-missed-activities
	:type t
	:initform t :initarg :ignore-missed-activities
	:accessor kirkrm-ignore-missed-activities
	:documentation "If true, the dispatcher will ignore missed activities.")
   (cost-per-priority 
	:type number
	:initform 100 :initarg :cost-per-priority
	:accessor kirkrm-cost-per-priority
	:documentation "The cost per priority level.")
   (num-priorities
	:type number
	:initform 10 :initarg :num-priorities
	:accessor kirkrm-num-priorities
	:documentation "The number of priority levels.")

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;; PLANNER STATE
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   (base-env
	:type t
	:initform nil :initarg :base-env
	:accessor kirkrm-base-env
	:documentation "The submitted start environment plus any obstacles added or removed.")
   (master-env
	:type t
	:initform nil :initarg :master-env
	:accessor kirkrm-master-env
	:documentation "The base env plus any features submitted.")
   (master-tpn
	:type t
	:initform nil :initarg :master-tpn
	:accessor kirkrm-master-tpn
	:documentation "The tpn created by merging the submissions.")
   (master-dispatch 
	:type t
	:initform nil :initarg :master-dispatch
	:accessor kirkrm-master-dispatch
	:documentation "The Drake dispatcher instance.")
   (dispatcher-start-time
	:type number
	:initform 0 :initarg :dispatcher-start-time
	:accessor kirkrm-dispatcher-start-time
	:documentation "The global starting time of the previous run of the dispatcher")

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;; SUBMISSION MANAGEMENT
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   (waiting-submissions
	:type linked-hashset
	:initform (make-linked-hashset :key-function #'submission-id :test #'equal)
	:initarg :waiting-submissions
	:accessor kirkrm-waiting-submissions
	:documentation "The set of submissions that have been submitted (waiting to start).")
   (started-submissions
	:type linked-hashset
	:initform (make-linked-hashset :key-function #'submission-id :test #'equal)
	:initarg :started-submissions
	:accessor kirkrm-started-submissions
	:documentation "The set of submissions that have been started (first event has been executed).")
   (completed-submissions
	:type linked-hashset
	:initform (make-linked-hashset :key-function #'submission-id :test #'equal)
	:initarg :completed-submissions
	:accessor kirkrm-completed-submissions
	:documentation "The set of submissions that have been completed (last event has been executed).")
   (event-dictionary
	:type dictionary
	:initform (make-dictionary`((:name :ldg-symbol  :key-fn ,#'event-data-ldg-symbol  :test ,#'eql)
											  (:name :master-name :key-fn ,#'event-data-master-name :test ,#'equal)
											  (:name :tpn-event   :key-fn ,#'event-data-tpn-event   :test ,#'eql)
											  ))
	:initarg :event-dictionary
	:accessor submission-event-dictionary
	:documentation "A mapping between events in the LDG, MASTER, and TPN"
	)

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;; THREADING
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   (running-flag
	:type mtk-utils-threading:flag
	:initform (mtk-utils-threading:make-flag :value nil)
	:accessor kirkrm-running-flag
	:documentation   "A boolean value. T if kirkrm has been started. 
      NIL if kirkrm has been stopped." )
   (lock 
	:initform (bordeaux-threads:make-recursive-lock) 
	:initarg :lock
	:accessor kirkrm-lock
	:documentation "Lock used to make access to kirkrm-state atomic.")
   ))


(defmethod reset ((kirkrm kirkrm-state))
  (bordeaux-threads:with-recursive-lock-held ((kirkrm-lock kirkrm))
	(with-slots (base-env master-env master-tpn master-dispatch dispatcher-start-time
						  waiting-submissions started-submissions completed-submissions
						  event-dictionary) kirkrm
	  (setf base-env nil)
	  (setf master-env nil)
	  (setf master-tpn nil)
	  (setf master-dispatch nil)
	  (setf dispatcher-start-time 0)
	  (coll-clear! waiting-submissions)
	  (coll-clear! started-submissions)
	  (coll-clear! completed-submissions)
	  ;; set up the dictionary
	  (dictionary-clear event-dictionary)
	  ;;;; add the master start
	  (dictionary-add  event-dictionary 
	   (make-instance 'event-data
					  :ldg-symbol (intern "master-start" :keyword)
					  :master-name "master-start"
					  :tpn-event nil :submission nil
					  :tpn-start-p nil :tpn-end-p nil
					  ))
	  ;;;; add the master end
	  (dictionary-add  event-dictionary 
	   (make-instance 'event-data
					  :ldg-symbol (intern "master-end" :keyword)
					  :master-name "master-end"
					  :tpn-event nil :submission nil
					  :tpn-start-p nil :tpn-end-p nil
					  ))
	  )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; METHODS TO KEEP TRACK OF SUBMISSIONS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro make-ldg-symbol (symbol-name)
  `(intern ,symbol-name :keyword))

(defmethod add-submission-to-dictionary ((kirkrm kirkrm-state)(submission submission))
  (with-slots (event-dictionary) kirkrm
	(with-slots (tpn-obj noop-name noopend-name) submission
	  ;; store events from the submission's tpn into the event dictionary
	  (dolist (event (tpn::tpn-events tpn-obj))
		(let* ((master-name (make-master-name submission event))
			   (event-data (make-instance 'event-data 
										  :ldg-symbol (make-ldg-symbol master-name)
										  :master-name master-name
										  :tpn-event event :submission submission
										  :tpn-start-p (eql event (tpn::tpn-start tpn-obj))
										  :tpn-end-p (eql event (tpn::tpn-end tpn-obj))  )))
		  (dictionary-add event-dictionary event-data)
		  ))
	  ;; store the noop event into the dictionary
	  (dictionary-add 
	   event-dictionary (make-instance 'event-data
									   :ldg-symbol (make-ldg-symbol noop-name)
									   :master-name noop-name
									   :tpn-event nil :submission submission
									   :tpn-start-p nil	:tpn-end-p nil))
	  ;; store the noopend event into the dictionary
	  (dictionary-add 
	   event-dictionary (make-instance 'event-data
									   :ldg-symbol (make-ldg-symbol noopend-name)
									   :master-name noopend-name
									   :tpn-event nil :submission submission
									   :tpn-start-p nil :tpn-end-p nil))
	  )))

(defmethod remove-submission-from-dictionary ((kirkrm kirkrm-state)(submission submission))
  (with-slots (event-dictionary) kirkrm
	(with-slots (tpn-obj noop-name noopend-name) submission
	  ;; remove events from the submission's tpn from the event dictionary
	  (dolist (event (tpn::tpn-events tpn-obj))
		(dictionary-remove event-dictionary :tpn-event event))
	  ;; remove the noop event from the dictionary
	  (dictionary-remove event-dictionary :master-name noop-name)
	  ;; remove the noopend event from the dictionary
	  (dictionary-remove event-dictionary :master-name noopend-name)
	  )))

(defvar submission-state-transitions 
  (make-array '(5 5) :initial-contents '((t   t   nil nil nil)
										 (nil t   t   t   t  )
										 (nil nil t   t   t  )
										 (nil nil nil t   t  )
										 (nil nil nil nil t  ))
			  ))

(defmethod submission-state! ((kirkrm kirkrm-state)(submission submission)(desired-state integer))
  (with-slots (waiting-submissions started-submissions completed-submissions) kirkrm
	(with-slots (state) submission
	  ;; check if we are transitioning between the same states
	  (when (= state desired-state)
		(return-from submission-state! nil))
	  ;; check if the transition is allowed
	  (unless (aref submission-state-transitions state desired-state)
		(error "A submission cannot transition from state ~a to ~a." state desired-state))
	  ;; handle the valid transition
	  ;; remove it from its current list
	  (cond ((= state state-waiting) (coll-remove! waiting-submissions submission))
			((= state state-started) (coll-remove! started-submissions submission))
			((= state state-completed) (coll-remove! completed-submissions submission)))
	  ;; add it to the new list
	  (cond ((= desired-state state-waiting) (coll-add! waiting-submissions submission))
			((= desired-state state-started) (coll-add! started-submissions submission))
			((= desired-state state-completed) (coll-add! completed-submissions submission)))
	  ;; add or remove the submission to the event dictionary
	  (let ((prev (or (= state state-waiting) (= state state-started)))
			(curr (or (= desired-state state-waiting) (= desired-state state-started))))
		(when (and prev (not curr))
		  (remove-submission-from-dictionary kirkrm submission))
		(when (and (not prev) curr)
		  (add-submission-to-dictionary kirkrm submission))
		)
	  ;; set the desired state
	  (setf state desired-state)
	  )))

(defmethod submission-waiting! ((kirkrm kirkrm-state)(submission submission))
  ;; a submission can only be waiting if it was in the none state.
  (with-slots (waiting-submissions started-submissions completed-submissions) kirkrm
	(cond ((or (coll-contains? waiting-submissions submission)
			   (coll-contains? started-submissions submission)
			   (coll-contains? completed-submissions submission))
		   (error "Received a duplicate submission. A submissison with id ~s already exists." (submission-id submission)) 
		   ;; return
		   nil)
		  (:otherwise
		   (submission-state! kirkrm submission state-waiting)
		   ;; return
		   t)
		  )))

(defmethod submission-started! ((kirkrm kirkrm-state)(submission submission))
  ;; a submission can only be started if it was waiting before
  (submission-state! kirkrm submission state-started) )

(defmethod submission-completed! ((kirkrm kirkrm-state)(submission submission))
  ;; a submission can only be completed if it was waiting or started before
  (submission-state! kirkrm submission state-completed) )

(defmethod submission-cancelled! ((kirkrm kirkrm-state)(submission submission))
  (submission-state! kirkrm submission state-cancelled) )

(defmethod submission-cancelled! ((kirkrm kirkrm-state)(submission-id string))
  (let ((submission (get-submission kirkrm submission-id)))
	(cond (submission
		   (submission-state! kirkrm submission state-cancelled)
		   t)
		  (:otherwise 
		   nil))
	))

(defmethod get-submission ((kirkrm kirkrm-state) (submission-id string))
  (bordeaux-threads:with-recursive-lock-held ((kirkrm-lock kirkrm))
	(with-slots (waiting-submissions started-submissions completed-submissions) kirkrm
	  (let ((submission))
		(setf submission (coll-get waiting-submissions submission-id))
		(unless submission
		  (setf submission (coll-get started-submissions submission-id)))
		(unless submission
		  (setf submission (coll-get completed-submissions submission-id)))
		submission))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; EVENT-DATA METHODS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod get-execution-time ((kirkrm kirkrm-state)(event tpn::event))
  "Returns the execution time of the given submission tpn event. 
   Returns NIL if the execution time is NOT FOUND or is less than 0."
  (with-slots (event-dictionary) kirkrm
	(let ((event-data (dictionary-get event-dictionary :tpn-event event))
		  (event-exec-time))
	  (cond (;; if the event-data exists, retrieve the paired ldg symbol.
			 event-data (setf event-exec-time (event-data-execution-time event-data))
						(if (< event-exec-time 0) nil event-exec-time))
			(;; if event-data doesn't exist, return nil.
			 :otherwise nil))
	  )))

(defmethod get-ldg-symbol ((kirkrm kirkrm-state)(event tpn::event))
  "Returns the symbol paired with the given submission tpn event
  Returns NIL if the event cannot be found."
  (with-slots (event-dictionary) kirkrm
	(let ((event-data (dictionary-get event-dictionary :tpn-event event)))
	  (cond (;; if the event-data exists, retrieve the paired ldg symbol.
			 event-data (event-data-ldg-symbol event-data))
			(;; if event-data doesn't exist, return nil.
			 :otherwise nil))
	  )))

(defmethod get-ldg-symbol ((kirkrm kirkrm-state)(master-name string))
  "Returns the symbol paired with the given submission tpn event
  Returns NIL if the master-name cannot be found."
  (with-slots (event-dictionary) kirkrm
	(let ((event-data (dictionary-get event-dictionary :master-name master-name)))
	  (cond (;; if the event-data exists, retrieve the paired ldg symbol.
			 event-data (event-data-ldg-symbol event-data))
			(;; if event-data doesn't exist, return nil.
			 :otherwise nil))
	  )))

(defmethod get-master-name ((kirkrm kirkrm-state)(event tpn::event))
  "Returns the symbol paired with the given submission tpn event
  Returns NIL if the event cannot be found."
  (with-slots (event-dictionary) kirkrm
	(let ((event-data (dictionary-get event-dictionary :tpn-event event)))
 	  (cond (;; if the event-data exists, retrieve the paired master-name.
			 event-data (event-data-master-name event-data))
			(;; if event-data doesn't exist, return nil.
			 :otherwise nil))
	  )))
