;;;; Copyright (c) 2009-2011 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:kirk)

;; Class TEST-RESOURCE-MANAGER
;; depends on Class RESOURCE-MANAGER-INTERFACE, which requires
;; abcl packages such as "java::". When compiling under allegro,
;; (or any other lisp runtime)
;; using the allegro variant means the resource-manager won't run
;; but, it will still compile.
#+abcl
(defclass test-resource-manager (resource-manager::resource-manager-interface) ())
#-abcl
(defclass test-resource-manager () ())
 
;;; global state
(defvar *kirkrm-state* (make-instance 'kirkrm-state :cost-per-priority 100 :num-priorities 10)
  "Maintains global state information used to run the Kirk Resource Manager.")
(defvar *resource-manager* nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; RUNNING METHODS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro with-running-flag-set (running-flag &rest body)
  "A macro that will only run the included body code if the RUNNING-FLAG is set to T.
The BODY will be run in mutex with other BODY code enclosed by this macro.
RUNNING-FLAG must be an instance of mtk-utils-threading:FLAG class."
  (let ((run-flag (gensym "FLAG-"))
		(flag-lock (gensym "FLAG-LOCK-")))
	`(let ((,run-flag ,running-flag)
		   (,flag-lock (mtk-utils-threading:flag-lock ,running-flag)))
	   ;; The above two let layers ensure the dispatch argument 
	   ;; is evaluated once, and there is no variable-name clash
	   ;; See "http://www.apl.jhu.edu/~hall/Lisp-Notes/Macros.html"
	   ;; for the challenges of writing macros.
	   (bordeaux-threads:with-recursive-lock-held (,flag-lock)
		 ;; Check the state of the run flag
		 (if (mtk-utils-threading:flag? ,run-flag :wait nil)
			 ;; The run flag is T, so we can execute the body
			 (progn ,@body)
			 ;; The run flag is NIL, signal an error
			 (error "The START method has not been called. The Kirk Resource Manager is not running."))
	   )
	)))

(defmacro set-running-flag (running-flag value &rest body)
  "A macro that will set the value of the RUNNING-FLAG to VALUE.
   If the value of RUNNING-FLAG changes, BODY will be executed.
   If the VALUE is the same, BODY will not be executed."
   (let ((run-flag (gensym "FLAG-"))
		 (flag-lock (gensym "FLAG-LOCK-"))
		 (val (gensym "VALUE-")))
	`(let ((,run-flag ,running-flag)
		   (,flag-lock (mtk-utils-threading:flag-lock ,running-flag))
		   (,val ,value))
	   (bordeaux-threads:with-recursive-lock-held (,flag-lock)
		 ;; Check the state of the run flag
		 (unless (eql (mtk-utils-threading:flag? ,run-flag :wait nil) ,val)
			 (mtk-utils-threading:flag! ,run-flag ,val)
			 (progn ,@body)))
	   )))

(defun kirkrm-running? (&key (wait nil))
  (mtk-utils-threading:flag? (kirkrm-running-flag *kirkrm-state*) :wait wait))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; DEBUG PRINT METHODS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *output-detail-level* 100)

(defun level-format (eval-level &rest body)
  (let ((length (length body)))
	 (setf eval-level (min eval-level length))
	 (when (> eval-level 0)
	   (format t "~a" (nth (- eval-level 1) body)))
	 ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; INTERFACE METHODS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod getFeasibility ((interface test-resource-manager) (id string) (tpn string) (env string) &key (priority 5) (start-time `(:relative 0)))
  (with-running-flag-set (kirkrm-running-flag *kirkrm-state*)
	(mtk-utils:format-time t "getFeasibility(tpn = ~s, env = ~s)~%" tpn env)

	;; kill the dispatcher
	(kill-dispatcher)

	(bordeaux-threads:with-recursive-lock-held ((kirkrm-lock *kirkrm-state*))
	  ;; parse the input strings into tpn objects
	  (let ((submission (make-submission :id id :tpn tpn :env env :priority priority :start-time start-time)))
		;; submit
		(submission-waiting! *kirkrm-state* submission)

		(multiple-value-bind 
			  (feasible temp-master-tpn temp-master-env temp-master-dispatch skipped-submissions) (calculate-plan)
		  (declare (ignore temp-master-env temp-master-dispatch))

		  ;; find out if the submission has been planned away
		  (when feasible
			(setf feasible (not (find id skipped-submissions :test #'equal))))

		  ;; remove submission
		  (submission-cancelled! *kirkrm-state* submission)

		  ;; start the dispatcher
		  (start-dispatcher)

		  (if feasible
			  (multiple-value-bind (lb-span ub-span)
				  (compute-span temp-master-tpn)
				(list "success" (remove id skipped-submissions :test #'equal) lb-span ub-span))
			  (list "fail" nil nil nil))
		  )
		))))

(defmethod submit ((interface test-resource-manager) (id string) (tpn string) (env string) &key (priority 5) (start-time `(:relative 0)))
  (with-running-flag-set (kirkrm-running-flag *kirkrm-state*)
    (mtk-utils:format-time t "submit(tpn = ~s env = ~s)~%" tpn env)	

	;; kill the dispatcher
	(kill-dispatcher)

	(bordeaux-threads:with-recursive-lock-held ((kirkrm-lock *kirkrm-state*))
	  (let ((submission (make-submission :id id :tpn tpn :env env :priority priority :start-time start-time)))
		;; submit
		(submission-waiting! *kirkrm-state* submission)

		;; ;; kill the dispatcher
		;; (kill-dispatcher)

		(multiple-value-bind 
			  (feasible temp-master-tpn temp-master-env temp-master-dispatch skipped-submissions) (calculate-plan)

		  ;; print out the temp-master-tpn
		  ;; (mtk-utils:format-time t "TEMP-MASTER-TPN:~%~a~%" 	(with-output-to-string (tpn-stream)
		  ;; 											  (tpn::write-tpn tpn-stream temp-master-tpn :force-labeling? t)))
		  ;; (mtk-utils:format-time t "SKIPPED-SUBMISSIONS: ~a~%" skipped-submissions)

		  ;; find out if the submission has been planned away
		  (when feasible
			(setf feasible (not (find (submission-id submission) skipped-submissions :test #'equal))))
		  (when feasible
			(with-slots (master-tpn master-env master-dispatch) *kirkrm-state*
			  (setf master-tpn temp-master-tpn)
			  (setf master-env temp-master-env)
			  (setf master-dispatch temp-master-dispatch)))
		  ;; remove the submission if it is not feasible
		  (unless feasible
			;; remove submission
			(submission-cancelled! *kirkrm-state* submission))

		  ;; start the dispatcher
		  (start-dispatcher)

		  ;; (format t "feasible: ~a ~%" feasible)

		  (if feasible
			  (list "success" (remove (submission-id submission) skipped-submissions :test #'equal))
			  (list "fail" (remove (submission-id submission) skipped-submissions :test #'equal)))
		  )
		))))

(defmethod cancel ((interface test-resource-manager) (id string) (tpn string))
  (with-running-flag-set (kirkrm-running-flag *kirkrm-state*)
    (mtk-utils:format-time t "cancel(id = ~s, tpn = ~s)~%" id tpn)

	;; kill the dispatcher
	(kill-dispatcher)

	(bordeaux-threads:with-recursive-lock-held ((kirkrm-lock *kirkrm-state*))

	  ;; remove submission
	  (if (submission-cancelled! *kirkrm-state* id)
		  (progn
			;; ;; kill the dispatcher
			;; (kill-dispatcher)
			(multiple-value-bind 
				  (feasible temp-master-tpn temp-master-env temp-master-dispatch skipped-submissions) (calculate-plan)
			  (when feasible
				(setf feasible (not (find id skipped-submissions :test #'equal))))
			  (when feasible
				(with-slots (master-tpn master-env master-dispatch) *kirkrm-state*
				  (setf master-tpn temp-master-tpn)
				  (setf master-env temp-master-env)
				  (setf master-dispatch temp-master-dispatch)))
			  ;; start the dispatcher
			  (start-dispatcher)  
			  (if feasible
				  "success"
				  "fail")
			  )
			)
		  "dne"
		  ))))

(defmethod start ((interface test-resource-manager) (env string) &key (time-tolerance 0.5) (ignore-missed-activities t))
  (set-running-flag (kirkrm-running-flag *kirkrm-state*) t
	(mtk-utils:format-time t "start(env = ~s)~%" env)

	;; Reset, master parameters
	(reset *kirkrm-state*)

	;; Save off test-resource-manager
	(setf *resource-manager* interface)
	;; Initialize the remaining two parameters in the global state (timetol and ignore-missed-activities)
	(cond ((< time-tolerance 0) (error "start: time-tolerance cannot be less than 0."))
		  (:otherwise (setf (kirkrm-time-tolerance *kirkrm-state*) time-tolerance)))
	(setf (kirkrm-ignore-missed-activities *kirkrm-state*) ignore-missed-activities)
	;; Read the environment string and create an environment. 
	(cond ((null env) (error "start: env string can't be null.")) 
		  (:otherwise 
		   (let ((temp-base-env (with-input-from-string (env-in-stream env)
								  (read-environment-stream env-in-stream))))
			 (cond ((null temp-base-env) (error "start: could not parse the environment-string.")) 
				   ((eql 0 (hash-table-count (env-vehicles temp-base-env))) (error "start: no vehicles defined in the environment."))
				   (:otherwise (setf (kirkrm-base-env *kirkrm-state*) temp-base-env)))
			 )))
	
	"success"))

(defmethod stop ((interface test-resource-manager))
  (set-running-flag (kirkrm-running-flag *kirkrm-state*) nil
	(mtk-utils:format-time t "stop()~%")

	;; Stop the dispatcher
	(kill-dispatcher)
  "success"))

(defmethod addObstacle ((interface test-resource-manager) (obstacle string) &key (replace t))
  (with-running-flag-set (kirkrm-running-flag *kirkrm-state*)
	(mtk-utils:format-time t "addObstacle(obstacle = ~s)~%" obstacle)

	(bordeaux-threads:with-recursive-lock-held ((kirkrm-lock *kirkrm-state*))
	  ;; Save off the obstacle
	  (with-slots (base-env) *kirkrm-state*
		(let* ((obstacle-obj (make-obstacle-from-lxml (parse-xml-string obstacle)))
			   (obstacle-name (env-obstacle-name obstacle-obj)))
		  (if (gethash obstacle-name (env-obstacles base-env))
			  (if replace
				  (progn (setf (gethash obstacle-name (env-obstacles base-env)) obstacle-obj)
						 "replaced")
				  "failed")
			  (progn (setf (gethash obstacle-name (env-obstacles base-env)) obstacle-obj)
					 "success")
			  ))))
	))

(defmethod removeObstacle ((interface test-resource-manager) (obstacle string))
  (with-running-flag-set (kirkrm-running-flag *kirkrm-state*)
	(mtk-utils:format-time t "removeObstacle(obstacle = ~s)~%" obstacle)

	(bordeaux-threads:with-recursive-lock-held ((kirkrm-lock *kirkrm-state*))
	  (let* ((obstacle-obj nil)
			 (obstacle-name nil))
		(handler-case
			(progn
			  (setf obstacle-obj (make-obstacle-from-lxml (parse-xml-string obstacle)))
			  (setf obstacle-name (env-obstacle-name obstacle-obj)))
		  (s-xml:xml-parser-error () (setf obstacle-name obstacle)))
		(if (remhash obstacle-name (env-obstacles (kirkrm-base-env *kirkrm-state*)))
			"success"
			"fail")))))

(defmethod replan ((interface test-resource-manager))
 (with-running-flag-set (kirkrm-running-flag *kirkrm-state*)
	(level-format *output-detail-level*
				  (mtk-utils:format-time nil "replan()~%"))
	;; kill the dispatcher
	(kill-dispatcher)
	(bordeaux-threads:with-recursive-lock-held ((kirkrm-lock *kirkrm-state*))
	  ;; parse the input strings into tpn objects
	  (multiple-value-bind 
			(feasible temp-master-tpn temp-master-env temp-master-dispatch skipped-submissions) (calculate-plan)
		(declare (ignore temp-master-env temp-master-dispatch))
		;; start the dispatcher
		(start-dispatcher)
		(if feasible
			(multiple-value-bind (lb-span ub-span)
				(compute-span temp-master-tpn)
			  (list "success" skipped-submissions lb-span ub-span))
			(list "fail" nil nil nil))
		)
		))
 )

(defmethod getTPNandENV ()
  "Returns a list containing two strings, the XML representations 
   of the MASTER-TPN and the MASTER-ENV."
  (with-running-flag-set (kirkrm-running-flag *kirkrm-state*)
	(bordeaux-threads:with-recursive-lock-held ((kirkrm-lock *kirkrm-state*))
	  (with-slots (master-tpn master-env) *kirkrm-state*
		(if (and master-tpn master-env)
			(list
			 (with-output-to-string (tpn-stream)
			   (tpn::write-tpn tpn-stream master-tpn :force-labeling? t))
			 (write-environment-string master-env))
			(list nil nil))))))

(defmethod test-notify ((interface test-resource-manager) (taskId string) (submitterId string) (status string) (description string) (resourceUri string) (resourceName string))
  #+abcl
  (resource-manager::notify-task-status interface taskId submitterId status description resourceUri resourceName)
  "success")

(defmethod test-notify-xml ((interface test-resource-manager) (xmlString string))
  #+abcl
  (resource-manager::notify-task-status-xml interface xmlString)
  "success")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; HELPER METHODS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun update-master-plan2 (executed-events)
  (bordeaux-threads:with-recursive-lock-held ((kirkrm-lock *kirkrm-state*))
	(with-slots (event-dictionary dispatcher-start-time) *kirkrm-state*
	  (loop for (ldg-symbol current-time) in executed-events do
		   (let ((event-data (dictionary-get event-dictionary :ldg-symbol ldg-symbol)))	   
			 (cond (;; if the event-data was found in the event-dictionary, we update it.
					event-data
					(with-slots (submission tpn-start-p tpn-end-p execution-time) event-data
					  (cond (;; mark the submission as completed
							 tpn-end-p
							 (submission-completed! *kirkrm-state* submission))
							(;; mark the submission as started
							 tpn-start-p
							 (submission-started! *kirkrm-state* submission))
							)
					  ;; save off the completion time
					  (setf execution-time (+ dispatcher-start-time current-time))))
				   (;; Ignore events not found in the event-dictionary.
					;; These events are from previous runs.
					;; OR are from events inserted by Kirk? - Should we ignore these?
					:otherwise nil)
				   ))
		   ))
	))

(defun show-master-tpn ()
  "Shows the global tpn"
  (with-output-to-string (tpn-out-stream)
	(tpn::write-tpn tpn-out-stream (kirkrm-master-tpn *kirkrm-state*) :force-labeling? t)))

(defun calculate-plan ()
  (let ((temp-master-tpn nil)
		(temp-master-env nil)
		(kirk-tpn nil)
		(kirk-env nil)
		(best-choice nil)
		(temp-master-dispatch nil)
		(feasible t))

	;; call update master-plan
	(when (kirkrm-master-dispatch *kirkrm-state*)
	  (update-master-plan2 (drake2::dispatch-trace (kirkrm-master-dispatch *kirkrm-state*))))

	;; merge master plan & environment with that extracted from dispatcher
	(multiple-value-setq (temp-master-tpn temp-master-env)
	  (merge-master-with-plan (kirkrm-waiting-submissions *kirkrm-state*)
							  (kirkrm-started-submissions *kirkrm-state*)
							  (kirkrm-completed-submissions *kirkrm-state*)
							  (kirkrm-base-env *kirkrm-state*)
							  (kirkrm-master-dispatch *kirkrm-state*)))

	;; call kirk to solve the problem
	(multiple-value-setq (feasible kirk-tpn kirk-env best-choice) ;;kirk-env
			  (kirk-vz-graph temp-master-tpn temp-master-env))
	;;    Even thought we don't use the kirk environment variable, 
	;;      we need to use it in order to suppress a warning.
	kirk-env 

	;; call drake to solve the problem
	(when feasible
	  (setf temp-master-dispatch (tpn-to-dispatch kirk-tpn (kirkrm-master-dispatch *kirkrm-state*)))
	  (setf feasible (not (null temp-master-dispatch))) )

	;; Find out which submissions were skipped
	(let ((environment (adjoin tpn::*always-active* 
							   (tpn::internalize-choice-assignment-descriptions best-choice temp-master-tpn)))
		  (skipped-submissions '()))
	  (tpn::label-tpn-with-activation-assignments! temp-master-tpn)
	  (coll-map (kirkrm-waiting-submissions *kirkrm-state*) 
							 (lambda (submission) 
							   (let ((episode (submission-master-tpn-noop-episode submission) ))
								 (when (and (not (null episode)) 
											(tpn::episode-enabled-by-environment? episode environment))
								   (push (submission-id submission) skipped-submissions)
								   )))	)
	  
	  (coll-map (kirkrm-waiting-submissions *kirkrm-state*)  
							 (lambda (submission)
							   (let ((episode (submission-master-tpn-noop-episode submission) ))
								 (if (and (not (null episode)) 
										  (tpn::episode-enabled-by-environment? episode environment))
									 (push (submission-id submission) skipped-submissions)
									 )
								 )))

	  ;; return answers
	  (if feasible
		  (values feasible kirk-tpn temp-master-env temp-master-dispatch skipped-submissions)
		  (values nil nil nil nil nil)))
	))

(defun merge-master-with-plan (waiting-submissions started-submissions completed-submissions base-env master-dispatch)
  (declare (ignore master-dispatch))
  (let (;; create a copy of the master-tpn
		(new-master-tpn)
		;; create a copy of the master-env
		(new-master-env (copy-env base-env))
		)
	 ;; build up lists of submissions that have started and not.
	(let ((waiting-submissions-list (coll-tolist waiting-submissions))
		  (started-submissions-list (coll-tolist started-submissions))
		  (completed-submissions-list (coll-tolist completed-submissions)))

	  (format t "Submissions waiting: ")
	  (dolist (s waiting-submissions-list) (format t "~a, "(submission-id s)))
	  (format t "~%Submissions started: ")
	  (dolist (s started-submissions-list) (format t "~a, "(submission-id s)))
	  (format t "~%Submissions completed: ")
	  (dolist (s completed-submissions-list) (format t "~a, "(submission-id s)))
	  (format t "~%")

	  (let ((decision-names-list '()))
		(flet ((add-decisions-to-name-list (submission)
				 (loop for decision in (tpn::tpn-decisions (submission-tpn-obj submission)) do
					  (push (tpn::event-name decision) decision-names-list))))
		  ;; add decisions from not-started-submissions
		  (dolist (submission waiting-submissions-list)
			(add-decisions-to-name-list submission)
			;; add the noop-decision
			(push (format nil "~a:noop" (submission-id submission)) decision-names-list))
		  )
		;; build the new master-tpn
		(setf new-master-tpn (tpn::make-tpn "master" "master-start" "master-end" decision-names-list)))
	  ;; add episodes to the master-tpn
	  (let ((running-submissions-list (append started-submissions-list (copy-list waiting-submissions-list))))
		(with-slots (dispatcher-start-time) *kirkrm-state*
		  (setf dispatcher-start-time (get-universal-time));; (+ dispatcher-start-time (drake2::dispatch-current-time master-dispatch))

		;;;; add episodes from started-submissions
		  (dolist (submission started-submissions-list)
			(add-started-submission-to-tpn submission new-master-tpn dispatcher-start-time))
		;;;; add episodes from not-started-submissions
		  (dolist (submission waiting-submissions-list)
			(add-waiting-submission-to-tpn submission new-master-tpn dispatcher-start-time))
		)

		;; Sort the submission so the one with the earliest timestamp comes first					 	   
		(setf running-submissions-list (sort running-submissions-list #'< :key (lambda (submission) (submission-timestamp submission))))
		;; create a total ordering on the submission based on the times they were submitted
		;; Loop over the sorted submissions and add episodes between them
		(loop for curr-submission in (rest running-submissions-list)
		   with prev-submission = (first running-submissions-list) do
			 (make-and-add-episode 
			  :start (tpn::get-event (submission-noopend-name prev-submission) new-master-tpn)
			  :end (tpn::get-event (submission-noop-name curr-submission) new-master-tpn)
			  :lb 0
			  :ub *infinity*
			  :cost 0
			  :tpn new-master-tpn)
			 (setf prev-submission curr-submission)
			 )
		
		;; merge the features from the environment
		(loop for submission in running-submissions-list do
			 (with-slots ((master-features features)) new-master-env
			   (setf master-features
					 (merge-hash-tables master-features 
										(env-features (submission-env-obj submission)))))))
	  ;; return the tpn copy and env
	  (values new-master-tpn new-master-env))

))

(defun add-waiting-submission-to-tpn (submission master-tpn current-global-time)
  (with-slots (num-priorities cost-per-priority) *kirkrm-state*
	(with-slots ((sub-id id) (sub-tpn tpn-obj) (sub-start-time start-time) (sub-priority priority)) submission

	  (let* ((noop-event (tpn::get-event (submission-noop-name submission) master-tpn))
			 (noop-end-event (tpn::get-event (submission-noopend-name submission) master-tpn))
			 (master-start (tpn::tpn-start master-tpn))
			 (master-end (tpn::tpn-end master-tpn))
			 (current-time current-global-time)
			 (start-delay-time (if (equal :relative (first sub-start-time))
								   (second sub-start-time)
								   (- (second sub-start-time) current-time)))
			 ;; calculate costs
			 (noop-cost (- (* (+ (- num-priorities sub-priority) 1) cost-per-priority) 1))
			 (plan-cost (- noop-cost 1))
			 )
		;; suppress negative start-delay-times
		;; this is not necessarily the right solution - what should happen when submissions are so quick, kirk can't process in time,
		;; as a result submission build up that have not been started?
		(when (< start-delay-time 0)
		  (setf start-delay-time 0))
		;; when the time is relative, switch the start time to absolute
		(when (equal :relative (first sub-start-time))
		  (setf sub-start-time `(:absolute ,(+ current-time (second sub-start-time)))))
		;; declare a function to convert from submission-tpn to master names
		(flet ((event-to-master (event) (tpn::get-event (get-master-name *kirkrm-state* event) master-tpn)))
		  ;; merge the tpn - add all episodes
		  (loop for episode in (tpn::tpn-episodes sub-tpn) do
			   (make-and-add-episode 
				:start (event-to-master (tpn::episode-start episode))
				:end (event-to-master (tpn::episode-end episode))
				:lb (to-binf (tpn::episode-lowerbound episode))
				:ub (to-binf (tpn::episode-upperbound episode))
				:cost (/ (if (typep (episode-cost episode) 'constant-episode-cost)
							 (tpn::constant-episode-cost-cost (episode-cost episode))
							 (error "Unable to encode the cost ~a for episode ~a." (episode-cost episode) episode))
						 (* cost-per-priority num-priorities))
				:description (copy-tree (tpn::episode-description episode))
				:tpn master-tpn)
			   )
		  
		  ;; add episode from master-start to noop decision
		  ;; the infinity ub is necessary for the new topological order based on submission
		  (make-and-add-episode :start master-start :end noop-event :lb 0 :ub *infinity* :tpn master-tpn)
		  ;; add episode from noop-event to tpn-start
		  (make-and-add-episode :start noop-event :end (event-to-master (tpn::tpn-start sub-tpn)) :lb start-delay-time :ub start-delay-time :cost plan-cost :tpn master-tpn)
		  ;; add episode from tpn-end noop-end-event
		  (make-and-add-episode :start (event-to-master (tpn::tpn-end sub-tpn)) :end noop-end-event :tpn master-tpn)
		  ;; add episode from noop-event to noop-end-event
		  (setf (submission-master-tpn-noop-episode submission)
				(make-and-add-episode 
				 :start noop-event 
				 :end noop-end-event
				 :ub *infinity*
				 :cost noop-cost
				 :tpn master-tpn))
		  ;; add episode from noop-end-event to master-end
		  (make-and-add-episode :start noop-end-event :end master-end :ub *infinity* :tpn master-tpn)
		  )))))

(defun add-started-submission-to-tpn (submission master-tpn current-global-time)
  (with-slots (num-priorities cost-per-priority master-dispatch) *kirkrm-state*
	(let* ((tpn (submission-tpn-obj submission))
		   (noop-event (tpn::get-event (submission-noop-name submission) master-tpn))
		   (noop-end-event (tpn::get-event (submission-noopend-name submission) master-tpn))
		   (master-start (tpn::tpn-start master-tpn))
		   (master-end (tpn::tpn-end master-tpn))
		   )
	  (flet ((event-to-master (event) (tpn::get-event (get-master-name *kirkrm-state* event) master-tpn)))
		;; merge the tpn - add all episodes
		(loop for episode in (tpn::tpn-episodes tpn) do
			 (let ((start-time (get-execution-time *kirkrm-state* (tpn::episode-start episode)))
				   (end-time (get-execution-time *kirkrm-state* (tpn::episode-end episode)) )
				   ;;(end-event (get-ldg-symbol *kirkrm-state* (tpn::episode-end episode)) )
				   (new-lb (to-pinf (tpn::episode-lowerbound episode)))
				   (new-ub (to-pinf (tpn::episode-upperbound episode)))
				   )
			   (cond (end-time 
							   (setf new-lb 0)
							   (setf new-ub 0))
					 (start-time 
								 ;;(setf new-lb (drake2::computed-lowerbound end-event new-lb current-time master-dispatch))
								 ;;(setf new-ub (drake2::computed-upperbound end-event new-ub current-time master-dispatch))
								 (let ((deltat (max (- current-global-time start-time) 0)))
								   (setf new-lb (max (- new-lb deltat) 0))
								   (setf new-ub (max (- new-ub deltat) 0))
								   )
								 ))
			   ;; delete the primitive if it has already been dispatched.
			   (when (or end-time start-time)
				 ;; (format t "!!!!!!!!!!!!!!!!!!!!!! 1: ~a~%" (rest (tpn::episode-description episode)))
				 (delete :PRIMITIVE (rest (tpn::episode-description episode)) :test #'equalp :key (lambda (x) (first x))) 
				 ;; (format t "!!!!!!!!!!!!!!!!!!!!!! 2: ~a~%" (rest (tpn::episode-description episode)))
				 )
			   ;;(format t "~a ~a: [~a ~a]~%" (tpn::event-name (tpn::episode-start episode)) (tpn::event-name (tpn::episode-end episode)) new-lb new-ub)

			   (make-and-add-episode 
				:start (event-to-master (tpn::episode-start episode))
				:end (event-to-master (tpn::episode-end episode))
				:lb (to-binf (max new-lb 0))
				:ub (to-binf (max new-ub 0))
				:cost (/ (if (typep (episode-cost episode) 'constant-episode-cost)
							 (tpn::constant-episode-cost-cost (episode-cost episode))
							 (error "Unable to encode the cost ~a for episode ~a." (episode-cost episode) episode))
						 (* cost-per-priority num-priorities))
				:description (tpn::episode-description episode)
				:tpn master-tpn)) )
		;; add episode from master-start to noop decision
		;; the infinity ub is necessary for the new topological order based on submission
		(make-and-add-episode :start master-start :end noop-event :lb 0 :ub *infinity*  :tpn master-tpn)
		;; add episode from noop-event to tpn-start
		(make-and-add-episode :start noop-event :end (event-to-master (tpn::tpn-start tpn)) :ub 0 :cost 0 :tpn master-tpn)
		;; add episode from tpn-end noop-end-event
		(make-and-add-episode :start (event-to-master (tpn::tpn-end tpn)) :end noop-end-event :tpn master-tpn)
		;; do NOT add episode from noop-event to noop-end-event
		;;
		;; add episode from noop-end-event to master-end
		(make-and-add-episode :start noop-end-event :end master-end :ub *infinity* :tpn master-tpn)
		))))


(defun make-and-add-episode (&key (start nil) (end nil) (lb 0) (ub 0) (cost 0) (description '()) (tpn nil) (support :CONTROL-PROGRAM))
  ;; (format t "make-and-add-episode: ~a->~a, [~a ~a], cost:~a~%" start end lb ub cost)
  (let ((episode (tpn::make-episode start end lb ub cost description tpn :support support)))
	(tpn::add-episode start end episode)
	episode))

(defun tpn-to-dispatch (tpn old-dispatch)
  "Given a TPN, returns the ONLINE-DISPATCH object
   If the TPN is infeasible (i.e. if the dispatchable form is infeasible), the value nil will be returned."
  ;; Convert the tpn into Drake's ldg + activites format
  (flet ((event-to-symbol (event) (let ((symbol (get-ldg-symbol *kirkrm-state* event))) 
									(unless symbol (setf symbol (make-ldg-symbol (tpn::event-name event))))
									symbol)))

	(multiple-value-bind (graph activities)
		(drake2:tpn-to-ldg tpn :event-to-symbol-function #'event-to-symbol) ;; DAVID HERE!

	  ;; Create the online-dispatch-instance
	  (let ((start-event-symbol (get-ldg-symbol *kirkrm-state* "master-start"))
			(dispatchable-graph nil)
			(dispatch nil))

		;; Create the dispatchable graph, or set it to NIL if the LDG is infeasible
		(setf dispatchable-graph 
			  (handler-case (drake2:construct-dispatchable-graph graph)
				(mtk-graph-algorithms::infeasible-problem () nil)))

		;; Check if the dispatchable form is feasible
		(when dispatchable-graph
		  (setf dispatch
				(if old-dispatch
					(make-instance 'drake2:online-ldg-dispatch 
								   :start start-event-symbol ;; get correct start
								   :ldg dispatchable-graph
								   :time-tol (kirkrm-time-tolerance *kirkrm-state*)
								   :mailbox (drake2:online-dispatch-mailbox old-dispatch)
								   :activities (drake2:activities-list activities)    
								   :active-activities (drake2::online-dispatch-active-activities old-dispatch))
					(make-instance 'drake2:online-ldg-dispatch 
								   :start start-event-symbol ;; get correct start
								   :ldg dispatchable-graph
								   :time-tol (kirkrm-time-tolerance *kirkrm-state*)
								   :activities (drake2:activities-list activities) )
					)))

		;; If dispatch was created, return it. Otherwise, return nil
		dispatch
		)))
  )

(defun merge-hash-tables (&rest tables)
  (let ((union
         (make-hash-table
          :test (first
                 (sort (mapcar #'hash-table-test tables) #'>
                       :key (lambda (test)
                              (ecase test
                                (eq 0)
                                (eql 1)
                                (equal 2)
                                (equalp 3)))))
          :size (reduce #'max (mapcar #'hash-table-size tables)))))
    (dolist (table tables)
      (maphash (lambda (key val) (setf (gethash key union) val)) table))
    union))

(defun copy-tpn (tpn)
  (let ((tpn-lxml (with-output-to-string (tpn-stream)
					(tpn::write-tpn tpn-stream tpn :force-labeling? t))))
	(first (tpn::parse-tpn-string tpn-lxml 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))
))

(defun copy-env (env)
  (let ((env-lxml (write-environment-string env)))
	(read-environment-string env-lxml)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; DISPATCHER METHODS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun print-dispatcher-trace (current-time event-list)
  (declare (ignore current-time event-list))
  (format t "dispatcher trace: ~a~%" (drake2::dispatch-trace (kirkrm-master-dispatch *kirkrm-state*)))
  )

(defun kill-dispatcher ()
  (with-slots (master-dispatch) *kirkrm-state*
	(when (and master-dispatch (drake2:dispatcher-running? master-dispatch))
	  (drake2::stop-dispatcher master-dispatch :wait t))
	))

(defun start-dispatcher ()
  (with-slots (master-dispatch ignore-missed-activities) *kirkrm-state*
	(when master-dispatch
	  (drake2::clear-stop-dispatcher master-dispatch)
	  (drake2:start-dispatcher master-dispatch
							   :event-callback #'print-dispatcher-trace ;;nil
							   :status-callback #'resource-manager-status-callback
							   :activity-callback #'resource-manager-callback
							   :ignore-missed-activities ignore-missed-activities
							   :terminate-after-last-event t))
	))

(defun dispatcher-running? ()
  (with-slots (master-dispatch) *kirkrm-state*
	(if master-dispatch
		(drake2:dispatcher-running? master-dispatch)
		nil)
	))

(defun resource-manager-callback (id primitive lb ub)
  "Publishes a message with four values delimited by pipes:
   ID - an integer value that uniquely identifies this primitive. 
       To be returned in a call to notify-activity-completed to 
       indicate that the associated primitve has completed executing.
   PRIMITIVE - a string to be executed
   LB - The numeric lowerbound time (in seconds) for which the primitive
        should be executed
   UB - The numeric upperbound time (in seconds), executing a primitive
        longer than this UB will definitley fail. However, executing a
        primitive for anything other than its lowerbound also has a
        probability of failing."
  (let ((message (format nil "primitive|~a|~a|~a|~a|~%" id primitive lb ub)))
	(test-notify-xml *resource-manager* message))

  ;; COMMENT OUT
  ;; This section will notify the resource-manager of completed activities at the lb time.
  ;; (flet ((done () 
  ;; 		   (sleep lb)
  ;; 		   ;; Format calls NOT made from the main thread may not print.
  ;; 		   ;; This behavior depends on the lisp runtime, and it's slime interface.
  ;; 		   (mtk-utils:format-time t "Sending completed message for primitive: ~s~%" primitive)
  ;; 		   (notify-activity-completed id)))
  ;; 	(bordeaux-threads:make-thread #'done))

  )

(defun resource-manager-status-callback (name &rest data)
  (let ((message (format nil "status|~a|" name)))
	(dolist (datum data)
	  (setf message (concatenate 'string message (format nil "~a" datum) "|")))
	(test-notify-xml *resource-manager* message)) )

(defun notify-activity-completed (id-or-primitive)
  "This is a callback, used by the resource-manager to tell the dispatcher
   that an activity has completed.
  ID-OR-PRIMITIVE can either be an integer ID associated with a primitive that
   completed execution, or the primitive string itself.
  Note: It is more reliable to use the ID to identify a completed primitive, 
        because a single primitive may be used more than once."
  (with-slots (master-dispatch) *kirkrm-state*
	(let ((id id-or-primitive))
	  (block get-id 
		;; see if the input is the id
		(when (numberp id-or-primitive)
		  (setf id id-or-primitive)
		  (return-from get-id))
		;; when the input is a string, it could be an id or a primitive
		(when (stringp id-or-primitive)
		  ;; see if the input is a string id
		  (setf id (handler-case (org.mapcar.parse-number::parse-number id-or-primitive)
					 (org.mapcar.parse-number::invalid-number nil)))
		  (when id (return-from get-id))
		  ;; see if the input is a string primitive
		  (let ((matching-activity 
				 (find id-or-primitive (drake2::online-dispatch-active-activities master-dispatch) 
					   :test #'equal :key (lambda (activity) (drake2:activity-primitive activity)))))
			(when matching-activity
			  (setf id (drake2:activity-id matching-activity)))
			(when id (return-from get-id))))
		;; if you didn't find anything, then error
		(error "Could not find an id associated with completed activity: ~s" id-or-primitive))
	  (mtk-utils:format-time t "NOTIFY-ACTIVITY-COMPLETED, id: ~s~%" id)
	  (drake2:notify-complete master-dispatch id))
	))

