;;;; Copyright (c) 2009-2011 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:kirk)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; EVENT-DATA
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass event-data ()
  ((ldg-symbol
	 :type symbol
	 :initform nil
	 :initarg :ldg-symbol
	 :accessor event-data-ldg-symbol
	 :documentation "The symbol used to represent this event in the MASTER LDG"
	 )
   (master-name
	 :type string
	 :initform nil
	 :initarg :master-name
	 :accessor event-data-master-name
	 :documentation "The name used to represent this event in the MASTER TPN"
	 )
   (tpn-event
	:type t
	:initform nil
	:initarg :tpn-event
	:accessor event-data-tpn-event
	:documentation "The event object used to represent this event in the SUBMISSION TPN"
	)
   (submission
	:type submission
	:initform nil
	:initarg :submission
	:accessor event-data-submission
	:documentation "The submission that created this event.")
   (tpn-start-p
	:type t
	:initform nil
	:initarg :tpn-start-p
	:accessor event-data-tpn-start-p
	:documentation "T if this event is the start of its submitted TPN."
	)
   (tpn-end-p
	:type t
	:initform nil
	:initarg :tpn-end-p
	:accessor event-data-tpn-end-p
	:documentation "T if this event is the end of its submitted TPN."
	)
   (execution-time
	:type number
	:initform -1
	:initarg :time
	:accessor event-data-execution-time
	:documentation "The execution time of this event.")
   )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; SUBMISSION CLASS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar state-none 0)
(defvar state-waiting 1)
(defvar state-started 2)
(defvar state-completed 3)
(defvar state-cancelled 4)

(defclass submission ()
  ((lock 
	:initform (bordeaux-threads:make-recursive-lock) 
	:initarg :lock
	:accessor  submission-lock
	:documentation "Lock used to make submission access atomic.")
   (id :type string 
	   :initform nil :initarg :id
	   :accessor submission-id
	   :documentation "ID of the submitted plan.")
   (timestamp :type :number 
			  :initform 0 :initarg :timestamp
			  :accessor submission-timestamp
			  :documentation "The time in seconds, relative to (get-universal-time), the plan was submitted.")
   (priority :type number
			 :initform 0 :initarg :priority
			 :accessor submission-priority
			 :documentation "The priority of this submission.")
   (start-time :type list
			   :initform `(:relative 0) :initarg :start-time
			   :accessor submission-start-time
			   :documentation "The start-time of this submission. This time can take two forms.
                       (:absolute 0) In the absolute form, the time-value is given in seconds 
                                     relative to the universal time, the number of seconds that 
                                     have elapsed since 00:00 of January 1, 1900 in the GMT time zone.
			           (:relative 0) In the relative form, the time-value is given in seconds
                                     relative to whenever the plan makes it into the dispatcher.")
   (tpn-obj :type tpn::temporal-plan-network 
			:initform nil :initarg :tpn-obj
			:accessor submission-tpn-obj
			:documentation "Submitted TPN object.")
   (env-obj :type kirk::environment 
			:initform nil :initarg :env-obj
			:accessor submission-env-obj
			:documentation "Submitted ENV object.")
   (noop-name :type string
			  :initform nil :initarg :noop-name
			  :accessor submission-noop-name
			  :documentation "The name of the noop event used in the master tpn.")
   (noopend-name :type string
				 :initform nil :initarg :noopend-name
				 :accessor submission-noopend-name
				 :documentation "The name of the noopend event used in the master tpn.")
   (master-tpn-noop-episode 
	:type t
	:initform nil
	:accessor submission-master-tpn-noop-episode
	:documentation "The noop episode used in the master-tpn.
     This value is not meant for general use.
     It is saved here for convenience, and is used to check 
     whether the noop episode or the submission was chosen by 
     Kirk to be included in the master tpn.")
   (state
	:type integer
	:initform state-none
	:initarg :state
	:accessor submission-state
	:documentation "Whether the submission is waiting to be dispatched, started, or completed. 
                    This state must correspond to the list the submission is in.")
   ))

(defun make-submission (&key (id nil) (tpn nil) (env nil) (priority 5) (start-time `(:relative 0)))
  (let ((timestamp_ (get-universal-time))
		(id_ nil)
		(tpn-obj_ nil)
		(env-obj_ nil)
		(noop-name_ nil)
		(noopend-name_ nil) )
	;; save off the id
	(cond ((stringp id) (setf id_ id))
		  (t (error "ID must be of type string and cannot be nil.")))
	;; save off the tpn
	(cond ((typep tpn 'tpn::temporal-plan-network)
		   (setf tpn-obj_ tpn))
		  ((stringp tpn)
		   ;; convert the tpn-string into a tpn object
		   (setf tpn-obj_ (first (tpn::parse-tpn-string tpn 'tpn::get-event 'tpn::make-episode 'tpn::add-episode 'tpn::make-tpn))))
		  (t (error "TPN must be of type TPN or a tpn-xml formatted string, and cannot be nil.")))
	;; save off the environment
	(cond ((typep env 'environment)
		   (setf env-obj_ env))
		  ((stringp env)
		   ;; convert the env-string into a env object
		   (setf env-obj_ (with-input-from-string (env-in-stream env)
							(read-environment-stream env-in-stream))))
		  (t (error "ENV must be of type ENVIRONMENT or an environment-xml formatted string, and cannot be nil.")))
	(setf noop-name_ (format nil "~a:noop" id_))
	(setf noopend-name_ (format nil "~a:noopend" id_))
	;; make the submission
	(make-instance 'submission :id id_ :timestamp timestamp_ 
				   :priority priority :start-time start-time
				   :tpn-obj tpn-obj_ :env-obj env-obj_
				   :noop-name noop-name_ :noopend-name noopend-name_)
	))

(defun make-master-name (submission event)
  (concatenate 'string (submission-id submission) ":" (format nil "~a" (tpn::event-name event))))
