;;; -*- Mode:Common-Lisp; Package:LP-MODEL; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :lp-model)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; Constructor for constraints.                                             ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((constraint linear-constraint) 
									   &rest args
									   &key (lb nil lb-p)
									   (ub nil ub-p)
									   (vars nil vars-p) 
									   (vals nil vals-p))
  "This is the initializer for a constraint if it is given VARS and VALS."

  (with-slots (num-elements element-indices element-values row-lb row-ub sense) constraint
	;; First, make sure the indices and values lists match in size.
	(unless (eql (length element-indices) (length element-values))
	  (error "ELEMENT-INDICES and ELEMENT-VALUES do not match in size for
              LINEAR-CONSTRAINT:~a~%" constraint))
	(setf num-elements (length element-indices))
	;; Determine the sense.
	(if (and lb-p ub-p)
		(progn
		  (if (eql lb ub)
			  ;; This is an equality constraint.
			  (setf sense *sense-=*)
			  ;; This is a range constraint.
			  (setf sense *sense-range*))
		  ;; Coerce into doubles.
		  (setf row-lb (coerce lb 'double-float))
		  (setf row-ub (coerce ub 'double-float)))
		(if ub-p
			;; This is a <= constraint.
			(progn
			  (setf sense *sense-<=*)
			  (setf row-ub (coerce ub 'double-float))
			  (setf row-lb nil))
			(if lb-p
				;; This is a >= constraint.
				(progn
				  (setf sense *sense->=*)
				  (setf row-lb (coerce lb 'double-float))
				  (setf row-ub nil))
				;; No bounds given!! Default to an equality constraint of zero.
				(progn
				  (setf sense *sense-=*)
				  (setf row-lb 0.0d0)
				  (setf row-ub 0.0d0))))))

;;   (let* ((lb-temp (lower-bound constraint))
;; 		 (lb-coerced (coerce lb-temp 'double-float)))
;; ;	(setf (lower-bound constraint) lb-temp)
;; )

  
  (if (and vars-p vals-p)
	  (progn
		;; The situation is we've been given a list of variables (or their 
		;; indices) and their values and we need to go enter them into the
		;; appropriate places.
		
		(loop for var in vars
		   for val in vals
		   do (setf (linear-part constraint var) val)))
	  (if (or vars-p vals-p)
		  (warn "INITIALIZE-INSTANCE given one of VARS and VALS. You probably meant
                 to give both.~%In LINEAR-CONSTRAINT ~a~%" constraint)))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; Functions on LP constraints.                                             ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro make-constraint-and-add-to-lp (lp &key 
										 (type :linear)
										 (name nil name-p)
										 (vars nil vars-p)
										 (vals nil vals-p)
										 (lb nil lb-p)
										 (ub nil ub-p))

  `(let ((c ,(case type
				   (:linear `(make-instance 'linear-constraint
											,@(if name-p (list :name name))
											,@(if vars-p (list :vars vars))
											,@(if vals-p (list :vals vals))
											,@(if lb-p (list :lb lb))
											,@(if ub-p (list :ub ub))))
				   (otherwise (error "Unrecognized constraint type: ~a~%" type)))))
	 (add-constraint-to-lp c ,lp)))

(defmethod add-constraint-to-lp ((constraint lp-constraint) (lp lp))
  "Adds the given constraint to the LP and returns its row index."
  (with-slots (constraints constraints-ht) lp
	(let ((idx (length constraints)))
	  (with-slots (row-index-in-formulation name) constraint
		;; Update CONSTRAINT with its index
		(setf row-index-in-formulation idx)
		;; Add CONSTRAINT to CONSTRAINTS list in LP
		(push constraint constraints)
		;; Add CONSTRAINT to CONSTRAINTS-HT hash-table in LP
		(setf (gethash name constraints-ht) constraint)
		;; FIRE any listeners
		(fire lp :constraint-added constraint))
	  ;; Return the index
	  (values idx constraint))))

(defmethod linear-part ((constraint linear-constraint) idx)
  "Returns the linear part of CONSTRAINT for IDX."
  
  (with-slots (element-indices element-values) constraint
	;; First check if the index is in the indices list.
	(let ((pos (position idx element-indices)))
	  (if pos
		  ;; If IDX is in the list, return its corresponding value.
		  (elt element-values pos)
		  ;; Otherwise, return 0.
		  0))))

(defmethod (setf linear-part) (val (constraint linear-constraint) idx)
  "Updates the coefficients of a linear constraint function."
  (with-slots (element-indices element-values) constraint

	(with-slots (num-elements element-indices element-values) constraint
	
	  ;; First, check if this index already has a coefficient.
	  ;; If so, update it and we're done.
	  (let ((pos (position idx element-indices)))
		(if pos 
			(setf (elt element-values pos) val)
		  
			;; If this IDX doesn't already have a value, just stick it on.
			(progn
			  (vector-push-extend idx element-indices)
			  (vector-push-extend (coerce val 'double-float) element-values)
			  ;; Increment the number of elements.
			  (incf num-elements)))))))

(defmethod activate-constraint ((constraint lp-constraint))
  "Activate CONSTRAINT."
  (with-slots (active?) constraint
	(unless nil;active?
	  (setf active? T)
	  (fire constraint :constraint-activated))))

(defmethod deactivate-constraint ((constraint lp-constraint))
  "Deactivate CONSTRAINT."
  (with-slots (active?) constraint
	(when T;active?
	  ;(break "Deactivating constraint ~a." constraint)
	  (setf active? NIL)
	  (fire constraint :constraint-deactivated))))