;;; -*- Mode:Common-Lisp; Package:LP-MODEL; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :lp-model)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; Constant Definitions                                                     ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconstant *infty* 'infty "Represents positive infinity.")
(defconstant *-infty* 'neg-infty "Represents negative infinity.")

(defconstant *sense-=* :sense-= "Represents an equality constraint.")
(defconstant *sense->=* :sense->= "Represents a greater than or equal constraint.")
(defconstant *sense-<=* :sense-<= "Represents a less than or equal constratint.")
(defconstant *sense-range* :sense-range "represents a range constraint.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; Class Definitions                                                        ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass packed-vector ()
  ((num-elements
	:type fixnum
	:initarg :num-el
	:initform 0
	:accessor num-elements
	:documentation "The number of non-zero left-hand side elements (columns).")
   
   (element-indices
	:type (vector fixnum)
	:initarg :indices
	:initform (make-array 5 :fill-pointer 0 :adjustable t :element-type 'fixnum)
	:accessor element-indices
	:documentation "Vector of column indices for left-hand side elements.")
   
   (element-values
	:type (vector double-float)
	:initarg :values
	:initform (make-array 5 :fill-pointer 0 :adjustable t :element-type 'double-float)
	:accessor element-values
	:documentation "Vector of values for left-hand side elements."))
  (:documentation "Represents a vector. Zero elements are omitted for memory conservation."))

(defclass lp (event-publisher)

  ((name
	:initarg :name
	:initform (gentemp "LP-")
	:reader name
	:documentation "The name given to this LP.")
   
   (variables
	:initarg :variables
	:initform nil
	:reader variables
	:documentation "A list of LP-VARIABLEs in this LP. This list is not
                    guaranteed to be in any specific order.")

   (variables-ht
	:initform (make-hash-table :test #'equal)
	:reader variables-ht
	:documentation "A hash table of the variables hashed by name.")
   
   (objective
	:initarg :objective
	:initform nil
	:reader objective
	:documentation "The objective of this LP.")
   
   (constraints
	:initarg :constraints
	:initform nil
	:reader constraints
	:documentation "List of LP-CONSTRAINTs used in the LP. This list is not
                    guaranteed to be in any specific order.")
   
   (constraints-ht
	:initform (make-hash-table :test #'equal)
	:reader constraints-ht
	:documentation "A hash table of the constraints hashed by name."))

  (:documentation "An instance of a linear program."))

(defclass lp-variable (event-publisher)

  ((name
	:initarg :name
	:initform (gentemp "LP-VARIABLE-")
	:reader name
	:documentation "The name of this variable.")
   
   (lower-bound
	:initarg :lower-bound
	:type double-float
	:initform *-infty*
	:accessor lower-bound
	:documentation "The lower bound of this variable. Defaults to *-infty*.")
   
   (upper-bound
	:initarg :upper-bound
	:type double-float
	:initform *infty*
	:accessor upper-bound
	:documentation "The upper bound of this variable. Defaults to *infty*.")

   (col-index-in-formulation
	:type fixnum
	:initarg :col-index
	:initform -1
	:reader col-index-in-formulation
	:documentation "The col index in the LP tableau."))

  (:documentation "Represents a single variable in an LP."))

(defclass lp-constraint (event-publisher)

  ((name
	:initarg :name
	:initform (gentemp "LP-CONSTRAINT-")
	:reader name
	:documentation "The name of this constraint.")

   (active?
	:initarg :active?
	:initform T
	:reader constraint-active?
	:documentation "T iff this constraint is active.")))

(defclass linear-constraint (lp-constraint packed-vector)
  
  ((name 
	:initarg :name
	:initform (gentemp "LINEAR-CONSTRAINT-")
	:accessor name)
   
   (row-lb
    :type double-float
	:initarg :lb
	:initform nil
	:accessor lower-bound
   	:documentation "Lower bound for constraint row.")
   
   (row-ub
	:type double-float
	:initarg :ub
	:initform nil
	:accessor upper-bound
   	:documentation "Upper bound for constraint row.")
   
   (row-index-in-formulation
	:type fixnum
	:initform 0
	:accessor row-index-in-formulation
	:documentation "The row index in the LP tableau.")

   (sense
	:accessor sense
	:documentation "The sense of the constraint. Is autopopulated with *SENSE-** on creation.")
   
   )
  
  )

(defclass lp-objective (event-publisher)

  ((name
	:initarg :name
	:initform (gentemp "LP-OBJECTIVE-")
	:reader name
	:documentation "The name of this objective."))
  
  (:documentation "The super class for LP-OBJECTIVEs"))

(defclass linear-objective (lp-objective packed-vector)

  ((name
	:initform (gentemp "LINEAR-OBJECTIVE-")
	:reader name))

  (:documentation "A linear objective function."))

(defclass diagonal-quadratic-objective (linear-objective)

  ((name
	:initform (gentemp "DIAGONAL-QUADRATIC-OBJECTIVE-")
	:reader name)
   
   (diag
	:type (vector double-float)
	:initarg :diag
	:initform (make-array 1 :fill-pointer 0 :adjustable t :element-type 'double-float)
	:accessor quadratic-objective-diag
	:documentation "Specifies the elements on the diagonal of the matrix."))

  (:documentation "A quadratic objective function with all non-zero elements on the diagonal."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; Generic Function Definitions                                             ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric add-variable-to-lp (var lp &key lin-obj-val)
  (:documentation
   "Add the variable to the specified LP. Returns the index of the added
    variable."))

(defgeneric linear-part (item var)
  (:documentation "Get the linear part of ITEM for VAR.")
  (:argument-precedence-order var item))

(defgeneric (setf linear-part) (val item var)
  (:documentation "Set the linear part of some object for VAR to VAL.")
  (:argument-precedence-order var item val))

(defmethod sort-problem ((lp lp))
  "Sorts the given LP so that VARIABLES and CONSTRAINTS are guaranteed to be in increasing order."
  (with-slots (variables constraints) lp
	(setf variables (sort variables #'< :key #'col-index-in-formulation))
	(setf constraints (sort constraints #'< :key #'row-index-in-formulation))))