;;; -*- Mode:Common-Lisp; Package:LP-MODEL; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :lp-model)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; Functions on LP objectives.                                              ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmacro make-objective-and-add-to-lp (lp &optional (type :linear) &rest args)
  "Makes an objective function and adds it to the LP. If args is given, then the type
   of objective function must be specified.

   Possible options for the objective function are:
     - :linear
     - :diag

   Returns the resulting objective function."
  `(let ((obj ,(case type 
					 (:linear `(make-instance 'linear-objective ,@args))
					 (:diag `(make-instance 'diagonal-quadratic-objective ,@args))
					 (otherwise (error "Unknown objective type!")))))
	 (add-objective-to-lp! obj ,lp)
	 obj))

(defmethod add-objective-to-lp! ((obj lp-objective) (lp lp))

  (with-slots (objective) lp
	(setf objective obj)))

(defmethod set-lp-linear-objective-idx! ((lp lp) idx val)
  "Changes the linear part of the LP's objective function so that the given
   variable (IDX) has coefficient VAL.

   If LP does not have an objective function, a linear objective function is
   created."

  ;; Check to see if an objective function exists, create one if necessary.
  (with-slots (objective) lp
	(unless (typep objective 'lp-objective)
	  (make-objective-and-add-to-lp lp :linear))
	
	;; We have a valid objective function. Now update it!
	(setf (linear-part objective idx) val)))

(defmethod linear-part ((objective linear-objective) idx)
  "Returns the linear part of the objective function for the corresponding variable."
  
  (with-slots (element-indices element-values) objective
	;; First check if the index is in the indices list.
	(let ((pos (position idx element-indices)))
	  (if pos
		  ;; If IDX is in the list, return its corresponding value.
		  (elt element-values pos)
		  ;; Otherwise, return 0.
		  0.0d0))))

(defmethod (setf linear-part) (val (objective linear-objective) idx)
  "Updates the coefficients of a linear objective function."
  (with-slots (element-indices element-values) objective
	
	
	(with-slots (num-elements element-indices element-values) objective
	  
	  ;; First, check if this index already has a coefficient.
	  ;; If so, update it and we're done.
	  (let ((pos (position idx element-indices)))
		(if pos 
			(setf (elt element-values pos) val)
			
			;; If this IDX doesn't already have a value, just stick it on.
			(progn
			  (vector-push-extend idx element-indices)
			  (vector-push-extend (coerce val 'double-float) element-values)
			  ;; Increment the number of elements.
			  (incf num-elements)))))))

