;;; -*- Mode:Common-Lisp; Package:LP-SOLVER; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :lp-solver)

;;; Define the variables that all implementations should fill in.
(defvar *max-solvers* nil 
  "The maximum number of simultaneous solvers that can be handled.")

(defclass lp-solver ()
  ((name
	:initarg :name
	:initform (gentemp "lp-solver-")
	:reader lp-solver-name
	:documentation "The name given to this solver instance.")
   
   (problems
	:initform nil
	:reader lp-solver-problems
	:documentation "A list of problems held by the solver.")

   (max-problems
	:initform nil
	:reader lp-solver-max-problems
	:documentation "The maximum number of problems this solver can hold."))
  
  (:documentation "Represents an instance of an LP Solver."))

(defclass lp-problem ()
  ((name
	:initarg :name
	:initform (gentemp "lp-problem-")
	:reader lp-problem-name
	:documentation "The name given to this problem.")
   
   (solver
	:initform nil
	:reader lp-problem-solver
	:documentation "Pointer to the solver this problem belongs to.")

   (model
	:initarg :model
	:initform nil
	:accessor model
	:documentation "The model backing this problem."))

  (:documentation "An instance of a problem that belongs to a solver."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                                          ;;;
;;; Method definitions                                                       ;;;
;;;                                                                          ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric solve (lp-problem)
  (:documentation "Solve the given LP-PROBLEM. Returns 0 if there is an optimal
                   solution."))

(defgeneric get-new-problem (solver)
  (:documentation "Get an LP-PROBLEM instance for the given SOLVER."))

(defgeneric get-objective-value (lp-problem)
  (:documentation "Get the objective value for the LP-PROBLEM. Should be called 
                   after SOLVE."))

(defgeneric get-column-solution (lp-problem)
  (:documentation "Get the column solution for LP-PROBLEM as an array. Should be
                   called after SOLVE."))

(defgeneric save-state (lp-problem filename)
  (:documentation "Output LP-PROBLEM as FILENAME. The supported file types vary 
                   depending on the implementation used. FILENAME is parsed by 
                   lisp, so logical paths can be used."))

(defgeneric free (object)
  (:documentation "When finished with an LP-PROBLEM or LP-SOLVER, call this
                   method to recover its memory."))

(defgeneric get-row-conflicts (lp-problem)
  (:documentation "Return the row indices implicated in a conflict."))

(defgeneric get-col-conflicts (lp-problem)
  (:documentation "Return the indices of variables implicated in a conflict. A 
                   variable is implicated if its bounds are involved in the 
                   conflict."))

(defgeneric get-all-conflicts (lp-problem)
  (:documentation "Returns a values structure. The first value is the indices
                   of implicated rows. The second is the indices of implicated
                   variables."))

(defgeneric get-objective-limit-conflicts (lp-problem obj-limit)
  (:documentation "Returns the constraints responsible for causing the
                   objective function to be greater than OBJ-LIMIT."))