;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:mtk-utils-2d-geometry)

;;-------------------------------------------------------------------------------
;; 2d-line-segment Class
;;-------------------------------------------------------------------------------

(defclass 2d-line-segment (2d-line)
  ((point1 :type 2d-point
		   :accessor point1
		   :initarg :point1
		   :initform (make-2d-point 0 0)
		   :documentation "One of the two points that define this line.")
   (point2 :type 2d-point
		   :accessor point2
		   :initarg :point2
		   :initform (make-2d-point 0 0)
		   :documentation "One of the two points that define this line.")
))

(defmethod print-object ((seg 2d-line-segment) stream)
  (format stream "<~a, from ~a to ~a>" (get-2d-line seg) (point1 seg) (point2 seg)))

;;-------------------------------------------------------------------------------
;; 2d-line-segment Constructors
;;-------------------------------------------------------------------------------

(defmethod make1-2d-line-segment ((x1 number) (y1 number) (x2 number) (y2 number) 
							  &optional (instance nil))
  "Make a line segment from point (x1,y1) to point (x2,y2)."
  (unless instance (setf instance (allocate-instance (find-class'2d-line-segment))))
  (setf (point1 instance) (make-2d-point x1 y1))
  (setf (point2 instance) (make-2d-point x2 y2))
  (make2-2d-line x1 y1 x2 y2 instance)
  instance)

(defmethod make2-2d-line-segment ((pt1 2d-point) (pt2 2d-point)
							  &optional (instance nil))
  "Make a line segment from 2d-point object pt1 to 2d-point object pt2."
  (unless instance (setf instance (allocate-instance (find-class '2d-line-segment))))
  (setf (point1 instance) pt1)
  (setf (point2 instance) pt2)
  (make3-2d-line pt1 pt2 instance)
  instance)

(defmethod copy-2d-line-segment ((seg 2d-line-segment))
  "Returns a copy of the 2d-line object."
  (make-instance '2d-line-segment :a (a seg) :b (b seg) :c (c seg) 
				 :point1 (copy-2d-point (point1 seg)) 
				 :point2 (copy-2d-point (point2 seg)) ))

(defmethod get-2d-line ((line 2d-line-segment))
  (make-instance '2d-line :a (a line)  :b (b line) :c (c line))
)

;;------------------------------------------------------------
;; 2d-line-segment Equality
;;------------------------------------------------------------

(defmethod equal-geo ((seg1 2d-line-segment) (seg2 2d-line-segment) &key (tolerance *geometry-tolerance*))
  (or (and (equal-geo (point1 seg1) (point1 seg2) :tolerance tolerance)
		   (equal-geo (point2 seg1) (point2 seg2) :tolerance tolerance)) 
	  (and (equal-geo (point1 seg1) (point2 seg2) :tolerance tolerance)
		   (equal-geo (point2 seg1) (point1 seg2) :tolerance tolerance))))

(defmethod equalp-geo ((seg1 2d-line-segment) (seg2 2d-line-segment) &key (tolerance *geometry-tolerance*))
  (or (and (equalp-geo (point1 seg1) (point1 seg2) :tolerance tolerance)
		   (equalp-geo (point2 seg1) (point2 seg2) :tolerance tolerance)) 
	  (and (equalp-geo (point1 seg1) (point2 seg2) :tolerance tolerance)
		   (equalp-geo (point2 seg1) (point1 seg2) :tolerance tolerance))))

(defmethod equalp-geo ((seg 2d-line-segment) (pt 2d-point) &key (tolerance *geometry-tolerance*))
  (and (equalp-geo (point1 seg) pt :tolerance tolerance)
	   (equalp-geo (point2 seg) pt :tolerance tolerance)))

(defmethod equalp-geo ((pt 2d-point) (seg 2d-line-segment) &key (tolerance *geometry-tolerance*))
  (equalp-geo seg pt :tolerance tolerance))

;;-------------------------------------------------------------------------------
;; 2d-line-segment Method
;;-------------------------------------------------------------------------------

(defmethod get-intersection ((seg1 2d-line-segment) (seg2 2d-line-segment) 
							 &key (include-seg1pt1 t) (include-seg1pt2 t) 
							 (include-seg2pt1 t) (include-seg2pt2 t) )
  "Returns the intersection of two infinite-length lines.
The returns could be:
- nil, no intersection
- 2d-line object, the overlapping line object
- 2d-point object, representing the intersection point"
  (cond
	;; preserve accuracy, check for identical lines and end-point intersections first.
	((equalp-geo seg1 seg2 :tolerance 0) (copy-2d-line-segment seg1))
	;; check for intersection
	(:otherwise
	 (let* ((line1 (get-2d-line seg1))
			(line2 (get-2d-line seg2))
			(line-intersect (get-intersection line1 line2)))
	(cond (; no intersection
		   (null line-intersect)  nil)
		   ; point intersection: check if point lies on both lines 
		  ((typep line-intersect '2d-point)
		   (cond
			 ; preserve accuracy, check for end-point equivalence
			 ((equalp-geo (point1 seg1) (point1 seg2) :tolerance 0)
			  (if (and include-seg1pt1 include-seg2pt1) (copy-2d-point (point1 seg1)) nil))
			 ((equalp-geo (point1 seg1) (point2 seg2) :tolerance 0)
			  (if (and include-seg1pt1 include-seg2pt2) (copy-2d-point (point1 seg1)) nil))
			 ((equalp-geo (point2 seg1) (point1 seg2) :tolerance 0)
			  (if (and include-seg1pt2 include-seg2pt1) (copy-2d-point (point2 seg1)) nil))
			 ((equalp-geo (point2 seg1) (point2 seg2) :tolerance 0)
			  (if (and include-seg1pt2 include-seg2pt2) (copy-2d-point (point2 seg1)) nil))
			 ; now, check for point-on-line
			 ((and (point-on-line? line-intersect seg1	
								   :include-pt1 include-seg1pt1 :include-pt2 include-seg1pt2 :tolerance *inf*)
				   (point-on-line? line-intersect seg2
								   :include-pt1 include-seg2pt1 :include-pt2 include-seg2pt2 :tolerance *inf*))
			  line-intersect)
			 (:otherwise nil) ))
		  ; line intersection
			 ((typep line-intersect '2d-line)
										; line intersection: check if they overlap
			  (let ((s1p1 (point1 seg1))
					(s1p2 (point2 seg1))
					(s2p1 (point1 seg2))
					(s2p2 (point2 seg2)))
				(let* ((ends (sort `#(,s1p1 ,s1p2, s2p1, s2p2) #'2d-point-compare))
					   (test-line1 (make2-2d-line-segment (elt ends 0) (elt ends 1))))
				  (if (or (equalp-geo test-line1 seg1) (equalp-geo test-line1 seg2))
										;no intersection
					  (if (equalp-geo (elt ends 1) (elt ends 2))
										; the intersection is a point
						  (copy-2d-point (elt ends 1))
										; there is no intersection
						  nil)
										;there is intersection
					  (if (equalp-geo (elt ends 1) (elt ends 2))
										; the intersection is a point
						  (copy-2d-point (elt ends 1))
										; the intersection is a line
						  (make2-2d-line-segment (elt ends 1) (elt ends 2)))
					  )))))))))


(defmethod get-intersection ((seg1 2d-line-segment) (seg2 2d-line) &key)
  (get-intersection seg2 seg1))

(defmethod get-intersection ((seg1 2d-line) (seg2 2d-line-segment) &key)
  (error "Implementation Not Complete")
)

(defmethod get-length ((line 2d-line-segment))
  "Returns the length of the line"
  (get-distance (point1 line) (point2 line)))

(defmethod point-on-line? ((pt 2d-point) (segment 2d-line-segment) &key (include-pt1 t) (include-pt2 t) (tolerance *geometry-tolerance*))
  "Returns t if the point lies on the line segment."
  (and (point-on-line? pt (get-2d-line segment) :tolerance tolerance)
	   (with-slots (point1 point2) segment
		 (if (< (abs (get-slope segment)) 1)
			 ;; closer to horizontal
			 (between? (x point1) (x point2) (x pt) :include-start include-pt1 :include-end include-pt2)
			 ;; closer to vertical
			 (between? (y point1) (y point2) (y pt) :include-start include-pt1 :include-end include-pt2)
			 )
)))

(defmethod degenerate? ((segment 2d-line-segment) &key (tolerance *geometry-tolerance*))
  "Returns t if the line segment has two equivalent end points"
  (equal-geo (point1 segment) (point2 segment) :tolerance tolerance))

(defmethod get-midpoint ((seg 2d-line-segment))
  (with-slots (point1 point2) seg
	(get-2d-point-midpoint point1 point2))
)

;;-------------------------------------------------------------------------------
;; Helper Methods
;;-------------------------------------------------------------------------------

(defun between? (start-number end-number test-number 
				 &key (include-start t) (include-end t))
  (assert (numberp start-number))
  (assert (numberp end-number))
  (assert (numberp test-number))
  (and 
   ; test start condition
   (or include-start (not (equalp start-number test-number)))
   ; test end condition
   (or include-end (not (equalp end-number test-number)))
   ; inbetween
   (equalp test-number (second (sort `(,start-number ,test-number ,end-number) #'<)))
))

