;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:mtk-utils-2d-geometry)

;;-------------------------------------------------------------------------------
;; 2d-line Class
;;-------------------------------------------------------------------------------

(defclass 2d-line ()
  ((a :type number
	  :accessor a
	  :initarg :a
	  :initform 0
	  :documentation "A coefficient from the equation Ax+By+C=0")
   (b :type number
	  :accessor b
	  :initarg :b
	  :initform 0
	  :documentation "B coefficient from the equation Ax+By+C=0")
   (c :type number
	  :accessor c
	  :initarg :c
	  :initform 0
	  :documentation "C coefficient from the equation Ax+By+C=0")
))

(defmethod print-object ((line 2d-line) stream)
  (format stream "<~ax+~ay+~a=0>" (a line) (b line) (c line)))

;;-------------------------------------------------------------------------------
;; 2d-line Constructors
;;-------------------------------------------------------------------------------

(defmethod make1-2d-line ((a number) (b number) (c number) &optional (instance nil))
  "Make a line by specifying the coefficients of the general form equation Ax+By+C=0"
  (unless instance (setf instance (allocate-instance (find-class '2d-line))))
  (setf (a instance) a)
  (setf (b instance) b)
  (setf (c instance) c)
  instance)

(defmethod make2-2d-line ((x1 number) (y1 number) (x2 number) (y2 number) &optional (instance nil))
  "Make a line from point (x1,y1) to point (x2,y2)."
  (unless instance (setf instance (allocate-instance (find-class '2d-line))))
  (setf (a instance) (- y1 y2))
  (setf (b instance) (- x2 x1))
  (setf (c instance) (- (* x1 y2) (* x2 y1)))
  instance)

(defmethod make3-2d-line ((pt1 2d-point) (pt2 2d-point) 
					  &optional (instance nil))
  "Make a line from 2d-point object pt1 to 2d-point object pt2."
  (make2-2d-line (x pt1) (y pt1) (x pt2) (y pt2) instance)
  instance)

(defmethod make4-2d-line ((pt 2d-point) (slope number)
					  &optional (instance nil))
  "Make a line with slope through a point"
  (unless instance (setf instance (allocate-instance '2d-line)))
  (setf (a instance) slope)
  (setf (b instance) -1)
  (setf (c instance) (- (y pt) (* slope (x pt))))
  instance)

(defmethod copy-2d-line ((line 2d-line))
  "Returns a copy of the 2d-line object."
  (make-instance '2d-line :a (a line) :b (b line) :c (c line)))

;;------------------------------------------------------------
;; 2d-line Equality
;;------------------------------------------------------------

(defmethod equal-geo ((line1 2d-line) (line2 2d-line) &key (tolerance *geometry-tolerance*))
  (and (equal-tol (a line1) (a line2) :tolerance tolerance)
	   (equal-tol (b line1) (b line2) :tolerance tolerance)
	   (equal-tol (c line1) (c line2) :tolerance tolerance)))

(defmethod equalp-geo ((line1 2d-line) (line2 2d-line) &key (tolerance *geometry-tolerance*))
  (equal-geo line1 line2 :tolerance tolerance))

;;-------------------------------------------------------------------------------
;; 2d-line Methods
;;-------------------------------------------------------------------------------

(defmethod get-slope ((line 2d-line)) 
  "Returns the slope of the line. If the line is vertical, 
returns positive infinity"
  (if (equalp (b line) 0)
	  *inf*
	  (- (/ (a line) (b line))))
)

(defmethod get-intersection ((line1 2d-line)(line2 2d-line) &key)
  "Returns the intersection of two infinite-length lines.
The returns could be:
- nil, no intersection
- 2d-line object, the first line object (line1)
- 2d-point object, representing the intersection point"
  (cond 
	;; one or both of the lines is degenerate - no intersection
	((or (degenerate? line1) (degenerate? line2)) nil)
	;; preserve accuracy,  check if both lines are identical before calculating
	((equalp-geo line1 line2 :tolerance 0) line1)
	;; continue with intersection detection
	(:otherwise
	 (let ((a1 (a line1)) 
		   (b1 (b line1))
		   (c1 (c line1))
		   (a2 (a line2))
		   (b2 (b line2))
		   (c2 (c line2)))
	   (let ((b1c2-b2c1 (- (* b1 c2) (* b2 c1)))
			 (c1a2-c2a1 (- (* c1 a2) (* c2 a1)))
			 (a1b2-a2b1 (- (* a1 b2) (* a2 b1))))
		 (if (not (equalp a1b2-a2b1 0))
										; the lines intersect, return the point
			 (make-2d-point (/ b1c2-b2c1 a1b2-a2b1) (/ c1a2-c2a1 a1b2-a2b1))
										; the lines do NOT intersect
			 (if (and (equalp b1c2-b2c1 0) (equalp c1a2-c2a1 0))
										; the lines are collinear, return the first line
				 line1
										; the lines are parallel, return nil
				 nil))
		 )))))

(defmethod parallel? ((line1 2d-line)(line2 2d-line))
  "Returns t if the two lines are parallel"
  (equalp (* (a line1) (b line2)) 
		  (* (a line2) (b line1))))

(defmethod perpendicular? ((line1 2d-line)(line2 2d-line))
  "Returns t if the two lines are perpendicular"
  (equalp (- (* (a line1) (b line2)))
		  (* (a line2) (b line1))))

(defmethod collinear? ((line1 2d-line)(line2 2d-line))
  "Returns t if the two lines are collinear"
  (let ((a1 (a line1)) 
		(b1 (b line1))
		(c1 (c line1))
		(a2 (a line2))
		(b2 (b line2))
		(c2 (c line2)))
	(and (equalp (* a1 b2) (* a2 b1)) 
		 (equalp (* b1 c2) (* b2 c1))	
		 (equalp (* a1 c2) (* a2 c1)))	
))

(defmethod vertical? ((line 2d-line) &key (tolerance *geometry-tolerance*))
  "Returns t if the line is vertical"
  (equal-tol (b line) 0 :tolerance tolerance))

(defmethod horizontal? ((line 2d-line) &key (tolerance *geometry-tolerance*))
  "Returns t if the line is horizontal"
  (equal-tol (a line) 0 :tolerance tolerance))

(defmethod get-x-intercept ((line 2d-line) )
  "Returns the x-value of the x intercept if it exists, o/w nil."
  (if (horizontal? line) 
	  nil 
	  (/ (- (c line)) (a line))))

(defmethod get-y-intercept ((line 2d-line) )
  "Returns the y-value of the y intercept if it exists, o/w nil."
  (if (vertical? line) 
	  nil 
	  (/ (- (c line)) (b line))))

(defmethod point-on-line? ((pt 2d-point) (line 2d-line) &key (tolerance *geometry-tolerance*))
  "Returns t if the point lies on the line, o/w nil."
  ; Ax+By+C=0
  (equal-tol (+ (* (a line) (x pt)) 
			 (* (b line) (y pt)) 
			 (c line)) 
		  0
		  :tolerance tolerance))

(defmethod degenerate? ((line 2d-line) &key (tolerance *geometry-tolerance*))
  "Returns t if the line has A and B coefficients equal to 0"
  (and (equal-tol (a line) 0 :tolerance tolerance) 
	   (equal-tol (b line) 0 :tolerance tolerance)))

;;-------------------------------------------------------------------------------
;; Helper Methods
;;-------------------------------------------------------------------------------
