;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:mtk-utils-2d-geometry)

;;------------------------------------------------------------
;; 2d-point CLASS
;;------------------------------------------------------------

(defclass 2d-point ()
  ((x :type number
	  :accessor x
	  :initarg :x
	  :initform 0
	  :documentation "x-coordinate")
   (y :type number
	  :accessor y
	  :initarg :y
	  :initform 0
	  :documentation "y-coordinate")
))

(defmethod print-object ((pt 2d-point) stream)
  (format stream "(~a,~a)" (x pt) (y pt)))

;;------------------------------------------------------------
;; 2d-point Constructor
;;------------------------------------------------------------

(defun make-2d-point (x y)
  (make-instance '2d-point :x x :y y))

(defmethod copy-2d-point ((pt 2d-point))
  (make-instance '2d-point :x (x pt) :y (y pt)))

;;------------------------------------------------------------
;; 2d-point Equality
;;------------------------------------------------------------

(defmethod equal-geo ((pt1 2d-point) (pt2 2d-point) &key (tolerance *geometry-tolerance*))
  (and (equal-tol (x pt1) (x pt2) :tolerance tolerance)
	   (equalp (y pt1) (y pt2)) :tolerance tolerance))

(defmethod equalp-geo ((pt1 2d-point) (pt2 2d-point) &key (tolerance *geometry-tolerance*))
  (equal-geo pt1 pt2 :tolerance tolerance))

;;------------------------------------------------------------
;; 2d-point Methods
;;------------------------------------------------------------

(defmethod get-distance ((pt1 2d-point) (pt2 2d-point))
  (let ((xdiff (- (x pt1) (x pt2)))
		(ydiff (- (y pt1) (y pt2))))
	(sqrt (+ (* xdiff xdiff) (* ydiff ydiff))))
)

(defmethod get-2d-point-midpoint ((pt1 2d-point) (pt2 2d-point))
  (make-2d-point 
   (/ (+ (x pt1) (x pt2))  2)
   (/ (+ (y pt1) (y pt2))  2))
)

(defmethod get-2d-point-slope ((pt1 2d-point) (pt2 2d-point))
  "Returns the slope of the line. If the line is vertical, 
returns positive infinity"
  (let ((diffx (- (x pt1) (x pt2)))
		(diffy (- (y pt1) (y pt2))))
	(if (equalp diffx 0)
		*inf*
		(/ diffy diffx))
	))

;;-------------------------------------------------------------------------------
;; Helper Methods
;;-------------------------------------------------------------------------------

(defmethod get-dist-from-origin ((pt 2d-point))
  "Returns the straight-line distance from a point to the origin."
  (sqrt (+ (* (x pt) (x pt)) (* (y pt) (y pt)))))

(defmethod 2d-point-compare ((pt1 2d-point) (pt2 2d-point) &key (predicate #'<))
	(if (< (abs (get-2d-point-slope pt1 pt2)) 1)
		;; closer to horizontal
		(funcall predicate (x pt1) (x pt2))
		;; closer to vertical
		(funcall predicate (y pt1) (y pt2))	)
	)