;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:mtk-utils-2d-geometry)

;;-------------------------------------------------------------------------------
;; polygon Class
;;-------------------------------------------------------------------------------

(defclass polygon ()
  ((point-list :type list
			   :accessor point-list
			   :initarg :point-list
			   :initform nil
			   :documentation "A list of 2d-point objects"))
  )

(defmethod print-object ((poly polygon) stream)
  (format stream "<~a>" (point-list poly)))

;;-------------------------------------------------------------------------------
;; polygon Constructors
;;-------------------------------------------------------------------------------

(defmethod make-polygon ()
  (make-instance 'polygon))

(defmethod make1-polygon ((point-list list))
  (make-instance 'polygon :point-list point-list))

(defmethod add-point ((poly polygon) (pt 2d-point)) 
  (setf (point-list poly) (nconc (point-list poly) `(,pt))))

;;-------------------------------------------------------------------------------
;; polygon Methods
;;-------------------------------------------------------------------------------


(defmethod get-intersection ((poly polygon) (seg 2d-line-segment) 
							 &key (include-edge t) (include-interior t) )
  "Returns a non-redundant list of all intersections between a polygon and a line segment.
   This method does NOT include the interior of the polygon. The intersections could either
   be line segments tangent to the edges of the polygon or points conincident on the vertices
   or edges of the polygon."

   (let* ((intersections nil)
		  (pointset '())
		  (linesegset '()))
         
	;; loop over the line segments
	(loop for poly-seg in (get-segments poly) do
		 (let ((intersection (get-intersection poly-seg seg 
											   :include-seg1pt1 t :include-seg1pt2 t 
											   :include-seg2pt1 t :include-seg2pt2 t)))
		   (cond 
			 ((typep intersection '2d-point) (pushnew-geo intersection pointset))
			 ((typep intersection '2d-line-segment) (pushnew-geo intersection linesegset)))))

    ;; include edge
    (when include-edge 
	  (setf intersections (nconc intersections (copy-list linesegset)))
	  (setf intersections (nconc intersections (copy-list pointset))))

	;; include iterior
	(when include-interior 
	  (loop for lineseg in linesegset do 
		   (pushnew-geo (point1 lineseg) pointset)
		   (pushnew-geo (point2 lineseg) pointset)))
	(pushnew-geo (point1 seg) pointset) ; add end-points of the line-seg
	(pushnew-geo (point2 seg) pointset) ; add end-points of the line-seg

	(let ((pointarray (make-array (length pointset) :initial-contents pointset)))	
	  ; sort the points so they are in order
	  (setf pointarray (stable-sort pointarray #'2d-point-compare ))

	  (loop for i from 0 to (- (length pointarray) 2) do
		   (let* ((pt1 (aref pointarray i))
				  (pt2 (aref pointarray (+ i 1)))
				  (lineseg (make2-2d-line-segment pt1 pt2)))
			 (when (and (contains? poly (get-2d-point-midpoint pt1 pt2))
						(not (member lineseg linesegset :test #'equalp-geo)))
			   (setf intersections (nconc intersections `(,lineseg)))))))

	; sort the points & lines they are in order
	(setf intersections (stable-sort intersections #'2d-point-compare
									 :key (lambda (x)
											(cond 
											  ((typep x '2d-point) x)
											  ((typep x '2d-line-segment) (point1 x)) ) )
									 ))

	intersections) )

(defmethod get-segments ((poly polygon))
  (if (<= (length (point-list poly)) 1)
	  ; True: polygon has 1 point or less
	  nil
	  (let ((curr (point-list poly))
			(prev nil)
			(seg-list nil))
		(loop 
		   (setf prev curr)
		   (setf curr (cdr curr))
		   (if (null curr)
			   (progn
				 (push (make2-2d-line-segment (car prev) (first (point-list poly)))
					   seg-list)
				 (return))
			   (push (make2-2d-line-segment (car prev) (car curr))
					 seg-list)))
		seg-list)
))

(defmethod intersects? ((poly polygon) (line 2d-line) &key)
  (error "Implementation Not Complete")
)


(defmethod intersects? ((poly polygon) (seg 2d-line-segment) 
						&key (include-edge t) (include-interior t))
  (let ((intersections (get-intersection poly seg 
										 :include-edge include-edge 
										 :include-interior include-interior)))
	(if intersections t nil)))


;; (defmethod contains? ((poly polygon) (pt 2d-point)) 
;; "Returns true if the polygon contains the points. Points along the border of the polygon
;; may or may not return true."
;;   (let ((point-list-length (length (point-list poly)))) 
;; 	(let ((point-array (make-array point-list-length :initial-contents (point-list poly)))
;; 		  (j (- point-list-length 1))
;; 		  (oddNodes nil)
;; 		  (ptx (x pt))
;; 		  (pty (y pt)))
;; 	  (loop for i from 0 to (- point-list-length 1) do
;; 		   (let ((xi (x (aref point-array i)))
;; 				 (yi (y (aref point-array i)))
;; 				 (xj (x (aref point-array j)))
;; 				 (yj (y (aref point-array j))))
;; 			 (when (or (and (< yi pty) (>= yj pty))
;; 					   (and (< yj pty) (>= yi pty)))
;; 			   (when (< (+ xi (* (/ (- pty yi) (- yj yi)) (- xj xi)))
;; 						ptx)
;; 				   (setf oddNodes (not oddNodes)))
;; 			   )
;; 			 )
;; 		   (setf j i))
;; 	   oddNodes))
;;   )

;; Code from http://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/pnpoly.html
;; Original Author: W. Randolph Franklin
;;
;; int pnpoly(int nvert, float *vertx, float *verty, float testx, float testy)
;; {
;;   int i, j, c = 0;
;;   for (i = 0, j = nvert-1; i < nvert; j = i++) {
;;     if ( ((verty[i]>testy) != (verty[j]>testy)) &&
;; 	 (testx < (vertx[j]-vertx[i]) * (testy-verty[i]) / (verty[j]-verty[i]) + vertx[i]) )
;;        c = !c;
;;   }
;;   return c;
;; }
;;
(defmethod contains? ((poly polygon) (pt 2d-point) &key (exterior nil)) 
"Returns true if the polygon contains the points. Points along the border of the polygon
may or may not return true."
  (let ((point-list-length (length (point-list poly)))) 
	(let ((point-array (make-array point-list-length :initial-contents (point-list poly)))
		  (j (- point-list-length 1))
		  (c exterior)
		  (tx (x pt))
		  (ty (y pt)))
	  (loop for i from 0 to (- point-list-length 1) do
	  	   (let ((vxi (x (aref point-array i)))
	  			 (vyi (y (aref point-array i)))
	  			 (vxj (x (aref point-array j)))
	  			 (vyj (y (aref point-array j))))
	  		  (when (not (eql (> vyi ty) (> vyj ty)))
				(when (< tx (+ (/ (* (- vxj vxi) (- ty vyi)) (- vyj vyi)) vxi))
				  (setf c (not c)))) )
	  	   (setf j i))
	   c))
  )


(defmethod adjacent-points? ((poly polygon) (pt1 2d-point) (pt2 2d-point) &key (tolerance *geometry-tolerance*))
  (let ((answerp nil))
	(loop for seg in (get-segments poly) do
		 (when (equalp-geo seg (make2-2d-line-segment pt1 pt2) :tolerance tolerance)
		   (setf answerp t)
		   (return)))
	answerp)
  )
