;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package :mtk-utils-bounded-set)

(defclass bound ()
  ((value :type number
          :accessor get-value
          :initarg :value 
          :documentation "")
   (inequality :type function
               :accessor get-inequality
               :initarg :inequality
               :documentation "")))

(defclass interval ()
  ((lower-bound :type bound
                :accessor get-lower-bound
                :initarg :lower-bound
                :documentation "")
   (upper-bound :type bound
                :accessor get-upper-bound
                :initarg :upper-bound
                :documentation "")))

(defmethod initialize-instance :after ((interval interval) &key)
  (let* ((lower-bound            (get-lower-bound interval))
         (upper-bound            (get-upper-bound interval))
         (lower-bound-value      (get-value      lower-bound))
         (lower-bound-inequality (get-inequality lower-bound))
         (upper-bound-value      (get-value      upper-bound))
         (upper-bound-inequality (get-inequality upper-bound)))
    (assert (<= lower-bound-value upper-bound-value))
    (assert (or (eql lower-bound-inequality #'>)
                (eql lower-bound-inequality #'>=)))
    (assert (or (eql upper-bound-inequality #'<)
                (eql upper-bound-inequality #'<=)))))

;;-------------------------------------------------------------------------------
;; Print object
;;-------------------------------------------------------------------------------

(defmethod print-object ((interval interval) stream)
  (format stream "(:lower-bound ~A :upper-bound ~A)"
	  (get-lower-bound interval)
	  (get-upper-bound interval)))

(defmethod print-object ((bound bound) stream)
  (multiple-value-bind (lambda-expression is-closure-p function-name)
      (function-lambda-expression (get-inequality bound))
    (declare (ignore is-closure-p))
    (format stream "(~A :inequality #'~A)" 
	    (let ((value (get-value bound)))
	      (cond ((eql value *inf*) "*inf*")
                ((eql value *-inf*) "*-inf*")
                (t value)))
	    (if lambda-expression lambda-expression function-name))))

(defmethod interval-to-string ((interval interval))
  (let* ((lb (get-lower-bound interval))
         (ub (get-upper-bound interval))
         (lb-inequality-string (if (eql (get-inequality lb) #'>) "(" "["))
         (lb-value (if (= (get-value lb) *-inf*) "-inf" (get-value lb)))
         (ub-inequality-string (if (eql (get-inequality ub) #'<) ")" "]"))
         (ub-value (if (= (get-value ub) *inf*) "inf" (get-value ub))))
    (format nil "~A~A,~A~A"
	    lb-inequality-string
	    lb-value
	    ub-value
	    ub-inequality-string)))

;;-------------------------------------------------------------------------------
;; Making class instances
;;-------------------------------------------------------------------------------

(defmethod make-bound ((value number) &key inequality)
  "The following describes the format of the input parameters:
      <inequality> := #'> | #'>= | #'< | #'<=."
  (assert (or (eql inequality #'>) (eql inequality #'>=)
              (eql inequality #'<) (eql inequality #'<=)))
  (make-instance 'bound :value value :inequality inequality))

(defmethod make-lower-bound ((value number) &key (inequality #'>=))
  "The following describes the format of the input parameters:
      <inequality> := #'> | #'>=
    Assumptions:
      * If no :inequality key-value pair is specified, the inequality is 
        assumed to be NOT strict by default, i.e. #'>=."
  (setf inequality (if (= value *-inf*) #'> inequality))
  (assert (or (eql inequality #'>) (eql inequality #'>=)))
  (make-instance 'bound :value value :inequality inequality))

(defmethod make-inf-lower-bound ()
  (make-lower-bound *-inf*))

(defmethod make-upper-bound ((value number) &key (inequality #'<=))
  "The following describes the format of the input parameters:
      <inequality> := #'< | #'<=
    Assumptions:
      * If no :inequality key-value pair is specified, the inequality is 
        assumed to be NOT strict by default, i.e. #'<=."
  (setf inequality (if (= value *inf*) #'< inequality))
  (assert (or (eql inequality #'<) (eql inequality #'<=)))
  (make-instance 'bound :value value :inequality inequality))

(defmethod make-inf-upper-bound ()
  (make-upper-bound *inf*))

(defmethod make-interval (&key lower-bound upper-bound)
  "The following describes the format of the input parameters:
      <lower-bound>   := (<number> :inequality <lb-inequality>) | (<number>)
      <upper-bound>   := (<number> :inequality <ub-inequality>) | (<number>)
      <lb-inequality> := #'> | #'>=
      <ub-inequality> := #'< | #'<=

    Assumptions:
      * If no :inequality key-value pair is specified for the <lower-bound>,
        the <lb-inequality> is assumed to be NOT strict by default, i.e. #'>=.
        Likewise, if no :inequality key-value pair is specified for the
        <upper-bound>, the <lb-inequality> is assumed to be NOT strict by
        default, i.e.  #'<=."
  (make-instance 'interval
    :lower-bound (apply #'make-lower-bound lower-bound)
    :upper-bound (apply #'make-upper-bound upper-bound)))

(defmethod make-inf-interval ()
  (make-instance 'interval
    :lower-bound (make-inf-lower-bound) 
    :upper-bound (make-inf-upper-bound)))

;;-------------------------------------------------------------------------------
;; +, =, <, <=, >, >=, min, max
;;-------------------------------------------------------------------------------

(defmethod bound+ ((x bound) (y bound))
  (let ((x-value (get-value x))
	(x-inequality (get-inequality x))
	(y-value (get-value y))
	(y-inequality (get-inequality y)))
    (cond
     ((eql x-inequality #'>) 
      (cond ((or (eql y-inequality #'>=) (eql y-inequality #'>)) 
	     (make-instance 'bound :value (+ x-value y-value) :inequality #'>))
	    (t (make-inf-interval))))
     ((eql x-inequality #'<) 
      (cond ((or (eql y-inequality #'<=) (eql y-inequality #'<))
	     (make-instance 'bound :value (+ x-value y-value) :inequality #'<))
	    (t (make-inf-interval))))
     ((eql x-inequality #'>=) 
      (cond ((eql y-inequality #'>=)
	     (make-instance 'bound :value (+ x-value y-value) :inequality #'>=))
	    ((eql y-inequality #'>)
	     (make-instance 'bound :value (+ x-value y-value) :inequality #'>))
	    (t (make-inf-interval))))
     ((eql x-inequality #'<=) 
      (cond ((eql y-inequality #'<=) 
	     (make-instance 'bound :value (+ x-value y-value) :inequality #'<=))
	    ((eql y-inequality #'<)
	     (make-instance 'bound :value (+ x-value y-value) :inequality #'<))
	    (t (make-inf-interval))))
     (t (error "Error in #'bound+: Unknown case detected.")))))

(defmethod bound= ((x bound) (y bound))
  (eql x y))

(defmethod bound< ((x bound) (y bound))
  (let ((x-value (get-value x))
	(x-inequality (get-inequality x))
	(y-value (get-value y))
	(y-inequality (get-inequality y)))
    (cond
     ((eql x-inequality #'<) 
      (cond ((eql y-inequality #'<) (if (< x-value y-value) t nil))
	    ((eql y-inequality #'<=) (if (<= x-value y-value) t nil))
	    (t t)))
     ((eql x-inequality #'<=) 
      (cond ((or (eql y-inequality #'<=) (eql y-inequality #'<))
	     (if (< x-value y-value) t nil))
	    (t t)))
     ((eql x-inequality #'>=) 
      (cond ((eql y-inequality #'>) (if (<= x-value y-value) t nil))
	    ((eql y-inequality #'>=) (if (< x-value y-value) t nil))
	    (t t)))
     ((eql x-inequality #'>) 
      (cond ((or (eql y-inequality #'<=) (eql y-inequality #'<)) nil)
	    (t (if (< x-value y-value) t nil)))))))

(defmethod bound<= ((x bound) (y bound))
  (or (bound= x y) (bound< x y)))

(defmethod bound> ((x bound) (y bound))
  (let ((x-value (get-value x))
	(x-inequality (get-inequality x))
	(y-value (get-value y))
	(y-inequality (get-inequality y)))
    (cond
     ((eql x-inequality #'>) 
      (cond ((eql y-inequality #'>) (if (> x-value y-value) t nil))
	    ((eql y-inequality #'>=) (if (>= x-value y-value) t nil))
	    (t t)))
     ((eql x-inequality #'>=) 
      (cond ((or (eql y-inequality #'>=) (eql y-inequality #'>))
	     (if (> x-value y-value) t nil))
	    (t t)))
     ((eql x-inequality #'<=) 
      (cond ((eql y-inequality #'<) (if (>= x-value y-value) t nil))
	    ((eql y-inequality #'<=) (if (> x-value y-value) t nil))
	    (t t)))
     ((eql x-inequality #'<) 
      (cond ((or (eql y-inequality #'>=) (eql y-inequality #'>)) nil)
	    (t (if (> x-value y-value) t nil)))))))

(defmethod bound>= ((x bound) (y bound))
  (or (bound= x y) (bound> x y)))

(defmethod bound-min ((x bound) (y bound))
  (if (bound< x y) x y))

(defmethod bound-max ((x bound) (y bound))
  (if (bound> x y) x y))

(defmethod interval= ((x interval) (y interval))
  (eql x y))

(defmethod interval-intersection ((x interval) (y interval))
  (if (and (<= (get-value (get-lower-bound x)) (get-value (get-upper-bound y)))
	   (<= (get-value (get-lower-bound y)) (get-value (get-upper-bound x))))
      (make-instance 'interval 
	:lower-bound (bound-max (get-lower-bound x) (get-lower-bound y))
	:upper-bound (bound-min (get-upper-bound x) (get-upper-bound y)))
    nil))
