;;;; Copyright (c) 2011 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David Wang

(in-package #:mtk-utils-threading)

(defclass mailbox ()
  ((lock 
	:initform (bordeaux-threads:make-recursive-lock) 
	:initarg :lock
	:accessor mailbox-lock
	:documentation "Lock used to make mailbox access atomic.")
   (condition 
	:initform (bordeaux-threads:make-condition-variable)
	:initarg :condition
	:accessor mailbox-condition
	:documentation "The conditional variable used to indicate the mailbox has mail.")
   (timeout-flag
	:type flag
	:initform nil
	:initarg :timeout-flag
	:accessor mailbox-timeout-flag
	:documentation "A flag indicating a wait has timed-out. Used in the read method. 
This flag exists in here because bordeaux-threads says in some lisp implementations
only global variables can be shared between threads. This means timeouts may only
work reliably on some implementations if this instance of mailbox is global.")
   (queue 
	:initform '()
	:initarg :queue
	:accessor mailbox-queue
	:documentation "Queue used to store items in the mailbox.")
   ))

(defun make-mailbox ()
  (make-instance 'mailbox))

(defun mailbox-empty-p (mailbox)
  "Returns non-NIL if the mailbox can be read from, NIL otherwise."
  (bordeaux-threads:with-recursive-lock-held ((mailbox-lock mailbox))
	(null (mailbox-queue mailbox))))

(defun mailbox-empty! (mailbox)
  "Clears the mailbox of all contents"
  (bordeaux-threads:with-recursive-lock-held ((mailbox-lock mailbox))
	(setf (mailbox-queue mailbox) '())))

(defun mailbox-send (mailbox item)
  "Sends an item into the mailbox, notifying 1 waiter
   to wake up for retrieval of that object. 
   Copied and fixed from threads.lisp"
  (bordeaux-threads:with-recursive-lock-held ((mailbox-lock mailbox))
	(setf (mailbox-queue mailbox) 
		  (nconc (mailbox-queue mailbox) `(,item)))
	(bordeaux-threads:condition-notify (mailbox-condition mailbox))))

(defun mailbox-read (mailbox &key (wait t) (timeout -1))
  "Reads mail from the mailbox. 
   When WAIT is t, the method will wait TIMEOUT seconds or 
     until the mailbox has mail, and then return two values. 
     The first is the mail. The value is t. A TIMEOUT value of
     0 or less means the method will not timeout.
   When WAIT is nil, the method will return immediately with two values.
     If the mailbox is empty both values will be nil.
     Otherwise the first is the mail, the second is t."
  (bordeaux-threads:with-recursive-lock-held ((mailbox-lock mailbox))
	(when wait
	  (setf (mailbox-timeout-flag mailbox) (make-flag :value nil))
	  ;; create a thread to trigger the conditional wait to end the timeout.
	  (when (> timeout 0)
		(bordeaux-threads:make-thread 
		 (lambda () (sleep timeout) 
				 (bordeaux-threads:with-recursive-lock-held ((mailbox-lock mailbox))
				   (flag! (mailbox-timeout-flag mailbox) t)
				   (bordeaux-threads:condition-notify (mailbox-condition mailbox))))))
	  (loop (when (or (not (mailbox-empty-p mailbox))
					  (flag? (mailbox-timeout-flag mailbox) :wait nil))
			  (return))
		 (bordeaux-threads:condition-wait 
		  (mailbox-condition mailbox)
		  (mailbox-lock mailbox))))
   (if (mailbox-empty-p mailbox)
	   (values nil nil)
	   (values (pop (mailbox-queue mailbox)) t))
   ))

(defun mailbox-peek (mailbox)
  "Returns two values. The second returns non-NIL when the mailbox
is empty. The first value is the next item to be read from the mailbox. 
This value may be nil if the mailbox is empty, or if the first item is nil. 
To distinguish between these two cases, the second value returns is NIL 
if the mailbox is empty, and T if the mailbox was is not empty.

Note that due to multi-threading, the first value returned upon
peek, may be different from the one returned upon next read in the
calling thread."
  (bordeaux-threads:with-recursive-lock-held ((mailbox-lock mailbox))
     (values (car (mailbox-queue mailbox))
             (not (null (mailbox-queue mailbox))))))
