;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

;;; Copyright (c) 1999 - 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :OPSAT)


;;;(eval-when (compile load eval)
;;;  (export '(
;;;			get-constraint-assignment-of-search-problem 
;;;			insert-assignment-into-constraint-system
;;;			remove-assignment-from-constraint-system
;;;            )
;;;          :OPSAT)
;;;  )

;;; Optimization:
;;; o For check-current-state-against-constraints:
;;;   - If only one set of constraints exist (sat or unsat), 
;;;   then we should maintain them as inserted between successive 
;;;   states, rather then reinsert them for each call.

;;;   - Unit clauses should inserted and removed as assignments,
;;;   rather than as clauses.

;;; To Do:
;;; o Extend to conditional CSPs:
;;;   - add in comments of functions to change.
;;;   - modify functions as indicated by their to do.

;;; o Clean up *bfs-debug* printing to be consistent
;;;   with debug-print, using ISAT. 
;;;   - Add flags for turning on and off tracing in different 
;;;     packages, along with trace levels.
;;;   - Currently there are too many messages printed to be
;;;     useful.  They have been commented out until they
;;;     can be rationalized.
;;; o Change terminology of insert and remove constraint, to
;;;   activate and deactivate.
;;; o Change terminology of new-constraint to associate-constraint.
;;; o Extend current state so that assignments have a TRUTH-VALUE.
;;; o Can we avoid using two different data structures for 
;;;   assignments and the mapping between them?

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; External Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (get-constraint-assignment-of-search-problem
;;;    (search-problem search-problem) variable-name value)
;;;
;;;  Returns the assignment-object associated with the constraint-system
;;;  of SEARCH-PROBLEM for (= VARIABLE-NAME VALUE), where VARIABLE-NAME 
;;;  refers to a state variable.
;;;
;;; (insert-assignment-into-constraint-system 
;;;    (search-problem search-problem) source assignment truth-value)
;;;  
;;;   Asserts that ASSIGNMENT has TRUTH-VALUE in SEARCH-PROBLEM 
;;;   CONSTRAINT-SYSTEM, based on SOURCE.
;;;
;;;   SEARCH-PROBLEM is a search problem object.
;;;   ASSIGNMENT is a constraint-assignment object of the constraint-system
;;;      of SEARCH-PROBLEM.
;;;   TRUTH-VALUE is one of (*true* or *false*)
;;;   SOURCE is a lisp object denoting the creator of the insertion.
;;;
;;; (remove-assignment-from-constraint-system
;;;     search-problem source assignment truth-value)
;;;  
;;;    Removes the assertion that ASSIGNMENT has TRUTH-VALUE in 
;;;    SEARCH-PROBLEM constraint-system, based on SOURCE.
;;;
;;;    SEARCH-PROBLEM is a search problem object.
;;;    ASSIGNMENT is a constraint-assignment object of the
;;;      constraint-system of SEARCH-PROBLEM.
;;;    TRUTH-VALUE is one of (*true* or *false*)
;;;    SOURCE is a lisp object denoting the creator of the insertion.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Module Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (CLEAR-CURRENT-STATE-FROM-CONSTRAINTS!
;;;    &OPTIONAL (SEARCH-PROBLEM *DEFAULT-SEARCH-PROBLEM*))
;;;  
;;;    Clears the current search state of SEARCH-PROBLEM and 
;;;    removes the assertion of current state into 
;;;    SEARCH-PROBLEM's constraint system.

;;; (SWITCH-PREVIOUS-AND-CURRENT-STATE-IN-CONSTRAINTS
;;;    &OPTIONAL (SEARCH-PROBLEM *DEFAULT-SEARCH-PROBLEM*))
;;;
;;;   Given a search-problem whose constraint system contains
;;;   the assignments of the previous search state,
;;;   updates constraint system to the assignments in the current 
;;;   search state.
;;;
;;;   Note: current and previous assignments are state-assignment
;;;   objects of the search problem (not it's constraint-system).

;;; (NEW-CONSTRAINT-FOR-CONSTRAINT-SYSTEM
;;;     SEARCH-PROBLEM SOURCE CONSTRAINT-DESCRIPTION)
;;;  
;;;    Creates constraint described by CONSTRAINT-DESCRIPTION
;;;    for the constraint-system of SEARCH-PROBLEM.  Associates constraint
;;;    with constraint system, but does not ACTIVATE the constraint.
;;;  
;;;    SOURCE is an object that produced the constraint.
;;;    SEARCH-PROBLEM is a search problem object.
;;;
;;;    The form of CONSTRAINT-DESCRIPTION is defined by 
;;;    the constraint system and involves state assignments over the 
;;;    variables defined in constraint system.
;;;
;;;  (INSERT-CONSTRAINT-INTO-CONSTRAINT-SYSTEM
;;;    SEARCH-PROBLEM CONSTRAINT)
;;;  
;;;   Inserts CONSTRAINT object (of constraint system) into 
;;;   the constraint-system of SEARCH-PROBLEM.
;;;   SEARCH-PROBLEM is a search problem object.
;;;   CONSTRAINT is an object created by 
;;;   NEW-CONSTRAINT.
;;;
;;;  (CHECK-CURRENT-STATE-AGAINST-CONSTRAINTS
;;;    &OPTIONAL (SEARCH-PROBLEM *DEFAULT-SEARCH-PROBLEM*)
;;;	       (PERFORM-SEARCH? T))
;;;  
;;;   Given a SEARCH-PROBLEM whose constraint system has asserted
;;;   the previous search state. Update constraint system to the 
;;;   assignments of the current search state, then check this 
;;;   current state against the BASE, SAT and UNSAT constraints.
;;;   
;;;   Returns values: FEASIBLE? and CONFLICT.
;;;   FEASIBLE? is True iff base+sat constraints are satisfied
;;;   and if base+unsat constraints are violated.
;;;   state.  If FEASIBLE? is false, then CONFLICT is a subset of 
;;;   the current assignment that makes constraint system
;;;   infeasible (violates base+sat or entails base+unsat).  
;;;   A conflict is a list of STATE-ASSIGNMENTS or
;;;   :CURRENT, representing the current state assignment.
;;;   Each state assignment is a state-assignment object.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; External Interface Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod get-constraint-assignment-of-search-problem
    ((search-problem opsat-problem) variable-name value)

  "Returns the assignment-object associated with the constraint-system
   of SEARCH-PROBLEM for (= VARIABLE-NAME VALUE), where VARIABLE-NAME 
   refers to a state variable."
  
  (let ((constraint-system (opsat-problem-constraint-system search-problem)))

	(get-assignment constraint-system
					variable-name
					value)
;;;    (apply (opsat-problem-get-constraint-assignment-fn search-problem)
;;;		   (list constraint-system
;;;				 variable-name
;;;				 value))
	))

;;; Check:
;;; o Is this redundant with remove-assignment-from-constraint-system?
;;;   It looks like the same.
(defmethod insert-assignment-into-constraint-system 
    ((search-problem opsat-problem) source assignment truth-value)
  
  "Asserts that ASSIGNMENT has TRUTH-VALUE in SEARCH-PROBLEM 
   CONSTRAINT-SYSTEM, based on SOURCE.

   SEARCH-PROBLEM is a search problem object.
   ASSIGNMENT is a constraint-assignment object of the constraint-system
      of SEARCH-PROBLEM.
   TRUTH-VALUE is one of (*true* or *false*)
   SOURCE is a lisp object denoting the creator of the insertion."

  (assert! (opsat-problem-constraint-system search-problem)
		   assignment
		   truth-value
		   source)
  )

(defmethod remove-assignment-from-constraint-system
    ((search-problem opsat-problem) source assignment truth-value)
  
  "Removes the assertion that ASSIGNMENT has TRUTH-VALUE in 
    SEARCH-PROBLEM constraint-system, based on SOURCE.

    SEARCH-PROBLEM is a search problem object.
    ASSIGNMENT is a constraint-assignment object of the
      constraint-system of SEARCH-PROBLEM.
    TRUTH-VALUE is one of (*true* or *false*)
    SOURCE is a lisp object denoting the creator of the insertion."
  (unassert! (opsat-problem-constraint-system search-problem)
			 assignment
			 truth-value
			 source)
  )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Module Interface Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod clear-current-state-from-constraints!
    ((search-problem opsat-problem))
  
  " Clears the current search state of SEARCH-PROBLEM and 
    removes the assertion of current state into 
    SEARCH-PROBLEM's constraint system."
  
  (loop for variable in (state-space-variables search-problem)
      for current = (state-variable-current-assignment variable)
      unless (null current)
      do (setf (state-variable-current-assignment variable) nil)
		 (remove-state-assignment-from-constraint-system
		  search-problem
		  search-problem				;the source
		  current
		  *true*))
  T)

;;; To Do: 
;;; o Extend current state so that assignments have a TRUTH-VALUE.

;;; Check:
;;; o If previous and current are the same, should be
;;;   still make a switch?
(defmethod switch-previous-and-current-state-in-constraints
    ((search-problem opsat-problem))

  "Given a search-problem whose constraint system contains
   the assignments of the previous search state,
   updates constraint system to the assignments in the current 
   search state.

   Note: current and previous assignments are state-assignment
   objects of the search problem (not its constraint-system)."

;;;  (when *bfs-debug*
;;;    (format t "~%  Changing from previous to current search state:"))
;(format t "******************************~%")
  (loop for variable in (state-space-variables search-problem)
      for current = (state-variable-current-assignment variable)
      for previous = (state-variable-previous-assignment variable)
;;;   do

;;;	(when *bfs-debug*
;;;	   (format t "~A" current))
      unless (eql previous current)
      do (unless (null previous)
				 (remove-state-assignment-from-constraint-system
				  search-problem
				  search-problem
				  previous
				  *true*))
	   (unless (null current)
		 (insert-state-assignment-into-constraint-system
		  search-problem
		  search-problem ;the source
		  current
		  *true*))))

;;; FIX:
;;; o Change how constraint system is passed in to deal with different
;;;   keyword lists.  The current strategy is wired to ISAT.

;;; To Do:
;;; o Name this new-constraint and make sure that the arguments are
;;;   consistent with essential-constraint-system.
(defmethod new-constraint-for-constraint-system
    ((search-problem opsat-problem) source constraint-description)
  
  "Creates constraint described by CONSTRAINT-DESCRIPTION
    for the constraint-system of SEARCH-PROBLEM.  Associates constraint
    with constraint system, but does not ACTIVATE the constraint.
  
    Returns the internalized constraint for CONSTRAINT-DESCRIPTION
    that has been added to SEARCH-PROBLEM.

    SOURCE is an object that produced the constraint.
    SEARCH-PROBLEM is a search problem object.

    The form of CONSTRAINT-DESCRIPTION is defined by 
    the constraint system and restricts state assignments over the 
    variables defined in constraint system."
  
;;;  (new-constraint
;;;   (opsat-problem-constraint-system search-problem)
;;;   source :clause constraint-description)
  
;  (format t "~s~% " `(,(opsat-problem-constraint-system search-problem)
;		   :source ,source . ,constraint-description))


  (apply #'new-constraint
		 `(,(opsat-problem-constraint-system search-problem)
			:source ,source . ,constraint-description)))

(defmethod insert-constraint-into-constraint-system
    ((search-problem opsat-problem) constraint)
  
  "Inserts CONSTRAINT object (of constraint system) into 
   the constraint-system of SEARCH-PROBLEM.
   SEARCH-PROBLEM is a search problem object.
   CONSTRAINT is an object created by 
   NEW-CONSTRAINT."
  (insert-constraint 
   (opsat-problem-constraint-system search-problem)
   constraint))

;;; Optimization:
;;; o If only one set of constraints exist (sat or unsat), 
;;;   then we should maintain them as inserted between successive 
;;;   states, rather then reinsert them for each call.

;;; o Unit clauses should inserted and removed as assignments,
;;;   rather than as clauses.

;;; FIX:
;;; o Make sure in the above function and in general that the
;;;   unsat base theory works.  It would be better to make it 
;;;   so that the set of constraints that need to be entailed 
;;;   are explicitly passed.  (particularly since state variable 
;;;   definition constraints need not be entailed).

;;; Check:
;;; o Are assignments state-assignments of constraint-system-assignments?

(defmethod check-current-state-against-constraints
    ((search-problem opsat-problem)
	 &optional (perform-search? T))
  
  "Given a SEARCH-PROBLEM with the previous search state 
   inserted into the problem's constraint system, update 
   constraint system to the assignments of the current 
   search state, then check this current state against 
   the BASE, SAT and UNSAT constraints.

   Returns values: FEASIBLE? and CONFLICT.
   FEASIBLE? is True iff base+sat constraints are satisfied
   and if base+unsat constraints are violated.
   state.  If FEASIBLE? is false, then CONFLICT is a subset of 
   the current assignment that makes constraint system
   infeasible (violates base+sat or entails base+unsat).  
   A conflict is a list of STATE-ASSIGNMENTS or
   :CURRENT, representing the current state assignment.
   Each state assignment is a state-assignment object."

  ;; Update constraint system with assignments changed from previous
  ;; to current search state.
  (switch-previous-and-current-state-in-constraints 
   search-problem)

  ;; Check consistency of the constraint system.
;;;  (when *bfs-debug*
;;;    (format t "~%  Test current state for consistency:"))

  (let ((satisfied? nil)
		(conflict nil))

	;; First check the SAT constraints...
    (multiple-value-setq (satisfied? conflict)
      (achieves-problem-SAT-constraints-list?
       search-problem perform-search?))

	;; ... then check the UNSAT constraints.
    (when (and satisfied? 
			   (bfig::all-state-variables-assigned?  
				(bfig::search-problem-current-search-node search-problem)))
      (multiple-value-setq (satisfied? conflict)
		(achieves-problem-UNSAT-constraints-list?
		 search-problem perform-search?)))

    (values satisfied? conflict)))

;;; Added:
;;; o When, during search, a new candidate is to be tested against
;;; constraints, we now provide the option to look for all consistent models
;;; for a theory. Such a theory should be BASE + a SAT-CONSTRAINT-DESCRIPTIONS
;;; from SAT-CONSTRAINT-DESCRIPTIONS-LIST.

;;; Similar to check-current-state-against-constraints but
;;; rather than just checking consistency for the
;;; SAT-CONSTRAINT-DESCRIPTIONS-LIST, tries to find all
;;; consistent models for each set BASE+SET-CONSTRAINT-DESCRIPTIONS

(defmethod check-constraints-and-find-all-models-if-any
    ((search-problem opsat-problem)
     &optional (perform-search? T))
  

  "Given a SEARCH-PROBLEM with the previous search state 
   inserted into the problem's constraint system, update 
   constraint system to the assignments of the current 
   search state, then find all consistent models for the
   theory.


   Returns values: FEASIBLE? and CONFLICT and LIST-OF-SETS-OF-MODELS

   FEASIBLE? is True iff base+sat constraints are satisfied
   and if base+unsat constraints are violated.
   state.  If FEASIBLE? is false, then CONFLICT is a subset of 
   the current assignment that makes constraint system
   infeasible (violates base+sat or entails base+unsat).  
   A conflict is a list of STATE-ASSIGNMENTS or
   :CURRENT, representing the current state assignment.
   Each state assignment is a state-assignment object.

  For each combination theory+sat-constraint-descriptions
	  with sat-constraint-descritions
				in sat-constraint-descriptions-list
  we collect the consistent models within LIST-OF-SET-OF-CONSISTENT-MODELS"
  
  ;; Update constraint system with assignments changed from previous
  ;; to current search state.
  (switch-previous-and-current-state-in-constraints 
   search-problem)
  
  (let ((satisfied?) (conflict) (list-of-sets-of-models))

	;; First check the UNSAT constraints. 
    (multiple-value-setq (satisfied? conflict) 
	  ;;(satisfied? = T ) -> unsat-const are inconsistent!
      (achieves-problem-UNSAT-constraints-list?
       search-problem perform-search?))

    (when satisfied?
	  ;; ...then enumerate all consistent models for each BASE+SAT-CONSTRAINTS set
      (multiple-value-setq (satisfied? conflict list-of-sets-of-models)
		(return-all-sets-of-models-for-SAT-constraints-list-if-any
		 search-problem)))
    
    (values satisfied? conflict list-of-sets-of-models)))

(defmethod return-all-sets-of-models-for-SAT-constraints-list-if-any
	((search-problem opsat-problem))

  "If all sets BASE+SAT-CONSTRAINT-DESCRIPTIONS is feasible, returns
   all consitent models for each set, else return the conflict.

   RETURNS: (values consistent conflict list-of-sets-of-models)
   Consistent is T iff search-problem has at least one consistent solution.
   If not consistent, returns a non-nil conflict, where a conflict is
    a list of inconsistent assignments, or :CURRENT.
   Otherwise returns a list of models, where each model is a list of assignments.
	"
  (let ((sat-constraints-list
		 (opsat-problem-sat-constraints-list search-problem)))
    (cond ((or (eql sat-constraints-list :no-sat-constraints)
			   (null sat-constraints-list))

		   ;; No SAT constraints to test, return success.

		   ;;	   (when *bfs-debug*
		   ;;	     (format t "~%  No Sat constraints to satisfy." ))
		   (values T nil nil))

		  (:else

		   ;; There are one or more sets of SAT constraints, 
		   ;; check each set against base constraints to determine
		   ;; consistency.

		   ;; 	   (when *bfs-debug*
		   ;;	     (format t "~%  Check if satisfies SAT constraints? ~a" 
		   ;;		     sat-constraints))
		   
		   (loop with consistent?
			   with constraint-system-conflict
			   with models-for-theory
			   for sat-constraints in sat-constraints-list
			   do 
				 ;; Check consistency of next set.
				 (multiple-value-setq (models-for-theory constraint-system-conflict)
				   (insert-constraints-and-find-all-models-if-any
					(opsat-problem-constraint-system
					 search-problem)
					:added-constraints sat-constraints
					:source search-problem
										;:perform-search? perform-search?
					:return-conflict? T)
				   )
				 
				 (setf consistent? models-for-theory)
						    
				 ;; (when *bfs-debug*
				 ;;   (format t "~%  Consistent = ~a for SAT constraint ~% ~a" 
				 ;;      consistent? sat-constraints))
			   collecting models-for-theory into list-of-sets-of-models
			   unless consistent?
			   do 
				 ;; Set is inconsistent, return failure
				 (cond ((null constraint-system-conflict)
						(return (values nil :CURRENT)))
					   
					   (:else
						(let ((candgen-conflict 
							   (convert-constraint-conflict-to-candgen-conflict 
								search-problem constraint-system-conflict)))
						  ;; (when *bfs-debug*
						  ;;	(format t "~%  ~%  Resulting conflict: ~a" 
						  ;;		candgen-conflict))
		      
						  (return (values nil candgen-conflict nil)))))
				 
				 
			   finally 
				 ;; All sets are consistent, return success.

				 (return (values T nil list-of-sets-of-models))
				 )))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod insert-state-assignment-into-constraint-system 
    ((search-problem opsat-problem) source assignment truth-value)
  
  "Asserts that ASSIGNMENT has TRUTH-VALUE in CONSTRAINT-SYSTEM, 
   based on SOURCE.

   SEARCH-PROBLEM is a search problem object.
   ASSIGNMENT is a state-assignment object of SEARCH-PROBLEM
     (not its constraint-system).
   TRUTH-VALUE is one of (*true* or *false*)
   SOURCE is a lisp object denoting the creator of the insertion."

  (assert! (opsat-problem-constraint-system search-problem)
		   (state-assignment-to-constraint-assignment assignment)
		   truth-value
		   source)
  )

;;; Check:
;;; o Is this redundant with remove-assignment-from-constraint-system?
;;;   It looks like the same.
(defmethod remove-state-assignment-from-constraint-system
    ((search-problem opsat-problem) source assignment truth-value)
  
  "Removes the assertion that ASSIGNMENT has TRUTH-VALUE in 
    CONSTRAINT-SYSTEM, based on SOURCE.

    SEARCH-PROBLEM is a search problem object.
    ASSIGNMENT is a state-assignment object of SEARCH-PROBLEM
     (not its constraint-system).
    TRUTH-VALUE is one of (*true* or *false*)
    SOURCE is a lisp object denoting the creator of the insertion."

  (unassert! (opsat-problem-constraint-system search-problem)
			 (state-assignment-to-constraint-assignment assignment)
			 truth-value
			 source)
  )


(defmethod achieves-problem-SAT-constraints-list? 
    ((search-problem conflict-directed-relaxation-problem) perform-search?)

  (let ((sat-constraints-list
		 (opsat-problem-sat-constraints-list search-problem)))
    (cond ((or (eql sat-constraints-list :no-sat-constraints)
			   (null sat-constraints-list))

		   ;; No SAT constraints to test, return success.
		   (values T nil))

		  (:else
		   (loop with consistent?
			  with constraint-system-conflict

			  for sat-constraints in sat-constraints-list
			  do 
				(multiple-value-setq (consistent? constraint-system-conflict)
				  (consistent-with? (opsat-problem-constraint-system
									 search-problem) 
									:perform-search? perform-search?
									:return-conflict? T
									:added-constraints sat-constraints
									:return-negative-cycle? t))
			;;	(format t "Has conflict ~A~%" constraint-system-conflict)
			;;	(setf constraint-system-conflict (pop constraint-system-conflict))
			;;	(format t "Reduced conflict ~A~%" constraint-system-conflict)

			  unless consistent?
			  do 
			  ;; Set is inconsistent, return failure
				(cond ((null constraint-system-conflict)
					   (return (values nil :CURRENT)))
					  (:else
					   (let ((candgen-conflict 
							  (convert-constraint-conflict-to-candgen-conflict 
							   search-problem (pop constraint-system-conflict))))
						 (return (values nil (list candgen-conflict (pop constraint-system-conflict)))))))
				
			  finally 
				(return (values T nil)))))))

;;; Check:
;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't nil denote the
;;;   conflict of empty, assignments, in which case everything is
;;;   inconsistent?  Check and correct this.

;;; To Do:
;;; o Rename added-clauses to added-constraints.
(defmethod achieves-problem-SAT-constraints-list? 
    ((search-problem opsat-problem) perform-search?)

  "Checks if the current search-state of SEARCH-PROBLEM is consistent
   with the problem's BASE + SAT constraints, for each set of
   Sat constraints.  
   If SAT constraints list is :NO-SAT-CONSTRAINTS or the empty list,
   then test immediately succeeds.  
   If Sat constraints list contains as an element the empty list '(), 
   then BASE is tested for consistency.

   If PERFORM-SEARCH? is true, tests consistency using full search,
   otherwise, performs constraint propagation only.
   ASSUMES the current search-state has already been inserted into 
   constraint system of SEARCH-PROBLEM. 

   RETURNS (values CONSISTENT? SEARCH-PROBLEM-CONFLICT).
   A conflict is a list of STATE-ASSIGNMENTS or
   :CURRENT, representing the current state assignment. 
   Each state assignment is a state-assignment object."

  (let ((sat-constraints-list
		 (opsat-problem-sat-constraints-list search-problem)))
    (cond ((or (eql sat-constraints-list :no-sat-constraints)
			   (null sat-constraints-list))

		   ;; No SAT constraints to test, return success.
		   ;;	   (when *bfs-debug*
		   ;;	     (format t "~%  No Sat constraints to satisfy." ))
		   (values T nil))

		  (:else

		   ;; There are one or more sets of SAT constraints, 
		   ;; check each set against base constraints do determine
		   ;; consistency.

		   ;; 	   (when *bfs-debug*
		   ;;	     (format t "~%  Check if satisfies SAT constraints? ~a" 
		   ;;		     sat-constraints))
		   
		   (loop with consistent?
			   with constraint-system-conflict

			   for sat-constraints in sat-constraints-list
			   do 
				 ;; Check consistency of next set.
				 ;; TO DO: 
				 ;; o Definitely a bad idea to tie this to isat this way,
				 ;;   but I think they're already checking all that

 ;; (format t " Checking consistency ~A~%" (opsat-problem-constraint-system   search-problem))

				 (multiple-value-setq (consistent? constraint-system-conflict)
				   (consistent-with? (opsat-problem-constraint-system
									  search-problem) 
									 :perform-search? perform-search?
									 :return-conflict? T
									 :added-constraints sat-constraints
									 :return-negative-cycle nil
									 ;; :save-learned-conflicts? t
									 ;; This is dangerous because
									 ;; there can be multiple, contradictory 
									 ;; sets of SAT constraints. Removed until
									 ;; dynamic CSPs are done correctly.
									 )
				   
;;;				       (consistent-with? (opsat-problem-constraint-system
;;;										  search-problem)
;;;										 :added-constraints sat-constraints
;;;										 :source search-problem
;;;										 :perform-search? perform-search?
;;;										 :return-conflict? T)
				   )
				 
				 ;; (when *bfs-debug*
				 ;;   (format t "~%  Consistent = ~a for SAT constraint ~% ~a" 
				 ;;      consistent? sat-constraints))
				 
			   unless consistent?
			   do 
				 ;; Set is inconsistent, return failure
				 (cond ((null constraint-system-conflict)
						(return (values nil :CURRENT)))
					   (:else
						(let ((candgen-conflict 
							   (convert-constraint-conflict-to-candgen-conflict 
								search-problem constraint-system-conflict)))
						  
						  ;; (when *bfs-debug*
						  ;;	(format t "~%  ~%  Resulting conflict: ~a" 
						  ;;		candgen-conflict))
		      
						  (return (values nil candgen-conflict)))))
				 
			   finally 
				 ;; All sets are consistent, return success.

				 (return (values T nil))
				 )))))

;;; Check:
;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't NIL denote the
;;;   conflict of empty, assignments, in which case everything is
;;;   inconsistent?  Check and correct this.

;;; To Do:
;;; o Rename added-clauses to added-constraints.
(defmethod achieves-problem-unsat-constraints-list? 
    ((search-problem opsat-problem) perform-search?)

  "Checks if the current search-state of SEARCH-PROBLEM is inconsistent
   with the problem's BASE + UnSat constraints, for each set of
   UnSat constraints.  
   If UnSat constraints list is :no-unsat-constraints or an empty list,
   then test immediately succeeds.  
   If UnSat constraint list is an empty list '(), then BASE
   is tested for inconsistency.

   If PERFORM-SEARCH? is true, tests using full search,
   otherwise, performs constraint propagation only.
   
   ASSUMES the current search-state has already been inserted into 
   constraint system of SEARCH-PROBLEM. 

   Note: For most applications the base constraints should be 
   consistent for all extensions to current state.

   RETURNS (values INCONSISTENT? SEARCH-PROBLEM-CONFLICT).
   The conflict is a subset of the current state sufficient
   to ensure that a set of unsat constraints is satisfied.
   In particular, a conflict is a list of STATE-ASSIGNMENTS or
   :CURRENT, representing the current state assignment. 
   Each state assignment is a state-assignment object. "

  (let ((unsat-constraints-list
		 (opsat-problem-unsat-constraints-list search-problem)))
    (cond ((or (eql unsat-constraints-list :no-unsat-constraints)
			   (null unsat-constraints-list))
		   
		   ;; No SAT constraints to test, return success.

;;;	       (when *bfs-debug*
;;;	          (format t "~%  No unsat constraints to satisfy." ))
		   (values T nil))

		  (:else
		   
		   ;; There are one or more sets of UNSAT constraints, 
		   ;; check each set against base constraints do determine
		   ;; inconsistency.

;;; 	   (when *bfs-debug*
;;;	          (format t "~%  Check if satisfies sets of UNSAT constraints ~a" 
;;;		        unsat-constraints-list))

		   (loop with inconsistent?
			   with constraint-system-conflict
			   for unsat-constraints in unsat-constraints-list
			   do 

				 ;; Check inconsistency of next set.
				 (multiple-value-setq 
					 (inconsistent? constraint-system-conflict)
				   (inconsistent-with? 
					(opsat-problem-constraint-system
					 search-problem)
					:added-constraints unsat-constraints
					:source search-problem
					:perform-search? perform-search?
					:return-conflict? T
										;	:base-theory-consistent-with-all-assignments? T
					)
				   
				   )

				 ;; (when *bfs-debug*
				 ;;   (format t "~%  Inconsistent? = ~a for unsat constraints ~% ~a" 
				 ;;      inconsistent? unsat-constraints))
			   unless inconsistent?
			   do 

				 ;; Set is NOT inconsistent, return failure.
				 (cond ((null constraint-system-conflict)
						(return (values nil :CURRENT)))
					   (:else
						(let ((candgen-conflict 
							   (convert-constraint-conflict-to-candgen-conflict 
								search-problem constraint-system-conflict)))
						  ;; (when *bfs-debug*
						  ;;	(format t "~%  ~%  Resulting conflict: ~a" 
						  ;;		candgen-conflict))
						  (return (values nil candgen-conflict)))))
			   finally
				 ;; All sets are inconsistent, return success.
				 (return (values T nil))
				 )))))

;;; Fix:
;;; o CONVERT-CONSTRAINT-CONFLICT-TO-CANDGEN-CONFLICT filters away any
;;;   assignment in a conflict that does not involve a decision
;;;   variable or that is assigned false.  Isn't it an error for
;;;   any such literal to appear?  If so, then merely filtering is
;;;   incorrect.

;;; FIX:
;;; o ISAT provides a truth assignment whose variable is a state-assignment.
;;;   ITC provides a state-assignment.
;;;   Both need to provide the same thing.
;;;   Make ISAT's KEY extract the state-assignment, 
;;;   rather than the truth-value assignment.
(defmethod convert-constraint-conflict-to-candgen-conflict 
    ((search-problem opsat-problem) constraint-system-conflict)

  "Takes as input a conflict CONSTRAINT-SYSTEM-CONFLICT
   that is encoded as a list of assignment objects in 
   the constraint-system of SEARCH-PROBLEM.

   RETURNS an equivalent conflict that is encoded 
   as a list of state-assignment objects of search-problem's
   candidate generator.

   CONSTRAINT-SYSTEM-CONFLICT is either :CURRENT or a list of 
   constraint-system assignments that do not achieve both the 
   SAT and UNSAT constraints."
  
  (if (eql constraint-system-conflict :CURRENT)
      (values :current)
    (loop for truth-assignment in constraint-system-conflict
		for cs-state-assignment = (assignment-variable truth-assignment)
		for truth-value = (assignment-value truth-assignment)
		for cg-state-assignment =
		  (convert-constraint-assignment-to-candgen-assignment 
		   search-problem cs-state-assignment)
		for cg-state-variable = (when cg-state-assignment
								  (state-assignment-variable cg-state-assignment))
		for cg-current-state-assignment = 
		  (when cg-state-variable 
			(state-variable-current-assignment cg-state-variable))
;;;		do (debug-print-for 
;;;			:opsat t "~%  Conflict literal ~s : ~s "
;;;			truth-assignment truth-value)
;;;		   (debug-print-for 
;;;			:opsat t "~%   -> ~s: var ~s,  cur asgn ~s."
;;;			state-assignment state-variable current-state-assignment)
;;;		   (debug-print-for 
;;;			:opsat t "~%     (eql bfs-asgn cur-asgn)? ~s: (eql truth: ~s *true*: ~s)? ~s."
;;;			(eql state-assignment current-state-assignment)  truth-value *true* (eql truth-value *true*))
		when (and cg-state-assignment
				  (eql cg-state-assignment cg-current-state-assignment)
				  (eql truth-value *true*))
			 
			 ;; Assignment is only part of the conflict if it is
			 ;; a member of the current assignments.
			 ;; Note: currently OPSAT only supports insertion of
			 ;; positive assignments, so a negative assignment 
			 ;; could not have been installed by OPSAT.
;;;		do (debug-print-for 
;;;			:opsat t "~%  Conflict literal: ~a, ~
;;;                      ~%    converts to assignment ~a and truth-value ~A."
;;;			truth-assignment state-assignment truth-value) and
		collect cg-state-assignment)))
