;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :OPSAT)

;;; This file contains procedures for setting and querying the current
;;; search state.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (SETUP-CURRENT-SEARCH-STATE-FROM-SEARCH-NODE
;;;    (SEARCH-PROBLEM OPSAT-PROBLEM) SEARCH-NODE )
;;;  
;;;   Sets the current state to be the best cost state of the
;;;   (best) kernel denoted by SEARCH-NODE.  Called for
;;;   effect, not value.
;;;    
;;;   The best cost state is comprised of the variable assignments 
;;;   of the (partial) kernel, augmented with the best cost 
;;;   assignments of all variables not mentioned in the kernel.

;;; (CURRENT-SEARCH-STATE
;;;    &OPTIONAL (SEARCH-PROBLEM *DEFAULT-SEARCH-PROBLEM*))
;;;  
;;;   Returns the current search state of SEARCH-PROBLEM, 
;;;   specified as a list of decision-assignment objects, one
;;;   for each decision variables.

;;;(CURRENT-SEARCH-CONSTRAINT-ASSIGNMENTS SEARCH-PROBLEM)
;;;  
;;;   Returns a list of assignments that represent the current 
;;;   search state of SEARCH-PROBLEM.
;;;   Each assignment is denoted by an assignment object of 
;;;   SEARCH-PROBLEM's constraint-system (not a decision-assignment 
;;;   object).

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Extend to conditional CSPs:
;;;   - add in comments of functions to change.
;;;   - modify functions as indicated by their to do.

;;; o Rename terms here to be consistent with names in the JDAM pseudo-code.  
;;;   Update comments appropriately.

;;; o Rethink the external output.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Optimization:
;;; o Rather than setting all variables to their default, incrementally 
;;;   switch between current assignments and default assignments as current 
;;;   states are added and removed.

;;; To Do:
;;; o Name should have !, since nothing is returned.
;;; o Get rid of optional, and make a method on opsat-problem.
(defmethod setup-current-search-state-from-search-node
    ((search-problem opsat-problem) search-node)
  
  "Sets the current state to be the best cost state of the
   (best) kernel denoted by SEARCH-NODE.  
   Function called for effect, not value.
    
   The best cost state is comprised of the variable assignments 
   of the (partial) kernel, augmented with the best cost assignments of 
   all variables not mentioned in the kernel."

;;  (declare (ignore full-assignment-supplied?))

  ;; Note: In the following, all assignments are decision-assignment 
  ;; objects.
  
  ;; Record the current state as the previous state.
  ;; Set up the new current state, 
  
  ;; 1) default the current state to the best assignment to each variable.
  ;; These assignments act as defaults for variable assignments
  ;; not specified by SEARCH-NODE.

  (loop for variable in (state-space-variables search-problem)
      do (setf (state-variable-previous-assignment variable)
		   (state-variable-current-assignment variable))
        
		 (setf (state-variable-current-assignment variable)
			   nil ;default to not assigning it!
;		   (decision-variable-best-assignment variable)
))
  
  ;; 2) Record over the current state the assignments of the best
  ;; (partial) kernel specified by the search-node.  Any variable 
  ;; not specified in the kernel gets its best cost value.

  ;; To extract the best kernel, note that a SEARCH-NODE 
  ;; is a search tree node comprised of a set of ancestor 
  ;; assignments, and a set of child assignments.  Each (partial) 
  ;; kernel is comprised of the ancestors plus one child 
  ;; assignment.  The best kernel is the ancestor assigments 
  ;; plus the first child assignment.

  ;; Record as current each ancestor-assignment plus ...
  (loop for new-assignment in
		(search-node-ancestor-assignments search-node)
      for variable = (state-assignment-variable new-assignment)
      do (setf (state-variable-current-assignment variable) 
		   new-assignment))

  ;; ... the best child assignment.
  (let ((new-assignment
		 (search-node-preferred-child-assignment search-node)))
    (unless (null new-assignment)
      (setf (state-variable-current-assignment
			 (state-assignment-variable new-assignment))
		new-assignment)))
  
  ;;  (debug-print-for :opsat t "Current state ~a"
  ;;	       (current-search-state search-problem))
  T)

(defmethod current-search-state
    ((search-problem opsat-problem))
  
  "Returns the current search state of SEARCH-PROBLEM, 
   specified as a list of decision-assignment objects, one
   for each decision variables."
  
  (loop for var in (state-space-variables search-problem)
      collecting (state-variable-current-assignment var)))

;;; Proposed Change:

;;; o Think through the external output format.  Do we want this 
;;;   to return assignment structures, arbitrary 
;;;   expressions or variable/value binding expressions?

;;; o Do we want to eliminate implicit assignments within OPSAT,
;;;   or perform extraction external to OPSAT?  This seems like 
;;;   an extra layer of complexity.

;;; Optimization:
;;; o when an search-node denotes a full assignment, it would
;;;   be faster to pass along the search-node's assignments
;;;   rather than consing up a new list from current search state.
(defmethod current-search-constraint-assignments 
	((search-problem opsat-problem))
  
  "Returns a list of assignments that represent the current 
   search state of SEARCH-PROBLEM.
   Each assignment is denoted by an assignment object of 
   SEARCH-PROBLEM's constraint-system (not a decision-assignment 
   object)."
  
  (loop for variable in (state-space-variables search-problem)
	 unless (null (state-variable-current-assignment variable))
	 collect (state-assignment-to-constraint-assignment
			  (state-variable-current-assignment variable))))
    
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; None