;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :OPSAT)


;;; FIX:
;;; o mtu::declare-debug-print-system!
;;;   has been exported from MTU in its package definition,
;;;   and included in bfig.  Why isn't the function external (: vs ::)
;;;   and why isn't it part of :OPSAT?

(mtu::declare-debug-print-system! :opsat)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			;; Truth values
;;;			*true* true
;;;			*false* false
;;;			*none* none
;;;
;;;			opsat-problem-constraint-system
;;;            *no-superset-solutions?*
;;;            )
;;;          :OPSAT)
;;;  )

;;; This file defines the data structure used to denote an OPSAT-PROBLEM.
;;; It also includes the definition of flags used to informed search. 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; (OPSAT-PROBLEM-CONSTRAINT-SYSTEM OPSAT-PROBLEM)

;;; A constraint system containing the constraints of 
;;; the Optimal CSP.

;;; *NO-SUPERSET-SOLUTIONS?* default value T.

;;;   This variable controls whether all consistent states are admitted as
;;;   solutions or only kernel assignments that are not a superset of 
;;;   another set of assignments.  Useful only when some assignments are 
;;;   eliminated as implicit assignments.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; o Extend to conditional CSPs:
;;;   - add in comments of functions to change.
;;;   - modify functions as indicated by their to do.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Global variable and class definitions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *no-superset-solutions?* t
  "This variable controls whether all consistent states are admitted as
   solutions or only kernel assignments that are not a superset of 
   another set of assignments.  Useful only when some assignments are 
   eliminated as implicit assignments.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Debugging global variables (for opsat-reader)
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *opsat-debug* nil "T when debugging opsat-Model code")

(defvar *opsat-debug-model* nil  
  "Stores current model being worked on when *opsat-debug* is T")

;;; Changed.
;;; o Class definitions
;;;   * OPSAT-PROBLEM just performs Constraint-based A*
;;;   * conflict-directed-opsat-problem now is just an extension of
;;;     (opsat-problem informed-search-problem-w-conflicts)
;;;   * conflict-directed-opsat-problem performs CDA*
;;; -- gustavop

(defclass opsat-problem (informed-search-problem)

  ((name
	:initarg :name
	:initform (gentemp "opsat-problem-")
	:accessor opsat-problem-name
	:documentation "Name of an Optimal CSP.")

   (constraint-system
	:initform nil
	:accessor opsat-problem-constraint-system
	:documentation "A constraint system containing the constraints of the 
                    Optimal CSP.")

   (base-constraints
	:initform '()						
	:accessor opsat-problem-base-constraints
	:documentation "Constraints shared by the SAT and UNSAT constraints.")

   (sat-constraints-list			
	:initform 'nil
	:accessor opsat-problem-sat-constraints-list
	:documentation "List of sets of constraints, such that each set, when 
                    combined with the Base constraints, define the SAT 
                    condition.")
  
   (unsat-constraints-list
	:initform '()
	:accessor opsat-problem-unsat-constraints-list
	:documentation "List of sets of constraints, such that each set, when 
                    combined with the Base constraints, define the UNSAT 
                    condition.")

   (convert-constraint-to-candgen-assignment-ht
	:initform nil
	:accessor opsat-problem-convert-constraint-to-candgen-assignment-ht
	:documentation "Hash table used to translate assignments encoded by the 
                    constraint system, to their equivalent assignments, as 
                    encoded by best first search.")
  
   (state-trace
	:initform nil
	:accessor opsat-problem-state-trace
	:documentation "Object encoding a trace of visited states for 
                    search problem."))


  (:documentation "An Optimal Satisfiability problem --, a
                   finite domain constraint optimization problem with
                   Sat and Unsat constraints specified in propositional
                   state logic -- that will be w/o conflict-direction,
                   i.e., through constraint-based a*"))

;;; Optimization:
;;; o How costly is the translation between constraint to candgen encoding of
;;;   assignments.  Is their a more efficient way to mediate between the two?

;;; Changed:
;;; o Added comment.

;;; Changed:
;;; o slots definitions moved to constraint-based-astar-problem def.
;;; -- gustavop
(defclass conflict-directed-opsat-problem (opsat-problem
										   informed-search-problem-w-conflicts)

  
  ()
  (:documentation "An Optimal Satisfiability problem --, a
                   finite domain constraint optimization problem with
                   Sat and Unsat constraints specified in propositional
                   state logic."))



;;; Added:
;;; o When asking for (leading) solutions, we may now not need
;;; just the decision variable assignments. We may also want
;;;   - Enumeration of all models consistent with solution
;;;   - Return cost of the solution
;;; Then it is more convenient to encapsulate all this info
;;; within opsat-solution.
;;; o The slot ''models'' is a list where all models related
;;; to the opsat-solution are stored. It may seem redundant
;;; to access it as consistent-models (since a model should
;;; be consistent by definition) but I believe it is clearer this way.
;;; - Who wrote this, Gustavo Pinheiro? - BCW
;;; - Yes, I added this. --gustavo
(defclass opsat-solution ()

  ((name
	:initarg :name
	:initform (gentemp "opsat-solution-")
	:accessor opsat-solution-name
	:documentation "Name of an OPSAT Solution.")

   (decision-assignments
	:initarg :decision-assignments
	:initform nil
	:accessor opsat-solution-decision-assignments
	:documentation "Decision variables assignments
			in this solution")

   (constraint-relaxations
	:initarg :constraint-relaxations
	:initform nil
	:accessor opsat-solution-constraint-relaxations
	:documentation "Relaxations for some constraints")

   (sets-of-consistent-models
	:initarg :sets-of-consistent-models
	:initform nil
	:accessor opsat-solution-sets-of-consistent-models
	:documentation "Consistent extensions to the partial-assignment
         			defined by decision variables. Defaults to nil, i.e.
        			we normally don't ask to enumerate all consistent models.")

   (objective-fns-costs-ht
	:initform (make-hash-table :test #'equal)
	:accessor opsat-solution-objective-fns-costs-ht
	:documentation "Maps the name of an Objective Function to its value for this opsat-solution"))

  (:documentation "A solution for an OPSAT problem. "))

(defmethod print-object ((opsat-sln opsat-solution) stream)
  (with-slots (decision-assignments) opsat-sln
	(format stream "~a" decision-assignments)))
   
;;; To Do:
;;; o Should state-variable-descriptions include guards?

;;; Changed:
;;; o Updated documentation for state/decision-variable-descriptions.
;;; o Added fix comments.

;;; Fix:
;;; o In tpn-start-event-name.  The term TPN should not appear within OPSAT.
;;;   OPSAT problems shouldn't in general have start events.  Instead, this
;;;   is only relevant to OPITC problems.  Rewrite.

(defmethod initialize-instance :after 
		   ((search-problem opsat-problem) 
			&key
			(decision-variable-descriptions '())
			(state-variable-descriptions '())
			(base-constraint-descriptions '())
			;; Defaults to testing consistency on the base constraints alone.
			(constraint-system-type 'theory)
			;; Initialize start-event-name with default value nil. 
			;; Pass the value to the start-event-name slot of constraint-system 
			;; in opsat-problem. - Peng Yu.
			(start-event-name 'start) 
			(sat-constraint-descriptions-list '(()) )
			(unsat-constraint-descriptions-list :no-unsat-constraints)
			(binary-objective-function #'+)
			(cost-offset 0)
			(cost-order-function #'>=))

  "Creates and returns a finite domain optimization problem with 
   constraints of type constraint-system-type.  Default to a problem that 
   maximizes the sum of the decision variable assignment costs, given a set
   of constraints in propositional state logic.  The 
   objective function can be modified by calling add-objective-function.

   State variables are finite domain variables that are mentioned 
   in state constraints.
   Decision variables is a subset of the state variables 
   that are being optimized over.

   CONSTRAINT-SYSTEM-TYPE is an object type with slot NAME and
     initarg :NAME. CONSTRAINT-SYSTEM-TYPE defaults to THEORY.
     CONSTRAINT-SYSTEM-TYPE supports the following methods:

      It supports the following methods: 

      NEW-CONSTRAINT is a method that creates a state constraint
       for CONSTRAINT-SYSTEM.  Its arguments are 
          (CONSTRAINT-SYSTEM :SOURCE SOURCE . CONSTRAINT-DESCRIPTION).

      INSERT-CONSTRAINT is a method that inserts a constraint
       into CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed by 
       CONSTRAINT-OBJECT as arguments.

      GET-ASSIGNMENT is a method that takes arguments
      THEORY, VARIABLE AND VALUE, and looks up the object representing 
       assignment within CONSTRAINT-SYSTEM.

      ASSERT! is a method that takes arguments:
        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
           &optional (TRUTH-VALUE *true*) (SOURCE NIL))
       It asserts that ASSIGNMENT 
       has TRUTH-VALUE (*true* or *false*) in CONSTRAINT-SYSTEM, 
       justified by SOURCE.
       It returns nothing useful.

      UNASSERT! is a method that takes arguments:
        (CONSTRAINT-SYSTEM CONSTRAINT-ASSIGNMENT-OBJECT 
          &optional (TRUTH-VALUE *true*)(SOURCE NIL) )
       It removes the assertion that ASSIGNMENT has TRUTH-VALUE 
       (*true* or *false*) in CONSTRAINT-SYSTEM due to justification SOURCE.
       It returns nothing useful.

      CONSISTENT-WITH? is a method that takes as arguments:
        (THEORY &OPTIONAL &KEY (ADDED-CONSTRAINTS NIL) (SOURCE NIL)
          (PERFORM-SEARCH? T) (RETURN-CONFLICT? T))
       It tests the consistency of the current constraints and assignments 
       in THEORY, conjoined with ADDED-CONSTRAINTS and ADDED-ASSIGNMENTS.

       It returns (VALUES CONSISTENT? CONFLICT),
       where CONSISTENT? is true unless theory is proven inconsistent.  

       Each constraint in ADDED-CONSTRAINTS is a constraint object.
       Each assignment in ADDED-ASSIGNMENTS is a list:
         (VARIABLE-NAME VALUE &OPTIONAL (TRUTH-VALUE *TRUE*))
       and specifies that proposition (= VARIABLE VALUE) has TRUTH-VALUE.

       It first tests consistency by propagation.
       If PERFORM-SEARCH? is TRUE, and propagation is consistent, then it 
       continues with a full DPLL-style satisfiability test.  

       If RETURN-CONFLICT? is TRUE and an inconsistency is found by unit 
       propagation, then it returns a subset of the current inserted 
       assignments that led to an inconsistency.
     
      INCONSISTENT-WITH? is a method that takes as arguments:
        (THEORY &optional &KEY (added-constraints nil) (source nil)
          (perform-search? T) (return-conflict? T)
          (base-theory-consistent-with-all-assignments? T))
       It tests the INCONSISTENCY of the current constraints and 
       assignments in THEORY, conjoined with ADDED-CONSTRAINTS.

       It RETURNS (values INCONSISTENT? CONFLICT).  INCONSISTENT? is 
       false unless theory is proven inconsistent.  It first tests 
       consistency by unit propagation.  Then if theory is consistent
       by unit propagation and PERFORM-SEARCH? is TRUE, 
       it continues with a full DPLL-style satisfiability test.  

       Each constraint in ADDED-CONSTRAINTS is a constraint object.
       Each assignment in ADDED-ASSIGNMENTS is a list:
         (VARIABLE-NAME VALUE &optional (TRUTH-VALUE *TRUE*))
       and specifies that proposition (= variable value) has TRUTH-VALUE.
 
       If RETURN-CONFLICT? is TRUE and theory is proven necessarily 
       true by unit propagation, then it returns a subset of the current 
       decision variable assignments that is sufficient to guarantee that 
       theory is satisfied.

       If BASE-THEORY-CONSISTENT-WITH-ALL-ASSIGNMENTS? is True, then 
       it assumes that the inserted assignments are consistent with the 
       constraints in theory (minus added-constraints).
       In this case a conflict is a subset of the decision variable assignments 
       that, together with THEORY (minus assignments) entail ADDED-CONSTRAINTS.

       A conflict is created by searching over a subset of the current 
       constraints that support the truth of all ADDED-CONSTRAINTS, or all 
       non-unit clauses, and collecting the set of all propositions in
       these clauses such that the proposition denotes an assignment to 
       a decision variable, and the proposition's truth assignment was
       made during search by splitting.

   STATE-VARIABLE-DESCRIPTIONS is a list:
     (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)* 
   where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of state variable.  It currently consists of
    one required keyword pair :DOMAIN DOMAIN where DOMAIN is a list of values.
   
   DECISION-VARIABLE-DESCRIPTIONS is a list:
    (VARIABLE-NAME . VARIABLE-KEYWORD-LIST)*
   where VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of decision variable.  It currently consists of
    one required keyword :DOMAIN and one optional keyword :GUARD:

      (:DOMAIN (DOMAIN-ELEMENT-DESCRIPTION+) 
       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *true*])

    where each DOMAIN-ELEMENT-DESCRIPTION is either an atom, denoting 
    a domain element, or list of the form:

      (DOMAIN-ELEMENT <KEYWORD-VALUE-PAIR>*)
  
   where DOMAIN-ELEMENT is the value of the assignment, and  
   KEYWORD-VALUE-PAIRS correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of the VARIABLE class.
   This currently consists of :COST.

   :GUARD is optional, and denotes an assignment that
   activates variable.

   The multi-attribute objective function of the optimization problem 
   is defined over the decision variables.  The attribute cost of
   each decision variable value is specified in DECISION-VARIABLE-DESCRIPTIONS.

   SEARCH-ORDER is the order in which search-tree nodes is expanded.
   It if one of the keywords in SEARCH-PROBLEM-SEARCH-ORDER, and defaults
   to :best-first.

   The composition of attribute costs specified jointly by 
   BINARY-OBJECTIVE-FUNCTION, COST-OFFSET and COST-ORDER-FUNCTION. 
   BINARY-OBJECTIVE-FUNCTION specifies how attribute costs are pair-wise
    combined, and defaults to +.  This function must satisfy preferential
    independence.
   COST-OFFSET specifies a constant value to be added to all costs.
   COST-ORDER-FUNCTION places a total order on values, and is used to
    define the best value.  The default, >=, corresponds to a solution
    that maximizes cost (utility).

   BASE-CONSTRAINT-DESCRIPTIONS is a list of state logic 
    constraint descriptions.  The BASE STATE CONSTRAINTS, should
    be consistent for every decision-variable assignment; the
    modeler violates this rule at his or her own risk.

   A CONSTRAINT DESCRIPTION is a constraint, denoted by
   a list of STATE LITERALS.

   A STATE LITERAL is a literal whose proposition is an assignment 
    (VARIABLE-NAME VALUE).  VARIABLE-NAME must be one described in
    STATE-VARIABLE DESCRIPTIONS, and VALUE must be an element of
    VARIABLE-NAME's DOMAIN in that description.

   SAT-CONSTRAINT-DESCRIPTIONS-LIST is either the keyword
   :NO-SAT-CONSTRAINTS, or a list of ONE or more 
   SAT-CONSTRAINT-DESCRIPTIONS, each of which denotes a list of
   ZERO or more constraints.  Each constraint is described by a 
   CONSTRAINT DESCRIPTION, as specified above.  
   If a list is provided, then for each list element SAT-CONSTRAINTS,
   the solutions generated must satisfy the union of BASE-CONSTRAINTS 
   and SAT-CONSTRAINTS.
   Note that this condition will be enforced even when the 
   SAT-CONSTRAINTS list is empty.

   If keyword :NO-SAT-CONSTRAINTS is provided, then no sat constraint
   is imposed, and the base constraints are not tested for consistency
   alone.  This is equivalent to SAT-CONSTRAINT-DESCRIPTIONS = ().

   SAT-CONSTRAINT-DESCRIPTIONS-LIST defaults to '(()), which
   imposes the condition that the solution satisfy the base constraints.
  
   UNSAT-CONSTRAINT-DESCRIPTIONS-list is either the keyword
   :NO-UNSAT-CONSTRAINTS, or a list of ONE or more 
   UNSAT-CONSTRAINT-DESCRIPTIONS, each of which denotes a list of
   ZERO or more constraints.  Each constraint is described by a 
   CONSTRAINT DESCRIPTION, as specified above.  
   If a list is provided, then for each list element UNSAT-CONSTRAINTS,
   the solutions generated must violate the union of BASE-CONSTRAINTS 
   and UNSAT-CONSTRAINTS.
   Note that this condition will be enforced even when the 
   UNSAT-CONSTRAINTS list is empty.

   SAT-CONSTRAINT-DESCRIPTIONS-LIST defaults to :NO-UNSAT-CONSTRAINTS
   specifying that no unsat constraint is imposed, and the base constraints 
   are not tested for consistency alone.  This is equivalent to 
   UNSAT-CONSTRAINTS = ().

   SOURCE is an object that invoked the creation of OPSAT problem.
  
   NAME is an sexpression naming the OPSAT problem."
  
  (with-slots (name) search-problem
	(let* ((source (state-space-source search-problem))
		   (constraint-system
			;; Creates constraint system, should be 
			;; parameterized to a type.
			(make-instance constraint-system-type
			  :name `(constraint-system ,constraint-system-type ,name)
			  ;;Added by Peng; Add the start tpn event name to opsat problem.
			  :start-name start-event-name)))
	  ;;(props (theory-propositions constraint-system))

	  (setup-opsat-problem! search-problem)
    
	  ;; o Initialize the multi-attribute cost function of 
	  ;;   the constraint optimization problem.  By default 
	  ;;   the cost function maximizes (minimizws?) the sum of decision variable 
	  ;;   assignment costs.

	  (add-objective-function! search-problem
							   binary-objective-function
							   cost-offset
							   cost-order-function)
	  
	  ;; o Setup the constraints for the OPSAT problem:
	  ;;   Specify the domains of each variable and 
	  ;;   the state constraints between variables.

	  (add-constraint-system! search-problem
							  constraint-system)
  4
	  ;; Introduce the decision and state variables into
	  ;; the OPSAT problem.
	  (add-state-variables! constraint-system
							state-variable-descriptions
							source)

	  (add-decision-variables! search-problem
							   decision-variable-descriptions
							   source)
    
	  ;; o Create instances of all constraints for SEARCH PROBLEM,
	  ;;   and associate them with CONSTRAINT SYSTEM.

	  ;;   Insert the BASE CONSTRAINTS in CONSTRAINT SYSTEM.

	  (add-base-constraints! search-problem
							 base-constraint-descriptions
							 source)
	
	  ;;   Create and associate the SAT constraints with constraint-system.
	  (add-sat-constraints! search-problem
							sat-constraint-descriptions-list 
							source)
    
	  ;;   Create and associate the UNSAT constraints with constraint-system.
	  (add-unsat-constraints! search-problem
							  unsat-constraint-descriptions-list
							  source)

	  (bfig:initialize-bit-ht! search-problem)

	  T)))

;;; Changed:
;;; o Added.
(defclass conditional-opsat-problem 
	(informed-conditional-search-problem-w-conflicts opsat-problem)

  ()
  
  (:documentation "An OPSAT problem with conditional variables."))


(defclass conflict-directed-relaxation-problem (conditional-opsat-problem)

  
  ()
  (:documentation "An Optimal Satisfiability problem that allows
constraints to be relaxed in order to restore consistency"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Macros Internal to OpSat
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro convert-constraint-assignment-to-candgen-assignment
    (search-problem cs-assignment)
  
  " Given an assignment object, CS-ASSIGNMENT, for the constraint
    system of SEARCH-PROBLEM, returns the equivalent assignment, as
    represented by the search problem itself."
  
  `(get-assignment-from-hash 
    ,cs-assignment
    (opsat-problem-convert-constraint-to-candgen-assignment-ht ,search-problem)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Clause and Constraint Descriptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; A clause description is a property list that includes:
;;; (:clause <list of literals>)
;;; Note that the list of literals is the second element, NOT rest.
;;; To Do:
;;; o Check if these duplicate 
(defconstant *clause-description-type* :clause
  "The type of a constraint description, indicating a clause.")

(defmacro constraint-description-type (description)

  "Returns a keyword denoting the type of constraint DESCRIPTION."
  
  `(first ,description))

(defmacro clause-description? (constraint-description)
  
  "Returns true iff CONSTRAINT-DESCRIPTION is a clause description."
  
  `(and (consp ,constraint-description)
		(eql (constraint-description-type ,constraint-description) *clause-description-type*)))

(defmacro make-clause-description (literals)
  
  "Given a list of literals, returns a description for a clause that is
   a disjunction of literals."
  
  `(list *clause-description-type* ,literals))

(defmacro clause-description-literals (clause-description)
  
  "Returns a list of the literals CLAUSE-DESCRIPTION."
    
  `(second ,clause-description))

(defun check-clause-descriptions! (constraint-descriptions)
 
  "Generates a warning for every element of constraint-descriptions
   that is not a clause description."
  
  (loop for description in constraint-descriptions
	  unless (clause-description? description)
	  do (warn t "~a is not a clause description" description)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Variable Descriptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro make-decision-variable-domain-description 
	(&key variable-name domain guard)
  
  "Creates a decision variable description for a variable
   named VARIABLE-NAME, with DOMAIN specified by a list of 
   DOMAIN element-descriptions, and optional GUARD, which
   enables the variable to be assigned."

  (if (null guard)
	  `(list ,variable-name 
			 :DOMAIN ,domain)
	`(list ,variable-name 
		   :DOMAIN ,domain
		   :GUARD ,guard)))

(defmacro make-decision-variable-domain-element-description 
	(&key value cost)
  
  "Creates a value description for an element of a decision variable,
   with VALUE and COST."
  
  `(list ,value :COST ,cost))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Indexed Variable Descriptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro add-index-to-variable-name (variable-name index)

  "Adds a temporal INDEX to VARIABLE-NAME."
  
  `(list :AT ,variable-name ,index))

(defmacro name-of-indexed-variable (indexed-variable-name)

  "Extracts VARIABLE-NAME from a time INDEXED-VARIABLE-NAME."
  
  `(second ,indexed-variable-name))

(defmacro index-of-indexed-variable (indexed-variable-name)

  "Extracts INDEX from a time INDEXED-VARIABLE-NAME."
  
  `(third ,indexed-variable-name))

(defun indexed-variable-< (variable1 variable2)
  
  "Returns true iff indexed VARIABLE1 precedes VARIABLE2 by index,
                 or they co-occur but VARIABLE1 precedes variable2 
                 lexicographically."

  (or (< (index-of-indexed-variable variable1)
		 (index-of-indexed-variable variable2))
	  
	  (string-lessp 
	   (format nil "~s" (name-of-indexed-variable variable1))
	   (format nil "~s" (name-of-indexed-variable variable2)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Indexed Assignment Descriptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun make-indexed-assignment (&key variable-name value index)
  
  "Returns an assignment expression with VARIABLE-NAME
   indexed by INDEX."
  
  (make-assignment :VARIABLE-NAME (add-index-to-variable-name
								   variable-name
								   index)
				   :VALUE value))

(defun add-index-to-assignment (assignment index)

  "Adds a temporal INDEX to the variable-name of ASSIGNMENT,
   Returning the INDEXED-ASSIGNMENT.
   ASSIGNMENT is of the form (= VARIABLE-NAME VALUE)
   indexed-assignment is of the form
   (= (VARIABLE-NAME INDEX) VALUE) "
  
  (make-indexed-assignment :variable-name (assignment-variable assignment)
						   :value (assignment-value assignment)
						   :index index))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; VCSP problem
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass conflict-directed-opsat-vcsp
    (opsat-problem informed-search-problem-w-conflicts-vcsp)
  ())



;;;;;;;;;;;;;;;
; OPSAT-MODEL ;
;;;;;;;;;;;;;;;

(defclass opsat-model ()

  ((name
	:initarg :name
	:initform (gentemp "opsat-model-")
	:accessor opsat-model-name
	:documentation "Name of an OPSAT Model complying with .ops
			format")

   (domain-ht
	:initform (make-hash-table :test #'equal)
	:accessor opsat-model-domain-ht
	:documentation "Maps a domain-name to a list of values, i.e, the domain's elements")

   (variable-ht
	:initform (make-hash-table :test #'equal)
	:accessor opsat-model-variable-ht
	:documentation "A hash table mappings every problem variable to its
			respective object.")

   (decision-variables
	:initform '()
	:accessor opsat-model-decision-variables
	:documentation "The variables that are related to utility of a solution.
			List of <(VARIABLE-NAME DOMAIN-NAME>*.")

   (state-variables
	:initform '()
	:accessor opsat-model-state-variables
	:documentation "The variables that are solely related to satisfiability.
			List of <(VARIABLE-NAME DOMAIN-NAME>*.")
   
   (hard-constraints
	:initform '()
	:accessor opsat-model-hard-constraints
	:documentation "The hard-constraints for the opsat-problem in CNF
			format. List of clauses. Clause represented as a list
			of literals")

   (objective-function-system
	:initform '()
	:accessor opsat-model-objective-function-system
	:documentation "List of objective functions to be
			optimized over."))

  (:documentation "An OPSAT model built according to .ops format."))


(defclass opsat-model-variable ()
  ((name
	:initarg :name
	:initform (gentemp "opsat-model-variable-")
	:accessor opsat-model-variable-name
	:documentation "A variable for opsat problem. May be either
			a decision or state variable.")
   (domain-name
	:initarg :domain-name
	:initform nil
	:accessor opsat-model-variable-domain-name)
   (decision-variable?
	:initform nil
	:accessor opsat-model-variable-decision-variable?))
  (:documentation "A variable for opsat problem. May be either
			a decision or state variable."))

(defclass opsat-objective-function ()

  ((name
	:initarg :name
	:initform (gentemp "opsat-objective-function-")
	:accessor opsat-objective-function-name
	:documentation "Name of an OPSAT Objective Fn that
			should comply with .ops format")

   (cost-order
	:initarg :cost-order
	:initform 'MAX
	:accessor opsat-objective-function-cost-order
	:documentation "Predicate for determining the preference
			among solutions, >= for Maximization and
			<= for Minimization.")

   (composition-operator
	:initarg :composition-operator
	:initform '+
	:accessor opsat-objective-function-composition-operator
	:documentation "Function for composing attribute costs.")

   (soft-constraint-system
	:initarg :soft-constraints
	:initform nil
	:accessor opsat-objective-function-soft-constraint-system
	:documentation "List of soft-constraints related to this
			objective. A soft-constraint is represented
			as <(SCOPE <(TUPLE COST)>*>"))

  (:documentation "An objective function that encapsulates a
		    preference, a composition operator, and
		    soft-constraints."))



;;; Define a cost system that maps the options directly to their costs. In the
;;; opsat-model, there is implicitly a three-dimensional cost array which looks like:
;;;
;;;                Option_1-1   Option_1-2  ...  Option_1-n1   Option_2-1  ...   Option_2-n2 ...
;;; Objective 1      c1-1-1       c1-1-2    ...    c1-1-n1       c1-2-1    ...     c1-2-n2   ...
;;; Objective 2      c2-1-1       c2-1-2    ...    c2-1-n1       c2-2-1    ...     c2-2-n2   ...
;;;   ... ...
;;;
;;; For example, in the toy model, we have
;;;
;;; Objective 1:
;;; ((((A TOY-INSTANCE)) (((A-2) 14) ((A-4) 28) ((A-7) 49)))
;;;  (((B TOY-INSTANCE)) (((B-3) -15) ((B-8) -40)))
;;;  (((C TOY-INSTANCE)) (((C-2) -4) ((C-5) -10) ((C-6) -12) ((C-9) -18))))
;;;
;;; Objective 2:
;;; ((((A TOY-INSTANCE)) (((A-2) -2) ((A-4) -4) ((A-7) -7)))
;;;  (((B TOY-INSTANCE)) (((B-3) 30) ((B-8) 80)))
;;;  (((C TOY-INSTANCE)) (((C-2) 4) ((C-5) 10) ((C-6) 12) ((C-9) 18))))
;;;
;;; The corresponding table looks like:
;;;                A-2   A-4   A-7   B-3   B-8   C-2   C-5   C-6   C-9
;;; Objective 1     14    28    49   -15   -40    -4   -10   -12   -18
;;; Objective 2     -2    -4    -7    30    80     4    10    12    18
;;;
;;; In the opsat-model class, it is organized as Objectives -> Scopes -> Costs,
;;; while what we usually need in multi-objective problems is Scopes ->
;;; Objective -> Costs. This class is designed in this form. There are two usage
;;; of this class: quickly build a new soft-constraint-system and quickly get
;;; the cost of a certain variable or a tuple.
;;;
;;; 'objective-names' stores all the objective names. 'scope-names' stores the
;;; all the scope names. 'option-names' stores all the option names in each
;;; scope. 'cost' is a three-dimensional array that stores the costs. 'cost-map'
;;; is a hash-table that can map the name of the option to its cost.
;;;
(defclass opsat-cost-system ()
  
  ((objective-names
	:initform nil
	:accessor opsat-objective-names
	:documentation "The objective names of the model. The sequence of costs in the value-tuple
                        will follow the sequence of this list.")
   
   (scope-names
	:initform nil
	:accessor opsat-scope-names
	:documentation "The scopes of the costs. Could be a single decision variable or a tuple.")
   
   (option-names
	:initform nil
	:accessor opsat-option-names
	:documentation "This is a two-dimensional array, storing the options names for each scope")
	 
   (costs
	:initform nil
	:accessor opsat-costs
	:documentation "The three-dimensional array of the costs")
   
   (cost-map
	:initform (make-hash-table :test #'equal)
	:accessor opsat-cost-map
	:documentation "The key of the hash table are the options in all domains. The mapping of
                         the key is an array of costs corresponding each the objective.")))

