 ;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

 (in-package :OPSAT)

 ;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

 ;;; This software may not be redistributed, and can only be retained
 ;;; and used with the explicit written consent of the author,
 ;;; Brian C. Williams.
 ;;; This software is made available as is; no guarantee is provided
 ;;; with respect to performance or correct behavior.
 ;;; This software may only be used for non-commercial, non-profit,
 ;;; research activities

 ;;; This file provides the necessary methods to setup an opsat-problem
 ;;; from an opsat-model description.


 ;; To Do:
 ;; o Move this recently add code (which creates opsat-models and problems from vcsp
 ;; modeled in .mof) to a new package. :OPSAT should not contain files unrelated to the search algorithm.


 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;   Call the multi-objective solver.
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod solve-multi-objective (mof-file
				  &optional (cutoff-delta-set nil))
  "Solve the multi-objective problem described in the mof-file. Create an
   opsat-problem for that description and call the find-fuzzy-pareto-solutions
   method to solve it. Finally parse the solutions and writ the result into xml
   file."

  (let* ((opsat-model (pre-process (opsat-load mof-file) cutoff-delta-set))
	 (solutions (post-process opsat-model 
				  (find-fuzzy-convex-pareto-front opsat-model
								  (first cutoff-delta-set)))))
    (format t "~%~%Solutions:~%Length: ~a ~%~%"(length solutions))
    (loop for solution in solutions do
	  (format t "    ~A~%" (opsat-solution-decision-assignments solution))
	  (maphash #'(lambda (k v)
		       (format t "    ~A = ~A" k v))
		   (opsat-solution-objective-fns-costs-ht solution))
	  (format t "~%~%"))))
    
;    (parse-solutions-to-string solutions)))


(defmethod pre-process ((model opsat-model) cutoff-delta-set)
  
  "Adjust the soft constraints' value to all '#+ and 'MAX."

  (when (not (null (first cutoff-delta-set)))
    (loop
	with obj-system = (opsat-model-objective-function-system model)
			  
	with total-objs = (length obj-system)
	with cutoff-value = (first cutoff-delta-set)
	with cutoff-names = (second cutoff-delta-set)

	with new-cutoff-delta = (make-array total-objs)
	with i = 0
				
	for obj in obj-system do

	  (loop with obj-name = (format nil "~A" (first (opsat-objective-function-name obj)))
	      for j from 0 to (- total-objs 1) do
		(when (equal (aref cutoff-names j)
			     obj-name)
		  (setf (aref new-cutoff-delta i) (aref cutoff-value j))))
	
	  (incf i)
	
	finally (setf (first cutoff-delta-set) new-cutoff-delta)))
  
  (loop
      with obj-system = (opsat-model-objective-function-system model)
      with soft-constraints = nil

      for obj in obj-system do
	(setf soft-constraints (opsat-objective-function-soft-constraint-system obj))
	
	; If the composition operator is product, do log on each value
	(when (equal (opsat-objective-function-composition-operator obj)
		     '#'*)
	  (loop for constraint in soft-constraints do
		(loop with tuples = (second constraint)
				    
		    for tuple in tuples do
		      (setf (second tuple) (log (second tuple))))))
	
	; If the cost order is MIN, negate the values
	(when (equal (opsat-objective-function-cost-order obj)
		     'MIN)
	  (loop for constraint in soft-constraints do
		(loop with tuples = (second constraint)
				      
		    for tuple in tuples do
		      (setf (second tuple) (- 0 (second tuple))))))
	
      finally (return model)))


(defmethod post-process ((model opsat-model) solutions)
  
  "Calculate the real objective value if the operator is '#'* or 'MAX."

  (loop with obj-system = (opsat-model-objective-function-system model)
			  
      for obj in obj-system do
	; If the cost order is MIN, negate the values
	(when (equal (opsat-objective-function-cost-order obj)
		   'MIN)
	    (loop with obj-name = (first (opsat-objective-function-name obj))
		for solution in solutions do
		  (let ((cost-map (opsat-solution-objective-fns-costs-ht solution)))
			(setf (gethash obj-name cost-map)
			  (- 0 (gethash obj-name cost-map))))))
		    

	; If the composition operator is product, do log on each value.
	(when (equal (opsat-objective-function-composition-operator obj)
		   '#'*)
	    (loop with obj-name = (first (opsat-objective-function-name obj))
		for solution in solutions do
		  (let ((cost-map (opsat-solution-objective-fns-costs-ht solution)))
			(setf (gethash obj-name cost-map)
			  (exp (gethash obj-name cost-map))))))

      finally (return solutions)))



 (defmethod find-leading-solutions-for-mof-input-problem (mof-file &optional
								   (cutoff-delta nil)
								   (maximum-solution-count nil)
								   (find-all-consistent-models? T))
   "Reads in the problem described in mof-file, creates an opsat-problem for that description
   and finds leading assignments that satisfy the given hard-constraints and have best cost.

   RETURNS the xml-file containing solution descriptions"

   
   (let* ((opsat-model (opsat-load mof-file))
	  (opsat-search-problem (new-vcsp-problem-from-opsat-model opsat-model))
		  (solutions
		   (find-solutions-w-costs opsat-search-problem
								   :cutoff-delta cutoff-delta
								   :maximum-solution-count maximum-solution-count
								   :find-all-consistent-models? find-all-consistent-models?)))
     (format t "~%~%Solutions:~%~%")
     (loop for solution in solutions do
	   (format t "    ~A~%" (opsat-solution-decision-assignments solution))
	   (maphash #'(lambda (k v)
			(format t "    ~A = ~A" k v))
		    (opsat-solution-objective-fns-costs-ht solution))
	   (format t "~%~%")
     )
     (write-xml (concatenate 'string (string-right-trim ".mof" mof-file) ".xml")
			    (parse-opsat-solutions solutions))))



 (defmethod new-vcsp-problem-from-opsat-model ((opsat-model opsat-model))
   "Creates and returns a conflict-directed-opsat-vcsp from the opsat-model"

   (multiple-value-bind (problem-name
						 decision-variable-descriptions
						 state-variable-descriptions
						 state-clauses
						 soft-cons-system
						 cost-order-function)
       (extract-opsat-model-info opsat-model)
     (new-vcsp-problem
	  decision-variable-descriptions
	  state-variable-descriptions
	  state-clauses
	  :source problem-name
	  :name problem-name
	  :cost-order-function cost-order-function
	  :soft-constraint-system soft-cons-system)))


 (defmethod extract-opsat-model-info ((opsat-model opsat-model))
   "Returns descriptions of all fields required to create a vcsp search problem"

   (let ((decision-variables '())
	 (state-variables '())
	 (state-clauses '())
	 (soft-cons-system '())
	 (domain-ht (opsat-model-domain-ht opsat-model)))

			 ; ((variable-ht (opsat-model-decision-variables opsat-model))
			 ; 	(domain-ht (opsat-model-domain-ht opsat-model))
			 ; 	(hard-constraints (opsat-model-hard-constraints opsat-model))
			 ; 	(objective-fns (opsat-model-objective-function-system opsat-model)))

	 ;; format the decision-variables and state-variables descriptions
     (do-ht (variable-name variable (opsat-model-variable-ht opsat-model))
       (if (opsat-model-variable-decision-variable? variable)
	   (setf decision-variables (push `(,variable-name :domain
							   ,(format-domain-for-opsat
							     (gethash (opsat-model-variable-domain-name
								       variable)
								      domain-ht)))
					  decision-variables)))
       (setf state-variables (push `(,variable-name :domain
						    ,(gethash (opsat-model-variable-domain-name
							       variable)
							      domain-ht))
				   state-variables)))
     
     (loop for clause in (opsat-model-hard-constraints opsat-model)
	 do (setf state-clauses 
	      (push `(:clause ,clause)
		    state-clauses)))
     
     (loop for soft-constraint in (opsat-objective-function-soft-constraint-system
				   (first (opsat-model-objective-function-system
					   opsat-model)))
	 do (setf soft-cons-system
	      (push (new-soft-constraint-list
		     (first soft-constraint)
		     (second soft-constraint)
		     #'>= )
		    soft-cons-system)))


     (values
      (opsat-model-name opsat-model)
      decision-variables
      state-variables
      state-clauses
      soft-cons-system
      #'>=)))


 (defmethod format-domain-for-opsat (domain-description)

   (let ((formatted-domain '()))
     (loop for value in domain-description
		 do (format t "~% value ~a" value)
		 do  (setf formatted-domain 
			   (push `(,value :cost 0) formatted-domain)))

     formatted-domain))


  (defmethod parse-opsat-solutions (solutions)

    (loop for opsat-solution in solutions
	with all-models
	with metrics = nil
	     
	do (loop for model in (first (opsat-solution-sets-of-consistent-models opsat-solution))
	       do (maphash #'(lambda (metric-name metric-value)
			       (push (list (format nil "~A" metric-name) (format nil "~A" metric-value)) metrics))
			   (opsat-solution-objective-fns-costs-ht opsat-solution))
		  (setf all-models (cons
				    (list metrics
					  (loop for true-proposition in (isat::theory-model-true-propositions model)
					      collect (let* ((assignment-description (isat:proposition-description (first true-proposition)))
							     (var-and-value (rest assignment-description)))
							(list	
					;car here is due mpl naming 
							 (format nil "~a"   (car (first var-and-value)))
							 (format nil "~a"  (second var-and-value))))))
				    all-models)))
	   (setf metrics nil)
	finally (return all-models)))

  (defmethod parse-solutions-to-string (solutions)

    (loop for opsat-solution in solutions
	with all-models
	with metrics = nil
	with sol-div-mark = '<S>  ; The mark that seperates the solutions
	with metric-assignment-div-mark = '<AS> ; The mark that seperates the metrices and assignments
	with metric-div-mark = '<M> ; The mark that seperates the metrics
	with assignment-div-mark = '<A> ; The mark that seperates the metrics
	with div-mark = '<V>        ; The mark that seperates the name and the value
	     
	do (loop for model in (first (opsat-solution-sets-of-consistent-models opsat-solution))
	       do (maphash #'(lambda (metric-name metric-value)
			       (push (list metric-div-mark (format nil "~A" metric-name) div-mark (format nil "~A" metric-value)) metrics))
			   (opsat-solution-objective-fns-costs-ht opsat-solution))
		  (setf all-models (cons
				    (list sol-div-mark
				          metrics
					  metric-assignment-div-mark					  
					  (loop for true-proposition in (isat::theory-model-true-propositions model)
					      collect (let* ((assignment-description (isat:proposition-description (first true-proposition)))
							     (var-and-value (rest assignment-description)))
							(list	
							 assignment-div-mark
					;car here is due mpl naming 
							 (format nil "~a"   (car (first var-and-value)))
							 div-mark
							 (format nil "~a"  (second var-and-value))))))
				    all-models)))
	   (setf metrics nil)
	finally (return (format nil "~A" all-models))))


;;;(defmethod parse-solutions-to-string (solutions)
;;;  
;;;  "Parse the solutions into a string which Java function in OPN can process."
;;;  
;;;  (loop 
;;;      with sol-string = ""
;;;      with sol-div-mark = " <SOLUTION> "   ; The mark that seperates the solutions
;;;      with metric-assignment-div-mark = " <ASSIGNMENTS> " ; The mark that seperates the metrices and assignments
;;;      with metric-div-mark = " <METRIC> "  ; The mark that seperates the metrics
;;;      with assignment-div-mark = " <ASSIGNMENT> "  ; The mark that seperates the metrics
;;;      with div-mark = " <VALUE> "  ; The mark that seperates the name and the value
;;;			
;;;      for solution in solutions do
;;;	(let ((metrics (first solution))
;;;	      (assignments (second solution)))
;;;
;;;	  (setf sol-string (concatenate 'string sol-string sol-div-mark))
;;;	  
;;;	  (loop for metric in metrics do
;;;		(setf sol-string (concatenate 'string sol-string metric-div-mark (first metric) div-mark (second metric))))
;;;	  
;;;	  (setf sol-string (concatenate 'string sol-string metric-assignment-div-mark))
;;;	  
;;;	  (loop for assignment in assignments do
;;;		(setf sol-string (concatenate 'string sol-string assignment-div-mark (first assignment) div-mark (second assignment)))))
;;;
;;;	  finally (return sol-string)))



 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;;;   Write XML files
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun write-xml(filename solution)
  (unless (null (probe-file filename))
    (delete-file filename))
  (let ((out (open filename :direction :output :if-does-not-exist :create)))
    (write-line "<?xml version = \"1.0\" encoding = \"ISO-8859-1\"?>" out)
    (do ((tokens solution (setq tokens (cdr tokens))))
	((eql tokens nil))
      (write-line "<token>" out)
      (write-line "  <dvars>" out)
      (write-dvars out (car (cdr (car tokens))))
      (write-line "  </dvars>" out)
      (write-line "  <metrics>" out)
      (write-metrics out (car (car tokens)))
      (write-line "  </metrics>" out)
      (write-line "</token>" out))
    (close out))

  filename)

(defun write-dvars(out dvars)
  (do ((pair dvars (setq pair (cdr pair))))
      ((eql pair nil))
    (write-line "    <dvar>" out)
    (write-string "      <name>" out)
    (write-string (car (car pair)) out)
    (write-line "</name>" out)
    (write-string "      <value>" out)
    (write-string (car (cdr (car pair))) out)
    (write-line "</value>" out)
    (write-line "    </dvar>" out)))

(defun write-metrics(out metrics)
  (do ((pair metrics (setq pair (cdr pair))))
      ((eql pair nil))
    (write-line "    <metric>" out)
    (write-string "      <name>" out)
    (write-string (write-to-string (car (car pair))) out)
    (write-line "</name>" out)
    (write-string "      <value>" out)
    (write-string (write-to-string (car (cdr (car pair)))) out)
    (write-line "</value>" out)
    (write-line "    </metric>" out)))

