;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

(in-package :OPSAT)

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities


;;; The code in this file is suitable for reading *.mof files modeling VCSP problem
;;; A Valued Constraint Satisfaction Problem representation in  MPL is as follows:
;;;
;;;
;;; Notice: It should be evident that this code will not work for mode estimation.


(defmacro get-opsat-model-variable (name opsat-model)

  "Maps variable NAME to an existing internal variable structure"

  `(gethash ,name (opsat-model-variable-ht ,opsat-model)))

(defmacro get-opsat-model-domain (name opsat-model)

  "Maps domain NAME to an existing internal variable structure"

  `(gethash ,name (opsat-model-domain-ht ,opsat-model)))


(defmethod OPSAT-load (&optional file-name)

  " Reads the model descriptions of the ops file and stores in
   a new opsat model (problem)  object as variables, constraints and objects.
   Returns the new opsat problem object."

  ;; Get file-name if not provided.
  (when (null file-name)
    (setq file-name (request-input-file-name)))

  ;; Open file as (reversible-read) stream to be read.
  ;; A reversive-read stream allows tokens to be read
  ;; from the stream and then pushed back on, when not
  ;; what was expected.
  (unless (null file-name)
    (let ((path-name (pathname file-name)
		     ;;	  The following doesn't produce the desired result 
		     ;;   of defaulting extension to .mof
		     ;;	 (if (pathnamep file-name)
		     ;;		       file-name
		     ;;	   (format nil "~A.mof" file-name))
		     ))
      (with-open-file (file-stream path-name 
					   :direction :input :if-exists :supersede)
	(let ((stream (create-reversible-read-stream file-stream)))
	  
	  ;; Read in the model name
	  (when *opsat-debug*
			(format t "~% Starting read of model."))

	  (let* ((opsat-model-name (OPSAT-read-opsat-model-name 
								 stream))
			 (opsat-model (make-instance 'opsat-model
					:name opsat-model-name)))
	    (when *opsat-debug*
	      (setq *opsat-debug-model* opsat-model))
	    
		  ;; Read in the variable (enumerated) type definitions.
	    (OPSAT-read-domain-types stream  opsat-model)

	    ; ignore multiple instances ; ignore instance name
	     (OPSAT-read-instance-variables stream opsat-model)
	    
		;; Reading instance constraint definitions
	    (OPSAT-read-hard-constraints stream opsat-model)


	    ;;if there is TRANSITION DEFINITIONS it should stand by itself
	    ;; with no transitions definitions at all.

	    (strip-symbol-sequence-if-next stream '(TRANSITION DEFINITIONS)) 

		;; Reading instance objective functions
	    (OPSAT-read-objective-functions-and-flag-decision-variables! stream opsat-model)

		;; Returning opsat model
	    (when *opsat-debug*
	      (format t "~% Finished model."))
	    opsat-model))))))

;;;   ************************************************************
;;;   **
;;;   **  Support Functions
;;;   **
;;;   ************************************************************

(defmethod OPSAT-read-opsat-model-name (stream)

  "Read a system model name from reversible STREAM if NAME-PROVIDED? 
   is true.  Otherwise, generates a symbol denoting name.
   Returns new name."

  ;;; Note: OPSAT name is put in the current package *package*.

  (when *OPSAT-debug*
	(format t "~% Looking for opsat model name"))
  (let* ((name-provided? (eql (strip-symbol-sequence-if-next 
							   stream '(SYSTEM MODEL NAMED))
							  :found))
		 (opsat-model-name (if name-provided?
								(reversible-read stream)
							  (intern (gensym "OPSAT-MODEL-")
										; :cl-user
									  ))))
	(when *opsat-debug*
	  (if name-provided?
		  (format t "~% model = ~A" opsat-model-name)
		(format t "~% No model name found")))
	opsat-model-name))



(defmethod OPSAT-read-domain-types (stream
				    (opsat-model opsat-model))

  " Process zero or more instances of:
       VARIABLE-DOMAIN-TYPE <domain-name> (<member-names>*)"

  (when *opsat-debug*
	(format t "~% Looking for variable domain type definition section"))
  (strip-symbol-sequence-if-next 
   stream '(VARIABLE DOMAIN TYPE DEFINITIONS) :signal-error)
  (when *opsat-debug*
	(format t "~% Found."))
  (loop while (eql (strip-symbol-if-next stream 'VARIABLE-DOMAIN-TYPE) :found)
      for domain-name = (reversible-read stream)
      for members = (reversible-read stream)
      do (create-variable-domain-type domain-name members opsat-model)))

;;; Note
;;; o There should not exist more than one instance if we are modeling a VCSP.
;;; For VCSP modeling, one should define a proble as a component.

(defmethod OPSAT-read-instance-variables (stream (opsat-model opsat-model))

  "Process zero or mode instances of
        <VARIABLE <variable-name> TYPE <[DEPENDENT]>
		DOMAIN-TYPE <domain-name>>*"

  (when *opsat-debug*
	(format t "~% Looking for  variable definitions section"))
  (strip-symbol-sequence-if-next 
   stream '(VARIABLE DEFINITIONS) :signal-error)
  (when *opsat-debug*
	(format t "~% Found."))
  ;(loop while
  (when (eql (strip-symbol-if-next stream 'INSTANCE ) :found)
    (reversible-read stream) ; get rid of instance name
    
    (loop while (eql (strip-symbol-if-next stream 'VARIABLE) :found)
	for var-name = (reversible-read stream)
	for category = (progn (strip-symbol-if-next 
			       stream 'CATEGORY :signal-error)
			      (reversible-read stream))
	for domain-name = (progn (strip-symbol-if-next 
				  stream 'DOMAIN-TYPE :signal-error)
				 (reversible-read stream))
	for value = (if (eql (strip-symbol-if-next stream 'VALUE)
			     :found)
			(reversible-read stream)
		      nil)
			 
	do (create-opsat-model-variable! var-name domain-name opsat-model
					 value category)))
  (when *opsat-debug*
    (format t  "~% Finished section."))
  t)


(defmethod OPSAT-read-hard-constraints (stream
					(opsat-model opsat-model))
  "Process a formula in CNF"

  (when *opsat-debug*
	(format t "~% Looking for HARD CONSTRAINT  definitions section" ))
  (strip-symbol-sequence-if-next 
   stream '(STATE CONSTRAINT DEFINITIONS) :signal-error)
  (when *opsat-debug*
	(format t "~% Found."))
  (when (eql (strip-symbol-if-next stream 'INSTANCE) :found)
    (reversible-read stream) ;ger rid of instance name
    (let ((hard-constraints-def (progn (strip-symbol-if-next stream 'STATIC-CONSTRAINT :signal-error)
				       (reversible-read stream))))
      (create-instance-hard-constraints hard-constraints-def opsat-model))))
    
    
(defmethod OPSAT-read-objective-functions-and-flag-decision-variables! (stream
					   (opsat-model opsat-model))
  "Process zero or more instances of OBJECTVE-FUNCTION
	  INSTANCE <OBJECTIVE-FN-NAME>
		<OBJECTIVE-FN-PREFERENCE>
		<OBJECTIVE-FN-COMPOSITION-OP>
		<SOFT-CONSTRAINT
		   SCOPE <SOFT-CONS-SCOPE>
		   TUPLE-LIST <(TUPLE COST)>*>*"

  (when *opsat-debug*
	(format t "~% Looking for UTILITY FUNCTION DEFINITIONS section"))
  (strip-symbol-sequence-if-next
   stream '(UTILITY FUNCTION DEFINITIONS) :signal-error)
  (when *opsat-debug*
    (format t "~% Found."))

  (let ((objective-fn-system '()))
    (when (eq (strip-symbol-if-next stream 'INSTANCE) :found)
      (reversible-read stream)		; get rid of instance name
      (loop while (eql (strip-symbol-if-next stream 'CRITERIA) :found)
	  for criteria-name = (reversible-read stream)
	  for preference = (progn (strip-symbol-if-next
				   stream 'PREFERENCE :signal-error)
				  (reversible-read stream))
	  for composition-operator = (progn (strip-symbol-if-next
					     stream 'COMPOSITION-OP
					     :signal-error)
					    (reversible-read stream))
	  do (setf objective-fn-system
	       (append objective-fn-system
		       (list
			(make-instance 'opsat-objective-function
			  :name criteria-name
			  :cost-order preference
			  :composition-operator composition-operator
			  :soft-constraints (get-soft-constraint-system-and-flag-decision-variables!
					     stream criteria-name opsat-model)))))))

    (setf (opsat-model-objective-function-system opsat-model)
      objective-fn-system)

    (when *opsat-debug*
	(format t "~% Done with UTILITY FUNCTION DEFINITIONS "))))
  


(defmethod get-soft-constraint-system-and-flag-decision-variables!
    (stream instance-name (opsat-model opsat-model))
  "Read and creat soft-constraints"


  (when *opsat-debug*
    (format t "~%  Reading SOFT-CONSTRAINTS for ~a OBJECTIVE"
				instance-name))

  (loop while (eql (strip-symbol-if-next stream 'UTILITY-RELATION) :found)
	      for scope = (progn (strip-symbol-if-next
				       stream 'SCOPE :signal-error)
				      (reversible-read stream))
	      for tuple-list = (progn (strip-symbol-if-next
				       stream 'VALUED-TUPLES :signal-error)
				      (reversible-read stream))
			 ;; i don't like to place this side effect here
			 ;; (this is what '-and-flag-decision-variables!' refers to
			 ;; to do: make this neater.
		      do ( flag-decision-variables-within-scope! scope opsat-model)
		      collect (list scope tuple-list)))
    
  

(defmethod flag-decision-variables-within-scope! (scope (opsat-model opsat-model))

  "Turn on flags for variables within scope. This should indicate these are decision variables."

  (loop for variable in scope
			do (setf (opsat-model-variable-decision-variable?
				  (get-opsat-model-variable variable opsat-model))
			     T)))

;;;; to be moved to a separate file (create-opsat-model.lisp)

(defmethod create-instance-hard-constraints
    (hard-constraints (opsat-model opsat-model))

  "Records hard-constraints inside opsat-model"
  
  ; (unless (legal-cnf-state-formula? static-constraint opsat-model)
  ; (warn "Static constraint for ~A is not in conjunctive normal form:
  ; ~A."  instance-name static-constraint))
  (setf (opsat-model-hard-constraints opsat-model)
    (internalize-cnf-formula hard-constraints opsat-model)))

(defmethod internalize-cnf-formula
    (cnf-state-formula (opsat-model opsat-model))

  "Returns T if CNF-FORMULA was successful interned,
   otherwise return nil."

  (let ((state-clauses
		 (cond ((tautology? cnf-state-formula)
				'())
			   ((conjunction? cnf-state-formula)
				(expression-operands cnf-state-formula))
			   (:otherwise
			    (error "~a is not a valid cnf formula of ~a."
				   cnf-state-formula opsat-model)))))
   ; state-clauses
					 (loop for clause in state-clauses
 			   ; do (format t "~% CLAUSE : ~a" clause)
 	    collecting
	    (expression-operands clause)
					; (internalize-state-clause clause 
; ; 											 system-model
; ; 											 (or part-of
; ; 												 cnf-state-formula))
 			   )
	))

; (defmethod internalize-state-clause
;     (state-clause  (opsat-model opsat-model))

;   "Internalize clause from hard-constraint"

;   (let ((assignment-literals
; 		 (cond ((disjunction? state-clause)
; 				(expression-operands state-clause))
; 			   ((assignment-literal? state-clause)
; 				(list state-clause))
; 			   (:otherwise
; 				(if part-of
; 					(error "~a is not a valid clause in ~a of ~a."
; 						   state-clause part-of opsat-model)
; 				  (error "~a is not a valid clause of ~a."
; 						 state-clause opsat-model))))))
; 	(internalize-assignment-literals assignment-literals 
; 									 opsat-model
; 									 (or part-of
; 										 state-clause))))



  
(defmethod create-variable-domain-type 
    (domain-name members (opsat-model opsat-model))
  
  " Creates a variable domain type called DOMAIN-NAME for OPSAT MODEL,
    with MEMBERS as mutually exclusive, collectivelely exhaustive states"
  (when *opsat-debug*
    (format t "~% Creating domain ~a with values ~a" domain-name members))
  (setf (gethash domain-name (opsat-model-domain-ht opsat-model))
    members))

;;; Note:
;;;	o Currently all soft-constraints (including unary soft-constraints)
;;;  are represented separately. Since this was not true before, we still
;;;  need to pass :keywords 'cost' for decision-variables.
;;;  e.g.: (a :domain ((hi :cost 2) (low :cost 3)))

(defmethod create-opsat-model-variable!
    (variable-name domain-name (opsat-model opsat-model)
     &optional (value nil) (category :DEPENDENT) (instance-name nil))
  
  "Creates a  decison variable called VARIABLE-NAME w/ DOMAIN
   for OPSAT MODEL."

  (declare (ignore value category instance-name))
  
  (when *opsat-debug*
    (format t "~% Creating Variable ~a with domain ~a" variable-name domain-name))
  (unless (null (get-opsat-model-variable variable-name opsat-model))
    (warn "variable with name ~S already exists: ~S, replacing it."
	  variable-name (get-opsat-model-variable variable-name opsat-model)))
  (unless (get-opsat-model-domain domain-name opsat-model)
    (warn "Domain ~a is not a defined domain type."
	  domain-name))
  
  (setf (get-opsat-model-variable variable-name opsat-model)
    (make-instance 'opsat-model-variable :name variable-name :domain-name domain-name)))
		
  
(defmethod create-decision-variable
    (variable-name domain-name (opsat-model opsat-model))
  
  "Creates and insert a  decison variable called VARIABLE-NAME w/ DOMAIN
   for OPSAT MODEL."
  
  ;(setf (opsat-model-decision-variables opsat-model)
    (push (list variable-name domain-name)
	  (opsat-model-decision-variables opsat-model));)
  )

(defmethod create-state-variable
    (variable-name domain-name (opsat-model opsat-model))
  
  "Creates a state variable called VARIABLE-NAME w/ DOMAIN
   for OPSAT MODEL."

  ;(setf (opsat-model-state-variables opsat-model)
    (push (list variable-name domain-name)
	  (opsat-model-state-variables opsat-model));)
 )
   




  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;   
;;      Utilities for OPSAT-READER
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Note:
;;;   o After going through the CCA-to-OPSAT manipulations functions
;;;     within CCA-model/create-cca-model.lisp it was not exactly clear
;;;     the need to use a internal representation for CCA.
;;;
;;;  "Given a list of assignment literals, where each literal is an assignment
;;;    or its negation, and an assignment is:
;;;     (= variable-name value)
;;;    returns an equivalent list of literals with each assignment replaced
;;;    by an assignment structure."

;;; this macro is just a get-arround for the current implementation

; (defmacro convert-decision-domain-to-OPSAT-format




; (defmethod OPSAT-read-decision-variables (stream
; 				    (opsat-model opsat-model))
  
;   "Process zero or more instances of:
; 	INSTANCE <decision-variable-name>
; 	   DOMAIN-TYPE <domain-name>"
;   (when *opsat-debug*
; 	(format t "~% Looking for decision variable definitions section"))
;   (strip-symbol-sequence-if-next 
;    stream '(DECISION VARIABLE DEFINITIONS) :signal-error)
;   (when *opsat-debug*
; 	(format t "~% Found."))
;   (loop while (eql (strip-symbol-if-next stream 'INSTANCE) :found)
;       for variable-name = (reversible-read stream)
;       for domain-name = (progn (reversible-read stream)
; 			       (reversible-read stream))
;       do (create-decision-variable variable-name domain-name opsat-model)))
  


; (defmethod OPSAT-read-variables (variable-tag stream
; 			       (opsat-model opsat-model))
  
;   "Process zero or more instances of:
; 	INSTANCE <decision-variable-name>
; 	   DOMAIN-TYPE <domain-name>"
  
;   (when *opsat-debug*
; 	(format t "~% Looking for ~a  definitions section" variable-tag))
;   (strip-symbol-sequence-if-next 
;    stream (cons variable-tag '(VARIABLE DEFINITIONS)) :signal-error)
;   (when *opsat-debug*
; 	(format t "~% Found."))
;   (loop while (eql (strip-symbol-if-next stream 'INSTANCE) :found)
;       for variable-name = (reversible-read stream)
;       for domain-name = (progn (reversible-read stream)
; 			       (reversible-read stream))
;       do (format t "~% Variable ~a read" domain-name)			
;       do (if (equal variable-tag 'decision)
; 	     (create-decision-variable variable-name domain-name opsat-model)
; 	     (create-state-variable variable-name domain-name opsat-model))))

