;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

 (in-package :OPSAT)

;;; Copyright (c) 1999 - 2009 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities

;;; This file provides the methods to provide solutions in a controllable
;;; region close to the Pareto front for multi-objective optimization problems.

;;; This file provides methods to find the Fuzzy Convex Pareto Front of
;;; a *Multi-Objective Valued Constraint Satisfaction Problem (MO-VCSP)*
;;; that is represented through an OpSat-Model.

;; To Do:
;; o Extend from Pareto-front only to a controllable region close to the
;;   Pareto front.
;; o Extend from 2-objective to multiple objectives.
;; o To accomadate multiplication, change #'* to #'+ by applying log.
;; o Make conflicts reusable.
;; o Analyze the soft constraint tuples??

;;; !!!! JUST CONVEX PARETO FOR NOW !!!!

;;; To Do:
;;; o Implement cutoff-delta-set. Right now we just return the Convex Pareto Points.


(defmethod find-fuzzy-convex-pareto-front ((opsat-model opsat-model)
										   &optional (cutoff-delta-set nil))
  
  "Takes an OPSAT MODEL as input. An OpSAT Model is a representation of 
    a Multi-Objective Valued-CSP (MO-VCSP). 

   RETURNS the Convex Pareto Solutions of the MO-VCSP and, optionally,
    solutions that are tolerable (i.e. a solution that, for each objective 
    is no farther than
    CUTOFF-DELTA-SET from the pareto front.
    We call this set of points *Fuzzy Convex Pareto Front*.

   CUTOFF-DELTA-SET defines the tolerance cutoff on each objective.
    CUTOFF-DELTA-SET defaults to nil, i.e., only the Convex Pareto Front is 
    returned. If a CUTOFF-DELTA-SET is provided, solutions near the Convex 
    Pareto Front will also be returned. "

  ;; optima generated through optimization of each objective individually
  (let* ((obj-fns (opsat-model-objective-function-system opsat-model))
		 (obj-name-vector (get-objective-name-vector obj-fns))
		 (obj-soft-constraint-list (get-objective-soft-constraint-list obj-fns))

		 (individual-optima-and-region (single-objective-solve
										opsat-model cutoff-delta-set))
		 (individual-optima (first individual-optima-and-region))
		 (individual-region (second individual-optima-and-region))

		 (cost-map (build-cost-map obj-soft-constraint-list)))

    
	;; for each solution previously found, evaluate remaining objectives
    (eval-all-objective-value-for-set! individual-region
									   obj-name-vector
									   obj-soft-constraint-list
									   cost-map)
    
    (format t "CUTOFF: ~A~%" cutoff-delta-set)
    
	;;     (format t "INITIAL BASE SOLUTIONS:~%")
	;;     (loop for base in individual-optima do
	;; 	  (maphash #'(lambda (k v)
	;; 		       (format t "   ~A = ~A~%" k v))
	;; 		   (opsat-solution-objective-fns-costs-ht base))
	;; 	  (format t "~%"))
	;;     (format t "~%")

    (append individual-region
			(recursive-knee-calculation individual-optima
										opsat-model
										obj-name-vector
										obj-soft-constraint-list
										cost-map
										cutoff-delta-set
										0)))) ; Recursively find convex points.
 

(defmethod recursive-knee-calculation (knee-base-solutions
									   (opsat-model opsat-model)
									   obj-name-vector
									   obj-soft-constraint-list
									   cost-map
									   cutoff-delta-set
									   level)
  
  " Takes KNEE-BASE-SOLUTIONS as input. Knee-Base-Solutions are the two ends of 
    a stretch of the convex front that will be further refined. 

    Such base solutions will determine the new direction over which the 
    optimization will continue. Given the new direction, a linear combination of 
    the original objective functions is built, yielding a new objective function 
    to be used.

   RETURNS the knee for the given base solutions and recursively calls itself 
   to generate knees for the stretches to be further refined."
  
  ;; new-direction represented through coefficients
  
  (let* ((new-direction (coefficients-of-normalized-knee-vector knee-base-solutions
																obj-name-vector))
		 (new-objective-fn (if (null new-direction)
							   nil
							 (knee-objective-fn new-direction
												obj-soft-constraint-list)))
		 (top-solutions nil)
		 (knee-solution nil))
    
    (if (infeasible-direction? new-direction)
		
		nil
      
	  ;; Otherwise:
      (progn 
		
		;; update opsat-model with the obj fn we've just derived
        (update-opsat-model-cost-system! new-objective-fn opsat-model)
	
	
		(setf top-solutions
		  (find-solutions-w-costs (new-vcsp-problem-from-opsat-model opsat-model)
								  :cutoff-delta 0
								  :find-all-consistent-models? t))
	
		(setf knee-solution (the-top-sln top-solutions))
		
		(eval-all-objective-value-for-set! top-solutions
										   obj-name-vector
										   obj-soft-constraint-list
										   cost-map)
		
		;; The points on the base are always valid solutions!
		(assert (not (null knee-solution)) nil "Knee-solution should not be null")
	
		;; 	(format t "## New Direction ##:   ~A~%" new-direction)
		;; 	(loop for fn in new-objective-fn do
		;; 	      (format t "   ~A~%~%" fn))
		;; 	(format t "~%LEVEL:  ~A   BASE SOLUTIONS:~%" level)
		;; 	(loop for base in knee-base-solutions do
		;; 	      (maphash #'(lambda (k v)
		;; 			   (format t "   ~A = ~A~%" k v))
		;; 		       (opsat-solution-objective-fns-costs-ht base))
		;; 	      (format t "~%"))

		;; 	(format t "KNEE SOLUTION:~%")
		;; 	(maphash #'(lambda (k v)
		;; 		     (format t "   ~A = ~A~%" k v))
		;; 		 (opsat-solution-objective-fns-costs-ht knee-solution))
		;; 	(format t "~%")
	
	
		(if (and (knee-better-than-its-base-solutions?
				  knee-solution knee-base-solutions new-direction obj-name-vector)
				 (< level 5))
			
			(loop with new-base-solutions = nil
											
				for knee-base-solution in knee-base-solutions 
				do (setf new-base-solutions 
					 (remove knee-base-solution knee-base-solutions))

				   (setf new-base-solutions 
					 (cons knee-solution new-base-solutions))
		  
				   (setf top-solutions 
					 (cons knee-solution
						   (recursive-knee-calculation new-base-solutions
													   opsat-model
													   obj-name-vector
													   obj-soft-constraint-list
													   cost-map
													   cutoff-delta-set
													   (+ level 1))))

				finally (return top-solutions))
		  
		  (if (null cutoff-delta-set)
			  
			  (filter-base-solutions knee-base-solutions
									 top-solutions
									 obj-name-vector
									 obj-soft-constraint-list
									 cost-map)
			
			(pull-back-to-cutoff-set opsat-model
									 new-direction
									 knee-base-solutions
									 cutoff-delta-set
									 obj-name-vector
									 obj-soft-constraint-list
									 cost-map)))))))


(defmethod get-objective-soft-constraint-list (obj-fns)

  (loop with soft-constraint-list = '()

	  for obj-fn in obj-fns do
		(setf soft-constraint-list
	      (append soft-constraint-list
				  (list (copy-list
						 (opsat-objective-function-soft-constraint-system
						  obj-fn)))))

	  finally (return soft-constraint-list)))


(defmethod pull-back-to-cutoff-set ((opsat-model opsat-model)
									direction
									base-solutions
									cutoff-delta-set
									obj-name-vector
									obj-soft-constraint-list
									cost-map)
  
  "Pull the front line back and find the fuzzy Pareto solutions 
   according to the cutoff-delta-set."

  (loop with cutoff-delta = (dot-multiply direction cutoff-delta-set)
	  with fuzzy-solutions = (find-solutions-w-costs
							  (new-vcsp-problem-from-opsat-model opsat-model)
							  :cutoff-delta cutoff-delta
							  :find-all-consistent-models? t)
	  with trimmed-solutions = '()
	
	  for solution in fuzzy-solutions 
	  do (eval-all-objective-value! solution
									obj-name-vector
									obj-soft-constraint-list
									cost-map)
		 
		 (if (in-the-box-region? solution base-solutions direction obj-name-vector)
			 
			 (push solution trimmed-solutions))

	  finally (return trimmed-solutions)))


(defmethod in-the-box-region? (solution base-solutions direction obj-name-vector)
  
  (let ((linking-vector-set '()))
	
    (loop with solution-obj-vector = (get-obj-vector solution obj-name-vector)
	  
		for base-solution in base-solutions do
		  (push (vector-substract
				 direction
				 (normalize (vector-substract solution-obj-vector
											  (get-obj-vector base-solution
															  obj-name-vector))))
				linking-vector-set))

    (loop with in-the-box? = t

		until (or (null linking-vector-set)
				  (not in-the-box?)) 
		
		do (let ((first-vector (first linking-vector-set)))
			 
			 (setf linking-vector-set (rest linking-vector-set))
			 
			 (loop for vector in linking-vector-set 
				 do (when (> (dot-multiply first-vector vector)
							 -0.0001)
					  (setf in-the-box? nil)
					  (return nil))))
	  
		finally (return in-the-box?))))


;;; Right now this function takes care of checking whether the solution is worse 
;;; than the base ones. We should ''fix'' opsat to get rid of this test.
;;; Fixed -Maokai

(defmethod filter-base-solutions (slns-to-be-removed
								  set-of-solutions
								  obj-name-vector
								  obj-soft-constraint-list
								  cost-map)
  
  "Filtering solutions that are worse than the base or are base sln themselves
   SET-OF-SOLUTIONS will be DESTRUCTIVELY modified and RETURNED without 
   SLNS-TO-BE-REMOVED."
  
  (let ((trimmed-solutions '()))
	
    (loop for potential-solution in set-of-solutions 
		until (null set-of-solutions)
		do (unless (duplicate-solution? potential-solution slns-to-be-removed)
			 (eval-all-objective-value! potential-solution
										obj-name-vector
										obj-soft-constraint-list
										cost-map)
			 (push potential-solution trimmed-solutions)))
	
    trimmed-solutions))


;;; Fix:
;;; o We should not need to provide the names of the obj fns as 
;;;   arguments for this method.
(defmethod coefficients-of-normalized-knee-vector 
	(knee-base-solutions objective-names)
  
  "Takes as input a set of KNEE-BASE-SOLUTIONS and 
   RETURNS the coefficients of the normalized vector that is orthogonal to 
   the hyperplane uniquely determined by knee-base-solutions. 
   (The sequence of the vector elements corresponds to the sequence of 
   the OBJECTIVE-NAMES.) 
   That vector is called normalized-knee-vector."

  (let* ((n-of-objs (length objective-names))
		 (matrix (make-array (list (- n-of-objs 1))))
		 (v0 (get-obj-vector (first knee-base-solutions) objective-names))
		 (i 0))
    
    (loop for solution in (rest knee-base-solutions) 
		do (let ((vi (get-obj-vector solution objective-names))
				 (obj-vector (make-array n-of-objs)))
			 
			 (loop for j from 0 to (- n-of-objs 1) 
				 do (setf (aref obj-vector j) (- (aref v0 j) (aref vi j)))
				 finally (setf (aref matrix i) obj-vector)))
		   (incf i)
		finally (return (null-space matrix)))))


(defmethod get-objective-name-vector (obj-fns)

  "Take a list of objectives and return the name vector of the objectives."

  (loop with name-vector = (make-array (length obj-fns))
	  with i = 0
	
	  for obj-fn in obj-fns 
	  do (setf (aref name-vector i)
		   (first (opsat-objective-function-name obj-fn)))
		 (incf i)

	  finally (return name-vector)))


(defmethod knee-objective-fn (new-direction soft-constraints-of-all-objs)
  
  "RETURNS a list of the soft constraints which is a combination of 
   soft constraints from each objective but the values are multiplied 
   by the corresponding factor in NEW-DIRECTION."

  (loop with new-objective = '()
	  with i = 0
	  with scopes = (make-hash-table :test #'equal)

	  for constraints-of-one-obj in soft-constraints-of-all-objs do
		(loop for constraint in constraints-of-one-obj do
			  (let ((tuples (second constraint))
					(coefficient (aref new-direction i))
					(existing-tuples (gethash (first constraint) scopes)))

				(if (null existing-tuples)

					;; If there is no such tuple in the NEW-OBJECTIVE.
					
					(loop with new-tuples = '()
						with new-tuple = nil
			  
						for tuple in tuples do
						  (setf new-tuple (copy-list tuple))
						  (setf (second new-tuple) (* coefficient (second tuple)))
						  (setf new-tuples (append new-tuples (list new-tuple)))
			  
			  
						finally (push (list (first constraint) new-tuples)
									  new-objective)
								(setf (gethash (first constraint) scopes)
								  new-tuples))

				  ;; If there is already such a tuple in the NEW-OBJECTIVE.

				  (loop with new-tuples = '()
					  with working-tuple-list = existing-tuples
					       
					  for tuple in tuples do
						(loop for existing-tuple in working-tuple-list do
							  (when (equal (first tuple) (first existing-tuple))
								(incf (second existing-tuple)
									  (* coefficient (second tuple)))
								(remove existing-tuple working-tuple-list
										:test #'equal)
								(return))
							finally (push tuple new-tuples))
					  finally (setf (gethash (first constraint) scopes)
								(append existing-tuples new-tuples))))))

		(incf i)
	  finally (return new-objective)))


(defmethod knee-better-than-its-base-solutions? (knee-solution
												 base-solutions
												 new-direction
												 obj-name-vector)
										
  
  "Checks whether we indeed got a knee from the last optimization."
  
  (let ((mixed-cost (dot-multiply (get-cost-vector (first base-solutions)
												   obj-name-vector)
								  new-direction))
		(knee-cost (gethash 'metric-name
							(opsat-solution-objective-fns-costs-ht knee-solution))))

    (> knee-cost (+ mixed-cost (* (abs mixed-cost) 0.001)))))


(defmethod update-opsat-model-cost-system! (new-objective 
											(opsat-model opsat-model))
  
  "Updates the Objective-Function-System of OPSAT-MODEL according to 
   NEW-OBJECTIVE. Called for effect."
  
  ;; the solution implemented here seems to rely on the fact that we when solving 
  ;; an MO-VCSP using OPSAT we just use the first obj of obj-fn-system.
  ;; THIS IS NOT ADEQUATE.
  
  (let ((template-objective 
		 (first (opsat-model-objective-function-system opsat-model))))
	
	(setf (opsat-objective-function-soft-constraint-system template-objective)
	  new-objective)))


(defmethod single-objective-solve ((opsat-model opsat-model) cutoff-delta-set)
  
  "Takes an OPSAT-MODEL (i.e. a MO-VCSP) as input and optimizes individually 
   for each objective, not considering trade-offs.

   RETURNS the set of individual optima , i.e. the union of the single 
   best solution considering one objective at a time."
  
										; Solve on each objective
  (loop 
      with objectives = (opsat-model-objective-function-system opsat-model)
      with opsat-search-problem
      with one-obj-solutions
      with cutoff-delta = 0
      with solutions = '()
      with all-solutions = '()
      with i = 0
			  
      for obj in objectives do 
		
		;; Make it single objective
		(setf (opsat-model-objective-function-system opsat-model) (list obj))
		
		;; Parse the opsat-model to opsat-search-problem (Use Gustavo's function)
		(setf opsat-search-problem (new-vcsp-problem-from-opsat-model opsat-model))

		;; If the cutoff delta is for '#*, we need to adjust it to the right value
		(when (and (not (null cutoff-delta-set))
				   (equal (opsat-objective-function-cost-order obj)
						  '#'*))
		  
		  (setf one-obj-solutions
			(find-solutions-w-costs opsat-search-problem
									:cutoff-delta 0
									:find-all-consistent-models? nil))

		  (let* ((cost-map (opsat-solution-objective-fns-costs-ht
							(first one-obj-solutions)))
				 (best-value (gethash 'metric-name cost-map))
				 (required-cutoff (aref cutoff-delta-set i))
				 (adjusted-cutoff (- best-value
									 (log (- (exp best-value)
											 required-cutoff)))))
			(setf (aref cutoff-delta-set i) adjusted-cutoff)))
	
		(unless (null cutoff-delta-set)
		  (setf cutoff-delta (aref cutoff-delta-set i)))
										; Get the optimal solutions
		(setf one-obj-solutions (find-solutions-w-costs 
								 opsat-search-problem
								 :cutoff-delta cutoff-delta
								 :find-all-consistent-models? t))
	
	
		(setf all-solutions (append all-solutions one-obj-solutions))
	
										; All all the solutions into the set
		(setf solutions (append solutions (list (first one-obj-solutions))))

		(incf i)

										; Recover the model and return the solutions
      finally (setf (opsat-model-objective-function-system opsat-model) objectives)
			  (return (list solutions all-solutions))))


(defmethod build-cost-map (obj-soft-constraint-list)

  "Build a cost mapping array for each objective from the possible tuples to 
   their values.  For example, in the following 1-objective soft constraint:
   ((((MPL::A MPL::TOY-INSTANCE)) (((MPL::A-2) 14) ((MPL::A-4) 28) ((MPL::A-7) 49)))
    (((MPL::B MPL::TOY-INSTANCE)) (((MPL::B-3) -15) ((MPL::B-8) -40)))
    (((MPL::C MPL::TOY-INSTANCE)) (((MPL::C-2) -4) ((MPL::C-5) -10) 
      ((MPL::C-6) -12) ((MPL::C-9) -18))))

  The cost mapping will be an array with 1 element, and the mapping will be:
  (((MPL::A) (MPL::A-2)) -> 14)
  (((MPL::A) (MPL::A-4)) -> 28) ... (((MPL::C) (MPL::C-6)) -> -12).

  If there exists tuples, a mapping would look like 
     (((MPL::A MPL::B) (MPL::A-2 MPL::B-3)) -> 32).

  RETURNS an array of such mappings for each objective."
  
  (loop with cost-map = (make-array (length obj-soft-constraint-list))
	  with i = 0

	  for soft-constraints in obj-soft-constraint-list 
	  do (setf (aref cost-map i) (make-hash-table :test #'equal))
	
		 (loop with mapping = (aref cost-map i)

			 for constraint in soft-constraints 
			 do (let ((scope (mapcar #'car (first constraint)))
					  (tuple-values (second constraint)))
				  
				  (loop for tuple-value in tuple-values 
					  do (setf (gethash (list scope (first tuple-value)) mapping)
						   (second tuple-value)))))
		 
		 ;; 	(maphash #'(lambda (k v)
		 ;; 		     (format t "~%    ~A = ~A" k v))
		 ;; 		 (aref cost-map i))
		 
		 (incf i)

	  finally (return cost-map)))


(defmethod eval-all-objective-value-for-set!
	(solutions obj-name-vector obj-soft-constraint-list cost-map)

  "Calculate all objective values for each solution in the 'solutions' set,
  i.e., call the method eval-all-objective-value! repeatedly for each solution in
  the set."

  (loop for solution in solutions do
		(eval-all-objective-value! solution obj-name-vector
								   obj-soft-constraint-list cost-map)))


(defmethod eval-all-objective-value!
	((solution opsat-solution) obj-name-vector obj-soft-constraint-list cost-map)

  "Evaluate all objective value for the solution. Preserve the value in the
   domain 'objective-fns-costs-ht' of the opsat-solution."

  ;; First make a hash map from the variables to their assigned options

  (let ((assignments (opsat-solution-decision-assignments solution))
		(var-to-value-map (make-hash-table)))
	
	;;    (format t "~%~A~%" assignments)
	
    (loop for assign in assignments do
		  (let ((assign-description (isat::proposition-description assign)))
			
			(setf (gethash (first (second assign-description)) var-to-value-map)
			  (first (last assign-description)))))
	
	;;     (maphash #'(lambda (k v)
	;; 		 (format t "~%    ~A = ~A" k v))
	;; 	     var-to-value-map)
	;;     (format t "~%")

	;; Then assess the value of all scopes in each objective function and sum them up
	;; to get the value for that objective function
    (loop with i = 0

		for soft-constraints in obj-soft-constraint-list 
		do (loop with cost = 0
			   with mapping = (aref cost-map i)

			   for constraint in soft-constraints 
			   do (let* ((scope (mapcar #'car (first constraint)))
						 (assigned-tuple (mapcar #'(lambda (var)
													 (gethash var var-to-value-map))
												 scope))
						 (scope-cost (gethash (list scope assigned-tuple) mapping)))
					
					(unless (null scope-cost)

					  ;; (format t "Tuple: ~A    Cost: ~A~%" 
					  ;;     scope assigned-tuple scope-cost)
					  
					  (incf cost scope-cost)))
	  
				  ;; Finally preserve the values of each objective in the solution
			   finally (setf (gethash (aref obj-name-vector i)
									  (opsat-solution-objective-fns-costs-ht solution))
						 cost))
		   (incf i))))


(defmethod get-obj-vector ((solution opsat-solution) names)

  "Get the objective vector [o1 o2 ... on].'"

  (let* ((len (length names))
		 (vector (make-array len))
		 (costs (opsat-solution-objective-fns-costs-ht solution)))
	
    (loop for i from 0 to (- len 1) 
		do (setf (aref vector i)
			 (gethash (aref names i) costs))
		finally (return vector))))


(defmethod get-cost-vector ((solution opsat-solution) names)

  "Get the vector of costs according to the sequance of the objective names.

   RETURNS the cost vector of the objectives."

  (let* ((len (- (length names) 1))
		 (costs (opsat-solution-objective-fns-costs-ht solution))
		 (cost-vector (make-array (length names))))
	
    (loop for i from 0 to len 
		do (setf (aref cost-vector i)
			 (gethash (aref names i) costs))
		   
		finally (return cost-vector))))


(defun null-space (matrix)

  "Get the normalized null vector of the matrix, using Gaussian elimination. The 
   matrix should be n-1 * n, where n is the number of objectives.

   RETURNS the normalized null vector of the matrix."
  
  (let* ((len (length matrix))
		 (order (make-array len))		; The altered order of elements
		 (null-vector (make-array (+ len 1))) ; The vector to return
		 (op-matrix matrix)				; The operated matrix
		 (last-index (- len 1)))		; The index of the last row
										; Initiate the position record
    (loop for i from 0 to last-index do
		  (setf (aref order i) i))
    
	;; Execute the Gaussian elimination
    (loop 
		with j
		with swap
		for i from 0 to last-index do
		  ;; Find the first row in op-matrix whose i'th column is non-zero
		  (setf j i)
		  (loop until (or (>= j len)
						  (> (aref (aref op-matrix j) i)
							 1.0e-10)
						  (< (aref (aref op-matrix j) i)
							 -1.0e-10))
			  do
				(incf j))
	  
										; If no non-zero row for a column, it means the matrix is not invertible
										; Return nil in this case
		  (when (= j len)
			(return nil))
	    
		  ;; Swap the two rows if j doesn't equal i
		  ;; Also swap the position records.
		  (unless (= i j)
			(setf swap (aref op-matrix i))
			(setf (aref op-matrix i) (aref op-matrix j))
			(setf (aref op-matrix j) swap)
	    
			(setf swap (aref order i))
			(setf (aref order i) (aref order j))
			(setf (aref order j) swap))

		  ;; Make the column i of row j 1
		  (let ((column-i (aref op-matrix i)))
			(setf (aref op-matrix i) (divide-vector-by column-i (aref column-i i))))
	  
		  ;; Make the column i of every other rows 0
		  (loop for j from 0 to (- len 1) do
				(unless (= j i)
				  (setf (aref op-matrix j)
					(gaussian-elimination (aref op-matrix j)
										  (aref op-matrix i)
										  i)))))
    
	;; Calculate the real vector and normalize it (just get the last 
	;; element of each row)
	
    (setf (aref null-vector len) -1)
    (loop for i from 0 to last-index do
		  (setf (aref null-vector i) (aref (aref matrix i) len))
		finally (return (normalize null-vector)))))


(defun gaussian-elimination (v1 v2 col)
  
  "Use Gaussian elimination to make the column 'col' of v1 0.

   RETURNS the new column v1."
  
  (let ((n (- 0
			  (aref v1 col))))
	
    (if (= n 0)
		v1
      (vector-sum v1 (divide-vector-by v2 (/ 1 n))))))


(defmethod infeasible-direction? (direction)
  
  (if (null direction)
	  
      t
	
    (loop for i from 0 to (- (length direction) 1) 
		do (when (<= (aref direction i) 1.0e-10)
			 (return t))
		   
		   (return nil))))


;;;;;;;;;
;;;
;;; Operations on Vectors
;;;
;;;;;;;;;


(defun list-to-array (list)
  
  "Takes as input a LIST.
   RETURNs an equivalent array."
  
  (let* ((len (length list))
		 (array (make-array len)))
	
    (loop for i from 0 to (- len 1) 
		do (setf (aref array i) (first list))
		   (setf list (rest list))
		   
		finally (return array))))


(defun array-equal (x y)

  "Let x and y be arrays.
   Returns true iff x and y have identical elements."

  (let ((lenx (length x))
		(leny (length y)))
	
    (if (= lenx leny)
		
		(let ((len (- lenx 1)))
		  
		  (loop for i from 0 to len 
			  do (unless (= (aref x i)
							(aref y i))
				   (return nil))
			  finally (return t)))
      nil)))


(defun dot-multiply (x y)
  
  "Dot multiply between two vectors, an auxiliary function. Notice: the matching of
   the dimensions of the vectors is not checked.
   RETURNS the result of the dot-multiply."

  (let ((len (- (length x) 1))
		(sum 0))
	
    (loop for i from 0 to len 
		do (incf sum (* (aref x i) (aref y i)))
		finally (return sum))))


(defun vector-sum (v1 v2)
  
  "Let v1 and v2 be arrays, denoting vectors.
   Returns the sum of v1 and v2."
  
  (let* ((len (length v1))
		 (op-v (make-array len)))
	
    (loop for i from 0 to (- len 1) 
		do (setf (aref op-v i) (+ (aref v1 i) 
								  (aref v2 i)))
		finally (return op-v))))


(defun vector-substract (v1 v2)
  
  "Let v1 and v2 be arrays, denoting vectors.
   Returns the difference of v1 and v2."
  
  (let* ((len (length v1))
		 (op-v (make-array len)))
	
    (loop for i from 0 to (- len 1) 
		do (setf (aref op-v i) (- (aref v1 i) 
								  (aref v2 i)))
		finally (return op-v))))


(defun divide-vector-by (vector n)
  
  "Let vector be an array.
   Divide every element of vector by n.
   RETURN the corresonding array."
  
  (let* ((len (length vector))
		 (op-v (make-array len)))
	
    (loop for i from 0 to (- len 1) 
		do (setf (aref op-v i)
			 (float (/ (aref vector i)
					   n)))
		finally (return op-v))))


(defun normalize (vector)
  
  "Normalize VECTOR by dividing every element by the vector's sum of squares.

   RETURNS the nomalized vector."
  
  (let ((square-sum (dot-multiply vector vector)))
	
    (if (= square-sum 0)
		vector
      (divide-vector-by vector
						(- 0 (sqrt square-sum))))))


;;;;;;;;;
;;;
;;; Operations on Opsat Solutions
;;;
;;;;;;;;;


(defmethod the-top-sln (top-solutions)
  
  "Get the top-sln from the set of top-solutions."

  
  (first top-solutions))


(defmethod duplicate-solution? ((potential-sln opsat-solution) reference-slns)
  
  "Let REFERENCE-SLNS be a list of unique opsat-solutions, and
   POTENTIAL-SLN be an opsat-solution.

   Returns Truee IFF the equivalent to POTENTIAL-SLN  already
      appears in REFERENCE-SLNS."
  
  (loop for reference-sln in reference-slns
	  do (when (same-opsat-solution? potential-sln reference-sln)
		   (return t))
	  finally (return nil)))
			 

(defmethod same-opsat-solution? ((sln1 opsat-solution) (sln2 opsat-solution))
  
  "Takes as input two Opsat Solutions, SLN1 and SLN2.
   Returns true iff the DECISION-ASSIGNMENTS of these solutions are equivalent."
  
  (equal (print-object sln1 nil)
		 (print-object sln2 nil)))


;;;;;;;;;
;;; Misc Utils
;;;;;;;;;

(defmethod generate-knee-base (base-sln knee-solution)
  
  (list base-sln knee-solution))


