;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

;;; Copyright (c) 1999 - 2007 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :OPSAT)

;;;(eval-when (compile load eval)
;;;  (export '(*default-state-trace*
;;;			state-trace
;;;;;;			new-state-trace
;;;			initialize-state-trace!
;;;			trace-search-state
;;;			print-checked-search-states)
;;;		  :OPSAT))

;;; This file contains code to support tracing of OPSAT's search procedure.

;;; To Do:
;;; o Convert from structs to objects.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  
;;; The main functions to be invoked in order to maintain a trace of
;;; search states are:

;;;   SET-DEBUG-LEVEL  
;;;      Sets the debugging level (described below).

;;;   DEBUG-LEVEL 
;;;      Returns the current debugging level.

;;;   PRINT-CHECKED-SEARCH-STATES 
;;;      Prints out the states that have been checked thus far.  
;;;      This function numbers each state (so that in the future 
;;;      checked states may be referenced by number) and prints
;;;      out the state's status.

;;; These three and related functions are described more precisely below.

;;; First, the following two functions are defined in :mtk-utils 
;;; (debug-support.lisp):

;;;  (SET-DEBUG-LEVEL &OPTIONAL (LEVEL 2))
;;;
;;;    Sets the debug level.  Must be an integer between 0 and 3, where:
;;;
;;;      0  means no debugging information is stored, and no trace information
;;;         is printed 
;;;      1  means some trace information is printed out, and no debugging
;;;         information is stored
;;;      2  means some trace information is printed out, and debugging
;;;         information is stored
;;;      3  more trace information is printed out, and debugging information
;;;         is stored "

;;;  (DEBUG-LEVEL)
;;;    Returns the current debug level.

;;; The remaining variables and functions are defined in this file:

;;;  *DEFAULT-STATE-TRACE* defaults to NIL
;;;    Stores the most recently tested search state, typically associated with 
;;;    the current search problem"

;;;  (MAKE-INSTANCE 'STATE-TRACE :KEY SEARCH-PROBLEM)
;;;    Creates a new state-trace for SEARCH-PROBLEM and returns it.  
;;;    Also sets *default-state-trace* to this newly created state-trace.

;;;  (INITIALIZE-STATE-TRACE!
;;;      &OPTIONAL (SEARCH-PROBLEM *DEFAULT-SEARCH-PROBLEM*))
;;;    Initializes tracing of search states for search-problem.

;;;  (TRACE-SEARCH-STATE (STATE-TRACE STATE-TRACE) SEARCH-STATE STATUS 
;;;      )
;;;    Inform the tracing system that SEARCH-STATE has been tested, and that
;;;    its test result is STATUS.  
;;;
;;;    SEARCH-STATE is a list of constraint-system assignment objects,
;;;    denoting assignments to a search problem's decision variables.
;;;
;;;    STATUS can be:
;;;      :ACCEPTED meaning this search-state was accepted,
;;;      :REJECTED meaning this search-state was rejected, or
;;;      :SUBSUMED meaning this search-state would have been accepted 
;;;               had it not also been deemed to be subsumed
;;;               by a kernel solution that is a subset of this state.

;;;    Depending on the value of (debug-level) performs one of three
;;;    actions:
;;;      = 0   Do nothing.
;;;      = 1   Print a short message.
;;;      = 2   Print a short message and record search state for play back.
;;;      = 3   Print a long message and record.

;;;  (PRINT-CHECKED-SEARCH-STATES
;;;      &OPTIONAL (STATE-TRACE *DEFAULT-STATE-TRACE*))
;;;    Prints the search-states that have been tested thus far.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; o Extend to conditional CSPs:
;;;   - add in comments of functions to change.
;;;   - modify functions as indicated by their to do.

;;;   o Find the debug-level and set-debug-level functions and pull them 
;;;     into an accessible file.  Rename debug to trace.
;;;   o Does debug state have to be a separate structure, or can it
;;;     just be a slot of search problem?
;;;   o Replace list manipulation with defstructs in the following fns.
;;;   o Write functions for tracing and stepping through search that
;;;     are similar to lisp tracing and stepping.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Global variable and structure definitions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *default-state-trace* nil
  "Stores the most recently tested search state, typically associated with 
  the current search problem")

(defclass state-trace ()
  
  ((search-problem
	:initarg :search-problem
	:initform nil
	:accessor state-trace-search-problem
	:documentation "The search problem being traced.")
   (search-states
	:initform nil
	:accessor state-trace-search-states
	:documentation "The search-states that have been tested and their ~
                    test results."))

  (:documentation "A structure containing a trace of the ~
                   search-states tested during search."))

(defmethod print-object ((state-trace state-trace) stream)
  (with-slots (search-problem) state-trace
	(format stream "<State trace of ~a>" search-problem)))

;;;(defstruct (state-trace
;;;			(:print-function print-state-trace))
;;;  "A structure containing a trace of the search-states tested
;;;   during search."
;;;  
;;;  (search-problem nil)					; The search problem being traced.
;;;  (search-states nil)					; The search-states that have been
;;;										; tested and their test results.
;;;  )
;;;
;;;(defun print-state-trace (state-trace stream depth)
;;;  
;;;  (declare (ignore depth))
;;;  (format stream
;;;	  "<State trace of ~a>"
;;;	  (ignore-errors (state-trace-search-problem state-trace))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Interface functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Doesn't this need to tell search-problem to point to this
;;;   new state trace?

;;; o Alternatively, should the state trace be what is supplied to the
;;;   other functions, so that this is generic to any search?

;;; o Make tracing coherent with *bfs-debug* and other
;;;   debugging flags.

(defmethod initialize-instance :after ((state-trace state-trace)
									   &rest extra-args)

  "Has side effect of setting *DEFAULT-STATE-TRACE* to be the newly 
   created state-trace."

  (declare (ignore Extra-Args))
  (setq *default-state-trace* state-trace))

;;;(defmethod new-state-trace ((search-problem opsat-problem))
;;;  
;;;  "Creates a new state-trace for SEARCH-PROBLEM and returns it.  
;;;   Also sets *default-state-trace* to this newly created state-trace."
;;;  
;;;  (setq *default-state-trace*
;;;    (make-instance 'state-trace
;;;	  :search-problem search-problem)))


(defmethod initialize-state-trace!
    ((search-problem opsat-problem))
  
  "Initializes tracing of search states for search-problem."
  
  (setf (state-trace-search-states (opsat-problem-state-trace search-problem))
	nil))

(defmethod trace-search-state ((state-trace state-trace) search-state status)
  
  "Inform the tracing system that SEARCH-STATE has been tested, and that
   its test result is STATUS.  

   SEARCH-STATE is a list of constraint-system assignment objects,
   denoting assignments to a search problem's decision variables.

   STATUS can be:
     :ACCEPTED meaning this search-state was accepted,
     :REJECTED meaning this search-state was rejected, or
     :SUBSUMED meaning this search-state would have been accepted 
               had it not also been deemed to be subsumed
               by a kernel solution that is a subset of this state.

   Depending on the value of *debug-level* performs one of three
   actions:
     = 0   Do nothing.
     = 1   Print a short message.
     = 2   Print a short message and record search state for play back.
     = 3   Print a long message and record."
  
  (if (member status '(:accepted :rejected :subsumed))
	  ;; only then proceed
      (ecase (debug-level)
		(0 ;; Do nothing
		 )
		(1 ;; Merely print, but do nothing else
		 (print-short-search-state-info search-state status))
		(2 ;; Print short info, but also store it away
		 (store-search-state-for-tracing  state-trace search-state status)
		 (print-short-search-state-info search-state status))
		(3 ;; Prints long info, and stores away the search-state
		 (store-search-state-for-tracing state-trace search-state status)
		 (print-long-search-state-info 
		  search-state 
		  status
		  (length (state-trace-search-states state-trace)))))
    (cerror "Continue" "Invalid search-state status: ~a" status)))

(defmethod print-checked-search-states
    ((state-trace state-trace))
  
  "Prints the search-states that have been tested thus far."
  
  (loop for cand-status in (state-trace-search-states state-trace)
	  for i from 1
	  do (print-long-search-state-info (first cand-status)
									   (second cand-status)
									   i)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Should this instead checkpoint the state, so that
;;;   search can pickup from that state?
;;; o Clean up this code, using accessors that are more
;;;   transparent.
(defmethod store-search-state-for-tracing
	((state-trace state-trace) search-state status)
  
  "Stores SEARCH-STATE with STATUS into STATE-TRACE."
  
  (if (state-trace-search-states state-trace)
      (setf (rest (last (state-trace-search-states state-trace)))
		(list (list search-state status)))
    (push (list search-state status) 
		  (state-trace-search-states state-trace))))

;;; To Do:
;;; o Should be method on search-state?  Defined where?
(defun print-long-search-state-info (search-state status number)
  
  "Prints out long information about search-state.  
   In all cases it prints out the search-state, its number, 
   and its status."
  
  (ecase status
    ((:rejected :subsumed)
     (debug-print t "~a. ~a : ~a"
				  number
				  status
				  search-state))
    (:accepted
     (debug-print t "~a. ~a : ~a"
				  number
				  status
				  search-state))))

;;; To Do:
;;; o Should be method on search-state?  Defined where?
(defun print-short-search-state-info (search-state status)
  
  "Prints out short information about SEARCH-STATE.  
   If the search-state is :REJECTED it prints a ., 
   if it is :SUBSUMED it prints a *, and if
   it is :ACCEPTED it prints out the search-state."
  
  (ecase status
    (:rejected
     (debug-print t "."))
    (:subsumed
     (debug-print t "*"))
    (:accepted
     (when search-state
	   ;; a non-empty search-state
       (debug-print t "accepted ~a~%" search-state)))))
