;;; -*- Mode:Common-Lisp; Package:OPSAT; Base:10 -*-

(in-package :OPSAT)

;;; Copyright (c) 1999 - 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                                               ;;;
;;;                     OPSAT Testing with ISAT                   ;;;
;;;                     State Clause Constraints                  ;;;
;;;                                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Add logging functions.
;;; (deftest-suite :OPSAT
;;;;;;    (log-OpSAT-statistics)
;;;;;;  (stop-and-print-OpSAT-statistics)
;;;    )

(deftest (test0 :opsat) ()

  ;; An optimal CSP with two decision variables and one 
  ;; clause to satisfy.  The best candidate satisfies 
  ;; the clause.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))))
		 (state-clauses '(
						  (:clause ((not (= a hi))))
						  ))
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test0
						  :name 'test0
						  )))
    (test '(((= A low)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "a={(hi 3),(low 2)}, 
                      b={hi,low},
                      clause: not a=hi")))

(deftest (test0-1 :opsat) ()

  ;; An optimal CSP with two decision variables and one 
  ;; clause to satisfy.  The best candidate satisfies 
  ;; the clause.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))))
		 (state-clauses '(
						  (:clause ((not (= a hi))))
						  ))
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test0
						  :name 'test0
						  )))
    (test '(((= A low)))
		  (find-solutions search-problem :maximum-solution-count 1)
		  :test #'equal-model-sequences
		  :fail-info "a={(hi 3),(low 2)}, 
                      b={hi,low},
                      clause: not a=hi")))

;;; Changed:
;;; Decision and State Variable specs updated for all tests,
;;; to include :domain keyword.
(deftest (test1 :opsat) ()

  ;; An optimal CSP with two decision variables and one 
  ;; clause to satisfy.  The best candidate satisfies 
  ;; the clause.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (state-clauses
		  '((:clause ((not (= a hi))(not (= b hi))))))
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test1
						  :name 'test1
						  )))
    (test '(((= A hi)(= B low))
			((= A low)(= B low))
			((= A low)(= B hi))
			)
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "a={(hi 3),(low 2)}, 
                      b={(hi 1),(low 2)},
                      c={hi,low},
                      clause: not a=hi or not b=hi")))

(deftest (test2 :opsat) ()

  ;; An optimal CSP with two decision variables and one 
  ;; clause to satisfy.  The best candidate violates the 
  ;; clause DIRECTLY, generating the negation of that clause, 
  ;; (a=hi,b=low), as a conflict.  This conflict is used to 
  ;; jump over next two candidates.  Last candidate satisfies 
  ;; clause.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 3)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (state-clauses
		  '(
			(:clause ((not (= a hi))(not (= b low))))
			))
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test2
						  :name 'test2
						  )))
    
    (test '(((= A low)(= B low))		;cost 2+3=5
			((= A hi)(= B hi))			;cost 3+1=4
			((= A low)(= B hi))			;cost 2+1=3
			)
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      B={hi,low}
                      clause: notA=hi or notB=low"
		  )))

(deftest (test3 :opsat) ()

  ;; An optimal CSP with two decision variables and two 
  ;; clauses to satisfy.  The partial assignment corresponding 
  ;; to the best candidate satisfies clauses, but no 
  ;; extension to this candidate satisfies both clauses
  ;; (since a=hi => not (c=low) and b=low => c=low).  
  ;; This results in the single conflict, (a=hi, b=low), 
  ;; same as Test 2.  Conflict used to jump over 
  ;; next two candidates; last candidate satisfies clause.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 3)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (state-clauses 
		  '(
			(:clause ((not (= a hi))(not (= c low))))
			(:clause ((= c low)(not (= b low))))
			))
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test3
						  :name 'test3
						  )))
    
    (test '(
			((= A low)(= B low))		;cost 2+3=5
			((= A hi)(= B hi))			;cost 3+1=4
			((= A low)(= B hi))			;cost 2+1=3
			)
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A=((hi 3)(low 2)) 
                      B=((hi 1)(low 3))
                      clauses: notA=hi or notC=low,
                               C=low or notB=low")))

(deftest (test4 :opsat) ()

  ;; An optimal CSP with two decision variables and 
  ;; three clauses to satisfy.  This is similar to 
  ;; Test 3, but the third clause results in a
  ;; unit clause conflict (a=hi), which subsumes 
  ;; conflict (a=hi, b=low).
  ;; That is, 
  ;;        a=hi      from candidate 1,
  ;;        c=hi      from clause 1
  ;;        b=hi      from clause 2   
  ;;           violating candidate assignment b=low
  ;;        c=low     from clause 3
  ;;           which is inconsistent

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 3)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (state-clauses
		  '(
			(:clause ((not (= a hi))(not (= c low))))
			(:clause ((= c low)(not (= b low))))
			(:clause ((= c low)(not (= b hi)))))
		  )
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test4
						  :name 'test4
						  )))

    (test  '(
			 ((= A low)(= B low))		;cost 2+3=5
			 ((= A low)(= B hi))		;cost 2+1=3
			 )
										;'(((= A low)(= B low)))
		   (find-solutions search-problem)
		   :test #'equal-model-sequences
		   :fail-info "A=((hi 3)(low 2)), 
                      B=((hi 1)(low 3)),
                      clauses:
                        notA=hi or notC=low,
                        C=low or notB=low,
                        C=low or notB=hi"
		   )))

(deftest (test5 :opsat) ()

  ;; An optimal CSP with two decision variables and four 
  ;; clauses to satisfy.  This is similar to Test 4, 
  ;; which produces conflict not a=hi.
  ;; However, the additional clause results in a 
  ;; second conflict, (a=low, b=low),  being discovered.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (state-clauses
		  '((:clause ((not (= a hi))(not (= c low))))
			(:clause ((= c low)(not (= b low))))
			(:clause ((= c low)(not (= b hi))))
			(:clause ((not (= a low))(not (= b low)))))
		  )
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test5
						  :name 'test5
						  )))
    (test '(((= A low)(= B hi)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A=((hi 3)(low 2)), 
                      B=((hi 1)(low 2)),
                      clauses:
                        notA=hi or notC=low,
                        C=low or notB=low,
                        C=low or notB=hi,
                        notA=low or notB=low")))

(deftest (test6 :opsat) ()

  ;; An optimal CSP with two decision variables and four 
  ;; clauses to satisfy.  This is similar to Test 5, but 
  ;; the fourth clause is tightened, resulting in two unit 
  ;; conflicts, (a=hi) and (b=low).  No feasible solution 
  ;; exists.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b  :domain ((hi :cost 1) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (state-clauses '(
						  (:clause ((not (= a hi))(not (= c low))))
						  (:clause ((= c low)(not (= b low))))
						  (:clause ((= c low)(not (= b hi))))
						  (:clause ((not (= a low))))
						  )
						)

		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test6
						  :name 'test6
						  )))

    (test '()
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A=((hi 3)(low 2)), 
                      B=((hi 1)(low 2)),
                      clauses: notA=hi or notC=low,
                               C=low or notB=low,
                               C=low or notB=hi,
                               notA=low
            -> Soln: none")))

(deftest (test7 :opsat) ()

  ;; An optimal CSP with two decision variables, two base 
  ;; (theory) clauses, and two unsat clauses.  
  ;; This is similar to Test 3, but with unsat clauses added.
  ;; One conflict is generated, (a=hi, b=low); the same as Test 3.
  ;; No conflicts are generated by the unsat clauses.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 3)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (base-clauses 
		  '(
			(:clause ((not (= a hi))(not (= c low))))
			(:clause ((= c low)(not (= b low))))
			)
		  )
	 
		 (unsat-clauses-list '(
							   ((:clause ((= c low)(not (= b hi))))
								(:clause ((not (= a low)))))
							   )
							 )

		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :unsat-constraint-descriptions-list
						  unsat-clauses-list
						  :source 'test7
						  :name 'test7
						  )))

    (test 
     '(((= A low)(= B low))				;cost 2+3=5
       ((= A hi)(= B hi))				;cost 3+1=4
       ((= A low)(= B hi)))				;cost 2+1=3

     (find-solutions search-problem)
     :test #'equal-model-sequences
     :fail-info "A=((hi 3)(low 2)), 
                 B=((hi 1)(low 2)),
                 base: notA=hi or notC=low,
                       C=low or notB=low,
                 unsat: C=low or notB=hi,
                        notA=low")))

(deftest (test8 :opsat) ()

  ;; An optimal CSP with two decision variables, one 
  ;; base (theory) clause, one sat clause and one 
  ;; unsat clause.  Three conflicts are generated, 
  ;; (a=hi), (b=low) and (a=low, b=hi).  
  ;; The first conflict is generated from the first
  ;; candidate, (a=hi,b=low), using the base clause
  ;; and the first disjunct of the unsat clause.
  ;; The second conflict is generated from the
  ;; candidate (a=low,b=low), using the second disjunct
  ;; of the unsat clause.
  ;; The third conflict is generated from the sat clause.
  ;; All candidates are eliminated.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
	 
		 (base-clauses 
		  '((:clause ((not (= a hi))(= c low)))))
		 (sat-clauses-list 
		  '(
			((:clause ((= a hi)(not (= b hi)))))
			)
		  )
		 (unsat-clauses-list 
		  '(
			((:clause ((= c low)(not (= b hi)))))
			)
		  )
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :sat-constraint-descriptions-list
						  sat-clauses-list
						  :unsat-constraint-descriptions-list
						  unsat-clauses-list
						  :source 'test8
						  :name 'test8
						  )))
    (test '()
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A=((hi 3)(low 2)), 
                      B=((hi 1)(low 2)),
                      base: notA=hi or C=low,
                      sat:  A=hi or notB=hi,
                      unsat: C=low or notB=hi
                      -> Soln: none")))

(deftest (test9 :opsat) ()

  ;; An optimal CSP with two decision variables, one 
  ;; base (theory) clause, one sat clause and one 
  ;; unsat clause.  This is similar to Test 8, but
  ;; with the sat clause relaxed.  One conflict, 
  ;; (b=low), is generated by the first candidate, 
  ;; (a=hi,b=low) using the unsat clause.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (base-clauses '(
						 (:clause ((not (= a hi))(not (= c low))))
						 ))
		 (sat-clauses-list '(
							 (
							  (:clause ((= c hi)(not (= b hi))))
							  )
							 )
						   )
		 (unsat-clauses-list '((
								(:clause ((= c low)(not (= b hi))))
								)
							   ))
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :sat-constraint-descriptions-list
						  sat-clauses-list
						  :unsat-constraint-descriptions-list
						  unsat-clauses-list
						  :source 'test9
						  :name 'test9
						  )))
    (test '(((= A hi)(= B hi)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A=((hi 3)(low 2)), 
                      B=((hi 1)(low 2)),
                      base: notA=hi or not C=low,
                      sat:  C=hi or notB=hi,
                      unsat: C=low or notB=hi")))

(deftest (test10 :opsat) ()

  ;; An optimal CSP with two decision variables and one 
  ;; clause to satisfy.  The best candidate satisfies 
  ;; the clause.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))
			(c :domain ((hi :cost 1) (low :cost 2)))
			(d :domain ((hi :cost 1) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))
			(d :domain (hi low))
			(e :domain (hi low))
			(f :domain (hi low))))
		 (base-clauses '(
						 (:clause ((not (= a hi))(not (= b hi))))
						 )
					   )
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :source 'test10
						  :name 'test10
						  )))

    (test '(((= A hi)(= B low)(= C low)(= D low)) ; 1 = 9
			((= A low)(= B low)(= C low)(= D low)) ;2 = 8
			((= A hi)(= B low)(= C hi)(= D low)) ;  3 = 8
			((= A hi)(= B low)(= C low)(= D hi)) ;  4 = 8
			((= A low)(= B low)(= C low)(= D hi)) ; 5 = 7
			((= A hi)(= B low)(= C hi)(= D hi)) ;   6 = 7
			((= A low)(= B low)(= C hi)(= D low)) ; 7 = 7
			((= A low)(= B hi)(= C low)(= D low)) ; 8 = 7
			((= A low)(= B hi)(= C hi)(= D low)) ;  9 = 6
			((= A low)(= B hi)(= C low)(= D hi)) ; 10 = 6 
			((= A low)(= B low)(= C hi)(= D hi)) ; 11 = 6
			((= A low)(= B hi)(= C hi)(= D hi)) ;  12 = 5
			)
		  (find-solutions search-problem
						  :cutoff-delta nil
						  :maximum-solution-count nil)
		  :test #'equal-model-sequences
		  :fail-info "a={(hi 3),(low 2)}, 
                      b=c=d={(hi 1),(low 2)},
                      e=f={hi,low},
                      Soln: <{a=h,b=l,c=l,d=l},              ;cost 9 *
                             {a=l,b=l,c=l,d=l},              ;cost 8 *
                             {a=h,b=l,c=h,d=l},              ;cost 8 *
                             {a=h,b=l,c=l,d=h},              ;cost 8 *
                             {a=l,b=l,c=l,d=h},              ;cost 7 *
                             {a=h,b=l,c=h,d=h},              ;cost 7 *

                             {a=l,b=l,c=l,d=h},              ;cost 7 *
                             {a=l,b=l,c=h,d=l},              ;cost 7 *
                             {a=l,b=l,c=h,d=l},              ;cost 7 *
                             {a=l,b=h,c=l,d=l}               ;cost 7 *
                             {a=l,b=l,c=h,d=h},              ;cost 6
                             {a=l,b=h,c=l,d=h}               ;cost 6
                             {a=l,b=h,c=h,d=l}               ;cost 6
                             {a=l,b=h,c=h,d=h}>              ;cost 5"
		  )))

(deftest (test11 :opsat) ()

  ;; An optimal CSP with two decision variables, one 
  ;; base (theory) clause, one sat clause and one 
  ;; unsat clause.  This is similar to Test 8, but
  ;; with the sat clause relaxed.  One conflict, 
  ;; (b=low), is generated by the first candidate, 
  ;; (a=hi,b=low) using the unsat clause.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (base-clauses '(
						 (:clause ((not (= a hi))(not (= c low))))
						 ))
		 (sat-clauses-list '(
							 (
							  (:clause ((= c hi)(not (= b hi))))
							  )
							 (
							  (:clause ((= a hi)(= b hi)))
							  )
							 ))
		 (unsat-clauses-list '(((:clause ((= c low) (not (= b hi)))))))

		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :sat-constraint-descriptions-list
						  sat-clauses-list
						  :unsat-constraint-descriptions-list
						  unsat-clauses-list
						  :source 'test11
						  :name 'test11
						  )))
    (test '(((= A hi)(= B hi)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A=((hi 3)(low 2)), 
                      B=((hi 1)(low 2)),
                      base: notA=hi or not C=low,
                      sat:  C=hi or notB=hi,
                      unsat: C=low or notB=hi")))

(deftest (test31 :opsat) ()

  ;; A conditional optimal CSP with two (unconditional) 
  ;; decision variables and one clause to satisfy.  
  ;; The best candidate satisfies the clause.

  (let* ((decision-variable-descriptions 
		  `((a :domain ((hi :cost 3) (low :cost 2))
			   :guard ,*always-active*)
			(b :domain ((hi :cost 1) (low :cost 2))
			   :guard ,*always-active*)))
		 
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 
		 (state-clauses
		  '((:clause ((not (= a hi))(not (= b hi))))))
		 
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test31
						  :name 'test31
						  :conditional? T)))
	
    (test '(((= A hi)(= B low))
			((= A low)(= B low))
			((= A low)(= B hi))
			)
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "a={(hi 3),(low 2)}, 
                      b={(hi 1),(low 2)},
                      c={hi,low},
                      clause: not a=hi or not b=hi")))


(deftest (test32 :opsat) ()

  ;; Similar to test10 but solved through Constraint-based A*
  ;; (no conflict-direction)

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 3) (low :cost 2)))
			(b :domain ((hi :cost 1) (low :cost 2)))
			(c :domain ((hi :cost 1) (low :cost 2)))
			(d :domain ((hi :cost 1) (low :cost 2)))))
		 
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))
			(d :domain (hi low))
			(e :domain (hi low))
			(f :domain (hi low))))
		 
		 (base-clauses '((:clause ((not (= a hi))(not (= b hi))))))
		 
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :source 'test32
						  :name 'test32
						  :conflict-directed? nil)))

    (test '(((= A hi)(= B low)(= C low)(= D low)) ; 1 = 9
			((= A low)(= B low)(= C low)(= D low)) ;2 = 8
			((= A hi)(= B low)(= C hi)(= D low)) ;  3 = 8
			((= A hi)(= B low)(= C low)(= D hi)) ;  4 = 8
			((= A low)(= B low)(= C low)(= D hi)) ; 5 = 7
			((= A hi)(= B low)(= C hi)(= D hi)) ;   6 = 7
			((= A low)(= B low)(= C hi)(= D low)) ; 7 = 7
			((= A low)(= B hi)(= C low)(= D low)) ; 8 = 7
			((= A low)(= B hi)(= C hi)(= D low)) ;  9 = 6
			((= A low)(= B hi)(= C low)(= D hi)) ; 10 = 6 
			((= A low)(= B low)(= C hi)(= D hi)) ; 11 = 6
			((= A low)(= B hi)(= C hi)(= D hi))) ;  12 = 5

		  (find-solutions search-problem
						  :cutoff-delta nil
						  :maximum-solution-count nil)
		  :test (get-test-fn-for-seq-of-sets '((1 3) (4 7) (8 10)))
										;:test #'equal-model-sequences
		  :fail-info "a={(hi 3),(low 2)}, 
                      b=c=d={(hi 1),(low 2)},
                      e=f={hi,low},
                      Soln: <{a=h,b=l,c=l,d=l},              ;cost 9 *
                             {a=l,b=l,c=l,d=l},              ;cost 8 *
                             {a=h,b=l,c=h,d=l},              ;cost 8 *
                             {a=h,b=l,c=l,d=h},              ;cost 8 *
                             {a=l,b=l,c=l,d=h},              ;cost 7 *
                             {a=h,b=l,c=h,d=h},              ;cost 7 *

                             {a=l,b=l,c=l,d=h},              ;cost 7 *
                             {a=l,b=l,c=h,d=l},              ;cost 7 *
                             {a=l,b=l,c=h,d=l},              ;cost 7 *
                             {a=l,b=h,c=l,d=l}               ;cost 7 *
                             {a=l,b=l,c=h,d=h},              ;cost 6
                             {a=l,b=h,c=l,d=h}               ;cost 6
                             {a=l,b=h,c=h,d=l}               ;cost 6
                             {a=l,b=h,c=h,d=h}>              ;cost 5"
		  )))

;;; TO DO:
;;; o Eliminete redundancy (when declaring decision variables, cost should not
;;;   be mentioned.
;;; -- gustavop

(deftest (test33 :opsat) ()

  ;; A Valued CSP that is similar to test2.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 0) (low :cost 0)))
			(b :domain ((hi :cost 0) (low :cost 0)))))
		 
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
	 
		 (state-clauses
		  '((:clause ((not (= a hi))(not (= b low))))))
		 
		 (valued-constraints (list
							  (new-soft-constraint-list '(A)
														'(((HI) 3)
														  ((LOW) 2)) #'<=)
							  (new-soft-constraint-list '(B)
														'(((HI) 1)
														  ((LOW) 3)) #'<=)))
	 
		 (search-problem (new-vcsp-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test33
						  :name 'test33
						  :cost-order-function #'<=
						  :soft-constraint-system 
						  valued-constraints)))
    
    (test (reverse '(((= A low)(= B low)) ;	cost 2+3=5
					 ((= A hi)(= B hi))	;	cost 3+1=4
					 ((= A low)(= B hi)) ;cost 2+1=3
					 ))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      B={hi,low}
                      clause: notA=hi or notB=low"
		  )))


(deftest (test34 :opsat) ()

  ;; A Valued CSP that is similar to test7.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 0) (low :cost 0)))
			(b :domain ((hi :cost 0) (low :cost 0)))))
		 
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))

		 (base-clauses 
		  '((:clause ((not (= a hi))(not (= c low))))
			(:clause ((= c low)(not (= b low))))))
	 
		 (unsat-clauses-list '(((:clause ((= c low)(not (= b hi))))
								(:clause ((not (= a low)))))))
	 
		 (valued-constraints
		  (list (new-soft-constraint-list '(A)
										  '(((HI) 3)
											((LOW) 2)) #'>=)
				(new-soft-constraint-list '(B)
										  '(((HI) 1)
											((LOW) 3)) #'>=)))
		 (search-problem (new-vcsp-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :unsat-constraint-descriptions-list
						  unsat-clauses-list
						  :source 'test34
						  :name 'test34
						  :soft-constraint-system
						  valued-constraints)))

    (test 
     '(((= A low)(= B low))				;cost 2+3=5
       ((= A hi)(= B hi))				;cost 3+1=4
       ((= A low)(= B hi)))				;cost 2+1=3

     (find-solutions search-problem)
     :test #'equal-model-sequences
     :fail-info "A=((hi 3)(low 2)), 
                 B=((hi 1)(low 2)),
                 base: notA=hi or notC=low,
                       C=low or notB=low,
                 unsat: C=low or notB=hi,
                        notA=low")))


(deftest (test35 :opsat) ()

  ;; A Valued CSP that is similar to test9.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 0) (low :cost 0)))
			(b :domain ((hi :cost 0) (low :cost 0)))))
	 
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
	 
		 (base-clauses '((:clause ((not (= a hi))(not (= c low))))))
		 (sat-clauses-list '(((:clause ((= c hi)(not (= b hi)))))))
		 (unsat-clauses-list '(((:clause ((= c low) (not (= b hi)))))))
	 
		 (valued-constraints
		  (list (new-soft-constraint-list '(A)
										  '(((HI) 3) ((LOW) 2)) #'<=)
				(new-soft-constraint-list '(B)
										  '(((HI) 1) ((LOW) 2)) #'<=)))
	 
		 (search-problem (new-vcsp-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :sat-constraint-descriptions-list
						  sat-clauses-list
						  :unsat-constraint-descriptions-list
						  unsat-clauses-list
						  :source 'test35
						  :name 'test35
						  :cost-order-function #'<=
						  :soft-constraint-system
						  valued-constraints)))
    (test '(((= A hi)(= B hi)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A=((hi 3)(low 2)), 
                      B=((hi 1)(low 2)),
                      base: notA=hi or not C=low,
                      sat:  C=hi or notB=hi,
                      unsat: C=low or notB=hi")))


(deftest (test36 :opsat) ()

  ;; A Valued CSP that is similar to test13.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 0) (low :cost 0))) ; hi 25 low 8
		    (b :domain ((hi :cost 0) (low :cost 0))) ; hi 40 low 12
		    (c :domain ((hi :cost 0) (low :cost 0))) ; hi 30 low 20
		    (d :domain ((hi :cost 0) (low :cost 0))))) ; hi 20 low 7
			
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
		    (c :domain (hi low))
		    (d :domain (hi low))))
			
		 (state-clauses
		  '((:clause ((not (= d hi))))))
	 
		 (valued-constraints (list
							  (new-soft-constraint-list '(A)
														'(((HI) 25)
														  ((LOW) 8)) #'>=)
							  (new-soft-constraint-list '(B)
														'(((HI) 40)
														  ((LOW) 12)) #'>=)
							  (new-soft-constraint-list '(C)
														'(((HI) 30)
														  ((LOW) 20)) #'>=)
							  (new-soft-constraint-list '(D)
														'(((HI) 20)
														  ((LOW) 7)) #'>=)))
			
		 (search-problem (new-vcsp-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test36
						  :name 'test36
						  :soft-constraint-system 
						  valued-constraints)))
    
    (test
     '(((= D LOW) (= C HI) (= B HI) (= A HI)) ((= D LOW) (= C LOW) (= B HI) (= A HI))
	   ((= D LOW) (= C HI) (= B HI) (= A LOW)) ((= D LOW) (= C LOW) (= B HI) (= A LOW))
	   ((= D LOW) (= C HI) (= B LOW) (= A HI)) ((= D LOW) (= C LOW) (= B LOW) (= A HI))
	   ((= D LOW) (= C HI) (= B LOW) (= A LOW)) ((= D LOW) (= C LOW) (= B LOW) (= A LOW)))
										; '(((= A hi)(= B hi)(= C hi)(= D low)) ;cost 25 + 40 + 30 + 7 = 102
										; 	((= A hi)(= B hi)(= C low)(= D low))	;cost 25 + 40 + 20 + 7 = 92
										; 	((= A low)(= B hi)(= C hi)(= D low))	;cost 8  + 40 + 30 + 7 = 85
										; 	((= A low)(= B hi)(= C low)(= D low))	;cost 8  + 40 + 20 + 7 = 75
										; 	((= A hi)(= B low)(= C hi)(= D low))	;cost 25 + 12 + 30 + 7 = 74
										; 	((= A hi)(= B low)(= C low)(= D low))	;cost 25 + 12 + 20 + 7 = 64
										; 	((= A low)(= B low)(= C hi)(= D low))   ;cost 8  + 12 + 30 + 7 = 57
										; 	((= A low)(= B low)(= C hi)(= D low)))  ;cost 8  + 12 + 20 + 7 = 47

	 (find-solutions search-problem)
	 :test #'equal-model-sequences
	 :fail-info "((a (hi :cost 25) (low :cost 8))
		    (b (hi :cost 40) (low :cost 12))
		    (c (hi :cost 30) (low :cost 20))
		    (d (hi :cost 20) (low :cost 7))
		     clause: (:clause ((not (= d hi))))"
	 )))

(deftest (test38 :opsat) ()

  ;; Ed Crawley's Apollo design example, framed as a valued CSP.

  (let* ((decision-variable-descriptions
		  '((lor :domain ((yes :cost 0) (no :cost 0)))
			(cm-crew :domain ((cm-2 :cost 0) (cm-3 :cost 0)))
			(lm-crew :domain ((lm-0 :cost 0) (lm-1 :cost 0) (lm-2 :cost 0) (lm-3 :cost 0)))
			(lm-fuel :domain ((lm-cryogenic :cost 0) (lm-storable :cost 0) (lm-na :cost 0)))
			(sm-fuel :domain ((sm-cryogenic :cost 0) (sm-storable :cost 0)))))
		 (state-variable-descriptions
		  '((eor :domain (yes no))
			(lor :domain (yes no))
			(earth-launch :domain (orbit direct))
			(moon-arrival :domain (orbit direct))
			(moon-departure :domain (orbit direct))
			(cm-crew :domain (cm-2 cm-3))
			(lm-crew :domain (lm-0 lm-1 lm-2 lm-3))
			(lm-fuel :domain (lm-cryogenic lm-storable lm-na))
			(sm-fuel :domain (sm-cryogenic sm-storable))))
		 (state-clauses
		  '((:clause ((= earth-launch orbit)(= eor no))) ;EORConstraint
			(:clause ((= moon-arrival orbit)(= lor no))) ;LORConstraint
			(:clause ((= cm-crew cm-3)(not (= lm-crew lm-3)))) ;lmcmcrew
			(:clause ((= moon-departure orbit)(= lor no))) ;moonLeaving
			(:clause ((not (= lm-crew lm-0))(not (= lor yes)))) ;lmexists
			(:clause ((= lm-crew lm-0)(= lor yes))) ;lmexists
			(:clause ((not (= lm-fuel lm-na))(not (= lor yes)))) ;lmFuelCons
			(:clause ((= lm-fuel lm-na)(= lor yes))))) ;lmFuelCons
		 (valued-constraints
		  (list
		   (new-soft-constraint-list '(SM-FUEL LOR LM-FUEL LM-CREW)
									 '(((nil NO nil nil) 0)
									   ((nil nil nil LM-0) 0)
									   ((SM-CRYOGENIC YES LM-CRYOGENIC LM-1) 5540.534)
									   ((SM-STORABLE YES LM-CRYOGENIC LM-1) 7953.0024)
									   ((SM-CRYOGENIC YES LM-STORABLE LM-1) 9737.675)
									   ((SM-STORABLE YES LM-STORABLE LM-1) 13977.67)
									   ((SM-CRYOGENIC YES LM-NA LM-1) 9737.675)
									   ((SM-STORABLE YES LM-NA LM-1) 13977.67)
									   ((SM-CRYOGENIC YES LM-CRYOGENIC LM-2) 7387.3784)
									   ((SM-STORABLE YES LM-CRYOGENIC LM-2) 10604.004)
									   ((SM-CRYOGENIC YES LM-STORABLE LM-2) 12983.566)
									   ((SM-STORABLE YES LM-STORABLE LM-2) 18636.895)
									   ((SM-CRYOGENIC YES LM-NA LM-2) 12983.566)
									   ((SM-STORABLE YES LM-NA LM-2) 18636.895)
									   ((SM-CRYOGENIC YES LM-CRYOGENIC LM-3) 9234.223)
									   ((SM-STORABLE YES LM-CRYOGENIC LM-3) 13255.004)
									   ((SM-CRYOGENIC YES LM-STORABLE LM-3) 16229.458)
									   ((SM-STORABLE YES LM-STORABLE LM-3) 23296.117)
									   ((SM-CRYOGENIC YES LM-NA LM-3) 16229.458)
									   ((SM-STORABLE YES LM-NA LM-3) 23296.117)) #'<=)
			    
			    
		   (new-soft-constraint-list '(SM-FUEL LM-CREW)
									 '(((nil LM-0) 0)
									   ((SM-CRYOGENIC LM-1) 2612.4048)
									   ((SM-STORABLE LM-1) 3749.9023) ((SM-CRYOGENIC LM-2) 3483.206)
									   ((SM-STORABLE LM-2) 4999.8696) ((SM-CRYOGENIC LM-3) 4354.008)
									   ((SM-STORABLE LM-3) 6249.8374)) #'<=)
			    
		   (new-soft-constraint-list '(SM-FUEL LOR CM-CREW)
									 '(((nil YES nil) 0)
									   ((SM-CRYOGENIC NO CM-2) 78119.43)
									   ((SM-STORABLE NO CM-2) 168699.5) ((SM-CRYOGENIC NO CM-3) 107414.22)
									   ((SM-STORABLE NO CM-3) 231961.78)) #'<=)
			    
		   (new-soft-constraint-list '(SM-FUEL CM-CREW)
									 '(((SM-CRYOGENIC CM-2) 15880.687)
									   ((SM-STORABLE CM-2) 24204.219) ((SM-CRYOGENIC CM-3) 21835.945)
									   ((SM-STORABLE CM-3) 33280.797)) #'<=)
			    
		   (new-soft-constraint-list '(LOR LM-FUEL LM-CREW)
									 '(((NO nil nil) 0)
									   ((YES nil LM-0) 0)
									   ((YES LM-CRYOGENIC LM-1) 15690.348)
									   ((YES LM-STORABLE LM-1) 27576.312) ((YES LM-NA LM-1) 27576.312)
									   ((YES LM-CRYOGENIC LM-2) 20920.465) ((YES LM-STORABLE LM-2) 36768.42)
									   ((YES LM-NA LM-2) 36768.42) ((YES LM-CRYOGENIC LM-3) 26150.578)
									   ((YES LM-STORABLE LM-3) 45960.523) ((YES LM-NA LM-3) 45960.523)) #'<=)
			    
		   (new-soft-constraint-list '(LM-CREW)
									 '(((LM-0) 0)
									   ((LM-1) 7398.1206) ((LM-2) 9864.16) ((LM-3) 12330.201)) #'<=)
			    
		   (new-soft-constraint-list '(CM-CREW)
									 '(((CM-2) 19728.32) ((CM-3) 27126.441)) #'<=)))
	 
	 
		 (search-problem (new-vcsp-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :source 'test38
						  :name 'test38
						  :cost-order-function #'<=
						  :soft-constraint-system 
						  valued-constraints
						  )))
    
    (test  '(((= LOR yes)(= CM-CREW CM-2)(= LM-CREW LM-1)(= LM-FUEL LM-CRYOGENIC)(= SM-FUEL SM-CRYOGENIC))
			 ((= LOR yes)(= CM-CREW CM-2)(= LM-CREW LM-2)(= LM-FUEL LM-CRYOGENIC)(= SM-FUEL SM-CRYOGENIC)))
	   
		   (find-solutions search-problem
						   :cutoff-delta nil
						   :maximum-solution-count 2)
										;:find-all-consistent-models? T)	;2)
		   :test #'equal-model-sequences
		   :fail-info "A={(hi 3),(low 2)} 
                      B={(hi 1),(low 3)}
                      B={hi,low}
                      clause: notA=hi or notB=low")))



(deftest (test39 :opsat) ()

  ;; Test9, but with the assignment costs swapped.

  (let* ((decision-variable-descriptions 
		  '((a :domain ((hi :cost 2) (low :cost 3)))
			(b :domain ((hi :cost 2) (low :cost 1)))))
		 (state-variable-descriptions 
		  '((a :domain (hi low))
			(b :domain (hi low))
			(c :domain (hi low))))
		 (base-clauses '((:clause ((not (= a hi))(not (= c low))))))
		 (sat-clauses-list '(((:clause ((= c hi)(not (= b hi)))))))
		 (unsat-clauses-list '(((:clause ((= c low)(not (= b hi)))))))
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  base-clauses
						  :sat-constraint-descriptions-list
						  sat-clauses-list
						  :unsat-constraint-descriptions-list
						  unsat-clauses-list
						  :source 'test39
						  :name 'test39
						  )))
    (test '(((= A hi)(= B hi)))
		  (find-solutions search-problem)
		  :test #'equal-model-sequences
		  :fail-info "A=((hi 3)(low 2)), 
                      B=((hi 1)(low 2)),
                      base: notA=hi or not C=low,
                      sat:  C=hi or notB=hi,
                      unsat: C=low or notB=hi")))

(deftest (test30 :opsat) ()
  "Test out a moderate sized n-queens problem."
  
  (let ((solution (nqueens-test 10 :print-on nil)))
	(test '(((= COL10 R1) (= COL9 R3) (= COL8 R6) (= COL7 R9) (= COL6 R7)
			 (= COL5 R10) (= COL4 R4) (= COL3 R2) (= COL2 R5) (= COL1 R8)))
		  solution
		  :test #'equal-model-sequences
		  :fail-info "Nqueens 10 test.")))

; (deftest (test42 :opsat) ()

;   ;; Testing the state-space-conflicts that are being generated
;   ;; It seems that we are identical conflicts multiple times

;   (let* ((decision-variable-descriptions 
; 		  '((a :domain ((hi :cost 0) (low :cost 0))) ; hi 25 low 8
; 		    (b :domain ((hi :cost 0) (low :cost 0))) ; hi 40 low 12
; 		    (c :domain ((hi :cost 0) (low :cost 0))) ; hi 30 low 20
; 		    )) 
			
; 		 (state-variable-descriptions 
; 		  '((a :domain (hi low))
; 			(b :domain (hi low))
; 		    (c :domain (hi low))
; 		    ))
			
; 		 (state-clauses
; 		  '((:clause ((not (= a hi))(not (= b hi))))))
	 
; 		 (valued-constraints (list
; 							  (new-soft-constraint-list '(A)
; 														'(((HI) 25)
; 														  ((LOW) 8)) #'>=)
; 							  (new-soft-constraint-list '(B)
; 														'(((HI) 40)
; 														  ((LOW) 12)) #'>=)
; 							  (new-soft-constraint-list '(C)
; 														'(((HI) 30)
; 														  ((LOW) 20)) #'>=)
; 							  ))
			
; 		 (search-problem (new-vcsp-problem
; 						  decision-variable-descriptions
; 						  state-variable-descriptions
; 						  state-clauses
; 						  :source 'test40
; 						  :name 'test40
; 						  :soft-constraint-system 
; 						  valued-constraints)))
	
;     (test
; 	 '( ~(A=HI B=HI))
; ;      '( ((= C HI) (= B HI) (= A LOW))
; ; 	    ((= C LOW) (= B HI) (= A LOW))
; ; 	   ((= C HI) (= B LOW) (= A HI)) ((= C LOW) (= B LOW) (= A HI))
; ; 	   ((= C HI) (= B LOW) (= A LOW)) ((= C LOW) (= B LOW) (= A LOW)))
	 
; 										; 	((= A low)(= B hi)(= C hi))	;cost 8  * 40 * 30 = 78
; 										; 	((= A low)(= B hi)(= C low)) ;cost 8  + 40 + 20 = 68	 
; 										; 	((= A hi)(= B low)(= C hi))	;cost 25 + 12 + 30  = 67
; 										; 	((= A hi)(= B low)(= C low))	;cost 25 + 12 + 20 = 57
; 										; 	((= A low)(= B low)(= C hi))   ;cost 8  + 12 + 30 = 50
; 										; 	((= A low)(= B low)(= C low))	;cost 8 + 12 + 20 = 40	 
	 
; 	 (progn 
; 	   (find-solutions search-problem)
; 	   (bfig::state-space-conflicts search-problem))
; 	 :test #'equal-model-sequences
; 	 :fail-info "((a (hi :cost 25) (low :cost 8))
; 		    (b (hi :cost 40) (low :cost 12))
; 		    (c (hi :cost 30) (low :cost 20)))
; 		     clause: (:clause ((not (= a hi))(not (= b hi))))"
; 	 )))
