;;; RUN "(nqueen-test 5)" from within the opsat package.

(in-package :OPSAT)

(defun make-column-var (i)
  (intern (concatenate 'string "COL" (format nil "~a" i))
		   :opsat)
)

(defun make-row-value (i)
  (intern (concatenate 'string "R" (format nil "~a" i))
		  		  :opsat)
)

;;; adds row constraints
(defun make-state-clauses-neqrows (column-vars row-values)
"Creates a list of clauses representing all of the row constraints"
  (loop for i1 from 0 to (- (length column-vars) 1) append
	   (loop for i2 from 0 to (- (length column-vars) 1) append
			(let ((var1 (nth i1 column-vars))
				  (var2 (nth i2 column-vars)))
				(loop for val in row-values append
					 (if (< i1 i2)			
						 `((:clause 
							((not (= ,var1 ,val))
							 (not (= ,var2 ,val)))))
						 nil)
					 )))))


;;; adds diagonal constraints
(defun make-state-clauses-neqdiag (column-vars row-values)
"Creates a list of clauses representing all of the diagonal constraints"
  (let* ((maxi (- (length column-vars) 1))
		 (diag-len (+ (* 2 maxi) 1))
		 (diagLR (loop for i from 1 to diag-len collect '()))
		 (diagRL (loop for i from 1 to diag-len collect '())))
	(loop for ivar from 0 to maxi do
		 (loop for ival from 0 to maxi do
			  (let ((var (nth ivar column-vars))
					(val (nth ival row-values))
					(iLR (+ (- ivar ival) maxi))
					(iRL (+ (- (- maxi ivar) ival) maxi)))
				(push `(not (= ,var ,val)) (nth iLR diagLR))
				(push `(not (= ,var ,val)) (nth iRL diagRL)))))
	(append (loop for atom-list in diagLR append (make-pairwise-clause atom-list)) 
			(loop for atom-list in diagRL append (make-pairwise-clause atom-list)))
))

(defun make-pairwise-clause (atom-list)
"Given a list of atoms '(A1 A2 A3), this method returns a list of clauses
representing all of the pairwise ORs of those atoms, with no repeats. i.e.:

((:CLAUSE (A1 A2))
 (:CLAUSE (A1 A3))
 (:CLAUSE (A2 A3)))
"
  (let* ((len (length atom-list)))
	(if (equal len 1) 
		nil
		(loop for i from 0 to (- len 2) append
			 (loop for j from (+ i 1) to (- len 1) append
				  `((:clause 
					 (,(nth i atom-list) ,(nth j atom-list))))
)))))



(defun nqueens-test (n &key (print-on t)) 
"This function creates and runs an OPSAT N-queens problem. The parameter N dictates
the number of queens in the problem. The problem is expressed as follows:

Variable - one variable per column (numbered starting from 1).
Domain - the domain of each variable are the rows (numbered starting from 1).
Constraint - there are two types of constraints: row and diagonal.

-- row constraint:
The expression that row 1 only has one queen looks as follows using infix notation:
  (col1!=r1) -> ((col2!=r1) & (col3!=r1) & ... )
which takes the following form in CNF, where each clause has two atomic propositions.
  (((col1!=r1) v (col2!=r1)) ^ ((col1!=r1) v (col3!=r1)) ^ ...for each column after col1 )
For an N-queens problem, the first row introduces (N-1) row-constraint clauses.
The second row introduces (N-2) new clauses, then (N-3), and so on. 
There are N(N-1)/2 row-constraint clauses in total.

-- diag constraint
A diagonal constraint looks similar to a row constraint, each clause consists of two 
atomic propositions.
"
  (let* ((column-vars (loop for i from 1 to n collect (make-column-var i)))
		 (row-values (loop for i from 1 to n collect (make-row-value i)))
		 (decision-variable-descriptions
		  (loop for i in column-vars collect 
			   `(,i :domain 
					,(loop for j in row-values collect `(,j :cost 1)))))
		 (state-variable-descriptions
		  (loop for i in column-vars collect 
			   `(,i :domain 
					,(loop for j in row-values collect j))))
		 (state-clauses (append (make-state-clauses-neqrows column-vars row-values)
								(make-state-clauses-neqdiag column-vars row-values)))
		 (search-problem (new-opsat-problem
						  decision-variable-descriptions
						  state-variable-descriptions
						  state-clauses
						  :cost-order-function #'>
						  ))
		 (time (get-internal-run-time))
		 (solution nil))
	(isat::dynamic-backtracking-on)
	(if print-on (format t "starting search... "))
	(setq time (get-internal-run-time))
	(setq solution (find-solutions search-problem :maximum-solution-count 1))
	(setq time (/ (- (get-internal-run-time) time) internal-time-units-per-second))
	(if print-on (format t "elapsed time: ~$ secs~%" time))	
	(values solution time)
 ))

#+allegro
(defun nqueens-profiling (n)
  (prof:with-profiling (:type :time :count 500 )
	  (nqueens-test n))
  (prof:show-call-graph)
  (prof:show-flat-profile)
  (prof:show-call-counts)
  )

(defun nqueens-data-collection ()
	(let ((queens '(4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20)))
	  (loop for queen in queens do
		   #+allegro (excl:gc t)
		   (multiple-value-bind (solution time)
			   (nqueens-test queen :print-on nil)
			   (declare (ignore solution))
			   (format t "~a, ~3$~%" queen time)))))


(defun proclaim-optimize ()
  "Call this method prior to compilation
to enable various compile-time optimizations"
  (proclaim '(optimize 
			  (safety 1)
			  (space 3)
			  (speed 3)
			  (debug 0))))
