;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung

(in-package :mtk-qsp)

(defclass qsp (stn)
  ((state-constraint-ht-edge :type hash-table
			     :accessor get-state-constraint-ht-edge
			     :initform (make-hash-table)
			     :documentation "")))

(defmethod get-state-constraints ((qsp qsp))
  (loop for c_i being the hash-values in (get-state-constraint-ht-edge qsp)
      collect c_i))

(defmethod get-edges-with-state-constraint ((qsp qsp))
  (loop for e_i being the hash-key in (get-state-constraint-ht-edge qsp)
       collect e_i))

(defmethod get-state-constraint ((e_i mtk-graph:directed-edge) (qsp qsp))
  (gethash e_i (get-state-constraint-ht-edge qsp)))

(defmethod mtk-graph-graph2dot:dot-print-graph
    ((qsp qsp)
     &key
     (stream t)
     (vertex-labeler #'mtk-graph-graph2dot:dot-generate-vertex-label)
     (edge-labeler #'dot-generate-qsp-edge-label)
     (rest qsp))
  ;;(format t "QSP dot-print-graph: ~A.~%" edge-labeler)
  (call-next-method qsp
		    :stream stream
		    :vertex-labeler vertex-labeler
		    :edge-labeler edge-labeler
		    :rest rest))

(defmethod dot-generate-qsp-edge-label ((e_i mtk-graph:directed-edge)
					&rest rest)
  ;;(format t "QSP Labeler was called.~%")
  (let* ((qsp (first rest))
	 (state-constraint (get-state-constraint e_i qsp)))
    (if state-constraint
	(format nil "~A \\n ~A"
		(mtk-csp-tcsp-tcsp2dot:dot-generate-tcsp-edge-label e_i qsp)
		(mtk-csp:get-relation state-constraint))
      (mtk-csp-tcsp-tcsp2dot:dot-generate-tcsp-edge-label e_i qsp))))

;;-------------------------------------------------------------------------------
;; Print object
;;-------------------------------------------------------------------------------

(defmethod print-object ((p qsp) stream)
  (format stream "(:variables ~A :temporal-constraints ~A :state-constraints ~A)"
	  (mapcar #'(lambda (v_i)
		      (format nil "(~A :domain ~A)" 
			      v_i (mtk-csp:get-domain v_i p)))
		  (mtk-csp:get-variables p))
	  (mtk-csp:get-constraints p)
	  (get-state-constraints p)))

;;-------------------------------------------------------------------------------
;; Delete variable(s) from a QSP
;;-------------------------------------------------------------------------------

(defmethod delete-variable (variable (qsp qsp))
  (dolist (e_i (mtk-graph:find-edges-with-vertex variable qsp))
    (remhash e_i (get-state-constraint-ht-edge qsp)))
  (call-next-method variable qsp))

(defmethod delete-vertex (vertex (qsp qsp))
  (delete-variable vertex qsp))

;;-------------------------------------------------------------------------------
;; Add constraint(s) to a QSP
;;-------------------------------------------------------------------------------

(defmethod make-state-constraint (&key scope ((:relation clause)))
  " Constraint description consists of two key-value pairs, where
      <state-constraint>  := (:scope <scope> :relation <clause>)
      <scope>             := (<variable> <variable>)
      <clause>            := (<literal>+)
      <literal>           := <atom> | (not <atom>)
      <atom>              := (= <variable> <value>)

      Assumptions:"
  ;; Must provide the scope and the relation
  (assert (and scope clause))
  ;; Only unary state constraint is supported.
  (assert (eq (length scope) 1))
  (let ((relation clause))
    ;; Only create a state constraint if the relation is not an empty list and
    ;; not a list of an empty set.
    (if (and relation (first relation))
	(make-instance 'mtk-csp:constraint :scope scope :relation relation)
	nil)))
  

(defmethod make-qsp-constraints (&key scope relation (state-constraint nil))
  " Constraint description consists of two key-value pairs, where
      <scope>             := (<variable>+)
      <relation>          := (<interval>+)
      <state-constraint>  := (:scope <scope> :relation <clause>)
      <interval>          := (:lower-bound <lower-bound>
                              :upper-bound <upper-bound>)
      <lower-bound>       := (<number> :inequality <lb-inequality>) | (<number>)
      <upper-bound>       := (<number> :inequality <ub-inequality>) | (<number>)
      <lb-inequality>     := #'> | #'>=
      <ub-inequality>     := #'< | #'<=
      <clause>            := (<literal>+)
      <literal>           := <atom> | (not <atom>)
      <atom>              := (= <variable> <value>)

      Assumptions:"
  (values (funcall #'make-temporal-constraint :scope scope :relation relation)
	  (if state-constraint
	      (apply #'make-state-constraint state-constraint)
	    nil)))
	   
(defmethod add-constraint ((constraint-description list) (qsp qsp))
  " Constraint description consists of two key-value pairs, where
      <constraint>        := (:scope (<variable> <variable>)
                              :relation (<interval>+)
                              :state-constaint <state-constraint>)
      <interval>          := (:lower-bound <lower-bound>
                              :upper-bound <upper-bound>)
      <lower-bound>       := (<number> :inequality <lb-inequality>) | (<number>)
      <upper-bound>       := (<number> :inequality <ub-inequality>) | (<number>)
      <lb-inequality>     := #'> | #'>=
      <ub-inequality>     := #'< | #'<=
      <state-constraint>  := (:scope <scope> :relation <clause>)
      <clause>            := (<literal>+)
      <literal>           := <atom> | (not <atom>)
      <atom>              := (= <variable> <value>)

      Assumptions:"
  (multiple-value-bind (temporal-constraint state-constraint)
      (apply #'make-qsp-constraints constraint-description)
    (let ((e_i (add-constraint temporal-constraint qsp)))
      (if state-constraint
	  (setf (gethash e_i (get-state-constraint-ht-edge qsp)) 
		state-constraint))
      e_i)))

;;-------------------------------------------------------------------------------
;; Delete constraint(s) from a QSP
;;-------------------------------------------------------------------------------

(defmethod delete-constraint ((c_i mtk-csp:constraint) (qsp qsp))
  (multiple-value-bind (c_i e_i lb-ub) (call-next-method c_i qsp)
    (remhash e_i (get-state-constraint-ht-edge qsp))
    (values c_i e_i lb-ub)))
