;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

(in-package :TPN)

;;; Copyright (c) 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

;;; ***********************************************************************
;;;
;;;                Explanation for a Temporal Plan Networks
;;;
;;; ***********************************************************************


;;; ***********************************************************************
;;;
;;;                          To Do
;;;
;;; ***********************************************************************

;;; General fixes:

;;; o If the user decides to replace exclusion with interval-relaxation in the TPN,
;;;   then we need to remove exlcusion and add interval-relaxation to
;;;   (opsat-problem-base-constraints search-problem).

;;; o Base-constraints in OPSAT has a different meaning then
;;;   we've used in constraint0system.  For constraint-system
;;;   change the name to something like fixed-constraints,
;;;   or required-constraints.

;;; o In essential-constraint and essential-constraint-system
;;;   replace :source with :support.

;;; o When TPN-representative-explanations-with-explained-exclusions
;;;   and related TPN selection problem functions are called, 
;;;   create a selection problem if it doesn't exist, and 
;;;   update it if the TPN has been modified.

;;; General Optimizations:
;;; o Add the discovered conflicts to the OPSAT selection problem.
;;    It can't hurt.

;;; To Do:
;;; o Should choice be in internal or external form?

;;; ***********************************************************************
;;;
;;;     Interface Function: Compute Representative Explanations
;;;
;;; ***********************************************************************

;;; Fix:
;;; o This explanation function should only work when we have a full
;;;   assignment of choices to TPN decision events.
;;;   - If only a partial assignment exists, then consistency would need
;;;     to search over unassigned variables, but ITC doesn't do this.
;;;   - Instead ITC is checking consistency against a relaxed problem.
;;;   -> The solution is to extend consistency checking to search over
;;;      the unassigned variables.
;;;      o Option 1:  Use OPSAT to check consistency, and elevate the
;;;          interface to OPSAT as the constraint system.  This is the
;;;          simplest, and most transparent option.  The problem is that
;;;          state information for consistency checking and OPSAT search
;;;          might collide.
;;;      - Option 2: Extend ITC to a DTP solver, by coupling ITC to a
;;;                  DFS version of BFIG/search.

;;; Check:
;;; o Assert and unassert take different representations for choices
;;;   as inputs.  Is this correct?
(defmethod TPN-representative-explanations-with-explained-exclusions

	((tpn tpn-w-selection) &key (choices '()) (perform-interval-relaxation? T))
  
  "TPN is a temporal plan network, for which a choice-selection 
   problem has already been formulated.  TPN has a corresponding
   set of constraints, divisible into relaxable and fixed.

   CHOICES is a list of assignments of the form:
          (= <decision-event-name> <choice-number>),

   where 
      <decision-event-name (normally a number) names a TPN choice, and 
      <choice-number>, a positive integer, denotes one of the choices.

   These assignments are not removed during relaxation.
   CHOICES defaults to nil, denoting no assignments.

   Computes explanations for the current TPN (i.e., currently inserted constraints)
   with CHOICES asserted.

   Returns (values REPRESENTATIVE_EXPLANATIONS MINIMAL_CONFLICTS)
   where
      MINIMAL-CONFLICTS is a (partial) list of minimal conflicts
      such that each MINIMIMAL_CONFLICT_I u BASE_CONSTRAINTS is inconsistent.


      EXPLANATION_I in REPRESENTATIVE_EXPLANATIONS
         is a triple <R_i, IR_i, Exp_IR_i>   if perform-interval-relaxation?
                     <R_i, Exc_i, Exp_Exc_i> Otherwise.
                     
   where R_i subset RELAXABLE-CONSTRAINTS, is a maximal relaxation of 
         such that R_i u BASE-CONSTRAINTS is consistent,

   for PERFORM-INTERVAL-RELAXATION? = nil:
         Exc_i is the exclusion set of R_i (relaxable constraints - R_i),
         Exp_Exc_i is a list of explanations EE_ij 
           for each exclusion Exc_ij in Exc_i
           where EE_IJ = <exc_ij, explanation_ij> such that
               explanation_ij implies (not exc_ij)

   for PERFORM-INTERVAL-RELAXATION? = T:
         IR_i is the interval relaxation of exclusion Exc_i.
         Exp_IR_i is ...

         IR_i is a list of relaxed STCs 
            whose [l,u] bounds have been minimally weakened,
            where R_i u IR_i u BASE-CONSTRAINTS is consistent.

		 Exp_IR_i justify IRI_i and are triples
            <INTERVAL-RELAXATION IR_ij, EXCLUSION E_ij, Exp_IR_ij>.

         where
            EXCLUSION E_ij is an STC (s,e)[l,u] that is inconsistent 
               with BASE u RELAXATION IR_i,

            INTERVAL-RELAXATION IR_ij is an STC (s,e)[l',u']
               IR_ij is a relaxation of E_ij (i.e., [l,u] subset [l',u']) 
                  that is minimal (i.e., no [l'',u''] exists that is a 
                  consistent relaxation.
                  NOTE: we then tighten l'' and u'' of IR_ij to eliminate values 
                  inconsistent with IR_i u R_i u BASE-CONSTRAINTS.

            Exp_IR_ij is a minimal subset of R_i
            such that BASE-CONSTRAINTS u Exp_IR_ij implies (not E_ij).

   A set of explanations is representative if it contains a relaxation containing
   constraint C when C appears in some maximal relaxation, and if it contains an
   exclusion set or interval relaxations containing constraint C, when C appears 
   in some minimal exclusion set.

   Constraints are relaxed by removing the constraints or, for simple temporal 
   constraints, by expanding their interval constraint (interval relaxation).

   Assumptions:
   - No choices are asserted within TPN selection problem upon entry.
   - Should only relax relaxable constraints that are currently inserted.

   Side Effects:
   - Creats a TPN selection problem if it doesn't already exist.
   - Inserted constraints are restored to the same state as upon entry.
   - All asserted assignments are removed, same as upon entry.
   - the state of the STN of the TPN selection problem is 
     changed during the function, but restored at the end. "

  ;; 0) Ensure that a TPN selection problem has been recorded.
  (establish-tpn-selection-problem! tpn)

  ;; 1) Assert the choices.  TPN's STN starts with
  ;;    all TPN STCs asserted, and no choices asserted.
  (let* ((STN (tpn-constraint-system tpn))

		 ;; Snapshot of the inserted constraints, 
		 ;; used to restore constraint-system state at function end.
		 (constraint-snapshot 
		  (copy-list (constraint-system-inserted-constraints STN)))

		 ;; Assignment objects corresponding to choices.
		 (choice-assignments
		  (assert-choices-for-representative-explanation! tpn choices)))
	
	(debug-print-for :tpn t "Explanation for inconsistent choices ~a of tpn ~a"
					 choice-assignments tpn)
	(debug-print-for :tpn t "  Setting up ...")
	(debug-print-for 
	 :tpn t "    Snap shot of inserted constraints before call = ~a"
	 constraint-snapshot)
	
	;; 2) Identify the base and relaxable constraints, corresponding to 
	;;    constraints introduced by required and relaxable episodes.
	;;    Note: assignments are not enforced by constraints, hence we don't
	;;          need to worry about them.
	(let (
		  ;; All base constraints should, by definition be inserted.
		  (base-constraints 
		   (constraint-system-base-constraints STN))
			
		  ;; Only consider relaxing relaxable constraints that are
		  ;; currently inserted.
		  (relaxable-constraints
		   (loop for c in (constraint-system-constraints STN)
			   unless (or (constraint-removed? c) (null (stc-lb-relaxable? c)) (null (stc-ub-relaxable? c))) collecting c)))
	  
	  (debug-print-for :tpn t "    Unchangeable base constraints = ~a."
					   base-constraints)
	  (debug-print-for :tpn t "    Relaxable constraints = ~a."
					   relaxable-constraints)

	  ;; 3) Find the representative explanations, by fully relaxing excluded STCs.

	  ;;    Each explanation is a triple <relaxation, exclusion, 
	  ;;    exclusion-explanations>.

	  ;;    Relaxation is a maximal subset of relaxable constraints that is
	  ;;    consistent with of base-constraints.

	  ;;    Exclusion is the set of constraints not included in relaxation.
		
	  ;;    Each exclusion explanation is a pair (exclusion_i, exc_explanation_i)
	  ;;    such that for exclusion_i in exclusion, 
	  ;;    exc_explanation_i implies (not exclusion_i).

	  ;; FIX: Check what the explanation functions assume about inserted 
	  ;;      constraints and check their side-effects.

	  (debug-print-for 
	   :tpn t "  Computing maximal relaxations as fully relaxed constraints.")
		
	  ;; Note: - The representative-explanation routines assume choices are 
	  ;;         already set.
	  ;;       - May modify the set of inserted constraints.
	  (multiple-value-bind (representative-explanations minimal-conflicts)
		  (Representative-Explanations-with-Explained-Exclusions
		   STN relaxable-constraints
		   :base-constraints base-constraints)

		(debug-print-for :tpn t "    Representative explanations = ~a."
						 representative-explanations)

		(debug-print-for :tpn t "    Conflicts = ~a."
						 minimal-conflicts)

		;; 4) Given representative explanations, compute a new set of 
		;;    tightened explanations that replaces each excluded constraint 
		;;    with a relaxed simple temporal constraints (added to the 
		;;    relaxation) whose interval has been widened.
		(when perform-interval-relaxation?
		  (debug-print-for 
		   :tpn t "  Tightening maximal relaxations to interval relaxations.")
		  (setq representative-explanations
			(tighten-exclusions-of-explanations STN base-constraints 
					representative-explanations :choices choices))

		  (debug-print-for
		   :tpn t "    Interval-relaxed representative explanations = ~a."
		   representative-explanations))

		;; 6) Map each explanation to its corresponding TPN elements.

		;;    For now, just return the explanations in terms of 
		;;    temporal constraints, while leaving it to the caller to perform 
		;;    further processing.

		;; 7) Unassert Choices.  Assume that STN started with no choices asserted,
		;;    OPSAT guarantees that all assignments it makes have been removed.
		(unassert-choices-for-representative-explanation! tpn choice-assignments)
		  
		;; 8) Restore inserted constraints.
		(establish-constraints! STN constraint-snapshot)

		;; 9) Return the representative explanations.
		(values representative-explanations minimal-conflicts)))))

;;; ***********************************************************************
;;;
;;;                          Support Functions
;;;
;;; ***********************************************************************

;;; ***********************************************************************
;;;
;;;          Asserting and Removing Choice Assignments
;;;
;;; ***********************************************************************

(defmethod assert-choices-for-representative-explanation!
	((tpn tpn-w-selection) choices)
  
  "Let TPN be a temporal plan network with corresponding choice
   selection problem.

   Let CHOICES be a list of TPN choice assignments to be asserted,
   where each choice assignment is of the form:
      (= <decision-event-name> <choice-value>)

   Returns the list of internalized choice-assignments that
   have been inserted.

   Side Effect: Ensures that Choices and TPN's STCs are asserted.

   Assumption: TPN's STCs are already asserted.
               No choices are asserted."
  
  ;; Note: Upon entry, OPSAT guarantees that it has unasserted
  ;;       any assignments asserted into selection problem.
  
  (let* ((problem (tpn-selection-problem tpn))
		 (choice-assignments

		  ;; Check that each choice is legal, and collect
		  ;; the corresponding assignment object.
		  (loop for choice in choices
			  for variable-name = (choice-assignment-variable choice)
			  for value = (choice-assignment-value choice)
			  collecting (get-constraint-assignment-of-search-problem
						  problem variable-name value))))

	;; Assert each assignment into problem's constraint-system.
	(loop for assignment in choice-assignments
		do (insert-assignment-into-constraint-system 
			problem
			'TPN-representative-explanations-with-explained-exclusions
			assignment
			*true*))
	;; Return the internalize choice assignments.
	(values choice-assignments)))

(defmethod unassert-choices-for-representative-explanation!
	((tpn tpn-w-selection) choice-assignments)
  
  "Let TPN be a temporal plan network with corresponding choice
   selection problem.

   Let CHOICE-ASSIGNMENTS be a list of internalized TPN choice assignment
   objects to be removed.

   Side Effect: Ensures that CHOICE-ASSIGNMENTS hace been removed.

   Assumption: TPN's STCs have already been restored to their
               original state."

  ;; Notes on restoring problem state:
  ;;    - Choice Selection (Opsat) guarantees that no assignments were asserted.
  ;;    - The explanation functions are designed to ensure that all relaxed
  ;;      constraints have been restored (Step 3).
  ;;    - Step 4 is designed to ensure that any STC partial relaxations have
  ;;      been removed, and their corresponding relaxed constraint has been 
  ;;;     restored.
  ;;    - ITC's incremental functions ensure that its state is properly
  ;;      maintained as constraints are added and removed, hence ITC's 
  ;;      state is correct.
  ;;    - This function removes the choice assignments, thus ensuring that 
  ;;      all assignments are removed, and Problem's state is correctly restored.
  
  (let ((problem (tpn-selection-problem tpn)))

	;; Removing each assignment from problem's constraint-system.
	(loop for assignment in choice-assignments
		do (remove-assignment-from-constraint-system 
			problem 'TPN-representative-explanations-with-explained-exclusions
			assignment *true*))
	
	T))

;;; ***********************************************************************
;;;
;;;  Computing Interval Relaxations, by Tightening Excluded Constraints
;;;
;;; ***********************************************************************

(defmethod tighten-exclusions-of-explanations
	((STN itc-stn) base-constraints explanations &key (choices '()))

  "Let STN be a Simple Temporal Network, and
         EXPLANATIONS be a list of explanations for STN (with asserted choices).
         Each explanation is a triple (R_i, Exc_i, Exp_Exc_i>

      where R_i subset relaxable constraints, is a maximal relaxation of 
               relaxable constraints u base constraints,
            Exc_i is the exclusion set of R_i (relaxable constraints - R_i), and
            Exp_Exc_i is a list of explanations EE_ij 
              for each exclusion Exc_ij in Exc_i
              where EE_IJ = <exc_ij, explanation_ij> such that
              explanation_ij implies (not exc_ij).

   Returns TIGHTENED-EXPLANATION of STN, a triple
      <RELAXATION, INTERVAL-RELAXATIONS, INTERVAL-RELAXATION-EXPLANATIONS>,
   where
         RELAXATION is a list of relaxed STCs that are unmodified,

         INTERVAL-RELAXATIONS is a list of relaxed STCs 
            whose [l,u] bounds have been minimally weakened,
            where RELAXATION u INTERVAL-RELAXATIONS u BASE-CONSTRAINTS 
              is consistent.

		 INTERVAL-RELAXATION-EXPLANATIONS justify INTERVAL-RELAXATIONS and are 
         triples
            <INTERVAL-RELAXATION I, EXCLUSION, IR_EXPLANATION>.

         where
            EXCLUSION E is an STC (s,e)[l,u] that is inconsistent with 
              BASE u RELAXATION,

            INTERVAL-RELAXATION I is an STC (s,e)[l',u']
               I is a relaxation of E (i.e., [l,u] subset [l',u']) 
                  that is minimal (i.e., no [l'',u''] exists that is a 
                  consistent relaxation.
                  NOTE: we then tighten l'' and u'' of I to eliminate values 
                  inconsistent with TIGHT-RELAXATION u BASE CONSTRAINTS.

            IR_EXPLANATION is a minimal subset of RELAXATION
            such that BASE u IR_EXPLANATION implies (not EXCLUSION).

   Side Effects: 
     Modifies the set of constraints inserted into STN."

  (loop for explanation in explanations
	  collecting (tighten-exclusions-of-explanation 
				  STN base-constraints explanation :choices choices)))

;;; To Do:
;;;  - we don't need exclusions and exclusion-explanations listed separately,
;;;    exclusions can be paired up with their explanations.

;;; NOTE: We could, but don't, tighten l'' and u'' of I to eliminate values 
;;;      inconsistent with TIGHT-RELAXATION u BASE CONSTRAINTS.


(defmethod continuous-constraint-relaxation
	(tpn explanation &key 
	 (choice-assignments '())
	 (original-choice-nodes-assignments '())
	 (active-constraints '())
	 (inactive-constraints '())
	 (event-list '())
	 (event-pair-value-table (make-hash-table :test 'equal))
	 (inactive-event-pair-value-table (make-hash-table :test 'equal))
	 (out-constraints '())
	 (relaxed-constraints '())
	 (constraint-snapshot nil)
	 (start-event nil)
	 (end-event nil))


  ;; 0. Make a copy of the current constraint system
  (setf constraint-snapshot 
   (copy-list (constraint-system-inserted-constraints 
			   (tpn-constraint-system tpn))))



  ;; 1. Get current choice assignments from explantion
  (setf choice-assignments
		
		;; Check that each choice is legal, and collect
		;; the corresponding assignment object.
		(loop for choice in explanation
		   for variable-name = (choice-assignment-variable choice)
		   for value = (choice-assignment-value choice)
		   collecting (get-constraint-assignment-of-search-problem
					   (tpn-selection-problem tpn) variable-name value)))

  (loop for choice-assignment in choice-assignments
	 do (when (and (not (equal (ITC::state-assignment-value choice-assignment) 'IN)) 
				   (not (equal (ITC::state-assignment-value choice-assignment) 'OUT)))
		  (push choice-assignment original-choice-nodes-assignments)))
  
  ;;(format t "Choice Assignments: ~A~%" choice-assignments)

  ;; 2. Assert the current choice assignments
  (loop for assignment in choice-assignments
	 do (insert-assignment-into-constraint-system 
		 (tpn-selection-problem tpn)
		 'TPN-representative-explanations-with-explained-exclusions
		 assignment
		 *true*))


  ;; 3. Get active constraints
  (loop for constraint in (constraint-system-constraints 
						   (opsat-problem-constraint-system (tpn-selection-problem tpn)))
	 do (if (equal T (ITC::stc-active? constraint))
		  (push constraint active-constraints)
		  (push constraint inactive-constraints)))

  ;;(format t "Active Constraints: ~A~%" active-constraints)

  ;; 3.1 Build a hash table to store inactive constraints
  (loop for constraint in inactive-constraints
	 do (setf (gethash 
			   (format nil "~A ~A" (ITC::event-name (ITC::stc-start constraint))
					   (ITC::event-name (ITC::stc-end constraint)))
			   inactive-event-pair-value-table) 
			   (ITC::stc-ub constraint))
	   
	   (setf (gethash 
			  (format nil "~A ~A" (ITC::event-name (ITC::stc-end constraint))
					  (ITC::event-name (ITC::stc-start constraint))) 
			  inactive-event-pair-value-table) 
			 (ITC::b- (ITC::stc-lb constraint))))

  ;; 4. Build hash table of all active constraints
  (loop for constraint in active-constraints
	 do (when (not (onlist? (ITC::stc-start constraint) event-list))
		  (push (ITC::stc-start constraint) event-list))
	   (when (not (onlist? (ITC::stc-end constraint) event-list)) 
		 (push (ITC::stc-end constraint) event-list))

	   (if (not (equal (ITC::stc-guard constraint) *true*))
		 
		 (if (equal (ITC::state-assignment-value (ITC::stc-guard constraint))
					'OUT)
			 (progn 
			   
			   (setf (gethash 
					  (format nil "~A ~A" (ITC::event-name (ITC::stc-start constraint))
							  (ITC::event-name (ITC::stc-end constraint)))
					  event-pair-value-table) 
					 ITC::*infinity*)
			   
			   (setf (gethash 
					  (format nil "~A ~A" (ITC::event-name (ITC::stc-end constraint))
							  (ITC::event-name (ITC::stc-start constraint))) 
					  event-pair-value-table) 
					 ITC::*infinity*)
			   
			   (push constraint out-constraints))


			 (progn
			   
			   (setf (gethash 
					  (format nil "~A ~A" (ITC::event-name (ITC::stc-start constraint))
							  (ITC::event-name (ITC::stc-end constraint))) 
					  event-pair-value-table) 
					 (ITC::stc-ub constraint))
			   
			   (setf (gethash 
					  (format nil "~A ~A" (ITC::event-name (ITC::stc-end constraint))
							  (ITC::event-name (ITC::stc-start constraint))) 
					  event-pair-value-table) 
					 (ITC::b- (ITC::stc-lb constraint)))))
			 
			 (progn
			   
			   (setf (gethash 
					  (format nil "~A ~A" (ITC::event-name (ITC::stc-start constraint))
							  (ITC::event-name (ITC::stc-end constraint))) 
					  event-pair-value-table) 
					 (ITC::stc-ub constraint))
			   
			   (setf (gethash 
					  (format nil "~A ~A" (ITC::event-name (ITC::stc-end constraint))
							  (ITC::event-name (ITC::stc-start constraint))) 
					  event-pair-value-table) 
					 (ITC::b- (ITC::stc-lb constraint))))))


  ;; 5. Fill the rest of the table with infinity
  (loop for start-event-name in event-list
	 do (loop for end-event-name in event-list
		   do (when (equal (gethash 
							(format nil "~A ~A" (ITC::event-name start-event-name) 
									(ITC::event-name end-event-name))
							event-pair-value-table) 'NIL)
				(setf (gethash 
					   (format nil "~A ~A" (ITC::event-name start-event-name) 
							   (ITC::event-name end-event-name))
					   event-pair-value-table) ITC::*infinity*))))


  ;; 6. Start Floyd-Warshall
  (loop for middle-event-name in event-list               
	 do (loop for end-event-name in event-list
		   do (loop for start-event-name in event-list
				   
				 do (when (ITC::b> (gethash 
									(format nil "~A ~A" start-event-name end-event-name) 
									event-pair-value-table) 
								   (ITC::b+ (gethash 
											 (format nil "~A ~A" start-event-name middle-event-name) 
											 event-pair-value-table) 
											(gethash 
											 (format nil "~A ~A" middle-event-name end-event-name) 
											 event-pair-value-table)))
					  
					  (setf (gethash 
							 (format nil "~A ~A" start-event-name end-event-name) 
							 event-pair-value-table) 
							(ITC::b+ (gethash 
									  (format nil "~A ~A" start-event-name middle-event-name) 
									  event-pair-value-table) 
									 (gethash (format nil "~A ~A" middle-event-name end-event-name) 
											  event-pair-value-table)))))))
  
  ;; 7. Generate Results
  

  (loop for constraint in out-constraints
	 do (setf start-event (ITC::stc-start constraint))
	   (setf end-event (ITC::stc-end constraint))

	   (push (list (list start-event end-event)
				   (list 
					(ITC::b- (gethash 
							  (format nil "~A ~A" end-event start-event) 
							  inactive-event-pair-value-table)) 
					(gethash 
					 (format nil "~A ~A" start-event end-event) 
					 inactive-event-pair-value-table)
					(ITC::b- (gethash 
							  (format nil "~A ~A" end-event start-event) 
							  event-pair-value-table)) 
					(gethash 
					 (format nil "~A ~A" start-event end-event) 
					 event-pair-value-table))) relaxed-constraints))
  
  (setf relaxed-constraints (list original-choice-nodes-assignments relaxed-constraints))

  ;; 8.  Unassert Choices.  Assume that STN started with no choices asserted,
  ;;    OPSAT guarantees that all assignments it makes have been removed.
  (unassert-choices-for-representative-explanation! tpn choice-assignments)
  
  ;; 9. Restore inserted constraints.
  (establish-constraints! (tpn-constraint-system tpn) constraint-snapshot)

relaxed-constraints)


(defmethod tighten-exclusions-of-explanation 
	((STN itc-stn) base-constraints explanation &key 
	 (choices '())
	 (relaxation nil)
	 (exclusions nil)
	 (interval-relaxation nil)
	 (interval-relaxation-explanation nil)
	 (interval-relaxations nil)
	 (interval-relaxation-explanations nil)
	 (start-event nil)
	 (end-event nil)
	 (lb nil)
	 (ub nil)
	 (event-pair-value-table nil)
	 (event-list nil))

  "Let STN be a Simple Temporal Network and,
       EXPLANATION of STN be a triple <R_i, Exc_i, Exp_Exc_i>.

      where R_i subset relaxable constraints, is a maximal relaxation of 
               relaxable constraints u base constraints,
            Exc_i is the exclusion set of R_i (relaxable constraints - R_i), and
            Exp_Exc_i is a list of explanations EE_ij 
              for each exclusion Exc_ij in Exc_i
              where EE_IJ = <exc_ij, explanation_ij> such that
              explanation_ij implies (not exc_ij).

   Returns TIGHTENED-EXPLANATION of STN, a triple
      <RELAXATION, INTERVAL-RELAXATIONS, INTERVAL-RELAXATION-EXPLANATIONS>,

   where
         RELAXATION is a list of relaxed STCs that are unmodified,

         INTERVAL-RELAXATIONS is a list of relaxed STCs 
            whose [l,u] bounds have been minimally weakened,
            where RELAXATION u INTERVAL-RELAXATIONS u BASE-CONSTRAINTS 
              is consistent.

		 INTERVAL-RELAXATION-EXPLANATIONS justify INTERVAL-RELAXATIONS and are 
         triples
            <INTERVAL-RELAXATION I, EXCLUSION, IR_EXPLANATION>.

         where
            EXCLUSION E is an STC (s,e)[l,u] that is inconsistent 
              with BASE u RELAXATION,

            INTERVAL-RELAXATION I is an STC (s,e)[l',u']
               I is a relaxation of E (i.e., [l,u] subset [l',u']) 
                  that is minimal (i.e., no [l'',u''] exists that is a 
                  consistent relaxation.

            IR_EXPLANATION is a minimal subset of RELAXATION u INTERVAL-RELAXATIONS
            such that BASE u IR_EXPLANATION implies (not EXCLUSION).

   Using Floyd Warshall to find the tightest bound now.
               
   Side Effects: 
     Modifies the set of constraints inserted into STN."

  ;; temporal call to variable base-constraint to avoid warnings.
  (push '() base-constraints)
  (first base-constraints)

     
   ;; Initialize variables
  (setf relaxation (first explanation))
  (setf exclusions (second explanation))
       ;;     (format t "Exclusions: ~A ~%" exclusions)
  
  ;; Iterate through all exclusion in the current explanation
  (loop for exclusion in exclusions
      ;; do (format t "~A ~%" (ITC::constraint-removed? exclusion))
      do (when (and (equal NIL NIL) (not (and (equal (ITC::stc-lb exclusion) 0.0) (equal (ITC::stc-ub exclusion) 0.0))))
        ;; Get the constraint that needs relaxation.
			(setf start-event (ITC::stc-start exclusion))
			(setf end-event (ITC::stc-end exclusion))
			(setf lb (ITC::stc-lb exclusion))
			(setf ub (ITC::stc-ub exclusion))

         
            ;; [0,0] constraint will not be relaxed

			(when (and (not (and (ITC::b= lb 0) (ITC::b= ub 0))) (not (and (ITC::b= lb 0) (ITC::b= ub ITC::*infinity*))))

      ;;     (format t "Relaxing ~A ~A:  ~A - ~A~%" start-event end-event lb ub)
           ;; Create a hash table to store APSP values
           (setq event-pair-value-table NIL) 
           (setq event-pair-value-table (make-hash-table :test 'equal))
           (setq event-list NIL) 

 ;;;                do (format t "Looping.....~A     start: ~A   end:  ~A   lb:  ~A   ub:   ~A~%" (ITC::constraint-removed? exclusion) start-event end-event lb ub)
       
           ;; Add existing constraint values to the table
           (loop for constraint in (append (first explanation) (second explanation))
               do (when (match (ITC::stc-guard constraint) choices)
                    (when (not (onlist? (ITC::stc-start constraint) event-list))
                      (push (ITC::stc-start constraint) event-list))
                    (when (not (onlist? (ITC::stc-end constraint) event-list)) 
                      (push (ITC::stc-end constraint) event-list))
                    ;;(format t "LIST     ~A~%" event-list)
;;;                    (format t "Adding ~A ~A~%" constraint (gethash 
;;;                                (format nil "~A ~A" (ITC::event-name (ITC::stc-start constraint))
;;;                                  (ITC::event-name (ITC::stc-end constraint))) event-pair-value-table))
                    (setf (gethash 
                           (format nil "~A ~A" (ITC::event-name (ITC::stc-start constraint))
                             (ITC::event-name (ITC::stc-end constraint))) event-pair-value-table) 
                      (ITC::stc-ub constraint))
                    (setf (gethash 
                           (format nil "~A ~A" (ITC::event-name (ITC::stc-end constraint))
                             (ITC::event-name (ITC::stc-start constraint))) event-pair-value-table) 
                      (ITC::b- (ITC::stc-lb constraint)))))
           
           (loop for start-event-name in event-list
               do (loop for end-event-name in event-list
                      do (when (equal (gethash 
                                           (format nil "~A ~A" (ITC::event-name start-event-name) (ITC::event-name end-event-name))
                                           event-pair-value-table) 'NIL)
						   (setf (gethash 
                                (format nil "~A ~A" (ITC::event-name start-event-name) (ITC::event-name end-event-name))
                                event-pair-value-table) ITC::*infinity*))))
           
           ;; Reset the constraint that needs relaxation
           (setf (gethash (format nil "~A ~A" start-event end-event) event-pair-value-table) ITC::*infinity*)
           (setf (gethash (format nil "~A ~A" end-event start-event) event-pair-value-table) ITC::*infinity*)
           
           ;; Start Floyd-Warshall
           (loop for middle-event-name in event-list               
                 do (loop for end-event-name in event-list
                      do (loop for start-event-name in event-list

                             do (when (ITC::b> (gethash (format nil "~A ~A" start-event-name end-event-name) 
                                                        event-pair-value-table) 
                                               (ITC::b+ (gethash (format nil "~A ~A" start-event-name middle-event-name) 
                                                                 event-pair-value-table) 
                                                        (gethash (format nil "~A ~A" middle-event-name end-event-name) 
                                                                 event-pair-value-table)))

                                  (setf (gethash (format nil "~A ~A" start-event-name end-event-name) 
                                                 event-pair-value-table) 
                                    (ITC::b+ (gethash (format nil "~A ~A" start-event-name middle-event-name) 
                                                      event-pair-value-table) 
                                             (gethash (format nil "~A ~A" middle-event-name end-event-name) 
                                                      event-pair-value-table)))))))
           
           ;; Get the updated constraint value, store it in interval-relaxation
             ;;(setf ubt (gethash (format nil "~A ~A" start-event end-event) event-pair-value-table))
            ;; (setf lbt (gethash (format nil "~A ~A" end-event start-event) event-pair-value-table))
          ;; (format t "~A--~A: ub ~A lb ~A~%" start-event end-event ubt lbt)
           (when (not (ITC::b< (gethash (format nil "~A ~A" start-event end-event) event-pair-value-table) (ITC::b- (gethash (format nil "~A ~A" end-event start-event) event-pair-value-table))))	
             (setf ub (gethash (format nil "~A ~A" start-event end-event) event-pair-value-table))
             (setf lb (gethash (format nil "~A ~A" end-event start-event) event-pair-value-table))
             (when (ITC::b< ub 0)
               (setf ub 0)
               (setf lb 0)
               )
           ;;(format t "~A--~A: ub ~A lb ~A~%" start-event end-event ub lb)
   
             (setf interval-relaxation (new-constraint stn :start start-event :end end-event :lb (ITC::b- lb) :ub ub))
             (setq interval-relaxation-explanation (list interval-relaxation (first exclusions) explanation))
             (push interval-relaxation interval-relaxations)
             (push interval-relaxation-explanation interval-relaxation-explanations))
           
)))

  ;; Return each of the following:
  
  ;; RELAXATION is a set of relaxable STCs that are retained with out 
  ;;   modification.
  
  ;; INTERVAL-RELAXATIONS are relaxable STCs whose [l,u] bounds have been 
  ;;   minimally weakened.
  
  ;; INTERVAL-RELAXATION-EXPLANATIONS explain each interval relaxation,
  ;;    and are of the form :
  ;;        (INTERVAL-RELAXATION EXCLUSION EXPLANATION)
  ;;    as described above.
  (values (list relaxation interval-relaxations interval-relaxation-explanations)))     


;; (defun match (stc-guard choices &key 
;; 			  (same? NIL) 
;; 			  (decision-event NIL) 
;; 			  (decision NIL)
;; 			  (choices-seq nil)
;; 			  (match nil)
;; 			  (dummy nil))
;;   (declare (ignore match dummy))

;;   (when (equal choices NIL)
;;     (setf same? T))
            
;;   (when (not (equal choices NIL))
;;     (loop for choice in choices
;; 	   do (setf choices-seq (format nil "~A" stc-guard))

;; 		 (multiple-value-setq (match dummy decision-event decision)
;; 		   (excl:match-regexp "\\(.*\\)=\\(.*\\)" (format nil "~a" choices-seq)))
		 
;; 	   ;;  (format t "~A ~A ~A~%" stc-guard decision-event decision)
		 
;; 		 (when (or (equal (format nil "(= ~a ~a)" decision-event decision) (format nil "~a" choice)) (equal stc-guard 'TRUE))
;; 		   (setf same? T)))
;;     )
;;   same?)


;; To Peng: Above this comment is your original function. It uses the Allegro-only package excl.
;; To make it implementation independent please switch to the function below.
;; I HAVE NOT TESTED IT YET!
;; My goal is to stop using AllegroCL dependent libraries, so not to use anything from package "excl".
;; The "cl-ppcre" package can be found in the folder "regexp-utils".


(defun match (stc-guard choices &key 
			  (same? NIL) 
			  (decision-event NIL) 
			  (decision NIL)
			  (choices-seq nil)
			  (match nil)
			  (regs nil))

  (when (equal choices NIL)
    (setf same? T))
            
  (when (not (equal choices NIL))
    (loop for choice in choices
	   do (setf choices-seq (format nil "~A" stc-guard))

		 (multiple-value-setq (match regs)
		   ;; The following regular expression is either correct or use this one:
		   ;;   "\\((.*)\\)=\\((.*)\\)" "(.*)=(.*)"
		   ;; Depending on whether you are supposed to be matching the extra parenthesis
		   (cl-ppcre:scan-to-strings "\\((.*)\\)=\\((.*)\\)" (format nil "~a" choices-seq)))
		 (when match
		   (setf decision-event (aref regs 0))
		   (setf decision (aref regs 1)) )
		 
	   ;;  (format t "~A ~A ~A~%" stc-guard decision-event decision)
		 
		 (when (or (equal (format nil "(= ~a ~a)" decision-event decision) (format nil "~a" choice)) (equal stc-guard 'TRUE))
		   (setf same? T)))
    )
  same?)



(defun onlist? (event event-list &key (on? NIL))
       ;; do  (format t "~A        ~A~%" event event-list)
  (loop for element in event-list
      do (when (equal (format nil "~A" (ITC::event-name element))
                      (format nil "~A" (ITC::event-name event)))        
           (setf on? T)))

  on?)

;;; ***********************************************************************
;;;
;;;             Minimally Weaken Simple Temporal Constraint
;;;
;;; ***********************************************************************

;;; To Do:

;;; o Can we prove whether or not interval relaxations influence 
;;;   successive interval relaxations?    
;;;   - I believe they can, because the interval relaxation 
;;;     tightens what is entailed.
;;;   - Prove whether our interval relaxations are minimal.

;;; Optimization:
;;; o To speed up the computation of explanations for interval relaxation,
;;;   - Start with the exclusion's explanation + preceding interval relaxations,
;;;     and search for a subset that is a minimal conflict.

;;; o In the following we use preferred-conflict-of-constraints
;;;   but we may want to use conflict-of-constraints, however, to do so
;;;   we need to be able to specify base-constraints to it.

;;; NOTE: We could but don't tighten l'' and u'' of I to eliminate values 
;;;       inconsistent with TIGHT-RELAXATION u BASE CONSTRAINTS.
(defmethod tightest-interval-relaxation
	
	((STN itc-stn) (exclusion itc-stc)
	 base-constraints tight-relaxation base+tight)
  
  "Let STN be a simple temporal network that supports incremental reasoning,

       BASE-CONSTRAINTS be a list of simple temporal constraints (STCs) 
          that can not be relaxed,

       TIGHT-RELAXATION be a list of STCs consistent with BASE-CONSTRAINTS.

       EXCLUSION E be an STC (s,e)[l,u] that is inconsistent with 
         TIGHT-RELAXATION u BASE-CONSTRAINTS.
 
   Returns (values INTERVAL-RELAXATION EXPLANATION) where
      INTERVAL-RELAXATION I is an STC (s,e)[l',u']
         such that I is consistent with tight relaxation and base constraints.
                   I is a relaxation of E (i.e., [l,u] subset [l',u']) 
                      that is minimal (i.e., no [l'',u''] exists that is a 
                      consistent relaxation.

      EXPLANATION is a minimal subset of TIGHT-RELAXATION 
          such that BASE u EXPLANATION implies (not EXCLUSION).

   Assumptions: Upon entry TIGHT-RELAXATION and BASE-CONSTRAINTS are inserted 
                into STN, while EXCLUSION is not.

   Side Effect:
      - STN is modified during the function.
      - BASE-CONSTRAINTS and TIGHT-RELAXATION are restored upon exit."

  ;;   Given Exclusion (s,e)[le, ur],
  ;; 5b) Compute exclusion's minimal interval relaxation.
  
  (multiple-value-bind (s e lr ur)
	  (compute-tightest-interval-relaxation exclusion stn)

	(let* (
		   ;; 5b3) Construct IR_EXPLANATION for interval-relaxation,
		   ;;      a minimal subset of TIGHT-RELAXATION such that
		   ;;      BASE u EXPLANATION implies (not EXCLUSION).
		   (IR-explanation (preferred-conflict-of-constraints
							STN tight-relaxation
							:base-constraints
							(adjoin exclusion base-constraints)))
		 
		   ;; 5b4) Assemble a support for this new STC based on explanation.
		   ;;      Support is the conflict, comprised of 
		   ;;      {exclusion} u IR_explanation.
		   (support `(interval-relaxation ,exclusion . ,IR-explanation))

		   ;; 5b5) Create a new STC for this interval relaxation.
		   (interval-relaxation (new-constraint stn :source support
												:start s :end e
												:lb lr :ub ur)))

	  ;; 5b6) Ensure that constraints are restored to the state upon entry.
	  (establish-constraints! STN base+tight)
	
	  ;; 5b7) Return the new STC and explanation.
	  (values interval-relaxation IR-explanation))))

(defmethod compute-tightest-interval-relaxation ((STC ITC-STC)(STN ITC-STN))

  "Let STN be a simple temporal network, and 
   STC be an inconsistent simple temporal constraint of STN,
   to be relaxed.

   Computes a minimal relaxation of STC that is consistent with STN.

   Returns a description of the relaxation in the form:
      (VALUES START END LB UB)
   denoting constraint (start,end [lr,ur]."

  (let ((s (stc-start STC))
		(e (stc-end STC))
		(le (stc-lb STC))
		(ue (stc-ub STC))
		lr ur)

	;; 5b1) use ITC to find the implied constraint (s,e)[li,ui].
	(multiple-value-bind (li ui)
		(event-pair-time-bound s e stn)
	  
	  ;; 5b2) compute the new relaxation (s,e)[lr,ur] as
	  ;;      - if   ue < li then (s,e)[le,li]
	  ;;      - else ui < le then (s,e)[ui,ue]
	  
	  (format t "~2%  Relaxing ~a" STC)
	  (format t "~%    Consistent range [~a,~a]" li ui)
	  (if (b< ue ui)
		  (setq lr le 
				ur ui)

		;; else li < le
		(setq lr li 
			  ur ui))
	  (format t "~%    Relaxed interval [~a,~a]" lr ur)
	  
	  (values s e lr ur))))


;;; FIX:
;;; o The following function was added to manipulate-itc-stn,lisp.
;;;   Once its confirmed that this function is in manipulate-itc-stn.lisp
;;;   AND successfully checked in to the repository delete from here.

;;;;;; Note:
;;;;;;  o This is a loose bound.  A tighter bound can be computed using
;;;;;;    SSSP from start.
;;;(defmethod event-pair-time-bound
;;;	((start itc-event) (end itc-event) (stn itc-stn))
;;;		
;;;  "Given STN with events START and END,
;;;   computes the implied constraint (start,end)[l,b].
;;;   Returns (values l b)."
;;;
;;;  ;; Note: The calls to EVENT-TIME-BOUNDS ensures that
;;;  ;;       propagate is called.
;;;  (multiple-value-bind (sl su)
;;;	  (event-time-bounds start STN)
;;;
;;;	(multiple-value-bind (el eu)
;;;		(event-time-bounds end STN)
;;;
;;;	  ;; [el-su, eu-sl] = [el, eu] - [sl, su] 
;;;	  (values (- el su) (- eu sl)))))