;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

;;; Copyright (c) 2008-2011

(in-package :TPN)

(defvar *reader-spec-version* nil "Holds the spec version that the
reader is currently working with.")

(defmacro read-with-spec-versions ((feature-name) &body body)
  (let ((cond-statements (list 'cond)))
	(loop
	   for statement in body
	   when (not (listp statement)) do
		 (error "Error in FOR-SPEC-VERSIONS. The body must
	   consist entirely of lists.")
	   else
	     when (numberp (first statement)) do
		   (push (append (list (list'eql '*reader-spec-version* (first statement)))
						 (rest statement)) cond-statements)
		 else do
		   (error "Error in FOR-SPEC-VERSIONS. The test must be a number!")
		 )
	(push `(t (error "The ~a feature is unsupported in this version of spec." (quote ,feature-name))) cond-statements)
	(nreverse cond-statements)))

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			;; Parse TPN
;;;			parse-tpn-file
;;;            )
;;;          :TPN)
;;;  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Reader for XML Format Spec2 TPN files
;;;
;;; Authors: Paul Robertson,
;;;          Brian Williams, and
;;;          David Wang
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Handle parameters within a TPN.  Where can these parameters appear?
;;;   When and where are the parameters bound?

;;; o Determine the convention for referring to the top level TPN. Is
;;;   it TopLevel or is it Main?

;;; o Generalize so that any type of TPN or TP can be read in.
;;;   Start by creating an instance of the TP(N) and then apply
;;;   methods to it to read in the relevant XML elements.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; An XML Spec2 TPN file is parsed using PARSE-TPN-FILE, 
;;; which takes the following 5 arguments:

;;; file-name: a string or pathname object that references the tpn file to be 
;;;           parsed.

;;; get-event:   a function of 2 arguments -- the event name and TPN object.

;;; make-episode: a function of seven positional arguments and one keyword argument: 
;;;               (1) a start event, (2) an end event, (3) a lower bound, (4) an upper bound, 
;;;               (5) cost, (6) an episode description, of the form (type . associations),
;;;               (7) a TPN, and keyword argument :support.

;;; add-episode:  a function of three arguments (1-2) start and end event objects 
;;;               and (3) an episode between the two events.

;;; make-tpn:     A function of four arguments (1) name (2) start event,
;;;               (3) end event and decision events.
;;;               If there are no decision events, then nil is supplied.
;;;

;;; PARSE-TPN-FILE parses an XML Spec2 TPN file in three steps:

;;; 1. Reads in the XML file to a string. 

;;; 2. Parses the LXML string, and generates an
;;;    equivalent LXML structure, a parenthesized lisp structure
;;;    corresponding to the XML. If necessary, the content of the 
;;;    input file is enclosed with <Spec2> ... </Spec2>.  This is performed 
;;;    in order to make the file legal XML.

;;; 3. Parse each component TPN in the LXML 
;;;    and return a list of the parsed TPN's.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun parse-tpn-file
	(file-name &optional (get-event #'get-event) (make-episode #'make-episode) (add-episode #'add-episode) (make-tpn #'make-tpn))
  "Returns a list of TPN objects, corresponding to the TPNs described by
   a SPEC2 XML formatted file named FILE-NAME.  

   The tpns are constructed by functions GET-EVENT, MAKE-EPISODE, ADD-EPISODE
   and MAKE-TPN."

  (let (;; Reads file-name into a String.
		(tpn-spec2-string (read-file-to-string (pathname file-name) :multibyte t)))
		;; Returns a list of TPN objects.
    (parse-tpn-string tpn-spec2-string get-event make-episode add-episode make-tpn)))

(defun parse-tpn-stream 
	(tpn-spec2-stream get-event make-episode add-episode make-tpn)
  (parse-tpn-string (read-stream-to-string tpn-spec2-stream) get-event make-episode add-episode make-tpn))

(defun parse-tpn-string
	(tpn-spec2-string get-event make-episode add-episode make-tpn)
  "Returns a list of TPN objects, corresponding to the TPNs described by
   a SPEC2 XML formatted string named TPN-SPEC2-STRING.  

   The tpns are constructed by functions GET-EVENT, MAKE-EPISODE, ADD-EPISODE
   and MAKE-TPN."

  (let (;; Convert the contents of the augmented XML TPN file to 
		;; list structures in LXML format, and ensures the content is 
		;; enclosed by <Spec2> ... </Spec2> tags.
		(tpn-spec2-lxml (make-spec2-string-legal-xml tpn-spec2-string))
		;; tpn-list will contain the TPN objects that
		;; are converted from the LXML structure.
		(tpn-list nil))
	  ;; We have a spec2 structure.
	  ;; Build a TPN object for each TPN expression in the LXML structure.

	(dolist (tpn-lxml-description (SPEC2-LXML-description-TPNS
								   tpn-spec2-lxml))
	  (cond ((TPN-LXML-description? tpn-lxml-description)
			 ;; Convert the component TPN from LXML to a TPN object.
			 (push (parse-lxml-component-tpn tpn-lxml-description
											 get-event 
											 make-episode
											 add-episode
											 make-tpn)
				   tpn-list))
			(:otherwise
			 (error "Unexpected top level item: ~a." tpn-lxml-description))))
    tpn-list))

;; ;;; Added by Pedro Santana
;; ;;;
;; ;;; This is a first attempt to create a parser for TPN SPEC3 files, which should~
;; ;;; be backwards compatible with SPEC2
;; ;;;
;; ;;; To Do:
;; ;;; o Write function that deals with both SPEC3 and SPEC2 and rename it to 
;; ;;;   parse-tpn-string


;; (defun parse-tpn-string-spec3
;; 	(tpn-string get-event make-episode add-episode make-tpn)
;;   "Returns a list of TPN objects, corresponding to the TPNs described by
;;    an XML formatted string named TPN-STRING.  

;;    The tpns are constructed by functions GET-EVENT, MAKE-EPISODE, ADD-EPISODE
;;    and MAKE-TPN."
;;   (let (;; Convert the contents of the augmented XML TPN file to 
;; 		;; list structures in LXML format, and ensures the content is 
;; 		;; enclosed by <Spec2> ... </Spec2> tags.
;; 		(tpn-lxml (make-string-legal-xml tpn-string))
;; 		;; tpn-list will contain the TPN objects that
;; 		;; are converted from the LXML structure.
;; 		(tpn-list nil))
;; 	  ;; We have a spec2 structure.
;; 	  ;; Build a TPN object for each TPN expression in the LXML structure.
;; 	(dolist (tpn-lxml-description (SPEC2-LXML-description-TPNS
;; 								   tpn-lxml))
;; 	  (cond ((TPN-LXML-description? tpn-lxml-description)
;; 			 ;; Convert the component TPN from LXML to a TPN object.
;; 			 (push (parse-lxml-component-tpn tpn-lxml-description
;; 												   get-event 
;; 												   make-episode
;; 												   add-episode
;; 												   make-tpn)
;; 				   tpn-list))
;; 			(:otherwise
;; 			 (error "Unexpected top level item: ~a." tpn-lxml-description))))
;;     tpn-list))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Wrap the file content with <Spec2> ... </Spec2> in order to make it
;;; a legal XML file
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; Fix: (added by Pedro Santana)
;;; o Is there any reason why we have to enclose the string xml with~
;;;   <SPEC2>? Doesn't the loop get rid of enclosing tags until it~
;;;   finds a <TPN> tag?
;;;

(defun make-spec2-string-legal-xml (file-string)
  "Given a SPEC2 TPN file as a string, named FILE-STRING,
   return the LXML (list-XML) whose content is equivalent, but is legal 
   XML format.  This is achieved by wrapping the file content 
   with <SPEC2> ... </SPEC2>."
  ;; The XML string we receive may already be legal SPEC2. Namely, there is ONLY
  ;; ONE top level tag, which contains the TPN tags. The standard implementation would 
  ;; be to parse the string into XML, maybe discover it's not legal, wrap it in 
  ;; one top-level tag, and then parse it again, but that would require two calls to parse-xml.
  ;; Instead, we wrap it in a tag first, parse it once, and then descend into the LXML structure
  ;; until we discover the tag nesting we want. Namely, this function will try to return
  ;; LXML equivalent to <SPEC2><TPN></TPN>...<TPN></TPN></SPEC2>.
  (let* ((xml (parse-xml-string (concatenate 'string "<SPEC2>" file-string "</SPEC2>"))))
	(loop while (not (eq (caadr xml) :TPN)) do
	 	 (setf xml (cadr xml)))
	xml))

;;; Added by Pedro Santana
;;;
;;; Same function as before, but just doesn't use the <SPEC2> tag
;;;
(defun make-string-legal-xml (file-string)
  "Given a TPN file as a string, named FILE-STRING,
   return the LXML (list-XML) whose content is equivalent, but is legal 
   XML format.  This is achieved by wrapping the file content 
   with <SPEC> ... </SPEC>."
  ;; The XML string we receive may already be legal. Namely, there is ONLY
  ;; ONE top level tag, which contains the TPN tags. The standard implementation would 
  ;; be to parse the string into XML, maybe discover it's not legal, wrap it in 
  ;; one top-level tag, and then parse it again, but that would require two calls to parse-xml.
  ;; Instead, we wrap it in a tag first, parse it once, and then descend into the LXML structure
  ;; until we discover the tag nesting we want. Namely, this function will try to return
  ;; LXML equivalent to <SPEC><TPN></TPN>...<TPN></TPN></SPEC>.
  (let* ((xml (parse-xml-string (concatenate 'string "<SPEC>" file-string "</SPEC>"))))
	(loop while (not (eq (caadr xml) :TPN)) do
	 	 (setf xml (cadr xml)))
	xml))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Reads an Episode from the LXML representation.
;;; Discards all irrelevent inter-xml-node text and 
;;; converts bound and event numbers to numbers.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Note:
;;; o TELL, ASK, PRIMITIVE and MACRO are extracted from the episode
;;;   description on demand.  They could be error checked upon 
;;;   parsing.

;;; Check:
;;; o An episode can contain zero or more of these fields.
;;;   Can multiples of each appear?

;;; To Do:
;;; o Handle episode parameters.
(defun parse-lxml-episode (episode-lxml-description TPN get-event make-episode add-episode)
  
  "Given a tpn EPISODE-LXML-DESCRIPTION in LXML representation,
   Returns an Episode object."
  (let ((start-event nil)
		(end-event nil)
		(lowerbound 0)
		(upperbound *infinity*)
		(cost (make-instance 'constant-episode-cost :cost 0))
		(relaxable nil)
		(isDummy nil)
		(isTemporalConstraint nil)
		(isStateConstraint nil)
		(isActivity nil)
		(executed nil)
		(disabled nil)
        (controllable t)
		(temporal-duration (make-instance 'bounded-temporal-duration :lower-bound 0 :upper-bound *infinity*)))
	;; For each association in EPISODE-LXML-DESCRIPTION,
	;; If attribute is one of (LOWERBOUND UPPERBOUND COST START END)
	;; then convert the attribute value from a string to an sexpression,
	;; otherwise, leave the association unchanged.
	
    (dolist (lxml-association (description-associations
							   episode-lxml-description))
      (when (association? lxml-association)

	;;	(format t "Attributes: ~A~%" (association-attribute lxml-association))

		(case (association-attribute lxml-association)
		  (:LOWERBOUND
		   (read-with-spec-versions (flat-lower-bound)
			 (2
			  (setf (btd-lower-bound temporal-duration) (convert-lxml-2-tpn-bound (convert-association-value lxml-association))))))
		  (:UPPERBOUND
		   (read-with-spec-versions (flat-upper-bound)
			 (2
			  (setf (btd-upper-bound temporal-duration) (convert-lxml-2-tpn-bound (convert-association-value lxml-association))))))
		  (:COST
		   (read-with-spec-versions (cost)
			 (2
			  (setf cost
					(if (convert-lxml-2-tpn-bound (convert-association-value lxml-association)) 
						(make-instance 'constant-episode-cost :cost (convert-lxml-2-tpn-bound (convert-association-value lxml-association))) 
						(make-instance 'constant-episode-cost :cost 0))))
			 (3
			  (setf cost
					(convert-lxml-2-cost-object lxml-association)))))
		  (:START
		   (setf start-event
				 (funcall get-event (convert-association-value lxml-association) TPN)))
		  (:END 
		   (setf end-event
				 (funcall get-event (convert-association-value lxml-association) TPN)))
		  (:RELAXABLE
		   (setf relaxable
				 (convert-lxml-2-t-nil (convert-association-value lxml-association))))

		  (when (and (= lowerbound 0) (= upperbound 0))

			(setf isDummy T))
		  
		  (:duration
		   (read-with-spec-versions (:duration)
			 (3
			  (setf temporal-duration
					(convert-lxml-2-temporal-duration lxml-association)))))
		  
		  (:TEMPORALCONSTRAINT
		   (setf isTemporalConstraint
				 (convert-lxml-2-t-nil (convert-association-value lxml-association))))
		  
		  (:STATECONSTRAINT
		   (setf isStateConstraint
				 (convert-lxml-2-t-nil (convert-association-value lxml-association))))
		  (:ACTIVITY
		   (setf isActivity
				 (convert-lxml-2-t-nil (convert-association-value lxml-association))))
		  (:EXECUTED
		   (setf executed
				 (convert-lxml-2-t-nil (convert-association-value lxml-association))))
		  (:DISABLED
		   (setf disabled
				 (convert-lxml-2-t-nil (convert-association-value lxml-association))))
		  (:CONTROLLABLE
		   (setf controllable
				 (convert-lxml-2-t-nil (convert-association-value lxml-association))))
		  ;; (:MEAN
		  ;;  (setf mean
		  ;; 		 (convert-lxml-2-tpn-bound (convert-association-value lxml-association))))
		  ;; (:VAR
		  ;;  (setf variance
		  ;; 		 (convert-lxml-2-tpn-bound (convert-association-value lxml-association))))

		  (:GUARD )
		  
		  )))
	;;; We're not quite finished converting to version 3 of the TPN
	;;; object definition, so hacks go here to bridge the gap.
	
	;; If the duration is bounded, copy out the lower and upper bounds.
	(when (typep temporal-duration 'bounded-temporal-duration)
	  (setf lowerbound (btd-lower-bound temporal-duration))
	  (setf upperbound (btd-upper-bound temporal-duration)))
	
	(funcall add-episode
			 start-event
			 end-event
			 (funcall make-episode 
					  start-event
					  end-event
					  lowerbound
					  upperbound
					  cost
					  (cons :EPISODE (description-associations episode-lxml-description))
					  TPN
					  :support :CONTROL-PROGRAM
					  :relaxable relaxable
					  :executed executed
					  :disabled disabled
					  :isDummy isDummy
					  :isTemporalConstraint isTemporalConstraint
					  :isStateConstraint isStateConstraint
					  :isActivity isActivity
					  :controllable controllable
					  ;; Assumes that all episodes specified in
					  ;; a SPEC2 file were generate by a control
					  ;; program.
					  :temporal-duration temporal-duration
					  ))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Reads a Decision from the LXML representation.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Added by Pedro Santana
;;; To do:
;;; o Incorporate the decision as an object within the TPN
;;;
(defun parse-lxml-decision (decision-lxml-description)
  
  "Given a tpn DECISION-LXML-DESCRIPTION in LXML representation,
   Returns the properties of the decision tag."

  (let ((name nil)	(controllablep t) (type nil) (attribute-string nil)
		(domain nil) (cost nil))

	;; For each association in DECISION-LXML-DESCRIPTION,	
    (dolist (lxml-association (description-associations
							   decision-lxml-description))
      (when (association? lxml-association)

		(case (association-attribute lxml-association)		  
		  (:NAME
		   (setf name (convert-association-value lxml-association)))
		  (:TYPE
		   (setf type (convert-association-value lxml-association))	   
		   (if (equalp type "UNCONTROLLABLE")
			   (setf controllablep nil)))		  		  
		  ;;This field probably belongs in the model file, not on the TPN
		  ;;description.
		  (:DOMAIN 
		   (setf attribute-string (convert-association-value lxml-association))	   
		   (setf domain (string-to-list attribute-string)))
		  (:COST
		   (setf attribute-string (convert-association-value lxml-association))	   
		   (setf cost (string-to-list attribute-string)))
		  )))

	(list :name name :controllable-p controllablep :domain domain :cost cost)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Reads a Chance-Constraint from the LXML representation.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Added by Pedro Santana
;;; To do:
;;; o Incorporate the chance-constraint as an object within the TPN
;;;
(defun parse-lxml-chance-constraint (chance-constraint-lxml-description)
  
  "Given a tpn CHANCE-CONSTRAINT-LXML-DESCRIPTION in LXML representation,
   Returns the properties of the tag."

  (let ((name nil) (temporal-list nil) (state-list nil) (risk nil))

	;; For each association in CHANCE-CONSTRAINT-LXML-DESCRIPTION,	
    (dolist (lxml-association (description-associations chance-constraint-lxml-description))
      (when (association? lxml-association)
		(case (association-attribute lxml-association)		  
		  (:NAME
		   (setf name (convert-association-value lxml-association)))
		  (:TEMPORAL-CONSTRAINT
		   (push (convert-association-value lxml-association) temporal-list))
		  (:STATE-CONSTRAINT
		   (push (convert-association-value lxml-association) state-list))
		  (:RISK
		   (setf risk (convert-association-value lxml-association)))
		  )))

	(list :name name :temp-list temporal-list :state-list state-list :risk risk)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Reads a Temporal-Constraint from the LXML representation.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Added by Pedro Santana
;;; To do:
;;; o Incorporate the temporal-constraint as an object within the TPN
;;; o Replace these very similar functions by macros
(defun parse-lxml-temporal-constraint (temporal-constraint-lxml-description)
  
  "Given a tpn TEMPORAL-CONSTRAINT-LXML-DESCRIPTION in LXML representation,
   Returns the properties of the tag."

  (let ((name nil) (start nil)(end nil) (lb nil) (ub nil))

	;; For each association in TEMPORAL-CONSTRAINT-LXML-DESCRIPTION,	
    (dolist (lxml-association (description-associations temporal-constraint-lxml-description))
      (when (association? lxml-association)
		(case (association-attribute lxml-association)		  
		  (:NAME
		   (setf name (convert-association-value lxml-association)))
		  (:START
		   (setf start (convert-association-value lxml-association)))
		  (:END
		   (setf end (convert-association-value lxml-association)))
		  (:LOWERBOUND
		   (setf lb (convert-association-value lxml-association)))
		  (:UPPERBOUND
		   (setf ub (convert-association-value lxml-association)))
		  )))

	(list :name name :start start :end end :lb lb :ub ub)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Read a Node from the LXML representation
;;; Discards all irrelevant inter-xml-node text and constructs the TPN.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Modified by Pedro Santana. 
;;;
;;; This function is a replacement to the old parse-lxml-component-tpn, which is
;;; commented below. It should be able to handle both SPEC2 and SPEC3 and defaults
;;; to SPEC2 if no format is specified.
(defun parse-lxml-component-tpn
	(tpn-lxml get-event make-episode add-episode make-tpn)
  
  "Given a TPN description TPN-LXML, in LXML format,returns an equivalent 
   TPN object.
   GET-EVENT, MAKE-EPISODE and MAKE-TPN are functions for creating TPN events,
   TPN episodes and TPN objects, respectively, while ADD-EPISODE is a function 
   for adding a tpn episode object to a tpn object:

   get-event:    a function of 2 argument -- the event name and TPN object.

   make-episode: a function of seven positional arguments: (1) a start event,  
                 (2) an end event, (3) a lower bound, (4) an upper bound, 
                 (5) a cost, (6) an episode description, of the form (type . associations),
                 (7) a TPN, and one keyword argument :support.

   add-episode:  a function of three arguments (1-2) start and end event objects 
                 and (3) an episode between the two events.

   make-tpn:     A function of four arguments (1) name (2) start event,
                 (3) end event and (4) decision events.  If there are no 
                 decision events, then nil is supplied."

  ;;Parses all top-level components of a TPN, except for EPISODEs (or ARCs in SPEC2),
  ;;which are handled by a second loop below (why?).
  (let ((format nil)
		(tpn-name nil)
		(start-name 0)
		(end-name 0)
		(decision-names nil)
		(tag-properties nil)		
		(episode-name :ARC))
	;; First, we must determine what version of TPN file we are
	;; parsing.
	
	(setf format (parse-lxml-component-tpn-format tpn-lxml))
	(let ((*reader-spec-version* (cond
							((equalp format "SPEC3")
							 3)
							(:otherwise
							 2))))
	  (if (eql *reader-spec-version* 2)
		  (setf episode-name :arc)
		  (setf episode-name :episode))
	  (dolist (association (description-associations tpn-lxml))
		(when (association? association) 
		  (case (association-attribute association)
			;; Format type
			;; (:FORMAT
			 
			;;  (setq format (convert-association-value association))
			;;  (cond
			;;    ((equalp format "SPEC2") 
			;; 	(setf episode-name :ARC))
			;;    ((equalp format "SPEC3") 
			;; 	(setf episode-name :EPISODE))
			;;    (:otherwise
			;; 	(format t "~% Ignoring TPN ~a, in format ~a, not SPEC2 or SPEC3." tpn-name format)
			;; 	(return-from parse-lxml-component-tpn nil))))
			(:NAME
			 ;; Name of the TPN
			 (setq tpn-name (convert-association-value association)))
			(:START 
			 ;; TPN start event ID
			 (setq start-name (convert-association-value association)))
			(:END 
			 ;; TPN end event ID
			 (setq end-name (convert-association-value association)))
			;; A Choice between alternative threads.
			;; Only one of decision's successor episodes
			;; is followed.
			(:DECISION
			 (read-with-spec-versions (decision)
			   (3
				(setf tag-properties (parse-lxml-decision association))			  
				;; (if (getf tag-properties :controllable-p)
				;; 	  (format t "Choice ~a is CONTROLLABLE~%" (getf tag-properties :name))
				;; 	  (format t "Choice ~a is UNCONTROLLABLE~%" (getf tag-properties :name)))
				
				;; (unless (null (getf tag-properties :prior))
				;; 	(format t "Choice ~a with priors ~a~%" (getf tag-properties :name) (getf tag-properties :prior)))
				
				(push (getf tag-properties :name) decision-names))
			   (2
				(push (convert-association-value association) decision-names))))
			
			;; Until the parser is reworked to fully parse episodes
			;; within this main loop, this is a hack to yell if the
			;; incorrect episode tag is found.
			(:arc
			 (read-with-spec-versions (episode-with-arc-tag)
			   (2
				nil)))
			
			(:episode
			 (read-with-spec-versions (episode-with-episode-tag)
			   (3
				nil)))
			
			(:CHANCE-CONSTRAINT
			 (read-with-spec-versions (chance-constraints)
			   (3
				(setf tag-properties (parse-lxml-chance-constraint association)))))
			
			(:TEMPORAL-CONSTRAINT
			 (read-with-spec-versions (temporal-constraints)
			   (3
				(setf tag-properties (parse-lxml-temporal-constraint association))))))))
	  
	  ;; Parses EPISODEs (or ARCs in SPEC2) only.
	  ;;  o Why can't this be done in the previous loop?
	  ;;
	  
	  (let ((TPN (funcall make-tpn
						  tpn-name
						  start-name
						  end-name
						  decision-names)))
		(dolist (association (description-associations tpn-lxml))
		  (when (association? association) 
			(when (eq (association-attribute association) episode-name)
			  ;; TPN episodes
			  (parse-lxml-episode association TPN get-event make-episode add-episode))))
		TPN))))

(defun parse-lxml-component-tpn-format (tpn-lxml)
  (convert-association-value
   (find :format (description-associations tpn-lxml) :key #'(lambda (x) (association-attribute x)))))

;; (defun parse-lxml-component-tpn
;; 	(tpn-lxml get-event make-episode add-episode make-tpn)
  
;;   "Given a TPN description TPN-LXML, in LXML format,returns an cquivalent 
;;    TPN object.
;;    GET-EVENT, MAKE-EPISODE and MAKE-TPN are functions for creating TPN events,
;;    TPN episodes and TPN objects, respectively, while ADD-EPISODE is a function 
;;    for adding a tpn episode object to a tpn object:

;;    get-event:    a function of 2 argument -- the event name and TPN object.

;;    make-episode: a function of seven positional arguments: (1) a start event,  
;;                  (2) an end event, (3) a lower bound, (4) an upper bound, 
;;                  (5) a cost, (6) an episode description, of the form (type . associations),
;;                  (7) a TPN, and one keyword argument :support.

;;    add-episode:  a function of three arguments (1-2) start and end event objects 
;;                  and (3) an episode between the two events.

;;    make-tpn:     A function of four arguments (1) name (2) start event,
;;                  (3) end event and (4) decision events.  If there are no 
;;                  decision events, then nil is supplied."

;;   (let ((format nil)
;; 		(tpn-name nil)
;; 		(start-name 0)
;; 		(end-name 0)
;; 		(decision-names nil))
;;     (dolist (association (description-associations tpn-lxml))
;;       (when (association? association) 
;; 		(case (association-attribute association)
;; 		  ;; Format type, must be SPEC2
;; 		  (:FORMAT
;; 		   (setq format (convert-association-value association)))
;; 		  (:NAME
;; 		   ;; Name of the TPN
;; 		   (setq tpn-name (convert-association-value association)))
;; 		  (:START 
;; 		   ;; TPN start event ID
;; 		   (setq start-name (convert-association-value association)))
;; 		  (:END 
;; 		   ;; TPN end event ID
;; 		   (setq end-name (convert-association-value association)))
;; 		  ;; A Choice between alternative threads.
;; 		  ;; Only one of decision's successor episodes
;; 		  ;; is followed.
;; 		  (:DECISION 
;; 		   (push (convert-association-value association) decision-names))
;; 		  )))

;; 	(if (or (null format)
;; 			(equalp format "SPEC2"))

;; 		(let ((TPN (funcall make-tpn
;; 							tpn-name
;; 							start-name
;; 							end-name
;; 							decision-names)))
;; 		  (dolist (association (description-associations tpn-lxml))
;; 			(when (association? association) 
;; 			  (when (eq (association-attribute association) :ARC)
;; 			  ;; TPN episodes
;; 				(parse-lxml-episode association TPN get-event make-episode add-episode))))
;; 		  TPN)

;; 		(format t "~% Ignoring TPN ~a, in format ~a, not SPEC2." tpn-name format)
;; 	  )))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Converting Interval Bounds
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun convert-lxml-2-tpn-bound (lxml-bound)
  
  "Takes an interval bound LXML-BOUND in LXML
   format and converts it to an equivalent bound
   in TPN format.  The distinction is the denotation
   of positive and negative infinity."

  (cond ((null lxml-bound) nil)
	    ((equalp lxml-bound "+INF") *infinity*)
		((equalp lxml-bound "-INF") *-infinity*)
		(t (read-from-string lxml-bound))))

(defun convert-tpn-2-lxml-bound (bound)
  
  "Takes an interval bound BOUND in TPN
   format and converts it to an equivalent bound
   in LXML format.  The distinction is the denotation
   of positive and negative infinity."
  
  (cond ((eql bound	*infinity*)
		 "+INF")
		((eql bound	*-infinity*)
		 "-INF")
		(:otherwise
		 (format nil "~s" bound))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Converting Temporal Bounds
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun convert-lxml-2-temporal-duration (tag)
  ;; We only look at the first temporal bound within this tag.
  (let ((instance-xml-description (first (description-associations tag))))
	(case (association-attribute instance-xml-description)
	  (:loopingduration
	   (convert-lxml-2-looping-temporal-duration instance-xml-description))
	  (:boundedduration
	   (convert-lxml-2-bounded-temporal-duration instance-xml-description))
	  (otherwise
	   (error "Unknown temporal duration type.")))))

(defun convert-lxml-2-bounded-temporal-duration (xml-description)
  (let (ub
		lb)
	(dolist (association (description-associations xml-description))
      (when (association? association) 
		(case (association-attribute association)
		  ;; Format type
		  (:lowerbound
		   (setq lb (convert-lxml-2-tpn-bound (convert-association-value association))))
		  (:upperbound
		   (setq ub (convert-lxml-2-tpn-bound (convert-association-value association)))))))
	(make-instance 'bounded-temporal-duration :lower-bound lb :upper-bound ub)))

(defun convert-lxml-2-looping-temporal-duration (xml-description)
  (let (n-ub
		n-lb
		loop-duration)
	(dolist (association (description-associations xml-description))
      (when (association? association)
		(case (association-attribute association)
		  (:nlowerbound
		   (setq n-lb (convert-lxml-2-tpn-bound (convert-association-value association))))
		  (:nupperbound
		   (setq n-ub (convert-lxml-2-tpn-bound (convert-association-value association))))
		  (:loopduration
		   (setq loop-duration (convert-lxml-2-temporal-duration association))))))
	(make-instance 'looping-temporal-duration :n-lower-bound n-lb :n-upper-bound n-ub :per-loop-duration loop-duration)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Converting to cost objects.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun convert-lxml-2-cost-object (tag)
  ;; We only look at the first cost within this tag.
  (let ((instance-xml-description (first (description-associations tag))))
	(case (association-attribute instance-xml-description)
	  (:linearloopingcost
	   (convert-lxml-2-linear-looping-cost instance-xml-description))
	  (:constantcost
	   (convert-lxml-2-constant-cost instance-xml-description))
	  (otherwise
	   (error "Unknown cost type.")))))

(defun convert-lxml-2-linear-looping-cost (xml-description)
  (let ((per-loop-cost 0)
		(setup-cost 0))
	(dolist (association (description-associations xml-description))
	  (when (association? association)
		(case (association-attribute association)
		  (:perloop
		   (setq per-loop-cost (convert-lxml-2-tpn-bound (convert-association-value association))))
		  (:setup
		   (setq setup-cost (convert-lxml-2-tpn-bound (convert-association-value association))))
		  (otherwise
		   (error "Unknown option for linear looping cost.")))))
	(make-instance 'linear-looping-episode-cost :cost-per-loop per-loop-cost :setup-cost setup-cost)))

(defun convert-lxml-2-constant-cost (xml-description)
  (let ((constant-cost 0))
	(dolist (association (description-associations xml-description))
	  (when (association? association)
		(case (association-attribute association)
		  (:cost
		   (setq constant-cost (convert-lxml-2-tpn-bound (convert-association-value association))))
		  (otherwise
		   (error "Unknown option for constant cost.")))))
	(make-instance 'constant-episode-cost :cost constant-cost)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Converting Relaxable Marker
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun convert-lxml-2-t-nil (lxml-binary &key 
					(t-or-nil NIL))

  "Parse a T or NIL in string, return T or NIL"


  (if (equal lxml-binary "T")
	  (setf t-or-nil T)
	  (setf t-or-nil NIL))

  t-or-nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Setting Episode Looping Marker 
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun convert-lxml-2-t-nil-loops (tag)
  ;; If we have a looping duration, record that the episode and TPN has loops. 
  (let ((instance-xml-description (first (description-associations tag))))
	(case (association-attribute instance-xml-description)
	  (:loopingduration
	   t))))



