;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

;;; TPN to XML Spec2 Writer.
;;;     Copyright (c) 2008-2009 Brian C. Williams, MIT.  All rights reserved.

;;;     Distribution for educational and non-commercial purposes only, 
;;;     by written permission of the author.

(in-package :TPN)

(defvar *writer-spec-version* nil "Holds the spec version that the
writer is currently working with.")

(defmacro write-with-spec-versions ((feature-name) &body body)
  (let ((cond-statements (list 'cond)))
	(loop
	   for statement in body
	   when (not (listp statement)) do
		 (error "Error in FOR-SPEC-VERSIONS. The body must
	   consist entirely of lists.")
	   else
	     when (numberp (first statement)) do
		   (push (append (list (list'eql '*writer-spec-version* (first statement)))
						 (rest statement)) cond-statements)
		 else do
		   (error "Error in FOR-SPEC-VERSIONS. The test must be a number!")
		 )
	(push `(t (error "The ~a feature is unsupported in this version of spec." (quote ,feature-name))) cond-statements)
	(nreverse cond-statements)))

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			write-tpns
;;;			)
;;;		  :TPN))

;;; To Do:
;;; o Document the name of the root TPN.
;;; o Document the use of parameters in Macros.
;;;   Where are the parameters specified and bound?

;;;   ************************************************************
;;;   **
;;;   **  Write a Temporal Plan Network (TPN) Object 
;;;   **  to an XML Spec2 File
;;;   **
;;;   **  (write-tpn <file-name> &optional (<TPN> *TPN*))
;;;   **
;;;   ************************************************************
;;; (WRITE-TPNS (FILENAME TPNS
;;;				   &KEY (ENVIRONMENT *ALL*)
;;;						(FORCE-LABELING? NIL))
;;;
;;;   Download TPNS to FILENAME in .tpn format.
;;;   The extension .tpn is ALWAYS added to FILENAME.
;;;
;;;   ENVIRONMENT is either *All* or a list of decision-event 
;;;   choice assignments.
;;;      If *All* then the complete TPNS are output.
;;;      If a choice assignment, then only outputs those episodes
;;;      that are enabled by that choice assignment.
;;;
;;;     Note, an episode will be labeled by *always-active* if it holds
;;;     independent of a choice.  *always-active* MUST appear in the
;;;     environment for this type of episode to be output.
;;;
;;;   If FORCE-LABELING? is true, then generates a new
;;;   labeling of each TPN with guard conditions, otherwise,
;;;   it only labels if a labeling doesn't already exist.


;;;   See TPN-File_Format.txt in RMPL2/documentation for a more
;;;   detailed description of the XML TPN Spec2 format.  this file
;;;   also appears in this directory as TPN_File_Format-duplicated.txt

;;;   XML TPN Spec2 Output format:

;;;   <tpn-description>+ where

;;;   <tpn-description> ::= 

;;;     <TPN>
;;;        <FORMAT> SPEC2 </FORMAT>
;;;        <NAME> <tpn-name> </NAME>
;;;        <START> <event-name> </START>
;;;        <END> <event-name> </END>
;;;        [<DECISION> <event-name> </DECISION>]*
;;;        <episode-description>+
;;;     </TPN>

;;;   <episode-description> ::= 

;;;     <ARC>
;;;        <START> <event-name> </START>
;;;        <END> <event-name> </END>
;;;        <LOWERBOUND> <bound> </LOWERBOUND>
;;;        <UPPERBOUND> <bound> </UPPERBOUND>
;;;        [<ASK> <state-constraint> </ASK>]
;;;        [<TELL> <state-constraint> </TELL>]
;;;        [<MACRO> <macro-invocation> </MACRO>]
;;;        [<PRIMITIVE> <primitive-invocation> </PRIMITIVE>]
;;;     </ARC>

;;;   <tpn-name> ::= any string?
;;;   <event-name> ::= Number or Symbol, any string?
;;;   <bound> ::= non-negative integer, +INF or macro parameter bound to 
;;;      non-negative integer or +INF
;;;   <state-constraint> ::= String of symbols, where a symbol may be a macro 
;;;      parameter.
;;;   <primitive-invocation> ::= <primitive-function-name>([arg1][, arg2]...)
;;;   <macro-invocation> ::= <macro-tpn-name>([arg1][, arg2]...)
;;;   <macro-parameter> ::= $<non-negative-integer>

;;;***************************************************************
;;;
;;;   **  To Do List  **
;;;
;;;***************************************************************

;;;***************************************************************
;;;
;;;   **  Interface Functions  **
;;;
;;;***************************************************************

;;; To Do:
;;; o Probably want to refine environment,
;;;   by passing in one environment for each TPN.
(defun write-tpns (filename tpns
				   &key (environment *all*)
						(force-labeling? nil)
					    (version 3))

  "Download TPNS to FILENAME in .tpn format.
   The extension .tpn is ALWAYS added to FILENAME.

   ENVIRONMENT is either *All* or a list of decision-event 
   choice assignments.
      If *All* then the complete TPNS are output.
      If a choice assignment, then only outputs those episodes
      that are enabled by that choice assignment.

     Note, an episode will be labeled by *always-active* if it holds
     independent of a choice.  *always-active* MUST appear in the
     environment for this type of episode to be output.

   If FORCE-LABELING? is true, then generates a new
   labeling of each TPN with guard conditions, otherwise,
   it only labels if a labeling doesn't already exist. 

   VERSION specifies the version of the SPEC file format to use."

  (format t "Saving TPN to ~a~%" filename) 

  (let ((path-name (if (pathnamep filename)
					   filename
                     (format nil "~a.tpn" filename))))

    (with-open-file (stream path-name
					 :direction :output 
					 :if-exists :supersede
					 :if-does-not-exist :create)
      (if (null stream)
		  (format t "Could not open file ~a~%" filename)
		  
		  (dolist (tpn tpns)
			(write-tpn stream tpn
					   :environment environment
					   :force-labeling? force-labeling?
					   :version version
					   ))))
	'T))

;;;***********
;;;
;;;   **  Write TPN
;;;
;;;***********

;;;   <tpn-description> ::= 

;;;     <TPN>
;;;        <FORMAT> SPEC2 </FORMAT>
;;;        <NAME> <tpn-name> </NAME>
;;;        <START> <event-name> </START>
;;;        <END> <event-name> </END>
;;;        [<DECISION> <event-name> </DECISION>]*
;;;        <episode-description>+
;;;     </TPN>

;;; Check:
;;; o When we output a single choice for a decision event, do we
;;;   still want to label the event as a decision event?  It shouldn't
;;;   do any damage.  Conversely, if we remove an event as a decision event,
;;;   we have to make sure that environment doesn't include multiple choices.

;;; To Do:
;;; o Generalize this so that any type of TPN or TP can be written out.
(defmethod write-tpn (stream (tpn temporal-plan-network)
					  &key (environment *all*)
						   (force-labeling? nil)
						(version 3))

  "Write XML description of TPN to STREAM, while indenting 
   INDENT characters.  Writes either the complete TPN, or
   a subset of TPN that denotes a valid plan, depending on the
   value of environment.

   ENVIRONMENT is either *All* or a list of decision-event 
   choice assignments.
      If *All* then the complete TPN is output.
      If a choice assignment, then only outputs those episodes
      that are enabled by that choice assignment.

     Note, an episode will be labeled by *always-active* if it holds
     independent of a choice.  *always-active* MUST appear in the
     environment for this type of episode to be output.

   If FORCE-LABELING? is true, then generates a new
   labeling of the TPN with guard conditions, otherwise,
   it only labels if a labeling doesn't already exist."
  (let ((*writer-spec-version* version))
	(unless (valid-tpn-environment? tpn environment)
	  (error "~a is an invalid environment for ~a."
			 environment tpn))
	
	;; Labels each TPN episode and decision-event with the
	;; most immediate decision-assignment that activates that
	;; episode/event.

	(unless (all? environment)
	  ;; Don't need to label TPN if we are printing out all
	  ;; of it.
	  (when (or force-labeling?
				(not (tpn-labeled? tpn)))
		(label-tpn-with-activation-assignments! tpn)))

	;; outputs TPN in XML format.
	(let ((xml-slist (make-smart-list)))
	  ;; add TPN tag
	  (coll-add! xml-slist :TPN)
	  ;; add FORMAT tag
	  (write-with-spec-versions (format-tag)
		(2 (add-tagged-item xml-slist :FORMAT "SPEC2"))
		(3 (add-tagged-item xml-slist :FORMAT "SPEC3")))
	  ;; add NAME tag
	  (add-tagged-item xml-slist :NAME (tpn-name tpn))
	  ;; add START tag
	  (add-tagged-item xml-slist :START (event-name (tpn-start tpn)))
	  ;; add END tag
	  (add-tagged-item xml-slist :END (event-name (tpn-end tpn)))
	  ;; add DECISION tagS
	  (loop for decision in (tpn-decisions tpn)
		 do (write-with-spec-versions (decision-tag)
			  (2 (add-tagged-item xml-slist :DECISION (event-name decision)))
			  (3 (add-tagged-item xml-slist :decision (list :name (event-name decision))))))
	  ;; add EPISODE tagS
	  (dolist (episode (tpn-episodes tpn))
		(add-item xml-slist (write-episode episode :environment environment)))

	  (print-xml (coll-tolist xml-slist) :stream stream :pretty t)))
)


;;;***********
;;;
;;;   **  Write Episode
;;;
;;;***********

;;;   <episode-description> ::= 

;;;     <ARC>
;;;        <START> <event-name> </START>
;;;        <END> <event-name> </END>
;;;        <LOWERBOUND> <bound> </LOWERBOUND>
;;;        <UPPERBOUND> <bound> </UPPERBOUND>
;;;        [<ASK> <state-constraint> </ASK>]
;;;        [<TELL> <state-constraint> </TELL>]
;;;        [<MACRO> <macro-invocation> </MACRO>]
;;;        [<PRIMITIVE> <primitive-invocation> </PRIMITIVE>]
;;;     </ARC>

;;; To Do:
;;; o Generalize this so that any type of Episode can be written out.
(defmethod write-episode 
	((episode episode) &key (environment *all*))
  
  "Writes XML description of EPISODE to STREAM, while indenting 
   INDENT characters, if EPISODE is enabled by ENVIRONMENT.

   ENVIRONMENT is either *All* or a list of decision-event 
   choice assignments.  EPISODE is output if environment is 
   *All*, or if choice assignment enables EPISODE.

   Note episode will be labeled by *always-active* if it holds
   independent of a choice.  *always-active* MUST appear in the
   environment for this type of episode to be written."

  (when (episode-enabled-by-environment? episode environment)
	
	(let ((xml-slist (make-smart-list)))
	  ;; add ARC tag
	  (write-with-spec-versions (arc-name)
		(2
		 (coll-add! xml-slist :ARC))
		(3
		 (coll-add! xml-slist :episode)))
	  ;; add START tag
	  (add-tagged-item xml-slist :START (event-name (episode-start episode)))
	  ;; add END tag
	  (add-tagged-item xml-slist :END (event-name (episode-end episode)))
	  ;; Write out the temporal duration
	  (write-with-spec-versions (episode-duration)
		(2
		 ;; add LOWERBOUND tag
		 (add-tagged-item xml-slist :LOWERBOUND (convert-tpn-2-lxml-bound (episode-lowerbound episode)))
		 ;; add UPPERBOUND tag
		 (add-tagged-item xml-slist :UPPERBOUND (convert-tpn-2-lxml-bound (episode-upperbound episode))))
		(3
		 (add-item xml-slist (temporal-duration-to-lxml (episode-temporal-duration episode)))))
	  ;; add COST tag
	  (add-item xml-slist (cost-to-lxml (episode-cost episode)))
	  ;(add-tagged-item xml-slist :COST (convert-tpn-2-lxml-bound (episode-cost episode)))
	  ;; add ASK tag
	  (add-tagged-item xml-slist :ASK (episode-ask episode))
	  ;; add TELL tag
	  (add-tagged-item xml-slist :TELL (episode-tell episode))
	  ;; add MACRO tag
	  (add-tagged-item xml-slist :MACRO (episode-macro episode))
	  ;; add PRIMITIVE tag
	  (add-tagged-item xml-slist :PRIMITIVE (episode-primitive episode))
	  ;; add TYPE tag
	  (add-tagged-item xml-slist :TYPE (episode-type episode))

	  ;;Added by Peng; Not yet know if the relaxable and executed tag should be separated from
	  ;;description tag.
	  (when (episode-relaxable episode)
		(add-tagged-item xml-slist :RELAXABLE (format nil "~A" (episode-relaxable episode))))
	  (when (episode-executed episode)
		(add-tagged-item xml-slist :EXECUTED (format nil "~A" (episode-executed episode))))
	  (when (episode-disabled episode)
		(add-tagged-item xml-slist :DISABLED (format nil "~A" (episode-disabled episode))))
	  ;; (when (episode-isDummy episode)
	  ;; 	(add-tagged-item xml-slist :DUMMY (format nil "~A" (episode-isDummy episode))))
	  (when (episode-isTemporalConstraint episode)
		(add-tagged-item xml-slist :TEMPORALCONSTRAINT (format nil "~A" (episode-isTemporalConstraint episode))))
	  (when (episode-isStateConstraint episode)
		(add-tagged-item xml-slist :STATECONSTRAINT (format nil "~A" (episode-isStateConstraint episode))))
	  (when (episode-isActivity episode)
		(add-tagged-item xml-slist :ACTIVITY (format nil "~A" (episode-isActivity episode))))

	  (coll-tolist xml-slist))))

;;;***********
;;;
;;;   **  TPN Environments
;;;
;;;***********

;;; To Do:
;;; o Check that assignments are valid decision-event assignments
;;;   for tpn, and if fails, print out an error message that specifies
;;;   which element of environment fails.
(defmethod valid-tpn-environment? ((tpn temporal-plan-network) environment)
  (or (all? environment)
	  (loop for e in environment
		  always (or (always-active? e)
					 (assignment? e)))))

(defmethod episode-enabled-by-environment? ((episode episode) environment)
  
  "Returns true iff EPISODE is enabled within ENVIRONMENT.

     ENVIRONMENT is *All* or a list of choice assignments to 
     zero or more decision-events of EPISODE's TPN.  

     EPISODE is enabled if environment is *All* or if the 
     choice assignments activate an execution thread to EPISODE.

     Note episode will be labeled by *always-active* if it holds
     independent of a choice.  *always-active* MUST appear in the
     environment for this type of episode to be enabled."

  (or (all? environment)
	  (and (episode-activation-assignment-is-conjunction? episode)
		   (every #'(lambda (assignment) (contains-activation-assignment?
										  assignment environment))
				  (episode-activation-assignment episode)))
	  (contains-activation-assignment?
	   (episode-activation-assignment episode)
	   environment)))

(defun contains-activation-assignment? (assignment environment)

  "Returns true iff ENVIRONMENT contains ASSIGNMENT
   ASSIGNMENT is an assignment of a value to a choice,
   or the value bound to *always-active*."
  (if (always-active? assignment)
	  (contains-always-active? environment)
	(loop for assignment2 in environment
		unless (always-active? assignment2)
		when (equal-assignments? assignment assignment2)
		return t
		finally (return nil))))

;;;***************************************************************
;;;***************************************************************
;;;
;;;   **  Support Functions  **
;;;
;;;***************************************************************
;;;***************************************************************

(defun add-item (xml-slist item)
  (unless (null item)
	(coll-add! xml-slist item)))

(defun add-tagged-item (xml-slist tag item)
  (unless (null item)
	(coll-add! xml-slist `(,tag ,item))))

(defun temporal-duration-to-lxml (duration)
  (let ((xml-slist (make-smart-list)))
	(coll-add! xml-slist :duration)
	(add-item xml-slist (temporal-duration-to-lxml-helper duration))
	(coll-tolist xml-slist)))

(defmethod temporal-duration-to-lxml-helper ((duration looping-temporal-duration))
  (let ((xml-slist (make-smart-list)))
	(coll-add! xml-slist :loopingduration)
	(add-tagged-item xml-slist :nupperbound (convert-tpn-2-lxml-bound (ltd-n-upper-bound duration)))
	(add-tagged-item xml-slist :nlowerbound (convert-tpn-2-lxml-bound (ltd-n-lower-bound duration)))
	(let ((loop-duration (make-smart-list)))
	  (coll-add! loop-duration :loopduration)
	  (add-item loop-duration (temporal-duration-to-lxml-helper (ltd-per-loop-duration duration)))
	  (add-item xml-slist (coll-tolist loop-duration)))
	(coll-tolist xml-slist)))

(defmethod temporal-duration-to-lxml-helper ((duration bounded-temporal-duration))
  (let ((xml-slist (make-smart-list)))
	(coll-add! xml-slist :boundedduration)
	(add-tagged-item xml-slist :upperbound (convert-tpn-2-lxml-bound (btd-upper-bound duration)))
	(add-tagged-item xml-slist :lowerbound (convert-tpn-2-lxml-bound (btd-lower-bound duration)))
	(coll-tolist xml-slist)))

(defun cost-to-lxml (cost)
  (let ((xml-slist (make-smart-list)))
	(coll-add! xml-slist :cost)
	(add-item xml-slist (cost-to-lxml-helper cost))
	(coll-tolist xml-slist)))

(defmethod cost-to-lxml-helper ((cost linear-looping-episode-cost))
  (write-with-spec-versions (linear-looping-episode-cost)
	(3
	 (let ((xml-slist (make-smart-list)))
	   (coll-add! xml-slist :linearloopingcost)
	   (add-tagged-item xml-slist :setup (convert-tpn-2-lxml-bound (looping-setup-cost cost)))
	   (add-tagged-item xml-slist :perloop (convert-tpn-2-lxml-bound (looping-cost-per-loop cost)))
	   (coll-tolist xml-slist)))))

(defmethod cost-to-lxml-helper ((cost constant-episode-cost))
  (write-with-spec-versions (constant-episode-cost)
	(2
	 (convert-tpn-2-lxml-bound (constant-episode-cost-cost cost)))
	(3
	 (let ((xml-slist (make-smart-list)))
	   (coll-add! xml-slist :constantcost)
	   (add-tagged-item xml-slist :cost (convert-tpn-2-lxml-bound (constant-episode-cost-cost cost)))
	   (coll-tolist xml-slist)))))

(defmethod cost-to-lxml-helper ((cost t))
  (unless (numberp cost)
	(error "Unknown cost type for ~a." cost))
  (cost-to-lxml-helper (make-instance 'constant-episode-cost :cost cost)))
