;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

;;; Copyright (c) 2008-2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

;;; ***********************************************************************
;;;
;;;                Choice Selection for a Temporal Plan Network
;;;
;;; ***********************************************************************

;;; This file contains code for selecting a feasible or optimal set of choices
;;; for a Temporal Plan Network (TPN).

;;; A TPN represents all executions of a concurrent timed program in which 
;;; the program involves a non-deterministic choice operator and interval 
;;; time constraints.

;;  Choice Selection finds an execution that is consistent with respect to the 
;;; temporal constraints.  This execution is described by a selection for 
;;; each non-deterministic choice encountered during the execution.

;;; A choice assignment is represented as an assignment expression whose
;;; variable is a decision-event and whose value is the number of a
;;  successor episode.  In internal form decision-event is an object, and
;;  externally decision-event is denoted by its event number.

;;; Optimal Choice Selection returns the feasible choice selection that 
;;; maximizes the utility of the selected execution.  Currently, the utility 
;;; of the selected execution is the sum of the utilities of the episodes 
;;; contained by the execution.

(in-package :TPN)

;;;(eval-when (compile load eval)
;;;  (export '(
;;;			select-tpn-choice!
;;;			frame-tpn-selection-problem
;;;            )
;;;          :TPN))

;;; ***********************************************************************
;;;
;;;     To Do:
;;;
;;; ***********************************************************************

;;; To Do:
;;; o For decision-variables, rename keyword :DOMAIN to :VALUATION.

;;; o Handle macros and multiple TPNs that are specified in a single TPN file.
;;;   How is this supposed to work?

;;; o Seem To Dos and Fixes in line.

;;; ***********************************************************************
;;;
;;;     Interface Functions
;;;
;;; ***********************************************************************


;;; *********************
;;;
;;;     Finds an Optimal Consistent Plan for TPN and Writes to File.
;;;
;;; *********************

;;; Fix:
;;; o Delete this comment once tested.
;;;   Renamed from write-best-plan-for-tpn-file.
(defun write-best-plan-of-tpn-to-file (filename)
  
  "Let FILENAME be a SPEC2 tpn file (with extension .tpn),
   Selects the highest utility plan from the TPN of FILENAME
   and outputs it to a same named file, but with extension .plan."
  
  (let* ((file-pathname (pathname filename))
		 (tpns (parse-tpn-file file-pathname
							   #'get-event
							   #'make-episode
							   #'add-episode
							   #'make-tpn))
		 (tpn1 (first tpns)))

	(let ((best-choice (tpn-best-choice tpn1))
		  (environment nil)
		  (plan-file-pathname
		   (make-pathname :host (pathname-host file-pathname)
						  :device (pathname-device file-pathname)
						  :directory (pathname-directory file-pathname)
						  :name (pathname-name file-pathname)
						  :type "plan")))

	  ;; the "best-choice" variable is nil if there are 
	  ;; no choices to be made, regardless of whether the plan is 
	  ;; temporally feasible. We need to check if the plan is 
	  ;; temporally feasible to determine what to write.
	  ;; If feasible - 'environment' is a list that contains the guard conditions
	  ;;               for all the episodes (arcs) that should be output.
	  ;; If NOT - leave 'environment' as nil, to output no arcs (just the start and end node).
	  (when (feasible-choices? tpn1 best-choice) 
		  (setf environment (adjoin *always-active* 
									(internalize-choice-assignment-descriptions best-choice tpn1))))

	  (write-tpns plan-file-pathname
				  (list tpn1)
				  ;; Include episodes that are active independent of a choice.
				  :environment environment
				  :force-labeling? t)
	  )))


;;; *********************
;;;
;;;     Initializes TPN Choice Selection,
;;;     throwing away previous problem framings and solutions.
;;;
;;; *********************


;;; Fix:
;;; o Check that I've initialized everything that is required.
;;;   Should we reinitialize the labeling?
(defmethod initialize-tpn-choice-selection! ((tpn tpn-w-selection))
  
  "Resets choice selection for TPN, so that when
   SELECT-TPN-CHOICE! is called, a new selection problem is
   formulated and a new best choice is found and recorded."

  (with-slots (selection-problem best-choice) tpn
	(setq selection-problem nil
		  best-choice nil)))

;;; *********************
;;;
;;;     Returns a Best Set of Choices for TPN.
;;;     Frames problem and solves if needed.
;;;
;;; *********************

(defmethod tpn-best-choice ((tpn tpn-w-selection))
  
  "Given a TPN, 
   Returns a description of the best (higest utility) choice selection for TPN.
   Best choice is a list of assignments to TPN decision events of the form:
          (= <decision-event-number> <choice-number>).
   The choice-number indicates a successor episode."

  (select-tpn-choice! tpn)

  (with-slots (best-choice) tpn
	best-choice))

;;; *********************
;;;
;;;     Returns a Best Set of explanations for TPN.
;;;     Frames problem and solves if needed.
;;;
;;; *********************

(defmethod tpn-best-explanation ((tpn tpn-w-selection))
  
  "Given a TPN, 
   Returns a description of the best (higest utility) explanation 
   for the overconstrained TPN."

  (tpn-explanations tpn :exp-count 1))



(defmethod tpn-explanations ((tpn tpn-w-selection) &key (exp-count 1))
  
  "Given a TPN as input, 
   ensures that a best choice (which is an explanation)
   has been computed for TPN,
   where a TPN choice denotes an execution of the TPN.
   
   All constraints will have IN or OUT assignments."



  (when (null (tpn-selection-problem tpn))
	;; Generate a tpn-selection-problem from TPN 
	;; if it doesn't already exist.
	(frame-tpn-explanation-problem tpn))

(reverse (explanations-to-selection-problem tpn :exp-count exp-count)))


;;; *********************
;;;
;;;     Records a Best Set of Choices for TPN
;;;     if not recorded already.
;;;
;;; *********************

;;; Fix:
;;; o Current method selection may only work when OPSAT (find-solutions)
;;;   performs A* search.  In order to perform branch and bound we need to
;;;   implement suboptimality conflicts for conditional CSPs.

;;; To Do:
;;; o Update the TPN, choice selection problem and best choice incrementally.
;;;   - When episodes of the TPN change, 
;;;   - record that the TPN-DOCSP is out of date.
;;;   - Update the TPN-DOCSP and choice selection on demand.
(defmethod select-tpn-choice! ((tpn tpn-w-selection))
  
  "Given a TPN as input, 
   ensures that a best choice has been computed for TPN,
   where a TPN choice denotes an execution of the TPN.

   Side Effects:
      Ensures that a TPN selection problem as a Dynamic OCSP
      has been recorded on TPN.

      Ensures that a best choice is recorded for TPN.

      Best choice is a list of choice assignment descriptions
      of the form:
          (= <decision-event-number> <choice-number>)."

  (when (null (tpn-selection-problem tpn))
	;; Generate a tpn-selection-problem from TPN 
	;; if it doesn't already exist.

	(frame-tpn-selection-problem tpn))

  (with-slots (best-choice) tpn
	;; If there is no choice in this TPN, skip the choice selection process
	;; and check consistency directly.
   
	(when (not (BFIG::state-space-variables (get-tpn-selection-problem tpn)))

	  ;;(format t "TPN feasible? ~A~%" (ITC::consistent? (OPSAT::opsat-problem-constraint-system (get-tpn-selection-problem tpn))))
	  	  
	  (if (ITC::consistent? (OPSAT::opsat-problem-constraint-system (get-tpn-selection-problem tpn)))
		  (setf best-choice (list (list 'T)))
		  (setf best-choice NIL)))

	(unless best-choice
	  (setq best-choice (best-solution-to-selection-problem tpn)))	  
	T))

;;; Check:
;;; o Assert and unassert take different representations for choices
;;;   as inputs.  Is this correct?
;;; o What is the form of the output conflict and is it a correct format?
;;;   Document.

;;; To Do:
;;; o Generalize to testing consistency for partial assignments.
;;; o Make sure that the returned conflict is in a meaningful form and
;;;   document that form.

;;; FIX:
;;; o Check that choices denotes a full assignment, and if not, flag an error.

;;; *********************
;;;
;;;     Records a Best Set of explanations for TPN
;;;     if not recorded already.
;;;
;;; *********************

(defmethod select-tpn-best-explanation! ((tpn tpn-w-selection))
  
  "Given a TPN as input, 
   ensures that a best choice (which is an explanation)
   has been computed for TPN,
   where a TPN choice denotes an execution of the TPN.
   
   All constraints will have IN or OUT assignments.

   Side Effects:
      Ensures that a TPN selection problem as a Dynamic OCSP
      has been recorded on TPN.

      Ensures that a best choice is recorded for TPN.

      Best choice is a list of choice assignment descriptions
      of the form:
          (= <decision-event-number> <choice-number>)."

  (when (null (tpn-selection-problem tpn))
	;; Generate a tpn-selection-problem from TPN 
	;; if it doesn't already exist.
	(frame-tpn-explanation-problem tpn))

  (with-slots (best-choice) tpn
	(unless best-choice
	  (setq best-choice (best-explanation-to-selection-problem tpn)))	  
	T))


(defmethod feasible-choices? ((tpn tpn-w-selection) choices
							  &key (return-conflict? T) 
								   (initialize? nil)
							  &allow-other-keys)
  
  "Let TPN be a temporal plan network, an
       CHOICES denote a COMPLETE assignment of choices 
         to decision events in TPN,
       where a choice assignment is of the form:
         (= <decision-event-name> <number-of-choice>)

   Returns (values CONSISTENT? CONFLICT).

   CONSISTENT? is non-NIL iff TPN is temporally consistent for CHOICES.

   CONFLICT denotes a set of assignments or constraints
     that is mutually inconsistent.

   If TPN with CHOICES is found inconsistent and RETURN-CONFLICT? is TRUE, 
   then a CONFLICT is returned; otherwise, nil is returned.

   When returned, a conflict is generated by applying function 
   KEY to the constraint or assignment objects of the conflict,
   and by collecting the non-nil, irredundant values returned by KEY.

   KEY is a function used to filter and transform the elements comprising
   a conflict.  It takes as input an object denoting a constraint or assignment,
   and returns any arbitrary expression, or nil, if the constraint/assigment
   is to be ignored.

   If INITIALIZE? is True, then resets all constraint solver derived values
   before testing consistency; otherwise, update incrementally from 
   previous values.

   Assumptions:
   - No choices are asserted within TPN selection problem upon entry.
   - The choice selection problem has already been formulated.

   Side Effects:
   - Inserted constraints are unchanged.
   - All asserted assignments are removed, same as upon entry.
   - the state of the STN of the TPN selection problem is 
     changed during the function, but restored at the end."

  (establish-tpn-selection-problem! tpn)

  ;; 1) Assert the choices.  TPN's STN starts with
  ;;    all TPN STCs asserted, and no choices asserted.
  (let ((STN (tpn-constraint-system tpn))

		;; Assignment objects corresponding to choices.
		(choice-assignments
		 (assert-choices-for-representative-explanation! tpn choices)))

	(debug-print-for :tpn t "Checking consistency of choice ~a for ~a."
					 choice-assignments tpn)
	
	;; 2) Check consistency of the STN.
	(multiple-value-bind (consistent? conflict)
		(consistent? STN :initialize? initialize?
					 :return-conflict? return-conflict?
;;; Fix: Figure out which is correct, and delete the other
					 :key #'stc-guard-to-state-assignment-w-truth-description
;;;					 :key #'stc-guard-to-truth-assignment
					 )

	  ;; 3) Report the results.
	  (if consistent?
		  (debug-print-for :tpn t "  ~a consistent for choice ~a."
						   tpn choices)
		(debug-print-for :tpn t "  ~a inconsistent for choice ~a."
						 tpn choices))

	  ;; 3) Unassert Choices.  Assume that STN started with no choices asserted.
	  (unassert-choices-for-representative-explanation! tpn choice-assignments)
	  
	  (values consistent? conflict))))

;;; ***********************************************************************
;;;
;;;                Support Functions
;;;
;;; ***********************************************************************

;;; Fix:
;;; o Record the source for each TPN episode 
;;;   (Implemented for Control Programs only).

;;; o Record for each constraint whether or not it can be relaxed, 
;;;   (Implemented for Control Programs only).

;;; To Do:
;;; o Update the TPN, choice selection problem and best choice incrementally.
;;;   - When episodes of the TPN change, 
;;;   - record that the TPN-DOCSP is out of date.
;;;   - Update the TPN-DOCSP and choice selection on demand.
(defmethod frame-tpn-selection-problem
	((tpn tpn-w-selection) &key (force-labeling? nil) (return-description? nil))

  "Given a TPN as input,

   RETURNS a TPN choice selection problem, framed
   as a dynamic optimal constraint satisfaction problem.

   Records the selection problem under TPN-SELECTION-PROBLEM.

   If FORCE-LABELING? is true, then generates a new
   labeling of the TPN with guard conditions, otherwise,
   it only labels if a labeling doesn't already exist.

   If RETURN-DESCRIPTION? is true, then returns a description
   of a choice-selection problem, rather than an instance
   of an opsat object, representing the problem."

  ;; Labels each TPN episode and decision-event with the
  ;; most immediate decision-assignment that activates that
  ;; episode/event.


  (when (or force-labeling?
			(not (tpn-labeled? tpn)))
(print "ddddddddddddddddddddddddddddddddddddd")
;;;	(check-that-tpn-is-well-formed! tpn)
	(label-tpn-with-activation-assignments! tpn))
(print "ddddddddddddddddddddddddddddddddddddd")
  ;; Create the dynamic optimal CSP denoting the
  ;; method selection problem.
  (let ((tpn-selection-problem
		 (construct-selection-problem-for-tpn 
		  tpn :return-description? return-description?)))
(print "ddddddddddddddddddddddddddddddddddddd")
	;;; Records the Optimal CSP for tpn selection on the TPN.
	(with-slots (selection-problem) tpn 
	  (setq selection-problem tpn-selection-problem))

	(values tpn-selection-problem))

  ;; Leave the activation-assignments in place, there is no
  ;; particular reason to remove them.
  ;;	(remove-tpn-activation-assignments! tpn)
  )

(defmethod frame-tpn-explanation-problem
	((tpn tpn-w-selection) &key (force-labeling? nil) (return-description? nil))

  "Given a TPN as input,

   RETURNS a TPN choice selection problem, framed
   as a dynamic optimal constraint satisfaction problem.
   All constraints will be decision/state variables with IN or OUT.

   Records the selection problem under TPN-SELECTION-PROBLEM.

   If FORCE-LABELING? is true, then generates a new
   labeling of the TPN with guard conditions, otherwise,
   it only labels if a labeling doesn't already exist.

   If RETURN-DESCRIPTION? is true, then returns a description
   of a choice-selection problem, rather than an instance
   of an opsat object, representing the problem."

  ;; Labels each TPN episode and decision-event with the
  ;; most immediate decision-assignment that activates that
  ;; episode/event.


  (when (or force-labeling?
			(not (tpn-labeled? tpn)))

;;;	(check-that-tpn-is-well-formed! tpn)
	(label-tpn-with-activation-assignments! tpn))

  ;; Create the dynamic optimal CSP denoting the
  ;; method selection problem.
  (let ((tpn-selection-problem
		 (construct-explanation-problem-for-tpn
		  tpn :return-description? return-description?)))

	;;; Records the Optimal CSP for tpn selection on the TPN.
	(with-slots (selection-problem) tpn 
	  (setq selection-problem tpn-selection-problem))

	(values tpn-selection-problem))

  ;; Leave the activation-assignments in place, there is no
  ;; particular reason to remove them.
  ;;	(remove-tpn-activation-assignments! tpn)
  )

(defmethod frame-tpn-planning-problem
	((tpn tpn-w-selection) &key (force-labeling? nil) (return-description? nil))

  "Given a TPN as input,

   RETURNS a TPN choice selection problem, framed
   as a dynamic optimal constraint satisfaction problem.
   All constraints will be decision/state variables with IN or OUT.

   Records the selection problem under TPN-SELECTION-PROBLEM.

   If FORCE-LABELING? is true, then generates a new
   labeling of the TPN with guard conditions, otherwise,
   it only labels if a labeling doesn't already exist.

   If RETURN-DESCRIPTION? is true, then returns a description
   of a choice-selection problem, rather than an instance
   of an opsat object, representing the problem."

  ;; Labels each TPN episode and decision-event with the
  ;; most immediate decision-assignment that activates that
  ;; episode/event.
  (when (or force-labeling?
			(not (tpn-labeled? tpn)))

;;;	(check-that-tpn-is-well-formed! tpn)
	(label-tpn-with-activation-assignments! tpn))

  ;; Create the dynamic optimal CSP denoting the
  ;; method selection problem.
  (let ((tpn-selection-problem
		 (construct-planning-problem-for-tpn
		  tpn :return-description? return-description?)))

	;;; Records the Optimal CSP for tpn selection on the TPN.
	(with-slots (selection-problem) tpn 
	  (setq selection-problem tpn-selection-problem))

	(values tpn-selection-problem))

  ;; Leave the activation-assignments in place, there is no
  ;; particular reason to remove them.
  ;;	(remove-tpn-activation-assignments! tpn)
  )

(defmethod get-tpn-selection-problem
	((tpn tpn-w-selection) &key 
						   (force-labeling? nil)
						   (return-description? nil))
  
  "Given a TPN as input,

   RETURNS a TPN choice selection problem, framed
   as a dynamic optimal constraint satisfaction problem.

   If a TPN selection problem already exists, then return it;
   else generate a new one.

   Records the selection problem under TPN-SELECTION-PROBLEM.

   If FORCE-LABELING? is true, AND a selectio problem 
   does not exist, then generates a new
   labeling of the TPN with guard conditions; 
   otherwise, it only labels if a labeling doesn't 
   already exist.

   If RETURN-DESCRIPTION? is true, then returns a description
   of a choice-selection problem, rather than an instance
   of an opsat object, representing the problem."

  (or (tpn-selection-problem tpn)

	  ;; Generate a tpn-selection-problem from TPN 
	  ;; if it doesn't already exist.
	  (frame-tpn-selection-problem tpn
								   :force-labeling? force-labeling?
								   :return-description? return-description?)))

(defmethod establish-tpn-selection-problem!
	((tpn tpn-w-selection) &key 
						   (force-labeling? nil)
						   (return-description? nil))
  
  "Given a TPN as input,

   Ensures that a choice selection problem is recorded for TPN,
   under TPN-SELECTION-PROBLEM.  Problem is framed as a dynamic 
   optimal constraint satisfaction problem.

   Returns T.

   If a TPN selection problem already exists then use it,
   else generates a new one.

   If FORCE-LABELING? is true, AND a selectio problem 
   does not exist, then generates a new
   labeling of the TPN with guard conditions; 
   otherwise, it only labels if a labeling doesn't 
   already exist.

   If RETURN-DESCRIPTION? is true, then returns a description
   of a choice-selection problem, rather than an instance
   of an opsat object, representing the problem."

  (get-tpn-selection-problem tpn
							 :force-labeling? force-labeling?
							 :return-description? return-description?)
  T)

(defmethod best-solution-to-selection-problem 
	((tpn tpn-w-selection))
  
  "Let TPN be a Temporal Plan Network for which 
   a selection-problem has already been framed.

   Find and return a best choice as a list of 
   choice assignment descriptions, each of the form:

      (= <decision-event-number> <choice-number>)"

  ;; A selection problem (formulated as an OPSAT problem)
  ;; is a constraint optimization problem with discrete dynamic 
  ;; variables, temporal events and simple temporal constraints.

  ;; Solve the OPSAT problem corresponding to the selection problem.
  ;; The solution, an assignment of choices, is an internal opsat assignment.
  (let* ((selection-problem (tpn-selection-problem tpn))
		 (solutions (find-solutions selection-problem :maximum-solution-count 1))
		 (best-solution (first solutions)))
	;;(format t "Best Solution to Selection Problem: ~A~%" best-solution)

	;; Returns best choice as a list of choice assignment descriptions
	;; of the form:
	;;    <assignment> | (NOT <assignment>) | (UNKNOWN <assignment>)
	;;    (note: only the first case should ever occur).
	
	;; where 
	
	;;    <assignment> ::= (= <decision-event-number> <choice-number>)
	
	(loop for assignment in best-solution
		collecting (constraint-system-assignment-description assignment)
				   ;; Note: 
				   ;; In the above we called CONSTRAINT-SYSTEM-ASSIGNMENT-DESCRIPTION
				   ;; rather than CONSTRAINT-SYSTEM-ASSIGNMENT-DESCRIPTION-W-TRUTH.
				   ;; There are two reasons for this: 1) we know that all assignments
				   ;; hold positively in the solution, and 2) the truth assignment on
				   ;; proposition was removed before it was returned.
				   )))

(defmethod best-explanation-to-selection-problem 
	((tpn tpn-w-selection))
  
  "Let TPN be a Temporal Plan Network for which 
   a selection-problem has already been framed.

   Find and return a best explanation as a list of 
   choice assignment descriptions, each of the form:

      (= <decision-event-number> <choice-number>)"

  ;; A selection problem (formulated as an OPSAT problem)
  ;; is a constraint optimization problem with discrete dynamic 
  ;; variables, temporal events and simple temporal constraints.

  ;; Solve the OPSAT problem corresponding to the selection problem.
  ;; The solution, an assignment of choices, is an internal opsat assignment.
  (let* ((selection-problem (tpn-selection-problem tpn))
		 (solutions (find-solutions selection-problem :maximum-solution-count 1))
		 (best-solution (first solutions)))

	;; Returns best choice as a list of choice assignment descriptions
	;; of the form:
	;;    <assignment> | (NOT <assignment>) | (UNKNOWN <assignment>)
	;;    (note: only the first case should ever occur).
	
	;; where 
	
	;;    <assignment> ::= (= <decision-event-number> <choice-number>)
	
	(loop for assignment in best-solution
		collecting (constraint-system-assignment-description assignment)
				   ;; Note: 
				   ;; In the above we called CONSTRAINT-SYSTEM-ASSIGNMENT-DESCRIPTION
				   ;; rather than CONSTRAINT-SYSTEM-ASSIGNMENT-DESCRIPTION-W-TRUTH.
				   ;; There are two reasons for this: 1) we know that all assignments
				   ;; hold positively in the solution, and 2) the truth assignment on
				   ;; proposition was removed before it was returned.
				   )))

(defmethod explanations-to-selection-problem 
	((tpn tpn-w-selection) &key 
						   (exp-count 1)
						   (explanation '())
						    (explanations '()))
  
  "Let TPN be a Temporal Plan Network for which 
   a selection-problem has already been framed.

   Find and return a best explanation as a list of 
   choice assignment descriptions, each of the form:

      (= constraint 'IN' or 'OUT')"

  ;; A selection problem (formulated as an OPSAT problem)
  ;; is a constraint optimization problem with discrete dynamic 
  ;; variables, temporal events and simple temporal constraints.

  ;; Solve the OPSAT problem corresponding to the selection problem.
  ;; The solution, an assignment of choices, is an internal opsat assignment.
  (let* ((selection-problem (tpn-selection-problem tpn))
		 (solutions (find-minimal-relaxations selection-problem :maximum-solution-count exp-count))
		 )


	(loop for solution in solutions
	do (setf explanation '())
	do (loop for assignment in solution
		do (push (constraint-system-assignment-description assignment) explanation))
	
	(push explanation explanations))
	
				  (values explanations))
  )

;;; The following isn't sufficicient as a test.
;;; The compiler will generate STCs that are along a thread of execution.
;;; This is okay.  The problem is with STCs that are not part of a thread.
;;; The issue is that we need to be able to identify those episodes that are
;;; the start of threads, versus those (STCs) that are not.  There is no
;;; easy way to detect this.
;;;(defmethod check-that-tpn-is-well-formed! ((tpn temporal-plan-network))
;;;  
;;;  "Given TPN, a temporal plan network, checks if TPN is well formed,
;;;   and issues warnings for any problems.
;;;
;;;   Choice selection requires that there are no simple temporal
;;;   constraints with a decision-event as a start.  Choice selection
;;;   fails if this occurs.  The RMPL compiler is required to enforce
;;;   this constraint."
;;;  
;;;  (loop for event in (tpn-decisions tpn)
;;;	  do (loop for episode in (event-successor-episodes event)
;;;			 when (episode-denotes-temporal-constraint? episode)
;;;			 do (warn "Decision event ~a is the start of simple temporal  ~
;;;                        constraint ~a. This is not allowed, Kirk will fail." 
;;;					  event episode))))

;;; ***********************************************************************
;;;
;;;     Functions For Testing Only
;;;
;;; ***********************************************************************

;;; To Do:
;;; o Move to a file test-support.lisp.
(defun read-write-tpns! (filename)
  
  "Let FILENAME be a SPEC2 tpn file, (with extension .tpn).
   Reads TPNs and writes them back out to a same named file, 
   but with extension .new-tpn.

   Used solely for debugging."
  
  (let* ((file-pathname (pathname filename))
		 (tpns (parse-tpn-file file-pathname
							   #'get-event
							   #'make-episode
							   #'add-episode
							   #'make-tpn))
		 (plan-file-pathname
		  (make-pathname :host (pathname-host file-pathname)
						 :device (pathname-device file-pathname)
						 :directory (pathname-directory file-pathname)
						 :name (pathname-name file-pathname)
						 :type "new-tpn")))
	(write-tpns plan-file-pathname
				tpns
				:environment *all*)))

(defun read-and-print-selection-problem! (filename)
    
  "Let FILENAME be a SPEC2 tpn file, (with extension .tpn).
   Reads TPNs and outputs a description of a selection problem
   for TPN."
  
  (let* ((file-pathname (pathname filename))
		 (tpns (parse-tpn-file file-pathname
							   #'get-event
							   #'make-episode
							   #'add-episode
							   #'make-tpn))
		 (tpn (first tpns))
		 (problem (frame-tpn-selection-problem 
				   tpn :return-description? T)))
	(print-tpn-problem problem)
	T))

(defun read-and-print-choice! (filename)
    
  "Let FILENAME be a SPEC2 tpn file, (with extension .tpn).
   Reads TPNs and outputs the optimal feasilbe choice to the
   selection problem for TPN."
  
  (let* ((file-pathname (pathname filename))
		 (tpns (parse-tpn-file file-pathname
							   #'get-event
							   #'make-episode
							   #'add-episode
							   #'make-tpn))
		 (tpn (first tpns)))

	(initialize-tpn-choice-selection! tpn)
	(format t "~% For TPN ~a, best choice = ~a."
			tpn (tpn-best-choice tpn))

	T))
