;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

;;; Copyright (c) 2008-9


(in-package :TPN)

;;; *********************************************************************
;;;
;;; ****  To Do ****
;;;
;;; *********************************************************************

;;; o Remove tpn- in other files, as the prefix to method calls and class instantiations.

;;; o Check whether any additional classes for event and episode are needed
;;;   to reflect the additional plan and plan-network classes that have
;;;   been added.

;;; o Correct the types of all methods, and organize methods with
;;;   classes, where appropriate.
;;;   - do this for all files within the TPN directory.

;;; o Rewrite creation and Make-instance support methods.
;;;   - do this for all files within the TPN directory.

;;; o For the XML reader, rewrite most construction functions to
;;;   be methods off of the plan and plan-network instance.

;;; *********************************************************************
;;;
;;; ****  Temporal Plan and Plan Network: Creation ****
;;;
;;; *********************************************************************

;;; This system reads in and creates objects representing a 
;;; Temporal Plan Network (TPN), as first introduced within the Kirk system.

;;; This file contains object definitions and macros for manipulating:

;;; o TPNs,
;;; o Events,
;;; o Episodes and
;;; o Choices.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Variables Specifying Classes used to Instantiate Temporal Plan Networks
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *read-tpn-type* 'TPN-W-SELECTION
  "Name of the class used to instantiate TPNs during XML read.")

(defvar *read-STN-type* ITC::'LTNP ;;'ITC-STN
  "Name of the class used to instantiate STNs during XML read.")

(defvar *read-event-type* 'event-w-selection
  "Name of the class used to instantiate TPN events during XML read.")

(defvar *read-episode-type* 'episode-w-activation-labeling
  "Name of the class used to instantiate a TPN's episodes during XML read.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Temporal Plan and Plan Network Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Defines Classes for the following types of Plans:

;;;    - Temporal Plan (a temporal plan without decisions).
;;;     - Supports Patrick Conrad's dispatcher, 
;;;     - A superclass of TPN.

;;;    - TP-w-histories
;;;     - A temporal Plan arranged into concurrent threads

;;;    - TP-w-STC-reasoning
;;;     - A temporal Plan arranged into concurrent threads

;;;    - Qualitative State Plan (a temporal plan with state episodes).
;;;      - supports Sulu.

;;;    - QSP w histories.

;;;    - QSP-w-STC-reasoning

;;;    - QSP w histories

;;;    - QSP w histories and flow tubes.
;;;      - supports Chekov.


;;; Defines Classes for the following types of Plan Networks
;;; (i.e., includes nested choice):

;;;    - temporal-plan-network

;;;    - TPN-w-histories

;;;    - TPN-w-STC-reasoning

;;;    - TPN-w-activation-labeling

;;;    - TPN-w-selection

;;;    - qualitative-state-plan-network
;;;      - a TPN with state episodes.
;;;      - supports Seung's planner.

;;;    - QSPN w histories.

;;;    - QSPN-w-STC-reasoning

;;;    - QSPN-w-activation-labeling

;;;    - QSPN-w-selection

;;;    - QSPN w flow-tubes.

;;;    - QSPN w flow-tubes and selection.

;;;    - QSPN w histories and flow-tubes.

;;;    - QSPN w histories, flow-tubes and selection.
;;;       - Supports the new Kirk/Chekov.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The Essential Temporal Plan Class
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Augment in order to explicitly represent the end event of each choice.
;;;   - How should we represent it?
;;;   - How should we extract it?
;;;   - How should we read and write it, when its represented explicitly in XML?

;;; o Should the TPN representation maintain the nested structure of the combinators,
;;;   or flatten, as it does now?

;;; o Currently we represent event equality through [0,0] constraints.
;;;   Should we instead represent it through explicit equality?

;;; o In spec2, temporal constraints and episodes are both represented as "arcs".
;;;   separate them out.  Also separate [0,0] constraints out into event equalities.

;;; o We may want a temporal-plan-fragment class that both temporal-plan and
;;;   temporal-plan-network build of, if we find our selves developing TP methods
;;;   that tpn wants to over ride.

;;; Optimization:
;;; o Make episodes a growable array, rather than a list, in order
;;;   to index the episode of a TPN more quickly.
(defclass temporal-plan () 

  ((name :initarg :name
		 :accessor tp-name
		 :documentation "Name of the instance.")

   (number-of-episodes 
	:initform 0 :documentation "The number of episodes currently 
                               associated with temporal-plan.")

   (start :initarg :start
		  :accessor tp-start
		  :documentation "The start event of the instance.  All of ~
                          its episodes should be reachable ~
                          from START.")

   (end :initarg :end
		:accessor tp-end
		:documentation "The end event of the instance.")
   
   (events :initform '()				 
		   :accessor tp-events
		   :documentation "Events of the instance.")						

   (event-ht :initform (make-hash-table :test #'equal)
			 :accessor tp-event-ht
			 :documentation "Hash table of event name/object pairs.")

   (episodes :initform '()				
			 :accessor tp-episodes
			 :documentation "Episodes of the instance, listed
                             in the order in which they were created.
                             The index of each episode corresponds to its
                             position in this list, starting with 0.")
   
   (last-of-episodes :initform '()				
;;;					 :accessor last-of-tp-episodes
					 :documentation "Last element of the list of episodes of the instance.
                                     This is used to maintain episodes in the order
                                     that they were created, so that reading and writing
                                     maintains the episode order."))

  (:documentation  "A class representing a Temporal Plan (TP)."))

(defmethod initialize-instance :after
		   ((TP Temporal-Plan)
			&key
			start-name
			end-name)
  (setup-temporal-constraint-recording! start-name TP)
  (with-slots (start end) TP
	(setq start (get-event start-name TP))
	(setq end (get-event END-name TP)))
  T)

(defmethod setup-temporal-constraint-recording! 
	(start-name (TP temporal-plan))
  
  "Sets up database to record Temporal events and constraints."

  (declare (ignore start-name))
  T)

(defmethod next-episode-index ((TP temporal-plan))
  
  "Returns the next continguous ineex to be assigned
   to a new episode of temporal plan TP."

  (with-slots (number-of-episodes) TP
	(let ((next-index number-of-episodes))
	  (incf number-of-episodes)
	  next-index)))

(defmethod nth-episode (index (TP temporal-plan))
  
  "Returns the episode of temporal plan TP with INDEX.
   where index is assigned to episodes of TP in the order created,
   starting form 0."   

  (with-slots (episodes) TP
	(nth index episodes)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The Essential Qualitative State Plan Class
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; Add state variables to this class.
(defclass qualitative-state-plan (temporal-plan)
  
  ()

  (:documentation  "A Temporal Plan with state variables and state episodes,
                    specifying constraints on the state variables."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Temporal and Qualitative State Plans with Histories
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass tp-w-histories (temporal-plan) 

  ((histories :initform '()				 
			  :reader tp-histories
			  :documentation "List of all histories of the temporal plan."))

  (:documentation  "A TP class that is organized as a set of concurrent histories"))



;;; To Do:
;;; Associate with each history the state variable whose behavior it defines.
(defclass qsp-w-histories (qualitative-state-plan tp-w-histories) 

  ()

  (:documentation  "A QSP class that is organized as a set of concurrent histories"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPs and QSPs that Supports Reasoning Over Simple Temporal Constraints
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass tp-w-stc-reasoning (temporal-plan)
  
  ((temporal-network 
;;;	:initarg :temporal-network
	:accessor tp-temporal-network
	:documentation "Network of temporal Constraints."))

  (:documentation  "A TP (tempooral plan) class that supports
                    reasoning over its simple temporal constraints"))

(defmethod setup-temporal-constraint-recording! 
	(start-name (TP tp-w-STC-reasoning))

  "Sets up database to record Temporal events and constraints."

  (with-slots (temporal-network name) tp

	(setq temporal-network (make-instance *read-STN-type*
										  :name `(STN ,name)
										  :start-name start-name))))


(defclass qsp-w-stc-reasoning (qualitative-state-plan tp-w-stc-reasoning)
  
  ()

  (:documentation  "A QSP (qualitative state plan) class that supports
                    reasoning over its simple temporal constraints"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; QSPs that Support Flow-tube Construction
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass qsp-w-flow-tubes (qsp-w-stc-reasoning)
  
  ()

  (:documentation  "A QSP (qualitative state plan) class that supports
                    flow tube construction and temporal reasoning."))


(defclass qsp-w-histories-and-flow-tubes (qsp-w-flow-tubes qsp-w-histories)
  
  ()

  (:documentation  "A QSP (qualitative state plan) with histories class
                    that supports flow tube construction and temporal 
                    reasoning."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Temporal Plan Network Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The Essential Temporal Plan Network Class
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass temporal-plan-network (temporal-plan copyable-mixin) 

  ((name :accessor tpn-name)

   (start :accessor tpn-start)

   (end :accessor tpn-end)
   
   (events :accessor tpn-events)						

   (event-ht :accessor tpn-event-ht)

   (decisions :initform '()				 
			  :accessor tpn-decisions
			  :documentation "Events that denote a set of choices.")

   (episodes :accessor tpn-episodes)

;;;   (last-of-episodes 				
;;;;;;					 :accessor last-of-tpn-episodes
;;;	)
   )

  (:documentation  "A class representing a Temporal Plan Network (TPN)."))

(defmethod initialize-instance :after
		   ((tpn Temporal-Plan-Network)
			&key decision-names)

  (unless (null decision-names)
	(dolist (decision-name decision-names)
	  (let ((decision-event (get-event decision-name TPN)))
		(decision-event! decision-event TPN))))
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The Essential Qualitative State Plan Network Class 
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass qualitative-state-plan-network
	(temporal-plan-network qualitative-state-plan)
  
  ()

  (:documentation  "A Qualitative State Plan with choice."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPNs and QSPNs comprised of Parallel Histories
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass tpn-w-histories (temporal-plan-network tp-w-histories) 

  ((histories :reader tpn-histories
			  :documentation "List of all histories of the temporal plan network."))

  (:documentation  "A TPN class that is organized as a set of concurrent histories"))


(defclass qspn-w-histories 
	(qualitative-state-plan-network tpn-w-histories qsp-w-histories) 

  ()

  (:documentation  "A QSPN class that is organized as a set of concurrent histories"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPN and QSPN Classes that Support Different Types of Reasoning
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPNs and QSPNs that Supports Reasoning Over Simple Temporal Constraints
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass tpn-w-stc-reasoning (temporal-plan-network tp-w-stc-reasoning)
  
  ((temporal-network 
	:accessor tpn-temporal-network
	:documentation "Network of temporal constraints."))

  (:documentation  "A TPN (tempooral plan network) class that supports
                    reasoning over its simple temporal constraints"))

(defclass qspn-w-stc-reasoning 
	(qualitative-state-plan-network tpn-w-stc-reasoning qsp-w-stc-reasoning)
  
  ()

  (:documentation  "A QSPN (qualitative state plan network) class that supports
                    reasoning over its simple temporal constraints"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPNS and QSPNS that Supports Labeling of Activation Assignments
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass tpn-w-activation-labeling (temporal-plan-network)
  
  ((activation-episode-ht :initform (make-hash-table :test #'equal)
						  :accessor tpn-activation-episode-ht
						  :documentation "Hash table of activation-assignment ~
                                          / episode list pairs, where each
                                          episode is activated by its affiliated
                                          assignment.")
   
   (labeled? :initform '()				 
			 :reader tpn-labeled?
			 :documentation "True iff TPN activities and events have been
                             labeled with activation assignments."))

  (:documentation  "A TPN (tempooral plan network) class that supports
                    automatic labeling of events and apisodes with the
                    choices that activate them."))
								  
(defmethod tpn-labeled! ((tpn tpn-w-activation-labeling))

  "Records that TPN has been labeled."
  
  (with-slots (labeled?) tpn
	(setq labeled? t)))

(defmethod tpn-unlabeled! ((tpn tpn-w-activation-labeling))

  "Records that TPN has NOT been labeled."
  
  (with-slots (labeled?) tpn
	(setq labeled? nil)))

(defclass qspn-w-activation-labeling
	(qualitative-state-plan-network tpn-w-activation-labeling)
  
  ()

  (:documentation  "A QSPN (qualitative state plan network) class that supports
                    automatic labeling of events and apisodes with the
                    choices that activate them."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPNs and QSPNs that support TPN Choice Selection
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass tpn-w-selection (tpn-w-activation-labeling tpn-w-stc-reasoning)
  
  ((selection-problem :initform nil
		 :reader TPN-selection-problem
		 :documentation "Dynamic optimal Constraint Satisfaction Problem 
                         that encodes the TPN selection problem.")
   (best-choice
	:initform nil
	:documentation "Is a description of a consistent assignment to choices
                    with the highest utility. description is a list of 
                    assignment descriptions:
                    (= <decision-event-number> <choice-number>).")


)

  (:documentation  "A TPN (temporal plan network) class that supports
                    plan generation by selecting optimal choices."))

;;; To Do:
;;; o Create a tpn-constraint-system if one doesn't already exist.
;;; o If TPN has been modified, then make sure that tpn-constraint-system
;;;   has been updated before this function returns.
(defmethod tpn-constraint-system ((tpn tpn-w-selection))
  "Returns the constraint system associated with TPN or nil if none."

  (let ((tpn-ocsp (tpn-selection-problem tpn)))
	(cond ((typep tpn-ocsp 'opsat-problem)
		   
		   (opsat-problem-constraint-system tpn-ocsp))
		  
		  (:otherwise nil))))

(defclass qspn-w-selection
	(qspn-w-activation-labeling qspn-w-stc-reasoning tpn-w-selection)
  
  ()

  (:documentation  "A QSPN (qualitative state plan network) class
                    that supports plan generation by selecting 
                    optimal choices."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; QSPNs that support Flow Tube Compilation
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass qspn-w-flow-tubes 
	(QSPN-w-selection QSP-w-flow-tubes) 

  ()

  (:documentation  "A QSPN class that supports flow tube construction
                    and assignment of choices to events and episodes."))

(defclass qspn-w-histories-and-flow-tubes
	(QSPN-w-histories QSPN-w-flow-tubes) 

  ()

  (:documentation
   "A QSPN class that is organized as a set of concurrent histories,
    and supports flow-tube compilation and selection."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; QSPN Class that supports Flow Tube Compilation AND Selection
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; The following defclass are conditioned on abcl because they reported 
;;; an "inconsistent precidence graph" error when compiling on 
;;; abcl 0.25. They do not give this error in Allegro 8.1. The reason
;;; for this error was not explored.

#-abcl
(defclass qspn-w-flow-tubes-and-selection
	(qspn-w-selection qspn-w-flow-tubes)
 
  ()

  (:documentation
   "A QSPN that supports compilation of flow-tubes and Selection"))

#-abcl
(defclass qspn-w-histories-flow-tubes-and-selection
	
	(qspn-w-flow-tubes-and-selection qspn-w-histories-and-flow-tubes) 

  ()

  (:documentation
   "A QSPN class that is organized as a set of concurrent histories,
    and supports flow-tube compilation and selection."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPN-Event Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Defines Classes for the following types of Events.

;;; o Event
;;; o Event-w-history
;;; o Event-w-STC-reasoning
;;; o Event-w-activation-labeling
;;; o Event-w-selection

;;; o State-Event
;;; o State-Event-w-history
;;; o State-Event-w-activation-labeling
;;; o State-Event-w-selection
;;; o State-event-w-flow-tube
;;; o State-event-w-history-and-flow-tube

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The Essential Event Class
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o A TPN event could be a specialization of an ITC event,
;;;   rather than having a pointer to an ITC event.
;;;   To do this, an ITC event needs to be a class.

(defclass event (copyable-mixin) 

  ((tp :initarg :tp
	   :accessor event-tp
	   :documentation "TP that event is associated with." )
   
   (name :initarg :name
		 :accessor event-name
		 :documentation "Name of the event." )
   
   (decision? :initform nil
			  :reader decision-event?
			  :documentation "True iff event denotes a choice of threads.")

   (successor-episodes :initform '()
					   :accessor event-successor-episodes
					   :documentation "List of episodes that start at ~
									    this event.")
   
   (predecessor-episodes :initform '()
						 :accessor event-predecessor-episodes
						 :documentation "List of episodes that end at ~
									    this event."))

  (:documentation "A class representing an event."))


(defmacro find-tp-event (event-name TP)

  " Looks up and returns the event object associated 
    with EVENT-NAME in TP.  If the event object
    does not exist, it returns NIL."
  
  `(gethash ,event-name (tp-event-ht ,TP)))

(defun find-tp-event-regex (regex tp)
  "Looks up and returns a list of events whose names match
regexp. List is empty if no matches found."

  (let ((results nil))
	(dolist (e (tp-events tp) results)
	  (when (scan regex (event-name e))
		(push e results)))))

(defmacro find-tpn-event (event-name TPN)

  " Looks up and returns the event object associated 
    with EVENT-NAME in TP.  If the event object
    does not exist, it returns NIL."
  
  `(gethash ,event-name (tpn-event-ht ,TPN)))



(defmethod initialize-instance :after
		   ((event event)
			&key)

  (with-slots (TP name itc-event) event

	;; Add to events of TP
	(push event (tp-events TP))

	;; insert event into network's hash table,
	(setf (find-tp-event name TP) event)

	;; Add to event a corresponding event
	;; within the incremental temporal consistency
	;; algorithm, ITC.
	(setq itc-event
	  (make-itc-event event TP)))
  T)


(defmethod decision-event! 
	
	((event event)(tpn temporal-plan-network))
  
  "Asserts that EVENT denotes a decision between
   alternative threads of a TPN."
  
  (with-slots (decision?) event
	(when (null decision?)
	  (setq decision? T)
	  (with-slots (decisions) tpn
		(push event decisions)))))


(defmethod event-has-predecessors? 
	
	((event event))
  
  "Returns True iff event has one or more predecessor episodes."
  
  (with-slots (predecessor-episodes) event
	(not (null (remove-if #'episode-istemporalconstraint predecessor-episodes)))))


(defmethod event-has-multiple-predecessors? ((event event))
  
  "Returns True iff event has TWO or more predecessor episodes."
  
  (with-slots (predecessor-episodes) event
	(two-or-more? predecessor-episodes)))

(defun two-or-more? (l)
  
  "Returns True iff list L has two or more elements."
  
  (not (null (cdr l))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; An Event of a State Plan
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass state-event (event) 

  ()

  (:documentation "A class representing a temporal event that is
                   part of a QSP or QSPN."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; An Event that is Part of a History
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass event-w-history (event) 

  ()

  (:documentation "A class representing a temporal event that is
                   part of a history."))


(defclass state-event-w-history (state-event event-w-history) 

  ()

    (:documentation "A class representing a temporal event of a QSP or QSPN
                     that is part of a history."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Event Classes that Support Different Types of Reasoning
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Event with Reasoning Over Simple Temporal Constraints
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass event-w-stc-reasoning (event) 

  ((itc-event :initform nil
			  :accessor event-itc-event
			  :documentation " The corresponding event object used by the ~
                               incremental temporal consistency (ITC) ~
                               algorithm to compute event bounds."))

  (:documentation "A class representing a temporal event that supports 
                   reasoning over its simple temporal constraints."))

(defclass state-event-w-stc-reasoning (state-event event-w-stc-reasoning) 

  ()

  (:documentation "A class representing a temporal state event that supports 
                   reasoning over its simple temporal constraints."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Event that Supports Labeling of Activation Assignments
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass event-w-activation-labeling (event) 

  ((activation-assignment :initform nil
						  :accessor event-activation-assignment
						  :documentation "A choice assignment that activates event.")
   (activation-label      :initform nil
						  :accessor event-activation-label
						  :documentation "A conjunction (AND) of all choice assignments that activate an event."))  

  (:documentation "A class representing an event of a TPN that can be labeled
                   with a choice assignment that activates it."))


(defclass state-event-w-activation-labeling
	
	(state-event event-w-activation-labeling) 

  ()

  (:documentation "A class representing an event of a QSPN that can be labeled
                   with a choice assignment that activates it."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Event that Supports TPN Choice Selection
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass event-w-selection 
	(event-w-activation-labeling event-w-stc-reasoning) 

  ()

  (:documentation "A class representing a temporal event of a TPN that supports 
                   selection of an optimal plan from alternative choices."))


(defclass state-event-w-selection 
	(state-event-w-activation-labeling state-event-w-stc-reasoning event-w-selection) 

  ()

  (:documentation "A class representing a temporal event of a QSPN that supports 
                   selection of an optimal plan from alternative choices."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Events that Support Flow Tube Construction 
;;; and Choice Selection
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass state-event-w-flow-tube
	(state-event-w-selection) 

  ()

  (:documentation "A class representing a temporal event of a TPN that supports 
                   flow tubes and selection of an optimal plan from alternative 
                   choices."))


(defclass state-event-w-history-and-flow-tube
	
	(state-event-w-flow-tube state-event-w-history) 

  ()

  (:documentation "A class representing a temporal event that is
                   part of a history."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Episode Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Defines the following classes for episode:

;;; o Episode
;;; o Episode-w-history
;;; o episode-w-stc-reasoning
;;; o episode-w-activation-labeling
;;; o episode-w-selection

;;; o state-episode
;;; o state-episode-w-history
;;; o state-episode-w-flow-tube
;;; o state-episode-w-history-and-flow-tube

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The Essential TPN-Episode Class
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass episode (copyable-mixin) 

  ((tp :initarg :tp
	   :accessor episode-tp
	   :documentation "TPN of TP that episode is associated with." )
   
   (index
	:initform nil
	:reader episode-index
	:documentation "Index used to refer to episode of TP." )
   
   (start :initarg :start
		  :accessor episode-start
		  :documentation "The event denoting the start of episode.")

   (end :initarg :end
		:accessor episode-end
		:documentation "The event denoting the end of episode.")
   
   (lowerbound :initarg :lowerbound
			   :accessor episode-lowerbound
			   :documentation "A lower bound on the duration of the episode.")

   (upperbound :initarg :upperbound
			   :accessor episode-upperbound
			   :documentation " An upper bound on the duration of the episode.")
   
   (temporal-duration :initarg :temporal-duration
					  :accessor episode-temporal-duration
					  :documentation "Stores an object representing
				      the temporal information for this episode. TODO:
				      Consolidate current lowerbound, upperbound,
				      mean, and variance slots into this slot.")
   
   (cost :initarg :cost
		 :accessor episode-cost
		 :initform 0
		 :documentation " Cost of executing episode.")

   (description :initarg :description
				:accessor episode-description
				:documentation "Description of episode, ~
                                of the form (EPISODE . alist)")
   
   (support
	:initarg :support
	:initform :control-program
	:reader episode-support
	:documentation "Specifies the ting for which episode is a consequence.
                    Either an atomic symbol, or a list of the form:

                     (<reason> . <antecedents>), 

                    where antecedents collectively support the existence of episode based
                    on reason.  Defaults to :CONTROL-PROGRAM,
                    meaning that episode was specified in a control program." )

   (relaxable
	:initarg :relaxable
	:accessor episode-relaxable
	:initform nil
	:documentation "Specifies if the episode can be relaxed in case the TPN is inconsistent.")

   (isStateConstraint
	:initarg :isStateConstraint
	:accessor episode-isStateConstraint
	:initform nil
	:documentation "Specifies if the episode represents a state constraint.")


   (isTemporalConstraint
	:initarg :isTemporalConstraint
	:accessor episode-isTemporalConstraint
	:initform nil
	:documentation "Specifies if the episode represents a temporal constraint.")

   (isActivity
	:initarg :isActivity
	:accessor episode-isActivity
	:initform nil
	:documentation "Specifies if the episode represents an activity.")

   (isDummy
	:initarg :isDummy
	:accessor episode-isDummy
	:initform nil
	:documentation "Specifies if the episode represents a redundant linking episode.")

   (disabled
	:initarg :disabled
	:accessor episode-disabled
	:initform nil
	:documentation "Specifies if the episode has been disabled.")
   
   (executed
	:initarg :executed
	:accessor episode-executed
	:initform nil
	:documentation "Specifies if the episode has been executed or not.")

   (controllable
	:initarg :controllable
	:accessor episode-controllable
	:initform t
	:documentation "Specifies if the episode is temporally controllable or not.")

   (mean
	:initarg :mean
	:accessor episode-mean
	:initform 0
	:documentation "If the episode is uncontrollable, the temporal mean of its distribution.")

   (variance
	:initarg :variance
	:accessor episode-variance
	:initform *infinity*
	:documentation "If the episode is uncontrollable, the temporal variance of its distribution.")

   )

  (:documentation "A class representing an episode."))

(defmethod initialize-instance :after
		   ((episode episode)
			&key)

  (with-slots (TP index) episode

	(setq index (next-episode-index TP))
	
	;; Add episode to episodes of TP whil ensuring
	;; that episodes are listed in the order that
	;; they are created.
	(with-slots (episodes last-of-episodes) TP
	  (let ((episode-list (list episode)))
		(cond ((null episodes)
			   (setf episodes episode-list))

			  (:otherwise
			   (setf (rest last-of-episodes)
				 episode-list)))
		
		(setq last-of-episodes episode-list))))
  T)

(defmethod episode-name ((episode episode))
  
  "The name of an episode is (TP-NAME INDEX),
   where TP-name is the name of the 
   temporal plan (network) (TP)
   that episode is part of, and INDEX is the unique
   index assigned episode within its TP."
  
  `(,(tp-name (episode-tp episode))
	,(episode-index episode)))


;;; To Do:
;;; o Not all of these should be supported for all episode types.
;;;  - They could be part of a "classic episode."
;;;  - All episodes share macro, but what else?
;;;  - some of these should be associated with "executable" episode.


(defmethod episode-ask ((episode episode))

  "Returns the ask constraint of EPISODE."
  
  (description-value :ask (episode-description episode)))


(defmethod episode-tell ((episode episode))

  "Returns the tell constraint of EPISODE."
  
  (description-value :tell (episode-description episode)))


(defmethod episode-primitive ((episode episode))

  "Returns the primitive function invocation of EPISODE."
  
  (description-value :primitive (episode-description episode)))

(defmethod (setf episode-primitive) (new-primitive (episode episode))
  (let ((primitive-list (find :primitive (cdr (episode-description episode)) :key #'car)))
	(when (null primitive-list)
	  (error "Attempting to change episode primitive for an episode
	  that has no primitive."))
	(setf (cadr primitive-list) new-primitive)))

(defmethod episode-type ((episode episode))

  "Returns the primitive function invocation of EPISODE."
  
  (description-value :type (episode-description episode)))


(defmethod episode-macro ((episode episode))

  "Returns the macro function invocation of EPISODE."
  
  (description-value :macro (episode-description episode)))


(defmethod source-is-an-episode? ((constraint essential-constraint))
  
  
  "Returns true iff constraint was sourced (generated) by an episode."
  
  (typep (constraint-source constraint) 'episode))


(defmethod episode-denotes-temporal-constraint? ((episode episode))

  "Returns True if episode is used to specify a temporal constraint.
   We take this to be true if episode not a constraint tell/ask or a
   function innvocation (primitive or macro)."

  (and (null (episode-ask episode))
	   (null (episode-tell episode))
	   (null (episode-primitive episode))
	   (null (episode-macro episode))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; An Episode that is Part of a History
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass episode-w-history (episode) 

  ()

  (:documentation "A class representing an episode that is
                   part of a history."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Episode Class for QSPs and TSPNs
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass state-episode (episode) 

  ((goal-region :initarg :goal-region
				:initform T
				:accessor episode-goal-region
				:documentation "The state constraint (a polygon) that
                                must hold at the end eent of the episode."))

  (:documentation "A class representing a state episode, which imposes
                   constraints on the system state over an interal of time"))

(defclass state-episode-w-history (state-episode episode-w-history) 

  ()

  (:documentation "A class representing a state episode that is
                   part of a history."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Episode Classes that Support Different Types of Reasoning
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Episode with Reasoning Over Simple Temporal Constraints
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass episode-w-stc-reasoning (episode) 

  ((itc-stc :initform nil
			:accessor episode-itc-stc
			:documentation "An ITC simple temporal constraint, denoting ~
                            the bounds on the duration of the episode."))

  (:documentation "A class representing an episode that supports 
                   reasoning over its simple temporal constraints."))

(defmethod initialize-instance :after
		   ((episode episode-w-STC-reasoning)
			&key)

  (with-slots (itc-stc TP start end lowerbound upperbound temporal-duration cost) 
	  episode
	
	
	(if (episode-contains-loops? episode)
		(setq itc-stc (make-itc-looping-temporal-constraint
					   start end 
					   temporal-duration
					   cost
					   TP))

		(setq itc-stc (make-itc-simple-temporal-constraint
					   start end 
					   lowerbound upperbound
					   TP))))
  T)

(defclass state-episode-w-stc-reasoning 
	(state-episode episode-w-STC-reasoning) 

  ()

  (:documentation "A class representing a state episode that supports 
                   reasoning over its simple temporal constraints."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Episode that Supports Labeling of Activation Assignments
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass episode-w-activation-labeling (episode) 

  ((activation-assignment :initform nil
						  :accessor episode-activation-assignment
						  :documentation "A choice assignment that
                                          activates episode. If more
                                          than one (in a list),
                                          assumed to be conjunction.")
   (activation-assignment-is-conjunction?
	:initform nil
	:accessor episode-activation-assignment-is-conjunction?
	:documentation "T iff ACTIVATION-ASSIGNMENT is a conjunction (list
	of clauses).")
   (activation-label      :initform nil
						  :accessor episode-activation-label
						  :documentation "A conjunction (AND) of all choice 
                                          assignments that activate an episode."))

  (:documentation "A class representing an episode that can be labeled
                   with a choice assignment that activates it."))

(defmethod episode-labeled? 
	((episode episode-w-activation-labeling))

  "Returns true iff EPISODE is labeled with a non-null 
   activation-assignment."

  (not (null 
		(episode-activation-assignment
		 episode))))

(defmacro episode-activation-decision-event (episode)
  
  "Given EPISODE with a non-null activation-assignment,
   returns the decision-event of the activation-assignment."
  
  `(assignment-variable
	(episode-activation-assignment ,episode)))

(defmacro episode-activation-choice (episode)

  "Given EPISODE with a non-null activation-assignment,
   returns the choice of the activation-assignment."
  
  `(assignment-value
	(episode-activation-assignment ,episode)))

(defclass state-episode-w-activation-labeling 
	(state-episode episode-w-activation-labeling) 

  ()

  (:documentation "A class representing a state episode that can be labeled
                   with a choice assignment that activates it."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Episode that supports TPN Choice Selection
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass episode-w-selection 
	(episode-w-activation-labeling episode-w-stc-reasoning) 

  ()

  (:documentation "A class representing an episode that can be labeled
                   with a choice assignment that activates it."))

(defclass state-episode-w-selection 
	(state-episode-w-activation-labeling state-episode-w-stc-reasoning episode-w-selection) 

  ()

  (:documentation "A class representing a state episode that can be labeled
                   with a choice assignment that activates it."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Episode that supports Flow Tube Construction 
;;; and Choice Selection
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Eventually we will want to split this into an episode w selection
;;;   and one without.  The later is all we need for Chekov.
(defclass state-episode-w-flow-tube
	(state-episode-w-selection) 

  ((flow-tube :initform nil
			  :accessor episode-flow-tube
			  :documentation "A flowtube object representing all possible
                              state trajectories during episode that satsify
                              episode's constraints and that of all successors.")
   
   (infeasible? :initform nil
				:reader episode-infeasible?
				:documentation "True iff no trajectory of the physical plant
                                exists that satisfies episode, and its 
                                surrounding constraints")
   
   (visited? :initform nil
			 :reader episode-visited?
			 :documentation "True iff episode has been visited during
                             flow-tube compilation."))

  (:documentation "A class representing a TPN state episode that has the 
                   ability to generate and represent flow tubes."))

(defclass state-episode-w-history-and-flow-tube
	
	(state-episode-w-history state-episode-w-flow-tube) 

  ()

  (:documentation "A class representing a state episode that is
                   part of a history."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Finding Episodes that correspond to a particular Activation Assignment
;;; of a TPN
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro tpn-activation-assignment-episodes 
	(activation-assignment TPN)

  " Looks up and returns the list of TPN episodes that
    are activated by ACTIVATION-ASSIGNMENT."
  
  `(gethash ,activation-assignment
			(tpn-activation-episode-ht ,TPN)))

(defmacro clear-tpn-activation-assignment-episodes! (tpn)

  " Removes all activation-assignment / activated episode
    associations for TPN."
  
  `(clrhash (tpn-activation-episode-ht ,tpn)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPN Flow Tube Class
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass flow-tube () 

  ((initial-region :initarg :initial-region
				   :accessor flow-tube-initial-region
				   :documentation "All states at start that satisfy the episode, 
                        dynamic and successor constraints." )
   
   (tightened-upperbound :initarg :upperbound
						 :accessor flow-tube-tightened-upperbound
						 :documentation "An upper bound on the duration of the flow tube, 
                               based on dynamical constraints."))

  (:documentation "A class representing the flow tube of a TPN episode."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; TPN History Class
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass history () 

  ((start :initarg :start
		  :accessor history-start
		  :documentation "The event denoting the start of history.")

   (end :initarg :end
		:accessor history-end
		:documentation "The event denoting the end of history."))

  (:documentation "A class representing a history."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Printing TPN Instances
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Clean up print out of tpns, episodes and events.
;;;   Make it print readably.

(defmethod print-object ((obj event) stream)
  (with-slots (name) obj 
;;;    (format stream "<event ~a>" name)
	(format stream "<evt ~a>" name)))

;;; Fix:
;;; o This doesn't appear to handle PRIMITIVE correctly.
;;;   Note that episode may be a PRIMITVE, MACRO, TELL or ASK.
;;;   Print should distinguish each.

(defmethod print-object ((self episode) stream)

  "Print short description of an EPISODE."
  
  (with-slots (start end upperbound lowerbound description) self

    (let ((primitive (description-value 'primitive description)))

      (cond (primitive
			 (format stream "<Episode ~a ~a [~a,~a] ~a>" 
					 start end lowerbound upperbound primitive))
			(:otherwise
			 (format stream "<Episode ~a ~a  [~a,~a]>"
					 start end lowerbound upperbound))))))

(defmethod print-details ((episode episode) 
						  &optional (stream t))

  "Print detailed description of a EPISODE."
  
  (with-slots (cost description activation-assignment itc-stc) episode

	(format stream "~%    ~a:" episode)

	(format stream "~%       Cost: ~a" cost)

	(unless (null description)
	  (format stream "~%       Description: ~a" description))

	(unless (null activation-assignment)
	  (format stream "~%       guard: ~a" activation-assignment))
	
	(unless (null itc-stc)
	  (format stream "~%       itc-stc: ~a" itc-stc))))

(defmethod print-details :after 
		   ((episode episode-w-stc-reasoning) &optional (stream t))

  "Print detailed description of an EPISODE with STC-reasoning."
  
  (with-slots (itc-stc) episode

	(unless (null itc-stc)
	  (format stream "~%       itc-stc: ~a" itc-stc))))

(defmethod print-details :after 
		   ((episode episode-w-activation-labeling) &optional (stream t))

  "Print detailed description of an EPISODE with activation-labeling."
  
  (with-slots (activation-assignment) episode

	(unless (null activation-assignment)
	  (format stream "~%       guard: ~a" activation-assignment))))

(defmethod print-object ((self temporal-plan) stream)

  "Print short description of TP."
  
  (with-slots (name start end) self
    (format stream "<TP ~a [~a ~a]>"
			name start end)))

(defmethod print-object ((self temporal-plan-network) stream)

  "Print short description of TPN."
  
  (with-slots (name start end) self
    (format stream "<TPN ~a [~a ~a]>"
			name start end)))

(defmethod print-details 
	((tp temporal-plan) &optional (stream t))

  "Print detailed description of a TP."
  
  (with-slots (events episodes) tp
	(format stream "~% ~a:" tp)
	(format stream "~%    Events:")
	(loop for e in events
		do (format stream "~%       ~a" e))
	(format stream "~%    Episodes:")
	(loop for e in episodes
		do (print-details e stream)
		   (format t "~%"))))

(defmethod print-details 
	((tpn temporal-plan-network) &optional (stream t))

  "Print detailed description of a TPN."
  
  (with-slots (decisions events episodes) tpn
	(format stream "~% ~a:" tpn)
	(format stream "~%    Decisions:")
    (loop for d in decisions
		do (format stream "~%       ~a" d))
	(format stream "~%    Events:")
	(loop for e in events
		do (format stream "~%       ~a" e))
	(format stream "~%    Episodes:")
	(loop for e in episodes
		do (print-details e stream)
		   (format t "~%"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Functions for Creating TPNs
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Creating TPNs

;;; To Do:
;;; o Handle parameters.
;;; o Think through which type of TPN to create here.

;;; Fix;
;;; o Need to be able to specify the type of the TP or TPN being
;;;   made.
(defun make-tpn (tpn-name start-name end-name decision-names)

  "Returns a tpn object with TPN-NAME and start, end and decision
   events, named START-NAME, END-NAME and DECISION-NAME, respectively.
   If there are no decision events, then nil is supplied.  A decision
   event denotes a non-deterministic choice between alternative threads."
  
  (make-instance *read-tpn-type*
	:name tpn-name
	:start-name start-name
	:end-name end-name
	:decision-names decision-names)
)

;;; Creating TPN Events

(defmethod get-event (event-name (tp temporal-plan))

  "Return a unique event object for EVENT-NAME of a TP or TPN.
   A new event object is created if one does not exist 
   named EVENT-NAME in TP."

  (or (find-tp-event event-name tp)
	  (make-tp-event event-name tp)))

;;; Fix:
;;; o The event type should come from a slot on TP.

(defmethod make-tp-event (event-name (tp temporal-plan))

  "Returns an event object with NAME."
  
  (make-instance *read-event-type*
	:name event-name
	:tp TP))

;;; Creating TPN Episodes

;;; Note:
;;; o This extracts TELL, ASK, PRIMITIVE or MACRO from episode description
;;;   on demand.  Might want to error check them on input.
;;;   We assume that there is at most one entry for each of these fields
;;;   per episode.

;;; Fix:
;;; o The episode type should come from a slot on
;;;   TP.

;;; To Do:
;;; o Handle parameters.

(defmethod make-episode 
	
	((start event) (end event) 
	 lowerbound upperbound cost
	 description (tp temporal-plan) 
	 &key (support :control-program) (relaxable nil) 
	 (executed nil) (disabled nil)
	 (isDummy nil) 
	 (isTemporalConstraint nil)
	 (isStateConstraint nil)
	 (isActivity nil)
     (controllable t)
	 (temporal-duration (make-instance 'bounded-temporal-duration :lower-bound lowerbound :upper-bound upperbound))
	 &allow-other-keys)

  "Returns an EPISODE object, which is part of temporal plan TP.
   DESCRIPTION specifies the behavior during episode.
   START and END are the start an end events of the episode.
   The duration of the episode is bounded by LOWERBOUND and UPPERBOUND,
   Support specifies what episode is a consequence of.
   Support is either an atomic symbol, or a list of the form:
      (<reason> . <antecedents>), where antecedents
   collectively support the existence of episode based
   on reason.  Defaults to :CONTROL-PROGRAM,
   meaning that episode was specified in a control program.
  SUPPORT defaults to :CONTROL-PROGRAM, specifying that episode
  is part of a control program, and is a goal behavior, specified
  by the user."

  (make-instance *read-episode-type*
	:tp TP
	:start start
	:end end
	:lowerbound  lowerbound
	:upperbound upperbound
	:cost cost
	:description description
	:support support
	:relaxable relaxable
	:executed executed
	:disabled disabled
	:isDummy isDummy
	:isTemporalConstraint isTemporalConstraint
	:isStateConstraint isStateConstraint
	:isActivity isActivity
    :controllable controllable
	:temporal-duration temporal-duration))

(defmethod add-episode 
	((start event) (end event) (episode episode))

  "Associates EPISODE with its START and END event."
  
  (with-slots (successor-episodes) start
    (setf successor-episodes
		  `(,episode . ,successor-episodes)))
  
  (with-slots (predecessor-episodes) end
    (setf predecessor-episodes
	  `(,episode . ,predecessor-episodes))))


(defmethod remove-episode (episode)

  "Removes episode from its associated tp."

  (let ((start-event (episode-start episode))
  		(end-event (episode-end episode)))
  	(with-slots (successor-episodes) start-event
  	  (setf successor-episodes
  			(remove episode successor-episodes)))
  	(with-slots (predecessor-episodes) end-event
  	  (setf predecessor-episodes
  			(remove episode predecessor-episodes)))
	;; Remove episode from tp
	(setf (tp-episodes (episode-tp episode)) 
		  (remove episode (tp-episodes (episode-tp episode))))
  ;; Decrement number-of-episodes and reset pointer to last episode.
	(with-slots (number-of-episodes last-of-episodes) (episode-tp episode)
	  (decf number-of-episodes)
	  (setf last-of-episodes 
			(nthcdr (- number-of-episodes 1) (tp-episodes (episode-tp episode)))))))



;;; Creating ITC Events

;;; FIX:
;;; o This should be TP-temporal-network,
;;;   Not TPN-temporal-network.  Rearrange classes.
(defmethod make-itc-event ((event event) (tp temporal-plan))

  "Returns an event within the Incremental Temporal 
   Consistency (ITC) package that corresponds to 
   EVENT."
  
  (let ((event-name (event-name event))
		(STN (TP-temporal-network TP)))
	(itc:get-event event-name STN)))


;;; Creating ITC Simple Temporal Constraints

(defmethod make-itc-simple-temporal-constraint

	((start-event event) 
	 (end-event event)
	 lowerbound 
	 upperbound 
	 (tp temporal-plan))

  "Returns a SIMPLE TEMPORAL CONSTRAINT that bounds the duration
   from START-EVENT to END-EVENT to interval [LOWERBOUND, UPPERBOUND]."

  (let ((itc-start (event-itc-event start-event))
		(itc-end (event-itc-event end-event)))
	(insert-new-constraint (TP-temporal-network TP)
						   :start itc-start
						   :end itc-end 
						   :lb lowerbound
						   :ub upperbound
						   :source TP)))

(defmethod make-itc-looping-temporal-constraint

	((start-event event) 
	 (end-event event)
	 temporal-duration
	 cost
	 (tp temporal-plan))

  "Returns a SIMPLE TEMPORAL CONSTRAINT that bounds the duration
   from START-EVENT to END-EVENT to interval [LOWERBOUND, UPPERBOUND]."

  (let ((itc-start (event-itc-event start-event))
		(itc-end (event-itc-event end-event)))
  	(ITC::insert-new-looping-constraint (TP-temporal-network TP)
										:start itc-start
										:end itc-end 
										:N-lb (ltd-n-lower-bound temporal-duration)
										:N-ub (ltd-n-upper-bound temporal-duration)
										:t-lb (btd-lower-bound(ltd-per-loop-duration temporal-duration))
										:t-ub (btd-upper-bound(ltd-per-loop-duration temporal-duration))
										:reward-function cost  
										:source TP)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Utilities for loops
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod episode-contains-loops? ((episode episode))
  (with-slots (temporal-duration) episode
	(typep temporal-duration 'looping-temporal-duration)))

(defmethod tp-contains-loops? ((tp temporal-plan))
  (with-slots (episodes) tp
	(some #'episode-contains-loops? episodes)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Structure Implementation:
;;; This is an example of how to build the TPN with structures
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Update the following functions, similar to what has been done
;;;   for the corresponding class functions.

;;;(defstruct tpn 
;;;  name									; Name of the TPN
;;;  start									; The start event of the TPN.
;;;										    ; All of the TPN episodes should
;;;										    ; be reachable from START.
;;;  end									; The end event of the TPN.
;;;  )

;;;(defstruct tpnepisode 
;;;  to									    ; The event that episode terminates on.
;;;  info									; An association list, describing 
;;;										    ; the activity that episode represents.
;;;  )

;;;(defstruct tpnevent 
;;;  name									; Identifier (name) of the event.
;;;  episodesout							; List of TPN episodes that start at
;;;										    ; this event.
;;;  )
;;;
;;;(defun mktpn (name startevent endevent)
;;;
;;;  "Returns a tpn object with NAME and START and END EVENTS."
;;;
;;;  (make-tpn :name name 
;;;			:start startevent 
;;;			:end endevent))
;;;
;;;(defun mkepisode (to info)
;;;  
;;;  "Returns a TPN EPISODE object that terminates on TO and
;;;   has associated information INFO, in the form of an association list."
;;;
;;;  (make-tpnepisode :to to 
;;;				   :info info))
;;;
;;;(defun mkevent (name)
;;;  
;;;  "Returns a TPN event object with identifier NAME."
;;;
;;;  (make-tpnevent :name name
;;;				 :episodesout nil))
;;;
;;;(defun add-episode (event episode)
;;;
;;;  "Associates EPISODE with its start event EVENT."
;;;  
;;;  (setf (tpnevent-episodesout event)
;;;	(cons episode (tpnevent-episodesout event))))
