;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

;;; Copyright (c) 2009

;;; Authors: Brian Williams and Paul Robertson

(in-package :TPN)

(mtu::declare-debug-print-system! :TPN)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; General Utility Macros for Temporal Plan Networks (TPNs)
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Macros:  All
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconstant *all* :ALL "Indicates that all elements are used.")

(defmacro all? (keyword)
  
  "Returns true if all elements are to be used."
  
  `(eql ,keyword ,*all*))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Macros:  Association Lists
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro association? (item)

  "Returns True iff ITEM is an association in an association list."
  
  `(consp ,item))

(defmacro make-association (attribute value)
  
  "Creates an ATTRIBUTE / VALUE assocation (pair)."
  
  `(list ,attribute ,value))

(defmacro association-attribute (association)
  
  "Returns the attribute of an ASSOCIATION."
  
  `(first ,association))

(defmacro association-value (association)

  "Returns the value of an ASSOCIATION."
  
  `(second ,association))

(defmacro make-association-list ()
  
  "Creates and returns an empty association list."
  
  '(list nil))

(defmacro alist-value (attribute associations)

  "Returns the value associated with ATTRIBUTE in ASSOCATIONS."
  
  (let ((gs (gensym)))
	`(let ((,gs (assoc ,attribute ,associations :test #'symbol-eql)))
	   (when ,gs
		 (association-value ,gs)))))

(defmacro add-association! (attribute value associations)

  "Returns ASSOCIATIONS list, augmented with ATTRIBUTE 
   associated with VALUE."
  
  `(push (make-association ,attribute ,value)
		 ,associations))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Macros:  Generic Descriptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro description? (item)

  "Returns true if ITEM is a description.  A description 
   consistents of a TYPE followed by a list of attribute/value
   ASSOCIATIONS."
  
  `(consp ,item))

(defmacro make-description (type associations)
  
  "A description is a TYPE plus a list of attribute/value
   ASSOCIATIONS."
  
  `(cons ,type ,associations))

(defmacro description-type (description)
  
  "Returns the TYPE of a DESCRIPTION."

  `(first ,description))

(defmacro description-associations (description)
  
  "Returns the list of associations of DESCRIPTION."
  
  `(rest ,description))

(defmacro description-value (attribute description)

  "Returns the value of ATTRIBUTE within DESCRIPTION's
   associations.  Returns nil if no such attribute exists."
  
  `(alist-value ,attribute (description-associations ,description)))

(defmacro description-attribute-associations (attribute description)

  "Returns the associations of ATTRIBUTE within DESCRIPTION's
   associations.  Returns nil if no such attribute exists."

  ;; NOTE: The use of the word associations is deceptive, an episode's
  ;; description actually takes the list form: (:episode (:primitive "veh.method()"))
  ;; It is necessary to use description-associations to return the
  ;; associations list portion of the descriptions.

  `(assoc ,attribute (description-associations ,description) :test #'symbol-eql))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Macros:  SPEC2 descriptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Check:
;;; o In the following its strange that the first item in the LXML
;;;   is the SPEC2 description.  What is the rest of the LXML?
;;;   Is this correct?
(defmacro TPN-SPEC2-LXML-description? (lxml-description)
  
  "Returns TRUE if LXML-DESCRIPTION is a SPEC2 description
   in LXML format, else nil."

  `(and (description? ,lxml-description)
		(consp (first ,lxml-description))
		(equal (symbol-name (description-type (first ,lxml-description)))
				(symbol-name '|Spec2|))))

(defmacro SPEC2-LXML-description-TPNS (lxml-description)

  "Given LXML-DESCRIPTION, an LXML SPEC2 description of a set of TPNs,
  `return the corresponding list of LXML TPN descriptions."

   (format t "~A~%" lxml-description)

   (format t "~A~%" `(cdr ,lxml-description))


  `(cdr ,lxml-description))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Macros:  TPN LXML descriptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro TPN-LXML-description? (lxml-description)
  
  "Returns TRUE if LXML-DESCRIPTION is a TPN description
   in LXML format, else nil."

  `(and (description? ,lxml-description)
		(equal (symbol-name (description-type ,lxml-description)) (symbol-name '|TPN|))))

;;; Converting description values.

(defmacro convert-association-value (association)
  
  "Converts the value of ASSOCIATION from a string
   to an sexpression and returns that value."

  `(string-trim " " (association-value ,association)))

(defmacro transfer-association!
	(attribute association target-associations)
  
  "Adds ASSOCIATION, containing ATTRIBUTE, to TARGET-ASSOCIATIONS, 
   while converting the value of ASSOCIATION from a string
   to an sexpression."

  `(add-association! ,attribute 
					 (convert-association-value ,association)
					 ,target-associations))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Support Macros:  Choice Assignments and their External Description
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro choice-assignment?
	(expression)

  "Returns true iff EXPRESSION denotes a choice assignment."
  
  `(assignment? ,expression))

(defmacro make-choice-assignment
	(variable value)

  "Returns an ASSIGNMENT of choice VARIABLE to VALUE."
  
  `(make-assignment :variable-name ,variable 
					:value ,value))

(defmacro choice-assignment-variable
	(choice-assignment)

  "Returns the variable of a CHOICE variable/value ASSIGNMENT."
  
  `(assignment-variable ,choice-assignment))

(defmacro choice-assignment-value
	(choice-assignment)

  "Returns the value of a CHOICE variabe/value ASSIGNMENT."
  
  `(assignment-value ,choice-assignment))

;;; Check:
;;; o Is this needed, besides for testing?
(defun choice-assignment-descriptions (choice-assignments)
  
  "Given choice assignments, each of the form:
    (= <decision-event-object> <value>)

   Returns a list of external descriptions of the form:
    (= <decision-event-name> <value>)"
	
  (loop for c in choice-assignments
	  collecting (choice-assignment-description c)))

;;; Check:
;;; o Is this needed, besides for testing?
(defun choice-assignment-description (choice-assignment)
  
  "Given a choice assignment of the form:
    (= <decision-event-object> <value>)

   Returns an external description of the form:
    (= <decision-event-name> <value>)"
  
  (if (choice-assignment? choice-assignment)
	  
	  (make-assignment
	   :variable-name (event-name 
					   (choice-assignment-variable choice-assignment))
	   :value (choice-assignment-value choice-assignment))
	
	choice-assignment))

;;; Added.
;;; Check:
;;; o Is the internal form correct?
;;; o Is this used?
(defconstant *decision-event-type* 'event
  "Type of an instance of a TPN decision-event")

(defun internalize-choice-assignment-description 
	(choice-assignment-description TPN)
  
  "Converts choice assignment description:
     (= <decision-event-name> <value>)

   to internal form:
     (= <decision-event-object> <value>)."
  
  (let* ((decision-event-name 
		  (choice-assignment-variable choice-assignment-description))
		 
		 (value (choice-assignment-value choice-assignment-description))
		 
		 (decision-event (if (typep decision-event-name *decision-event-type*)
							 decision-event-name
						   (get-event decision-event-name TPN))))
	
	(make-choice-assignment decision-event value)))

(defun internalize-choice-assignment-descriptions
  (choice-assignment-descriptions TPN)
  
  "Converts a set of choice assignment descriptions
     each of the form:
        (= <decision-event-name> <value>)

     to internal form:
       (= <decision-event-object> <value>)."
	
  (loop for c in choice-assignment-descriptions
	  collecting (internalize-choice-assignment-description c tpn)))
