(in-package :tpn)

;;;Implements the code necessary to try to change an TPN into an STN and then dispatch it
;;;correctly, using temporal reasoning from the ITC class and a user defined interface.
;;; DISPATCH-TPN-WITH-EPISODES is the primary method.
       
(defun dispatch-tpn-with-episodes (filename callback callback-object total-duration)

  "Dispatches a TPN according to a provided callback and object. This algorithm handles 
  non-uncertain cases correctly, but will not extend correctly to STNUs.

Input:
FILENAME - a string with the file name of a plan, relative to the TPN directory 
CALLBACK - a method handle with argument list (object cmd time), where the cmd is the string of
the primitive in the TPN file and TIME is the time to execute the action.
TOTAL-DURATION - the longest amount of time it should run."

  ;;grab the stn
  (multiple-value-bind (stn tpn) (extract-stn filename)
	(format t "~s~%~s~%" stn tpn)
	(format t "~s~%" (tpn-episodes tpn))
	(setf (stn-start stn) (find-event "0" stn))
	

	;;make a dispatching instance
	(let ((stn-dispatch (make-instance 'stn-dispatch :stn stn :incomplete-episodes (tpn-episodes tpn))) ;create the dispatch instance
		  (start-time (get-universal-time))
		  (episodes-ongoing nil)
		  (completed-episodes nil)
		  (record nil))

	   (format t "AHHHH! ~s~%" (stn-start stn))
	  (format t "AHHHH! ~s~%" (itc::dispatch-a stn-dispatch))

	  ;;loop for as long as you're supposed to
	  (loop while (b< (- (get-universal-time) start-time) total-duration)  
		  do (setf completed-episodes nil)
		   do (format t "HERE~%")

			 ;;Handle completed episodes. Assume the complete when you assigned them.
		  do (loop for i in episodes-ongoing ;find ongoing epsiodes that are done
				 do (when (b>= (get-universal-time) (second i)) ;if the episode is done
					  (setf episodes-ongoing (remove i episodes-ongoing)) ;it's no longer ongoing
					  (push (first i) completed-episodes))) ;and it is complete

			 ;;Find the NEXT-TIME-WITH-EPISODES and destructure
		  do (multiple-value-bind (exec-event next-times current-time starting-episodes)
				 (next-time-with-episodes stn-dispatch (- (get-universal-time) start-time) completed-episodes)

			   (format t "~s ~s ~s ~%" stn-dispatch (- (get-universal-time) start-time) completed-episodes)
			   (format t "~s ~s ~s ~s ~%" exec-event next-times current-time starting-episodes)

			   (when (eql nil exec-event) (return)) ;when there isn't an event to execute, we're done, so quit

			   (push (list exec-event current-time) record) ;note execution time
			   (sleep-if-positive (bminus current-time (bminus (get-universal-time) start-time))) ;wait for the "current time"
			  
			   ;;Run the starting episodes
			   (loop for i in starting-episodes
							  ;;Make callback
				   do (funcall callback callback-object (episode-primitive (first i)) (bminus (second i)current-time))
					  ;;Mark them as ongoing
				   do (push (list (first i) (b+ start-time (second i))) episodes-ongoing))
			   (sleep-if-positive (bminus (first next-times) (bminus (get-universal-time) start-time))))) ;wait for the time to start again
	  record))) 

(defmethod extract-stn (filename)

  "Try to extract a STN from the input TPN. It throws a warning if there are decision
   variables in the TPN, but then proceedes because the planner might do this intentionally.
   A better check to make sure there really aren't choices in the plan would be helpful, 
   but if choices remain the plan will probably just be inconsistent. Ignores state variables. "

  (let ((tpn (car  (parse-tpn-file (merge-pathnames (pathname filename) 
													(pathname (concatenate 'string  cl-user::*mars-toolkit-pathname* "/")))
								   #'tpn::get-event #'tpn::make-episode #'tpn::add-episode #'tpn::make-tpn))))
	
	(when (not (null (tpn-decisions tpn)))
	  (warn "The TPN has decisions. Incorrect behavior might result if these were not planned out."))

										;Return the network
	(values (tpn-temporal-network tpn) tpn)))

(defun sleep-if-positive (time)

  "Calls sleep with the input value, but only if the value is positive. Negative values would cause
   and error. Negatives are ignored."

  (when (< 0 time)
	(sleep time)))


(defun print-for-dispatcher (object cmd time-cmd)

  "A print function to be used with the dispatcher. It just prints the names of objects
  to the screen when they are 'executed'. For debugging purposes"

  (declare (ignore object))
  (format t "~a ~a ~%" time-cmd cmd))


(defun test-tpn-dispatching ()

  "Try dispatching two plans. Again, since these actually run in real time we can't
   put them in a test."

  (dispatch-tpn-with-episodes "TPN/Examples/test-stn.tpn" #'print-for-dispatcher nil 20)
;  (dispatch-tpn-with-episodes "TPN/Examples/test.plan" #'print-for-dispatcher nil 200)
)

(defun test-extract ()

  "Try extracting. TODO: make a test?"

  (let ((stn (extract-stn "tpn/Examples/test.plan")))
 
	(print  (itc::offline-dispatch-stn stn))))


(defmethod next-time-with-episodes ((stn-dispatch stn-dispatch) &optional new-time complete-episodes)

  "Determines a new set of episodes of a STN problem to execute. Takes a STN-DISPATCH object, which holds
   the state information required for dispatching. Also takes NEW-TIME, the current time, and COMPLETE-EPISODES, 
   a list of TPN-EPISODES which have completed - which should be episodes previously returned as starting.

   Returns:
      EXEC-EVENT - the event to execute
      NEXT-TIMES - a list (l u), the possible times at which the dispatcher (this algorithm) can be correctly called again
      CURRENT-TIME - the time at which EXEC-EVENT is executed
      STARTING-EPISODES - a list of episodes to start, each episode is a list (ep time) where the time
                         is the scheduled duration and ep is a TPN-EPISODE. "
  
  (with-slots (stn A waiting-for-episode incomplete-episodes) stn-dispatch

	(format t "HERE0 ~s ~s ~%" new-time complete-episodes)
	(format t "HERE1 ~s ~s ~s ~s ~%" stn A waiting-for-episode incomplete-episodes)

	;;0. For any completed episodes, remove them from the incomplete list and mark any events waiting on
	;;them as active. Don't worry if we miss another dependency because 2. cleans that up.
	(loop for ep in complete-episodes
		do (setf incomplete-episodes ( remove ep incomplete-episodes)) ;delete from incomplete
		   (let ((end-event (convert-tpn-event-to-stn-event stn (episode-end ep)))) ;find end event
			 (loop for i in waiting-for-episode ;look for it in the waiting list
				 thereis (when (eql i end-event) ;if there
						   (setf waiting-for-episode (remove end-event waiting-for-episode)) ;remove from waiting
						   (push end-event A))))) ;make it active

	(format t "HERE2 ~s ~s ~s ~s ~%" stn A waiting-for-episode incomplete-episodes)

	;;1. Find the next time point
	(multiple-value-bind (exec-event next-times current-time)  (itc::next-stn-time-point stn-dispatch new-time)

	  (format t "HERE3 ~s ~s~%" stn-dispatch new-time)
	  (format t "HERE4 ~s ~s ~s ~%" exec-event next-times current-time )

	  ;;2. Take all events that are the end of incomplete episodes that are "active"
	  ;;and move them to the waiting-for-episode list
	  (loop for ep in incomplete-episodes
		  do (loop for i in A
				 thereis (when (eql (convert-tpn-event-to-stn-event stn (episode-end ep))  i) 
						   (progn (setf A (remove i A))
								  (push i waiting-for-episode)
;;;								(format t "not ready yet: ~a because of: ~a ~%" i ep)
								  t))))

	  (let ((starting-episodes nil) (episode-end-time nil))
		;;3. If the event scheduled starts an episode, pick a time to end the episode
		(loop for ep in incomplete-episodes 
			do (when (eql exec-event (convert-tpn-event-to-stn-event stn (episode-start ep)))
				 (progn
;;;					   (format t "starting ep: ~a ~%" ep)
				   (let ((end-event (convert-tpn-event-to-stn-event stn (episode-end ep))))
					 (setf episode-end-time (event-lb end-event)) ;just use the lower bound
					 ;; Warning: ITC::PROPAGATE-EXEC-TIME should be given exactly 2 arguments.
					 ;; It was given 3 arguments. Delete the episode-end-time
					 ;; (itc::propagate-exec-time stn-dispatch end-event episode-end-time)
					 (itc::propagate-exec-time stn-dispatch end-event)

					 (push (list ep episode-end-time) starting-episodes))
				   t)))					;and propagate it to help make sure other things are consistent

		(values exec-event next-times current-time starting-episodes)))))

(defmethod convert-tpn-event-to-stn-event ((stn stn) (tpn-event event-w-stc-reasoning))

  "Simple helper function. Takes a tpn event and finds the equivalent stn event from the
   input stn."
  
  (find-event (event-name tpn-event) stn))
