;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

;;; Copyright (c) 2008-10 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

;;; ***********************************************************************
;;;
;;;  Labeling Guards (Activation-Assignments) of a Temporal Plan Network
;;;
;;; ***********************************************************************

;;; This file contains code for labeling the episodes and events of a
;;; TPN with the guard assignments that activate those episodes and events.

;;; A TPN represents all executions of a concurrent timed program in which 
;;; the program involves a non-deterministic choice operator and interval 
;;; time constraints.  A TPN is comprised of a set of nested choice, parallel
;;; and sequential compositions.

;;; Events and episodes are activated when a sequence of choices have been
;;; made that results in a thread of execution from the start to that event or
;;; episode.  Each event and episode is guarded with an assignment corresponding
;;; to its innermost nested choice.


;;; A choice assignment is represented as an assignment expression whose
;;; variable is a decision-event and whose value is the number of a
;;  successor episode.

(in-package :TPN)

;;; To Do:
;;; o Consider renaming activation-assignment to guard.

;;; *********************
;;;
;;;     Interface Function: Label-TPN-Activation-Assignments
;;; 
;;;     Labels Episodes and Decision Events with Activation Assignments.
;;;
;;; *********************

;;; For each decision-event and episode, records the
;;; the innermost CHOICE ASSIGNMENT that activates it, called
;;; an activation-assignment.  The activation-assignment corresponds
;;; to the :GUARD condition of a conditional CSP/COP.
;;; A choice assignment is an object generated using MAKE-ASSIGNMENT.

;;;   Records activation-assignment (guard) on slot:
;;;
;;;      tpn-[episode | event]-activation-assignment.
;;;
;;;   Maintains a mapping from activation assignments to
;;;   episodes that are activated using slot:
;;;
;;;      tpn-activation-assignment-episodes.

;;; Pseudo code:

;;; Let Q be a queue of TPN events to be processed.
;;; Q = {start(TPN)}
;;; Repeat until Q is empty
;;;   Pop event E from Q
;;;   If all of E's predecessors are labeled
;;;   Then
;;;     If E is a decision event,
;;;     Then
;;;       Label E's successors with distinct choice assignments
;;;       for decision event.
;;;     Else,
;;;       Let label = if decision-end-event?(E)
;;;                   label-for-preceding-choice(E)
;;;                   label(E)
;;;       Label E's successors with label.

;;; Check:
;;; o Do we check to see if an event was already expanded,
;;;   and if so, ignore?  We should do this also when
;;;   queued, as well as dequeued.
(defmethod label-tpn-with-activation-assignments! 

	((tpn tpn-w-activation-labeling))
  
  "Given a TPN as input, this function
   labels each episode and decision-event of TPN with the 
   innermost choice assignment that activates it, called
   the episode's activation-assignment.  A choice assignment is an
   object generated using MAKE-ASSIGNMENT.

   Each activation-assignment is recorded on the following
   slot, respectively, for an episode or event:
      tpn-[episode | event]-activation-assignment

   Maintains a mapping from activation assignments to
   episodes they activate.  This is maintained at:
      tpn-activation-assignment-episodes.

   This function is called for effect only.

   Note that the complete set of choice assignments required to
   activate an object can be extracted by tracing
   backwards through activations to the start event."

  ;; Removes any pre-existing activation assignments before
  ;; constructing a new labeling.
  (remove-tpn-activation-assignments! tpn)
  
  ;; TPN is labeled with activation-assignments by propagating
  ;; forward from the start event.  Q is a list of events to be
  ;; propagated, and denote the fringe of this propagation.

  ;; Repeatedly propagates an event on the fringe of propagation.
  (loop 
	  with Q = (list (tpn-start tpn))
			   
	  until (null Q)
			
	  for e = (pop Q)
		  
	  do (debug-print-for :tpn t "")
		 (debug-print-for :tpn t "Processing event ~a." e)

		 (cond ((null (event-activation-assignment e))
				
				;; Event e has not already been labeled,
				;; making it a candidate to be labeled.
				
				;; All predecessors of event E must be labeled before 
				;; labeling event E.  This is required in order to 
				;; determine when a set of choices reconverge.
		   
				(cond ((all-event-predecessors-labeled? e)
		   
					   ;; All predecessors of E are labeled,
					   ;; hence we can now label E and propagate.
					   
					   (debug-print-for 
						:tpn t "  All predecessor episodes of ~a are labeled," e)

					   ;; Queue for labeling, each of e's successor events that hasn't been labeled.
					   (debug-print-for 
						:tpn t "  ... propagate")
					   (setq Q (queue-event-successor-events-for-labeling e Q))

					   ;; Label event E and its immediate successor episodes
					   ;; with an activation-assignment.
					   
					   (debug-print-for :tpn t "  ... and label.")
					   (label-event-and-immediate-successor-episodes! e tpn))
					  
					  (:otherwise
					   (debug-print-for 
						:tpn t "  Some predecessor of ~a is unlabeled, skip it." 
						e))))
			   
			   (:otherwise
				(debug-print-for 
				 :tpn t "  Event ~a is already labeled, ignore it." e))))
  
  ;; Now that all events and non-temporal constraint episodes have
  ;; been labelled, we need to go through and label each pure temporal
  ;; constraint episode.
  
  (loop
	 for tc in (remove-if-not #'episode-istemporalconstraint (tpn-episodes tpn))
	 do (label-episode-based-on-events tc tpn))
  (tpn-unlabeled! tpn)
  
  T)

(defmethod label-episode-based-on-events ((episode episode-w-activation-labeling) (tpn tpn-w-activation-labeling))
  "Label EPISODE with the conjunction of the activation labels of its
  neighboring events."
  (let ((start-activation (event-activation-assignment (episode-start episode)))
		 (end-activation (event-activation-assignment (episode-end episode))))
	(cond ((not (and (eql start-activation *always-active*) (eql start-activation *always-active*)))
		   (label-episode-with-activation-assignment! episode (list start-activation end-activation) tpn :conjunction? t))
		  ((and (not (eql start-activation *always-active*))
				(eql end-activation *always-active*))
		   (label-episode-with-activation-assignment! episode start-activation tpn :conjunction? nil))
		  ((and (not (eql end-activation *always-active*))
				(eql start-activation *always-active*))
		   (label-episode-with-activation-assignment! episode end-activation tpn :conjunction? nil))
		  (:otherwise
		   (label-episode-with-activation-assignment! episode *always-active* tpn :conjunction? nil)))))
;;; ***************************************************************
;;;
;;;   Support: Labeling an Event and its (Immediate) Successors
;;;            with Activation Assignments
;;;
;;; ***************************************************************

;;; *********************
;;;
;;;   Top Level Labeling of Events and their Immediate Successors
;;;
;;; *********************

(defmethod label-event-and-immediate-successor-episodes! 
	
	((e event-w-activation-labeling) (tpn tpn-w-activation-labeling))

  "Label event E of TPN and its immediate successor episodes
   with an activation-assignment."

  (let (assignment label activation-list)
	(cond
	 ((decision-event? e)

	  ;; Event E is a decision-event (possibly a start).
	  (label-decision-event-and-immediate-successor-episodes! e tpn))
			   
	 ((not (event-has-predecessors? e))
				
	  ;; E denotes a (non-decision) START EVENT
	  ;; when event has no predecessors.

	  (label-start-event-and-immediate-successor-episodes! e tpn))

	 (:otherwise
				
	  (debug-print-for 
	   :tpn t "  Processing ~a as a non-decision, non-start event." 
	   e)

	  ;; E is a non-decision event.
	  ;; We only label decision-events, hence E will not be labeled here.
				
	  ;; Propagate activation-assignments from predecessors to
	  ;; successors.
				
	  (cond 
	   ;; The label used for successors depends on what the
	   ;; event e represents:
	   
	   ((event-predecessors-have-same-activation-assignment? e)

		;;  E denotes the join of 1 or more PARALLEL THREADS,
		;;  since the predecessor episodes of E are all labeled with 
		;;   the same activation assignment,

		(label-parallel-join-event-and-immediate-successor-episodes! e tpn))
	   
	   ((event-predecessor-labels-exhaust-choices? e)
		   
		;; E denotes the JOIN of the choices of a PRECEDING
		;; DECISION-EVENT, since the activation-assignments of 
		;; E's predecessor episodes collectively exhaust the choices 
		;; for some decision event DE,
		
		(label-decision-join-event-and-immediate-successor-episodes! e tpn))	   
	   
	   ((first (setf activation-list (multiple-value-list (threads-of-multiple-decision-events-collapse-to-one?  e))))   
		
		(setf assignment (first activation-list))
		(setf label (second activation-list))

		(label-complex-join-event-and-immediate-successor-episodes! e assignment tpn)
		(label-complex-join-event-and-immediate-successor-episodes! e label tpn :label-p t))
	   
	   (:otherwise
		
		;; An unexpected case.		
		(debug-print-event-predecessor-activation-pairs! e)
				  
		(error "The activation assignments of the predecessors of event ~
                   ~A are not identical and are not the complete set of ~
                   alternative choices for one or more decision events." 
			   e)))))))


;;; *********************
;;;
;;;   Fanning Out During Labeling of Events and their Immediate Successors
;;;
;;; *********************

(defmethod label-decision-event-and-immediate-successor-episodes! 
	
	((e event-w-activation-labeling) (tpn tpn-w-activation-labeling))

  "Event E is a decision-event (possibly a start).

   Label decision event E of TPN and its immediate successor 
   episodes with an activation-assignment.  The successor
   episodes of E denote choices of E, and are labeled by
   respective assignments to E."

  (debug-print-for :tpn t "  Processing ~a as a decision event." e)

  ;; The predecessor episodes of a decision event 
  ;; are expected to all have the same activation assignment.
							  
  ;; This is because the threads of choices of a preceding decision-event 
  ;; can not end (reconverge) at a successor decision-event, 
  ;; but must converge BEFORE that decision-event.
							  
  (unless (event-predecessors-have-same-activation-assignment? e)

	(error "The activation-assignments of the predecessor episodes of ~
            decision-event ~a should be identical, but aren't."
		   e))

  ;; Label decision event E with the label given to its predecessors
  ;; or label it *always-active*, if event has no predecessor.

  (let (assignment label)
	
	(cond 
	  ((event-has-predecessors? e)
	   (setf assignment (predecessor-activation-assignment e))
	   (setf label (predecessor-activation-label e)))
	  (:otherwise
	   (setf assignment *always-active*)
	   (setf label *always-active*)))
	
	(label-event-with-activation-assignment! e assignment)    ;Propagates assignment	
	(label-event-with-activation-assignment! e label :label-p t) ;Propagates label											
		
	;; The choices of decision-event E are E's immediate successor episodes.
	;; Label each successor episode by assignment (= E N), where N is the 
	;; position in which the episode appears in E's successor list.
	
	(debug-print-for 
	 :tpn t "    Sequencing through choices of decision event ~a." e)
	
	(loop for ep in (remove-if #'episode-istemporalconstraint (event-successor-episodes e))
	   for n from 1 ;Fix: changes this line to take into account the decision's domain
	   for activation-assignment = (make-assignment :variable-name e :value n)
		 
	   do 
		 (debug-print-for :tpn t "      Labeling ~a with choice ~a." 
						  ep activation-assignment)
	   
		 (label-episode-with-activation-assignment! ep activation-assignment tpn)		 
		 (label-episode-with-activation-assignment! ep (add-to-label! activation-assignment label) tpn :label-p t))))


(defmethod label-start-event-and-immediate-successor-episodes! 
	
	((e event-w-activation-labeling) (tpn tpn-w-activation-labeling))

  "E denotes a (non-decision) START EVENT, since E has no predecessors.
   Label E and its successor episodes with *always-active*."
  
  (debug-print-for :tpn t "  Processing ~a as a start event." e)

  ;; Label start E, indicating that it is *always-active*.
  (label-event-with-activation-assignment!
   e *always-active*)

  (label-event-with-activation-assignment!
   e *always-active* :label-p t)
				
  ;; Pass label to e's successor episodes.
  (label-successors-with-activation-assignment!
   e  *always-active* tpn)

  (label-successors-with-activation-assignment!
   e  *always-active* tpn :label-p t))

;;; *********************
;;;
;;;   Fanning Out During Labeling of Events and their Immediate Successors
;;;
;;; *********************

(defmethod label-parallel-join-event-and-immediate-successor-episodes! 
	
	((e event-w-activation-labeling) (tpn tpn-w-activation-labeling))

  "E denotes the join of 1 or more PARALLEL THREADS,
   since the predecessor episodes of E are all labeled with 
   the same activation assignment.
   Label the activation assignment of E's successor episodes with this
   predecessor assignment."
				  
  (debug-print-for :tpn t "    ~a's predecessors have the same label." e)

  ;; THEN label the activation assignment of E's 
  ;;      successor episodes with this predecessor assignment.
				  
  (let ((activation-assignment
		 (predecessor-activation-assignment e))
		(activation-label
		 (predecessor-activation-label e)))
					
	(when (null activation-assignment)
	  (error "Predecessor of ~a has no activation ~
              assignment.  This shouldn't happen." e))
								  
	;; Label E and its successor episodes with 
	;; E's predecessor activation-assignment.
	(label-event-with-activation-assignment!
	 e activation-assignment)

	(label-event-with-activation-assignment!
	 e activation-label :label-p t)
					  
	(label-successors-with-activation-assignment!
	 e activation-assignment tpn)
	
	(label-successors-with-activation-assignment!
	 e activation-label tpn :label-p t)
	))


(defmethod label-decision-join-event-and-immediate-successor-episodes! 
	
	((e event-w-activation-labeling) (tpn tpn-w-activation-labeling))

  "E denotes the JOIN of the choices of a PRECEDING DECISION-EVENT,
   (the activation-assignments of e's predecessor episodes
   collectively exhaust the choices for some decision event DE).

   Label decision event E of TPN and its immediate successor 
   episodes with the activation-assignment of DE."

  ;; E denotes the JOIN of the choices of a preceding
  ;; decision-event DE, hence E denotes the end event of DE.

  (debug-print-for 
   :tpn t "  ~a's predecessors denote reconverging choices." e)

  ;; Label the activation-assignment of e's successor episodes
  ;; with DE's activation-assignment.
								
  (let ((assignment
		 (predecessor-decision-event-activation-assignment e))
		(label
		 (predecessor-decision-event-activation-label e)))
								  
	(label-event-with-activation-assignment! e assignment)
	(label-event-with-activation-assignment! e label :label-p t)
								
	(label-successors-with-activation-assignment! e assignment tpn)
	(label-successors-with-activation-assignment! e label tpn :label-p t)))


(defmethod label-complex-join-event-and-immediate-successor-episodes! 

	((e event-w-activation-labeling) label (tpn tpn-w-activation-labeling) &key (label-p nil))

  "E denotes the JOIN of the choices of several decision-events.
   (the activation-assignments of e's predecessor episodes
   collectively exhaust the choices of one or more decision events),
   and collapse to a thread with LABEL as its activation assignment.

   Label decision event E of TPN and its immediate successor 
   episodes with LABEL."

  ;; E denotes the JOIN of the choices of one or more
  ;; decision-events, hence E denotes the end event of each of
  ;; those decision-events.

  (debug-print-for 
   :tpn t "  ~a's predecessors denote multiple reconverging choices." 
   e)

  ;; Label the activation-assignment of e's successor episodes
  ;; with activation-assignment LABEL.  
  (label-event-with-activation-assignment! e label :label-p label-p)  
  (label-successors-with-activation-assignment! e label tpn :label-p label-p))


;;; ***************************************************************
;;;
;;;   Support: Querying Properties of TPN Activation Assignments
;;;
;;; ***************************************************************

;;; Tests on the activation assignments of episodes and events.
;;; Here we use LABEL as a short form for activation assignment.

(defmethod all-event-predecessors-labeled? 
	
	((event event-w-activation-labeling))
  
  "Returns true iff every predecessor of EVENT has a non-empty
   activation-assignment or is a 'temporal constraint', i.e. not part
   of the path of execution.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT."
  
  (loop for e in (event-predecessor-episodes event)
	  unless (or (episode-istemporalconstraint e) (episode-labeled? e))
	  do (debug-print-for :tpn t "    Predecessor ~a unlabeled." e)
	  and return nil
	  finally (return t))

;;; Alternatively:  
;;;  (every #'episode-labeled?
;;;		 (event-predecessor-episodes event))

  )

(defmethod event-predecessors-have-same-activation-assignment? 

	((event event-w-activation-labeling))
  
  "Returns true iff every predecessor of EVENT is labeled with the
   same activation assignment or is a pure temporal constraint.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT."

  (if (event-has-multiple-predecessors? event)

	  (let* ((episodes (remove-if #'episode-istemporalconstraint (event-predecessor-episodes event)))
			 (activation-assignment
			  (episode-activation-assignment (first episodes))))
  
		(every #'(lambda (ep)
				   (equal (episode-activation-assignment ep)
						  activation-assignment))
			   (rest episodes)))

	;; Event has zero or one predecessor.
	T))

(defmethod event-predecessor-labels-exhaust-choices? 

	((event event-w-activation-labeling))
  
  "Returns true iff the activation assignments of the predecessors 
   of EVENT exhaust all choices for the same decision-event.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT."

  (and (same-decision-event-in-predecessor-labels? event)
	   (predecessor-labels-exhaust-choice-numbers? event)))


(defmethod same-decision-event-in-predecessor-labels? 

	((event event-w-activation-labeling))
  
  "Returns true iff the activation-assignment of every predecessor 
   of EVENT is an assignment to a decision event and every such 
   assignment has the same decision-event listed as its variable.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT."

  (let* ((episodes (remove-if #'episode-istemporalconstraint (event-predecessor-episodes event)))
		 (activation-assignment 
		  (episode-activation-assignment
		   (first episodes))))
	(when (assignment? activation-assignment)
	  (let ((decision-event
			 (assignment-variable activation-assignment)))
	
		(every #'(lambda (ep)
			   
				   (let ((activation-assignment
						  (episode-activation-assignment ep)))
				 
					 (and (assignment? activation-assignment)
						  (equal (assignment-variable 
								  activation-assignment)
								 decision-event))))
		   
			   (rest episodes))))))

(defmethod predecessor-labels-exhaust-choice-numbers? 

	((event event-w-activation-labeling))
  
  "Returns true iff the activation assignments of
   the predecessors of EVENT contain a
   contiguous list of choice numbers, starting at 1.

   Assumes that the activation-assignments of the predecessors 
   of EVENT are all labeled with ASSIGNMENTS to the
   SAME decision-event.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT.  The value of an activation 
   assignment is a positive integer."

  (let* ((episodes (remove-if #'episode-istemporalconstraint (event-predecessor-episodes event)))
		 (decision-event
		  (episode-activation-decision-event (first episodes))))
	
	(loop for ep in episodes
		collecting (episode-activation-choice ep) into choices
		finally (return (all-decision-event-choices-present? 
						 decision-event choices)))))
	   
(defmethod all-decision-event-choices-present? 
	
	((decision-event event-w-activation-labeling) choices)
  
  "Given DECISION-EVENT, representing a set of options, and
   a set of choices, each a number representing an option of
   DECISION-EVENT that has been selected, 
   Returns T iff CHOICES exhaustively selects every option
   of decision-event."

  (and (same-length? (event-successor-episodes 
					  decision-event)
					 choices)

	   ;; Consists of n elements 1 ...n.
	   (let ((sorted-choices (sort choices #'<)))
;;;	     (debug-print-for :tpn t "  *** sorted choices = ~a" sorted-choices)
		 (loop for n from 1 by 1
			 for c in sorted-choices
			 always (eql n c)))))

	
;;; -Create Q containing all decision-events in sorting.
;;; -loop until queue is empty.
;;;  - Dequeue decision-event and its choices.
;;;  - Test if all choices associated with decision-event is in sorting.
;;;    - Success
;;;      - delete the bin of decision-event from sorting.
;;;      - get activation-assignment of decision-event
;;;        and insert it into sorting.
;;;      - add decision-event to queue.

;;; Finally,
;;;  - If *always-active* was not found, one decision-event remains, 
;;;    and it has one choice,
;;;      then return event/choice as the activation-assignment for event.
;;;  - If no decision events remain, but *always-active* was found,
;;;    then return *always-active* as the activation-assignment for event.
;;;  - Else return failure.
(defmethod threads-of-multiple-decision-events-collapse-to-one? 
	
	((event event-w-activation-labeling))
  
  "Returns non-nil if the threads, represented by EVENT'S successors
   collapse to a single thread.  If collapsed, returns the activation
   assignment that the threads collapse to.

   A set of threads (predecessors) collapse to the activation assignment
   of a decision-event if the threads corresponding to all choices of
   decision-event are present."

  (let (activation)
	(multiple-value-bind (sorting always-active-labeled? total-label)
		(sort-predecessor-activations event)
	  (loop 
		 ;; SORTING contains all activation-assignments of threads 
		 ;; (predecessors) going into event, sorted by decision event.
		   
		 ;; Q contains all decision-events in sorting that need to be
		 ;; examined to see if all its choices reconverge at event.
		   
		 ;; ALWAYS-ACTIVE-LABELED? is True iff if its an
		 ;; activation-assignment of one of the threads going into event.
		 with Q = (mapcar #'first sorting)
		   
		 until (null Q)
		 ;; Stop when all bins in sorting have been examined,
		 ;; since they were lasts modified.

		 ;; A decision-event and all its choices that converge on event.
		 for decision-event = (pop Q)
		 for choices = (values-of-bin decision-event sorting)
										 
		 when (all-decision-event-choices-present? decision-event choices)
		 ;; All possible choices of decision-event converge at event.
		   
		 do 
		 ;; The threads collapse, so that their activation assignment 
		 ;; is that of their decision event.  Remove decision event from 
		 ;; sorting and add its activation assignment.
		   
		 ;; Note: if the activation assignment is *always-active*
		 ;;   its not a regular assignment, and we keep track of it
		 ;;   on the side using always-active-labeled?.
		   (setq sorting (delete-from-sorting decision-event sorting)
				 activation (event-activation-assignment decision-event))
		   
		   (remove-decision-from-label! decision-event total-label)
		   		   
		   (cond ((eql activation *always-active*)
				  (setq always-active-labeled? T))
				 
				 (:otherwise
				  (setq sorting (add-activation-to-sorting activation sorting))
				  (push (assignment-variable activation) Q)))
		   
		 finally (cond ((and always-active-labeled?
							 (null sorting))
						
						;; Predecessors collapse to the activation
						;; assignment *always-active* (same as start).
						(return (values *always-active* (list *always-active*))))
					   
					   ((eql (length sorting) 1)
						(destructuring-bind (decision-event . choices) 
							(first sorting)
						  (cond ((eql (length choices) 1)
								 
								 ;; Predecessors collapse to the activation
								 ;; assignment decision-event/choice,
								 ;; return it.
								 
								 (let ((choice (first choices)))
								   (return (values (make-assignment :variable-name
																	decision-event
																	:value choice)
												   total-label))))
								(:otherwise
								 
								 ;; Event is preceded by multiple 
								 ;; activation-assignments that haven't 
								 ;; collapsed.  This shouldn't occur.
								 (return nil)))))
					   
					   (:otherwise
						;; Event is preceded by multiple activation-assignments
						;; that haven't collapsed.  This shouldn't occur.
						(return nil)))))))


;;; ***************************************************************
;;;
;;;   Support Utilities: Manipulating Acivation Assignments
;;;
;;; ***************************************************************

;;; Utilities for queueing propagation, and 
;;; for manipulating activation assignments (guard conditions), 
;;; including
;;;    o accessing, 
;;;    o adding, 
;;;    o removing and 
;;; activation assignments.

;;; *********************
;;;
;;;   Retrieving Activation Assignments
;;;
;;; *********************

;;; Basic utilities for accessing activation assignments of
;;; preceding episodes and decision events.

(defmethod predecessor-activation-assignment 
	
	((event event-w-activation-labeling))
  
  "Returns the activation assignment shared among
   EVENT's predecessors.

   It is assumed that all predecessor episodes
   of EVENT have been given the same activation assignment.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT."

  (episode-activation-assignment
   (first 
	(remove-if #'episode-istemporalconstraint (event-predecessor-episodes event)))))

(defmethod predecessor-activation-label
	
	((event event-w-activation-labeling))
  
  "Returns the activation label shared among
   EVENT's predecessors.

   It is assumed that all predecessor episodes
   of EVENT have been given the same activation label.

   An activation label is a conjunction of objects 
   generated using MAKE-ASSIGNMENT."

  (episode-activation-label
   (first 
	(remove-if #'episode-istemporalconstraint (event-predecessor-episodes event)))))


(defmethod predecessor-decision-event-activation-assignment 
	
	((event event-w-activation-labeling))

  "Return the activation-assignment of the decision-event DE
   that activates event EVENT.

   Note that EVENT might be the end event of DECISION-EVENT."
  
  ;; The decision event that activates EVENT or that
  ;; EVENT ends.
  (let ((decision-event
		 (assignment-variable 
		  (predecessor-activation-assignment event))))
	
	;; The assignment that activates DECISION-EVENT.
	(event-activation-assignment decision-event)

;;; FIX:
;;; o Delete when works.
;;;   This looks at the assignment of event's predecessor,
;;;   but this doesn't work for the start event.
;;;   However, it appears that all events are labeled with
;;;   activation assignments, hence going to the predecessor
;;;;  is unnecessary.	
;;;	(predecessor-activation-assignment decision-event)
	))

(defmethod predecessor-decision-event-activation-label
	
	((event event-w-activation-labeling))

  "Return the activation-label of the decision-event DE
   that activates event EVENT.

   Note that EVENT might be the end event of DECISION-EVENT."
  
  ;; The decision event that activates EVENT or that
  ;; EVENT ends.
  (let ((decision-event
		 (assignment-variable 
		  (predecessor-activation-assignment event))))
	
	;; The assignment that activates DECISION-EVENT.
	(event-activation-label decision-event)))


(defun add-to-label! (assignment label)
  "Adds a new assignment to a pre-existing label."
  (if (equalp label *always-active*)
	  (setf label (list *always-active*)))

  (pushnew assignment label :test #'equalp))

(defun remove-from-label! (assignment label)
  "Removes an specific assignment from a label."
  (if (equalp label *always-active*)
	  (setf label (list *always-active*)))
  (remove-if #'(lambda (x) (equalp x assignment)) label))

(defun remove-decision-from-label! (decision-event label)
  "Removes all assignments to an specific decision event
   from a label."
  (remove-if 
   #'(lambda (assig) 
	   (equalp decision-event (assignment-variable assig))) 
   label))

(defun combine-labels! (label-src label-dst)
  "Adds the assignments from label-src to label-dst without
   repetition."
  (loop for assignment in label-src do
	   (add-to-label! assignment label-dst)))

(defun label-contains-assignment? (assignment label)
  "Boolean function that tests whether a label contains a particular
   assignment."
  (if (equalp assignment *always-active*)
	  (setf assignment (list *always-active*)))

  (member assignment label))


;;; *********************
;;;
;;;   Recording Activation Assignments
;;;
;;; *********************

;;; Basic utilities for recording activation assignments of
;;; episodes and decision events.

(defmethod label-event-with-activation-assignment!

	((event event-w-activation-labeling) activation-assignment &key (label-p nil))

  "Label EVENT with ACTIVATION-ASSIGNMENT.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT."
  
  (debug-print-for :tpn t "    Labeling event ~a with ~a." 
				   event activation-assignment)
  
  (if label-p
	(setf (event-activation-label event) activation-assignment)
	(setf (event-activation-assignment event) activation-assignment)))

(defmethod label-successors-with-activation-assignment! 

	((event event-w-activation-labeling) 
	 activation-assignment 
	 (tpn tpn-w-activation-labeling) &key (label-p nil))

  "Label all successor episodes of EVENT with ACTIVATION-ASSIGNMENT.

   An activation assignment is an object generated using MAKE-ASSIGNMENT."

  (debug-print-for :tpn t "      Labeling successor episodes of ~a with ~a:" 
				   event activation-assignment)
  
  (loop 
	 for successor in (remove-if #'episode-istemporalconstraint (event-successor-episodes event))
					   
	  do 
		(debug-print-for :tpn t "        ~a <- ~a." 
						 successor activation-assignment) 
	
	   (label-episode-with-activation-assignment!
			successor activation-assignment tpn :label-p label-p)))


(defmethod label-episode-with-activation-assignment!
	((episode episode-w-activation-labeling) activation-assignment
	 (tpn tpn-w-activation-labeling) &key (label-p nil) (conjunction? nil))

  "Label EPISODE of TPN with ACTIVATION-ASSIGNMENT.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT."

;;;  (debug-print-for :tpn t "      Labeling ~a with ~a." 
;;;				   episode activation-assignment)
  
  (cond
	(label-p ;label
	 (setf (episode-activation-label episode) 
		   activation-assignment))

	(:otherwise ;innermost activation assignment
	 ;; Associate episode with its activation-assignment.
	 (if conjunction?
	   (loop
		  for assignment in activation-assignment
		  do
			(push episode
				  (tpn-activation-assignment-episodes 
				   assignment TPN)))
	   (push episode
			 (tpn-activation-assignment-episodes
			  activation-assignment TPN)))
	 
	 ;; Label episode with activation-assignment.
	 (setf (episode-activation-assignment episode) 
		   activation-assignment)
	 (setf (episode-activation-assignment-is-conjunction? episode)
		   conjunction?))))

;;; *********************
;;;
;;;   Removing Activation Assignments
;;;
;;; *********************

;;; Basic utilities for removing activation assignments of
;;; episodes and decision events.

;;; Move forward from start
;;;   If episode/event is labeled
;;;   Then
;;;     Remove label
;;;     Continue to all successors.

(defmethod remove-tpn-activation-assignments! 

	((tpn tpn-w-activation-labeling))
  
  "Removes (sets to nil) all activation-assignments of TPN's episodes. 
   These were recorded by label-tpn-with-activation-assignments!.

   An activation assignment is an object generated 
   using MAKE-ASSIGNMENT."

  (debug-print-for :tpn t "Removing labeling for ~a." 
				   tpn)
  (loop 
	  with Q = (list (tpn-start tpn))
	  with next-e
			   
	  until (null Q)
			
	  for e = (pop Q)

	  do (debug-print-for :tpn t "  Checking event ~a." e)
		 (cond ((not (null (event-activation-assignment e)))

				(debug-print-for 
				 :tpn t "    Unlabeling event ~a and its successor episodes." 
				 e)
				
				;; Clear the activation assignment and label of each event.
				(setf (event-activation-assignment e) nil)
				(setf (event-activation-label e) nil)
		 
				;; Clear the activation assignment and label of every
				;; successor episode of e.
				(dolist (ep (event-successor-episodes e))
				  
				  (setf (episode-activation-assignment ep) nil)
				  (setf (episode-activation-label ep) nil)
			 
				  ;; Queue e's successor events to be cleared.
				  (setq next-e (episode-end ep))
				  
				  (unless (null (event-activation-assignment next-e))
					(push next-e Q))))
			   (:otherwise
				(debug-print-for 
				 :tpn t "    Event ~a is not labeled." 
				 e)
				))
		 
		 ;; and clear the mapping from activation assignments 
		 ;; to episodes.
	  finally (clear-tpn-activation-assignment-episodes! tpn))
  
  (tpn-unlabeled! tpn)
  T)

;;; *********************
;;;
;;;   Debug printing activation assignments
;;;
;;; *********************

(defmethod debug-print-event-predecessor-activation-pairs! 

	((e event-w-activation-labeling))
			 
  "Print predecessors of event E and their corresponding 
   activation assignments if debug-print is enabled."
		 
  (debug-print-for 
   :tpn t "  Event ~a has predecessor : activation-assignment pairs:" e)
		   
  (loop for ep in (event-predecessor-episodes e)
	  for a = (episode-activation-assignment ep)
	  do (debug-print-for :tpn t "    ~a : ~a." ep a)))

;;; ***************************************************************
;;;
;;;   Support Utilities: Queueing Events for Propagation
;;;
;;; ***************************************************************

;;; Labeling is performed by a propagation process that
;;; begins at the start event and moves (propagates) for
;;; to successor episodes, hence following the threads of execution.
;;; The propagation process does not follow pure temporal constraints
;;; (i.e. constraints that are not part of the execution path).
;;; The queue maintains the current fringe of this propagation.

(defmethod queue-event-successor-events-for-labeling 
	((e event-w-activation-labeling) Q)
						 
  "Let E be an event that is being labeled with an activation
   assignment, and let Q b a list of events to be labeled.
   Adds to Q all successor events of E that need to be labeled."
						 
  (loop for ep in (event-successor-episodes e)
	  for next-e = (episode-end ep)
	  do (cond ((episode-istemporalconstraint ep)
				(debug-print-for
				 :tpn t "    Found non-execution thread episode leading to Event ~a. Skipping."
				 next-e))
			   ((null (event-activation-assignment next-e))
				(push next-e Q)
				(debug-print-for 
				 :tpn t "    Queuing successor Event ~a for labeling." 
				 next-e))
			   
			   (:otherwise
				
				(debug-print-for 
				 :tpn t "    Successor Event ~a is already labeled, don't queue."
				 next-e))))
  Q)

;;; ***************************************************************
;;;
;;;   Support Utilities: Sorting Decision Events
;;;
;;; ***************************************************************

;;; *********************
;;;
;;;   General Utility:  Sorting of Values into Bins.
;;;
;;; *********************

(defconstant *empty-sorting* nil
  "Empty list representing a sorting of values into bins.")

(defun values-of-bin 
	
	(bin sorting &key (test #'eql))

  "Given SORTING, representing a sorting of values (partitioning) into bins, 
   returns the VALUES associated with BIN of SORTING.
   BIN is an arbirary lisp expression, and TEST compares BIN 
   with the bins of SORTING.

   SORTING is an association list, where each association is:
         (bin . values)
   where VALUES are all values added to BIN of SORTING."

  (rest (assoc bin sorting :test test)))


(defun add-to-sorting 
	
	(bin value sorting &key (test #'eql))

  "Given SORTING, representing a sorting of values into bins, 
   adds VALUE to the BIN of SORTING.  VALUE and BIN are
   arbirary lisp expressions, and TEST compares BIN with 
   the bins of SORTING.

   SORTING is an association list, where each association is:

         (bin . values)

   where VALUES are all values added to BIN of SORTING."

  (let ((sorting-bin (assoc bin sorting :test test)))
	
	(cond ((null sorting-bin)
		   (push `(,bin ,value) sorting))

		  (:otherwise
		   (push value (rest sorting-bin))))
		  
	sorting))


(defun delete-from-sorting 
	
	(bin sorting &key (test #'eql))
  
  "Given SORTING, representing a sorting of values into bins, 
   removes all values associated with BIN from SORTING.  BIN 
   can be an arbirary lisp expression, and TEST compares BIN 
   with the bins of SORTING.

   SORTING is an association list, where each association is:
         (bin . values)
   where VALUES are all values added to BIN of SORTING."

  (delete bin sorting :key #'first :count 1 :test test))


;;; *********************
;;;
;;;   Sorting Activation Assignments by Decision Event
;;;
;;; *********************

(defmethod sort-predecessor-activations 
	
	((event event-w-activation-labeling))
  
  "Given EVENT,
   Return (SORTING ALWAYS-ACTIVE-PRESENT?) 
   Where ALWAYS-ACTIVE-PRESENT? is true iff one of the predecessors of EVENT
     is labeled with *always-active*, and
   SORTING represent the sort the activation assignments (= e choice)
   of the predecessors of event into an association list,
     where each association is (d . d-choices), where
     d is a decision event appearing in one of the activation assignments,
     and d-choices is the list of all choices for d that appear in
     one of the activation assignments."
  
  (loop with sorting = *empty-sorting*
	 with total-label = nil
	 with always-active-present? = nil
	 for episode in (remove-if #'episode-istemporalconstraint (event-predecessor-episodes event))
	 for activation = (episode-activation-assignment episode)
	 for label = (episode-activation-label episode)  
	 do 
	   (if (eql activation *always-active*)
		   (setq always-active-present? T) ;If true
		   (setq sorting ;Else
				 (add-activation-to-sorting activation sorting)))
	   (combine-labels! label total-label)
	 finally (return (values sorting always-active-present? total-label))))


(defun add-activation-to-sorting 
	
	(activation-assignment sorting)

  "Given an activation-assignment SORTING, representing a
     sorting of activation assignments into bins by decision event,
     adds ACTIVATION-ASSIGNMENT to the appropriate bin within
     SORTING.

     ACTIVATION-ASSIGNMENT is an assignment, of the form:


   (= decision-event choice)

     SORTING is an association list (denoting bins and their values).
     Each association is:

           (decision-event choices)

     where choices is a list of choices to decision-event that appear
     in activation-assignments added to SORTING."

  (let ((decision-event (assignment-variable activation-assignment))
		(choice (assignment-value activation-assignment)))
	
	(add-to-sorting decision-event choice sorting)))
