;;; -*- Mode:Common-Lisp; Package:TPN; Base:10 -*-

;;; Copyright (c) 2008 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :TPN)

;;; *********************
;;;
;;;   Constructing Descriptions of TPN Choice Selection Problems
;;;   Frames as Dynamic OCSPs
;;;
;;; *********************

;;; Input: A TPN with
;;;   - A list of decision events.
;;;
;;;   - Each decision event labeled with an activation assignment
;;;     (event-activation-assignment).
;;;     
;;;   - Each episode labeled with an activation assignment
;;;     (episode-activation-assignment).
;;;     
;;;   - A mapping from activation assignments to corresponding episodes
;;;     (tpn-activation-assignment-episodes).

;;;   - May want to collect decision events in the order in which
;;;     they are visited by the labeling function; this is the 
;;;     order in which a dynamic CSP routine should search these
;;;     decision.

;;; Output: An (Optimal) Dynamic Temporal Problem is a triple
;;;  ODTP = <D,G,CSP>, where:

;;; o A decision (design) variable d in D is created for each decision 
;;;   event e, where:
;;;   - The domain of d is a list of integers from 1 to n, where
;;;     n = length of successors of e.
;;;   - d is a conditional variable, activated by the 
;;;     activation-assignment g of e.
;;;   - Attribute cost g(d=v) is the combined cost of the episodes activated 
;;;     by d=v; Sum_{e in activated-episodes(d)} Cost(e).

;;; o CSP is a triple <S,E,C>

;;; o S, the discrete state variables of CSP, are exactly the decision
;;;   variables D of ODTP.
;;; o E, the real-value state variables of CSP, are the events of TPN.
;;; o C, the constraints of CSP are conditional, simple temporal constraints
;;;   of the form:
;;;
;;;                   s_i = v_ik -> e_l-e_m in [l,u]
;;;

;;; o Heuristic cost is defined as follows:
;;;   - f(D=V) = Sum_{d=v in D=V} f(d=v)
;;;   - Assumes that assignment costs f,g,h can be decomposed
;;;     per attribute, and then combined.
;;;   - Heuristic attribute cost of assignment d=v:
;;;     - f(d=v) = g(d=v) + h(d=v)
;;;     - h(d=v) = sum_{s in immediate_succesor-decision-events(e)}
;;;                  max_{v' in Domain(s)}(g(s=v')+h(s=v'))
;;;     This can be determined as the upper bound on the costs of
;;;     the decision-even

;;; Fix:
;;; o Update new-opsat-problem with the new variable spec format.
;;;   (done?).

;;; To Do:
;;; o Compute heuristic costs and provide to OPSAT.
(defmethod construct-selection-problem-for-tpn
	((tpn tpn-w-selection) &key (return-description? nil))

  "Given a TPN and corresponding CHOICE-VARIABLE-SPECIFICATION 
   as input, 

   Return a dynamic OCSP that frames the TPN selection problem.

   If RETURN-DESCRIPTION? is true, then returns a description
   of a choice-selection problem, rather than an instance
   of an opsat object, representing the problem."

  (multiple-value-bind (decision-variable-specs
						state-variable-specs
						constraint-specs)
	  (construct-dynamic-OCSP-specs-from-TPN tpn)

	;; (format t "decision-variable-specs: ~s~%" decision-variable-specs)
	;; (format t "state-variable-specs: ~s~%" state-variable-specs)
	;; (format t "constraint-specs: ~s~%" constraint-specs)

	;; (multiple-value-bind (decision-variable-specs2
	;; 					  state-variable-specs2
	;; 					  constraint-specs2)
	;; 	(construct-dynamic-OCSP-specs-from-TPN-ask-tell tpn)
	;;   (setf decision-variable-specs (append decision-variable-specs 
	;; 										decision-variable-specs2))
	;;   (setf state-variable-specs (append state-variable-specs 
	;; 									 state-variable-specs2))
	;;   (setf constraint-specs (append constraint-specs 
	;; 								 constraint-specs2)) )
	
	(if (tp-contains-loops? tpn)

	  (if return-description?
		  `(new-opsat-problem ,decision-variable-specs
							  ,state-variable-specs
							  ,constraint-specs
							  :constraint-system-type ITC::LTNP
							  :start-event-name (event-name (tpn-start ,tpn));;;Added by Peng; Add a new slot in opsat-problem to store start event name from tpn.
							  :conditional? T
							  :source ,tpn
							  :name (list :choice-selection ,tpn))

		  (new-opsat-problem decision-variable-specs
							 state-variable-specs
							 constraint-specs
							 :constraint-system-type ITC::'LTNP
							 :start-event-name (event-name (tpn-start tpn));;;Added by Peng; Add a new slot in opsat-problem to store start event name from tpn.
							 :conditional? T
							 :source tpn
							 :name (list :choice-selection tpn)
							 
							 :binary-objective-function #'+
							 :cost-offset 0
							 :cost-order-function #'<=
							 ))
	  
	  ;; Else: TPN without loops
	  (if return-description?
		  `(new-opsat-problem ,decision-variable-specs
							  ,state-variable-specs
							  ,constraint-specs
							  :constraint-system-type ITC-STN
							  :start-event-name (event-name (tpn-start ,tpn));;;Added by Peng; Add a new slot in opsat-problem to store start event name from tpn.
							  :conditional? T
							  :source ,tpn
							  :name (list :choice-selection ,tpn))

		  (new-opsat-problem decision-variable-specs
							 state-variable-specs
							 constraint-specs
							 :constraint-system-type 'ITC-STN
							 :start-event-name (event-name (tpn-start tpn));;;Added by Peng; Add a new slot in opsat-problem to store start event name from tpn.
							 :conditional? T
							 :source tpn
							 :name (list :choice-selection tpn)
							 
							 :binary-objective-function #'+
							 :cost-offset 0
							 :cost-order-function #'<=
							 )))))

(defmethod construct-explanation-problem-for-tpn
	((tpn tpn-w-selection) &key (return-description? nil))

  "Given a TPN and corresponding CHOICE-VARIABLE-SPECIFICATION 
   as input, 

   Return a dynamic OCSP that frames the TPN selection problem.

   If RETURN-DESCRIPTION? is true, then returns a description
   of a choice-selection problem, rather than an instance
   of an opsat object, representing the problem."

  (multiple-value-bind (decision-variable-specs
						state-variable-specs
						constraint-specs)
	  (construct-explanation-dynamic-OCSP-specs-from-TPN tpn)

;;	(format t "Creating new OPSAT problem~%")
;;	(format t "Constraint-specs ~a~%" constraint-specs)


	;;  (format t "Decision Variable Specs ~A~%" decision-variable-specs)
	;;  (format t "State Variable Specs ~A~%" state-variable-specs)
	;; (format t "Constarint Specs ~A~%" constraint-specs)


	;; (format t "Decision Variable Specs ~A~%" decision-variable-specs)
	;; (format t "State Variable Specs ~A~%" state-variable-specs)
	;; (format t "Constarint Specs ~A~%" constraint-specs)
	
	(if return-description?
		`(new-opsat-problem ,decision-variable-specs
							,state-variable-specs
							,constraint-specs
							:constraint-system-type ITC-STN
							:start-event-name (event-name (tpn-start tpn));;;Added by Peng; Add a new slot in opsat-problem to store start event name from tpn.
							:conditional? T
							:source ,tpn
							:name (list :choice-selection ,tpn))

	  (new-opsat-problem decision-variable-specs
						 state-variable-specs
						 constraint-specs
						 :constraint-system-type 'ITC-STN
						 :start-event-name (event-name (tpn-start tpn));;;Added by Peng; Add a new slot in opsat-problem to store start event name from tpn.
						 :conditional? T
						 :source tpn
						 :name (list :choice-selection tpn)
						 
						 :binary-objective-function #'+
						 :cost-offset 0
						 :cost-order-function #'<=
                      ))))


(defmethod construct-planning-problem-for-tpn
	((tpn tpn-w-selection) &key (return-description? nil))

  "Given a TPN and corresponding CHOICE-VARIABLE-SPECIFICATION 
   as input, 

   Return a dynamic OCSP that frames the TPN selection problem.

   If RETURN-DESCRIPTION? is true, then returns a description
   of a choice-selection problem, rather than an instance
   of an opsat object, representing the problem."

  (multiple-value-bind (decision-variable-specs
						state-variable-specs
						constraint-specs)
	  (construct-planning-dynamic-OCSP-specs-from-TPN tpn)

;;	(format t "Creating new OPSAT problem~%")
;;	(format t "Constraint-specs ~a~%" constraint-specs)

	;; (format t "Decision Variable Specs ~A~%" decision-variable-specs)
	;; (format t "State Variable Specs ~A~%" state-variable-specs)
	;; (format t "Constarint Specs ~A~%" constraint-specs)
	
	(if return-description?
		`(new-opsat-problem ,decision-variable-specs
							,state-variable-specs
							,constraint-specs
							:constraint-system-type ITC-STN
							:start-event-name (event-name (tpn-start tpn));;;Added by Peng; Add a new slot in opsat-problem to store start event name from tpn.
							:conditional? T
							:source ,tpn
							:name (list :choice-selection ,tpn))

	  (new-opsat-problem decision-variable-specs
						 state-variable-specs
						 constraint-specs
						 :constraint-system-type 'ITC-STN
						 :start-event-name (event-name (tpn-start tpn));;;Added by Peng; Add a new slot in opsat-problem to store start event name from tpn.
						 :conditional? T
						 :source tpn
						 :name (list :choice-selection tpn)
						 
						 :binary-objective-function #'+
						 :cost-offset 0
						 :cost-order-function #'<=
                      ))))

(defmethod construct-dynamic-OCSP-specs-from-TPN-ask-tell ((tpn tpn-w-selection))
  (let ((decision-variable-spec nil)
		(state-variable-spec nil)
		(constraint-spec nil))
	  ;; find episodes with like variables
	  (loop for episode in (tpn::tpn-episodes tpn) do
		   (let* ((tell-list (read-from-string (tpn::episode-tell episode)))
				  (var (second tell-list))
				  (val (third tell-list)))
			 ;; Added the line below to suppress compiler warnings.
			 (declare (ignore var val)))
		   

		   ;;nconcing (make-constraint-spec-episode episode) into constraint-specs

		   )
	(values decision-variable-spec state-variable-spec constraint-spec)
	))

(defmethod construct-dynamic-OCSP-specs-from-TPN ((tpn tpn-w-selection))

  "Given a TPN, Returns the elements of a Dynamic Optimal CSP 
   specification that frames the TPN selection problem:

   (VALUES DECISION-VARIABLE-SPEC+ STATE-VARIABLE-SPEC+ CONSTRAINT-SPEC+)

   where

   DECISION-VARIABLE-SPEC is of the form:
      (:NAME DECISION-EVENT-NAME 
       :VALUATION ((1 COST_1) ... (N cost_N))
      [:GUARD ACTIVATION-ASSIGNMENT])

   STATE-VARIABLE-SPEC is of the form:
      (:NAME DECISION-EVENT-NAME :DOMAIN (1 2 ...N)

     where N is the number of choices associated with
     DECISION-EVENT.

   CONSTRAINT-SPEC is of the form:
      (:START START-EVENT-NAME :END END-EVENT-NAME 
       :LB EPISODE-DURATION-LOWER-BOUND :UB EPISODE-DURATION-UPPER-BOUND
       :GUARD CHOICE-ASSIGNMENT :SOURCE EPISODE :CAN-BE-RELAXED? <TRUTH-VALUE>)"

  (multiple-value-bind (decision-variable-specs state-variable-specs)
	  (loop 
		 with decision-variable-spec
		 with state-variable-spec
		   
		 ;; Decision, state variable and constraint specs are built on an
		 ;; event by event basis.
		   
		 for decision-event in (tpn-decisions tpn)
		 ;;do (format t "Decision Event Name: ~a~%" decision-event)
		 do (multiple-value-setq (decision-variable-spec state-variable-spec)
			  
			  (make-variable-specs-for-decision-event 
			   decision-event tpn))
		   
		 collecting decision-variable-spec into decision-variable-specs

		 collecting state-variable-spec into state-variable-specs

		 finally (return (values decision-variable-specs
								 state-variable-specs)))
	(print "ddddddddddddddddddddddddddddddddddddd00")
	(loop
	   for episode in (tpn-episodes tpn)
	   collecting (make-constraint-spec episode) into constraint-specs
	   finally (return (values decision-variable-specs
							   state-variable-specs
							   constraint-specs)))
	
))

 
(defmethod construct-explanation-dynamic-OCSP-specs-from-TPN ((tpn tpn-w-selection) &key
															  (spec nil)
															  (collected-constraint-specs (list '()))
															  (collected-decision-variable-specs (list '()))
															  (collected-state-variable-specs (list '()))
															  (assignment-description-in nil)
															  (assignment-description-out nil)
															  (decision-variable-spec nil)
															  (state-variable-spec)
															  (constraint-spec-in nil)
															  (constraint-spec-out nil))

  "Given a TPN, Returns the elements of a Dynamic Optimal CSP 
   specification that frames the TPN selection problem:

   (VALUES DECISION-VARIABLE-SPEC+ STATE-VARIABLE-SPEC+ CONSTRAINT-SPEC+)

   where

   DECISION-VARIABLE-SPEC is of the form:
      (:NAME DECISION-EVENT-NAME 
       :VALUATION ((1 COST_1) ... (N cost_N))
       [:GUARD ACTIVATION-ASSIGNMENT])

   STATE-VARIABLE-SPEC is of the form:
      (:NAME DECISION-EVENT-NAME :DOMAIN (1 2 ...N)

     where N is the number of choices associated with
     DECISION-EVENT.

   CONSTRAINT-SPEC is of the form:
      (:START START-EVENT-NAME :END END-EVENT-NAME 
       :LB EPISODE-DURATION-LOWER-BOUND :UB EPISODE-DURATION-UPPER-BOUND
       :GUARD CHOICE-ASSIGNMENT :SOURCE EPISODE :CAN-BE-RELAXED? <TRUTH-VALUE>)"


  (pop collected-decision-variable-specs)
  (pop collected-state-variable-specs)
  


  (loop for episode in (tpn-activation-assignment-episodes 
                        *always-active* tpn)

      do (when (episode-relaxable episode)

		;; (multiple-value-bind (match vehicle-name action-name arguments)
		;; 	(match-primitive (episode-primitive episode))
		
		   ;; (when (equal action-name "duration_constraint")
			 ;; (format t "Episode Primitive ~A~%" action-name)
			 ;;  (format t "~A ~A~%" (episode-primitive episode) (episode-support episode))
			 ;;  (format t "adding ~A ~A ~A ~A~%" (episode-start episode) (episode-end episode) (episode-lowerbound episode) (episode-upperbound episode))
			 
			 
			 (setf assignment-description-in
				   (make-assignment
					:variable-name (format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
					:value 'in))
			 
			 (setf assignment-description-out
				   (make-assignment
					:variable-name (format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
					:value 'out))

			 (setf decision-variable-spec 
				   `(,(format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
					  :domain ((IN :cost 0) (OUT :cost ,(if (typep (episode-cost episode) 'constant-episode-cost)
															(constant-episode-cost-cost (episode-cost episode))
															(error "Unable to encode the cost ~a for episode ~a." (episode-cost episode) episode)))) 
					  :guard ,*always-active*))
			 
			 (setf state-variable-spec
				   `(,(format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
					  :domain (IN OUT)))
			 
			 (setf constraint-spec-in `(:start ,(event-name (episode-start episode))
											   :end ,(event-name (episode-end episode))
											   :lb ,(episode-lowerbound episode)
											   :ub ,(episode-upperbound episode)
											   :guard ,(list assignment-description-in)
											   :source ,(episode-name episode)
											   :can-be-relaxed? ,(can-episode-be-relaxed? episode)))
			 
			 (setf constraint-spec-out `(:start ,(event-name (episode-start episode)) 
												:end ,(event-name (episode-end episode)) 
												:lb 0
												:ub 1000000
												:guard ,(list assignment-description-out)
												:source ,(episode-name episode)
												:can-be-relaxed? ,(can-episode-be-relaxed? episode)))
			 
			 (push decision-variable-spec collected-decision-variable-specs)
			 (push state-variable-spec collected-state-variable-specs)
			 (push constraint-spec-in collected-constraint-specs)
			 (push constraint-spec-out collected-constraint-specs)
			 
			 ))

  (nbutlast collected-constraint-specs)
;; (format t "Collected ~A~%" collected-constraint-specs)
  (loop 
	  with decision-variable-spec
	  with state-variable-spec
		   
		   ;; Decision, state variable and constraint specs are built on an
		   ;; event by event basis.
		   
	  for decision-event in (tpn-decisions tpn)

	  do (multiple-value-setq (decision-variable-spec state-variable-spec)
		   
		   (make-variable-specs-for-decision-event 
			decision-event tpn))


	;;  collecting decision-variable-spec into decision-variable-specs

    ;;  collecting state-variable-spec into state-variable-specs
    
    ;;(push decision-variable-spec collected-decision-variable-specs)
    (nconc collected-decision-variable-specs (list decision-variable-spec))

    ;;(format t "   Adding: ~A~%" decision-variable-spec)
    ;;(push state-variable-spec collected-state-variable-specs)
    (nconc collected-state-variable-specs (list state-variable-spec))

    (setf spec (make-explanation-constraint-specs-activated-by-decision-event 
                decision-event tpn))
    
    (nconc collected-constraint-specs (first spec))
     ;; (pop collected-constraint-specs)
   ;; (format t "Got constraint-specs ~a~%" collected-constraint-specs)
    (nconc collected-decision-variable-specs (second spec))
   ;; (format t "Got decision-variable-specs ~a~%" collected-decision-variable-specs)
    (nconc collected-state-variable-specs (third spec))
    ;; (format t "Got state-variable-specs ~a~%" collected-state-variable-specs)
    )


	(when (equal NIL (first collected-constraint-specs))
	  (pop collected-constraint-specs))
	(when (equal NIL (first collected-decision-variable-specs))
	  (pop collected-decision-variable-specs))
	(when (equal NIL (first collected-state-variable-specs))
	  (pop collected-state-variable-specs))

	  (values collected-decision-variable-specs
							  collected-state-variable-specs
							  (nconc
          collected-constraint-specs
          (make-non-explanation-constraint-specs-activated-by-assignment
           *always-active* tpn)))
  
  )




(defmethod construct-planning-dynamic-OCSP-specs-from-TPN ((tpn tpn-w-selection) &key
															  (spec nil)
															  (collected-constraint-specs (list '()))
															  (collected-decision-variable-specs (list '()))
															  (collected-state-variable-specs (list '()))
															  (assignment-description-in nil)
															  (assignment-description-out nil)
															  (decision-variable-spec nil)
															  (state-variable-spec)
															  (constraint-spec-in nil)
															  (constraint-spec-out nil))

  "Given a TPN, Returns the elements of a Dynamic Optimal CSP 
   specification that frames the TPN planning problem."


  (pop collected-decision-variable-specs)
  (pop collected-state-variable-specs)
  

  ;; Get the first event
  ;;(setf start-event (tpn-start tpn))
  

  (loop for episode in (tpn-activation-assignment-episodes 
                        *always-active* tpn)

      do (when (episode-relaxable episode)

		;; (multiple-value-bind (match vehicle-name action-name arguments)
		;; 	(match-primitive (episode-primitive episode))
		
		   ;; (when (equal action-name "duration_constraint")
			 ;; (format t "Episode Primitive ~A~%" action-name)
			 ;;  (format t "~A ~A~%" (episode-primitive episode) (episode-support episode))
			 ;;  (format t "adding ~A ~A ~A ~A~%" (episode-start episode) (episode-end episode) (episode-lowerbound episode) (episode-upperbound episode))
			 
			 
			 (setf assignment-description-in
				   (make-assignment
					:variable-name (format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
					:value 'in))
			 
			 (setf assignment-description-out
				   (make-assignment
					:variable-name (format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
					:value 'out))
			;; (format t  "Episode cost ~A~%" (episode-cost episode))
			 (setf decision-variable-spec 
				   `(,(format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
					  :domain ((IN :cost 0) (OUT :cost ,(if (typep (episode-cost episode) 'constant-episode-cost)
															(constant-episode-cost-cost (episode-cost episode))
															(error "Unable to encode the cost ~a for episode ~a." (episode-cost episode) episode)))) 
					  :guard ,*always-active*))
			 
			 (setf state-variable-spec
				   `(,(format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
					  :domain (IN OUT)))
			 
			 (setf constraint-spec-in `(:start ,(event-name (episode-start episode))
											   :end ,(event-name (episode-end episode))
											   :lb ,(episode-lowerbound episode)
											   :ub ,(episode-upperbound episode)
											   :guard ,assignment-description-in
											   :source ,(episode-name episode)
											   :can-be-relaxed? ,(can-episode-be-relaxed? episode)))
			 
			 (setf constraint-spec-out `(:start ,(event-name (episode-start episode)) 
												:end ,(event-name (episode-end episode)) 
												:lb 0
												:ub 1000000
												:guard ,assignment-description-out
												:source ,(episode-name episode)
												:can-be-relaxed? ,(can-episode-be-relaxed? episode)))
			 
			 (push decision-variable-spec collected-decision-variable-specs)
			 (push state-variable-spec collected-state-variable-specs)
			 (push constraint-spec-in collected-constraint-specs)
			 (push constraint-spec-out collected-constraint-specs)
			 
			 ))

  (nbutlast collected-constraint-specs)
;; (format t "Collected ~A~%" collected-constraint-specs)
  (loop 
	  with decision-variable-spec
	  with state-variable-spec
		   
		   ;; Decision, state variable and constraint specs are built on an
		   ;; event by event basis.
		   
	  for decision-event in (tpn-decisions tpn)

	  do (multiple-value-setq (decision-variable-spec state-variable-spec)
		   
		   (make-variable-specs-for-decision-event 
			decision-event tpn))

;;	   (format t "Processing ~A~%" decision-variable-spec)

	;;  collecting decision-variable-spec into decision-variable-specs

    ;;  collecting state-variable-spec into state-variable-specs
    
    ;;(push decision-variable-spec collected-decision-variable-specs)
    (nconc collected-decision-variable-specs (list decision-variable-spec))

    ;;(format t "   Adding: ~A~%" decision-variable-spec)
    ;;(push state-variable-spec collected-state-variable-specs)
    (nconc collected-state-variable-specs (list state-variable-spec))

    (setf spec (make-explanation-constraint-specs-activated-by-decision-event 
                decision-event tpn))
    
    (nconc collected-constraint-specs (first spec))
     ;; (pop collected-constraint-specs)
   ;; (format t "Got constraint-specs ~a~%" collected-constraint-specs)
    (nconc collected-decision-variable-specs (second spec))
   ;; (format t "Got decision-variable-specs ~a~%" collected-decision-variable-specs)
    (nconc collected-state-variable-specs (third spec))
    ;; (format t "Got state-variable-specs ~a~%" collected-state-variable-specs)
    )


;;    (format t "Done~% ~A" collected-constraint-specs)
	(when (equal NIL (first collected-constraint-specs))
	  (pop collected-constraint-specs))


	  (values collected-decision-variable-specs
							  collected-state-variable-specs
							  (nconc
          collected-constraint-specs
          (make-non-explanation-constraint-specs-activated-by-assignment
           *always-active* tpn)))
  
  )


(defun match-primitive (primitive)
  "Matches a primitive string of the form VEHICLE.ACTION(ARG,ARG,...)
   and returns multiple values '(t VEHICLE ACTION (ARG ARG ...)).
   The first argument is t if there is a match, nil if not."
 (let ((match nil)
	   (regs nil))
       (multiple-value-setq (match regs)
         (cl-ppcre:scan-to-strings "\\s*(.*)\\s*\\.\\s*(.*)\\s*\\(\\s*(.*)\\s*\\)\\s*"  primitive))
       ;(format t "~s~%" match)
       ;(format t "~s~%" regs)
	   (if match
		   (let ((vehicle (string-trim '(#\Space #\Tab #\Newline) (aref regs 0)))
				 (action (string-trim '(#\Space #\Tab #\Newline) (aref regs 1)))
				 (arguments (aref regs 2)))
			 (setf vehicle (if (string= vehicle "") nil vehicle))
			 (setf action (if (string= action "") nil action))
			 (setf arguments (cl-ppcre:split "\\s*,\\s*" arguments))
			 (setf arguments (mapcar #'(lambda (x) (string-trim " " x))  arguments))
			 (values t vehicle action arguments))
		   (values nil nil nil nil))
	   ))


;;; *********************
;;;
;;;     Creating State and Decision Variable Specifications
;;;
;;; *********************

(defmethod make-variable-specs-for-decision-event
	((decision-event event-w-selection) (tpn tpn-w-selection))
  
  "Given DECISION-EVENT of TPN,
   Returns (VALUES DECISION-VARIABLE-SPEC STATE-VARIABLE-SPEC)
   the decision and state variable specifications of a DOCSP
   for DECISION-EVENT, encoding TPN decision selection.

   DECISION-VARIABLE-SPEC is of the form:
      (DECISION-EVENT-NAME 
       :VALUATION ((1 COST_1) ... (N cost_N))
       [:GUARD ACTIVATION-ASSIGNMENT-DESCRIPTION])

   STATE-VARIABLE-SPEC is of the form:
      (DECISION-EVENT-NAME :DOMAIN (1 2 ...N)

   where N is the number of choices associated with
   DECISION-EVENT."

  (let ((name (event-name decision-event))
		(activation-assignment
		 (event-activation-assignment decision-event)))
	
	(multiple-value-bind (domain valuation)

		;; Extract the valuation and domain for the variable
		;; corresponding to decision event.
		(decision-event-domain-and-valuation decision-event TPN)

	  (let* ((guard-spec 
			  
			  ;; Build the variable's guard condition.
			  (when activation-assignment
				`(:guard ,(choice-assignment-description activation-assignment))))
			 
			 ;; Place the domain and valuations in the desired format.
			 (decision-variable-spec 
			  `(
;;;				:name
				,name
;;;				:valuation
				:domain
				,valuation . ,guard-spec))
			 
			 (state-variable-spec
			  `(
;;;				:name
				,name :domain ,domain)))
 
		(values decision-variable-spec
				state-variable-spec)))))

(defmethod decision-event-domain-and-valuation 
	((decision-event event-w-selection) (tpn tpn-w-selection))
  
  "Given DECISION-EVENT of TPN,
   Return (VALUES DOMAIN VALUATIONS), where
   DOMAIN is the list of CHOICES that may be assigned to the decision
     (a list of integers 1 - N, where N is the number of choices).
   VALUATION is a list of (CHOICE CHOICE-COST) pairs."
  
  (loop
	 ;; Extract each choice assignment of DECISION-EVENT
	 for successor in (event-successor-episodes decision-event)
	 for assignment = (episode-activation-assignment successor)

	 ;; and its corresponding value and cost.
	 for value = (assignment-value assignment)
	   
	 for cost = (choice-assignment-cost assignment tpn)

	 collecting value into domain
	 collecting `(,value :cost ,cost) into valuation
	   
	 finally (return (values domain valuation))))

;;; Fix:
;;; o Need to be able to read in from XML, the rewards associated
;;;   with episodes.

;;; To Do:
;;; o Generalize the operation used by Kirk to combine episode costs from
;;;   addition  to any preferentially independent operation.
(defmethod choice-assignment-cost (assignment (tpn tpn-w-selection))
  
  "Returns the immediate cost of ASSIGNMENT of TPN.
   The cost of a choice-assignment is the accumulated cost of
   the episodes directly activated by choice."  

  (loop for episode in (tpn-activation-assignment-episodes 
						assignment tpn)					   
	  summing (get-episode-cost episode)))

;; Added by James Paterson to handle looping costs.
(defun get-episode-cost (episode)
  "Currently just return the upper bound of the looping episode cost."

  (cond
	((typep (episode-cost episode) 'looping-episode-cost)
	 (calculate-looping-cost (episode-cost episode) (ltd-n-upper-bound (episode-temporal-duration episode))))
	((typep (episode-cost episode) 'constant-episode-cost)
	 (constant-episode-cost-cost (episode-cost episode)))
	(t
	 (error "Unable to encode cost ~a." (episode-cost episode)))))

;;; *********************
;;;
;;;     Creating Constraint Specifications
;;;
;;; *********************

(defmethod make-constraint-specs-activated-by-decision-event 
	((decision-event event-w-selection) (tpn tpn-w-selection))
  
  "Returns a list of DOCSP constraint specifications
   corresponding to all episodes activated by on of the
   choice assignments of DECISION-EVENT."
  
  (loop 
	  for successor in (event-successor-episodes decision-event)
	  for assignment = (episode-activation-assignment
						successor)
							  
	  nconcing (make-constraint-specs-activated-by-assignment
				assignment tpn)))

(defmethod make-explanation-constraint-specs-activated-by-decision-event 
	((decision-event event-w-selection) (tpn tpn-w-selection) &key
	 (spec nil)
	 (current-constraint-specs (list ()))
	 (current-decision-variable-specs (list ()))
	 (current-state-variable-specs (list ())))
  
  "Returns a list of DOCSP constraint specifications
   corresponding to all episodes activated by on of the
   choice assignments of DECISION-EVENT."
  

  
  (loop 
	  for successor in (event-successor-episodes decision-event)
	  for assignment = (episode-activation-assignment
                     successor)
    

    do (setf spec (make-explanation-constraint-specs-activated-by-assignment
                   assignment tpn))

	 ;;  (format t "Adding ~a~%" spec)
						  
    (nconc current-constraint-specs (first spec))
    (nconc current-decision-variable-specs (second spec))
    (nconc current-state-variable-specs (third spec)))
  (pop current-constraint-specs)
  (pop current-decision-variable-specs)
  (pop current-state-variable-specs)
 ;; (format t "-Done with: ~A~%" decision-event)
    ;; (format t "decision-variable-specs ~a~%" current-decision-variable-specs)
  ;; (format t "state-variable-specs ~a~%" current-state-variable-specs)
 ;; (format t "constraint-specs ~a~%" current-constraint-specs)
  
  (values (list current-constraint-specs current-decision-variable-specs current-state-variable-specs)))


(defmethod make-constraint-specs-activated-by-assignment 
	(assignment (tpn tpn-w-selection))

  "Returns a list of DOCSP constraint specifications
   corresponding to all episodes activated by choice
   ASSIGNMENT."

  (loop for episode in (tpn-activation-assignment-episodes 
						assignment tpn)
	  collecting (make-constraint-spec episode)))

(defmethod make-non-explanation-constraint-specs-activated-by-assignment 
	(assignment (tpn tpn-w-selection) &key
	 (returned-specs nil))

  "Returns a list of DOCSP constraint specifications
   corresponding to all episodes activated by choice
   ASSIGNMENT."
  
  (loop for episode in (tpn-activation-assignment-episodes 
                        assignment tpn)
        do (when (not (episode-relaxable episode))
             (push (make-constraint-spec episode) returned-specs)))
  
  (values returned-specs))
  

(defmethod make-explanation-constraint-specs-activated-by-assignment 
    (assignment (tpn tpn-w-selection) &key
	 (spec nil)
	 (constraint-specs nil)
	 (decision-variable-specs nil)
	 (state-variable-specs nil))
  

  "Returns a list of DOCSP constraint specifications
   corresponding to all episodes activated by choice
   ASSIGNMENT."
  
  

  
  (loop for episode in (tpn-activation-assignment-episodes 
                        assignment tpn)


	 do (if (not (episode-relaxable episode))
			(push (make-constraint-spec episode) constraint-specs)
			
			(progn
			  
			  (setf spec (make-explanation-constraint-spec episode))
			  (nconc constraint-specs (first spec))
			  (push (second spec) decision-variable-specs)
			  (push (third spec) state-variable-specs)
			  )))
  
  ;; (format t "--decision-variable-specs ~A~%" decision-variable-specs) 
  
  (values (list constraint-specs decision-variable-specs state-variable-specs)))

(defmethod make-explanation-constraint-spec ((episode episode-w-selection) &key
											 (assignment-description nil))

  "Returns a DOCSP constraint specification corresponding to EPISODE,
   of the form:

   (:START START-EVENT-NAME :END END-EVENT-NAME 
    :LB EPISODE-DURATION-LOWER-BOUND :UB EPISODE-DURATION-UPPER-BOUND
    :GUARD ASSIGNMENT-DESCRIPTION :SOURCE EPISODE :CAN-BE-RELAXED? <TRUTH-VALUE>)"
  
  (setf assignment-description
		(choice-assignment-description
		 (episode-activation-assignment episode)))
  (let
      ;; Extract the elements of simple temporal constraint.
      
      ;; Start and end event.
      ((start-event-name (event-name
                          (episode-start episode)))
       (end-event-name (event-name
                        (episode-end episode)))
       
       ;; Upper and lower bound.
       (episode-duration-lower-bound 
        (episode-lowerbound episode))
       (episode-duration-upper-bound 
        (episode-upperbound episode))
       
       ;; Constraint is guarded by the activation-assignment
       ;; of episode.
       (assignment-description-in
        (make-assignment
         :variable-name (format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
	   :value 'in))
    
       (assignment-description-out
        (make-assignment
         :variable-name (format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
         :value 'out))
       
       (decision-variable-spec 
        `(,(format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
            :domain ((IN :cost 0) (OUT :cost ,(if (typep (episode-cost episode) 'constant-episode-cost)
												  (constant-episode-cost-cost (episode-cost episode))
												  (error "Unable to encode the cost ~a for episode ~a." (episode-cost episode) episode)))) 
            :guard ,assignment-description))
       
       (state-variable-spec
        `(,(format nil "~A-~A" (event-name (episode-start episode)) (event-name (episode-end episode)))
          :domain (IN OUT)))
       
       )
    
    ;; Build and return the simple temporal constraint spec,
    ;; conditioned on the guard of the episode.
    (values (list (list `(:start ,start-event-name 
                           :end ,end-event-name 
                           :lb ,episode-duration-lower-bound
                           :ub ,episode-duration-upper-bound
                           :guard ,(list assignment-description-in)
                           :source ,(episode-name episode)
                           :can-be-relaxed? ,(can-episode-be-relaxed? episode))
                  
                  `(:start ,start-event-name 
                           :end ,end-event-name 
                           :lb 0
                           :ub 100000
                           :guard ,(list assignment-description-out)
                           :source ,(episode-name episode)
                           :can-be-relaxed? ,(can-episode-be-relaxed? episode)))
            decision-variable-spec
            state-variable-spec)
            )))


(defmethod make-constraint-spec ((episode episode-w-activation-labeling))

  "Returns a DOCSP constraint specification corresponding to EPISODE,
   of the form:

   (:START START-EVENT-NAME :END END-EVENT-NAME 
    :LB EPISODE-DURATION-LOWER-BOUND :UB EPISODE-DURATION-UPPER-BOUND
    :GUARD ASSIGNMENT-DESCRIPTION :SOURCE EPISODE :CAN-BE-RELAXED? <TRUTH-VALUE>)"

  (let
	  ;; Extract the elements of simple temporal constraint.
	  
	  ;; Start and end event.
	  ((start-event-name (event-name
						  (episode-start episode)))
	   (end-event-name (event-name
						(episode-end episode)))
		


	   ;; Constraint is guarded by the activation-assignment
	   ;; of episode.
	   (assignment-description
		(if (eql *always-active* (episode-activation-assignment episode))
			*always-active*
			(if (episode-activation-assignment-is-conjunction? episode)
				(map 'list #'choice-assignment-description (episode-activation-assignment episode))
				(list (choice-assignment-description
					   (episode-activation-assignment episode))))
			))

	   ;; Upper and lower bound.
	   ; The two function calls, (brational) and (to-rinf) are important.
	   ; (brational) - changes the type from float to rational,
	   ;   a list type in which accuracy is maintained. Lack of accuracy was 
	   ;   causing trouble in the relaxation methods  STN-ub-scan, in propagate-ub.lisp
	   ;   and STN-lb-update!, in propagate-lb.lisp. In these methods rigid components
	   ;   would sometimes be detected as negative cycles, if one of the bounds was
	   ;   updated through some arithmetic, but the resulting rounding due to floating
	   ;   point error resulted in a number some small delta greater or smaller than
	   ;   the original value.
	   ; (to-rinf) - changes the infinity values to a large rational number.
	   ;   This fixes a problem with negative cycle detection. The problem is simple:
	   ;   We label each event with a value corresponding to its distance from the start
	   ;   node. If one event's distance from the start node is infinity, how will you 
	   ;   detect negative cycles in any subsequent events? Afterall, infinity +/- any 
	   ;   other number is still infinity, and not comparable.

	   (episode-duration-lower-bound (brational (to-rinf (episode-lowerbound episode))))
	   (episode-duration-upper-bound (brational (to-rinf (episode-upperbound episode))))

	   )

	 	   (when (> episode-duration-lower-bound 100000000)
		 (setf episode-duration-lower-bound 100000000))

		   (when (> episode-duration-upper-bound 100000000)
		 (setf episode-duration-upper-bound 100000000)	   )
	   (if (episode-contains-loops? episode)
		   (values `(:start ,start-event-name 
							:end ,end-event-name 
							:N-lb ,(ltd-n-lower-bound (episode-temporal-duration episode))
							:N-ub ,(ltd-n-upper-bound (episode-temporal-duration episode))
							:t-lb ,(btd-lower-bound (ltd-per-loop-duration (episode-temporal-duration episode)))
							:t-ub ,(btd-upper-bound (ltd-per-loop-duration (episode-temporal-duration episode)))
							:lb ,(* (ltd-n-lower-bound (episode-temporal-duration episode)) 
									(btd-lower-bound (ltd-per-loop-duration (episode-temporal-duration episode))))
							:ub ,(* (ltd-n-upper-bound (episode-temporal-duration episode)) 
									(btd-upper-bound (ltd-per-loop-duration (episode-temporal-duration episode))))
							:reward-function ,(episode-cost episode)
							:guard ,assignment-description
							:source ,(episode-name episode)
							:can-be-relaxed? ,(can-episode-be-relaxed? episode)))
		   ;; Build and return the simple temporal constraint spec,
		   ;; conditioned on the guard of the episode.
		   (values `(:start ,start-event-name 
							:end ,end-event-name 
							:lb ,episode-duration-lower-bound
							:ub ,episode-duration-upper-bound
							:guard ,assignment-description
							:source ,(episode-name episode)
							:can-be-relaxed? ,(can-episode-be-relaxed? episode))))))

;;; *********************
;;;
;;;   Determining if Episode Constraints Can Be Relaxed
;;;
;;; *********************

;;; To Do:
;;; o Generalize the conditions under which an episode's temporal constraint
;;;   can be relaxed, once we understand the ways that an episode can be
;;;   created, other than through a control program.
;;;   - This should be determined based on episode's support.
(defmethod can-episode-be-relaxed? ((episode episode))
  
  "Returns True iff episode and its corresponding constraints can be relaxed.
   Currently an episode can be relaxed only if the episode was
   created by a control program."
  
  (if (eql (episode-support episode) :control-program)
	  T
	nil))

;;; *********************
;;;
;;;   Print Dynamic OCSP for TPN
;;;
;;; *********************

;;; To Do:
;;; o This belongs in another file.

;;; A TPN problem is of the form:
;;; (NEW-OPSAT-PROBLEM <decision-variable-spec> <state-variable-spec>
;;;    <constraint-spec> . <keyword/value-property-list>)

(defun tpn-problem? (tpn-cop)
  "Returns t iff TPN-COP denotes an OPSAT problem description.
   This is an expression of the form (NEW-OPSAT-PROBLEM ...)"
  
  (eql (first tpn-cop) 'new-opsat-problem))

(defun tpn-problem-decision-variable-specifications (tpn-cop)
  (second tpn-cop))

(defun tpn-problem-state-variable-specifications (tpn-cop)
  (third tpn-cop))

(defun tpn-problem-constraint-specifications (tpn-cop)
  (fourth tpn-cop))

(defun tpn-problem-keyword-arguments (tpn-cop)
  (nthcdr 4 tpn-cop))
  
(defun print-tpn-problem (tpn-cop &optional (stream t))
  
  "Prints to STREAM the specification of an OPSAT problem
   that implements a TPN method selection problem.  This
   specification is in the format accepted by NEW-OPSAT-PROBLEM."
  (format stream "~% TPN OpSat Problem:")
  
  (format stream "~2%   Decision Variable Specifications:")
  (dolist (s (tpn-problem-decision-variable-specifications tpn-cop))
	(format stream "~%   ~a" s))
  
  (format t "~2%   State Variable Specifications:")
  (dolist (s (tpn-problem-state-variable-specifications tpn-cop))
	(format stream "~%   ~a" s))
  
  (format t "~2%   Constraints:")
  (dolist (s (tpn-problem-constraint-specifications tpn-cop))
	(format stream "~%   ~a" s)))
