;;;; Copyright (c) 2013

;;;; This file contains methods used to manipulate TPNs in place.
;;;;

(in-package :TPN)

(defun scale-tpn-costs! (tpn scale-factor)
  "Modifies `TPN` in place so that all costs are multiplied by `SCALE-FACTOR`."
  (flet ((scale-cost! (episode)
		   (let ((cost (episode-cost episode)))
			 (cond
			   ((typep cost 'constant-episode-cost)
				(setf (constant-episode-cost-cost cost) (* (constant-episode-cost-cost cost) scale-factor)))
			   ((typep cost 'linear-looping-episode-cost)
				(setf (looping-setup-cost cost) (* (looping-setup-cost cost) scale-factor))
				(setf (looping-cost-per-loop cost) (* (looping-cost-per-loop cost) scale-factor)))
			   (:default
				(error "I don't know how to scale ~a." cost))))))
	(mapc #'scale-cost! (tpn-episodes tpn)))
  tpn)

(defun join-tpns-in-series! (tpn1 tpn2 &key (lb 0) (ub 100000))
  "Merges `TPN1` and `TPN2` into a single TPN where `TPN1`'s end event
  is connected to `TPN2`'s start event with a [`LB`,`UB`] arc.
  
  `TPN1` is modified in place, `TPN2` is destroyed."
  (let ((end1 (tpn-end tpn1))
		(start2 (tpn-start tpn2))
		(end2 (tpn-end tpn2)))
	(merge-tpns! tpn1 tpn2)
	(setf (tpn-end tpn1) end2)
	(add-episode end1 start2 (make-episode end1 start2
										   lb
										   ub
										   (make-instance 'constant-episode-cost)
										   nil
										   tpn1
										   :temporal-duration (make-instance 'bounded-temporal-duration :lower-bound lb :upper-bound ub)))
	tpn1))

(defun insert-macro-tpn! (tpn tpn-to-insert start-event end-event)
  "Merges `TPN-TO-INSERT` into `TPN` with [0, 0] bounds between
  `START-EVENT` and `TPN-TO-INSERT`'s start episode and
  `END-EVENT` and `TPN-TO-INSERT`'s end episode.

  Modifies `TPN` in place and `TPN-TO-INSERT` is destroyed (no longer
  self consistent) and should be discarded."
  (let ((sub-start-event (tpn-start tpn-to-insert))
		(sub-end-event (tpn-end tpn-to-insert)))
	(merge-tpns! tpn tpn-to-insert)
	(add-episode start-event sub-start-event
				 (make-episode start-event
							   sub-start-event
							   0
							   0
							   (make-instance 'constant-episode-cost)
							   (list)
							   tpn
							   :temporal-duration (make-instance 'bounded-temporal-duration :lower-bound 0 :upper-bound 0)))
	(add-episode sub-end-event end-event
				 (make-episode sub-end-event
							   end-event
							   0
							   0
							   (make-instance 'constant-episode-cost)
							   (list)
							   tpn
							   :temporal-duration (make-instance 'bounded-temporal-duration :lower-bound 0 :upper-bound 0))))
  ;; Mark the TPN as unlabelled
  (tpn-unlabeled! tpn)
  tpn)

(defun merge-tpns! (tpn &rest tpns-to-merge)
  "Merges each TPN in `TPNS-TO-MERGE` into `TPN`. Modifies `TPN` and
  `TPNS-TO-MERGE` in place.

  Checks that there are no name clashes.

  **WARNING** The resulting TPN shares structure with the TPNs in
  `TPNS-TO-MERGE`. Additionally, invariants in `TPNS-TO-MERGE` are
  broken. Pointers to the TPNs in `TPNS-TO-MERGE` should be discarded
  after calling this function."
  (labels ((add-event-to-ht (tpn event)
			 ;; First, check to see if the event already exists in the
			 ;; HT.
			 (let ((event-ht (tpn-event-ht tpn))
				   (event-name (event-name event)))
			   (when (nth-value 1 (gethash event-name event-ht))
				 (error "Unable to merge TPNs! Name clash for event ~a." event-name))
			   (setf (gethash event-name event-ht) event)
			   (setf (event-tp event) tpn)))
		   (update-episode-pointer (tpn episode)
			 (setf (slot-value episode 'index) (next-episode-index tpn))
			 (setf (episode-tp episode) tpn))
		   (merge-tpn2-into-tpn1 (tpn1 tpn2)
			 (setf (tpn-events tpn1) (nconc (tpn-events tpn1) (tpn-events tpn2)))
			 (setf (tpn-decisions tpn1) (nconc (tpn-decisions tpn1) (tpn-decisions tpn2)))
			 (setf (tpn-episodes tpn1) (nconc (tpn-episodes tpn1) (tpn-episodes tpn2)))
			 (mapc #'(lambda (e) (add-event-to-ht tpn1 e))
				   (tpn-events tpn2))
			 (mapc #'(lambda (e) (update-episode-pointer tpn1 e))
				   (tpn-episodes tpn2))
			 (setf (slot-value tpn1 'last-of-episodes)
				   (slot-value tpn2 'last-of-episodes))
			 tpn1))
	;; Mark the TPN as unlabelled
	(tpn-unlabeled! tpn)
	(reduce #'merge-tpn2-into-tpn1 tpns-to-merge :initial-value tpn)))

(defun transform-tpn-names! (tpn transform-fn)
  "This function modifies `TPN` in place with new names for all
  events. The new name is decided by calling `TRANSFORM-FN` with the
  existing name and expecting a new name in return."
  ;; Iterate through the events, renaming them.
  (flet ((rename-event (event)
		   (let* ((old-name (event-name event))
				  (new-name (funcall transform-fn old-name)))
			 (remhash old-name (tpn-event-ht tpn))
			 (setf (gethash new-name (tpn-event-ht tpn)) event)
			 (setf (event-name event) new-name))))
	(mapc #'rename-event (tpn-events tpn)))
  ;; Mark the TPN as unlabelled
  (tpn-unlabeled! tpn)
  ;; Return the modified TPN
  tpn)

(defun flatten-tpn-loops! (tpn episodes-with-loops)

  "Flattens loops in the tpn by converting looping episodes into 
   episodes in series.
  
  Returns the modified TPN."

  (loop for episode-with-assigned-loops in episodes-with-loops
	 do 
	   (let* ((episode-with-loops (first episode-with-assigned-loops))
			  (number-of-loops (second episode-with-assigned-loops))
			  (temporal-duration (ltd-per-loop-duration (episode-temporal-duration episode-with-loops)))
			  (looping-start-event (episode-start episode-with-loops))
			  (looping-end-event (episode-end episode-with-loops))		  
			  (looping-start-event-name (event-name looping-start-event))
			  (looping-end-event-name (event-name looping-end-event))
			  (looping-episode-primitive (episode-primitive episode-with-loops)))
		 
		 (loop for loop-number from 1 to number-of-loops
			do 
			  (let* ((start-event-name (concatenate 'string looping-start-event-name " " "start loop" " " (write-to-string loop-number)))
					 (end-event-name (concatenate 'string looping-end-event-name " " "end loop" " " (write-to-string loop-number)))
					 (start-event (get-event start-event-name tpn))
					 (end-event (get-event end-event-name tpn))
					 (cost (make-instance 'constant-episode-cost)))

				(let ((episode (make-episode start-event 
											 end-event 
											 (btd-lower-bound temporal-duration)
											 (btd-upper-bound temporal-duration) 
											 cost
											 (list :episode (list :primitive looping-episode-primitive))
											 tpn
											 :temporal-duration temporal-duration)))   
				  (add-episode end-event looping-end-event episode))

				(when (eql loop-number 1)
				  (let ((first-episode (make-episode looping-start-event 
													 start-event 
													 0
													 0 
													 cost 
													 (list :episode (list :primitive looping-episode-primitive))
													 tpn
													 :temporal-duration (make-instance 'bounded-temporal-duration :lower-bound 0 :upper-bound 0))))   
					(add-episode looping-start-event start-event  first-episode)))

				(when (eql loop-number number-of-loops)
				  (let ((final-episode (make-episode end-event 
													 looping-end-event 
													 0
													 0 
													 cost 
													 (list :episode (list :primitive looping-episode-primitive))
													 tpn
													 :temporal-duration (make-instance 'bounded-temporal-duration :lower-bound 0 :upper-bound 0))))   
					(add-episode end-event looping-end-event final-episode)))))
		 (remove-episode episode-with-loops)))
  tpn)


(defun relax-tpn-loops! (tpn)
  "First, the function calls the LTNP scheduler to find the optimal number of loops for
  every looping episode."

  "Then, it modifies looping episodes to be non-looping episodes, with relaxed temporal bound
  [N*tl,N*tu], where N is the optimal loopnumber for that episode found by the scheduler. 
  In particular, the function modifies the episode's temporal duration from a looping
  temporal duration to a bounded temporal duration.
  
  Returns the modified TPN."
 
  (let ((ltnp (make-instance 'ITC::LTNP :name  'LTNP1 :start-name (tp-start tpn)))
		(looping-constraint-to-episode-ht (make-hash-table :test #'equal)))
	(loop for episode in (tp-episodes tpn)
	   do (with-slots (start end temporal-duration cost) episode
			(if (typep temporal-duration 'looping-temporal-duration)
				(progn
				  (let ((constraint(ITC::insert-new-looping-constraint 
									ltnp 
									:source 'LTNP1 
									:start start
									:end end
									:t-lb (btd-lower-bound (ltd-per-loop-duration(episode-temporal-duration episode))) 
									:t-ub (btd-upper-bound (ltd-per-loop-duration(episode-temporal-duration episode)))
									:N-lb (ltd-n-lower-bound(episode-temporal-duration episode)) 
									:N-ub (ltd-n-upper-bound(episode-temporal-duration episode)) 
									:reward-function cost)))
					(setf (gethash constraint looping-constraint-to-episode-ht) episode)))
				
				(ITC::insert-new-constraint 
				 ltnp 
				 :source 'LTNP1 
				 :start start
				 :end end
				 :lb (btd-lower-bound (episode-temporal-duration episode)) 
				 :ub (btd-upper-bound (episode-temporal-duration episode))))))
	(ITC::solve-ltnp! ltnp)
		
	(loop for constraint being the hash-keys of looping-constraint-to-episode-ht 
	   using (hash-value episode) do
		 (when (not (eql (ITC::N-lb constraint) (ITC::N-ub constraint))) 
		   (error "Number of loops in the constraint is a range, not an Integer."))
		 (let ((per-loop-lb (btd-lower-bound (ltd-per-loop-duration(episode-temporal-duration episode))))
			   (per-loop-ub (btd-upper-bound (ltd-per-loop-duration(episode-temporal-duration episode))))
			   (number-of-loops (ITC::N-lb constraint)))
		   (setf (episode-temporal-duration episode) (make-instance 
													  'bounded-temporal-duration 
													  :lower-bound (* per-loop-lb number-of-loops)
													  :upper-bound (* per-loop-ub number-of-loops)))
		   (ground-primitive! episode "$NUM_LOOPS$" (write-to-string number-of-loops))
		   (print (episode-primitive episode)))))
  tpn)

(defun ground-primitive! (episode from to)
  (let ((new-primitive (episode-primitive episode)))
	(setf new-primitive (replace-all new-primitive from to))
	(setf (episode-primitive episode) new-primitive)
	;; Check grounding	
	(when (not (search to new-primitive)) (error "Primitive not grounded."))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Utility Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun replace-all (string part replacement &key (test #'char=))
"Returns a new string in which all the occurences of the part 
is replaced with replacement."
    (with-output-to-string (out)
      (loop with part-length = (length part)
            for old-pos = 0 then (+ pos part-length)
            for pos = (search part string
                              :start2 old-pos
                              :test test)
            do (write-string string out
                             :start old-pos
                             :end (or pos (length string)))
            when pos do (write-string replacement out)
            while pos))) 
