;;;; Copyright (c) 2013

;;;; This file contains methods used to transform an existing TPN into
;;;; a different TPN without destructively modifying the first.
;;;;
;;;; Currently, the main function in this file is a method used to
;;;; rename all events. This is useful for things like macro
;;;; expansion.

(in-package :TPN)

(defun merge-tpns (&rest tpns)
  "Merges each TPN in `TPNS`. This function does a lot of copying so
  it does not need to modify any TPN in place. If you care about
  performance, you may want to use `MERGE-TPNS!` instead."
  (let ((copies (mapcar #'copy-tpn tpns)))
	(apply #'merge-tpns! (car copies) (cdr copies))))

(defun prefix-tpn-names (tpn prefix)
  "This function returns a copy of `TPN` with all the event names
  prefixed with `PREFIX`. The second value this function returns is
  the function responsible for the prefixing (so it can be used to
  transform other strings into the new namespace used by the TPN)."
  (let* ((transform-fn #'(lambda (old)
						   (concatenate 'string prefix old)))
		 (transformed-tpn (transform-tpn-names tpn transform-fn)))
	(values transformed-tpn transform-fn)))

(defun postfix-tpn-names (tpn postfix)
  "This function returns a copy of `TPN` with all the event names
  postfixed with `POSTFIX`. The second value this function returns is
  the function responsible for the postfixing (so it can be used to
  transform other strings into the new namespace used by the TPN)."
  (let* ((transform-fn #'(lambda (old)
						   (concatenate 'string old postfix)))
		 (transformed-tpn (transform-tpn-names tpn transform-fn)))
	(values transformed-tpn transform-fn)))

(defun transform-tpn-names (tpn transform-fn)
  "This function takes `TPN` and creates a new TPN with new names for
  all events. The new name is decided by calling `TRANSFORM-FN` with
  the existing name and expecting a new name in return."
  ;; Make a copy of the TPN so we can manipulate it.
  (let ((copy (copy-tpn tpn)))
	(transform-tpn-names! copy transform-fn)
	;; Return the modified copy.
	copy))
