;;; -*- Mode:Common-Lisp; Package:CS; Base:10 -*-

;;; Copyright (c) 1999 - 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :CS)

;;; FIX:
;;; o mtu::declare-debug-print-system!
;;;   has been exported from MTU in its package definition,
;;;   and included in :BFIG.  Why isn't the function external (: vs ::)
;;;   and why isn't it part of :ITC?

(mtu::declare-debug-print-system! :CS)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Fix:
;;; o Check out stn-consistent?, and make sure that it maps to
;;;   consistent? properly.  Use the interface function consistent?
;;;   in tests.

;;; o Fix Consistent? and externalize-conflict.
;;;   - rewrite consistent? for itc and isat.
;;;   - change the documentation for externalize-conflict and key,
;;;     specifying that key returns *true* if the element is to be ignored.
;;      Currently it says nil, but the implementation assumes *true*.
;;;   - make sure that externalize removes redundant elements, if so required.

;;; o When conflict is nil, it is interpreted as the current
;;;   state being the conflict.  However, couldn't nil denote the
;;;   conflict of empty assignments, in which case everything is
;;;   inconsistent?  Check and correct this.

;;; To Do:

;;; o Incorporate explanation mixin into ITC.

;;; o Incorporate explanation mixin into ISAT.

;;; o Create a standard initialize! method.

;;; o Create an essential-constraint[-system] class with finite
;;;   domain state variables, implement :around methods for 
;;;   interface functions assert!, unassert! assign! and unassign!,
;;;   and incorporate into ISAT and ITC.

;;; o Add ! to the names for insert-constraint and remove-constraint,
;;;   and update callers: ITC, ISAT and OPSAT.

;;; o For consistent? Rethink a uniform representation for returned conflicts and the
;;;   Key function.  
;;;   - Should the elements of a conflict always be constraints?  In this
;;;     In this case an assignment object needs to be a specialization of
;;;     a constraint.  Alternatively, the elements of a conflict can either be
;;;     be constraints or assignments.  Either way, there needs to be "essential-"
;;;     classes that support this.

;;;   - Should key be a method of constraint system, or a function
;;;      supplied by the caller?  should it be a function of the conflict element
;;;      only, or constraint-system as well?

;;;   - We need a consistent way of identifying base constraints, and for 
;;;     specifying whether or not they should be included in a conflict.

;;; o Clean externalize-conflict up, by making conflict an instance of a class,
;;;   and by providing externalize-conflict (and key?) methods.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                Run-time Flags
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *default-to-incremental-consistency?* nil

  "True iff CONSISTENT-CONSTRAINTS? should perform incremental
   consistency checks.  Note: this does not set the default
   for interface function CONSISTENT?"
  )

(defun incremental-consistency! ()
  (setq  *default-to-incremental-consistency?* T))

(defun non-incremental-consistency! ()
  (setq  *default-to-incremental-consistency?* nil))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;          Essential Constraint and Constraint System Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:

;;; o The term BASE-CONSTRAINT has a different meaning in OPSAT than from
;;;   here.  This could create a name conflict when extending OPSAT to
;;;   support explanation.  We can avoid this name conflict by changine
;;;   base-constraints to required-constraints within 
;;;   essential-constraint-system.

;;; o Here we say that all constraint systems should support the addition
;;;   and removal of constraints.  While we require this for most problems
;;;   solved within the MERS group, this isn't a necessary feature of all
;;;   constraint solvers.  Could separate our solvers into those that
;;;   support constraint modification and those that don't.
(defclass Essential-Constraint-System () 

  ((name
	:initform (gentemp "CS-")
	:initarg :name
;;;	:reader constraint-system-name

	:documentation "NAME is any lisp object or expression 
                    that names constraint system.")

   (constraints
	:initform '()
	:reader constraint-system-constraints
	:documentation "List of all constraint objects that are members of 
                    CONSTRAINT SYSTEM.  Each element must be an
                    instance of essential-constraint.")

   (base-constraints
	:initform '()
	:reader constraint-system-base-constraints
	:documentation "List of all constraint objects of system
                    in which can-be-relaxed? is marked nil.")
   
   ;; (relaxable-constraints
   ;; 	:initform '()
   ;; 	:reader constraint-system-relaxable-constraints
   ;; 	:documentation "List of all constraint objects of system
   ;;                  in which can-be-relaxed? is marked True.")
   
   (inserted-constraints
	:initform '()
	:reader constraint-system-inserted-constraints
	:documentation "List of all constraint objects that are members of 
                    the constraints of CONSTRAINT SYSTEM, and are currently
                    indicated as inserted into CONSTRAINT SYSTEM.")

;;;   (source 
;;;	:initarg :source
;;;	:initform nil
;;;;;;	:accessor constraint-source
;;;	:documentation 
;;;	"SOURCE stores domain specific information about the constraint, 
;;;     and can be any lisp object.  Typically used to specify the source or
;;;     type of a constraint.")
;;;   
;;;   (removed?
;;;	;; Indicate that constraint starts as removed from constraint-system.
;;;	:initform T
;;;	:reader constraint-removed?
;;;	:documentation
;;;	"True iff CONSTRAINT is currently removed from its constraint-system.")
;;;  
;;;   (to-be-inserted?
;;;	;; Indicate that constraint starts as not needing to be inserted.
;;;	:initform nil
;;;	:documentation
;;;	"True iff CONSTRAINT needs to be inserted, and hasn't yet")
   )

  (:documentation "An essential class for all constraint systems
                   that supports the basic constraint system
                   interface of creating, adding and removing contraints
                   and for querying constraint consistency."))

(defmethod print-object ((obj essential-constraint-system) stream)
  
  (with-slots (name) obj 
    (format stream "~a" name)))

;;; To Do:
;;; o Define and document to-be-inserted? as a :reader slot?
(defclass Essential-Constraint () 

  ((name
	:initform (gentemp "C-")
	:initarg :name

	:documentation "NAME is any lisp object or expression 
                    that names constraint.")

   (constraint-system
	:initarg :constraint-system
	:initform nil
	:reader constraint-constraint-system
	:documentation "The CONSTRAINT SYSTEM that constraint is part of.")

   (source 
	:initarg :source
	:initform nil
	:reader constraint-source
	:documentation 
	"SOURCE stores domain specific information about the constraint, 
     and can be any lisp object.  Typically used to specify the source or
     type of a constraint.")

   
   ;; (can-be-relaxed?
   ;; 	:initarg :can-be-relaxed?
   ;; 	:initform nil
   ;; 	:reader constraint-can-be-relaxed?
   ;; 	:documentation 
   ;; 	"True iff stc can be relaxed by the problem solver.")
   
   (removed?
	;; Indicate that constraint starts as removed from constraint-system.
	:initform T
	:reader constraint-removed?
	:documentation
	"True iff CONSTRAINT is currently removed from its constraint-system.")
  
   (to-be-inserted?
	;; Indicate that constraint starts as not needing to be inserted.
	:initform nil
	:documentation
	"True iff CONSTRAINT needs to be inserted, and hasn't yet"))

  (:documentation "An essential constraint class for all constraint systems
                   that supports the basic constraint system
                   interface of creating, adding and removing contraints."))

(defmethod print-object ((obj essential-constraint) stream)
  (with-slots (name) obj 
    (format stream "~a" name)))

(defmethod initialize-instance :after 
		   ((constraint essential-constraint) 
			&rest extra-args)

  "Creates a new constraint."

  (declare (ignore extra-args))
  
  ;; Associate constraint with NETWORK (although do not insert).
  (let ((cs (constraint-constraint-system constraint)))
	(with-slots (constraints) cs
	  (push constraint constraints))
  
	(with-slots (base-constraints) cs
	  (push constraint base-constraints))
	
	))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;          Interface Function: New Constraint
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod new-constraint :around
		   ((constraint-system essential-constraint-system)
			&key &allow-other-keys)

  "Creates a new constraint for CONSTRAINT-SYSTEM, which is described by
   DESCRIPTION, and whose source is SOURCE.

   The function creates a new constraint, but DOES NOT insert it into 
   CONSTRAINT-SYSTEM, i.e., the constraint is a removed constraint.  
   Constraint may be inserted later by calling INSERT-CONSTRAINT.  

   This function DOES NOT check for duplicate constraints, 
   but merely proceeds to create additional copies.

   The function returns the newly created constraint."

  (let ((constraint (call-next-method)))

	(debug-print-for :cs t "Created constraint ~A" ;; into ~A." 
					 constraint
;;;				   constraint-system
					 )
	
	;; Return the new constraint.
	constraint))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;          Interface Function: Insert-Constraint
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Implementation of Standard Constraint System Interface for Dummy

;;;      INSERT-CONSTRAINT is a method that inserts a constraint
;;;       into CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed by 
;;;       CONSTRAINT-OBJECT as arguments.

(defmethod insert-constraint :around 
		   ((constraint-system essential-constraint-system)
			(constraint essential-constraint))
  
  "Inserts a CONSTRAINT into CONSTRAINT-SYSTEM.

   When called, this :around method ensures that the primary, :before
   and :after methods are only invoked if constraint was not already inserted.
   In addition, ensures that constraint has been place on the 
   constraint list, and removed? is set to nil.

   CONSTRAINT itself must have been created earlier using NEW-CONSTRAINT.
   If CONSTRAINT is already inserted, this function does nothing.  

   The function Returns T if constraint is newly added and nil if it
   is already inserted into constraint system."

  (debug-print-for :cs t "Inserting ~A" ;; into ~A." 
				   constraint
;;;				   constraint-system
				   )

  (with-slots (removed?) constraint
	(cond ((not removed?)

		   ;; Constraint is already in constraint-system, 
		   ;; nothing needs to be done.
		   (debug-print-for :cs t " Already inserted, do nothing.")
		   nil)
 
		  (:otherwise

		   (debug-print-for :cs t " Inserting ~a" constraint)
		   
		   ;; Add mark as inserted and
		   (setf removed? nil)

		   ;; add to the list of inserted constraints.
		   (with-slots (inserted-constraints) constraint-system
			 (push constraint inserted-constraints))
		 
		   ;; This is where a specialization of essential-constraint-system
		   ;; performs additional analysis.
	  
		   (call-next-method)
		   
		   ;; Return that constraint had been removed and insertion succeeded.
		   T)))
  )

(defmethod insert-constraint ((constraint-system essential-constraint-system)
							  (constraint essential-constraint))
  
  "Inserts a CONSTRAINT into CONSTRAINT-SYSTEM.  

   When called, the :around method ensures that this method is only 
   invoked if constraint was not already inserted.
   In addition, ensures that constraint has been place on the 
   constraint list, and removed? is set to nil.

   CONSTRAINT itself must have been created earlier using NEW-CONSTRAINT.
   If CONSTRAINT is already inserted, this function does nothing.  

   The function Returns T if constraint is newly added and nil if it
   is already inserted into constraint system."
  
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;          Interface Function: Remove-Constraint
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;  REMOVE-CONSTRAINT is a method that removes a constraint
;;;      from CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed by 
;;;      CONSTRAINT-OBJECT as arguments.  If constraint is already removed,
;;;      does nothing.

(defmethod remove-constraint :around
		   ((constraint-system essential-constraint-system)
			(constraint essential-constraint))
  
  "Removes CONSTRAINT from its CONSTRAINT-SYSTEM.
  
   When called, this :around method ensures that the primary, :before
   and :after methods are only invoked if constraint is currently inserted.
   In addition, ensures that constraint has been removed from the 
   constraint list, and removed? is set to T.

   If CONSTRAINT is already removed, this function does nothing.

   The function Returns T if constraint is newly removed and nil if it
   is already removed from constraint system."

  (debug-print-for :CS t "Removing ~A." ;; from ~A." 
				   constraint 
;;;				   constraint-system
				   )

  (with-slots (removed?) constraint
	(unless removed?

	  ;; Nothing needs to be done if the constraint is already removed.

	  ;; Set the removed status of the constraint, and
	  (setf removed? t)

	  ;; Remove from the list of inserted constraints.
	  (with-slots (inserted-constraints) constraint-system
		(setq inserted-constraints
		  (delete constraint inserted-constraints)))

	  ;; This is where a specialization of essential-constraint-system
	  ;; performs additional analysis.

	  (call-next-method)
	  
	  ;; Return that constraint was inserted and removal succeeded.
	  T))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;          Interface Function: Establish-Constraints!
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod establish-constraints!
	((constraint-system essential-constraint-system) 
	 constraints-to-establish)
  
  "Let CONSTRAINTS-TO-ESTABLISH be a subset of the constraints of CONSTRAINT-SYSTEM.
   Ensures that CONSTRAINTS-TO-ESTABLISH and only those constraints are
   inserted into constraint-system.

   Returns nothing meaningful."
  
  ;; Mark the constraints that should be inserted.
  (loop for constraint in constraints-to-establish
	  do (with-slots (to-be-inserted?) constraint
		   (setq to-be-inserted? T)))

  ;; Remove any constraint that is currently inserted but shouldn't.
  (with-slots (inserted-constraints) constraint-system
	(loop for constraint in inserted-constraints
		do (with-slots (to-be-inserted?) constraint
			 (unless to-be-inserted?
			   (remove-constraint constraint-system constraint)))))
		   
  ;; Add any constraint that should be inserted but currently isn't,
  ;; and clean up the marks.
  (loop for constraint in constraints-to-establish
	  do (with-slots (removed? to-be-inserted?) constraint
		   (when removed?
			 (insert-constraint constraint-system constraint))
		   (setq to-be-inserted? nil)
		   ))
  
  T)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;          Interface Function: Consistent?
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Rethink a uniform representation for returned conflicts and the
;;;   Key function.  
;;;   - Should the elements of a conflict always be constraints?  In this
;;;     In this case an assignment object needs to be a specialization of
;;;     a constraint.  Alternatively, the elements of a conflict can either be
;;;     be constraints or assignments.  Either way, there needs to be "essential-"
;;;     classes that support this.

;;;   - Should key be a method of constraint system, or a function
;;;      supplied by the caller?  should it be a function of the conflict element
;;;      only, or constraint-system as well?

;;;   - We need a consistent way of identifying base constraints, and for 
;;;     specifying whether or not they should be included in a conflict.
(defmethod consistent? :around

		   ((constraint-system essential-constraint-system)
			&key 
			(return-conflict? T) 
			(initialize? nil)
			&allow-other-keys)

  "Check consistency of inserted constraints in CONSTRAINT-SYSTEM.
   Returns (values CONSISTENT? CONFLICT).

   CONSISTENT? is non-NIL iff there exists an
   assignment to the CONSTRAINT-SYSTEM's variables
   that satisfies all inserted constraints.

   CONFLICT denotes a set of constraints or assignments that is mutually
   inconsistent, some times in combination with a base set of constraints.

   If constraint-system is found inconsistent and RETURN-CONFLICT? is TRUE, 
   then a CONFLICT is returned; otherwise, nil is returned.

   When returned, a conflict is generated by applying function 
   KEY to the constraint or assignment objects of the conflict,
   and by collecting the non-nil, irredundant values returned by KEY.

   KEY is a function used to filter and transform the elements comprising
   a conflict.  It takes as input an object denoting a constraint or assignment,
   and returns any arbitrary expression, or nil, if the constraint/assigment
   is to be ignored.

   If INITIALIZE? is True, then reset all constraint solver derived values;
   otherwise, update incrementally from previous values.

   Note: This :around method passes its values for KEY and RETURN-CONFLICT to
   conflict?'s primary methods to ensure that the same value is used.  To
   change the default values for KEY and RETURN-CONFLICT define another :around
   method for class specializations.

   Side Effect:  None"

  (declare (ignore initialize?))
  
  (debug-print-for :CS t "Test consistency of constraint system ~A." 
				   constraint-system)

  (multiple-value-bind (consistent? conflict)
	  (call-next-method 
;;;	                    constraint-system 
;;;						:return-conflict? return-conflict?
;;;						:initialize? initialize?
						)

	(cond ((not consistent?)

		   ;; Inconsistent
		   (debug-print-for :CS t "~A is inconsistent." constraint-system)
			 
		   (cond (return-conflict?

				  ;; Asked to return a conflict.
				  (debug-print-for :CS t "Returns conflict ~A." 
								   conflict)
					
				  (values nil conflict))
				 
				 (:otherwise
				  
				  ;; Not asked to return a conflict.
				  (values nil nil))))

		  (:otherwise

		   ;; Consistent
		   (debug-print-for :CS t "~A is consistent." constraint-system)
			 
		   (values T nil)))))

;;; To Do:  
;;; o Clean up externalize-conflict, by making conflict an instance of a class,
;;;   and by providing externalize-conflict (and key?) methods.
(defmethod externalize-conflict ((constraint-system essential-constraint-system)
								 conflict &key (key #'identity) (test #'eql) (key-returns-conjunction? nil)
								 &allow-other-keys)
  
  "Maps elements of CONFLICT of CONSTRAINT-SYSTEM from internal to external form.
   Applies KEY to each element of conflict and returns a list of those results
   that are not bound to *true*.  Removes duplicate elements, where equivalence
   of items is determined using TEST.

   KEY-RETURNS-CONJUNCTION? is a little bit of an ugly hack until
   externalize-conflict is cleaned up. (On a side note, we seriously
   need to do that). If T then it is assumed that KEY returns a
   conjunction of clauses in a list as opposed to a single clause."
  
  (if key-returns-conjunction?
	  (loop for element in conflict
		 for externalized-element = (funcall key element)
		 unless (eql externalized-element *true*)
		 appending externalized-element into external-conflict
		 finally (return (values (remove-duplicates external-conflict
													:test test))))
	  (loop for element in conflict
		 for externalized-element = (funcall key element)
		 unless (eql externalized-element *true*)
		 collecting externalized-element into external-conflict
		 finally (return (values (remove-duplicates external-conflict
													:test test))))))

(defmethod consistent? ((constraint-system essential-constraint-system)
						&key
						(return-conflict? T)
						(key #'identity)
						(test #'eql)
						(initialize? nil)
						&allow-other-keys)

  "Check consistency of inserted constraints in CONSTRAINT-SYSTEM.
   Returns (values CONSISTENT? CONFLICT).

   CONSISTENT? is non-NIL iff there exists an
   assignment to the CONSTRAINT-SYSTEM's variables
   that satisfy all inserted constraints.

   CONFLICT is a set denoting the source of the inconsistency.

   If CONSTRAINT-SYSTEM is found inconsistent and RETURN-CONFLICT? is TRUE, 
   then a CONFLICT is returned; otherwise, nil is returned.

    Internally a CONFLICT is a set of constraints that is inconsistent.  
    The returned conflict is generated by applying KEY to each element of
    an internal conflict, collecting those result that are not equal to *TRUE*,
    and removing duplicate elements using TEST.

    KEY defaults to IDENTITY.
   
    TEST takes as input pairs of externalized elements, and returns true
    iff the two elements are equal.  TEST defaults to EQL.

   If INITIALIZE? is True, then reset all constraint solver derived values;
   otherwise, update incrementally from previous values.

   The default method assumes that constraint-system is always consistent.
   
    Side Effect:  None"
  
  (declare (ignore return-conflict? key test initialize?))
  
  (values t nil))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;          Interface Function: Consistent-Constraints?
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; FIX:
;;; o Asking the constraint solver to initialize for every consistency test.
;;;   This is less efficient, but ITC produces a bug when asked to update
;;;   incrementally.
(defmethod consistent-constraints?
	((constraint-system essential-constraint-system) constraints-to-test)
  
  "Let CONSTRAINTS be a subset of the constraints of CONSTRAINT-SYSTEM.
   Returns True unless an assignment exists that satisfies CONSTRAINTS."
  
  (establish-constraints! constraint-system constraints-to-test)  
 
  (consistent? constraint-system 
  			   :initialize? (not *default-to-incremental-consistency?*)))

;;; Optimization:
;;; o Do this without consing.
(defmethod consistent-constraints-without? 
	((constraint-system essential-constraint-system)
	 (Constraint essential-constraint) Constraints)
  
  "Let CONSTRAINT be a constraint object, and 
       CONSTRAINTS be a list, denoting a set of constraint objects.
   Returns True iff no assignment exists that satisfies CONSTRAINTS - {CONSTRAINT}."

  ;; Important: Used remove rather than delete to avoid
  ;;    side-effecting Constraints.
  (let ((test-constraints (remove constraint constraints)))
	(consistent-constraints? constraint-system test-constraints)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;¯
;;;                Essential Constraint System Class
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; FIX:
;;; o Brought over from the explanation-class, rewrite.

;;;(defmethod consistent? ((constraint-system essential-constraint-system)
;;;						&optional &key
;;;								  (return-conflict? T)
;;;								  (key #'identity)
;;;						&allow-other-keys)
;;;
;;;  "Checks consistency of inserted constraints in CONSTRAINT-SYSTEM.
;;;   Returns (values CONSISTENT? CONFLICT).
;;;
;;;   CONSISTENT? is non-NIL iff there exists an
;;;   assignment to the CONSTRAINT-SYSTEM's variables
;;;   that satisfy all inserted constraints.
;;;
;;;   If constraint-system is found inconsistent and RETURN-CONFLICT? 
;;;   is TRUE, then a CONFLICT is returned.
;;;
;;;   A CONFLICT denotes a partial state variable assignment
;;;   that is inconsistent.  It is generated by applying function 
;;;   KEY to each state variable assignment that 
;;;   supports an inconsistent constraint of STN, and by 
;;;   collecting the non-nil, irredundant values returned by KEY.
;;;
;;;    By default, KEY returns the assignment.
;;;   
;;;    Side Effect:  None"
;;;
;;;  (debug-print-for :CS t "Test consistency of constraint system ~A." constraint-system)
;;;
;;;  ;; Check consistency.
;;;  (multiple-value-bind (consistent? conflict)
;;;	  (consistent-stn? STN)
;;;
;;;    (cond ((not consistent?)
;;;	   
;;;		   ;; Inconsistent
;;;		   (debug-print-for :CS t "~A is inconsistent." constraint-system)
;;;		   
;;;		   (cond (return-conflict?
;;;				  
;;;				  (setq conflict
;;;					(externalize-conflict conflict key))
;;;				  
;;;				  (debug-print-for :CS t "Returns conflict ~A." 
;;;								   conflict)
;;;				  (values nil conflict))
;;;				 
;;;				 (:otherwise
;;;				  (values nil nil))))
;;;
;;;		  ;; Consistent
;;;		  (:otherwise
;;;
;;;		   (debug-print-for :CS t "~A is consistent." constraint-system)
;;;		   
;;;		   (values T nil)))))
