;; -*- Mode:Common-Lisp; Package:CS; Base:10 -*-

;;; Copyright (c) 1999 - 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :CS)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                Explanations = Conflicts and Relaxations
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This file contains functions for extracting conflicts and
;;; relaxations (aka explanations) given a basic CSP solver.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                To Do
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Divide into separate files for the different types of explanation.
;;; o See below.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                Explainable Constraint System Mixin
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass essential-constraint-system-with-explanation 
	(essential-constraint-system)

  ()
  
  (:documentation "A mix-in for a constraint system that enables it to
                   generate explanations.  Requires that the class that
                   that it is included in supports the following methods:

                  INSERT-CONSTRAINT is a method that inserts a constraint
                    into CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed 
                    by CONSTRAINT-OBJECT as arguments.  If constraint is already 
                    inserted, does nothing.

                  REMOVE-CONSTRAINT is a method that removes a constraint
                    from CONSTRAINT-SYSTEM.  It takes CONSTRAINT-SYSTEM followed 
                    by CONSTRAINT-OBJECT as arguments.  If constraint is already 
                    removed, does nothing.

                 CONSISTENT? is a method that takes as arguments:
                   (CONSTRAINT-SYSTEM ESSENTIAL-CONSTRAINT-SYSTEM)
				   &key (return-conflict? T) (key #'identity)
				   (test #'eql) (initialize? nil) &allow-other-keys)

                   Check consistency of inserted constraints in CONSTRAINT-SYSTEM.
                   Returns (values CONSISTENT? CONFLICT).

                   CONSISTENT? is non-NIL iff there exists an
                   assignment to the CONSTRAINT-SYSTEM's variables
                   that satisfy all inserted constraints.

                   CONFLICT is a set denoting the source of the inconsistency.

                   If CONSTRAINT-SYSTEM is found inconsistent and RETURN-CONFLICT?
                   is TRUE,then a CONFLICT is returned; otherwise, nil is 
                   returned.

                   Internally a CONFLICT is a set of constraints that is 
                   inconsistent.  
                   The returned conflict is generated by applying KEY to each 
                   element of an internal conflict, collecting those result that 
                   are not equal to *TRUE*,and removing duplicate elements 
                   using TEST.

                   KEY defaults to IDENTITY.
   
                   TEST takes as input pairs of externalized elements, and 
                   returns true iff the two elements are equal.  TEST defaults 
                   to EQL.

                   If INITIALIZE? is True, then reset all constraint solver 
                   derived values; otherwise, update incrementally from previous 
                   values.

                   The default method assumes that constraint-system is always 
                   consistent.
   
                   Side Effect:  None

                  "))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;     1)      Returning A Single Minimal Conflict
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Modify to handle a base set of constraints.
;;;   All constraints within the base must appear in the returned conflict.
(defmethod conflict-of-constraints
	((constraint-system essential-constraint-system-with-explanation)
	 constraints &key (method :rePlayXplain))
	   
  "Let CONSTRAINTS be a list of constraint objects.
   Returns a minimal CONFLICT of CONSTRAINTS;
           that is, a subset of CONSTRAINTS such that 
           no proper subset of CONFLICT is inconsistent.

   Method is the method used to extract the conflict.
        :rePlayXplain - The method presented on slide 77,
           `Computing Explanations in Problem Solving,'
            Sullivan and Junker IJCAI09 Tutorial."
  
  (case method

	(:rePlayXplain
	 (rePlayXplain constraint-system constraints))

	(otherwise
	 (error "Method ~a for conflict-of-constraints is not supported.
           for constraint-system ~a and constraints ~a." 
			method constraint-system constraints))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                rePlayXplain
;;;
;;;   Returns a single minimal conflict,
;;;   using an algorithm that repeatedly extends
;;;   a candidate until it conflicts.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; From Slide 77, "Computing Explanations in Problem Solving,"
;;; Sullivan and Junker IJCAI09 Tutorial.

;;; To Do:
;;; o Modify to handle a base set of constraints.
;;;   All constraints within the base must appear in the returned conflict.
(defmethod rePlayXplain
	((constraint-system essential-constraint-system-with-explanation)
	 Constraints)
	   
  "Let CONSTRAINTS be a list of inconsistent constraint objects.
   Returns a minimal CONFLICT of CONSTRAINTS;
           that is, a subset of CONSTRAINTS such that 
           no proper subset of CONFLICT is inconsistent.
   Flags an error if constraints are consistent."

  (debug-print-for :cs t " Extracting minimal conflict via rePlayXplain for ~%~s"
				   constraints)

  (when (consistent-constraints? constraint-system constraints)
	
	(error "Constraints ~s should be inconsistent, but are consistent." 
		   constraints))
  
  (debug-print-for :cs t " Start with () and add constraints until inconsistent.")
  (loop with minimal-conflict = '()

	  for 
		;; Discover elements of minimal conflict, one at a time.
		;; For each iteration,
		;; accumulate elements of CONSTRAINTS, in order 
		;; until an inconsistent subset is found.
		;; This subset (denoted CONFLICT) is a conflict, but might not be minimal.
						  
		;; However, the last element added to CONFLICT must be part of the
		;; MINIMAL-CONFLICT, since this last element resulted in a 
		;; transition from consistent to inconsistent.
		;; Add the last element of CONFLICT to MINIMAL-CONFLICT,
		;; and recurse.
						  
		;; During successive calls, include the elements of MINIMAL-CONFLICT,
		;; before adding through elements of constraints sequentially.
		c = (loop with conflict = minimal-conflict
				for c2 in constraints
				do (debug-print-for :cs t "   Adding ~a." c2)
				   (push c2 conflict)
;;;				   (print conflict)
				   (debug-print-for :cs t "   Testing consistency.")
				unless (consistent-constraints? constraint-system conflict)
				return c2)
			  
	  do (debug-print-for :cs t "~%   ~a created inconsistency, add to minimal conflict."
						  c)
		 (push c minimal-conflict)
		 (debug-print-for :cs t "~%   Minimal conflict includes: ~s." 
						  minimal-conflict)
		 
		 (debug-print-for :cs t " Find next element of conflict.")
		 
		 ;; Stop as soon as these accumulated last elements
		 ;; are inconsistent.  These form a minimal conflict.
	  unless (consistent-constraints? constraint-system minimal-conflict)
		   
	  do (debug-print-for :cs t " Minimal conflict = ~a." minimal-conflict)
		(return minimal-conflict)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;     2)      Returning A Single Preferred
;;;               Conflict or Relaxation
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                preferred-conflict-of-constraints
;;;
;;;   Returns the most preferred minimal conflict,
;;;   using a sequential two pass algorithm.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod preferred-conflict-of-constraints
	((constraint-system essential-constraint-system-with-explanation)
	 ordered-constraints &key (method :LexQuicXplain) (Base-Constraints '()))

  "Let ORDERED-CONSTRAINTS and BASE-CONSTRAINTS be lists of constraint objects, where
    BASE-CONSTRAINTS + ORDERED-CONSTRAINTS is inconsisent,
    BASE-CONSTRAINTS typically denotes a set of consistent constraints, and
    ORDERED-CONSTRAINTS denotes a preference order on constraints, 
      from most to least preferred.

   Method is the method used to extract the conflict.

     :PreferredConflict - The method presented on slide 93,
        `Computing Explanations in Problem Solving,'
         Sullivan and Junker IJCAI09 Tutorial.
         Note: This doesn't yet support BASE-CONSTRAINTS constraints.

     :LexQuicXplain - Divide an conquer method based on QuickSort.
         The QuickXplain Algorithm, by Ulrich Junker. 
         `Quickxplain: Preferred explanations and relaxations
         for over-constrained problems.' In AAAI, pages 167-172, 2004.

   Returns the minimal conflict of BASE-CONSTRAINTS + ORDERED-CONSTRAINTS.
      This is the lexicographically most preferred amongst all
      minimal conflicts, that is, the one that maximizes the 
      selection of more preferred constraints."
  
  (case method

	(:PreferredConflict
	 (PreferredConflict constraint-system ordered-constraints base-constraints))

	(:LexQuicXplain
	 (LexQuicXplain constraint-system ordered-constraints base-constraints))

	(otherwise
	 (error "Method ~a for preferred-conflict-of-constraints is not supported.
           for constraint-system ~a and constraints ~a." 
			method constraint-system ordered-constraints))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   3)           PreferredConflict
;;;
;;;   Returns the most preferred minimal conflict,
;;;   using a serial two pass algorithm.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; From Slide 93, "Computing Explanations in Problem Solving,"
;;; Sullivan and Junker IJCAI09 Tutorial.

;;; To Do:
;;; o Modify to handle a base set of constraints.
;;;   All constraints within the base must appear in the returned conflict.
(defmethod PreferredConflict
	((constraint-system essential-constraint-system-with-explanation)
	 Ordered-Constraints &optional (Base-Constraints '()))
  
  "Let ORDERED-CONSTRAINTS and BASE-CONSTRAINTS be lists of constraint objects, where
    BASE-CONSTRAINTS + ORDERED-CONSTRAINTS is inconsisent,
    BASE-CONSTRAINTS typically denotes a set of consistent constraints, and
    ORDERED-CONSTRAINTS denotes a preference order on constraints, 
      from most to least preferred.

   Returns the minimal conflict of BASE-CONSTRAINTS + ORDERED-CONSTRAINTS.
      This is the lexicographically most preferred amongst all
      minimal conflicts, that is, the one that maximizes the 
      selection of more preferred constraints.

   Note:  Base-Constraints is not currently supported.  A non-nil Base-Constraints will
   result in an error.  Use LexQuicXplain instead."
  
  (unless (null base-constraints)
	(error
	 "Base constraints not yet supported by PreferredConflict, use LexQuicXplain."))
  
  ;; 1) Add constraints from ORDERED-CONSTRAINTS, from most to least preferred 
  ;;    until inconsistent.
  (let ((candidate-conflict'()))
	;; Candidate-conflict is a list of constraints,
	;;    ordered from least to most preferred,
	;;    such that the set of constraints is inconsistent.

	(loop for c in ordered-constraints
		do (push c candidate-conflict)
		until (not (consistent-constraints? constraint-system candidate-conflict)))
  
	;; 2) Remove constraints from least to most preferred, 
	;;    leaving the constraint out if inconsistency is maintained.
	(let ((candidate-conflict2 (copy-list candidate-conflict)))

	  (loop for c in candidate-conflict
		  unless (consistent-constraints-without?
				  constraint-system c candidate-conflict2)
		  do (delete c candidate-conflict2))

	  ;; candidate-conflict2 is a minimal conflict
	  ;; order from least to most preferred.

	  ;; Return minimal conflict in order of most to least preferred.
	  (nreverse candidate-conflict2))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                LexQuicXplain
;;;
;;;   Returns the most preferred minimal conflict, 
;;;   using a quicksort style algorithm (like QuickExplain).
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; From Slide 96, Computing Explanations in Problem Solving,"
;;; Sullivan and Junker IJCAI09 Tutorial.
;;; Called LeXplain in the tutorial, but renamed LexQuicXplain here
;;; to make explicit the connection to QuicXplain.

;;; QuickXplain Algorithm, by 
;;; Ulrich Junker. "Quickxplain: Preferred explanations and relaxations
;;; for over-constrained problems." In AAAI, pages 167-172, 2004.
(defmethod LexQuicXplain 
	((constraint-system essential-constraint-system-with-explanation)
	 Ordered-Constraints &optional (Base-Constraints '()))
	   
  "Let ORDERED-CONSTRAINTS and BASE-CONSTRAINTS be lists of constraint objects, where
    BASE-CONSTRAINTS + ORDERED-CONSTRAINTS is inconsisent,
    BASE-CONSTRAINTS typically denotes a set of consistent constraints, and
    ORDERED-CONSTRAINTS denotes a preference order on constraints, 
      from most to least preferred.

   Returns the minimal conflict of BASE-CONSTRAINTS + ORDERED-CONSTRAINTS.
      This is the lexicographically most preferred amongst all
      minimal conflicts, that is, the one that maximizes the 
      selection of more preferred constraints."
   
  (cond ((not (consistent-constraints? constraint-system Base-Constraints))
		 nil)
		
		((singleton? Ordered-Constraints)
		 Ordered-Constraints)
		
		(:otherwise
		 
		 (multiple-value-bind (Ca Cb)
			 (split-constraints ordered-constraints)

		   (let* ((E2 (LexQuicXplain 
					   constraint-system Cb (union-disjoint-constraints Base-Constraints Ca)))
				  (E1 (LeXQuicXplain
					   constraint-system Ca (union-disjoint-constraints Base-Constraints E2)))
				  (E3 (union-disjoint-constraints E1 E2)))
;;;			 (print E3)
			 E3)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   4)           Preferred-Relaxation-of-Constraints
;;;
;;;   Returns the most preferred maximal relaxation.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Modify to handle a base set of constraints.
;;;   All constraints within the base must appear in the returned relaxation.
(defmethod preferred-relaxation-of-constraints
	((constraint-system essential-constraint-system-with-explanation)
	 ordered-constraints &key (method :PreferredRelax))
	   
  "Let ORDERED-CONSTRAINTS be a list of mutually inconsistent constraint objects,
       where the list denotes a total preference order, 
       from most to least preferred.

   Method is the method used to extract the conflict.
        :rePlayXplain - The method presented on slide 93,
           of `Computing Explanations in Problem Solving,'
            Sullivan and Junker IJCAI09 Tutorial.
  
   Returns the maximal relaxation of ORDERED-CONSTRAINTS
           that is lexicographically most preferred amongst all
           maximal relaxations, that is, maximizes the selection of
           more preferred constraints."

  (case method

	(:PreferredRelax
	 (PreferredRelax constraint-system ordered-constraints))

	(otherwise
	 (error "Method ~a for preferred-relaxation-of-constraints is not supported.
           for constraint-system ~a and constraints ~a." 
			method constraint-system ordered-constraints))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                PreferredRelax
;;;
;;;   Returns the most preferred maximal relaxation.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; From Slide 93 of "Computing Explanations in Problem Solving,"
;;; Sullivan and Junker IJCAI09 Tutorial.

;;; To Do:
;;; o Check that this is a correct interpretation.
;;; o Modify to handle a base set of constraints.
;;;   All constraints within the base must appear in the returned relaxation.
(defmethod PreferredRelax
	((constraint-system essential-constraint-system-with-explanation)
	 Ordered-Constraints)
	   
  "Let ORDERED-CONSTRAINTS be a list of mutually inconsistent constraint objects,
       where the list denotes a total preference order, 
       from most to least preferred.

   Returns the maximal relaxation of ORDERED-CONSTRAINTS
           that is lexicographically most preferred amongst all
           maximal relaxations, that is, maximizes the selection of
           more preferred constraints."

  ;; Add each constraint of ORDERED-CONSTRAINTS, to the relaxation.
  ;; Keep if consistent, else remove.
  (loop with reversed-preferred-relaxation = '()
	  for c in ordered-constraints
	  do (push c reversed-preferred-relaxation)
	  unless (consistent-constraints? constraint-system reversed-preferred-relaxation)
	  do (pop reversed-preferred-relaxation)
	  finally (return (nreverse reversed-preferred-relaxation))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;     5)      Returning ALL Minimal Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Modify to handle a base set of constraints.
;;;   All constraints within the base must appear in the returned relaxation.
(defmethod all-conflicts-of-constraints
	((constraint-system essential-constraint-system-with-explanation)
	 Constraints &key (method :DAA-Min-Unsat) (test #'eql))

  "Let CONSTRAINTS be a set of constraints of CONSTRAINT-SYSTEM, and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns all minimal conflicts of CONSTRAINTS.

   A conflict of CONSTRAINTS is an inconsistent subset of CONSTRAINTS.
   A minimal conflict is a conflict, no subset of which is a conflict.

   Method is the method used to extract ALL conflicts.
        :Min-Unsat - method searches systematically through all inconsistent
          (and maximal relaxations) of Constraints.  Implements the method 
          described on pps. 4-5 of 
          James Bailey and Peter J. Stuckey, `Discovery of 
          Minimal Unsatisfiable Subsets of Constraints Using 
          Hitting Set Dualization,' In Manuel V. Hermenegildo 
          and Daniel Cabeza, editors, PADL, volume 3350 of 
          Lecture Notes in Computer Science, pages 174-186. 
          Springer, 2005.

       :DAA-Min-Unsat - Method exploits a dual relationship between 
          minimal conflicts (unsatisfiable sets of constraints and 
          maximal satisfiable sets of constraints to efficiently search
          for all minimal conflicts.  Taken directly from pp 5-8 
          (pseudo code on 7) of James Bailey and Peter J. Stuckey, listed above."
  
  (case method

	(:Min-Unsat
	 (min-unsat constraint-system constraints :test test))
	
	(:DAA-Min-Unsat
	 (multiple-value-bind (explanations conflicts)
		 (daa-min-unsat constraint-system constraints :test test)
	   (declare (ignore explanations))
	   conflicts))

	(otherwise

	 (error "Method ~a for all-conflict-of-constraints is not supported.
           for constraint-system ~a and constraints ~a with test ~a." 
			method constraint-system constraints test))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                Min-Unsat
;;;
;;; Returns all minimal conflicts of D u P,
;;; by searching systematically through all inconsistent
;;; (and maximal relaxations) of D u P.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Taken directly from pp 4-5 of 
;;;    James Bailey and Peter J. Stuckey, "Discovery of Minimal Unsatisfiable Subsets of 
;;;    Constraints Using Hitting Set Dualization," In Manuel V. Hermenegildo and Daniel 
;;;    Cabeza, editors, PADL, volume 3350 of Lecture Notes in Computer Science,
;;;    pages 174-186. Springer, 2005.

;;; Due to:
;;;   Maria Garcia de la Banda, Peter J. Stuckey, and Jeremy Wazny. Finding all mini- 
;;;   mal unsatisfiable subsets. In Proceedings of the 5th ACM SIGPLAN international 
;;;   conference on Principles and Practice of Declarative Programming, pages 32-43. 
;;;   ACM Press, 2003. 

;;; who generalized from :

;;;13. A. Hou. A theory of measurement in diagnosis from first principles. Artificial 
;;;    Intelligence, 65:281-328, 1994.

;;;12. B. Han and S-J. Lee. Deriving minimal conflict sets by CS-trees with mark set in 
;;;    diagnosis from first principles. IEEE Transactions on Systems, Man, and Cyber- 
;;;    netics, 29(2):281-286, 1999. 

;;; To Do:
;;; Implement the optimizations listed at the end of min-unsat-internal.
(defmethod min-unsat 
	((constraint-system essential-constraint-system-with-explanation)
	 P &key (test #'eql))

  "Let P be a set of constraints, and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns all minimal conflicts of P.

   A conflict of P is an inconsistent subset of P.
   A minimal conflict is a conflict, no subset of which is a conflict."

  (min-unsat-internal constraint-system
					  '() P '() :test test))

(defmethod min-unsat-internal 
	((constraint-system essential-constraint-system-with-explanation)
	 D P A &key (test #'eql))

  "Let D and P be disjoint sets of constraints,
   A be a set of minimal conflicts, and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns A, augmented with all minimal conflicts of D union P
   that contain D as a subset.

   A conflict of constraints U is an inconsistent subset of U.
   A minimal conflict is a conflict, no subset of which is a conflict."


  (unless (consistent-constraints? constraint-system (union-disjoint-constraints D P))

	;; Searches systematically through all proper subsets of 
	;; P that contain D, from smallest to largest.

	;; Adds to A, a subset of D union P that constitutes a minimal conflict
	;; (i.e., inconsistent and not already in A).


	(loop with P-prime = (copy-list P)
		for c in P
		do (setq P-prime (delete c P-prime :test test))
		   (setq A (min-unsat-internal constraint-system D P-prime A))
		   (setq D (add-unique-element c D)))

	;; By the end of the loop, D := D u P.
	;; which is inconsistent.
	
	;; If not subsumed by a conflict in A,
	;; then record D as a minimal conflict.
	(setq A (add-set-if-minimal (reverse D) A :test test)))
		 
  A)

;;; Min-UnSat is optimized by Garcial et al. by:

;;; (a) detecting constraints that are present in all minimal unsatisfiable subsets 
;;;     by preprocessing, 

;;; (b) taking into account constraints that must always be satisfiable once other 
;;;     constraints are not present, 

;;; (c) using reasoning about independence of constraints to reduce 
;;;     the number of subsets examined, and 

;;; (d) most importantly, using incremental constraint solving to select which elements
;;;      c to select first in the while loop. 

;;; The last modification is the most important in terms of reducing the amount of 
;;; satisfaction checking and subsets examined. Essentially by performing the 
;;; satisfiability check sat(D union P) incrementally, we find the first constraint ci where 
;;; D union {c1 , . . . , ci-1 } is satisfiable and D union {c1 , . . . , ci } is not. 
;;; This guarantees that ci appears in some minimal unsatisfiable set. By choosing 
;;; c = ci in the while loop we (hopefully) quickly find large satisfiable subsets thus 
;;; reducing the search. 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                Dualize and Advance: DAA-Min-Unsat
;;;
;;; Returns all minimal conflicts of D u P.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Taken directly from pp 5-8 (pseudo code on 7) of 
;;;    James Bailey and Peter J. Stuckey, "Discovery of Minimal Unsatisfiable Subsets of 
;;;    Constraints Using Hitting Set Dualization," In Manuel V. Hermenegildo and Daniel 
;;;    Cabeza, editors, PADL, volume 3350 of Lecture Notes in Computer Science,
;;;    pages 174-186. Springer, 2005.

;;; Key idea: Exploit a dual relationship exists between the minimal conflicts (unsatisfiable 
;;; sets of constraints and the maximal satisfiable sets of constraints. In particular, 
;;; given set G of satisfiable constraint sets, let HST(G') denote the collection of 
;;; smallest sets not contained in any set from G. If we let G be the collection of all 
;;; maximal satisfiable constraint sets, then HST(G') is the collection of all smallest sets 
;;; not contained in any maximal satisfiable set (i.e. the minimal conflicts - unsatisfiable 
;;; constraint sets). Furthermore, if G is a collection of some, but not all maximal 
;;; satisfiable constraint sets, then HST(G') must contain at least one set which is satisfiable
;;; and is not contained in any set in G.

;;; For universe U, G' denotes the set of complements of each element of G, and HST(G')
;;; denotes the hitting sets of these complements; for example, let Universe
;;; U = {c1 , c2 , c3 , c4 } and let maximal satisfiable sets of constraints
;;; G = {{c3 }, {c4 }, {c2 }}. Then the complement sets are 
;;; G' = {{c1 , c2 , c3 }, {c1 , c3 , c4 }, {c1 , c2 , c4 }} and the hitting sets of G',
;;; equivalently the minimal unsatisfiable sets are 
;;; HST(G') = {{c1 }, {c2 , c4 }, {c2 , c3 }, {c3 , c4 }}.

;;; DAA-min-unsat jointly generates the minimal conflicts and the maximal satisfiable sets.
;;; Variables employed:
;;;  X is a satisfiable set, which is grown into a maximal satisfiable set M by the 
;;;    GROW procedure, which simply adds new constraints that do not cause unsatisfiability; 
;;;  A is the set of minimal conflicts currently found; 
;;;  Xc contains the complements of the maximal satisfiable sets currently found; and 
;;;  N are the hitting sets for Xc and are the candidates for minimal unsatisfiable subsets.

;;; Initially N (the set of minimal conflicts), Xc (the complements of maximal 
;;; satisfiable sets), and X (the candidate max sat set we are building) are empty. 

;;; In the repeat loop, the algorithm, repeatedly grows M to a maximal satisfiable subset, 
;;; then adds its complement to X' and calculates the new hitting sets for X',
;;; which is the set N, the candidate minimal conflicts.  For each element S of N not already 
;;; recognised as a minimal conflict (in A) we check satisfiability. If the set S 
;;; is satisfiable (not a conflict) then it is the starting point for a new maximal satisfiable 
;;; set, and we break the for loop and continue. Otherwise S is added to the minimal conflicts
;;; A. When we find no satisfiable S, then we have discovered all minimal unsatisfiable subsets. 

;;; Check:
;;  o set difference appears to reverse the order of remaining set elements.
;;;   Is this consistent?  Do we want to switch the order back?
(defmethod daa-min-unsat 
	((constraint-system essential-constraint-system-with-explanation)
	 U &key (test #'eql))

  "Let U denote a set of constraints (the Universe), and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns (value explanations minimal-conflicts)
   Where each explanation is a maximal relaxation / exclusion set pair,
   and each minimal conflict is a minimal inconsistent subset of U."
  
  ;;  R is a relaxed (satisfiable) set, typically not maximal.

  ;;  M is a maximal relaxed set, generated from X by the grow procedure.
  ;;    This simply adds new constraints that do not cause unsatisfiability; 
  
  ;;  Xc-set contains exclusion sets, which are the complements of the maximal relaxations
  ;;    currently found;

  ;;  H-set are the hitting sets for Xc-set and are the candidates for minimal unsatisfiable subsets;

  ;;  E-set contains explanations currently found, where each explanation is a maximal relaxation 
  ;;    and its corresponding exclusion set.
  
  ;;  C-set is the set of minimal conflicts currently found (the inconsistent subset of the 
  ;;    hitting sets).

  (loop 
	  with H-set = '(())
	  with R = '()						; Initial seed of set to grow to max sat set M.
	  with C-set = '()					; Minimal conflicts discovered.
	  with Xc-set = '()					; Exclusion sets (omplements of maximal relaxations 
										; discovered.
					
	  for M = (daa-min-unsat-grow constraint-system
								  R U :test test) ; Grow R to a new maximal relaxation M.
	  for Xc = (set-difference U M :test test) ; Compute the complement.
			   ;; Note: set-difference appears to be non-destructive, 
			   ;;    but seems like it reverses the order of U.  
	  for E = (list M Xc)
			  ;; An explanation, comprised of the maximal relaxation an its exclusion.
	  collecting E into E-set
						
	  do (setq
			 ;; Accumulate the complement of M (exclusion set) with the exclusion sets of 
			 ;; the other discovered maximal relaxations.
			 ;; Note: each maximal relaxation is unique, and hence its exclusion set as well.
			 Xc-set (add-unique-element Xc Xc-set) 
			 
			 ;; Incrementally compute the hitting sets of the exclusion sets;
			 ;; by duality, these are candidate conflicts,
			 ;; and consistent constraints that are not subsets of any
			 ;; extension discovered thus far.
			 H-set (extend-hitting-set Xc H-set :test test)
;;;			 H-set (Hitting-Set Xc-set :test test)  ; non-incremental version.

			 ;; Set the default value of the seed R for constructing the 
			 ;; next candidate maximal satisfiable set to nil, indicating no seed exists.
			 ;; We'll look through H-set for a new seed in a moment.
			 R '())					
		 
		 ;; Elements of the hitting set H-set are a) (minimal?) conflicts and
		 ;; b) (non-minimal) relaxations that are not subsets of any maximal relaxation found thus 
		 ;; far.
		 
		 ;; Search the hitting sets for a relaxation, to be used as the next
		 ;; seed R, passed to GROW, and record any conflicts found along the way.
		 (loop for H in H-set
			 unless (contains-set? H C-set :test test)
			 do (cond ((consistent-constraints? constraint-system H)
					   (setq R H)
					   (return))
					  (:otherwise
					   (setq C-set (add-unique-element H C-set)))))

		 ;; If no next seed found (consistent hitting set),
		 ;; then all minimal conflicts, maximal relaxations and exclusion sets have been found.
		 ;; Return the explanations and conflicts found.
		 (when (eql R '())
		   (return (values E-set C-set)))))

(defmethod daa-min-unsat-grow 
	((constraint-system essential-constraint-system-with-explanation)
	 S U &key (test #'eql))

  "Let S and U be (possibly overlapping) sets of constraints, and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns a set that extends S with each constraint of U (in order) 
   that doesn't result in an inconsistency."
  
  (loop for c in U
	  for Sp = (adjoin c S :test test)
			   
	  when (and (not (eql S Sp))
				(consistent-constraints? constraint-system Sp))
		   ;; If C is not already in S and 
		   ;; the addition of C to S is consistent then keep.
	  do (setq S Sp)
	  finally (return S)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;     6)      Returning All Explanations and Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod all-explanations-and-conflicts-of-constraints
	((constraint-system essential-constraint-system-with-explanation)
	 Constraints &key (method :DAA-Min-Unsat) (test #'eql))

  "Let CONSTRAINTS be a set of constraints of CONSTRAINT-SYSTEM, and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns all explanations and minimal conflicts of CONSTRAINTS,
   where each explanation is a maximal relaxation / exclusion set pair, and

   A conflict of CONSTRAINTS is an inconsistent subset of CONSTRAINTS.
   A minimal conflict is a conflict, no subset of which is a conflict.

   Method is the method used to extract ALL explanations and conflicts.

       :DAA-Min-Unsat - Method exploits a dual relationship between 
          minimal conflicts (unsatisfiable sets of constraints and 
          maximal satisfiable sets of constraints to efficiently search
          for all minimal conflicts.  Taken directly from pp 5-8 
          (pseudo code on 7) of James Bailey and Peter J. Stuckey, listed above."
  
  (case method

	(:DAA-Min-Unsat
	 (daa-min-unsat constraint-system constraints :test test))

	(otherwise

	 (error "Method ~a for all-conflict-of-constraints is not supported.
           for constraint-system ~a and constraints ~a with test ~a." 
			method constraint-system constraints test))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;     7)      Returning Representative Explanations
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod representative-explanations-of-constraints
	((constraint-system essential-constraint-system-with-explanation)
	 constraints 
	 &key (method :representativeXPlain) (test #'eql) (base-constraints '()))

  "Let CONSTRAINT-SYSTEM be a constraint system that includes 
   two sets of constraints: BASE-CONSTRAINTS and CONSTRAINTS, 
   where base is a list of consistent constraints, and
   BASE-CONSTRAINTS + CONSTRAINTS is inconsistent.

   Let TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns (values E-set and C-set).
   C-set is the set of Conflict that are discovered while
      constructing representative explanations.
   E-set is the minimal set of representative explanations,
     where each explanation is a pair, consisting of a maximal 
     relaxation R, (a subset of CONSTRAINTS), and its corresponding 
   minimal exclusion set (CONSTRAINTS - R).
  
   A set of explanations is representative if it contains a relaxation containing
   constraint C when C appears in some maximal relaxation, and if it contains an
   exclusion set containing constraint C when C appears in some minimal exclusion set.

   Method is the method used to extract a representative set of explanations.
        :representativeXPlain - a variant of the DAA-min-unsat with a greedy
          heuristic, biased towards enumerating a representative set first.
          Method taken directly from pp 326 of Barry O'Sullivan, Alexandre 
          Papadopoulos, Boi Faltings and Pearl Pu, `Representative Explanations
          for Over Constrained Problems,'  In AAAI, pages 323-328. AAAI Press, 2007."
  
  (case method

	(:representativeXPlain
	 (representativeXPlain constraint-system 
						   base-constraints constraints :test test))

	(otherwise
	 (error "Method ~a for all-conflict-of-constraints is not supported.
           for constraint-system ~a and constraints ~a with test ~a." 
			method constraint-system constraints test))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                Representative XPlain
;;;
;;; Returns a minimal representative set of explanations for 
;;; inconsistent constraints B u U, where B is consistent.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Taken directly from pp 326 of 
;;;    Barry O'Sullivan, Alexandre Papadopoulos, Boi Faltings and Pearl Pu, 
;;;    "Representative Explanations for Over Constrained Problems," 
;;;    In AAAI, pages 323-328. AAAI Press, 2007.

;;; To Do:
;;; o Determine according to the paper what should be returned.
(defmethod representativeXPlain 
	((constraint-system essential-constraint-system-with-explanation)
	 B U &key (test #'eql))
  
  "Let problem P = B union U denote a set of inconsisent constraints,
   where B, the base constraints, is consistent, and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Return (values E-set and C-set).  C-set is the set of discovered Conflicts.
   E-set is the minimal representative set of explanations,
   where each explanation is a pair, consisting of a maximal relaxation R subset U of P,
   and its corresponding minimal exclusion set U - R.
  
   A set of explanations is representative if it contains a relaxation containing
   constraint C when C appears in some maximal relaxation, and if it contains an
   exclusion set containing constraint C when C appears in some minimal exclusion set."
  
  ;; R denotes a particular maximal relaxation of P = B u U.
  ;; Xc-set denotes a set of minimal exclusion sets.
  ;; I-set denotes the set of (representative) explanations,
  ;;    comprised of relaxation / exclusion pairs (R, U \ R).

  (multiple-value-bind (E-set C-set)
	  (RepresentativeDA constraint-system B U)

	;; First generate representative minimimal exclusion sets
	;; that attempt to collectively contain all elements of U.

	;; Next, generate additional exclusions, so that every constraint
	;; in C is included in the corresponding maximum relaxation.
	(loop 
		with R
		with Xc
		for c in U
		when (and (loop for E in E-set
					  for Xc = (second E)
					  always (member c Xc :test test))
				  (consistent-constraints? constraint-system
										   (adjoin c B :test test)))
			 
			 ;; For each c that is not yet in a maximal relaxation,
			 ;; but is consistent with the Base.

			 ;; Find a maximal relaxation contain constraint,
			 ;; by extending Base + constraint to a maximal relaxation.
		do (setq R (representativeXplain-Grow constraint-system
											  B 
											  (list c) 
											  (remove c U :test test))
				 ;; and find its corresponding exclusion set.
				 Xc (set-difference U R :test test))

		   ;; Fix: The paper's approach seems unclear, and has been commented out.

		   ;; Records pairs comprised of relaxation R's exclusion set,
		   ;; and the minimal representative exclusion sets.
;;;		   (push (list (set-difference U R :test test) Xc-set) X-set)
		   
		   ;; Our Approach:
		   ;; Collect explanations for new representative maximal relaxations 
		   ;; to cover constraints contained in maximal relaxations.
		and collect (list R Xc) into new-E-set
		finally (setq E-set (nconc new-E-set E-set))
				;; Add new explanations to those covering all excluded constraints.
				)
	
	;; Finally, greedily remove explanations (exclusion sets)
	;; not required to maintain representativeness.
	(setq E-set (representativeXplain-minimize E-set))
	
	;; Return representative explanations and discovered conflicts.
	(values E-set C-set)))

;;; To Do:
;;; o Clean up the incremental hitting set computation in this function,
;;;   see comments below.
;;; o Do the same for the sister DA version.
(defmethod RepresentativeDA 
	((constraint-system essential-constraint-system-with-explanation)
	 B U &key (test #'eql))

  "Let P = B union U denote a set of base (B) and User (U) constraints,
   Such that B is consistent and B union U is inconsistent, and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns the sets of 1) representative explanations, and 2) conflicts found.
   An explanation is a maximal relaxation and its exclusion set.  
   Th property of representative explanations is that a) if a constraint C is contained
   in a maximal relaxation, then it is contained in a representative maximal relaxation,
   and if a constraint D is contained in a minimum exclusion set, then its is contained
   in a representative minimum exclusion.  This function only guarantees that
   constraints appear in the exclusions, not the maximal relaxations."
  
  ;;  R is a satisfiable set, which is grown into a maximal satisfiable set by the 
  ;;    GROW procedure, which simply adds new constraints that do not cause unsatisfiability; 

  ;;  C-set is the set of minimal conflicts currently found; 

  ;;  Xc-set contains the representative minimal exclusion sets, 
  ;;  the complements of the maximal satisfiable sets currently found, and 

  ;;  H-set are the hitting sets for Xc-set and are the candidates for minimal conflicts

  (loop 
	  with H-set = '(())
	  with R = '()						; max relaxation and initial seed used to grow to it.
	  with Xc = '()						; Minimum exclusion corresponding to R.
	  with Xc-set = '()					; minimal exclusions, complements of max sat sets discovered.
	  with C-set = '()					; minimal conflicts discovered.
	  with uncovered = U				; constraints not yet contained in an exclusion.
				   
	  do (setq
			 ;; Grow R to a new maximal satisfied set.
			 ;; first with constraints already covered in exclusion sets,
			 ;; and then those uncovered, to increase the likelihood that the exclusion of
			 ;; the maximal satisfied set will contain uncovered constraints.
			 R (RepresentativeXplain-Grow constraint-system
										  B R (set-difference U Uncovered :test test)
										  :test test) 
			 R (RepresentativeXplain-Grow constraint-system
										  B R U :test test)

			 ;; Create the minimum exclusion set for the maximum relaxation R.
			 ;; Note: set-difference appears to be non-destructive, 
			 ;;    but seems like it reserses the order of U.  
			 Xc (set-difference U R :test test))

	  collecting (list R Xc) into E-set
										; Explanations discovered, where each explanation
										; is a pair of a maximal relaxation and corresponding 
										; exclusion set.					   
	  do (unless (null (intersection Xc Uncovered :test test))
		   ;; Accumulate the new exclusion set with the exclusion sets for the other
		   ;; discovered maximal relaxations
		   ;; Note: Each maximal set is unique, and hence its complement as well.
		   
		   (setq Uncovered (set-difference Uncovered Xc)
				 Xc-set (add-unique-element Xc Xc-set))) 

		 ;; Incrementally compute the hitting sets of these complements;			 
		 ;; by duality, these are candidate conflicts,
		 ;; and consistent constraints that are not subsets of any
		 ;; extension discovered thus far.
		 (setq H-set (extend-hitting-set Xc H-set :test test)
;;;	           H-set (Hitting-Set Xc-set :test test) )  ; non-incremental version.

			   ;; Set the default value of the seed for constructing the 
			   ;; next candidate maximal relaxation
			   ;; to nil, indicating no seed exists.
			   ;; We'll look through h-set for a new seed in a moment.
			   R '())
		 
		 ;; Elements of the hitting set h-set are a) (minimal?) conflicts and
		 ;; b) consistent sets that are not subsets of any maximal sat sets found thus far.
		 
		 ;; Search the hitting sets for a consistent set, to be used as the next
		 ;; seed R, passed to GROW, and record any conflicts found along the way.
		 
		 ;; Fix:
		 ;; o Should we dequeue processed conflicts?
		 ;; o Should we dequeue processed seeds?
		 ;; o Are there redundant or non-minimal conflicts?
		 (loop for H in H-set
			 unless (contains-set? H C-set :test test)
			 do (cond ((consistent-constraints? constraint-system (union B H :test test))
					  
					   (setq R H)
					   (return))
					  
					  (:otherwise
					   
					   (setq C-set (add-unique-element H C-set)))))

		 ;; Stop either when all constraints represented in exclusions OR
		 ;; all mutual exclusions hae been found (no consistent hitting sets
		 ;; exist, to be used as a next seed).
		 (when (or (eql Uncovered '())
				   (eql R '()))
		   
		   ;; Return explanations and conflicts found.
		   (return (values E-set C-set)))))

(defmethod RepresentativeXplain-Grow 
	((constraint-system essential-constraint-system-with-explanation)
	 B S M &key (test #'eql))
  
  "Let S and M be (possibly overlapping) sets of constraints,
   B be a set of base constraints, such that B Union S is consistent, and
   TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Extends and returns S with each constraint
   of M (in order) that doesn't result in B union S being inconsistent."
  
  (loop 
	  with BuS = (union B S :test test)
	  for c in M
	  for BuSp = (adjoin c BuS :test test)
			   
	  when (and (not (eql BuS BuSp))
				(consistent-constraints? constraint-system BuSp))

		   ;; If C is not already in S and 
		   ;; the addition of C to S is consistent then keep.
	  do (setq BuS BuSp
			   S (add-unique-element c S))

		 
	  finally (return S)))

;;; FIX:
;;; o Implement this function, its currently a flow through.

;;; To Do:
;;; o A better algorithm might greedily select exclusions sets that
;;;   include / exclude the greatest number of constraints that have
;;;   yet to be included / excluded.

;;; Note: No pseudo code is provided in the paper this is our best guess
(defun RepresentativeXplain-minimize (minimal-exclusion-sets)
  
  "Let MINIMAL-EXCLUSION-SETS be a set of minimal-exclusion-sets.
   Greedily removes explanations not required to maintain representativeness, and
   Returns the reduced set of minimal exclusions.

   A representative set of exclusion sets contains and excludes every
   constraint that is included and excluded by a complete set of minimial
   excluaions."

  minimal-exclusion-sets)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;     8)      Conflicts of Explanation
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod Preferred-Conflicts-of-Explanation 
	((constraint-system essential-constraint-system-with-explanation)
	 ordered-maximal-relaxation exclusion-set &key (base-constraints '()))

  "Let ORDERED-MAXIMAL-RELAXATION, EXCLUSION-SET and BASE-CONSTRAINTS be disjoint 
   sets of constraints, 
   ORDERED-MAXIMAL-RELAXATION, is in order of decreasing preference, and
   BASE-CONSTRAINTS + ORDERED-MAXIMAL-RELAXATION is a maximal relaxation with
   complement EXCLUSION-SET.

   Returns an explanation for each excluded value e_i in EXCLUSION-SET,
      comprised of pairs <e_i, explanation_i> such that
      <e_i> o explanation_i o BASE-CONSTRAINTS
      forms the most preferred minimal conflict that is a subset of 
      <e_i> o ORDERED-MAXIMAL-RELAXATION o BASE-CONSTRAINTS."
  
  (loop for e in exclusion-set
	  for e-base = (cons e base-constraints)
	  for e-explanation = (preferred-conflict-of-constraints 
						   constraint-system ordered-maximal-relaxation
						   :method :LexQuicXplain :Base-constraints e-base)
	  collecting (list e e-explanation)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;     9)      Representative Explanations and their Conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Optimization:
;;; o Could improve efficiency by searching the returned conflicts first,
;;;   to explain each exclusion, and then generating the conflicts via
;;;   consistency tests if a suitable conflict doesn't already exist.

(defmethod Representative-Explanations-with-Explained-Exclusions
	((constraint-system essential-constraint-system-with-explanation)
	 Constraints &key (base-constraints '()) (test #'eql))

  "Let CONSTRAINTS and BASE-CONSTRAINTS be disjoint constraints of CONSTRAINT-SYSTEM,
   whose union is inconsistent, and let TEST be a binary relation, used to determine if two
   constraints are equivalent.

   Returns (values REPRESENTATIVE_EXPLANATIONS MINIMAL_CONFLICTS)
   where
      minimal_conflicts is a (partial) list of minimal conflicts
         of CONSTRAINTS u BASE_CONSTRAINTS, and

      explanation_i in representative explanations is a 
         is a triple (R_i, Exc_i, Exp_Exc_i>

   where R_i subset CONSTRAINTS, is a maximal relaxation of CONSTRAINTS u BASE-CONSTRAINTS,
         Exc_i is the exclusion set of R_i (constraints - R_i), and
         Exp_Exc_i is a list of explanations EE_ij for each exclusion Exc_ij in Exc_i
            where EE_IJ = <exc_ij, explanation_ij> such that
               {exc_ij} u explanation_ij forms a minimal conflict of
               {exc_ij} u R_i u BASE-CONSTRAINTS.
  
   A set of explanations is representative if it contains a relaxation containing
   constraint C when C appears in some maximal relaxation, and if it contains an
   exclusion set containing constraint C when C appears in some minimal exclusion set. "

  (multiple-value-bind (explanations conflicts)
	  (representative-explanations-of-constraints constraint-system
												  constraints
												  :base-constraints base-constraints
												  :test test)

	(values 
	 (loop for (relaxation exclusion) in explanations
		 collecting (list relaxation 
						  exclusion
						  (preferred-conflicts-of-explanation constraint-system
															  relaxation exclusion
															  :base-constraints
															  base-constraints)))
	 conflicts)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Support Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Manipulating Constraints
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun split-constraints (Ordered-Constraints)
  
  "Let ORDERED-CONSTRAINTS denote a list of constraints, 
      ordered from most to least preferred.
   Returns two lists of constraints, each ordered from most to least preferred,
      such that the two lists are roughly of equal length, and
         each element of the first list is preferred over
         each element of the second list."

  (let* ((split-length (round (* (length ordered-constraints) 0.5)))
		 (split1 (copy-list ordered-constraints))
		 (split2 (nthcdr split-length split1)))
	(setf (cdr 
		   (nthcdr (1- split-length) split1))
	  nil)
	(values split1 split2)))

;;; Optimization:
;;; o May allow union to be distructive, check and modify.
(defun union-disjoint-constraints (Constraints1 Constraints2)
  
  "Let CONSTRAINTS1 and CONSTRAINTS2 denote disjoint sets of constraints, 
      ordered from most to least preferred.
   Returns the union of the two sets, while maintaining order of preferrence."

  (append Constraints1 Constraints2))
