;;; -*- Mode:Common-Lisp; Package:CS; Base:10 -*-

;;; Copyright (c) 1999 - 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :CS)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Manipulating Sets
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun singleton? (set)
  
  "let SET be a list of elements.
   Returns true iff SET contains one element."

  (eql (length set) 1))

(defun equal-sets? (set1 set2 &key (test #'eql))
  
  "Let SET1 and SET2 denote sets of elements, and
   TEST be a binary relation, used to determine 
   if two set elements are equivalent.

   Returns T iff set1 and set2 are equivalent and nil otherwise.
   Set elements do not need to appear in the same order."
  
  (and (subsetp set1 set2 :test test)
	   (subsetp set2 set1 :test test)))

(defun contains-set? (set set-of-sets &key (test #'eql))
  
  "Let SET be a set of elements,
   set-OF-SETS be a set whose elements are sets of elements, and
   TEST be a binary relation, used to determine 
   if two set elements are equivalent.

   Returns T iff SET is an element of SET-OF-SETS."
  
  (loop for e in set-of-sets
	  when (equal-sets? set e :test test)
	  do (return t)
	  finally (return nil)))

;;; Check (???):
;;; o Do the input sets denote disjoint sets of constraints?
;;; o Are the inputs in any special order?
;;; o Is the output in any special order?

(defun add-unique-element (e set)
  
  "Adds E to set SET, where E is guaranteed to
   not already be contained in SET."
  
  (cons e set))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Sets of Minimal Sets
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun superset-of-set-element? (D A  &key (test #'eql))
	   
  "Let D be a set of elements,
   A be a set of sets of elements, and
   TEST be a binary relation, used to determine
   if two set elements are equivalent.

   Returns T iff an element of A exists that is
   a subset of D."
	   
  (loop for s in A
	  when (subsetp s D :test test)
	  return T
	  finally (return nil)))

(defun add-set-if-minimal (set sets &key (test #'eql))
  
  "Let SET be a set,
   SETS be a set of sets, and
   TEST be a binary relation, used to determine if two
   set elements are equivalent.

   Returns SETS, augmented by SET,
   unless SET is a superset of some element of SETS."
  
  (unless (superset-of-set-element? set sets :test test)
	(setq sets
	  (add-unique-element set sets)))
  sets)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;   Hitting Sets
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; To Do:
;;; o Look to merge with similar functions in OpSat,
;;;   to place in a separate system, and to make more efficient.

(defun Hitting-Set (sets &key (test #'eql))

  "Let SETS denote a set of sets of elements, and
   TEST be a binary relation, used to determine if two
   set elements are equivalent.

   Returns Hitting Sets, a set of minimal hitting sets of SET.
   A hitting set of SETS contains at least one element from each set in SETs.
   A minimal hitting set is a hitting set, no proper subset of which is a hitting set."
  
  (loop with HS = '(())
	  for set in sets
	  do (setq HS (extend-hitting-set set HS :test test))
	  finally (return HS)))

(defun Extend-Hitting-Set (set hitting-sets &key (test #'eql))
  
  "Let HITTING-SETS, be a set of minimal hitting sets, covering some unnamed set of sets,
   SET denote a set of elements not yet covered by HITTING-SETS, and
   TEST be a binary relation, used to determine if two
   set elements are equivalent.

   Returns a set of minimal hitting sets that HIT (cover) SET, 
   while augmenting HITTING-SETS.
   
   A hitting set of SETs is one containing at least one element from each set in SETS,
   A minimal hitting set of SET is a hitting set of SET such that no proper subset 
   is a hitting set of SET."

  (loop with augmented-hitting-sets = '()
	  for e in set

	  do (loop for hs in hitting-sets
			 for augmented-hs = (adjoin e hs :test test)
			 do (setq augmented-hitting-sets
				  (add-set-if-minimal augmented-hs augmented-hitting-sets :test test)))

	  finally (return augmented-hitting-sets)))
