;;; -*- Mode:Common-Lisp; Package:CS; Base:10 -*-

;;; Copyright (c) 1999 - 2010 Brian C. Williams, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Brian C. Williams.
;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.
;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :CS)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;             Support for Stand Alone Testing: 
;;;                Dummy Constraint System and Constraint
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass Dummy-Constraint-System (essential-constraint-system)

  ((name
	:initform (gentemp "Dummy-CS-")
	:documentation "NAME is any lisp object or expression 
                    that names dummy constraint system.")
   
   (constraint-conflicts
	:initarg :constraint-conflicts
	:initform '()
	:documentation "Subsets of constraints that are mutually inconsistent."))   
  
  (:documentation "A dummy constraint system, used for testing.
                   constraints are symbols, and ocnsistency specified through
                   conficts over constraint symbols."))

(defclass Dummy-Constraint (Essential-Constraint) 

  ()

  (:documentation "A dummy constraint, used for testing.
                   constraints are symbols, and consistency is
                   specified through conficts over constraint symbols."))

(defmethod new-constraint
	((constraint-system dummy-constraint-system)
	 &key name source &allow-other-keys)

  "Creates a new constraint for CONSTRAINT-SYSTEM named NAME, 
   which is created by source SOURCE.

   The function creates a new constraint, but DOES NOT insert it into 
   CONSTRAINT-SYSTEM, i.e., the constraint is a removed constraint.  
   Constraint may be inserted later by calling INSERT-CONSTRAINT.  

   This function DOES NOT check for duplicate constraints, 
   but merely proceeds to create additional copies.

   The function returns the newly created constraint."
  
  (make-instance 'Dummy-Constraint
	:name name
	:source source
	:constraint-system constraint-system))

(defmethod set-conflicts! ((constraint-system Dummy-Constraint-System) new-conflicts)

  "Records NEW-CONFLICTS as the complete set of conflicts for CONSTRAINT-SYSTEM.
   NEW-CONFLICTS is a list of conflicts, where each conflict is a list of
   constraints of constraint system that are unsatisfiable.

   Note:  This is not a standard interface function, but special to 
   dummy-constraint-system."

  (with-slots (constraint-conflicts) constraint-system

	(setq constraint-conflicts new-conflicts)))

(defmethod remove-constraint ((constraint-system dummy-constraint-system)
							  (constraint dummy-constraint))
  
  "Removes CONSTRAINT from its CONSTRAINT-SYSTEM.
  
   When called, this :around method ensures that the primary, :before
   and :after methods are only invoked if constraint is currently inserted.
   In addition, ensures that constraint has been removed from the 
   constraint list, and removed? is set to T.

   If CONSTRAINT is already removed, this primary method is not invoked.

   The function Returns T if constraint is newly removed and nil if it
   is already removed from constraint system
   "

  T)

(defmethod consistent? ((constraint-system dummy-constraint-system)
;;;						&optional 
						&key
;;;								  (return-conflict? T)
;;;								  (key #'identity)
						&allow-other-keys)

  "Checks consistency of inserted constraints in CONSTRAINT-SYSTEM.
   Returns (values CONSISTENT? CONFLICT).

   CONSISTENT? is non-NIL iff there exists an
   assignment to the CONSTRAINT-SYSTEM's variables
   that satisfy all inserted constraints.

   CONFLICT is always nil.  Dummy doesn't generate conflicts.
   
    Side Effect:  None"

  ;; Check consistency.
  (let ((consistent?
		 (with-slots (inserted-constraints constraint-conflicts) constraint-system
		   (not (superset-of-set-element?
				 inserted-constraints constraint-conflicts)))))

	(values consistent? nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;             Support for Stand Alone Testing: 
;;;                Dummy Constraint System
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass Dummy-CS-w-Explanation 
	(Dummy-Constraint-System essential-constraint-system-with-explanation)

  ()
  
  (:documentation "A dummy constraint system, used for testing, that can generate explanations.
                   constraints are symbols, and ocnsistency specified through
                   conficts over constraint symbols."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;             Stand Alone Testing
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftest-suite :CS
;;;    (log-CS-statistics)
;;;  (stop-and-print-CS-statistics)
	)

(deftest (test1 :CS) ()

  (let* ((constraint-system
		  (make-instance 'Dummy-CS-w-Explanation
			:name 'test-1))
		 (a (new-constraint constraint-system :name 'a))
		 (b (new-constraint constraint-system :name 'b))
		 (c (new-constraint constraint-system :name 'c))
		 (d (new-constraint constraint-system :name 'd))
		 (conflicts (list (list a c)))
		 (constraints (list a b c d)))

	(loop for constraint in constraints
		do (insert-constraint constraint-system constraint))
	
	(set-conflicts! constraint-system conflicts)
	
;;;	(format t "~% Constraints = ~a, Conflicts = ~A" constraints conflicts)

;;;	(format t "~% Min-Unsat Conflicts = ~A" 
;;;			(min-unsat constraint-system constraints))
	
	(test `((,A ,C))
		  (all-conflicts-of-constraints constraint-system constraints
										:method :min-unsat)
;;;		  (min-unsat constraint-system constraints)

		  :multiple-values nil
		  :test #'equal
		  :fail-info "Min-Unsat Conflicts = ((A C))")

;;;	(format t "~% rePlayXplain Conflict = ~A" 
;;;			(rePlayXplain constraint-system constraints))

	(test `(,A ,C)
		  (conflict-of-constraints constraint-system constraints
								   :method :rePlayXplain)
;;;		  (rePlayXplain constraint-system constraints)
		  :multiple-values nil
		  :test #'equal
		  :fail-info "rePlayXplain Conflict = (A C)")

;;;	(format t "~% Preferred-conflict-of-constraints = ~A" 
;;;			(Preferred-conflict-of-constraints constraint-system constraints))
	
	(test `(,A ,C)
		  (Preferred-conflict-of-constraints constraint-system constraints
											 :method :PreferredConflict)
;;;		  (PreferredConflict constraint-system constraints)
		  :multiple-values nil
		  :test #'equal
		  :fail-info "PreferredConflict = (A C)")

;;;	(format t "~% preferred-relaxation-of-constraints = ~A" 
;;;			(PreferredRelax constraint-system constraints))
	
	(test `(,A ,B ,D)
		  (preferred-relaxation-of-constraints constraint-system constraints
											   :method :PreferredRelax)
;;;		  (PreferredRelax constraint-system constraints)
		  :multiple-values nil
		  :test #'equal
		  :fail-info "PreferredRelax = (A B)")

;;;	(format t "~% LexQuicXplain Preferred Conflict = ~A" 
;;;			(LexQuicXplain constraint-system constraints))

	(test `(,A ,C)
		  (preferred-conflict-of-constraints constraint-system constraints
											 :method :LexQuicXplain)
;;;		  (LexQuicXplain constraint-system constraints)
		  :multiple-values nil
		  :test #'equal
		  :fail-info " LexQuicXplain Preferred Conflict = (A C)")
	
;;;	(multiple-value-bind (E-set C-set)
;;;		(daa-min-unsat constraint-system constraints)
;;;	  (format t "~% daa-min-unsat explanations = ~A, conflicts = ~A" 
;;;			  E-set C-set))
	
	(test `((((,D ,B ,A) (,C)) ((,D ,B ,C) (,A))) ((,A ,C)))
		  (all-explanations-and-conflicts-of-constraints
		   constraint-system constraints
		   :method :daa-min-unsat)
;;;		  (daa-min-unsat constraint-system constraints)
		  :multiple-values t
		  :test #'equal
		  :fail-info "daa-min-unsat explanations = (((D B A) (C)) ((D B C) (A))), conflicts = ((A C))")
	
;;;	(multiple-value-bind (E-set C-set)
;;;		(representativeXPlain constraint-system '() constraints)
;;;	  (format t "~% representativeXPlain explanations = ~A, conflicts = ~a" 
;;;			  E-set C-set))
	
	(test `((((,D ,B ,A) (,C)) ((,D ,B ,C) (,A))) ((,A ,C)))
		  (representative-explanations-of-constraints constraint-system
													  constraints
													  :method :representativeXPlain)
;;;		  (representativeXPlain constraint-system '() constraints)
		  :multiple-values t
		  :test #'equal
		  :fail-info "representativeXPlain explanations = (((D B A) (C)) ((D B C) (A))), conflicts = ((A C))")
	
;;;	(format t "~% Conflicts of representative explanations:")
;;;	(loop for (relaxation exclusion e-conflicts) 
;;;		in (Representative-Explanations-with-Explained-Exclusions constraint-system constraints)
;;;		do (format t "~% relaxation ~A, excludes ~a." 
;;;				   relaxation exclusion)
;;;		   (loop for (excluded conflict) in e-conflicts
;;;			   do (format t "~%   It excludes ~A because of ~a" 
;;;						  excluded conflict)))

	(test `((((,D ,B ,A) (,C)
						 ((,C (,A))))
			 ((,D ,B ,C) (,A)
						 ((,A (,C)))))
			((,A ,C)))
		  (Representative-Explanations-with-Explained-Exclusions constraint-system constraints)
		  :multiple-values T
		  :test #'equal
		  :fail-info "conflicts-of-representative explanations = ~
                      '(((D B A) (C) ((C (A)))) ((D B C) (A) ((A (C)))))")
	))
