(in-package #:ros-interface)

;;;; Socket Client using USOCKET

;;Test function
;;
;;1 - Connects to a socket server
;;2 - Sends a message
;;3 - Reads the response
;;4 - Closes the socket
(defun lisp-socket-client-test (port hostname)
  (let (client-socket)
	(setf client-socket (ros-connect-socket-server port hostname))
	(ros-send-to-server client-socket "Hi, MERS!")
	(ros-read-from-server client-socket)
	(ros-disconnect-socket-server client-socket)))

;;Connects to a given host and returns an socket object that should be stored for 
;;further communication
(defun ros-connect-socket-server (port hostname)
  (let ( (cli-socket nil) )
	(loop
	   ;;Tries to connect to the server
	   (setf cli-socket (usocket:socket-connect hostname port :protocol :stream :element-type 'character))
	   (return-from ros-connect-socket-server cli-socket))))


;;Closes the client socket (Syntatic sugar to make the library consistent)
(defun ros-disconnect-socket-server (cli-socket)
  (usocket:socket-close cli-socket))

;;Sends "message" to the server connected to the client socket "cli-socket"
;;
;;NOTE: write-line appends an extra new line character (~%) to the end of the string,
;;which makes it convenient to use along with read-line
;;(defun ros-send-to-server (cli-socket message)
;;	(write-line message (usocket:socket-stream cli-socket))
;;	(finish-output (usocket:socket-stream cli-socket)))

(defun ros-send-to-server (cli-socket message)
  (let* ((message-header (format nil "~4,'0X" (length message)))
		 (msg (format nil "~a~a" message-header message)))
	(format (usocket:socket-stream cli-socket) msg)
	(finish-output (usocket:socket-stream cli-socket))))

;; The following helper function reads a specified length of data from a stream. It
;; then returns that data. If, however, an EOF is encountered before finishing (i.e.,
;; perhaps the socket closed?), then returns nil.
(defun read-data (stream n-desired)
  (let* ((buff (make-array n-desired :element-type 'character :initial-element #\0))
		 (n-read (read-sequence buff stream :start 0 :end n-desired)))
	(if (< n-read n-desired)
		nil
		buff)))


;;Tries to read a message from the input buffer and automatically closes the socket
;;if the connection has been terminated by the server.
;;
;; Returns: * the message (a string), if it gets one (no longer returns a list if more than one)
;;          * nil, if no data to read and non-blocking
;;          * -1, if the socket connection was closed by the other side.
;;
(defun ros-read-from-server (cli-socket &key (blocking? nil))
  (let* ((message-header nil)
		 (message-len 0)
		 (message-data nil))
	(if (or blocking? (listen (usocket:socket-stream cli-socket)))
		(progn
		  ;; Attempt to read the header (which is a hexadecimal four bytes signifying
		  ;; the length of the data to follow.
		  (setf message-header (read-data (usocket:socket-stream cli-socket) 4))
		  (when (equalp message-header nil)
			(return-from ros-read-from-server -1))
		  ;; Parse the message header to extract the length of the message.
		  (setf message-len (parse-integer message-header :radix 16))
		  ;; Now read the rest of the message since we know it's length.
		  (setf message-data (read-data (usocket:socket-stream cli-socket) message-len))
		  (when (equalp message-data nil)
			(return-from ros-read-from-server -1))
		  ;; Return the data we just read!
		  message-data)
		;; If non-blocking and no data to read, just return immediately.
		(return-from ros-read-from-server nil))))


;; A nice macro that Eric defined - guarantees that the socket gets cleaned up
;; if an error occurs.
(defmacro with-ros-socket ((socket-name &rest open-args) &body body)
  `(let ((,socket-name (ros-connect-socket-server ,@open-args)))
    (unwind-protect (progn ,@body)
	  (ros-disconnect-socket-server ,socket-name))))