(in-package :state-space)

;;;Provides the tools to create and manipulate clauses.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Making clauses from specifications
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod get-universal-clause ((state-space state-space))
  "Simple wrapper to get the universal clause for a state-space."
  ;;Simply ask for the clause with no assignments
  (get-clause-from-assignments state-space nil))

(defmethod get-clause-from-descriptions ((state-space state-space) assignment-descriptions)
"Primary user method for getting clauses. Takes the descriptions and get a clause for it if 
the clause is valid. Sample: '((= A 1)(= B 2))"
(with-slots (assignment-ht) state-space
  (let ((assignment-list
		 (loop for description in assignment-descriptions
			when (null (gethash description assignment-ht))
			do (error "Bad description in clause creation.")
			end
			collect (gethash description assignment-ht))))
	(get-clause-from-assignments state-space assignment-list))))

(defmethod get-clause-from-assignments ((state-space state-space) assignment-list)
  "Primary method for getting a clause from the list of assignments. Returns the existing one or creates
a new one.

Return the object for the clause specified by a list of assignment objects within a state space.
Returns nil if the object is not found."

  (with-slots (clause-ht) state-space
	(let* ((sorted-assignments (sort-assignments assignment-list)) ;sort the list to match the ht
		   (existing-clause (gethash sorted-assignments clause-ht))) ;try the ht to see if exists already

	  (if (not (null existing-clause))	;if exists
		  existing-clause				;return it
		  (construct-clause-from-assignments state-space sorted-assignments))))) ; otherwise make a new one


(defmethod construct-clause-from-assignments ((state-space state-space) assignment-list)
"Construct a new clause object from the list of assignments. Return nil if it fails or if
the clause is inconsistent (if the state-space allows inconsistency)."

;;Make the instance
(let ((new-clause (make-instance (state-space-clause-type state-space) :state-space state-space)))

  ;;Unless all the assignments are added successfully, return nil
  (unless (and (loop for assignment in assignment-list
				  always (add-assignment-to-new-clause! new-clause assignment))
			   (state-clause-valid? new-clause))

	(return-from construct-clause-from-assignments nil))

  ;;The new clause is valid, so add it to the state-space
  (add-clause-to-state-space! state-space new-clause)

  ;;Return the new clause
  new-clause))

(defmethod add-assignment-to-new-clause! ((new-clause state-clause) (new-assignment state-assignment))
  "Adds an assignment to the clause. Returns t if successful and nil if there is already an
assignment to the variable of the assignment. No change is made if nil is returned.

NOT A USER FUNCTION! Must only be used on clauses not yet registered"
  ;;Get the current assignment to the variable this new-assignment is trying to assign
  (let ((curr-assignment (get-clause-assignment-to-var new-clause 
													   (state-assignment-variable new-assignment))))

	;;the current one must be the same or non-existent to add the new one, so:
	(if (and (not (null curr-assignment)) ;if both fail
			(not (eql curr-assignment new-assignment)))
		(return-from add-assignment-to-new-clause! nil) ;return failure
		(set-clause-assignment! new-clause new-assignment))) ;add the assignment

  ;;and return success
  t)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Clauses and the state-space
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod add-clause-to-state-space! ((state-space state-space) (clause state-clause))
  "Put the clause into the ht of all clauses."

  (let ((clause-key (state-clause-key clause))) ;get the hash key
	(setf (gethash clause-key (state-space-clause-ht state-space)) clause))) ;and put the clause in the ht

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Tests on clauses
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;(defun clause-valid-safe? (clause)
  ;;(unless (null clause)
;;	(let ((safe (state-clause-valid? clause))
;;		  (non-safe (clause-valid? clause)))
;;	  (assert (eql safe non-safe))
;;	  safe)))

(defun clause-valid? (clause)
  "Test whether the clause is valid. A function (not a method) so that a nil can be 
handled correctly. Note that an incomplete clause may be marked as valid even if there
is no possible completion."
  
  (when (not (null clause))
	(assert (typep clause 'state-clause))
	(state-clause-valid? clause)))

;;careful to use the fully qualified name
;;does not work! dunno why
(memoize-function 'state-space::clause-valid?)

(defmethod state-clause-valid? ((clause state-clause))
  "Without conflicts, any clause created is true."
  t)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Unary operations on clauses
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod clause-list-of-assignments ((clause state-clause))
  "Return the list version of the assignments of this clause."
  (loop for assignment being the hash-value of (state-clause-assignment-ht clause)
	   collect assignment))

(defmethod get-clause-assignment-to-var ((clause state-clause) (var state-variable))
  "Return the assignment to the input variable of the clause. Return false if none."
  (gethash var (state-clause-assignment-ht clause)))

(defmethod get-clause-assigned-vars ((clause state-clause))
  "Return a list of the variables assigned by the clause."
  (loop for variable being the hash-keys of (state-clause-assignment-ht clause)
	 collect variable))

(defmethod set-clause-assignment! ((clause state-clause) (assignment state-assignment))
"Puts the assignment in the clause. NOT a user function. Assumes the error checking has
already been done and that the clause hasn't been registered yet."
(setf (gethash (state-assignment-variable assignment)
			   (state-clause-assignment-ht clause)) assignment))

(defmethod clause-universal? ((clause state-clause))
  (eql clause (get-universal-clause (state-clause-state-space clause))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Binary operations on clauses
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod clause-subsumed-method? ((subsumed-clause state-clause) (subsuming-clause state-clause))
  "Test if the subsumed-clause is subsumed by the subsuming-clause. This is the case if all variables
assigned in the subsuming-clause have the same assignments in the subsumed-clause. Return t/nil."

;;All subsuming assignments must be assigned in the subsumed clause.
(loop for subsuming-assignment in (clause-list-of-assignments subsuming-clause)
	 always (eql subsuming-assignment 
				 (get-clause-assignment-to-var subsumed-clause 
											   (state-assignment-variable subsuming-assignment)))))

(defun clause-subsumed? (x y)
"A function to allow memoized calls."
  (clause-subsumed-method? x y))
;;;Memoize the subsumption function. Use the whole input list with equalp.

(memoize-function 'state-space::clause-subsumed? :key #'identity :test #'equal)

(defmethod clauses-contradictory? ((clause1 state-clause) (clause2 state-clause))
  "Clauses are contradictory if they have differing assignments to any variable. The clauses may differ
in which variables are assigned without being contradictory. Invalid clauses are contradictory with everything."
  
  ;;The clauses are not contradictory if they have a non-empty intersection.
  (null (clause-intersection clause1 clause2)))

(defmethod clause-intersection-method ((clause1 state-clause) (clause2 state-clause))
  "Return the clause that is the intersection of the two inputs. Returns nil if
the result is empty or invalid. The intersection is the concatenation of the state
assignments of both clauses."

  ;;Simply append the lists of assignments, removing duplicates in case they overlap.
  ;;get-clause-from-assignments provides the needed error checking for inconsistencies.

  (get-clause-from-assignments (state-clause-state-space clause1)
							   (remove-duplicates
								(append (clause-list-of-assignments clause1)
										(clause-list-of-assignments clause2)))))

(defun clause-intersection-no-valid-checking (x y)
  "A function to allow memoized calls."
  (clause-intersection-method x y))

(memoize-function 'state-space::clause-intersection-no-valid-checking :key #'identity :test #'equal)

(defun clause-intersection (x y)
  "User function for intersections."
  (unless (or (null x) (null y))
	(let ((intersect (clause-intersection-no-valid-checking x y)))
	  (when (clause-valid? intersect)
		intersect))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Groups of clauses
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod remove-subsumed-clauses ((clause state-clause) clause-list)
  "Return a new list with clause-list pruned of all clauses subsumed by the new clause."
  (remove clause clause-list 
		  :test #'(lambda (clause list-member) 
					(clause-subsumed? list-member clause))))

(defun clause-list-intersection (list1 list2)
  "Return a list with the intersection of every pair from the two lists."

  (delete nil ;remove any nils
		  ;;loop through both lists and intersect
		  (loop for clause1 in list1
			   append (loop for clause2 in list2
						   collect (clause-intersection clause1 clause2)))))

(defun make-list-non-subsuming (clause-list)
  "Returns the clause-list, removed of entries so that no entry subsumes another."
  
  (loop for potential-clause in clause-list 
	 ;;someplace to put the final list
	 with new-list = nil
	 ;;when there is nothing in the new-list yet
	 when (or (null new-list)			  ;;or the potential-clause isn't subsumed by an existing one
			  (loop for accepted-clause in new-list
				 never (clause-subsumed? potential-clause accepted-clause)))
	 ;;and remove any kernels that this subsumes (the potential-clause is now accepted)
	 do (setf new-list (remove-subsumed-clauses potential-clause new-list))
	 ;;save it into the new-list (must come after the subsumption test!)
	  (push potential-clause new-list)
	 

	 finally (return new-list)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Utilities
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod state-clause-key ((clause state-clause))
  "The hash-key of a clause is the assignments in a list, sorted by their variable indices."
  (sort-assignments (loop for assignment being the hash-value of (state-clause-assignment-ht clause)
					   collect assignment)))

(defun sort-assignments (assignment-list)
  "Helper function. Sorts a list of assignments by the indices of their variables."
  (sort assignment-list '<	;lowest first
		:key #'(lambda (assignment) ;sort by the index of the assignment's var
			   (state-variable-index 
				(state-assignment-variable assignment)))))


(defmethod state-clause-length ((clause state-clause))
  (let ((total-length 0))
	(loop for assignment being the hash-keys of (state-clause-assignment-ht clause)
		 do (list assignment) ;don't do anything, ignore the warning
		 do (setf total-length (+ 1 total-length)))
	total-length))