(in-package :state-space)

;;;Adds an extension to state-spaces to allow conflicts and consistency checking.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State space with conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass state-space-w-conflicts (state-space)

  ((conflicts
	:initform nil
	:accessor state-space-conflicts
	:documentation "List of conflict clauses.")

   (clause-type
	:allocation :class
	:initform 'state-clause-w-conflicts
	:accessor state-space-clause-type
	:documentation "Class type of the clauses. Indicate that the clause
is part of a state-space with conflicts.")

   (kernels
	:initform :nil
	:accessor state-space-kernels
	:documentation "A list of kernels for the state-space (of all the known conflicts)."))

  (:documentation "Extends state-spaces to allow for conflict clauses and provides
support for checking consistency against those conflicts. Also maintains a set of 
complete kernels for the state space. This may be very slow for large problems with
many conflicts/vars, but should make testing whether adding a new conflict empties the 
state-space of solutions."))

(defmethod initialize-instance :after ((state-space-w-conflicts state-space-w-conflicts) 
									   &key)
  "Initialize the kernels to be the universal set, since there are no known conflicts yet."
  (with-slots (kernels) state-space-w-conflicts
	(setf kernels (list (get-universal-clause state-space-w-conflicts)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State clause with conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass state-clause-w-conflicts (state-clause)

  ()

  (:documentation "Differentiated to allow methods to differentiate between
clauses created for state-spaces with or without conflicts."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Adding conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod add-conflict! ((state-space state-space-w-conflicts) (new-conflict state-clause))
  "User method to add a new conflict. Avoids redundant conflicts by making sure that the 
new one not subsumed by any known conflicts."
 
  ;;clear the memoization of valid! be careful to use a fully qualified name!
 
   (org.tfeb.hax.memoize:unmemoize-function 'state-space::clause-valid?)
  (debug-print-for :state-space t "Adding conflict: ~a" new-conflict)
  
  (with-slots (conflicts) state-space
	;;when the new clause is never subsumed by an existing conflict
	(when (loop for existing-conflict in conflicts
			 never (clause-subsumed? new-conflict existing-conflict))
	  
	  ;;prune all subsumed conflicts from the active set by using remove and a 
	  ;;lambda mapping to a subsumption test
	  (setf conflicts (remove-subsumed-clauses new-conflict conflicts))
	  
	  ;;insert the conflict (make sure to prune before adding the new one!)
	  (insert-conflict! state-space new-conflict)

	  (debug-print-for :state-space t "New conflict added, new kernels: ~a" (state-space-kernels state-space))))

  (org.tfeb.hax.memoize:memoize-function 'state-space::clause-valid?)
  )

(defmethod insert-conflict! ((state-space state-space-w-conflicts) (new-conflict state-clause))
  "Non-user method to insert the conflict into the state-space object. Assumed to be
non-redundant. Conflicts are stored in a simple list."

  (push new-conflict (state-space-conflicts state-space))
  (update-kernels-with-conflict! state-space new-conflict))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Clauses and conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod state-space-invalid? ((state-space state-space-w-conflicts))
  "Returns whether the state-space is currently invalid, i.e. it has no kernels."
  (null (state-space-kernels state-space)))

(defmethod get-valid-full-assignment ((state-space state-space-w-conflicts))
  "Return an arbitrary extension of one of the kernels, which is guaranteed to be valid."

  ;;pick a kernel to extend
  (let ((working-clause (first (state-space-kernels state-space))))
	(assert (not (null working-clause))) ;;can't give one if there's no kernel
	(loop for variable in (state-space-variables state-space)
		 ;;when a variable is not assigned
		 when (null (get-clause-assignment-to-var working-clause variable))
		 ;;assign it arbitrarily to the first in the domain
		 do (setf working-clause 
				 (clause-intersection working-clause 
									  (get-clause-from-assignments state-space 
																   (list (first (state-variable-domain variable))) ))))
	working-clause))

(defmethod get-all-consistent-clauses ((state-space state-space-w-conflicts))
  "Get every consistent clause of a state-space, in no particular order."

  ;;start with the kernels
  (let ((working-list (state-space-kernels state-space)))
	
	;;then extend with every possible assignment
	(loop for variable in (state-space-variables state-space)
	   with variable-assignments
	   ;;get the list of clauses for each variable's assignments
	   do (setf variable-assignments
				(loop for assignment in (state-variable-domain variable)
				   collect (get-clause-from-assignments state-space (list assignment))))
	   ;;intersect the lists, throwing away any nils from invalid clauses
	   do (setf working-list (remove nil (clause-list-intersection working-list variable-assignments))))

	;;remove the dupes and return
		
	(remove-duplicates working-list)))

(defmethod state-clause-valid? ((clause state-clause-w-conflicts))
  "Check if the state-clause is subsumed by any of the conflicts. 
idea."
;;(format t "deriving ~a~%" clause)
  (with-slots (state-space) clause
	;;True if clause is never subsumed by a conflict
	;;the unless is there to catch a universal clause when there are no kernels,
	;;there won't be a conflict that subsumes it, but it's definitely bad
	(unless (null (state-space-kernels (state-clause-state-space clause))) 
	  (loop for conflict in (state-space-conflicts state-space)
		 never (clause-subsumed? clause conflict)))))

(defmethod clause-avoids-all-conflicts? ((clause state-clause-w-conflicts))
  "Check if the state-clause avoids all possible conflicts if it is subsumed 
by a kernel, since they are kernels of the entire state-space. If true, any assignment
may be given to unassigned variables in the clause. Naturally, new conflicts may
invalidate this result."
  (with-slots (state-space) clause
	"True if clause is never subsumed by a conflict"
	(loop for kernel in (state-space-kernels state-space)
	   thereis (clause-subsumed? clause kernel))))

(defmethod conflicts-to-commit-to-clause ((clause state-clause-w-conflicts))
"Return the conflicts necessary to commit to the input clause. Returns nil
if it the clause is already invalid."
(if (state-clause-valid? clause)
	(loop for conflict in (kernels-of-conflict clause)
	   collect conflict)
	nil)
)

(defmethod commit-to-clause! ((clause state-clause-w-conflicts))
"If the clause is valid, commit to the clause. We can do this by inserting conflicts
on all other possiblities for the assignments. This is the same as generating the kernels
of the clause and making them all conflicts. If the clause is valid, then this process cannot
remove all assignments."
  (if (state-clause-valid? clause)
	  (progn (loop for conflict in (kernels-of-conflict clause)
				do (add-conflict! (state-clause-state-space conflict) conflict)
				  ;(format t "commit to ~a" conflict)
)
		t)
	  nil))

(defmethod commit-to-clause-allowed? ((clause state-clause-w-conflicts))
  "Tests whether we can commit to the input clause without removing all kernels."
  (not (would-conflicts-remove-all-kernels? (loop for conflict in (kernels-of-conflict clause)
											   collect conflict))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Kernels
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun would-conflicts-remove-all-kernels? (new-conflicts)
  "Test a list of potential new-conflicts (or a single one). Determine whether the new conflict would invalidate all 
the current kernels. If true, this new conflict would remove all solutions to the state-space."
  ;;if no conflicts, it would not remove all kernels
  (when (null new-conflicts)
	(return-from would-conflicts-remove-all-kernels? nil))
  ;;make a singleton into a list
  (when (not (listp new-conflicts))
	(setf new-conflicts (list new-conflicts)))

  (assert (not (null (state-space-kernels (state-clause-state-space (first new-conflicts)))))
		  ((state-clause-state-space (first new-conflicts)))
		  "If this function is called, there should be a kernel.")

  ;;Attempt generating the new set of kernels. Make sure it's not null.
  (let ((new-kernels (state-space-kernels (state-clause-state-space (first new-conflicts)))))
	(loop for conflict in new-conflicts
		 do (setf new-kernels (update-input-kernels-w-conflict conflict new-kernels)))
	(null new-kernels)))


(defmethod update-kernels-with-conflict! ((state-space state-space-w-conflicts) 
										 (new-conflict state-clause))
  "Update the known global kernels to reflect the new conflicts."

 
  (with-slots (kernels) state-space
	;;uses another function to do the work, just saves the result to keep it
	(setf kernels (update-input-kernels-w-conflict new-conflict kernels) ;set the new kernels
		  )))						 ;and the kernels of this conflict

(defmethod update-input-kernels-w-conflict ((new-conflict state-clause) kernels)
  "Calculates the change to the input list of kernels if a conflict is added."

 ;;The update requires intersecting the kernels of the new-conflict with all
  ;;the existing kernels. Those intersections are the new kernels. The new kernels
  ;;can also be checked for subsumption among them. 
  
  ;;Generate all the possible kernels of this conflict and intersect them all
  ;;against all the old kernels. Then build up a new minimal set of kernels by making sure
  ;;that no member of the set subsumes another.
  ;;NB: The following approach checks the validity of the kernels against the old ones, maybe
  ;;that step can be skipped?
  ;;for each potential kernel is the cross product
  
  (make-list-non-subsuming ;to the non-subsuming version of the list
   (clause-list-intersection ;defined by the powerset of intersections
	kernels ;of the existing kernels
	(kernels-of-conflict new-conflict))))

(defmethod kernels-of-conflict ((conflict state-clause))
  "Returns a list of clauses that are possible resolutions of this conflict. These
are the clauses defined by a single assignment that contradicts an assignment in the conflict,
i.e. an assignment to one variable that has an assignment in the conflict, but which is
different than the assignment used in the conflict."
  
  ;;if the conflict is the universal clause, there are no kernels
  (when (eql conflict (get-universal-clause (state-clause-state-space conflict)))
	(return-from kernels-of-conflict nil))

  ;;for every variable in the conflict
  (loop for assigned-var in (get-clause-assigned-vars conflict)
	 with assigned-value
	 ;;stash the value used in the conflict
	 do (setf assigned-value (get-clause-assignment-to-var conflict assigned-var))
	 ;;append the clauses from this variable into the result
	 append (loop for value in (state-variable-domain assigned-var) ;every value of this var
			   unless (eql value assigned-value) ;unless it's the one used
			   ;;collect the clause associated with just that assignment (all that's needed
			   ;;to resolve the conflict)
			   collect (get-clause-from-assignments (state-clause-state-space conflict)
													(list value)) 
			   end)))

(defmethod split-clause-around-clauses ((initial-clause state-clause) clauses-to-avoid)

  (let ((old-working-clauses (list initial-clause)))

	(loop for clause-to-avoid in clauses-to-avoid
		 with new-working-clauses
		 do (setf new-working-clauses nil)
		 do (loop for clause-avoidance in (kernels-of-conflict clause-to-avoid)
				 do (loop for working-clause in old-working-clauses
						 do (let ((possible-new-clause 
								   (clause-intersection 
									working-clause 
									clause-avoidance)))
							  (when (clause-valid? possible-new-clause)
								(pushnew possible-new-clause new-working-clauses)))))
		 do (setf old-working-clauses new-working-clauses))

	old-working-clauses))