(in-package :state-space)

;;;Holds the definitions of the basic classes of the state-space package.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Space
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass state-space ()
  
  ((name
	:initarg :name
	:initform nil;(gentemp "STATE-SPACE-")
	:accessor state-space-name		
	:documentation "The name of a state space.")
   
   (source
	:initarg :source
	:initform nil
	:accessor state-space-source
	:documentation "Object that created state space.")

   (variable-type
	:allocation :class
	:reader state-space-variable-type
	:initform 'state-variable
	:documentation "Class type of variables of state space.")

   (variables
	:initform nil
	:accessor state-space-variables
	:documentation "The variables of the state space.")
										
   (assignment-ht
	:initform (make-hash-table :test #'equal)
	:accessor state-space-assignment-ht
	:documentation "A hash table mapping assignment ~
                    expressions to assignment objects.")

   (clause-type
	:allocation :class
	:initform 'state-clause
	:accessor state-space-clause-type
	:documentation "Class type of the clauses.")

   (clause-ht
	:initform (make-hash-table :test #'equal)
	:accessor state-space-clause-ht
	:documentation "A ht of all the clauses used within the state space. 
Hashed from sorted assignment lists to clause objects."))
  
  (:documentation   "A state space, defined as a set of variables,
                     their corresponding domains."))

(defmethod print-object 

	((state-space state-space) stream)

  (with-slots (name) state-space
	(format stream "<SS: ~a>" name)))

(defmacro get-assignment-from-hash 

	(assignment assignment-ht)

  "Returns the bsf-assignment object of ASSIGNMENT-HT 
   corresponding to ASSIGNMENT expression.  
   ASSIGNMENT-HT is an assignment hash table.
   ASSIGNMENT is an expression (= variable value)."
  
  `(gethash ,assignment ,assignment-ht))

(defmacro get-state-assignment 

	(state-assignment state-space)

  "Returns the bsf-assignment object of STATE-SPACE
   corresponding to ASSIGNMENT.  
   ASSIGNMENT is an expression (= variable-name value)."

  
  `(get-assignment-from-hash
    ,state-assignment (state-space-assignment-ht
					   ,state-space)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Variable
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass state-variable ()
  
  ((name 
	:initarg :name
	:accessor state-variable-name
	:documentation "Name of the decision variable.")

   (domain
	:initform nil
	:accessor state-variable-domain
	:documentation "Possible assignments to variable.")

   (assignment-type
	:allocation :class
	:reader state-variable-assignment-type
	:initform 'state-assignment
	:documentation "Class type of assignments to variable.")
   
   (index
	:accessor state-variable-index
	:documentation "Allows for sorting."))

  (:documentation "A variable of a state space with ~
                   associated domain."))

(defmethod print-object 

	((variable state-variable) stream)
  
  "Prints short descriptor for a variable."
  
  (with-slots (name) variable
	(format stream "~a" name)))

(defmethod initialize-instance :after 
		   ((variable state-variable) 
			&key state-space domain)
  
  "Creates a new variable for STATE-SPACE with the 
   specified name.  DOMAIN is a list of domain 
   element descriptions.
   Each DOMAIN element description is either an atom, 
   denoting an element, or list of the form:

   (ELEMENT . <keyword-value-pair>*)
  
   where ELEMENT is the value of an assignment, and the 
   keyword-value-pairs correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of VARIABLE."

  (let ((intern-domain (make-variable-domain
						state-space variable
						domain)))
    
	;; Setup the assignments in variable's domain.
    (state-variable-install-domain! variable 
									intern-domain)

	;;Give the new variable an index. If there aren't any yet, start with 1. Otherwise,
	;;use 1+ previous-index
	(with-slots (index) variable
	  (if (null (state-space-variables state-space))
		  (setf index 1)
		  (setf index (+ 1 (state-variable-index (first (state-space-variables state-space)))))))

	;; Add VARIABLE to the variables of the search problem
    (push variable
		  (state-space-variables state-space))

	T))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Assignment Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass state-assignment ()

  ((variable
	:initarg :variable
	:accessor state-assignment-variable
	:documentation "The variable of this assignment.")

   (value
	:initarg :value
	:accessor state-assignment-value
	:documentation "The value of this assignment."))
  
  (:documentation "A variable/value assignment."))

(defmethod print-object 

	((state-assignment state-assignment) stream)
  
  "Prints short descriptor for a state variable assignment."
  
  (with-slots (variable value) state-assignment
	(format stream "~a=~a" variable value)))

(defmacro state-assignment-description
	
	(state-assignment)
  
  "Given a STATE-ASSIGNMENT object,
   Returns an external description of assignment, in the form:
     (= variable-name value)."
  
  `(make-assignment
	:variable-name (state-variable-name
					(state-assignment-variable ,state-assignment))
	:value (state-assignment-value ,state-assignment)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; State Clause
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass state-clause ()
  
  ((assignment-ht 
	:initform (make-hash-table :test 'eql)
	:accessor state-clause-assignment-ht
	:documentation "The map between the variables and the assignments.")

   (state-space
	:initarg :state-space
	:accessor state-clause-state-space
	:documentation "Holds the reference back to the state space this clause
is part of."))

  (:documentation "A clause - a partial set of assignments."))


(defmethod print-object ((state-clause state-clause) stream)
  "Prints short descriptor for a state variable assignment."
  (with-slots (assignment-ht valid?) state-clause
	(let ((all-assignments
		   (loop for one-assignment being the hash-values of assignment-ht
			  collecting one-assignment)))
	  (if (null all-assignments)
		  (format stream "\{\}")
		  (format stream "\{~{~a~^, ~}\}" all-assignments)))))
