(in-package :state-space)

;;;Hold the functions for basic manipulations of a state space, creating variables,
;;;initializing, and such.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Initialize State Space
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((state-space state-space) 
									   &key variable-descriptions)
  "The initialization - takes a description of the vars and creates the variables
and their domains."
  (add-variables-from-descriptions! state-space variable-descriptions)
  )


(defmethod add-variables-from-descriptions! ((state-space state-space)  variable-descriptions)
  "User callable function to add variables to a state space. Called
by the initialize method."
  (loop for (variable-name . variable-keyword-list) 
	 in variable-descriptions
	 do (add-variable! 
		 state-space variable-name variable-keyword-list))
  )

(defmethod add-variable! 

	((state-space state-space)
	 variable-name variable-keyword-plist)

  "Creates a new variable for STATE-SPACE with the 
   specified name.  Variable will be of the type specified
   in the VARIABLE-TYPE slot of STATE-SPACE.

   VARIABLE-KEYWORD-LIST is a property list of keyword value pairs
    supplied to make-instance of variable.  It currently consists of
    one required keyword :DOMAIN and one optional keyword :GUARD:

      (:DOMAIN (DOMAIN-ELEMENT-DESCRIPTION+) 
       [:GUARD (= GVARIABLE-NAME GVALUE) | :GUARD *always-active*])

    where each DOMAIN-ELEMENT-DESCRIPTION is either an atom, denoting 
    a domain element, or list of the form:

      (DOMAIN-ELEMENT <KEYWORD-VALUE-PAIR>*)
  
   where DOMAIN-ELEMENT is the value of the assignment, and  
   KEYWORD-VALUE-PAIRS correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of the VARIABLE class.
   This currently consists of :COST.

   :GUARD is optional, and denotes an assignment that
   activates variable.

   Note: If variable is of type decision-variable, 
   then this function ensures that variable's domain is placed in best 
   first order."

  (apply 'make-instance (state-space-variable-type state-space)
		 :name variable-name
		 :state-space state-space
		 variable-keyword-plist)
  )



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Installing Decision Variable Domains
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod make-variable-domain
    ((state-space state-space) 
	 (variable state-variable)
	 value-descriptions)

  "Takes as input a STATE VARIABLE of STATE-SPACE, whose
   (valued) domain is described by VALUE-DESCRIPTIONS.
   VALUE-DESCRIPTIONS is a list of elements, each of
   which describes one value.  

   Each VALUE-DESCRIPTION is either
   an atom, denoting a value, or list of the form:
   (VALUE . <keyword-value-pair>*)  where VALUE is the
   value of the assignment, and the keyword-value-pairs
   correspond to the remaining keyword arguments to 
   make-instance for ASSIGNMENT-TYPE of VARIABLE.

   RETURNS a new list of assignment objects for VARIABLE, 
   that corresponds to VALUE-DESCRIPTIONS.

   Note: This list may be destructively modified; 
   for example, by sort."
      
  ;; create the assignments and add them to the assignment 
  ;; lookup hashtable.
  (loop 
      with ht = (state-space-assignment-ht state-space)
      with variable-name = (state-variable-name variable)
      for value-description in value-descriptions
      for assignment = (make-variable-assignment
						variable value-description)
	  for value = (state-assignment-value assignment)
      do (setf (get-assignment-from-hash
				`(= ,variable-name ,value) ht)
		   assignment)
      collecting assignment))

(defmethod make-variable-assignment 
	((variable state-variable)
	 value-description)
  
  "Given VARIABLE and VALUE-DESCRIPTION, returns
   an object denoting an assignment of variable to value.

   VALUE-DESCRIPTION is either an atom, denoting a value, 
   or list of the form:

   (VALUE . <keyword-value-pair>*)
  
   where VALUE is the value of the assignment, and the 
   keyword-value-pairs correspond to the remaining keyword 
   arguments to make-instance for ASSIGNMENT-TYPE of VARIABLE."

  (let ((assignment-type
		 (state-variable-assignment-type variable)))
	
	(if (atom value-description)

		(make-instance assignment-type
		  :variable variable
		  :value value-description)
	  
	  (destructuring-bind (value . keyword-value-pairs)
		  value-description
;;;		(format t "~% value = ~a, keyword pairs = ~a"
;;;				value keyword-value-pairs)
		(apply 'make-instance
		 `(,assignment-type
		   :variable ,variable
		   :value ,value
		   . ,keyword-value-pairs)))
	  )))

(defmethod state-variable-install-domain! 
	((variable state-variable) domain)
  
  " Connects a previously defined VARIABLE object to DOMAIN,
    a previously defined list of assignments object."

  ;; Point each assignment to its variable, and
  ;; clear any previous alternate assignments.
  (loop for assignment in domain
      do (setf (state-assignment-variable assignment) 
		   variable))

  ;; Point variable to its new domain.
  (setf (state-variable-domain variable) domain)
  T)

