(in-package :state-space)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; test-state-spaces
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftest (test-state-spaces :state-space) ()
  "Try making a state space with a couple of variables. Make sure the 
right number of variables and assignments get created."
  (let* ((var-descriptions
		 '((a :domain ((1) (2) (3)))
		   (b :domain ((1) (2)))))
		(state-space (make-instance 'state-space :variable-descriptions var-descriptions)))

	(with-slots (variables assignment-ht) state-space
	  (test 5 (hash-table-count assignment-ht)
			:fail-info "Wrong number of assignments.")
	  (test 2 (length variables))
	  (test 2 (state-variable-index (first (state-space-variables state-space))))))
  
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; test-clause-creation
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftest (test-clause-creation :state-space) ()
"Try making some clauses. Make sure that the unique objects are created correctly. Also make sure
universal clauses work and that improper clauses are rejected correctly."
(let* ((var-descriptions
		'((a :domain ((1) (2) (3)))
		  (b :domain ((1) (2)))
		  (c :domain ((1) (2) (3) (4)))))
	   ;;create the state space
	   (state-space (make-instance 'state-space :variable-descriptions var-descriptions))
	   
	   ;;allocate some clauses
	   (universal1 (get-clause-from-descriptions state-space nil)) ;two copies of the universal
	   (universal2 (get-clause-from-descriptions state-space nil))
	   (testa-1 (get-clause-from-descriptions state-space '((= a 1) (= b 2)))) ;two copies of an arbitrary clause
	   (testa-2 (get-clause-from-descriptions state-space '((= b 2) (= a 1))))
	   (testb (get-clause-from-descriptions state-space '((= a 1) (= a 2))))) ;an invalid clause - two assignments to a
  
  ;;throwaway valid clause
  (get-clause-from-descriptions state-space '((= a 1) (= b 2) (= c 4)))

  ;;bad description, tests for an error
  (test-error (get-clause-from-descriptions state-space '((= a 1) (= b 2) (= c 400)))
						:fail-info "Fails to catch bad assignment")

  (test t (not (null universal1)) :fail-info "Universal clause was nil.")
  (test universal1 universal2 :fail-info "Universal clauses weren't the same.")
  (test testa-1 testa-2 :fail-info "Created two version of the same clause.")
  (test nil testb :fail-info "Allocated a clause for an invalid clause.")
  (test 3 (hash-table-count (state-space-clause-ht state-space))
		:fail-info "Should only have three clauses.")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; test-intersections
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftest (test-intersections :state-space) ()
  "Test several types of intersections. Subtests:
a: {A=1}int{B=1}={A=1,B=1} ;simple intersection (making the result first)
b: {A=1}int{nil}={A=1}     ;simple intersection with a universal clause
c: {A=1}int{A=2}=nil       ;inconsistency checking
d: {A=1,B=1}int{B=1,C=2}={A=1,B=1,C=2} ;int with overlap

Clauses below are defined with the letter of the test they first appear in"
  (let* ((var-descriptions
		  '((a :domain ((1) (2) (3)))
			(b :domain ((1) (2)))
			(c :domain ((1) (2) (3) (4)))))
		 ;;create the state space
		 (state-space (make-instance 'state-space :variable-descriptions var-descriptions))
		 (universal (get-clause-from-descriptions state-space nil))
		 (clause-a0 (get-clause-from-descriptions state-space '((= a 1) (= b 1))))
		 (clause-a1 (get-clause-from-descriptions state-space '((= a 1))))
		 (clause-a2 (get-clause-from-descriptions state-space '((= b 1))))
		 (clause-c (get-clause-from-descriptions state-space '((= a 2))))
		 (clause-d1 (get-clause-from-descriptions state-space '((= b 1) (= c 2))))
		 (clause-d2 (get-clause-from-descriptions state-space '((= a 1) (= b 1) (= c 2)))))

	(test clause-a0 (clause-intersection clause-a1 clause-a2) :fail-info "Failed:  {A=1}int{B=1}={A=1,B=1}")
	(test clause-a1 (clause-intersection universal clause-a1) :fail-info "Failed:  {A=1}int{nil}={A=1} ")
	(test nil (clause-intersection clause-a1 clause-c) :fail-info "Failed:  {A=1}int{A=2}=nil")
	(test clause-d2 (clause-intersection clause-a0 clause-d1) :fail-info "Failed:  {A=1,B=1}int{B=1,C=2}={A=1,B=1,C=2}")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; test-subsumed-contradictory
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftest (test-subsumed-contradictory :state-space) ()
  "Test the subsumed and contradictory functions. Tests:
a:{A=1}subsumed-by{A=1,B=1} = nil
b:{A=1,B=1}subsumed-by{A=1} = t
c:{B=1}subsumed-by{A=1} = nil
d:{A=1}subsumed-by{nil} = t  (note that {nil} is the universal clause, everything is subsumed by it)
e:{nil}subsumed-by{A=1} = nil
f:{A=1}subsumed-by{A=1} = t
g:{A=1}subsumed-by{A=2} = nil
h:{nil}subsumed-by{nil} = t  

i: {A=1}contradicts{A=1,B=1} = nil
j: {A=1}contradicts{B=1} = nil
k: {A=1}contradicts{A=2} = t
l: {A=1,B=1}contradicts{B=2,C=2} = t
m: {A=1}contradicts{nil} = nil"

  (let* ((var-descriptions
		  '((a :domain ((1) (2) (3)))
			(b :domain ((1) (2)))
			(c :domain ((1) (2) (3) (4)))))
		 ;;create the state space
		 (state-space (make-instance 'state-space :variable-descriptions var-descriptions))
		 (universal (get-clause-from-descriptions state-space nil))
		 (clause-a1 (get-clause-from-descriptions state-space '((= a 1))))
		 (clause-a2 (get-clause-from-descriptions state-space '((= a 2))))
		 (clause-b1 (get-clause-from-descriptions state-space '((= b 1))))
		 (clause-a1b1 (get-clause-from-descriptions state-space '((= a 1) (= b 1))))
		 (clause-b2c2 (get-clause-from-descriptions state-space '((= b 2) (= c 2)))))
  
	(test nil (clause-subsumed? clause-a1 clause-a1b1) :fail-info "a:{A=1}subsumed-by{A=1,B=1} = nil")
	(test t (clause-subsumed? clause-a1b1 clause-a1) :fail-info "b:{A=1,B=1}subsumed-by{A=1} = t")
	(test nil (clause-subsumed? clause-b1 clause-a1) :fail-info "c:{B=1}subsumed-by{A=1} = nil")
	(test t (clause-subsumed? clause-a1 universal) :fail-info "d:{A=1}subsumed-by{nil} = t  (note that {nil} is the universal clause, everything is subsumed by it)")
	(test nil (clause-subsumed? universal clause-a1) :fail-info "e:{nil}subsumed-by{A=1} = nil")
	(test t (clause-subsumed? clause-a1 clause-a1) :fail-info "f:{A=1}subsumed-by{A=1} = t")
	(test nil (clause-subsumed? clause-a1 clause-a2) :fail-info "g:{A=1}subsumed-by{A=2} = nil")
	(test t (clause-subsumed? universal universal) :fail-info "h:{nil}subsumed-by{nil} = t  ")

	(test nil (clauses-contradictory? clause-a1 clause-a1b1) :fail-info "i: {A=1}contradicts{A=1,B=1} = nil")
	(test nil (clauses-contradictory? clause-a1 clause-b1) :fail-info "j: {A=1}contradicts{B=1} = nil")
	(test t (clauses-contradictory? clause-a1 clause-a2) :fail-info "k: {A=1}contradicts{A=2} = t")
	(test t (clauses-contradictory? clause-a1b1 clause-b2c2) :fail-info "l: {A=1,B=1}contradicts{B=2,C=2} = t")
	(test nil (clauses-contradictory? clause-a1 universal) :fail-info "m: {A=1}contradicts{nil} = nil")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; test-conflicts
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftest (test-conflicts :state-space) ()
  "Test the conflict support. Create a state-space, throw in some conflicts, try
the basic functions, then make the space completely infeasible and try again."
  (let*  ((var-descriptions
		   '((a :domain ((1) (2) (3)))
			 (b :domain ((1) (2)))
			 (c :domain ((1) (2) (3) (4)))))
		  ;;create the state space
		  (state-space (make-instance 'state-space-w-conflicts :variable-descriptions var-descriptions))
		  (universal (get-universal-clause state-space))
		  (clause-a1 (get-clause-from-descriptions state-space '((= a 1))))
		  (clause-a2 (get-clause-from-descriptions state-space '((= a 2))))
		  (clause-a3 (get-clause-from-descriptions state-space '((= a 3))))
		  (clause-b1 (get-clause-from-descriptions state-space '((= b 1))))
		  (clause-b2 (get-clause-from-descriptions state-space '((= b 2))))
		  (clause-a1b1 (get-clause-from-descriptions state-space '((= a 1) (= b 1))))
		  (clause-b1c2 (get-clause-from-descriptions state-space '((= b 1) (= c 2))))
		  (clause-a3b1 (get-clause-from-descriptions state-space '((= a 3) (= b 1))))
		  (clause-a3b2 (get-clause-from-descriptions state-space '((= a 3) (= b 2))))
		  (clause-a2b1 (get-clause-from-descriptions state-space '((= a 2) (= b 1))))
		  (clause-a3c1 (get-clause-from-descriptions state-space '((= a 3) (= c 1))))
		  (clause-a2b1c3 (get-clause-from-descriptions state-space '((= a 2) (= b 1) (= c 3)))))

	;;create clauses for one of the compound ones
	(test (list clause-b2 clause-a2 clause-a3) (kernels-of-conflict clause-a1b1)
		  :test #'lists-eql-out-of-order
		  :fail-info "If the clauses are out of order, nothing has gone wrong.")

	(test t (clause-avoids-all-conflicts? universal)
		  :fail-info "Universal clause should pass with no conflicts yet.")

	(test t (would-conflicts-remove-all-kernels? (list universal))
		  :fail-info "The universal clause should remove all kernels.")

	;;add conflicts
	(add-conflict! state-space clause-a1)
	(add-conflict! state-space clause-b2)

	;;test the conflicts
	(test (list clause-b2 clause-a1) (state-space-conflicts state-space)
		  :test #'lists-eql-out-of-order
		  :fail-info "Conflicts should be a1 and b2.")

	;;test the kernels
	(test (list clause-a2b1 clause-a3b1) (state-space-kernels state-space)
		  :test #'lists-eql-out-of-order
		  :fail-info "Kernels should be {a=2,b=1}, {a=3, b=1}")
	
	;;test valid checking
	(test t (clause-valid? clause-b1c2)
		  :fail-info "Clause {b=1,c=2} should be a kernel, should be valid.")
	(test nil (clause-valid? clause-a3b2) 
		  :fail-info "Clause {a=3,b=2} should be invalid, subsumed by conflict b=2.")
	(test t (clause-valid? universal)
		  :fail-info "Universal clause should be valid.")

	;;test conflict avoidance
	(test nil (clause-avoids-all-conflicts? clause-a3c1)
		  :fail-info "The clause {a=3,c=1} does not avoid the conflict {b=2}")
	(test t (clause-avoids-all-conflicts? clause-a2b1c3)
		  :fail-info "Clause {a=2,b=1,c=3} is subsumed by a kernel.")
	(test nil (clause-avoids-all-conflicts? universal)
		  :fail-info "Universal clause cannot avoid any existing conflits.")

	;;Test check on the effect of new conflcits
	(test nil (would-conflicts-remove-all-kernels? (list clause-a1)) 
		  :fail-info "{a=1} is already a conflict, should not subsume conflicts")
	(test t (would-conflicts-remove-all-kernels? (list clause-b1))
		  :fail-info "{b=1} is used to make the space infeasible, should invalidate")
	(test t (would-conflicts-remove-all-kernels? (list clause-b1 clause-a1))
		  :fail-info "{b=1} is used to make the space infeasible, should invalidate regardless of others")

	(test t (not (null (get-valid-full-assignment state-space))))

	(test 8 (length (get-all-consistent-clauses state-space)) 
		  :fail-info "Should include 8 clauses.")

	;;now make the whole space infeasible
	(add-conflict! state-space clause-b1)
	(test nil (state-space-kernels state-space)
		  :fail-info "There should not be any kernels.")
	;;make sure the universal clause fails correctly
	(test nil (clause-valid? universal)
		  :fail-info "Even the universal clause is now invalid.")
	(test nil (clause-valid? clause-b1c2)
		  :fail-info "Clause {b=1,c=2} now invalid.")
	(test nil (clause-avoids-all-conflicts? universal)
		  :fail-info "No kernels, so universal can't be subsumed by one.")
	(test nil (clause-avoids-all-conflicts? clause-a2b1c3)
		  :fail-info "No kernels, so clause {a=2,b=1,c=3} fails")))

(deftest (test-committing :state-space) ()
  "Test the conflict support. Create a state-space, throw in some conflicts, try
the basic functions, then make the space completely infeasible and try again."
  (let*  ((var-descriptions
		   '((a :domain ((1) (2) (3)))
			 (b :domain ((1) (2)))
			 (c :domain ((1) (2) (3) (4)))))
		  ;;create the state space
		  (state-space (make-instance 'state-space-w-conflicts :variable-descriptions var-descriptions))
		  (universal (get-universal-clause state-space))
		  (clause-a1 (get-clause-from-descriptions state-space '((= a 1))))
		  (clause-b1 (get-clause-from-descriptions state-space '((= b 1))))
		  (clause-b2 (get-clause-from-descriptions state-space '((= b 2))))
		  (clause-a2b1 (get-clause-from-descriptions state-space '((= a 2) (= b 1))))
		  (clause-a3b1 (get-clause-from-descriptions state-space '((= a 3) (= b 1))))
		  (clause-b1c1 (get-clause-from-descriptions state-space '((= b 1) (= c 1))))
		  (clause-a2b1c1 (get-clause-from-descriptions state-space '((= a 2) (= b 1) (= c 1))))
		  (clause-a3b1c1 (get-clause-from-descriptions state-space '((= a 3) (= b 1) (= c 1)))))

	;;add conflicts
	(add-conflict! state-space clause-a1)

	(test t (commit-to-clause! clause-b1)
		  :fail-info "We should be able to commit to {b=1}.")

	;;test the conflicts
	(test (list clause-b2 clause-a1) (state-space-conflicts state-space)
		  :test #'lists-eql-out-of-order
		  :fail-info "Conflicts should be a1 and b2.")

	;;test the kernels
	(test (list clause-a2b1 clause-a3b1) (state-space-kernels state-space)
		  :test #'lists-eql-out-of-order
		  :fail-info "Kernels should be {a=2,b=1}, {a=3, b=1}")
	
	;;Now try another restriction
	(test t (commit-to-clause! clause-b1c1)
		  :fail-info "We should be able to commit to {b=1,c=1}.")

	(test nil (commit-to-clause! clause-b2)
		  :fail-info "We should be not able to commit to {b=2}.")

	;;Test a valid clause
	(test t (clause-valid? clause-a2b1c1)
		  :fail-info "Clause {a=1,b=1,c=1} should be valid.")

	;;test the kernels
	(test (list clause-a2b1c1 clause-a3b1c1) (state-space-kernels state-space)
		  :test #'lists-eql-out-of-order
		  :fail-info "Kernels should be {a=2,b=1,c=1}, {a=3, b=1, c=1}")


	(test nil (clause-avoids-all-conflicts? universal)
		  :fail-info "Universal clause cannot avoid any existing conflits.")
	(test nil (clause-avoids-all-conflicts?  clause-b1c1)
		  :fail-info "A clause we've committed to should be fine.")

	(test 2 (length (get-all-consistent-clauses state-space)) 
		  :fail-info "Should include 2 valid clauses.")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; utility function
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun lists-eql-out-of-order (list1 list2)
  "Utility function, tests if the two input lists are equal, but permuted."
  ;;remove all the elements of list1 from list2
  (loop for element1 in list1
	 do (setf list2 (remove element1 list2)))
  ;;list2 should now be empty
  (null list2))
