;;;; Copyright (c) 2013 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Author:
;;;;   Pedro H. Santana

(in-package :stochastic)

;;; ***********************************************************************
;;;
;;;          ***  Utilities for Combinatorics  ***
;;;
;;; ***********************************************************************

(defun to-list (arg)
  "If the argument isn't a list, listifies it."
  (if (listp arg)
	  (return-from to-list arg)
	  (return-from to-list (list arg))))


(defun remove-index (lst index)
  "Removes the element with a given index from a list.
  This function is non-destructive"
  (remove (elt lst index) lst :start index :count 1))


(defun delete-index! (lst index)
  "Removes the element with a given index from a list.
  This is a destructive function."
  (delete (elt lst index) lst :start index :count 1))


(defun range (&key (lower 0) (upper 0))
  "Analogous to the range(l,u) in Python"
  (loop for i from lower to (- upper 1) collecting i))


(defun cart-prod (set1 &rest other-sets)
  "Performs the Cartesian product of any number of sets *in the order*
  in which they are passed to the function."
  
  (let (set-prod)
	(setf set-prod (to-list set1))
	  
	(loop for set2 in other-sets do 
		 (setf set-prod (cart-prod-bin set-prod set2)))

	set-prod))

(defun cart-prod-bin (set1 set2)
  "Performs the Cartesian product between two sets of elements"
  (let ((set-prod nil))
	(loop for el1 in set1 do	   
		 (loop for el2 in set2 do 			  
			  (push (append (to-list el1) (to-list el2)) set-prod)))	
	(reverse set-prod)))


(defun n-perm-k (n-list k)  
  "Returns all k permutations of elements from a list"
  (let (n)
	(setf n (length n-list))

	;;Takes care of the special cases
	(cond
	  ((or (< n k) (< k 1)) 
	   (return-from n-perm-k nil)))
	
	;;Here we are guaranteed that n>k and k>0
	(n-perm-k-recursive n-list k)))


(defun n-perm-k-recursive (n-list k)
  "Recursive function that returns all k permutations of elements from a list
  with n elements. For efficiency purposes, it assumes that n > k"
  (let ((permutations nil) (all-permutations nil))
	;; Halting condition. Returns all single element sets in the list.
	(if(= k 1)
	  (return-from n-perm-k-recursive (loop for el in n-list collecting (list el))))
	
	(loop for idx from 0 to (- (length n-list) 1)  do
		 (setf permutations (n-perm-k-recursive (remove-index n-list idx) (- k 1)))		 
		 (setf all-permutations (append all-permutations (push-into-elements (elt n-list idx) permutations))))
	
	all-permutations))

(defun n-choose-k (n-list k)  
  "Returns all k combinations of elements from a list"
  (let (n)
	(setf n (length n-list))

	;;Takes care of degenerate cases
	(if (or (< n k) (< k 1)) 
	   (return-from n-choose-k nil))
	
	;;Here we are guaranteed that n>=k and k>0
	(n-choose-k-recursive n-list k)))

(defun push-into-elements (el lst-of-lsts)
  "Pushes a new element into every entry of a list of lists."
  (loop for i from 0 to (- (length lst-of-lsts) 1) do
	   (push el (elt lst-of-lsts i)))
  lst-of-lsts)

(defun n-choose-k-recursive (n-list k)
  "Recursive function that returns all k combinations of elements from a list
  with n elements. For efficiency purposes, it assumes that n > k"
  (let ((combinations nil) (all-combinations nil) el)

	;; Halting condition. 	
	(if (= k 1) ;Returns all single element sets in the list.
	   (return-from n-choose-k-recursive (loop for el in n-list collecting (list el))))	
	
	(loop while (> (length n-list) k) do
		 (setf el (elt n-list 0)) ; Selects first element
		 (setf n-list (delete-index! n-list 0)) ;Removes from the list
		 (setf combinations (n-choose-k-recursive n-list (- k 1))) ;All k-1 element combinatios involving the rest of the list								
		 (setf all-combinations (append all-combinations (push-into-elements el combinations)))) ;All k element combinations with the current first element
	
	;;The last combination comprises all remaining elements from the list 
	(setf all-combinations (append all-combinations (list n-list)))
	
	all-combinations))


(defun permutations (lst)
  "Returns all possible permutations of a list"
  (n-perm-k lst (length lst)))