;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;; Defines useful functions for converting between different datatypes used
;; in the :stochastic package.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Generic Definition
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric convert-datatype (A out)
  (:documentation "Converts A to datatype OUT. Supported datatypes are:
                   :SCALAR
                   :LIST
                   :VECTOR
                   :MATLISP"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Convenience Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun get-datatype (A)
  "Return the datatype of A."
  (cond ((listp A)
		 :list)
		((arrayp A)
		 :vector)
		((numberp A)
		 :scalar)
		#+:matlisp
		((typep A 'real-matrix)
		 :matlisp)
		(T
		 (error "Unknown datatype for ~a." A))))

(defun generic-size (A)
  "Determine the size of A, regardless of what datatype A is.

   ========================================
   Preconditions
   ========================================
   A is one of the recognized datatypes:
   :list
   :scalar
   :vector
   :matlisp

   ========================================
   Effects
   ========================================
   None

   ========================================
   Returns
   ========================================
   List of length two. First element is number of rows, second is
   number of columns."
  (cond ((numberp A)
		 '(1 1))
		((listp A)
		 (if (listp (first A))
			 `(,(length A) ,(length (first A)))
			 `(,(length A) 1)))
		((vectorp A)
		 `(,(array-dimension A 0) 1))
		((arrayp A)
		 `(,(array-dimension A 0) ,(array-dimension A 1)))
		#+:matlisp
		((typep A 'real-matrix)
		 (size A))
		(T
		 (error "~a is an unrecognized datatype." A)))
)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Implementations
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod convert-datatype (A (out (eql :scalar)))
  (cond ((numberp A)
		 A)
		((and (listp A) (eql (length A) 1))
		 (first A))
		((and (arrayp A) (eql (array-dimension A 0) 1))
		 (aref A 0))
		(T
		 (error "Cannot convert ~a to :SCALAR." A))))

(defmethod convert-datatype (A (out (eql :list)))
  (cond ((listp A)
		 A)
		((numberp A)
		 (list A))
		((vectorp A)
		 (loop with rows = (- (array-dimension A 0) 1)
			  for i from 0 to rows
			  collecting (aref A i)))
		((arrayp A)
		 (loop with rows = (- (array-dimension A 0) 1) and cols = (- (array-dimension A 1) 1)
			for i from 0 to rows
			collecting (loop for j from 0 to cols
						  collecting (aref A i j))))
		(T
		 (error "Cannot convert ~a to a :LIST." A))))

(defmethod convert-datatype (A (out (eql :vector)))
  (cond ((arrayp A)
		 A)
		((numberp A)
		 (make-array 1 :initial-element A))
		((listp A)
		 ;; Determine if it is one or two dimensional.
		 (if (listp (first A))
			 (make-array (list (length A) (length (first A))) :initial-contents A)
			 (make-array (length A) :initial-contents A)))
		(T
		 (error "Cannot convert ~a to a :VECTOR." A))))

#+:matlisp
(defmethod convert-datatype (A (out (eql :matlisp)))
  (cond ((listp A)
		 (make-real-matrix A))
		((numberp A)
		 (make-real-matrix (list A)))
		((arrayp A)
		 (make-real-matrix A))
		(T
		 (error "Unable to convert ~a to :MATLISP object." A))))

#+:matlisp
(defmethod convert-datatype ((A real-matrix) (out (eql :matlisp)))
  A)

#+:matlisp
(defmethod convert-datatype ((A real-matrix) (out (eql :vector)))
  (convert-to-lisp-array A))

#+:matlisp
(defmethod convert-datatype ((A real-matrix) (out (eql :list)))
  (convert-datatype (convert-to-lisp-array A) :list))

#+:matlisp
(defmethod convert-datatype ((A real-matrix) (out (eql :scalar)))
  (if (and (eql (number-of-rows A) 1) (eql (number-of-cols A) 1))
	  (matrix-ref A 0)
	  (error "Unable to convert ~a to a :SCALAR." A)))

