;;;; Copyright (c) 2013 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Pedro H. Santana

(in-package :stochastic)

(defconstant *is-above* '>)
(defconstant *is-below* '<)
(defconstant *is-equal* '=)
(defconstant *inconclusive* '?)

;;; ***********************************************************************
;;;
;;;    ***  Functions for computing probabilities of discrete events ***
;;; 
;;; ***********************************************************************

(defun prob-union (prob-model cond-prob)  
  "Exact union probability."  
  (prob-union-truncated prob-model cond-prob (length (get-assignment-list cond-prob))))


(defun prob-union-lb (prob-model cond-prob order)
  "According to the Bonferroni inequalities, every truncated evaluation
  of the Inclusion-Exclusion principle that ends on an even number of
  combinations is a lower bound of the exact probability."   
  (prob-union-truncated prob-model cond-prob (* 2 order)))


(defun prob-union-ub (prob-model cond-prob order)
  "According to the Bonferroni inequalities, every truncated evaluation
  of the Inclusion-Exclusion principle that ends on an odd number of
  combinations is an upper bound of the exact probability."  
  (prob-union-truncated prob-model cond-prob (- (* 2 order) 1)))


(defun prob-union-boole (prob-model cond-prob)
  "Boole's famous (and quick) upper bound for the probability of a union
  of events."  
  (prob-union-ub prob-model cond-prob 1))


(defun prob-union-threshold (prob-model cond-prob threshold)
  "This function tries to determine whether the probability of a union of events
  is above, below, or is equal to a given threshold."
  (let (result)
	(setf result (prob-union-truncated prob-model cond-prob (length (get-assignment-list cond-prob)) :threshold threshold))

	;; If the function ends up computing the whole expression of the Inclusion-Exclusion principle, 
	;; then we know that the probability has to be equal to the threshold. 
	(if (equal result *inconclusive*)
		*is-equal*
		result)))


(defun prob-union-truncated (prob-model prob-query order &key (threshold 0 thres-p))
  "Takes a probability query in the form of a conditional disjunctions of events and computes
  the probability value  of their union  by implementing the Inclusion-Exclusion principle. 
  The second argument is an object that represents a probability model." 
  (let ((prob 0) (combinations nil) (comb-prob 0))

	(loop for k from 1 to order do
	   ;; all combinations of k elements from the
	   ;; assignment list. This might take A LONG time 
	   ;; to compute.
		 (setf combinations (n-choose-k (get-assignment-list prob-query) k))

	   ;;Sums the probability of all combinations
		 (setf comb-prob 
		 	   (loop for comb in combinations summing 
		 			(compute-joint-probability prob-model (make-cond-prob comb (get-conditional-list prob-query)))))
		 
	   ;;Inclusion-Exclusion principle
		 (if (evenp k)
			 (decf prob comb-prob)
			 (incf prob comb-prob))

	   ;; A probability threshold has been provided for early termination. In this case,
	   ;; instead of returning a probability, the function returns whether it's above,
	   ;; below or is equal to the threshold.
		 (when thres-p
		   (cond
			 ((evenp k) ;;prob is a lower bound			  
			  (if (> prob threshold)
				  (return-from prob-union-truncated *is-above*)))
			 (t ;; prob is an upper bound
			  (if (< prob threshold)
				  (return-from prob-union-truncated *is-below*))))))
	
	;; Returns the computed probability or the symbol saying that it's equal to the 
	;; given probability threshold. 	
	(if thres-p
		(return-from prob-union-truncated *inconclusive*) 
		(return-from prob-union-truncated prob))))