;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;;; This file describes the interface for distributions.
;;; It contains the base class for distributions and the
;;; definition of generic functions. Some common methods are
;;; also described here.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass probability-distribution ()
  ((domain :initform '()
		   :initarg domain
		   :accessor distribution-domain
		   :documentation "The domain object describing the domain for this
                           distribution. Should be a subclass of DOMAIN.")
   (partition :initform '()
			  :initarg partition
			  :accessor partition-function
			  :documentation "The partition function for this distribution.")
   (query-variables :initform '()
					:initarg :query-variables
					:accessor get-query-variables
					:documentation "The variables over which probability mass
                                    is defined.")
   (conditioning-variables :initform '()
						   :initarg :conditioning-variables
						   :accessor get-conditioning-variables
						   :documentation "The variables that this distribution
                                           is conditioned upon."))
  (:documentation "The base class for any probability distribution."))


;(defclass discrete-probability-distribution (probability-distribution))

;(defclass continuous-probability-distribution (probability-distribution))


(defgeneric pmf (distro element)
  (:documentation "Evaluate the probability mass of ELEMENT according to
                   DISTRO's probability mass function."))

(defgeneric cmf (distro element-set)
  (:documentation "Evaluate the probability mass of all elements in ELEMENT-SET
                   according to DISTRO's probability mass function."))

(defgeneric pdf (distro point)
  (:documentation "Evaluate the probability density of DISTRO at the domain POINT."))

(defgeneric cdf (distro point)
  (:documentation "Evaluate the cumulative density of DISTRO at the domain POINT."))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Generic Function Definitions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric evaluate-distribution (distribution value)
  (:documentation
   "Evaluate the DISTRIBUTION at the given VALUE. If VALUE is not a member
    of DISTRIBUTION's DOMAIN, implementations can either return 0 or 
    signal an error. Typically distributions with bounded domains will
    return 0 unless VALUE is not a member of the unbounded version of the
    domain.

    ======================================================================
    Preconditions
    ======================================================================
    VALUE is a valid member of DISTRIBUTION's DOMAIN. If DOMAIN is a
    bounded domain, VALUE must be a valid member of the unbounded version
    of the DOMAIN.

    DISTRIBUTION must be properly initialized. If any slots of
    DISTRIBUTION are modified, CACHE-DISTRIBUTION-SLOTS must be called on
    the DISTRIBUTION to ensure proper caching of any helper slots.

    ======================================================================
    Effects
    ======================================================================
    None.

    ======================================================================
    Returns
    ======================================================================
    Scalar. The result of evaluating DISTRIBUTION at VALUE."))

(defgeneric evaluate-log-distribution (distribution value)
  (:documentation
   "Evaluate the DISTRIBUTION at the given VALUE and return the natural
    logarithm of that. If VALUE is not a member of DISTRIBUTION's DOMAIN,
    implementations can either return -Inf or signal an error. Typically 
    distributions with bounded domains will return -Inf unless VALUE is not
    a member of the unbounded version of the domain.

    A generic implementation of this is in distribution-utilities.lisp.
    Distributions may override this for greater efficiency.

    ======================================================================
    Preconditions
    ======================================================================
    VALUE is a valid member of DISTRIBUTION's DOMAIN. If DOMAIN is a
    bounded domain, VALUE must be a valid member of the unbounded version
    of the DOMAIN.

    DISTRIBUTION must be properly initialized. If any slots of
    DISTRIBUTION are modified, CACHE-DISTRIBUTION-SLOTS must be called on
    the DISTRIBUTION to ensure proper caching of any helper slots.

    ======================================================================
    Effects
    ======================================================================
    None.

    ======================================================================
    Returns
    ======================================================================
    Scalar. The natural log of DISTRIBUTION at VALUE."))

(defgeneric sample-distribution (distribution)
  (:documentation
   "Sample a member of of DISTRIBUTION's DOMAIN with probability defined
    by the DISTRIBUTION.

    ======================================================================
    Preconditions
    ======================================================================
    DISTRIBUTION must be properly initialized. If any slots of
    DISTRIBUTION are modified, CACHE-DISTRIBUTION-SLOTS must be called on
    the DISTRIBUTION to ensure proper caching of any helper slots.

    ======================================================================
    Effects
    ======================================================================
    None.

    ======================================================================
    Returns
    ======================================================================
    Member of DISTRIBUTION's DOMAIN."))

(defgeneric cache-distribution-slots (distribution)
  (:documentation
   "Some distributions will need to pre-compute values such as the 
    partition function, mean, variance, etc. Pre-computation is necessary
    to avoid unneccessary repeated calculations (such as repeated
    inversion of the same matrix. When any slot values are modified in
    DISTRIBUTION, CACHE-DISTRIBUTION-SLOTS should be called to ensure
    DISTRIBUTION's pre-computed values are valid.

    ======================================================================
    Preconditions
    ======================================================================
    None.

    ======================================================================
    Effects
    ======================================================================
    Slots of DISTRIBUTION are modified.

    ======================================================================
    Returns
    ======================================================================
    DISTRIBUTION."))

(defgeneric learn-model (distribution data-set &key ml-estimate?)
  (:documentation
   "Create a new distribution of the same class as DISTRIBUTION
    such that it is the maximum-likelihood OR unbiased
    estimate of the true distribution that generated the data. Each class
    of distribution may choose to ignore the ML-ESTIMATE? flag. See doc
    string for class of DISTRIBUTION in question for more info.

    DATA-SET is a list of data-points.

    ======================================================================
    Preconditions
    ======================================================================
    DATA-SET is a list of valid members of DISTRIBUTION's domain.

    DATA-SET is NOT empty.

    ======================================================================
    Effects
    ======================================================================
    None.

    ======================================================================
    Returns
    ======================================================================
    A new, valid distribution of the same class as DISTRIBUTION."))

(defgeneric log-likelihood (distribution data-set)
  (:documentation
   "Calculate the log-likelihood of observing DATA-SET. A generic
    method for this is given in distribution-utilities.lisp. Distributions
    may override it to provide greater efficiency.

    DATA-SET is a list of data-points.

    ======================================================================
    Preconditions
    ======================================================================
    DATA-SET is a list of valid members of DISTRIBUTION's domain.

    DATA-SET is NOT empty.

    ======================================================================
    Effects
    ======================================================================
    None.

    ======================================================================
    Returns
    ======================================================================
    Scalar. The log-likelihood of observing the data in DATA-SET."))

(defgeneric distribution-moment (distribution index)
  (:documentation
   "Calculate and return the INDEXth moment of DISTRIBUTION. All moments
    except the first are about the mean.

    When there is no closed form way to calculate the moments,
    implementations may signal a condition of type 
    DISTRIBUTION-MOMENT-ERROR.

    ======================================================================
    Preconditions
    ======================================================================
    INDEX is an integer.

    ======================================================================
    Effects
    ======================================================================
    None.

    ======================================================================
    Returns
    ======================================================================
    The INDEXth moment of DISTRIBUTION."))