;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;;; To Do:
;;; o Move to a more general file.
(mtu::declare-debug-print-system! :stochastic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Manipulating Gaussian Normal Distributions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This file contains classes and methods for representing and
;;; manipulating a univariate normal distribution.

;;; It should be extended to multi-variate normal distributions.

;;; Note: This file is a work in progress.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Normal Distribution
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass normal-distribution (probability-distribution)
  
  ((index :initarg :index
		  :accessor normal-distribution-index)
		
   (domain :initform '()
		   :accessor domain
		   :initarg :domain
		   :documentation "Instances of variables that distribution 
                           is defined over.")
   
   (mean :initform '()
		 :initarg :mean
		 :accessor normal-distribution-mean
		 :documentation "Mean of the distribution,
                         A vector of size D = |Domain|.")

   (internal-mean :initform '()
				  :reader normal-distribution-internal-mean
				  :documentation "Not meant to be accessed outside of this class.
                                  Stores mean in a convenient data structure for 
                                  processing.")

   (variance :initform '()
			 :initarg :variance
			 :accessor normal-distribution-variance
			 :documentation "Variance of the distribution.
                             A 2-D vector of size DxD, where D = |Domain|.")

   (internal-variance :initform '()
					  :reader internal-variance
					  :documentation "Not meant to be accessed outside of
                                      this class.")
   
   (partition :initform '()
			  :accessor normal-distribution-partition
			  :documentation "Partition function (normalization constant)
                              for the distribution.")

   (inverse-variance :initform '()
					 :accessor normal-distribution-inverse-variance
					 :documentation "Precomputed inverse of the variance matrix.")

   (addition-fun :initform '()
				 :reader normal-distribution-addition-fun
				 :documentation "Function handle for a function that takes two arguments
                                 of the same type as INTERNAL-MEAN or INTERNAL-VARIANCE
                                 and returns a single object which is the addition of
                                 them.")

   (inner-product-fun :initform '()
					  :reader normal-distribution-inner-product-fun
					  :documentation "Function handle for a function that takes two
                                      arguments the same type as INTERNAL-MEAN or floats 
                                      and returns the inner product.")

   (outer-product-fun :initform '()
					  :reader normal-distribution-outer-product-fun
					  :documentation "Function handle for a function that takes two
                                      arguments the same type as INTERNAL-MEAN or floats 
                                      and returns the outer product.")

   (variance-sqrt :initform '()
				  :reader normal-distribution-variance-sqrt
				  :documentation "Internal slot for caching the square root of the variance.")

   (native-datatype :initform '()
					:accessor distribution-native-datatype
					:documentation "The native datatype returned if
                                    *distribution-return-native-datatype* is T.")

   (internal-datatype :initform '()
					  :accessor distribution-internal-datatype
					  :documentation "The internal datatype used in the implementation."))
  
  (:documentation "Represents a normal distribution over real-values 
                  variables, with MEAN and VARIANCE. The specialized
                  versions (UNIVARIATE-NORMAL-DISTRIBUTION and
                  MULTIVARIATE-NORMAL-DISTRIBUTION) should be used
                  instead."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Generic Method Definitions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defgeneric sample-unit-variance-gaussian (dist)
  (:documentation "Every subclass of NORMAL-DISTRIBUTION needs to provide
                   this method to sample from a zero mean, unit variance
                   normal distribution."))


(defun randn ()
  "Samples a value from a zero mean, unit variance normal distribution."
  (let* ((u (+ 1 (* -1 (random 1.0))))
		 (v (random 1.0))
		 (z (* (sqrt (* -2 (log u))) (cos (* 2 pi v)))))
	z))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Constructor
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((dist normal-distribution) &key (verify? T))
  "Calculate PARTITION and INVERSE-VARIANCE to make later calculations easier."
  (cache-distribution-slots dist)
  (if verify? (validate-slots dist))
  dist)

(defmethod cache-distribution-slots ((dist normal-distribution))
  "Copy the mean and variance fields into the internal representations.
   Implementations of normal-distribution are responsible for providing an
   :AFTER method to finish caching slots."

  (with-slots (internal-datatype
			   mean
			   variance
			   internal-mean
			   internal-variance
			   native-datatype) dist
	(setf native-datatype (get-datatype mean))
	(setf internal-mean (convert-datatype mean internal-datatype))
	(setf internal-variance (convert-datatype variance internal-datatype)))
  dist)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Methods
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod print-object ((dist normal-distribution) stream)
  (format stream "Normal Distribution ~%Mean: ~a~%Variance: ~a"
		  (normal-distribution-mean dist)
		  (normal-distribution-variance dist)))

(defmethod evaluate-distribution ((dist normal-distribution) value) 
  "Evaluate the DISTRIBUTION at the given VALUE. If VALUE is not a member
   of DISTRIBUTION's DOMAIN, implementations can either return 0 or 
   signal an error. Typically distributions with bounded domains will
   return 0 unless VALUE is not a member of the unbounded version of the
   domain.

   ======================================================================
   Preconditions
   ======================================================================
   VALUE is a valid member of DISTRIBUTION's DOMAIN. If DOMAIN is a
   bounded domain, VALUE must be a valid member of the unbounded version
   of the DOMAIN.

   DISTRIBUTION must be properly initialized. If any slots of
   DISTRIBUTION are modified, CACHE-DISTRIBUTION-SLOTS must be called on
   the DISTRIBUTION to ensure proper caching of any helper slots.

   ======================================================================
   Effects
   ======================================================================
   None.

   ======================================================================
   Returns
   ======================================================================
   Scalar. The result of evaluating DISTRIBUTION at VALUE."
  (let ((value (convert-datatype value (distribution-internal-datatype dist))))

	(multiple-value-bind (success? details) (domain-member? (domain dist) value)
	  (ccase details
		(:not-a-member (error "Value ~a is not a member of the domain of ~a."
							  value dist))
		(:out-of-bounds 0)
		(:member (with-slots (internal-mean 
							  partition 
							  addition-fun 
							  inner-product-fun
							  outer-product-fun
							  inverse-variance) dist
				   (handler-case
					   (let ((residual (funcall addition-fun (funcall inner-product-fun -1 internal-mean) value)))
						 (/ (exp (/ (- (convert-datatype (funcall inner-product-fun 
																  residual 
																  (funcall outer-product-fun 
																		   inverse-variance 
																		   residual)) 
														 :scalar)) 
									2)) 
							partition))
					 (floating-point-underflow () 0.0))))))))

(defmethod evaluate-log-distribution ((dist normal-distribution) value)  
  "Evaluate the DISTRIBUTION at the given VALUE and return the natural
   logarithm of that. If VALUE is not a member of DISTRIBUTION's DOMAIN,
   implementations can either return -Inf or signal an error. Typically 
   distributions with bounded domains will return -Inf unless VALUE is not
   a member of the unbounded version of the domain.

   A generic implementation of this is in distribution-utilities.lisp.
   Distributions may override this for greater efficiency.

   ======================================================================
   Preconditions
   ======================================================================
   VALUE is a valid member of DISTRIBUTION's DOMAIN. If DOMAIN is a
   bounded domain, VALUE must be a valid member of the unbounded version
   of the DOMAIN.

   DISTRIBUTION must be properly initialized. If any slots of
   DISTRIBUTION are modified, CACHE-DISTRIBUTION-SLOTS must be called on
   the DISTRIBUTION to ensure proper caching of any helper slots.

   ======================================================================
   Effects
   ======================================================================
   None.

   ======================================================================
   Returns
   ======================================================================
   Scalar. The natural log of DISTRIBUTION at VALUE."

  (let ((value (convert-datatype value (distribution-internal-datatype dist))))

	(multiple-value-bind (success? details) (domain-member? (domain dist) value)
	  (ccase details
		(:not-a-member (error "Value ~a is not a member of the domain of ~a."
							  value dist))
		(:out-of-bounds 0)
		(:member (with-slots (internal-mean 
							  partition 
							  addition-fun 
							  inner-product-fun
							  outer-product-fun
							  inverse-variance) dist
				   (let ((residual (funcall addition-fun 
											(funcall inner-product-fun 
													 -1 
													 internal-mean) 
											value)))
					 (- (- (/ (funcall inner-product-fun
									   residual
									   (funcall outer-product-fun
												inverse-variance
												residual)) 2))
						(log partition)))))))))

(defmethod sample-distribution ((dist normal-distribution))
  "Sample a member of of DISTRIBUTION's DOMAIN with probability defined
   by the DISTRIBUTION.

   ======================================================================
   Preconditions
   ======================================================================
   DISTRIBUTION must be properly initialized. If any slots of
   DISTRIBUTION are modified, CACHE-DISTRIBUTION-SLOTS must be called on
   the DISTRIBUTION to ensure proper caching of any helper slots.

   ======================================================================
   Effects
   ======================================================================
   None.

   ======================================================================
   Returns
   ======================================================================
   Member of DISTRIBUTION's DOMAIN."

  (with-slots (addition-fun
			   internal-mean
			   outer-product-fun
			   variance-sqrt
			   native-datatype) dist
	(convert-datatype (funcall addition-fun
							   internal-mean
							   (funcall outer-product-fun
										variance-sqrt 
										(sample-unit-variance-gaussian dist)))
					  native-datatype)))

(defmethod learn-model ((distribution (eql :normal-distribution)) data-set &key (ml-estimate? nil))
  "Create a new distribution of the same class as DISTRIBUTION
   such that it is the maximum-likelihood OR unbiased
   estimate of the true distribution that generated the data. Respects
   ML-ESTIMATE? flag.

   DATA-SET is a list of data-points.

   ======================================================================
   Preconditions
   ======================================================================
   DATA-SET is a list of valid members of DISTRIBUTION's domain.

   DATA-SET is NOT empty.

   ======================================================================
   Effects
   ======================================================================
   None.

   ======================================================================
   Returns
   ======================================================================
   A new, valid distribution of the same class as DISTRIBUTION."

  ;; Determine if this is scalar or not.
  (let ((converted-element (convert-datatype (first data-set) :list)))
	;; If scalar, this is easy.
	(if (eql (length converted-element) 1)
		(learn-model :scalar-normal-distribution data-set :ml-estimate? ml-estimate?)
		;; If not scalar, convert to matlisp and work that way.
		#+:matlisp
		(learn-model :matlisp-normal-distribution data-set :ml-estimate? ml-estimate?))))


(defmethod distribution-moment ((dist normal-distribution) index)
  "Calculate and return the INDEXth moment of DISTRIBUTION. All moments
   except the first are about the mean.

   When there is no closed form way to calculate the moments,
   implementations may signal a condition of type 
   DISTRIBUTION-MOMENT-ERROR.

   ======================================================================
   Preconditions
   ======================================================================
   INDEX is an integer.

   ======================================================================
   Effects
   ======================================================================
   None.

   ======================================================================
   Returns
   ======================================================================
   The INDEXth moment of DISTRIBUTION."
  (cond ((equal index 1)
		 (slot-value dist 'mean))
		((oddp index)
		 0)
		((evenp index)
		 (with-slots (internal-variance
					  native-datatype
					  inner-product-fun) dist
		   (let ((double-factorial (reduce #'* 
										   (loop for i from (- index 1) downto 3 by 2
												collecting i)))
				 (exponent (loop with prod = internal-variance
							  repeat (- (/ index 2) 1)
							  do (setf prod (funcall inner-product-fun prod internal-variance))
							  finally (return prod))))
			 (convert-datatype (funcall inner-product-fun double-factorial exponent)
							   native-datatype))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Factory Method
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-normal-distribution (&key mean variance (domain nil) (verify? T))
  (let ((converted-mean (convert-datatype mean :list)))
	(if (eql (length converted-mean) 1)
		(make-instance 'scalar-normal-distribution 
					   :mean mean 
					   :variance variance 
					   :domain domain
					   :verify? verify?)
		#+:matlisp
		(make-instance 'matlisp-normal-distribution 
					   :mean mean 
					   :variance variance 
					   :domain domain
					   :verify? verify?))))