;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Manipulating Gaussian Normal Distributions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This file contains classes and methods for representing and
;;; manipulating a univariate normal distribution.

;;; It should be extended to multi-variate normal distributions.

;;; Note: This file is a work in progress.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Normal Distribution
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass scalar-normal-distribution (normal-distribution)
  
  ((addition-fun :initform #'+)

   (inner-product-fun :initform #'*)

   (outer-product-fun :initform #'*)

   (internal-datatype :initform :scalar))
  
  (:documentation "Represents a normal distribution that uses scalar floats
                   for calculations and internal representations."))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Constructor
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((dist scalar-normal-distribution) &key domain)
  "Calculate PARTITION and INVERSE-VARIANCE to make later calculations easier."
  (if (null domain)
	  (setf (slot-value dist 'domain) (make-instance 'continuous-domain :dims 1)))
  dist)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Functions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod sample-unit-variance-gaussian ((dist scalar-normal-distribution))
  (randn))

(defmethod cache-distribution-slots :after ((dist scalar-normal-distribution))
  "Precompute DIST's partition function, inverse variance, and variance sqrt.

   ========================================
   Preconditions
   ========================================
   MEAN and VARIANCE of DIST are populated.

   ========================================
   Effects
   ========================================
   DIST is modified.

   ========================================
   Returns
   ========================================
   DIST."

  (with-slots (internal-mean
			   internal-variance
			   inverse-variance
			   variance-sqrt
			   partition) dist
	;; Now calculate the partition and inverse variance.
	(setf partition (sqrt (* 2 pi internal-variance)))
	(setf inverse-variance (/ internal-variance))
	(setf variance-sqrt (sqrt internal-variance)))
  dist)

(defmethod validate-slots ((dist scalar-normal-distribution))
  (with-slots (internal-mean internal-variance) dist
	;; Make sure variance is not negative.
	(if (<= internal-variance 0)
		(error "Variance must be > 0. It is ~a." internal-variance))))

(defmethod learn-model ((distribution (eql :scalar-normal-distribution)) data-set &key (ml-estimate? nil))
  "Create a new distribution of the same class as DISTRIBUTION
   such that it is the maximum-likelihood OR unbiased
   estimate of the true distribution that generated the data. Respects
   ML-ESTIMATE? flag.

   DATA-SET is a list of data-points.

   ======================================================================
   Preconditions
   ======================================================================
   DATA-SET is a list of valid members of DISTRIBUTION's domain.

   DATA-SET is NOT empty.

   ======================================================================
   Effects
   ======================================================================
   None.

   ======================================================================
   Returns
   ======================================================================
   A new, valid distribution of the same class as DISTRIBUTION."

  ;; Determine the original datatype.
  (let* ((original-datatype (get-datatype (first data-set)))
		 ;; If original datatype isn't :scalar, convert everything.
		 (data-set (if (eql :scalar original-datatype) 
					   data-set 
					   (map 'list #'(lambda (x) (convert-datatype x :scalar)) data-set)))
		 (n (length data-set))
		 (emp-average (/ (reduce #'+ data-set) n))
		 (emp-variance (/ (reduce #'+ (map 'list #'(lambda (x) 
													 (expt (- x emp-average) 2)) data-set)) 
						  (if ml-estimate? n (- n 1)))))
	(make-instance 'scalar-normal-distribution
				   :mean (convert-datatype emp-average original-datatype)
				   :variance (convert-datatype emp-variance original-datatype))))