;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;;; This file describes classes and functions for manipulating 
;;; univariate normal distributions.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass univariate-normal-distribution (normal-distribution)
  ()
  (:documentation "Represents a normal distribution over real-values
                   variables of one dimension."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Constructor
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((dist univariate-normal-distribution) &key mean variance (verify? T) (empty? nil))
  "Calculate PARTITION and INVERSE-VARIANCE to make later calculations easier."

  (when (or (not verify?) (not empty?))
	;; Check the variance is nonnegative and mean and variance are numbers.
	(if (not (numberp mean))
		(error "Distribution has an invalid mean of ~a." mean))
	(if (or (<= variance 0) (not (numberp variance)))
		(error "Distribution has an invalid variance of ~a." variance)))
  (unless empty?
	(cache-distribution-slots dist))
  dist)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Methods
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod cache-distribution-slots ((dist univariate-normal-distribution))
  "Precompute DIST's partition function and inverse variance.

   ========================================
   Preconditions
   ========================================
   MEAN and VARIANCE of DIST are populated.

   ========================================
   Effects
   ========================================
   DIST is modified.

   ========================================
   Returns
   ========================================
   DIST."
  (with-slots (partition inverse-variance variance) dist
	(setf partition (sqrt (* 2 pi variance)))
	(setf inverse-variance (/ 1 variance)))
  dist)

(defmethod evaluate-distribution
	((distribution univariate-normal-distribution) value)
  "Evaluate DISTRIBUTION at VALUE.

   ========================================
   Preconditions
   ========================================
   VALUE is a valid value in DISTRIBUTION's domain.

   ========================================
   Effects
   ========================================
   None.

   ========================================
   Returns
   ========================================
   Scalar."
  (check-distribution-argument! distribution value)
  (with-slots (partition inverse-variance mean) distribution
	(let* ((exponent (* -0.5 (expt (- value mean) 2) inverse-variance))
		   (base (handler-case (exp exponent) 
				   (floating-point-underflow () 0))))
	  (/ base partition))))

(defmethod sample-distribution
	((distribution univariate-normal-distribution))
  "Take a sample from DISTRIBUTION.

   ========================================
   Preconditions
   ========================================
   DISTRIBUTION has been properly initialized.

   ========================================
   Effects
   ========================================
   None.

   ========================================
   Returns
   ========================================
   Scalar sample from DISTRIBUTION."
  (with-slots (mean variance) distribution
	(sample-normal-distribution mean variance)))

(defmethod check-distribution-argument! 
	((distribution univariate-normal-distribution) value)
  (with-slots (mean) distribution
	(if (not (numberp value))
		(error "Check value of normal distribution ~a at ~a." 
			   distribution value))))

(defun sample-normal-distribution (mean variance)
  (let* ((u (+ 1 (* -1 (random 1.0))))
		 (v (random 1.0))
		 (z (* (sqrt (* -2 (log u))) (cos (* 2 pi v)))))
	(+ mean (* (sqrt variance) z))))

(defmethod learn-model ((distribution univariate-normal-distribution) data-set &key (ml-estimate? nil))
  "Calculate the estimates of MEAN and VARIANCE from the DATA-SET and
   return a new distribution so that it represents the learned MEAN and VARIANCE.

   The default behavior is to calculate the unbiased (not ML) estimate.

   ========================================
   Preconditions
   ========================================
   DATA-SET is a list of scalars.

   DATA-SET is not empty.

   ========================================
   Effects
   ========================================
   None

   ========================================
   Returns
   ========================================
   New DISTRIBUTION."

  ;; Calculate the unbiased mean and variance.
  (let* ((n (length data-set))
		 (emp-average (/ (reduce #'+ data-set) n))
		 (emp-variance (/ (reduce #'+ (map 'list #'(lambda (x) 
													 (expt (- x emp-average) 2)) data-set)) 
						  (if ml-estimate? n (- n 1)))))
	(make-instance 'univariate-normal-distribution
				   :domain (distribution-domain distribution)
				   :mean emp-average
				   :variance emp-variance)))