;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;;; This file describes classes and functions for manipulating 
;;; multivariate normal distributions.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass uniform-distribution (probability-distribution)
  
  ((index :initarg :index
		  :accessor uniform-distribution-index)
		
   (domain :initform '()
		   :accessor uniform-distribution-domain
		   :initarg :domain
		   :documentation "Instances of variables that distribution 
                           is defined over.")
   (partition :initform '()
			  :accessor uniform-distribution-partition
			  :documentation "This distribution's partition function.")

   (native-datatype :initform '()
					:accessor distribution-native-datatype
					:documentation "The datatype used to construct this dist."))
  
  (:documentation "Represents a uniform distribution over the DOMAIN."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Constructor
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((dist uniform-distribution) &key)
  "Calculate PARTITION to make later calculations easier."
  (with-slots (domain native-datatype) dist
	(if (typep domain 'continuous-domain)
		(with-slots ((native-datatype1 native-datatype)) domain
		  (setf native-datatype native-datatype1))))
  (cache-distribution-slots dist)
  dist)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Factory Methods
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro make-uniform-distribution (&key (hint nil hint-p) (lower-bound nil lower-bound-p) 
									 (upper-bound nil upper-bound-p) 
									 (elements nil elements-p))
  "Macro that automatically creates a UNIFORM-DISTRIBUTION with the correct
   domain based on the input arguments. If lisp is unable to determine the type
   of the arguments at compile time, HINT can be used to tell the macro which
   domain to use."

  ;; First decide if we are making a continuous or discrete distribution.
  (if elements-p
	  ;; We're making a discrete distribution.
	  `(make-instance 'uniform-distribution
					  :domain (make-instance 'discrete-domain
											 :elements ,elements))
	  ;; If elements isn't provided, we're making a continuous domain.
	  `(make-instance 'uniform-distribution
					  :domain (make-continuous-bounded-domain
							   ,@(if hint-p `(:hint ,hint))
							   :lower-bound ,lower-bound
							   :upper-bound ,upper-bound))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Methods
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod cache-distribution-slots ((dist uniform-distribution))
  "Cache the partition function."
  (with-slots (partition domain) dist
	(setf partition (domain-volume domain))))

(defmethod evaluate-distribution
	((distribution uniform-distribution) value)
  "Evaluate DISTRIBUTION at VALUE. If VALUE is not a member of DOMAIN,
   returns zero.

   ========================================
   Preconditions
   ========================================
   None.

   ========================================
   Effects
   ========================================
   None.

   ========================================
   Returns
   ========================================
   Scalar."

  (with-slots (partition domain) distribution
	(if (domain-member? domain value)
		(/ 1 partition)
		0)))

(defmethod check-distribution-argument!
	((distribution uniform-distribution) value)
  (with-slots (domain) distribution
	(if (not (domain-member? domain value))
		(error "Check value of uniform distribution ~a at ~a." 
			   distribution value))))

(defmethod sample-distribution
	((distribution uniform-distribution))
  "Take a sample from DISTRIBUTION.

   ========================================
   Preconditions
   ========================================
   DISTRIBUTION has been properly initialized.

   ========================================
   Effects
   ========================================
   None.

   ========================================
   Returns
   ========================================
   Member of domain sampled from DISTRIBUTION."
  (with-slots (domain native-datatype) distribution
	(convert-datatype (uniform-sample-from-domain domain) native-datatype)))

(defmethod learn-model ((distribution (eql :uniform-distribution)) data-set &key ml-estimate?)
  ;; First determine if the data is scalar or not.
  (let ((element-size (generic-size (first data-set))))
	(if (eql 1 (first element-size))
		;; Delegate to a scalar-uniform-distribution
		(learn-model :scalar-uniform-distribution data-set :ml-estimate? ml-estimate?)
		;; Delegate to a multivariate distribution.
		#+:matlisp
		(learn-model :matlisp-uniform-distribution data-set :ml-estimate? ml-estimate?))))

(defmethod learn-model ((distribution (eql :scalar-uniform-distribution)) data-set &key ml-estimate?)
  (declare (ignore ml-estimate?))
  (let* ((converted-values (map 'list #'(lambda (x) (convert-datatype x :scalar)) data-set))
		 (native-datatype (get-datatype (first data-set)))
		 (max-value (convert-datatype (reduce #'max converted-values) native-datatype))
		 (min-value (convert-datatype (reduce #'min converted-values) native-datatype)))
	(make-uniform-distribution :hint :scalar :lower-bound min-value :upper-bound max-value)))

#+:matlisp
(defmethod learn-model ((distribution (eql :matlisp-uniform-distribution)) data-set &key ml-estimate?)
  (declare (ignore ml-estimate?))
  (let* ((converted-values (map 'list #'(lambda (x) (convert-datatype x :list)) data-set))
		 (native-datatype (get-datatype (first data-set)))
		 (max-value nil)
		 (min-value nil))
	(multiple-value-bind (maxes mins) (loop with maxes = (copy-list (first converted-values)) 
										 and mins = (copy-list (first converted-values)) 
										 and dims = (length (first converted-values))
										 for data in (cdr converted-values)
										 do (loop for idx from 0 to (- dims 1)
											   do (if (> (nth idx data) (nth idx maxes))
													  (setf (nth idx maxes) (nth idx data))
													  (if (< (nth idx data) (nth idx mins))
														  (setf (nth idx mins) (nth idx data)))))
										 finally (return (values maxes mins)))
	  (setf max-value (convert-datatype maxes native-datatype))
	  (setf min-value (convert-datatype mins native-datatype)))
	(make-uniform-distribution :hint :matlisp :lower-bound min-value :upper-bound max-value)))


(defmethod uniform-sample-from-domain ((domain discrete-domain))
  (with-slots (size elements) domain
	;;; Choose an index at random and return that element.
	(nth (random size) elements)))

(defmethod uniform-sample-from-domain ((domain scalar-continuous-bounded-domain))
  (with-slots (lower-bound upper-bound) domain
	(+ lower-bound (random (float (- upper-bound lower-bound))))))

#+:matlisp
(defmethod uniform-sample-from-domain ((domain matlisp-continuous-bounded-domain))
  (with-slots (dimensions lower-bound upper-bound) domain
	(let ((diff (m- upper-bound lower-bound)))
	  (m+ lower-bound (map-matrix #'random diff)))))

(defmethod print-object ((dist uniform-distribution) stream)
  (print-unreadable-object (dist stream :identity T :type T)
	(princ "Domain: " stream)
	(princ (slot-value dist 'domain) stream)))