;;; -*- Mode:Common-Lisp; Package:stochastic; Base:10 -*-

;;; Copyright (c) 2012 Eric M. Timmons, MIT.

;;; This software may not be redistributed, and can only be retained
;;; and used with the explicit written consent of the author,
;;; Eric M. Timmons.

;;; This software is made available as is; no guarantee is provided
;;; with respect to performance or correct behavior.

;;; This software may only be used for non-commercial, non-profit,
;;; research activities.

(in-package :stochastic)

;;; This file describes classes and functions for domains.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classes
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass continuous-domain (domain)
  ((dimensions :initform 0
			   :initarg :dims
			   :accessor domain-dimensions
			   :documentation "The number of dimensions in domain.")
   (native-datatype :initform nil
					:accessor domain-native-datatype))
  (:documentation "Variables over all real numbers."))

(defclass bounded-domain ()
  ((lower-bound :initform '()
				:accessor domain-lower-bound
				:documentation "The lower bound of a bounded domain.")
   
   (upper-bound :initform '()
				:accessor domain-upper-bound
				:documentation "The upper bound of a bounded domain."))
  (:documentation "The base class for a bounded domain."))

(defclass continuous-bounded-domain (bounded-domain continuous-domain)
  ((internal-datatype :initform nil))
  (:documentation "Base class for a bounded domain over a continuous space."))

(defclass scalar-continuous-bounded-domain (continuous-bounded-domain)
  ((lower-bound :initform nil
				:initarg :lower-bound)
   (upper-bound :initform nil
				:initarg :upper-bound)

   (internal-datatype :initform :scalar))
  (:documentation "Domain for a single, continuous, bounded domain."))

#+:matlisp
(defclass matlisp-continuous-bounded-domain (continuous-bounded-domain)
  ((lower-bound :initform nil
				:initarg :lower-bound)
   (upper-bound :initform nil
				:initarg :upper-bound)
   (internal-datatype :initform :matlisp))
  (:documentation "Domain for a continuous, bounded domain."))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Constructors
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod initialize-instance :after ((dist continuous-bounded-domain) 
									   &key lower-bound upper-bound)
  (with-slots ((lower lower-bound) (upper upper-bound) 
			   internal-datatype native-datatype dimensions) dist
	(setf native-datatype (get-datatype lower-bound))
	(setf dimensions (first (generic-size lower-bound)))
	(setf lower (convert-datatype lower-bound internal-datatype))
	(setf upper (convert-datatype upper-bound internal-datatype))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Methods
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod domain-member? ((domain continuous-domain) value)
  "Determine if VALUE is a member of DOMAIN.

   ========================================
   Preconditions
   ========================================
   None.

   ========================================
   Effects
   ========================================
   None.

   ========================================
   Returns
   ========================================
   T iff VALUE is a member of DOMAIN."
  (with-slots (dimensions) domain
	(let ((success? T))
	  (handler-case (setf success? (equal `(,dimensions 1) 
										  (generic-size value)))
		(error () (setf success? nil)))
	  (values success? (if success? :member :not-a-member)))))


(defmethod domain-member? ((domain scalar-continuous-bounded-domain) value)
  "Determine if VALUE is a member of DOMAIN.

   ========================================
   Preconditions
   ========================================
   None.

   ========================================
   Effects
   ========================================
   None.

   ========================================
   Returns
   ========================================
   T iff VALUE is a member of DOMAIN."
  (let ((value (convert-datatype value :scalar)))
	(with-slots (lower-bound upper-bound) domain
	  (and (<= lower-bound value)
		   (>= upper-bound value)))))

#+:matlisp
(defmethod domain-member? ((domain matlisp-continuous-bounded-domain) value)
  "Determine if VALUE is a member of DOMAIN.
   VALUE is part of the domain only if VALUE is a matlisp matrix.

   ========================================
   Preconditions
   ========================================
   None.

   ========================================
   Effects
   ========================================
   None.

   ========================================
   Returns
   ========================================
   T iff VALUE is a member of DOMAIN."
  (let ((value (convert-datatype value :matlisp)))
	(with-slots (lower-bound upper-bound dimensions) domain
	  (and (col-vector-p value)
		   (eq dimensions (number-of-rows value))
		   (loop for k from 0 to (- dimensions 1)
			  always (<= (matrix-ref lower-bound k) 
						 (matrix-ref value k) 
						 (matrix-ref upper-bound k)))))))

(defmacro make-continuous-domain (size)
  "Makes a CONTINUOUS-DOMAIN of SIZE."
  `(make-instance 'continuous-domain :dims ,size))

(defun make-continuous-bounded-domain (&key (hint nil hint-p) lower-bound upper-bound)
  "Function to automatically choose the correct implementation of
   CONTINUOUS-BOUNDED-DOMAIN given the lower and upper bounds. HINT can be
   used to tell the macro which implementation to use. HINT can be one of:
      :SCALAR
      :MULTI-VARIATE
      :MATLISP"
  (let ((dist-name (if hint-p
					   (case hint
						 (:scalar 'scalar-continuous-bounded-domain)
						 (:multivariate #+:matlisp 'matlisp-continuous-bounded-domain)
						 #+:matlisp (:matlisp 'matlisp-continuous-bounded-domain)
						 (T (error "Unrecognized hint: ~a." hint)))
					   (if (eql 1 (first (generic-size lower-bound)))
						   'scalar-continuous-bounded-domain
						   #+:matlisp 'matlisp-continuous-bounded-domain))))
	(make-instance dist-name :lower-bound lower-bound :upper-bound upper-bound)))

(defmethod domain-volume ((domain scalar-continuous-bounded-domain))
  (with-slots (lower-bound upper-bound) domain
	(- upper-bound lower-bound)))

#+:matlisp
(defmethod domain-volume ((domain matlisp-continuous-bounded-domain))
  (with-slots (lower-bound upper-bound dimensions) domain
	(reduce #'* (convert-to-lisp-array (m- upper-bound lower-bound)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Print Methods
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmethod print-object ((domain continuous-domain) stream)
  (print-unreadable-object (domain stream :identity T :type T)
	(princ "Dimensions: " stream)
	(princ (domain-dimensions domain) stream)))

(defmethod print-object ((domain continuous-bounded-domain) stream)
  (print-unreadable-object (domain stream :identity T :type T)
	(princ "Lower Bound: " stream)
	(princ (domain-lower-bound domain) stream)
	(princ " Upper Bound: " stream)
	(princ (domain-upper-bound domain) stream)))