;;;; Copyright (c) 2013 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Pedro H. Santana

(in-package :stochastic)

;;; ***********************************************************************
;;;
;;;   ***  Data structures and functions for probabilistic inference  ***
;;; 
;;; ***********************************************************************


(defun joint-to-list (cond-prob)
  "Takes a conditional probability with several query assignments and converts into a list
  of single assignment conditional probability with the same conditioning."
  (let ((single-var-cond-prob-list nil))

	(loop for assig in (get-assignment-list cond-prob) do
		 (push (make-cond-prob (list assig) (get-conditional-list cond-prob)) single-var-cond-prob-list))

	single-var-cond-prob-list))


(defun has-double-assignment? (prob-query)
  "Efficiently detects if any RV in the list been 
  assigned two values at the same time."
  (let (assig-rvs (assig-set nil))

	(setf assig-rvs (get-assig-rvs-from-cond-prob prob-query)) 	
	
	(loop for rv in assig-rvs do	 
		 
		  (if (member rv assig-set :test #'equal)
	 	  	 (return-from has-double-assignment? t)
	 	  	 (push rv assig-set)))
	nil))



(defmethod compute-joint-probability ((prob-model probability-model) (prob-query conditional-probability))
  "Takes in a probability model and a generic conditional probability and returns the desired 
  conditional probability."
  
  ;;Removes duplicate assignments to the same variables from the query
  (setf (get-assignment-list prob-query) (remove-duplicates (get-assignment-list prob-query) :test #'equal))
  
  ;;If we still have more than one assignment to the same RV, it means that
  ;;it has been assigned two different values at the same time. Therefore, the
  ;;joint probability is necessarily 0.0.  
   (if (has-double-assignment? prob-query)
   	  (return-from compute-joint-probability 0.0))
  
  ;;From this point on, we now that the query has no duplicate or inconsistent
  ;;assignments.

  ;(format t "~%****WARNING: this function currently only implements the case where all query variables are conditionally independent.~%")
  
  ;;This should be replaced by something more general in the future.
  (compute-joint-probability-all-conditionally-independent prob-model prob-query))



(defmethod compute-joint-probability-all-conditionally-independent ((prob-model probability-model) (prob-query conditional-probability))
  "Computes the value of a conditional probability assuming that all
  events (assignments to RVs) are independent."
  
  (let ((joint-prob 1) prob cond-prob-list)
	
	;;Breaks the joint query into a list of single variable queries with the
	;;same conditioning.
	(setf cond-prob-list (joint-to-list prob-query))
		
	;; For every assignment 
	(loop for cond-prob in cond-prob-list do
	   ;;Tries to get probability from the model
		 (setf prob (get-prob-value-from-model prob-model cond-prob)) 
	   
		 (unless prob
		   (format t "~% Couldn't find ~a" (cond-prob-to-string cond-prob))
		   (return-from compute-joint-probability-all-conditionally-independent nil))

	   ;;Multiplies all probabilities together
		 (setf joint-prob (* joint-prob prob)))

	;;Returns the joint probability of the independent variables
	joint-prob))
