;;;; Copyright (c) 2013 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Pedro H. Santana

(in-package :stochastic)

;;; ***********************************************************************
;;;
;;;     ***  Data structures for representing probabilistic models  ***
;;; 
;;; ***********************************************************************

;;; ***********************************************************************
;;;
;;;     ***  Classes ***
;;; 
;;; ***********************************************************************

(defclass probability-model ()
  ((random-variables
	:initarg :random-variables
	:initform nil
	:accessor get-random-variables
	:documentation "List of random variables that form this probability model.")
   (size
	:initarg :size
	:initform 0
	:accessor get-size
	:documentation "Number of conditional probabilities that form the model.")
   (conditional-prob
	:initarg :conditional-prob
	:initform nil
	:accessor get-conditional-prob
	:documentation "Conditional probabilities that form a general probability model.")
   ))

(defclass probability-model-ht (probability-model)
  ((conditional-prob
	:initarg :conditional-prob-ht
	:initform (make-hash-table :test #'equal)
	:accessor get-conditional-prob-ht
	:documentation "Implementation of a probability model that uses multiple hash tables to store the values.")))


(defclass conditional-probability ()
  ((assignment-list
	:initarg :assignment-list
	:initform nil
	:accessor get-assignment-list
	:documentation "List (conjunction) of assignments 'on the left' of the conditional probability."
	)
   (conditional-list	
	:initarg :conditional-list
	:initform nil
	:accessor get-conditional-list
	:documentation "List (conjunction) of conditional assignments")))


(defmethod print-object ((prob-model probability-model) stream)
  (print-unreadable-object (prob-model stream :identity T :type T)
	(with-slots (size random-variables conditional-prob) prob-model
	  (format stream "~%:size= ~a ~%:random-variables=~%~a" size random-variables)
	  (format stream "~%:conditional-probabilities=~% ~a"
			  (maphash #'(lambda (k v) (format t "~a => ~a~%" k v)) conditional-prob)))))


(defmacro assignment-to-string (rv-assig)
  "Converts an RV assignment into a string in an aesthetically pleasing way."
  `(format nil "(~a ~a)" (get-name (get-rv-from-assignment ,rv-assig))  (get-value-from-assignment ,rv-assig)))


(defun assignment-list-to-string (rv-assig-lst &key (separator ""))
  "Converts a list of assigments to RVs into a string."
  (let ((str nil))
	(loop with count = 1 for assig in rv-assig-lst do
		 (if (= count (length rv-assig-lst))
			 (setf separator ""))
		 (setf str (concatenate 'string str (format nil "~a" (assignment-to-string assig)) separator))
		 (incf count))	
	str))


(defun cond-prob-to-string (cond-prob &key (separator ""))
  "Converts a conditional probability into a string in an aesthetically pleasing way."
  (let (str)
	(setf str (format nil "Pr[~a" (assignment-list-to-string (get-assignment-list cond-prob) :separator separator)))		
	(when (conditional-p cond-prob)
	  (setf str (concatenate 'string str "|" (assignment-list-to-string (get-conditional-list cond-prob)))))	
	(concatenate 'string str "]")))

;;; ***********************************************************************
;;;
;;;     ***  Functions and Macros ***
;;; 
;;; ***********************************************************************

(defun make-cond-prob (lst-assignment lst-conditional)
  "Creates a conditional probability object."
  (make-instance 'conditional-probability 
				  :assignment-list lst-assignment  :conditional-list lst-conditional))

(defun make-uncond-prob (lst-assignment)
  "Syntatic sugar for creating unconditional probability objects."
  (make-instance 'conditional-probability 
				  :assignment-list lst-assignment  :conditional-list nil))


(defmacro get-rvs-from-assignment-list (assig-list)
  "Get the RVs in a list of assignments."
  `(loop for assig in ,assig-list collecting (get-rv-from-assignment assig)))

(defmacro get-values-from-assignment-list (assig-list)
  "Get the values assigned to RVs in a list of assignments."
  `(loop for assig in ,assig-list collecting (get-value-from-assignment assig)))


(defmacro get-assig-rvs-from-cond-prob (cond-prob)
  "Gets the random variables involved in the assignment list."
  `(get-rvs-from-assignment-list (get-assignment-list ,cond-prob)))

(defmacro get-cond-rvs-from-cond-prob (cond-prob)
  "Gets the random variables involved in the conditioning list."
  `(get-rvs-from-assignment-list (get-conditional-list ,cond-prob)))


(defmacro get-assig-values-from-cond-prob (cond-prob)
  "Gets the value assigned to the random variables not involved in the conditioning list."
  `(get-values-from-assignment-list (get-assignment-list ,cond-prob)))

(defmacro get-cond-values-from-cond-prob (cond-prob)
  "Gets the values assigned to the random variables in the conditioning list."
  `(get-values-from-assignment-list (get-conditional-list ,cond-prob)))


(defmacro conditional-p (cond-prob)
  "Determines if a conditional probability is has an empty
  conditioning list or not."
  `(if (get-conditional-list ,cond-prob)
	   t
	   nil))


;; (defmacro get-domain-from-cond-prob (cond-prob)
;;   "Gets the domain object for the RV not involved in the conditioning."
;;   `(get-domain (get-rv-from-cond-prob ,cond-prob)))

;; (defmacro get-domain-elements-from-cond-prob (cond-prob)
;;   "Gets the domain elements for the RV not involved in the conditioning."
;;   `(domain-elements (get-domain-from-cond-prob ,cond-prob)))



(defmacro parse-cond-prob (cond-prob)
  "Parses a conditional probability into a two element meta list.
   
   ( ((assig-rvs) (cond-rvs)) ((assig-values) (cond-values)) ).
   
   The purpose is to use each piece in order to efficiently index the
   conditional models on a hash table."
  `(list
 	(list (get-assig-rvs-from-cond-prob ,cond-prob) (get-cond-rvs-from-cond-prob ,cond-prob)) ;RVs
	(list (get-assig-values-from-cond-prob ,cond-prob) (get-cond-values-from-cond-prob ,cond-prob)))) ;Values

(defmacro get-rvs-from-parsed (parsed-cond-prob) 
  `(first ,parsed-cond-prob))

(defmacro get-values-from-parsed (parsed-cond-prob)
   `(second ,parsed-cond-prob))

(defmacro get-assig-rvs-from-parsed (parsed-cond-prob)
  `(first (get-rvs-from-parsed ,parsed-cond-prob)))

(defmacro get-cond-rvs-from-parsed (parsed-cond-prob)
  `(second (get-rvs-from-parsed ,parsed-cond-prob)))

(defmacro get-assig-values-from-parsed (parsed-cond-prob)
  `(first (get-values-from-parsed ,parsed-cond-prob)))

(defmacro get-cond-values-from-parsed (parsed-cond-prob)
  `(second (get-values-from-parsed ,parsed-cond-prob)))






(defmacro make-prob-model ()
  "Creates an empty probability model that uses hash tables for storage."
  `(make-instance 'probability-model-ht))

(defmacro add-rvs-to-model (rvs-list prob-model)
  "Adds a list of random variables to the set of random variables
  in the probability model."
  `(loop for rv in (concatenate 'list (first ,rvs-list) (second ,rvs-list)) do
		(pushnew rv (get-random-variables ,prob-model))))



(defmacro make-ht ()
  "Creates a Hash Table that can use strings as keys. Should
  be moved elsewhere (but where???)"
  `(make-hash-table :test #'equal))

(defmacro read-from-ht (key table)
  "Syntatic sugar for reading from a Hash Table (HT). Should
  be moved elsewhere (but where???)."
  `(gethash ,key ,table))

(defmacro write-to-ht (key value table)
  "Syntatic sugar for writing to a Hash Table (HT). Should
  be moved elsewhere (but where???)."
  `(setf (gethash ,key ,table) ,value))




;;; ***********************************************************************
;;;
;;;     ***  Functions and Methods ***
;;; 
;;; ***********************************************************************


(defmethod add-to-prob-model ((prob-model probability-model-ht) cond-prob prob-value)
  "Adds elements to a probability model constructed as a series of hash tables." 
  (let (parsed all-cond-ht a-cond-ht a-cond-values-ht)
	
	(setf parsed (parse-cond-prob cond-prob)) ;first=RVs, secod=values

	;;All currently available conditional probability models
	(setf all-cond-ht (get-conditional-prob-ht prob-model))
	
	;;Creates a new empty hash table if this conditional probability has never
	;;been added to the model before.
	(unless (read-from-ht (get-rvs-from-parsed parsed) all-cond-ht) 
	  (write-to-ht (get-rvs-from-parsed parsed) (make-ht) all-cond-ht))
	
	;;The hash table corresponding to the specific conditional model being
	;;added.
	(setf a-cond-ht (read-from-ht (get-rvs-from-parsed parsed) all-cond-ht))

	;;Creates a new empty hash table if this particular assignment to the
	;;conditional probability has never bee added to the model before.
	(unless (read-from-ht (get-cond-values-from-parsed parsed) a-cond-ht) 
	  (write-to-ht (get-cond-values-from-parsed parsed) (make-ht) a-cond-ht))

	;;The hash table corresponding to the specific assignment to the conditional 
	;;model being added.
	(setf a-cond-values-ht (read-from-ht (get-cond-values-from-parsed parsed) a-cond-ht))

	;;FINALLY stores the probabilitiy value on the HT corresponding to the specific
	;;assignment to the conditioning variables. This overwrites any previously-
	;;written value.
	(write-to-ht (get-assig-values-from-parsed parsed) prob-value a-cond-values-ht)
	
	;;Increments the number of conditional probabilities in the model
	(incf (get-size prob-model))

	;;Adds the random variables to the probability model
	(add-rvs-to-model (get-rvs-from-parsed parsed) prob-model)))



(defmethod get-prob-value-from-model ((prob-model probability-model-ht) cond-prob)
  "Retrieves the probability value stored in the model."
  (let (parsed all-cond-ht a-cond-ht a-cond-values-ht)
	
	(setf parsed (parse-cond-prob cond-prob)) ;first=RVs, secod=values

	;;All currently available conditional probability models
	(setf all-cond-ht (get-conditional-prob-ht prob-model))
	
	;;No conditional probability model is present.
	(unless (read-from-ht (get-rvs-from-parsed parsed) all-cond-ht) 
	  (return-from get-prob-value-from-model nil))

	;;The hash table corresponding to the specific conditional model being
	;;queried.
	(setf a-cond-ht (read-from-ht (get-rvs-from-parsed parsed) all-cond-ht))

	;;No value associated with this particular assignment to the conditional
	;;RVs.
	(unless (read-from-ht (get-cond-values-from-parsed parsed) a-cond-ht) 
	  (return-from get-prob-value-from-model nil))

	;;The hash table corresponding to the specific assignment to the conditional 
	;;model being added.
	(setf a-cond-values-ht (read-from-ht (get-cond-values-from-parsed parsed) a-cond-ht))

	;;FINALLY reads the probabilitiy value on the HT corresponding to the specific
	;;assignment to the conditioning variables.
	(read-from-ht (get-assig-values-from-parsed parsed) a-cond-values-ht)))



(defmethod check-sanity-prob-model ((prob-model probability-model-ht))
  "Checks if a conditional probability model implemented with hash tables
   is sane, i.e., all of its conditional probability models are well-formed."
  ;; (let (sum cond-count domain-elements rv rv-value random-variable index)
	
  ;; 	;; We should have one probability value per assignment
  ;; 	(unless (= (length lst-cond-probabilities) (length lst-prob-values))
  ;; 	  (return-from check-sanity-prob-models nil))

  ;; 	;;Probability values need to sum to one
  ;; 	(setf sum (loop for i in lst-prob-values summing i))
  ;; 	(unless (= sum 1)	  
  ;; 	  (return-from check-sanity-prob-models nil))

  ;; 	;;All models should have the same conditional list
  ;; 	(setf cond-count
  ;; 		  (count 
  ;; 		   (get-conditional-list (first lst-cond-probabilities)) lst-cond-probabilities :test #'equal))


  ;; 	(unless (= cond-count (length lst-cond-probabilities))
  ;; 	  (return-from check-sanity-prob-models nil))
	

  ;; 	;; The assignments should be all for the same random variable 
  ;; 	;; and contain all elements in its domain
  ;; 	(setf domain-elements (get-domain-elements-from-cond-prob (first lst-cond-probabilities)))
  ;; 	(setf random-variable (get-rv-from-cond-prob (first lst-cond-probabilities)))

  ;; 	(loop for cond-prob in lst-cond-probabilities do
  ;; 		 (setf rv (get-rv-from-cond-prob cond-prob)) ;current rv
  ;; 		 (setf rv-value (get-value-from-cond-prob cond-prob)) ;assigned value

  ;; 		 (unless (equal rv random-variable) ;different RVs being assigned
  ;; 		   (return-from check-sanity-prob-models nil))
		 
  ;; 		 ;;Searches for assignment value on the domain elements. If not found,
  ;; 		 ;;return an error. Otherwise, remove it from the list.
  ;; 		 (setf index (position rv-value domain-elements))
  ;; 		 (unless index
  ;; 		   (return-from check-sanity-prob-models nil))		 
  ;; 		 (setf domain-elements (remove rv-value domain-elements :start index :count 1)))
  	
  ;; 	;;Checks if all domain elements have been covered
  ;; 	(unless (= (length domain-elements) 0)
  ;; 	  (return-from check-sanity-prob-models nil)))

  ;; Model passed all sanity checks
  t)


