;;;; Copyright (c) 2013 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Pedro H. Santana

(in-package :stochastic)

(defun test-discrete-union-prob ()
  (let (A B C D E F dom (A-cond-B nil) (B-C-uncond nil) (D-uncond nil) (E-uncond nil) (F-uncond nil)
		  A-cond-B-probs B-C-uncond-probs D-uncond-probs E-uncond-probs F-uncond-probs
		  prob-model cond-probs prob-values prob-query)
	
	;; Creates discrete domain
	(setf dom (make-discrete-domain '(0 1)))

	;; Creates a number of binary random variables
	(setf A (make-discrete-rv "A" dom))
	(setf B (make-discrete-rv "B" dom))
	(setf C (make-discrete-rv "C" dom))
	(setf D (make-discrete-rv "D" dom))
	(setf E (make-discrete-rv "E" dom))
	(setf F (make-discrete-rv "F" dom))

	;; Hand-crafted probability models
	(push (make-cond-prob `((,A 0))  `((,B 0))) A-cond-B)
	(push (make-cond-prob `((,A 1))  `((,B 0))) A-cond-B)
	(push (make-cond-prob `((,A 0))  `((,B 1))) A-cond-B)
	(push (make-cond-prob `((,A 1))  `((,B 1))) A-cond-B)	
	(setf A-cond-B (reverse A-cond-B))

	(push (make-uncond-prob `((,B 0) (,C 0))) B-C-uncond)
	(push (make-uncond-prob `((,B 0) (,C 1))) B-C-uncond)
	(push (make-uncond-prob `((,B 1) (,C 0))) B-C-uncond)
	(push (make-uncond-prob `((,B 1) (,C 1))) B-C-uncond)
	(setf B-C-uncond (reverse B-C-uncond))
	
	(push (make-uncond-prob `((,D 0))) D-uncond)
	(push (make-uncond-prob `((,D 1))) D-uncond)
	(setf D-uncond (reverse D-uncond))

	(push (make-uncond-prob `((,E 0))) E-uncond)
	(push (make-uncond-prob `((,E 1))) E-uncond)
	(setf E-uncond (reverse E-uncond))

	(push (make-uncond-prob `((,F 0))) F-uncond)
	(push (make-uncond-prob `((,F 1))) F-uncond)
	(setf F-uncond (reverse F-uncond))

	;;Probability values (use double precision DOES make a difference)
	(setf A-cond-B-probs (list 0.2d0 0.8d0 0.1d0 0.9d0))
	(setf B-C-uncond-probs (list 0.4d0 0.6d0 0.3d0 0.7d0))
	(setf D-uncond-probs (list 0.45d0 0.55d0))
	(setf E-uncond-probs (list 0.1d0 0.9d0))
	(setf F-uncond-probs (list 0.8d0 0.2d0))
		
	;;Concatenation of all models
	(setf cond-probs (append A-cond-B B-C-uncond D-uncond E-uncond F-uncond))
	(setf prob-values (append A-cond-B-probs B-C-uncond-probs D-uncond-probs E-uncond-probs F-uncond-probs))

	;;Empty probability model
	(setf prob-model (make-prob-model))
	
	;;Populates the model
	(loop for i from 0 to (- (length cond-probs) 1) do
	 	 (add-to-prob-model prob-model (elt cond-probs i) (elt prob-values i)))

	;;Prints the model on the screen so that we can see whether it has been
	;;correctly stored. 
	(cond
	  ;;The sanity checker still has to be implemented!
	  ((check-sanity-prob-model prob-model)
	   ;;Retrieves the probabilities that have just been added
	   (format t "The probability model is sane. Here it is:~%")	   
	   (loop for cp in cond-probs do		 
			(format t "~%~a = ~a" (cond-prob-to-string cp) (get-prob-value-from-model prob-model cp))))
	  (t
	   (format t "Malformed probability model.")))
	
	;;Some example probabilistic queries
	(setf prob-query (make-uncond-prob (list (make-assignment-rv D 0) (make-assignment-rv D 1)) ))
	(format t "~%~%Query: ~a=~a" (cond-prob-to-string prob-query :separator "^") (compute-joint-probability prob-model prob-query))

	(setf prob-query (make-uncond-prob (list (make-assignment-rv D 0) (make-assignment-rv E 1)) ))
	(format t "~%~%Query: ~a=~a" (cond-prob-to-string prob-query :separator "v") (prob-union prob-model prob-query))

	(setf prob-query (make-uncond-prob (list (make-assignment-rv D 1) (make-assignment-rv E 0)) ))
	(format t "~%~%Query: ~a=~a" (cond-prob-to-string prob-query :separator "v") (prob-union prob-model prob-query))

	(setf prob-query (make-uncond-prob (list (make-assignment-rv D 1) (make-assignment-rv E 0) (make-assignment-rv E 1)) ))
	(format t "~%~%Query: ~a=~a" (cond-prob-to-string prob-query :separator "v") (prob-union prob-model prob-query))

	(setf prob-query (make-uncond-prob (list (make-assignment-rv D 0) (make-assignment-rv E 0) (make-assignment-rv F 0)) ))
	(format t "~%~%Query: ~a=~a" (cond-prob-to-string prob-query :separator "v") (prob-union prob-model prob-query))

	;;Returns the model for inspection purposes
	;;prob-model	
))