;;;; Copyright (c) 2012 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Andrew Wang and Pedro Santana

(in-package #:stochastic)

;;; Pedro: I would argue that nondeterministic variables shouldn't 'contain' a sampling function.
;;; In fact, I would say that they don't even have to necessarily contain a model, since we might
;;; not have information about a random variable directly. For example, let's say we have P(A|B) 
;;; and P(B). From this information, we can derive a model for P(A) through marginalization, which
;;; might be a pretty complicated task by itself. In my understanding, a random variable should only
;;; have a type (discrete, continuous, mixed) and a domain description. The actual probability values
;;; should be part of a probabilistic model, where the same assignment to an RV might have different
;;; probabilities depending on the model. What do you think?


(defclass random-variable ()
  ((name
	:documentation "Human-readable name of the random variable."	
	:initarg :name
	:initform nil
	:accessor get-name
	)   
   (domain
	:documentation "The list of symbols that can be assigned to this random variable"
	:type domain
	:initform nil
	:initarg :domain
	:accessor get-domain
   )))

(defclass discrete-random-variable (random-variable)
  ((domain
	:documentation "A discrete list of symbols"
	:type discrete-domain   
   )))

(defmethod print-object ((rv random-variable) stream)
  (print-unreadable-object (rv stream :identity T :type T)
	(with-slots (name domain) rv
	  (format stream "~%:name= ~a~%:domain= ~a" name domain))))


(defun make-discrete-rv (name domain)
  "Syntatic sugar for creating discrete random variable"
  (make-instance 'discrete-random-variable :name name :domain domain))

(defmethod make-assignment-rv((var random-variable) value)
  "Creates valid assignments to random variables according to their domains."
  (cond 
	((domain-member? (get-domain var) value)
	 (list var value))
	(t
	 (format t "Value ~a is not part of the domain of ~a." value (get-name var))
	 nil)))

(defmacro get-rv-from-assignment (rv-assignment)
  `(first ,rv-assignment))

(defmacro get-value-from-assignment (rv-assignment)
  `(second ,rv-assignment))