;;;; Copyright (c) 2011 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Seung H. Chung, David C. Wang

(in-package #:tburton)

(defmethod plan ((qsp mtk-qsp:qsp) cca s_0 &optional (problem-name nil))
  "Returns a qualitative control plan (QCP) that is generated using the
constraint and causal order decomposition.
;;;   I. Initialize
;;;      1. eQCP <- InsertInitialState(QSP,s(0));
;;;      2. s(0:t) <- null;
;;;      3. newGoals <- null;
;;;  II. Decompose CCA
;;;      1. R^{DD} <- ComputeCausalConstraintDecomposition(CCA);
;;; III. While and open goal exists
;;;      while openGoals(eQCP) != null | newGoals != null do
;;;      1. Select an open goal
;;;         a. g <- ChooseOpenGoal(eQCP):
;;;         b. c <- getComponent(g);
;;;         c. s(0:t) <- chooseTrajectory(s0(c),g);
;;;         d. if s(0:t) = null then
;;;         e.   return failed;
;;;      2. Check consistency
;;;      3. Update the plan
;;;         a. eQCP <- InsertTrajectory(eQCP,s0(c),s(0:t));
;;;         b. s0(c) <- s(t);
;;;      4. Extract new subgoals
;;;         a. newGoals <- ChooseSubgoals(s(0:t));
;;;  IV. Return QCP = proj_{X^c}(eQCP);
"  
;;  (mtk-graph2dot:dot-graph qsp "test-qsp.ps")

;;  (format t "Initial State: ~A~%" s_0)
;;;   I. Initialize
  (let* ((i 0)
;;;      3. newGoals <- null;
	 ;;(newGoals nil)
	 (goals (mtk-qsp:get-edges-with-state-constraint qsp))
;;;      1. eQCP <- InsertInitialState(QSP,s(0));
	 (s0-edge-ht-component (make-hash-table :test #'equal))
	 (eqcp (insert-initial-state s_0 qsp s0-edge-ht-component))
	 (eqcp-transitive-closure (mtk-graph-algorithms:transitive-closure eqcp))
;;;      2. s(0:t) <- null;
	 ;;(s[0-t] nil)
;;;  II. Decompose CCA
;;;      1. R^{DD} <- ComputeCausalConstraintDecomposition(CCA);
	 (decomposition (compute-constraint-causal-order-decomposition CCA))
	 (transitive-closure-causal-graph (mtk-graph-algorithms:transitive-closure
					   decomposition))
	 (component-ht-variable (make-hash-table :test #'equal))
	 (is-better-goal-function
	  #'(lambda (e_i e_j)
		  ;; A component is the name of the variable clusters formed
		  ;; by making the component graph.
	      (let ((c_i (get-component (get-state-constraint e_i eqcp)
					component-ht-variable))
		    (c_j (get-component (get-state-constraint e_j eqcp)
					component-ht-variable)))
		(if (equal c_i c_j)
		    (funcall eqcp-transitive-closure
			     (mtk-graph:get-source-vertex e_i)
			     (mtk-graph:get-target-vertex e_j))
		    (funcall transitive-closure-causal-graph
			     c_j c_i)))))
	 )
    ;; For now a decomposition is simply a strongly connected component graph
    ;; of the causal graph of the cca. This information can be used to 
    ;; generate a hash table that maps a variable to its component.
;;    (mtk-graph2dot:dot-print-graph decomposition)
    (dolist (component (mtk-graph:get-vertices decomposition))
      (dolist (variable component)
;;	(format t "variable ~A component ~A~%" variable component)
	(setf (gethash variable component-ht-variable) component)))
    ;; Ideally i should use a priority queue here.
    ;; Should sort the goals here.
    (setf goals 
	  (sort (copy-list goals)
		#'(lambda (e_i e_j)
		    (let ((c_i (get-component (get-state-constraint e_i eqcp)
					      component-ht-variable))
			  (c_j (get-component (get-state-constraint e_j eqcp)
					      component-ht-variable)))
		      (if (equal c_i c_j)
			  (funcall eqcp-transitive-closure
				   (mtk-graph:get-source-vertex e_i)
				   (mtk-graph:get-target-vertex e_j))
			  (funcall transitive-closure-causal-graph
				   c_j c_i))))))
;; (format t "Goals: ~A~%"
;; (mapcar #'(lambda (g)
;; 		(mtk-csp:get-relation (get-state-constraint g eqcp)))
;; 	    goals))

;;; III. While and open goal exists
;;;      while openGoals(eQCP) != null | newGoals != null do
				(loop
				   while goals
				   do
;;;      1. Select an open goal
				   (let* (
;;;         a. g <- ChooseOpenGoal(eQCP):
						  (g-edge (pop goals))
						  (g (get-state-constraint g-edge eqcp))
;;;         b. c <- getComponent(g);
						  (c (get-component g component-ht-variable))
;;;         c. s(0:t) <- chooseTrajectory(s0(c),g);
						  (s[0]-edge (gethash c s0-edge-ht-component))
						  (s[0] (get-state-constraint s[0]-edge eqcp))
						  (s[0-t] (choose-trajectory c s[0] g cca))
						  (subgoal-edges nil))
;;					 (format t "Trajectory: ~A~%" s[0-t])
;;;         d. if s(0:t) = null then
					 (if (null s[0-t])
;;;         e.   return failed;
						 (return nil))
;;;      2. Check consistency
;;;      3. Update the plan
;;;         a. eQCP <- InsertTrajectory(eQCP,s0(c),s(0:t));
					 (multiple-value-setq (s[0]-edge subgoal-edges)
					   (insert-trajectory s[0-t] eqcp s[0]-edge g-edge
										  component-ht-variable cca))
;;;         b. s0(c) <- s(t);
					 (setf (gethash c s0-edge-ht-component) s[0]-edge)
;;;      4. Extract new subgoals
;;;         a. newGoals <- ChooseSubgoals(s(0:t));
					 (incf i)
					 ;; (mtk-graph2dot:dot-graph eqcp
					 ;; 							(concatenate 'string 
					 ;; 										 "test" (format nil "~A" i) ".ps"))
					 (setf goals (append goals subgoal-edges))
					 (setf eqcp-transitive-closure
						   (mtk-graph-algorithms:transitive-closure eqcp))
					 (setf is-better-goal-function
						   #'(lambda (e_i e_j)
							   (let ((c_i (get-component (get-state-constraint e_i eqcp)
														 component-ht-variable))
									 (c_j (get-component (get-state-constraint e_j eqcp)
														 component-ht-variable)))
								 (if (equal c_i c_j)
									 (funcall eqcp-transitive-closure
											  (mtk-graph:get-source-vertex e_i)
											  (mtk-graph:get-target-vertex e_j))
									 (funcall transitive-closure-causal-graph
											  c_j c_i)))))
					 ;;	   (setf goals (sort-goals goals is-better-goal-function))
					 (setf goals (sort (copy-list goals) is-better-goal-function))

					 ;; 	   (let ((trans-g (make-instance 'mtk-graph:directed-graph)))
					 ;; 	     (mtk-graph:add-edges 
					 ;; 	      (remove-duplicates
					 ;; 	       (loop
					 ;; 		  for k being the hash-keys in eqcp-transitive-closure-ht
					 ;; 		  using (hash-value v)
					 ;; 		  collect (if v k nil))
					 ;; 	       :test #'equal)
					 ;; 	      trans-g)
					 ;; 	     (mtk-graph2dot:dot-graph
					 ;; 	      trans-g
					 ;; 	      (concatenate 'string 
					 ;; 			   "test" (format nil "~A" i) "trans.ps")))
					 ;; 	   (format t "Goals: ~A~%"
					 ;;  		   (mapcar #'(lambda (g)
					 ;;  			       (mtk-csp:get-relation
					 ;;  				(get-state-constraint g eqcp)))
					 ;;  			   goals))
					 ;; 	   (let ((e_i (first goals)))
					 ;; 	     (dolist (e_j (cdr goals))
					 ;; 	       (let ((c_i (get-component (get-state-constraint e_i eqcp)
					 ;; 					 component-ht-variable))
					 ;; 		     (c_j (get-component (get-state-constraint e_j eqcp)
					 ;; 					 component-ht-variable)))
					 ;; 		 (format t "** ~A ~A -> ~A ~A: "
					 ;; 			 (mtk-graph:get-vertices e_i) 
					 ;; 			 (mtk-csp:get-relation (get-state-constraint e_i eqcp)) 
					 ;; 			 (mtk-graph::get-vertices e_j)
					 ;; 			 (mtk-csp:get-relation (get-state-constraint e_j eqcp)) )
					 ;; 		 (format t "same component? ~A " (equal c_i c_j))
					 ;; 		 (format t "causally before? ~A "
					 ;; 			 (funcall transitive-closure-causal-graph c_j c_i))
					 ;; 		 (format t "temporally before? ~A~%"
					 ;; 			 (funcall eqcp-transitive-closure
					 ;; 				  (mtk-graph:get-source-vertex e_i)
					 ;; 				  (mtk-graph:get-target-vertex e_j)))
					 ;; 		 (format t "temporally later? ~A~%"
					 ;; 			 (funcall eqcp-transitive-closure
					 ;; 				  (mtk-graph:get-source-vertex e_j)
					 ;; 				  (mtk-graph:get-target-vertex e_i))))))
					 ))
				;;	   (setq newGoals (ChooseSubgoals s[0-t]))))
;;;  IV. Return QCP = proj_{X^c}(eQCP);
				(values (project eqcp) eqcp)
				))

(defmethod insert-initial-state ((s0 list)
				 (qsp qsp)
				 (s0-edge-ht-component hash-table))
  (dolist (s0_i s0)
    (let* ((c_i (second s0_i))
	   (e_i (add-constraint `(:scope
				  (0 (,c_i 0))
				  :relation 
				  ((:lower-bound (0) :upper-bound (,*inf*)))
				  :state-constraint
				  (:scope (,c_i) :relation (,s0_i)))
				qsp)))
;;      (format t "Adding edge ~A with hash key ~A~%" e_i `(,c_i))
      (setf (gethash `(,c_i) s0-edge-ht-component) e_i)))
  qsp)

(defmethod sort-goals (goals is-better-function)
  (setf goals (sort (copy-list goals) is-better-function))
  (let ((unordered-goals nil)
	(goal (pop goals)))
    (loop
       while (and goals (not (funcall is-better-function goal (first goals))))
       do (if (funcall is-better-function (first goals) goal)
	      (sort-goals goals is-better-function)
	      (push (pop goals) unordered-goals)))
    (sort (copy-list (append (push goal goals) unordered-goals))
	  is-better-function)))
	 

;; For now, we only perform causal-order-decomposition.
;; For simplicity and testing, currently a cca is not really a cca.
;; A cca a struct that contains a slot for causal-graph.
(defmethod compute-constraint-causal-order-decomposition (cca)
  (mtk-graph-algorithms:generate-component-graph (cca-causal-graph cca)))

;; (defmethod opengoals(eQCP)
;;   nil)

(defmethod get-component ((g mtk-csp:constraint)
			  (component-ht-variable hash-table))
  ;; For now, a goal only contains a unary constraint. This means that
  ;; the length of the scope associated with the goal constraint should be one.
  (let ((scope (mtk-csp:get-scope g)))
    (assert (= (length scope) 1))
    (gethash (first scope) component-ht-variable)))

;; For simplicity and testing, currently a cca is not really a cca.
;; A cca a struct that contains a slot for a planner. Note that
;; the planner must be decomposed appropriately. In this case,
;; we assume only decomposed problem.
(defmethod choose-trajectory (component
			      (s0 mtk-csp:constraint)
			      (g mtk-csp:constraint)
			      cca)
;; (format t "Choosing a trajectory from ~A to ~A~%"
;; 		  (mtk-csp:get-relation s0) (mtk-csp:get-relation g))
  (let ((planning-function (cca-planning-function cca))
		(state-variables (cca-state-variables cca))
		;; Choose only first... we assume no disjunctive constraints.
		(from (first (mtk-csp:get-relation s0)))
		(to   (first (mtk-csp:get-relation g))))
    ;; If the goal is on a state variable
    (if (some #'(lambda (x) (member x state-variables)) component)
		;; Find a plan that achieves the goal
		(funcall planning-function from to component)
		;; Otherwise, the goal is on the control variable.
		(list (list nil
					`(= ,(first (mtk-csp:get-scope g)) no-cmd))
			  (list nil
					(first (mtk-csp:get-relation g)))))))

;; (defmethod chooseSubgoals(s_[0-t])
;;   nil)

(defmethod insert-trajectory (s[0-t] eqcp s[0]-edge g-edge component-ht-variable cca)
  ;;(format t "Current edge: ~S~%" s[0]-edge)
  ;;(format t "Trajectory s[0-t]: ~S~%" s[0-t])
  (let ((subgoal-edges nil)
	(state-variables (cca-state-variables cca)))
    (if (not (equal s[0-t] t))
	(dolist (a-s[i] s[0-t])
	  (let* ((v_source (mtk-graph:get-target-vertex s[0]-edge))
		 (v_source-name (first v_source))
		 (v_source-index (second v_source))
		 (v_target (list v_source-name (+ v_source-index 1)))
		 (constraint-description 
		  (list :scope
			`(,v_source ,v_target)
			:relation
			`((:lower-bound (0 :inequality ,#'>) :upper-bound (,*inf*)))
			:state-constraint
			(list :scope (list v_source-name)
			      :relation (list (second a-s[i]))))))
	    ;;(format t "Adding Constraint w/ Description: ~S~%" constraint-description)
	    (setf s[0]-edge (add-constraint constraint-description eqcp))
	    ;; Add subgoals
	    (dolist (state-constraint (first a-s[i]))
	      (let ((component (get-component state-constraint component-ht-variable))
		    (sub-source (gensym))
		    (sub-target (gensym))
		    (state-constraint-description
		     (list :scope (mtk-csp:get-scope state-constraint)
			   :relation (mtk-csp:get-relation state-constraint)))
		    (ub-description (list 0 :inequality #'<=)))
		(if (some #'(lambda (x) (member x state-variables)) component)
		    (setf ub-description (list *inf*)))
		(let ((sub-constraint-description
		       (list :scope
			     (list sub-source sub-target)
			     :relation
			     (list (list :lower-bound `(0 :inequality ,#'>)
					 :upper-bound `(,*inf*)))
			     :state-constraint
			     state-constraint-description))
		      (sub-source-constraint-description
		       (list :scope
			     (list sub-source v_source)
			     :relation
			     (list (list :lower-bound `(0)
					 :upper-bound ub-description))))
		      (sub-target-constraint-description
		       (list :scope
			     (list v_source sub-target)
			     :relation
			     (list (list :lower-bound `(0)
					 :upper-bound `(,*inf*))))))
		  (push (add-constraint sub-constraint-description eqcp) subgoal-edges)
		  (add-constraint sub-source-constraint-description eqcp)
		  (add-constraint sub-target-constraint-description eqcp)))))))
    ;; Add goal achievement temporal constraint.
    ;;   With metric temporal constraints, one two of the four edges are sufficient.
    ;;   For qualitative, however, four edges are necessary to represent the
    ;;   directionality of a the two [0,0] bounds.
    (add-constraint `(:scope
		      (,(mtk-graph:get-source-vertex g-edge)
			,(mtk-graph:get-source-vertex s[0]-edge))
		      :relation
		      ((:lower-bound (0) :upper-bound (0))))
		    eqcp)
     (add-constraint `(:scope
		      (,(mtk-graph:get-source-vertex s[0]-edge)
			,(mtk-graph:get-source-vertex g-edge))
		      :relation
		      ((:lower-bound (0) :upper-bound (0))))
		    eqcp)
    (add-constraint `(:scope
		      (,(mtk-graph:get-target-vertex g-edge)
			,(mtk-graph:get-target-vertex s[0]-edge))
		      :relation
		      ((:lower-bound (0) :upper-bound (0))))
		    eqcp)
    (add-constraint `(:scope
		      (,(mtk-graph:get-target-vertex s[0]-edge)
			,(mtk-graph:get-target-vertex g-edge))
		      :relation
		      ((:lower-bound (0) :upper-bound (0))))
		    eqcp)
    (values s[0]-edge subgoal-edges)))

(defmethod project (eqcp)
  eqcp)

(defun domain-to-cca (problem-symbol)
  (values (make-instance 'qsp))
  (let* ((qsp (make-qsp-bus))
	 (causal-graph (make-causal-graph-bus))
	 (policy-function-ht-component (make-hash-table :test #'equal))
	 (state-variables '(mtk-qsp::bus))
	 (s0 '((= mtk-qsp::bus mtk-qsp::off)
	       (= mtk-qsp::bus-cmd mtk-qsp::no-cmd))))
    (setf (gethash '(mtk-qsp::bus) policy-function-ht-component)
	  (make-policy-function #'policy-description-bus))
    (let* ((planning-function (make-planning-function policy-function-ht-component))
	   (cca (make-cca :state-variables state-variables
			  :causal-graph causal-graph
			  :planning-function planning-function)))
	  (values qsp cca s0))))