;;;; Copyright (c) 2014 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   Steve Levine

(in-package :tburton)

(defun schedule-offline (plan)
  (let* ((episodes (tpn::tpn-episodes plan))
		 (g (make-instance 'mtk-graph:weighted-directed-graph))
		 (actions nil)
		 (plan-event-start (event-to-symbol-fn (tpn::tpn-start plan))) ;; Starting event in the action graph
		 (events (loop for e in (tpn::tp-events plan) 
					collecting (event-to-symbol-fn e)))
		 (action-schedule nil)

		 ;; For the optimized version
		 (event-ht-int (make-hash-table :test #'equalp))
		 (N 0)
		 (d nil)
		 (edges nil)
		 )

	;; Next, we will compute an APSP using all of our events, based on the temporal constraints of the graph. 
	;; First, we construct a distance graph in g. While we're looping, we also build up a list of actions.
	(loop for e in events do (mtk-graph:add-vertex e g)) ;; Add each action to the graph

	(loop for e in events do
		 (setf (gethash e event-ht-int)
			   N)
		 (incf N))


	(loop for episode in episodes do
		 (let* ((lb (to-pinf (tpn::episode-lowerbound episode)))
				(ub (to-pinf (tpn::episode-upperbound episode)))
				(e_start (event-to-symbol-fn (tpn::episode-start episode)))
				(e_end (event-to-symbol-fn (tpn::episode-end episode)))
				(primitive (tpn::episode-primitive episode)))
		   (mtk-graph:add-weighted-edge (list (list e_start e_end) ub) g)
		   (mtk-graph:add-weighted-edge (list (list e_end e_start) (- lb)) g)

		   (push (list (gethash e_start event-ht-int)
					   (gethash e_end event-ht-int)
					   ub)
				 edges)
		   (push (list (gethash e_end event-ht-int)
					   (gethash e_start event-ht-int)
					   (- lb))
				 edges)




		   (when primitive
			 (push (list primitive e_start e_end)
				   actions))))

	;; Compute the APSP via Floyd Warshall
	;;(setf distance-ht-edge (mtk-graph-algorithms:floyd-warshall g))
	(setf d (floyd-warshall-optimized N edges))


	;; Now, construct an action-schedule, containing pairs of the form (action-name execution-time)
	;; for each action in the plan.
	(loop for (action-name e-start nil) in actions do
		 (push (list action-name
					 (- (aref d 
							  (gethash e-start event-ht-int)
							  (gethash plan-event-start event-ht-int))))
			   action-schedule))

	;; Now, that we have an action schedule, sort it by execution time.
	(setf action-schedule (sort action-schedule
								#'(lambda (x y)
									(< (second x)
									   (second y)))))

	;; Now, print them out!
	(format t "Plan Schedule:~%")
	(loop for (action-name execution-time) in action-schedule do
		 (format t "~,4f: ~a~%" execution-time action-name))



    ))


(defun floyd-warshall-optimized (N edges)
  (let* ((d (make-array (list N N)
						:adjustable nil
						:initial-element *inf*)))
	(dotimes (i N)
	  (setf (aref d i i) 
			0.0))

	(loop for (i j w) in edges do
		 (setf (aref d i j)
			   w))

	(dotimes (k N)
	  (dotimes (i N)
		(dotimes (j N)
		  (when (> (aref d i j)
				   (+ (aref d i k)
					  (aref d k j)))
			(setf (aref d i j)
				  (+ (aref d i k)
					 (aref d k j)))))))
      
	d))


(defun event-to-symbol-fn (event)
  "This function takes in an event (i.e., from a TPN), and converts it to a symbol.
   This symbol is treated as case-insensitive (not the defautl for LISP symbols!), and is 
   put in the keyword package for ease. So, when they're printed out, an event with name "0"
   will be printed like :|0| fyi."
  (let* ((event-name (tpn::event-name event))
		 (event-symbol (string-to-symbol-fn event-name)))
	event-symbol))

(defun string-to-symbol-fn (str)
  (intern str :keyword))