;;;; Copyright (c) 2013 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:tburton)

#+allegro
(defmacro exit-on-error (&rest body)
	"If an error in BODY occurs, this form will print the error 
     and force the application to exit with code 0."
	`(unwind-protect 
		  (handler-case
			  (progn ,@body)
			(error (e) (format *terminal-io* "Error: ~a~%" e)))
	   (cl-user::exit 0)))

#+allegro
(defun tburton-main ()
  (exit-on-error
   ;; (format t "~a~%" (system:command-line-arguments))
   ;; (format t "~a~%" (system:command-line-argument-count))
   (system:with-command-line-arguments
	   (("d"      :short domain-path  :required-companion)
		("p"      :short problem-path  :required-companion)
        ("plan"   :long plan-path :optional-companion)
        ("tpn"    :long tpn-path :optional-companion)
        ("dot"    :long dot-path :optional-companion)
		("z"      :short use-zero-episodes)
		("h"      :short shelp)
		("help"   :long  lhelp))
	   (rest)
	 ;; (continue) will continue executions, but first giving the
	 ;; person running the app a chance to poke around.
	 
	 ;; if the user asked for help, just print the help then return.
	 (when (or shelp lhelp (= (system:command-line-argument-count) 3))
	   (print-tburton-executable-help :stream *terminal-io*)
	   (cl-user::exit 0))
	 
	 ;; continue with algorithm
	 (let ((app-path (pathname (sys:command-line-argument 0)))
		   (curr-path (excl:current-directory)))
	   ;; get the path from which the current application was run
	   (unless (pathname-directory app-path)
		 (setf app-path (translate-logical-pathname
						 (concatenate 'string "sys:" (sys:command-line-argument 0)))))

	   ;; logical handling of arguments
	   (let ((abs-domain-path nil)
			 (abs-problem-path nil)
			 (abs-plan-path nil)
             (abs-tpn-path nil)
             (abs-dot-path nil)
			 (tpn-solution nil))
		 ;; make sure use-zero-episodes variable is a boolean
		 (setf use-zero-episodes (and use-zero-episodes t))
		 ;; clean up the domain and problem paths so they are absolute or nil.
		 (when domain-path
		   (setf abs-domain-path (merge-pathnames domain-path curr-path)))
		 (when problem-path
		   (setf abs-problem-path (merge-pathnames problem-path curr-path)))
		 ;; handle non optioned arguments (i.e. arguments without the "-a option" format)
		 (cond (;; if the user specified -d and not -p option
				(and domain-path (not problem-path))
				 (error "Missing problem file. When using the -d option to specify the domain file. The -p option must also be used to specify the problem file."))
				((and (not domain-path) problem-path)
				 (error "Missing domain file. When using the -p option to specify the problem file. The -d option must also be used to specify the domain file."))				
				((and domain-path problem-path)
				;; we do not need to parse non option arguments, so continue
				 )
			   (;; interpret single argument as csv file
				(< (length rest) 2)
				(error "Too few arguments provided."))
			   (;; interpret two arguments as domain and problem file
				(= (length rest) 2)
				(setf abs-domain-path (merge-pathnames (first rest) curr-path))
				(setf abs-problem-path (merge-pathnames (second rest) curr-path))
				)
			   (:otherwise 
				(error "Too many arguments provided.")))
		 (unless (probe-file abs-domain-path)
		   (error "Could not find domain file \"~a\"." domain-path))
		 (unless (probe-file abs-problem-path)
		   (error "Could not find problem file \"~a\"." problem-path))
		 ;;;; RUN TBURTON!
		 (setf tpn-solution (tburton-file abs-domain-path abs-problem-path nil :use-zero-episodes use-zero-episodes))
		 (when tpn-solution
		   ;; if no options are specified, then just print the tpn
		   (unless (or plan-path tpn-path dot-path)
			 (setf tpn-path t))
		   ;; print the plan
		   (cond (;; if only the plan option was specified, we print the plan.
				  (eq plan-path t)
				  (write-plan *terminal-io* tpn-solution))
				 (;; if a path was specified, we write the plan to that file.
				  plan-path
				  (setf abs-plan-path (merge-pathnames plan-path curr-path))
				  (with-open-file (stream abs-plan-path :direction :output :if-exists :supersede)
					(write-plan stream tpn-solution))
				  )
				 (:otherwise))
		   ;; print the tpn
		   (cond (;; if only the tpn option was specified, we print the tpn.
				  (eq tpn-path t)
				  (tpn::write-tpn *terminal-io* tpn-solution)
				  (format *terminal-io* "~%"))
				 (;; if a path was specified, we write the tpn to that file.
				  tpn-path
				  (setf abs-tpn-path (merge-pathnames tpn-path curr-path))
				  (with-open-file (stream abs-tpn-path :direction :output :if-exists :supersede)
					(tpn::write-tpn stream tpn-solution))
				  )
				 (:otherwise))
		   ;; print the dot
		   (cond (;; if only the tpn option was specified, we print the tpn.
				  (eq dot-path t)
				  (dot-print-tpn tpn-solution :stream *terminal-io* :rankdir nil)
				  )
				 (;; if a path was specified, we write the tpn to that file.
				  dot-path
				  (setf abs-dot-path (merge-pathnames dot-path curr-path))
				  (with-open-file (stream abs-dot-path :direction :output :if-exists :supersede)
					(dot-print-tpn tpn-solution :stream stream :rankdir nil))
				  )
				 (:otherwise)))
		 (unless tpn-solution
		   (format *terminal-io* "No solution found!~%"))
		 (format *terminal-io* "Completed Successfully!~%")
		 )))
   ))

(defun print-tburton-executable-help (&key (stream t))
  (format stream
"USAGE
   tburton [PDDL-DOMAIN-FILE] [PDDL-PROBLEM-FILE]
   tburton -d [PDDL-DOMAIN-FILE] -p [PDDL-PROBLEM-FILE]
   tburton -d [PDDL-DOMAIN-FILE] -p [PDDL-PROBLEM-FILE]

DESCRIPTION 
   Run the tBurton planner on PDDL domain and problem files. tBurton supports PDDL 2.1 requirements: \":strips, :typing, :durative-actions, and :duration-inequalities\". When none of the output options (--plan, --tpn, --dot) are used, a TPN in XML format will be displayed on stdout.

OPTIONS
   -d [DOMAIN_FILE]
       Path to the domain file.
   -p [PROBLEM_FILE]
       Path to the problem file.
   --plan  --plan=[PLAN_PATH]
       output the plan in standard IPC format. Use this option alone to write to stdout, or with an argument to specify output to a file.
   --tpn  --tpn=[TPN_PATH]
       output the plan in TPN XML format. Use this option alone to write to stdout, or with an argument to specify output to a file.
   --dot  --dot=[DOT_PATH]
       output the plan (TPN) in the graphviz DOT language. Use this option alone to write to stdout, or with an argument to specify output to a file.
   -z
       When specified, episodes with actions are padded with episodes of 0 lower and upper bound times. (Default: zero-length episodes are not used).
   -h  --help
       Print this help message.
"))

(defmethod write-plan ((stream t) (tpn tpn::temporal-plan-network))
  "Writes out a TPN in the IPC standard textual plan representation"
  (let ((event-time-ht (compute-plan tpn)))
	(format stream "; Length ~a~%" (length (tpn::tpn-episodes tpn)))
	(format stream "; Makespan ~a~%" (gethash (tpn::tpn-end tpn) event-time-ht))
	(loop for episode in (tpn::tpn-episodes tpn) do
		 (let ((start_time (gethash (tpn::episode-start episode) event-time-ht))
			   (end_time (gethash (tpn::episode-end episode) event-time-ht))
			   (action (tpn::episode-primitive episode)))
		   (format stream "~a: ~a [~a]~%" start_time action (- end_time start_time))
		   ))
	)
  )

(defmethod compute-plan ((tpn tpn::temporal-plan-network) &key (rational-inf most-positive-double-float) (-rational-inf most-negative-double-float))
  "Creates a schedulable plan by hoosing event times that minimize
   the makespan time between the start and end of the TPN."
  (let ((event-ht (make-hash-table))
		(num-events 0) ;; the number is incremented in the loop
		(num-episodes (length (tpn::tpn-episodes tpn))))
	;; Create a hash table that maps events to an index number.
	(loop for event in (tpn::tpn-events tpn) do
		 (setf (gethash event event-ht) num-events)
		 (incf num-events))
	;; Create the A, b, and c  matricies.
	(let*  (;; Each episode has a lower and upperbound constraint
			(m (* 2 num-episodes))
			;; Each event has an occurence time
			;; plus two slack variable, one each for the lower and upper bounds.
			(n (+ num-events (* 2 num-episodes)))
			(A (make-array `(,m ,n) :element-type 'number :initial-element 0))
			(b (make-array m :element-type 'number :initial-element 0))
			(c (make-array n :element-type 'number :initial-element 0))
			(tpn-starti (gethash (tpn::tpn-start tpn) event-ht))
			(tpn-endi (gethash (tpn::tpn-end tpn) event-ht)))
	  ;; Create the A and b matrix
	  (loop for episode in (tpn::tpn-episodes tpn)
		   with mi = 0 do
		   (let ((starti (gethash (tpn::episode-start episode) event-ht))
				 (endi (gethash (tpn::episode-end episode) event-ht))
				 (ub (to-rinf (tpn::episode-upperbound episode) :rational-inf rational-inf :-rational-inf -rational-inf))
				 (lb (to-rinf (tpn::episode-lowerbound episode) :rational-inf rational-inf :-rational-inf -rational-inf)))
			 ;; b - a + s1 = ub, where s1>=0
			 (setf (aref A mi endi) 1)
			 (unless (eql tpn-starti starti)
			   (setf (aref A mi starti) -1))
			 (setf (aref A mi (+ num-events mi)) 1)
			 (setf (aref b mi) ub)
			 (incf mi)
			 ;; b - a - s2 = lb, where s2>=0
			 (setf (aref A mi endi) 1)
			 (unless (eql tpn-starti starti)
			   (setf (aref A mi starti) -1))
			 (setf (aref A mi (+ num-events mi)) -1)
			 (setf (aref b mi) lb)
			 (incf mi)))
	  ;; Create the c matrix
	  (setf (aref c tpn-endi) 1)
	  ;; 
	  (let ((min-sol nil)
	  		(min-c nil))
	  	;; Solve the simplex for the minimum case
	  	(multiple-value-setq (min-sol min-c)
	  	  (simplex:simplex A b c))
	  	;; Return the event times
		(if min-sol
			(let ((event-time-ht (make-hash-table))
				  (y (first min-c)))
			  (loop for event in (tpn::tpn-events tpn) do
				 ;; map the event-object to its time from the solutions
				   (setf (gethash event event-time-ht) 
						 (aref y (gethash event event-ht)))
				   )
			;; (loop for episode in (tpn::tpn-episodes tpn) do
			;; 	 ;; map the event-object to its time from the solutions
			;; 	 (setf (gethash event-time-ht (tpn::episode-start episode)) 
			;; 		   (aref y (gethash (tpn::episode-start episode) event-ht)))
			;; 	 ;; map the event-object to its time from the solutions.
			;; 	 (setf (gethash event-time-ht (tpn::episode-end episode)) 
			;; 		   (aref y (gethash (tpn::episode-end episode) event-ht)))
			;; 	 )
			  event-time-ht)
			nil)
		))))