;;;; Copyright (c) 2012 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang

(in-package #:tburton)

(defun tburton-file (pddl-domain-pathname pddl-problem-pathname &optional problem-name &key (use-zero-episodes nil))
  "Given a pddl domain and problem file, it returns the TPN object 
   representing the solution, or nil if no solution exists.
  PDDL-DOMAIN-PATHNAME - a String or Path to an existing pddl domain file.
  PDDL-PROBLEM-PATHNAME - a String or Path to an existing pddl problem file.
  PROBLEM-NAME - the symbol name of the problem to be solved in the problem file.
  "
  (let ((*package* (find-package 'domains)))
	(load pddl-domain-pathname)
	(load pddl-problem-pathname)
	(unless problem-name
	  (setf problem-name domains:*the-problem*))
	)
  (multiple-value-bind (qsp cca s_0) (domain-to-cca problem-name)
	(multiple-value-bind (proj-sol solution) (plan qsp cca s_0 problem-name)
	  (declare (ignore proj-sol))
	  (cond (solution
			 (make-tpn-from-solution solution problem-name :use-zero-episodes use-zero-episodes))
			(t nil))))
  )

(defun make-tpn-from-solution (solution problem-name &key (use-zero-episodes nil))
  (let* ((num-activities (if use-zero-episodes 
							 (reduce #'+ (mapcar (lambda (prims) (length prims))  solution))
							 (length solution)))
		 (tpn (tpn::make-tpn (symbol-name problem-name) (format nil "0") (format nil "~a" num-activities) '()))
		 (domain (domains:get-domain (domains:problem-domain (domains:get-problem problem-name))) ))
	;; (format t "num-activities: ~a~%" num-activities)
	;; create a function that will make it easier to add episodes for this tpn
	(flet ((make-add-episode (start end lb ub primitive)
			 (let ((description (list :episode))) ; Note: Use LIST, do NOT use the quote (')
			   (when primitive
				 (nconc description (list (tpn::make-association :primitive primitive))))
			   (tpn::add-episode start end (tpn::make-episode start end lb ub 0 description tpn :support :CONTROL-PROGRAM ))))  )
	  ;; add the episodes
	  (if use-zero-episodes
		  (loop for primitives in solution
			 with i = 0 do
			   (let* ((offset (+ (* 2 (length primitives)) 1))
					  (start-event (tpn::get-event (format nil "~a" i) tpn))
					  (end-event (tpn::get-event (format nil "~a" (+ i offset)) tpn)))
				 ;; (format t "~a ~a->~a~%" offset start-event end-event)
				 (loop for primitive in primitives do
					  (let ((primitive-string (format nil "~A" primitive))
							(duration (compute-duration-interval (domains:get-operator (first primitive) domain)))
							(act-start-event (tpn::get-event (format nil "~a" (incf i)) tpn))
							(act-end-event (tpn::get-event (format nil "~a" (incf i)) tpn)))
						(make-add-episode start-event act-start-event 0 0 nil)
						(make-add-episode act-start-event act-end-event (to-binf (car duration)) (to-binf (cadr duration)) primitive-string)
						(make-add-episode act-end-event end-event 0 *infinity* nil))  )
				 (incf i)) )
		  (loop for primitives in solution
			 with i = 0 do
			   (let ((start-event (tpn::get-event (format nil "~a" i) tpn))
					 (end-event (tpn::get-event (format nil "~a" (+ i 1)) tpn)))
				 (loop for primitive in primitives do
					  (let ((primitive-string (format nil "~A" primitive))
							(duration (compute-duration-interval (domains:get-operator (first primitive) domain))))
						(make-add-episode start-event end-event (to-binf (car duration)) (to-binf (cadr duration)) primitive-string)))
				 (incf i 1)))
		  )
	  )
	tpn))


(defun compute-duration-interval (operator)
  "Returns a list of the lower and upperbounds for the given operator using the portable infinity representation."
  (let ((duration (domains:op-duration operator))
		(lb 0)
		(ub *inf*))
	(flet ((check-bounds (time)
			 (and (>= time lb) (<= time ub)) ))
	  (when (not (eql (car duration) 'domains::and))
		(setf duration `(domains::and ,duration)))
	  (map nil 
		   (lambda (simp-dur-const)
			 (case (car simp-dur-const)
			   (domains::= (if (check-bounds (caddr simp-dur-const)) 
					  (setf lb (caddr simp-dur-const) 
							ub (caddr simp-dur-const))
					  (error "The duration constraint ~a of action ~a is temporaly infeasible." simp-dur-const (domains:op-name operator))))
			   (domains::<= (if (check-bounds (caddr simp-dur-const)) 
					   (setf ub (caddr simp-dur-const))
					   (error "The duration constraint ~a of action ~a is temporaly infeasible." simp-dur-const (domains:op-name operator))) )
			   (domains::>= (if (check-bounds (caddr simp-dur-const)) 
					   (setf lb (caddr simp-dur-const))
					   (error "The duration constraint ~a of action ~a is temporaly infeasible." simp-dur-const (domains:op-name operator))))
			   ))
		   (cdr duration)))
	`(,lb ,ub)))