;;;; Copyright (c) 2009 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David Wang
;;;;   Alborz Geramifard

;;;; This file contains all the data structures and algorithms to run 
;;;; floyd-warshall's all pairs shortest path algorithm. There are two
;;;; versions of the algorithm implemented, a 'full' and a 'symmetrical'
;;;; version. 
;;;;
;;;; The 'full' version should be used for a directed graph. 
;;;; Given two nodes, A and B, there can be one edge weight for the arc
;;;; from A to B, and a different weight for the arc from B to A.
;;;; The 'symmetrical' (abbreviated 'symm') version should be used
;;;; for an undirected graph, or a graph in which the edge weight
;;;; between A and B is the same regardless of direction. The 
;;;; symmetrical variant has the benefit that it uses half the memory
;;;; and runs twice as fast as the 'full' version.
;;;;
;;;; Both implementations start with a data structure that is used both
;;;; for representing the initial problem and stores the final solution.
;;;; 
;;;; Use MAKE-APSP to create a problem. Here we create a data structure
;;;; for an undirected graph of 10 nodes.
;;;;   (defparameter *APSP-PROBLEM* (make-apsp 10 :symmetric t))
;;;;
;;;; Use SET-COST to initialize the data structure with edge-costs/weights.
;;;;   (set-cost *APSP-PROBLEM* 0 0 15.6)
;;;;   (set-cost *APSP-PROBLEM* 1 0 2.3)
;;;;   ...
;;;;   (set-cost *APSP-PROBLEM* 9 9 17.9)
;;;; Note: The data structure is zero-indexed, and row-major, i.e. the 
;;;; indices are: (row, column). For symmetric matricies, the indices
;;;; should be  'column <= row'. If the column-index is greater than the
;;;; row index, the two indices will be swapped.
;;;;
;;;; Use floyd-warshall to solve the APSP problem.
;;;;  (floyd-warshall *APSP-PROBLEM*)
;;;;  
;;;; Finally, you can use GET-COST, GET-COST-FROM-PATH, and GET-PATH,
;;;; to recover solutions from the object.
;;;; To get the lowest cost from source 1 to destination 2:
;;;;   (get-cost *APSP-PROBLEM* 1 2)
;;;; To get the cost of the path leading from 1 to 7 to 2:
;;;;   (get-cost-from-path *APSP-PROBLEM* '(1 7 2))
;;;; To get the shortest path from source 1 to destination 2:
;;;;   (get-path *APSP-PROBLEM* 1 2)   =>   '(1 8 9 2)
;;;; 


;;;; IMPLEMENTATION NODES:
;;;; This implementation purposefully uses one array to store the 
;;;; cost matrix and another to store the shortcut matrix. It was 
;;;; found in testing, that in most cases (for arrays under 1000-by-1000) 
;;;; lisp allocates these monolithic arrays faster than smaller arrays.
;;;; i.e. Allocating one n-by-n array with each element being a 
;;;; size 2 array, an object, or a list, were all slower. 
;;;; Using a n-by-n-by-2 array was also considered, but then we would 
;;;; not have been able to gain the memory efficiency by creating arrays
;;;; of one type.
;;;;
;;;; These assumptions may not hold true for other implementations of
;;;; lisp, larger arrays, or even slower computers.


(in-package #:visibility-graph)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Class APSP-FULL
;; This class represents a weight and shortcut matrix
;; for a directed graph, by using full n-by-n arrays.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass apsp-full ()
  ((num-nodes
	:type integer
	:reader apsp-full-num-nodes
	:initarg :num-nodes
	:documentation "The number of nodes in this apsp graph.")
   (cost-matrix
    :reader apsp-full-cost-matrix
	:initarg :cost-matrix
    :documentation "Cost Matrix is a n-by-n matrix of weights. The weight between source and destination is given by the index (source index, destination index).")
   (shortcut-matrix
    :reader apsp-full-shortcut-matrix
	:initarg :shortcut-matrix
    :documentation "Shortcut Matrix is a n-by-n matrix of indices. The index of the next intermediate node to visit to travel from source to destination node is given by the index (source index, destination index).")
   ))

(defun make-apsp-full (num-nodes &key (element-type 'number) (initial-contents nil) (initial-element *inf*))
  "Returns a APSP-FULL object, capable of storing the 
  cost/weights of a directional graph with NUM-NODES."
  (let* ((dim (list num-nodes num-nodes))
		 (cost-matrix-args (list dim)))
	;; build the argument list to cost-matrix
	(when element-type
	  (nconc cost-matrix-args `(:element-type ,element-type)))
	(if initial-contents
		(nconc cost-matrix-args `(:initial-contents ,initial-contents))
		(nconc cost-matrix-args `(:initial-element ,initial-element)))
	(make-instance 'apsp-full 		 
				   :num-nodes num-nodes
				   :cost-matrix  (apply #'make-array cost-matrix-args)
				   ;; lazy initialization - we reinitialize the shortcut
				   ;; matrix when the algorithm is run. This helps maintain
				   ;; the invariant if the cost matrix is modified and the
				   ;; algorithm is re-run.
				   :shortcut-matrix nil)))

(defmethod set-cost ((apsp apsp-full) (src integer) (des integer) (cost number))
  (with-slots ((cost-matrix cost-matrix)) apsp
	(setf (aref cost-matrix src des) cost)))

(defmethod get-cost ((apsp apsp-full) (src integer) (des integer))
  (with-slots ((cost-matrix cost-matrix)) apsp
	(aref cost-matrix src des )))

(defmethod get-cost-from-path ((apsp apsp-full) (path list))
  "Returns the cost of the path"
  (with-slots ((c cost-matrix)) apsp
	(if (null (rest path))
		0
		(+ (aref c (first path) (second path)) 
		   (get-cost-from-path apsp (rest path))))))

(defmethod get-path ((apsp apsp-full) (src integer) (des integer))
  "Returns the path from src to des"
  (with-slots ((c cost-matrix) (s shortcut-matrix)) apsp
    (let ((shortcut (aref s src des)))
          (if (minusp shortcut)
              (list src des)
              (append (get-path apsp src shortcut) (rest (get-path apsp shortcut des)))))))

(defmethod floyd-warshall ((apsp apsp-full) &key (add-function #'+) (less-function #'<) &allow-other-keys)
  "All pair shortest path using floyd-warshall"
  (with-slots ((c cost-matrix) (s shortcut-matrix) (n num-nodes)) apsp
    (setf s (make-array (list n n) :initial-element -1))  
    (dotimes (k n)
      (dotimes (i n)
        (dotimes (j n)
		  (let ((ikkj_val (funcall add-function (aref c i k) (aref c k j)) ))
			(when (funcall less-function ikkj_val (aref c i j))
			  (setf (aref c i j) ikkj_val)
			  (setf (aref s i j) k)
			 ;(format t "~d => ~d Shortcut: ~d Cost: ~d~%" i j k (aref c i j))
			  ))))))
  apsp)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Class APSP-SYMM
;; This class represents a weight and shortcut matrix
;; for a symmetric directed graph (weights are the same
;; going in either direction, i.e. an undirected graph.
;;
;; The representation only stores the LOWER-TRIANGLE, 
;; of matricies, but stores this information in a 
;; linear array. For a graph with n nodes, a lower-
;; triangular symmetric matrix only has n*(n+1)/2
;; cells. This number is not exactly half of a full
;; matrix, because the major-diagnonal still needs to
;; be stored. The mapping from lower-triangular to linear
;; is as follows:
;; 
;; LOWER-TRAINGULAR MATRIX:
;; A
;; B  C
;; D  E  F
;; G  H  I  J
;;
;; LINEAR ARRAY:
;; A B C D E F G H I J
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass apsp-symm ()
  ((num-nodes
	:type integer
	:reader apsp-symm-num-nodes
	:initarg :num-nodes
	:documentation "The number of nodes in this apsp graph.")
   (cost-matrix
    :reader apsp-symm-cost-matrix
	:initarg :cost-matrix
    :documentation "Cost Matrix is a lower-matrix of weights. The weight between source and destination is given by the index (source index, destination index).")
   (shortcut-matrix
    :reader apsp-symm-shortcut-matrix
	:initarg :shortcut-matrix
    :documentation "Shortcut Matrix is a lower-matrix of indices. The index of the next intermediate node to visit to travel from source to destination node is given by the index (source index, destination index).")
   ))

(defun make-apsp-symm (num-nodes &key (element-type 'number) (initial-contents nil) (initial-element *inf*))
  "Returns a APSP-SYMM object, capable of storing the 
  cost/weights of an undirected graph with NUM-NODES."
  (let* ((dim (round (* 0.5 num-nodes (+ num-nodes 1))))
		 (cost-matrix-args (list dim)))
	;; build the argument list to cost-matrix
	(when element-type
	  (nconc cost-matrix-args `(:element-type ,element-type)))
	(if initial-contents
		(nconc cost-matrix-args `(:initial-contents ,initial-contents))
		(nconc cost-matrix-args `(:initial-element ,initial-element)))
	;; make-instance
	(make-instance 'apsp-symm	 
				   :num-nodes num-nodes
				   :cost-matrix  (apply #'make-array cost-matrix-args)
				   ;; lazy initialization - we reinitialize the shortcut
				   ;; matrix when the algorithm is run. This helps maintain
				   ;; the invariant if the cost matrix is modified and the
				   ;; algorithm is re-run.
				   :shortcut-matrix nil)))

(defun apsp-symm-calc-index (i j)
  (if (> j i)
	  (round (+ (* 0.5 j (+ j 1)) i))
	  (round (+ (* 0.5 i (+ i 1)) j))))

(defmethod set-cost ((apsp apsp-symm) (src integer) (des integer) (cost number))
  (with-slots ((cost-matrix cost-matrix)) apsp
	(setf (aref cost-matrix (apsp-symm-calc-index src des)) cost)))

(defmethod get-cost ((apsp apsp-symm) (src integer) (des integer))
  (with-slots ((cost-matrix cost-matrix)) apsp
	(aref cost-matrix (apsp-symm-calc-index src des))))

(defmethod get-cost-from-path ((apsp apsp-symm) (path list))
  "Returns the cost of the path"
  (with-slots ((c cost-matrix)) apsp
	(if (null (rest path))
		0
		(+ (aref c (apsp-symm-calc-index (first path) (second path))) 
		   (get-cost-from-path apsp (rest path))))))

(defmethod get-path ((apsp apsp-symm) (src integer) (des integer))
  "Returns the path from src to des"
  (with-slots ((c cost-matrix) (s shortcut-matrix)) apsp
    (let ((shortcut (aref s (apsp-symm-calc-index src des))))
          (if (minusp shortcut)
              (list src des)
              (append (get-path apsp src shortcut) (rest (get-path apsp shortcut des)))))))


;; for (int k = 0; k < N; ++k)
;;     for (int i = 0; i < N; ++i)
;;         for (int j = 0; j <= i; ++j)
;;             dist[j][i] = dist[i][j] = min(dist[i][j], dist[i][k] + dist[k][j]);
(defmethod floyd-warshall ((apsp apsp-symm) &key (add-function #'+) (less-function #'<) &allow-other-keys)
  "All pair shortest path using floyd-warshall on a symmetric graph"
  (with-slots ((c cost-matrix) (s shortcut-matrix) (n num-nodes)) apsp
    (setf s (make-array (array-dimensions c) :initial-element -1))
	  (dotimes (k n)
		(dotimes (i n)
		  (dotimes (j (+ i 1))
			(let ((ij (apsp-symm-calc-index i j))
				  (ikkj_val (funcall add-function (aref c (apsp-symm-calc-index i k)) 
									 (aref c (apsp-symm-calc-index k j)))))
			  (when (funcall less-function ikkj_val (aref c ij))
				(setf (aref c ij) ikkj_val)
				(setf (aref s ij) k)))))))
  apsp)


(defun make-apsp (num-nodes &key (symmetric nil) (element-type 'number) (initial-contents nil) (initial-element *inf*))
  "This function provides a convenient way of creating
   either a full or symmetric APSP data structure.

   NUM-NODES must be an integer, the number of nodes that your graph will
   contain.
  
   :SYMMETRIC can be either T or NIL. NIL to create a full matrix. 
   T to create a symmetric matrix. By default, the value is NIL."
	;; build the argument list to cost-matrix
  (if symmetric
	  (make-apsp-symm num-nodes 
					  :element-type element-type 
					  :initial-contents initial-contents
					  :initial-element initial-element)
	  (make-apsp-full num-nodes 
					  :element-type element-type 
					  :initial-contents initial-contents
					  :initial-element initial-element)))

