;;;; Copyright (c) 2008-2010 Brian C. Williams, MIT.

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the authors, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Original Author:
;;;;   David Wang

(in-package #:visibility-graph)

(mtk-utils:deftest-suite :test-floyd-warshall)

;;; Test the creation of a full floyd-warshall problem
;;; via MAKE-APSP and quick initialization
(mtk-utils:deftest (test-fw-1 :test-floyd-warshall) ()
  (let* ((contents '((0  1  3  5  2)
					 (2  0  4  8  5)
					 (1  3  0  2  1)
					 (6  5  4  0  9)
					 (1  9  7  2  0)))
		 (n (length contents))
		 (apsp-problem (make-apsp n 
					   :initial-contents contents)))
	;; Tests the creation and initialization
    (mtk-utils:test (make-array (list n n) :initial-contents contents)
		  (apsp-full-cost-matrix apsp-problem)
		  :test #'equalp
		  :fail-info "")))


;;; Test the creation of a full floyd-warshall problem
;;; via SET-COST
(mtk-utils:deftest (test-fw-2 :test-floyd-warshall) ()
  (let* ((contents '((0  1  3  5  2)
					 (2  0  4  8  5)
					 (1  3  0  2  1)
					 (6  5  4  0  9)
					 (1  9  7  2  0)))
		 (n (length contents))
		 (apsp-problem (make-apsp n)))
	(dotimes (i n)
	  (dotimes (j n)
		(set-cost apsp-problem i j (nth j (nth i contents)))))
    (mtk-utils:test (make-array (list n n) :initial-contents contents)
					(apsp-full-cost-matrix apsp-problem)
					:test #'equalp
					:fail-info "")))

;;; Test the full floyd-warshall algorithm
(mtk-utils:deftest (test-fw-3 :test-floyd-warshall) ()
  (let* ((contents '((0    1  1e20 4)
					 (1    0  8    6)
					 (1e20 8  0    2)
					 (4    6  2    0)))
		 (cost-answer '((0 1 6 4)
						(1 0 7 5) 
						(6 7 0 2) 
						(4 5 2 0)))
		 (shortcut-answer '((-1 -1 3 -1) 
							(-1 -1 3 0)
							(3 3 -1 -1)
							(-1 0 -1 -1)))
		 (n (length contents))
		 (apsp-problem (make-apsp n 
					   :initial-contents contents)))
	;; Run the floyd-warshall algorithm
	(floyd-warshall apsp-problem)
	;; Tests the cost answer
    (mtk-utils:test (make-array (list n n) :initial-contents cost-answer)
					(apsp-full-cost-matrix apsp-problem)
		  :test #'equalp
		  :fail-info "")
	;; Tests the shortcut answer
    (mtk-utils:test (make-array (list n n) :initial-contents shortcut-answer)
					(apsp-full-shortcut-matrix apsp-problem)
		  :test #'equalp
		  :fail-info "")))

;;; Test GET-COST, GET-COST-FROM-PATH, and GET-PATH
(mtk-utils:deftest (test-fw-4 :test-floyd-warshall) ()
  (let* ((cost-answer '((0 1 6 4)
						(1 0 7 5) 
						(6 7 0 2) 
						(4 5 2 0)))
		 (shortcut-answer '((-1 -1 3 -1) 
							(-1 -1 3 0)
							(3 3 -1 -1)
							(-1 0 -1 -1)))
		 (n 4)
		 (apsp-problem (make-apsp n)))
	;; initialize the problem with the answers 
	(with-slots ((c cost-matrix) (s shortcut-matrix) (n num-nodes)) apsp-problem
	  (setf c (make-array (list n n) :initial-contents cost-answer))
	  (setf s (make-array (list n n) :initial-contents shortcut-answer)))
	;; Test GET-COST
    (mtk-utils:test (nth 2 (nth 1 cost-answer))
					(get-cost apsp-problem 1 2)
		  :test #'equalp
		  :fail-info "")
	;; Test GET-COST-FROM-PATH
    (mtk-utils:test 15
					(get-cost-from-path apsp-problem '(0 1 2 3 1))
		  :test #'equalp
		  :fail-info "")
	;; Test GET-PATH
    (mtk-utils:test '(1 0 3 2)
					(get-path apsp-problem 1 2)
		  :test #'equalp
		  :fail-info "")))



;;; Test the creation of a symmetric floyd-warshall problem
;;; via MAKE-APSP and quick initialization
(mtk-utils:deftest (test-fw-5 :test-floyd-warshall) ()
  (let* ((contents '(0
					 2  0
					 1  3  0
					 6  5  4  0
					 1  9  7  2  0))
		 (n 5)
		 (apsp-problem (make-apsp n
								  :symmetric t
								  :initial-contents contents)))
	;; Tests the creation and initialization
    (mtk-utils:test (make-array (length contents) :initial-contents contents)
		  (apsp-symm-cost-matrix apsp-problem)
		  :test #'equalp
		  :fail-info "")))


;;; Test the creation of a symmetric floyd-warshall problem
;;; via SET-COST
(mtk-utils:deftest (test-fw-6 :test-floyd-warshall) ()
  (let* ((contents '(0
					 2  0
					 1  3  0
					 6  5  4  0
					 1  9  7  2  0))
		 (n 5)
		 (apsp-problem (make-apsp n :symmetric t)))
	(dotimes (i n)
	  (dotimes (j (+ i 1))
		(set-cost apsp-problem i j (nth (round (+ (* 0.5 i (+ i 1)) j)) contents))))
    (mtk-utils:test (make-array (length contents) :initial-contents contents)
					(apsp-symm-cost-matrix apsp-problem)
					:test #'equalp
					:fail-info "")))


;;; Test the symmetric floyd-warshall algorithm
(mtk-utils:deftest (test-fw-7 :test-floyd-warshall) ()
  (let* ((contents '(0
					 1    0
					 1e20 8  0
					 4    6  2    0))
		 (cost-answer '(0
						1 0 
						6 7 0 
						4 5 2 0))
		 (shortcut-answer '(-1
							-1 -1
							 3  3 -1
							-1  0 -1 -1))
		 (n 4)
		 (apsp-problem (make-apsp n 
					   :symmetric t
					   :initial-contents contents)))
	;; Run the floyd-warshall algorithm
	(floyd-warshall apsp-problem)
	;; Tests the cost answer
    (mtk-utils:test (make-array (length cost-answer) :initial-contents cost-answer)
					(apsp-symm-cost-matrix apsp-problem)
		  :test #'equalp
		  :fail-info "")
	;; Tests the shortcut answer
    (mtk-utils:test (make-array (length shortcut-answer) :initial-contents shortcut-answer)
					(apsp-symm-shortcut-matrix apsp-problem)
		  :test #'equalp
		  :fail-info "")))

;;; Test GET-COST, GET-COST-FROM-PATH, and GET-PATH for symmetric.
(mtk-utils:deftest (test-fw-8 :test-floyd-warshall) ()
  (let* ((cost-answer '(0
						1 0 
						6 7 0 
						4 5 2 0))
		 (shortcut-answer '(-1
							-1 -1
							 3  3 -1
							-1  0 -1 -1))
		 (n 4)
		 (apsp-problem (make-apsp n
								  :symmetric t)))
	;; initialize the problem with the answers 
	(with-slots ((c cost-matrix) (s shortcut-matrix) (n num-nodes)) apsp-problem
	  (setf c (make-array (length cost-answer) :initial-contents cost-answer))
	  (setf s (make-array (length shortcut-answer) :initial-contents shortcut-answer)))
	;; Test GET-COST
    (mtk-utils:test (nth 4 cost-answer)
					(get-cost apsp-problem 1 2)
		  :test #'equalp
		  :fail-info "")
	;; Test GET-COST-FROM-PATH
    (mtk-utils:test 15
					(get-cost-from-path apsp-problem '(0 1 2 3 1))
		  :test #'equalp
		  :fail-info "")
	;; Test GET-PATH
    (mtk-utils:test '(1 0 3 2)
					(get-path apsp-problem 1 2)
		  :test #'equalp
		  :fail-info "")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; MANUAL TEST SCRIPTS AND FUNCTIONS

(defmethod vg-add-episode ((apsp apsp-full) (src integer) (des integer) (lb number) (ub number))
  (with-slots (cost-matrix) apsp
	(setf (aref cost-matrix src des) ub)
	(setf (aref cost-matrix des src) (- lb))))

(defun test-fw-9 ()
  (let* ((n 4)
		 (apsp (make-apsp n :symmetric nil)))
	(vg-add-episode apsp 0 1 1 10)
	(vg-add-episode apsp 0 2 1 10)
	(vg-add-episode apsp 2 3 0 5)
	(vg-add-episode apsp 1 3 0 5)
	(format t "~s~%" (apsp-full-cost-matrix apsp))
	;; run floyd-warshall
	(floyd-warshall apsp)
	(apsp-full-cost-matrix apsp)
	)
  )
