;;;; Copyright (c) 2010 Massachusetts Institute of Technology

;;;; This software may not be redistributed, and can only be retained and used
;;;; with the explicit written consent of the author, subject to the following
;;;; conditions:

;;;; The above copyright notice and this permission notice shall be included in
;;;; all copies or substantial portions of the Software.

;;;; This software may only be used for non-commercial, non-profit, research
;;;; activities.

;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESSED
;;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
;;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;;; DEALINGS IN THE SOFTWARE.

;;;; Authors:
;;;;   David C. Wang
;;;;   Alborz Geramifard

(in-package #:visibility-graph)

(defclass visibility-graph ()
  ((vertex-ht
	:reader vg-vertex-ht
	:initarg :vg-vertex-ht
	:documentation "Maps from a node in the visibility-graph to its integer index 
                    in the apsp object.")
   (vertex-array
	:reader vg-vertex-array
	:initarg :vg-vertex-array
	:documentation "Maps from the integer index in the apsp object to a node in 
                    the visibility-graph.")
   (apsp-object
	:reader vg-apsp
	:initarg :vg-apsp
	:documentation "Either an apsp-full or apsp-symm object.")))

;;-------------------------------------------------------------------------------
;; Visibility Graph
;;-------------------------------------------------------------------------------

;;; obstacle-list - is a list of polygon objects
;;; waypoint-list - is a list of ("waypoint-name" 2d-point-obj).

;; TODO: return an error if the start is in an obstacle.
(defun make-visibility-graph (obstacle-list waypoint-list)
  "Returns a visibility-graph object that represents routing from 
   points in WAYPOINT-LIST and OBSTACLE-LIST, around obstacles 
   contanined in OBSTACLE-LIST.

   WAYPOINT-LIST is a list of 2d-geometry.2d-point objects.

   OBSTACLE-LIST is a list of 2d-geometry.polygon objects.

   It is assumed that the distance between 2 points is identical in both directions.
   This assumption is important because the cost-matrix given to the floyd-warshall algorithm
   only has calculations done for its lower-triangle. The resulting costs are flipped to 
   the upper traingle. This symmetry is exploited for speed."

  ;; Finding all pairs shortest path (APSP) is at best a O(N^3) operation, where N is the
  ;; number of verticies to consider, so we want to prune as many of those nodes from 
  ;; consideration as possible before we initialize the APSP problem. 

  (let ((vx-list (make-smart-list))
		(vx-ht (make-hash-table)))
	(let ((add-vx (lambda (vx) 
					(setf (gethash vx vx-ht) (coll-size vx-list))
					(coll-add! vx-list vx)))
		  (filter-waypoint  (lambda (waypoint) 
							  (dolist (obstacle obstacle-list t) 
								(when (contains? obstacle waypoint) (return nil)))))
		  (filter-obstacle-vertex  (lambda (vertex) 
									 (dolist (obstacle obstacle-list t)
									   (when (and (not (member vertex (point-list obstacle) :test #'equalp-geo)) 
												  (contains? obstacle vertex)) (return nil))))))
	  ; remove points contained in the polygon
	  (dolist (waypoint waypoint-list)
		(when (funcall filter-waypoint waypoint) 
		  (funcall add-vx waypoint)))
	  ; remove points contained in the polygon
	  ; doing this step here ensures overlapping polygons are handled correctly
	  ; namely, if one polygon's point lies in another polygon
	  (dolist (obstacle obstacle-list)
		(dolist (obstacle-vertex (point-list obstacle))
		  (when (funcall filter-obstacle-vertex obstacle-vertex)
			(funcall add-vx obstacle-vertex))))
	  
	  ;; Create and initialize the floyd-warshall object.
	  (let ((vx-array (make-array (coll-size vx-list) :initial-contents (coll-tolist vx-list)))
			(vx1 nil)
			(vx2 nil)
			(value 0)
			(vg-apsp (make-apsp (coll-size vx-list) :symmetric t)))
		;; add in the remaining lengths, making sure not
		;; to reset any values already set.
		(dotimes (i (coll-size vx-list))
			 (dotimes (j (+ i 1))
			   (setf vx1 (aref vx-array i))
			   (setf vx2 (aref vx-array j))
			   (setf value 
					 (if (intersects-obstacle? 
						  obstacle-list 
						  (make2-2d-line-segment vx1 vx2))
						 *inf*
						 (get-distance vx1 vx2)))
			   (set-cost vg-apsp i j value)))
		;; Run the floyd-warshall algorithm.
		(floyd-warshall vg-apsp)
		;; Create and return the visibility-graph object.
		(make-instance 'visibility-graph 
					   :vg-vertex-ht vx-ht 
					   :vg-vertex-array vx-array
					   :vg-apsp vg-apsp)
		))))

 
(defun intersects-obstacle? (obstacle-list seg)

  "Returns TRUE if a line segment, SEG, intersects any of the polygons in 
   OBSTACLE-LIST. Returns NIL otherwise. Note that this method was designed
   for use in a visibility graph: Intersections are only tested with the
   interior of the polygon, i.e. a line segment co-linear with an obstacle's
   edge will not count as an intersection.
   
   SEG is a '2d-line-segment' object from mtk-util's 2d-geometry library.
  
   OBSTACLE-LIST is a list of 'polygon' objects from mtk-util's 2d-geometry library."

  (dolist (obstacle obstacle-list nil)
	(when (intersects? obstacle seg :include-edge nil :include-interior t) 
	  (return t))))

(defmacro vg-get-index (value vg)
  "If VALUE is an integer, we assume it is an index and return it.
   Otherwise, we lookup value in VG's vg-node-ht to retrieve the index.
   If no index is found, nil is returned."
	`(if (typep ,value 'integer)
		,value
		(gethash ,value (vg-vertex-ht ,vg))))

(defmethod get-cost ((vg visibility-graph) src des)
  "Returns the cost of getting from SRC to DES. If SRC or DES 
   are not an index into VG's APSP field or a point in the waypoint or
   obstacle-list used to make VG, NIL is returned.

  SRC is an index into the visibility-graph object's APSP field OR
  one of the points in the wapoint-list or obstacle-list.

  DES is an index into the visibility-graph object's APSP field OR
  one of the points in the wapoint-list or obstacle-list."

  (let ((src-index (vg-get-index src vg))
		(des-index (vg-get-index des vg)))
	(if (and src-index des-index)
		(get-cost (vg-apsp vg) src-index des-index)
		nil)))

(defmethod get-cost-from-path ((vg visibility-graph) (path list))
  "Returns the cost of PATH, 0 if PATH is empty, or NIL if PATH 
   contains one or more points not used to make the 
   visibility-graph problem."
  (let ((path-indices nil)
		(bad-path nil))
	;; get the path-indices, but break with no indices if the path
	;; contains points that cannot be found in the vertex hash table.
	(setf path-indices
		  (loop for vx in path
			 for vx-index = (vg-get-index vx vg)
			 if vx-index
			 collect vx-index
			 else do (setf bad-path t) (return nil)))
	(if bad-path
		nil
		(get-cost-from-path (vg-apsp vg) path-indices))))

(defmethod get-path ((vg visibility-graph) src des)
  "Returns the path for getting from SRC to DES. If SRC or DES 
   are not an index into VG's APSP field or a point in the waypoint or
   obstacle-list used to make VG, NIL is returned.

  SRC is an index into the visibility-graph object's APSP field OR
  one of the points in the wapoint-list or obstacle-list.

  DES is an index into the visibility-graph object's APSP field OR
  one of the points in the wapoint-list or obstacle-list."

  (let ((src-index (vg-get-index src vg))
		(des-index (vg-get-index des vg)))
	(if (and src-index des-index)
		(mapcar (lambda (x) (aref (vg-vertex-array vg) x))  
				(get-path (vg-apsp vg) src-index des-index))
		nil)))



